# Copyright (c) Microsoft Corporation. All rights reserved.
#

# If MSVCDIR is defined we use the preinstalled compiler.
!if DEFINED(MSVCDIR)
CC_DIR=$(MSVCDIR)\bin
!if DEFINED(VSINSTALLDIR)
# VS7 installed
PATH=$(PATH);$(VSINSTALLDIR)
WIN32_INCS=$(CC_INC)"$(MSVCDIR)\PlatformSDK\include" 
!else
# VC6 and Platform SDK installed
PATH=$(PATH);$(MSDEVDIR)\bin
WIN32_INCS=$(CC_INC)"$(MSSDK)\include" 
!endif
!else
!if DEFINED(VCINSTALLDIR)
# VS8 installed
PATH=$(PATH);$(VCINSTALLDIR)\bin
WIN32_INCS=$(CC_INC)"$(VCINSTALLDIR)\PlatformSDK\include" 
CC_DIR=$(VCINSTALLDIR)\bin
CC_FLAGS=$(CC_FLAGS) -GS-
!else
# VC6 and platform files dumped into MMLITE_SDK\tools
WIN32_INCS=$(CC_INC)$(MMLITE_SDK)\tools\i386\inc\SDK
CC_DIR=$(MMLITE_SDK)\tools\i386\bin\vc6
!endif
!endif

CC=$(CC_DIR)\cl.exe
CPLUS=$(CC_DIR)\cl.exe

CC_DEF=-D
CC_UNDEF=-U
CC_INC=-I
CC_OUT=-Fo

CC_SBR=-FR$*.sbr

CPRE="$(CC)" -P

CC_FLAGS=-nologo $(CC_FLAGS)
!if DEFINED(COVERAGE_BUILD)
CC_FLAGS=$(CC_FLAGS) -Gm
!endif

#   -Gy enables function-level linking (intended to be used with
#   linker option -opt:ref).
#   -GF enables string pooling, which should reduce size.
#   The rest of the MS i386 compiler options are fairly standard....
CC_FLAGS=$(CC_FLAGS) -c -X -Gy -GF -W3 -WX $(CC_DEF)i386=1 $(_CFLAGS)
CPLUS_FLAGS=$(CC_FLAGS)

#
#   Assembler settings
#
AS_OUT=-Fo
AS_SBR=-FR$*.sbr


#
#   Linker settings
#
LINK=$(CC_DIR)\link.exe
LINK_FLAGS=$(LINK_FLAGS) -machine:$(TARGETCPU) $(_LFLAGS) $(LINK_XXX_FLAGS)
LINK_FLAGS=$(LINK_FLAGS) -nologo
LINK_DEF=-def:
LINK_OUT=-out:
LINK_IMPLIB=-implib:

!ifdef COMPONENT_ENTRY
LINK_ENTRY=-entry:$(COMPONENT_ENTRY)
!else
COB_ENTRY=-entry:_CobMainCRT
!if DEFINED(_UNICODE)
EXE_ENTRY=-entry:wmainCRTStartup1
!else
EXE_ENTRY=-entry:mainCRTStartup1
!endif
!endif

#   -subsystem:console,3.27 makes the linker happy with unrecognized
#   binary types (eg, MMLITE COB files), and 3.27 happens to pack the
#   sections tighter (eg, 512-byte alignment instead of the default
#   4K);  manually changing the alignment to 512 (using -align:512)
#   not only generates a linker warning but also doesn't always
#   produce a working binary, so the version specification appears to
#   affect more than just section alignment....
!if !DEFINED(LINK_IMAGE)
# This makes images smaller but does not run on NT
#LINK_XXX_FLAGS=-map -subsystem:console,3.27 -incremental:no -align:256 -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text
LINK_XXX_FLAGS=-map -subsystem:console,3.27 -incremental:no
!else
# Special link flags for an image that gets glued together with
# _dosxtndt and mk86boot
LINK_XXX_FLAGS=-map -align:1024 -base:0x00000000 -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text
!endif

!if "$(TARGETTYPE)"=="release"
LINK_REF= -opt:ref
!else
LINK_REF= -opt:noref
!endif

LINK_DLL_FLAGS=-dll $(LINK_REF) $(LINK_FLAGS)
LINK_EXE_FLAGS= $(LINK_REF) $(LINK_FLAGS)

#   We have to specify -debug and -debugtype in order for -pdb to
#   generate anything.  BUGBUG: We don't currently specify these DEBUG
#   settings for release builds, because debug info is still being
#   generated inside the binary.
#   We need to fix that, either with the right set of linker options
#   or a post-link "stripsym" utility. -JP
!if DEFINED(COVERAGE_BUILD)
LINK_DBG_FLAGS=-debug:full -debugtype:cv,fixup -incremental:no -pdb:$*.pdb
!else
LINK_DBG_FLAGS=-debug:full -pdb:$*.pdb
!endif


#
#   Librarian settings
#
LIBRARIAN=$(CC_DIR)\link.exe
LIB_FLAGS=-lib -machine:$(TARGETCPU) -nologo
LIB_DEF=-def:
LIB_OUT=$(LINK_OUT)


#
#   Source browser settings
#
BSCMAKE=$(CC_DIR)\bscmake
BSCMAKE_FLAGS=-nologo -n -o

#
#   Resource compiler settings
#
RC_DIR=$(CC_DIR)
!if DEFINED(MSVCDIR) && !DEFINED(VSINSTALLDIR)
# VC6 preinstallation finds it from a different place
RC_DIR=$(MSDEVDIR)\bin
!endif

RC=$(RC_DIR)\rc.exe

RC_DEF=-d
RC_UNDEF=-U
RC_INC=-i
RC_OUT=-fo
RC_NODEFAULTINC=-x

# 16-bit and image manipulation tools
MKMEMFS=$(MMLITE_SDK)\tools\bin\mkmemfs
MK86BOOT=$(MMLITE_SDK)\tools\bin\mk86boot
ASM16=$(MMLITE_SDK)\tools\i386\bin\masm611\ml
LINK16=$(MMLITE_SDK)\tools\i386\bin\masm611\link

!if !DEFINED(LINK_IMAGE)
LINK_EXE_FLAGS=$(LINK_EXE_FLAGS) -fixed:NO
!endif

LINK_FLAGS=$(LINK_FLAGS) -nodefaultlib

#   No entry point means no entry point: turn off default
!ifndef LINK_ENTRY
#LINK_DLL_FLAGS=$(LINK_DLL_FLAGS) -noentry
LINK_DLL_FLAGS=$(LINK_DLL_FLAGS) $(COB_ENTRY)
!endif


#
#   TARGETTYPE-specific settings for these tools
#
!if "$(TARGETTYPE)"=="release"

# VC6 compilerbug.
# -Ox==/Ogityb1 /Gs  but the "g" one screws up INT64 type promotion.
#CC_FLAGS=$(CC_FLAGS) -Ox -MD
CC_FLAGS=$(CC_FLAGS) -Osiyb1y -Gs -MD

!if DEFINED(_RELEASE_SYMBOLS)
CC_FLAGS= $(CC_FLAGS) -Zi
LINK_DLL_FLAGS=$(LINK_DLL_FLAGS) $(LINK_DBG_FLAGS)
LINK_EXE_FLAGS=$(LINK_EXE_FLAGS) $(LINK_DBG_FLAGS)
!endif

!else

CC_FLAGS=$(CC_FLAGS) -Zi -Od -MDd

!if !DEFINED(LINK_IMAGE)
LINK_DLL_FLAGS=$(LINK_DLL_FLAGS) $(LINK_DBG_FLAGS)
LINK_EXE_FLAGS=$(LINK_EXE_FLAGS) $(LINK_DBG_FLAGS)
!endif

!endif  # TARGETTYPE

TLBIMP=$(MMLITE_SDK)\tools\bin\tlbimp
