# Copyright (c) Microsoft Corporation. All rights reserved.
#

# This file is included by common.mk

# .obs defines _SPECIAL_CFLAGS (used as -DSIM=1 for some NTU simulator files).
# .obw makes a unicode version from a _TCHAR file and
# .oba makes a non-unicode version respectively
# .bin is used for header stripped images in arm
#
.SUFFIXES:
.SUFFIXES: .asm .s .c .cpp .cs .h .i .$(O) .oba .obs .obw .def .lib .dll .exe .sbr .bsc .rc .res .bin .size .mk

!if !DEFINED(__MACHDEP_AS_RULES)
#
#   rules for .s assembly language files
#
{$(SRC_DIR)\md\$(TARGETCPU)\}.s.$(O):
    copy $< $(OBJECT_DIR)\$(*B).c
    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(CC_SBR) $(*B).c | $(SEDIT) -b -d "#line" | $(SEDIT) -b -z ZAPTOHERE > $(*B).ss
    $(_SILENT)$(_CMD)$(AS) $(AS_FLAGS) $(AS_OUT)$*.$(O) $(AS_SBR) $(*B).ss

{$(SRC_DIR)\}.s.$(O):
    copy $< $(OBJECT_DIR)\$(*B).c
    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(CC_SBR) $(*B).c | $(SEDIT) -b -d "#line" | $(SEDIT) -b -z ZAPTOHERE > $(*B).ss
    $(_SILENT)$(_CMD)$(AS) $(AS_FLAGS) $(AS_OUT)$*.$(O) $(AS_SBR) $(*B).ss

{$(SRC_DIR)\}.s{$(OBJECT_DIR)\}.$(O):
    copy $< $(OBJECT_DIR)\$(*B).c
    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(CC_SBR) $(OBJECT_DIR)\$(*B).c | $(SEDIT) -b -d "#line" | $(SEDIT) -b -z ZAPTOHERE > $(OBJECT_DIR)\$(*B).ss
    $(_SILENT)$(_CMD)$(AS) $(AS_FLAGS) $(AS_OUT)$*.$(O) $(AS_SBR) $(OBJECT_DIR)\$(*B).ss
!endif

!if !DEFINED(__MACHDEP_CC_RULES)
#
#   Default rules for C language files
#
{$(SRC_DIR)\}.c.$(O):
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(CC_OUT)$*.$(O) $(CC_SBR) $<

{$(SRC_DIR)\md\$(TARGETCPU)\}.c.$(O):
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(CC_OUT)$*.$(O) $(CC_SBR) $<

.c.$(O):
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(CC_OUT)$*.$(O) $(CC_SBR) $<

# rule top make ASCII version from TCHAR file
{$(SRC_DIR)\}.c.oba:
    @echo ASCII $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(CC_UNDEF)_UNICODE $(CC_OUT)$*.oba $(CC_SBR) $<
!if "$(TARGETCPU)"=="bf" # compiler is "too clever" and renames file
    -del $*.oba
    ren $*.doj $*.oba
!endif

# rule top make UNICODE version from TCHAR file
{$(SRC_DIR)\}.c.obw:
    @echo UNICODE $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(CC_DEF)_UNICODE=1 $(CC_OUT)$*.obw $(CC_SBR) $<
!if "$(TARGETCPU)"=="bf" # compiler is "too clever" and renames file
    -del $*.obw
    ren $*.doj $*.obw
!endif

# rule to make SIMULATOR build out of given file
{$(SRC_DIR)\}.c.obs:
    @echo SPECIAL options "$(_SPECIAL_CFLAGS)" $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(_SPECIAL_CFLAGS) $(CC_OUT)$*.obs $(CC_SBR) $<
!if "$(TARGETCPU)"=="bf" # compiler is "too clever" and renames file
    -del $*.obs
    ren $*.doj $*.obs
!endif

{$(SRC_DIR)\md\$(TARGETCPU)\}.c.obs:
    @echo SPECIAL options "$(_SPECIAL_CFLAGS)" $<
    $(_SILENT)$(_CMD)"$(CC)" $(CC_FLAGS) $(_SPECIAL_CFLAGS) $(CC_OUT)$*.obs $(CC_SBR) $<
!if "$(TARGETCPU)"=="bf" # compiler is "too clever" and renames file
    ren $*.doj $*.obs
!endif

# Rule to compile C++
{$(SRC_DIR)\}.cpp.$(O):
    @echo $<
    $(_SILENT)$(_CMD)"$(CPLUS)" $(CPLUS_FLAGS) $(CC_OUT)$*.$(O) $<

# Rules to generate assembly into .s
{$(SRC_DIR)\}.c.s:
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" -S $(CC_FLAGS) $(CC_OUT)$*.s $(CC_SBR) $<

{$(SRC_DIR)\md\$(TARGETCPU)\}.c.s:
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" -S $(CC_FLAGS) $(CC_OUT)$*.s $(CC_SBR) $<

{$(SRC_DIR)\}.cpp.s:
    @echo $<
    $(_SILENT)$(_CMD)"$(CC)" -S $(CC_FLAGS) $(CC_OUT)$*.s $(CC_SBR) $<

!endif # !DEFINED(__MACHDEP_CC_RULES)

# Rule to preprocess file into .i
{$(SRC_DIR)\}.c.i:
    @echo $<
    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(CC_OUT)$*.i $(CC_SBR) $<

{$(SRC_DIR)\}.cpp.i:
    @echo $<
    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(CC_OUT)$*.i $(CC_SBR) $<

###   rule for IMPORT LIBRARY files (derived from DEF files)
##{$(SRC_DIR)\}.def{$(LIBRARY_DIR)\}.lib:
##    @echo creating IMPORT library $(@F)
##    $(_SILENT)copy $< $(OBJECT_DIR)\$(<B).c $(_NUL)
##    $(_SILENT)$(_CMD)$(CPRE) $(CC_FLAGS) $(OBJECT_DIR)\$(<B).c >$(OBJECT_DIR)\$(<B).def
##    $(_SILENT)"$(LIBRARIAN)" $(LIB_FLAGS) $(LIB_DEF)$(OBJECT_DIR)\$(<B).def $(LIB_OUT)$*.lib


!if !DEFINED(__MACHDEP_LIB_RULES)
#
#   Inference rules for building libraries.
#   Requires an object with the same name as the library.
#
.$(O){$(LIBRARY_DIR)\}.lib:
    -del $@
    @echo Creating LIBRARY $@
    $(_SILENT)"$(LIBRARIAN)" $(LIB_FLAGS) $(LIB_OUT)$@ $**

!endif # !DEFINED(__MACHDEP_LIB_RULES)


.$(O){$(LIBRARY_DIR)\}.$(O):
    copy $** $@

!if !DEFINED(__MACHDEP_LINK_RULES)

!if DEFINED(PMDLL)
.$(O){$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$*.$(REL) $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    $(_SILENT)$(PMDLL) $*.$(REL) $@
    $(_SILENT)$(MAYBEDELETE) $*.$(REL)

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@ from libraries
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$*.$(REL) $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    $(_SILENT)$(PMDLL) $*.$(REL) $@
    $(_SILENT)$(MAYBEDELETE) $*.$(REL)

.$(O){$(SYSTEM_DIR)}.cob:
    @echo Linking COB $@
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(COB_ENTRY) $(LINK_OUT)$*.$(REL) $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    $(_SILENT)$(PMDLL) $*.$(REL) $@
    $(_SILENT)$(MAYBEDELETE) $*.$(REL)

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.cob:
    @echo Linking COB $@ from libraries
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(COB_ENTRY) $(LINK_OUT)$*.$(REL) $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    $(_SILENT)$(PMDLL) $*.$(REL) $@
    $(_SILENT)$(MAYBEDELETE) $*.$(REL)
    
!else
.$(O){$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@ from libraries
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$*.@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

.$(O){$(SYSTEM_DIR)}.cob:
    @echo Linking COB $@
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(COB_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.cob:
    @echo Linking COB $@ from libraries
    $(_SILENT)"$(LINK)" $(LINK_EXE_FLAGS) $(COB_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    
!endif    

!if "$(TARGETCPU)"=="i386"
.$(O){$(SYSTEM_DIR)}.dll:
    @echo Linking DLL $@
    $(_SILENT)"$(LINK)" $(LINK_DLL_FLAGS) $(LINK_ENTRY) $(LINK_DEF)$(SRC_DIR)\$(@B).def $(LINK_IMPLIB)$(LIBRARY_DIR)\$(@B)_exp.lib $(LINK_OUT)$*.dll $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!else
.$(O){$(SYSTEM_DIR)}.dll:
    @echo Linking DLL $@ (XXX not really)
    $(_SILENT)"$(LINK)" $(LINK_DLL_FLAGS) $(LINK_ENTRY) $(LINK_OUT)$*.dll $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!endif

!endif # !DEFINED(__MACHDEP_LINK_RULES)

#   rule for resource files
{$(SRC_DIR)\}.rc.res:
!if "$(RC)" != ""
    $(_SILENT)echo $(<F)
    $(_SILENT)"$(RC)" $(RC_FLAGS) $(RC_OUT)$*.res $<
!else
    $(_SILENT)echo No resource compiler for this configuration.
!endif

# Rule for installing generated include files
.h{$(GEN_INC_DIR)\}.h:
	$(_SILENT)$(MMLITE_SDK)\tools\bin\mvifchng $< $@

# Some of them have a .c suffix
.c{$(GEN_INC_DIR)\}.c:
	$(_SILENT)$(MMLITE_SDK)\tools\bin\mvifchng $< $@

# Rule for installing generated package files
.c{package\}.c:
	$(_SILENT)$(MMLITE_SDK)\tools\bin\mvifchng $< $@


# Generate size summary
{$(SYSTEM_DIR)\}.exe{$(SYSTEM_DIR)\}.size:
        %SYSTEMROOT%\system32\cscript //Nologo $(MMLITE_SDK)\conf\scripts\sizes.js -t $(TARGETCONFIG) -c $(MMLITE_SDK)\conf\fileclass $(SYSTEM_DIR)\$(@B).map > $(SYSTEM_DIR)\$(@B).size
