# Copyright (c) Microsoft Corporation. All rights reserved.
#

all: allfirst $(CREATE_DIRS) cdobj $(TARGETS) $(SUBDIRS)

allfirst:
    $(_SILENT)set GCC_EXEC_PREFIX=$(GCC_EXEC_PREFIX)
    $(_SILENT)set INCLUDE=
    $(_SILENT)set LIB=
    $(_SILENT)set TMPDIR=$(TMP)
    $(_SILENT)set TCS=$(TOOLS_DIR)
    $(_SILENT)set PATH=$(PATH)
    $(_SILENT)set _MAKE_TARGET_=all
    $(_SILENT)set LM_LICENSE_FILE=$(LM_LICENSE_FILE)

cdobj:
    $(_SILENT)cd $(OBJECT_DIR)

## Use this one to hand build: e.g. nmake setflags foo.obj
setflags:
    $(_SILENT)set GCC_EXEC_PREFIX=$(GCC_EXEC_PREFIX)
    cd $(OBJECT_DIR)    

help:
    @echo Usage: "nmake [options] [component(s) and/or configuration(s)]"
    @echo.
    @echo Options include:
    @echo.
    @echo   TARGETCPU={i386,arm,tm1,h8} ($(TARGETCPU) assumed)
    @echo   TARGETTYPE={debug,release}  ($(TARGETTYPE) assumed)
    @echo   _VERBOSE={0,1}              (0 assumed, 1 for verbose)
    @echo.
    @echo Look for your binaries in $(SYSTEM_DIR).
    @echo Echo to hand build, use the setflags target, e.g.
    @echo nmake setflags foo.i
    @echo The result goes to $(OBJECT_DIR)

env:
    @set

$(CREATE_DIRS):
!if "$(PROCESSOR_ARCHITECTURE)"==""
    $(_SILENT)if not exist $@\..\..\..\..\..\nul mkdir $@\..\..\..\..\..
    $(_SILENT)if not exist $@\..\..\..\..\nul mkdir $@\..\..\..\..
    $(_SILENT)if not exist $@\..\..\..\nul mkdir $@\..\..\..
    $(_SILENT)if not exist $@\..\..\nul mkdir $@\..\..
    $(_SILENT)if not exist $@\..\nul mkdir $@\..
    $(_SILENT)if not exist $@\nul mkdir $@
!endif
    $(_SILENT)if not exist $@\nul mkdir $@
    $(_SILENT)if not exist $(SYSTEMDRIVE)\tmp\nul mkdir $(SYSTEMDRIVE)\tmp


# The clean target. Delete here and recurse
#
clean: cleanfirst $(SUBDIRS)
!if DEFINED(TARGETS)
    -$(_SILENT)del /q $(TARGETS) $(_NUL)
!endif
    -$(_SILENT)del /q $(OBJECT_DIR) $(_NUL)

cleanfirst:
    $(_SILENT)set _MAKE_TARGET_=clean

#
#   Rules to support chaining makefiles into subdirectories.
#   If you are building directories, you must define "SUBDIRS" BEFORE
#   including this file
#
#   Example for executing makefiles in subdirectories A, B, C:
#
#   SUBDIRS=A.dir B.dir C.dir
#   !include {....}\Common.mk
#
#   Notes:
#   * the dummy extension ".dir" is required
#

!ifdef SUBDIRS
$(SUBDIRS):
    $(_SILENT)cd $(SRC_DIR)\$*
    $(_SILENT)nmake /nologo $(DEFAULT_MAKE_ARGS) /$(MAKEFLAGS) %%_MAKE_TARGET_%%
    $(_SILENT)cd $(SRC_DIR)
!endif

!if "$(COPY_FILES)"!=""
$(COPY_FILES) : $(COPY_SRC_DIR)\$(@F)
    $(_SILENT)echo copy $(@F)
    $(_SILENT)copy $(COPY_SRC_DIR)\$(@F) $@ > nul
!endif

#
# Dependency processing
#
TMPDEP=$(OBJECT_DIR)\tmp1.dep
TMPDEP2=$(OBJECT_DIR)\tmp2.dep
DEPENDENCY_FILE=$(DEPENDENCY_DIR)\$(TARGETCPU).dep

MDFLAGS = -x .$(O) $(MMLITE_INCS) -I$(GEN_INC_DIR) $(CC_INCS) -r $(GEN_INC_DIR) $$(GEN_INC_DIR) -r $(MMLITE_BUILD_DIR) $$(MMLITE_BUILD_DIR) -r $(MMLITE_SDK) $$(MMLITE_SDK)

!if DEFINED(WIN32_INCS)
MDFLAGS= $(MDFLAGS) -K $(WIN32_INCS)
!endif

# Run mkdpend on all the files.
# Sort results so as to make result more stable and overwrite dependency 
# file only if it actually changed.
# Turn all slashes into backslashed or nmake will ignore the dependency.
# Bugs: does not account for .obs .oba .obw rules.
# Bugs: mkdepend generates dependencies for files with trailing tildes but
#       drops the tilde.  Workaround is to del/s *~ before running this.
#
depend_old: depfirst $(CREATE_DIRS) $(SUBDIRS)
    $(_SILENT)if exist $(TMPDEP) del $(TMPDEP)
    -$(_SILENT)$(MMLITE_SDK)\tools\bin\mkdepend $(MDFLAGS) -s $(SRC_DIR)\ *.c??  >> $(TMPDEP)
    -$(_SILENT)$(MMLITE_SDK)\tools\bin\mkdepend $(MDFLAGS) -s $(SRC_DIR)\md\$(TARGETCPU) *.c?? >> $(TMPDEP)
    $(_SILENT)$(SEDIT) -nocr -z / -s \ < $(TMPDEP) | %SYSTEMROOT%\system32\sort -o $(TMPDEP2)
    $(_SILENT)$(MMLITE_SDK)\tools\bin\mvifchng $(TMPDEP2) $(DEPENDENCY_FILE)

# This one is slightly improved and uses jscript.
# Fixes the tilde problem, generates more suffixes, and accounts for files
# in subdirs.  The script reads each file only once so it might even be faster.
MKDEP=%SYSTEMROOT%\system32\cscript //nologo $(MMLITE_SDK)\conf\scripts\mkdep.js -v
depend: depfirst $(CREATE_DIRS) $(SUBDIRS)
    $(_SILENT)$(MKDEP) -o $(DEPENDENCY_FILE) -bs -x .obw -x .oba -x .obs -xp *~ $(MDFLAGS) -s $(SRC_DIR)\ *.c *.c?? *.s md\$(TARGETCPU)\*.c md\$(TARGETCPU)\*.c?? md\$(TARGETCPU)\*.s

depfirst:
    $(_SILENT)set _MAKE_TARGET_=depend

