# Copyright (c) Microsoft Corporation. All rights reserved.
#

# Settings for Philips Trimedia processor

#
#   Environment settings
#
TOOLS_DIR=$(MMLITE_SDK)\tools\tm1
#PATH=$(TOOLS_DIR);$(PATH)
INCLUDE=
LIB=


ENDIAN=l

# Trimedia processor description file
MACHINE = $(TOOLS_DIR)\lib\tm1.md

#
#   Compiler settings
#
CPRE= $(TOOLS_DIR)\bin\cpp


CC_DEF=-D
CC_UNDEF=-U
CC_INC=-I
CC_OUT=-o $(_DUMMY)

CC_FLAGS= -undef -nostdinc $(CC_FLAGS) -D__TCS__=1

TMCCOM=$(TOOLS_DIR)\bin\tmccom
TMCCOM_FLAGS=-e$(ENDIAN) -O3 -no_min_stabs

TMSCHED      = $(TOOLS_DIR)\bin\tmsched
TMSCHED_FLAGS = -noloadspec -e$(ENDIAN) $(MACHINE)

TMVER=$(MMLITE_SDK)\tools\bin\lifever -q -V $(BASERTL_VERSION)

CCECHO=echo

!if "$(ENDIAN)" == "b"
CC_FLAGS       = $(CC_FLAGS) -D__BIG_ENDIAN__
!endif

# Versioning
CC_FLAGS=$(CC_FLAGS) $(CC_DEF)MMLITE_VERSION=$(BASERTL_VERSION)

#
#   Assembler settings
#
AS           = $(TOOLS_DIR)\bin\tmas
AS_FLAGS     = -e$(ENDIAN) $(MACHINE)
AS_OUT= -o=

#
#   Linker settings
#
LINK         = $(TOOLS_DIR)\bin\tmld
LINK_FLAGS=-e$(ENDIAN) -exec -literal -start=__start
LINK_OUT=-o $(_DUMMY)
LINK_START_GROUP=--start-group
LINK_END_GROUP=--end-group

LINK_EXE_FLAGS=$(LINK_FLAGS)

!ifndef EXE_ENTRY
EXE_ENTRY= $(LIBRARY_DIR)\_crt0.$(O)
!endif
COB_ENTRY=-e_CobMainCRT -u_CobMainCRT


#
#   Librarian settings
#
LIBRARIAN=$(TOOLS_DIR)\bin\tmar
LIB_FLAGS=-rc $(LIB_FLAGS)
LIB_OUT=


TMCOMP       = $(TOOLS_DIR)\bin\tmcomp
TMCOMP_FLAGS  = -e$(ENDIAN)

TMCFE        = $(TOOLS_DIR)\bin\tmcfe
TMMM         = $(MMLITE_SDK)\tools\bin\tmmm

LOAD         = $(LINK)
!ifndef TM_LOAD_BASE
TM_LOAD_BASE=0xfa019440
!endif
LOAD_FLAGS    = -e$(ENDIAN) -literal -load=$(TM_LOAD_BASE),0xFA800000

### Override default build rules
__MACHDEP_AS_RULES=1
__MACHDEP_CC_RULES=1
__MACHDEP_LINK_RULES=1


#
#   rules for .s assembly language files
#
#  no preprocessing on this platform...
#
{$(SRC_DIR)\md\$(TARGETCPU)\}.s.$(O):
    $(_SILENT)$(_CMD)$(AS) $(AS_OUT)$*.$(O) $(AS_FLAGS) $**

{$(SRC_DIR)\}.s.$(O):
    $(_SILENT)$(_CMD)$(AS) $(AS_OUT)$*.$(O) $(AS_FLAGS) $**


.c.$(O):
    @echo compiling $(<B).c
    $(_SILENT)$(CPRE) $(CC_FLAGS) $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

{$(SRC_DIR)\}.c.$(O):
    @echo compiling $(<B).c
    $(_SILENT)$(CPRE) $(CC_FLAGS) $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

{$(SRC_DIR)\md\$(TARGETCPU)\}.c.$(O):
    @echo compiling $(<B).c
    $(_SILENT)$(CPRE) $(CC_FLAGS) $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

{$(SRC_DIR)\}.c.oba:
    @echo compiling ASCII $(<B).c
    $(_SILENT)$(CPRE) $(CC_FLAGS) $(CC_UNDEF)_UNICODE $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

{$(SRC_DIR)\}.c.obw:
    @echo compiling UNICODE $(<B).c
    $(_SILENT)$(CPRE) $(CC_FLAGS) $(CC_DEF)_UNICODE $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

{$(SRC_DIR)\}.c.obs:
    @echo SPECIAL options "$(_SPECIAL_CFLAGS)" $<
    $(_SILENT)$(CPRE) $(CC_FLAGS) $(_SPECIAL_CFLAGS) $< $*.i
    $(_SILENT)$(TMCCOM)  -o $*.t $(TMCCOM_FLAGS)  $*.i
    $(_SILENT)$(TMSCHED) -o=$*.s  $(TMSCHED_FLAGS) $*.t
    $(_SILENT)$(AS)    -o=$@ $(AS_FLAGS) $*.s

.$(O){$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@
!if !EXIST($(SYSTEM_DIR)\rtl)
    echo nothing > $(SYSTEM_DIR)\rtl
!endif
    $(_SILENT)$(TOOLS_DIR)\bin\tmstupid $(LINK) -o=$*.lx $(LINK_EXE_FLAGS) $(EXE_ENTRY) $**
    $(_SILENT)$(TMCOMP) -o=$*.cx $(TMCOMP_FLAGS) $*.lx
    $(_SILENT)$(LOAD) -o=$@ $(LOAD_FLAGS) $*.cx
    $(_SILENT)$(TMMM) $(SYSTEM_DIR)\rtl $@
    $(_SILENT)$(TMVER) $@

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.exe:
    @echo Linking EXE $@ from libraries
!if !EXIST($(SYSTEM_DIR)\rtl)
    echo nothing > $(SYSTEM_DIR)\rtl
!endif
    $(_SILENT)$(TOOLS_DIR)\bin\tmstupid $(LINK) -o=$*.lx $(LINK_EXE_FLAGS) $(EXE_ENTRY) $**
    $(_SILENT)$(TMCOMP) -o=$*.cx $(TMCOMP_FLAGS) $*.lx
    $(_SILENT)$(LOAD) -o=$@ $(LOAD_FLAGS) $*.cx
    $(_SILENT)$(TMMM) $(SYSTEM_DIR)\rtl $@
    $(_SILENT)$(TMVER) $@

.$(O){$(SYSTEM_DIR)}.cob:
    @echo NOTYET Linking COB $@

{$(LIBRARY_DIR)}.lib{$(SYSTEM_DIR)}.cob:
    @echo NOTYET Linking COB $@ from libraries
