# Copyright (c) Microsoft Corporation. All rights reserved.
#
# Build cerf image using ARM tools
# separate file because of different link address

SUBDIR=conf\package

!if "$(TOOLS)"=="arm"
EXE_ENTRY=
_LFLAGS=$(_LFLAGS) -aif -ro-base 0xBFFFFF80 -reloc
FSIMAGE=_cerfs.$(O)
LOADER=$(LIBRARY_DIR)\aif_ldr.lib 

!else if "$(TOOLS)"=="ads"
EXE_ENTRY=
_LFLAGS=$(_LFLAGS) -ro-base 0xBFFFFF80
FSIMAGE=_cerfs.$(O)
LOADER=$(LIBRARY_DIR)\elf_ldr.lib 

!else
# GCC
LINK_RELOCATABLE=
EXE_ENTRY=-e_start -u_start -Ttext 0xc0000000
FSIMAGE=cerfs.$(O)
LOADER=$(LIBRARY_DIR)\elf_ldr.lib 
!endif

TARGETS = $(SYSTEM_DIR)\_fredCerf.bin

!include common.mk
!include $(MMLITE_SDK)/conf/common.mk

!if exist (private.mk)
!include private.mk
!endif

#dependecies&configs
CNF_ROMFS_POLL= \
        $(LIBRARY_DIR)\_fredCerf.lib $(COMMON_BASE_LIBS) \
        $(FSIMAGE) $(LOADER) \
	defcobtab.$(O) \
        $(LIBRARY_DIR)\romfs.lib \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB)

#USBDEV=usbdmouse
#USBDEV=usbdkeybd
#USBDEV=usbdserial
USBDEV=$(LIBRARY_DIR)\usbdserial.lib $(LIBRARY_DIR)\usbdkeybd.lib $(LIBRARY_DIR)\usbdmouse.lib

#why isnt this defined already?!?
PROTOCOL_LIBS=	$(LIBRARY_DIR)\protocol.lib \
	$(LIBRARY_DIR)\ip.lib $(LIBRARY_DIR)\dhcp.lib \
	$(LIBRARY_DIR)\tcp.lib $(LIBRARY_DIR)\igmp.lib \
	$(LIBRARY_DIR)\route.lib $(LIBRARY_DIR)\udp.lib \
	$(LIBRARY_DIR)\dns.lib $(LIBRARY_DIR)\sockets.lib \


CNF_HOSTFS_NO_POLL= \
    $(LIBRARY_DIR)\_firstCerf.$(O) $(LIBRARY_DIR)\_fredCerf.lib \
	$(COMMON_BASE_LIBS) $(LOADER) \
	cerfcobtab.$(O) cerf_cfg.$(O) \
    $(LIBRARY_DIR)\hostfs.lib $(LIBRARY_DIR)\serplex.lib \
    $(LIBRARY_DIR)\drivers.lib $(LIBRARY_DIR)\drivercfg.lib \
    $(LIBRARY_DIR)\sa.lib $(LIBRARY_DIR)\null.lib \
    $(LIBRARY_DIR)\udc.lib $(USBDEV) \
    $(LIBRARY_DIR)\ucb_cerf.lib \
    $(LIBRARY_DIR)\cs8900.lib $(PROTOCOL_LIBS) \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
    $(LIBRARY_DIR)\mmcrt.lib

!if DEFINED(_CLR)
CNF_HOSTFS_NO_POLL=$(CNF_HOSTFS_NO_POLL) $(CLR_LIBS)
!endif

$(SYSTEM_DIR)\_fredCerf.exe: $(LIBRARY_DIR)\_startCerf.$(O) \
	runinit.$(O) \
        $(CNF_HOSTFS_NO_POLL)
#        $(CNF_ROMFS_POLL)

#build rule
$(SYSTEM_DIR)\_fredCerf.exe:
    $(_SILENT)$(LINK) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

# .bin rule now generic

ALLTESTS=\
        $(SYSTEM_DIR)\machinfo.exe \
        $(SYSTEM_DIR)\cmdecho.exe \
        $(SYSTEM_DIR)\lsmodule.exe \
        $(SYSTEM_DIR)\algnflt.exe \
        $(SYSTEM_DIR)\tcis.exe \
        $(SYSTEM_DIR)\echo.exe \
        $(SYSTEM_DIR)\himom.exe \
        $(SYSTEM_DIR)\tstshf64.exe \
        $(SYSTEM_DIR)\tstxor64.exe \
        $(SYSTEM_DIR)\tstads64.exe \
        $(SYSTEM_DIR)\tstuads64.exe \
        $(SYSTEM_DIR)\tstdiv64.exe \
        $(SYSTEM_DIR)\tstudiv64.exe \
        $(SYSTEM_DIR)\tstmul64.exe \
        $(SYSTEM_DIR)\tstumul64.exe \
        $(SYSTEM_DIR)\tstmul64s.exe \
        $(SYSTEM_DIR)\tstumul64s.exe \
        $(SYSTEM_DIR)\tstmod64.exe \
        $(SYSTEM_DIR)\tstumod64.exe \
        $(SYSTEM_DIR)\tstmemfn.exe \
        $(SYSTEM_DIR)\heaptest.exe \
        $(SYSTEM_DIR)\heaptstx.exe \
        $(SYSTEM_DIR)\heaptstc.exe \
        $(SYSTEM_DIR)\limage.exe \
        $(SYSTEM_DIR)\nametest.exe \
        $(SYSTEM_DIR)\t1.exe \
        $(SYSTEM_DIR)\t2.exe \
        $(SYSTEM_DIR)\t3.exe \
        $(SYSTEM_DIR)\t4.exe \
        $(SYSTEM_DIR)\t5.exe \
        $(SYSTEM_DIR)\t6.exe \
        $(SYSTEM_DIR)\t7.exe \
        $(SYSTEM_DIR)\raid135.exe \
        $(SYSTEM_DIR)\raid164.exe \
        $(SYSTEM_DIR)\raid182.exe \
        $(SYSTEM_DIR)\ttcreate.exe \
        $(SYSTEM_DIR)\tcswitch.exe \
        $(SYSTEM_DIR)\constrn2.exe \
        $(SYSTEM_DIR)\tmutex.exe \
        $(SYSTEM_DIR)\sleeptst.exe \
        $(SYSTEM_DIR)\constrnt.exe \
        $(SYSTEM_DIR)\alltests.exe \
        $(SYSTEM_DIR)\fileio.exe \
        $(SYSTEM_DIR)\memfunc.exe \
        $(SYSTEM_DIR)\sleep.exe \

!ifndef __NO_CPLUSPLUS
CPLUSTESTS= $(SYSTEM_DIR)\tstcpp.exe \
        $(SYSTEM_DIR)\getargs.exe
!endif

EXES=   $(SYSTEM_DIR)\init.exe \
        $(SYSTEM_DIR)\soaptest.exe \
        $(SYSTEM_DIR)\hostfstest.exe \
        $(SYSTEM_DIR)\morsetest.exe \
        $(SYSTEM_DIR)\buttontest.exe \
        $(SYSTEM_DIR)\romfstest.exe \
        $(SYSTEM_DIR)\hp_state.exe \
        $(SYSTEM_DIR)\lsmodule.exe \
        $(SYSTEM_DIR)\sock.exe \
        $(SYSTEM_DIR)\ttcp.exe \
        $(SYSTEM_DIR)\ping.exe \
        $(SYSTEM_DIR)\route.exe \
        $(SYSTEM_DIR)\cerfslp.exe \
        $(SYSTEM_DIR)\sinout.exe \
        $(SYSTEM_DIR)\calctest.exe \
        $(SYSTEM_DIR)\sensormain.exe \

COBS=   $(SYSTEM_DIR)\http.cob \
        $(SYSTEM_DIR)\sax.cob \
        $(SYSTEM_DIR)\soap.cob \
        $(SYSTEM_DIR)\calc.cob \
        $(SYSTEM_DIR)\sensor.cob \
        $(SYSTEM_DIR)\cobsamp.cob \
        $(SYSTEM_DIR)\tokenizer.cob \
        $(SYSTEM_DIR)\drivers.cob \
        $(OBJECT_DIR)\drivercfg.cob \
        $(SYSTEM_DIR)\protocol.cob \
        $(SYSTEM_DIR)\hicob.cob \
        $(SYSTEM_DIR)\null.cob \
        $(SYSTEM_DIR)\cs8900.cob \
        $(SYSTEM_DIR)\sa.cob \
        $(SYSTEM_DIR)\ucb_cerf.cob \
        $(SYSTEM_DIR)\udc.cob \
        $(CERFARM_PRIVATE_COBS) \

DATAFILES= \
        $(SRC_DIR)\drivers.dat \
        $(SRC_DIR)\init.tzk \
        $(SRC_DIR)\default.htm \
        $(SRC_DIR)\soaptest2.xml \
        $(SRC_DIR)\soap.html \


MMVMTEST= \
	$(SYSTEM_DIR)\init.exe \
	$(CERFARM_PRIVATE_EXES)


FSFILES=$(ALLTESTS) $(EXES) $(COBS) $(DATAFILES) $(CPLUSTESTS) $(MMVMTEST)
#FSFILES=$(EXES) $(COBS) $(SRC_DIR)\init.tzk

# fs.flp differs across machines.  Unfortunately, both GCC
# and the ARM tools know of that name, one way or other.
# Hence the obsession with deleting it. Sigh.
_fs.$(O): fs.flp

_cerfs.$(O): _fs.$(O)
        -del *.flp
        -del _cerfs.$(O)
        ren _fs.$(O) _cerfs.$(O)

cerfs.$(O): fs.flp
        $(LINK) -r -b binary $** -o $@
        -del *.flp

fs.flp: $(FSFILES)
        -del *.flp
        $(MMLITE_SDK)\tools\bin\mkmemfs -rom -debug -ascii -align 4 boot.exe boot.exe $(FSFILES)
        ren 0.flp fs.flp
        del boot.exe

fs.flp: $(SRC_DIR)\cerfarm.mk

# Hackery to get a different drivercfg.cob into each platform
# XXX causes some unnecessary linking and packaging
$(OBJECT_DIR)\drivercfg.cob: always
	copy $(SYSTEM_DIR)\cerf_cfg.cob $@

always:
