# Copyright (c) Microsoft Corporation. All rights reserved.
#
# Build ATMEL eb55/EB63 board image
# separate file because of different link address, and other differences

SUBDIR=conf\package

EXE_ENTRY=
EXE_ENTRY3=
USING_GCC=0

!if "$(TOOLS)"=="arm"
_LFLAGS=$(_LFLAGS) -aif -reloc
# Run from RAM
RAM_ENTRY=-ro-base 0x02008000
RAM_ENTRY3=-ro-base 0x03008000
# Run from Flash
# The flash starts at 0x01000000 but the header needs to be stripped so
# link it 0x80 (header size) below the flash start.
FLASH_ENTRY=-ro-base 0x00ffff80 -rw-base 0x02000000
LOADER=$(LIBRARY_DIR)\aif_ldr.lib 
# _eb63fs.$(O)

!else if "$(TOOLS)"=="ads"
EXE_ENTRY=-first _start$(EB).$(O)
EXE_ENTRY3=-first _start$(EB)ext.$(O)
# Run from RAM
_LFLAGS2=-ro-base 0x02008000
_LFLAGS3=-ro-base 0x03008000
LOADER=$(LIBRARY_DIR)\elf_ldr.lib
# _eb63fs.$(O)

!else if "$(TOOLS)"=="evc"
LINK_RELOCATABLE=
# Run from RAM
EXE_ENTRY=-entry:_start /FIXED /BASE:0x02000000 /ALIGN:32 /SECTION:.pdata,D /MAP:$*.map  _bss.$(O)
EXE_ENTRY3=-entry:_start /FIXED /BASE:0x03000000 /ALIGN:32 /SECTION:.pdata,D /MAP:$*.map  _bss.$(O)
LOADER=$(LIBRARY_DIR)\pe_ldr.lib _bss.$(O)

!else if "$(TARGETCPU)"=="ppc"
LINK_RELOCATABLE=
BYTESEX=-bigendian
# Run from RAM
EXE_ENTRY=-entry:_start /FIXED /BASE:0x02000000 /ALIGN:32 /SECTION:.pdata,D /MAP:$*.map
EXE_ENTRY3=$(EXE_ENTRY)
LINK_XXX_FLAGS=-release -driver -map -debug -debugtype:cv -incremental:no -opt:ref -opt:icf -nodefaultlib -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text
LOADER=$(LIBRARY_DIR)\pe_ldr.lib

!else
#
# GCC
# NB: This is now used by the mips simulator also
#
USING_GCC=1
!if "$(TARGETCPU)"=="arm"
LINK_RELOCATABLE=-T $(SRC_DIR)\arm.linkscript
FLASH_ADDRESS=0x01000000
!else
!if "$(TARGETCPU)"=="mips"
LINK_RELOCATABLE=-T $(SRC_DIR)\mips.linkscript
FLASH_ADDRESS=0xbfc00000
BYTESEX=-bigendian
!endif
!endif
# Run from RAM
RAM_ENTRY=-e_start -u_start -Ttext 0x02008000
RAM_ENTRY3=-e_start -u_start -Ttext 0x03008000
# Run from Flash
FLASH_ENTRY=-e_start  -u_start -Ttext $(FLASH_ADDRESS) -Tdata 0x02000000
LOADER=$(LIBRARY_DIR)\elf_ldr.lib
# eb63fs.$(O)
!endif

TARGETS= \
	$(SYSTEM_DIR)\$(EB).bin \
	$(SYSTEM_DIR)\$(EB)big.bin \
	$(SYSTEM_DIR)\$(EB)big2.bin \
	$(SYSTEM_DIR)\$(EB)big.size \
	$(SYSTEM_DIR)\$(EB)soap.bin \
	$(SYSTEM_DIR)\$(EB)soap.size \
	$(SYSTEM_DIR)\$(EB)sec.bin \
	$(SYSTEM_DIR)\$(EB)sec.size \
	$(SYSTEM_DIR)\$(EB)ds.bin \
	$(SYSTEM_DIR)\$(EB)ds.size \
	$(SYSTEM_DIR)\$(EB)man.bin \
	$(SYSTEM_DIR)\$(EB)man.size \
	$(SYSTEM_DIR)\$(EB)manu.bin \
	$(SYSTEM_DIR)\$(EB)manu.size \
	$(SYSTEM_DIR)\$(EB)gs.bin \
	$(SYSTEM_DIR)\$(EB)gs.size \
	$(SYSTEM_DIR)\$(EB)ext.bin \
	$(SYSTEM_DIR)\$(EB)ext.size \
	$(SYSTEM_DIR)\$(EB)fs.flp

!include common.mk
!include $(MMLITE_SDK)/conf/common.mk

#dependecies
# _startEb63 is explicit cuz ARM tools ifnot.
# _firstEb63 is explicit cuz different
STARTFILE=$(LIBRARY_DIR)\_start$(EB).$(O)
!if ("$(TARGETCPU)"=="mips") || ("$(TARGETCPU)"=="ppc")
STARTFILE3=$(STARTFILE)
!else
STARTFILE3=$(LIBRARY_DIR)\_start$(EB)ext.$(O)
!endif
BASICS= \
    $(LIBRARY_DIR)\_first$(EB).$(O) $(LIBRARY_DIR)\_fred$(EB).lib \
    $(COMMON_BASE_LIBS) $(LOADER) $(LIBRARY_DIR)\serplex.lib \
    $(LIBRARY_DIR)\hostfs.lib $(LIBRARY_DIR)\romfs.lib \
    $(LIBRARY_DIR)\drivers.lib $(LIBRARY_DIR)\drivercfg.lib \
    $(LIBRARY_DIR)\null.lib \
    $(LIBRARY_DIR)\db161.lib $(LIBRARY_DIR)\bv16.lib \
    $(LIBRARY_DIR)\eb63usart.lib \
	$(LIBRARY_DIR)\gianofb.lib $(LIBRARY_DIR)\gsound.lib \
	$(LIBRARY_DIR)\gianokbd.lib $(LIBRARY_DIR)\gianoms.lib \
    $(LIBRARY_DIR)\sl811.lib $(LIBRARY_DIR)\usbhub.lib \
    $(LIBRARY_DIR)\usbstor.lib  $(LIBRARY_DIR)\pnp.lib 

!if DEFINED(_CLR)
BASICS=$(BASICS) $(CLR_LIBS)
!endif

$(SYSTEM_DIR)\$(EB).exe: $(STARTFILE) $(BASICS) $(EB).$(O) $(EB)_cfg.$(O) $(EB)cobtab.$(O) runinit.$(O) \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib

$(SYSTEM_DIR)\$(EB)big.exe: $(STARTFILE) $(BASICS) $(EB)big.$(O) $(EB)_cfg.$(O) runinit.$(O) \
	everyns.$(O) $(EB63BIG_LIBS) \
	$(PRIVATE_SOAP_LIBS) $(PRIVATE_PROTOCOL_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

$(SYSTEM_DIR)\$(EB)big2.exe: $(STARTFILE) $(BASICS) $(EB)big2.$(O) runinit.$(O) \
	everyns.$(O) $(EB63BIG_LIBS) \
	$(PRIVATE_SOAP_LIBS) $(PRIVATE_PROTOCOL_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

#	$(SOAP_LIBS) $(LIBRARY_DIR)\sernet.lib $(NET_LIBS) \
#	$(LIBRARY_DIR)\wsock32.lib \

$(SYSTEM_DIR)\$(EB)soap.exe: $(STARTFILE) $(BASICS) \
	$(EB)big.$(O) $(EB)_cfg.$(O) $(EB)soap.$(O) wsmanns.$(O) \
	$(EB63BIG_LIBS) $(PRIVATE_SOAP_LIBS) $(PRIVATE_PROTOCOL_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

$(SYSTEM_DIR)\$(EB)sec.exe: $(STARTFILE) $(BASICS) $(EB)sec.$(O) wsmanns.$(O) \
	$(EB)_cfg.$(O) $(SOAP_UDP_LIBS) $(LIBRARY_DIR)\sernet.lib $(NET_UDP_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

$(SYSTEM_DIR)\$(EB)ds.exe: $(STARTFILE) $(BASICS) \
	$(EB)ds.$(O) dsns.$(O) $(EB)_cfg.$(O) \
	$(EB63DS_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \
	
$(SYSTEM_DIR)\$(EB)man.exe: $(STARTFILE) $(BASICS) \
	$(EB)man.$(O) wsmanns.$(O) $(EB)_cfg.$(O) \
	$(LIBRARY_DIR)\sernet.lib $(WS_MAN_TCP_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

$(SYSTEM_DIR)\$(EB)manu.exe: $(STARTFILE) $(BASICS) \
	$(EB)manu.$(O) $(EB)soap.$(O) wsmanns.$(O) $(EB)_cfg.$(O) \
	$(LIBRARY_DIR)\sernet.lib $(WS_MAN_UDP_LIBS) \
	$(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib \

$(SYSTEM_DIR)\$(EB)gs.exe: $(STARTFILE) $(BASICS) $(EB)gs.$(O) runinit.$(O) \
	$(EB)_cfg.$(O) $(LIBRARY_DIR)\sernet.lib $(NET_LIBS) $(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib $(LIBRARY_DIR)\basecrt.lib

$(SYSTEM_DIR)\$(EB)ext.exe: $(STARTFILE3) $(BASICS) $(EB).$(O) $(EB)cobtab.$(O) runinit.$(O) \
	$(EB)_cfg.$(O) $(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib

#build rules. Special link addresses.
$(SYSTEM_DIR)\$(EB).exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(RAM_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!if "$(USING_GCC)"=="1"
    $(_SILENT)$(LINK) $(_LFLAGS2) -r $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(RAM_ENTRY) $(LINK_OUT)$@.rel $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!endif

$(SYSTEM_DIR)\$(EB)big.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)big2.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)soap.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)sec.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)ds.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)
    
$(SYSTEM_DIR)\$(EB)man.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)manu.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

$(SYSTEM_DIR)\$(EB)gs.exe:
    $(_SILENT)$(LINK) $(_LFLAGS2) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!if "$(USING_GCC)"=="1"
    $(_SILENT)$(LINK) $(_LFLAGS2) -r $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(FLASH_ENTRY) $(LINK_OUT)$@.rel $(LINK_START_GROUP) $** $(LINK_END_GROUP)
!endif

$(SYSTEM_DIR)\$(EB)ext.exe:
    $(_SILENT)$(LINK) $(_LFLAGS3) $(LINK_EXE_FLAGS) $(EXE_ENTRY3) $(RAM_ENTRY3) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)


#embedded filesystem, if any.
# MAKE SURE IT FITS. Current expectation is < 512KB!
FSFILES= $(PRIVATE_ROMFS_FILES) \
	$(SYSTEM_DIR)\sock.exe \
	$(SYSTEM_DIR)\runsoap.exe \
	$(SYSTEM_DIR)\runsntp.exe \
	$(SYSTEM_DIR)\proxytest2.exe \
	$(SYSTEM_DIR)\eb63_led.exe \
	$(SYSTEM_DIR)\init.tzk \
	$(SYSTEM_DIR)\datetime.exe \
	$(SYSTEM_DIR)\sr0.exe \
	$(SYSTEM_DIR)\hostfstest.exe \
	$(SYSTEM_DIR)\tzk.cob
#	$(SYSTEM_DIR)\sectest.exe \
#	$(SYSTEM_DIR)\sensormain.exe \
#	$(SYSTEM_DIR)\proxytest3.exe \
#	$(SYSTEM_DIR)\ping.exe \
#	$(SYSTEM_DIR)\s.tzk \
#	$(SYSTEM_DIR)\dnsserver.exe \
#	$(SYSTEM_DIR)\host_database.txt \


$(SYSTEM_DIR)\host_database.txt: $(SRC_DIR)\host_database.txt
        copy $** $@

# fs.flp differs across machines.  Unfortunately, both GCC
# and the ARM tools know of that name, one way or other.
# Hence the obsession with deleting it. Sigh.
_fs.$(O): fs.flp

_$(EB)fs.$(O): _fs.$(O)
        -del fs.flp ?.flp
        -del _$(EB)fs.$(O)
        ren _fs.$(O) _$(EB)fs.$(O)

$(EB)fs.$(O): fs.flp
        $(LINK) -r -b binary $** -o $@
        -del fs.flp ?.flp

$(SYSTEM_DIR)\$(EB)fs.flp fs.flp: $(FSFILES)
        -del fs.flp ?.flp
        $(MMLITE_SDK)\tools\bin\mkmemfs $(BYTESEX) -rom -debug -ascii -align 4 boot.exe boot.exe $(FSFILES)
        ren 0.flp fs.flp
        del boot.exe
	copy fs.flp $(SYSTEM_DIR)\$(EB)fs.flp
	@echo The following should be less than 512KB (524288/0x80000)
	dir $(SYSTEM_DIR)\$(EB)fs.flp

fs.flp: $(SRC_DIR)\$(EB).mk

# Hackery to get a different drivercfg.cob into each platform
# XXX causes some unnecessary linking and packaging
$(OBJECT_DIR)\drivercfg.cob: always
        copy $(SYSTEM_DIR)\$(EB)_cfg.cob $@

$(SYSTEM_DIR)\s.tzk: $(SRC_DIR)\s.tzk
        copy $** $@

$(SYSTEM_DIR)\init.tzk: always
        copy $(SRC_DIR)\eb63_init.tzk $@

always:
