# Copyright (c) Microsoft Corporation. All rights reserved.
#
# Build Xilinx ML40x board image for eMIPS processor
# separate file because of different link address, and other differences

SUBDIR=conf\package

EXE_ENTRY=

#
# GCC
#
!if "$(TARGETCPU)"!="mips"
error "This only works with the mips toolchain"
!endif

LINK_RELOCATABLE=-T $(SRC_DIR)\mips.linkscript
FLASH_ADDRESS=0xbfc00000
BYTESEX=-bigendian

# Run from RAM
RAM_ENTRY=-e_start -u_start -Ttext 0x80000000
# Run from Flash
FLASH_ENTRY=-e_start  -u_start -Ttext $(FLASH_ADDRESS) -Tdata 0x80000000
#Use SE loader
#LOADER=$(LIBRARY_DIR)\elf_ldr.lib
LOADER=$(LIBRARY_DIR)\se_elf_ldr.lib

TARGETS= \
	$(SYSTEM_DIR)\Ml.bin \
	$(SYSTEM_DIR)\Ml.size \
	$(SYSTEM_DIR)\Mlfs.flp

!include common.mk
!include $(MMLITE_SDK)/conf/common.mk

#dependecies
BASICS= $(LIBRARY_DIR)\_startMl.$(O) \
    $(LIBRARY_DIR)\_firstMl.$(O) $(LIBRARY_DIR)\_fredMl.lib \
    $(COMMON_BASE_LIBS) $(LOADER) $(LIBRARY_DIR)\serplex.lib \
    $(LIBRARY_DIR)\hostfs.lib $(LIBRARY_DIR)\romfs.lib \
	Ml_cfg.$(O) \
    $(LIBRARY_DIR)\drivers.lib $(LIBRARY_DIR)\drivercfg.lib \
    $(LIBRARY_DIR)\null.lib \
    $(LIBRARY_DIR)\neilsart.lib $(LIBRARY_DIR)\sysace.lib

!if DEFINED(_CLR)
BASICS=$(BASICS) $(CLR_LIBS)
!endif

$(SYSTEM_DIR)\Ml.exe: $(BASICS) Ml.$(O) Mlcobtab.$(O) runinit.$(O) \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib

#build rules. Special link addresses.
$(SYSTEM_DIR)\Ml.exe:
    $(_SILENT)$(LINK) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(RAM_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

#embedded filesystem, if any.
FSFILES= $(PRIVATE_ROMFS_FILES) \
	$(SYSTEM_DIR)\sock.exe \
	$(SYSTEM_DIR)\tzk.cob \
	$(SYSTEM_DIR)\eb63_led.exe \
	$(SYSTEM_DIR)\datetime.exe \
	$(SYSTEM_DIR)\hostfstest.exe

# fs.flp differs across machines.  Unfortunately, both GCC
# and the ARM tools know of that name, one way or other.
# Hence the obsession with deleting it. Sigh.
_fs.$(O): fs.flp

_Mlfs.$(O): _fs.$(O)
        -del fs.flp ?.flp
        -del _Mlfs.$(O)
        ren _fs.$(O) _Mlfs.$(O)

Mlfs.$(O): fs.flp
        $(LINK) -r -b binary $** -o $@
        -del fs.flp ?.flp

$(SYSTEM_DIR)\Mlfs.flp fs.flp: $(FSFILES)
        -del fs.flp ?.flp
        $(MMLITE_SDK)\tools\bin\mkmemfs $(BYTESEX) -rom -debug -ascii -align 4 boot.exe boot.exe $(FSFILES)
        ren 0.flp fs.flp
        del boot.exe
	copy fs.flp $(SYSTEM_DIR)\Mlfs.flp
	dir $(SYSTEM_DIR)\Mlfs.flp

fs.flp: $(SRC_DIR)\Ml.mk

# Hackery to get a different drivercfg.cob into each platform
# XXX causes some unnecessary linking and packaging
$(OBJECT_DIR)\drivercfg.cob: always
        copy $(SYSTEM_DIR)\Ml_cfg.cob $@

always:
