# Copyright (c) Microsoft Corporation. All rights reserved.
#
SUBDIR=conf\package
!include $(MMLITE_SDK)/conf/make/first.mk

!if exist (private.mk)
!include private.mk
!endif

# Macro for m32.exe (32 bit base image) so as not to have to repeat it.
_M32_=$(OBJECT_DIR)\m32.exe

MMDEFLT_IMAGES=\
    $(_M32_)       \
    $(SYSTEM_DIR)\init.exe      \
    $(SYSTEM_DIR)\protocol.cob  \
    $(SYSTEM_DIR)\drivers.cob   \
    $(OBJECT_DIR)\drivercfg.cob \

#    $(SYSTEM_DIR)\http.exe      \
#    $(SYSTEM_DIR)\wsock32.dll   \
#    $(SYSTEM_DIR)\mscoree.dll   \
#    $(SYSTEM_DIR)\mscorlib.dll  \
#    $(SYSTEM_DIR)\hello.exe

MMTEST_IMAGES=\
    $(_M32_)       \
    $(SYSTEM_DIR)\init.exe      \
    $(SYSTEM_DIR)\alltests.exe  \
    $(SYSTEM_DIR)\machinfo.exe  \
    $(SYSTEM_DIR)\lsmodule.exe  \
    $(SYSTEM_DIR)\memfunc.exe  \
    $(SYSTEM_DIR)\heaptest.exe

COMPONENT_ENTRY=main
LINK_IMAGE=1
CUSTOM_TARGET=1

#
# Build multiple configurations for the same platform
#
MMLITE=$(SYSTEM_DIR)\mmlite.exe
MMSMALL=$(SYSTEM_DIR)\mmsmall.exe
MMTEST=$(SYSTEM_DIR)\mmtest.exe

TARGETS= $(PC_PRIVATE_EXES) $(MMLITE) $(MSMALL) $(MMTEST)

!include common.mk
!include $(MMLITE_SDK)/conf/common.mk

NAME=_start
LIBS= 	$(LIBRARY_DIR)\$(NAME).$(O) $(LIBRARY_DIR)\$(NAME).lib \
	runinit.$(O) $(COMMON_BASE_LIBS) \
	$(LIBRARY_DIR)\pe_ldr.lib $(LIBRARY_DIR)\romfs.lib \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
	$(NO_CLR_LIB)

$(SYSTEM_DIR)\$(NAME).exe: $(LIBS)

SMALL_LIBS= $(LIBRARY_DIR)\$(NAME).$(O) $(LIBRARY_DIR)\$(NAME).lib \
	$(LIBRARY_DIR)\base_sml.lib $(LIBRARY_DIR)\$(NOSCHED).lib \
	$(LIBRARY_DIR)\no_heap.lib $(LIBRARY_DIR)\no_ldr.lib \
	$(LIBRARY_DIR)\basecrt.lib $(LIBRARY_DIR)\util.lib $(INTERFACES_LIB)

_DOSXTND_= $(LIBRARY_DIR)\_dosxtnd.exe

# Rule for building default m32.exe (32-bit base image).
# Alignment and base must be right or the image will not work
# because the 16-bit code makes assumptions.
$(_M32_): $(OBJECT_DIR)\defcobtab.$(O) $(LIBS)
	$(_SILENT)echo Linking image $@
	$(_SILENT)"$(LINK)" -nologo -align:1024 -base:0x00000000 -opt:ref -nodefaultlib -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text -entry:"main" $(LINK_OUT)$@ $**

!if DEFINED(VS71COMNTOOLS)
# VS7.1 added some more stuff in the header? UNTESTED.
$(OBJECT_DIR)\mmsmall32.exe: $(SMALL_LIBS)
	$(_SILENT)echo Linking image $@
	$(_SILENT)"$(LINK)" -nologo -map -align:32 -base:0x00000000 -opt:ref -nodefaultlib -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text -entry:"main" $(LINK_OUT)$@ $**
!else
$(OBJECT_DIR)\mmsmall32.exe: $(SMALL_LIBS)
	$(_SILENT)echo Linking image $@
	$(_SILENT)"$(LINK)" -nologo -map -align:16 -base:0x00000000 -opt:ref -nodefaultlib -merge:.bss=.data -merge:.rdata=.text -merge:.data=.text -entry:"main" $(LINK_OUT)$@ $**
!endif

# Default configuration
$(MMLITE): $(_M32_) $(_DOSXTND_) $(MMDEFLT_IMAGES)
	$(_SILENT)echo Creating image $@
	$(_SILENT)$(MKMEMFS) -debug -ascii -rom -align 4 $(OBJECT_DIR)\mmdeflt.img $(MMDEFLT_IMAGES)
	$(_SILENT)$(MK86BOOT) $(OBJECT_DIR)\mmdeflt.img $(_DOSXTND_) $@

# Other configurations

# Small2
# XXX Need to specify subcomponents in more detail.  This one will not
# XXX be small or even functional.

$(MMSMALL): $(OBJECT_DIR)\mmsmall32.exe $(_DOSXTND_)
	$(_SILENT)echo Creating image $@
	$(_SILENT)$(MK86BOOT) $** $@

# Tests
$(MMTEST): $(_M32_) $(_DOSXTND_) $(MMTEST_IMAGES)
	$(_SILENT)echo Creating image $@
	$(_SILENT)$(MKMEMFS) -debug -ascii -rom -align 4 $(OBJECT_DIR)\mmtest.img $(MMTEST_IMAGES)
	$(_SILENT)$(MK86BOOT) $(OBJECT_DIR)\mmtest.img $(_DOSXTND_) $@


# Hackery to get a different drivercfg.cob into each platform
# XXX causes some unnecessary linking and packaging
$(OBJECT_DIR)\drivercfg.cob: always
	copy $(SYSTEM_DIR)\pc_cfg.cob $@

always:
