# Build MSFT Zenith Platform-2 board image
# separate file because of different link address, and other differences
#
# NB: 128K RAM @00100000, 1meg FLASH @01000000

SUBDIR=conf\package

!include $(MMLITE_SDK)/conf/make/first.mk

!if "$(TOOLS)"=="arm"
EXE_ENTRY=
_LFLAGS=$(_LFLAGS) -aif -ro-base 0x000FFF80 -reloc
#COMMON=_ebfs.$(O) $(LIBRARY_DIR)\aif_ldr.lib 
COMMON=$(LIBRARY_DIR)\aif_ldr.lib 

!else if "$(TOOLS)"=="ads"
EXE_ENTRY=-first _startZ1.$(O)
_LFLAGS=$(_LFLAGS) -ro-base 0x00100000
COMMON=$(LIBRARY_DIR)\elf_ldr.lib 

!else if "$(TOOLS)"=="evc"
LINK_RELOCATABLE=
EXE_ENTRY=-entry:_start /FIXED /BASE:0x00100000 /ALIGN:32 /MAP:$(SYSTEM_DIR)\_fredZ1.map /SECTION:.pdata,D
COMMON=$(LIBRARY_DIR)\pe_ldr.lib _bss.$(O)
#COMMON= $(COMMON) _ebfs.$(O)

!else
# GCC
LINK_RELOCATABLE=

!ifdef _WITH_MINIE
# external RAM
EXE_ENTRY=-e_start -u_start -Ttext 0x02000000
!else
EXE_ENTRY=-e_start -u_start -Ttext 0x00100000
!endif

#COMMON=ebfs.$(O) $(LIBRARY_DIR)\elf_ldr.lib 
COMMON=$(LIBRARY_DIR)\elf_ldr.lib 
#COMMON=$(LIBRARY_DIR)\pe_ldr.lib 
!endif

TARGETS = $(SYSTEM_DIR)\_fredZ1.bin fs.flp

!if exist (private.mk)
!include private.mk
!endif

!include common.mk
!include $(MMLITE_SDK)/conf/common.mk

#dependecies
# _startZ1 is explicit cuz ARM tools ifnot.
# _firstZ1 is explicit cuz different
FREDZ1= \
	$(LIBRARY_DIR)\_startZ1.$(O) \
	$(LIBRARY_DIR)\_firstZ1.$(O) $(LIBRARY_DIR)\_fredZ1.lib \
	$(COMMON_BASE_LIBS) $(COMMON) \
    $(LIBRARY_DIR)\serplex.lib \
	$(LIBRARY_DIR)\hostfs.lib $(LIBRARY_DIR)\romfs.lib \
	z1_cfg.$(O) \
        $(LIBRARY_DIR)\drivers.lib $(LIBRARY_DIR)\drivercfg.lib \
        $(LIBRARY_DIR)\at91x40.lib $(LIBRARY_DIR)\fmradio.lib \
	$(LIBRARY_DIR)\basecrt.lib $(INTERFACES_LIB) \
	$(LIBRARY_DIR)\mmcrt.lib

!ifdef _WITH_MINIE
$(SYSTEM_DIR)\_fredZ1.exe: \
	$(FREDZ1) z1cobtab-demo.$(O) $(LIBRARY_DIR)\demo2.lib $(MINIE_LIBS)
!else
$(SYSTEM_DIR)\_fredZ1.exe: $(FREDZ1) z1cobtab.$(O)
!endif

#build rules
$(SYSTEM_DIR)\_fredZ1.exe:
    $(_SILENT)$(LINK) $(LINK_EXE_FLAGS) $(EXE_ENTRY) $(LINK_OUT)$@ $(LINK_START_GROUP) $** $(LINK_END_GROUP)

!if "$(TOOLS)"=="gnu" || "$(TOOLS)"=="gnu4"
$(SYSTEM_DIR)\_fredZ1.bin: $(SYSTEM_DIR)\_fredZ1.exe 
        $(TOOLS_DIR)\arm-elf-objcopy -O binary $(SYSTEM_DIR)\_fredZ1.exe $(SYSTEM_DIR)\_fredZ1.bin
!else if "$(TOOLS)"=="evc"
$(SYSTEM_DIR)\_fredZ1.bin: $(SYSTEM_DIR)\_fredZ1.exe
        $(SEDIT) -skip 688 $(SYSTEM_DIR)\_fredZ1.exe $(SYSTEM_DIR)\_fredZ1.bin
!else
$(SYSTEM_DIR)\_fredZ1.bin: $(SYSTEM_DIR)\_fredZ1.exe
        $(SEDIT) -skip 128 $(SYSTEM_DIR)\_fredZ1.exe $(SYSTEM_DIR)\_fredZ1.bin
!endif

#embedded filesystem, if any
DATAFILES= $(SRC_DIR)\z1.tzk $(SRC_DIR)\default.htm
FSFILES= $(SYSTEM_DIR)\sernet.cob \
        $(SYSTEM_DIR)\hostfstest.exe $(SYSTEM_DIR)\sr0.exe \
        $(SYSTEM_DIR)\sock.exe $(SYSTEM_DIR)\protocol.cob \
        $(SYSTEM_DIR)\http.cob $(SYSTEM_DIR)\tokenizer.cob \
        $(SYSTEM_DIR)\hp_state.exe \
        $(DATAFILES)

# fs.flp differs across machines.  Unfortunately, both GCC
# and the ARM tools know of that name, one way or other.
# Hence the obsession with deleting it. Sigh.
_fs.$(O): fs.flp

_ebfs.$(O): _fs.$(O)
        -del *.flp
        -del _ebfs.$(O)
        ren _fs.$(O) _ebfs.$(O)

ebfs.$(O): fs.flp
        $(LINK) -r -b binary $** -o $@
        -del *.flp

fs.flp: $(FSFILES)
        -del *.flp
        $(MMLITE_SDK)\tools\bin\mkmemfs -rom -debug -ascii -align 4 boot.exe boot.exe $(FSFILES)
        ren 0.flp fs.flp
        del boot.exe

fs.flp: $(SRC_DIR)\z1.mk

# Hackery to get a different drivercfg.cob into each platform
# XXX causes some unnecessary linking and packaging
$(OBJECT_DIR)\drivercfg.cob: always
        copy $(SYSTEM_DIR)\z1_cfg.cob $@

always:
