<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Interface inheritance expansion -jvh 2000 -->
<xsl:stylesheet version="1.0" exclude-result-prefixes="msxsl local" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:msxsl="urn:schemas-microsoft-com:xslt" xmlns:local="#local-functions" xmlns:mic="x-schema:expanded-schema.xml">

  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>

  <!-- This stylesheet expands interface inheritance (extends) -->
  <!-- Identity transformation template -->
  <xsl:template match="@*|node()"><xsl:copy><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <!-- Some simple templates for sticking in missing defaults. Schema SHOULD do it. -->
  <xsl:template match="mic:has-device"><xsl:copy><xsl:if test="not(@optional)"><xsl:attribute name="optional">true</xsl:attribute></xsl:if><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <xsl:template match="//mic:interface/mic:method"><xsl:copy><xsl:if test="not(@virtual)"><xsl:attribute name="virtual">true</xsl:attribute></xsl:if><xsl:if test="not(@stack-size)"><xsl:attribute name="stack-size">0</xsl:attribute></xsl:if><xsl:choose><xsl:when test="@SCODE"/><xsl:otherwise><xsl:attribute name="type">SCODE</xsl:attribute></xsl:otherwise></xsl:choose><xsl:attribute name="xname"><xsl:choose><xsl:when test="@serialize-as and not(contains(@serialize-as, '()')) and @serialize-as != '@' and @serialize-as != '@wsa:To' and not(contains(@serialize-as, '!'))"><xsl:value-of select="translate(@serialize-as, '@', '')"/></xsl:when><xsl:otherwise><xsl:value-of select="@name"/></xsl:otherwise></xsl:choose></xsl:attribute><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <xsl:template match="mic:constructor"><xsl:copy><xsl:if test="not(@prefix)"><xsl:attribute name="prefix"></xsl:attribute></xsl:if><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <xsl:template match="mic:device"><xsl:copy><xsl:if test="not(@is-net)"><xsl:attribute name="is-net">false</xsl:attribute></xsl:if><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <!-- @serialize-as @foo => foo; foo => foo; else @name -->
  <xsl:template match="mic:struct|mic:field|mic:arg"><xsl:copy><xsl:attribute name="xname"><xsl:choose><xsl:when test="@serialize-as and not(contains(@serialize-as, '()')) and @serialize-as != '@' and @serialize-as != '@wsa:To' and not(contains(@serialize-as, '!'))"><xsl:value-of select="translate(@serialize-as, '@', '')"/></xsl:when><xsl:otherwise><xsl:value-of select="@name"/></xsl:otherwise></xsl:choose></xsl:attribute><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>


  <!-- Add tentative string-table for remoting -->
  <xsl:template match="mic:interface[mic:namespace]">
    <xsl:variable name="current-interface" select="."/>
    <xsl:copy>
      <!-- avoid duplicating existing strings -->
      <xsl:apply-templates select="@*|node()[not(self::mic:string-entry)]"></xsl:apply-templates>

      <!-- Print relevant strings. Sort. Filter later. -->

      <!-- All known names in interface (skip non-remotable fields) -->
      <xsl:for-each select="(.|mic:method|mic:method/mic:arg|mic:struct|mic:struct/mic:field[not(@serialize-as = 'none()')])/@name|mic:method/@responsename|mic:method/@resultname|mic:string-entry|(mic:struct|mic:method|mic:arg|mic:field)/@serialize-as[not(contains(., '()') or . = '@' or @serialize-as = '@wsa:To')]">
        <xsl:sort select="."/>
          <xsl:element name="string-entry" namespace="x-schema:expanded-schema.xml"><xsl:if test="not(parent::mic:method|parent::mic:method-response|parent::mic:struct)"><xsl:attribute name="qualified">no</xsl:attribute></xsl:if><xsl:attribute name="interface-name"><xsl:value-of select="$current-interface/@name"/></xsl:attribute><xsl:attribute name="inheritance-depth">0</xsl:attribute><xsl:value-of select="."/></xsl:element>
      </xsl:for-each>
      <!-- Add on missing responsenames -->
      <xsl:for-each select="mic:method[not(@responsename) and not(@type = 'none')]">
        <xsl:sort select="."/>
          <xsl:element name="string-entry" namespace="x-schema:expanded-schema.xml"><xsl:attribute name="qualified">no</xsl:attribute><xsl:attribute name="interface-name"><xsl:value-of select="$current-interface/@name"/></xsl:attribute><xsl:attribute name="inheritance-depth">0</xsl:attribute><xsl:value-of select="@name"/>Response</xsl:element>
      </xsl:for-each>
      <!-- another default... -->
      <xsl:element name="string-entry" namespace="x-schema:expanded-schema.xml"><xsl:attribute name="qualified">no</xsl:attribute><xsl:attribute name="interface-name"><xsl:value-of select="$current-interface/@name"/></xsl:attribute><xsl:attribute name="inheritance-depth">0</xsl:attribute>Result</xsl:element>
    </xsl:copy>
  </xsl:template>

</xsl:stylesheet>
