<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Interface inheritance expansion -jvh 2000 -->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>

  <!-- This stylesheet expands interface inheritance (extends) -->
  <!-- Identity transformation template -->
  <xsl:template match="/|@*|node()"><xsl:copy><xsl:apply-templates select="@*|node()" /></xsl:copy></xsl:template>

  <!-- Top level template for expanding interface prototypes -->
  <xsl:template match="//mic:interface">
    <xsl:variable name="current-interface" select="."/>
    <xsl:copy>
      <!-- Print all fields that are not methods (do methods later) -->
      <xsl:apply-templates select="@*|node()[name() != 'method' and name() != 'method-group']" />
      <xsl:for-each select="@extends">
        <xsl:apply-templates mode="ExtendInterface" select="//mic:interface[@name = current()]">
          <xsl:with-param name="current-interface" select="$current-interface"/>
          <xsl:with-param name="inheritance-depth" select="1"/>
        </xsl:apply-templates>
      </xsl:for-each>
      <xsl:for-each select="mic:method|mic:method-group">
        <xsl:copy>
          <!-- Put in absolute numbering for methods for genhtml use -->
          <!-- NOTE: this is no longer strictly "absolute" but has a definitive
            -  ordering between methods and method-groups
            -->
          <xsl:attribute name="absolute-index">
            <xsl:value-of select="position()"/>
          </xsl:attribute>
          <!-- Record the name of the interface -->
          <xsl:attribute name="interface-name">
            <xsl:value-of select="$current-interface/@name" />
          </xsl:attribute>
          <xsl:apply-templates select="@*|node()" />
        </xsl:copy>
      </xsl:for-each>
      
      <!-- Copy strings that are not shared with super-class -->
      <!--
      <xsl:for-each select="mic:string-table">
        <xsl:copy>
        </xsl:copy>
      </xsl:for-each>
      -->
    </xsl:copy>
  </xsl:template>

  <!-- sub-template for interfaces -->
  <xsl:template mode="ExtendInterface" match="//mic:interface">
    <xsl:param name="current-interface"/>
    <xsl:param name="inheritance-depth"/>
    <xsl:element name="inherits-interface" namespace="x-schema:expanded-schema.xml">
      <xsl:value-of select="@name"/>
    </xsl:element>
    <xsl:for-each select="@extends">
      <xsl:apply-templates mode="ExtendInterface" select="//mic:interface[@name = current()]">
        <xsl:with-param name="current-interface" select="$current-interface"/>
        <xsl:with-param name="inheritance-depth" select="$inheritance-depth + 1"/>
      </xsl:apply-templates>
    </xsl:for-each>
    <xsl:for-each select="mic:string-entry">
      <xsl:copy>
        <xsl:attribute name="inheritance-depth">
          <xsl:value-of select="$inheritance-depth"></xsl:value-of>
        </xsl:attribute>
        <xsl:apply-templates select="@*[not(name() = 'inheritance-depth')]|node()"/>
      </xsl:copy>
    </xsl:for-each>
    <xsl:for-each select="mic:struct">
      <xsl:element name="extended-struct" namespace="x-schema:expanded-schema.xml">
        <xsl:attribute name="extended-interface">
          <xsl:value-of select="../@name"></xsl:value-of>
        </xsl:attribute>
        <xsl:attribute name="inheritance-depth">
          <xsl:value-of select="$inheritance-depth"></xsl:value-of>
        </xsl:attribute>
        <xsl:apply-templates select="@*|node()" />
      </xsl:element>
    </xsl:for-each>
    <xsl:for-each select="mic:method">
      <xsl:element name="extended-method" namespace="x-schema:expanded-schema.xml">
        <xsl:attribute name="extended-interface">
          <xsl:value-of select="../@name"></xsl:value-of>
        </xsl:attribute>
        <xsl:attribute name="interface-name">
          <xsl:value-of select="$current-interface/@name"></xsl:value-of>
        </xsl:attribute>
        <xsl:attribute name="inheritance-depth">
          <xsl:value-of select="$inheritance-depth"></xsl:value-of>
        </xsl:attribute>
        <xsl:apply-templates select="@*|node()" />
      </xsl:element>
    </xsl:for-each>
  </xsl:template>

  <!-- Template for setting @src-ptr and @implementation-name for subclassed methods -->
  <xsl:template match="mic:subclassed-method">
    <xsl:copy>
      <xsl:if test="not(@source-ptr)">
        <xsl:attribute name="source-ptr">
          <xsl:value-of select="../@source-dir"></xsl:value-of>/<xsl:value-of select="../mic:source-file"></xsl:value-of>#<xsl:value-of select="../mic:implements/@prefix"></xsl:value-of><xsl:value-of select="@name"></xsl:value-of>
        </xsl:attribute>
      </xsl:if>
      <xsl:if test="not(@implementation-name)">
        <xsl:attribute name="implementation-name">
          <xsl:value-of select="../mic:implements/@prefix"/><xsl:value-of select="@name"/>
        </xsl:attribute>
      </xsl:if>
      <xsl:apply-templates select="@*|node()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Template for copying @base in constant definitions -->
  <xsl:template match="mic:constant-value">
    <xsl:copy>
      <xsl:if test="not(@nobase)">
        <xsl:if test="../@base">
          <xsl:attribute name="base">
            <xsl:value-of select="../@base"></xsl:value-of>
          </xsl:attribute>
        </xsl:if>
      </xsl:if>
      <xsl:apply-templates select="@*|node()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Top-level template for determining vtable-name (for gencclass.xsl) -->
  <xsl:template match="//mic:concrete-class/mic:implements">
    <xsl:copy><xsl:attribute name="vtable-name"><!-- First one is "v", the rest "PREFIXv" --><xsl:choose><xsl:when test="../mic:implements[1]/@prefix[. = current()/@prefix]">v</xsl:when><xsl:otherwise><xsl:value-of select="@prefix"></xsl:value-of>v</xsl:otherwise></xsl:choose></xsl:attribute><xsl:attribute name="vtable-implementation-name"><xsl:if test="../mic:implements[2]"><xsl:value-of select="@prefix"></xsl:value-of></xsl:if><xsl:value-of select="../@name"></xsl:value-of>Vtbl</xsl:attribute>
      <!-- copy the rest -->
      <xsl:apply-templates select="@*|node()"></xsl:apply-templates></xsl:copy>
  </xsl:template>
  <!-- Top-level template for adding attributes to local constructors -->
  <xsl:template match="//mic:concrete-class/mic:constructor">
    <!-- It is the first implements -->
    <xsl:copy>
      <xsl:attribute name="interface-type"><xsl:for-each select="../mic:implements[1]"><xsl:value-of select="//mic:interface[@name = current()]/@type"></xsl:value-of></xsl:for-each></xsl:attribute>
      <xsl:attribute name="type"><xsl:choose><xsl:when test="@type"><xsl:value-of select="@type"></xsl:value-of></xsl:when><xsl:otherwise><xsl:for-each select="../mic:implements[1]"><xsl:value-of select="//mic:interface[@name = current()]/@type"></xsl:value-of></xsl:for-each></xsl:otherwise></xsl:choose></xsl:attribute>
      <xsl:attribute name="implements-prefix"><xsl:value-of select="../mic:implements[1]/@prefix"></xsl:value-of></xsl:attribute>
      <xsl:attribute name="name"><xsl:choose><xsl:when test="@name"><xsl:value-of select="@name"></xsl:value-of></xsl:when><xsl:otherwise><xsl:value-of select="@prefix"></xsl:value-of><xsl:value-of select="../@name"></xsl:value-of>New</xsl:otherwise></xsl:choose></xsl:attribute>
      <!-- copy the rest -->
      <xsl:apply-templates select="@*[name() != 'type' and name() != 'name']|node()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Zap component specs, they are expanded below -->
  <!--  <xsl:template match="//mic:component"></xsl:template> -->
  <!--  <xsl:template match="//mic:hardware-platform"></xsl:template> -->
  <!--  <xsl:template match="//mic:profile"></xsl:template> -->
  <!-- But skip the schema for later stages (Zaps all attributes) -->
  <xsl:template match="/mic:mmlite-spec/@*"></xsl:template>

  <!-- Top level template for expanding product specifications -->
  <xsl:template match="//mic:product">
    <xsl:copy>
      <xsl:apply-templates select="@*"></xsl:apply-templates>
      <xsl:for-each select="mic:hardware-platform-name">
        <xsl:for-each select="//mic:hardware-platform[@name = current()]">
          <xsl:apply-templates select="@cpu|*[name() != 'description']"></xsl:apply-templates>
        </xsl:for-each>
      </xsl:for-each>
      <!-- expand component containment recursively from our profile -->
      <xsl:for-each select="mic:profile-name">
        <xsl:for-each select="//mic:profile[@name = current()]">
          <xsl:for-each select="mic:contains-component">
            <xsl:copy>
              <xsl:value-of select="."></xsl:value-of>
            </xsl:copy>
            <xsl:for-each select="//mic:component[@name = current()]">
              <xsl:apply-templates mode="mode_IDAESAT"/>
            </xsl:for-each>
          </xsl:for-each>
        </xsl:for-each>
      </xsl:for-each>
      <xsl:apply-templates select="*|comment()|processing-instruction()|text()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode -->
  <xsl:template mode="mode_IDAESAT" match="mic:contains-component">
    <xsl:copy>
      <xsl:value-of select="@*|node()"></xsl:value-of>
    </xsl:copy>
    <xsl:for-each select="//mic:component[@name = current()]">
      <xsl:apply-templates mode="mode_IDAESAT" />
    </xsl:for-each>
  </xsl:template>

  <!-- Set component/@prefix if not set already -->
  <xsl:template match="//mic:component">
    <xsl:copy><xsl:if test="not(@prefix)"><xsl:attribute name="prefix"><xsl:value-of select="@name"/></xsl:attribute></xsl:if><xsl:apply-templates select="@*|node()" /></xsl:copy>
  </xsl:template>

  <!-- Recognize "*" in uses-namespace -->
  <xsl:template match="//mic:uses-namespace[@prefix = '*']">
    <xsl:for-each select="//mic:namespace"><xsl:element
       name="uses-namespace" namespace="x-schema:expanded-schema.xml"><xsl:attribute name="prefix"><xsl:value-of select="@prefix"/></xsl:attribute></xsl:element>
    </xsl:for-each>
  </xsl:template>

  <!-- Recognize "*" in contains-component XXX no can do w/ idref -->
  <xsl:template match="//mic:contains-component[. = '*']">
    <xsl:for-each select="//mic:component"><xsl:element
       name="contains-component" namespace="x-schema:expanded-schema.xml"><xsl:value-of select="@name"/></xsl:element>
    </xsl:for-each>
  </xsl:template>

</xsl:stylesheet>
