<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Class inheritance expansion -jvh 2000 -->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>

  <!-- Simple interface to concrete-class inheritance (implements) -->
  <!-- Identity transformation template -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy></xsl:template>

  <!-- Top level template for expanding concrete class implemented methods -->
  <xsl:template match="//mic:concrete-class">
    <xsl:variable name="current-class" select="."/>
    <xsl:copy>
      <!-- Print all fields that are not methods (do methods by hand) -->
      <xsl:apply-templates select="@*" />
      <xsl:apply-templates select="node()[name() != 'implements-method']" />
      <!-- Generate concrete methods for the methods of the immediate
           interface we implement.  Recursion is handled by later stages.
       -->
      <xsl:for-each select="mic:implements">
        <xsl:variable name="current-implements" select="."/>
        <xsl:for-each select="//mic:interface[@name = current()]">
        <!-- If the interface is the same as the super class then we let
             later stages inherit them rather than subclassing them here.
          -->
        <xsl:variable name="current-interface" select="."/>
        <xsl:choose>
          <xsl:when test="//mic:concrete-class[@name = $current-class/@super-class]/mic:implements[. = current()/@name]">
           <xsl:comment>COMMON INTERFACE WITH SUPERCLASS</xsl:comment>
          </xsl:when>
          <xsl:otherwise>
          <!-- treat the first implements as special, extend the rest -->
          <!-- XXX should match inherited concrete classes with super -->
          <!-- XXX need to handle case where super-class covers only part of inherit -->
            <xsl:for-each select="mic:method | mic:extended-method[not($current-class/@super-class) or (@extended-interface != 'IUnknown' and ($current-class/@super-class = 'DefCob' or $current-class/@super-class = 'CUnknown')) or ($current-implements/preceding-sibling::mic:implements and $current-class/@super-class != 'DefCob' and $current-class/@super-class != 'CUnknown')]">
              <xsl:element name="implements-method" namespace="x-schema:expanded-schema.xml">
              <xsl:choose>
                <xsl:when test="$current-class/mic:implements-method[@name = current()/@name]">
                  <xsl:for-each select="$current-class/mic:implements-method[@name = current()/@name]">
                    <xsl:apply-templates select="@*|node()"/>
                  </xsl:for-each>
                </xsl:when>
                <xsl:otherwise>
                  <xsl:attribute name="implemented-interface">
                    <xsl:value-of select="$current-implements" />
                  </xsl:attribute>
                  <xsl:attribute name="implementation-name">
          <xsl:choose><xsl:when test="$current-class/mic:unimplemented-method[@name = current()/@name]">UNIMPLEMENTED</xsl:when>
            <xsl:otherwise>
    <xsl:value-of select="$current-implements/@prefix"/><xsl:value-of select="@name"/>
          </xsl:otherwise></xsl:choose>
  </xsl:attribute>
                      <!-- Save @prefix and @vtable-name for gencclass.xsl use -->
                      <xsl:attribute name="implementation-prefix">
                        <xsl:value-of select="$current-implements/@prefix"/>
                      </xsl:attribute>
                      <xsl:attribute name="vtable-name">
                        <xsl:value-of select="$current-implements/@vtable-name"/>
                      </xsl:attribute>
                      <xsl:attribute name="implementation-class">
                        <xsl:value-of select="$current-class/@name"/>
                      </xsl:attribute>
        <xsl:attribute name="private"><xsl:choose><xsl:when test="$current-class/mic:subclassed-method[@name = current()/@name]/@private"><xsl:value-of select="$current-class/mic:subclassed-method[@name = current()/@name]/@private"/></xsl:when><xsl:when test="$current-implements/@private"><xsl:value-of select="$current-implements/@private"/></xsl:when><xsl:otherwise>true</xsl:otherwise></xsl:choose></xsl:attribute>
                      <xsl:apply-templates select="@*|node()"/>
                    </xsl:otherwise>
                  </xsl:choose>
                </xsl:element>
              </xsl:for-each>
            </xsl:otherwise>
          </xsl:choose>
          <!-- constructors ONLY get inherited from IMMEDIATE interface -->
          <!-- keep it simple... -->
          <xsl:for-each select="mic:constructor">
            <xsl:choose>
              <!-- if the prefix is the same, the concrete one overrides -->
              <xsl:when test="@prefix = $current-class/mic:constructor/@prefix">
                <xsl:comment>Concrete constructor overrides iface</xsl:comment>
              </xsl:when>
              <xsl:otherwise>
                <xsl:copy>
                  <xsl:attribute name="interface-type"><xsl:value-of select="//mic:interface[@name = $current-implements]/@type"></xsl:value-of></xsl:attribute>
                  <xsl:attribute name="type"><xsl:choose><xsl:when test="@type"><xsl:value-of select="@type"></xsl:value-of></xsl:when><xsl:otherwise><xsl:value-of select="//mic:interface[@name = $current-implements]/@type"></xsl:value-of></xsl:otherwise></xsl:choose></xsl:attribute>
                  <xsl:attribute name="implements-prefix"><xsl:value-of select="$current-implements/@prefix"></xsl:value-of></xsl:attribute>
                  <xsl:attribute name="name"><xsl:choose><xsl:when test="@name"><xsl:value-of select="@name"></xsl:value-of></xsl:when><xsl:otherwise><xsl:value-of select="@prefix"></xsl:value-of><xsl:value-of select="$current-class/@name"></xsl:value-of>New</xsl:otherwise></xsl:choose></xsl:attribute>
                  <xsl:apply-templates select="@*[name() != 'type' and name() != 'name']|node()"/>
                </xsl:copy>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:for-each>
        </xsl:for-each>
      </xsl:for-each>
    </xsl:copy>
  </xsl:template>

  <!-- Back-propagate inheritance to interface -->
  <xsl:template match="//mic:interface">
    <xsl:copy>
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
      <!-- Expand all extends into extended-by -->
      <xsl:apply-templates mode="mode_IDARGAT" select="//mic:interface[@extends = current()/@name]"></xsl:apply-templates>
      <!-- Expand all implements into implemented-by -->
      <!-- only traverse concrete hierarchy -->
      <xsl:apply-templates mode="mode_IDAFHAT" select="//mic:concrete-class[ mic:implements = current()/@name]"></xsl:apply-templates>
      <!-- Expand products -->
      <xsl:apply-templates mode="mode_IDAXHAT" select="."></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode --><xsl:template mode="mode_IDARGAT" match="//mic:interface">
    <xsl:element name="extended-by" namespace="x-schema:expanded-schema.xml">
      <xsl:value-of select="@name"></xsl:value-of>
    </xsl:element>
    <xsl:apply-templates mode="mode_IDARGAT" select="//mic:interface[@extends = current()/@name]"></xsl:apply-templates>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode --><xsl:template mode="mode_IDAFHAT" match="//mic:concrete-class">
    <xsl:element name="implemented-by" namespace="x-schema:expanded-schema.xml">
      <xsl:value-of select="@name"></xsl:value-of>
    </xsl:element>
    <xsl:apply-templates mode="mode_IDAFHAT" select="//mic:concrete-class[@super-class = current()/@name]"></xsl:apply-templates>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode --><xsl:template mode="mode_IDAXHAT" match="//mic:interface">
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:interface[@extends = current()/@name]"></xsl:apply-templates>
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:concrete-class[ mic:implements = current()/@name]"></xsl:apply-templates>
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:concrete-class[ mic:uses-interface/@name = current()/@name]"></xsl:apply-templates>
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:interface[ mic:uses-interface/@name = current()/@name]"></xsl:apply-templates>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode --><xsl:template mode="mode_IDAXHAT" match="//mic:concrete-class">
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:concrete-class[@super-class = current()/@name]"></xsl:apply-templates>
    <xsl:apply-templates mode="mode_IDAXHAT" select="//mic:product[ mic:contains-concrete-class/@name = current()/@name]"></xsl:apply-templates>
  </xsl:template>
  <!-- [XSL-XSLT] Converted nested template to a mode --><xsl:template mode="mode_IDAXHAT" match="//mic:product">
    <xsl:element name="product-name" namespace="x-schema:expanded-schema.xml">
      <xsl:value-of select="@name"></xsl:value-of>
    </xsl:element>
  </xsl:template>
</xsl:stylesheet>
