<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Subclass expansion -jvh 2000 -->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>
  <!-- This stylesheet expands subclassed concrete classes (super-class) -->

  <!-- Identity transformation template -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|node()"></xsl:apply-templates></xsl:copy></xsl:template>

  <!-- Top level template for expanding concrete class implemented methods -->
  <xsl:template match="//mic:concrete-class">
    <xsl:variable name="current-class" select="."/>
    <xsl:copy>
      <!-- Print all fields that are not methods (do methods later) -->
      <xsl:apply-templates select="@*|node()[name() != 'implements-method']" />
      <!-- copy inherited-interface from interface -->
      <xsl:for-each select="mic:implements">
        <xsl:variable name="current-implements" select="."/>
        <xsl:element name="inherits-interface" namespace="x-schema:expanded-schema.xml"><xsl:attribute name="prefix"><xsl:value-of select="$current-implements/@prefix"/></xsl:attribute>
           <xsl:value-of select="."/>
        </xsl:element>
        <xsl:for-each select="id(.)/mic:inherits-interface"><xsl:copy><xsl:attribute name="prefix"><xsl:value-of select="$current-implements/@prefix"/></xsl:attribute><xsl:value-of select="."/></xsl:copy></xsl:for-each>
      </xsl:for-each>
      <xsl:for-each select="@super-class">
        <xsl:element name="inherited-class" namespace="x-schema:expanded-schema.xml"><xsl:value-of select="."/></xsl:element>
        <xsl:apply-templates mode="mode_IDAGBAT" select="//mic:concrete-class[@name = current()]">
          <xsl:with-param name="current-class" select="$current-class" />
        </xsl:apply-templates>
      </xsl:for-each>
      <xsl:for-each select="mic:implements-method">
        <xsl:copy>
          <!-- Allow overriding @implementation-name & more with subclassed-method
       -->
          <xsl:attribute name="implementation-name">
            <xsl:choose>
              <!-- if the subclassed name is specified use it -->
              <xsl:when test="$current-class/mic:subclassed-method[@name = current()/@name]/@implementation-name">
                <xsl:value-of select="$current-class/mic:subclassed-method[@name = current()/@name]/@implementation-name"></xsl:value-of>
              </xsl:when>
              <!-- Use name of inherited implementation -->
              <xsl:otherwise>
                <xsl:value-of select="@implementation-name"></xsl:value-of>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:attribute>
          <xsl:attribute name="private">
            <xsl:choose>
              <!-- if the subclassed has @private use it -->
              <xsl:when test="$current-class/mic:subclassed-method[@name = current()/@name]/@private">
                <xsl:value-of select="$current-class/mic:subclassed-method[@name = current()/@name]/@private"></xsl:value-of>
              </xsl:when>
              <!-- if the implements had a @private use that one -->
              <xsl:when test="$current-class/mic:implements[. = current()/@implemented-interface]/@private">
                <xsl:value-of select="$current-class/mic:implements[. = current()/@implemented-interface]/@private"></xsl:value-of>
              </xsl:when>
                <!-- when inherited make it public -->
              <xsl:when test="$current-class/mic:inherited-method[@name = current()/@name]">
                false
              </xsl:when>
              <!-- By default the method should be PRIVATE -->
              <xsl:otherwise>
                true
              </xsl:otherwise>
            </xsl:choose>
          </xsl:attribute>
          <xsl:apply-templates select="@*[name() != 'implementation-name' and name() != 'source-ptr']|node()" />
        </xsl:copy>
      </xsl:for-each>
      <!-- Record which classes this is a super-class of
         and backpropagate product names
      -->
      <xsl:apply-templates mode="SuperClass" select="//mic:concrete-class[@super-class = current()/@name]"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Repeat identity transformation in this scope -->
  <xsl:template match="@*|/|node()" mode="mode_IDAGBAT">
    <xsl:param name="current-class"/>
    <xsl:copy><xsl:apply-templates mode="mode_IDAGBAT" select="@*|node()"><xsl:with-param name="current-class" select="$current-class"/></xsl:apply-templates></xsl:copy></xsl:template>

  <!-- Make all inherited methods @private="false" -->
  <xsl:template match="mic:implements-method" mode="mode_IDAGBAT">
    <xsl:param name="current-class"/>
    <xsl:copy><xsl:attribute name="private">false</xsl:attribute><xsl:apply-templates mode="mode_IDAGBAT" select="@*|node()"><xsl:with-param name="current-class" select="$current-class"/></xsl:apply-templates></xsl:copy></xsl:template>

  <xsl:template mode="mode_IDAGBAT" match="@implemented-interface">
    <xsl:param name="current-class"/>
    <xsl:attribute name="implemented-interface">
      <!-- Limitation: If a class has a super-class it can only
           have one implements
        -->
      <xsl:value-of select="$current-class/mic:implements"/>
    </xsl:attribute>
  </xsl:template>

  <!-- For late binding make implementation-name NULL -->
  <xsl:template mode="mode_IDAGBAT" match="@implementation-name">
    <xsl:param name="current-class"/>
    <xsl:attribute name="implementation-name">
      <xsl:choose>
        <xsl:when test="$current-class/mic:unimplemented-method[@name = current()/../@name]">UNIMPLEMENTED</xsl:when>
        <xsl:when test="$current-class[@binding='compile']">
          <xsl:value-of select="."/>
        </xsl:when>
        <xsl:otherwise>NULL</xsl:otherwise>
      </xsl:choose>      
    </xsl:attribute>
  </xsl:template>

  <!-- [XSL-XSLT] Converted nested template to a mode -->
  <xsl:template mode="mode_IDAGBAT" match="//mic:concrete-class">
    <xsl:param name="current-class"/>
    <xsl:for-each select="@super-class">
      <xsl:element name="inherited-class" namespace="x-schema:expanded-schema.xml"><xsl:value-of select="."/></xsl:element>
      <xsl:apply-templates mode="mode_IDAGBAT" select="//mic:concrete-class[@name = current()]"><xsl:with-param name="current-class" select="$current-class"/></xsl:apply-templates>
    </xsl:for-each>
    <xsl:for-each select="mic:implements-method">
      <xsl:choose>
        <xsl:when test="not($current-class/mic:inherited-method[@name = current()/@name]) and not($current-class/mic:unimplemented-method[@name = current()/@name])">
<!-- OLD: inherit is default        <xsl:when test="$current-class/mic:subclassed-method[@name = current()/@name]"> -->
          <xsl:element name="implements-method" namespace="x-schema:expanded-schema.xml">
            <!-- Override implementation-name with that from subclasser
                     Allow overriding it or make it up.
                 -->
            <xsl:attribute name="implementation-name">
              <xsl:choose>
                <xsl:when test="$current-class/mic:subclassed-method[@name = current()/@name]/@implementation-name">
                  <xsl:value-of select="$current-class/mic:subclassed-method[@name = current()/@name]/@implementation-name"></xsl:value-of>
                </xsl:when>
                <xsl:when test="$current-class/mic:inherited-method[@name = current()/@name]/@implementation-name">
                  <xsl:value-of select="$current-class/mic:implements/@prefix"/>
                </xsl:when>
                <xsl:otherwise>
                  <xsl:value-of select="$current-class/mic:implements/@prefix"/><xsl:value-of select="@name"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:attribute>
            <xsl:attribute name="implementation-class">
              <xsl:choose>
                <xsl:when test="$current-class/mic:inherited-method[@name = current()/@name]">
                  <!-- XXX should figure out recursively -->
                  <xsl:value-of select="$current-class/@super-class" />
                </xsl:when>
                <xsl:otherwise>
                  <xsl:value-of select="$current-class/@name"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:attribute>
            <!-- Copy @src-ptr when present -->
            <xsl:if test="$current-class/mic:subclassed-method[@name = current()/@name]/@source-ptr">
              <xsl:attribute name="source-ptr">
                <xsl:value-of select="$current-class/mic:subclassed-method[@name = current()/@name]/@source-ptr"/>
              </xsl:attribute>
            </xsl:if>
            <xsl:apply-templates mode="mode_IDAGBAT" select="@*[name() != 'implementation-name' and name() != 'implementation-class' and name() != 'source-ptr']|node()"><xsl:with-param name="current-class" select="$current-class"/></xsl:apply-templates>
          </xsl:element>
        </xsl:when>
        <xsl:otherwise>
          <xsl:apply-templates mode="mode_IDAGBAT" select="."><xsl:with-param name="current-class" select="$current-class"/></xsl:apply-templates>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>
  </xsl:template>

  <xsl:template mode="SuperClass" match="//mic:concrete-class">
    <xsl:apply-templates mode="SuperClass" select="//mic:concrete-class[@super-class = current()/@name]"></xsl:apply-templates>
    <xsl:element name="superclass-of" namespace="x-schema:expanded-schema.xml">
      <xsl:value-of select="@name" />
    </xsl:element>
    <xsl:for-each select="mic:product-name">
      <xsl:copy>
        <xsl:value-of select="."></xsl:value-of>
      </xsl:copy>
    </xsl:for-each>
  </xsl:template>

  <!-- sort string entries -->
  <xsl:template match="mic:interface">
    <xsl:copy>
      <xsl:apply-templates select="@*|node()[not(name() = 'string-entry')]"/>
      <xsl:for-each select="mic:string-entry">
        <xsl:sort select="@inheritance-depth" order="descending"/>
        <xsl:sort select="text()"/>
        <xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy>
      </xsl:for-each>
    </xsl:copy>
  </xsl:template>

</xsl:stylesheet>
