<?xml version='1.0'?>
<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Generate list of components and their contents -jvh 2007 -->

<xsl:stylesheet version="1.0" exclude-result-prefixes="msxsl local" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:msxsl="urn:schemas-microsoft-com:xslt" xmlns:local="#local-functions" xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>

  <xsl:template match="/">
      <xsl:for-each select="//mic:component">
        /* @FILE:<xsl:value-of select="@name"/>-component */
        <xsl:variable name="current-component" select="."/>
        <xsl:element name="component">
          <xsl:attribute name="name"><xsl:value-of select="@name"/></xsl:attribute>
          <xsl:for-each select="mic:concrete-class">
            <xsl:variable name="current-class" select="."/>
            <xsl:element name="concrete-class">
              <xsl:attribute name="name"><xsl:value-of select="@name"/></xsl:attribute>
              <xsl:apply-templates select="mic:implements-method" />
              <xsl:apply-templates select="id(@super-class)" mode="SUPER"/>
              <xsl:for-each select="mic:implements">
                <xsl:apply-templates select="id(text())" mode="SUPER"/>
              </xsl:for-each>
            </xsl:element>
          </xsl:for-each>
        </xsl:element>
      </xsl:for-each>

  </xsl:template>

  <xsl:template match="mic:concrete-class" mode="INHERITED-METHOD">
    <xsl:param name="method-name"/>
    <xsl:apply-templates select="mic:implements-method[@name = $method-name]"
                         mode="INHERITED-METHOD"/>
  </xsl:template>

  <xsl:template match="mic:implements-method" mode="INHERITED-METHOD">
    <xsl:if test="@implementation-name != 'UNIMPLEMENTED' and @implementation-name != 'NULL'">
      <xsl:element name="inherited-method">
        <xsl:attribute name="name"><xsl:value-of select="@implementation-name"/>
        </xsl:attribute>
        <xsl:attribute name="super-class"><xsl:value-of select="../@name"/>
        </xsl:attribute>
      </xsl:element>
    </xsl:if>
    <xsl:if test="@implementation-name = 'NULL'">
      <xsl:apply-templates select="id(../@super-class)" mode="INHERITED-METHOD">
        <xsl:with-param name="method-name"><xsl:value-of select="@name"/></xsl:with-param>
      </xsl:apply-templates>
    </xsl:if>
  </xsl:template>

  <xsl:template match="mic:implements-method">
    <xsl:if test="@implementation-name != 'UNIMPLEMENTED' and @implementation-name != 'NULL'">
      <xsl:element name="method">
        <xsl:attribute name="name"><xsl:value-of select="@implementation-name"/>
        </xsl:attribute>
      </xsl:element>
    </xsl:if>
    <xsl:if test="@implementation-name = 'NULL'">
      <xsl:apply-templates select="id(../@super-class)" mode="INHERITED-METHOD">
        <xsl:with-param name="method-name"><xsl:value-of select="@name"/></xsl:with-param>
      </xsl:apply-templates>
    </xsl:if>
  </xsl:template>

  <xsl:template match="mic:concrete-class" mode="SUPER">
    <xsl:element name="super-class">
      <xsl:attribute name="name"><xsl:value-of select="@name"/></xsl:attribute>
      <xsl:apply-templates select="id(@super-class)" mode="SUPER"/>
      <xsl:for-each select="mic:implements">
        <xsl:apply-templates select="id(text())" mode="SUPER"/>
      </xsl:for-each>
    </xsl:element>
  </xsl:template>

  <xsl:template match="mic:interface" mode="SUPER">
    <xsl:element name="interface">
      <xsl:attribute name="name"><xsl:value-of select="@name"/></xsl:attribute>
      <xsl:apply-templates select="id(@extends)" mode="SUPER"/>
    </xsl:element>
  </xsl:template>

</xsl:stylesheet>
