<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Generate cross reference file for finding methods etc. -jvh 2000 -->
<xsl:stylesheet version="1.0" exclude-result-prefixes="msxsl local" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:msxsl="urn:schemas-microsoft-com:xslt" xmlns:local="#local-functions" xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="html" indent="yes" omit-xml-declaration="yes"/>

  <!-- Identity transformation template (for copying html tags) -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|node()"></xsl:apply-templates></xsl:copy></xsl:template>
  <!-- Template for printing see-also etc. links -->
  <xsl:template match="mic:see-also|mic:implemented-by|mic:extended-by|@extends|mic:implements|@super-class|mic:superclass-of">
    <A><xsl:attribute name="HREF">
        <xsl:choose>
          <xsl:when test="@link">
            <xsl:value-of select="@link"></xsl:value-of>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="."></xsl:value-of>.htm
        </xsl:otherwise>
        </xsl:choose>
      </xsl:attribute><xsl:value-of select="."/></A>
  </xsl:template>
  <xsl:template match="mic:example">
    <p><b>Example <xsl:value-of select="@name"/>
      </b></p>
    <PRE><CODE><xsl:value-of select="."/></CODE></PRE>
  </xsl:template>
  <xsl:template match="mic:note">
    <p><b>NOTE:</b>
      <xsl:value-of select="."/></p>
  </xsl:template>
  <xsl:template match="mic:description">
    <p><xsl:value-of select="."/></p>
  </xsl:template>
  <!-- HTML header generation -->
  <xsl:template match="mic:interface/@name|mic:concrete-class/@name">
    <xsl:variable name="current-name" select="."/>
    <HEAD>
      <META HTTP-EQUIV="Content-Type" Content="text/html; charset=Windows-1252"></META>
      <TITLE><xsl:value-of select="." /></TITLE>
      <!-- make relative reference to root for scripts -->
      <SCRIPT><xsl:attribute name="SRC"><xsl:for-each select="$current-name">../<xsl:value-of select="local:eval_IDA3CAT(.)"/>/linkcss.js</xsl:for-each></xsl:attribute>
      </SCRIPT>
      <SCRIPT><xsl:attribute name="SRC"><xsl:for-each select="$current-name">../<xsl:value-of select="local:eval_IDAHDAT(.)"/>/langref.js</xsl:for-each></xsl:attribute>
      </SCRIPT>
    </HEAD>
  </xsl:template>
  <!-- more HTML header generation through quasi-template -->
  <xsl:template match="/mic:mmlite-spec">
    <TABLE CLASS="buttonbarshade" CELLSPACING="0"><TR><TD>
        </TD></TR></TABLE>
    <TABLE CLASS="buttonbartable" CELLSPACING="0" ><TR ID="hdr"><TD CLASS="runninghead">MMLite Software Development Kit</TD></TR></TABLE>
    <br/>
    <font color="ff0000" size="1"><i>This information has not yet been publicly disclosed and is Microsoft Confidential.</i></font>
  </xsl:template>
  <!-- HTML generation -->
  <xsl:template match="/">
    <xsl:for-each select="/mic:mmlite-spec/mic:concrete-class">
      <!-- Split files at first occurence only -->
      <xsl:if test="/mic:mmlite-spec/mic:concrete-class[@source-dir = current()/@source-dir][1]/@name[. = current()/@name]">
        /* @FILE:src/<xsl:value-of select="@source-dir"></xsl:value-of>/docindex */
        <HTML>
          <xsl:apply-templates select="@name"></xsl:apply-templates>
          <xsl:apply-templates select="/mic:mmlite-spec"></xsl:apply-templates>
          <BODY TOPMARGIN="0">
            <H1><xsl:value-of select="@source-dir"></xsl:value-of> Index</H1>
            <xsl:for-each select="//mic:concrete-class[@source-dir = current()/@source-dir]">
              <xsl:variable name="current-class" select="."/>
              <xsl:if test="mic:implements-method">
                <B><xsl:value-of select="@name" /> Methods</B>
                <DL>
                  <xsl:for-each select="mic:implements-method">
                    <xsl:variable name="current-method" select="."/>
                    <xsl:choose>
                      <xsl:when test="@implementation-class = $current-class/@name">
                        <LI>
                          <A><xsl:attribute name="HREF"><xsl:for-each select="$current-method"><!-- make relative reference to root -->../<xsl:value-of select="local:eval_IDADGAT(.)"/></xsl:for-each>/<xsl:value-of select="@implementation-class"></xsl:value-of>.htm#<xsl:value-of select="@implementation-name"></xsl:value-of></xsl:attribute><xsl:attribute name="NAME"><xsl:value-of select="@implementation-name"></xsl:value-of></xsl:attribute><xsl:value-of select="@implementation-name"></xsl:value-of>
                          </A>
                        </LI>
                      </xsl:when>
                    </xsl:choose>
                  </xsl:for-each>
                </DL>
              </xsl:if>
            </xsl:for-each>
          </BODY>
        </HTML>
      </xsl:if>
    </xsl:for-each>
  </xsl:template>
  <msxsl:script implements-prefix="local"><![CDATA[
  
    // ******************************************************* //
    // **  xsl:script                                          //
    // **                                                      //
    // **  functions converted from xsl:script blocks          //
    // ******************************************************* //
    
  
  
  
    // ******************************************************* //
    // **  xsl:eval                                            //
    // **                                                      //
    // **  functions representing xsl:eval statements          //
    // ******************************************************* //

    // !! Warning - "this" changed to "__this" throughout, possible undesireable consequences !!  

  
    function eval_IDA3CAT(_contextNodeList)
    {
      var __this = _contextNodeList.item(0);
      return "evalFOO"; // + __this.getAttribute("source-dir").replace(/[a-zA-Z0-9-_]+/g, "..")
                  
    }
    
    
    function eval_IDAHDAT(_contextNodeList)
    {
      var __this = _contextNodeList.item(0);
      return "evalBAR"; //__this.getAttribute("source-dir").replace(/[a-zA-Z0-9-_]+/g, "..")
                  
    }
    
    
    function eval_IDADGAT(_contextNodeList)
    {
      var __this = _contextNodeList.item(0);
      return "evalBAZ"; //__this.getAttribute("source-dir").replace(/[a-zA-Z0-9-_]+/g, "..")
                  
    }
    
    
  
  
    // ******************************************************* //
    // **  expr functions                                      //
    // **                                                      //
    // **  functions representing xsl:if expr="" or            //
    // **  xsl:when expr="" statements                         //
    // ******************************************************* //

    // !! Warning - "this" changed to "__this" throughout, possible undesireable consequences !!  

  
    
  
    // ******************************************************* //
    // **  Boilerplate follows                                 //
    // **                                                      //
    // **  implementations of XTLRuntime functions             //
    // ******************************************************* //
    
    function absoluteChildNumber(pNode)
    {
      var n = 1;
      while (pNode = pNode.previousSibling)
        n++;
      return n;
    }
    function xtlruntime_absoluteChildNumber(pNodeList)
    {
      if (pNodeList)
        return absoluteChildNumber(pNodeList.item(0));
      else
        return "";
    }
  
    
    function ancestorChildNumber(bstrNodeName, pNode)
    {
      // BUGBUG - needs XPath syntax
      var a = pNode.selectSingleNode("ancestor(" + bstrNodeName + ")");
      if (a)
        return childNumber(a);
      else
        return null;
    }
    function xtlruntime_ancestorChildNumber(bstrNodeName, pNodeList)
    {
      if (pNodeList)
      {
        var n = ancestorChildNumber(bstrNodeName, pNodeList.item(0));
        if (n)
          return n;
      }
      return "";
    }
      
    function childNumber(pNode)
    {
      var n = 1;
      var name = pNode.nodeName;
      while (pNode = pNode.previousSibling)
        if (pNode.nodeName == name)
          n++;
      return n;
    }
    function xtlruntime_childNumber(pNodeList)
    {
      if (pNodeList)
        return childNumber(pNodeList.item(0));
      else
        return "";
    }

    function depth(pNode)
    {
      var n = 0;
      while (pNode = pNode.parentNode)
        n++;
      return n;
    }
    function xtlruntime_depth(pNodeList)
    {
      if (pNodeList)
        return depth(pNodeList.item(0));
      else
        return 0;
    }
    
    function formatDate(varDate, bstrFormat, varDestLocale)
    {
      // BUGBUG ignores locale
      var date = varDate;  // BUGBUG need conversion from varDate
      
      var result = "";
      for (var i=0; i<bstrFormat.length; i++)
      {
        switch (bstrFormat.charAt(i)) {
        case "d":
          if (bstrFormat.charAt(i+1) == "d") {
            i++;
            if (bstrFormat.charAt(i+1) == "d") {
              i++;
              if (bstrFormat.charAt(i+1) == "d") {
                //dddd
                i++;
                var longMonthName = new Array("Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday");
                result += longMonthName[date.getDay()];
              }
              else { //ddd
                var shortMonthName = new Array("Sun","Mon","Tue","Wed","Thu","Fri","Sat");
                result += shortMonthName[date.getDay()];
              }
            }
            else //dd
               result += (date.getDate()<10 ? "0" : "") + (date.getDate());
          }
          else //dd
            result += date.getDate();
          break;
        case "M":
          if (bstrFormat.charAt(i+1) == "M") {
            i++;
            if (bstrFormat.charAt(i+1) == "M") {
              i++;
              if (bstrFormat.charAt(i+1) == "M") {
                //MMMM
                i++;
                var longMonthName = new Array("January","February","March","April","May","June","July","August","September","October","November","December");
                result += longMonthName[date.getMonth()];
              }
              else { //MMM
                var shortMonthName = new Array("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec");
                result += shortMonthName[date.getMonth()];
              }
            }
            else //MM
              result += (date.getMonth()<9 ? "0" : "") + (date.getMonth() + 1);
          }
          else //M
            result += date.getMonth() + 1;
          break;
        case "y":
          if (bstrFormat.charAt(i+1) == "y") {
            i++;
            if (bstrFormat.charAt(i+1) == "y" && bstrFormat.charAt(i+2) == "y") {
              // yyyy
              i += 2;
              result += date.getFullYear();
            }
            else //yy
              result += (date.getFullYear() + "").substr(2,2);
          }
          else //y
            result += date.getFullYear() % 100;
          break;
        case "'":
          i++;
          while (bstrFormat.charAt(i) != "'" && i < bstrFormat.length)
            result += bstrFormat.charAt(i++);
          // Handle quoted quotes ('') in the literal
          if (bstrFormat.charAt(i+1) == "'")
            result += "'";
          break;
        default:
          result += bstrFormat.charAt(i);
        }
      }
      
      return result;
    }
    
    function formatIndex(lIndex, bstrFormat)
    {
      var flag = false;
      switch (bstrFormat) {
      case "1":
        return lIndex.toString();
        break;
      case "A":
      case "a":
        var n = (lIndex - 1) % 26;
        var r = Math.floor((lIndex - 1) / 26);
        var digit = String.fromCharCode(bstrFormat.charCodeAt(0) + n);
        var result = "";
        for (var i=0; i<=r; i++)
          result += digit;
        return result;
        break;
      case "i":
        flag = true;
      case "I":
        if (lIndex < 1) return null;
        switch (lIndex % 10) {
        case 0: var ones = ""; break;
        case 1: var ones = "I"; break;
        case 2: var ones = "II"; break;
        case 3: var ones = "III"; break;
        case 4: var ones = "IV"; break;
        case 5: var ones = "V"; break;
        case 6: var ones = "VI"; break;
        case 7: var ones = "VII"; break;
        case 8: var ones = "VIII"; break;
        case 9: var ones = "IX"; break;
        }
        switch (Math.floor(lIndex/10) % 10) {
        case 0: var tens = ""; break;
        case 1: var tens = "X"; break;
        case 2: var tens = "XX"; break;
        case 3: var tens = "XXX"; break;
        case 4: var tens = "XL"; break;
        case 5: var tens = "L"; break;
        case 6: var tens = "LX"; break;
        case 7: var tens = "LXX"; break;
        case 8: var tens = "LXXX"; break;
        case 9: var tens = "XC"; break;
        }
        switch (Math.floor(lIndex/100) % 10) {
        case 0: var hundreds = ""; break;
        case 1: var hundreds = "C"; break;
        case 2: var hundreds = "CC"; break;
        case 3: var hundreds = "CCC"; break;
        case 4: var hundreds = "CD"; break;
        case 5: var hundreds = "D"; break;
        case 6: var hundreds = "DC"; break;
        case 7: var hundreds = "DCC"; break;
        case 8: var hundreds = "DCCC"; break;
        case 9: var hundreds = "CM"; break;
        }
        var thousands = "MMMMMMMMM".substring(0,Math.floor(lIndex/1000) % 10);
        var digits = thousands + hundreds + tens + ones;
        if (flag)
          return digits.toLowerCase();
        else
          return digits;
        break;
      default: // 01, 001, etc.
        // No error checking done.  formatStrings like ".....1" produce some interesting results
        var n = lIndex.toString();
        var leadingZeroes = bstrFormat.substring(0, bstrFormat.length - n.length);
        return leadingZeroes + n;
        break;
      }
    }
    

    function tenToTheNth(n)
    {
      if (n == 0)
        return 1;
      else
        return 10 * tenToTheNth(n-1);
    }

    function formatNumber(number, bstrFormat)
    {
      var afterDecimal = false;
      var afterDecimalPrecision = 0;
      
      // parse the format string, building a template determining precision required for
      //  rounding, whether there is a thousands separator, and performing scaling.
      var thousandsSeparator = false;
      var afterDecimalMinPrecision = 0;
      var afterDecimalMaxPrecision = 0;
      var beforeDecimalMinPrecision = 0;
      var template = "";
      for (var i=0; i<bstrFormat.length; i++)
      {
        switch (bstrFormat.charAt(i)) {
        case "0":
          if (afterDecimal)
          {
            afterDecimalPrecision++;
            afterDecimalMinPrecision = afterDecimalPrecision;
            afterDecimalMaxPrecision = afterDecimalPrecision;
          }
          else
          {
            beforeDecimalMinPrecision++;
          }
          template += "0";
          break;
        case "#":
          if (afterDecimal)
          {
            afterDecimalPrecision++;
            afterDecimalMaxPrecision = afterDecimalPrecision;
          }
          else
          {
            if (beforeDecimalMinPrecision > 0)
              beforeDecimalMinPrecision++;
          }
          template += "#";
          break;
        case ".":
          afterDecimal = true;
          template += ".";
          break;
        case ",":
          if (afterDecimal)
            template += ",";
          else
          {
            if (bstrFormat.charAt(i+1) == "0" || bstrFormat.charAt(i+1) == "#")
              thousandsSeparator = true;
            else
              number /= 1000;
          }
          break;
        case "%":
          number *= 100;
          template += "%";
          break;          
       case '"':
          i++;
          while (bstrFormat.charAt(i) != '"' && i < bstrFormat.length)
            template += '"' + bstrFormat.charAt(i++);
          break;
        default:
          template += bstrFormat.charAt(i);
        }
      }

      // Round the number to the correct number of digits and convert it to to a string;
      number = (Math.round(number*tenToTheNth(afterDecimalMaxPrecision)))/tenToTheNth(afterDecimalMaxPrecision);
      var numberString = number + "";
      if (number < 1)
        numberString = numberString.substring(1);

      // Where is the decimal point?
      quotedTemplate = template.replace(/"\./g,'""');
      var decimalPosition = quotedTemplate.indexOf(".");
      if (decimalPosition < 0)
        decimalPosition = template.length;

      // Fill in the template starting at the decimal point and working left
      var j = numberString.indexOf(".");
      if (j < 0)
        j = numberString.length;
      var insertionPoint = 0;
      var thousands = 0;
      for (var i=decimalPosition; i >=0; i--)
      {
        if (template.charAt(i-1) != '"')
        {
          if ((template.charAt(i) == "0" || template.charAt(i) == "#") && j > 0)
          {
            template = template.substring(0,i) + numberString.charAt(--j) + template.substring(i+1);
            if (thousandsSeparator && j > 0)
            {
              thousands++;
              if (thousands > 2) {
                thousands = 0;
                template = template.substring(0,i) + "," + template.substring(i);
              }
            }
            insertionPoint = i;
          }
          if (template.charAt(i) == "#" && j <= 0)
            template = template.substring(0,i) + template.substring(i+1);
        }
      }
      
      // Insert remaining left digits, with thousands separators as necessary
      if (j > 0)
      {
        for (j; j>0; j--)
        {
          template = template.substring(0,insertionPoint) + numberString.substring(j-1,j) + template.substring(insertionPoint);
          if (thousandsSeparator)
          {
            if (thousands > 2) {
              thousands = 0;
              template = template.substring(0,insertionPoint+1) + "," + template.substring(insertionPoint+1);
            }
            thousands++;
          }
        }
      }
      

      // Where is the decimal point now?
      quotedTemplate = template.replace(/"\./g,'""');
      var decimalPosition = quotedTemplate.indexOf(".");
      if (decimalPosition < 0)
        decimalPosition = template.length;

      // Fill in the template starting at the decimal point and working right
      var j = numberString.indexOf(".");
      if (j < 0)
        j = numberString.length;
      
      for (var i=decimalPosition + 1; i < template.length; i++)
      {
        if (template.charAt(i) == "0" && j < numberString.length)
        {
          j++
          template = template.substring(0,i) + numberString.charAt(j) + template.substring(i+1);
        }
        if (template.charAt(i) == "#")
        {
          if (j < numberString.length)
          {
            j++
            template = template.substring(0,i) + numberString.charAt(j) + template.substring(i+1);
          }
          else
          {
            template = template.substring(0,i) + template.substring(i+1);
            i--;
          }
        }
        if (template.charAt(i) == '"')
          i++;
      }

      template = template.replace(/"/g,"");
      return template;      
    }

    function formatTime(varDate, bstrFormat)
    {
      var time = varDate;  // BUGBUG need conversion from varDate
      
      var result = "";
      for (var i=0; i<bstrFormat.length; i++)
      {
        switch (bstrFormat.charAt(i)) {
        case "H":
          if (bstrFormat.charAt(i+1) == "H") {
            // HH
            i++;
            result += (time.getHours()<10 ? "0" : "") + (time.getHours());
          }
          else // H
            result += time.getHours();
          break;
        case "h":
          var hours = time.getHours() % 12;
          if (hours == 0) hours = 12;
          if (bstrFormat.charAt(i+1) == "h") {
            // hh
            i++;
            result += (hours<10 ? "0" : "") + hours;
          }
          else // h
            result += hours;
          break;
        case "m":
          if (bstrFormat.charAt(i+1) == "m") {
            // mm
            i++;
            result += (time.getMinutes()<10 ? "0" : "") + (time.getMinutes());
          }
          else // m
            result += time.getMinutes();
          break;
       case "s":
          if (bstrFormat.charAt(i+1) == "s") {
            // ss
            i++;
            result += (time.getSeconds()<10 ? "0" : "") + (time.getSeconds());
          }
          else // s
            result += time.getSeconds();
          break;
       case "t":
          if (bstrFormat.charAt(i+1) == "t") {
            // tt
            i++;
            result += time.getHours() < 12 ? "AM" : "PM";
          }
          else // t
            result += time.getHours() < 12 ? "A" : "P";
          break;
       case "'":
          i++;
          while (bstrFormat.charAt(i) != "'" && i < bstrFormat.length)
            result += bstrFormat.charAt(i++);
          // Handle quoted quotes ('') in the literal
          if (bstrFormat.charAt(i+1) == "'")
            result += "'";
          break;
        default:
          result += bstrFormat.charAt(i);
        }
      }
      
      return result;
    }


    // Note that this now returns a string instead of a number
    function uniqueID(pNode)
    {
      if (pNode)
      {
        if (pNode.nodeType == 2)
          return uniqueID(pNode.selectSingleNode("..")) + "_" + pNode.nodeName;
        else if (pNode.parentNode)
          return uniqueID(pNode.parentNode) + "_" + absoluteChildNumber(pNode);
        else
          return absoluteChildNumber(pNode);
      }
      return "";
    }
    function xtlruntime_uniqueID(pNodeList)
    {
      if (pNodeList)
        return uniqueID(pNodeList.item(0));
      else
        return "";
    }

  ]]></msxsl:script>
</xsl:stylesheet>
