<?xml version='1.0'?>
<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Generate cross reference file for finding methods etc. -jvh 2000 -->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" omit-xml-declaration="yes"/>

  <!-- Identity transformation template (for copying html tags) -->
  <xsl:template match="@*|node()"><xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy></xsl:template>

  <!-- Template for printing see-also etc. links -->
  <xsl:template match="mic:see-also|mic:implemented-by|mic:extended-by|@extends|mic:implements|@super-class|mic:superclass-of">
    <A><xsl:attribute name="HREF">
      <xsl:choose>
        <xsl:when test="@link">
          <xsl:value-of select="@link"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="."/>.htm
        </xsl:otherwise>
      </xsl:choose>
    </xsl:attribute><xsl:value-of select="."/></A>
  </xsl:template>  

  <xsl:template match="mic:example" xml:space="preserve">
    <p><b>Example <xsl:value-of select="@name"/> </b></p>
    <PRE><CODE><xsl:value-of select="."/></CODE></PRE>
  </xsl:template>

  <xsl:template match="mic:note">
    <p><b>NOTE:</b> <xsl:value-of select="."/></p>
  </xsl:template>

  <xsl:template match="mic:description">
    <p><xsl:value-of select="."/></p>
  </xsl:template>

  <!-- HTML header generation -->
  <xsl:template match="/mic:mmlite-spec">
    <HEAD>
    <META HTTP-EQUIV="Content-Type" Content="text/html; charset=Windows-1252"/>
    <TITLE>Class documentation index</TITLE>
    <!-- make relative reference to root for scripts -->
    <SCRIPT SRC="linkcss.js"></SCRIPT>
    <SCRIPT SRC="langref.js"></SCRIPT>
    </HEAD>

    <TABLE CLASS="buttonbarshade" CELLSPACING="0"><TR><TD> </TD></TR></TABLE>
    <TABLE CLASS="buttonbartable" CELLSPACING="0"><TR ID="hdr"><TD CLASS="runninghead">MMLite Software Development Kit</TD></TR></TABLE>
    <br/>
    <font color="ff0000" size="1"><i>This information has not yet been publicly disclosed and is Microsoft Confidential.</i></font>
  </xsl:template>

  <!-- HTML generation -->
  <xsl:template match="/" xml:space="preserve">
    <HTML>
    <xsl:apply-templates select="/mic:mmlite-spec"/>
    <BODY TOPMARGIN="0">
    <H1>Class Documentation Index</H1>

    <xsl:for-each select="//mic:concrete-class">
      <xsl:variable name="current-class" select="."/>
        <H2><xsl:value-of select="@source-dir"/> directory</H2>
        <xsl:for-each select="//mic:concrete-class[@source-dir = $current-class/@source-dir]">
          <xsl:if test="mic:implements-method">
            <B><xsl:value-of select="@name"/> Methods</B>
            <DL>
             <xsl:for-each select="mic:implements-method">
              <xsl:choose>
                <!-- Make  link for each implemenation-name (HPExtract)
                     dot-name (CHeap.Extract) c-name (CHeap_Extract)
                     cplus-name (CHeap::Extract)
                -->
                <xsl:when test="@implementation-class = $current-class/@name">
                  <LI>
                  <A><xsl:attribute name="HREF"><xsl:value-of select="@implementation-class"/>.htm#<xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:attribute name="NAME"><xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:value-of select="@implementation-name"/></A>

                  <A><xsl:attribute name="HREF"><xsl:value-of select="@implementation-class"/>.htm#<xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:attribute name="NAME"><xsl:value-of select="@dot-name"/></xsl:attribute><xsl:value-of select="@dot-name"/></A>

                  <A><xsl:attribute name="HREF"><xsl:value-of select="@implementation-class"/>.htm#<xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:attribute name="NAME"><xsl:value-of select="@c-name"/></xsl:attribute><xsl:value-of select="@c-name"/></A>

                  <A><xsl:attribute name="HREF"><xsl:value-of select="@implementation-class"/>.htm#<xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:attribute name="NAME"><xsl:value-of select="@cplus-name"/></xsl:attribute><xsl:value-of select="@cplus-name"/></A>
                  </LI>
                </xsl:when>
              </xsl:choose>
            </xsl:for-each>
            </DL>
          </xsl:if>
        </xsl:for-each>
    </xsl:for-each>
    </BODY>
    </HTML>
  </xsl:template>
</xsl:stylesheet>
