<?xml version='1.0'?>
<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Generate file index files -jvh 2002 -->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
  <xsl:output method="html"/>

  <!-- Identity transformation template (for copying html tags) -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy></xsl:template>

  <!-- Fixup some A tags to save unnecessary writing -->
  <xsl:template match="A">
    <A><xsl:attribute name="HREF">
      <xsl:choose>
        <xsl:when test="@HREF and @HREF = '#'">
          #<xsl:value-of select="." />
        </xsl:when>
        <xsl:when test="@HREF">
          <xsl:value-of select="@HREF"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="." />.htm
        </xsl:otherwise>
      </xsl:choose>
    </xsl:attribute><xsl:value-of select="." /></A></xsl:template><!-- no space -->

  <xsl:template match="note">
    <p><b>NOTE:</b> <xsl:value-of select="." /></p>
  </xsl:template>

  <xsl:template match="description">
    <p><xsl:value-of select="." /></p>
  </xsl:template>

  <!-- HTML header generation -->
  <xsl:template name="MakeHeader">
    <xsl:param name="title" />
    <xsl:param name="rootpath" />
    <HEAD>
    <META HTTP-EQUIV="Content-Type" Content="text/html; charset=Windows-1252"/>
    <TITLE><xsl:value-of select="$title" /></TITLE>
    <SCRIPT>
       <xsl:attribute name="SRC"><xsl:value-of select="$rootpath"/>/refcss.js</xsl:attribute>
     </SCRIPT>
    </HEAD>
    <SCRIPT>PrintHeader(null);</SCRIPT>
  </xsl:template>

  <xsl:template match="dir/@name">
    <A><xsl:attribute name="HREF"><xsl:value-of select="."/>/files.htm</xsl:attribute><xsl:value-of select="."/>/</A>
  </xsl:template>

  <xsl:template match="file/@name">
    <A><xsl:attribute name="HREF"><xsl:value-of select="../@htmlname"/></xsl:attribute><xsl:value-of select="."/></A>
  </xsl:template>

  <xsl:template match="dir|file">
    <xsl:choose>
      <xsl:when test="P"><!-- DT when there is text, else simple list -->
        <DT><B>
          <xsl:apply-templates select="@name"/>
        </B></DT>
      </xsl:when>
      <xsl:otherwise>
        <xsl:apply-templates select="@name"/>
      </xsl:otherwise>
    </xsl:choose>
    <xsl:apply-templates select="*"/>
  </xsl:template>

  <!-- HTML generation -->
  <xsl:template match="/">
    <!-- Generate one file for all interfaces -->
    <HTML>
    <xsl:call-template name="MakeHeader">
      <xsl:with-param name="title">TITLE</xsl:with-param>
      <xsl:with-param name="rootpath"><xsl:value-of select="/source-files/@rootpath"/></xsl:with-param>
    </xsl:call-template>
    <BODY>
    <DIV CLASS="BodyDiv">
    <H1>Source Files</H1>
    <xsl:for-each select="/source-files" xml:space="preserve">
      <xsl:apply-templates select="P|discussion"/>
      <DL>
      <xsl:for-each select="dir">
        <xsl:apply-templates select="."/>
      </xsl:for-each>
      </DL>
      <DL>
      <xsl:for-each select="file">
        <xsl:apply-templates select="."/>
      </xsl:for-each>
      </DL>
    </xsl:for-each>
    <xsl:apply-templates select="see-also"/>
    </DIV>
    <SCRIPT>PrintFooter("..");</SCRIPT>
    </BODY>
    </HTML>
  </xsl:template>
</xsl:stylesheet>
