<?xml version='1.0'?>
<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Generate documentation HTML pages -jvh 2000 -->
<xsl:stylesheet version="1.0" exclude-result-prefixes="msxsl local" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:msxsl="urn:schemas-microsoft-com:xslt" xmlns:local="#local-functions" xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="html"/>

  <!-- Identity transformation template (for copying html tags) -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|node()"/></xsl:copy></xsl:template>

  <!-- Template for printing see-also etc. links -->
  <xsl:template match="mic:see-also|mic:implemented-by|mic:extended-by|@extends|mic:implements|@super-class|mic:superclass-of"  xml:space="preserve">
    <A><xsl:attribute name="HREF">
      <xsl:choose>
        <xsl:when test="@HREF and @HREF = '#'">
          #<xsl:value-of select="." />
        </xsl:when>
        <xsl:when test="@HREF">
          <xsl:value-of select="@HREF"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="." />.htm
        </xsl:otherwise>
      </xsl:choose>
    </xsl:attribute><xsl:value-of select="." /></A>
  </xsl:template>

  <!-- Fixup some A tags to save unnecessary writing -->
  <xsl:template match="mic:A">
    <A><xsl:attribute name="HREF">
      <xsl:choose>
        <xsl:when test="@HREF and @HREF = '#'">
          #<xsl:value-of select="." />
        </xsl:when>
        <xsl:when test="@HREF">
          <xsl:value-of select="@HREF"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="." />.htm
        </xsl:otherwise>
      </xsl:choose>
    </xsl:attribute><xsl:value-of select="." /></A></xsl:template><!-- no space -->

  <xsl:template match="mic:example" xml:space="preserve">
    <P><B>Example <xsl:value-of select="@name"/> </B></P>
    <PRE xml:space="preserve"><CODE><xsl:value-of select="." /></CODE></PRE>
  </xsl:template>

  <xsl:template match="mic:note">
    <P><B>NOTE:</B> <xsl:value-of select="." /></P>
  </xsl:template>

  <xsl:template match="mic:description">
    <P><xsl:value-of select="." /></P>
  </xsl:template>

  <!-- HTML header generation -->
  <xsl:template name="MakeHeader">
    <xsl:param name="title" />
    <HEAD>
    <META HTTP-EQUIV="Content-Type" Content="text/html; charset=Windows-1252"/>
    <TITLE><xsl:value-of select="$title" /></TITLE>
    <!-- <SCRIPT SRC="refcss.js"></SCRIPT> -->
    <LINK REL="stylesheet" HREF="ref4.css"/>
    </HEAD>
      <TABLE CLASS='buttonbarshade' CELLSPACING='0'><TR><TD>\n</TD></TR></TABLE>
      <TABLE CLASS='buttonbartable' CELLSPACING='0'><TR ID='hdr'><TD CLASS='runninghead'>
      Microsoft Invisible Computing
      </TD>
      </TR></TABLE><br/>
  </xsl:template>

  <xsl:template name="MakeFooter">
    <TABLE CLASS='footerbar' CELLSPACING='0'><TR><TD>
</TD></TR></TABLE>
	<TABLE CLASS='footerbartable' CELLSPACING='0'><TR ID='hdr'>
	<TD CLASS='runninghead'>&#xa9;2006 Microsoft Corporation. All rights reserved.</TD>
	<TD CLASS='runninghead'><A HREF='http://www.microsoft.com/isapi/gomscom.asp?target=/info/cpyright.htm' TARGET='_top'>Terms of Use</A></TD>
	<TD CLASS='runninghead'><A HREF='http://www.microsoft.com/isapi/gomscom.asp?target=/info/privacy.htm' TARGET='_top'>Privacy Statement</A></TD>
	<TD CLASS='runninghead'><A HREF='http://www.microsoft.com/isapi/gomscom.asp?target=/enable/' TARGET='_top'>Accessibility</A></TD>
	<TD CLASS='runninghead'><A HREF='../EULA.htm'>End User License Agreement</A></TD>
    </TR></TABLE>
  </xsl:template>

  <xsl:template match="mic:constant">
    <xsl:variable name="current-constant" select="."/>
    <H3><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute></A><xsl:value-of select="@name"/> Constant</H3>
    <xsl:apply-templates select="mic:P"/>
    <xsl:apply-templates select="mic:description"/>
    <xsl:if test="@type"><P>Underlying C type: <CODE><xsl:value-of select="@ctype"/></CODE></P></xsl:if>
    <DL>
    <xsl:for-each select="mic:constant-value">
      <DT xml:space="preserve"><B><A><xsl:attribute name="NAME"><xsl:value-of select="$current-constant/@prefix"/><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="$current-constant/@prefix"/><xsl:value-of select="@name"/></A></B> <sp/> <xsl:choose><xsl:when test="@base">
        <xsl:value-of select="local:eval_FormatConstantValue(.)" />
        </xsl:when><xsl:otherwise><xsl:value-of select="@ordinal"/></xsl:otherwise></xsl:choose></DT>
      <xsl:if test="mic:P"><DD><xsl:apply-templates select="mic:P"/></DD></xsl:if>
      <xsl:if test="mic:description"><DD><xsl:value-of select="mic:description"/></DD></xsl:if>
      <xsl:for-each select="mic:notes"><DD><B>NOTE:</B> <xsl:value-of select="." /></DD></xsl:for-each>
    </xsl:for-each>
    </DL>
    <xsl:apply-templates select="mic:discussion"/>
    <xsl:apply-templates select="mic:notes"/>
    <xsl:apply-templates select="mic:example" />
    <xsl:if test="mic:see-also">
      <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
    </xsl:if>
  </xsl:template>  

  <xsl:template match="mic:struct">
    <H2><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A> Structure</H2>
    <xsl:apply-templates select="mic:P"/>
    <xsl:apply-templates select="mic:description"/>
    <xsl:if test="mic:see-also">
      <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
    </xsl:if>
    <B>Fields</B>
    <DL>
      <xsl:for-each select="mic:field">
        <DT><xsl:value-of select="@name"/></DT>
        <DD>C/C++ type: <CODE><xsl:value-of select="@ctype"/>
          <xsl:if test="@array"> [<xsl:value-of select="@array"/>]</xsl:if>
          </CODE>
        </DD>
        <xsl:if test="../../mic:namespace and @xtype != 'none'">
          <DD>Serializer Type: <CODE><xsl:value-of select="@xtype"/>
            <xsl:if test="@array"> []</xsl:if><xsl:if test="@serialize-as"><!-- (as <xsl:value-of select="@serialize-as"/>)
            --></xsl:if></CODE>
          </DD>
        </xsl:if>
        <DD> <xsl:apply-templates select="mic:P"/> </DD>
        <DD> <xsl:apply-templates select="mic:description"/> </DD>
      </xsl:for-each>
    </DL>
    <xsl:apply-templates select="mic:discussion"/>
    <xsl:apply-templates select="mic:example"/>
    <xsl:apply-templates select="mic:notes"/>
  </xsl:template>

  <xsl:template match="mic:typedef">
    <DL>
    <!-- Print typedef. @arglist means function pointer. C syntax is messy -->
    <DT><CODE xml:space="preserve">typedef <xsl:value-of select="@type"/> <xsl:if test="@arglist">(</xsl:if><xsl:if test="@linkage"><xsl:value-of select="@linkage"/> </xsl:if><xsl:if test="@arglist"> *</xsl:if><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A><xsl:if test="@arglist">)(<xsl:value-of select="@arglist"/>)</xsl:if><xsl:if test="@array"> [<xsl:value-of select="@array"/>]</xsl:if>;</CODE></DT>
    <DD> <xsl:apply-templates select="mic:P"/> </DD>
    <DD> <xsl:apply-templates select="mic:description"/> </DD>
    </DL>
  </xsl:template>

  <xsl:template match="mic:method-group">
    <xsl:variable name="current-group" select="."/>
    <H2><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A> Method Group</H2>
    <xsl:apply-templates select="mic:P"/>
    <xsl:apply-templates select="mic:description"/>
    <xsl:apply-templates select="mic:discussion"/>
    <xsl:apply-templates select="mic:example"/>
    <xsl:apply-templates select="mic:notes"/>
    <xsl:if test="mic:see-also">
      <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
    </xsl:if>
    <B>Methods</B>
    <DL xml:space="preserve">
      <!-- all the methods after this group -->
      <xsl:for-each select="../mic:method[number(@absolute-index) &gt; number(current()/@absolute-index)]">
        <!-- but not after the next -->
        <xsl:if test="../mic:method-group[number(@absolute-index) &lt; number(current()/@absolute-index)][position()=last()]/@name = $current-group/@name" >
          <DT>
            <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:if>
      </xsl:for-each>
    </DL>
  </xsl:template>

  <xsl:template match="mic:method">
    <xsl:param name="current-interface"/>
    <H3><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><!-- <xsl:attribute name="HREF"><xsl:value-of select="$current-interface/@name"/>.htm</xsl:attribute> -->
    <xsl:value-of select="$current-interface/@name"/></A>.<xsl:value-of select="@name"/> Method</H3>
    <xsl:apply-templates select="mic:P"/>
    <xsl:apply-templates select="mic:description"/>
      <PRE xml:space="preserve"><CODE><xsl:value-of select="@ctype"/> <xsl:value-of select="@name"/>( <xsl:choose><xsl:when test="not(mic:this-arg|mic:arg)">void )</xsl:when><xsl:otherwise><xsl:for-each select="mic:this-arg|mic:arg">
    /*<xsl:value-of select="@direction"/><xsl:if test="@optional = 'true'">,optional</xsl:if><xsl:if test="self::mic:this-arg">,this</xsl:if><xsl:if test="@size-is">,size-is</xsl:if><xsl:if test="@length-is">,length-is</xsl:if><xsl:if test="@type-is">,type-is</xsl:if><xsl:if test="@selector-is">,selector-is</xsl:if><xsl:if test="@name = ../mic:arg/@size-is">,size-of</xsl:if><xsl:if test="@name = ../mic:arg/@length-is">,length-of</xsl:if><xsl:if test="@name = ../mic:arg/@type-is">,type-of</xsl:if><xsl:if test="@name = ../mic:arg/@selector-is">,selector-of</xsl:if>*/ <xsl:value-of select="@ctype"/> <xsl:value-of select="@name"/><xsl:if test="position()!=last()">,</xsl:if></xsl:for-each>)</xsl:otherwise></xsl:choose></CODE></PRE>

    <UL type="disc">
      <xsl:choose>
        <xsl:when test="@type = 'SCODE'">
          <LI>Returns <A HREF="IUnknown.htm#S_OK">S_OK</A> if the method is successful, otherwise an <A HREF="IUnknown.htm#SCODE">error code</A>. </LI>
          <xsl:for-each select="mic:return-value">
            <LI><A><xsl:attribute name="HREF">IUnknown.htm#<xsl:value-of select="@value"/></xsl:attribute><xsl:value-of select="@value"/></A></LI><xsl:if test="mic:P"><DD><xsl:apply-templates select="mic:P"/></DD></xsl:if>
          </xsl:for-each>
        </xsl:when>
        <xsl:otherwise>
          <xsl:for-each select="mic:return-value">
            <LI><xsl:value-of select="@value"/></LI>
            <xsl:if test="mic:P"><DD><xsl:apply-templates select="mic:P"/></DD></xsl:if>
          </xsl:for-each>
        </xsl:otherwise>
      </xsl:choose>
    </UL>
    <DL>
    <xsl:for-each select="mic:arg">
      <DT><B><xsl:value-of select="@name"/></B></DT>
      <DD><xsl:apply-templates select="mic:P"/></DD>
      <DD><xsl:value-of select="mic:description"/></DD>
      <xsl:for-each select="mic:notes"><DD><B>NOTE:</B> <xsl:value-of select="." /></DD></xsl:for-each>
    </xsl:for-each>
    </DL>
    <xsl:apply-templates select="mic:discussion"/>
    <xsl:apply-templates select="mic:notes"/>
    <xsl:apply-templates select="mic:example" />
    <xsl:if test="mic:see-also">
      <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
    </xsl:if>

      <H4>C++</H4><PRE xml:space="preserve"><CODE><xsl:value-of select="@ctype"/> <xsl:value-of select="../@name"/>::<xsl:value-of select="@name"/>( <xsl:choose><xsl:when test="not(mic:arg)">void )</xsl:when><xsl:otherwise><xsl:for-each select="mic:arg">
    /*<xsl:value-of select="@direction"/><xsl:if test="@optional = 'true'">,optional</xsl:if><xsl:if test="self::mic:this-arg">,this</xsl:if><xsl:if test="@size-is">,size-is</xsl:if><xsl:if test="@length-is">,length-is</xsl:if><xsl:if test="@type-is">,type-is</xsl:if><xsl:if test="@selector-is">,selector-is</xsl:if><xsl:if test="@name = ../mic:arg/@size-is">,size-of</xsl:if><xsl:if test="@name = ../mic:arg/@length-is">,length-of</xsl:if><xsl:if test="@name = ../mic:arg/@type-is">,type-of</xsl:if><xsl:if test="@name = ../mic:arg/@selector-is">,selector-of</xsl:if>*/ <xsl:value-of select="@ctype"/> <xsl:value-of select="@name"/><xsl:if test="position()!=last()">,</xsl:if></xsl:for-each>)</xsl:otherwise></xsl:choose></CODE></PRE>

<!-- XXX BEGIN UNFINISHED XXX
    < ! - - SOAP method - - >
    <xsl:if test="../mic:namespace">
      <xsl:variable name="ns" select="../mic:namespace/@prefix"/>

      <H4>Request SOAP message</H4><PRE xml:space="preserve"><CODE>&lt;s:Envelope&gt;
  &lt;s:Header&gt;
    &lt;wsa:Action&gt;<xsl:value-of select="$ns/../text()"/>/<xsl:value-of select="@name"/>&lt;/wsa:Action&gt;
    &lt;wsa:MessageId&gt;uuid:ec044ac4-a457-48cf-b8e1-708bcbbaa24f&lt;/wsa:MessageId&gt;
    &lt;wsa:To&gt;http://www.example.com/wsman<xsl:if test="mic:arg[@serialize-as = '@wsa:To']">?<xsl:value-of select="mic:arg[@serialize-as = '@wsa:To']/@name"/>=<xsl:value-of select="mic:arg[@serialize-as = '@wsa:To']/@sampletext"/></xsl:if>&lt;/wsa:To&gt;
  &lt;/s:Header&gt;
  &lt;s:Body&gt;
<xsl:if test="not(@serialize-as = 'text()')">    &lt;<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/><xsl:for-each select="mic:arg[@serialize-as = '@']"> <xsl:value-of select="@name"/>="<xsl:value-of select='@sampletext'/>"</xsl:for-each>&gt;</xsl:if><xsl:for-each select="mic:arg[not(@direction = 'out') and not(substring(@serialize-as,1,1) = '@' or @serialize-as = 'text()' or @serialize-as = 'none()' or ../mic:arg[@size-is = current()/@name] or @name = ../mic:arg/@size-is or @name = ../mic:arg/@length-is or @name = ../mic:arg/@type-is or @name = ../mic:arg/@selector-is)]"><xsl:if test="not(@serialize-as = 'text()')">
      &lt;<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/>&gt;<xsl:value-of select="@sampletext"/>&lt;/<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/>&gt;</xsl:if></xsl:for-each><xsl:if test="not(@serialize-as = 'text()')">
    &lt;/<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/>&gt;</xsl:if>
  &lt;/s:Body&gt;
&lt;/s:Envelope&gt;</CODE></PRE>

      <H4>Response SOAP message</H4><PRE xml:space="preserve"><CODE>&lt;s:Envelope&gt;
  &lt;s:Header&gt;
    &lt;wsa:Action&gt;<xsl:value-of select="$ns/../text()"/>/<xsl:value-of select="@responsename"/><xsl:if test="not(@responsename)"><xsl:value-of select="@name"/>Response</xsl:if>&lt;/wsa:Action&gt;
    &lt;wsa:MessageId&gt;uuid:1f01b7f3-6942-4c35-bd69-b36d29ac6def&lt;/wsa:MessageId&gt;
    &lt;wsa:RelatesTo&gt;uuid:ec044ac4-a457-48cf-b8e1-708bcbbaa24f&lt;/wsa:RelatesTo&gt;
  &lt;/s:Header&gt;
  &lt;s:Body&gt;<xsl:if test="not(@serialize-as = 'text()')">
    &lt;<xsl:value-of select="$ns"/>:<xsl:value-of select="@responsename"/><xsl:if test="not(@responsename)"><xsl:value-of select="@name"/>Response</xsl:if><xsl:for-each select="mic:arg[@serialize-as = '@']"> <xsl:value-of select="@name"/>="<xsl:value-of select='@sampletext'/>"</xsl:for-each>&gt;</xsl:if><xsl:for-each select="mic:arg[(@direction = 'out' or @direction = 'inout') and not(substring(@serialize-as,1,1) = '@' or @serialize-as = 'text()' or @serialize-as = 'none()' or ../mic:arg[@size-is = current()/@name] or @name = ../mic:arg/@size-is or @name = ../mic:arg/@length-is or @name = ../mic:arg/@type-is or @name = ../mic:arg/@selector-is)]"><xsl:if test="not(@serialize-as = 'text()')">
      &lt;<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/>&gt;<xsl:value-of select="@sampletext"/>&lt;/<xsl:value-of select="$ns"/>:<xsl:value-of select="@name"/>&gt;</xsl:if></xsl:for-each><xsl:if test="not(@serialize-as = 'text()')">
    &lt;/<xsl:value-of select="$ns"/>:<xsl:value-of select="@responsename"/><xsl:if test="not(@responsename)"><xsl:value-of select="@name"/>Response</xsl:if>&gt;</xsl:if>
  &lt;/s:Body&gt;
&lt;/s:Envelope&gt;</CODE></PRE>
    </xsl:if>
XXX END UNFINISHED XXX -->

  </xsl:template>

  <!-- SOAP documentation -->
  <xsl:template match="mic:namespace" mode="printxns">
    <P>XML namespace <xsl:value-of select="@prefix"/>=<xsl:value-of select="text()"/> <xsl:if test="@type = 'STRUCT'"> (used for structs)</xsl:if></P>
  </xsl:template>

  <!-- HTML generation -->
  <xsl:template match="/">
    <!-- Generate one file for all interfaces -->
    /* @FILE:interfaces */
    <HTML>
    <xsl:call-template name="MakeHeader">
      <xsl:with-param name="title" select="Interfaces"/>
    </xsl:call-template>
    <BODY>
    <DIV CLASS="BodyDiv">
    <H1>Interfaces</H1>
    <xsl:for-each select="//mic:interface-group" xml:space="preserve">
      <P><A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/> Interfaces:</A>
        <xsl:for-each select="//mic:interface[@mygroup = current()/@name]">
          <A><xsl:attribute name="HREF"><xsl:value-of select="@name"/>.htm</xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each>
      </P>
    </xsl:for-each>
    <xsl:for-each select="//mic:interface-group">
      <H2><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A> Interfaces</H2>
      <xsl:apply-templates select="mic:P|mic:description"/>
      <xsl:apply-templates select="mic:discussion"/>
      <DL>
        <xsl:for-each select="//mic:interface[@mygroup = current()/@name]">
          <DT>
            <A><xsl:attribute name="HREF"><xsl:value-of select="@name"/>.htm</xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
      </DL>
    </xsl:for-each>
    </DIV>
    <xsl:call-template name="MakeFooter"/>
    </BODY>
    </HTML>

    <xsl:for-each select="//mic:interface">
      <xsl:variable name="current-interface" select="."/>
      /* @FILE:<xsl:value-of select="@name"/> */
      <HTML>
      <xsl:call-template name="MakeHeader">
        <xsl:with-param name="title" select="@name"/>
      </xsl:call-template>
      <BODY>
      <DIV CLASS="BodyDiv">

      <H1><xsl:value-of select="@name"/> Interface</H1>
      <xsl:apply-templates select="mic:P"/>
      <xsl:apply-templates select="mic:description"/>
      <xsl:apply-templates mode="printxns" select="mic:namespace" />
      <P>C type for interface pointer: <CODE><xsl:value-of select="@ctype"/></CODE></P>
      <xsl:if test="@extends">
        <P> Extends <xsl:apply-templates select="@extends"/> </P>
      </xsl:if>
      <xsl:if test="mic:extended-by">
        <P> Extended by <xsl:apply-templates select="mic:extended-by" /> </P>
      </xsl:if>
      <xsl:if test="mic:implemented-by">
        <P> Implemented by <xsl:apply-templates select="mic:implemented-by" /> </P>
      </xsl:if>
      <xsl:if test="mic:see-also">
        <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
      </xsl:if>
      <xsl:if test="mic:method-group">
        <P xml:space="preserve"><A HREF="#Method groups">Method groups</A>:
        <xsl:for-each select="mic:method-group">
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each></P>
      </xsl:if>
      <xsl:if test="mic:method and not(mic:method-group)">
        <P xml:space="preserve"><A HREF="#Methods">Methods</A>:
        <xsl:for-each select="mic:method">
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each></P>
      </xsl:if>
      <xsl:if test="mic:constant">
        <P xml:space="preserve"><A HREF="#Constants">Constants</A>:
        <xsl:for-each select="mic:constant">
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each></P>
      </xsl:if>
      <xsl:if test="mic:struct">
        <P xml:space="preserve"><A HREF="#Structs">Structs</A>:
        <xsl:for-each select="mic:struct">
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each></P>
      </xsl:if>
      <xsl:if test="mic:typedef">
        <P xml:space="preserve"><A HREF="#Types">Types</A>:
        <xsl:for-each select="mic:typedef">
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each></P>
      </xsl:if>
      <xsl:apply-templates select="mic:discussion"/>
      <xsl:apply-templates select="mic:example"/>
      <xsl:apply-templates select="mic:notes" />


      <xsl:if test="mic:method-group">
        <H3><A><xsl:attribute name="NAME">Method groups</xsl:attribute>Method groups</A></H3>
        <DL xml:space="preserve">
        <xsl:for-each select="mic:method-group">
          <xsl:variable name="current-group" select="."/>
          <DT>
          <B><A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/>
            </xsl:attribute><xsl:value-of select="@name"/> Methods:</A></B>
            <!-- all the methods after this group -->
            <xsl:for-each select="../mic:method[number(@absolute-index) &gt; number(current()/@absolute-index)]">
              <!-- but not after the next -->
              <xsl:if test="../mic:method-group[number(@absolute-index) &lt; number(current()/@absolute-index)][position()=last()]/@name = $current-group/@name" >
                <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
              </xsl:if>
            </xsl:for-each>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
        </DL>
      </xsl:if>

      <xsl:if test="mic:method and not(mic:method-group)">
        <H3><A><xsl:attribute name="NAME">Methods</xsl:attribute>Methods</A></H3>
        <DL>
        <xsl:for-each select="mic:method">
          <DT>
          <A><xsl:attribute name="HREF">
              <xsl:value-of select="$current-interface/@name"/>.htm#<xsl:value-of select="@name"/>
            </xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
        </DL>
      </xsl:if>
      <xsl:if test="mic:constant">
        <H3><A><xsl:attribute name="NAME">Constants</xsl:attribute>Constants</A></H3>
        <DL>
        <xsl:for-each select="mic:constant">
          <DT>
          <A><xsl:attribute name="HREF">
              <xsl:value-of select="$current-interface/@name"/>.htm#<xsl:value-of select="@name"/>
            </xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
        </DL>
      </xsl:if>
      <xsl:if test="mic:struct/mic:field">
        <H3><A><xsl:attribute name="NAME">Structs</xsl:attribute>Structs</A></H3>
        <DL>
        <xsl:for-each select="mic:struct">
          <DT>
          <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
        </DL>
      </xsl:if>
      <xsl:if test="mic:typedef">
        <H3><A><xsl:attribute name="NAME">Types</xsl:attribute>Types</A></H3>
        <xsl:apply-templates select="mic:typedef" />
      </xsl:if>
      <xsl:apply-templates select="mic:method|mic:method-group">
        <xsl:with-param name="current-interface" select="$current-interface"/>
      </xsl:apply-templates>
      <xsl:apply-templates select="mic:constant" />
      <xsl:if test="mic:struct/mic:field">
        <xsl:apply-templates select="mic:struct" />
      </xsl:if>

      <P/>
      </DIV>
      <xsl:call-template name="MakeFooter"/>
      </BODY>
      </HTML>
    </xsl:for-each>

    <xsl:for-each select="//mic:concrete-class">
      <xsl:variable name="current-class" select="."/>
      /* @FILE:<xsl:value-of select="@name"/> */
      <HTML>
      <xsl:call-template name="MakeHeader">
        <xsl:with-param name="title" select="@name"/>
      </xsl:call-template>
      <BODY>
      <DIV CLASS="BodyDiv">

      <H1><xsl:value-of select="@name"/> Concrete Class</H1>
      <xsl:apply-templates select="mic:P|mic:description"/>

      <xsl:if test="mic:implements">
        <P> Implements <xsl:apply-templates select="mic:implements"/> </P>
      </xsl:if>
      <xsl:if test="@super-class">
        <P> Subclass of <xsl:apply-templates select="@super-class" /> </P>
      </xsl:if>
      <xsl:if test="mic:superclass-of">
        <P> Superclass of <xsl:apply-templates select="mic:superclass-of" /> </P>
      </xsl:if>
      <xsl:if test="mic:constructor">
        <H3>Constructors</H3>
        <xsl:for-each select="mic:constructor">
<PRE xml:space="preserve"><CODE><xsl:value-of select="@type"/> <xsl:value-of select="@name"/>( <xsl:choose><xsl:when test="not(mic:arg)">void )</xsl:when><xsl:otherwise><xsl:for-each select="mic:arg">
    /*<xsl:value-of select="@direction"/><xsl:if test="@optional = 'true'">,optional</xsl:if>*/ <xsl:value-of select="@type"/> <xsl:value-of select="@name"/><xsl:if test="position()!=last()">,</xsl:if></xsl:for-each>
)</xsl:otherwise></xsl:choose></CODE></PRE>
          <xsl:apply-templates select="mic:P" />
          <DL>
          <xsl:for-each select="mic:arg">
            <DT><B><xsl:value-of select="@name"/></B></DT>
            <DD><xsl:apply-templates select="mic:P"/></DD>
            <DD><xsl:value-of select="mic:description"/></DD>
            <xsl:for-each select="mic:notes"><DD><B>NOTE:</B> <xsl:value-of select="." /></DD></xsl:for-each>
          </xsl:for-each>
          </DL>
          <xsl:apply-templates select="mic:discussion"/>
          <xsl:apply-templates select="mic:notes"/>
          <xsl:apply-templates select="mic:example" />
          <xsl:if test="mic:see-also">
            <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
          </xsl:if>
        </xsl:for-each>
      </xsl:if>
      <xsl:if test="mic:implements-method">
        <H3>Methods</H3>
        <DL>
        <xsl:for-each select="mic:implements-method">
          <xsl:variable name="current-method" select="."/>
          <xsl:choose>
            <xsl:when test="@implementation-class = $current-class/@name">
              <DT>
              <A><xsl:attribute name="NAME"><xsl:value-of select="@implementation-name"/></xsl:attribute><xsl:value-of select="@implementation-name"/>
              </A>
              implements
              <A><xsl:attribute name="HREF"><xsl:choose><xsl:when test="@extended-interface"><xsl:value-of select="@extended-interface"/></xsl:when><xsl:otherwise><xsl:value-of select="@interface-name"/></xsl:otherwise></xsl:choose>.htm#<xsl:value-of select="@name"/>
                </xsl:attribute><xsl:choose><xsl:when test="@extended-interface"><xsl:value-of select="@extended-interface"/></xsl:when><xsl:otherwise><xsl:value-of select="@interface-name"/></xsl:otherwise></xsl:choose>.<xsl:value-of select="@name"/></A>
              </DT>
              <!-- Generate list of source references as generated by canal.js
               -->
              <xsl:if test="/mic:mmlite-spec/mic:tag[@name = current()/@implementation-name]">
                <DD xml:space="preserve">
                Source in
                <!-- XXX This loop is EXTREMELY SLOW (~ 10 min on my box) -->
                <xsl:for-each select="/mic:mmlite-spec/mic:tag[@name = current()/@implementation-name or @name = current()/@dot-name or @name = current()/@c-name or @name = current()/@cplus-name]">
                  <A><xsl:attribute name="HREF"><xsl:value-of select="@path"/>/<xsl:value-of select="@file"/>.htm#<xsl:value-of select="@ref"/></xsl:attribute><xsl:value-of select="@file"/></A>
                </xsl:for-each>
                </DD>
              </xsl:if>
              <DD> <xsl:apply-templates select="mic:P"/> </DD>
            </xsl:when>
            <xsl:otherwise>
              <DT>
              Inherits
              <A><xsl:attribute name="HREF">
                <xsl:value-of select="@implementation-class"/>.htm#<xsl:value-of select="@name"/>
                </xsl:attribute><xsl:value-of select="@dot-name"/>
              </A>              
              </DT>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
        </DL>
      </xsl:if>
      <xsl:apply-templates select="mic:discussion"/>
      <xsl:apply-templates select="mic:example"/>
        <xsl:if test="mic:see-also">
          <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
        </xsl:if>
      <xsl:apply-templates select="mic:notes" />

      <P/>
      </DIV>
      <xsl:call-template name="MakeFooter"/>
      </BODY>
      </HTML>
    </xsl:for-each>

    <xsl:for-each select="//mic:component">
      <xsl:variable name="current-component" select="."/>
      /* @FILE:_<xsl:value-of select="@name"/> */
      <HTML>
      <xsl:call-template name="MakeHeader">
        <xsl:with-param name="title" select="@name"/>
      </xsl:call-template>
      <BODY>
      <DIV CLASS="BodyDiv">

      <xsl:if test="@prefix">
        <P>Prefix: <xsl:apply-templates select="@prefix" /> </P>
      </xsl:if>

      <H1><xsl:value-of select="@name"/> Component</H1>
      <xsl:apply-templates select="mic:P|mic:description"/>

      <xsl:if test="mic:concrete-class">
        <H3>Concrete classes</H3>
        <DL>
        <xsl:for-each select="mic:concrete-class">
          <xsl:variable name="current-class" select="."/>
              <DT>
              <A><xsl:attribute name="HREF">
                <xsl:value-of select="@name"/>.htm
                </xsl:attribute><xsl:value-of select="@name"/>
              </A>
              <xsl:apply-templates select="mic:P"/>
              </DT>
        </xsl:for-each>
        </DL>
      </xsl:if>
      <xsl:apply-templates select="mic:discussion"/>
      <xsl:apply-templates select="mic:example"/>
      <xsl:if test="mic:see-also">
        <P> See also <xsl:apply-templates select="mic:see-also" /> </P>
      </xsl:if>
      <xsl:apply-templates select="mic:notes" />

      <P/>
      </DIV>
      <xsl:call-template name="MakeFooter"/>
      </BODY>
      </HTML>
    </xsl:for-each>

    <!-- Component groups -->
    /* @FILE:components */
    <HTML>
    <xsl:call-template name="MakeHeader">
      <xsl:with-param name="title" select="Components"/>
    </xsl:call-template>
    <BODY>
    <DIV CLASS="BodyDiv">
    <H1>Component Groups</H1>

    <xsl:for-each select="//mic:component-group" xml:space="preserve">
      <P><A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/> Components:</A>
        <xsl:for-each select="//mic:component[@mygroup = current()/@name]">
          <A><xsl:attribute name="HREF">_<xsl:value-of select="@name"/>.htm</xsl:attribute><xsl:value-of select="@name"/></A>
        </xsl:for-each>
      </P>
    </xsl:for-each>

    <xsl:for-each select="//mic:component-group">
      <H2><A><xsl:attribute name="NAME"><xsl:value-of select="@name"/></xsl:attribute><xsl:value-of select="@name"/></A> Components</H2>
      <xsl:apply-templates select="mic:P|mic:description"/>
      <xsl:apply-templates select="mic:discussion"/>
      <DL>
        <xsl:for-each select="//mic:component[@mygroup = current()/@name]">
          <DT>
            <A><xsl:attribute name="HREF">_<xsl:value-of select="@name"/>.htm</xsl:attribute><xsl:value-of select="@name"/></A>
          </DT>
          <DD> <xsl:apply-templates select="mic:P"/> </DD>
          <DD>Concrete classes:
             <xsl:for-each select="mic:concrete-class" xml:space="preserve">
          <A><xsl:attribute name="HREF"><xsl:value-of select="@name"/>.htm</xsl:attribute><xsl:value-of select="@name"/></A>
             </xsl:for-each>
          </DD>
          <DD> <xsl:apply-templates select="mic:description"/> </DD>
        </xsl:for-each>
      </DL>
    </xsl:for-each>

<!--
    <H1>Used Components</H1>
    <xsl:for-each select="/mic:mmlite-spec/mic:profile[@name = 'all']/mic:contains-component" xml:space="preserve">
      <A><xsl:attribute name="HREF">#<xsl:value-of select="@name"/>
             </xsl:attribute><xsl:value-of select="@name"/></A>
    </xsl:for-each>
    <xsl:for-each select="/mic:mmlite-spec/mic:profile[@name = 'all']/mic:contains-component">
      <DL>
      <H2><A><xsl:attribute name="NAME"><xsl:value-of select="."/></xsl:attribute><xsl:value-of select="."/></A> Component</H2>
      <xsl:for-each select="//mic:component[@name = current()]">
        <xsl:apply-templates select="." mode="ComponentList"/>
      </xsl:for-each>
      </DL>
    </xsl:for-each>
-->
    </DIV>
    <xsl:call-template name="MakeFooter"/>
    </BODY>
    </HTML>
  </xsl:template>

  <xsl:template match="//mic:component" mode="ComponentList">
    <xsl:apply-templates select="mic:P"/>
    <xsl:if test="mic:contains-concrete-class|mic:concrete-class">
      Concrete classes: 
        <xsl:for-each select="mic:contains-concrete-class|mic:concrete-class" xml:space="preserve">
        <A><xsl:attribute name="HREF"><xsl:value-of select="@name"/>.htm
               </xsl:attribute><xsl:value-of select="@name"/></A>
      </xsl:for-each>
    </xsl:if>
    <xsl:for-each select="mic:contains-component">
      <DL>
      <H3><A><xsl:attribute name="NAME"><xsl:value-of select="."/>
             </xsl:attribute><xsl:value-of select="."/></A> Component</H3>
        <xsl:for-each select="//mic:component[@name = current()]">
          <xsl:apply-templates select="." mode="ComponentList"/>
        </xsl:for-each>
      </DL>
    </xsl:for-each>
  </xsl:template>

  <xsl:template match="//mic:component" mode="CompList">
    <DL>
    <H3><xsl:value-of select="@name"/> Component.</H3>
    <xsl:apply-templates select="mic:P|mic:discussion" />
    <xsl:if test="mic:concrete-class|mic:contains-concrete-class">
      Concrete classes:
      <xsl:for-each select="mic:concrete-class|id(mic:contains-concrete-class/@name)">
        <xsl:apply-templates select="." mode="CompList"/>
      </xsl:for-each>
    </xsl:if>
    </DL>
  </xsl:template>

  <xsl:template match="//mic:concrete-class" mode="CompList" xml:space="preserve">
    <A><xsl:attribute name="HREF"><xsl:value-of select="@name"/>.htm
       </xsl:attribute><xsl:value-of select="@name"/></A>
  </xsl:template>

  <msxsl:script implements-prefix="local"><![CDATA[

    function eval_FormatConstantValue(_contextNodeList)
    {
      var __this = _contextNodeList.item(0);
      var result = "0x" + (parseInt(__this.getAttribute("base")) + parseInt(__this.getAttribute("ordinal"))).toString(16)

      return result;
    }

  ]]></msxsl:script>
</xsl:stylesheet>
