<!-- Copyright (c) Microsoft Corporation. All rights reserved. -->
<!-- Munge spec into soap typeinfo, part 1 -->

<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:mic="x-schema:expanded-schema.xml">
  <xsl:output method="xml" indent="yes" omit-xml-declaration="yes"/>

  <!-- Identity transformation template -->
  <xsl:template match="@*|/|node()"><xsl:copy><xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates></xsl:copy></xsl:template>

  <!-- Filter out stuff that is not needed -->
  <xsl:template match="mic:interface|mic:arg|mic:this-arg|mic:method|mic:method-response|mic:extended-method|mic:extended-method-response|mic:struct|mic:struct/mic:field" /><!-- except below -->

  <!-- Add type-tag to field -->
  <xsl:template match="mic:field" mode="doit">
    <xsl:copy>
      <!-- Add @iface-index - of total count of stuff in this interface -->
      <xsl:attribute name="iface-index"><xsl:value-of select="position() +count(../../mic:method|../../mic:method-response|../../mic:struct|../../mic:method/mic:arg|../preceding-sibling::mic:struct/mic:field)+count(../../mic:method[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')]) "/></xsl:attribute>
      <!-- figure out type -->
      <xsl:variable name="tag-type1"><xsl:apply-templates select="@type" mode="argtype"/></xsl:variable>
      <!-- If this is an array and not identified as such XXX add _ARRAY -->
      <xsl:variable name="tag-type"><xsl:value-of select="$tag-type1"/><xsl:if test="not(contains($tag-type1, '_ARRAY')) and @array">_ARRAY</xsl:if></xsl:variable>
      <!-- Remove _ARRAY. The typeinfo no longer wants it -->
      <xsl:attribute name="type-tag">ARG_<xsl:value-of select="substring-before($tag-type, '_ARRAY')"/><xsl:if test="not(contains($tag-type, '_ARRAY'))"><xsl:value-of select="$tag-type"/></xsl:if></xsl:attribute>
      <xsl:apply-templates select="@type" mode="fieldbyref"><xsl:with-param name="tag-type"><xsl:value-of select="$tag-type"/></xsl:with-param></xsl:apply-templates>
      <xsl:if test="@array"><xsl:attribute name="is-array">true</xsl:attribute></xsl:if>
      <!-- Add size-of,length-of,type-of,selector-of to some arguments -->
      <xsl:if test="../mic:field[@size-is = current()/@name]">
        <xsl:attribute name="size-of"><xsl:value-of select="../mic:field[@size-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:field[@length-is = current()/@name]">
        <xsl:attribute name="length-of"><xsl:value-of select="../mic:field[@length-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:field[@type-is = current()/@name]">
        <xsl:attribute name="type-of"><xsl:value-of select="../mic:field[@type-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:field[@selector-is = current()/@name]">
        <xsl:attribute name="selector-of"><xsl:value-of select="../mic:field[@selector-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <!-- size-is-pos, length-is-pos: relative positions of argument -->
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Add interfacecount to interface-group, only those with <namespace> XXXX  NEVER USED since there is no namespace here XXX -->
  <xsl:template match="mic:interface-group[mic:namespace]">
    <!-- First stuff in some struct stuff XXX lets not
    <xsl:element name="struct-group" namespace="x-schema:expanded-schema.xml">
      <xsl:attribute name="structcount"><xsl:value-of select="count(//mic:interface[@mygroup = current()/@name]/mic:struct)"/></xsl:attribute>      
      <xsl:apply-templates select="//mic:interface[@mygroup = current()/@name]/mic:struct"></xsl:apply-templates>
    </xsl:element>
end XXX -->
    <!-- The group -->
    <xsl:copy>
      <xsl:attribute name="interfacecount"><xsl:value-of select="count(//mic:interface[@mygroup = current()/@name])"/></xsl:attribute>  
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- Add method-count to interface -->
  <xsl:template match="mic:interface[mic:namespace]">
    <xsl:copy>
      <xsl:attribute name="methodcount"><xsl:value-of select="count(mic:method|mic:method-response|mic:struct)"/></xsl:attribute>
      <xsl:attribute name="iface-index-count"><xsl:value-of select="count(mic:method|mic:method-response|mic:struct|mic:method/mic:arg|mic:struct/mic:field)+1+count(mic:method[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')])"/></xsl:attribute>
      <!-- count of stuff that gets inherited from other interfaces -->
      <xsl:attribute name="inherited-count"><xsl:value-of select="count(mic:extended-method[not(@extended-interface = 'IUnknown')]|mic:extended-method-response[not(@extended-interface = 'IUnknown')]|mic:extended-struct[not(@extended-interface = 'IUnknown')]|mic:extended-method[not(@extended-interface = 'IUnknown')]/mic:arg|mic:extended-struct[not(@extended-interface = 'IUnknown')]/mic:field) + count(mic:extended-method[not(@extended-interface = 'IUnknown')][not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')]) + count(mic:string-entry[not(@inheritance-depth = 0)])"/></xsl:attribute>
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
      <!-- Deal with methods separately so as to get ordinals right.
       - - Also ends up putting methods before structs, which is what
       - - the struct indexing expects.
       -->
      <xsl:apply-templates select="mic:method | mic:method-response | mic:extended-method | mic:extended-method-response" mode="doit"></xsl:apply-templates>
      <!-- Print struct -->
      <xsl:for-each select="mic:struct">
        <xsl:copy>
          <xsl:attribute name="index"><xsl:value-of select="position()+1+count(../mic:method|../mic:method-response)"/></xsl:attribute>
        <!-- Add @iface-index - of total count of stuff in interface -->
        <xsl:attribute name="iface-index"><xsl:value-of select="position()+count(../mic:method|../mic:method-response)"/></xsl:attribute>
          <xsl:attribute name="fieldcount"><xsl:value-of select="count(mic:field)"/></xsl:attribute>
          <xsl:attribute name="myinterface"><xsl:value-of select="../@name"/></xsl:attribute>
          <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
          <xsl:apply-templates select="mic:field" mode="doit"></xsl:apply-templates>
        </xsl:copy>
      </xsl:for-each>
    </xsl:copy>
  </xsl:template>

  <!-- Add to method -->
  <xsl:template match="mic:method | mic:method-response | mic:extended-method | mic:extended-method-response" mode="doit">
    <xsl:copy>
      <!-- Add default responsename XXX remove later to make typeinfo smaller -->
      <xsl:if test="not(@responsename) and not(@type = 'none')">
        <xsl:attribute name="responsename"><xsl:value-of select="@name"/>Response</xsl:attribute>
      </xsl:if>
      <!-- Add @argcount -->
      <xsl:attribute name="argcount"><xsl:value-of select="count(mic:arg|mic:this-arg)"/></xsl:attribute>
      <!-- Add @ordinal -->
      <xsl:attribute name="ordinal"><xsl:value-of select="position()-1-count(preceding-sibling::mic:method-response)-count(preceding-sibling::mic:extended-method-response)-count(self::mic:method-response)-count(self::mic:extended-method-response)"/></xsl:attribute>
      <xsl:if test="self::mic:method | self::mic:method-response">
        <!-- Add @iface-index - of total count of stuff in interface -->
        <xsl:attribute name="iface-index"><xsl:value-of select="position()-count(preceding-sibling::mic:extended-method)-count(preceding-sibling::mic:extended-method-response)"/></xsl:attribute>
      </xsl:if>
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"/>
      <!-- Add out argument for return value -->
      <xsl:if test="not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')">
        <xsl:variable name="tag-type"><xsl:apply-templates select="@type" mode="argtype"/></xsl:variable>
        <xsl:element name="arg" namespace="x-schema:expanded-schema.xml">
          <xsl:if test="self::mic:method">
            <!-- Add @iface-index - of total count of stuff in interface -->
            <xsl:attribute name="iface-index"><xsl:value-of select="1 + count(../mic:method|../mic:method-response|../mic:struct|preceding-sibling::mic:method/mic:arg)+count(preceding-sibling::mic:method[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')])"/></xsl:attribute>
          </xsl:if>
          <xsl:if test="self::mic:method-response">
            <!-- Add @iface-index - of total count of stuff in interface -->
            <xsl:attribute name="iface-index"><xsl:value-of select="1 + count(../mic:method|../mic:method-response|../mic:struct|preceding-sibling::mic:method-response/mic:arg)+count(preceding-sibling::mic:method-response[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')])+count(../preceding-sibling::mic:method[@type = 'none']/mic:arg)"/></xsl:attribute>
          </xsl:if>
          <xsl:attribute name="type"><xsl:value-of select="@type"/></xsl:attribute>
          <!-- Remove _ARRAY. The typeinfo no longer wants it -->
          <xsl:attribute name="type-tag">ARG_<xsl:value-of select="substring-before($tag-type, '_ARRAY')"/><xsl:if test="not(contains($tag-type, '_ARRAY'))"><xsl:value-of select="$tag-type"/></xsl:if></xsl:attribute>
          <xsl:attribute name="name"><xsl:value-of select="@resultname"/><xsl:if test="not(@resultname)">Result</xsl:if></xsl:attribute><xsl:attribute name="direction">out</xsl:attribute>
          <xsl:attribute name="xname"><xsl:value-of select="@resultname"/><xsl:if test="not(@resultname)">Result</xsl:if></xsl:attribute><xsl:attribute name="direction">out</xsl:attribute>
          <xsl:attribute name="is-return-value">true</xsl:attribute>
          <!-- copy @array attribute -->
          <xsl:apply-templates select="@array"/>
          <xsl:call-template name="argbyref"><xsl:with-param name="tag-type"><xsl:value-of select="$tag-type"/></xsl:with-param></xsl:call-template>
        </xsl:element>
      </xsl:if>
      <xsl:apply-templates select="mic:arg|mic:this-arg" mode="doit"/>
    </xsl:copy>
  </xsl:template>

  <!-- Add to arg -->
  <xsl:template match="mic:arg|mic:this-arg" mode="doit">
    <xsl:copy>
      <xsl:variable name="tag-type"><xsl:apply-templates select="@type" mode="argtype"/></xsl:variable>
      <!-- Add type-tag -->
      <!-- Remove _ARRAY. The typeinfo no longer wants it -->
      <xsl:attribute name="type-tag">ARG_<xsl:value-of select="substring-before($tag-type, '_ARRAY')"/><xsl:if test="not(contains($tag-type, '_ARRAY'))"><xsl:value-of select="$tag-type"/></xsl:if></xsl:attribute>
      <xsl:attribute name="index"><xsl:value-of select="position()"/></xsl:attribute>
      <xsl:if test="self::mic:arg and parent::mic:method">
        <!-- Add @iface-index - of total count of stuff in interface -->
        <xsl:attribute name="iface-index"><xsl:value-of select="position() - count(preceding-sibling::mic:this-arg) + count(../../mic:method|../../mic:method-response|../../mic:struct|../preceding-sibling::mic:method/mic:arg) + count(../preceding-sibling::mic:method[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')])+(not(../@type = 'void') and not(../@type = 'SCODE') and not(../@type = 'none'))"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="self::mic:arg and parent::mic:method-response">
        <!-- Add @iface-index - of total count of stuff in interface -->
        <xsl:attribute name="iface-index"><xsl:value-of select="position() - count(preceding-sibling::mic:this-arg) + count(../../mic:method|../../mic:method-response|../../mic:struct|../preceding-sibling::mic:method-response/mic:arg) + count(../preceding-sibling::mic:method-response[not(@type = 'void') and not(@type = 'SCODE') and not(@type = 'none')])+(not(../@type = 'void') and not(../@type = 'SCODE') and not(../@type = 'none'))+count(../preceding-sibling::mic:method[@type = 'none']/mic:arg)"/></xsl:attribute>
      </xsl:if>
      <xsl:call-template name="argbyref"><xsl:with-param name="tag-type"><xsl:value-of select="$tag-type"/></xsl:with-param></xsl:call-template>
      <!-- Add size-of,length-of,type-of,selector-of to some arguments -->
      <xsl:if test="../mic:arg[@size-is = current()/@name]">
        <xsl:attribute name="size-of"><xsl:value-of select="../mic:arg[@size-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:arg[@length-is = current()/@name]">
        <xsl:attribute name="length-of"><xsl:value-of select="../mic:arg[@length-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:arg[@type-is = current()/@name]">
        <xsl:attribute name="type-of"><xsl:value-of select="../mic:arg[@type-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <xsl:if test="../mic:arg[@selector-is = current()/@name]">
        <xsl:attribute name="selector-of"><xsl:value-of select="../mic:arg[@selector-is = current()/@name]/@name"/></xsl:attribute>
      </xsl:if>
      <!-- size-is-pos, length-is-pos: relative positions of argument -->
      <xsl:apply-templates select="@*|*|comment()|processing-instruction()|text()"></xsl:apply-templates>
    </xsl:copy>
  </xsl:template>

  <!-- figure out correct type tag (best guess really...) -->
  <xsl:template match="@type" mode="argtype"><xsl:choose><xsl:when test="../@list">LISTHEAD</xsl:when><xsl:when test=". = 'string'">STRING</xsl:when><xsl:when test=". = 'UINT32' or . = 'UINT16' or . = 'UINT8' or . = 'UINT' or . = 'SCODE' or . = 'INT32' or . = 'INT16' or . = 'INT8' or . = 'INT'">INT</xsl:when><xsl:when test=". = 'PSEUDOFLOAT'">FLOAT</xsl:when><xsl:when test=". = 'TIME' or . = 'TIME*'">DATETIME</xsl:when><xsl:when test=". = 'BOOL' or . = 'BOOL*'">BOOL</xsl:when><xsl:when test=". = 'INT64' or . = 'PINT64' or . = 'INT64*' or . = 'UINT64'">INT64</xsl:when><xsl:when test=". = 'IID'">GUID</xsl:when><xsl:when test=". = 'URI'">URI</xsl:when><xsl:when test=". = 'PICONTINUATION'">CONTINUATION</xsl:when><xsl:when test=". = 'void*'">BLOB</xsl:when><xsl:when test=". = 'PTR'">BLOB</xsl:when><xsl:when test="(. = 'UINT*' or . = 'PUINT') and (../@direction != 'in') and not(../@size-is) and not(../@length-is) and not(../@array)">INT</xsl:when><xsl:when test=". = 'UINT*' or . = 'PUINT'">INT_ARRAY</xsl:when><xsl:when test=". = 'PSEUDOFLOAT*' and (../@direction != 'in') and not(../@size-is) and not(../@length-is)">FLOAT</xsl:when><xsl:when test=". = 'PSEUDOFLOAT*'">FLOAT_ARRAY</xsl:when><xsl:when test=". = 'char**' or . = 'string[]'">STRING_ARRAY</xsl:when><xsl:when test=". = 'LISTNODE' or . = 'QNAME' or . = 'ANYTYPE' or . = 'BITMAP'"><xsl:value-of select="."/></xsl:when><xsl:when test=". = 'PIBUFFER'">IBUFFER</xsl:when><xsl:when test="../@size-is or ../@length-is">STRUCT_ARRAY</xsl:when><xsl:otherwise>STRUCT</xsl:otherwise></xsl:choose></xsl:template>

  <xsl:template match="mic:this-arg/@type" mode="argtype">THIS</xsl:template>

  <xsl:template name="argbyref"><xsl:param name="tag-type"/><xsl:if test="not($tag-type = 'INT' or $tag-type = 'BOOL' or $tag-type = 'FLOAT' or $tag-type = 'INT64' or $tag-type = 'TIME'  or $tag-type = 'DATETIME' or $tag-type = 'THIS' or $tag-type = 'CONTINUATION') or contains($tag-type, '_ARRAY') or @direction = 'out' or @direction = 'inout'"><xsl:attribute name="byref">true</xsl:attribute></xsl:if><xsl:if test="contains($tag-type, '_ARRAY')"><xsl:attribute name="is-array">true</xsl:attribute></xsl:if></xsl:template>

  <!-- Fields are by ref when the type is a pointer type (or ANYTYPE/LISTHEAD)
       or a string
    -->
  <xsl:template mode="fieldbyref" match="@type"><xsl:param name="tag-type"/><xsl:variable name="type-ns"><xsl:choose><xsl:when test="contains(., ':')"><xsl:value-of select="substring-before(., ':')"/></xsl:when><xsl:otherwise><xsl:value-of select="../../../@name"/></xsl:otherwise></xsl:choose></xsl:variable><xsl:variable name="type-local"><xsl:choose><xsl:when test="contains(., ':')"><xsl:value-of select="substring-after(., ':')"/></xsl:when><xsl:otherwise><xsl:value-of select="."/></xsl:otherwise></xsl:choose></xsl:variable><xsl:if test="$tag-type = 'ANYTYPE' or $tag-type = 'QNAME' or $tag-type = 'LISTHEAD' or $tag-type = 'STRING' or $tag-type = 'BITMAP' or //mic:interface[@name = $type-ns]/mic:struct[@ptype = $type-local]"><xsl:attribute name="byref">true</xsl:attribute></xsl:if><xsl:if test="contains($tag-type, '_ARRAY')"><xsl:attribute name="is-array">true</xsl:attribute></xsl:if><xsl:attribute name="TYPE"><xsl:value-of select="."/></xsl:attribute></xsl:template>

</xsl:stylesheet>
