/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Concrete type of continuation -jvh */

#ifndef __CONT_H_
#define __CONT_H_ 1

#include <mmlite.h>
#include <web/typeinfo.h>

typedef struct CONTINFO *PCONTINFO;
typedef const struct CONTINFO *CPCONTINFO;

struct CONTINFO {
    const struct IContinuationVtbl *v;
    UINT RefCnt;
    ListNodeDeclare(CONTINFO, ListNode);
#define contloff offsetof(struct CONTINFO, ListNode)
    UINT Flags;
#ifdef _CLR
    BOOL isCLR;
    _TCHAR *CLRObjectName;
#endif

    /* Response generation */
    PISOAPSERIALIZER Serializer; /* Where to write result message */
    PIHTTPCONNECTOR Request;    /* For SendReply() */
    BOOL MsgStarted;            /* did we already call ser->StartMessage() */

    /* Messaging delay tracking */
    TIME BeginTime; /* when this cont was created (serv) or called (client) */
    TIME sendTime;  /* when the message to this  continuation was sent */

    /* method info */
    _TCHAR *Uri;
    _TCHAR *WsaTo;              /* Uri w/ possible arguments */
    _TCHAR *SoapAction;         /* by default use method ns QName */
    CPXMLNAMESPACE methodNS;
    CPMETHOD_DESCR metd;
    INT32 RetVal;
    void *Method;

    PBYTE Frame;             /* Stack frame, in ArgSpace or on client stack */
    UINT FrameSize;          /* Size of Frame. Tracks reuse on server. */
    PIUNKNOWN pThis;         /* The object, to be released afterwards. */
    PIHTTPCONNECTOR pConnector;
    CONDITION cd;            /* Wait point for future */
    MUTEX mx;                /* To protect wait point for c_Broadcast */
    enum CONT_STATE {CONT_INIT, CONT_RPC, CONT_READY} State;

    /* Temp heap for allocating arguments */
    PIHEAP TmpHeap;
    PBYTE ArgSpace;         /* IHeap impl */
    PBYTE ArgSpaceLast;     /* IHeap impl */
    INT ArgSpaceAvail;      /* IHeap impl */

    /* Current thread info */
    PITHREAD Thread;
    PBYTE StackBase;
    UINT StackSize;
    PBYTE StackCurrent;

    /* Deserializer state saved for response processing */
    CPXMLNAMESPACE encodingNS;
    CPXMLNAMESPACE envelopeNS;
    PXMLNAMESPACE namespaces;

    PWSCONTEXT pContext; /* wsa context made available to services */
    /* WS-Addressing, WS-Routing, WS-Reservation, and related fields */
    BOOL MayOverrideAddr; /* FALSE: Proxy or already got addr */
    const _TCHAR *To;
    _TCHAR *resname;
};

#define iCONT(_p_) ((PICONTINUATION)(_p_))
#define pCONT(_i_) ((PCONTINFO)(_i_))

#endif /* __CONT_H_ */
