/* Copyright (c) Microsoft Corporation. All rights reserved. */


#ifndef __TOKENIZER_INTERFACE__ 
#define __TOKENIZER_INTERFACE__ 1


#include <mmlite.h>


/* Parsing utilities */


#define CR                      13
#define LF                      10
#define SP                      32
#define HT                      9
#define COLON                   58
                      
/* The parameters are of type PBYTE (c) and char (n) */

#define IsASCII(c, n)           (((UINT) (*(c))) == (n))
#define IsCR(c)                 (IsASCII(c, CR))        /* carriage return, '\r' */
#define IsLF(c)                 (IsASCII(c, LF))        /* linefeed, '\n' */
#define IsSP(c)                 (IsASCII(c, SP))        /* whitespace, ' ' */
#define IsHT(c)                 (IsASCII(c, HT))        /* horizontal tab, '\t' */
#define IsCOLON(c)              (IsASCII(c, COLON))     /* colon, ':' */

#define IsWhitespace(c)         ( IsCR((c)) || IsLF((c)) || IsSP((c)) || IsHT((c)))

/* The parameters are of type PBYTE* (c) and char (n) */
                            
#define EatASCII(c, n) \
{ \
    while (IsASCII(*(c), (n))) { \
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 1, NULL); \
        if ((sc) != S_OK) goto err_eat_macro; \
    } \
}

#define EatWHITESPACE(c) \
{ \
    while ( IsWhitespace(*(c)) ) { \
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 1, NULL); \
        if ((sc) != S_OK) goto err_eat_macro; \
    } \
}

#define EatSP(c)                { EatASCII((c), SP); }


#define EatCR(c)                { EatASCII((c), CR); }

#define EatLF(c)                { EatASCII((c), LF); }

#define EatHT(c)                { EatASCII((c), HT); }

#define EatOneCRLF_OLD(c) \
{ \
    UINT Match = 1; \
    sc = Tokenizer->v->CmpToken(Tokenizer, c, (const PBYTE) "\r\n", 2, &Match); \
    if (FAILED(sc)) goto err_eat_macro; \
    if (sc == S_OK) { \
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 2, NULL); \
        if (FAILED(sc)) goto err_eat_macro; \
    } \
}

#define EatOneCRLF(c) \
{ \
	UINT Match = 1; \
    sc = Tokenizer->v->CmpToken(Tokenizer, c, (const PBYTE) "\r", 1, &Match); \
    if (sc == S_OK) { \
	    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 1, NULL); \
        if (FAILED(sc)) goto err_eat_macro; \
	} \
    sc = Tokenizer->v->CmpToken(Tokenizer, c, (const PBYTE) "\n", 1, &Match); \
    if (FAILED(sc)) goto err_eat_macro; \
    if (sc == S_OK) { \
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 1, NULL); \
        if (FAILED(sc)) goto err_eat_macro; \
    } \
}

#define EatLWS(c) \
{ \
    EatOneCRLF(c); \
    while (IsSP(*(c)) || IsHT(*(c))) { \
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, c, 1, NULL); \
        if ((sc) != S_OK) goto err_eat_macro; \
    } \
}


#endif /* __TOKENIZER_INTERFACE__ */
