/* Copyright (c) Microsoft Corporation. All rights reserved. */
	#include <arm_asm.h>

	_CODE(C$$code)

/*
 * ARM assembly source file for machine-dependent atomic functions
 */

        _EXPORT(AtomicCmpAndSwap)
        _EXPORT(AtomicSwap)

/*
 * UINT AtomicCmpAndSwap(UINT *ptr, UINT oldVal, UINT newVal)
 *
 * Atomic compare and swap.  If oldVal matches *ptr, write newVal
 * to *ptr and return 1; otherwise, return 0.  Temporarily disable
 * interrupts to make read-modify-write operation indivisible.
 */
_FUNCTION(AtomicCmpAndSwap)
        mrs      r12, cpsr          /* get current PSR */
        orr      r3, r12, #0xc0     /* set IRQ/FIQ disable bits */
        msr      cpsr_c, r3        /* disable interrupts */
        ldr      r3, [r0]          /* read word pointed to by arg1 */
        cmp      r1, r3            /* does arg2 == *arg1 ? */
	moveq    r3, #1		   /* if arg2 == *arg1, return 1 */
	streq    r2, [r0]	   /* and set *arg1 = arg3 */
        movne    r3, #0            /* else if arg2 != *arg1, return 0 */
        msr      cpsr_c, r12        /* restore original PSR */
	mov      r0, r3		   /* set return reg = r0 */
        mov      pc, lr            /* return from subroutine */


/*
 * UINT AtomicSwap(UINT *ptr, UINT newVal)
 *
 * Atomic swap.  Write newVal to *ptr.  Return original value 
 * of *ptr.  Temporarily disable interrupts to make read-modify-
 * write operation indivisible.
 */

_FUNCTION(AtomicSwap)
	swp      r0, r1, [r0]      /* all in one shot */
        mov      pc, lr            /* return from subroutine */

        _END
