/* Copyright (c) Microsoft Corporation. All rights reserved. */
	#include <arm_asm.h>

	_CODE(C$$code)

/*
 * Common low level code for ARM MMLite.
 * Not in C for a variety of reasons,
 * esp that asm() statements dont agree across tools.
 */

    _EXPORT(IntOn)
    _EXPORT(IntOff)
    _EXPORT(SetPsr)
    _EXPORT(GetPC)
    _EXPORT(ProcessImage)
    _EXPORT(ProcessArgs)
    _EXPORT(DelayLoop)

    _IMPORT(ProcessImageFromPC)
    _IMPORT(ProcessArgsFromPC)

/*
 * UINT32 IntOff(void)
 * 
 * Disable interrupts, return previous PSR value
 */
_FUNCTION(IntOff)
    mrs     r0,cpsr         /* get current bits (return value) */
    orr     r1,r0,#0xC0     /* disable IRQ&FIQ */
    msr     cpsr_c,r1       /* set */
    mov     pc,lr           /* return */

/*
 * UINT32 IntOn(void)
 * 
 * Enable interrupts, return previous PSR value
 */
_FUNCTION(IntOn)
    mrs     r0,cpsr         /* get current bits (return value) */
    bic     r1,r0,#0xC0     /* enable IRQ&FIQ */
    msr     cpsr_c,r1
    mov     pc,lr

/*
 * void SetPsr(UINT32 PsrValue)
 *
 * Set the current PSR's control field to the given value
 */
_FUNCTION(SetPsr)
    msr     cpsr_c,r0     /* (re)install value */
    mov     pc,lr         /* return  */

/*
 * ADDRESS GetPC(void)
 *
 * Returns the PC at the call site.
 */
_FUNCTION(GetPC)
    mov     r0,lr       /* dont bother to sub 4 */
    mov     pc,lr       /* return  */

/*
 * _TCHAR * ProcessImage(void)
 *
 * Returns the process' ImageName pointer
 */
_FUNCTION(ProcessImage)
    mov     r0,lr       /* dont bother to sub 4 */
    b       ProcessImageFromPC   /* let it return to caller directly */

/*
 * _TCHAR * ProcessArgs(void)
 *
 * Returns the process' Arguments pointer
 */
_FUNCTION(ProcessArgs)
    mov     r0,lr       /* dont bother to sub 4 */
    b       _F(ProcessArgsFromPC)   /* let it return to caller directly */

/*
 * void DelayLoop(INT32 LoopCount)
 *
 * Spins as fast as possible for the given number of times.
 * If argument negative same as zero (or one, no matter).
 */
_FUNCTION(DelayLoop)
    subs    r0,r0,#1
    bge     DelayLoop
    mov     pc,lr         /* return  */

    _END
