/* Copyright (c) Microsoft Corporation. All rights reserved. */
	#include <arm_asm.h>

	_CODE(C$$code)

/*
 *
 * Trap dispatching code
 */

    _EXPORT(InstallTrapBranch)
    _EXPORT(TrapInit)
    _EXPORT(TrapHandler)
    _EXPORT(TrapPAEntry)
    _EXPORT(TrapDAEntry)
    _EXPORT(HWLoadThreadContext)
    _EXPORT(Reschedule)

    _IMPORT(TrapFIQ)
    _IMPORT(TrapIRQ)
    _IMPORT(TrapUI)
    _IMPORT(TrapPA)
    _IMPORT(TrapDA)
    _IMPORT(_Reschedule)
    _IMPORT(FlushCache)

    _EXPORT(IntStackPtr)

_LABEL(VecInstr)    _DCD     0xE59FF018      /* ldr pc,[pc,#0x18] */

/*
 * void InstallTrapBranch(UINT32 Offset, UINT32 FuncPtr)
 * {
 *    *((UINT32 *)Offset) = "ldr pc,[pc,#0x18]"
 *    *((UINT32 *)(Offset+0x20) = FuncPtr;
 *    return;
 * }
 */
_FUNCTION(InstallTrapBranch)
    ldr     r2,VecInstr
    str     r2,[r0]
    str     r1,[r0,#0x20]
    mov     pc,lr


#if 1 || defined(NEED_EXITUSERMODE)

    _EXPORT(ExitUserMode)

_FUNCTION(ExitUserMode)
    mov     r1,lr
    mov     r0,#0x08
    bl      InstallTrapBranch
    swi     0                       /* return to caller in SVC mode */
    mov     pc,r1                   /* just in case, simulator.. */

#endif


/* Code to setup abort handling vectors */

_FUNCTION(TrapInit)
    mov     r3,lr   /* save return address */

    /* setup stacks for each exception */
    mrs     r1,cpsr

    /* fiq save area */
    msr     cpsr_c,#0xD1
    ldr     r0,TrapDASavePtr
    sub     r13,r0,#(16*4)        /* how to get to TrapFIQSave */

    /* irq save area */
    msr     cpsr_c,#0xD2
    sub     r13,r0,#(16*3)        /* how to get to TrapIRQSave */

    /* undefined instr save area */
    msr     cpsr_c,#0xDB
    sub     r13,r0,#(16*2)        /* hot to get to TrapUISave */

    /* but not data/prefetch aborts- they are setup dynamically) */

    msr     cpsr_c,r1

#if 1 /* unnecessary */
    /* Next setup branch vectors for each exception */

    /* FIQ traps */
    mov     r0,#0x1C
    adr     r1,_F(TrapHandler)
    bl      InstallTrapBranch

    /* IRQ traps */
    mov     r0,#0x18
    adr     r1,_F(TrapHandler)
    bl      InstallTrapBranch
#if defined(NO)
    adr     r1,_F(TrapHandler)
    ldr     r0,=0x38
    str     r1,[r0]
#endif

    /* undefined instr branch vector */
    mov     r0,#0x04
    adr     r1,_F(TrapHandler)
    bl      InstallTrapBranch   /* Install trap vector */

    /* prefetch abort branch vector */
    mov     r0,#0x0C
    adr     r1,TrapPAEntry
    bl      InstallTrapBranch   /* Install trap vector */

    /* data abort branch vector */
    mov     r0,#0x10
    adr     r1,TrapDAEntry
    bl      InstallTrapBranch   /* Install trap vector */

    bl      _F(FlushCache)
#endif

    mov     pc,r3

_FUNCTION(TrapDAEntry)
    ldr     r13,TrapDASavePtr     /* how to get to TrapDASave */
    b       _F(TrapHandler)

_FUNCTION(TrapPAEntry)
    ldr     r13,TrapDASavePtr
    sub     r13,r13,#(16*1)       /* how to get to TrapPASave */
    b _F(TrapHandler)

/*
 * The following routine should eventually go into ROM
 * On entry, r13 should point to a 4-word 'save' area,
 * where the first word is a vector to branch to after
 * saving the svc context.
 */
_FUNCTION(TrapHandler)
    stmfd   r13, {r0,r14}       /* save r0, pc */
    mrs     r0, spsr            /* save PSR */
    str     r0, [r13, #-12]

    sub     r0, r13, #12        /* save pointer to reg save area */

    msr     cpsr_c, #0x93       /* to SVC mode */

    stmfd   r13!, {r13-r15}     /* save some regs, get SP right */
    stmfd   r13!, {r1-r12}      /* save some more regs */

    ldmfd   r0, {r0-r3}         /* r0=PSR, r1=r0, r2=pc, r3=handler */
    stmfd   r13!, {r0,r1}

    str     r2, [r13, #16*4]

    mov     r0, r13             /* arg0= pContext */
    adr     lr, _F(HWLoadThreadContext)
    mov     pc, r3              /* call handler */



/*
 * Voluntary context-switch code.
 * [Did not prove we *have* to switch, hence we might not]
 * Only have to save enough state for a function return.
 * Reload might be from interrupt above, hence PSR too.
 */

_FUNCTION(Reschedule)
    /*
     * Big register save, on current stack.
     * TODO: ignore caller-saved registers.
     * assume we can trash r0, and we were in svc mode
     *
     * The following code ensures the correct stack
     * value is stored by making the stack the 1st reg
     */
    stmfd   r13!,{r13-r15}   /* save most regs, leave space for pc */
    stmfd   r13!,{r0-r12}    /* save most regs */

    /* save psr */
    mrs     r0,cpsr
    stmfd   r13!,{r0}

    /* save pc */
    str     r14,[r13,#64]

    /* Saved state pointer (r0) is argument to _Reschedule, */
    mov     r0,r13
    msr     cpsr_c, #0xD3       /* SVC mode, interrupts off */
    bl      _F(_Reschedule)

/* Note that we're already in svc mode so there is no mode switch */

_FUNCTION(HWLoadThreadContext)
    ldmfd   r0!, {r1}
 #if defined(__GNUC__)
    msr     spsr_all, r1
 #else
    msr     spsr_csxf, r1
 #endif
    ldmfd   r0, {r0-r15}^

/*
 * Global variables
 */

/*  NOTE: Be careful about how much stack to reserve... */
IntStackSize _EQU 1280

_LABEL(IntStackPtr)
    _DCD     IntStack+IntStackSize
_LABEL(TrapDASavePtr)
    _DCD     TrapDASave
    
    _BSS(C$$zidata)

_LABEL(IntStack)
    _SKIP IntStackSize

/*
 * Must be at end to make some tools happy
 */

    _DATA(C$$rwdata)

_LABEL(TrapSaveAreas)
    _DCD 0   /* psr */
    _DCD 0   /* r0 */
    _DCD 0   /* pc */
_LABEL(TrapFIQSave)
    _DCD _F(TrapFIQ)

    _DCD 0   /* psr */
    _DCD 0   /* r0 */
    _DCD 0   /* pc */
_LABEL(TrapIRQSave)
    _DCD _F(TrapIRQ)

    _DCD 0   /* psr */
    _DCD 0   /* r0 */
    _DCD 0   /* pc */
_LABEL(TrapUISave)
    _DCD _F(TrapUI)

/* Special handling for aborts, since data & prefetch share a reg set */
    _DCD 0   /* psr */
    _DCD 0   /* r0 */
    _DCD 0   /* pc */
_LABEL(TrapPASave)
    _DCD _F(TrapPA)

    _DCD 0   /* psr */
    _DCD 0   /* r0 */
    _DCD 0   /* pc */
_LABEL(TrapDASave)
    _DCD _F(TrapDA)

    _END
