/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _CODE(C$$code)

    _IMPORT(BaseInit)
    _IMPORT(TrapInit)
    _IMPORT(ExitUserMode)
    _IMPORT(IntStackPtr)
    _IMPORT(DebugBreak)
    _IMPORT(InitMMU)

    _EXPORT(_start)
    _EXPORT(_FlushCache)

/* 
 * Entry point
 */
    _ENTRY

_FUNCTION(_start)

/*
 * Make sure we're in SVC mode
 * The ROM monitor (ANGEL) starts us in USER mode.
 * If you reflash the AEB1 and rid of ANGEL rid of this too.
 */
   bl       ExitUserMode

/*
 * Disable interrupts for now, SVC mode
 */
    msr     cpsr_c, #0xD3

/*
 * Set stack pointer for SVC mode
 */
    ldr     sp, = _L(IntStackPtr)
    ldr     sp, [sp]

/*
 * Initialize the Sharp 790 MMU.
 */
    bl      _F(InitMMU)

/*
 * Initialize interrupt dispatching
 */
    bl      _F(TrapInit)

#if defined(__GNUC__)
/*
 * We are ready for a breakpoint
 */
    bl      _F(DebugBreak)
#endif

#if defined(ANGELX)
/*
 * Enable interrupts.  They are masked in the PIC, so we
 * won't get any that we own.
 */
    msr     CPSR_c, #0x13
#endif

/*
 * Go straight for MI init
 */
    b       _F(BaseInit)

/*
 * Cache flushing code.  In assembly because it needs to
 * control tightly its own use of memory.
 */
_LABEL(CCR)
    _DCD   0xffffa400
_FUNCTION(_FlushCache)

    /* Build the two constants address limits
     */
    mov    r0,#0x60000000
    orr    r1,r0,#0xa00         // 0x60000a00 : last cache set
    orr    r0,r0,#0x800         // 0x60000800 : first cache set

    /* CCR address
     */
    ldr    r2, CCR

    /* Put the cache in flush (force writeback) mode
     */
    mov    r3,#4
    strb   r3, [r2]
    nop                      // just in case..

    /* Loop through all sets
     */
_LABEL(FlushOneSet)
    ldr    r3, [r0]
    add    r0, r0, #4
    cmp    r0, r1
    bne    FlushOneSet

    /* Writebacks should have happened.
     * Now invalidate before we re-enable it.
     */
    mov    r3,#8
    strb   r3, [r2]
    nop                      // just in case..
    nop                      // just in case..

    /* Put the cache back in enabled mode
     */
    mov    r3,#1
    strb   r3, [r2]
    nop                      // just in case..
    nop                      // just in case..

    /* And return
     */
    mov    pc,r14

    _END
