/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Support routines
 */

    _CODE(C$$code)

    _EXPORT(MMUFaultAddress)
    _EXPORT(MMUFaultStatus)
    _EXPORT(ArmIdle)
    _EXPORT(FlushCache)
    _EXPORT(EnableMMU)
    _EXPORT(DisableMMU)
#if 0	
   _EXPORT(SetDataBreakPoint)
   _EXPORT(SetInstructionBreakPoint)
   _EXPORT(MMUGetTTB)
   _EXPORT(MMUGetCr0)
   _EXPORT(MMUGetCr1)
#endif

_FUNCTION(FlushCache)
    mov     r0,#0xE0000000
    add     r1,r0,#8192
_LABEL(l1)
    ldr     r2,[r0],#32
    teq     r1,r0
    bne     l1

    mcr     p15,0,r0,c7,c7,0    /* flush I+D caches */
    mov     pc,lr

_FUNCTION(MMUFaultAddress)
    mrc     p15,0,r0,c6,c0,0
    mov     pc,lr

_FUNCTION(MMUFaultStatus)
    mrc     p15,0,r0,c5,c0,0
    mov     pc,lr

/*
 * void ArmIdle(int xx, ADDRESS UncachedAddress)
 *
 * Performs the enter-idle-mode code sequence
 */
    _ALIGN(32)
_FUNCTION(ArmIdle)
    /* This is exactly like Intel says
     */
    mcr     p15,0,r0,c15,c2,2      /* disable clock switching */
    ldr     r1,[r1]                /* read UDCR, its uncached */
    mcr     p15,0,r0,c15,c8,2      /* wait for interrupt      */

    /* Out of Idle sequence
     * Intel says "Note that the user should re-enable clock switching",
     * but it doesnt say how.  The description of coprocessor 15 says
     * that OPC2=2 CRm=1 "Enables clock switching".  
     * So one guesses that the instruction below does the job.
     * Alas, it drives power consumption up by 70mA !!
     *
     * We have found that doing nothing seems to work fine, the machine
     * still works and we get the powersave [reduces by about 120mA].
     */
#if 0
    mcr     p15,0,r0,c15,c1,2      /* DONT! enable clock switching */
#endif
    mov     pc,lr

#if 0
_FUNCTION(MMUGetTTB)
    mrc     p15,0,r0,c2,c0,0
    mov     pc,lr
        
_FUNCTION(MMUGetCr1)
    mrc     p15,0,r0,c1,c0,0
    mov     pc,lr
        
_FUNCTION(MMUGetCr0)
    mrc     p15,0,r0,c0,c0,0
    mov     pc,lr
        

/*
 * void SetDataBreakPoint(ADDRESS DataAddress, UINT32 Value, UINT32 Mask)
 */
_FUNCTION(SetDataBreakPoint)
   mcr     p14,0,r0,c14,c0,0       /* set Data Breakpoint Address Register */
   mcr     p14,0,r1,c14,c1,0       /* set Data Breakpoint Value Register */
   mcr     p14,0,r2,c14,c2,0       /* set Data Breakpoint Mask Register */
   mov     r0,#6
   mcr     p14,0,r0,c14,c3,0       /* set Data Breakpoint Control Register */
   mov     pc,lr

/*
 * void SetInstructionBreakPoint(ADDRESS InstructionAddress)
 */
_FUNCTION(SetInstructionBreakPoint)
   bic     r0,r0,#2                /* make sure reserved bit is clear */
   orr     r0,r0,#1                /* enable instruction breakpoint */
   mcr     p14,0,r0,c14,c8,0       /* set Data Breakpoint Control Register */
   mov     pc,lr
#endif	

/*
 * void EnableMMU(ADDRESS PageTableBase)
 */
_FUNCTION(EnableMMU)
    mvn     r1, #0                  /* setup domain access (-1) */
    mcr     p15, 0, r1, c3, c0, 0   /* domain register */
    mcr     p15, 0, r0, c2, c0, 0   /* first level page table */
    mcr     p15, 0, r0, c8, c7, 0   /* flush I+D TLBs */
    mov     r1, #0x007D             /* Enable: MMU */
/*    orr     r1, r1, #0x1000         /* Enable icache */
    orr     r1, r1, #0x1200         /* Enable icache, S=0 R=1 */

    ldr     r0, =VirtualStart

    cmp     r0, #0                  /* make sure no stall on "mov pc,r0" below */
    mcr     p15, 0, r1, c1, c0, 0   /* enable the MMU & Caches */
    mov     pc, r0                  /*  & jump to new virtual address */
    nop

    /* MMU & caches now enabled.
     */
_FUNCTION(VirtualStart)
    nop
    mov     pc,r14
    
/*
 * void DisableMMU(void)
 */
_FUNCTION(DisableMMU)
    mov     r0,#0070               /* Nothing but 32bit addressing */
    mcr     p15, 0, r0, c1, c0, 0  /* Disable the MMU & Caches */
    nop                            /* let it take */
    nop
    mcr     p15, 0, r0, c8, c7, 0  /* flush I+D TLBs */
    mov     pc,r14

    _END
