/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 * Version for the Intrinsyc board.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _CODE(C$$code)

#if !defined (ADS)
    _IMPORT(_end)
#endif
    _IMPORT(BaseInit)
    _IMPORT(TrapInit)
     _IMPORT(TrapHandler)
     _IMPORT(TrapPAEntry)
     _IMPORT(TrapDAEntry)
    _IMPORT(PageTableInit)
    _IMPORT(IntStackPtr)
    _IMPORT(DebugBreak)


    _EXPORT(_start)
    _EXPORT(RegsAtEntry)

/*
 * Hard Vectors
 */
_FUNCTION(ExceptionVectors)
    ldr pc,[pc,#0x18]	/* 0x00: Reset vector */
    ldr pc,[pc,#0x18]	/* 0x04: Undefined instruction vector */
    ldr pc,[pc,#0x18]	/* 0x08: Software Interrupt vector */
    ldr pc,[pc,#0x18]	/* 0x0c: Prefetch Abort vector */
    ldr pc,[pc,#0x18]	/* 0x10: Data Abort vector */
    ldr pc,[pc,#0x18]	/* 0x14: ??? vector */
    ldr pc,[pc,#0x18]	/* 0x18: IRQ vector */
    ldr pc,[pc,#0x18]	/* 0x1c: FIQ vector */
    /*
     * And this is where the above guys jump to
     */
_LABEL(ExceptionHandlersTable)
    _DCD	_start		/* 0x00: Reset handler */
    _DCD	TrapHandler	/* 0x04: Undefined instruction handler */
    _DCD	TrapHandler	/* 0x08: Software Interrupt handler */
    _DCD	TrapPAEntry	/* 0x0c: Prefetch Abort handler */
    _DCD	TrapDAEntry	/* 0x10: Data Abort handler */
    _DCD	TrapHandler	/* 0x14: ??? handler */
    _DCD	TrapHandler	/* 0x18: IRQ handler */
    _DCD	TrapHandler	/* 0x1c: FIQ handler */

_LABEL(RegsAtEntry)
    _DCD 0
    _DCD 0
    _DCD 0
    _DCD 0

/* 
 * Entry point
 */
    _ENTRY

_FUNCTION(_start)

/*
 * Disable interrupts for now, SVC mode
 */
    msr     cpsr_c, #0xD3

/*
 * Save initial register values, the boot loader
 * can pass arguments to us and we might use them.
 */
    str     r0, RegsAtEntry+0
    str     r1, RegsAtEntry+4
    str     r2, RegsAtEntry+8
    str     r3, RegsAtEntry+12

/* Zero BSS
 */
    /* EndPtr = r3 */
    ldr     r3, _L(DataEnd)+0
    ldr     r4, _L(BssEnd)+0
    sub     r5, r5, r5
    sub     r6, r6, r6
_LABEL(ZroLoop)
    cmp     r3, r4
    stmloia r3!, {r5,r6}
    blo     ZroLoop

/*
 * Set stack pointer for SVC mode
 */
    ldr     sp, = _L(IntStackPtr)
    ldr     sp, [sp]

/*
 * Initialize the page tables 
 */
    bl      _F(PageTableInit)

/*
 * Initialize interrupt dispatching
 */
    bl      _F(TrapInit)

/*
 * Drop into debugger if required.
 */
#if 1
    ldr     r0, RegsAtEntry+0
    ands    r0,r0,#1
    blne    _F(DebugBreak)
#else
    bl      _F(DebugBreak)
#endif

/*
 * Go straight for MI init
 */
    b       _F(BaseInit)

    mov     pc,lr    /* amusme */

_LABEL(DataEnd)
    /* Data end == bss start */
#if defined(__GNUC__)
    _IMPORT(_edata)
    _DCD    _edata
#elif defined(_MSC_VER)
     _IMPORT(__bss_start)       /* or __bss_start__ */
     _DCD   __bss_start
#else /* one of the ARM toolsets */
    _IMPORT(|Image$$ZI$$Base|)
    _DCD    |Image$$ZI$$Base|
#endif

_LABEL(BssEnd)
    /* Bss end */
#if defined (ADS)
    _IMPORT(|Image$$ZI$$Limit|)
    _DCD    |Image$$ZI$$Limit|
#else
    _DCD    _end
#endif

    _END
