/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 * Version for the Atmel EB63 board.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _CODE(C$$code)

    _EXPORT(_start)

/*
 * Entry point
 *
 * This startup code is executed at chip powerup/reset.
 * Lives in FLASH, is visible at address 0x0 on powerup,
 * and at address 0x1000000 after TheEbi.RemapRegister
 * has been written to.
 *
 * One desireable feature of this code is to be executable
 * at both of the addresses indicated above. This way a
 * system reset can be performed by jumping at 0x1000000.
 * [NB: we could use the watchdog timer instead]
 *
 * This very same code can also be downloaded at its intended
 * execution place in RAM, and started and debugged in there.
 */
    _ENTRY

_FUNCTION(_start)
    msr     cpsr_c, #0xD3       /* disable ints, SVC mode */
                                /* NB: under Angel only turns intrs off */
    /* Setup the EBI:
     * - enable FLASH by writing to CS0
     * - enable RAM by writing to CS1
     *   and setting TheEbi->MemoryControl = EBIMC_ALE_CS4_6;
     * - stop remapping 
     */
    adr     r1, _L(TheEbi)     /* Setup regs with Ebi ptr and values */
    ldmia   r1, {r1,r2,r3,r4}

    stmia   r1, {r2,r3}        /* Set CS0&CS1 */
    str     r4, [r1, #+0x24]   /* Set MemoryControl */

    mov     r2, #1             /* for RemapControl and 0x1000000 */
    adr     r3, _L(Relocate)
    orr     r3, r3, r2, lsl #24/* r3 = Relocate+0x1000000 */
                               /* This is in FLASH, ONLY good at boot */
    movs    r4, pc, lsr #24    /* FLASH is lower than 0x1000000 at reset */
    str     r2, [r1, #+0x20]   /* stop remapping (if we were doing it) */
    moveq   pc, r3             /* .. and use new addresses if we have to */

    /* At this point:
     * - Memory mapping is the normal one, with FLASH at 1mil,
     *   RAM at 2mil, and SRAM at 0.
     * - If this code is in FLASH, we are running with pc=0x010000xx
     * - If this code is debugged in RAM, pc=0x020080xx (or similar).
     */
_LABEL(Relocate)
    b       _L(ArmStart)        /* PC relative: does not matter where we are */
    b       _start

    /* Static data needed above.
     */
_LABEL(TheEbi)
    _DCD    0xffe00000
    _DCD    0x01002529         /* CS0 for FLASH (see _fredEb63.c) */
    _DCD    0x02002121         /* CS0 for RAM   (see _fredEb63.c) */
    _DCD    0x00000006         /* MemoryControl (see _fredEb63.c) */
    
#include "../_armstart.s"
    _END
