/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

//
// Support routines
//

    _CODE(C$$code)

    _EXPORT(MMUFaultAddress)
    _EXPORT(MMUFaultStatus)
    _EXPORT(ArmMachineIdle)
    _EXPORT(FlushCache)
    _EXPORT(EnableMMU)
//   _EXPORT(SetDataBreakPoint)
//   _EXPORT(SetInstructionBreakPoint)

_FUNCTION(FlushCache)
    mov     r0,#0xE0000000
    add     r1,r0,#8192
_LABEL(l1)
    ldr     r2,[r0],#32
    teq     r1,r0
    bne     l1

    mcr     p15,0,r0,c7,c7,0    // flush I+D caches
    mov     pc,lr

_FUNCTION(MMUFaultAddress)
    mrc     p15,0,r0,c6,c0,0
    mov     pc,lr

_FUNCTION(MMUFaultStatus)
    mrc     p15,0,r0,c5,c0,0
    mov     pc,lr

_FUNCTION(ArmMachineIdle)
    /* Nothing yet */
    mov     pc,lr

//SetDataBreakPoint
//   mcr     p14,0,r0,c14,c0,0       // set DBAR (data breakpoint address register)
//   mcr     p14,0,r1,c14,c1,0       // set DBVR (data breakpoint value register)
//   mcr     p14,0,r2,c14,c2,0       // set DBVR (data breakpoint mask register)
//   mov     r0,#6
//   mcr     p14,0,r0,c14,c3,0       // set DBCR (data breakpoint control register)
//   mov     pc,lr
//
//SetInstructionBreakPoint
//   bic     r0,r0,#2                // make sure reserved bit is clear
//   orr     r0,r0,#1                // enable instruction breakpoint
//   mcr     p14,0,r0,c14,c8,0       // set DBCR (data breakpoint control register)
//   mov     pc,lr


//
// void EnableMMU(ADDRESS PageTableBase)
//
_FUNCTION(EnableMMU)
    mvn     r1, #0                  // setup domain access (-1)
    mcr     p15, 0, r1, c3, c0, 0   // domain register
    mcr     p15, 0, r0, c2, c0, 0   // first level page table
    mcr     p15, 0, r0, c8, c7, 0   // flush I+D TLBs
    mov     r1, #0x007D             // Enable: MMU
    orr     r1, r1, #0x1000         // Enable icache

    ldr     r0, =VirtualStart

    cmp     r0, #0                  // make sure no stall on "mov pc,r0" below
    mcr     p15, 0, r1, c1, c0, 0   // enable the MMU & Caches
    mov     pc, r0                  //  & jump to new virtual address
    nop

// MMU & caches now enabled.

_FUNCTION(VirtualStart)
    nop
    mov     pc,r14
    
    _END
