#include <arm_asm.h>

/*
 * Support routines
 */

    _CODE(C$$code)

    _EXPORT(MMUFaultAddress)
    _EXPORT(MMUFaultStatus)
    _EXPORT(FlushCache)
    _EXPORT(Uint32TimesUint32ToInt64)
    _EXPORT(ReadWrite)

/* We got no cache, and no MMU.
 * The EBI doesnt know what its doing, doesnt remember fault addresses.
 */
_FUNCTION(FlushCache)
_FUNCTION(MMUFaultAddress)
_FUNCTION(MMUFaultStatus)
    sub     r0,r0,r0    
    mov     pc,lr

/*
 * INT64 Uint32TimesUint32ToInt64(UINT32 Mul1, UINT32 Mul2)
 *
 * Returns the 64bit result of a 32x32bit multiplication
 * Arguments:
 * r0 = pointer to spot for returned value 
 * r1 = Mul1
 * r2 = Mul2
 */
_FUNCTION(Uint32TimesUint32ToInt64)
#if defined(__GNUC__)
    /* Compiler doesnt understand UMULL ?!? */
    .int    0xe0831192
#else
    umull   r1,r3,r2,r1
#endif
    stmia   r0, {r1,r3}
    mov     pc,lr

/*
 * UINT32 ReadWrite(PUINT32 ReadAddress,
 *                  PUINT32 WriteAddress,
 *                  UINT32 WriteValue);
 * 
 * In well-known time, read the I/O location at READ_ADDRESS and then write
 * the WRITE_VALUE at I/O location WRITE_ADDRESS. Return the value read.
 * NB: This should live in scratch RAM.
 */
_FUNCTION(ReadWrite)
    ldr     r0,[r0]
    str     r2,[r1]
    mov     pc,lr

    _END
