/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

        .set noreorder
        .set nomacro
        .set noat
        .set volatile

#if TIGON
        _IMPORT(SchedInt)
        _IMPORT(_Reschedule)
#else
        _IMPORT(Trap)
	    _IMPORT(ExtensionControl)
#endif
        _IMPORT(ProcessImageFromPC)
        _IMPORT(ProcessArgsFromPC)

        _EXPORT(HWLoadThreadContext)
        _EXPORT(HWSaveThreadContext)
#ifdef TIGON
        _EXPORT(SchedulingInterrupt)
#endif
        _EXPORT(Exception)
        _EXPORT(Exception_end)
        _EXPORT(UserModeTLBMiss)
        _EXPORT(UserModeTLBMiss_end)
        _EXPORT(DebugBreak)
        _EXPORT(AtomicCmpAndSwap)
        _EXPORT(DebuggerNotifyLoad)
        _EXPORT(DebuggerNotifyUnload)
        _EXPORT(ProcessImage)
        _EXPORT(ProcessArgs)
        _EXPORT(DelayLoop)
	

/* noreturn HWLoadThreadContext(PCXTINFO pCxt);
 *
 * Loads all cpu registers (except k0,k1) from struct pointed to by arg.
 * "longjmp".
 *
 * Never returns. The stack is unwound.
 */
ENTRY(HWLoadThreadContext)

#ifdef TIGON
	.set macro
	/* put IntsDisabled into k1 */
	la	$k1, IntsDisabled
	.set nomacro
#endif
        
	/* First load most registers */
	lw	$at, TS_AT($a0)
	lw	$v0, TS_V0($a0)
	lw	$v1, TS_V1($a0)
	/* a0 later */
	lw	$a1, TS_A1($a0)
	lw	$a2, TS_A2($a0)
	lw	$a3, TS_A3($a0)
	lw	$t0, TS_T0($a0)
	lw	$t1, TS_T1($a0)
	lw	$t2, TS_T2($a0)
	lw	$t3, TS_T3($a0)
	lw	$t4, TS_T4($a0)
	lw	$t5, TS_T5($a0)
	lw	$t6, TS_T6($a0)
	lw	$t7, TS_T7($a0)
	lw	$s0, TS_S0($a0)
	lw	$s1, TS_S1($a0)
	lw	$s2, TS_S2($a0)
	lw	$s3, TS_S3($a0)
	lw	$s4, TS_S4($a0)
	lw	$s5, TS_S5($a0)
	lw	$s6, TS_S6($a0)
	lw	$s7, TS_S7($a0)
	lw	$t8, TS_T8($a0)
	lw	$t9, TS_T9($a0)
    /* k0,k1 not restored */
	lw	$gp, TS_GP($a0)
	/* sp later */
	lw	$fp, TS_FP($a0)
	lw	$ra, TS_RA($a0)

#ifdef TIGON
	/* We can use K0, K1 on the Tigon */
	lw	$sp, TS_SP($a0)
	/* Put pc into k0 */
	lw	$k0, TS_PC($a0)
	lw	$a0, TS_A0($a0)
	j	$k0
	sw	$zero, 0($k1)	/* Enable interrupts. XXX unconditional */
	break 0
#else
    lw  $k1, TS_HI($a0)
    lw  $k0, TS_LO($a0)
    mthi $k1
    mtlo $k0
	/* eMIPS Extension control register */
#if 0
    lw  $k1, TS_EC($a0)
#else
	.set macro
	la	$k1, ExtensionControl
	.set nomacro
    lw  $k1, 0($k1)
    lw  $k0, TS_EC($a0)
	or  $k1,$k0,$k1
#endif	
    mtc0 $k1, $cp0_ectl
#if 0
    mfc0 $k1, $cp0_status
    nop
    ori  $k1,4 // IEp
#else
    lw  $k1, TS_SR($a0)
#endif
    mtc0 $k1, $cp0_status
     /* NB: After this instruction we cannot take any interrupts or traps
      */
	lw	$sp, TS_SP($a0)
	lw	$k1, TS_K1($a0)
	/* Put pc into k0 */
	lw	$k0, TS_PC($a0)
	lw	$a0, TS_A0($a0)
	j	$k0
    rfe
#endif
END(HWLoadThreadContext)

/* INT HWSaveThreadContext(PCXTINFO state)
 *
 * Saves current registers into struct pointed to by arg.
 * "setjmp".
 *
 * Caller saved regs at,v1,a0-3,t0-7,ra,mdlo,mdhi are not saved but zeroed.
 * v0 (return value) is not saved but is set to one.
 * Returns zero normally but may return anything when "longjmped" to.
 */
ENTRY(HWSaveThreadContext)
	/* caller pc is now in ra. save it as pc */
	sw	$ra, TS_PC($a0)
	/* save callee saved registers */
	sw	$s0, TS_S0($a0)
	sw	$s1, TS_S1($a0)
	sw	$s2, TS_S2($a0)
	sw	$s3, TS_S3($a0)
	sw	$s4, TS_S4($a0)
	sw	$s5, TS_S5($a0)
	sw	$s6, TS_S6($a0)
	sw	$s7, TS_S7($a0)
	sw	$t8, TS_T8($a0)
	sw	$t9, TS_T9($a0)
	sw	$gp, TS_GP($a0)
	sw	$sp, TS_SP($a0)
	sw	$fp, TS_FP($a0)

	/* save one as return value */
	lui	$t0, 1
	sw	$t0, TS_V0($a0)
	/* caller saved registers: save as zero */
	sw	$zero, TS_AT($a0)
	sw	$zero, TS_V1($a0)
	sw	$zero, TS_A0($a0)
	sw	$zero, TS_A1($a0)
	sw	$zero, TS_A2($a0)
	sw	$zero, TS_A3($a0)
	sw	$zero, TS_T0($a0)
	sw	$zero, TS_T1($a0)
	sw	$zero, TS_T2($a0)
	sw	$zero, TS_T3($a0)
	sw	$zero, TS_T4($a0)
	sw	$zero, TS_T5($a0)
	sw	$zero, TS_T6($a0)
	sw	$zero, TS_T7($a0)
	sw	$zero, TS_RA($a0)
#ifdef TIGON
#else
    mfc0 $t0, $cp0_ectl
    mfc0 $t1, $cp0_status
	sw	$t0,   TS_EC($a0)
	sw	$t1,   TS_SR($a0)
    mflo $t0
    mfhi $t1
	sw	$t0,   TS_LO($a0)
	sw	$t1,   TS_HI($a0)
	sw	$k1,   TS_K1($a0)
	/* k0 trashed by (exiting) traps, not usable */
	sw	$zero, TS_K0($a0)
#endif

	j	$ra
	move	$v0, $zero		/* return zero */

END(HWSaveThreadContext)

#ifdef TIGON
/* void SchedulingInterrupt(void)
 *
 * Second CPU switches first one here on timer interrupt
 */
ENTRY(SchedulingInterrupt)
	/* Save state on stack */
	addiu   $sp, $sp, -SIZEOF_CXTINFO
	/* save callee saved registers */
	sw	$at, TS_AT($sp)
	sw	$v0, TS_V0($sp)
	sw	$v1, TS_V1($sp)
	sw	$a0, TS_A0($sp)
	sw	$a1, TS_A1($sp)
	sw	$a2, TS_A2($sp)
	sw	$a3, TS_A3($sp)
	sw	$t0, TS_T0($sp)
	sw	$t1, TS_T1($sp)
	sw	$t2, TS_T2($sp)
	sw	$t3, TS_T3($sp)
	sw	$t4, TS_T4($sp)
	sw	$t5, TS_T5($sp)
	sw	$t6, TS_T6($sp)
	sw	$t7, TS_T7($sp)
	sw	$s0, TS_S0($sp)
	sw	$s1, TS_S1($sp)
	sw	$s2, TS_S2($sp)
	sw	$s3, TS_S3($sp)
	sw	$s4, TS_S4($sp)
	sw	$s5, TS_S5($sp)
	sw	$s6, TS_S6($sp)
	sw	$s7, TS_S7($sp)
	sw	$t8, TS_T8($sp)
	sw	$t9, TS_T9($sp)
	sw	$k0, TS_K0($sp)
	sw	$k1, TS_K1($sp)
	sw	$gp, TS_GP($sp)
	/* sp: later */
	sw	$fp, TS_FP($sp)
	sw	$ra, TS_RA($sp)

.set macro
	/* Fetch exception PC and put it where it should */
	la	$a1, ExceptionPC
	lw	$a2, 0($a1)

#if notyet /* handle LO/HI! */
	sw	$zero, TS_MDLO($sp)
	sw	$zero, TS_MDHI($sp)
    mfc0 $a2, $cp0_epc
#endif
	sw	$a2, TS_PC($sp)

	/* Save original stack */
	addiu   $a3, $sp, SIZEOF_CXTINFO
	sw	$a3, TS_SP($sp)
	move	$a0, $sp

	/* noreturn SchedInt(pCxtInfo *a0) */
	jal	SchedInt
	/* Space for arg */
	addiu   $sp, -4
END(SchedulingInterrupt)

#else
/*
 * void Exception(void)
 * LABEL Exception_end
 *
 */
ENTRY(Exception)
	/* Save state on stack */
	addiu   $sp, $sp, -SIZEOF_CXTINFO
	/* save registers */
	sw	$at, TS_AT($sp)
	sw	$v0, TS_V0($sp)
	sw	$v1, TS_V1($sp)
	sw	$a0, TS_A0($sp)
	sw	$a1, TS_A1($sp)
	sw	$a2, TS_A2($sp)
	sw	$a3, TS_A3($sp)
	sw	$t0, TS_T0($sp)
	sw	$t1, TS_T1($sp)
	sw	$t2, TS_T2($sp)
	sw	$t3, TS_T3($sp)
	sw	$t4, TS_T4($sp)
	sw	$t5, TS_T5($sp)
	sw	$t6, TS_T6($sp)
	sw	$t7, TS_T7($sp)
	sw	$s0, TS_S0($sp)
	sw	$s1, TS_S1($sp)
	sw	$s2, TS_S2($sp)
	sw	$s3, TS_S3($sp)
	sw	$s4, TS_S4($sp)
	sw	$s5, TS_S5($sp)
	sw	$s6, TS_S6($sp)
	sw	$s7, TS_S7($sp)
	sw	$t8, TS_T8($sp)
	sw	$t9, TS_T9($sp)
	sw	$k0, TS_K0($sp)
	sw	$k1, TS_K1($sp)
	sw	$gp, TS_GP($sp)
	/* sp: later */
	sw	$fp, TS_FP($sp)
	sw	$ra, TS_RA($sp)

    mfc0 $t0, $cp0_ectl
    mfc0 $t1, $cp0_status
	sw	$t0,   TS_EC($sp)
	sw	$t1,   TS_SR($sp)
	
    mflo $t0
    mfhi $t1
    mfc0 $a2, $cp0_epc
	sw	$t0, TS_LO($sp)
	sw	$t1, TS_HI($sp)
	sw	$a2, TS_PC($sp)

	/* Save original stack */
	addiu   $t0, $sp, SIZEOF_CXTINFO
    bgezal  $zero, _L(GotoTrap)
	sw	    $t0, TS_SP($sp)

_LABEL(AddressOfTrap)
     _DCD   Trap

_LABEL(GotoTrap)
    lw   $a2,0($ra)
    mfc0 $a1,$cp0_cause
    jr   $a2
    move $a0, $sp

_LABEL(Exception_end)
END(Exception)

/*
 * void UserModeTLBMiss(void);
 * LABEL UserModeTLBMiss_end
 *
 */
ENTRY(UserModeTLBMiss)
   /* We dont do this (yet)
    */
    break 1
    break 1
_LABEL(UserModeTLBMiss_end)
END(UserModeTLBMiss)

#endif

ENTRY(DebugBreak)
    break 0
	j $ra
	nop
END(DebugBreak)

#ifdef TIGON
#else
/*
 * BOOL AtomicCmpAndSwap(PUINT pTarget, UINT OldVal, UINT NewVal)
 *
 * Compare *pTarget with OldVal.  If equal, set *pTarget to NewVal,
 * return TRUE; else return FALSE.
 */
ENTRY(AtomicCmpAndSwap)
        ll      $v0,0($a0)
        bne     $v0,$a1,ACASDone
        move    $v0,$zero
        move    $v0,$a2
        sc      $v0,0($a0)
        beq     $v0,$zero,AtomicCmpAndSwap
        nop
ACASDone:
        j       $ra
        nop
END(AtomicCmpAndSwap)

/*
 * PUBLIC UINT DebuggerNotifyLoad(const char *ModuleName,
 *                               ADDRESS LoadAddress,
 *                               ADDRESS LoadEnd,
 *                               PTR Unused)
 * Notify an attached debugger of the loading of a new module.
 * Returns a handle to give back in the NotifyUnload below.
 */
ENTRY(DebuggerNotifyLoad)
        break   1
        j       $ra
        move    $v0,$zero
END(DebuggerNotifyLoad)

/*
 * void DebuggerNotifyUnload(const _TCHAR *ModuleName, ADDRESS LoadStart,
 *                          UINT Token)
 * Notify an attached debugger of the unloading of a module.
 */
ENTRY(DebuggerNotifyUnload)
        break   2
        j       $ra
        nop
END(DebuggerNotifyUnload)


#endif
	
ENTRY(ProcessImage)
        j       ProcessImageFromPC
        move    $a0,$ra
END(ProcessImage)

ENTRY(ProcessArgs)
        j       ProcessArgsFromPC
        move    $a0,$ra
END(ProcessArgs)

/*
 * extern void DelayLoop(INT32 LoopCount);
 *
 * Spins as fast as possible for the given number of times.
 * If argument negative same as zero (or one, no matter).
 */
ENTRY(DelayLoop)
        bgtz    $a0, _L(DelayLoop)
        addi    $a0, $a0, -1
        j       $ra
        nop
END(DelayLoop)

