/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

/*
 * Startup code for MIPS MMLite.
 * Version for the Atmel EB63 board (Giano simulator)
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _EXPORT(_start)

/*
 * Entry point
 *
 * This startup code is executed at chip powerup/reset.
 * Lives in FLASH, is visible at address 0x0 on powerup,
 * and at address 0x1000000 after TheEbi.RemapRegister
 * has been written to.
 *
 * One desireable feature of this code is to be executable
 * at both of the addresses indicated above. This way a
 * system reset can be performed by jumping at 0x1000000.
 * [NB: we could use the watchdog timer instead]
 *
 * This very same code can also be downloaded at its intended
 * execution place in RAM, and started and debugged in there.
 */
    _ENTRY

ENTRY(_start)
	/* Find out where we are running from
     */      
	.set noreorder
    bgezal  $zero, _L(_start2) /* Always jumps */
    nop
        
    /* Static data needed below.
     */
_LABEL(TheEbi)
    _DCD    0xffe00000
    _DCD    0x29250001         /* CS0 for FLASH (see _fredEb63.c) */
    _DCD    0x21210002         /* CS0 for RAM   (see _fredEb63.c) */
    _DCD    0x06000000         /* MemoryControl (see _fredEb63.c) */

_LABEL(_start2)
        
    /* Setup the EBI:
     * - enable FLASH by writing to CS0
     * - enable RAM by writing to CS1
     *   and setting TheEbi->MemoryControl = EBIMC_ALE_CS4_6;
     * - stop remapping 
     */
	lw      $t0, 0($ra)
	lw      $t1, 4($ra)
	lw      $t2, 8($ra)
	lw      $t3, 12($ra)
	
    sw      $t1, 0($t0)        /* Set CS0&CS1 */
    sw      $t2, 4($t0)
    sw      $t3, 0x24($t0)     /* Set MemoryControl */

    srl     $t3,$t3,1          /* ..well, yes not so nice */
    sw      $t3, 0x20($t0)     /* stop remapping (if we were doing it) */
	
    /* At this point:
     * - Memory mapping is the normal one, with FLASH at 1mil,
     *   RAM at 2mil, and SRAM at 0.
     * - If this code is in FLASH, we are running with pc=0x010000xx
     * - If this code is debugged in RAM, pc=0x020080xx (or similar).
     */
    b       _L(MipsStart)      /* PC relative: does not matter where we are */
    addi    $a0,$ra,-8         /* Compensate for the first two instructions */

END(_start)
    
#include "../_mipsstart.s"

