/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define DBGLVL 0
#define DEVICE_ORDER BIG_ENDIAN
#include <mmlite.h>
#include <mmhal.h>
#include <mmmacros.h>
#include <fred.h>
#include <mips/ml40x.h>

extern void _start(void); /* truly our image's start */
extern void MMUEnable(void);
extern void DbgPortInit(void);

#define _p2v(x) (0x80000000 | (x))

#define ThePmc ((struct _Pmc*)POWER_MGR_DEFAULT_ADDRESS)

void BoardInit(void)
{
#if 0
    /* Spin a bit to let the serplex start.
     * It runs on the same serial line we have been downloaded from
     */
    Delay(50000);
#endif

    /* Turn the clock on to all the peripherals we (know we'll) use
     */
    ThePmc->PeripheralPowerEnable = _bs32(PMCPC_USART);

    /* Debug serial line
     */
    DbgPortInit();

    /* ..Any other peripherals that need initialization
     */
    MMUEnable();
}


/* See how much SRAM we have
 * Get it from the SRAM controller.
 */
static struct _Sram *SramCtl(void)
{
    static struct _Sram *Sram = NULL;

    if (Sram == NULL) {
        struct _Pmt *Pmt = ThePmt;
        for (;;Pmt--) {
            if (Pmt->Tag == PMTTAG_SRAM) {
                Sram = (struct _Sram *)(Pmt->TopOfPhysicalAddress << 16);
                break;
            }
            if (Pmt->Tag == PMTTAG_END_OF_TABLE)
                return NULL; /* and good luck to you */
        }
    }
    return Sram;
}

static ADDRESS SRAM_Base(void)
{
    static ADDRESS Base = ~0;

    if (Base == ~0) {
        struct _Sram *Ctl = SramCtl();
        if (Ctl == NULL)
            Ctl = (struct _Sram *)SRAM_0_DEFAULT_ADDRESS;//return 0;  /* and good luck to you */
        Base = Ctl->BaseAddressAndTag & SRAMBT_BASE;
        Base = _p2v(Base);
    }
    return Base;
}

UINT MemorySize(void)
{
    struct _Sram *Ctl = SramCtl();
    if (Ctl == NULL)
            Ctl = (struct _Sram *)SRAM_0_DEFAULT_ADDRESS;//return 0;  /* and good luck to you */
    return Ctl->Control & SRAMST_SIZE;
}

/* Create the system heap.  Left to machdep code because messy otherwise.
 * The heap should account for all available memory, carved out of memory
 * that is already used or unavailable (like the 640K-1meg hole on PCs).
 * It does not have to account for all the memory in the system.
 * Returns a PIHEAP pointer.
 */

PIHEAP MachineHeapCreate( void )
{
    PIHEAP pIHeap;
    ADDRESS MemBase = (ADDRESS) _end;
    ADDRESS ImageBase = (ADDRESS) _start;
    UINT MemSize = MemorySize();
    ADDRESS SRAM = SRAM_Base();
    ADDRESS MemTop = (SRAM+MemSize);

    DBGME(1,printf("MachineHeapCreate: SRAM=%x..%x start=%x end=%x delta=",
                   SRAM, MemTop, ImageBase, MemBase));

    if ((SRAM == 0x80000000) || (ImageBase < SRAM) || (ImageBase > MemTop)) {
        DBGME(1,printf("%x ROM/Flash\n", MemBase-SRAM));
        /* If we are running from ROM/Flash, assume .data is at the
         * beginning of RAM so start the heap after it.
         * Does not account for image in heap size (the heap is .data + .bss
         * smaller than SRAM size).
         */
        MemSize -= MemBase - SRAM;
        pIHeap = CreateHeapFrom(MemBase, 0, MemSize, MemSize);
    } else {
        DBGME(1,printf("%x RAM\n", MemBase-ImageBase));
        /* If instead .text is in RAM then the image might be in the middle
         * of it so make a heap out of the entire RAM and then carve out the
         * image. This assumes there is space for the heap metadata at
         * start of RAM.
         * Accounts for image in heap size (the heap size is that of SRAM size)
         */
        pIHeap = CreateHeapFrom(SRAM, 0, MemSize, MemSize);

        if (MemTop > MemBase) {
            MemSize = MemBase - ImageBase;
            pIHeap->v->Extract(pIHeap, 0, (PTR) ImageBase, MemSize);
        }
    }

    return pIHeap;
}

void TheEnd(void)
{
    void (*Reset)(void) = (void (*)(void))BRAM_DEFAULT_ADDRESS;
    UINT32 psr;

    /* Say we are resetting
     */
    printf("Jumping to %x cuz I'm stoop (seeya!)\n", (ADDRESS)Reset);
    /* Turn off interrupts and cause a reset
     */
    TURN_INTERRUPTS_OFF(psr);
    psr |= (1 << 7);
    RESTORE_INTERRUPTS(psr);
    /* Will never get here but
     */
    for(;;)
        Reset();
}


/* .. */
void MipsMachineIdle(ADDRESS Arg)
{
    UnusedParameter(Arg);
    ThePmc->SystemPowerDisable = _bs32(PMCSC_CPU);
}

/* Say we are #1
 */
#undef CurrentProcessorOEM
UINT CurrentProcessorOEM(void)
{
    return 1;
}

void DebuggerOutputString( const unsigned char *String, int nBytes )
{
    int i;
    for (i = 0; i < nBytes; i++)
        DCPutChar(String[i]);
}
