/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <ppc_asm.h>

         _CODE(text)

        _IMPORT(ProcessImageFromPC)
        _IMPORT(ProcessArgsFromPC)
        
        _IMPORT(DelayMultiplier)
        _IMPORT(DelayShifter)
        _IMPORT(DelayOverhead)

/*
 *  const _TCHAR * ProcessArgs(void);
 *
 * Turns interrupts off, returns previous value of PSR
 */
ENTRY(ProcessArgs)
        mflr    r3
        b       _F(ProcessArgsFromPC)
END(ProcessArgs)

/*
 *  const _TCHAR * ProcessImage(void);
 *
 * Turns interrupts off, returns previous value of PSR
 */
ENTRY(ProcessImage)
        mflr    r3
        b       _F(ProcessImageFromPC)
END(ProcessImage)

/*
 * void TuneDelayLoop(INT32 LoopCount)
 *
 * Spins as fast as possible for the given number of times.
 * If argument negative same as zero (or one, no matter).
 */
ENTRY(TuneDelayLoop)
	    mr.     r4,r3
        mtctr   r3
	    blelr
        /* Simulator wont allow self-branches */
lp:     
        bdzlr
        b       lp
END(TuneDelayLoop)

/*
 * void Delay(UINT uSec)
 *
 * Spends AT LEAST uSec microsecond spinning
 * Interrupts should not be touched, so thats why.
 */
ENTRY(Delay)
     /* The implementation in C is:
      *        volatile INT nLoops = (INT)uSec;
      *        nLoops = (nLoops * DelayMultiplier) - DelayOverhead;
      *        nLoops = nLoops >> DelayShifter;
      *        TuneDelayLoop(nLoops)
      */
     lis       r11,DelayMultiplier
     lwz       r11,DelayMultiplier(r11)
     mulld     r3,r11,r3
     lis       r11,DelayOverhead
     lwz       r11,DelayOverhead(r11)
     subf      r3,r11,r3
     lis       r11,DelayShifter
     lwz       r11,DelayShifter(r11)
     srad      r3,r3,r11
     b         TuneDelayLoop
END(Delay)



/*
 *  BOOL AtomicCmpAndSwap(PUINT pTarget, UINT OldVal, UINT NewVal)
 *
 * Atomically compare *pTarget to OldVal and if equal write NewVal to it.
 * Otherwise do nothing.
 *
 * Returns TRUE iff the write did happen
 */
ENTRY(AtomicCmpAndSwap)
        lwarx   r6,r0,r3
        cmplw   r6,r4
        bne     _L(Fail)
        stwcx.  r5,r0,r3
        bne     _L(Fail)
        li      r3,1
        blr
Fail:
        li      r3,0
        blr
END(AtomicCmpAndSwap)

/*
 *   void WriteDecrementer(UINT32 Value);
 *
 * Write to the decrementer register
 */
ENTRY(WriteDecrementer)
        mtdec   r3
        blr
END(WriteDecrementer)

/*
 *  UINT32 ReadDecrementer(void);
 *
 * Returns the current value of the decrementer register
 */
ENTRY(ReadDecrementer)
        mfdec   r3
        blr
END(ReadDecrementer)

/*
 *   void WriteTimeBase(UINT64 Value);
 *
 * Write to the timebase registers
 */
ENTRY(WriteTimeBase)
#if 0
        /* Assembler does not parse these?!?! */
        sradi   r4,r3,32
#else
        rldicl  r4,r3,32,0
#endif
        li      r5,0
#if 0
        /* Assembler does not parse these?!?! */
        mttbl   r5
        mttbu   r4
        mttbl   r3
#else
        mtspr   284,r5
        mtspr   285,r4
        mtspr   284,r3
#endif
        blr
END(WriteTimeBase)

/*
 *  UINT64 ReadTimeBase(void);
 *
 * Returns the current value of the timebase registers
 */
ENTRY(ReadTimeBase)
        mftb   r3
        blr
END(ReadTimeBase)

/*
 *  __noreturn void HWLoadThreadContext(PCXTINFO pCxt);
 *
 * Load the CPU state pointed to by pCxt.
 * Does not return, switches stack (unwinding it if same)
 */
ENTRY(HWLoadThreadContext)
        ld      r4, CXTLR(r3)
        mtlr    r4
        ld      r4, CXTPC(r3)
        mtsrr0  r4
        ld      r4, CXTMSR(r3)
        mtsrr1  r4
        ld      r4, CXTCTR(r3)
        mtctr   r4
        ld      r4, CXTCR(r3)
        mtcr    r4
        ld      r4, CXTXER(r3)
        mtxer   r4

        ld      r0, CXTR0(r3)
        ld      r1, CXTR1(r3)
        ld      r2, CXTR2(r3)

        ld      r4, CXTR4(r3)
        ld      r5, CXTR5(r3)
        ld      r6, CXTR6(r3)
        ld      r7, CXTR7(r3)
        ld      r8, CXTR8(r3)
        ld      r9, CXTR9(r3)
        ld      r10, CXTR10(r3)
        ld      r11, CXTR11(r3)
        ld      r12, CXTR12(r3)
        ld      r13, CXTR13(r3)
        ld      r14, CXTR14(r3)
        ld      r15, CXTR15(r3)
        ld      r16, CXTR16(r3)
        ld      r17, CXTR17(r3)
        ld      r18, CXTR18(r3)
        ld      r19, CXTR19(r3)
        ld      r20, CXTR20(r3)
        ld      r21, CXTR21(r3)
        ld      r22, CXTR22(r3)
        ld      r23, CXTR23(r3)
        ld      r24, CXTR24(r3)
        ld      r25, CXTR25(r3)
        ld      r26, CXTR26(r3)
        ld      r27, CXTR27(r3)
        ld      r28, CXTR28(r3)
        ld      r29, CXTR29(r3)
        ld      r30, CXTR30(r3)
        ld      r31, CXTR31(r3)

        ld      r3, CXTR3(r3)

        rfid

END(HWLoadThreadContext)

/*
 *  UINT HWSaveThreadContext(PCXTINFO pCxt);
 *
 * Save CPU register state where pCxt points.
 * Caller-saved registers are *not* saved, just zeroed. BUGBUG
 *
 * Returns zero, unless it gets long-jumped to by the above
 * in which case it returns 1.
 */
ENTRY(HWSaveThreadContext)
        std     r0, CXTR0(r3)
        std     r1, CXTR1(r3)
        std     r2, CXTR2(r3)
#if 0
        std     r3, CXTR3(r3)
        std     r4, CXTR4(r3)
#else
        std     r4, CXTR4(r3)
        li      r4, 1
        std     r4, CXTR3(r3)
#endif
        std     r5, CXTR5(r3)
        std     r6, CXTR6(r3)
        std     r7, CXTR7(r3)
        std     r8, CXTR8(r3)
        std     r9, CXTR9(r3)
        std     r10, CXTR10(r3)
        std     r11, CXTR11(r3)
        std     r12, CXTR12(r3)
        std     r13, CXTR13(r3)
        std     r14, CXTR14(r3)
        std     r15, CXTR15(r3)
        std     r16, CXTR16(r3)
        std     r17, CXTR17(r3)
        std     r18, CXTR18(r3)
        std     r19, CXTR19(r3)
        std     r20, CXTR20(r3)
        std     r21, CXTR21(r3)
        std     r22, CXTR22(r3)
        std     r23, CXTR23(r3)
        std     r24, CXTR24(r3)
        std     r25, CXTR25(r3)
        std     r26, CXTR26(r3)
        std     r27, CXTR27(r3)
        std     r28, CXTR28(r3)
        std     r29, CXTR29(r3)
        std     r30, CXTR30(r3)
        std     r31, CXTR31(r3)
        mflr    r4
        std     r4, CXTLR(r3)
        std     r4, CXTPC(r3)
     /* This is called with interrupts disabled, but it really
      * wants to keep interrupts enabled.
      * BUGBUG: Should change the caller to do a GetPsr() before
      * BUGBUG  disabling interrupts and sticking that value in.
      */
        mfmsr   r4
        ori     r4,r4,MSR_EE_MASK
        std     r4, CXTMSR(r3)
        mfctr   r4
        std     r4, CXTCTR(r3)
        mfcr    r4
        std     r4, CXTCR(r3)
        mfxer   r4
        std     r4, CXTXER(r3)
        li      r3,0
        blr
END(HWSaveThreadContext)

/*
 * UINT64 SetDebugAddressBreakpointRegister(UINT64 Value)
 *
 * Sets the Data Address Breakpoint Register to the given value.
 * Returns the previous value
 */
ENTRY(SetDebugAddressBreakpointRegister)
        mfspr   r4,1013
        mtspr   1013,r3
        mr      r3,r4
        blr
END(SetDebugAddressBreakpointRegister)

/*
 * UINT64 SetDebugAddressBreakpointRegisterPriviledge(UINT64 Value)
 *
 * Sets the Data Address Breakpoint Register Priviledge to the given value.
 * Returns the previous value
 */
ENTRY(SetDebugAddressBreakpointRegisterPriviledge)
        mfspr   r4,1015
        mtspr   1015,r3
        mr      r3,r4
        blr
END(SetDebugAddressBreakpointRegisterPriviledge)

/*
 *  UINT64 _get_<FOO>(void);
 *
 * Returns the current value of the special register <FOO>
 */
GETSPR MACRO nReg
       mfspr r3,nReg
       blr
ENDM
ENTRY(_get_dabrx)
       GETSPR 1015
END(_get_dabrx)

ENTRY(_get_hdec)
       GETSPR 310
END(_get_hdec)

ENTRY(_get_hior)
       GETSPR 311
END(_get_hior)

ENTRY(_get_hsrr0)
       GETSPR 314
END(_get_hsrr0)

ENTRY(_get_hsrr1)
       GETSPR 315
END(_get_hsrr1)

ENTRY(_get_hsprg0)
       GETSPR 304
END(_get_hsprg0)

ENTRY(_get_hsprg1)
       GETSPR 305
END(_get_hsprg1)

ENTRY(_get_asr)
       GETSPR 280
END(_get_asr)


        _ENDCODE
        _END
