/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <ppc_asm.h>
#define __ASSEMBLER__ 1
#include "../ppc_traps.h"
#include "eb63.h"

        _CODE(text)


        _IMPORT(BootProcessorStartup)
        _IMPORT(NonBootProcessorStartup)
        _IMPORT(FlushCurrentEntireTb)
        _IMPORT(FlushCache)
        _IMPORT(Real0SectionStart)
        _IMPORT(Real0SectionEnd)

        _DATA(0)

        _EXPORT(ProcessorControlRegion)

/*
 * Export to C the address of the PCR array
 */
ProcessorControlRegion _EQU PCR_ARRAY_PHYSICAL_ADDRESS

/*
 * Initial stack for the boot processor.
 */
BootStack              _EQU ProcessorControlRegion+4000H
BOOT_STACK_TOP_FRAME   _EQU 0

/*
 * Stores the contents of the HID registers from the boot processor.
 */
        _BSS(BootProcessorHIDRegisters,4*8)
BootHid0 _EQU 0
BootHid1 _EQU 8
BootHid4 _EQU 16
BootHid5 _EQU 24

        _ENDDATA
        _CODE(text)

/*
 *
 * VOID
 * InitializeSegmentRegisters(
 *     VOID
 *     )
 *
 * Routine Description:
 *
 *     This function initializes the segment registers.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Notes:
 *
 *     Modifies condition register field 0.
 *
 */

ENTRY(InitializeSegmentRegisters)

        li      r4,0Fh                  // RB[index]=15
        oris    r4,r4,0800h             //    ; RB[V]=1
        slbia                           // invalidate SLB (except index 0); invalidate ERATs
SlbNextEntry:
        rlwinm. r3,r4,12,16,19          // RS[VSID]=RB[index],RS[Ks]=0,RS[Kp]=0,RS[N]=0,RS[L]=0,RS[C]=0
        rlwimi  r4,r4,28,0,3            // RB[ESID]=RB[index]
        slbmte  r3,r4                   // ESID -> VSID
        subi    r4,r4,1                 // RB[index] -= 1
        bne+    SlbNextEntry            // Continue for all 16 entries
        blr

END(InitializeSegmentRegisters)

/*
 *
 * VOID
 * InitializePageDirectory(
 *     VOID
 *     )
 *
 * Routine Description:
 *
 *     This function initializes the page directory page.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 */

ENTRY(InitializePageDirectory)

/*
 * Zero the contents of the page directory page.
 */

        li      r3,PAGE_DIRECTORY_PHYSICAL_ADDRESS
        li      r4,PAGE_SIZE SHR 7
        mtctr   r4

ZeroNextPdeCacheBlock:
        dcbz128 r0,r3
        addi    r3,r3,128
        bdnz+   ZeroNextPdeCacheBlock

/*
 * Fill the page directory with large page mappings of the physical address
 * range 0x00000000-0x1FFFFFFF at the cached virtual address range
 * 0x00000000-0x1FFFFFFF.  Max mem=512 MB.
 */

        li      r3,000000000h SHR PDI_SHIFT SHL 2
        li      r4,VALID_CACHED_LARGE_PDE_BITS
        li      r5,128
        mtctr   r5

StoreNextCachedPde:
        stw     r4,PAGE_DIRECTORY_PHYSICAL_ADDRESS(r3)
        addis   r4,r4,HIGHWORD(PAGE_SIZE_LARGE)
        addi    r3,r3,4
        bdnz+   StoreNextCachedPde

/*
 * Fill the page directory with large page mappings of the physical address
 * range 0x80000000-0xFFFFFFFF at the uncached virtual address range
 * 0x80000000-0xFFFFFFFF.
 */

        li      r3,080000000h SHR PDI_SHIFT SHL 2
        lis     r4,HIGHWORD(080000000h)
        ori     r4,r4,VALID_UNCACHED_LARGE_PDE_BITS
        li      r5,512
        mtctr   r5

StoreNextUncachedPde2:
        stw     r4,PAGE_DIRECTORY_PHYSICAL_ADDRESS(r3)
        addis   r4,r4,HIGHWORD(PAGE_SIZE_LARGE)
        addi    r3,r3,4
        bdnz+   StoreNextUncachedPde2
	    blr

END(InitializePageDirectory)

/*
 *
 * VOID
 * _start(
 *     VOID
 *     )
 *
 * Routine Description:
 *
 *     This function is the first point of entry for the system.
 *
 *     Although relocations are disabled, the branches below will still
 *     work because they are position independent instructions.
 *     Any called code also does not rely on any absolute addresses.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 */

        _ENTRY

ENTRY(_start)

/*
 * Disable interrupts.
 */

	    li      r31,0
        mtmsree r31

/*
 * Disable instruction and data relocation if they are enabled.  We assume
 * that if we are running at a virtual address, then the virtual address equals
 * the physical address.
 */

        mfmsr   r31
        li      r30,MSR_DR_MASK+MSR_IR_MASK
        andc    r31,r31,r30
        mtmsr   r31
        sync
        isync

/*
 * Store the values of the hardware implementation dependent registers 
 * for later use when starting non-boot processors.
 */
        lis     r27,BootProcessorHIDRegisters
        addi    r27,r27,BootProcessorHIDRegisters
        clrlwi  r27,r27,1                    /* clear high order address bit */
        mfspr   r28,SPR_HID0
        mfspr   r29,SPR_HID1
        mfspr   r30,SPR_HID4
        mfspr   r31,SPR_HID5
        std     r28,BootHid0(r27)
        std     r29,BootHid1(r27)
        std     r30,BootHid4(r27)
        std     r31,BootHid5(r27)

/*
 * Load SPRG0 with the physical address of this processor's PCR block.
 */

        lis     r3,0
        ori     r3,r3,PCR_ARRAY_PHYSICAL_ADDRESS        // base of array
        mfspr   r4,1023                                 // get PIR
                                                        // NB: Its shifted left by 1
        slwi    r4,r4,DCACHE_BLOCK_SIZE_BITS-1          // times sizeof(PCR)
        add     r3,r3,r4                                // PCR[PIR]
        mtspr   272,r3

/*
 * Initialize the segment registers.
 */

        bl      InitializeSegmentRegisters

/*
 * Initialize the page directory page.
 */

        bl      InitializePageDirectory


/*
 * Initialize the hardware tablewalk structures and registers for TLB misses.
 */

        lis     r3,HIGHWORD(HASHED_PAGE_TABLE_PROCESSOR0_PHYSICAL_ADDRESS)
        mtsdr1  r3                         /* SDR1[HTABSIZE]=0 (256KB) */

/*
 * Flush the translation lookaside buffer.
 */

        li      r3,1
        bl      FlushCurrentEntireTb

/*
 * Copy the trap handlers to the bottom of memory and invalidate 
 * the instruction cache for the modified range.
 */

        lis     r31,Real0SectionStart
        addi    r31,r31,Real0SectionStart
        lis     r30,Real0SectionEnd
        addi    r30,r30,Real0SectionEnd
        sub     r30,r30,r31                  // compute number of BYTEs to copy
        clrlwi  r31,r31,1                    // clear high order address bit
        li      r29,0
        srwi    r28,r30,2                    // compute number of DWORDs to copy
        mtctr   r28

CopyNextTrapHandlerDword:
        lwzx    r28,r31,r29
        stwx    r28,r0,r29
        addi    r29,r29,4                    // advance to next DWORD
        bdnz+   CopyNextTrapHandlerDword

        li      r3,0                         // BaseAddress
        mr      r4,r30                       // Length
        bl      FlushCache

/*
 * Switch to the boot stack
 */

        lis     sp,HIGHWORD(BootStack)
        ori     sp,sp,LOWWORD(BootStack)

        lis     r31,BootProcessorStartup
        addi    r31,r31,BootProcessorStartup

/*
 * Common code for processor startup
 * NB: Do not use r3 below
 */
_FUNCTION(ProceedToNextStageBoot)

/*
 * Terminate the call chain for the debugger.
 */

        li      r0,0
        mtlr    r0
        subi    sp,sp,BOOT_STACK_TOP_FRAME

/*
 * Enable instruction and data relocation, floating-point and vector unit
 * support, trap handling from the bottom of memory, and continue 
 * initialization in the new context
 */

        mtsrr0  r31
        mfmsr   r31
        li      r30,MSR_IP_MASK
        andc    r31,r31,r30
        ori     r31,r31,MSR_IR_MASK + MSR_DR_MASK + MSR_FP_MASK
        oris    r31,r31,HIGHWORD(MSR_VEC_MASK)
        mtsrr1  r31
        sync
        isync
        rfid

END(_start)

/*
 *
 * VOID
 * NonBootProcessorEntry(
 *     VOID
 *     )
 *
 * Routine Description:
 *
 *     This function is the first point of entry for a non-boot processor.
 *     It is invoked by the system reset handler when a processor is reset.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 */

ENTRY(NonBootProcessorEntry)

/*
 * Initialize the hardware implementation dependent registers using the values
 * from the boot processor.
 */

        lis     r27,BootProcessorHIDRegisters
        addi    r27,r27,BootProcessorHIDRegisters
        clrlwi  r27,r27,1                    /* clear high order address bit */
        ld      r28,BootHid0(r27)
        ld      r29,BootHid1(r27)
        ld      r30,BootHid4(r27)
        ld      r31,BootHid5(r27)
        mtspr   SPR_HID0,r28
        sync
        isync
        mtspr   SPR_HID1,r29
        sync
        isync
        mtspr   SPR_HID4,r30
        sync
        isync
        mtspr   SPR_HID5,r31
        sync
        isync

/*
 * The following non-standard SPR appears to control the base physical address
 * for the zeroth level exception handlers.  HIOR.
 */

        li      r0,0
        mtspr   311,r0

/*
 * Load SPRG0 with the physical address of this processor's PCR block.
 */

        lis     r3,0
        ori     r3,r3,PCR_ARRAY_PHYSICAL_ADDRESS        // base of array
        mfspr   r4,1023                                 // get PIR
                                                        // NB: Its shifted left by 1
        slwi    r4,r4,DCACHE_BLOCK_SIZE_BITS-1          // times sizeof(PCR)
        add     r3,r3,r4                                // PCR[PIR]
        mtspr   272,r3

/*
 * Initialize the segment registers.
 */

        bl      InitializeSegmentRegisters

/*
 * Initialize the hardware tablewalk structures and registers for TLB misses.
 */

        lis     r3,HIGHWORD(HASHED_PAGE_TABLE_PROCESSOR1_PHYSICAL_ADDRESS)
        mtsdr1  r3                         /* SDR1[HTABSIZE]=0 (256KB) */

/*
 * Flush the translation lookaside buffer.
 */
        li      r3,1
        bl      FlushCurrentEntireTb

/*
 * Switch to the thread stack allocated on our behalf.
 *
 */

        mfspr   r4,272                         /* real mode PCR address */
        ld      sp,PCR_Gpr5Save(r4)

/*
 * Pick up argument, before we start taking TLB misses
 */
        ld      r3,PCR_Gpr6Save(r4)

/*
 * Enable instruction and data relocation, floating-point and vector unit
 * support, trap handling from the bottom of memory, and continue with
 * initialization of this processor.
 */

        lis     r31,NonBootProcessorStartup
        addi    r31,r31,NonBootProcessorStartup
	    b       ProceedToNextStageBoot

END(NonBootProcessorEntry)

        _ENDCODE

        _END
