/* Implementation of mutex for cb_sched2.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define OPTIMIZE 
/* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base/schedul.h>
#include <schedulif.h>
#include <base/mutex2.h>
#include <base/thread2.h>

/* ======== Exported Interface Functions ======== 
 */

void Mutex_Init(
     PMUTEX pMutex
     )
{
    Mutex_Prologue();
    Mutex_ListNodeInit(pMutexArg);
    Mutex_SetSchedObject(pMutexArg, NULL);
    Mutex_WaiterQueueInit(pMutexArg);
}

void Mutex_Destroy(
     PMUTEX pMutex
     )
{
    Mutex_Prologue();
    Mutex_SetBad(pMutexArg);
    Mutex_WaiterQueueClear(pMutexArg);
}

BOOL Mutex_TryLock(
     PMUTEX pMutex
     )
{
  PTHDINFO pThdCurrent;
    Mutex_Prologue();

    pThdCurrent = (PTHDINFO)CurrentThread();
    /* BUGBUG needs new AtomicCmpAndSwapAddress() */
    if (AtomicCmpAndSwap((PUINT)Mutex_GetSchedObjectAddr(pMutexArg), 0,
			 (UINT)pThdCurrent)) {
/*       Mutex_SetTaken(pMutexArg); */
      if (pThdCurrent != NULL) {
	Thread_MutexListAdd(pThdCurrent, pMutexArg);
      }
      return TRUE;
    }
    return FALSE;
}

void Mutex_Lock(
     PMUTEX pMutex
     )
{
    PTHDINFO pThdCurrent;

    Mutex_Prologue();

    /* Scheduler lock *might* have been taken already, e.g. from
     * within the condition code
     */
    pThdCurrent = (PTHDINFO)CurrentThread();

    /* BUGBUG needs new AtomicCmpAndSwapAddress() */
    if (AtomicCmpAndSwap((PUINT) Mutex_GetSchedObjectAddr(pMutexArg), 0,
			 (UINT) pThdCurrent)) {
      /* mutex is not already taken */
      if (pThdCurrent != NULL) {
	Thread_MutexListAdd(pThdCurrent, pMutexArg);
      }
      return;
    }

    /* Mutex taken, take slow path */
    
    assert(!Mutex_IsBad(pMutexArg));
    
    /* Try to acquire a mutex on behalf of the current thread.
     * Returns S_OK if mutex is taken, S_FALSE if already owned.
     */
    pTheScheduler->v->DisablePreemption(pTheScheduler, TRUE);
    
    assert((UINT)pThdCurrent != (UINT)Mutex_GetSchedObject(pMutexArg));

/*     if (Mutex_GetSchedObject(pMutexArg) == NULL) { */
/*       Mutex_SetSchedObject(pMutexArg, pThdThis); */
/*     } */

    Mutex_SetTaken(pMutexArg);
    pTheScheduler->v->WaitForMutex(pTheScheduler,
				   (PMUTEX)pMutexArg);
    
    pTheScheduler->v->EnablePreemption(pTheScheduler, TRUE);

}

void Mutex_Unlock(
     PMUTEX pMutex
     )
{
    PTHDINFO pThdCurrent = (PTHDINFO)CurrentThread();

    Mutex_Prologue();
    /* Scheduler lock *might* have been taken already, e.g. from
     * within the condition code
     */

    assert(pMutexArg != NULL);
    assert(Mutex_GetSchedObject(pMutexArg) == (PTR)pThdCurrent);

    if (pThdCurrent != NULL) {
      Thread_MutexListRemove(pThdCurrent, pMutexArg);
    }
      
    /* Is anybody waiting */
    if (AtomicCmpAndSwap((PUINT) Mutex_GetSchedObjectAddr(pMutex),
                         (UINT)pThdCurrent, 0)) {
      return;
    }

    assert(!Mutex_WaiterQueueIsEmpty(pMutexArg));

/*     Mutex_ClearTaken(pMutexArg); */

/*     if (Mutex_WaiterQueueIsEmpty(pMutexArg)) { */
/*       Mutex_SetSchedObject(pMutexArg, NULL); */
/*       return; */
/*     } */

    /* Yes, take slower path */

    pTheScheduler->v->DisablePreemption(pTheScheduler,TRUE);

    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

/*      DBGOUT(DBGMUTEX,MXLVL,printf( */
/*             "UntakeMutex MX %08x MX.Owner %08x Self %08x\n", */
/*             pMutex, pThdOwner, CurrentThread())); */

    /* MP BUGBUG On a multiprocessor, a spinlock protecting
     * the thread queue is needed. This spinlock should reside
     * in the owner thread object.
     */

    /* Recheck there are infact waiters
     */

    pTheScheduler->v->WakeUpOnMutex(pTheScheduler, (PMUTEX)pMutexArg);

    pTheScheduler->v->EnablePreemption(pTheScheduler,TRUE);

}

/* Return TRUE if the current thread owns the mutex.
 */
BOOL Mutex_Locked(
     const MUTEX *pMutex
     )
{
    Mutex_Prologue();
    return ((PTR)CurrentThread() == Mutex_GetSchedObject(pMutexArg));
}


