/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <base/schedul.h>
#include <schedulif.h> /* for IPic */
#include <fred.h>

#ifdef _MSC_VER
#pragma warning(disable:4055)
#endif

SCODE AddDevice(PTR d, PTR i, UINT Flags, UINT Irq, PTR *pOldObject)
{
    struct IDevice *Device = (struct IDevice *)d;
    INTERRUPT_SERVICE_ROUTINE Isr = (INTERRUPT_SERVICE_ROUTINE)i;

    if (Flags & DEVICE_FLAGS_TRAP) {
#if defined (i386) || defined(arm)
        InstallIsr(Irq, (ADDRESS) i,
                   !!(Flags & DEVICE_FLAGS_RAW_NUMBER),
                   !!(Flags & DEVICE_FLAGS_USER_ACCESS));
#else
        return E_NOT_IMPLEMENTED;
#endif
    } else {
        /* Start by disabling the interrupt.
         * Save the old interrupt if requested.
         */
        PTR Old;
        pThePic->v->DisableInterrupt( pThePic, Irq,
                                      pOldObject ? pOldObject : &Old);

        /* NULL ISR means disable: we are done. Otherwise enable new one. */
        if (Isr != NULL)
            pThePic->v->EnableInterrupt(pThePic, Flags, Irq, Isr, Device);
    }
    return S_OK;
}


/* Get current absolute time
 */
#if defined(CurrentTime)
#undef CurrentTime
#endif
TIME CurrentTime(void)
{
    return GetKernelTime();
}

/* Generic unimplemented methods */
SCODE UNIMPLEMENTEDMethodSCODE(void)
{
    return E_NOT_IMPLEMENTED;
}

ADDRESS UNIMPLEMENTEDMethodZero(void)
{
    return 0;
}

/*
 * If something is not implemented and gets used..
 */
void UnimplementedRtlCall(UINT CalledFrom)
{
    UnusedParameter(CalledFrom);
#ifdef _DEBUG
    printf("Bad RTL call, called from x%x\n", CalledFrom);
#endif
    DebugBreak();
}
