/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* This file contains FirstThread */

#include <mmlite.h>
#include "fred.h"
#include <base/cons.h>
#include <tchar.h>
#include <base/loader.h>

#ifdef _MSC_VER
#pragma warning(disable:4100)
#endif

SCODE MCT PeGetFunctionAddress (
        PTR ExportData,
        UINT ExportSize,
        const _TCHAR * ModName,
        UINT Ordinal,
        /* [out] */ ADDRESS *pAddress)
{
    return E_NOT_IMPLEMENTED;
}

SCODE LoadImage(
           PINAMESPACE pNameSpace,
           const _TCHAR * pImage,
           const _TCHAR * pArgs,
           UINT Flags,
           IModule **ppIModule)
{
    return E_NOT_IMPLEMENTED;
}

PINAMESPACE GetImageNameSpace(void)
{
    return NULL;
}

PTR RtlTable[] = {0};
#define CLOCK_APP 1

#ifdef CLOCK_APP

#ifdef __NO_BUILTIN_INT64
const TIME TIME_STEP = Int64Initializer(0,TIME_SECONDS(10));
#else
#define TIME_STEP TIME_SECONDS(10)
#endif

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    CONDITION c;
    MUTEX m;
    TIME t;
    int i;

    UnusedParameter(arg);

    Mutex_Init(&m);
    Condition_Init(&c);

    t = CurrentTime();
    /* round time up to a second */
    //t = (t + TIME_SECONDS(2)) % TIME_MILLIS(1000);
    //Mutex_Lock(&m);
    for (i = 1; i < 10; i++) {
        //Condition_Wait(&c, &m, &t);
        putchar('.');           /* output a dot every second */
        SleepUntil(t);
        t = TimeAdd(TIME_STEP, t);
    }
    //Mutex_Unlock(&m);
    BaseDelete();
}

#else /* CLOCK_APP */

/* Which test(s) should we perform
 */
#define TEST_COND 1
#define TEST4

/* Temporary hack to print times on machines lacking 64 bit support */
#ifdef i386
#define TIME_LOWPART(t) ((UINT) (t))
#else
#define TIME_LOWPART(t) ((t).LowPart)
#endif

/* Print names for tests
 */
#if defined(TEST1)
#define P_TEST1 "  TimeShare #1\n"
#else
#define P_TEST1 ""
#endif
#if defined(TEST2)
#define P_TEST2 "  TimeShare #2\n"
#else
#define P_TEST2 ""
#endif
#if defined(TEST3)
#define P_TEST3 "  TimeShare #3\n"
#else
#define P_TEST3 ""
#endif
#if defined(TEST4)
#define P_TEST4 "  SingleThreaded #4\n"
#else
#define P_TEST4 ""
#endif
#if defined(TEST_MUTEX)
#define P_TEST_MUTEX "  Mutex objects\n"
#else
#define P_TEST_MUTEX ""
#endif
#if defined(TEST_COND)
#define P_TEST_COND "  Condition variables\n"
#else
#define P_TEST_COND ""
#endif
#if defined(TEST_ATOMIC)
#define P_TEST_ATOMIC "  AtomicXXX functions\n"
#else
#define P_TEST_ATOMIC ""
#endif
#if defined(TIME_OVERHEADS)
#define P_TIME_OVERHEADS "  Basic overheads timings\n"
#else
#define P_TIME_OVERHEADS ""
#endif
#if defined(TEST_SLEEP)
#define P_TEST_SLEEP "  SleepUntil\n"
#else
#define P_TEST_SLEEP ""
#endif
#if defined(TEST_CONSTR)
#define P_TEST_CONSTR "  Time constraints\n"
#else
#define P_TEST_CONSTR ""
#endif

CONDITION fooc, bazc;
MUTEX foom;

int nn = 0;
UINT x1 = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      printf("Will perform the following tests:\n" \
             P_TIME_OVERHEADS P_TEST_ATOMIC P_TEST_SLEEP \
             P_TEST_COND P_TEST_CONSTR P_TEST_MUTEX \
             P_TEST1 P_TEST2 P_TEST3 P_TEST4
             );

    if (me == 0) {

#if defined(TIME_OVERHEADS)
        TIME t0, t1, tn, t;

        Int32ToInt64(tn,10000000);
        for (i = 0; i < 10; i++) {
            t0 = CurrentTime();
            t1 = CurrentTime();
            t1 = TimeSubtract(t1,t0);
            if (TimeLess(t1,tn)) {
                tn = t1;
            }
            /*printf("CurrentTime() costs %d.%x (%d-th loop)\n",
                       t1, i);*/
        }
        printf("Minimum CurrentTime() cost is %d\n", TIME_LOWPART(tn));

        Int32ToInt64(t,TIME_RELATIVE(10000)); /* 1milli */
        Int32ToInt64(tn,100000); /* 10milli */
        for (i = 0; i < 10; i++) {
            t0 = CurrentTime();
            SleepUntil(t);
            t1 = CurrentTime();
            t1 = TimeSubtract(t1,t0);
            t1 = TimeAdd(t1,t);
            if (TimeLess(t1,tn)) {
                /*printf("SL %d.%x\n", t1);*/
                tn = t1;
            }
        }
        printf("Minimum SleepUntil() cost is %d\n", TIME_LOWPART(tn));
#endif

#if defined(TEST_COND)
        Condition_Init(&fooc);
        Condition_Init(&bazc);
#endif
#if defined(TEST_MUTEX)
        Mutex_Init(&foom);
#endif
#if !defined(TEST4)
        printf("Initial thread (x%x) started.\n", CurrentThread());
    {
        PIPROCESS pPrc = CurrentProcess();
        (void) pPrc->lpVtbl->CreateThread(pPrc,
                                          FirstThread,
                                          NULL, 0, NULL, NULL);
printf("created second thread\n");
        (void) pPrc->lpVtbl->CreateThread(pPrc,
                                          FirstThread,
                                          NULL, 0, NULL, NULL);
printf("created third thread\n");
    }
#endif
    }

#if defined(TEST2)
    Delay(3000);
#endif

    printf("Thread %d is x%x\n", me, CurrentThread());

#if defined(TEST_ATOMIC)
    for (i = 0; i < 1000; i++) {
      AtomicInc(&x1);
      AtomicAdd(&x1,20);
      AtomicDec(&x1);
      AtomicInc(&x1);
      AtomicSub(&x1,10);
      AtomicDec(&x1);
      AtomicSub(&x1,10);
    }
    printf("%d found x1 = %d\n", me, x1);
#endif

#if defined(TEST_SLEEP)
    /* This tests the IdleThread logic too, since all threads sleep
     */
    {
        TIME t, te;
        for (i = 0; i < 10; i++) {
            Int32ToInt64(t,TIME_RELATIVE(10000)); /* 1milli */
            te = CurrentTime();
            SleepUntil(t);
            t = CurrentTime();
            printf("%d slept from %d to %d.\n",
                   me, TIME_LOWPART(te), TIME_LOWPART(t));
        }
    }
#endif

#if defined(TEST_COND)
    if (me == 0) {
        TIME t, t1;
        for (i = 0; i < 3; i++) {
            Int32ToInt64(t,TIME_RELATIVE(20000)); /* 2milli */
            printf("%d Sleeping for %d\n", me, TIME_LOWPART(t));
            t1 = CurrentTime();
printf("A\n");
            SleepUntil(t);
printf("B\n");
            t = CurrentTime();
            printf("%d slept from %d to %d.\n",
                   me, TIME_LOWPART(t1), TIME_LOWPART(t));
            if (i == 2) {
                printf("%d signalled all other threads.\n", me);
                Condition_Broadcast(&fooc);
            } else {
                Condition_Signal(&fooc);
                printf("%d signalled 1 other thread.\n", me);
            }
        }
    } else {
        for (i = 0; i < 2; i++) {
            printf("%d into Wait\n", me);
            Condition_Wait(&fooc, NULL, NULL);
            printf("%d back from Wait\n", me);
        }
    }
printf("C\n");

    Delay(10000);

    {
      TIME t, t0, t1;
      Int32ToInt64(t,TIME_RELATIVE(1000*10));
      printf("%d into Wait for %d\n", me, TIME_LOWPART(t));
      t0 = CurrentTime();
      Condition_Wait(&bazc, NULL, &t);
      t1 = CurrentTime();
      printf("%d was in Wait from %d to %d\n",
             me, TIME_LOWPART(t0), TIME_LOWPART(t1));
      if (t1 - t0 < t)
          printf("Too short!\n");
    }

#endif

#if defined(TEST_CONSTR)
    {
        TIME_CONSTRAINT MyCos;
        TIME t0, t1;
        SCODE sc0, sc1;
        MyCos.Criticality = /*NON*/CRITICAL;
        /* start time of now plus "me" millisecs */
        Int32ToInt64(MyCos.Start,TIME_RELATIVE(me*1000*10));
        /* run time of 1msec */
        Int32ToInt64(MyCos.Estimate,1000*10);
        /* deadline me+1 millisecs */
        Int32ToInt64(MyCos.Deadline,TIME_RELATIVE((me+1)*1000*10));

        sc0 = BeginConstraint( FALSE, &MyCos, NULL);
        t0 = CurrentTime();
        printf("%d is inside a constraint at %d\n",
               me, TIME_LOWPART(t0));
        sc1 = EndConstraint(&t1);

        printf("%d time results are %x %x %d\n",
               me, sc0, sc1, TIME_LOWPART(t1));
        if (sc0 != S_OK || sc1 != S_OK)
            printf("    CONSTRAINT TEST FAILED !!\n");
        Delay(100000);
    }
#endif

#if defined(TEST1)
#define NLOOPS 12
#elif defined(TEST4)
#define NLOOPS 120*3
#else
#define NLOOPS 120
#endif
    for (i = 0;i < NLOOPS; i++) {
#if !defined(TEST3)
        TIME t;
#if defined(TEST_MUTEX)
        Mutex_Lock(&foom);
#endif
        t = CurrentTime();
        printf("%d at %d\n", me, TIME_LOWPART(t));
#if defined(TEST_MUTEX)
        Mutex_Unlock(&foom);
#endif
#if defined(TEST1)
        Delay(3000/*000*/);
#endif
#if defined(TEST2) || defined(TEST4)
        Delay(3);
#endif
#else /*TEST3*/
      {
          volatile int *p = (volatile int *)0x20000;
          volatile int *e = p + (64*1024 / 4);

          while (p < e) {
              i = *p;
              p += 8;
          }
      }
#endif
    }

    /* terminate normally */

    BaseDelete();
}
#endif /* CLOCK_APP */
