/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implementation of urgency comparators for kernel time constraints. */

#include <assert.h>
#include <base/urgency2.h>

/* Returns TRUE if both constraint urgencies are equal, otherwise FALSE. */
BOOL Urgency_EQ(
     PURGENCY pu1,
     PURGENCY pu2
     )
{
    TIME tRestartBy1 = Urgency_GetRestartBy(pu1);
    TIME tRestartBy2 = Urgency_GetRestartBy(pu2);
    assert(!TimeLess(tRestartBy1, TIME_ORIGIN) );
    assert(!TimeLess(tRestartBy2, TIME_ORIGIN) );

    if (Urgency_GetCriticality(pu1) != Urgency_GetCriticality(pu2)) {
      return FALSE;
    }

    return TimeEqual(tRestartBy1, tRestartBy2);
}

/* Returns TRUE if constraint pkc1 is more urgent than pkc2 */
BOOL Urgency_GT(
     PURGENCY pu1,
     PURGENCY pu2
     )
{
    TIME tRestartBy1 = Urgency_GetRestartBy(pu1);
    TIME tRestartBy2 = Urgency_GetRestartBy(pu2);

    assert(!TimeLess(tRestartBy1, TIME_ORIGIN) );
    assert(!TimeLess(tRestartBy2, TIME_ORIGIN) );

    /* time-share implies noncritical */
/*     if (UrgencyGetCriticality(pu1) == UrgencyGetCriticality(pu2)) { */
      return TimeLess(tRestartBy1, tRestartBy2);
/*     } */

/*     return UrgencyIsTimeSharing(pu2); */
}

/* Return true if constraint pkc1 is no less urgent than pkc2. */
BOOL Urgency_GTEQ(
     PURGENCY pu1,
     PURGENCY pu2
     )
{
    TIME tRestartBy1 = Urgency_GetRestartBy(pu1);
    TIME tRestartBy2 = Urgency_GetRestartBy(pu2);

    assert(!TimeLess(tRestartBy1, TIME_ORIGIN) );
    assert(!TimeLess(tRestartBy2, TIME_ORIGIN) );

    /* time-share implies noncritical */
/*     if (UrgencyGetCriticality(pu1) == UrgencyGetCriticality(pu2)) { */
      return !TimeLess(tRestartBy2, tRestartBy1);
/*     } */

    /* If they are not the same criticality they cant be EQ,
     * ==> same as above */
/*     return UrgencyIsTimeSharing(pu2); */
}
