/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <string.h>
#include <stdlib.h>

/* Get an object of type Iid from a string.
 */
SCODE __cdecl BindToObject(PINAMESPACE pINameSpace,
       const _TCHAR *FileName, UINT Flags, REFIID Iid, void **ppObject)
{
    SCODE sc;
    PIUNKNOWN pUnk;

    /* XXX Allow sloppy usage */
    *ppObject = NULL;
    if (pINameSpace == NULL)
        return E_INVALID_PARAMETER;

    sc = pINameSpace->v->Bind(pINameSpace, FileName, Flags, &pUnk);

    if (SUCCEEDED(sc)) {
        sc = pUnk->v->QueryInterface(pUnk, Iid, ppObject);
        pUnk->v->Release(pUnk);

        /* If we created an object of the wrong type it is time to undo. */
        if (FAILED(sc) && (Flags & NAME_SPACE_CREATE))
            pINameSpace->v->Unregister(pINameSpace, FileName);
    }

    return sc;
}

#ifdef _UNICODE
/* ASCII version of BindToObject */
SCODE __cdecl BindToObjectA(PINAMESPACE pINameSpace,
                            const char *FileName, UINT Flags, REFIID Iid,
                            void **ppObject)
{
    SCODE sc;
    wchar_t *UFileName = _strtow(FileName);

    if (UFileName == NULL) {
        *ppObject = NULL;
        return E_NOT_ENOUGH_MEMORY;
    }

    sc = BindToObject(pINameSpace, UFileName, Flags, Iid, ppObject);

    free(UFileName);
    return sc;
}
#endif
