/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>

/* C++ Memory allocation
 */
void * operator new( size_t size)
{
    PIHEAP pHeap = CurrentHeap();
    return pHeap->Alloc(HEAP_DEBUG_NAME("operator_new") | HEAP_ZERO_MEMORY,
                        size, 0);
}

/* C++ Memory release
 */
void operator delete( void *pThis)
{
    PIHEAP pHeap = CurrentHeap();
    (void) pHeap->Free(0, pThis);
}

#if defined(ADS) /* arm */
void *operator new[](size_t size)
{
    PIHEAP pHeap = CurrentHeap();
    return pHeap->Alloc(HEAP_DEBUG_NAME("operator_new[]") | HEAP_ZERO_MEMORY,
                        size, 0);
}

void __vecmap1c(void* arg1, void* arg2, int arg3, void (*arg4)(void*))
{
    DebugBreak();
}

void __vecmap1ci(void* arg1, void* arg2, int arg3, void (*arg4)(void*, int))
{
    DebugBreak();
}

/* Vector constructor
 * Allocate memory. Iterate per-element, calling CTOR.
 * If one of them fails, and DTOR is non-null, loop
 * calling DTOR on each succeffully CTOR-ed element.
 */
/* XXX Inconsistent use of last two arguments */
void * __vec_ctor (unsigned int nElements, unsigned int nElementSize, bool arg3, void (*ctor)(void*), void* (*dtor)(unsigned int))
{
    void *pThis, *v;
    BYTE *iter;
    unsigned int i;
    PIHEAP pHeap = CurrentHeap();
    i = (nElements*nElementSize) + sizeof(int);
    pThis = pHeap->Alloc(HEAP_DEBUG_NAME("vec_ctor") | HEAP_ZERO_MEMORY, i, 0);
    if (pThis) {
        /* Remember array size */
        *(int *)pThis = nElements;
        pThis = (void*) ((BYTE*)pThis + sizeof(int));
        iter = (BYTE*)pThis;
        for (i = 0; i < nElements; i++) {
            v = (ctor(iter), (void*)TRUE); /* XXX */
            if (v == NULL)
                goto Fail;
            iter += nElementSize;
        }
    }
    return pThis;
 Fail:
    if (dtor) {
        nElements = i;
        iter = (BYTE*)pThis;
        for (i = 0; i < nElements; i++) {
            dtor((UINT)iter);   /* XXX */
            iter += nElementSize;
        }
    }
    (void) pHeap->Free(0, pThis);
    return NULL;
}

/* XXX Inconsistent use of last argument */
void __vec_dtor (void *pThis, unsigned int nElementSize, void (*dtor)(void*pThis,int FreeThis), void (*FreeEachElement)(void*))
{
    PIHEAP pHeap;
    BYTE *iter;
    unsigned int i, nElements;

    if (pThis == NULL)
        return;

    iter = (BYTE*)pThis;
    /* retrieve array size */
    nElements = ((int*)pThis)[-1];
    for (i = 0; i < nElements; i++) {
        dtor(iter,(UINT)FreeEachElement); /* XXX */
        iter += nElementSize;
    }

    pHeap = CurrentHeap();
    (void) pHeap->Free(0, pThis);
}
#endif
