/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>

EXTERN_C INT MODENTRY CrtInit(INT Op);

/* Default COB Object */
typedef struct _DEFCOB {
    const struct IUnknownVtbl *v;
    UINT RefCnt;
} *PDEFCOB;

/* IUnknown::<this> points to v-table
 * pDC (pointer of Default Cob) converts interface ptr to implementation.
 * iDC (interface of DC)  converts implementation ptr to interface ptr.
 */
#define pDefCob(_i_) ((PDEFCOB)(_i_))
#define iDefCob(_p_) ((PIUNKNOWN)(_p_))

/* IUnknown::QI.  Return a pointer of the right kind. */
static SCODE MCT DefCobQueryInterface(PIUNKNOWN pThis, REFIID Iid,
                                       void **ppObj)
{
    return GenericQueryInterface(pThis, Iid, ppObj, &IID_IUnknown);
}

/* IUnknown::AddRef */
static UINT MCT DefCobAddRef(PIUNKNOWN pThis)
{
    PDEFCOB x = pDefCob(pThis);

    return AtomicInc(&x->RefCnt);
}

/* IUnknown::Release */
static UINT MCT DefCobRelease(PIUNKNOWN pThis)
{
    PDEFCOB x = pDefCob(pThis);
    UINT RefCnt = AtomicDec(&x->RefCnt);

    if (RefCnt == 0) {
        /* Nothing special to finalize or deallocate in this implementation
         * but give C runtime a chance to run global destructors etc.
         */
#ifndef h8
        CrtInit(DLL_PROCESS_DETACH);
#endif
    }
    return RefCnt;
}

static const struct IUnknownVtbl vtab = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    DefCobQueryInterface,
    DefCobAddRef,
    DefCobRelease
};

/* Global can be used by apps (or defcobmain.c) */
struct _DEFCOB _DefaultCobObject = { &vtab, 1 };
