/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "_stdio.h"
#include <io.h>

/* returns 0 if success, EOF if failure */
INT __cdecl fflush(FILE *stream)
{
    /* we dont do this */
    if (stream == NULL)
        return EOF;

    /* If non writeable its undefined */
    if ((stream->_flag & _IOWRT) == 0)
        return EOF;

    /* if in error */
    if (stream->_flag & (_IOERR | _IOEOF))
        return EOF;

    if (stream->_cnt > 0) {
#ifdef __USE_IFILE
        UINT cnt;
        SCODE sc = ReadWriteFile(stream->_file, TRUE, stream->Seek,
                                 (void*) stream->_base, stream->_cnt, &cnt);
        if (FAILED(sc)) {
            stream->_flag |= _IOERR;
            return EOF;
        }
        stream->Seek = Uint64AddUint32(stream->Seek, cnt);
#else
        int i;
        i = write((FILE_HANDLE)stream->_file,
                  stream->_base,
                  stream->_cnt);
        if (i < 0) {
            stream->_flag |= _IOERR;
            return EOF;
        }
#endif
        stream->_cnt = 0;
        stream->_ptr = stream->_base;
    }

    return 0;
}
