/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <string.h>
#include <tchar.h>
#include "_stdio.h"
#include <stdlib.h>

#ifdef __USE_IFILE

FILE* __cdecl _tfopen(const _TCHAR *filename, const _TCHAR *mode)
{
    FILE *f;
    UINT Uflags;
    NAME_SPACE_FLAGS Flags;
    SCODE sc;
    BOOL Trunc = FALSE;

    f = (FILE *) malloc( sizeof *f);
    if (f == NULL)
        return NULL;

    memset( f, 0, sizeof *f);

    Flags = Uflags = 0;
    while (*mode != _EOS)
        switch (*mode++) {
        case 'r' : Flags |= NAME_SPACE_READ;
            Uflags |= _IOREAD;
            break;
        case 'w' : Flags = NAME_SPACE_WRITE | NAME_SPACE_CREATE;
                   Flags &= ~NAME_SPACE_READ;
                   Trunc = TRUE;
            Uflags |= _IOWRT;
            break;
        case 'a' : Flags |= NAME_SPACE_APPEND;
            Uflags |= _IOAPPEND;
            break;
//        case 'b' : Flags |= O_BINARY;
//            break;
        case '+' : Flags |= NAME_SPACE_READ | NAME_SPACE_WRITE;
            Uflags |= _IOWRT;
            Trunc = FALSE;
            break;
        }

    sc = BindToObject(CurrentNameSpace(), filename, Flags, &IID_IFile,
                      &f->_file);

    if (FAILED(sc)) {
        free(f);
        return NULL;
    }

    if (Trunc) {
        PIFILE File = f->_file;
        sc = File->v->SetSize(File, Int64ToUint64(INT64_ZERO));
    }

    if (Uflags & _IOAPPEND) {
        PIFILE File = f->_file;
        sc = File->v->GetSize(File, &f->Seek);
    }

    f->_flag = Uflags;

    return f;
}
#else
#include <io.h>

FILE* __cdecl _tfopen(const _TCHAR *filename, const _TCHAR *mode)
{
    FILE *f;
    UINT Flags, Uflags;
    int fileno;

    f = (FILE *) malloc( sizeof *f);
    if (f == NULL)
        return NULL;

    memset( f, 0, sizeof *f);

    Flags = Uflags = 0;
    while (*mode != _EOS)
        switch (*mode++) {
        case 'r' : Flags = (Flags & ~3) | O_RDONLY;
            Uflags |= _IOREAD;
            break;
        case 'w' : Flags = (Flags & ~3) | O_WRONLY | O_CREAT | O_TRUNC;
            Uflags |= _IOWRT;
            break;
        case 'a' : Flags |= O_APPEND;
            Uflags |= _IOAPPEND;
            break;
        case 'b' : Flags |= O_BINARY;
            break;
        case '+' : Flags = (Flags & ~3) | O_RDWR;
            Uflags |= _IOWRT;
            break;
        }

    fileno = _topen(filename, Flags, 0666);
    if (fileno <= 0) {
        free(f);
        return NULL;
    }

    if (Uflags & _IOAPPEND) {
        INT pos = filelength(fileno);
        lseek(fileno,pos,SEEK_SET);
    }

    f->_file = (void *) fileno;
    f->_flag = Uflags;

    return f;
}
#endif
