/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "_stdio.h"
#include <io.h>

size_t __cdecl fread(void *ptr,
                    size_t size,
                    size_t nmemb,
                    FILE *stream)
{
    size_t      n, tot;
    int         i;
    int         c;
    char *      buf = (char *)ptr;

    /* Mumble, asking for zero bytes should be ok, sez ANSI
     */
    if (size == 0 || nmemb == 0)
        return 0;

    /* if in error */
    if (stream->_flag & (_IOERR | _IOEOF))
        return 0;

    /* unbuffered case, simple (xxx: we dont really do LBF)*/
    if (stream->_flag & (_IONBF|_IOLBF)) {
        n = 0;

        while(n < nmemb) {
#ifdef __USE_IFILE
            UINT cnt;
            SCODE sc = ReadWriteFile(stream->_file, FALSE, stream->Seek,
                                     (void*) buf, 1, &cnt);
            if (FAILED(sc) || sc == S_FALSE) {
                stream->_flag |= (sc != S_FALSE) ? _IOERR : _IOEOF;
                return n;
            }
            stream->Seek = Uint64AddUint32(stream->Seek, cnt);
            i = (int) cnt;
#else
            i = read((FILE_HANDLE) stream->_file, buf, 1);
            if ((size_t) i != size) {
                stream->_flag |= (i < 0) ? _IOERR : _IOEOF;
                return n;
            }
#endif
            c = *buf;

            switch(c) {
            case '\n':
                putc('\n', stdout);
                return n + 1;
                break;
            case '\r':          /* Text mode: discard returns */
                break;
            case '\b':          /* Cooked mode: crude line editing */
                /* XXX This probably belongs in shell or somewhere else */
                if ( buf != ptr ) {
                    *buf = '\0'; /* Replace the backspace with a NULL. */
                    buf -= size; /* Adjust the buffer pointers. */
                    n--;
                    putc('\b', stdout); /* Display the backspace */
                    putc(' ', stdout);
                    putc('\b', stdout);
                }
                break;

            default:
                putc(c, stdout);
                buf += size;
                n++;
                continue;
            }
        }

        return n;
    }

    /* buffered */
    for (n = 0; n < nmemb; n++) {
        tot = size;
        /* BUGBUG multithread unsafe */
        i = stream->_cnt;
        while (tot > (size_t)i) {
            /* the stream's buffer is smaller than SIZE */
            memcpy(buf, stream->_ptr, i);
            stream->_ptr += i;
            stream->_cnt -= i;
            buf += i;
            tot -= i;
            if (frefill(stream))
                return n;
            i = stream->_cnt;
        }
        if (tot) {
            memcpy(buf,stream->_ptr,tot);
            stream->_ptr += tot;
            stream->_cnt -= tot;
            buf += tot;
        }
    }

    return n;
}
