/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <mmlite.h>
#include "_stdio.h"

/* Refill a buffer, create one if needed
 * Returns 0 or EOF.
 */
INT __cdecl frefill( FILE *stream)
{
    /* must have inited if we get here,
     * else user stupid -> duly punished
     */

    stream->_cnt = 0;   /* drop whatever was left, if any */

    if (stream->_flag & _IOEOF)
        return EOF;

    /* NOTE: issues about read/write files and buffering
     * are totally ignored here. BUGBUG.
     */


    /* Allocate a buffer if we dont have one
     */
    if (stream->_base == NULL) {
        PTR buf;

        if (stream->_bufsize == 0)
            stream->_bufsize = BUFSIZ;
        buf = malloc(stream->_bufsize);
        if (buf == NULL)
            goto GotErr;

        /* Someone else raced us to it ? */
        if (!AtomicCmpAndSwap( (PUINT) &stream->_base, (UINT) NULL, (UINT) buf))
            free(buf);
    }

    stream->_ptr = stream->_base;

    {
#ifdef __USE_IFILE
        UINT cnt;
        SCODE sc = ReadWriteFile(stream->_file, FALSE, stream->Seek,
                                 (void*)stream->_base, stream->_bufsize,
                                 &cnt);
        if (FAILED(sc))
            goto GotErr;

        stream->Seek = Uint64AddUint32(stream->Seek, cnt);
#else
        FILE_HANDLE fileno = (FILE_HANDLE) stream->_file;
        int cnt = stream->_bufsize;

        cnt = read(fileno, (BYTE*) stream->_base, cnt);

        if (cnt < 0)
            goto GotErr;
#endif

        stream->_cnt = cnt;

        if (cnt == 0) {
            stream->_flag |= _IOEOF;
            return EOF;
        }
    }

    return 0;

 GotErr:
    stream->_flag |= _IOERR | _IOEOF;
    return EOF;
}

