/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "_stdio.h"
#include <io.h>
#include <mmlite.h>

/* returns number of elements written.
 * A return < NMEMB is error.
 */
size_t __cdecl fwrite(const void *ptr,
                     size_t size,
                     size_t nmemb,
                     FILE *stream)
{
    size_t n, tot;
    int i;
    char *buf;

    /* If non writeable */
    if ((stream->_flag & _IOWRT) == 0)
        return 0;

    /* if in error */
    if (stream->_flag & (_IOERR | _IOEOF))
        return 0;

    buf = (char *) ptr;

    /* unbuffered case, simple (xxx: we dont really do LBF)*/
    if (stream->_flag & (_IONBF|_IOLBF)) {
        for (n = 0; n < nmemb; n++) {
#ifdef __USE_IFILE
            UINT cnt;
            SCODE sc = ReadWriteFile(stream->_file, TRUE, stream->Seek,
                                     (void*) buf, size, &cnt);
            if (FAILED(sc) || sc == S_FALSE) {
                stream->_flag |= (sc != S_FALSE) ? _IOERR : _IOEOF;
                return n;
            }
            stream->Seek = Uint64AddUint32(stream->Seek, cnt);
#else
            i = write((FILE_HANDLE)stream->_file,buf,size);
            if ((size_t)i != size) {
                stream->_flag |= (i < 0) ? _IOERR : _IOEOF;
                return n;
            }
#endif
            buf += size;
        }
        return n;
    }

    /* buffered, fill up to one buf at a time. flush as needed.
     */
    if (stream->_base == 0) {

        if (stream->_bufsize == 0)
            stream->_bufsize = BUFSIZ;
        ptr = malloc(stream->_bufsize);
        if (ptr == NULL)
            goto GotErr;

        /* Someone else raced us to it ? */
        if (!AtomicCmpAndSwap( (PUINT) &stream->_base, (UINT) NULL, (UINT)ptr))
            free((void*)ptr);

        stream->_ptr = stream->_base;
    }

    for (n = 0; n < nmemb; n++) {
        for (tot = 0; tot < size;) {
            i = stream->_bufsize - stream->_cnt;
            if (i <= 0) {
                fflush(stream);
                i = stream->_bufsize;
            }
            i = (i > (int)(size - tot)) ? (size - tot) : i;
            memcpy(stream->_ptr,buf,i);
            stream->_cnt += i;
            stream->_ptr += i;
            buf += i;
            tot += i;
        }
    }

    return n;

 GotErr:
    stream->_flag |= _IOERR | _IOEOF;
    return 0;
}


