/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <crypto.h>
#include <diagnostics.h>

PICYPHER
GetCypher(
          _TCHAR *CypherCobName,
          UINT KeyStrength
          )
{
    SCODE sc;
    PICYPHER Cypher;
    PICYPHERCONSTRUCTOR Factory;
    PIUNKNOWN pUnk;

    sc = BindToObject(CurrentNameSpace(), CypherCobName,
                      NAME_SPACE_READ, &IID_ICypherConstructor,
                      (void *) &Factory);
    if (FAILED(sc)) {
        DBG(("Could not find the %s factory sc=%x\n\r", CypherCobName, sc));
        return NULL;
    }

    sc = Factory->v->Constructor(Factory,&pUnk,CypherCobName,KeyStrength);
    (void) Factory->v->Release(Factory);

    if (FAILED(sc)) {
        DBG(("Constructor failed kl=%d sc=%x\n\r", KeyStrength, sc));
        return NULL;
    }

    sc = pUnk->v->QueryInterface(pUnk,&IID_ICypher,(void**)&Cypher);
    pUnk->v->Release(pUnk);

    if (FAILED(sc)) {
        DBG(("%s is not a cypher sc=%x\n\r", CypherCobName, sc));
        return NULL;
    }

    return Cypher;
}
