/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Portable 64-bit integer math functions for MMLite
 */
#define __NO_BUILTIN_INT64 1
#if defined(MMLITE)
#include <mmlite.h>
#include <machdep.h>
#else
#define INT64 MYINT64
#define UINT64 MYUINT64
#include <windows.h>
#undef INT64
#undef UINT64
typedef struct _INT64 INT64;
typedef struct _UINT64 UINT64;
#define EXTERN_C extern
#include "int64.h"
#endif

/* Cannot (and should not) compile this if there is compiler support.
 * Dont think you can have it both ways because of alignment issues.
 * Someday the unions might get smarter.
 */
/* Well then, we'll give it another shot.  If there is compiler support
 * we'll just make sure the compiler support routines work.  So the code
 * in there is only used by the test programs e.g. tstdiv64.c & friends.
 */

/* Nullify macros for functions we define. If we get compiled then
 * it means that those macros are needed as functions.
 */
#undef Int64Add
#undef Int64AddInt32
#undef Int64Subtract
#undef Int64SubtractInt32

#if !defined(__TCS__)
#define carry(a,b) (((UINT32)(a) + (UINT32)(b)) < (a))
#define borrow(a,b) ((a) < (b))
#endif


/* MMLite runtime library function -- Add two signed,
 * 64-bit integers and return the signed, 64-bit sum.
 */
INT64 Int64Add(INT64 x, INT64 y)
{
    x.HighPart += y.HighPart + carry(x.LowPart, y.LowPart);
    x.LowPart += y.LowPart;
    return (x);
}

/* MMLite runtime library function -- Add one signed,
 * 64-bit integer and one signed 32-bit integer and
 * return the signed, 64-bit sum.
 */
INT64 Int64AddInt32(INT64 x, INT32 y)
{
    INT64 z;

    Int32ToInt64(z,y);
    return Int64Add(x,z);
}

/* MMLite runtime library function -- Add one unsigned,
 * 64-bit integer and one unsigned 32-bit integer and
 * return the unsigned, 64-bit sum.
 */
UINT64 Uint64AddUint32(UINT64 x, UINT32 y)
{
    x.HighPart += carry(x.LowPart, y);
    x.LowPart += y;
    return (x);
}

/* MMLite runtime library function -- Subtract a signed,
 * 64-bit integer (2nd arg) from a signed, 64-bit integer
 * (first arg) and return the signed, 64-bit difference.
 */
INT64 Int64Subtract(INT64 x, INT64 y)
{
    x.HighPart -= y.HighPart + borrow(x.LowPart, y.LowPart);
    x.LowPart -= y.LowPart;
    return (x);
}

/* MMLite runtime library function -- Subtract an unsigned,
 * 32-bit integer (2nd arg) from an unsigned, 64-bit integer
 * (first arg) and return the unsigned, 64-bit difference.
 */
UINT64 Uint64SubtractUint32(UINT64 x, UINT32 y)
{
    x.HighPart -= borrow(x.LowPart, y);
    x.LowPart -= y;
    return (x);
}
