/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Portable 64-bit integer math functions for MMLite
 */
#define __NO_BUILTIN_INT64 1
#if defined(MMLITE)
#include <mmlite.h>
#include <machdep.h>
#else
#define INT64 MYINT64
#define UINT64 MYUINT64
#include <windows.h>
#undef INT64
#undef UINT64
typedef struct _INT64 INT64;
typedef struct _UINT64 UINT64;
#define EXTERN_C extern
#include "int64.h"
#endif

/* Cannot (and should not) compile this if there is compiler support.
 * Dont think you can have it both ways because of alignment issues.
 * Someday the unions might get smarter.
 */
/* Well then, we'll give it another shot.  If there is compiler support
 * we'll just make sure the compiler support routines work.  So the code
 * in there is only used by the test programs e.g. tstdiv64.c & friends.
 */

/* Nullify macros for functions we define. If we get compiled then
 * it means that those macros are needed as functions.
 */
#undef Int64ModulusInt64
#undef Int64ModulusInt32
#undef Uint64ModulusUint64
#undef Uint64ModulusUint32

#if !defined(__TCS__)
#define MUX(a,b,c)     ((a) ? (b) : (c))
#define iabs(a) ((INT32)(a) < 0 ? -(INT32)(a) : (a))
/* umulm returns the high 32 bits of 32x32->64 multiplication */
UINT32 umulm(UINT32 a, UINT32 b);
#endif

#if defined(_MSC_VER)
#pragma warning(disable:4146)  /* unary minus to unsigned type, result still unsigned */
#endif

extern void Uint64DivMod(UINT64 *pq, UINT64 x, UINT64 y, UINT64 *pr);
extern void Int64DivMod(INT64 *pq, INT64 x, INT64 y, INT64 *pr);

/* MMLite runtime library function -- Remainder of the division
 * of a signed, 64-bit integer by a signed, 64-bit integer.  The
 * return value (the modulus) is a signed 64-bit integer.
 */
INT64 Int64ModulusInt64(INT64 x, INT64 y)
{
    INT64 z;

    Int64DivMod(NULL,x,y,&z);

    return (z);
}


/* MMLite runtime library function -- Remainder of the division
 * of a signed, 64-bit integer by a signed, 32-bit integer.  The
 * return value (the modulus) is a signed 64-bit integer.
 */
INT64 Int64ModulusInt32(INT64 x, INT32 y)
{
    INT64 y64;

    y64.LowPart = y;
    y64.HighPart = y >> 31;
    return Int64ModulusInt64(x, y64);
}

/* MMLite runtime library function -- Remainder of the division
 * of an unsigned, 64-bit integer by an unsigned, 64-bit integer.
 * The return value (the modulus) is an unsigned 64-bit integer.
 */
UINT64 Uint64ModulusUint64(UINT64 x, UINT64 y)
{
    UINT64 z;

    Uint64DivMod(NULL,x,y,&z);

    return (z);
}


/* MMLite runtime library function -- Remainder of the division
 * of an unsigned, 64-bit integer by an unsigned, 32-bit integer.
 * The return value (the modulus) is an unsigned 64-bit integer.
 */
UINT64 Uint64ModulusUint32(UINT64 x, UINT32 y)
{
    UINT64 z;

    z.HighPart = 0;
    z.LowPart = y;
    Uint64DivMod(NULL,x,z,&z);

    return (z);
}

