/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Portable 64-bit integer math functions for MMLite
 */
#define __NO_BUILTIN_INT64 1
#if defined(MMLITE)
#include <mmlite.h>
#include <machdep.h>
#else
#define INT64 MYINT64
#define UINT64 MYUINT64
#include <windows.h>
#undef INT64
#undef UINT64
typedef struct _INT64 INT64;
typedef struct _UINT64 UINT64;
#define EXTERN_C extern
#include "int64.h"
#endif

/* Cannot (and should not) compile this if there is compiler support.
 * Dont think you can have it both ways because of alignment issues.
 * Someday the unions might get smarter.
 */
/* Well then, we'll give it another shot.  If there is compiler support
 * we'll just make sure the compiler support routines work.  So the code
 * in there is only used by the test programs e.g. tstdiv64.c & friends.
 */

/* Nullify macros for functions we define. If we get compiled then
 * it means that those macros are needed as functions.
 */
#undef Int64Or
#undef Uint64Or
#undef Int64Xor
#undef Uint64Xor
#undef Int64And
#undef Uint64And

/* MMLite runtime library function -- Or two signed,
 * 64-bit integers, return the signed 64-bit result.
 */
INT64 Int64Or(INT64 a, INT64 b)
{
    a.LowPart = a.LowPart | b.LowPart;
    a.HighPart = a.HighPart | b.HighPart;
    return (a);
}

/* MMLite runtime library function -- Or two unsigned,
 * 64-bit integers, return the unsigned 64-bit result.
 */
UINT64 Uint64Or(UINT64 a, UINT64 b)
{
    a.LowPart = a.LowPart | b.LowPart;
    a.HighPart = a.HighPart | b.HighPart;
    return (a);
}

/* MMLite runtime library function -- Xor two signed,
 * 64-bit integers, return the signed 64-bit result.
 */
INT64 Int64Xor(INT64 a, INT64 b)
{
    a.LowPart = a.LowPart ^ b.LowPart;
    a.HighPart = a.HighPart ^ b.HighPart;
    return (a);
}

/* MMLite runtime library function -- Xor two unsigned,
 * 64-bit integers, return the unsigned 64-bit result.
 */
UINT64 Uint64Xor(UINT64 a, UINT64 b)
{
    a.LowPart = a.LowPart ^ b.LowPart;
    a.HighPart = a.HighPart ^ b.HighPart;
    return (a);
}

/* MMLite runtime library function -- And two signed,
 * 64-bit integers, return the signed 64-bit result.
 */
INT64 Int64And(INT64 a, INT64 b)
{
    a.LowPart = a.LowPart & b.LowPart;
    a.HighPart = a.HighPart & b.HighPart;
    return (a);
}

/* MMLite runtime library function -- And two unsigned,
 * 64-bit integers, return the unsigned 64-bit result.
 */
UINT64 Uint64And(UINT64 a, UINT64 b)
{
    a.LowPart = a.LowPart & b.LowPart;
    a.HighPart = a.HighPart & b.HighPart;
    return (a);
}
