/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Portable 64-bit integer math functions for MMLite
 */
#define __NO_BUILTIN_INT64 1
#if defined(MMLITE)
#include <mmlite.h>
#include <machdep.h>
#else
#define INT64 MYINT64
#define UINT64 MYUINT64
#include <windows.h>
#undef INT64
#undef UINT64
typedef struct _INT64 INT64;
typedef struct _UINT64 UINT64;
#define EXTERN_C extern
#include "int64.h"
#endif

/* Cannot (and should not) compile this if there is compiler support.
 * Dont think you can have it both ways because of alignment issues.
 * Someday the unions might get smarter.
 */
/* Well then, we'll give it another shot.  If there is compiler support
 * we'll just make sure the compiler support routines work.  So the code
 * in there is only used by the test programs e.g. tstdiv64.c & friends.
 */

/* Nullify macros for functions we define. If we get compiled then
 * it means that those macros are needed as functions.
 */
#undef Int64RShift
#undef Uint64RShift
#undef Int64LShift
#undef Uint64LShift

extern INT64 INT64_NEGATIVE_ONE;

/* MMLite runtime library function -- Shift a signed,
 * 64-bit integer right by the specified number of bits.
 * The first argument contains the value to be shifted
 * and the second argument specifies the shift amount.
 * The return value is the shifted version of the first
 * argument.
 */
INT64 Int64RShift(INT64 x, INT32 n)
{
    /* check bad cases: ((n > 63) || (n < 0)) */
    if ((n & 63) != n)
        return (x.HighPart >= 0) ? INT64_ZERO : INT64_NEGATIVE_ONE;

    if (n > 31) {
        x.LowPart = x.HighPart >> (n-32);
        x.HighPart = x.HighPart >> 31;
    } else
    if (n) {
        x.LowPart = (x.LowPart >> n) | (x.HighPart << (32-n));
        x.HighPart = x.HighPart >> n;
    }
    return (x);
}


/* MMLite runtime library function -- Shift an unsigned,
 * 64-bit integer right by the specified number of bits.
 * The first argument contains the value to be shifted
 * and the second argument specifies the shift amount.
 * The return value is the shifted version of the first
 * argument.
 */
UINT64 Uint64RShift(UINT64 x, INT32 n)
{
    /* check bad cases: ((n > 63) || (n < 0)) */
    if ((n & 63) != n)
        return Int64ToUint64(INT64_ZERO);

    if (n > 31) {
        x.LowPart = x.HighPart >> (n-32);
        x.HighPart = 0;
    } else
    if (n) {
        x.LowPart = (x.LowPart >> n) | (x.HighPart << (32-n));
        x.HighPart = x.HighPart >> n;
    }
    return (x);
}


/* MMLite runtime library function -- Shift a signed,
 * 64-bit integer left by the specified number of bits.
 * The first argument contains the value to be shifted
 * and the second argument specifies the shift amount.
 * The return value is the shifted version of the first
 * argument.
 */
INT64 Int64LShift(INT64 x, INT32 n)
{
    /* check bad cases: ((n > 63) || (n < 0)) */
    if ((n & 63) != n)
        return INT64_ZERO;
    if (n > 31) {
        x.HighPart = x.LowPart << (n-32);
        x.LowPart = 0;
    } else
    if (n) {
        x.HighPart = (x.HighPart << n) | (x.LowPart >> (32-n));
        x.LowPart = x.LowPart << n;
    }
    return (x);
}

/* MMLite runtime library function -- Shift an unsigned,
 * 64-bit integer left by the specified number of bits.
 * The first argument contains the value to be shifted
 * and the second argument specifies the shift amount.
 * The return value is the shifted version of the first
 * argument.
 */
UINT64 Uint64LShift(UINT64 x, INT32 n)
{
    Uint64ToInt64(x) = Int64LShift(Uint64ToInt64(x),n);
    return (x);
}
