/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include <io.h>
#include "_stdio.h"

/* Regular, low level I/O stuff
 */

struct SeekTable {
    INT64 SeekPos;
    PIFILE TheFile;
};

#define MAX_FILES FOPEN_MAX
struct SeekTable TheStupidSeekTable[MAX_FILES];

long int lseek(FILE_HANDLE fileno, long Offset, int Origin)
{
    INT64 NewPos, Offset64;
    PIFILE pFile;

    /* Make sure inited ifso */
    if (fileno < 3)
        (void) __stdfile(0);

    Int32ToInt64(Offset64,Offset);

    NewPos = TheStupidSeekTable[fileno].SeekPos;

    switch (Origin) {
    case SEEK_END:
        pFile = TheStupidSeekTable[fileno].TheFile;
        if (FAILED(pFile->v->GetSize( pFile, (UINT64*)&NewPos)))
            break;

        /* else fall through */

    case SEEK_CUR:
        NewPos = Int64Add(NewPos,Offset64);
        break;

    case SEEK_SET:
        NewPos = Offset64;
        break;
    }

    TheStupidSeekTable[fileno].SeekPos = NewPos;

    Offset = Int64ToInt32(NewPos);
    return (INT) Offset;

}

#ifdef _UNICODE /* we need both versions XXX */
FILE_HANDLE open(const char *FileName, int Flags, ...)
{
    PIFILE pFile;
    int i;

    i = NAME_SPACE_READ;/* unless*/
    if (Flags & O_WRONLY)
        i = NAME_SPACE_WRITE;
    if (Flags & O_RDWR)
        i |= NAME_SPACE_READ|NAME_SPACE_WRITE;

    if (Flags & O_APPEND)
        /* BUGBUG Seeking done correctly only if fopen() */
        i |= NAME_SPACE_APPEND;

    if ((Flags & (O_CREAT|O_EXCL)) == (O_CREAT|O_EXCL))
        i |= NAME_SPACE_FAILIFEXIST;
    if (Flags & O_CREAT)
        i |= NAME_SPACE_CREATE;

    BindToObjectA(CurrentNameSpace(), FileName, i, &IID_IFile, (void**)&pFile);
    if (pFile == NULL)
        return -1;

#ifdef notneeded
    if (TheStupidSeekTable == NULL) {
        struct SeekTable *t;

        t = (struct SeekTable *) malloc( MAX_FILES * sizeof(struct SeekTable));
        if (t == NULL)
            return -1;
        memset( t, 0, MAX_FILES * sizeof(struct SeekTable));

        if (!AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable, (UINT) NULL, (UINT) t))
            free(t);
    }
#endif /* notneeded */

    /* leave stdin/out/err alone */
    for (i = 3; i < MAX_FILES; i++)
        if (TheStupidSeekTable[i].TheFile == NULL)
            if (AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable[i].TheFile,
                                   (UINT) NULL,
                                   (UINT) pFile)) {
                Int32ToInt64(TheStupidSeekTable[i].SeekPos,0);
                /* BUGBUG if O_APPEND should getsize() it */
                return i;
            }

    pFile->v->Release( pFile);
    return -1;
}
#endif

FILE_HANDLE _topen(const _TCHAR *FileName, int Flags, ...)
{
    PIFILE pFile;
    int i;

    i = NAME_SPACE_READ;/* unless*/
    if (Flags & O_WRONLY)
        i = NAME_SPACE_WRITE;
    if (Flags & O_RDWR)
        i |= NAME_SPACE_READ|NAME_SPACE_WRITE;

    if (Flags & O_APPEND)
        /* BUGBUG Seeking done correctly only if fopen() */
        i |= NAME_SPACE_APPEND;

    if ((Flags & (O_CREAT|O_EXCL)) == (O_CREAT|O_EXCL))
        i |= NAME_SPACE_FAILIFEXIST;
    if (Flags & O_CREAT)
        i |= NAME_SPACE_CREATE;

    BindToObject(CurrentNameSpace(), FileName, i, &IID_IFile, (void**)&pFile);
    if (pFile == NULL)
        return -1;

#ifdef notneeded
    if (TheStupidSeekTable == NULL) {
        struct SeekTable *t;

        t = (struct SeekTable *) malloc( MAX_FILES * sizeof(struct SeekTable));
        if (t == NULL)
            return -1;
        memset( t, 0, MAX_FILES * sizeof(struct SeekTable));

        if (!AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable, (UINT) NULL, (UINT) t))
            free(t);
    }
#endif /* notneeded */

    /* leave stdin/out/err alone */
    for (i = 3; i < MAX_FILES; i++)
        if (TheStupidSeekTable[i].TheFile == NULL)
            if (AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable[i].TheFile,
                                   (UINT) NULL,
                                   (UINT) pFile)) {
                Int32ToInt64(TheStupidSeekTable[i].SeekPos,0);
                /* BUGBUG if O_APPEND should getsize() it */
                return i;
            }

    pFile->v->Release( pFile);
    return -1;
}

static INT __openstdfile(int fileno)
{
    PIFILE pFile;
    char *FileName;
    UINT Flags = 0;

    if (fileno == 0) {
        FileName = "stdin";
        /* Flags = O_RDONLY; */
    } else if (fileno == 1) {
        FileName = "stdout";
        /* Flags = O_WRONLY; */
    } else if (fileno == 2) {
        FileName = "stderr";
        /* Flags = O_RDWR; */
    } else {
        DebugBreak();
        return -1;
    }

    /* XXX Mode if O_CREAT */

    BindToObjectA(CurrentNameSpace(), FileName, Flags, &IID_IFile,
                  (void**)&pFile);
    if (pFile == NULL)
        return -1;

#ifdef notneeded
    if (TheStupidSeekTable == NULL) {
        struct SeekTable *t;

        t = (struct SeekTable *) malloc( MAX_FILES * sizeof(struct SeekTable));
        if (t == NULL)
            return -1;
        memset( t, 0, MAX_FILES * sizeof(struct SeekTable));

        if (!AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable, (UINT) NULL, (UINT) t))
            free(t);
    }
#endif /* notneeded */

    if (TheStupidSeekTable[fileno].TheFile == NULL)
      if (AtomicCmpAndSwap( (PUINT) &TheStupidSeekTable[fileno].TheFile,
                           (UINT) NULL,
                           (UINT) pFile)) {
          Int32ToInt64(TheStupidSeekTable[fileno].SeekPos,0);
          return fileno;
      }

    pFile->v->Release( pFile);
    return fileno;              /* ok, somebody else did it for us. */
}

int read(FILE_HANDLE fileno, void *buf, size_t nbytes)
{
    PIFILE pFile;
    INT64 Pos, Btr64;
    SCODE Sc;
    UINT BytesThisRound;
    int BytesRead = 0;

    /* Make sure inited ifso */
    if (fileno < 3)
        (void) __stdfile(0);

  Restart:
    pFile = TheStupidSeekTable[fileno].TheFile;
    Pos = TheStupidSeekTable[fileno].SeekPos;

    if (pFile == NULL) {
        if (fileno < 3 && fileno >= 0) {
            if (__openstdfile(fileno) == -1)
              return -1;
            else
              goto Restart;
        }
        return -1;
    }

    do {
        Sc = pFile->v->ReadAt( pFile, Int64ToUint64(Pos), buf, nbytes, &BytesThisRound);

        if (FAILED(Sc))
            goto Exit;

        if (Sc == S_FALSE && BytesThisRound == 0)
            return BytesRead;           /* EOF */

        Int32ToInt64(Btr64,BytesThisRound);
        Pos = Int64Add(Pos,Btr64);
        TheStupidSeekTable[fileno].SeekPos = Pos;

        BytesRead += BytesThisRound;
        buf = (UINT8 *) buf + BytesThisRound;
        nbytes -= BytesThisRound;

    } while (nbytes > 0);

 Exit:
    return BytesRead ? BytesRead : -1;
}

int write(FILE_HANDLE fileno, const void *buf, size_t nbytes)
{
    PIFILE pFile;
    INT64 Pos, Btr64;
    SCODE Sc;
    UINT BytesThisRound;
    int BytesWritten = 0;

    /* Make sure inited ifso */
    if (fileno < 3)
      (void) __stdfile(fileno);

  Restart:
    pFile = TheStupidSeekTable[fileno].TheFile;
    Pos = TheStupidSeekTable[fileno].SeekPos;

    if (pFile == NULL) {
        if (fileno < 3 && fileno >= 0) {
            if (__openstdfile(fileno) == -1)
              return -1;
            else
              goto Restart;
        }
        return -1;
    }

    do {
        Sc = pFile->v->WriteAt( pFile, Int64ToUint64(Pos), buf, nbytes, &BytesThisRound);
        if (FAILED(Sc))
            return BytesWritten ? BytesWritten : -1;

        Int32ToInt64(Btr64,BytesThisRound);
        Pos = Int64Add(Pos,Btr64);
        TheStupidSeekTable[fileno].SeekPos = Pos;

        BytesWritten += BytesThisRound;
        buf = (UINT8 *) buf + BytesThisRound;
        nbytes -= BytesThisRound;

    } while (nbytes > 0);

    return BytesWritten;
}

int close(FILE_HANDLE fileno)
{
    int r = 0;
    PIFILE pFile = TheStupidSeekTable[fileno].TheFile;

    if (pFile) {
        r = pFile->v->Release(pFile);
        if (r == 0) {
            r = 1;
            TheStupidSeekTable[fileno].TheFile = NULL;
        }
    }
    return r;
}

int filelength (FILE_HANDLE fileno)
{
    UINT64 NewPos;
    PIFILE pFile;

    pFile = TheStupidSeekTable[fileno].TheFile;
    if (FAILED(pFile->v->GetSize( pFile, &NewPos)))
        return -1;

    return Int64ToInt32(NewPos);
}


