/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include "_stdio.h"
#include <mmlite.h>
#include <mmmacros.h>

/* Return the Unix "_iob" thing
 */

#define STDIN_FLAGS     (_IOREAD | _IONBF)
#define STDOUT_FLAGS    (_IOWRT  | _IOLBF)
#define STDERR_FLAGS    (_IOWRT  | _IONBF)

FILE *__stdfile( int index )
{
    PIPROCESS pPrc = CurrentProcess();
    FILE *f;
    PIFILE File;
    SCODE sc;

    if (index < 0 || index > 2)
        return NULL;

 Restart:
    f = (FILE*) pPrc->v->GetValue(pPrc, index);
    if (f)
        return f;

    /* Open for CONSOLE
     */
    sc = BindToObjectA(CurrentNameSpace(),
                       index == 0 ? "stdin" : index == 1 ? "stdout" : "stderr",
                       index == 0 ? NAME_SPACE_READ
                                  : NAME_SPACE_READ|NAME_SPACE_WRITE,
                       &IID_IFile,
                       (void **) &File);
    if (FAILED(sc))
        return NULL;

    f = malloc(sizeof(*f));
    if (!f) {
        RELEASE(File);
        return NULL;
    }
    memset(f, 0, sizeof(*f));
    f->_file = File;
    f->_flag = index == 0 ? STDIN_FLAGS : index == 1 ? STDOUT_FLAGS : STDERR_FLAGS;

    f = (FILE*) pPrc->v->SetValue(pPrc, index, (ADDRESS) f);
    if (f) {
        /* There already was one, zap it */
        if (f)
            RELEASE((PIFILE) f->_file);
        free(f);
    }

    goto Restart;
}

void __closestdfile(void)
{
    PIPROCESS pPrc = CurrentProcess();
    UINT index;
    for (index = 0; index < 3; index++) {
        FILE *f = (FILE*) pPrc->v->GetValue(pPrc, index);
        if (f)
            RELEASE((PIFILE) f->_file);
        free(f);
    }
}
