/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *  64bit math support routines required by the ADS toolset.
 */
	#include <arm_asm.h>

	_CODE(C$$code)

#if !defined(__GNUC__)

        _EXPORT(_ll_cmpu)
        _EXPORT(_ll_cmpge)
        _EXPORT(_ll_ushift_r)
        _EXPORT(_ll_shift_l)
        _EXPORT(_ll_sshift_r)
        _EXPORT(_ll_sdiv)
        _EXPORT(_ll_sdiv10)
        _EXPORT(_ll_udiv)
        _EXPORT(_ll_udiv10)
        _EXPORT(_ll_mul)
        _EXPORT(_ll_mulls)
        _EXPORT(_ll_srdv)

#define _USE_PORTABLE_INT64 0

/*
 * Function  _LL_CMPU( INT64 a, INT64 b)
 * Return    Condition codes
 * Description
 *    Tests if (a==b)
 */
_FUNCTION(_ll_cmpu)
    cmp    r1,r3
    cmpeq  r0,r2
    mov    pc,lr

/*
 * Function  _LL_CMPGE( INT64 a, INT64 b)
 * Return    Condition codes 
 * Description
 *    Tests if (a>=b) 
 */
_FUNCTION(_ll_cmpge)
    subs   r12,r0,r2
    sbcs   r12,r1,r3
    mov    pc,lr


/*
 * Function  _LL_USHIFT_R(UINT64 a, UINT b)
 * Return    UINT64 in r0,r1
 * Description
 *    Shifts (a>>b) unsigned
 */
_FUNCTION(_ll_ushift_r)

    /* If (b>63)||(b<0) return 0 */
    bics   r3,r2,#0x3f
    subne  r0,r0,r0
    subne  r1,r1,r1
    movne  pc,lr

    /* if (b>31) return {hi>>(b-32),0} */
    subs   r12,r2,#0x20
    movge  r0,r1,lsr r12
    subge  r1,r1,r1
    movge  pc,lr

    /* if (b) return {(lo>>b)|(hi<<(32-b),hi>>b} */
    cmp    r2,#0
    rsbne  r12,r2,#0x20
    movne  r0,r0,lsr r2
    orrne  r0,r0,r1,lsl r12
    movne  r1,r1,lsr r2
    
    /* Done */
    mov    pc,lr


/*
 * Function  _LL_SHIFT_L(UINT64 a, UINT b)
 * Return    UINT64 in r0,r1
 * Description
 *    Shifts (a<<b) signed or unsigned
 */
_FUNCTION(_ll_shift_l)
    /* If (b>63)||(b<0) return 0 */
    bics   r3,r2,#0x3f
    subne  r0,r0,r0
    subne  r1,r1,r1
    movne  pc,lr

    /* if (b>31) return {0,lo<<(b-32)} */
    subs   r12,r2,#0x20
    movge  r1,r0,lsl r12
    subge  r0,r0,r0
    movge  pc,lr

    /* if (b) return {lo<<b,(hi<<b)|(lo>>(32-b) */
    cmp    r2,#0
    rsbne  r12,r2,#0x20
    movne  r1,r1,lsl r2
    orrne  r1,r1,r0,lsr r12
    movne  r0,r0,lsl r2
    
    /* Done */
    mov    pc,lr


/*
 * Function  _LL_SSHIFT_R(UINT64 a, UINT b)
 * Return    UINT64 in r0,r1
 * Description
 *    Shifts (a>>b) signed
 */
_FUNCTION(_ll_sshift_r)
    /* If (b>63)||(b<0) dup high bit in all positions */
    bics   r3,r2,#0x3f
    movne  r0,r1,asr #31
    movne  r1,r1,asr #31
    movne  pc,lr

    /* if (b>31) return {hi>>(b-32),hi>>31} */
    subs   r12,r2,#0x20
    movge  r0,r1,asr r12
    movge  r1,r1,asr #31
    movge  pc,lr

    /* if (b) return {(lo>>b)|(hi<<(32-b),hi>>b} */
    cmp    r2,#0
    rsbne  r12,r2,#0x20
    movne  r0,r0,lsr r2
    orrne  r0,r0,r1,lsl r12
    movne  r1,r1,asr r2
    
    /* Done */
    mov    pc,lr


/*
 * Function  _LL_SDIV( INT64 a, INT64 b)
 * Return    INT64 in r0,r1, INT64 in r2,r3
 * Description
 *    Performs the signed division (a/b),
 *    returns the result in r0/r1 and the remainder in r2/r3
 */
_FUNCTION(_ll_sdiv)
#if _USE_PORTABLE_INT64
        _IMPORT(Int64DivMod)
    /* Save LR, make room for results  */
    stmdb  sp!,{r0-r3,lr}

    /* Pointer to quotient result      */
    mov    r0,sp

    /* Pointer to remainder result     */
    add    r12,sp,#8

    /* Place it in the proper spot     */
    stmdb  sp!,{r12}

    /* Also place b.high on stack */
    stmdb  sp!,{r3}

    /* Shuffle registers     */
    ldmia  r0,{r1,r2,r3}

    /* Make call */
    bl     Int64DivMod

    /* Pop stack and return
     */
    ldmia  sp!,{r0,r1}
    ldmia  sp!,{r0-r3,pc}
#else

#define aLO r0
#define aHI r1
#define bLO r2
#define bHI r3

//    _DCD 0xe7ffdefe

    /* Save regs [x==a==r0.lo,r1.hi  y==b==r2.lo,r3.hi] */
    stmdb  sp!,{r4,lr}

    /* Remember sign of result:
     * sign = (x.HighPart ^ y.HighPart) >> 31;
     * aneg = x.HighPart >> 31;
     * tmp = sign | (aneg << 1)
     */
    eor    r4,aHI,bHI
    mov    r4,r4, lsr #31
    movs   ip,aHI, lsr #31
    orr    r4,r4,ip, lsl #1

    /* Take absolute values of dividend x and divisor y
     */
    beq    AnotNeg
    
    mvn    ip,aHI
    rsbs   aHI,aLO,#1
    movcc  aHI,#0
    add    aHI,ip,aHI
    rsb    aLO,aLO,#0

_LABEL(AnotNeg)
    movs   ip,bHI, lsr #31
    beq    BnotNeg
    
    mvn    ip,bHI
    rsbs   bHI,bLO,#1
    movcc  bHI,#0
    add    bHI,ip,bHI
    rsb    bLO,bLO,#0

_LABEL(BnotNeg)
    /* Perform the unsigned division, cast the results.
     */
    bl     _ll_udiv

    /* Do we have to negate the quotient/remainder ?
     * NB: after the lsl we get r4.31==sign C==Aneg
     */
    movs   r4,r4, lsl #31
    bcc    CheckQuo

    /* Negate remainder
     */
    mvn    ip,bHI
    rsbs   bHI,bLO,#1
    movcc  bHI,#0
    add    bHI,ip,bHI
    rsb    bLO,bLO,#0

    cmp    r4, #0

_LABEL(CheckQuo)
    ldmplia sp!,{r4,pc}  /* done ? */

    /* Negate quotient
     */
    mvn    ip,aHI
    rsbs   aHI,aLO,#1
    movcc  aHI,#0
    add    aHI,ip,aHI
    rsb    aLO,aLO,#0

    /* Pop stack and return
     */
    ldmia  sp!,{r4,pc}

#undef aLO
#undef aHI
#undef bLO
#undef bHI

#endif


/*
 * Function  _LL_SDIV10( INT64 a)
 * Return    INT64 in r0,r1
 * Description
 *    Performs the signed division (a/10)
 */
_FUNCTION(_ll_sdiv10)

    mov    r2,#10
    mov    r3,#0
    b      _ll_sdiv


/*
 * Function  _LL_UDIV( UINT64 a, UINT64 b)
 * Return    UINT64 in r0,r1, INT64 in r2,r3
 * Description
 *    Performs the unsigned division (a/b),
 *    returns the result in r0/r1 and the remainder in r2/r3
 */
_FUNCTION(_ll_udiv)
#if _USE_PORTABLE_INT64
        _IMPORT(Uint64DivMod)

    /* Save LR, make room for results  */
    stmdb  sp!,{r0-r3,lr}

    /* Pointer to quotient result      */
    mov    r0,sp

    /* Pointer to remainder result     */
    add    r12,sp,#8

    /* Place it in the proper spot     */
    stmdb  sp!,{r12}

    /* Also place b.high on stack */
    stmdb  sp!,{r3}

    /* Shuffle registers     */
    ldmia  r0,{r1,r2,r3}

    /* Make call */
    bl     Uint64DivMod

    /* Pop stack and return
     */
    ldmia  sp!,{r0,r1}
    ldmia  sp!,{r0-r3,pc}

#else

/* Register assignments
 * NB: after prologue!
 */
#define zLO r0
#define zHI r1
#define xLO r2
#define xHI r3
#define yLO r4
#define yHI r5
#define t1 r12
#define t2 r0   /* alias zLO */
#define t3 r1   /* alias zHI */
#define t4 r14
#define t6 r6
#define t7 r7
//saved: r8,r9,r10

    stmdb    r13!,{r4-r7,r14}
//    _DCD 0xe7ffdefe

    /* shuffle regs */
    mov      yHI,r3
    mov      yLO,r2
    mov      xHI,r1
    mov      xLO,r0
    
   /* end of prologue */

    /* y == 0 ? */
    orrs     t1,yLO,yHI         /* t1 scratch */
    beq      UDivZero

    /* x1ne0 = x.hi != 0 */
    movs     t2,xHI
    movne    t2,#1
    /* x32 = (x1ne0) ? x.hi : x.lo */
    moveq    t1,xLO
    movne    t1,xHI

    /* Get index of leftmost bit of x32 into I (t4) */
    movs     t4,t1,ASR #16
    moveq    t4,#8
    movne    t4,#24
    movs     t7,t1,LSR t4
    subeq    t4,t4,#4
    addne    t4,t4,#4
    movs     t7,t1,LSR t4
    subeq    t4,t4,#2
    addne    t4,t4,#2
    movs     t7,t1,LSR t4
    subeq    t4,t4,#1
    addne    t4,t4,#1
    movs     t1,t1,LSR t4
    movne    t1,#0
    moveq    t1,#1
    sub      t4,t4,t1

    /* ALIVE: t2,t4,x,y */

    /* y1ne0 = y.hi != 0 */
    movs     t1,yHI
    movne    t1,#1
    /* x1ne0 = x1ne0 - y1ne0 (used later on) */
    sub      t2,t2,t1

    /* y32 = (y1ne0) ? y.hi : y.lo */
    moveq    t1,yLO
    movne    t1,yHI

    /* Get index of leftmost bit of y32 into J (t6) */
    movs     t6,t1,ASR #16
    moveq    t6,#8
    movne    t6,#24
    movs     t7,t1,LSR t6
    subeq    t6,t6,#4
    addne    t6,t6,#4
    movs     t7,t1,LSR t6
    subeq    t6,t6,#2
    addne    t6,t6,#2
    movs     t7,t1,LSR t6
    subeq    t6,t6,#1
    addne    t6,t6,#1
    movs     t1,t1,LSR t6
    movne    t1,#0
    moveq    t1,#1
    sub      t6,t6,t1

    /* ALIVE: t2,t4,t6,x,y */

    /* N = I - J */
    subs     t3,t4,t6
    /* I,J unused => t4,t6 scratch */
    /* ALIVE: t2,t3,x,y */

    /* if (N != 0) temp(t1) = y.lo >> ((32-N)&31); */
    rsbne    t1,t3,#0x20
    andne    t1,t1,#0x1f
    movne    t1,yLO,LSR t1

    /* y.hi = y.hi << (N&31) */
    and      t6,t3,#0x1f
    mov      yHI,yHI,LSL t6
    /* if (N != 0) y.hi |= temp */
    orrne    yHI,yHI,t1

    /* t1 scratch */
    /* ALIVE: t2,t3,t6,x,y */

    /* y.lo <<= (n&31) */
    mov      yLO,yLO,LSL t6

    /* t6 scratch */
    /* ALIVE: t2,t3,x,y */

    /* N += (x1ne0 - y1ne0) << 5; */
    add      t6,t3,t2,LSL #5

    /* N is now t6 */
    /* t2,t3 scratch */
    /* ALIVE: t6,x,y */

    /* Swap parts of Y if necessary, e.g. (n >= 32) */
    /* if (n >= 32) { y.hi = y.lo; y.lo = 0; } */
    cmp      t6,#0x20
    movge    yHI,yLO
    movge    yLO,#0

   /* BEWARE from here down z is active, dont use t2,t3 aliases */

    /* z.lo = z.hi = 0 */
    mov      zHI,#0
    mov      zLO,#0

    /* ALIVE: t6,x,y,z */

    /* while (n-- >= 0) */
    b        L3

_LABEL(L9)
    /* xgey = (x >= y) */
    cmp      xHI,yHI
    cmpeq    xLO,yLO
    bcc      L8

    /* x.hi -= y.hi + (x.lo < y.lo) */
    cmp      xLO,yLO
    sbc      xHI,xHI,yHI

    /* x.lo -= y.lo */
    sub      xLO,xLO,yLO

    /* set carry again! */
    cmp      xLO,xLO
_LABEL(L8)
    /* z = (z << 1) + xgey */
    adcs     zLO,zLO,zLO
    adc      zHI,zHI,zHI

    /* y >>= 1 */
    mov      yLO,yLO,lsr #1
    orr      yLO,yLO,yHI,lsl #31
    mov      yHI,yHI,lsr #1

_LABEL(L3)
    cmp      t6,#0
    sub      t6,t6,#1
    bge      L9

_LABEL(L2)
    ldmia    r13!,{r4-r7,pc}

_LABEL(UDivZero)
    /* is (x != 0) ? */
    orrs     xLO,xLO,xHI
    /* x = 0; */
    mov      xLO,#0
    mov      xHI,#0
    /* if (x was != 0) z = ~0; else z = 0 */
    moveq    zHI,xLO
    mvnne    zHI,xLO
    mov      zLO,zHI
    b        L2

#undef zLO
#undef zHI
#undef xLO
#undef xHI
#undef yLO
#undef yHI
#undef t1
#undef t2
#undef t3
#undef t4
#undef t6
#undef t7
#endif


/*
 * Function  _LL_MUL( UINT64 a, UINT64 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Performs the unsigned multiplication (a*b)
 */
_FUNCTION(_ll_mul)

#if _USE_PORTABLE_INT64 /* portable code */

        _IMPORT(Uint64TimesUint64)
    /* Save LR, make room for result   */
    stmdb  sp!,{r0-r1,lr}

    /* Pointer to result     */
    mov    r0,sp

    /* Place R3 in the proper spot     */
    stmdb  sp!,{r3}

    /* Shuffle registers     */
    mov    r3,r2
    ldmia  r0,{r1,r2}

    /* Make call */
    bl     Uint64TimesUint64

    /* Pop stack and return
     */
    ldmia  sp!,{r3}
    ldmia  sp!,{r0-r1,pc}

#else

    /* Use multiply long instructions */
    stmdb  sp!,{lr}    /* that leaves 2 scratch: r12,lr */

    /* a.low*b.low -> t.low,t.high (t.low==result.low) */
    umull  r12,lr,r0,r2

    /* t.high+a.low*b.high -> t.high */
    mla    lr,r3,r0,lr

    /* put result.low in place */
    mov    r0,r12

    /* t.high+a.high*b.low -> result.high */
    mla    r1,r2,r1,lr

    ldmia  sp!,{pc}

#endif

/*
 * Function  _LL_MULLS( INT64 a, INT32 b)
 * Return    INT64 in r0,r1
 * Description
 *    Performs the signed multiplication (a*(INT64)b)
 */
_FUNCTION(_ll_mulls)

    /* sign extend and do the obvious */
    mov    r3,r2,asr #31
    b      _ll_mul


/*
 * Function  _LL_SRDV( INT64 a, INT64 b)
 * Return    INT64 in r0,r1
 * Description
 *    Performs the signed division (b/a)
 * "Signed Reversed Division" ?
 */
_FUNCTION(_ll_srdv)

    /* horrible quick hack: just swap 'em */
    stmdb  sp!,{r0-r3}
    ldmia  sp!,{r2-r3}
    ldmia  sp!,{r0-r1}

    /* And do the regular thing */
    b      _ll_sdiv

/*
 * Function  _LL_UDIV10( UINT64 a)
 * Return    UINT64 in r0,r1, INT64 in r2,r3
 * Description
 *    Performs the unsigned division (a/10),
 *    returns the result in r0/r1 and the remainder in r2/r3
 */
_FUNCTION(_ll_udiv10)

    /* do the obvious */
    sub    r3,r3,r3
    add    r2,r3,#10
    b      _ll_udiv


#endif /* !__GNUC__ */


       _END
