/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *  More 64bit math support routines required by the ADS toolset.
 *  These should only showup in a debug build, not in release.
 */
	#include <arm_asm.h>

	_CODE(C$$code)

#if !defined(__GNUC__)
        
        _EXPORT(_ll_from_l)
        _EXPORT(_ll_from_u)
        _EXPORT(_ll_to_l)
        _EXPORT(_ll_or)
        _EXPORT(_ll_eor)
        _EXPORT(_ll_and)
        _EXPORT(_ll_add)
        _EXPORT(_ll_addlu)
        _EXPORT(_ll_addls)
        _EXPORT(_ll_sub)
        _EXPORT(_ll_sublu)
        _EXPORT(_ll_subls)
        _EXPORT(_ll_rsblu)
        _EXPORT(_ll_subuu)
        _EXPORT(_ll_mullu)
        _EXPORT(_ll_cmple)


/*
 * Function  _LL_FROM_L( INT32 a)
 * Return    INT64 in r0,r1
 * Description
 *    Typecast a to an INT64 (with sign-extension)
 */
_FUNCTION(_ll_from_l)
    /* Slide high bit into all positions of the high part */
    mov    r1,r0,asr #31
    mov    pc,lr


/*
 * Function  _LL_FROM_U( UINT32 a)
 * Return    UINT64 in r0,r1
 * Description
 *    Typecast a to an UINT64 (without sign-extension)
 */
_FUNCTION(_ll_from_u)
    /* zero the high part */
    sub    r1,r1,r1
    mov    pc,lr


/*
 * Function  _LL_TO_L( INT64 a)
 * Return    INT32 in r0
 * Description
 *    Typecast a to an INT32 (just chop it)
 */
_FUNCTION(_ll_to_l)
    /* nothing to do.. sheez */
    mov    pc,lr


/*
 * Function  _LL_OR( UINT64 a, UINT64 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the logical or (a|b)
 */
_FUNCTION(_ll_or)
    /* does not need to set the cond codes, saves one inst */
    orr    r0,r0,r2
    orr    r1,r1,r3
    mov    pc,lr


/*
 * Function  _LL_EOR
 */
_FUNCTION(_ll_eor)
    eor    r0,r0,r2
    eor    r1,r1,r3
    mov    pc,lr


/*
 * Function  _LL_AND
 */
_FUNCTION(_ll_and)
    and    r0,r0,r2
    and    r1,r1,r3
    mov    pc,lr


/*
 * Function  _LL_ADD( UINT64 a, UINT64 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the sum (a+b)
 * Notes
 *    Same as signed version.
 */
_FUNCTION(_ll_add)
    adds   r0,r0,r2
    adc    r1,r1,r3
    mov    pc,lr


/*
 * Function  _LL_ADDLU( UINT64 a, UINT32 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the sum (a+(UINT64)b)
 */
_FUNCTION(_ll_addlu)
    adds   r0,r0,r2
    /* Zero entension so.. */
    adc    r1,r1,#0
    mov    pc,lr


/*
 * Function  _LL_ADDLS( INT64 a, INT32 b)
 * Return    INT64 in r0,r1
 * Description
 *    Returns the sum (a+(INT64)b)
 */
_FUNCTION(_ll_addls)
    adds   r0,r0,r2
    /* sign extension and carry */
    adc    r1,r1,r2,asr #31
    mov    pc,lr


/*
 * Function  _LL_SUB( INT64 a, INT64 b)
 * Return    INT64 in r0,r1
 * Description
 *    Returns the difference (a-b)
 * Notes
 *    Same as signed version.
 */
_FUNCTION(_ll_sub)
    subs   r0,r0,r2
    sbcs   r1,r1,r3
    mov    pc,lr


/*
 * Function  _LL_SUBLU( UINT64 a, UINT32 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the difference (a-(UINT64)b)
 */
_FUNCTION(_ll_sublu)
    subs   r0,r0,r2
    sbc    r1,r1,#0
    mov    pc,lr


/*
 * Function  _LL_SUBLS( INT64 a, INT32 b)
 * Return    INT64 in r0,r1
 * Description
 *    Returns the difference (a-(INT64)b)
 */
_FUNCTION(_ll_subls)
    subs   r0,r0,r2
    sbc    r1,r1,r2,asr #31
    mov    pc,lr


/*
 * Function  _LL_RSBLU( UINT64 a, UINT32 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the difference ((UNT64)b-a)
 */
_FUNCTION(_ll_rsblu)
    rsbs   r0,r0,r2
    rsc    r1,r1,#0
    mov    pc,lr


/*
 * Function  _LL_SUBUU( UINT32 a, UINT32 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the difference ((UINT64)a-(UINT64)b)
 */
_FUNCTION(_ll_subuu)
    subs   r0,r0,r1
    sbc    r1,r0,#0
    mov    pc,lr


/*
 * Function  _LL_MULLU( UINT64 a, UINT32 b)
 * Return    UINT64 in r0,r1
 * Description
 *    Returns the product (a*(UINT64)b)
 */
_FUNCTION(_ll_mullu)
    /* r3 scratch */
    umull  r0,r3,r2,r0
    mla    r1,r2,r1,r3
    mov    pc,lr


/*
 * Function  _LL_CMPLE( INT64 a, INT64 b)
 * Return    Condition codes
 * Description
 *    Tests if (a<=b)
 */
_FUNCTION(_ll_cmple)
    subs   r0,r0,r2
    sbceqs r1,r1,r3
    mov    pc,lr


#endif

       _END
