/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

/* Byteorder issues first.
 * An __int64 (a.k.a. INT64) as first parameter is passed in the pair a0/a1,
 * but it depends on the byteorder which is low32 and which is high32.
 * Similarly for the second parameter and a2/a3.
 * And the result and v0/v1.
 */
#ifdef __MIPSEB__
#define p0h $a0
#define p0l $a1
#define p1h $a2
#define p1l $a3
#define resh $v0
#define resl $v1
#else
#define p0h $a1
#define p0l $a0
#define p1h $a3
#define p1l $a2
#define resh $v1 
#define resl $v0
#endif

/* unsigned __int64 __ull_div(unsigned __int64 Dividend, unsigned __int64 Divisor)
 *
 * Dividend = p0h:p0l, Divisor = p1h:p1l
 * Quotient = resh:resl
 *
 * This routine uses at,v0-v1,a0-a3,t0-t3.
 * It does not uses t4 - t9; the other functions below depend on that.
 */

ENTRY(__udivdi3)
_FUNCTION(__ull_div)

    .set noreorder
    li   $t2,63           /* Set loop count                    */
    move $t0,$zero        /* Clear partial remainder           */
    move $t1,$zero        /* done.                             */
next:
    sll  $t1,$t1,1        /* Shift next dividend bit           */
    srl  $t3,$t0,31       /* into the partial remainder        */
    or   $t1,$t1,$t3      /*                                   */
    sll  $t0,$t0,1        /* so far:  t1:t0<<1                 */
    srl  $t3,p0h,31       /*                                   */
    or   $t0,$t0,$t3      /* done.                             */
    sll  p0h,p0h,1        /* Left shift one the QWORD dividend */
    srl  $t3,p0l,31       /*                                   */
    or   p0h,p0h,$t3      /* done.                             */
    sltu $t3,$t1,p1h      /* Is remainder < divisor.high ?     */
    beq  $zero,$t3,rmge   /* If eq, partial remainder not less */
    sll  p0l,p0l,1        /*                                   */
    bne  $zero,$t2,next   /* If ne, more iterations to go      */
    subu $t2,$t2,1        /* Decrement iteration count         */
    b    done             /*                                   */
    nop                   /*                                   */
rmge:
    bne  p1h,$t1,rmgt     /* If ne, partial remainder greater  */
    sltu $t3,$t0,p1l      /* Is remainder < divisor.low ?      */
    bne  $zero,$t3,rmls   /* If ne, partial remainder is less  */
    nop                   /*                                   */
rmgt:
    or   p0l,p0l,1        /* Merge quotient bit                */
    subu $t1,$t1,p1h      /* Subtract out divisor high part    */
    sltu $t3,$t0,p1l      /* Set borrow from high part         */
    subu $t1,$t1,$t3      /* Subtract borrow from high part    */
    subu $t0,$t0,p1l      /* Subtract out divisor low          */
rmls:
    bne  $zero,$t2,next   /* If ne, more iterations to go      */
    subu $t2,$t2,1        /* Decrement iteration count         */
done:
    move resl,p0l         /* Done, collect low-part of result  */
    jr      $ra
    move resh,p0h         /* Collect high-part of result       */
    .set reorder
    END(__udivdi3)

/* __int64 __ll_div(__int64 Dividend, __int64 Divisor)
 *
 * Dividend = p0h:p0l, Divisor = p1h:p1l
 * Quotient = resh:resl
 */
ENTRY(__divdi3)
_FUNCTION(__ll_div)

    .set noreorder

    slt  $t5,p0h,$zero   /* t5 == 1 iff Dividend < 0        */
    beq  $t5,$zero,dvdpos
    move $t4,$ra         /* Save ra                         */
    sltu $t7,$zero,p0l
    subu p0l,$zero,p0l
    subu $t8,$zero,p0h
    subu p0h,$t8,$t7
dvdpos:
    slt  $t6,p1h,$zero   /* t6 == 1 iff Divisor < 0         */
    beq  $t6,$zero,dvrpos
    sltu $t7,$zero,p1l
    subu p1l,$zero,p1l
    subu $t8,$zero,p1h
    subu p1h,$t8,$t7
dvrpos:
    jal  _L(__ull_div)
    nop
    beq  $t5,$t6,noneg
    sltu $t7,$zero,resl
    subu resl,$zero,resl
    subu $t8,$zero,resh
    subu resh,$t8,$t7
noneg:
    jr   $t4
    move $ra,$t4
    END(__divdi3)

/* unsigned __int64 __umoddi3(unsigned __int64 Dividend, unsigned __int64 Divisor)
 *
 * Dividend = p0h:p0l, Divisor = p1h:p1l
 * Remainder = resh:resl
 *
 */

ENTRY(__umoddi3)
    move $t4,$ra        /* Save ra                         */
    jal  _L(__ull_div)
    nop
    move resh,$t1
    move resl,$t0
    jr   $t4
    move $ra,$t4
END(__umoddi3)

/* __int64 __moddi3( __int64 Dividend,  __int64 Divisor)
 *
 * Dividend = p0h:p0l, Divisor = p1h:p1l
 * Remainder = resh:resl
 *
 */

ENTRY(__moddi3)
    move $t9,$ra        /* Save ra                         */
    jal  _L(__ll_div)
    nop
    move resh,$t1
    beq  $t5,$zero,noneg1
    move resl,$t0
    sltu $t7,$zero,resl
    subu resl,$zero,resl
    subu $t8,$zero,resh
    subu resh,$t8,$t7
noneg1:
    jr   $t9
    move $ra,$t9
END(__moddi3)
                

/* BOOL __cmpdi2( __int64 A,  __int64 B)
 *
 * A = p0h:p0l, B = p1h:p1l
 * Result = $v0 TRUE if equal
 *
 */

ENTRY(__cmpdi2)
    .set noat
    subu $v0,p0h,p1h
    subu $at,p0l,p1l
    jr   $ra
    addu $v0,$v0,$at
END(__cmpdi2)
                
