/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

    _EXPORT(longjump)
    _EXPORT(setjump)

	.set noreorder
	.set noat

/*
 * void longjump(PCXTINFO regs, UINT32 retval);
 *
 * Reload the register state and return the indicated value
 * (e.g. in returning from the setjump() call that saved them regs)
 */
ENTRY(longjump)
	/* Set the return value */
	move $v0,$a1
	/* First load most registers */
	/* a0 later */
	lw	$a1, TS_A1($a0)
	lw	$a2, TS_A2($a0)
	lw	$a3, TS_A3($a0)
	lw	$t0, TS_T0($a0)
	lw	$t1, TS_T1($a0)
	lw	$t2, TS_T2($a0)
	lw	$t3, TS_T3($a0)
	lw	$t4, TS_T4($a0)
	lw	$t5, TS_T5($a0)
	lw	$t6, TS_T6($a0)
	lw	$t7, TS_T7($a0)
	lw	$s0, TS_S0($a0)
	lw	$s1, TS_S1($a0)
	lw	$s2, TS_S2($a0)
	lw	$s3, TS_S3($a0)
	lw	$s4, TS_S4($a0)
	lw	$s5, TS_S5($a0)
	lw	$s6, TS_S6($a0)
	lw	$s7, TS_S7($a0)
	lw	$t8, TS_T8($a0)
	lw	$t9, TS_T9($a0)
    /* k0,k1 not restored */
	lw	$gp, TS_GP($a0)
	/* sp later */
	lw	$fp, TS_FP($a0)
	lw	$ra, TS_RA($a0)

    lw  $v1, TS_HI($a0)
    lw  $at, TS_LO($a0)
    mthi $v1
    mtlo $at
	lw	$sp, TS_SP($a0)
	/* Put pc into v1 */
	lw	$v1, TS_PC($a0)
	lw	$at, TS_AT($a0)
	j	$v1
	lw	$a0, TS_A0($a0)
END(longjump)

/*
 * UINT32 setjump(PCXTINFO regs);
 *
 * Save the register state where indicated.
 * Returns 0 from this one call.
 * Upon invocation of longjump() it will appear as if setjump()
 * instead returns, with the value passed to longjump().
 *
 * Caller saved regs at,v1,a0-3,t0-7,ra,mdlo,mdhi are not saved but zeroed.
 * v0 (return value) is not saved but is set to one.
 * Returns zero normally but may return anything when "longjmped" to.
 */
ENTRY(setjump)
	/* caller pc is now in ra. save it as pc */
	sw	$ra, TS_PC($a0)
	/* save callee saved registers */
	sw	$s0, TS_S0($a0)
	sw	$s1, TS_S1($a0)
	sw	$s2, TS_S2($a0)
	sw	$s3, TS_S3($a0)
	sw	$s4, TS_S4($a0)
	sw	$s5, TS_S5($a0)
	sw	$s6, TS_S6($a0)
	sw	$s7, TS_S7($a0)
	sw	$t8, TS_T8($a0)
	sw	$t9, TS_T9($a0)
	sw	$gp, TS_GP($a0)
	sw	$sp, TS_SP($a0)
	sw	$fp, TS_FP($a0)

	/* caller saved registers: save as zero */
	sw	$zero, TS_V0($a0)
	sw	$zero, TS_AT($a0)
	sw	$zero, TS_V1($a0)
	sw	$zero, TS_A0($a0)
	sw	$zero, TS_A1($a0)
	sw	$zero, TS_A2($a0)
	sw	$zero, TS_A3($a0)
	sw	$zero, TS_T0($a0)
	sw	$zero, TS_T1($a0)
	sw	$zero, TS_T2($a0)
	sw	$zero, TS_T3($a0)
	sw	$zero, TS_T4($a0)
	sw	$zero, TS_T5($a0)
	sw	$zero, TS_T6($a0)
	sw	$zero, TS_T7($a0)
	sw	$zero, TS_RA($a0)
    mflo $t0
    mfhi $t1
	sw	$t0,   TS_LO($a0)
	sw	$t1,   TS_HI($a0)

	j	$ra
	move	$v0, $zero		/* return zero */
END(setjump)
        
