/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

/* Byteorder issues first.
 * An __int64 (a.k.a. INT64) as first parameter is passed in the pair a0/a1,
 * but it depends on the byteorder which is low32 and which is high32.
 * Similarly for the second parameter and a2/a3.
 * And the result and v0/v1.
 */
#ifdef __MIPSEB__
#define p0h $a0
#define p0l $a1
#define p1h $a2
#define p1l $a3
#define resh $v0
#define resl $v1
#else
#define p0h $a1
#define p0l $a0
#define p1h $a3
#define p1l $a2
#define resh $v1 
#define resl $v0
#endif

    .set noat
    .set noreorder

/* __uint64 __lshrdi3( __uint64 A, int32 B)
 *
 * A = p0h:p0l, B = a2
 * Result = resh:resl = A >> B
 *
 */

ENTRY(__lshrdi3)
    .set noat
    .set noreorder
    sll     $a3,$a2,26 /* more than 32 bit positions? */
    bgez    $a3,L01
    srlv    resl,p0h,$a2  /* yes, easy */
    jr      $ra
    move    resh,$zero
L01:
    beqz    $a3,L02
    srlv    resl,p0l,$a2
    negu    $a3,$a2
    sllv    $a3,p0h,$a3
    or      resl,resl,$a3
L02:
    jr      $ra
    srlv    resh,p0h,$a2

END(__lshrdi3)
                
/* __int64 __ashldi3( __int64 A, int32 B)
 *
 * A = p0h:p0l, B = a2
 * Result = resh:resl = A << B
 *
 */

ENTRY(__ashldi3)
    .set noat
    .set noreorder
    sll     $a3,$a2,26 /* more than 32 bit positions? */
    bgez    $a3,L1
    sllv    resh,p0l,$a2  /* yes, easy */
    jr      $ra
    move    resl,$zero
L1:
    beqz    $a3,L2
    sllv    resh,p0h,$a2
    negu    $a3,$a2
    srlv    $a3,p0l,$a3
    or      resh,resh,$a3
L2:
    jr      $ra
    sllv    resl,p0l,$a2

END(__ashldi3)
                
/* __int64 __ashrdi3( __int64 A, int32 B)
 *
 * A = p0h:p0l, B = a2
 * Result = resh:resl = A >> B
 *
 */

ENTRY(__ashrdi3)
    .set noat
    .set noreorder
    sll     $a3,$a2,26 /* more than 32 bit positions? */
    bgez    $a3,L11
    srav    resl,p0h,$a2  /* yes, easy */
    jr      $ra
    sra     resh,p0h,31
L11:
    beqz    $a3,L12
    srlv    resl,p0l,$a2
    negu    $a3,$a2
    sllv    $a3,p0h,$a3
    or      resl,resl,$a3
L12:
    jr      $ra
    srav    resh,p0h,$a2

END(__ashrdi3)
                
