/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>

#include <stdlib.h>
#include <tchar.h>
#include <stdio.h>
#include "_stdio.h"

#define DLL_PROCESS_CREATE -1
/* WinNT compatible dll entry Op values */
#define DLL_PROCESS_ATTACH 1
#define DLL_THREAD_ATTACH  2
#define DLL_THREAD_DETACH  3
#define DLL_PROCESS_DETACH 0


/* Declarations for entry.c, initterm.c, crtinit.c, msvcrt.c, main.c, atexit.c
 */

typedef void (__cdecl *PF)(void);

EXTERN_C void __cdecl _initterm(PF *pfbegin, PF *pfend);

EXTERN_C INT MODENTRY _CRT_INIT(void *xxx, INT Op, INT Arg3);

EXTERN_C BOOL ProcessExiting;

EXTERN_C INT MODENTRY DllMain(void *xxx, INT Op, INT Arg3);

/* Marker variables in the .CRT section.
 * Defined in markers.c, used by _CRT_INIT in crtinit.c.
 */

EXTERN_C PF __xi_a[], __xi_z[]; /* C initializers */
EXTERN_C PF __xc_a[], __xc_z[]; /* C++ initializers */
EXTERN_C PF __xp_a[], __xp_z[]; /* C pre-terminators */
EXTERN_C PF __xt_a[], __xt_z[]; /* C terminators */ 


/* Data types and functions prototypes used in atexit.c, doatexit.c and exit.c
 */
typedef void (__cdecl * EXIT_FUNCTION)(void);

typedef struct _ATEXIT_DATA {
    UINT Entries;
    EXIT_FUNCTION Array[1];
} ATEXIT_DATA, *PATEXIT_DATA;

EXTERN_C INT __cdecl _doatexit(EXIT_FUNCTION, ATEXIT_DATA **);


/* crtinit.c
 *
 * This file is included in ansic.lib because it needs access
 * to the marker variables in the image that is calling _CRT_INIT.
 */

INT MODENTRY CrtInit(INT Op)
{
    if (Op == DLL_PROCESS_ATTACH) {
        /* call C++ global constructors */
        _initterm(__xc_a, __xc_z);
    }
    else if (Op == DLL_PROCESS_DETACH) {
        /* clean up */
        _initterm(__xt_a, __xt_z);
        __closestdfile();
    }

    return TRUE;
}


/* init.c
 *
 * This file is included in ansic.lib
 */

/* Walk an array of function pointers, call non-NULL ones.
 */
void __cdecl _initterm(PF *pfbegin, PF *pfend)
{
    while (pfbegin < pfend) {
        if (*pfbegin != NULL)
            (**pfbegin)();
        pfbegin++;
    }
}


/* markers.c
 *
 * This file is included in ansic.lib
 */

/* Need to put the following marker variables into the .CRT section.
 * The .CRT section contains arrays of function pointers.
 * The compiler creates functions and adds pointers to this section
 * for things like C++ global constructors.
 *
 * The XIA, XCA etc are group names with in the section.
 * The compiler sorts the contributions by the group name.
 * For example, .CRT$XCA followed by .CRT$XCB, ... .CRT$XCZ.
 * The marker variables below let us get pointers
 * to the beginning/end of the arrays of function pointers.
 *
 * For example, standard groups are
 *      XCA     used here, for begin marker
 *      XCC     "compiler" inits
 *      XCL     "library" inits
 *      XCU     "user" inits
 *      XCZ     used here, for end marker
 */

#pragma data_seg(".CRT$XCA", "DATA")
PF __xc_a[] = { 0 };                    /* C++ initializers */

#pragma data_seg(".CRT$XCZ", "DATA")
PF __xc_z[] = { 0 };

#if 0 /* not used yet */

#pragma data_seg(".CRT$XIA", "DATA")
PF __xi_a[] = { 0 };                    /* C initializers */

#pragma data_seg(".CRT$XIZ", "DATA")
PF __xi_z[] = { 0 };

#pragma data_seg(".CRT$XPA", "DATA")
PF __xp_a[] = { 0 };                    /* C pre-terminators */

#pragma data_seg(".CRT$XPZ", "DATA")
PF __xp_z[] = { 0 };

#endif

#pragma data_seg(".CRT$XTA", "DATA")
PF __xt_a[] = { 0 };                    /* C terminators */

#pragma data_seg(".CRT$XTZ", "DATA")
PF __xt_z[] = { 0 };
