/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <string.h>
#include <machdep.h>            /* for __MACHDEP_MEMFUNC */

/*
 * Quite often these functions have a strong effect on performance. 
 * They are good candidates for machine-dependent rewrites.
 */
#if !defined(__MACHDEP_MEMFUNC)

#ifdef _MSC_VER
#pragma function(memset)
#undef __MEMFUNC_ARE_INLINED
#endif

#if !defined(__MEMFUNC_ARE_INLINED)

/* Set C bytes at address D to the (8 bit) value V
 */
EXTERN_C void * __cdecl memset(void *d, int v, size_t c)
{
    if ((((ADDRESS) d) | c) & (sizeof(UINT) - 1)) {

        BYTE *pD = (BYTE *) d;
        BYTE *pE = (BYTE *) (((ADDRESS) d) + c);

        while (pD != pE)
            *(pD++) = (BYTE) v;

    }
    else {

        UINT *pD = (UINT *) d;
        UINT *pE = (UINT *) (BYTE *) (((ADDRESS) d) + c);
        UINT uv;

        uv = ((UINT) (v & 0xff)) | (((UINT) (v & 0xff)) << 8);

        /* Our processors are at least 32 bits
         */
        uv |= uv << 16;

#if (_UINTSIZE == 64)
        /* They might be 64 bits
         */
        uv |= uv << 32;
#endif

        while (pD != pE)
            *(pD++) = uv;
    }
    return d;
}

#endif /* ! __MEMFUNC_ARE_INLINED */
#endif /* ! __MACHDEP_MEMFUNC */
