/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <string.h>
#include <tchar.h>

/* Split string into tokens.                                           -jvh
 * Similar to strtok but without re-entrancy and concurrency problems.
 * - the input parameter is zero terminated at the next delimiter.
 * - the found delimiter is overwritten by a zero.
 * - the return value points to the rest of the string (not yet delimited).
 * - the function return NULL if no delimiters were found.
 * - if the first character is a delimiter the input will become empty "".
 * - give the return value as the input for the next round.
 */

_TCHAR * __cdecl _tcssplit(_TCHAR *s1, const _TCHAR *delimit)
{
    /* Find end of segment */
    _TCHAR *next = _tcspbrk(s1, delimit);
    if (next) {
        /* Found another delimiter, split string and return state. */
        *next = '\0';
        return next + 1;
    } else {
        /* Last segment, say so. */
        return NULL;
    }
}
