/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <crypto.h>
#include <mmmacros.h>
#include <base/cobdesc.h>

typedef struct _base64Cypher {
    CONST_VTBL struct ICypherVtbl *v;
    UINT RefCnt;
    UINT Position;
} base64Cypher, *Pbase64_CYPHER;

#define base64CypherConstructorDestroy NULL
typedef struct _DefCob base64CypherConstructor;

#include <s_base64Cypher_v.c>
#include <s_base64CypherConstructor_v.c>


PRIVATE const char Map[] = {
    /* 0 */ 'A',
    /* 1 */ 'B',
    /* 2 */ 'C',
    /* 3 */ 'D',
    /* 4 */ 'E',
    /* 5 */ 'F',
    /* 6 */ 'G',
    /* 7 */ 'H',
    /* 8 */ 'I',
    /* 9 */ 'J',
    /* 10 */ 'K',
    /* 11 */ 'L',
    /* 12 */ 'M',
    /* 13 */ 'N',
    /* 14 */ 'O',
    /* 15 */ 'P',
    /* 16 */ 'Q',
    /* 17 */ 'R',
    /* 18 */ 'S',
    /* 19 */ 'T',
    /* 20 */ 'U',
    /* 21 */ 'V',
    /* 22 */ 'W',
    /* 23 */ 'X',
    /* 24 */ 'Y',
    /* 25 */ 'Z',
    /* 26 */ 'a',
    /* 27 */ 'b',
    /* 28 */ 'c',
    /* 29 */ 'd',
    /* 30 */ 'e',
    /* 31 */ 'f',
    /* 32 */ 'g',
    /* 33 */ 'h',
    /* 34 */ 'i',
    /* 35 */ 'j',
    /* 36 */ 'k',
    /* 37 */ 'l',
    /* 38 */ 'm',
    /* 39 */ 'n',
    /* 40 */ 'o',
    /* 41 */ 'p',
    /* 42 */ 'q',
    /* 43 */ 'r',
    /* 44 */ 's',
    /* 45 */ 't',
    /* 46 */ 'u',
    /* 47 */ 'v',
    /* 48 */ 'w',
    /* 49 */ 'x',
    /* 50 */ 'y',
    /* 51 */ 'z',
    /* 52 */ '0',
    /* 53 */ '1',
    /* 54 */ '2',
    /* 55 */ '3',
    /* 56 */ '4',
    /* 57 */ '5',
    /* 58 */ '6',
    /* 59 */ '7',
    /* 60 */ '8',
    /* 61 */ '9',
    /* 62 */ '+',
    /* 63 */ '/'
};

PRIVATE char PadChar = '=';

/* Return the position in array above or a special token as appropriate */
#define SKIP 0x100
#define PAD  0x200
#define JUNK 0x400
#define SPECIAL 0xf00
PRIVATE UINT Decode(const _TCHAR c)
{
    if (c >= 'A' && c <= 'Z')
        return c - 'A';
    if (c >= 'a' && c <= 'z')
        return c - 'a' + 26;
    if (c >= '0' && c <= '9')
        return c - '0' + 52;
    if (c == '+')
        return 62;
    if (c == '/')
        return 63;
    if (isspace(c))
        return SKIP;
    if (c == PadChar)
        return PAD;
    /* else might signal error... */
    assert(FALSE);
    return JUNK;
}

/* Read one char, decode it, and update count */
PRIVATE UINT GetChar(const _TCHAR **ppData, UINT *pRemains)
{
    UINT Count = *pRemains, val = PAD;
    const _TCHAR *pData = *ppData;
    while (Count > 0) {
        val = Decode(*pData++);
        Count--;
        if (val == PAD) {
            /* Reached end. Stop. */
            //Count = 0;
            break;
        }

        if (val != SKIP || Count < 4)
            break;
        val = PAD;
    }
    *pRemains = Count;
    *ppData = pData;
    return val;
}

PRIVATE SCODE MCT base64CypherCypher(PICYPHER iThis,
    /*in*/ CYPHER_OPS Operation,
    /*in*/ const BYTE * pInBuffer,
    /*in*/ UINT InSize,
    /*out*/ UINT *pSizeConsumed,
    /*out*/ BYTE * pOutBuffer,
    /*in*/ UINT OutSize,
    /*out*/ UINT *pSizeProduced)
{
    UINT d1,d2,d3,d4, count = 0, total = 0;
    Pbase64_CYPHER This = (Pbase64_CYPHER) iThis;
    UINT InSize_Orig = InSize;
    UINT OutSize_Orig = OutSize;

    if (Operation & CYPHER_ENCRYPT) { /* encode */
        _TCHAR *pOut = (_TCHAR*) pOutBuffer;
        /* Check that it fits up front. Unless variable length.
         * 1/3 overhead rounded up to 4 bytes (1/4 overhead seen from output).
         * More overhead from \r\n (64/2)
         */
        if (OutSize < 4)
            return E_BUFFER_TOO_SMALL;
        if (!pSizeConsumed && __TRUNC(OutSize,4) - OutSize/4 - OutSize/32 < InSize)
            return E_BUFFER_TOO_SMALL;

        count = This->Position;
        while ((InSize >= 3 || (Operation & CYPHER_LAST)) && (OutSize > 4)) {
            /* Max 76 chars per line. */
            if (count++ >= 18) { /* 4 at a time */
                *pOut++ = '\r';
                *pOut++ = '\n';
                count = 0;
                total += 2;
                OutSize -= 2;
                continue;
            }
            InSize--;
            OutSize -= 4;
            total += 4;

            d1 = *pInBuffer++;
            /* 6 bits of first input */
            *pOut++ = Map[d1 >> 2];

            d1 &= 0x03;
            d1 <<= 4;
            if (InSize == 0) {
                *pOut++ = Map[d1];
                *pOut++ = PadChar;
                *pOut++ = PadChar;
                break;
            }
            InSize--;

            d2 = *pInBuffer++;
            /* 2 bits of first input, 4 of second */
            *pOut++ = Map[d1 | ((d2 & 0xf0) >> 4)];

            d2 &= 0x0f;
            d2 <<= 2;
            if (InSize == 0) {
                *pOut++ = Map[d2];
                *pOut++ = PadChar;
                break;
            }
            InSize--;

            d3 = *pInBuffer++;
            /* 4 bits of second input, 2 of third */
            *pOut++ = Map[d2 | ((d3 & 0xc0) >> 6)];
            /* 6 bits of third input */
            *pOut++ = Map[d3 & 0x3f];
        }

        /* Put in newline at end if chunk size was right. */
        if (count >= 18 && OutSize >= 2) { /* 4 at a time */
            *pOut++ = '\r';
            *pOut++ = '\n';
            total += 2;
            count = 0;
            OutSize -= 2;
        }

        assert(OutSize == OutSize_Orig - total);
        *pSizeProduced = total;
        if (pSizeConsumed)
            *pSizeConsumed = InSize_Orig - InSize;
        This->Position = count;

    } else { /* decode */
        const _TCHAR *pIn = (const _TCHAR*) pInBuffer;
        if (InSize < 4 && (Operation & CYPHER_LAST))
            return E_INVALID_PARAMETER;
        while (InSize >= 4) {
            d1 = GetChar(&pIn, &InSize);
            if (d1 & SPECIAL) {
                /* ERROR */
                break;
            }
            d2 = GetChar(&pIn, &InSize);
            if (d2 & SPECIAL)
                break;

            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 6,2 */
            *pOutBuffer++ = (UINT8) ((d1 << 2) | (d2 >> 4));

            d3 = GetChar(&pIn, &InSize);
            if (d3 & SPECIAL) {
                d4 = GetChar(&pIn, &InSize);
                /* if not PAD, then ERROR */
                break;
            }
            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 4,4 */
            *pOutBuffer++ = (UINT8) (((d2 & 0xf) << 4) | (d3 >> 2));

            d4 = GetChar(&pIn, &InSize);
            if (d4 & SPECIAL)
                break;

            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 2,6 */
            *pOutBuffer++ = (UINT8) (((d3 & 0x3) << 6) | d4);
        }

        *pSizeProduced = total;
        if (pSizeConsumed)
            *pSizeConsumed = InSize_Orig - InSize;
    }

    return S_OK;
}

PRIVATE SCODE MCT base64CypherGetBlockLengths (
                                 /*in*/ PICYPHER pThis,
                                 /*inout*/ CYPHER_OPS * pOps,
                                 /*inout*/ UINT * pInputBlockLength,
                                 /*inout*/ UINT * pOutputBlockLength
)
{
    UnusedParameter(pThis);

    if (pOps) {
        if (*pOps & CYPHER_ENCRYPT) {
            *pOps |= CYPHER_SIZE_GROWS;
            /* Minimum block size is 4 */
            if (pInputBlockLength)
                *pInputBlockLength = 3;
            if (pOutputBlockLength)
                *pOutputBlockLength = 4;
        }
        if (*pOps & CYPHER_DECRYPT) {
            *pOps |= CYPHER_SIZE_SHRINKS;
            if (pInputBlockLength)
                *pInputBlockLength = 4;
            if (pOutputBlockLength)
                *pOutputBlockLength = 3;
        }
    }

    return S_OK;
}

PRIVATE SCODE MCT base64CypherSetKey (
                 /*in*/ PICYPHER pThis,
                 /*in*/ const BYTE * Key,
                 /*in*/ UINT Size,
                 /*in*/ CYPHER_OPS Operations
)
{
    DBGME(0,_tprintf(_TEXT("base64CypherSK %x %x %d %x\n"),
                     (ADDRESS)pThis, (ADDRESS)Key, Size, Operations));
    return S_OK;
}

PRIVATE SCODE MCT base64CypherDestroy (
        PIUNKNOWN pThis)
{
    PIHEAP pHeap = CurrentHeap();
    DBGME(0,_tprintf(_TEXT("base64CypherDestroy %x\n"),(ADDRESS)pThis));
    return pHeap->v->Free(pHeap,0,pThis);
}

PRIVATE SCODE MCT base64CypherConstructorConstructor(
                             PICYPHERCONSTRUCTOR pThis,
                             /*out*/ PIUNKNOWN* ppUnk,
                             const _TCHAR * Name,
                             UINT KeyStrength)
{
    Pbase64_CYPHER base64Cypher;
    PIHEAP pHeap = CurrentHeap();

    UnusedParameter(Name);

    DBGME(0,_tprintf(_T("base64CypherConstructor(%x %x %s %d)\n"),
                     (ADDRESS) pThis, (ADDRESS) ppUnk, Name, KeyStrength));

    base64Cypher = pHeap->v->Alloc(pHeap,HEAP_ZERO_MEMORY, sizeof(*base64Cypher), 0);
    if (base64Cypher) {
        base64Cypher->v = &base64CypherVtbl.v;
        base64Cypher->RefCnt = 1;
        base64Cypher->Position = 0;
        *ppUnk = (PIUNKNOWN) base64Cypher;
        DBGME(0,_tprintf(_T("base64CypherConstructor returns %x\n"), (ADDRESS) base64Cypher));
        return S_OK;
    }

    DBGME(3,_tprintf(_T("base64CypherConstructor failed\n")));
    return E_NOT_ENOUGH_MEMORY;
}
