/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    drivers.cpp
 *
 * Purpose:
 *
 *    Initialization and control code for MMOSA device drivers.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <mmhal.h>
#define DLL_LINKAGE DLL_EXPORT
#include <drivers/drivers.h>
#include "config.h"

/* `C' style interfaces, defined in _drivers.c
 */
EXTERN_C DLL_EXPORT void * operator_new( UINT size);
EXTERN_C DLL_EXPORT void operator_delete( void *pThis);
EXTERN_C DLL_EXPORT SCODE MCT CDevice_QueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew);
EXTERN_C DLL_EXPORT UINT MCT CDevice_AddRef ( struct CDevice *pThis );
EXTERN_C DLL_EXPORT UINT MCT CDevice_Release ( struct CDevice *pThis );
EXTERN_C DLL_EXPORT void MCT CDevice_ThreadLoop(  struct CDevice *pThis );

EXTERN_C DLL_EXPORT SCODE MCT 
CNetDriver_QueryInterface (struct CNetDriver *pThis, REFIID Iid, void* *ppNew);
EXTERN_C INT MCT CNetDriver_Output(struct CNetDriver *pThis,
                                   struct mbuf *Pkt,
                                   struct sockaddr *Addr,
                                   struct rtentry *Route);

void * __cdecl operator new(size_t size)
{
    return operator_new(size);
}

void __cdecl operator delete(void *p)
{
    operator_delete(p);
}

/* Methods for the CDevice class
 */

/* This truly is IFile::QueryInterface, but the decl for
 * the IFile class does not allow it.
 * The other catch is reference counting.
 */
SCODE MCT CDevice::QueryInterface (REFIID Iid, void* *ppNew)
{
    return CDevice_QueryInterface(this, Iid, ppNew);
}

UINT MCT CDevice::AddRef (void)
{
    return CDevice_AddRef( this );
}

UINT MCT CDevice::Release (void)
{
    return CDevice_Release(this);
}

SCODE MCT CDevice::ReadAt (UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeRead);

    return E_ACCESS_DENIED; /* ? */
}

SCODE MCT CDevice::WriteAt (UINT64 Position, const BYTE *pBuffer,
                               UINT Size, PUINT pSizeWritten)
{
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeWritten);

    return E_ACCESS_DENIED; /* ? */
}

SCODE MCT CDevice::SetSize( UINT64 Size)
{
    UnusedParameter(Size);
    return E_ACCESS_DENIED; /* ? */
}

SCODE MCT CDevice::GetSize( PUINT64 pSize)
{
    UnusedParameter(pSize);
    return E_ACCESS_DENIED; /* ? */
}

BOOL MCT CDevice::Isr(PBOOL pWakeup)
{
    UnusedParameter(pWakeup);
    return FALSE;
}

BOOL MCT CDevice::HandleInterrupt()
{
    return FALSE;
}

INT  MCT CDevice::Initialize()
{
    return 0;
}

void MCT CDevice::ThreadLoop()
{
    CDevice_ThreadLoop(this);
}

/* Methods for the CNetDriver class
 */
SCODE MCT CNetDriver::QueryInterface (REFIID Iid, void* *ppNew)
{
    return CNetDriver_QueryInterface(this, Iid, ppNew);
}

INT MCT CNetDriver::Output(struct mbuf *Pkt, struct sockaddr *Addr, struct rtentry *Route)
{
    return CNetDriver_Output(this, Pkt, Addr, Route);
}

void MCT CNetDriver::Start(void)
{
}

INT  MCT CNetDriver::Ioctl( UINT Cmd, PTR Data)
{
    UnusedParameter(Cmd);
    UnusedParameter(Data);
    return 0;
}

INT  MCT CNetDriver::Watchdog(void)
{
    return 0;
}
