/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <driverif.h>
#include <assert.h>
#include <diagnostics.h>
#include <drivers/drivers.h>
#include "config.h"

extern DRIVER_CONFIG DriverConfigs[];

void ConfigureDrivers( PIMODULE pModule,
                       _TCHAR *DriverName,
                       BOOL AutoLoad )
{
    PDRIVER_CONFIG pc;
    PINAMESPACE ns = CurrentNameSpace();
    PIUNKNOWN pUnk;
    PIDEVICECONSTRUCTOR pDevCons;
    PIDEVICE pDev;

    SCODE sc;

    for (pc = DriverConfigs; pc->Name; pc++) {
        /* Skip this entry when appropriate */
        if (AutoLoad
            ? (pc->Flags & DEVICE_FLAGS_MANUAL_START)
            : _tcscmp(DriverName, pc->Name))
            continue;

        /* See if it already loaded and skip if so */
        sc = ns->v->Bind(ns, pc->Name, 0, &pUnk);
        if (SUCCEEDED(sc)) {
            DBGME(3,printf("ConfigureDrivers: %s already loaded, skipped\n",
                           pc->Name));
            pUnk->v->Release(pUnk);
            continue;
            
        }

        sc = ns->v->Bind(ns, pc->CobName, 0, &pUnk);
        if (FAILED(sc)) {       /* cob not found */
            DBGME(3,printf("ConfigureDrivers: %s not loaded sc=%x\n", pc->CobName, sc));
            continue;
        }
        sc = pUnk->v->QueryInterface(pUnk, &IID_IDeviceConstructor,
                                     (void **) &pDevCons);
        pUnk->v->Release(pUnk);
        if (FAILED(sc)) {
            DBGME(3,printf("cob %s missing constructor\n",pc->CobName));
            continue;
        }
        sc = pDevCons->v->Constructor(pDevCons, pc->Name, pc->RegisterBase, pc->Interrupt, &pUnk, pc->s1, pc->s2, pc->u1, pc->u2, (ADDRESS) pc->u3, pc->u4, pc->u5);
        if (FAILED(sc)) {    /* constructor failed */
            DBGME(3,printf("ConfigureDrivers: %s constructor failed sc=%x\n",
                   pc->CobName, sc));
            continue;
        }
        pDevCons->v->Release(pDevCons);

        if (pc->Flags & DEVICE_FLAGS_NET_DRIVER)
            sc = pUnk->v->QueryInterface(pUnk, &IID_INetDriver,
                                         (void **) &pDev);
        else
            sc = pUnk->v->QueryInterface(pUnk, &IID_IDevice,
                                         (void **) &pDev);
        if (FAILED(sc)) {
            DBGME(3,printf("cob %s not a device driver (sc=%x). Skipping ConnectDevice()\n",
                           pc->CobName,sc));
            pDev = (PIDEVICE) pUnk;
        } else {
            pUnk->v->Release(pUnk);
            sc = ConnectDevice((PCDEVICE) pDev, pc->Name, pc->Flags, pc->RegisterBase, pc->Interrupt);
            if (FAILED(sc)) {
                DBGME(3,printf("Could not connect %s, sc=x%x, ignoring\n",
                            pc->Name, sc));
                /* ignore, continue */
            }
        }

        /* Register new device in name space (XXX should be flag?) */
        sc = ns->v->Register(ns, pc->Name, (PIUNKNOWN) pDev,
                             NAME_SPACE_READ | NAME_SPACE_WRITE, NULL);
        if (FAILED(sc)) {
            DBGME(3,printf("Failed to register %s, sc=x%x, ignoring.\n",
                           pc->Name, sc));
        }
        pDev->v->Release(pDev);

        /* Say what happened */
        DBGME(2,printf("%s started ok.\n", pc->Name));
    }
}
