/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    pnp_table.c
 *
 * Purpose:
 *
 *    Plug aNd Play manager, lookup table.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

const DRIVER_NAME DriverTable[] = {
#define USB_MOUSE 0
    _T("usbmouse"),
#define USB_KEYBOARD 1
    _T("usbkeybd"),
#define USB_HUB 2
    _T("usbhub")
#if defined(SMALL_CUSTOM_TABLE)
#else
    ,
#define USB_STOR 3
    _T("usbstor"),
#define USB_KVM 4
    _T("usbkvm")
#endif
};


/* WARNING:
 * This table most be sorted or it wont work.
 * It MUST also NOT have duplicates
 */
const PNP_ENTRY PnpTable[] = {
#if defined(SMALL_CUSTOM_TABLE)
    { 0x0400, 0x1237, ANY_ID, USB_HUB }, // National Semi USB Hub
    { 0x045e, 0x0009, 0x0105, USB_MOUSE },   // Msft Intellimouse
    { 0x045e, 0x000b, 0x0207, USB_KEYBOARD } // Msft Natural Keyboard Elite
#else
{0x03ee,0x5609,0x0000,USB_KEYBOARD},//Mitsumi Japanese USB Keyboard
{0x03F9,0x0100,ANY_ID,USB_KEYBOARD},//Key Tronic USB Keyboard
{0x03F9,0x0101,ANY_ID,USB_KEYBOARD},//Key Tronic USB Keyboard
{0x03F9,0x0102,0x0000,USB_KEYBOARD},//Key Tronic USB Keyboard
{0x0400,0x1237,ANY_ID,USB_HUB},//National Semiconductor USB Hub
{0x0405,0x0000,ANY_ID,USB_HUB},//Sand USB Hub
{0x0409,0x0011,0x0000,USB_KEYBOARD},//NEC PC98 Series Layout USB Keyboard
{0x0409,0x0014,0x0000,USB_KEYBOARD},//NEC 109 Japanese USB Keyboard
{0x0409,0x0019,ANY_ID,USB_HUB},//NEC 109 Japanese USB Keyboard with Bus-Powered Hub
{0x0409,0x001A,ANY_ID,USB_HUB},//NEC PC98 Series Layout USB Keyboard with Bus-Powered Hub
{0x0409,0x0025,ANY_ID,USB_HUB},//NEC USB Mini Keyboard with Bus-Powered Hub
{0x0409,0x0034,ANY_ID,USB_KEYBOARD},//NEC 109 Japanese USB Keyboard with One-touch start buttons
{0x0409,0x003F,0x0000,USB_KEYBOARD},//NEC Wireless USB Keyboard with One-touch start buttons
{0x0409,0x004F,0x0000,USB_KEYBOARD},//NEC Wireless USB Keyboard with One-touch start buttons
{0x0409,0x0094,ANY_ID,USB_KEYBOARD},//NEC Japanese USB Keyboard with One-touch start buttons
{0x0409,0x0095,ANY_ID,USB_HUB},//NEC 109 USB Small Keyboard with Bus-Powered Hub
{0x0409,0x55AA,ANY_ID,USB_HUB},//NEC USB Hub
{0x0409,0x8010,ANY_ID,USB_HUB},//NEC Intellibase USB Hub
{0x0409,0x8011,ANY_ID,USB_HUB},//NEC Intellibase USB Hub
{0x0416,0x0101,ANY_ID,USB_HUB},//Winbond USB Hub
{0x0419,0x8001,ANY_ID,USB_HUB},//Samsung USB Hub
{0x0423,0x1237,ANY_ID,USB_HUB},//Computer Access Technology Corporation USB Hub
{0x042B,0x9316,ANY_ID,USB_HUB},//Intel 8x931Hx Customer Hub
{0x0430,0x0002,0x0000,USB_KEYBOARD},//Fujitsu USB Composite Keyboard
{0x0430,0x000A,0x0000,USB_KEYBOARD},//Fujitsu 109 Japanese USB Keyboard
{0x0430,0x000B,ANY_ID,USB_KEYBOARD},//Fujitsu 109 Japanese USB Keyboard
{0x0430,0x0082,ANY_ID,USB_KEYBOARD},//Fujitsu 109 Japanese USB Keyboard
{0x0430,0x0083,0x0000,USB_KEYBOARD},//Fujitsu 109 Japanese USB Keyboard
{0x0430,0x36BA,ANY_ID,USB_HUB},//Fujitsu USB Bus Powered Hub
{0x0431,0x0100,ANY_ID,USB_MOUSE},//Evolution USB Mouse-Trak by ITAC
{0x0433,0xABAB,ANY_ID,USB_KEYBOARD},//ALPS USB Keyboard
{0x0446,0x6781,ANY_ID,USB_MOUSE},//NMB USB Keyboard Mouse
{0x0446,0x6782,ANY_ID,USB_KEYBOARD},//NMB USB Keyboard
{0x044e,0x1104,ANY_ID,USB_KEYBOARD},//ALPS Japanese USB Keyboard
{0x0451,0x1446,ANY_ID,USB_HUB},//General Purpose USB Hub
{0x0458,0x0001,ANY_ID,USB_MOUSE},//Genius USB Mouse
{0x0458,0x0002,ANY_ID,USB_MOUSE},//Genius USB Net Mouse Pro
{0x0458,0x0003,ANY_ID,USB_MOUSE},//Genius USB Wheel Mouse
{0x045E,0x0009,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse
{0x045E,0x000B,ANY_ID,USB_KEYBOARD},//Microsoft USB Natural Keyboard
{0x045E,0x001C,ANY_ID,USB_HUB},//Microsoft Integrated USB Hub
{0x045E,0x001D,0x0000,USB_KEYBOARD},//Microsoft USB Natural Keyboard Pro
{0x045E,0x001E,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse Explorer
{0x045E,0x0025,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse with IntelliEye
{0x045E,0x0029,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse Web
{0x045E,0x002B,0x0000,USB_KEYBOARD},//Microsoft USB Internet Keyboard Pro
{0x045E,0x002D,0x0000,USB_KEYBOARD},//Microsoft Internet Keyboard
{0x045E,0x002F,ANY_ID,USB_HUB},//Microsoft Integrated USB Hub
{0x045E,0x0039,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse Optical
{0x045E,0x0040,ANY_ID,USB_MOUSE},//Microsoft USB Wheel Mouse Optical
{0x045E,0x0047,ANY_ID,USB_MOUSE},//Microsoft USB IntelliMouse Explorer 3.0
{0x045E,0x0059,ANY_ID,USB_MOUSE},//Microsoft USB Wireless IntelliMouse Explorer
{0x045E,0x005C,0x0000,USB_KEYBOARD},//Microsoft USB Office Keyboard (106/109)
{0x046A,0x0001,ANY_ID,USB_KEYBOARD},//Cherry GmbH USB Keyboard
{0x046A,0x0003,ANY_ID,USB_HUB},//Cherry GmbH USB-Hub
{0x046D,0xC000,ANY_ID,USB_MOUSE},//Logitech USB First/Pilot Mouse
{0x046D,0xC001,ANY_ID,USB_MOUSE},//Logitech USB First/Pilot Mouse+
{0x046D,0xC002,ANY_ID,USB_MOUSE},//Logitech USB MouseMan Wheel
{0x046D,0xC003,ANY_ID,USB_MOUSE},//Logitech USB MouseMan
{0x046D,0xC004,ANY_ID,USB_MOUSE},//Logitech USB WingMan Gaming Mouse
{0x046D,0xC005,ANY_ID,USB_MOUSE},//Logitech USB WingMan Gaming Wheel Mouse
{0x046D,0xC00B,ANY_ID,USB_MOUSE},//Logitech USB MouseMan Wheel+
{0x046D,0xC00C,ANY_ID,USB_MOUSE},//Logitech USB WheelMouse
{0x046D,0xC00D,ANY_ID,USB_MOUSE},//Logitech USB MouseMan Wheel+
{0x046D,0xC00E,ANY_ID,USB_MOUSE},//Logitech USB Wheel Mouse
{0x046D,0xC030,ANY_ID,USB_MOUSE},//Logitech USB iFeel Mouse
{0x046D,0xC031,ANY_ID,USB_MOUSE},//Logitech USB iFeel Mouse+
{0x046D,0xC032,ANY_ID,USB_MOUSE},//Logitech USB iFeel MouseMan
{0x046D,0xC033,ANY_ID,USB_MOUSE},//Logitech USB iFeel MouseMan+
{0x046D,0xC402,ANY_ID,USB_MOUSE},//Logitech USB Marble Mouse
{0x046D,0xC501,ANY_ID,USB_MOUSE},//Logitech USB Cordless Mouse
{0x046D,0xC502,ANY_ID,USB_MOUSE},//Logitech HID Cordless Mouse+
{0x046D,0xC503,0x0000,USB_KEYBOARD},//Logitech USB Cordless Keyboard
{0x046E,0x0100,ANY_ID,USB_KEYBOARD},//BTC USB Keyboard
{0x046E,0x6782,0x0000,USB_KEYBOARD},//BTC USB Keyboard
{0x0471,0x0201,ANY_ID,USB_HUB},//Philips USB Hub
{0x0471,0x0401,ANY_ID,USB_KEYBOARD},//Philips Semiconductors CICT USB Keyboard
{0x0471,0x0402,0x0000,USB_KEYBOARD},//Philips Semiconductors CICT USB Keyboard
{0x0471,0x0601,0x0000,USB_KEYBOARD},//Philips USB Wireless Keyboard
{0x0471,0x0700,ANY_ID,USB_HUB},//Philips Semiconductors CICT USB Hub
{0x047B,0x0001,ANY_ID,USB_KEYBOARD},//Silitek USB Keyboard
{0x047B,0x0002,0x0000,USB_KEYBOARD},//Silitek USB Keyboard
{0x04A5,0x0003,ANY_ID,USB_MOUSE},//API Generic USB K/B Mouse
{0x04A6,0x0180,ANY_ID,USB_HUB},//Nokia USB Hub Type P
{0x04a8,0x0101,ANY_ID,USB_HUB},//Multi{0xeo Labs USB Hub
{0x04B4,0x0001,ANY_ID,USB_MOUSE},//Cypress USB Mouse
{0x04C3,0x1102,ANY_ID,USB_MOUSE},//Maxi Switch, Inc. #1102 Mouse
{0x04C3,0x2102,ANY_ID,USB_MOUSE},//Maxi Switch, Inc. #2102 Mouse
{0x04c5,0x1018,ANY_ID,USB_KEYBOARD},//Fujitsu Japanese USB Keyboard Hot Buttons
{0x04c5,0x1020,0x0000,USB_KEYBOARD},//Fujitsu Japanese USB Composite Keyboard
{0x04c5,0x1022,0x0000,USB_KEYBOARD},//Fujitsu 109 Japanese USB Keyboard
{0x04CA,0x9304,ANY_ID,USB_HUB},//Lite On USB Hub
{0x04E1,0x0201,ANY_ID,USB_HUB},//Iiyama USB Hub
{0x04F2,0x0001,0x0000,USB_KEYBOARD},//Chicony USB Keyboard
{0x04F2,0x0002,ANY_ID,USB_KEYBOARD},//Chicony USB Keyboard
{0x0500,0x0001,0x0000,USB_KEYBOARD},//SUH DART USB Keyboard
{0x0500,0x0002,ANY_ID,USB_KEYBOARD},//SUH DART-2 USB Keyboard
{0x050F,0x0001,ANY_ID,USB_HUB},//KC Technology USB Hub
{0x050D,0x0217,ANY_ID,USB_HUB},//Belkin USB 2.0 Hub
{0x0510,0x0001,ANY_ID,USB_KEYBOARD},//Sejin USB Keyboard
{0x0510,0x1000,ANY_ID,USB_MOUSE},//Sejin USB Keyboard Mouse
{0x0510,0xE001,ANY_ID,USB_MOUSE},//Sejin USB Mouse
{0x054C,0x0001,ANY_ID,USB_HUB},//Sony USB HUB
{0x054C,0x0002,ANY_ID,USB_HUB},//Standard USB HUB
{0x0557,0x2205,ANY_ID,USB_KVM},//ATEN Keyboard in USB KVM switch
{0x055D,0x0001,ANY_ID,USB_KEYBOARD},//Samsung USB Keyboard
{0x055D,0x6780,ANY_ID,USB_KEYBOARD},//Samsung USB Keyboard V1
{0x055D,0x6781,0x0000,USB_KEYBOARD},//Samsung USB Keyboard V2
{0x055D,0x8001,ANY_ID,USB_HUB},//Samsung E.M. USB Hub
{0x0566,0x2801,ANY_ID,USB_MOUSE},//MIC USB K/B Mouse
{0x056D,0x0000,ANY_ID,USB_HUB},//EIZO USB Hub
{0x058F,0x9380,ANY_ID,USB_STOR},// JMTEK UsbDrive v1.1
{0x05A4,0x9720,0x0000,USB_KEYBOARD},//ORTEK Keyboard
{0x05A4,0x9722,ANY_ID,USB_KEYBOARD},//ORTEK Keyboard
{0x05AC,0x020B,ANY_ID,USB_KEYBOARD},//Apple Extended USB keyboard (Mitsumi)
{0x05AC,0x0307,ANY_ID,USB_MOUSE},//Apple Optical USB Mouse (Logitech)
{0x05AC,0x1003,ANY_ID,USB_HUB},//Mitsumi HUB in Apple keyboard
{0x05AF,0x9267,ANY_ID,USB_MOUSE},//KB 9251B - 678 Mouse
{0x05B2,0x6200,ANY_ID,USB_KEYBOARD},//Focus USB Keyboard
{0x05B2,0x7200,0x0000,USB_KEYBOARD},//Focus USB Keyboard
{0x05BC,0x0001,ANY_ID,USB_MOUSE},//Forward USB K/B Mouse
{0x05C7,0x1001,ANY_ID,USB_MOUSE},//Qtronix USB Mouse
{0x05C7,0x2001,ANY_ID,USB_KEYBOARD},//Qtronix USB Keyboard
{0x05d5,0x6782,ANY_ID,USB_MOUSE},//SUPGERGATE USB Keyboard Mouse
{0x05F2,0x0010,ANY_ID,USB_MOUSE},//Dexin USB AQ Mouse
{0x05FA,0x3301,ANY_ID,USB_MOUSE},//STSL USB Keyboard Mouse
{0x05FA,0x3302,ANY_ID,USB_KEYBOARD},//STSL USB Keyboard
{0x05FA,0x3303,ANY_ID,USB_MOUSE},//STSL USB Keyboard Mouse
{0x0605,0x0001,ANY_ID,USB_MOUSE},//Anchor USB Mouse
{0x060B,0x1006,0x0000,USB_KEYBOARD},//Japanese USB Keyboard - 260U
{0x060B,0x2101,ANY_ID,USB_MOUSE},//Solid Year USB Keyboard Mouse
{0x060B,0x5903,ANY_ID,USB_KEYBOARD},//Japanese USB Keyboard - 595U
{0x060B,0x6003,0x0000,USB_KEYBOARD},//Japanese USB Keyboard - 600HM
{0x0654,0x0006,ANY_ID,USB_HUB},//Granite Microsystems USB Hub
{0x0654,0x0016,ANY_ID,USB_HUB},//Granite Microsystems USB Hub
{0x0698,0x9999,ANY_ID,USB_HUB},//CTX USB Hub
{0x06D5,0x4000,ANY_ID,USB_KEYBOARD},//Toshiba USB 109 Japanese keyboard
{0x0a91,0x3001,0x0001,USB_KEYBOARD},//Globlink USB wireless keyboard
{0x8086,0x9303,ANY_ID,USB_HUB}    //Intel 8x930Hx USB Hub
#endif
};

