/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements GianoMs concrete class */
#include <mmlite.h>
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>

extern PIUNKNOWN GianoMsNew(
    ADDRESS BaseAddress);

/* ** 
*/
SCODE __cdecl GianoMsConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS RegisterBase,
        ADDRESS Unused,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    DBGME(0,_tprintf(_T("GianoMsConstructor(%x %s %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk));

    pUnk = GianoMsNew(RegisterBase);

    DBGME(1,_tprintf(_T("GianoMsConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("GianoMsConstructor failed\n")));
        return E_FAIL;
    }
}

/* IUnknown::QI.  Return a pointer of the right kind. */
PRIVATE SCODE MCT DriverCobQueryInterface(PIDEVICECONSTRUCTOR pThis, REFIID Iid,
                                       void **ppObj)
{
    return GenericQueryInterface((PIUNKNOWN) pThis, Iid, ppObj,
                                 &IID_IDeviceConstructor);
}

/* IUnknown::AddRef */
PRIVATE UINT MCT DriverCobAddRef(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    return AtomicInc(&x->RefCnt);
}

/* IUnknown::Release */
PRIVATE UINT MCT DriverCobRelease(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    /* NOTE: If driver has global destructors, it will have to override
     * this method with one that calls CrtExit() when RefCnt == 0.
     */
    return AtomicDec(&x->RefCnt);
}

static struct IDeviceConstructorVtbl GianoMsDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    DriverCobQueryInterface,
    DriverCobAddRef,
    DriverCobRelease,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) GianoMsConstructor
};

static struct _DRIVERCOB GianoMsDriverCobObject = { &GianoMsDriverCobVtable, 1 };
extern struct IDeviceVtbl GianoMsVtbl;

PIUNKNOWN GianoMsCobMain(void)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("GianoMsCobMain()\n")));

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &GianoMsDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &GianoMsVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    GianoMsVtbl.ThreadLoop = NULL;

    return (PIUNKNOWN) &GianoMsDriverCobObject;
}
