/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Defines for CS8900A chip, according to Cirrus Logic specs.
 * Symbolic names as per their document.
 */

/* I/O Mode register defines
 */
#define CS_IO_DATA              0x0000
#define CS_IO_DATA1             0x0002
#define CS_IO_TxCMD             0x0004
#define CS_IO_TxLength          0x0006
#define CS_IO_ISQ               0x0008
#define CS_IO_PP_PTR            0x000a
#define CS_IO_PP_DATA           0x000c
#define CS_IO_PP_DATA1          0x000e

/* PacketPage memory map
 */

#define CS_PP_BUS_INTERFACE     0x0000
#define CS_PP_STATUS            0x0100
#define CS_PP_XMIT              0x0140
#define CS_PP_FILTER            0x0150
#define CS_PP_RECEIVE_FRAME     0x0400
#define CS_PP_XMIT_FRAME        0x0a00

/* Register Offsets
 */

#define CS_PID                  0x0000
#define CS_REV                  0x0002
#define CS_IO_BASE              0x0020
#define CS_INTR                 0x0022
#define CS_DMA_CH               0x0024
#define CS_DMA_SOF              0x0026
#define CS_DMA_FCOUNT           0x0028
#define CS_DMA_BCOUNT           0x002a
#define CS_MEMORY_BASE          0x002c
#define CS_PROM_BASE            0x0030
#define CS_PROM_SIZE            0x0034
#define CS_EE_CMD               0x0040
#define CS_EE_DATA              0x0042
#define CS_RECV_BCOUNT          0x0044

#define CS_RxCFG                0x0102
#define CS_RxCTL                0x0104
#define CS_TxCFG                0x0106
#define CS_TxCMD                0x0108 /* same as TxCMD_Req, r-o */
#define CS_BufCFG               0x010A
#define CS_LineCTL              0x0112
#define CS_SelfCTL              0x0114
#define CS_BusCTL               0x0116
#define CS_TestCTL              0x0118

#define CS_ISQ                  0x0120
#define CS_RxEvent              0x0124
#define CS_TxEvent              0x0128
#define CS_BufEvent             0x012c
#define CS_RxMISS               0x0130
#define CS_TxCOL                0x0132
#define CS_LineST               0x0134
#define CS_SelfST               0x0136
#define CS_BusST                0x0138
#define CS_TDR                  0x013c

#define CS_TxCMD_Req            0x0144 /* same as TxCMD, w-o */
#define CS_TxLength             0x0146 /* w-o */
#define CS_LAFIlter             0x0150 /* 8 bytes */
#define CS_StationAddress       0x0158 /* 6 bytes */

#define CS_RxStatus             0x0400 /* Same bits as RxEvent */
#define CS_RxLength             0x0402

/* Product ID
 */
#define CS_PID_CS8900           0x630e /* says 0e630000 but reads so */
#define CS_PID_REV_MASK         0x1F00 /* again, bswapped            */
#define CS_PID_REV_B            0x0700
#define CS_PID_REV_C            0x0800

/* IO Base
 */
#define CS_IO_BASE_RESET        0x0300

/* Interrupt Number
 */
#define CS_INTR_0               0x0000  /* IRQ5 on ISA */
#define CS_INTR_1               0x0001  /* IRQ10 on ISA */
#define CS_INTR_2               0x0002  /* IRQ11 on ISA */
#define CS_INTR_3               0x0003  /* IRQ12 on ISA */
#define CS_INTR_NONE            0x0004

/* DMA Channel Number
 */
#define CS_DMA_CH_0             0x0000  /* DMA5 on ISA */
#define CS_DMA_CH_1             0x0001  /* DMA6 on ISA */
#define CS_DMA_CH_2             0x0002  /* DMA7 on ISA */
#define CS_DMA_CH_NONE          0x0003

/* Boot PROM Address Mask (Size really)
 */
#define CS_PROM_SIZE_MASK       0x000ff000
# define CS_PROM_SIZE_4Kbits    0x000ff000
# define CS_PROM_SIZE_8Kbits    0x000fe000
# define CS_PROM_SIZE_16kbits   0x000fc000

/* EEPROM Command
 */
#define CS_EE_CMD_ADD_MASK      0x00ff
#define CS_EE_CMD_CMD_MASK      0x0300
# define CS_EE_CMD_READ_REG      0x0200
# define CS_EE_CMD_WRITE_REG     0x0100
# define CS_EE_CMD_ERASE_REG     0x0300
# define CS_EE_CMD_ERASE         0x0000
#define CS_EE_CMD_ELSEL         0x0400


/* Receiver configuration
 */
#define CS_RxCFG_ID             0x0003
#define CS_RxCFG_Skip_1         0x0040
#define CS_RxCFG_StreamE        0x0080
#define CS_RxCFG_RxOKiE         0x0100
#define CS_RxCFG_RxDMAonly      0x0200
#define CS_RxCFG_AutoRxDMAE     0x0400
#define CS_RxCFG_BufferCRC      0x0800
#define CS_RxCFG_CRCerroriE     0x1000
#define CS_RxCFG_RuntiE         0x2000
#define CS_RxCFG_ExtradataiE    0x4000

/* Receiver control
 */
#define CS_RxCTL_ID             0x0005
#define CS_RxCTL_IAHashA        0x0040
#define CS_RxCTL_PromiscuousA   0x0080
#define CS_RxCTL_RxOKA          0x0100
#define CS_RxCTL_MulticastA     0x0200
#define CS_RxCTL_IndividualA    0x0400
#define CS_RxCTL_BroadcastA     0x0800
#define CS_RxCTL_CRCerrorA      0x1000
#define CS_RxCTL_RuntA          0x2000
#define CS_RxCTL_ExtradataA     0x4000


/* Transmit configuration
 */
#define CS_TxCFG_ID             0x0007
#define CS_TxCFG_LostCRSiE      0x0040
#define CS_TxCFG_SQEerroriE     0x0080
#define CS_TxCFG_TxOKiE         0x0100
#define CS_TxCFG_OutOfWindow    0x0200
#define CS_TxCFG_JabberiE       0x0400
#define CS_TxCFG_AnycolliE      0x0800
#define CS_TxCFG_16colliE       0x8000

/* Transmit command
 */
#define CS_TxCMD_ID             0x0009
#define CS_TxCMD_TxStart_5      0x0000
#define CS_TxCMD_TxStart_381    0x0040
#define CS_TxCMD_TxStart_1021   0x0080
#define CS_TxCMD_TxStart_full   0x00c0
#define CS_TxCMD_Force          0x0100
#define CS_TxCMD_Onecoll        0x0200
#define CS_TxCMD_InhibitCRC     0x1000
#define CS_TxCMD_TxPadDis       0x2000

/* Buffer configuration
 */
#define CS_BufCFG_ID            0x000b
#define CS_BufCFG_SWintX        0x0040
#define CS_BufCFG_RxDMAiE       0x0080
#define CS_BufCFG_Rdy4TxiE      0x0100
#define CS_BufCFG_TxUnderruniE  0x0200
#define CS_BufCFG_RxMissiE      0x0400
#define CS_BufCFG_Rx128iE       0x0800
#define CS_BufCFG_TxColOvfliE   0x1000
#define CS_BufCFG_MissOvfliE    0x2000
#define CS_BufCFG_RxDestiE      0x8000

/* Line control
 */
#define CS_LineCTL_ID           0x0013
#define CS_LineCTL_SerRxON      0x0040
#define CS_LineCTL_SerTxON      0x0080
#define CS_LineCTL_AUIOnly      0x0100
#define CS_LineCTL_AutoSelect   0x0300
#define CS_LineCTL_10BaseT      0x0000
#define CS_LineCTL_ModBackoff   0x0800
#define CS_LineCTL_PolarityDis  0x1000
#define CS_LineCTL_2partDefDis  0x2000
#define CS_LineCTL_LoRxSquelch  0x4000

/* Self control (sic)
 */
#define CS_SelfCTL_ID           0x0015
#define CS_SelfCTL_RESET        0x0040
#define CS_SelfCTL_SWSuspend    0x0100
#define CS_SelfCTL_HWSleepE     0x0200
#define CS_SelfCTL_HWStandbyE   0x0400
#define CS_SelfCTL_HC0E         0x1000
#define CS_SelfCTL_HC1E         0x2000
#define CS_SelfCTL_HCB0         0x4000
#define CS_SelfCTL_HCB1         0x8000

/* Bus control
 */
#define CS_BusCTL_ID            0x0017
#define CS_BusCTL_ResetRxDMA    0x0040
#define CS_BusCTL_DMAextend     0x0100
#define CS_BusCTL_UseSA         0x0200
#define CS_BusCTL_MemoryE       0x0400
#define CS_BusCTL_DMABurst      0x0800
#define CS_BusCTL_IOCHRDYE      0x1000
#define CS_BusCTL_RxDMASize     0x2000
#define CS_BusCTL_EnableIRQ     0x8000

/* Test control
 */
#define CS_TestCTL_ID           0x0019
#define CS_TestCTL_DisableLT    0x0080
#define CS_TestCTL_ENDECLoop    0x0200
#define CS_TestCTL_AUILoop      0x0400
#define CS_TestCTL_DisBackoff   0x0800
#define CS_TestCTL_FDX          0x4000

/* Interrupt Status Queue
 */
#define CS_ISQ_RegNum           0x003f
#define CS_ISQ_RegContent       0xffc0

/* Receiver Event
 */
#define CS_RxEvent_ID           0x0004
#define CS_RxEvent_IAHash       0x0040
#define CS_RxEvent_DribbleBits  0x0080
#define CS_RxEvent_RxOK         0x0100
#define CS_RxEvent_Hashed       0x0200
#define CS_RxEvent_HashIxMask   0xfc00
#define CS_RxEvent_Individual   0x0400
#define CS_RxEvent_Broadcast    0x0800
#define CS_RxEvent_CRCError     0x1000
#define CS_RxEvent_Runt         0x2000
#define CS_RxEvent_Extradata    0x4000

/* Transmitter Event
 */
#define CS_TxEvent_ID           0x0008
#define CS_TxEvent_LossOfCRS    0x0040
#define CS_TxEvent_SQEerror     0x0080
#define CS_TxEvent_TxOK         0x0100
#define CS_TxEvent_OutOfWindow  0x0200
#define CS_TxEvent_Jabber       0x0400
#define CS_TxEvent_nTxColl      0x7100
# define CS_TxEvent_nTxCollShift 11
#define CS_TxEvent_16coll       0x8000

/* Buffer Event
 */
#define CS_BufEvent_ID          0x000c
#define CS_BufEvent_SWint       0x0040
#define CS_BufEvent_RxDMAFrame  0x0080
#define CS_BufEvent_Rdy4Tx      0x0100
#define CS_BufEvent_TxUnderrun  0x0200
#define CS_BufEvent_RxMiss      0x0400
#define CS_BufEvent_Rx128       0x0800
#define CS_BufEvent_RxDest      0x8000

/* Receiver Miss Count
 */
#define CS_RxMISS_ID            0x0010
#define CS_RxMISS_Shift         6
#define CS_RxMISS_Mask          0x3ff   /* after shift */

/* Transmit Collision Counter
 */
#define CS_TxCOL_ID             0x0012
#define CS_TxCOL_Shift          6
#define CS_TxCOL_Mask           0x3ff   /* after shift */

/* Line Status
 */
#define CS_LineST_ID            0x0014
#define CS_LineST_LinkOK        0x0080
#define CS_LineST_AUI           0x0100
#define CS_LineST_10BT          0x0200
#define CS_LineST_PolarityOK    0x1000
#define CS_LineST_CRS           0x4000

/* Self Status
 */
#define CS_SelfST_ID            0x0016
#define CS_SelfST_3p3VActive    0x0040
#define CS_SelfST_INITD         0x0080
#define CS_SelfST_SIBUSY        0x0100
#define CS_SelfST_EEPresent     0x0200
#define CS_SelfST_EEOK          0x0400
#define CS_SelfST_ELPresent     0x0800
#define CS_SelfST_EEsize        0x1000

/* Bus Status
 */
#define CS_BusST_ID             0x0018
#define CS_BusST_TxBidErr       0x0080
#define CS_BusST_Rdy4YxNOW      0x0100

/* AUI Time Domain Reflectometer
 */
#define CS_TDR_ID               0x001c
#define CS_TDR_Shift            6
#define CS_TDR_Mask             0x3ff   /* after shift */

