/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Defines for the 8255x Ethernet COntroller.
 * Applicable to the 82551/0/9/8/7/2 chips, according to Intel manuals.
 */

/* PCI signature
 */
#define IX_8255x_PCI_SIGNATURE  0x12298086

/* Revision numbers (in the PCI config revision register)
 */
#define IX_82557_A              0x01
#define IX_82557_B              0x02
#define IX_82557_C              0x03
#define IX_82558_A              0x04
#define IX_82558_B              0x05
#define IX_82559_A              0x06
#define IX_82559_B              0x07
#define IX_82559_C              0x08 /* NB: used when testing the driver */
#define IX_82559ER_A            0x09
#define IX_82550_A              0x0c
#define IX_82550_B              0x0d
#define IX_82550_C              0x0e
#define IX_82551_A              0x0f
#define IX_82551_B              0x10
#define IX_MAXREV               0x11

/* Control / Status Register
 */
typedef struct _IX_CSR {

    struct _IX_SCB {

        volatile UINT8 Status;              /* @00 */
#define IX_ST_CUS_MASK          0xc0
# define IX_ST_CUS_IDLE         0x00
# define IX_ST_CUS_SUSPENDED    0x40
# define IX_ST_CUS_LPQ_ACTIVE   0x80
# define IX_ST_CUS_HQP_ACTIVE   0xc0
#define IX_ST_RUS_NO_RBDs       0x20    /* '58 */
#define IX_ST_RUS_MASK          0x1c
# define IX_ST_RUS_IDLE         0x00
# define IX_ST_RUS_SUSPENDED    0x04
# define IX_ST_RUS_NO_RESOURCES 0x08
# define IX_ST_RUS_READY        0x10

        volatile UINT8 StatAck;             /* @01 */
#define IX_SA_CX_MASK           0xfd
#define IX_SA_CX                0x80
# define IX_SA_TNO               0x80
#define IX_SA_FR                0x40
#define IX_SA_CNA               0x20
#define IX_SA_RNR               0x10
#define IX_SA_MDI               0x08
#define IX_SA_SWI               0x04
#define IX_SA_FCP               0x01

        volatile UINT8 Command;             /* @02 */
#define IX_CMD_CU_MASK          0xf0
# define IX_CMD_CU_NOP          0x00
# define IX_CMD_CU_START        0x10
# define IX_CMD_CU_RESUME       0x20
# define IX_CMD_CU_LOAD_DCA     0x40
# define IX_CMD_CU_DUMP_SC      0x50
# define IX_CMD_CU_LOAD_BASE    0x60
# define IX_CMD_CU_DUMPR_SC     0x70
# define IX_CMD_CU_S_RESUME     0xa0
#define IX_CMD_RU_MASK          0x07
# define IX_CMD_RU_NOP          0x00
# define IX_CMD_RU_START        0x01
# define IX_CMD_RU_RESUME       0x02
# define IX_CMD_RU_REDIRECT     0x03
# define IX_CMD_RU_ABORT        0x04
# define IX_CMD_RU_LOAD_HDS     0x05
# define IX_CMD_RU_LOAD_BASE    0x06
# define IX_CMD_RU_RBD_RESUME   0x07   /* '58 */

        volatile UINT8 InterruptMask;       /* @03 */
#define IX_MSK_CX               0x80
#define IX_MSK_FR               0x40
#define IX_MSK_CNA              0x20
#define IX_MSK_RNR              0x10
#define IX_MSK_ER               0x08
#define IX_MSK_FCP              0x04
#define IX_MSK_SI               0x02
#define IX_MSK_M                0x01

        volatile UINT32 GeneralPointer;     /* @04 */

    } SCB;

    volatile UINT32 Port;                   /* @08 */
#define IX_P_PTR_MASK           0xfffffff0
#define IX_P_SOFT_RESET         0x0  /* no ptr */
#define IX_P_SELF_TEST          0x1
#define IX_P_SELECTIVE_RESET    0x2  /* no ptr */
#define IX_P_DUMP               0x3
#define IX_P_DUMP_WAKEUP        0x7  /* 82559 and later */

    UINT16  Reserved;

    volatile UINT16 EepromControl;          /* @0e */
#define IX_EC_EEDO              0x08
#define IX_EC_EEDI              0x04
#define IX_EC_EECS              0x02
#define IX_EC_EESK              0x01
# define IX_EEPROM_WRITE        1
# define IX_EEPROM_READ         2

    volatile UINT32 MdiControl;             /* @10 */
#define IX_MC_INTEN             0x20000000
#define IX_MC_READY             0x10000000
#define IX_MC_OPCODE_MASK       0x0c000000
# define IX_MC_READ             0x08000000
# define IX_MC_WRITE            0x04000000
#define IX_MC_PHY_ADDRESS       0x03e00000
# define PHY_DEFAULT_ADDRESS    0x00200000
#define IX_MC_REG_ADDRESS       0x001f0000
# define IX_MC_ADDR_SHIFT       16
#define IX_MC_DATA_MASK         0x0000ffff

    volatile UINT32 RxDmaByteCount;         /* @14 */
#define IX_RC_MASK              0x000007f8

    volatile UINT8  EarlyRxInt;             /* @18 */

    volatile UINT8  FcThreshold;            /* @19 */      /* '58 and later */
#define IX_FC_0_5_KB            0x00
#define IX_FC_1_KB              0x01
#define IX_FC_1_25_KB           0x02
#define IX_FC_1_5_KB            0x03
#define IX_FC_1_75_KB           0x04
#define IX_FC_2_KB              0x05
#define IX_FC_2_25_KB           0x06
#define IX_FC_2_5_KB            0x07

    volatile UINT8  FcXonXoff;              /* @1a */      /* '58 and later */
#define IX_XX_PAUSED_LOW        0x10
#define IX_XX_PAUSED            0x08
#define IX_XX_FULL              0x04
#define IX_XX_XOFF              0x02
#define IX_XX_XON               0x01

    volatile UINT8 PMDR;                    /* @1b */
#define IX_PMDR_LINK_STATUS     0x80  /* rc '59*/
#define IX_PMDR_MAGIC_PKT       0x40  /* rc '59 !er*/
#define IX_PMDR_NICE_PKT        0x20  /* rc '59*/
#define IX_PMDR_TCO_READY       0x08  /* rw '58b*/
#define IX_PMDR_FORCE_TCO       0x04  /* ro '58b & '59*/
#define IX_PMDR_TCO_REQUEST     0x02  /* ro '58b & '59*/
#define IX_PMDR_PME_STATUS      0x01  /* rc '58 & ro '59*/

    volatile UINT8 GeneralControl;          /* @1c */      /* '59 only */
#define IX_GC_DEEP_DOWN         0x02
#define IX_GC_CLOCKRUN          0x01

    volatile UINT8 GeneralStatus;           /* @1d */      /* '59 only */
#define IX_GS_FULL_DUPLEX       0x04
#define IX_GS_100Mbps           0x02
#define IX_GS_LINK_UP           0x01

    UINT8   Reserved3[2+16];

    volatile UINT32 FunctionEvent;          /* @30 */

    volatile UINT32 FunctionEventMask;      /* @34 */

    volatile UINT32 FunctionEventState;     /* @38 */

    volatile UINT32 ForceEvent;             /* @3c */

} IX_CSR, *PIX_CSR;


/* Statistical Counters
 */
#define IX_NUMBER_OF_COUNTERS_MAX 21
typedef struct _IX_STATISTICAL_COUNTERS {
    UINT32 Counter[IX_NUMBER_OF_COUNTERS_MAX];
    /* Make this one cache line on PPC */
    UINT8  AlignmentPadding[128-(21*4)]; 
} IX_STATISTICAL_COUNTERS, *PIX_STATISTICAL_COUNTERS;
#define IX_SC_TGF               0
#define IX_SC_MAXCOL            1
#define IX_SC_LATECOL           2
#define IX_SC_TX_UNDERRUN       3
#define IX_SC_TX_LOST_CRS       4
#define IX_SC_DEFERRED          5
#define IX_SC_COLLISION_S       6
#define IX_SC_COLLISION_M       7
#define IX_SC_COLLISION_T       8
#define IX_SC_RGF               9
#define IX_SC_RX_CRC            10
#define IX_SC_RX_ALIGNMENT      11
#define IX_SC_RX_RESOURCES      12
#define IX_SC_RX_OVERRUN        13
#define IX_SC_RX_CDT            14
#define IX_SC_RX_SHORT          15
#define IX_SC_TX_FCP            16
#define IX_SC_RX_FCP            17
#define IX_SC_BAD_FCP           18
#define IX_SC_TCO               19
# define IX_SC_82557_STATUS     16
# define IX_SC_82558_STATUS     19
# define IX_SC_82559_STATUS     20
# define IX_SCDUMP_COMPLETE     0xa005
# define IX_SCDUMPR_COMPLETE    0xa007


/* Self-test results
 */
typedef struct _IX_SELFTEST_RESULTS {
    UINT32   RomSignature;
    UINT32   ResultMask;
#define IX_STR_FAIL             0x1000
#define IX_STR_DIAG             0x20
#define IX_STR_REGISTERS        0x8
#define IX_STR_ROM_FAIL         0x4
# define IX_STR_OK              0x0
} IX_SELFTEST_RESULTS, *PIX_SELFTEST_RESULTS;

/* Dump results
 */
#define IX_DUMP_SIZE            150
#define IX_DUMP_BYTES           (IX_DUMP_SIZE * 4)
typedef UINT32 IX_DUMP_RESULTS[IX_DUMP_SIZE];
#define IX_DUMP_STATUS          149
# define IX_DUMP_COMPLETE       0xa006

/* Action command header, common to all commands
 */
typedef struct _IX_COMMAND_HEADER {

    UINT16 Status;
#define IX_CBS_COMPLETE         0x8000
#define IX_CBS_OK               0x2000
#define IX_CBS_CRC              0x0800
#define IX_CBS_ALIGNMENT        0x0400
#define IX_CBS_BUF_TOO_SMALL    0x0200
#define IX_CBS_OVERRUN          0x0100
#define IX_CBS_RUNT_FRAME       0x0080
#define IX_CBS_TYPE             0x0020
#define IX_CBS_PHY_ERROR        0x0010
#define IX_CBS_NONE_MATCH       0x0004
#define IX_CBS_IA_NO_MATCH      0x0002
#define IX_CBS_COLLISION        0x0001
#define IX_CBS_TCO              0x0001

    UINT16 Command;
#define IX_CBC_ELAST            0x8000
#define IX_CBC_SUSPEND          0x4000
#define IX_CBC_INTERRUPT        0x2000
#define IX_CBC_CID              0x1f00
#define IX_CBC_NC               0x0010
#define IX_CBC_H                0x0010
#define IX_CBC_SF               0x0008
#define IX_CBC_CMD_MASK         0x0007
# define IX_CBC_NOP             0x0
# define IX_CBC_MAC_ADDRESS     0x1
# define IX_CBC_CONFIGURE       0x2
# define IX_CBC_MULTICAST       0x3
# define IX_CBC_TRANSMIT        0x4
# define IX_CBC_LOAD_UCODE      0x5
# define IX_CBC_DUMP            0x6
# define IX_CBC_DIAGNOSE        0x7

    UINT32 LinkAddress;
#define IX_LINK_ADDRESS_NULL 0xffffffff

} IX_COMMAND_HEADER, *PIX_COMMAND_HEADER;

/* Action commands
 */
typedef union _IX_COMMAND {

    IX_COMMAND_HEADER Nop;

    struct {
        IX_COMMAND_HEADER h;
        UINT8 data[24];
    } Configure;

    struct {
        IX_COMMAND_HEADER h;
        UINT8 data[6];
    } MacAddress;

    struct {
        IX_COMMAND_HEADER h;
        UINT16 MulticastByteCount;
        UINT8 data[1][6]; /* varsized */
    } Multicast;

    struct _IX_TCB {
        IX_COMMAND_HEADER h;
        UINT32 TbdArrayAddress;
        UINT16 TcbByteCount;
#define IX_BC_EOF               0x8000
        UINT8  Threshold;
        UINT8  TbdNumber;
    } Transmit;

#if 0
    /* We are not loading ucode.
     * BEWARE: Fix the sizing below if you enable this
     */
    struct {
        IX_COMMAND_HEADER h;
        UINT32 data[64];
    } LoadUcode;
#endif

    struct {
        IX_COMMAND_HEADER h;
        UINT32 BufferAddress;
    } Dump;

    IX_COMMAND_HEADER Diagnose;

    /* Make it at least 2 cache lines on PPC
     */
    UINT8 SizeMe[_DCACHELINESIZE*2];

} IX_COMMAND, *PIX_COMMAND;


/* Receive Frame Descriptor
 */
typedef struct _IX_RFD {

    IX_COMMAND_HEADER h;

    UINT32 RbdAddress;

    UINT16 ActualCount;
#define IX_AC_MASK              0x3fff
#define IX_AC_F                 0x4000
#define IX_AC_EOF               0x8000

    UINT16 Size;

} IX_RFD, *PIX_RFD;

/* Transmit frame descriptor (flexible mode)
 */
typedef struct _IX_TFD {

    struct _IX_TCB TCB;

    struct _IX_TBD {

        UINT32 BufferAddress;

        UINT16 ActualCount;
#define IX_AC_EL                0x8000  /* MASK same */

        UINT16 Reserved;
#define IX_MAX_TBD 4 /* or more, but thats what we use anyways */

    } TBD[IX_MAX_TBD]; 

} IX_TFD, *PIX_TFD;

/* MDI Registers and their bits
 */

/* MDI Control register
 */
#define IX_MDI_CONTROL          0x00
# define IX_MCT_RESET            0x8000
# define IX_MCT_LOOPBACK         0x4000
# define IX_MCT_100Mbps          0x2000
# define IX_MCT_AN_ENABLE        0x1000
# define IX_MCT_POWERDOWN        0x0800
# define IX_MCT_ISOLATE          0x0400
# define IX_MCT_RESTART_AN       0x0200
# define IX_MCT_FULL_DUPLEX      0x0100
# define IX_MCT_COLL_TEST        0x0080
# define IX_MCT_10Mbps           0x0040

/* MDI Status register
 */
#define IX_MDI_STATUS           0x01
# define IX_MST_T4               0x8000
# define IX_MST_FULL_DUPLEX      0x4000
# define IX_MST_HALF_DUPLEX      0x2000
# define IX_MST_10_FULL_DUPLEX   0x1000
# define IX_MST_10_HALF_DUPLEX   0x0800
# define IX_MST_T2_FULL_DUPLEX   0x0400
# define IX_MST_T2_HALF_DUPLEX   0x0200
# define IX_MST_EXT_STATUS       0x0100
# define IX_MST_NO_PREAMBLE      0x0040
# define IX_MST_AN_COMPLETE      0x0020
# define IX_MST_REMOTE_FAULT     0x0010
# define IX_MST_AN_ENABLED       0x0008
# define IX_MST_LINK_UP          0x0004
# define IX_MST_JABBER           0x0002
# define IX_MST_EXT_CAP          0x0001

/* MDI Identification registers
 */
#define IX_MDI_ID0              0x02
#define IX_MDI_ID1              0x03 /* low16 */
# define PHY_INTEL_82555_0        0x02a80150
# define PHY_INTEL_82555_1        0x02a80151
# define PHY_INTEL_82555_4        0x02a80154

/* Auto-negotiation advertisement register
 */
#define IX_MDI_AN_ADVERTISE     0x04
# define IX_AAA_NEXT_PAGE        0x8000
# define IX_AAA_ACK              0x4000
# define IX_AAA_REMOTE_FAULT     0x2000
# define IX_AAA_ASYM_PAUSE       0x0800
# define IX_AAA_PAUSE            0x0400
# define IX_AAA_T4               0x0200
# define IX_AAA_TX_FULL_DUPLEX   0x0100
# define IX_AAA_TX_HALF_DUPLEX   0x0080
# define IX_AAA_10T_FULL_DUPLEX  0x0040
# define IX_AAA_10T_HALF_DUPLEX  0x0020
# define IX_AAA_PROT_MASK        0x001f
# define IX_AAA_IEEE802_3          0x01

/* Auto-negotiation link partner ability register
 * (Same bits as above)
 */
#define IX_MDI_AN_PARTNER       0x05

/* Auto-negotiation expansion register
 */
#define IX_MDI_AN_EXPANSION     0x06
# define IX_AEX_PAR_DETECT       0x0010
# define IX_AEX_NEXT_PAGE_REMOTE 0x0008
# define IX_AEX_NEXT_PAGE_LOCAL  0x0004
# define IX_AEX_PAGE_RECEIVED    0x0002
# define IX_AEX_AN_REMOTE        0x0001

/* 82555 specific registers (on '57 and beyond)
 * Status and control register
 */
#define IX_PHY55_CSR            0x10
# define IX_5CS_FC               0x8000
# define IX_5CS_T4_ENABLE        0x4000
# define IX_5CS_REPEATER         0x2000
# define IX_5CS_SYNC             0x0800
# define IX_5CS_100_DOWN         0x0400
# define IX_5CS_10_DOWN          0x0200
# define IX_5CS_REV_POLARITY     0x0100
# define IX_5CS_AN_T4            0x0004
# define IX_5CS_AN_100           0x0002
# define IX_5CS_AN_FULL_DUPLEX   0x0001

/* Special control (one) register
 */
#define IX_PHY55_CTL            0x11
# define IX_5C1_NO_SCRAMBLER     0x8000
# define IX_5C1_NO_4_TO_5        0x4000
# define IX_5C1_FORCE_H          0x2000
# define IX_5C1_FORCE_34         0x1000
# define IX_5C1_GOOD_LINK        0x0800
# define IX_5C1_TRISTATE         0x0200
# define IX_5C1_NO_DYN_DOWN      0x0100
# define IX_5C1_AN_LOOPBACK      0x0080
# define IX_5C1_NO_FILTER        0x0020
# define IX_5C1_NO_AUTO_POLARITY 0x0010
# define IX_5C1_NO_SQUELCH       0x0008
# define IX_5C1_EXT_SQUELCH      0x0004
# define IX_5C1_NO_LINK_TEST     0x0002
# define IX_5C1_NO_JABBER       0x0001

/* Clock sythesis test and control register
 */
#define IX_PHY55_CLOCK          0x12
# define IX_5CK_SHIFT            0x8000
# define IX_5CK_LOAD             0x4000
# define IX_5CK_BREAK_ENABLE     0x2000
# define IX_5CK_EQL_PROBE_ENABLE 0x1000
# define IX_5CK_10_PROBE_ENABLE  0x0800
# define IX_5CK_PHY_ADDRESS_MASK 0x001f

/* 100BASE-TX receive false carrier counter register
 */
#define IX_PHY55_FALSE_CARRIER  0x13

/* 100BASE-TX receive disconnect counter register
 */
#define IX_PHY55_DISCONNECT     0x14

/* 100BASE-TX receive error frame counter register
 */
#define IX_PHY55_FRAME          0x15

/* Receive symbol error counter register
 */
#define IX_PHY55_SYMBOL         0x16

/* 100BASE-TX receive EOF error counter register
 */
#define IX_PHY55_100_EOF        0x17

/* 10BASE-T receive EOF error counter register
 */
#define IX_PHY55_10_EOF         0x18

/* 10BASE-T transmit jabber detect counter register
 */
#define IX_PHY55_JABBER         0x19

/* Equalizer control and status register
 */
#define IX_PHY55_EQUALIZER      0x1a
# define IX_5EQ_ENCODE(_opcode_,_data_)  ((UINT16)((_opcode_<<13)|(_data_&0x1fff)))
# define IX_5EQ_NOP              0x0
# define IX_5EQ_WRITE_ASD0       0x1
# define IX_5EQ_WRITE_ASD1       0x2
# define IX_5EQ_WRITE_ASD2       0x3
# define IX_5EQ_READ_STATUS      0x4
# define IX_5EQ_READ_JITTER      0x5
# define IX_5EQ_READ_CLOCK       0x6

/* Special control (two) register
 * Legend: First token is for ACTLED# pin, second for LILED# pin
 */
#define IX_PHY55_CTL2           0x1b
# define IX_5C2_ACTIVITY_LINKUP  0x0
# define IX_5C2_SPEED_COLLIDE    0x1
# define IX_5C2_SPEED_LINKUP     0x2
# define IX_5C2_ACTIVITY_COLLIDE 0x3
# define IX_5C2_OFF_OFF          0x4
# define IX_5C2_OFF_ON           0x5
# define IX_5C2_ON_OFF           0x6
# define IX_5C2_ON_ON            0x7
