/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define USE_INTRINSICS
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include <mmhal.h>
#include <drivers/drivers.h>
#include <assert.h>

/* XXXXXXXXXXX in machdep header file XXXXXXXXXXXX */
#define LanWrite16BitAttribute(_b_,_r_,_v_) \
     IOSpaceWriteInt16((IO_PORT)((_b_)+(_r_)),(UINT16)(_v_))
#define LanRead16BitRegister(_b_,_r_) \
     IOSpaceReadInt16((IO_PORT)((_b_)+(_r_)))
#define LanWrite16BitRegister(_b_,_r_,_v_) \
     IOSpaceWriteInt16((IO_PORT)((_b_)+(_r_)),(UINT16)(_v_))
#define LanWrite8BitRegister(_b_,_r_,_v_) \
     IOSpaceWriteInt8((IO_PORT)((_b_)+(_r_)),(UINT8)(_v_))
#define LanRead16BitWords(_b_,_r_,_w_,_n_) \
     IOSpaceReadFIFO16((IO_PORT)((_b_)+(_r_)),_w_,_n_)
#define LanWrite16BitWords(_b_,_r_,_w_,_n_) \
     IOSpaceWriteFIFO16((IO_PORT)((_b_)+(_r_)),_w_,_n_)
#define LanDisableBoardInts(_l_) \
     IOSpaceWriteInt8((IO_PORT)((_l_)->IoBase+LAN_MSK),0)
#define LanEnableBoardInts(_l_) \
     IOSpaceWriteInt8((IO_PORT)((_l_)->IoBase+LAN_MSK),0x73) /* xxx */
/* XXXXXXXXXXX in machdep header file XXXXXXXXXXXX */

#include <lan9000.h>

#ifdef _MSC_VER
#pragma warning(disable:4028)   /* formal parameter 1 different.. */
#endif

typedef struct _LAN9000 { /* Inherits from CDevice */

    struct arpcom ac;            /* standard ether interface header */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define CollisionCount ac.ac_if.if_data.ifi_collisions
#define RuntPackets  ac.ac_if.if_data.ifi_ierrors
#define XmitFailures  ac.ac_if.if_data.ifi_oerrors
#define MissedPackets ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define CompleteResets ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define XmitTimeouts ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define MemoryErrors ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]

    /* Driver's private data
     */
    IO_PORT IoBase;               /* Passed to driver      */
    IO_PORT AttributeBase;        /* Passed to driver      */
    UINT PhyType;                 /* Passed to driver      */
#define LAN_PHY_AUI 1
#define LAN_PHY_FDX 2

} LAN9000, *PLAN9000;

/*  Process packets that have been completely received.
 */
void
Lan9000Recv(PLAN9000 pLan)
{
    IO_PORT         IoBase = pLan->IoBase;
    struct ifnet   *ifp = &pLan->is_if;
    UINT            Length;
    UINT16          Reg, Status;
    struct mbuf *Pkt;
    struct ether_header eh;
    PIENDPOINTFACTORY mbf = ifp->MBufFact;
    SCODE sc;

    /* Select start of next ready packet
     */
AnotherPacket:
    Pkt = NULL;
    LanWrite16BitRegister(IoBase,LAN_PTR,
                          LAN_PTR_RCV |
                          LAN_PTR_AUTO_INCR |
                          LAN_PTR_READ);
    /* Get status word, check for errors
     */
    Status = LanRead16BitRegister(IoBase,LAN_DATA0);

    if (Status & LAN_RS_ERRMASK) {
        pLan->RuntPackets++;
        goto ReleaseFrame;
    }

    /* Get frame length
     */
    Reg = LanRead16BitRegister(IoBase,LAN_DATA0);
    Length = Reg - 5 - ((Status & LAN_RS_ODDFRM) ? 0 : 1);

    /* Get a buffer to copy to
     */
    sc = mbf->v->AllocateBuffer(mbf, Length - sizeof(struct ether_header),
                                (PINETDRIVER) ifp, (PIBUFFER *) &Pkt);
    if (FAILED(sc)) {
        /* No room, drop the packet
         */
        goto DropPacket;
    }

    /* Read&fix the ether header
     */
    LanRead16BitWords(IoBase,LAN_DATA0,(UINT16 *)&eh,(sizeof eh) >> 1);
    eh.ether_type = ntohs(eh.ether_type);

    /* Either way..  and its time to copy now.
     */
    assert(Pkt->m_pkthdr.len <= MCLBYTES);
    Pkt->m_len = Pkt->m_pkthdr.len;
    LanRead16BitWords(IoBase,LAN_DATA0,mtod(Pkt,PUINT16),Pkt->m_len >> 1);

    /* Control word and possibly last byte
     */
    Reg = LanRead16BitRegister(IoBase,LAN_DATA0);
    if (Status & LAN_RS_ODDFRM) /* && Reg & (LAN_CTL_ODD << 8) */
      (mtod(Pkt,PUINT8))[Pkt->m_len-1] = (UINT8)Reg;

    /* Release packet memory
     */
ReleaseFrame:
    LanWrite16BitRegister(IoBase,LAN_MMUCR,LAN_CMD_POPREL_RX_FIFO);

    if (Pkt) {

        UINT OldResetCount = pLan->CompleteResets;
        /*
         * Hand the packet to the network module.
         */
        ifp->if_ipackets++;
        mbf->v->EtherInput(mbf, (PINETDRIVER) ifp, &eh, (PIBUFFER) Pkt);

        /*
         * Watch out for the case the interface was reset
         * during the previous upcall.
         */
        if (OldResetCount != pLan->CompleteResets)
            return;
    }

    /* Make sure MMU is done (should be!) before checking for another packet
     */
    for (Length = 0;;) {
      if ((LanRead16BitRegister(IoBase,LAN_MMUCR) & LAN_MMUCR_BUSY) == 0)
        break;
      /* If its stuck there's probably an interrupt waiting to tell us why
       */
      if (++Length == 100)
        return;
      Delay(10);
    }

    /* See if we have another packet ready
     */
    Reg = LanRead16BitRegister(IoBase,LAN_IST);
    if (Reg & LAN_IST_RCV)
      goto AnotherPacket;

    /* Nope, nothing more to do
     */
    return;

DropPacket:
    if (Pkt)
        Pkt->m_hdr.v->Release((PIBUFFER) Pkt);
    Pkt = NULL;
    ifp->if_iqdrops++;
    ifp->if_ierrors++;
    goto ReleaseFrame;
}

/*
 * Start output on interface.
 */
void MCT
Lan9000Xmit( PLAN9000 pLan)
{
    struct mbuf     *Request, *m;
    UINT            len, alloc, i;
    UINT16          Reg;
    struct ifnet    *ifp = &pLan->is_if;
    IO_PORT         IoBase = pLan->IoBase;
    PIENDPOINTFACTORY mbf = ifp->MBufFact;

    mbf->v->NetLock(mbf);

    /* If we came from interrupt, that must be the last xmit command
     */
    ifp->if_flags &= ~IFF_OACTIVE;

    /* As long as there is some to do
     */
    while ((Request = pLan->is_if.if_snd.ifq_head) != NULL) {

        /* Count the bytes in the frame, we need
         * to ask the MMU for xmit memory allocation.
         * Value has already been checked to be less than ethermax.
         */
        len = Request->m_pkthdr.len;

        /* Roundup funny and make min
         */
        alloc = (len & ~1) + 6;
        if (alloc < 70)
          alloc = 70;

        /* Ask for memory. NB: bank2 always selected.
         */
        LanWrite16BitRegister(IoBase,LAN_MMUCR,
                              LAN_CMD_TX_ALLOCATE | (alloc >> 8));

        /* Wait for reply
         */
        for (i = 0; i < 100; i++) {
          Reg = LanRead16BitRegister(IoBase,LAN_IST);
          if (Reg & LAN_IST_ALLOC)
            break;
          Delay(10);
        }

        /* BUGBUG Should only spin a tiny bit and ifnot
         * wait for corresponding interrupt..
         */
        if (i == 100) DebugBreak();

        /* See what we got.
         */
        Reg = (UINT16) (LanRead16BitRegister(IoBase,LAN_PNR) >> 8);
        if (Reg & LAN_ARR_FAILED)
          break; /* will try later, FIFO must be full */

        /* Dequeue the next transmit request, make sure coherency..
         */
        IF_DEQUEUE(&pLan->is_if.if_snd, m);
        if ((m == 0) || (m != Request))
          continue;

        pLan->is_if.if_opackets++;

        /* Copy the frame out to the chip's memory
         */
        LanWrite16BitRegister(IoBase,LAN_PNR,Reg);
        LanWrite16BitRegister(IoBase,LAN_PTR,LAN_PTR_AUTO_INCR);

        /* Status word & length (BUGBUG use 32bit writes ??)
         */
        LanWrite16BitRegister(IoBase,LAN_DATA0,0);
        LanWrite16BitRegister(IoBase,LAN_DATA0,alloc);
        alloc -= 4;

        /* Sweep the buffers and copy
         */
        for (;m;m=m->m_next) {
          LanWrite16BitWords(IoBase,LAN_DATA0,
                             mtod(m,PUINT16),m->m_len >> 1);
          if (m->m_len & 1)
            LanWrite8BitRegister(IoBase,LAN_DATA0,
                                 (mtod(m,PUINT8))[m->m_len-1]);
          alloc -= m->m_len;
        }

        /* Pad and control byte
         */
        for (;alloc > 2;alloc -= 2)
          LanWrite16BitRegister(IoBase,LAN_DATA0,0);
        LanWrite16BitRegister(IoBase,LAN_DATA0,
                              (LAN_CTL_CRC | ((len & 1) << LAN_CTL_ODD_SHIFT))
                               << 8);

        /* Say go.
         */
        LanWrite16BitRegister(IoBase,LAN_MMUCR,LAN_CMD_ENQ_PKT);
        ifp->if_flags |= IFF_OACTIVE;
        
        Request->m_hdr.v->Release((PIBUFFER) Request);

        /* BUGBUG check the receive FIFO, stop if some pending
         */

    }

    mbf->v->NetUnlock(mbf);
}

/*
 * Interrupt service routine.
 * Return TRUE if rescheduling needed.
 */
BOOL MCT
Lan9000Isr(PCDEVICE pDev, PBOOL pNotMyInterrupt)
{
    PLAN9000 pLan = (PLAN9000) pDev;

    /* Disable interrupts here.  Will be enabled by Lan9000Interrupt()
     * after its done processing all requests.  It also acks the chip.
     */
    LanDisableBoardInts(pLan);

    /* Wakeup handler thread.
     */
    return Condition_InterruptSignal(&pLan->is_if.IntrPoint);
}


/* Thread-level interrupt handling
 */
BOOL MCT
Lan9000Interrupt(PCDEVICE pDev)
{
    PLAN9000 pLan = (PLAN9000) pDev;
    IO_PORT IoBase = pLan->IoBase;
    UINT16 Reg;

    for (;;) {
        Reg = LanRead16BitRegister(IoBase,LAN_IST);
        if (Reg == 0)
            break;

        if (Reg & LAN_IST_RCV)
            Lan9000Recv(pLan);

        if (Reg & LAN_IST_TX) {
            Lan9000Xmit(pLan);
            /* Ack */
            LanWrite16BitRegister(IoBase,LAN_ACK,LAN_ACK_TX);
        }

        /* BUGBUG write more code */
        if (Reg & (0x70)) DebugBreak();

    }

    LanEnableBoardInts(pLan);
    return TRUE;
}

/* Random other things..
 */
SCODE MCT
Lan9000Ioctl( PLAN9000 pLan, UINT Prop, caddr_t Data)
{
    if (Prop == SIOCGIFSTATS) {

#if 0
        /* This gets in the way a lot. Sigh.
         */
        mbf->v->NetLock(mbf);
        LanDisableBoardInts(pLan);
//      Lan9000ReadBufferFromCardMemory(pLan,&Scb,sizeof Scb, OFFSET_SCB);
        LanEnableBoardInts(pLan);
        mbf->v->NetUnlock(mbf);

        /* You'd think there is a quick way to reset these 16bit counters..
         * alas there is not.  Do we have to care for overflows ? Nyaa.
         */
        pLan->MissedPackets = Scb.ResourceErrors;
        pLan->MemoryErrors = Scb.OverrunErrors;
        pLan->RuntPackets = Scb.AlignmentErrors + Scb.CrcErrors;
#endif
    }

    /* We care for requests to turn the interface down/up.
     */
    if (Prop == SIOCSIFFLAGS) {

        if ((pLan->is_if.if_flags & IFF_UP) == 0)
            /*
             * Down. Take the easy way out: just disable interrupts.
             * If/when we get back up we'll just handle the errors.
             */
            LanDisableBoardInts(pLan);
        else
            /*
             * Up. Strictly speaking, needs protection.
             */
            LanEnableBoardInts(pLan);
    }
    return NO_ERROR;
}


/* Reset/Reinitialize the card
 */
INT MCT
Lan9000Reset(PCDEVICE pDev)
{
    PLAN9000 pLan = (PLAN9000) pDev;
    IO_PORT IoBase = pLan->IoBase;
    UINT16 Reg;
    struct ifnet *ifp;

    DBGME(2,printf(_T("Enter Lan9000Reset\n")));

    /* Stop network activity 
     */
    LanDisableBoardInts(pLan);

    ifp = &pLan->is_if;
    if (ifp->if_flags & IFF_RUNNING) {
        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
    }
    pLan->CompleteResets++;

    /* BUGBUG need more code to actually reset it
     * Like, RCR:SOFT_RST, MMUCR:RESET
     */

    /* Set the SQUELCH bit
     * This changes the threshold for a valid signal from 300mV to 180mV.
     * This is to fix problems we have seen with some (usually 8 port) hubs.
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,1);
    Reg = LanRead16BitRegister(IoBase,LAN_CR);
    Reg |= LAN_CR_NO_WAIT | LAN_CR_SET_SQLCH;
    LanWrite16BitRegister(IoBase,LAN_CR,Reg);

    DBGME(2,printf(_T("SMC config reg val: %x\n"),Reg));

    /* Initialize the control register
     */
    LanWrite16BitRegister(IoBase, LAN_CTR,
                          LAN_CTR_FIXED |
                          LAN_CTR_AUTO_RELEASE |
                          LAN_CTR_TE_ENABLE |
                          LAN_CTR_CR_ENABLE );

    /* Initialize memory configuration register
     * Reserve xmit room for one full size frame.
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,0);
    LanWrite16BitRegister(IoBase, LAN_MCR, 6 ); /* x256 */

    /* Initialize transmit control register
     * Its said PAD doesnt really work but we ask.
     */
    LanWrite16BitRegister(IoBase, LAN_TCR, LAN_TCR_PAD_EN | LAN_TCR_TXENA );

    /* Acknowledge anything that might be pending
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,2);
    LanWrite16BitRegister(IoBase,LAN_ACK,
                          LAN_ACK_ERCV |
                          LAN_ACK_RX_OVRN |
                          LAN_ACK_TX_EMPTY |
                          LAN_ACK_TX);

    /* Ready to receive frames.
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,0);
    LanWrite16BitRegister(IoBase,LAN_RCR, LAN_RCR_STRIP_CRC | LAN_RCR_RXEN );

    /* Select Bank2 for the duration.
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,2);

    /* We are up
     */
    ifp->if_flags |= (IFF_UP | IFF_RUNNING);

    LanEnableBoardInts(pLan);

    return 0;

}

struct INetDriverVtbl Lan9000DriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    (_INetDriver_Isr_METHOD_TYPE) Lan9000Isr,
    (_INetDriver_HandleInterrupt_METHOD_TYPE) Lan9000Interrupt,
    (_INetDriver_Initialize_METHOD_TYPE) Lan9000Reset,
    NULL,
    NULL,
    (_INetDriver_Start_METHOD_TYPE) Lan9000Xmit,
    (_INetDriver_Ioctl_METHOD_TYPE) Lan9000Ioctl,
    NULL
};

/* Once-only initialization
 */
PLAN9000
Lan9000Initialize(IO_PORT IoBase, IO_PORT AttributeBase, UINT PhyType)
{
    UINT16 Reg;
    INT i;
    PLAN9000 pLan;
    TIME Idle;
    static short unit=0;

    DBGME(2,printf(_T("Enter Lan9000Initialize %x %x\n"),IoBase,AttributeBase));
    /* Take chip out of powerdown mode and wait a bit.
     */
    LanWrite16BitAttribute(AttributeBase,LAN_ATT_ECOR,
                           LAN_ECOR_ENABLE);
    Int32ToInt64(Idle,TIME_RELATIVE(TIME_MILLIS(1001)));
    SleepUntil(Idle);

    /* Make sure its functional
     */
    Reg = LanRead16BitRegister(IoBase,LAN_BSR);
    if ((Reg & LAN_BSR_FIXED) != LAN_BSR_SIGNATURE) {
        DBGME(1,printf(_T("Lan9000 not detected at %x (%x)\n"),IoBase,Reg));
        goto ErrorExit;
    }

    /* By now the EEPROM info should have been read in, but make sure.
     * Select Bank1 to start
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,1);
    for (i = 0;i < 100; i++) {
        Reg = LanRead16BitRegister(IoBase,LAN_CTR);
        if ((Reg & (LAN_CTR_RELOAD | LAN_CTR_STORE)) == 0)
            break;
        Delay(100);
    }
    if (Reg & (LAN_CTR_RELOAD | LAN_CTR_STORE)) {
        DBGME(1,printf(_T("Lan9000 at %x seems stuck (%x)\n"),IoBase,Reg));
        goto ErrorExit;
    }

    /* Ok we are doing it for real.
     */
    pLan = (PLAN9000) malloc(sizeof(LAN9000));
    if (pLan == NULL)
        goto ErrorExit;

    memset(pLan, 0, sizeof(LAN9000));

    /* Set basic info.
     */
    pLan->IoBase = IoBase;
    pLan->AttributeBase = AttributeBase;
    pLan->PhyType = PhyType;

    /* Read station address off chip
     */
    Reg = LanRead16BitRegister(IoBase,LAN_IA01);
    pLan->is_addr[0] = (UINT8) (Reg >> 8);
    pLan->is_addr[1] = (UINT8) Reg;

    Reg = LanRead16BitRegister(IoBase,LAN_IA23);
    pLan->is_addr[2] = (UINT8) (Reg >> 8);
    pLan->is_addr[3] = (UINT8) Reg;

    Reg = LanRead16BitRegister(IoBase,LAN_IA45);
    pLan->is_addr[4] = (UINT8) (Reg >> 8);
    pLan->is_addr[5] = (UINT8) Reg;

    /* More init
     */
    pLan->is_if.v = &Lan9000DriverVtbl;
    pLan->is_if.RefCnt = 1;
    pLan->is_if.if_unit = unit++;
    pLan->is_if.if_name = "lan";
    pLan->is_if.if_type = IFT_ETHER;
    pLan->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    pLan->is_if.if_baudrate = 10*1000*1000; /* 10Mbps */

#if _DEBUG
    printf(_T("Lan9000 CardAddr %02x%02x%02x%02x%02x%02x Base %x Attrib %x\n"),
           pLan->is_addr[0], pLan->is_addr[1],
           pLan->is_addr[2], pLan->is_addr[3],
           pLan->is_addr[4], pLan->is_addr[5],
           pLan->IoBase, pLan->AttributeBase);
#endif

    /* Leave it set at Bank2, even if more init is needed.
     * That way messing with interrupts works.
     */
    LanWrite16BitRegister(IoBase,LAN_BSR,2);

    return pLan;

ErrorExit:
    return NULL;
}


SCODE __cdecl Lan9000Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * a1,
        const _TCHAR * a2,
        UINT AttribBase,
        UINT PhyType,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);
    UnusedParameter(a1);
    UnusedParameter(a2);

    /* Lan9000::Lan9000New:NetworkDriver 0x30000000 5 0x38000000 0 */
    DBGME(2, printf("Lan9000Constructor(%x %s %x %x %x attrib=%x phyt=%x)\n",
           pThis, Name, RegisterBase, Interrupt, ppUnk, AttribBase, PhyType));

    pUnk = (PIUNKNOWN) Lan9000Initialize((IO_PORT) RegisterBase,
                                         (IO_PORT) AttribBase, PhyType);

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(1, printf("Lan9000Constructor failed\n"));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Lan9000DriverCobVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Lan9000Constructor
};

static struct _DRIVERCOB Lan9000DriverCobObject = { &Lan9000DriverCobVtbl, 1 };

PIUNKNOWN Lan9000CobMain(void)
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Lan9000DriverCobVtbl,
                          "CDeviceConstructor",
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Lan9000DriverVtbl, "CNetDriver",
                          &IID_INetDriver);
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &Lan9000DriverCobObject;
}
