/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Simulator network driver defs */

#include <import_windows.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include "ntddpack.h"
#include <drivers/drivers.h>

typedef struct _DRIVER_INSTANCE * PDRIVER_INSTANCE;

#pragma pack(1)
typedef struct _ETHERNET_HEADER {
    UINT8 DstAddr[6];
    UINT8 SrcAddr[6];
    union {
        UINT16 Etype;
        UINT16 Length;
    };
} ETHERNET_HEADER, *PETHERNET_HEADER;
#pragma pack()

#define MIN_ETHERNET_PACKET_SIZE 14
#define MAX_ETHERNET_DATA_SIZE 1500
#define MAX_ETHERNET_PACKET_SIZE (MIN_ETHERNET_PACKET_SIZE + MAX_ETHERNET_DATA_SIZE)

typedef enum _PACKET_MODE {
    PacketModeInvalid = 0,
    PacketModeReceiving = 1,
    PacketModeTransmitting = 2
} PACKET_MODE;

/* Packet descriptor. */

typedef struct _PACKET PACKET, *PPACKET;

typedef struct _PACKET {
    OVERLAPPED  Overlapped;                     
    PACKET      *Next;
    PACKET_MODE Mode;
    SCODE       Result;
    UINT32      Length;
    UINT32      nBytesAvail;
    union {
        UINT8   Buffer[MAX_ETHERNET_PACKET_SIZE];
        ETHERNET_HEADER Ethernet;
    };
};

/* Packet queue. */

typedef struct _PACKETQUEUE {
    ULONG           Signature;
    PACKET          *Head;
    PACKET          *Tail;
    UINT32          Length;
    volatile UINT32 Lock;
} PACKETQUEUE, * PPACKETQUEUE;

#define MAX_RECEIVE_POOL_SIZE           32
#define MAX_TRANSMIT_POOL_SIZE          16

#define FREE_QUEUE_SIG  0xABABABAB
#define RECV_QUEUE_SIG  0xFACEFACE
#define SEND_QUEUE_SIG  0x0ACE0ACE

/* Network interface  */
typedef struct _NETWORK {
    BOOL bInitialized;

    /* The queues. */
    PACKETQUEUE FreePacketQueue;
    PACKETQUEUE ReceivePacketQueue;
    PACKETQUEUE TransmitPacketQueue;

    /* NT handles for sending. */
    HANDLE hFile;
    HANDLE IoCompletionPort;
    HANDLE hThread;

    /* The network speed&address. */
    ULONG BaudRate;
    UINT8 EthernetAddress[6];
} NETWORK, * PNETWORK;

extern SCODE NetworkInit(PNETWORK Network);
extern void NetworkThread(void * ThreadContext);

/* Packet class methods */
extern void PacketInit(PACKET * Packet);
extern BOOL PacketAllocate(PACKET ** Packet);
extern SCODE PacketWriteFile(HANDLE hFile, PACKET * Packet);
extern SCODE PacketReadFile(HANDLE hFile, PACKET * Packet);

/* Packet queue class methods. */
extern void PacketQueueInit(PACKETQUEUE * PacketQueue);
extern void AcquirePacketQueueLock(PACKETQUEUE * PacketQueue);
extern void ReleasePacketQueueLock(PACKETQUEUE * PacketQueue);
extern void EnqueuePacket(PACKETQUEUE * PacketQueue, PACKET * Packet);
extern PACKET * DequeuePacket(PACKETQUEUE * PacketQueue);
extern UINT32 GetPacketQueueLength(PACKETQUEUE * PacketQueue);
extern UINT32 GetMbufLength(struct mbuf * MBuf);
extern void CopyMbufToBuffer(void * Buffer, struct mbuf * MBuf);
