/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    eb63usart.h
 *
 * Purpose:
 *
 *    Portable Device Driver for serial lines, Atmel m63200 USART customization
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <mmhal.h>

/* Register Offsets
 */
#define USART_CONTROL(_p_)      ((IO_PORT)((_p_)+4*0))
#define USART_MODE(_p_)         ((IO_PORT)((_p_)+4*1))
#define USART_INTR_ENABLE(_p_)  ((IO_PORT)((_p_)+4*2))
#define USART_INTR_DISABLE(_p_) ((IO_PORT)((_p_)+4*3))
#define USART_INTR_MASK(_p_)    ((IO_PORT)((_p_)+4*4)) //10
#define USART_STATUS(_p_)       ((IO_PORT)((_p_)+4*5))
#define USART_RX_DATA(_p_)      ((IO_PORT)((_p_)+4*6))
#define USART_TX_DATA(_p_)      ((IO_PORT)((_p_)+4*7))
#define USART_BAUD(_p_)         ((IO_PORT)((_p_)+4*8)) //20
#define USART_RX_TIMEOUT(_p_)   ((IO_PORT)((_p_)+4*9))
#define USART_TX_TIMEOUT(_p_)   ((IO_PORT)((_p_)+4*10))
#define USART_RX_POINTER(_p_)   ((IO_PORT)((_p_)+4*12)) //30
#define USART_RX_COUNTER(_p_)   ((IO_PORT)((_p_)+4*13))
#define USART_TX_POINTER(_p_)   ((IO_PORT)((_p_)+4*14))
#define USART_TX_COUNTER(_p_)   ((IO_PORT)((_p_)+4*15))

#define PMC_PERIPH_CLOCK_ENABLE ((IO_PORT)0xffff4010)
#define                PMCPC_USART0    0x00000004

#if defined(AT91x40)
#define PIO_DISABLE           ((IO_PORT)0xffff0004)
#define                PIO_TXD0        0x00004000
#define                PIO_RXD0        0x00008000
#define                PIO_TX_SHIFT   7

#define AT_USART_0     ((ADDRESS)0xfffd0000)
/* there are 2 lines, addresses are topsy-turvy */
#define AT_USARTNO(_addr_) (((ADDRESS)_addr_ == AT_USART_0) ? 0 : 1)

#else /* AT91M63200 */

#define PIO_DISABLE           ((IO_PORT)0xfffec004)
#define                PIO_TXD0        0x00008000
#define                PIO_RXD0        0x00010000
#define                PIO_TX_SHIFT   3

#define AT_USART_0     ((ADDRESS)0xfffc0000)
#define AT_USART_SIZE  ((UINT)0x4000)

#define AT_USARTNO(_addr_) (((ADDRESS)_addr_ - AT_USART_0) / AT_USART_SIZE)

#endif

/*   Per-channel Registers    */

/* Control Register
 */
#define                USC_RSTRX       0x00000004
#define                USC_RSTTX       0x00000008
#define                USC_RXEN        0x00000010
#define                USC_RXDIS       0x00000020
#define                USC_TXEN        0x00000040
#define                USC_TXDIS       0x00000080
#define                USC_RSTSTA      0x00000100
#define                USC_STTBRK      0x00000200
#define                USC_STPBRK      0x00000400
#define                USC_STTO        0x00000800
#define                USC_SENDA       0x00001000

/* Mode Register
 */
#define                USM_CLK_MCKI    0x00000000
#define                USM_CLK_MCKI_8  0x00000010
#define                USM_CLK_EXT     0x00000020
#define                USM_BPC_5       0x00000000
#define                USM_BPC_6       0x00000040
#define                USM_BPC_7       0x00000080
#define                USM_BPC_8       0x000000c0
#define                USM_BPC_9       0x00020000
#define                USM_SYNC        0x00000100
#define                USM_EVEN        0x00000000
#define                USM_ODD         0x00000200
#define                USM_SPACE       0x00000400 /* forced 0 */
#define                USM_MARK        0x00000600 /* forced 1 */
#define                USM_NONE        0x00000800
#define                USM_MDROP       0x00000c00
#define                USM_1STOP       0x00000000
#define                USM_1_5STOP     0x00001000
#define                USM_2STOP       0x00002000
#define                USM_ECHO        0x00004000 /* rx->tx */
#define                USM_LOOPBACK    0x00008000 /* tx->rx */
#define                USM_ECHO2       0x0000c000 /* ?same as echo? */
#define                USM_SCK_ENABLE  0x00040000

/* Interrupt related registers
 */
#define                USI_RXRDY       0x00000001
#define                USI_TXRDY       0x00000002
#define                USI_RXBRK       0x00000004
#define                USI_ENDRX       0x00000008
#define                USI_ENDTX       0x00000010
#define                USI_OVRE        0x00000020
#define                USI_FRAME       0x00000040
#define                USI_PARE        0x00000080
#define                USI_TIMEOUT     0x00000100
#define                USI_TXEMPTY     0x00000200
#define                USI_COMMTX      0x40000000 /* ICE debug */
#define                USI_COMMRX      0x80000000 /* ICE debug */

/* Defines to customize the common code
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
  { \
         _i_ = IOSpaceReadInt32(USART_STATUS(_p_)) & \
               IOSpaceReadInt32(USART_INTR_MASK(_p_)); \
         _i_ = _bs32(_i_); \
  }
#define SER_RECEIVE_INTERRUPT(_i_) \
         ((_i_) & (USI_RXRDY|USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         ((_i_) & USI_TXRDY)

#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_ENABLE(_p_), \
                _bs32(USI_RXRDY|USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_ENABLE(_p_),_bs32(USI_TXRDY))
#define SER_DISABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_DISABLE(_p_),_bs32(USI_TXRDY))
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_DISABLE(_p_),~0)

#define SER_ACK_INTERRUPTS(_p_)

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECEIVE_INTERRUPT(_i_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_))

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_STATUS(_p_)) & _bs32(USI_TXRDY))
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_STATUS(_p_)) & _bs32(USI_RXRDY))

#define SER_READ_RECEIVE_BUFFER(_p_) \
         IOSpaceReadInt8(USART_RX_DATA(_p_))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) \
         IOSpaceWriteInt8(USART_TX_DATA(_p_),_c_)

/* BaudRate better be slower than faster, roundup 
 * NB: ASYNC, else no x16;
 */
#define SER_SET_BAUD_RATE(_p_,_b_) \
  { \
         struct MACHINE_INFO mi; \
         UINT _divisor_; \
         PIPROCESS pPrc = CurrentProcess(); \
         (void) pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi ); \
         _divisor_ = Uint64ToUint32(mi.ProcessorSpeed); \
         _divisor_ = (_divisor_/8)/(_b_); \
         _divisor_ = (_divisor_/2) + (_divisor_&1); \
         IOSpaceWriteInt32(USART_BAUD(_p_),_bs32(_divisor_)); \
  }

/* Line parameter encoding: stright from MODE register.
 * Default is 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS \
  (USM_CLK_MCKI|USM_BPC_8|USM_NONE|USM_1STOP)

/* Logic: disable txrx, reset. Set mode.
 * Enable txrx and clear errors.
 */
#define SER_SET_LINE_PARAMS(_p_,_l_) \
  { \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RXDIS|USC_TXDIS)); \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RSTRX|USC_RSTTX)); \
         IOSpaceWriteInt32(USART_MODE(_p_),_bs32(_l_)); \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RXEN|USC_TXEN|USC_RSTSTA|USC_STPBRK)); \
  }

/* We dont do modem control
 */
#define SER_SET_MODEM_BITS(_p_,_b_)

/* Transmitter interrupts when empty
 * [Or, thats the way we want to use it]
 */
#define SER_INTERRUPTS_ON_XMIT_EMPTY 1

/* Error handling, we got to
 */
#define SER_ERRORS(_i_,_p_) \
         ((_i_) & (USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_CLEAR_ERRORS(_i_,_p_) \
  { \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RSTSTA)); \
         _i_ = IOSpaceReadInt32(USART_STATUS(_p_)); \
         _i_ = _bs32(_i_); \
  }

/* Must enable clock to the part, and enable the PIO pins.
 * Find out which USART index it is, to select the right bits.
 * Also, clearup DMAs
 */
#define SER_SPECIAL_INIT(_p_) \
  { \
        UINT _PortNo_ = AT_USARTNO(_p_); \
        SER_DISABLE_INTERRUPTS(_p_); \
        IOSpaceWriteInt32(USART_TX_COUNTER(_p_),0); \
        IOSpaceWriteInt32(USART_RX_COUNTER(_p_),0); \
        IOSpaceWriteInt32(USART_RX_TIMEOUT(_p_),0); \
        IOSpaceWriteInt32(PMC_PERIPH_CLOCK_ENABLE,_bs32(PMCPC_USART0<<_PortNo_));\
        IOSpaceWriteInt32(PIO_DISABLE,_bs32((PIO_TXD0|PIO_RXD0)<<(_PortNo_*PIO_TX_SHIFT)));\
  }

/* HACKALERT needs VM */
#if defined(mips)
#define _v2p(v) ((ADDRESS)(v) & ~0x80000000)
#else
#define _v2p(v) ((ADDRESS)(v))
#endif

/* BUGBUG later */
#define SER_TEST_FOR_MORE_INTERRUPTS(_p_,_i_,_l_)

/* Definitions for DMA usage
 */
#define SER_START_XMIT_DMA(_p_,_b_,_s_) \
  { \
        ADDRESS _phys_ = _v2p(_b_); \
        IOSpaceWriteInt32(USART_TX_POINTER(_p_),_bs32(_phys_)); \
        IOSpaceWriteInt32(USART_TX_COUNTER(_p_),_bs32(_s_)); \
        IOSpaceWriteInt32(USART_INTR_ENABLE(_p_),_bs32(USI_ENDTX)); \
  }

#define SER_STOP_XMIT_DMA(_p_,_s_) \
  { \
        IOSpaceWriteInt32(USART_INTR_DISABLE(_p_),_bs32(USI_ENDTX)); \
        _s_ = IOSpaceReadInt32(USART_TX_COUNTER(_p_)); \
        _s_ = _bs32(_s_); \
        IOSpaceWriteInt32(USART_TX_COUNTER(_p_),0); \
  }

#define SER_ISR_XMIT_COMPLETE(_i_,_p_) \
         ((_i_) & USI_ENDTX)

#define SER_XMIT_DMA_COMPLETE(_p_) \
        (IOSpaceReadInt32(USART_STATUS(_p_)) & _bs32(USI_ENDTX))

/* NB: a timeout value of 1 means intr on every char, each unit is
 * worth 4 bit periods. A minimum of 3 is required to cover one byte.
 */
#define SER_LONG_DMA_TIMEOUT(_p_)  IOSpaceWriteInt32(USART_RX_TIMEOUT(_p_),_bs32(4))
#define SER_SHORT_DMA_TIMEOUT(_p_) IOSpaceWriteInt32(USART_RX_TIMEOUT(_p_),_bs32(1))

#if defined(USE_RCV_DMA_TIMEOUT)
#define _TIMEO SER_LONG_DMA_TIMEOUT
#define SER_START_RECV_DMA_TIMEOUT(_p_) \
  { \
        SER_LONG_DMA_TIMEOUT(_p_); \
        IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_STTO)); \
  }

#else
#define _TIMEO SER_SHORT_DMA_TIMEOUT
#define SER_START_RECV_DMA_TIMEOUT(_p_) \
  { \
        IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_STTO)); \
  }

#endif

#define SER_START_RECV_DMA(_p_,_b_,_s_) \
  { \
        ADDRESS _phys_ = _v2p(_b_); \
        _TIMEO(_p_); \
        IOSpaceWriteInt32(USART_RX_POINTER(_p_),_bs32(_phys_)); \
        IOSpaceWriteInt32(USART_RX_COUNTER(_p_),_bs32(_s_)); \
        IOSpaceWriteInt32(USART_INTR_ENABLE(_p_),_bs32(USI_ENDRX|USI_TIMEOUT)); \
  }

#define SER_RECV_DMA_COUNT(_p_,_s_) \
  { \
        _s_ = IOSpaceReadInt32(USART_RX_COUNTER(_p_)); \
        _s_ = _bs32(_s_); \
  }
#define SER_RECV_DMA_INTERRUPT(_i_,_p_) \
        ((_i_) & (USI_ENDRX|USI_TIMEOUT))

#define SER_RECV_DMA_COMPLETE(_i_,_p_) \
        ((_i_) & USI_ENDRX)

#define SER_RECV_DMA_TIMEOUT(_i_,_p_) \
        ((_i_) & USI_TIMEOUT)

/* Device driver proper
 * Leave it to other files to allow multiple versions at once.
 */
