/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <diagnostics.h>
#include <assert.h>
#include <drivers/serp.h>
#include <drivers/serplex.h>

extern BOOL IsLongTimeout;

/* BUGBUG per link ! */
extern struct _LINK_ERRORS LinkErrors;
static const UINT64 Zero = Int64Initializer(0,0);

#define EXPAND_IN_PLACE 1

/* Put a packet out, obey protocol
 */

UINT PutPacket(LINK_HANDLE hLine, UINT8 *Pkt, UINT Len, UINT BufferSize, UINT8 SerplexAddr)
{
    UINT Size = Len + 1;
    UINT8 *pNext, Byte, PrevByte, ByteNum;
    UINT16 Sum;
    UINT nn = 0;

    /* Make a pass over data to (a) compute xsum (b) check we'll fit in
     */
        
    pNext = Pkt;
    Sum = 0;

// SlowPath:
    Petabyte(hLine,STH_CHAR);
    Petabyte(hLine,SerplexAddr);
    nn = 2;
    ByteNum = 0;
    PrevByte = *pNext;
    for (;Size-- > 0;) {
        if (Size == 0) 
            Byte = ~(*(pNext-1));
        else
            Byte = *pNext++;
        if (Byte == PrevByte && ByteNum < 0xff) {
            ByteNum++;
        } else {
            if (ByteNum < 4) { /* no compressing */
                UINT i;
                for(i=0;i<ByteNum;i++) {
                    if ((PrevByte == STH_CHAR) ||
                        (PrevByte == EOT_CHAR) || 
                        (PrevByte == ESC_CHAR) ||
                        (PrevByte == COMP_CHAR)) 
                    {
                        Petabyte(hLine,ESC_CHAR);
                        nn++;
                    }
                    Petabyte(hLine,PrevByte);
                    Sum += PrevByte;
                    nn++;
                }
            } else {
                Petabyte(hLine, COMP_CHAR);
                nn++;
                Petabyte(hLine, ByteNum);
                Sum += ByteNum;
                nn++;
                Petabyte(hLine,PrevByte);
                Sum += PrevByte;
                nn++;
            }
            PrevByte = Byte;
            ByteNum = 1;
        }
    }

    Petabyte(hLine,EOT_CHAR);
    nn++;
    
    /* Fix checksum on account of STX chars */
    if ((Sum & 0xff) == STH_CHAR)
        Sum++;
    if ((Sum >> 8) == STH_CHAR)
        Sum += (1 << 8);/* this must not rollover */

    Petabyte(hLine,(UINT8)Sum);
    nn++;
    Petabyte(hLine,(UINT8)(Sum>>8));
    nn++;

    /* In case there is buffering down below
     */
    Pushout(hLine,nn);

#if !defined(__HOSTFS_CLIENT)
    DPRINT(("(%d)\n",nn));
#endif
    /* BUGBUG xmit error checking */
    return nn;
}

/* Pull a packet in, obey protocol
 */
BOOL GetPacket(PTRANSPORT pT) /* Compress For client */
{
    LINK_HANDLE hLine = pT->file;
    //, UINT8 *Pkt, UINT Num, UINT8 *SerplexAddr)

    UINT8 *pNext = NULL, *Pkt = NULL, Byte;
    UINT State;
    UINT16 Sum;
    SCODE sc;
    BOOL isCon = FALSE;
    UINT i, ByteNum = 0;
    
    LongTimeout(hLine);

StartOfPacket:

#define STX_WAIT            0
#define STX_ESCAPE          1
#define ADDR_WAIT           2
#define DATA_PULL           3
#define DATA_ESCAPE         4
#define CSUM0               5
#define CSUM1               6
#define COMP_NUM_WAIT       7
#define COMP_BODY_WAIT      8

    Sum = 0;
    State = STX_WAIT;
    
    for (;;) {

        /* Look for input
         */
        sc = Getabyte(hLine,&Byte);

        /* On timeouts, only give up after a long wait
         */
        if (FAILED(sc)) {
            if (IsLongTimeout) {
                LinkErrors.LongTimeout++;
                return FALSE;
                //goto StartOfPacket;
            }
            LongTimeout(hLine);
            continue;
        }

        /* Got a valid byte, see whatnot
         */
        switch (State) {

        case STX_WAIT: /* Wait for STX */
            if ((Byte == STH_CHAR)) {
                State = ADDR_WAIT;/* get address now */
                ShortTimeout(hLine);
            /* else ignore till we get it */
            } else if (Byte == ESC_CHAR) {
                State = STX_ESCAPE;
                LinkErrors.BeforeSTX++;
            } //else /* else ignore till we get it */
                //printf("SYN: x%x ",Byte);
            break;

        case STX_ESCAPE:
            State = STX_WAIT;
            break;

        case ADDR_WAIT:
            State = DATA_PULL;
            pT->IntendedReceiver = Byte;
            if (Byte == CONSerplexAddr) {
                isCon = TRUE;
                pNext = &pT->ConReadBuffer[pT->ConAvail];
                Pkt = pNext;
            } else {
                assert(pT->Avail == 0);
                isCon = FALSE;
                pNext = pT->ReadBuffer;
                Pkt = pT->ReadBuffer;
            }
            break;

        case DATA_PULL: /* Pull data in, end at EOT */
            if (Byte == COMP_CHAR) {
                State = COMP_NUM_WAIT; /* compressed data coming up */
            } else if (Byte == EOT_CHAR) {
                State = CSUM0; /* done with data */
                /* Fix expected checksum on account of STX chars */
                if ((Sum & 0xff) == STH_CHAR)
                    Sum++;
                if ((Sum >> 8) == STH_CHAR)
                    Sum += (1 << 8);/* this must not rollover */
            } else if (Byte == STH_CHAR) {
                LinkErrors.Retransmit++;
                DBGME(0,_tprintf(_T("Retransmission detected..\n")));
                goto StartOfPacket;
            } else if(Byte == ESC_CHAR) {
                    State = DATA_ESCAPE;
            } else if (isCon) {
                if (pNext - pT->ConReadBuffer < MAX_CON_BUF) {
                    *pNext++ = Byte;
                }
                /* else just drop the bytes for overflow */
                Sum += Byte;
            } else {
                if (pNext - pT->ReadBuffer < SERPLEX_MTU)
                    *pNext++ = Byte;
                /* else just drop the bytes for overflow */
                Sum += Byte;
            }
            break;
        
        case DATA_ESCAPE:
            if (isCon) {
                if (pNext - pT->ConReadBuffer < MAX_CON_BUF)
                    *pNext++ = Byte;
                /* otherwise just drop the bytes for overflow */
            } else {
                if (pNext - pT->ReadBuffer < SERPLEX_MTU)
                    *pNext++ = Byte;
            }
            Sum += Byte;
            State = DATA_PULL;
            break;
        
        case COMP_NUM_WAIT: /* uncompress */
            State = COMP_BODY_WAIT;
            ByteNum = Byte;
            Sum += Byte;
            break;
        
        case COMP_BODY_WAIT: /* uncompress */
            for(i=0; i<ByteNum; i++) {
                if (isCon) {
                    if (pNext - pT->ConReadBuffer < MAX_CON_BUF) {
                        *pNext++ = Byte;
                    }
                /* else just drop the bytes for overflow */
                } else {
                    if (pNext - pT->ReadBuffer < SERPLEX_MTU)
                        *pNext++ = Byte;
                    /* else just drop the bytes for overflow */
                }
            }
            Sum += Byte;
            State = DATA_PULL;
            ByteNum = 0;
            break;
        
        case CSUM0: /* get and verify checksum, first byte */

            if ((Sum & 0xff) == Byte)
                State = CSUM1;
            else {
                LinkErrors.FirstByteSum++;
                DBGME(0,_tprintf(_T("Bad pkt checksum, 1st byte %x != %x [%x]\n"),
                       Byte, Sum & 0xff, Sum>>8));
#if defined(__HOSTFS_CLIENT)
                /* We saw the EOT, we expect no more than 2 bytes after that.
                 * ==> this one is the second checksum byte, then no more.
                 */
                return FALSE;
#else
                /* wait for retransmit
                 */
                LongTimeout(hLine);
                goto StartOfPacket;
#endif
            }
            break;

        case CSUM1: /* get and verify checksum, second byte */

            if ((Sum >> 8) != Byte) {
                LinkErrors.SecondByteSum++;
                DBGME(0,_tprintf(_T("Bad pkt checksum, 2nd byte %x != %x\n"),
                       Byte, Sum & 0xff));
#if defined(__HOSTFS_CLIENT)
                /* We saw the EOT, and the first good checksum byte.
                 * ==> this one got garbled, and no more is coming.
                 */
                return FALSE;
#else
                /* wait for retransmit.
                 */
                LongTimeout(hLine);
                goto StartOfPacket;
#endif
            }

            /* NB: We cannot detect missing zeroes with the check-sum,
             * NB: should have shifted also for proper polinomial sum.
             * We'll check the number of bytes instead.
             */

            LongTimeout(hLine);
            /* Done */
            if (isCon) {
                pT->ConAvail += pNext - Pkt;
                //pT->ConConsumed = 0;
            } else {
                pT->Avail = pNext - Pkt;
                pT->Consumed = 0;
            }
            return TRUE;
        }
    }
    return FALSE;/*amusme*/
}
