/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <diagnostics.h>

extern BOOL IsLongTimeout;

/* BUGBUG per link ! */
struct _LINK_ERRORS LinkErrors = {0,};
void PrintLinkErrors(struct _LINK_ERRORS *Errors)
{
#ifdef _DEBUG
    _tprintf(_T("{LE"));
    if (Errors->LongTimeout)
        _tprintf(_T(" timeo %d"), Errors->LongTimeout);
    if (Errors->BeforeSTX)
        _tprintf(_T(" stx %d"), Errors->BeforeSTX);
    if (Errors->Retransmit)
        _tprintf(_T(" rep %d"), Errors->Retransmit);
    if (Errors->FirstByteSum)
        _tprintf(_T(" 1st %d"), Errors->FirstByteSum);
    if (Errors->SecondByteSum)
        _tprintf(_T(" 2nd %d"), Errors->SecondByteSum);
    if (Errors->OutOfSequence)
        _tprintf(_T(" seq %d"), Errors->OutOfSequence);
    if (Errors->BadLength)
        _tprintf(_T(" len %d"), Errors->BadLength);
    if (Errors->Reset)
        _tprintf(_T(" rst %d"), Errors->Reset);
    _tprintf(_T("}\n"));
#endif
}

/* If the connection is reliable and b/w not an issue we can go naked
 */

/* Put a packet out
 */
UINT PutPacketRaw(LINK_HANDLE hLine, struct packet *Pkt, UINT BufferSize)
{
    UnusedParameter(BufferSize);
    Petabytes(hLine,(UINT8*)Pkt,_bs16(Pkt->packet_length));
    return _bs16(Pkt->packet_length);
}

/* Pull a packet in
 */
UINT GetPacketRaw(LINK_HANDLE hLine, struct packet *Pkt, UINT8 ExpectedSeqNo)
{
    UINT8 *pNext, Byte;
    int State, Length;
    SCODE sc;

    LongTimeout(hLine);
 StartOfPacket:
    State = 0;
    pNext = (UINT8*)Pkt;
    for (;;) {

        /* Look for input
         */
        sc = Getabyte(hLine,&Byte);

        /* On timeouts, only give up after a long wait
         */
        if (FAILED(sc)) {
            if (IsLongTimeout) {
                LinkErrors.LongTimeout++;
                return 0;
            }
            LongTimeout(hLine);
            continue;
        }

        /* Got a valid byte, see whatnot
         */
        ShortTimeout(hLine);
        *pNext++ = Byte;
        Length = pNext - (UINT8*)Pkt;

        switch (State) {

        case 0: /* Wait for packet_length to be known */

            if (Length >= 4)
                State = 1;/* get the rest of the data now */

            break;

        case 1: /* Pull data in, only stop at length reached */

#if !defined(__HOSTFS_CLIENT)
            /* Client gets it in native byteorder, server maybe not
             */
            if (Length < _bs16(Pkt->packet_length))
                break;
#else
            if (Length < Pkt->packet_length)
                break;
#endif

            /* Verify packet flow logic, check reconnection/target restart
             */
            if ((Pkt->seqno == ExpectedSeqNo) ||
                (Pkt->packet_type == PACKET_TYPE_RESET)) {
                /* Done */
                return Length;
            }

            LinkErrors.OutOfSequence++;
#if !defined(__HOSTFS_CLIENT)
            _tprintf(_T("ORD: %d != %d\n"),Pkt->seqno, ExpectedSeqNo);
#endif

            /* Send a NACK and repeat
             */
            Pkt->packet_type = PACKET_TYPE_RESET;
            Pkt->seqno = ExpectedSeqNo;
            Pkt->sc = E_RESTART;

            /* Send back, fix length */
            Pkt->packet_length = MIN_PACKET_SIZE;
            (void) PutPacketRaw(hLine, Pkt, Length);

            LongTimeout(hLine);
            goto StartOfPacket;
        }
    }
    return 0;/*amusme*/
}
