/* Copyright (c) Microsoft Corporation. All rights reserved. */

typedef struct _MY_SERIAL_STATE {
  CONNECTION MiState;
  HANDLE hLine;
  UINT BaudRate;
} MY_SERIAL_STATE, *PMY_SERIAL_STATE;

void SerialShortTimeout(PCONNECTION pThis)
{
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    COMMTIMEOUTS Timeouts;

    if (hSer->BaudRate != 0) {

        /* Setup receive timeouts (none on xmit)
         * Knows we read one char at a time, change if that changes.
         */
        Timeouts.ReadIntervalTimeout = 0;
        Timeouts.ReadTotalTimeoutConstant = 1 + 10000/hSer->BaudRate;
        Timeouts.ReadTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutConstant = 0;        

        if (!SetCommTimeouts( hSer->hLine, &Timeouts)) {    
            DPRINT(("SetCommTimeouts() failed (x%x), ignoring..\n",
                    GetLastError()));
        }
    }

    IsLongTimeout = FALSE;
}

void SerialLongTimeout(PCONNECTION pThis)
{
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    COMMTIMEOUTS Timeouts;

    if (hSer->BaudRate != 0) {

        /* Setup receive timeouts (none on xmit)
         * First char might take a long time.
         */
        Timeouts.ReadIntervalTimeout = 0;
        Timeouts.ReadTotalTimeoutConstant = 0;
        Timeouts.ReadTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutConstant = 0;        

        if (!SetCommTimeouts( hSer->hLine, &Timeouts)) {    
            DPRINT(("SetCommTimeouts() failed (x%x), ignoring..\n",
                    GetLastError()));
        }
    }

    IsLongTimeout = TRUE;
}

SCODE SerialGetabyte(PCONNECTION pThis, UINT8 *pByte)
{
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    UINT nBytes;
    OVERLAPPED ol;
    UINT error;

    *(UINT64*)&ol.Offset = 0;
    ol.hEvent = NULL;
    nBytes = 0;

    if (!ReadFile(hSer->hLine,pByte,1,&nBytes, &ol)) {
        error = GetLastError();
        printf("Read failed (x%x)\n",error);
        exit(1);/* aurghfixme*/
    }
    /* mumble.. do we want to make timeouts visible.. ? */
    if (nBytes == 0)
        return E_TIMED_OUT;/*timeout*/
    NOISE((" %x",*pByte));
    return NO_ERROR;
}

SCODE SerialPetabyte(PCONNECTION pThis, UINT8 Byte)
{
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    OVERLAPPED ol;
    UINT nBytes;

    *(UINT64*)&ol.Offset = 0;
    ol.hEvent = NULL;

    NOISE(("%x ",Byte));
    nBytes = 0;
    if (!WriteFile(hSer->hLine,&Byte,1,&nBytes, &ol)) {
        EPRINT(("Write failed (x%x)\n",GetLastError()));
    }
/* for a reeeeally slow board, no flow */
if (hostfsd_debug_level > 3) Sleep(10);
    return NO_ERROR;
}

SCODE SerialPetabytes(PCONNECTION pThis, UINT8 *pBytes, UINT ByteCount)
{
#if 1
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    OVERLAPPED ol;
    UINT nBytes;

    *(UINT64*)&ol.Offset = 0;
    ol.hEvent = NULL;

    NOISE(("%x .. ",*pBytes));
    nBytes = 0;
    if (!WriteFile(hSer->hLine,pBytes,ByteCount,&nBytes, &ol)) {
        EPRINT(("Write failed (x%x)\n",GetLastError()));
    }
    return NO_ERROR;
#else
    SCODE sc = NO_ERROR;
    while (ByteCount-- > 0) {
        sc = SerialPetabyte(pThis,*pBytes++);
        if (FAILED(sc))
            break;
    }
    return sc;
#endif
}

SCODE SerialPushout(PCONNECTION pThis, UINT Count)
{
    /* nothing to do */
    return NO_ERROR;
}

void SerialClose(PCONNECTION pThis)
{
    PMY_SERIAL_STATE hSer = (PMY_SERIAL_STATE)pThis;
    CloseHandle(hSer->hLine);
    hSer->hLine = INVALID_HANDLE_VALUE;
}

const struct IConnectionVtbl SerialVtable = {
    SerialShortTimeout,
    SerialLongTimeout,
    SerialGetabyte,
    SerialPetabyte,
    SerialPetabytes,
    SerialPushout,
    SerialClose
};

CONNECTION *SerialNew(_TCHAR *Arg1, UINT Arg2, UINT Arg3)
{
    HANDLE hLine;
    DCB ControlBlock;
    MY_SERIAL_STATE *Ret;

    hLine = CreateFile( Arg1, GENERIC_READ | GENERIC_WRITE,
                       0, /* exclusive access */
                       NULL, /* no security attributes */
                       OPEN_EXISTING,
                       0, /* Not overlapped */
                       NULL  /* hTemplate - must be NULL for comms */
                       );
    if (hLine == INVALID_HANDLE_VALUE) {
        error = -1;
        printf("Could not open %s\n",Arg1);
        return NULL;
    }

    /* Setup line parameters
     */
    if (Arg2 == 0) /* BaudRate = 0 --> not a true serial line, e.g. USB */
        goto DontBaud;

#if 0
    /* A small xmit Q might help our SLOW partner. 
     */
    (void) SetupComm(hLine, 2*sizeof(struct Packet), 8);
#endif
    /* Get what we wont change
     */
    ControlBlock.DCBlength = sizeof(ControlBlock);
    if (!GetCommState(hLine,&ControlBlock)) {
        error = -2;
        EPRINT(("GetCommState() failed on %s (x%x)\n",Arg1,GetLastError()));
        /* might be a named pipe, ignore */
    }

    /* No parity, no flow..
     */
    ControlBlock.BaudRate = Arg2;
    ControlBlock.fBinary = TRUE;     /* no EOF check */
    ControlBlock.fParity = FALSE;
    ControlBlock.fOutxCtsFlow = FALSE;
    ControlBlock.fOutxDsrFlow = FALSE;
    ControlBlock.fDtrControl = DTR_CONTROL_ENABLE;
    ControlBlock.fDsrSensitivity = FALSE;
    ControlBlock.fTXContinueOnXoff = TRUE;
    ControlBlock.fOutX = FALSE;
    ControlBlock.fInX = FALSE;
    ControlBlock.fErrorChar = FALSE;
    ControlBlock.fNull = FALSE;
    ControlBlock.fRtsControl = RTS_CONTROL_ENABLE;
    ControlBlock.fAbortOnError = FALSE;
    ControlBlock.ByteSize = 8;
    ControlBlock.Parity = NOPARITY;
    ControlBlock.StopBits = ONESTOPBIT;
    if (!SetCommState(hLine,&ControlBlock)) {
        error = -3;
        EPRINT(("SetCommState() failed on %s (x%x)\n",Arg1,GetLastError()));
        /* might be a named pipe, ignore */
    }

 DontBaud:
    Ret = malloc(sizeof(MY_SERIAL_STATE));
    if (Ret == NULL) {
        error = -4;
        EPRINT(("Out of memory?!? (x%x)\n",GetLastError()));
        CloseHandle(hLine);
        return NULL;
    }

    Ret->MiState.v = &SerialVtable;
    Ret->MiState.SequenceNumber = 0;
    Ret->hLine = hLine;
    Ret->BaudRate = Arg2;

    return &Ret->MiState;
}

