/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Adapted from UDP version.
 *
 * NOTE: 
 * The serial line version works for USB too, but performance
 * might suffer. Specifically, this server code sends down one
 * byte at a time, assuming (a) the serial line is much slower
 * at sending than us at producing bytes and (b) the driver
 * uses a fifo  (either software or hardware fifo) to optimize
 * the loading of bytes.
 * In the USB case, with at least one driver, this turned out
 * not to be true.  The driver ended up sending one USB packet
 * per byte, and performance suffered [factor of 4 slowdown].
 * Additionally, the use of timeouts might confuse things.
 */

#define MAX_PKTSIZE (2*1024) /* its more like 500 or so */

typedef struct _MY_USB_STATE {
  CONNECTION MiState;
  HANDLE hUsb;
  UINT InBufferSize;
  UINT InBufferPtr;
  BYTE InBuffer[MAX_PKTSIZE];
  UINT OutBufferPtr;
  BYTE OutBuffer[MAX_PKTSIZE];
} MY_USB_STATE, *PMY_USB_STATE;

/* We dont need timeouts
 */
void UsbShortTimeout(PCONNECTION pThis)
{
    IsLongTimeout = FALSE;
}

void UsbLongTimeout(PCONNECTION pThis)
{
    IsLongTimeout = TRUE;
}

SCODE UsbGetabyte(PCONNECTION pThis, UINT8 *pByte)
{
    PMY_USB_STATE Usb = (PMY_USB_STATE)pThis;
    UINT nBytes;
    OVERLAPPED ol;

    if (Usb->InBufferSize == 0) {

        *(UINT64*)&ol.Offset = 0;
        ol.hEvent = NULL;
        nBytes = 0;

//#define NN MAX_PKTSIZE
#define NN 1
        if (!ReadFile(Usb->hUsb,Usb->InBuffer,NN,&nBytes, &ol)) {
            error = GetLastError();
            printf("Read failed (x%x)\n",error);
            exit(1);/* aurghfixme*/
        }
        /* mumble.. do we want to make timeouts visible.. ? */
        if (nBytes == 0) {
            NOISE((" rcvtimeout"));
            return E_TIMED_OUT;/*timeout*/
        }

#if (NN>1)
        DPRINT(("got %d bytes\n", nBytes));
#endif

        Usb->InBufferSize = nBytes;
        Usb->InBufferPtr = 0;
    }

    /* Getabyte, last byte ?
     */
    *pByte = Usb->InBuffer[Usb->InBufferPtr++];
    if (Usb->InBufferPtr == Usb->InBufferSize)
        Usb->InBufferSize = 0;

    NOISE((" %x",*pByte));
    return NO_ERROR;
}

SCODE UsbPetabyte(PCONNECTION pThis, UINT8 Byte)
{
    PMY_USB_STATE Usb = (PMY_USB_STATE)pThis;

    if (Usb->OutBufferPtr >= MAX_PKTSIZE) {
        EPRINT(("Xmit buffer overflow, dropping!\n"));
        return E_NOT_ENOUGH_MEMORY;
    }

    Usb->OutBuffer[Usb->OutBufferPtr++] = Byte;

    NOISE(("%x ",Byte));

    return NO_ERROR;
}

SCODE UsbPetabytes(PCONNECTION pThis, UINT8 *pBytes, UINT nBytes)
{
    SCODE sc = NO_ERROR;
    while (nBytes-- > 0) {
        sc = UsbPetabyte(pThis,*pBytes++);
        if (FAILED(sc))
            break;
    }
    return sc;
}

SCODE UsbPushout(PCONNECTION pThis, UINT PacketSize)
{
    PMY_USB_STATE Usb = (PMY_USB_STATE)pThis;
    OVERLAPPED ol;
    UINT nBytes;

    if (Usb->OutBufferPtr != PacketSize) {
        EPRINT(("Inconsistent xmit sizes %d != %d\n",
                Usb->OutBufferPtr,PacketSize));
        /* use what we know */
        PacketSize = Usb->OutBufferPtr;
    }

    if (PacketSize == 0) {
        EPRINT(("Zero xmit !?!?\n"));
        return NO_ERROR;
    }

    *(UINT64*)&ol.Offset = 0;
    ol.hEvent = NULL;
    nBytes = 0;
    error = NO_ERROR;
    if (!WriteFile(Usb->hUsb,Usb->OutBuffer,Usb->OutBufferPtr,&nBytes, &ol)) {
        EPRINT(("Write failed (x%x)\n",GetLastError()));
    }

    if (nBytes != Usb->OutBufferPtr) {
        error = GetLastError();
        EPRINT(("WriteFile says %d (!= %d) err = x%x\n",
                nBytes,Usb->OutBufferPtr,error));
    }

    /* Always consume buffer, regardless
     */
    Usb->OutBufferPtr = 0;
    return error;
}

void UsbClose(PCONNECTION pThis)
{
    PMY_USB_STATE Usb = (PMY_USB_STATE)pThis;
    CloseHandle(Usb->hUsb);
    Usb->hUsb = INVALID_HANDLE_VALUE;
}

const struct IConnectionVtbl UsbVtable = {
    UsbShortTimeout,
    UsbLongTimeout,
    UsbGetabyte,
    UsbPetabyte,
    UsbPetabytes,
    UsbPushout,
    UsbClose
};

CONNECTION *UsbNew(_TCHAR *Arg1, UINT Arg2, UINT PortNumber)
{
    MY_USB_STATE *Ret;
    HANDLE hUsb;
    _TCHAR ComN[3+9];

    if (PortNumber == 4444)
        PortNumber = 1;
    _sntprintf(ComN,sizeof ComN, _T("COM%d"), PortNumber);
    hUsb = CreateFile( ComN, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL, /* no security attributes */
                       OPEN_EXISTING,
                       0, /* Not overlapped */
                       NULL  /* hTemplate - must be NULL for comms */
                       );
    if (hUsb == INVALID_HANDLE_VALUE) {
        error = GetLastError();
        printf("Could not open %s on %s (err=x%x)\n",Arg1,ComN,error);
        return NULL;
    }

    Ret = malloc(sizeof(MY_USB_STATE));
    if (Ret == NULL) {
        error = -4;
        printf("Out of memory?!? (x%x)\n",GetLastError());
        goto Bad;
    }

    memset(Ret,0, sizeof *Ret);
    Ret->MiState.v = &UsbVtable;
    Ret->MiState.SequenceNumber = 0;/*doc*/
    Ret->hUsb = hUsb;

    return &Ret->MiState;
 Bad:
    CloseHandle(hUsb);
    return NULL;
}

