/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Reference: "Enhanced Host Controller Interface Specification
 *             for Universal Serial Bus"  Rev 1.0, March 2002.
 */

/* PCI Configuration Space
 */

#define PCIID_NEC_UPD720101               0x00e01033

#define EHCI_CLASSC                       0x0c032000
#define EHCI_SBRN                         0x20

/* Capability Registers
 */
typedef struct _EHCI_CAPS {
    volatile UINT32 VersionAndLength;
#define EHCI_V_CAPLENGTH                  0x000000ff
#define EHCI_V_HCIVERSION                 0xffff0000

    volatile UINT32 HcsParams;
#define EHCI_HCS_DEBUG_PORTNO             0x00f00000
#define EHCI_HCS_P_INDICATOR              0x00010000
#define EHCI_HCS_N_CC                     0x0000f000
#define EHCI_HCS_N_PCC                    0x00000f00
#define EHCI_HCS_PR_RULES                 0x00000080
#define EHCI_HCS_PPC                      0x00000010
#define EHCI_HCS_N_PORTS                  0x0000000f

    volatile UINT32 HccParams;
#define EHCI_HCS_EECP                     0x0000ff00
#define EHCI_HCS_CACHES_WHOLE_FRAME       0x00000080
#define EHCI_HCS_UFRAMES_CACHED           0x00000070
#define EHCI_HCS_ASYNCH_SCHEDULE_PARK     0x00000004
#define EHCI_HCS_PROG_FRAME_LIST_LENGTH   0x00000002
#define EHCI_HCS_64BITS                   0x00000001

    volatile UINT32 HcspPortRoute[64/32]; /* 15 nibbles */
#define EHCI_PR_PORTROUTE_MAX 15
} EHCI_CAPS, *PEHCI_CAPS;

/* Operational Registers
 */
typedef struct _EHCI_OPS {
    volatile UINT32 UsbCmd;
#define EHCI_UC_INTERRUPT_THRESHOLD       0x00ff0000 /* valid values are: */
# define EHCI_UC_1uf                      0x00010000
# define EHCI_UC_2uf                      0x00020000
# define EHCI_UC_4uf                      0x00040000
# define EHCI_UC_8uf                      0x00080000 /* 1ms */
# define EHCI_UC_16uf                     0x00100000
# define EHCI_UC_32uf                     0x00200000
# define EHCI_UC_64uf                     0x00400000
#define EHCI_UC_PARK_MODE                 0x00000800
#define EHCI_UC_PARK_COUNT                0x00000300
#define EHCI_UC_LIGHT_RESET               0x00000080
#define EHCI_UC_ASYNCH_INT_DOORBELL       0x00000040
#define EHCI_UC_ASYNCH_ENABLE             0x00000020
#define EHCI_UC_PERIODIC_ENABLE           0x00000010
#define EHCI_UC_FRAMELIST_SIZE            0x0000000c /* valid: */
# define EHCI_UC_FRAMELIST_1024           0x00000000
# define EHCI_UC_FRAMELIST_512            0x00000004
# define EHCI_UC_FRAMELIST_256            0x00000008
#define EHCI_UC_HCRESET                   0x00000002
#define EHCI_UC_RS                        0x00000001 /* run/stop */
# define EHCI_UC_RUN                      0x00000001

    volatile UINT32 UsbSts;
#define EHCI_STS_ASYNC_ENABLED            0x00008000
#define EHCI_STS_PERIODIC_ENABLED         0x00004000
#define EHCI_STS_RECLAMATION              0x00002000
#define EHCI_STS_HCHalted                 0x00001000
/* The following 6 bits are for UsbIntr also */
#define EHCI_STS_ASYNC_INT                0x00000020
#define EHCI_STS_HOST_SYSTEM_ERROR        0x00000010
#define EHCI_STS_FRAMELIST_ROLLOVER       0x00000008
#define EHCI_STS_PORT_CHANGE              0x00000004
#define EHCI_STS_USBERRINT                0x00000002
#define EHCI_STS_USBINT                   0x00000001
# define EHCI_STS_IMASK                   0x0000005f

    volatile UINT32 UsbIntr;

    volatile UINT32 FrIndex;
#define EHCI_FRI_MASK                     0x00001fff

    volatile UINT32 CtrlDsSegment;

    volatile UINT32 PeriodicListBase;
#define EHCI_PLB_ALIGNMENT                0x00000fff  /* 4kB */

    volatile UINT32 AsyncListAddr;
#define EHCI_LPL_ALIGNMENT                0x0000001f  /* 32B */

    volatile UINT32 Reserved[9];

    volatile UINT32 ConfigFlag;
#define EHCI_CF                           0x00000001

    volatile UINT32 PortSc[1];   /* varsize N_PORTS */
#define EHCI_PS_WKOC_E                    0x00400000 /* rw */
#define EHCI_PS_WKDSCNNT_E                0x00200000 /* rw */
#define EHCI_PS_WKCNNT_E                  0x00100000 /* rw */
#define EHCI_PS_PORT_TEST                 0x000f0000 /* rw valid: */
# define EHCI_PS_J_STATE                  0x00010000
# define EHCI_PS_TEST_K_STATE             0x00020000
# define EHCI_PS_TEST_SE0_NAK             0x00030000
# define EHCI_PS_TEST_PACKET              0x00040000
# define EHCI_PS_TEST_FORCE_ENABLE        0x00050000
#define EHCI_PS_PORT_INDICATOR            0x0000c000 /* rw valid: */
# define EHCI_PS_PI_OFF                   0x00000000
# define EHCI_PS_PI_AMBER                 0x00004000
# define EHCI_PS_PI_GREEN                 0x00008000
#define EHCI_PS_COMPANION_HC_OWNED        0x00002000 /* rw */
#define EHCI_PS_PP                        0x00001000 /* rw valid IFF PPC=1 */
#define EHCI_PS_LINE_STATUS               0x00000c00 /* ro meanings: */
# define EHCI_PS_LINE_SE0                 0x00000000
# define EHCI_PS_LINE_J_STATE             0x00000800
# define EHCI_PS_LINE_K_STATE             0x00000400
#define EHCI_PS_PORT_RESET                0x00000100 /* rw */
#define EHCI_PS_PORT_SUSPEND              0x00000080 /* rw */
#define EHCI_PS_PORT_RESUME               0x00000040 /* rw */
#define EHCI_PS_OVERCURRENT_CHANGE        0x00000020 /* rwc */
#define EHCI_PS_OVERCURRENT               0x00000010 /* ro */
#define EHCI_PS_ENABLED_CHANGE            0x00000008 /* rwc */
#define EHCI_PS_ENABLED                   0x00000004 /* rw */
#define EHCI_PS_CONNECTED_CHANGE          0x00000002 /* rwc */
#define EHCI_PS_CONNECTED                 0x00000001 /* ro */

} EHCI_OPS, *PEHCI_OPS;

/* Data Structures
 */

/* The framelist is an array of pointers:
 */
typedef UINT32 EHCI_FRAMELIST_LINK_POINTER;

#define EHCI_POINTER_MASK                (~EHCI_LPL_ALIGNMENT)
#define EHCI_POINTER_TYPE_TERMINATOR     0x00000001
#define EHCI_POINTER_TYPE_iTD            0x00000000
#define EHCI_POINTER_TYPE_QH             0x00000002
#define EHCI_POINTER_TYPE_siTD           0x00000004
#define EHCI_POINTER_TYPE_FSTN           0x00000006
# define EHCI_POINTER_TYPE_MASK          0x0000000e

/* Isochronous Transfer Descriptor
 * NB: Fields marked 'y'ellow are modified by the HC
 */
typedef struct _EHCI_ITD {
    EHCI_FRAMELIST_LINK_POINTER Next;
    volatile UINT32             Transaction[8];     /* Y */
#define EHCI_ITD_STATUS                  0xf0000000 /* Y meaning: */
# define EHCI_ITD_ACTIVE                 0x80000000 /* Y */
# define EHCI_ITD_DATA_ERROR             0x40000000 /* Y */
# define EHCI_ITD_BABBLE                 0x20000000 /* Y */
# define EHCI_ITD_XactErr                0x10000000 /* Y */
# define EHCI_ITD_STATUS_SHIFT           28
#define EHCI_ITD_BYTELENGTH              0x0fff0000 /* Y */
# define EHCI_ITD_LSHIFT                 16
#define EHCI_ITD_IOC                     0x00008000
#define EHCI_ITD_PG                      0x00007000
# define EHCI_ITD_PGSHIFT                12
#define EHCI_ITD_OFFSET                  0x00000fff
    UINT32                      Pointer[7];
#define EHCI_ITD_BUFFER                  0xfffff000
#define EHCI_ITD_ENDPT                   0x00000f00 /* in Pointer[0] */
#define EHCI_ITD_DEVICE                  0x0000007f /* in Pointer[0] */
#define EHCI_ITD_INPUT                   0x00000800 /* in Pointer[1] */
#define EHCI_ITD_OUTPUT                  0x00000000 /* in Pointer[1] */
#define EHCI_ITD_MAX_PACKET_SIZE         0x000007ff /* in Pointer[1] */
#define EHCI_ITD_MULT                    0x00000003 /* in Pointer[2] */
#define EHCI_ITD_MULT_1xuf               0x00000001
#define EHCI_ITD_MULT_2xuf               0x00000002
#define EHCI_ITD_MULT_3xuf               0x00000003

} EHCI_ITD, *PEHCI_ITD;

#define EHCI_ITD_ALIGNMENT               0x0000003f  /* 64B */

/* Split Transaction Isochronous Transfer Descriptor
 */
typedef struct _EHCI_SITD {
    EHCI_FRAMELIST_LINK_POINTER Next;

    UINT32                      EndpointCaps;
#define EHCI_SITD_INPUT                  0x80000000
#define EHCI_SITD_OUTPUT                 0x00000000
#define EHCI_SITD_PORT_MASK              0x7f000000
#define EHCI_SITD_HUB_MASK               0x007f0000
#define EHCI_SITD_ENDPT_MASK             0x00000f00
#define EHCI_SITD_DEVICE_MASK            0x0000007f

    UINT32                      uFrameSchedule;
#define EHCI_SITD_uFrame_mbz             0xffff0000
#define EHCI_SITD_uFrame_Cmask           0x0000ff00
# define EHCI_SITD_Cmask_SHIFT           8
#define EHCI_SITD_uFrame_Smask           0x000000ff

    volatile UINT32             TransferState;      /* Y */
#define EHCI_SITD_IOC                    0x80000000
#define EHCI_SITD_P                      0x40000000 /* Y meaning: */
# define EHCI_SITD_PAGE0                 0x00000000 /* Y */
# define EHCI_SITD_PAGE1                 0x40000000 /* Y */
#define EHCI_SITD_BYTELENGTH             0x03ff0000 /* Y */
# define EHCI_SITD_LSHIFT                16
#define EHCI_SITD_uFrame_Pmask           0x0000ff00 /* Y */
#define EHCI_SITD_STATUS                 0x000000ff /* Y meaning: */
# define EHCI_SITD_ACTIVE                0x00000080 /* Y */
# define EHCI_SITD_ERR                   0x00000040 /* Y */
# define EHCI_SITD_DATA_ERROR            0x00000020 /* Y */
# define EHCI_SITD_BABBLE                0x00000010 /* Y */
# define EHCI_SITD_XactErr               0x00000008 /* Y */
# define EHCI_SITD_uFrame_Miss           0x00000004 /* Y */
# define EHCI_SITD_END_SPLIT             0x00000002 /* Y */
# define EHCI_SITD_START_SPLIT           0x00000000 /* Y */
# define EHCI_SITD_xxx                   0x00000001 /* Y reserved */

    volatile UINT32             Page0;              /* Y */
#define EHCI_SITD_CURRENT_OFFSET         0x00000fff /* Y (optional) */
#define EHCI_SITD_PAGEMASK               0xfffff000

    volatile UINT32             Page1;              /* Y */
#define EHCI_SITD_TP                     0x00000018 /* Y meaning:*/
# define EHCI_SITD_TP_ALL                0x00000000 /* Y */
# define EHCI_SITD_TP_BEGIN              0x00000008 /* Y */
# define EHCI_SITD_TP_MID                0x00000010 /* Y */
# define EHCI_SITD_TP_END                0x00000018 /* Y */
#define EHCI_SITD_Tcount                 0x00000007 /* Y max:6*/

    EHCI_FRAMELIST_LINK_POINTER Previous;           /* Only T bit, 'invalid' */

} EHCI_SITD, *PEHCI_SITD;

#define EHCI_SITD_ALIGNMENT              0x0000001f  /* 32B */

/* Queue Element Transfer Descriptor
 */
typedef struct _EHCI_QTD {
    EHCI_FRAMELIST_LINK_POINTER Next;               /* T bit only */
    EHCI_FRAMELIST_LINK_POINTER Alternate;          /* T bit only */
#define EHCI_QTD_NakCnt                  0x0000001e /* Overlay only */

    volatile UINT32             Token;              /* Y */
#define EHCI_QTD_DT                      0x80000000 /* Y */
#define EHCI_QTD_LEN                     0x7fff0000 /* Y */
# define EHCI_QTD_LENSHIFT               16
#define EHCI_QTD_IOC                     0x00008000
#define EHCI_QTD_C_Page                  0x00007000 /* Y optional */
#define EHCI_QTD_CERR                    0x00000c00 /* Y */
#define EHCI_QTD_PID                     0x00000300 /* meaning: */
# define EHCI_QTD_PID_OUT                0x00000000
# define EHCI_QTD_PID_IN                 0x00000100
# define EHCI_QTD_PID_SETUP              0x00000200
#define EHCI_QTD_STATUS                  0x000000ff /* Y meaning: */
# define EHCI_QTD_ACTIVE                 0x00000080 /* Y */
# define EHCI_QTD_HALTED                 0x00000040 /* Y */
# define EHCI_QTD_DATA_ERROR             0x00000020 /* Y */
# define EHCI_QTD_BABBLE                 0x00000010 /* Y */
# define EHCI_QTD_XactErr                0x00000008 /* Y */
# define EHCI_QTD_uFrame_Miss            0x00000004 /* Y */
# define EHCI_QTD_END_SPLIT              0x00000002 /* Y */
# define EHCI_QTD_START_SPLIT            0x00000000 /* Y */
# define EHCI_QTD_ERR                    0x00000001 /* Y */
# define EHCI_QTD_PING                   0x00000001

    volatile UINT32             Page[5];            /* Y but only Page0 */
#define EHCI_QTD_CURRENT_OFFSET          0x00000fff /* Y (optional) Page0 */
#define EHCI_QTD_PAGEMASK                0xfffff000

#define EHCI_QTD_C_progmask              0x000000ff /* Page1, overlay */

#define EHCI_QTD_Sbytes                  0x00000fe0 /* Page2, overlay */
#define EHCI_QTD_FrameTag                0x0000001f /* Page2, overlay */

} EHCI_QTD, *PEHCI_QTD;

/* NB: The alignment of this structure is 32bytes but there is a
 * small note in the specs that says that no data structure known
 * to the hardware shall cross a 4KB boundary.
 * Which means we have to bump this up to 64bytes.
 */
#define EHCI_QTD_ALIGNMENT               0x0000003f  /* 64B */

/* Queue Head
 */
typedef struct _EHCI_QH {
    EHCI_FRAMELIST_LINK_POINTER Next;

    UINT32                      EndpointCaps;
#define EHCI_QH_RL                       0xf0000000 /* NAK count reload */
# define EHCI_QH_RL_SHIFT                28
#define EHCI_QH_C                        0x08000000 /* Control endpoint */
#define EHCI_QH_MAX_PACKET_SIZE          0x07ff0000
# define EHCI_QH_MPS_SHIFT               16
#define EHCI_QH_H                        0x00008000 /* Reclamation list head */
#define EHCI_QH_DTC                      0x00004000 /* Data toggle ctl */
#define EHCI_QH_EPS                      0x00003000 /* EPSpeed meaning: */
# define EHCI_QH_12Mbps                  0x00000000
# define EHCI_QH_1_5Mbps                 0x00001000
# define EHCI_QH_480Mbps                 0x00002000
#define EHCI_QH_ENDPT_MASK               0x00000f00
# define EHCI_QH_ENDPT_SHIFT             8
#define EHCI_QH_I                        0x00000080 /* Inactivate next xact */
#define EHCI_QH_DEVICE_MASK              0x0000007f

    UINT32                      EndpointCaps2;
#define EHCI_QH_MULT                     0xc0000000 /* high b/w multiplier: */
# define EHCI_QH_MULT_1xuf               0x40000000
# define EHCI_QH_MULT_2xuf               0x80000000
# define EHCI_QH_MULT_3xuf               0xc0000000
#define EHCI_QH_PORT_MASK                0x3f800000
# define EHCI_QH_PORT_SHIFT              23
#define EHCI_QH_HUB_MASK                 0x007f0000
# define EHCI_QH_HUB_SHIFT               16
#define EHCI_QH_uFrame_Cmask             0x0000ff00
# define EHCI_QH_uFC_SHIFT               8
#define EHCI_QH_uFrame_Imask             0x000000ff

    EHCI_FRAMELIST_LINK_POINTER Current;            /* Y no type */
    EHCI_QTD                    TransferOverlay;    /* Y */

} EHCI_QH, *PEHCI_QH;

/* NB: See above
 */
#define EHCI_QH_ALIGNMENT                 0x0000003f  /* 64B */

/* Periodic Frame Span Traversal Node
 */

typedef struct _EHCI_FSTN {
    EHCI_FRAMELIST_LINK_POINTER Next;
    EHCI_FRAMELIST_LINK_POINTER Previous;           /* Must be QH,2 */
} EHCI_FSTN, *PEHCI_FSTN;


