/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Reference: "OpenHCI Open Host Controller Interface Specification for USB"
 *            Rev 1.0a, September 1999.
 */

/* PCI Configuration Space
 */

#define PCIID_NEC_UPD720101               0x00351033

#define OHCI_CLASSC                       0x0c031000

/* Operational Registers
 */
typedef struct _OHCI_OPS {

    volatile UINT32 Revision;
#define OHCI_REV_MASK                     0x000000ff /* valid values are: */
# define OHCI_REV_1p0                     0x00000010
#define OHCI_REV_LEGACY                   0x00000100

    volatile UINT32 Control;
#define OHCI_C_CBSR                       0x00000003 /* meaning: */
# define OHCI_C_1to1                      0x00000000
# define OHCI_C_2to1                      0x00000001
# define OHCI_C_3to1                      0x00000002
# define OHCI_C_4to1                      0x00000003
#define OHCI_C_PLE                        0x00000004
#define OHCI_C_IE                         0x00000008
#define OHCI_C_CLE                        0x00000010
#define OHCI_C_BLE                        0x00000020
#define OHCI_C_HCFS                       0x000000c0 /* meaning: */
# define OHCI_C_USBRESET                  0x00000000
# define OHCI_C_USBRESUME                 0x00000040
# define OHCI_C_USBOPERATIONAL            0x00000080
# define OHCI_C_USBSUSPENDED              0x000000c0
#define OHCI_C_IR                         0x00000100 /* 1: SMIO */
#define OHCI_C_RWC                        0x00000200
#define OHCI_C_RWE                        0x00000400

    volatile UINT32 CommandStatus;
#define OHCI_CS_HCR                       0x00000001
#define OHCI_CS_CLF                       0x00000002
#define OHCI_CS_BLF                       0x00000004
#define OHCI_CS_OCR                       0x00000008
#define OHCI_CS_SOC                       0x00030000

    /* Same bits for these three guys */
    volatile UINT32 InterruptStatus;
#define OHCI_I_SO                         0x00000001
#define OHCI_I_WDH                        0x00000002
#define OHCI_I_SF                         0x00000004
#define OHCI_I_RD                         0x00000008
#define OHCI_I_UE                         0x00000010
#define OHCI_I_FNO                        0x00000020
#define OHCI_I_RHSC                       0x00000040
#define OHCI_I_OC                         0x40000000
#define OHCI_I_MIE                        0x80000000 /* but not on status */
    volatile UINT32 InterruptEnable;
    volatile UINT32 InterruptDisable;

    volatile UINT32 HCCA;
#define OHCI_HCCA_ALIGNMENT               0x000000ff  /* 256B */

    /* Alignment is common to all EDs and TDs */
    volatile UINT32 PeriodCurrentED;
#define OHCI_ED_ALIGNMENT                0x0000000f  /* 16B */
    volatile UINT32 ControlHeadED;
    volatile UINT32 ControlCurrentED;
    volatile UINT32 BulkHeadED;
    volatile UINT32 BulkCurrentED;
    volatile UINT32 DoneHead;

    volatile UINT32 FmInterval;
#define OHCI_FI_FI                       0x00003fff
# define OHCI_FI_DEFAUL                  0x00002edf
#define OHCI_FI_FSMPS                    0x7fff0000
#define OHCI_FI_FIT                      0x80000000

    volatile UINT32 FmRemaining;
#define OHCI_FM_FR                       0x00003fff
#define OHCI_FM_FRT                      0x80000000

    volatile UINT32 FmNumber;
#define OHCI_FN_FN                       0x0000ffff

    volatile UINT32 PeriodicStart;
#define OHCI_P_PS                        0x00003fff

    volatile UINT32 LSThreshold;
#define OHCI_LS_LST                      0x00000fff

    volatile UINT32 RhDescriptorA;
#define OHCI_RH_NDP                      0x000000ff
#define OHCI_RH_PSM                      0x00000100 /* 1: separately */
#define OHCI_RH_NPS                      0x00000200 /* 1: always on */
#define OHCI_RH_DT                       0x00000400
#define OHCI_RH_OCPM                     0x00000800 /* 1: separately */
#define OHCI_RH_NOCP                     0x00001000 /* 1: no curr prot */
#define OHCI_RH_POTPG                    0xff000000 /* 2ms per */
# define OHCI_RH_POTPG_SHIFT             24

    volatile UINT32 RhDescriptorB;
#define OHCI_RH_DR                       0x0000000f
#define OHCI_RH_PPCM                     0x000f0000

    volatile UINT32 RhStatus;
#define OHCI_ST_LPS                      0x00000001
# define OHCI_ST_CGP                     0x00000001 /* w */
#define OHCI_ST_OCI                      0x00000002 /* ro */
#define OHCI_ST_DRWE                     0x00008000 /* rw */
#define OHCI_ST_LPSC                     0x00010000
#define OHCI_ST_SGP                      0x00010000 /* w */
#define OHCI_ST_OCIC                     0x00020000 /* rw */
#define OHCI_ST_CRWE                     0x80000000 /* w */

#define OHCI_PORT_MAX 15
    volatile UINT32 RhPortStatus[OHCI_PORT_MAX];
#define OHCI_PS_CCS                      0x00000001 /* ro */
#define OHCI_PS_PES                      0x00000002
#define OHCI_PS_PSS                      0x00000004
#define OHCI_PS_POCI                     0x00000008
#define OHCI_PS_PRS                      0x00000010
#define OHCI_PS_PPS                      0x00000100
#define OHCI_PS_LSDA                     0x00000200

#define OHCI_PS_CPE                      0x00000001 /* wo */
#define OHCI_PS_SPE                      0x00000002
#define OHCI_PS_SPS                      0x00000004
#define OHCI_PS_CSS                      0x00000008
#define OHCI_PS_SPR                      0x00000010
#define OHCI_PS_SPP                      0x00000100
#define OHCI_PS_CPP                      0x00000200

#define OHCI_PS_CSC                      0x00010000 /* rw */
#define OHCI_PS_PESC                     0x00020000
#define OHCI_PS_PSSC                     0x00040000
#define OHCI_PS_OCIC                     0x00080000
#define OHCI_PS_PRSC                     0x00100000
# define OHCI_PS_CHANGE                  0x001f0000

} OHCI_OPS, *POHCI_OPS;

/* Data Structures
 */

/* All pointers are 16-bytes aligned
 */
typedef UINT32 OHCI_POINTER;

#define OHCI_POINTER_MASK                (~OHCI_ED_ALIGNMENT)
#define OHCI_POINTER_HALTED              0x00000001
#define OHCI_POINTER_DATA_TOGGLE         0x00000002


/* Endpoint descriptor
 */
typedef struct _OHCI_ED {
    UINT32       Caps;
#define OHCI_ED_FA                       0x0000007f
#define OHCI_ED_EN                       0x00000780
# define OHCI_ED_EN_SHIFT                7
#define OHCI_ED_D                        0x00001800 /* meaning: */
# define OHCI_ED_FROM_TD                 0x00000000
# define OHCI_ED_OUT                     0x00000800
# define OHCI_ED_IN                      0x00001000
# define OHCI_ED_FROM_TDalso             0x00001800
#define OHCI_ED_S                        0x00002000 /* 1: low speed */
#define OHCI_ED_K                        0x00004000 /* 1: sKip */
#define OHCI_ED_F                        0x00008000 /* 1: isoch */
#define OHCI_ED_MPS                      0x07ff0000
# define OHCI_ED_MPS_SHIFT               16

    OHCI_POINTER Tail;
    OHCI_POINTER Head;  /* H,C applies */
    OHCI_POINTER Next;

} OHCI_ED, *POHCI_ED;

/* General transfer descriptor
 */
typedef struct _OHCI_TD {
    UINT32       Caps;
/* The specs say that bits 0-17 are "unused" and the HC should not change them.
 * This means they are reserved for software use and the HC is required to preserve them.
 * Note that only bits 16-17 overlap with similar software bits in a iTD.
 */
#define OHCI_TD_SOFTWARE                 0x0003ffff
#define OHCI_TD_SF                       OHCI_TDI_SF
#define OHCI_TD_ISOCHRONOUS              0x00010000

#define OHCI_TD_R                        0x00040000
#define OHCI_TD_DP                       0x00180000 /* meaning: */
# define OHCI_TD_SETUP                   0x00000000
# define OHCI_TD_OUT                     0x00080000
# define OHCI_TD_IN                      0x00100000
#define OHCI_TD_DI                       0x00e00000
# define OHCI_TD_NO_INTERRUPT            0x00e00000
#define OHCI_TD_T                        0x03000000 /* meaning: */
# define OHCI_TD_DATA0                   0x00000000
# define OHCI_TD_DATA1                   0x01000000
# define OHCI_TD_T_FROM_ED               0x00000000
# define OHCI_TD_T_VALID                 0x02000000
#define OHCI_TD_EC                       0x0c000000
# define OHCI_TD_EC_SHIFT                26
#define OHCI_TD_CC                       0xf0000000 /* meaning: */
# define OHCI_NO_ERROR                   0x00000000
# define OHCI_CRC_ERROR                  0x10000000
# define OHCI_BIT_STUFFING_ERROR         0x20000000
# define OHCI_TOGGLE_MISMATCH_ERROR      0x30000000
# define OHCI_STALL_ERROR                0x40000000
# define OHCI_NO_RESPONSE_ERROR          0x50000000
# define OHCI_PID_CHECK_ERROR            0x60000000
# define OHCI_PID_ERROR                  0x70000000
# define OHCI_DATA_OVERRUN_ERROR         0x80000000
# define OHCI_DATA_UNDERRUN_ERROR        0x90000000
# define OHCI_BUFFER_OVERRUN_ERROR       0xc0000000
# define OHCI_BUFFER_UNDERRUN_ERROR      0xd0000000
# define OHCI_NOT_ACCESSED               0xe0000000
# define OHCI_NOT_ACCESSED_EITHER        0xf0000000
# define OHCI_TD_CC_SHIFT                28

    UINT32       CurrentBufferPointer;

    OHCI_POINTER Next;

    UINT32       BufferEnd;
} OHCI_TD, *POHCI_TD;

/* Isochronous transfer descriptor
 */
typedef struct _OHCI_ITD {
/* The specs say that bits 16-20 are "unused" and the HC should not change them.
 * This means they are reserved for software use and the HC is required to preserve them.
 * Note that only bits 16-17 overlap with similar software bits in a general TD.
 */
#define OHCI_TDI_SOFTWARE                0x001f0000
#define OHCI_TDI_ISOCHRONOUS             OHCI_TD_ISOCHRONOUS

    UINT32       Caps;
#define OHCI_TDI_SF                      0x0000ffff
#define OHCI_TDI_DI                      OHCI_TD_DI
#define OHCI_TDI_FC                      0x07000000
#define OHCI_TDI_CC                      OHCI_TD_CC

    UINT32       BufferPage0;

    OHCI_POINTER Next;

    UINT32       BufferEnd;

    UINT16       OffsetPsw[8];
} OHCI_ITD, *POHCI_ITD;

/* Host controller communication area
 */
typedef struct _OHCI_HCCA {
#define OHCI_NINTED 32
    OHCI_POINTER InterruptTable[OHCI_NINTED];

    UINT32       FrameNumber;

    OHCI_POINTER DoneHead;

    UINT8        Reserved[120]; /* roundup to 256B */

} OHCI_HCCA, *POHCI_HCCA;

