/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    sl811.h
 *
 * Purpose:
 *
 *    Definitions for Cypress's SL811HS USB master.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

typedef struct {
    volatile UINT8  Address;
    volatile UINT8  Data;
} SL811REGS, *PSL811REGS;

/* Register numbers
 */
#define SL_HCR_A      0x00                // rw
#define SL_HBA_A      0x01                // rw
#define SL_HBL_A      0x02                // rw
#define SL_HPID_A     0x03                // wo
# define SL_STATUS_A  0x03                // ro
#define SL_HDA_A      0x04                // wo
#define SL_TC_A       0x04                // ro
#define SL_CTRL1      0x05                // rw
#define SL_INTEN      0x06                // rw
#define SL_SAR        0x07                // rw
#define SL_HCR_B      0x08                // rw
#define SL_HBA_B      0x09                // rw
#define SL_HBL_B      0x0a                // rw
#define SL_HPID_B     0x0b                // wo
# define SL_STATUS_B  0x0b                // ro
#define SL_HDA_B      0x0c                // wo
#define SL_TC_B       0x0c                // ro
#define SL_ISTATUS    0x0d                // rw
#define SL_SOF_LOW    0x0e                // wo
# define SL_REV       0x0e                // ro
#define SL_SOF_HI     0x0f                // rw
#define SL_MEM        0x10                // rw

/* Host Control Register (0x0,0x8)
 */
#define SL_HCR_Arm       0x01
#define SL_HCR_Enable    0x02
#define SL_HCR_Out       0x04
#define SL_HCR_NextIsB   0x08
#define SL_HCR_ISO       0x10
#define SL_HCR_SOF       0x20
#define SL_HCR_DToggle   0x40
#define SL_HCR_Preamble  0x80

/* Host Base Address     (0x1,0x9)
 * NB: Pointer to internal RAM
 */

/* Host Base Length Register  (0x2,0xa)
 */
#define SL_HBL_MAX (256-16)

/* Host PID Register(wo) (0x3,0xb)
 */
#define SL_HPID_EPMASK     0x0f
#define SL_HPID_PMASK      0xf0
#define SL_MAKE_HPID(p,e)  ((UINT8)((((p)&0xf)<<4) | ((e)&0xf)))

#define SL_HPID_SETUP      0xd
#define SL_HPID_IN         0x9
#define SL_HPID_OUT        0x1
#define SL_HPID_SOF        0x5
#define SL_HPID_PREAMBLE   0xc
#define SL_HPID_NAK        0xa
#define SL_HPID_STALL      0xe
#define SL_HPID_DATA0      0x3
#define SL_HPID_DATA1      0xb

/* Status Register A/B (ro)  (0x3,0xb)
 */
#define SL_STATUS_ACK      0x01
#define SL_STATUS_Error    0x02
#define SL_STATUS_Timeout  0x04
#define SL_STATUS_Sequence 0x08
#define SL_STATUS_Setup    0x10
#define SL_STATUS_Overflow 0x20
#define SL_STATUS_NAK      0x40
#define SL_STATUS_STALL    0x80

/* Host Device Address Register  (0x4,0xc)
 */
#define SL_HDA_MASK        0x7f

/* Host Transfer Count Register  (0x4,0xc)
 */

/* Control Register 1     (0x5)
 */
#define SL_CTRL1_SOF_en    0x01
#define SL_CTRL1_reserved  0x82
#define SL_CTRL1_EOF2      0x04
#define SL_CTRL1_Reset     0x08
#define SL_CTRL1_force_J   0x10
#define SL_CTRL1_force_K   0x18
#define SL_CTRL1_LowSpeed  0x20
#define SL_CTRL1_Suspend   0x40

/* Interrupt Enable Register  (0x6)
 */
#define SL_INTEN_DONE_A    0x01
#define SL_INTEN_DONE_B    0x02
#define SL_INTEN_reserved  0x0c
#define SL_INTEN_SOF       0x10
#define SL_INTEN_Insert    0x20
#define SL_INTEN_Detect    0x40
#define SL_INTEN_donno     0x80

/* Slave Address Register (ro)  (0x7)
 */

/* Interrupt Status (rw1c)      (0xd)
 * NB: Same as INTEN above, plus
 */
#define SL_INTEN_D_PLUS    0x80

/* SOF Counter LOW (wo)         (0xe)
 */
#define SL_SOFL_1ms        0xe0  // at 12MHz

/* Hardware Revision Register (ro)   (0xe)
 */
#define SL_REV_MASK        0xf0
# define SL_REV_11H         0x00
# define SL_REV_11HS1p2     0x10
# define SL_REV_11HS1p5     0x20

/* SOF High + Control 2 Register     (0xf)
 */
#define SL_SOFH_SOFmask    0x3f
# define SL_SOFH_set(v)     (((v)>>6)&SL_SOFH_SOFmask)
# define SL_SOFH_get(v)     ((v)<<6)
# define SL_SOFH_1ms        0x2e
#define SL_SOFH_Polarity   0x40
#define SL_SOFH_Master     0x80




