/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbkbtable.c
 *
 * Purpose:
 *
 *    Tables and doc for USB keyboard driver.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


#if 0 /* documentation */
/*

Keycode table, from USB's HUT_1.11 specs.

0 00 Reserved (no event indicated)
1 01 Keyboard ErrorRollOver
2 02 Keyboard POSTFail
3 03 Keyboard ErrorUndefined
4 04 Keyboard a and A
5 05 Keyboard b and B 50
6 06 Keyboard c and C
7 07 Keyboard d and D
8 08 Keyboard e and E
9 09 Keyboard f and F
10 0A Keyboard g and G
11 0B Keyboard h and H
12 0C Keyboard i and I
13 0D Keyboard j and J
14 0E Keyboard k and K
15 0F Keyboard l and L
16 10 Keyboard m and M
17 11 Keyboard n and N
18 12 Keyboard o and O
19 13 Keyboard p and P
20 14 Keyboard q and Q
21 15 Keyboard r and R
22 16 Keyboard s and S
23 17 Keyboard t and T
24 18 Keyboard u and U
25 19 Keyboard v and V 
26 1A Keyboard w and W
27 1B Keyboard x and X
28 1C Keyboard y and Y
29 1D Keyboard z and Z
30 1E Keyboard 1 and !
31 1F Keyboard 2 and @
32 20 Keyboard 3 and #
33 21 Keyboard 4 and $
34 22 Keyboard 5 and %
35 23 Keyboard 6 and ^
36 24 Keyboard 7 and &
37 25 Keyboard 8 and *
38 26 Keyboard 9 and (
39 27 Keyboard 0 and )
40 28 Keyboard Return (ENTER)
41 29 Keyboard ESCAPE 
42 2A Keyboard DELETE (Backspace)
43 2B Keyboard Tab 
44 2C Keyboard Spacebar 
45 2D Keyboard - and (underscore)
46 2E Keyboard = and +
47 2F Keyboard [ and {
48 30 Keyboard ] and }
49 31 Keyboard \ and | 
50 32 Keyboard Non-US # and ~
51 33 Keyboard ; and :
52 34 Keyboard ' and "
53 35 Keyboard Grave Accent and Tilde
54 36 Keyboard, and <
55 37 Keyboard . and >
56 38 Keyboard / and ?
57 39 Keyboard Caps Lock
58 3A Keyboard F1 
59 3B Keyboard F2 
60 3C Keyboard F3 
61 3D Keyboard F4 
62 3E Keyboard F5 
63 3F Keyboard F6 
64 40 Keyboard F7 
65 41 Keyboard F8 
66 42 Keyboard F9 
67 43 Keyboard F10 
68 44 Keyboard F11 
69 45 Keyboard F12 
70 46 Keyboard PrintScreen
71 47 Keyboard Scroll Lock
72 48 Keyboard Pause
73 49 Keyboard Insert
74 4A Keyboard Home
75 4B Keyboard PageUp
76 4C Keyboard Delete Forward
77 4D Keyboard End
78 4E Keyboard PageDown
79 4F Keyboard RightArrow
80 50 Keyboard LeftArrow
81 51 Keyboard DownArrow
82 52 Keyboard UpArrow
83 53 Keypad Num Lock and Clear
84 54 Keypad /
85 55 Keypad * 
86 56 Keypad - 
87 57 Keypad + 
88 58 Keypad ENTER
89 59 Keypad 1 and End 
90 5A Keypad 2 and Down Arrow 
91 5B Keypad 3 and PageDn 
92 5C Keypad 4 and Left Arrow 
93 5D Keypad 5 
94 5E Keypad 6 and Right Arrow 
95 5F Keypad 7 and Home 
96 60 Keypad 8 and Up Arrow 
97 61 Keypad 9 and PageUp 
98 62 Keypad 0 and Insert 
99 63 Keypad . and Delete 
100 64 Keyboard Non-US \ and |
101 65 Keyboard Application
102 66 Keyboard Power
103 67 Keypad = 
104 68 Keyboard F13 
105 69 Keyboard F14 
106 6A Keyboard F15 
107 6B Keyboard F16 
108 6C Keyboard F17 
109 6D Keyboard F18 
110 6E Keyboard F19 
111 6F Keyboard F20 
112 70 Keyboard F21 
113 71 Keyboard F22 
114 72 Keyboard F23 
115 73 Keyboard F24 
116 74 Keyboard Execute 
117 75 Keyboard Help 
118 76 Keyboard Menu 
119 77 Keyboard Select 
120 78 Keyboard Stop 
121 79 Keyboard Again 
122 7A Keyboard Undo 
123 7B Keyboard Cut 
124 7C Keyboard Copy 
125 7D Keyboard Paste 
126 7E Keyboard Find 
127 7F Keyboard Mute 
128 80 Keyboard Volume Up 
129 81 Keyboard Volume Down 
130 82 Keyboard Locking Caps Lock
131 83 Keyboard Locking Num Lock
132 84 Keyboard Locking Scroll Lock
133 85 Keypad Comma
134 86 Keypad Equal Sign
135 87 Keyboard International1
136 88 Keyboard International2
137 89 Keyboard International3
138 8A Keyboard International4
139 8B Keyboard International5
140 8C Keyboard International6
141 8D Keyboard International7
142 8E Keyboard International8
143 8F Keyboard International9
144 90 Keyboard LANG1
145 91 Keyboard LANG2
146 92 Keyboard LANG3
147 93 Keyboard LANG4
148 94 Keyboard LANG5
149 95 Keyboard LANG6
150 96 Keyboard LANG7
151 97 Keyboard LANG8
152 98 Keyboard LANG9
153 99 Keyboard Alternate Erase
154 9A Keyboard SysReq/Attention
155 9B Keyboard Cancel 
156 9C Keyboard Clear 
157 9D Keyboard Prior 
158 9E Keyboard Return 
159 9F Keyboard Separator 
160 A0 Keyboard Out 
161 A1 Keyboard Oper 
162 A2 Keyboard Clear/Again 
163 A3 Keyboard CrSel/Props 
164 A4 Keyboard ExSel 
165-175 A5-CF Reserved 
176 B0 Keypad 00 
177 B1 Keypad 000 
178 B2 Thousands Separator
179 B3 Decimal Separator
180 B4 Currency Unit
181 B5 Currency Sub-unit
182 B6 Keypad ( 
183 B7 Keypad ) 
184 B8 Keypad { 
185 B9 Keypad } 
186 BA Keypad Tab 
187 BB Keypad Backspace 
188 BC Keypad A 
189 BD Keypad B 
190 BE Keypad C 
191 BF Keypad D 
192 C0 Keypad E 
193 C1 Keypad F 
194 C2 Keypad XOR 
195 C3 Keypad ^ 
196 C4 Keypad % 
197 C5 Keypad < 
198 C6 Keypad > 
199 C7 Keypad & 
200 C8 Keypad && 
201 C9 Keypad | 
202 CA Keypad || 
203 CB Keypad : 
204 CC Keypad # 
205 CD Keypad Space 
206 CE Keypad @ 
207 CF Keypad ! 
208 D0 Keypad Memory Store 
209 D1 Keypad Memory Recall 
210 D2 Keypad Memory Clear 
211 D3 Keypad Memory Add 
212 D4 Keypad Memory Subtract 
213 D5 Keypad Memory Multiply 
214 D6 Keypad Memory Divide 
215 D7 Keypad +/- 
216 D8 Keypad Clear 
217 D9 Keypad Clear Entry 
218 DA Keypad Binary 
219 DB Keypad Octal 
220 DC Keypad Decimal 
221 DD Keypad Hexadecimal 
222-223 DE-DF Reserved 
224 E0 Keyboard LeftControl 
225 E1 Keyboard LeftShift 
226 E2 Keyboard LeftAlt 
227 E3 Keyboard Left GUI
228 E4 Keyboard RightControl 
229 E5 Keyboard RightShift 
230 E6 Keyboard RightAlt 
231 E7 Keyboard Right GUI
232-65535 E8-FFFF Reserved 

*/
#endif /* documentation */

#define StateToKey(_s_)  (0x100-(_s_))
#define KeyToState(_k_)  StateToKey(_k_)
#define NUM_LOCK     StateToKey(ST_NUMLOCK)
#define CAPS_LOCK    StateToKey(ST_CAPSLOCK)
#define SCROLL_LOCK  StateToKey(ST_SCRLLOCK)
#define SYS_REQ   3
#define ESCAPE    '\033'

const UINT8 UsbRegularTable[256] = {
    /*0*/ 0,    /*1*/ 0,    /*2*/ 0,    /*3*/ 0,    /*4*/ 'a',
    /*5*/ 'b',  /*6*/ 'c',  /*7*/ 'd',  /*8*/ 'e',  /*9*/ 'f',
    /*10*/ 'g', /*11*/ 'h', /*12*/ 'i', /*13*/ 'j', /*14*/ 'k',
    /*15*/ 'l', /*16*/ 'm', /*17*/ 'n', /*18*/ 'o', /*19*/ 'p',
    /*20*/ 'q', /*21*/ 'r', /*22*/ 's', /*23*/ 't', /*24*/ 'u',
    /*25*/ 'v', /*26*/ 'w', /*27*/ 'x', /*28*/ 'y', /*29*/ 'z',
    /*30*/ '1', /*31*/ '2', /*32*/ '3', /*33*/ '4', /*34*/ '5',
    /*35*/ '6', /*36*/ '7', /*37*/ '8', /*38*/ '9', /*39*/ '0',
    /*40*/ '\r',/*41*/ ESCAPE,/*42*/ '\b',/*43*/ '\t',/*44*/ ' ',
    /*45*/ '-', /*46*/ '=', /*47*/ '[', /*48*/ ']', /*49*/ '\\',
    /*50*/ '#', /*51*/ ';', /*52*/ '\'',/*53*/ '`', /*54*/ ',',
    /*55*/ '.', /*56*/ '/', /*57*/ CAPS_LOCK,/*58*/ 0,/*59*/ 0,
    /*60*/ 0,   /*61*/ 0,   /*62*/ 0,   /*63*/ 0,   /*64*/ 0,
    /*65*/ 0,   /*66*/ 0,   /*67*/ 0,   /*68*/ 0,   /*69*/ 0,
    /*70*/ 0,   /*71*/ SCROLL_LOCK,/*72*/ 0,/*73*/ 0,/*74*/ 0,
    /*75*/ 0,   /*76*/ 0,   /*77*/ 0,   /*78*/ 0,   /*79*/ 0,
    /*80*/ 0,   /*81*/ 0,   /*82*/ 0,   /*83*/ NUM_LOCK,/*84*/ '/',
    /*85*/ '*', /*86*/ '-', /*87*/ '+', /*88*/ '\r',/*89*/ '1',
    /*90*/ '2', /*91*/ '3', /*92*/ '4', /*93*/ '5', /*94*/ '6',
    /*95*/ '7', /*96*/ '8', /*97*/ '9', /*98*/ '0', /*99*/ '.',
    /*100*/ '\\',/*101*/ 0, /*102*/ 0,  /*103*/ '=',/*104*/ 0,
    /*105*/ 0,  /*106*/ 0,  /*107*/ 0,  /*108*/ 0,  /*109*/ 0,
    /*110*/ 0,  /*111*/ 0,  /*112*/ 0,  /*113*/ 0,  /*114*/ 0,
    /*115*/ 0,  /*116*/ 0,  /*117*/ 0,  /*118*/ 0,  /*119*/ 0,
    /*120*/ 0,  /*121*/ 0,  /*122*/ 0,  /*123*/ 0,  /*124*/ 0,
    /*125*/ 0,  /*126*/ 0,  /*127*/ 0,  /*128*/ 0,  /*129*/ 0,
    /*130*/ CAPS_LOCK,/*131*/ NUM_LOCK,/*132*/ 0,/*133*/ ',',/*134*/ '=',
    /*135*/ 0,  /*136*/ 0,  /*137*/ 0,  /*138*/ 0,  /*139*/ 0,
    /*140*/ 0,  /*141*/ 0,  /*142*/ 0,  /*143*/ 0,  /*144*/ 0,
    /*145*/ 0,  /*146*/ 0,  /*147*/ 0,  /*148*/ 0,  /*149*/ 0,
    /*150*/ 0,  /*151*/ 0,  /*152*/ 0,  /*153*/ 0,  /*154*/ SYS_REQ,
    /*155*/ 0,  /*156*/ 0,  /*157*/ 0,  /*158*/ '\r',/*159*/ 0,
    /*160*/ 0,  /*161*/ 0,  /*162*/ 0,  /*163*/ 0,  /*164*/ 0,
};
const UINT8 UsbShiftTable[256] = {
    /*0*/ 0,    /*1*/ 0,    /*2*/ 0,    /*3*/ 0,    /*4*/ 'A',
    /*5*/ 'B',  /*6*/ 'C',  /*7*/ 'D',  /*8*/ 'E',  /*9*/ 'F',
    /*10*/ 'G', /*11*/ 'H', /*12*/ 'I', /*13*/ 'J', /*14*/ 'K',
    /*15*/ 'L', /*16*/ 'M', /*17*/ 'N', /*18*/ 'O', /*19*/ 'P',
    /*20*/ 'Q', /*21*/ 'R', /*22*/ 'S', /*23*/ 'T', /*24*/ 'U',
    /*25*/ 'V', /*26*/ 'W', /*27*/ 'X', /*28*/ 'Y', /*29*/ 'Z',
    /*30*/ '!', /*31*/ '@', /*32*/ '#', /*33*/ '$', /*34*/ '%',
    /*35*/ '^', /*36*/ '&', /*37*/ '*', /*38*/ '(', /*39*/ ')',
    /*40*/ '\r',/*41*/ ESCAPE,/*42*/ '\b',/*43*/ '\t',/*44*/ ' ',
    /*45*/ '_', /*46*/ '+', /*47*/ '{', /*48*/ '}', /*49*/ '|', 
    /*50*/ '~', /*51*/ ':', /*52*/ '\"',/*53*/ '~', /*54*/ '<',
    /*55*/ '>', /*56*/ '?', /*57*/ CAPS_LOCK,/*58*/ 0,/*59*/ 0,
    /*60*/ 0,   /*61*/ 0,   /*62*/ 0,   /*63*/ 0,   /*64*/ 0,
    /*65*/ 0,   /*66*/ 0,   /*67*/ 0,   /*68*/ 0,   /*69*/ 0,
    /*70*/ 0,   /*71*/ SCROLL_LOCK,/*72*/ 0,/*73*/ 0,/*74*/ 0,
    /*75*/ 0,   /*76*/ 0,   /*77*/ 0,   /*78*/ 0,   /*79*/ 0,
    /*80*/ 0,   /*81*/ 0,   /*82*/ 0,   /*83*/ NUM_LOCK,/*84*/ '/',
    /*85*/ '*', /*86*/ '-', /*87*/ '+', /*88*/ '\r',/*89*/ 0,
    /*90*/ 0,   /*91*/ 0,   /*92*/ 0,   /*93*/ 0,   /*94*/ 0,
    /*95*/ 0,   /*96*/ 0,   /*97*/ 0,   /*98*/ 0,   /*99*/ 0,
    /*100*/ '|',/*101*/ 0,  /*102*/ 0,  /*103*/ '=',/*104*/ 0,
    /*105*/ 0,  /*106*/ 0,  /*107*/ 0,  /*108*/ 0,  /*109*/ 0,
    /*110*/ 0,  /*111*/ 0,  /*112*/ 0,  /*113*/ 0,  /*114*/ 0,
    /*115*/ 0,  /*116*/ 0,  /*117*/ 0,  /*118*/ 0,  /*119*/ 0,
    /*120*/ 0,  /*121*/ 0,  /*122*/ 0,  /*123*/ 0,  /*124*/ 0,
    /*125*/ 0,  /*126*/ 0,  /*127*/ 0,  /*128*/ 0,  /*129*/ 0,
    /*130*/ CAPS_LOCK,/*131*/ NUM_LOCK,/*132*/ 0,/*133*/ ',',/*134*/ '=',
    /*135*/ 0,  /*136*/ 0,  /*137*/ 0,  /*138*/ 0,  /*139*/ 0,
    /*140*/ 0,  /*141*/ 0,  /*142*/ 0,  /*143*/ 0,  /*144*/ 0,
    /*145*/ 0,  /*146*/ 0,  /*147*/ 0,  /*148*/ 0,  /*149*/ 0,
    /*150*/ 0,  /*151*/ 0,  /*152*/ 0,  /*153*/ 0,  /*154*/ SYS_REQ,
    /*155*/ 0,  /*156*/ 0,  /*157*/ 0,  /*158*/ '\r',/*159*/ 0,
    /*160*/ 0,  /*161*/ 0,  /*162*/ 0,  /*163*/ 0,  /*164*/ 0,
};
const UINT8 UsbControlTable[256] = {
    /*0*/ 0,    /*1*/ 0,    /*2*/ 0,    /*3*/ 0,    /*4*/ 1,
    /*5*/ 2,    /*6*/ 3,    /*7*/ 4,    /*8*/ 5,    /*9*/ 6,
    /*10*/ 7,   /*11*/ 8,   /*12*/ 9,   /*13*/ 10,  /*14*/ 11,
    /*15*/ 12,  /*16*/ 13,  /*17*/ 14,  /*18*/ 15,  /*19*/ 16,
    /*20*/ 17,  /*21*/ 18,  /*22*/ 19,  /*23*/ 20,  /*24*/ 21,
    /*25*/ 22,  /*26*/ 23,  /*27*/ 24,  /*28*/ 25,  /*29*/ 26,
    /*30*/ 27,  /*31*/ 28,  /*32*/ 29,  /*33*/ 30,  /*34*/ 31,
    /*35*/ 32,  /*36*/ 33,  /*37*/ 34,  /*38*/ 35,  /*39*/ 36,
    /*40*/ 37,  /*41*/ 38,  /*42*/ 39,  /*43*/ 40,  /*44*/ 41,
    /*45*/ 42,  /*46*/ 43,  /*47*/ 44,  /*48*/ 45,  /*49*/ 46,
    /*50*/ 47,  /*51*/ 48,  /*52*/ 49,  /*53*/ 50,  /*54*/ 51,
    /*55*/ 52,  /*56*/ 53,  /*57*/ 54,  /*58*/ 55,  /*59*/ 56,
    /*60*/ 57,  /*61*/ 58,  /*62*/ 59,  /*63*/ 60,  /*64*/ 61,
    /*65*/ 62,  /*66*/ 63,  /*67*/ 64,  /*68*/ 65,  /*69*/ 66,
    /*70*/ 67,  /*71*/ 68,  /*72*/ 69,  /*73*/ 70,  /*74*/ 71,
    /*75*/ 72,  /*76*/ 73,  /*77*/ 74,  /*78*/ 75,  /*79*/ 76,
    /*80*/ 77,  /*81*/ 78,  /*82*/ 79,  /*83*/ 80,  /*84*/ 81,
    /*85*/ 82,  /*86*/ 83,  /*87*/ 84,  /*88*/ 85,  /*89*/ 86,
    /*90*/ 87,  /*91*/ 88,  /*92*/ 89,  /*93*/ 90,  /*94*/ 91,
    /*95*/ 92,  /*96*/ 93,  /*97*/ 94,  /*98*/ 95,  /*99*/ 96,
    /*100*/ 97, /*101*/ 98, /*102*/ 99, /*103*/ 100,/*104*/ 101,
    /*105*/ 102,/*106*/ 103,/*107*/ 104,/*108*/ 105,/*109*/ 106,
    /*110*/ 107,/*111*/ 108,/*112*/ 109,/*113*/ 110,/*114*/ 111,
    /*115*/ 112,/*116*/ 113,/*117*/ 114,/*118*/ 115,/*119*/ 116,
    /*120*/ 117,/*121*/ 118,/*122*/ 119,/*123*/ 120,/*124*/ 121,
    /*125*/ 122,/*126*/ 123,/*127*/ 124,/*128*/ 125,/*129*/ 126,
    /*130*/ 127,/*131*/ 0, 
};
const UINT8 UsbMetaTable[256] = {
    0, /* nothing here yet */
};

