/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    scsilib.h
 *
 * Purpose:
 *
 *    SCSI protocol library definitions
 *
 * References:
 *    "Small Computer Systems Interface (SCSI)", ANSI Draft
 *     X3T9.2/82-2 - Rev 17B December 1985   (SCSI-1)
 *
 *    "Small Computer System Interface - 2 (SCSI-II)", ANSI Draft
 *     X3T9.2/86-109 -  Rev 10C March 1990
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

/* Utilities to build command blocks
 */
void ScsiComposeInquiry(
    PCDB Cdb,
    UINT8 CodePage,
    UINT8 DataLength
    );

void ScsiComposeRequestSense(
    PCDB Cdb,
    UINT8 DataLength
    );

void ScsiComposeModeSense(
    PCDB  Cdb,
	UINT8 Page,
    UINT8 Length
    );

void ScsiComposeStartUnit(
    PCDB  Cdb,
    UINT8 StartFlags
    );

void ScsiComposeTestUnitReady(
    PCDB Cdb
    );

void ScsiComposeRead(
    PCDB Cdb,
    UINT32 SectorNumber,
    UINT8 NumberOfBlocks
    );

void ScsiComposeWrite(
    PCDB Cdb,
    UINT32 SectorNumber,
    UINT8 NumberOfBlocks
    );

void ScsiComposeReadCapacity(
    PCDB Cdb,
    UINT32  LogicalBlockNumber
    );

void ScsiComposeMediumRemoval(
    PCDB Cdb,
    BOOL AllowRemoval
    );

void ScsiComposeLongRead(
    PCDB Cdb,
    UINT32 SectorNumber,
    UINT16 NumberOfBlocks
    );

void ScsiComposeLongWrite(
    PCDB Cdb,
    UINT32 SectorNumber,
    UINT16 NumberOfBlocks
    );


/* Utilities to look at data
 */
void ScsiDumpInquiryData(
    SCSI2_INQUIRY_DATA *Data,
    BOOL Verbose,
    UINT8 *DeviceName  /* NB: Must be at least 31 bytes */
    );

void ScsiDumpCapacityData(
    SCSI_RCAP_DATA *Data
    );

void ScsiDumpSenseData(
    SCSI_SENSE_DATA *Data,
    UINT SenseDataLength
    );

void ScsiDumpAdditionalSenseData(
     UINT8 Key,
     UINT8 Qualifier
     );

/* Error hanlding and printing
 */
SCODE ScsiIsSenseErrorCorrectable(
     SCSI_SENSE_DATA *SenseData
     );

SCODE ScsiQuickTranslateSenseCodes(
    UINT8 SenseKey,
    UINT8 Asc,
    UINT8 Ascq
    );

void ScsiPrintError(
    UINT8 Flags,
    UINT8 ErrorCode,
    UINT8 *AdditionalData
    );

/* Flags */
#define SCSI_ERR_GRAVITY(x) ((UINT8)(x)&0xc0)
#define SCSI_ERR_FATAL      0x80
#define SCSI_ERR_BAD        0x40
#define SCSI_ERR_IGNORE     0x00

#define SCSI_ERR_CLASS(x)   ((UINT8)(x)&0x3f)
#define SCSI_ERR_STATUS     0x01
#define SCSI_ERR_SENSE      0x02
#define SCSI_ERR_MSEL       0x04

#define SCSI_ERR(g,c)       ((UINT8)(g|c))

