/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    transport.h
 *
 * Purpose:
 *
 *    Transport protocol interface definition
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include "cdb.h"

/* Transport Vtable
 */
typedef struct _TRANSPORT *PTRANSPORT;

struct ITransportVtbl {

    SCODE (*ResetRecovery) (
        PTRANSPORT this);

    SCODE (*SendCommand) (
        PTRANSPORT this,
        PCDB Cdb,
        BOOL WaitForCompletion);

    SCODE (*SendData) (
        PTRANSPORT this,
        PCDB Cdb,
        BOOL WaitForCompletion);

    SCODE (*ReadData) (
        PTRANSPORT this,
        PCDB Cdb,
        BOOL WaitForCompletion);

    SCODE (*ReadStatus) (
        PTRANSPORT this,
        PCDB *pCdb);
};

/* The Transport object
 */
typedef struct _TRANSPORT {
    struct ITransportVtbl *v;

    PIUSBHUB Hub;

    PUSBM_REQUEST_BLOCK ControlUrb;
    PUSBM_REQUEST_BLOCK DataUrb;
    PUSBM_REQUEST_BLOCK StatusUrb;
    PUSBM_REQUEST_BLOCK InterruptUrb;

    /* Allow separate MPS on each endpoint, even though.. */
    UINT16 MaximumPacketSizeIn;
    UINT16 MaximumPacketSizeOut;

    UINT16 MaximumPacketSizeInt;
    UINT8 MyUsbAddress;
    UINT8 PortNo;

    UINT8 InEndpoint;
    UINT8 OutEndpoint;
    UINT8 InterruptEndpoint;
    UINT8 MaximumControlPacketSize;

#define DISK_NAME_LENGTH 32
    UINT8 InquiryName[DISK_NAME_LENGTH];
} TRANSPORT;

/* Constructors
 */
extern SCODE CreateBulkTransport(PTRANSPORT this);
extern SCODE CreateCbiTransport(PTRANSPORT this);
extern SCODE CreateCbTransport(PTRANSPORT this);

/* Base utilities
 */
