/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    log.c
 *
 * Purpose:
 *
 *    Simplistic logging tool
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


typedef _TCHAR *NAME;

PRIVATE void PrintReq(int n);
PRIVATE void PrintIntr(UINT8 St);

#define nLOG (256*2)

PRIVATE UINT TheLog[nLOG];
PRIVATE UINT LogHead, LogTail;

PRIVATE void Log(UINT n)
{
    if (LogHead >= nLOG)
        LogHead = 0;
    TheLog[LogHead++] = n;
}

PRIVATE void PrintLog(void)
{
    UINT i;
    int LastReq = 0;

    if (LogHead >= nLOG) {
        _tprintf(_T("!!LogHead %x\n"), LogHead);
        LogHead = 0;
    }

    if (LogTail >= nLOG) {
        _tprintf(_T("!!LogTail %x\n"), LogTail);
        LogTail = 0;
    }

    if (LogHead == LogTail)
        goto Out;

    _tprintf(_T("Log[%d]:"), LogTail);
    for (i = LogTail;i != LogHead;) {

        UINT l = TheLog[i];
        UINT8 St, Tag, Ph, nReq;
        St = (UINT8)l;
        Tag = (UINT8)(l >> 8);
        Ph = (UINT8)(l >> 16);
        nReq = (UINT8)(l >> 24);
        switch (Tag) {
        case 0:
            PrintIntr(St); break;
        case 1:
            _tprintf(_T(" St%x Ph%d "),St,Ph);
            if (nReq > LastReq) {
                PrintReq(LastReq);
                LastReq = nReq;
            }
            break;
        case 2:
            _tprintf(_T(" %x %x %d "),nReq,St,Ph);
            break;
        case 9:
            _tprintf(_T(" %c %d "),St,Ph);
            break;
        default:
            _tprintf(_T(" %x "),l);
            break;
        }
        if (++i == nLOG)
            i = 0;
    }
    LogTail = i;
 Out:
    _tprintf(_T("\n"));
}

/* Other logging cruft
 */

void TheUdcState(PUdc this)
{
    _tprintf(_T("Ctl %x "),this->Control);
    _tprintf(_T("Addr %x "),this->Address);
    _tprintf(_T("MxO %x "),this->MaxOut);
    _tprintf(_T("MxI %x "),this->MaxIn);
    _tprintf(_T("Ctl02 [%x %x %x] "),
             this->EndpointZeroControl,this->EndpointOneControl,
             this->EndpointTwoControl);
    _tprintf(_T("Wc %x Status %x\r\n"),this->WriteCountZero,
             this->Status);
}

PRIVATE void PrintIntr(UINT8 St)
{
    NAME Bitsy[8] = {
        _T("EIR"),         _T("RIR"), 
        _T("TIR"),         _T("SUSIR"), 
        _T("RESIR"),       _T("RSTIR"), 
        _T("40"),          _T("80")
    };
    int i;
    _tprintf(_T("\n"));
    for (i = 0; i < 8; i++)
        if (St & (1 << i))
            _tprintf(_T("%s|"),Bitsy[i]);
}

#define MAXREQ 64
USB_REQUEST_PACKET ToMe[MAXREQ];

PRIVATE void PrintReq(int n)
{
    UINT8 *p;
    UINT i;
    NAME ReqNames[13] = {
        _T("GET_STATUS"), _T("CLEAR_FEATURE"), 
        _T("2"), _T("SET_FEATURE"), 
        _T("4"), _T("SET_ADDRESS"), 
        _T("GET_DESCRIPTOR"), _T("SET_DESCRIPTOR"), 
        _T("GET_CONFIG"), _T("SET_CONFIG"), 
        _T("GET_IFACE"), _T("SET_IFACE"), 
        _T("SYNCH_FRAME")
    };
    NAME DescNames[7] = {
        _T("0"),        _T("DEVICE"), 
        _T("CONFIG"),   _T("STRING"), 
        _T("IFACE"),    _T("ENDP"),
        _T("DEV_QUALIFIER")
    };
    NAME HidReqNames[13] = {
        _T("0"), _T("GET_REPORT"), 
        _T("GET_IDLE"), _T("GET_PROTOCOL"), 
        _T("4"), _T("5"), 
        _T("6"), _T("7"), 
        _T("8"), _T("SET_REPORT"), 
        _T("SET_IDLE"), _T("SET_PROTOCOL"), 
        _T("12")
    };
    NAME HidDescNames[4] = {
        _T("0"),        _T("HID"), 
        _T("REPORT"),   _T("PHYSICAL")
    };

    if (n >= MAXREQ) {
        _tprintf(_T("Index %d too big for Req!\r\n"),n);
        return;
    }
    p = ToMe[n];
    _tprintf(_T("Req[%d]: "),n);
    _tprintf(_T(" %x "), p[0]);
    if (p[1] < 13)
        _tprintf(_T("%s "),((p[0] & 0x7f) == 0x21) ? 
                 HidReqNames[p[1]] : ReqNames[p[1]]);
    else if (p[1] == 0x22)
        _tprintf(_T("SET_CONTROL_LINE_STATE "));
    else
        _tprintf(_T("%x "), p[1]);
    i = 4;
    if (p[1] == USB_R_GET_DESCRIPTOR) {
        if (p[3] < 7)
            _tprintf(_T(" %s[%d]"), DescNames[p[3]], p[2]);
        else if ((p[3] & 0xf0) == 0x20)
            _tprintf(_T(" %s[%d]"), HidDescNames[p[3]&3], p[2]);
        else
            _tprintf(_T(" %x [%d]"), p[3], p[2]);
    } else
        i = 2;

    for (; i < 8; i++) /* they are always 8 */
        _tprintf(_T(" %x"), p[i]);
}

typedef UINT8 ONESHOT[1+8];
#define MAXREP 256
ONESHOT FromMe[MAXREP];

void PrintRep(int n)
{
    UINT8 *p;
    UINT i;

    if (n >= MAXREP) {
        _tprintf(_T("Index %d too big for Rep!\r\n"),n);
        return;
    }
    p = &FromMe[n][1];
    _tprintf(_T("Rep[%d]: "),n);
    for (i = 0; i < FromMe[n][0]; i++)
        _tprintf(_T(" %x"), p[i]);
}


PRIVATE void UdcState(PUDC this, UINT n)
{
    UINT j;

    TheUdcState(this->Registers);

    _tprintf(_T("Sw[%d]: ph%d r%d s%d i%d f%d c%d"), n,
             this->XactPhase, this->nRecvd, this->nSent, 
             this->nInts, this->nReset, this->ConnectionState);
    PrintLog();

    /* Dump request if any */
    if (this->nRecvd) {
#if 0 /* done already in PrintLog() */
        for (j = 0; j < this->nReqs; j++) {
            PrintReq(j);
            _tprintf(_T("\r\n"));
        }
#endif
        this->nRecvd = 0;
        this->nReqs = 0;
    }

    /* Dump reply if any */
    if (this->nReps) {
        for (j = 0; j < this->nReps; j++) {
            PrintRep(j);
            _tprintf(_T("\n"));
        }
        this->nReps = 0;
    }

#if 1
    /* Dump in-data if any */
    if (!CQueueEmpty(this->OneQue)) {
        UINT Avail;
        Avail = this->OneQue.Head; /* cache against interrupts */
        Avail = CQueueCountEval(Avail,this->OneQue.Tail,this->OneQue.Size);

        _tprintf(_T("In[%d]:"),Avail);
        for (j = this->OneQue.Tail + 1;
             Avail && (j < this->OneQue.Size);
             j++) {
            _tprintf(_T(" %x"),this->OneQue.Buffer[j]);
            Avail--;
        }
        for (j = 0; Avail; j++) {
            _tprintf(_T(" %x"),this->OneQue.Buffer[j]);
            Avail--;
        }
        _tprintf(_T("\n"));
    }
#endif
}

