/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    udc.h
 *
 * Purpose:
 *
 *    Definitions for the Intel's UDC slave Universal Serial Bus device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 *
 */

#define UDCCR (*(volatile UINT *)0x80000000)
#define UDCAR (*(volatile UINT *)0x80000004)
#define UDCOMP (*(volatile UINT *)0x80000008)
#define UDCIMP (*(volatile UINT *)0x8000000c)
#define UDCCS0 (*(volatile UINT *)0x80000010)
#define UDCCS1 (*(volatile UINT *)0x80000014)
#define UDCCS2 (*(volatile UINT *)0x80000018)
#define UDCD0 (*(volatile UINT *)0x8000001c)
#define UDCWC (*(volatile UINT *)0x80000020)
#define UDCDR (*(volatile UINT *)0x80000028)
#define UDCSR (*(volatile UINT *)0x80000030)

typedef struct _Udc {
  volatile UINT Control;
# define                UDC_C_UDD       0x01
# define                UDC_C_UDA       0x02
# define                UDC_C_RES       0x04
# define                UDC_C_EIM       0x08
# define                UDC_C_RIM       0x10
# define                UDC_C_TIM       0x20
# define                UDC_C_SRM       0x40
# define                UDC_C_REM       0x80

  volatile UINT Address;
# define                UDC_A_MASK      0x7f

  volatile UINT MaxOut;
  volatile UINT MaxIn;

  volatile UINT EndpointZeroControl;
# define                UDC_C0_OPR      0x01
# define                UDC_C0_IPR      0x02
# define                UDC_C0_SST      0x04
# define                UDC_C0_FST      0x08
# define                UDC_C0_DE       0x10
# define                UDC_C0_SE       0x20
# define                UDC_C0_SO       0x40
# define                UDC_C0_SSE      0x80

  volatile UINT EndpointOneControl;
# define                UDC_C1_RFS      0x01
# define                UDC_C1_RPC      0x02
# define                UDC_C1_RPE      0x04
# define                UDC_C1_SST      0x08
# define                UDC_C1_FST      0x10
# define                UDC_C1_RNE      0x20

  volatile UINT EndpointTwoControl;
# define                UDC_C2_TFS      0x01
# define                UDC_C2_TPC      0x02
# define                UDC_C2_TPE      0x04
# define                UDC_C2_TUR      0x08
# define                UDC_C2_SST      0x10
# define                UDC_C2_FST      0x20

  volatile UINT DataZero;

  volatile UINT WriteCountZero;
# define                UDC_WC_MASK     0x0f

  volatile UINT res1;
  volatile UINT Data;
  volatile UINT res2;

  volatile UINT Status;
# define                UDC_ST_EIR      0x01
# define                UDC_ST_RIR      0x02
# define                UDC_ST_TIR      0x04
# define                UDC_ST_SUSIR    0x08
# define                UDC_ST_RESIR    0x10
# define                UDC_ST_RSTIR    0x20

} Udc, *PUdc;

#define TheUdc ((struct _Udc *)0x80000000)

