/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements GianoFb concrete class */
#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
//#define _DEBUG 1
//#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>
#include <framebuffer.h>

#if 0
#define FB_320x200_BASE 0xffd00000
#define FB_PALETTE_LOAD 0xffd80000
#endif

typedef struct _PGIANOFB {
    const struct IFrameBufferVtbl *v;
    UINT RefCnt;
    ADDRESS VramBase;
#define VramSize (320 * 200) /* at least */
    ADDRESS *PaletteBase;
} *PGIANOFB;

/* Convert between interface and instance pointer and vice versa */
#define pGianoFb(_i_) ((PGIANOFB)((_i_)))
#define iGianoFb(_p_) ((PIFRAMEBUFFER)((_p_)))

PRIVATE SCODE GianoFbQueryInterface (
    PIFRAMEBUFFER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    DBGME(0,_tprintf(_T("GianoFbQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                               &IID_IFrameBuffer);
    return sc;
}

PRIVATE UINT GianoFbAddRef (
    PIFRAMEBUFFER pThis)
{
    DBGME(0,_tprintf(_T("GianoFbAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pGianoFb(pThis)->RefCnt);
}

PRIVATE UINT GianoFbRelease (
    PIFRAMEBUFFER pThis)
{
    PGIANOFB This = pGianoFb(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);

    DBGME(0,_tprintf(_T("GianoFbRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}

PRIVATE SCODE GianoFbMap (
    PIFRAMEBUFFER pThis,
    /*inout*/ PTR * Where,
    /*inout*/ UINT * Length)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbMap(%x %x[%x] %x[%x])\n"), (ADDRESS)This,
                     (ADDRESS) Where, (ADDRESS) *Where,
                     (ADDRESS) Length, (ADDRESS) *Length));

    *Length = VramSize;

    if (*Where != NULL) {
        return E_INVALID_PARAMETER;
    }

    *Where = (PTR) This->VramBase;
    return S_OK;
}

PRIVATE SCODE GianoFbUnmap (
    PIFRAMEBUFFER pThis,
    /*inout*/ PTR Where)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbUnmap(%x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Where));

    if (Where != (PTR) This->VramBase)
        return E_INVALID_PARAMETER;

    return S_OK;
}

PRIVATE SCODE GianoFbGetProperty (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_PROPERTY Property,
    /*inout*/ UINT32 * Value)
{
    PGIANOFB This = pGianoFb(pThis);
    UINT32 v;

    DBGME(1,_tprintf(_T("GianoFbGetProp(%x %x %x)\n"), (ADDRESS)This,
                     Property, (ADDRESS)Value));

    switch (Property) {
    case FB_PROP_TOTAL_SIZE:
        v = VramSize; break;
    case FB_PROP_HORIZONTAL_LENGTH:
        v = 320; break;
    case FB_PROP_HORIZONTAL_VISIBLE_LENGTH:
        v = 320; break;
    case FB_PROP_HORIZONTAL_PORCH:
        v = 0; break;
    case FB_PROP_VERTICAL_LENGTH:
        v = 400; break;
    case FB_PROP_VERTICAL_VISIBLE_LENGTH:
        v = 200; break;
    case FB_PROP_VERTICAL_PORCH:
        v = 0; break;
    case FB_PROP_PIXEL_BITSIZE:
        v = 8; break;
    case FB_PROP_PIXEL_BYTESIZE:
        v = 1; break;
    case FB_PROP_PIXEL_ENCODING:
        v = PIXEL_TYPE_PALETTE_INDEX; break;
    case FB_PROP_MODE:
        v = FB_MODE_320x200_in_320x400_8bpp_PALETTE; break;
    case FB_PROP_PALETTE_LENGTH:
        v = 256; break;
    case FB_PROP_PALETTE_BITSIZE:
        v = 24; break;
    case FB_PROP_PALETTE_BYTESIZE:
        v = 3; break;
    case FB_PROP_PALETTE_ENCODING:
        v = PIXEL_TYPE_R8G8B8; break;
    default:
        return E_NOT_IMPLEMENTED;
    }

    *Value = v;
    return S_OK;
}

PRIVATE SCODE GianoFbSetMode (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_MODE Mode)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbSetM(%x %x) is %x\n"), (ADDRESS)This,
                     Mode, FB_MODE_320x200_in_320x400_8bpp_PALETTE));

    if (Mode != FB_MODE_320x200_in_320x400_8bpp_PALETTE)
        return E_NOT_IMPLEMENTED;

    return S_OK;
}

PRIVATE SCODE GianoFbGetPalette (
    PIFRAMEBUFFER pThis,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbGetP(%x %x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Buffer, Length));

    if ((Buffer == NULL) ||
        (Length < (256 * 3)) ||
        (ExpectedPixelFormat != PIXEL_TYPE_R8G8B8))
        return E_INVALID_PARAMETER;

#if notyet
    memcpy(Buffer,This->PaletteBase+1,256*3);
    return S_OK;
#else
    return E_NOT_IMPLEMENTED;
#endif
}

PRIVATE SCODE GianoFbGetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbGetPE(%x %x %x %x)\n"), (ADDRESS)This,
                     Index, (ADDRESS) Buffer, Length));

    if ((Index >= 256) ||
        (Buffer == NULL) ||
        (Length < 3) ||
        (ExpectedPixelFormat != PIXEL_TYPE_R8G8B8))
        return E_INVALID_PARAMETER;

#if notyet
    memcpy(Buffer,((UINT8*)(This->PaletteBase+1))+(Index * 3),3);
    return S_OK;
#else
    return E_NOT_IMPLEMENTED;
#endif
}

PRIVATE SCODE GianoFbSetPalette (
    PIFRAMEBUFFER pThis,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbSetP(%x %x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Buffer, Length));

    if ((Buffer == NULL) ||
        (Length < (3 * 256)) ||
        (PixelFormat != PIXEL_TYPE_R8G8B8))
        return E_INVALID_PARAMETER;

    *This->PaletteBase = _bs32((ADDRESS) Buffer);
    return S_OK;
}

PRIVATE SCODE GianoFbSetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PGIANOFB This = pGianoFb(pThis);

    DBGME(1,_tprintf(_T("GianoFbSetPE(%x %x %x %x)\n"), (ADDRESS)This,
                     Index, (ADDRESS) Buffer, Length));

    if ((Index >= 256) ||
        (Buffer == NULL) ||
        (Length < 3) ||
        (PixelFormat != PIXEL_TYPE_R8G8B8))
        return E_INVALID_PARAMETER;

#if 1 /*horrible hack */
    {
        static UINT8 Palette[256][3];
        memcpy(&Palette[Index],Buffer,3);
        if (Index == 255)
            *This->PaletteBase = _bs32((ADDRESS) Palette);
        return S_OK;
    }
#elif notyet
    memcpy(((UINT8*)(This->PaletteBase+1))+(Index * 3), Buffer,3);
    return S_OK;
#else
    return E_NOT_IMPLEMENTED;
#endif
}

#include "s_gianofb_v.c"

/* Constructor */
PIFRAMEBUFFER GianoFbNew(
    ADDRESS FramebufferBase,
    ADDRESS PaletteBase)
{
    PIHEAP pHeap = CurrentHeap();
    PGIANOFB This;

    DBGME(1,_tprintf(_T("GianoFbNew(%x %x)\n"),
                     FramebufferBase, PaletteBase));

    This = (PGIANOFB) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *This, 0);
    if (!This)
        return NULL;

    /* Initialize non-zero fields
     */
    This->v = &GianoFbVtbl;
    This->RefCnt = 1;
    This->VramBase = FramebufferBase;
    This->PaletteBase = (ADDRESS*) PaletteBase;

    return iGianoFb(This);
}

/* ** 
*/
SCODE __cdecl GianoFbConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS RegisterBase,
        ADDRESS PaletteBase,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    DBGME(0,_tprintf(_T("GianoFbConstructor(%x %s %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) GianoFbNew(RegisterBase,PaletteBase);

    DBGME(1,_tprintf(_T("GianoFbConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("GianoFbConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl GianoFbDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) GianoFbConstructor
};

static struct _DRIVERCOB GianoFbDriverCobObject = { &GianoFbDriverCobVtable, 1 };

#define GianoFbCobMain gianofbCobMain /* bug in XML scripts */
PIUNKNOWN GianoFbCobMain(void)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("GianoFbCobMain()\n")));

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &GianoFbDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &GianoFbDriverCobObject;
}
