/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements SimFb concrete class */
#include <mmlite.h>
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>

extern PIUNKNOWN SimFbNew(
    ADDRESS BaseAddress,
    ADDRESS PaletteAddress);

/* ** 
*/
SCODE __cdecl SimFbConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS RegisterBase,
        ADDRESS PaletteBase,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    DBGME(0,_tprintf(_T("SimFbConstructor(%x %s %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk));

    pUnk = SimFbNew(RegisterBase,PaletteBase);

    DBGME(1,_tprintf(_T("SimFbConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("SimFbConstructor failed\n")));
        return E_FAIL;
    }
}

/* IUnknown::QI.  Return a pointer of the right kind. */
PRIVATE SCODE MCT DriverCobQueryInterface(PIDEVICECONSTRUCTOR pThis, REFIID Iid,
                                       void **ppObj)
{
    return GenericQueryInterface((PIUNKNOWN) pThis, Iid, ppObj,
                                 &IID_IDeviceConstructor);
}

/* IUnknown::AddRef */
PRIVATE UINT MCT DriverCobAddRef(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    return AtomicInc(&x->RefCnt);
}

/* IUnknown::Release */
PRIVATE UINT MCT DriverCobRelease(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    /* NOTE: If driver has global destructors, it will have to override
     * this method with one that calls CrtExit() when RefCnt == 0.
     */
    return AtomicDec(&x->RefCnt);
}

static struct IDeviceConstructorVtbl SimFbDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    DriverCobQueryInterface,
    DriverCobAddRef,
    DriverCobRelease,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SimFbConstructor
};

static struct _DRIVERCOB SimFbDriverCobObject = { &SimFbDriverCobVtable, 1 };

PIUNKNOWN simfbCobMain(void)
{
    DBGME(0,_tprintf(_T("SimFbCobMain()\n")));

    return (PIUNKNOWN) &SimFbDriverCobObject;
}
