/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * hostname
 *
 * Says what this host's name might be.
 * Which is meaningless for multi-homed hosts but.
 */

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <winsock.h>

/*
 * Whining support.
 */
void PrintError(char *s)
{
    _tprintf(_T("%hs failed, errno x%x\n"), s, WSAGetLastError());
}

WSADATA WsaData;
BOOL Verbose = FALSE;
#define MAXNAMELEN 128
char name[MAXNAMELEN];

int _tmain(INT ArgC, _TCHAR *ArgV[])
{
    int err = 1, i;
    struct hostent *host;

    ParseCmdLine(&ArgC,&ArgV);

    /*
     * Parse options.
     */
    ArgV++; ArgC--;
    while(ArgC > 0 && ArgV[0][0] == _T('-')) {
        switch (ArgV[0][1]) {

        case 'v':
            Verbose = !Verbose;
            break;
        default:
            /* ignore */
            break;
        }
        ArgV++; ArgC--;
    }

    /*
     * Fire up winsock.
     */
    if (WSAStartup(0x0101, &WsaData) == SOCKET_ERROR) {
        PrintError("WSAStartup");
        goto Out;
    }

    /* Find our name one way
     */
    err = gethostname(name,MAXNAMELEN);
    if (err) {
        PrintError("gethostname");
        goto Out;
    }

    /* Make it better if we are asked to
     * NB: This still doesnt really cut it but..
     */
    if (Verbose) {
        /* If this does not work warn and revert to the simple way
         */
        host = gethostbyname(name);
        if (!host) {
            PrintError("gethostbyname");
            goto Simple;
        }

        /* Say what we are known as
         */
        _tprintf(_T("My official name is '%hs'\n"), host->h_name);

        if (host->h_aliases) {
            for (i = 0; host->h_aliases[i]; i++) {
                if (i == 0)
                    _tprintf(_T("My other names are:\n"));
                _tprintf(_T("\t'%hs'\n"),host->h_aliases[i]);
            }
        }

        /* Say what addresses we use
         */
        if (host->h_addr_list) {
            for (i = 0; host->h_addr_list[i]; i++) {
                if (i == 0)
                    _tprintf(_T("My addresses are:\n"));
                _tprintf(_T("\t'%hs'\n"),
                         inet_ntoa(*(struct in_addr *)host->h_addr_list[i]));
            }
        }        

    } else {
    Simple:
        _tprintf(_T("%hs\n"), name);
    }

    /*
     * All done.
     */
    err = 0;
 Out:
    WSACleanup();
    FreeCmdLine(&ArgC,&ArgV);
    return err;
}
