/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <ctype.h>

#define PRIVATE static

PRIVATE struct MORSEMAP {
    char c;
    char *beeps;
} morsemap [] = {
    {'a', ".-"},
    {'b', "-..."},
    {'c', "-.-."},
    {'d', "-.."},
    {'e', "."},
    {'f', "..-."},
    {'g', "--."},
    {'h', "...."},
    {'i', ".."},
    {'j', ".---"},
    {'k', "-.-"},
    {'l', ".-.."},
    {'m', "--"},
    {'n', "-."},
    {'o', "---"},
    {'p', ".--."},
    {'q', "--.-"},
    {'r', ".-."},
    {'s', "..."},
    {'t', "-"},
    {'u', "..-"},
    {'v', "...-"},
    {'w', ".--"},
    {'x', "-..-"},
    {'y', "-.--"},
    {'z', "--.."},
    {'0', "-----"},
    {'1', ".----"},
    {'2', "..---"},
    {'3', "...--"},
    {'4', "....-"},
    {'5', "....."},
    {'6', "-...."},
    {'7', "--..."},
    {'8', "---.."},
    {'9', "----."},
    {'.', ".-.-.-"},
    {',', "--..--"},
    {':', "---..."},
    {'?', "..--.."},
   {'\'', ".----."},
    {'-', "-....-"},
    {'/', "-..-."},
    {'(', "-.--.-"},
    {'"', ".-..-."},

    {'', ".-.-"},
    {'', ".--.-"},
    {'', ".--.-"},
    //{'Ch', "----"},
    {'', "..-.."},
    //{'n~', "--.--"},
    {'', "---."},
    {'', "..--"},

    {'\0', NULL},

};

#ifdef arm
/* CerfBoard specific */
typedef struct {
    volatile UINT32 Level;
    volatile UINT32 Direction;
    volatile UINT32 Set;
    volatile UINT32 Clear;
    volatile UINT32 DetectRaising;
    volatile UINT32 DetectFalling;
    volatile UINT32 DetectStatus;
    volatile UINT32 Alternate;
} GPIO;
#define gpio ((GPIO*)0x90040000)

PRIVATE void SetLed()
{
    gpio->Set = 3;
}

PRIVATE void ClearLed()
{
    gpio->Clear = 3;
}

/* TIME: dot = 1, space = 1, dash = 3, charspace = 3, wordspace = 7
 * mistake = "........" (8 dots)
 */
#endif

PRIVATE TIME BeepSleep;

PRIVATE void beep(UINT n)
{
    //printf("!%d", n);
#ifdef arm
    SetLed();
    for (; n; n--){
        SleepUntil(BeepSleep);
    }
    ClearLed();
#else
    for (; n; n--)
        printf("1");
#endif
}

PRIVATE void pause(UINT n)
{
#ifdef arm
    //printf("_%d", n);
    for (; n; n--) {
        SleepUntil(BeepSleep);
    }
#else
    for (; n; n--)
        printf("0");
#endif
}

PRIVATE void beeps(char *s)
{
    for (; *s != '\0'; s++) {
        if (*s == '-')
            beep(3);
        else
            beep(1);
        pause(1);               /* 1 */
    }
}

PRIVATE void morse(char c)
{
    struct MORSEMAP *mm;

    for (mm = morsemap; mm->c != '\0'; mm++)
        if (mm->c == tolower(c)) {
            beeps(mm->beeps);
            pause(2);           /* 3 */
            break;
        }
}

void MorseString(char *Str)
{
    Int32ToInt64(BeepSleep, TIME_RELATIVE(TIME_MILLIS(250)));

    for (; *Str != _T('\0'); Str++) {
        if (*Str == ' ') {
            pause(4);           /* 7 */
        } else {
            morse((char) *Str);
        }
    }
    pause(4);
}

void DumpGpio(void)
{
#ifdef arm
    printf("GPIO:\n");
    printf("L=%08x D=%08x DR=%08x DF=%08x DS=%08x A=%08x\n",
           gpio->Level, gpio->Direction, gpio->DetectRaising,
           gpio->DetectFalling, gpio->DetectStatus, gpio->Alternate);
    gpio->Set = 1 << 5;         /* make pin J2/19 +5V */
    gpio->Direction |= (1 << 5);   /* output */
    printf("L=%08x D=%08x DR=%08x DF=%08x DS=%08x A=%08x\n",
           gpio->Level, gpio->Direction, gpio->DetectRaising,
           gpio->DetectFalling, gpio->DetectStatus, gpio->Alternate);
#endif
}
