/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "snake.h"

#define MY_MAP_WIDTH 40
#define MY_MAP_HEIGHT 24

void ConsoleInitScreen(PPLATFORM This)
{}

/* These are generic..
 */
UINT8 GenericGetPixel(PPLATFORM pThis, MAP Map, UINT PixNum)
{
    return Map[PixNum];
}

void GenericSetPixel(PPLATFORM pThis, MAP Map, UINT PixNum, UINT8 Val)
{
    Map[PixNum] = Val;
}

void GenericFillMap(PPLATFORM pThis, MAP Map, UINT8 Val)
{
    memset(Map,Val,NUM_PX);
}

/* not generic
 */
void ConsoleShowMap(PPLATFORM This, MAP Map)
{
    UINT8 p;
    int i, x = 0;
    _tprintf(_T("\r"));
    for (i = 0; i < NUM_PX; i++) {
        p = GetPixel(Map,i);
        _tprintf(_T("%c"), (p == PX_FREE) ? ' ' :
               ((p == PX_GOODY) ? 'X' : 'o'));
            
        if (++x == MAP_WIDTH) {
            x = 0;
            _tprintf(_T("|\r\n"));
        }
    }
}

void NoSmile(PPLATFORM This, MAP Map)
{}


BOOL ButtonThreadShouldRun = TRUE;

void THREAD_LINKAGE ButtonThread(THREAD_ARGUMENT arg)
{
    int c;

    while (ButtonThreadShouldRun) {
        c = _gettchar();
        OnButtonPress(c);
    }
}

void ConsoleCreateButtonThread(PPLATFORM This)
{
    SCODE hr;
    PIPROCESS pProcess = CurrentProcess();

    hr = pProcess->v->CreateThread(pProcess, 0,
                                   ButtonThread, This,
                                   2048, /* use 2k stack */
                                   NULL, NULL);
    _tprintf(_T("Button thread created (%x)\r\n"),hr);
}

void ConsoleStopButtonThread(PPLATFORM This)
{
    ButtonThreadShouldRun = FALSE;
}



PlatformVtbl ConsoleVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,

    ConsoleInitScreen,
    GenericGetPixel,
    GenericSetPixel,
    GenericFillMap,
    ConsoleShowMap,
    NoSmile,
    ConsoleCreateButtonThread,
    ConsoleStopButtonThread
};

PPLATFORM CreateConsole(int PlatformType)
{
    _tprintf(_T("Console mode\r\n"));

    MAP_WIDTH = MY_MAP_WIDTH;
    MAP_HEIGHT = MY_MAP_HEIGHT;
    CurrentMap = malloc(MY_MAP_WIDTH*MY_MAP_HEIGHT);
    
    return &ConsoleVtbl;
}
