/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "snake.h"

#include <arm/at91m63200.h>

#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif

BOOL MCT P1PioIsr(void *This, PBOOL pNotMyInterrupt)
/*---------------------------------------------------------------------------*\
 * 
\*---------------------------------------------------------------------------*/
{
    struct _Pio *Pio = (struct _Pio *)This;
    BOOL rv;
    UINT Pins;

    /* See what ISR-enabled button was pressed
     *      NB: This clears the interrupt also.
     */
    Pins = Pio->IntrStatus & Pio->IntrMask;

    /* sanity check */
    if (Pins == 0) {
        *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    /* There might be more than one pending, dispatch all we got */
    rv = FALSE;

    if (Pins & PIOB_PB6) {
        rv = ButtonIsr((void*)3,pNotMyInterrupt);
    }

    if (Pins & PIOB_PB7) {
        rv |= ButtonIsr((void*)4,pNotMyInterrupt);
    }

    if (Pins & PIOB_PB16) {
        rv |= ButtonIsr((void*)5,pNotMyInterrupt);
    }
    return rv;
}

/* Actually.. dont need a thread
 */
void P1CreateButtonThread(PPLATFORM This)
{
    SCODE hr;

    _tprintf(_T("P1: Initializing Buttons\r\n"));

    PioA->Disable = PIOA_IRQ1|PIOA_IRQ2|PIOA_IRQ3;
    PioA->FilterEnable = PIOA_IRQ1|PIOA_IRQ2|PIOA_IRQ3;
    hr = AddDevice((PTR)0, (PTR)ButtonIsr, 0, 30, NULL);
    hr = AddDevice((PTR)1, (PTR)ButtonIsr, 0, 29, NULL);
    hr = AddDevice((PTR)2, (PTR)ButtonIsr, 0, 28, NULL);

    PioB->Disable = PIOB_PB6 | PIOB_PB7 | PIOB_PB16;
    PioB->FilterEnable = PIOB_PB6 | PIOB_PB7 | PIOB_PB16;
    hr = PioB->IntrStatus; /* clear pending */
    hr = AddDevice((PTR)PioB, (PTR)P1PioIsr, DEVICE_FLAGS_LEVEL_TRIGGER, 14,
                   NULL);
    PioB->IntrEnable = PIOB_PB6 | PIOB_PB7 | PIOB_PB16; 

    LcdCreateButtonThread(This);
}

void P1StopButtonThread(PPLATFORM This)
{
    RemoveDevice((PTR)0,30);
    RemoveDevice((PTR)1,29);
    RemoveDevice((PTR)2,28);
    PioB->IntrDisable = PIOB_PB6 | PIOB_PB7 | PIOB_PB16;
    RemoveDevice((PTR)PioB,14);
    LcdStopButtonThread(This);
}
