/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "snake.h"

#include <arm/at91x40.h>

#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif

BOOL MCT P2PioIsr(void *This, PBOOL pNotMyInterrupt)
/*---------------------------------------------------------------------------*\
 * 
\*---------------------------------------------------------------------------*/
{
    struct _Pio *Pio = (struct _Pio *)This;
    BOOL rv;
    UINT Pins;

    /* See what ISR-enabled button was pressed
     *      NB: This clears the interrupt also.
     */
    Pins = Pio->IntrStatus & Pio->IntrMask;

    /* sanity check */
    if (Pins == 0) {
        *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    /* There might be more than one pending, dispatch all we got */
    rv = FALSE;
#define _B0 0x10000
#define _B1 0x80
#define _B2 0x100
#define _B3 0x80000
#define _B4 0x40000
#define _B5 0x20000
#define BUTTS (_B0|_B1|_B2|_B3|_B4|_B5)

    if (Pins & _B1) {
        rv  = ButtonIsr((void*)1,pNotMyInterrupt);
    }
    if (Pins & _B2) {
        rv |= ButtonIsr((void*)2,pNotMyInterrupt);
    }
    if (Pins & _B0) {
        rv |= ButtonIsr((void*)0,pNotMyInterrupt);
    }
    if (Pins & _B5) {
        rv |= ButtonIsr((void*)5,pNotMyInterrupt);
    }
    if (Pins & _B4) {
        rv |= ButtonIsr((void*)4,pNotMyInterrupt);
    }
    if (Pins & _B3) {
        rv |= ButtonIsr((void*)3,pNotMyInterrupt);
    }
    return rv;
}

/* Actually.. dont need a thread
 */
void P2CreateButtonThread(PPLATFORM This)
{
    SCODE hr;
    _tprintf(_T("P2: Initializing Buttons (x%x)\r\n"),BUTTS);
    ThePio->Enable = BUTTS;
    ThePio->OutDisable = BUTTS;
    ThePio->FilterEnable = BUTTS;
    hr = ThePio->IntrStatus; /* clear pending */
    hr = AddDevice((PTR)ThePio, (PTR)P2PioIsr, DEVICE_FLAGS_LEVEL_TRIGGER, 8,
                   NULL);
    ThePio->IntrEnable = BUTTS; 
    LcdCreateButtonThread(This);
}

void P2StopButtonThread(PPLATFORM This)
{
    int x;
    ThePio->IntrDisable = BUTTS;
    x = ThePio->IntrStatus; /* clear pending */
    RemoveDevice((PTR)ThePio,8);

    LcdStopButtonThread(This);
}
