/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "snake.h"

#define MY_MAP_WIDTH (120/2)
#define MY_MAP_HEIGHT (96/2)

UINT8 *FrameBufferBase;
#define FB_WIDTH 320   /* we use 120 */
#define FB_HEIGHT 200  /* we use 96 */

void SimLcdInitScreen(PPLATFORM This)
{
    PINAMESPACE pNs = CurrentNameSpace();
    PIFRAMEBUFFER pDisplay = NULL;
    UINT Size;
    SCODE Sc;
    _TCHAR *Msg;

    Msg = _T("open");
    Sc = BindToObject(pNs, _T("display"), NAME_SPACE_READ | NAME_SPACE_WRITE,
                 &IID_IFrameBuffer,
                 (void**)&pDisplay);
    if (FAILED(Sc)) goto Out;

    /* Set it the way we want it
     */
    Msg = _T("set mode");
    Sc = pDisplay->v->SetMode(pDisplay,FB_MODE_320x200_in_320x400_8bpp_PALETTE);
    if (FAILED(Sc)) goto Out;

    /* Map it
     */
    Msg = _T("map");
    Size = 0;
    Sc = pDisplay->v->Map(pDisplay,(PTR*)&FrameBufferBase,&Size);
    if (FAILED(Sc)) goto Out;

    memset(FrameBufferBase,0xff,FB_WIDTH*FB_HEIGHT);
 Out:
    if (FAILED(Sc))
        _tprintf(_T("Could not %s 'display' (%x)\r\n"), Msg, Sc);
}

typedef UINT8 DOTS[2][2];
PRIVATE DOTS Icons[4] = {
{ {0x00,0x00}, {0x00,0x00} }, /* FREE */
{ {0xff,0xff}, {0xff,0xff} }, /* BUSY */
{ {0xff,0xff}, {0xff,0xff} }, /* unused */
{ {0xff,0x00}, {0x00,0xff} }  /* GOODY */
};

void SimLcdShowMap(PPLATFORM This, MAP Map)
{
    int i, j, b;
    BYTE c, *p;
    p = FrameBufferBase;
    b = 0;
    for (i = 0; i < MAP_HEIGHT; i++) {
        for (j = 0; j < MAP_WIDTH; j++) {
            c = GetPixel(Map,b++);
            p[0] = Icons[c][0][0];
            p[1] = Icons[c][0][1];
            p[FB_WIDTH] = Icons[c][1][0];
            p[FB_WIDTH+1] = Icons[c][1][1];
            p += 2;
        }
        p += FB_WIDTH + (FB_WIDTH-(2*MAP_WIDTH));
    }
}

PlatformVtbl SimLcdVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,

    SimLcdInitScreen,
    GenericGetPixel,
    GenericSetPixel,
    GenericFillMap,
    SimLcdShowMap,
    NoSmile,
    ConsoleCreateButtonThread,
    ConsoleStopButtonThread
};

PPLATFORM CreateSimLcd(int PlatformType)
{
    _tprintf(_T("Lcd on simulator\r\n"));

    MAP_WIDTH = MY_MAP_WIDTH;
    MAP_HEIGHT = MY_MAP_HEIGHT;
    CurrentMap = malloc(MY_MAP_WIDTH*MY_MAP_HEIGHT);
    
    return &SimLcdVtbl;
}

