/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "snake.h"

typedef _TCHAR *string;
string dirnames[6] = {_T("left"), _T("right"), _T("up"), _T("down"), _T("exit"), _T("paused")};

int CurrentDirection = PAUSE;

void OnButtonPress(int ButtonPressed)
{
    switch (ButtonPressed) {
    case Button1: /* oause */
        CurrentDirection = PAUSE;
        break;
    case Button2: /* exit */
        CurrentDirection = TERMINATE;
        break;
    case Button4: /* left */
        CurrentDirection = LEFT;
        break;
    case Button5: /* right */
        CurrentDirection = RIGHT;
        break;
    case Button3: /* up */
        CurrentDirection = UP;
        break;
    case Button6: /* down */
        CurrentDirection = DOWN;
        break;
    }
}



/* Stupid but quick */
UINT16 Random(void)
{
    static UINT16 Previous = 0;
    TIME Now = CurrentTime();
    UINT16 Next;

    /* in 25 usecs, overflows in 1.6 secs */
    Next = (UINT16)(Uint64ToUint32(Now) >> 8);
    if (Previous == 0) {
        Previous = Next;
        return Random();
    }
    Previous = Previous + Next;
    return Previous;
}

#define MAX_LEVEL 5
#define MAX_GRADE 3
typedef struct {
    UINT16 NumGoodies;
    UINT16 Speed;
} LEVEL_INFO;
typedef LEVEL_INFO GRADE_INFO[MAX_LEVEL];
GRADE_INFO LevelInfo[MAX_GRADE] = {
  { /* trivial */
    { 1, 200 },
    { 1, 200 },
    { 1, 200 },
    { 1, 200 },
    { 1, 200 }
  },
  { /* normal */
    { 1, 400 },
    { 2, 200 },
    { 3, 150 },
    { 4, 100 },
    { 4, 80 }
  },
  { /* harder */
    { 3, 150 },
    { 4, 100 },
    { 4, 100 },
    { 5, 80 },
    { 6, 80 }
  }
};
UINT8 CurLevel = 0;
UINT8 CurGrade = 1;
#define CurSpeed LevelInfo[CurGrade][CurLevel].Speed
#define CurGoodies LevelInfo[CurGrade][CurLevel].NumGoodies

UINT16 GoodiesLeft = 0;
UINT16 CurPos = 0;
UINT16 CurTime;
UINT CurScore, HighScore;
UINT TimeLimit;

void CheckHighScore(UINT Score)
{
    if (Score > HighScore) {
        _tprintf(_T("Previous HighScore %d\r\n"),HighScore);
        HighScore = Score;
    }
}

void YouWin(void)
{
    Smile();
    _tprintf(_T("You win! Score %d\r\n"),CurScore);
    CheckHighScore(CurScore);
}

void YouLose(int Level)
{
    _tprintf(_T("You lost on level %d, score %d\r\n"),
           Level,CurScore);
    CheckHighScore(CurScore);
}

BOOL AskIfPlayAgain(int GameNumber)
{
#if 0
    int yn;
    _tprintf(_T("Wanna play%s ? "), (GameNumber) ? _T(" again") : _T(""));
    yn = _gettchar();
    _tprintf(_T("\r\n"));
    return ((yn == _T('y')) || (yn == _T('Y')));
#else
    _tprintf(_T("New game (%d)\r\n"),GameNumber);
    return TRUE;
#endif
}

void NewMap(void)
{
    int i;
    UINT16 r = 0;

    _tprintf(_T("Level %d\r\n"),CurLevel);

    ZeroMap(CurrentMap,PX_FREE);

    GoodiesLeft = CurGoodies;
    for (i = GoodiesLeft; i >= 0; i--) {
        r = Random() % NUM_PX;
        if (GetPixel(CurrentMap,r) != PX_FREE) {
            i++;
            continue;
        }
        TRACE(printf("Goody at %d (%d,%d)",r,
                     r - ((r / MAP_WIDTH) * MAP_WIDTH), 
                     r / MAP_WIDTH));
        SetPixel(CurrentMap,r, (UINT8)((i == 0) ? PX_BUSY : PX_GOODY));
    }
    CurPos = r;
    CurrentDirection = PAUSE;
    TimeLimit = NUM_PX/8;
    CurTime = 0;
}

/* TRUE -> not done */
BOOL GameTick(void)
{
    UINT8 CurX, CurY;

    CurY = CurPos / MAP_WIDTH;
    CurX = CurPos - (CurY * MAP_WIDTH);
    TRACE(printf("x %d y %d ",CurX,CurY));

    switch (CurrentDirection) {
    case LEFT: if (CurX) CurX--; break;
    case RIGHT: if (CurX < (MAP_WIDTH-1)) CurX++; break;
    case UP: if (CurY < (MAP_HEIGHT-1)) CurY++; break;
    case DOWN: if (CurY) CurY--; break;
    case PAUSE: CurTime--; return TRUE;
    default:
        return FALSE;
    }
    TRACE(printf(" -> x %d y %d\r\n",CurX,CurY));
    CurPos = (CurY * MAP_WIDTH) + CurX;

    switch (GetPixel(CurrentMap,CurPos)) {
    case PX_BUSY:
        /* you lose */
        return FALSE;
    case PX_GOODY:
        CurScore += (TimeLimit > CurTime) ?
            (TimeLimit - CurTime) * (CurLevel + 1) : 1;
        _tprintf(_T("\n\rScore: %d Goodies left: %d "),CurScore,GoodiesLeft-1);
        if (--GoodiesLeft == 0) 
            /* level clear */
            return FALSE;
        /* fall through */
    case PX_FREE:
        SetPixel(CurrentMap,CurPos,PX_BUSY);
        return TRUE;
    }
    return FALSE;
}

void Sleep(int nMsecs)
{
    TIME SleepTime;
    Int32ToInt64(SleepTime, TIME_RELATIVE(TIME_MILLIS(CurSpeed)));
    SleepUntil(SleepTime);

    TRACE(printf("Goodies %d CurPos %d CurDir %s\r\n", 
                 GoodiesLeft, CurPos, dirnames[CurrentDirection]));
}

/* TRUE -> level complete */
BOOL NewGame(void)
{
    NewMap();

    do {
        ShowMap(CurrentMap);
        Sleep(CurSpeed);
        CurTime++;
    } while (GameTick());

    if (GoodiesLeft)
        YouLose(CurLevel);
    return (GoodiesLeft == 0);
}

PPLATFORM Hw = NULL;
PLATFORM_NEW Constructors[nPLATFORMS] = {
    CreateConsole, /* 0 */
    CreateSimLcd,  /* 1 */
    CreateZ,       /* 2 P1 */
    CreateZ,       /* 3 P2 */
    CreateZ        /* 4 P2b */
};

MAP CurrentMap = NULL;
int MAP_WIDTH = 0;
int MAP_HEIGHT = 0;

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    UINT i = 0;

    /* First argument, plaform type is ~~mandatory
     */
    sc = ParseCmdLine(&argc, &argv);
    if ((FAILED(sc)) || (argc == 1)) {
        _tprintf(_T("No args, assuming console mode\r\n"));
        Hw = CreateConsole(0);
    } else {
        i = _ttoi(argv[1]);
        if (i < nPLATFORMS) {
            Hw = Constructors[i](i);
        } else {
            _tprintf(_T("Usage: %s [PlatformType <0..%d>] [Level <0..%d>]\r\n"),
                   argv[0],nPLATFORMS,MAX_GRADE);
            return -1;
        }
        argc--,argv++;
    }

    if (argc > 1) {
        i = _ttoi(argv[1]);
        if (i < MAX_GRADE)
            CurGrade = (UINT8)i;
        argc--,argv++;
    }

    if (argc > 1) {
        RemoteName = argv[1];
    }

    InitScreen();
    CreateButtonThread();
    i = 0;
    while (AskIfPlayAgain(i++)) {

        CurScore = 0;
        CurLevel = 0;
        while (NewGame()) {
            if (++CurLevel < MAX_LEVEL)
                continue;
            YouWin();
            break;
        }
    }
    StopButtonThread();
    return 0;
}
