/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <drivers/drivers.h>
#include <framebuffer.h>

/* Buttons, if we got them.
 */
#define nBUTTONS 6

#define Button4 'l'
#define Button5 'r'
#define Button3 'd'
#define Button6 'u'
#define Button1 'p'
#define Button2 'q'

extern void OnButtonPress(int ButtonPressed);
extern void OnRemoteButtonPress(int ButtonPressed);

/* They control which way the snake goes
 */
#define LEFT 0
#define RIGHT 1
#define UP 2
#define DOWN 3
#define TERMINATE 4
#define PAUSE 5
#define DIR_BITS 0x3

extern int CurrentDirection;

/* The map is device-specific, but
 */
typedef UINT8 * MAP;
#define PX_FREE   ((UINT8)0)
#define PX_BUSY   ((UINT8)1)
#define PX_GOODY  ((UINT8)3)

extern MAP CurrentMap;
extern int MAP_WIDTH;
extern int MAP_HEIGHT;
#define NUM_PX (MAP_WIDTH*MAP_HEIGHT)


/* Utils
 */
#define TRACE(x)  //x

extern _TCHAR *RemoteName;

/* Interface to the display machinery, platform-specific
 * Bypass vtable for now.
 */
typedef struct PlatformVtbl *PPLATFORM;
typedef struct PlatformVtbl {
    __VTABLE_COMPATIBILITY_FILLER_DECLARATION

    SCODE   (MCT *QueryInterface)(PPLATFORM pThis, REFIID Iid, void* *ppNew);
    UINT    (MCT *AddRef)(PPLATFORM pThis);
    UINT    (MCT *Release)(PPLATFORM pThis);

    void    (MCT *InitScreen)(PPLATFORM pThis);

    UINT8   (MCT *GetPixel)(PPLATFORM pThis, MAP Map, UINT PixNum);
    void    (MCT *SetPixel)(PPLATFORM pThis, MAP Map, UINT PixNum, UINT8 Val);
    void    (MCT *FillMap)(PPLATFORM pThis, MAP Map, UINT8 Val);
    void    (MCT *ShowMap)(PPLATFORM pThis, MAP Map);
    void    (MCT *Smile)(PPLATFORM pThis, MAP Map);

    void    (MCT *CreateButtonThread)(PPLATFORM pThis);
    void    (MCT *StopButtonThread)(PPLATFORM pThis);
} PlatformVtbl;

extern PPLATFORM Hw;

#if 1
#define GetPixel(_m_,_n_) _m_[_n_]
#define SetPixel(_m_,_n_,_v_) _m_[_n_] = _v_
#define ZeroMap(_m_,_v_) memset(_m_,_v_,NUM_PX)
#else
#define GetPixel(_m_,_n_) Hw->GetPixel(Hw,_m_,_n_)
#define SetPixel(_m_,_n_,_v_) Hw->SetPixel(Hw,_m_,_n_,_v_)
#define ZeroMap(_m_,_v_) Hw->FillMap(Hw,_m_,_v_)
#endif

#define ShowMap(_m_) Hw->ShowMap(Hw,_m_)
#define Smile() Hw->Smile(Hw,CurrentMap)
#define InitScreen() Hw->InitScreen(Hw)
#define CreateButtonThread() Hw->CreateButtonThread(Hw)
#define StopButtonThread() Hw->StopButtonThread(Hw)

/* Common/shared methods
 */
extern UINT8 GenericGetPixel(PPLATFORM pThis, MAP Map, UINT PixNum);
extern void GenericSetPixel(PPLATFORM pThis, MAP Map, UINT PixNum, UINT8 Val);
extern void GenericFillMap(PPLATFORM pThis, MAP Map, UINT8 Val);
extern void NoSmile(PPLATFORM This, MAP Map);
extern void ConsoleCreateButtonThread(PPLATFORM This);
extern void ConsoleStopButtonThread(PPLATFORM This);
extern void LcdCreateButtonThread(PPLATFORM This);
extern void LcdStopButtonThread(PPLATFORM This);
extern BOOL MCT ButtonIsr(void *This, PBOOL pNotMyInterrupt);

/* Constructors
 */
typedef PPLATFORM (*PLATFORM_NEW)(int PlatformType);
extern PPLATFORM CreateConsole(int PlatformType);
extern PPLATFORM CreateSimLcd(int PlatformType);
extern PPLATFORM CreateZ(int PlatformType);

#define nPLATFORMS 5

