/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>
#include <base/cons.h>

#include "sntp.c"

SCODE MCT StartSntpClient(const _TCHAR *ServerName, UINT Period);
SCODE MCT StopSntpClient(void);
SCODE MCT AskSntpClient(void);

int _tmain(int argc, _TCHAR** argv)
{
    SCODE sc;
    const _TCHAR *Url = _T("http");

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("runsntp.c:_tmain: No command line args sc=x%x\n"), sc);
        goto Out;
    }

    switch (argc) {
        case 1:
            sc = StartSntpClient(DEFAULT_TIME_SERVER, SNTP_CLIENT_MIN_PERIOD);
            break;
        case 2:
            if (0 == _tcscmp(_T("stop"), argv[1]))
                sc = StopSntpClient();
            else if (0 == _tcscmp(_T("start"), argv[1]))
                sc = StartSntpClient(DEFAULT_TIME_SERVER, SNTP_CLIENT_MIN_PERIOD);
            else if (0 == _tcscmp(_T("lookup"), argv[1]))
                sc = AskSntpClient();
            else
                sc = StartSntpClient(argv[1], SNTP_CLIENT_MIN_PERIOD);
            break;
        case 3:
            sc = StartSntpClient(argv[1], _ttoi(argv[2]));
            break;
        default:
            _tprintf(_T("runsntp.c:_tmain: Invalid command line, argc=%d\n"), argc);
    }

    if (FAILED(sc)) {
        _tprintf(_T("runsntp.c:_tmain: StartSntpServer failed sc=x%x\n"), sc);
        goto Out;
    }

Out:
    FreeCmdLine(&argc, &argv);
    return sc;
}
    
SCODE MCT StartSntpClient(const _TCHAR *ServerName, UINT Period)
{
    SCODE sc;
    PINAMESPACE     ns = CurrentNameSpace();
    PISNTPCLIENT    SntpClient= NULL;

    sc = BindToObject(ns, _T("COB/sntp.cob"), NAME_SPACE_READ,
                      &IID_ISntpClient, (void **) &SntpClient);
    if (FAILED(sc)) {
        DBG(("runsntp.c:StartSntpClient: could not load sntp.cob sc=0x%x\n", sc));
        goto Out;
    }

    sc = SntpClient->v->Start(SntpClient, ServerName, Period);
    if (FAILED(sc)) {
        DBG(("runsntp.c:StartSntpClient: could not start SNTP client sc=0x%x\n", sc));
        goto Out;
    }

    printf("Started SNTP client\n");

    sc = S_OK;

Out:
    /* The SNTP client thread holds a reference too... */
    if (SntpClient)
        SntpClient->v->Release(SntpClient);

    return sc;
}
    
SCODE MCT StopSntpClient(void) 
{
    SCODE sc;
    PINAMESPACE     ns = CurrentNameSpace();
    PISNTPCLIENT    SntpClient= NULL;

    sc = BindToObject(ns, _T("COB/sntp.cob"), NAME_SPACE_READ,
                      &IID_ISntpClient, (void **) &SntpClient);
    if (FAILED(sc)) {
        DBG(("runsntp.c:StartSntpClient: could not load sntp.cob sc=0x%x\n", sc));
        goto Out;
    }

    sc = SntpClient->v->Stop(SntpClient);
    if (FAILED(sc)) {
        printf("Failed to stop SNTP client\n");
        goto Out;
    }

    printf("SNTP client is stopping\n");

Out:
    if (SntpClient)
        SntpClient->v->Release(SntpClient);

    return sc;
}
    
SCODE MCT AskSntpClient(void)
{
    SCODE sc;
    PINAMESPACE     ns = CurrentNameSpace();
    PISNTPCLIENT    SntpClient= NULL;
    TIME            time;

    sc = BindToObject(ns, _T("COB/sntp.cob"), NAME_SPACE_READ,
                      &IID_ISntpClient, (void **) &SntpClient);
    if (FAILED(sc)) {
        DBG(("runsntp.c:StartSntpClient: could not load sntp.cob sc=0x%x\n", sc));
        goto Out;
    }

    sc = SntpClient->v->CurrentTime(SntpClient, &time);
    if (FAILED(sc)) {
        printf("Failed to look up current time\n");
        goto Out;
    }

    printf("Current time (Windows format) is \n\t  SNTP=0x%llx\n\tsystem=0x%llx\n", 
        time, CurrentTime());

Out:
    if (SntpClient)
        SntpClient->v->Release(SntpClient);

    return sc;
}

