/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <mmhal.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <sampleif.h>
#include <winsock.h>
    
#define SNTP_CLIENT_MIN_PERIOD 60
#define SNTP_CLIENT_MAX_PERIOD 1024

/* Microsoft Stratum 1 NTP/SNTP server, see comments in sntp.c */ 
#define MICROSOFT_TIME_SERVER   _T("172.31.40.19")
#define DEFAULT_TIME_SERVER     MICROSOFT_TIME_SERVER

/* Send/receive buffer handling */
#define SNTP_REQUEST_WORDS 12
#define SNTP_REQUEST_BYTES (4*SNTP_REQUEST_WORDS)
#define SNTP_ORIGINATE_TIMESTAMP_OFFSET 6
#define SNTP_RECEIVE_TIMESTAMP_OFFSET (SNTP_ORIGINATE_TIMESTAMP_OFFSET + 2)
#define SNTP_TRANSMIT_TIMESTAMP_OFFSET (SNTP_RECEIVE_TIMESTAMP_OFFSET + 2)

/* The number of consecutive failures to withstand before stopping client */
#define SNTP_MAX_FAILURES 3

/* Should the SNTP client set the time offset in the SOAP implementation? */
#define SNTP_SOAP_INTEGRATE 1

typedef struct _SntpClient SntpClient;

struct _SntpClient {
    struct ISntpClientVtbl const *v;
    /* Don't cache an address unless also caching a valid socket */
    SOCKET Socket;
    SOCKADDR_IN SockAddr; 
    /* Adjustment to CurrentTime() output */
    TIME RangeAdjustment; /* Moves the system clock into the SNTP range */
    TIME LocalClockOffset; /* This is relative; RangeAdjustment is absolute */
    BOOL SubtractOffset;
    TIME LastSynchronized; /* Destination timestamp of last server reply */
    /* Synchronization thread */
    _TCHAR *ServerName;
    UINT Period;
    BOOL StopThread;
    BOOL ThreadIsRunning;
    /* Housekeeping */
    MUTEX Mx;
    UINT Refs;
};

#define pSntpClient(_x_) ((SntpClient *)_x_)
    
