/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <winsock.h>
struct rtentry;
#include <if.h>
#include <dhcp/dhcp_var.h>

struct ifreq        ifr;
struct ifaliasreq   addreq;
struct if_data      ifdata;
struct DhcpInfo     dhcpdata;

int _tmain(INT argc, _TCHAR *argv[])
{
    SOCKET s = INVALID_SOCKET;
    _TCHAR *Name, *why, *Ip;
    char *ip = NULL, *name = NULL;
    int upndown, DoDhcp, dbglvl, porttype, authmode, wep, adhoc;
    UINT8 DhcpFlags;

    ParseCmdLine(&argc,&argv);

    if (argc < 2 || argc > 3) {
        why = _T(
          "Usage: <interface> [<IP address> | down | up | dhcp | debug<N> | port<N> | auth<N> | wep<N>]"
          );
        goto ErrorExit;
    }

    s = socket(AF_INET, SOCK_DGRAM, 0);

    if (s == INVALID_SOCKET) {
        why = _T("Failed to create socket (Protocol Server not started ?)");
        goto ErrorExit;
    }

    Name = argv[1];

    DoDhcp = 0;
    DhcpFlags = 0;
    upndown = 0;
    dbglvl = -1;
    porttype = -1;
    authmode = -1;
    wep = -1;
    adhoc = 0;

    Ip = NULL;
    if (argc == 3) {
        if (_tcscmp(argv[2], _T("down")) == 0)
            upndown = 1;
        else if (_tcscmp(argv[2], _T("up")) == 0)
            upndown = 2;
        else if (_tcscmp(argv[2], _T("dhcp")) == 0)
            DoDhcp = 1;
        else if (_tcsncmp(argv[2], _T("debug"), 5) == 0) {
            dbglvl = _ttoi(argv[2]+5);
        } else if (_tcsncmp(argv[2], _T("port"), 4) == 0) {
            porttype = _ttoi(argv[2]+4);
        } else if (_tcsncmp(argv[2], _T("auth"), 4) == 0) {
            authmode = _ttoi(argv[2]+4);
        } else if (_tcsncmp(argv[2], _T("wep"), 3) == 0) {
            wep = _ttoi(argv[2]+3);
        } else if (_tcscmp(argv[2], _T("adhoc")) == 0) {
            adhoc = 1;
        } else
            Ip = argv[2];
    }

#if defined(_UNICODE)
    {
    int i;

    name = malloc(_tcsclen(Name) + 1);
    for (i = 0; Name[i]; i++)
        name[i] = (char) Name[i]; /* Adapter names contain only ASCII */
    name[i] = 0;

    if (Ip != NULL) {
        ip = malloc(_tcsclen(Ip) + 1);
        for (i = 0; Ip[i]; i++)
            ip[i] = (char) Ip[i]; /* IP addresses contain only ASCII */
        ip[i] = 0;
    } else
        ip = NULL;
    }
#else
    name = Name;
    ip = Ip;
#endif

/* Macro for accessing sockaddr as sockaddr_in */
#define sin(_sa_) (*(struct sockaddr_in *)&(_sa_))

    if (Ip != NULL) {

        _tcsncpy(addreq.ifra_name, Name, sizeof ifr.ifr_name);

        sin(addreq.ifra_addr).sin_len = 16;
        sin(addreq.ifra_addr).sin_family = AF_INET;
        sin(addreq.ifra_addr).sin_addr.s_addr = inet_addr(ip);

        sin(addreq.ifra_mask).sin_len = 16;
        sin(addreq.ifra_mask).sin_family = AF_INET;
        sin(addreq.ifra_mask).sin_addr.s_addr = ntohl(0xffff0000); /* XXXXXX */

        sin(addreq.ifra_broadaddr).sin_len = 16;
        sin(addreq.ifra_broadaddr).sin_family = AF_INET;
        sin(addreq.ifra_broadaddr).sin_addr.s_addr =
        sin(addreq.ifra_addr).sin_addr.s_addr &
        sin(addreq.ifra_mask).sin_addr.s_addr;

#if 1
        _tprintf(_T("ip = \"%hs\", a = 0x%08x b = 0x%08x m = 0x%08x\r\n"),
               ip,
               sin(addreq.ifra_addr).sin_addr.s_addr,
               sin(addreq.ifra_broadaddr).sin_addr.s_addr,
               sin(addreq.ifra_mask).sin_addr.s_addr);
#endif

        if (ioctlsocket(s, SIOCAIFADDR, (u_long*)&addreq) < 0) {
            why = _T("Ioctl SIOCAIFADDR failed");
            goto ErrorExit;
        }

        _tprintf(_T("Enabled IP address %hs for interface %hs.\r\n"),
                inet_ntoa(sin(addreq.ifra_addr).sin_addr), name);

        goto Exit;
    }

    if (upndown) {
        struct ifreq *rp = &ifr;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));

        if (ioctlsocket(s, SIOCGIFFLAGS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCGIFFLAGS failed");
            goto ErrorExit;
        }

        if (upndown == 1)
            rp->ifr_flags &= ~IFF_UP;
        else
            rp->ifr_flags |= IFF_UP;

        if (ioctlsocket(s, SIOCSIFFLAGS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCSIFFLAGS failed");
            goto ErrorExit;
        }

        _tprintf(_T("Interface %s is now %s\r\n"),
               Name, (upndown == 1) ? _T("down") : _T("up"));

        goto Exit;
    }

    if (adhoc) {
        struct ifreq *rp = &ifr;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));

        if (ioctlsocket(s, SIOCGIFFLAGS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCGIFFLAGS failed");
            goto ErrorExit;
        }

        rp->ifr_flags |= IFF_ADHOC;

        if (ioctlsocket(s, SIOCSIFFLAGS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCSIFFLAGS failed");
            goto ErrorExit;
        }

        _tprintf(_T("Interface %s: set flag IFF_ADHOC\r\n"), Name);
        goto Exit;
    }

    if (DoDhcp) {
        struct ifreq *rp = &ifr;
        UINT i;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        rp->ifr_dhcpflags = DhcpFlags;

        /*
         * Initiate DHCP transaction, and wait for it to finish.
         */
        if (ioctlsocket(s, SIOCIDHCP, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCIDHCP failed");
            goto ErrorExit;
        }

        if (ioctlsocket(s, SIOCGIFADDR, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOGGIFADDR failed");
            goto ErrorExit;
        }

        _tprintf(_T("Interface %hs has IP address %hs.\r\n"),
                name,
                inet_ntoa(((struct sockaddr_in *)&rp->ifr_addr)->sin_addr));

        memset(&dhcpdata,0,sizeof dhcpdata);
        *((struct ifreq *)&dhcpdata) = *rp;
        if (ioctlsocket(s, SIOCGDHCPINFO, (u_long*)&dhcpdata) == 0) {
            struct in_addr a;
            struct sockaddr_in *dns;

            _tprintf(_T("The DHCP information is %svalid, our lease "),
                     (dhcpdata.LeaseTimer) ? _T("") : _T("in"));
            if (dhcpdata.LeaseTimer)
                _tprintf(_T("will expire in %d seconds.\r\n"),
                         dhcpdata.LeaseTimer);
            else
                _tprintf(_T("has expired!\r\n"));

            a.s_addr = dhcpdata.LeasedIpAddr;
            _tprintf(_T("The IP address %hs "),inet_ntoa(a));
            a.s_addr = dhcpdata.DhcpSrvrIpAddr;
            _tprintf(_T("was leased from the DHCP Server %hs\r\n"),
                    inet_ntoa(a));

            a.s_addr = dhcpdata.DefaultRouter;
            _tprintf(_T("Default route through gateway %hs "),
                    inet_ntoa(a));
            a.s_addr = dhcpdata.SubnetMask;
            _tprintf(_T("with subnet mask %hs\r\n"),
                    inet_ntoa(a));

            _tprintf(_T("There are %d DNS server(s)"),dhcpdata.nscount);
            for (i = 0; i < dhcpdata.nscount; i++) {
                _tprintf(_T("\r\n\t"));
                dns = dhcpdata.nsaddr_list + i;
                if (dns->sin_family == AF_INET) {
                    _tprintf(_T("%hs"),inet_ntoa(dns->sin_addr));
                } else {
                    UINT16 *p = (UINT16 *) &dns->sin_port;
                    int j;
                    _tprintf(_T("? AF_%d (%d)"), dns->sin_family,dns->sin_len);
                    for (j = 0; j < (dns->sin_len/2); j++)
                        _tprintf(_T(" %4x"),*p++);
                }
            }
            _tprintf(_T("\r\n"));

            if (dhcpdata.HostName)
                _tprintf(_T("The host name on this interface is `%hs'\r\n"),
                         dhcpdata.HostName);

            if (dhcpdata.DomainName)
                _tprintf(_T("The domain name on this interface is `%hs'\r\n"),
                         dhcpdata.DomainName);

            if (dhcpdata.TimeServer) {
                a.s_addr = dhcpdata.TimeServer;
                _tprintf(_T("The time server is %hs\r\n"),inet_ntoa(a));
            }

        } else {
            why = _T("Ioctl SIOCGDHCPINFO failed");
            goto ErrorExit;
        }

        goto Exit;
    }

    if (dbglvl != -1) {
        struct ifreq *rp = &ifr;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        rp->ifr_debug.opcode = 0;
        rp->ifr_debug.param = (UINT16)dbglvl;

        if (ioctlsocket(s, SIOCSIFDEBUG, (u_long*)rp) == 0) {
            _tprintf(_T("DebugLevel set to %d for interface %s.\r\n"),
                   dbglvl, Name);
        } else {
            why = _T("Ioctl SIOCSIFDEBUG failed");
            goto ErrorExit;
        }
    }

    if (porttype != -1) {
        struct ifreq *rp = &ifr;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        rp->ifr_debug.opcode = 1;
        rp->ifr_debug.param = (UINT16)porttype;

        if (ioctlsocket(s, SIOCSIFDEBUG, (u_long*)rp) == 0) {
            _tprintf(_T("PortType set to %d for interface %s.\r\n"),
                   porttype, Name);
        } else {
            why = _T("Ioctl SIOCSIFDEBUG failed");
            goto ErrorExit;
        }
    }

    if (authmode != -1) {
        struct ifreq *rp = &ifr;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        rp->ifr_debug.opcode = 2;
        rp->ifr_debug.param = (UINT16)authmode;

        if (ioctlsocket(s, SIOCSIFDEBUG, (u_long*)rp) == 0) {
            _tprintf(_T("AuthMode set to %d for interface %s.\r\n"),
                   authmode, Name);
        } else {
            why = _T("Ioctl SIOCSIFDEBUG failed");
            goto ErrorExit;
        }
    }

    if (wep != -1) {
        struct ifreq *rp = &ifr;

        switch (wep) {
        case 1: why = _T("64-bit WEP"); break;
        case 2: why = _T("128-bit WEP"); break;
        default: why = _T("disabled"); break;
        }

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        rp->ifr_debug.opcode = 3;
        rp->ifr_debug.param = (UINT16)wep;

        if (ioctlsocket(s, SIOCSIFDEBUG, (u_long*)rp) == 0) {
            _tprintf(_T("WEP mode set to %s for interface %s.\r\n"),
                   why, Name);
        } else {
            why = _T("Ioctl SIOCSIFDEBUG failed");
            goto ErrorExit;
        }
    }

    if (*name) {
        struct {INT op; const char *name;} siocname[] = {
            {SIOCGIFADDR, "IP"},
            {SIOCGIFBRDADDR, "Broadcast"},
            {SIOCGIFDSTADDR, "Destination"},
            {SIOCGIFNETMASK, "Netmask"},
        };
        struct ifreq *rp = &ifr;
        int flags, metric;
        int i;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));

        for (i = 0; i < sizeof(siocname)/sizeof(siocname[0]); i++) {
            if (ioctlsocket(s, siocname[i].op, (u_long*)rp) == 0) {
                _tprintf(_T("Interface %s has %s address %hs.\r\n"), Name,
                  siocname[i].name,
                  inet_ntoa(((struct sockaddr_in *)
                  &rp->ifr_addr)->sin_addr));
            } else {
                _tprintf(_T("Interface %s has no %s address.\r\n"), Name,
                  siocname[i].name);
            }
        }

        if (ioctlsocket(s, SIOCGDHCPINFOPTR, (u_long*)rp) == 0) {
            UINT8 *HdwrAddr = (*(struct DhcpInfo **)rp)->HdwrAddr;
            _tprintf(_T("Interface %s has ethernet address %02x:%02x:%02x:%02x:%02x:%02x\r\n"), Name,
                   HdwrAddr[0],HdwrAddr[1],HdwrAddr[2],HdwrAddr[3],HdwrAddr[4],HdwrAddr[5]);
        } else {
            _tprintf(_T("Interface %s has no ethernet address.\r\n"), Name);
        }
        
        rp = (struct ifreq *)&ifdata;
        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));
        
        if (ioctlsocket(s, SIOCGIFFLAGS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCGIFFLAGS failed");
            goto ErrorExit;
        }
        flags = rp->ifr_flags;

        if (ioctlsocket(s, SIOCGIFMETRIC, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCGIFMETRIC failed");
            goto ErrorExit;
        }
        metric = rp->ifr_metric;

        rp = (struct ifreq *)&ifdata;
        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));

        if (ioctlsocket(s, SIOCGIFSTATS, (u_long*)rp) < 0) {
            why = _T("Ioctl SIOCGIFSTATS failed");
            goto ErrorExit;
        }
        rp = NULL;

        if (metric != (int)ifdata.ifi_metric) {
            why = _T("SIOCGIFMETRIC and SIOCGIFSTATS disagree on metric?!?");
            goto ErrorExit;
        }

        _tprintf(_T("Interface %s is %s and %srunning. Statistics (decimal, unless):\r\n"),
               Name, (flags & IFF_UP) ? _T("up") : _T("down"),
               (flags & IFF_RUNNING) ? _T("") : _T("not "));
        _tprintf(_T("type 0x%x addrlen %d hdrlen %d mtu %u flags 0x%x\r\n"),
               ifdata.ifi_type, ifdata.ifi_addrlen,
               ifdata.ifi_hdrlen, ifdata.ifi_mtu, flags);
        _tprintf(_T("metric %d baud %u collisions %u qdrop %u\r\n"),
               ifdata.ifi_metric, ifdata.ifi_baudrate,
               ifdata.ifi_collisions, ifdata.ifi_iqdrops);
        _tprintf(_T("in_pkts %u o_pkts %u in_bytes %u o_bytes %u\r\n"),
               ifdata.ifi_ipackets, ifdata.ifi_opackets,
               ifdata.ifi_ibytes, ifdata.ifi_obytes);
        _tprintf(_T("imcast %u omcast %u noproto %u i_err %u o_err %u\r\n"),
               ifdata.ifi_imcasts, ifdata.ifi_omcasts,
               ifdata.ifi_noproto, ifdata.ifi_ierrors, ifdata.ifi_oerrors);
        _tprintf(_T("Errors: miss %u reset %u xtimo %u mem %u others 0x%x 0x%x 0x%x 0x%x\r\n"),
               ifdata.ifi_errors[0], ifdata.ifi_errors[1],
               ifdata.ifi_errors[2], ifdata.ifi_errors[3],
               ifdata.ifi_errors[4], ifdata.ifi_errors[5],
               ifdata.ifi_errors[6], ifdata.ifi_errors[7]);
    }

    goto Exit;

ErrorExit:
    _tprintf(_T("%s, LastErr %x\r\n"), why, WSAGetLastError());
    _tprintf(_T("This test FAILED\r\n"));
Exit:
#if defined(_UNICODE)
    if (name != NULL)
        free(name);
    if (ip != NULL)
        free(ip);
#endif
    if (s != INVALID_SOCKET)
        closesocket(s);
    FreeCmdLine(&argc,&argv);
    return 0;
}
