/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <tchar.h>
#include <mmlite.h>

#include <drivers/drivers.h>
#if 0
#include "config.h"
#else
EXTERN_C void ConfigureDrivers( PIMODULE pModule,
                                _TCHAR *ConfigurationFileName,
                                BOOL LoadDllDrivers );
#endif
/*
 * --------------------------------------------------------------------------
 * Common device driver handling loop
 */
EXTERN_C void THREAD_LINKAGE DeviceThread(THREAD_ARGUMENT arg)
{
    PCDEVICE pThis = (PCDEVICE)arg;

    pThis->v->ThreadLoop(pThis);
}

/*
 * --------------------------------------------------------------------------
 * Builds each one of the statically/dynamically configured drivers, and
 * installs them as appropriate.
 */
EXTERN_C void PerformConfiguration( BOOL LoadDlls )
{
    PIMODULE pCurMod;

    GetCurrentModule(&pCurMod);
    ConfigureDrivers(pCurMod, _T("fs/drivers.dat"), LoadDlls );
    if (pCurMod)
        pCurMod->v->Release(pCurMod);

}

/* We can only do one driver at a time
 */
DRIVER_CONFIG DriverConfigs[] = {
  /* Initialize compatibly with previous use
   */
  {_T("sr0"), _T("COB/sernet.cob"), DEVICE_FLAGS_NET_DRIVER, 0, 0, _T("com2"), NULL, },
  {NULL,}
};

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    int i;
    ADDRESS *pArgs;

    /* Do we have args
     */
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc) || (argc < 2))
        goto Config;

    /* Are they enough
     */
    if (argc < 3) {
        _tprintf(_T("Usage: %s DriverName DriverCob [xFlags xRegisterBase InterruptNo [Args ..]]\n\r"), argv[0]);
        goto Done;
    }

    /* Mandatory ones
     */
    DriverConfigs[0].Name = argv[1];
    DriverConfigs[0].CobName = argv[2];
    DriverConfigs[0].Flags = 0;

    /* Optional ones
     */
    if (argc > 3)
        DriverConfigs[0].Flags = _tcstoul(argv[3],NULL,16);

    if (argc > 4)
        DriverConfigs[0].RegisterBase = _tcstoul(argv[4],NULL,16);

    if (argc > 5)
        DriverConfigs[0].Interrupt = _tcstoul(argv[5],NULL,10);

    /* Arguments to driver ? max=5
     */
    pArgs = (ADDRESS*) &DriverConfigs[0].s1;

    for (i = 0; (i < 5) && (argc > (6+i)); i++) {
        _TCHAR *Arg = argv[6+i], c = Arg[0];

        /* Is it a number or a string.
         * A hex ?
         */
        if (c == _T('0') && (Arg[1] == _T('x') || Arg[1] == _T('X'))) {
            *pArgs++ = _ttoh(Arg+2);
        }

        /* A decimal number ?
         */
        else if (_istdigit(c)) {
            *pArgs++ = _ttoi(Arg);
        }

        /* Must be a string then
         */
        else {
            *pArgs++ = (ADDRESS) Arg;
        }
    }

#if 0
    printf("config: %s %s %x %x %d [%x %x %x %x %x]\r\n",
           DriverConfigs[0].Name,
           DriverConfigs[0].CobName,
           DriverConfigs[0].Flags,
           DriverConfigs[0].RegisterBase,
           DriverConfigs[0].Interrupt,
           DriverConfigs[0].u1,
           DriverConfigs[0].u2,
           DriverConfigs[0].u3,
           DriverConfigs[0].u4,
           DriverConfigs[0].u5);
#endif

 Config:
    /* Now go do the config thing
     */
    PerformConfiguration( USE_DLLS );

 Done:
    /* Let args go (AFTER we done config)
     */
    (void) FreeCmdLine(&argc,&argv);

    return 0;
}
