/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Dont take this too seriously, but yes its a teenieweenie shell
 */
#define LINE_EDIT 0

#include <mmlite.h>

#if defined(_CLR)
#include <clr/clr.h>
#include <clrif.h>
#endif

#if LINE_EDIT
#include <stdio.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#define MAX_LINE 127  /* _TCHARs ! */
#ifndef MAX_PATH
#define MAX_PATH MAX_LINE
#endif
#define SCREEN_LINES 25

/* Lexicon
 */
#define BS '\b'
#define DEL BS
#define CR '\r'
#define LF '\n'
#define TAB '\t'
#define SP ' '
#define BEL '\007'
#define CZ '\032'
#define CD '\004'
#define D 'd'
#define Q 'q'
#define Z 'z'
#define PREV '\020' /* ^P */
#define NEXT '\016' /* ^N */
#define KILL '\025' /* ^U */
#define REM '!'
#define MINUS '-'
#define EOS ((_TCHAR)0)
#define EOT ((_TCHAR)EOF)

/* History
 */
#define MAX_HIST 4
typedef struct _LINE {
    UINT16 len;
    _TCHAR chars[MAX_LINE+1];
} LINE;

#if (LINE_EDIT==0)

extern int read(int,void*,size_t);
extern int write(int,const void *,size_t);
#define EOF (-1)

/*+++++++++++++++++
 * I/O
 */
PRIVATE _TCHAR getchar(void)
{
    _TCHAR c;
    int r = read(0,&c,sizeof(_TCHAR));
    return (r == -1) ? EOT : c;
}

PRIVATE INT putchar(_TCHAR c)
{
    return (write(1,&c,sizeof(_TCHAR)) == -1) ? EOF : c;
}

PRIVATE INT puts(const _TCHAR *s)
{
    size_t n = _tcslen(s);
    int r = write(1,s,n*sizeof(_TCHAR));
    return (r == -1) ? EOF : r;/*silly*/
}

#define fwrite(_b_,_s_,_n_,_f_) write(1,_b_,(_s_)*(_n_))
#endif

PRIVATE void * __cdecl mymalloc(size_t size)
{
    PIHEAP pIHeap = CurrentHeap();

    return pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("tzk"), size, 0);
}

/*+++++++++++++++++
 * Shell context
 */
typedef struct _CONTEXT {
    _TCHAR *curdir; /* NULL if root (e.g. CurentNameSpace()) */
    PINAMESPACE pNs;
    /* later: tree root, bindings, environ */
} CONTEXT;

/*+++++++++++++++++
 * Builtin commands
 */

/* All builtin commands are implemented by an action function 
 * that has the following signature
 */
typedef int (ACTION)(CONTEXT *pC, int argc, _TCHAR **argv);

/* Some commands iterate over lists, invoking action functions of this type
 */
typedef int (COMMAND)(void *pContext, _TCHAR **cmd, int *cmdlen);

typedef struct _COPY_CONTEXT {
    UINT64 CurrentOffset;
    PIFILE Destination;
} COPY_CONTEXT;

typedef enum {
  NoEof, CtlD, CtlZ
} EOF_MODE;

/* Return values.
 */
#define rOK 0
#define rQUIT (-1)

/* Forward decls for actions we implement
 */
PRIVATE ACTION doCd;
PRIVATE ACTION doDir;
PRIVATE ACTION doEcho;
PRIVATE ACTION doExec;
PRIVATE ACTION doStart;
PRIVATE ACTION doQuit;
PRIVATE ACTION doHelp;
PRIVATE ACTION doType;
PRIVATE ACTION doMore;
PRIVATE ACTION doPwd;
PRIVATE ACTION doLink;
PRIVATE ACTION doDel;
PRIVATE ACTION doSource;
PRIVATE ACTION doCopy;
//PRIVATE ACTION doNothing;
PRIVATE ACTION doDebug;
PRIVATE ACTION doSleep;
PRIVATE ACTION doMemory;

PRIVATE COMMAND command;
PRIVATE COMMAND type;
PRIVATE COMMAND more;

/* Strings: error messages, help messages, punctuation
 */
static _TCHAR eMemory[] = _TEXT("Out of memory.\r\n");
static _TCHAR eFound[] = _TEXT(" not found.\r\n");
static _TCHAR eNodir[] = _TEXT(" not a namespace.\r\n");
static _TCHAR eNofile[] = _TEXT(" not a file.\r\n");
static _TCHAR eNocreate[] = _TEXT(" not created.\r\n");
static _TCHAR eLoad[] = _TEXT(" did not load.\r\n");
static _TCHAR ePipe[] = _TEXT("Broken pipe.\r\n");
static _TCHAR eArgs[] = _TEXT("Missing argument(s).\r\n");
static _TCHAR eReg[] = _TEXT("Object registration failed.\r\n");
static _TCHAR eDeReg[] = _TEXT("Object de-registration failed.\r\n");
static _TCHAR eDebug[] = _TEXT("Subcommand missing, one of 'r8' 'r16' 'r32' 'w8' 'w16' 'w32' is required.\r\n");
static _TCHAR eInternal[] = _TEXT("Internal error.\r\n");

static _TCHAR sMemBanner[] = _TEXT("   Total     Free Reserved   Commit     Used  MaxUsed\r\n");

static _TCHAR Prompt[] = _TEXT("> ");
static _TCHAR OVWR[] = _TEXT("\b \b");
static _TCHAR CRLF[] = _TEXT("\r\n");
static _TCHAR COLON[] = _TEXT(": ");
static _TCHAR SPACE[] = _TEXT(" ");
static _TCHAR EMPTY[] = _TEXT("");
static _TCHAR DOT[] = _TEXT(".");
static _TCHAR DOTDOT[] = _TEXT("..");

static _TCHAR hTerminate[] = _TEXT("terminate this program");
static _TCHAR hDel[] = _TEXT("delete <file>") ;
static _TCHAR hEcho[] = _TEXT("echo (eval) <arguments>") ;
static _TCHAR hExec[] = _TEXT("execute a <program> with <arguments>");
static _TCHAR hStart[] = _TEXT("start a background <program> with <arguments>");
static _TCHAR hList[] = _TEXT("list content of a [<namespace>]");
static _TCHAR hLink[] = _TEXT("link object <src> to <dst> (Bind & Register)");
static _TCHAR hType[] = _TEXT("display a file <object>");
static _TCHAR hMore[] = _TEXT("display a file <object>, one page at a time");
static _TCHAR hPwd[] = _TEXT("print working directory");
static _TCHAR hCd[] = _TEXT("change current namespace to <argument>");
static _TCHAR hSource[] = _TEXT("execute one or more <script files>");
static _TCHAR hCopy[] = _TEXT("copy one or more <source file objects> to a <destination file object>");
static _TCHAR hDebug[] = _TEXT("memory accesses for debugging: 'r'/'w' <hex-address> [<hex-value>] ");
static _TCHAR hSleep[] = _TEXT("sleep for <seconds>");
static _TCHAR hMemory[] = _TEXT("system heap state");
static _TCHAR hHelp[] = _TEXT("builtin command list and description");
static _TCHAR noHelp[] = _TEXT("No help for ");
static _TCHAR hElse[] = _TEXT("Anything else we'll try to load it & execute as image.");

static _TCHAR BootScript[] = _TEXT("source fs/init.tzk");

#if defined(_CLR)
PRIVATE ACTION doClix;
static _TCHAR hClix[] = _TEXT("launch a CLI program with <arguments>");
#endif

/* Builtin command table.
 * NB:  Must be alphabetical cuz searching quicker
 */
typedef struct _BUILTIN {
    _TCHAR *name;
    ACTION *action;
    _TCHAR *description;
} BUILTIN;
BUILTIN builtins[] = {
    {_TEXT("."), doSource, hSource },
    {_TEXT("?"),     doHelp,  hHelp },
    {_TEXT("cat"),   doType,  hType },
    {_TEXT("cd"),    doCd,    hCd },
#if defined(_CLR)
    {_TEXT("clix"),  doClix,  hClix },
#endif
    {_TEXT("copy"),  doCopy,  hCopy },
    {_TEXT("cp"),    doCopy,  hCopy },
    {_TEXT("debug"), doDebug, hDebug },
    {_TEXT("del"),   doDel,   hDel },
    {_TEXT("dir"),   doDir,   hList },
    {_TEXT("echo"),  doEcho,  hEcho },
    {_TEXT("exec"),  doExec,  hExec },
    {_TEXT("exit"),  doQuit,  hTerminate },
    {_TEXT("help"),  doHelp,  hHelp },
    {_TEXT("link"),  doLink,  hLink },
    {_TEXT("ln"),    doLink,  hLink },
    {_TEXT("ls"),    doDir,   hList },
    {_TEXT("memory"),doMemory,hMemory },
    {_TEXT("more"),  doMore,  hMore },
    {_TEXT("pwd"),   doPwd,   hPwd },
    {_TEXT("quit"),  doQuit,  hTerminate },
    {_TEXT("rm"),    doDel,   hDel },
    {_TEXT("sleep"), doSleep, hSleep },
    {_TEXT("source"), doSource, hSource },
    {_TEXT("start"), doStart, hStart },
    {_TEXT("type"),  doType,  hType }
};
#define Nbuiltins (sizeof builtins / sizeof builtins[0])

/* Find a builtin command.
 * Quickly find if its not one.
 */
PRIVATE BUILTIN *findcmd(_TCHAR *name)
{
    int i,j;
    _TCHAR ch, ch0, *p;

    /* Use first char as pivot
     */
    ch0 = _totlower(name[0]);

    /* Simple sequential scan, dont have that many
     */
    for (i = 0; i < Nbuiltins; i++) {

        p = builtins[i].name;
        ch = p[0];

        /* List is ordered, can we quit early ?
         */
        if (ch > ch0)
            return NULL;

        /* Not yet up there ?
         */
        if (ch != ch0)
            continue;

        /* Maybe its the one, check the rest of the string
         */
        for (j = 1; ;j++) {
            ch = p[j];

            /* Mismatch ?
             */
            if (ch != _totlower(name[j]))
                break;

            /* Full match ?
             */
            if (ch == EOS)
                return builtins+i;
        }
    }
    return NULL;
}

/*+++++++++++++++++
 * FORNOW HACK
 */
#define NTOA_MAX_DIGITS 16

static void __cdecl _ntoa( void *Where, _TCHAR * String )
{
    _tcsncpy((_TCHAR*) Where,String,NTOA_MAX_DIGITS);
}

PRIVATE int ntoa(_TCHAR *r, _TCHAR *f, ...)
{
    va_list arg;

    va_start(arg,f);

    StrFormat( _ntoa, r, f, arg);
    r[NTOA_MAX_DIGITS-1] = _T('\0');

    va_end(arg);
    return 0;
}

/*+++++++++++++++++
 * Parsing
 */

/* Free up at the end of evaluation.
 * Its a list and we free it in reverse
 */
PRIVATE void arfree(_TCHAR **argv, int argc)
{
    while (argc-- > 0)
        if (argv[argc] != NULL)
            free(argv[argc]);
    free(argv);
}

/* Make a copy of a singleton token.
 */
PRIVATE _TCHAR *tokcpy(_TCHAR *tok, int len)
{
    _TCHAR *ntok = mymalloc((len+1)*sizeof(_TCHAR));
    if (ntok != NULL) {
        memcpy(ntok,tok,len*sizeof(_TCHAR));
        ntok[len] = 0;
    }
    return ntok;
}

/* Cleanup whitespaces from a command line.
 * Returns an array of pointers to the individual strings,
 * and a count of them.  Array is NOT NULL-terminated.
 * Caller responsible for freeing up *ptokens memory (with arfree).
 * Also returns the number of characters left unparsed, in case we
 * encounter a CR/LF/EOS.
 */
int tokenize(_TCHAR *line, int *plen, _TCHAR ***ptokens)
{
    int i, n, ll, nchars;
    _TCHAR ch, *p, *firstp;
    BOOL t = FALSE;

    *ptokens = NULL;
    nchars = *plen;
    if ((nchars == 0) || (line[0] == 0))
        return 0;

    /* Count how many tokens, cleanup
     */
    n = 0;
    firstp = NULL;
    for (i = 0; i < nchars; i++) {
        /* Just any joe char ?
         */
        ch = line[i];
        if ((ch == SP) || (ch == TAB) || (ch == CR) ||
            (ch == LF) || (ch == EOS)) {

            /* Change into standard sep for quicker testing later
             */
            line[i] = SP;

            /* Nope, blank or end of parse
             * Do we terminate a token ?
             */
            if (t) {
                n++;
                t = FALSE;
            }

            /* If just blank continue
             */
            if ((ch == TAB) || (ch == SP) || (ch == CR)) {
                continue;
            }
            /* LF/EOS: Terminate parsing here
             * Return count of unparsed
             */
            *plen = nchars - i - 1;
            nchars = i + 1;
            break;
        }
        /* good char
         */
        if (!t) {
            t = TRUE;
            if (firstp == NULL)
                firstp = line + i;
        }
    }
    /* Tokens end here no matter what
     */
    if (t)
        n++;

    /* So does parsing
     */
    if (i == nchars)
        *plen = 0;

    /* Got anything ?  Drop comments.
     */
    if ((n == 0) || (*firstp == REM))
        return 0;

    /* Allocate array for pointers
     */
    *ptokens = mymalloc(n*sizeof(_TCHAR*));
    if (*ptokens == NULL) {
        puts(eMemory);
        return 0;
    }

    /* Make a second pass, copying tokens
     */
    t = FALSE;
    ll = n = 0;
    p = line;
    for (i = 0; i < nchars; i++) {

        /* Blank ?
         */
        ch = line[i];
        if (ch == SP) {

            /* End of token ?
             */
            if (t) {
 lastok:
                /* Yes, and possibly the last one
                 * Make a copy.
                 */
                p = tokcpy(p,ll);
                if (p == NULL) {
                    /* outtamem, return what we got
                     */
                    return n;
                }
                (*ptokens)[n++] = p;
                t = FALSE;

                /* Last token ?
                 */
                if (i >= nchars)
                    return n;
            }
            /* Done with blank, keep going
             */
            continue;
        }
        /* A good char.  If start of token, remember the start
         * and start counting chars.
         */
        if (!t) {
            p = line+i;
            t = TRUE;
            ll = 1;
        } else {
            /* More chars to this token
             */
            ll++;
        }
    }
    /* If we didnt end clean, go finish up.
     */
    if (t)
        goto lastok;

    /* Clean end
     */
    return n;
}

/*+++++++++++++++++
 * Evaluation
 */

/* Evaluate a list of arguments
 */
PRIVATE int eval(CONTEXT *pC, _TCHAR **argv, int argc)
{
    BUILTIN *a;

    /* Is it a builtin ? If not we exec it
     */
    a = findcmd(argv[0]);
    return (a != NULL) ?
        a->action(pC,argc-1,argv+1) : doExec(pC,argc,argv);
}

/* Parse and evaluate a command line
 */
PRIVATE int command(void * pContext, _TCHAR **cmd, int *pcmdlen)
{
    CONTEXT *pC = (CONTEXT *)pContext;
    int argc, r;
    _TCHAR **argv;

    /* Split it in pieces
     */
    argc = tokenize(*cmd,pcmdlen,&argv);

    /* Make sure something is left
     */
    if (argc == 0)
        return rOK;

    /* Some to do, evaluate it
     */
    r = eval(pC,argv,argc);

    /* Free up resources and go home
     */
    arfree(argv,argc);
    return r;
}

/* Type one buffer of *cmdlen BYTES
 * Specially checks for stdout disappearing, returns rQUIT in that case.
 * BUGBUG Should have a -batch sort of switch to control this
 */
PRIVATE int type(void * pContext, _TCHAR **cmd, int *cmdlen)
{
    size_t n;

    /* One idea could be to pass in STDOUT but .
     */
    UnusedParameter(pContext);

    /* Just type it out
     */
    n = fwrite((BYTE*)(*cmd),
               sizeof(_TCHAR),
               (UINT32)(*cmdlen),
               stdout);

    /* Make sure our pipe is still alive and well
     */
    if (n != (size_t)(*cmdlen)) {
        puts(ePipe);
        return rQUIT;
    }

    /* We ate it all
     */
    *cmdlen = 0;
    return rOK;
}

/* More one buffer of *cmdlen CHARS
 * Specially checks for stdout disappearing, returns rQUIT in that case.
 * BUGBUG Should have a -batch sort of switch to control this
 */
PRIVATE int more(void * pContext, _TCHAR **cmd, int *cmdlen)
{
    UINT32  i, nChars;
    UINT  *nLines = (UINT*)pContext;
    _TCHAR ch, c;
    

    /* We do it one screen at a time
     * Scan for EOL chars
     */
    nChars = (UINT32)*cmdlen;
    for (i = 0; i < nChars; ++i) {

        /* When we get one, count one line.
         * When we get to one screenful, pause.
         */
        ch = cmd[0][i];
        if ( ch == LF ) {
            if (++(*nLines) == SCREEN_LINES ) {
                
                /* wait for LF input
                 * Bonus: 'q' quits early
                 */
                c = getchar();
                if (c == Q) {
                    puts(CRLF);
                    return E_NOT_COMMITTED; /* :-) */
                }
                *nLines = 0;
            }
        }

        /* Put the char out
         * Dont check against EOF in case we are putting one out.
         * We'll fail eventually on some decent char.
         */
        if (putchar(ch) != ch) {
            puts(ePipe);
            return rQUIT;
        }
    }

    /* We ate it all
     */
    *cmdlen = 0;
    return rOK;
}

/* Copy one buffer of *cmdlen BYTES
 * Specially checks for output file disappearing, returns rQUIT in that case.
 * BUGBUG Should have a -batch sort of switch to control this
 */
PRIVATE int copy(void * pContext, _TCHAR **cmd, int *cmdlen)
{
    SCODE r;
    UINT nBytes;
    COPY_CONTEXT *cc;

    /* Recover our argument
     */
    cc = (COPY_CONTEXT*) pContext;

    /* Do the writing
     */
    nBytes = (UINT32)(*cmdlen) * sizeof(_TCHAR);
    r = cc->Destination->v->WriteAt(cc->Destination,
                                    cc->CurrentOffset,
                                    (BYTE*)(*cmd),
                                    nBytes,
                                    &nBytes);
    if (FAILED(r)) {
        puts(ePipe);
        nBytes = 0;
        r = rQUIT;
    }

    /* Adjust our position pointer and say how much we ate
     */
    cc->CurrentOffset = Uint64AddUint32(cc->CurrentOffset,nBytes);
    *cmdlen -= nBytes/sizeof(_TCHAR);

    return r;
}


/*+++++++++++++++++
 * Commands
 */

/* Ancillary for a few: how to handle EOF
 */
PRIVATE EOF_MODE EofOption(int *aargc, _TCHAR ***aargv)
{
    int argc = *aargc;
    _TCHAR **argv = *aargv;

    if ((argc < 1) || (argv[0][0] != MINUS))
        return NoEof;
    *aargc = argc - 1;
    *aargv = argv + 1;
    switch (argv[0][1]) {
    case D: return CtlD;
    case Z: return CtlZ;
    }
    return NoEof;
}


/* Ancillary for a few: get a NS pointer
 */
PRIVATE SCODE BindToNs(PINAMESPACE pNs, _TCHAR *name, PINAMESPACE *pNewNs, BOOL Warn)
{
    SCODE r;
    PIUNKNOWN pUnk;

    r = pNs->v->Bind(pNs,
                     name,
                     0,
                     &pUnk);
    if (FAILED(r)) {
        if (Warn) {
            puts(name);
            puts(eFound);
        }
        return r;
    }

    /* Ok, but is it a namespace
     */
    r = pUnk->v->QueryInterface(pUnk,
                                &IID_INameSpace,
                                (void **) pNewNs);

    (void) pUnk->v->Release(pUnk);

    if (FAILED(r) && Warn) {
        puts(name);
        puts(eNodir);
    }

    return r;
}

/* Ancillary for a few: get a FILE pointer
 */
PRIVATE SCODE BindToFile(PINAMESPACE pNs, _TCHAR *name, PIFILE *pNewFile, NAME_SPACE_FLAGS Mode, BOOL Warn)
{
    SCODE r;
    PIUNKNOWN pUnk;

    r = pNs->v->Bind(pNs,
                     name,
                     Mode,
                     &pUnk);
    if (FAILED(r)) {
        if (Warn) {
            puts(name);
            puts(eFound);
        }
        return r;
    }

    /* Ok, but is it a file
     */
    r = pUnk->v->QueryInterface(pUnk,
                                &IID_IFile,
                                (void **) pNewFile);

    (void) pUnk->v->Release(pUnk);

    if (FAILED(r) && Warn) {
        puts(name);
        puts(eNofile);
    }

    return r;
}

/* DIR
 * List the content of a namespace
 * If no argument given, lists current direc.. ahem namespace
 */
PRIVATE int doDir(CONTEXT *pC, int argc, _TCHAR **argv)
{
    _TCHAR *buffer;
    PINAMESPACE pNs = pC->pNs;
    UINT length;
    SCODE r;

    /* Get a buffer to work with
     */
    buffer = mymalloc(MAX_PATH * sizeof(_TCHAR));
    if (buffer == NULL) {
        puts(eMemory);
        return rOK;
    }

    if (argc > 0) {

        /* List a specific directory
         * BUGBUG Its a list of !
         */
        r = BindToNs(pNs,argv[0],&pNs,TRUE);
        if (FAILED(r)) {
            free(buffer);
            return r;
        }

    }

    /* Get started with the listing
     */
    length = 0;
    r = pNs->v->FindNext(pNs,
                         EMPTY,
                         NULL,
                         buffer,
                         MAX_PATH-sizeof(_TCHAR),
                         &length);

    /* Loop while we find some
     */
    while (!FAILED(r) && (length > 0)) {

        /* Say what, one per line
         */
        buffer[length] = EOS;
        puts(buffer);
        puts(CRLF);

        /* Next ?
         */
        r = pNs->v->FindNext(pNs,
                             EMPTY,
                             buffer,
                             buffer,
                             MAX_PATH-sizeof(_TCHAR),
                             &length);
    }
    

    /* Release NS ref if we took one
     */
    if (argc > 0)
        (void) pNs->v->Release(pNs);

    /* Free resources and go home
     */
    free(buffer);
    return rOK;
}

/* CD
 * Change current namespace to ARG.
 * BUGBUG Doesnt handle ARG==path
 */
PRIVATE int doCd(CONTEXT *pC, int argc, _TCHAR **argv)
{
    UINT s0, s1;
    _TCHAR *cur = pC->curdir, *next, ch;
    PINAMESPACE pNs;
    SCODE r;

    /* Sanity
     */
    if ((argc < 1) ||
        (_tcscmp(DOT,argv[0]) == 0))
        return rOK;

    /* Going up ?
     */
    if (_tcscmp(DOTDOT,argv[0]) == 0) {
        /* Silently ignore if at root
         */
        if (cur == NULL)
            return rOK;

        /* Find last path separator, either \ or /
         */
        s0 = 0; s1 = ~0;
        do {
            ch = cur[s0];
            if (IS_PATH_SEPARATOR(ch))
                s1 = s0;
            s0++;
        } while (ch != EOS);

        /* If none found, we are going to the top
         */
        if (s1 == ~0) {
            /* Null the name
             */
            free(pC->curdir);
            pC->curdir = NULL;

            /* Switch to Current()
             */
            pNs = CurrentNameSpace();
            goto switch_ns;
        }

        /* Chop off string, and rebind
         */
        ch = cur[s1+1];
        cur[s1+1] = EOS;

        r = BindToNs(CurrentNameSpace(),cur,&pNs,TRUE);
        if (FAILED(r)) {
            /* Undo the (little) damage
             */
            cur[s1+1] = ch;
            return r;
        }

        /* Worked.  Release current NS, switch.
         */
        goto switch_ns;
    }

    /* Ignore possible extra stuff, just use first arg.
     * Make sure it exists
     */
    r = BindToNs(pC->pNs,argv[0],&pNs,TRUE);
    if (FAILED(r))
        return r;

    /* Special case for root
     */
    if (cur == NULL) {
        /* Current() carries no refs
         */
        pC->pNs = pNs;

        /* Steal arg string
         */
        pC->curdir = argv[0];
        argv[0] = NULL;

        goto trim_curdir;
    }

    /* Catenate strings
     */
    s1 = _tcslen(argv[0]);
    s0 = _tcslen(cur);
    next = realloc(cur, (s1+s0+2)*sizeof(_TCHAR));

    if (next == NULL) {
        /* Urgh, undo.. even if the system is mighty screwed anyways
         */
        (void) pNs->v->Release(pNs);
        puts(eMemory);
        return 0;
    }

    /* We have resources, do it.
     */
    next[s0++] = PATH_SEPARATOR1; /* pick one, anyone */
    memcpy(next+s0,argv[0],s1*sizeof(_TCHAR));
    next[s0+s1] = EOS;
    pC->curdir = next;

    /* Release old NS, change to new one
     */
  switch_ns:
    (void) pC->pNs->v->Release(pC->pNs);
    pC->pNs = pNs;

  trim_curdir:
    /* Make sure there are no trailing slashes, return
     */
    cur = pC->curdir;
    if (cur) {
        s0 = _tcslen(cur) - 1;
        if (IS_PATH_SEPARATOR(cur[s0]))
            cur[s0] = EOS;
    }
    return rOK;
}

/* Ancillary for the next ancillary:
 * turns a list into a single string
 * BUGBUG: this way we lose structure e.g. sub-strings
 */
PRIVATE _TCHAR *Linearize(int argc, _TCHAR **argv)
{
    int i, l;
    _TCHAR *line;

    /* See how it will be
     */
    l = 0;
    for (i = 0; i < argc; i++)
        l += sizeof(_TCHAR) * (_tcslen(argv[i]) + 1);

    /* Allocate it, caller will free
     */
    line = mymalloc(l);
    if (line == NULL) {
        puts(eMemory);
        return NULL;
    }

    /* Initialize it, then append all strings, SPACE separated
     */
    line[0] = EOS;
    for (i = 0; i < (argc-1); i++) {
        /* BUGBUG very inefficient
         */
        line = _tcscat(line,argv[i]);
        line = _tcscat(line,SPACE);
    }

    /* Last one, zero-terminate.
     */
    line = _tcscat(line,argv[i]);

    /* Done
     */
    return line;
}

PRIVATE _TCHAR *GetSuffix(_TCHAR *arg)
{
    _TCHAR *p;
    /* find end */
    for (p = arg; *p; p++)
        ;
    /* find dot going backwards */
    for ( ; *p != _T('.'); p--)
        if (p == arg)
            return NULL;
    return p+1; /* return next after dot */
}

/* Unpleasantness from having just one arg available */
struct RunMainArg {
    PIPROGRAM prog;
    _TCHAR *cmdline;
};

PRIVATE SCODE RunMain(struct RunMainArg *a)
{
    SCODE sc = a->prog->v->Main(a->prog, a->cmdline);

    /* Cleanup */
    a->prog->v->Release(a->prog);
    free(a->cmdline);
    free(a);
    return sc;
}

/* Ancillary for the next two guys:
 * common code for creating a new process/module.
 */
PRIVATE int LoadImage(CONTEXT *pC, int argc, _TCHAR **argv, BOOL Wait)
{
    PIPROCESS pPrc = CurrentProcess();
    _TCHAR *Arguments = NULL, *Suffix;
    _TCHAR Number[NTOA_MAX_DIGITS];
    SCODE r;

    /* Check we still got game
     */
    if (argc <= 0)
        return rOK;

    /* Linearize arguments (sigh)
     */
    if (argc > 1)
        Arguments = Linearize(argc-1,argv+1);

    Suffix = GetSuffix(argv[0]);
 
    if (Suffix && !_tcscmp(Suffix, _T("cob"))) {
        /* Load it COB style
         */
        PINAMESPACE ns;
        PIPROGRAM prog;
        if (!Arguments)
            Arguments = _tcsdup(_T("")); /* not optional */

        /* Bind to "COB/" */
        r = BindToObject(CurrentNameSpace(), _T("COB"), NAME_SPACE_READ,
                         &IID_INameSpace, (void **) &ns);
        if (FAILED(r))
            goto Out;

        /* Get the Cob's Main method */
        r = BindToObject(ns, argv[0], NAME_SPACE_READ, &IID_IProgram,
                         (void **) &prog);
        ns->v->Release(ns);
        if (FAILED(r))
            goto Out;

        if (Wait) {
            r = prog->v->Main(prog, Arguments ? Arguments : _T(""));
            prog->v->Release(prog);
        } else {
            struct RunMainArg *a = malloc(sizeof(*a));
            if (!a) {
                r = E_NOT_ENOUGH_MEMORY;
                prog->v->Release(prog);
                goto Out;
            }
            a->prog = prog;
            a->cmdline = Arguments;
            r = pPrc->v->CreateThread(pPrc, 0,
                                      (THREAD_FUNCTION) RunMain, a,
                                      0, NULL, NULL);
            if (FAILED(r))
                prog->v->Release(prog);
            else
                Arguments = NULL;
        }

    } else {
        /* Load it EXE style
         */
        PIMODULE pMod;
        r = pPrc->v->LoadImage(pPrc,
                               argv[0],
                               Arguments,
                               LOAD_IMAGE_CREATE_THREAD|LOAD_IMAGE_CALL_ENTRY_POINT,
                               &pMod);
        /* If it didnt work say so */
        if (FAILED(r))
            goto Out;

        /* Get the new process's IProcess */
        r = pMod->v->QueryInterface(pMod, &IID_IProcess, (void**) &pPrc);
        pMod->v->Release(pMod);
        if (FAILED(r))
            goto Out;

        /* Do we have to wait for it ?
         */
        if (Wait)
            r = pPrc->v->WaitFor(pPrc, TIME_FOREVER);

        /* Dont leak refs.. */
        pPrc->v->Release(pPrc);
    }

    /* Done, free resources and go home
     */
 Out:
    if (FAILED(r)) {
        puts(argv[0]);
        puts(eLoad);
        ntoa(Number,_T("%8x"),r);
        puts(Number);
    }

    if (Arguments)
        free(Arguments);
    return r;
}

/* START
 * Creates a parallel process to execute a command,
 * doesnt wait for it and returns immediately.
 */
PRIVATE int doStart(CONTEXT *pC, int argc, _TCHAR **argv)
{
    return LoadImage(pC, argc,argv,FALSE);
}

/* EXEC
 * Creates a new process to execute a command,
 * waits for it to complete execution.
 */
PRIVATE int doExec(CONTEXT *pC, int argc, _TCHAR **argv)
{
    return LoadImage(pC,argc,argv,TRUE);
}

#if defined(_CLR)
/* CLIX
 * Creates a parallel CLR process to execute a CLI program
 */
PRIVATE int doClix(CONTEXT *pC, int argc, _TCHAR **argv)
{
    PIPROCESS pPrc = CurrentProcess();
    PIMODULE pMod = NULL;
    _TCHAR *Arguments = NULL;
    SCODE r;
    PICLR pTheCLR;
    PINAMESPACE pTheNameSpace;
    BUILTIN *a;

    /* Check we still got game
     */
    if (argc <= 0)
        return rOK;

    pTheNameSpace = CurrentNameSpace();

    if (pTheNameSpace->v->Bind(pTheNameSpace, CLR_NAME,
			       0, (PIUNKNOWN*)&pTheCLR) != S_OK) {
      a = findcmd("clix");
      puts(a->description);
      goto Out;
    }
     /* Linearize arguments (sigh)
     */
//    if (argc > 1) {
//        Arguments = Linearize(argc-1,argv+1);  
//    } 
    /* Load it
     */
//    r = pTheCLR->v->LoadImage(pTheCLR, argv[0], Arguments,
//			      LOAD_IMAGE_CREATE_THREAD| LOAD_IMAGE_CALL_ENTRY_POINT,
//			      &pMod);
    
    /* 
     * If there is no argument, call LoadImage. 
     * If there is any arguments, call InvokeMethod.
     * by jhkang
     */
    /* Linearize arguments (sigh)
     */
    if (argc > 1) {
        if (_tcscmp("-r", argv[0]) == 0) { //registration
            if (argc > 3) Arguments = Linearize(argc-3,argv+3);
            r = pTheCLR->v->Register(pTheCLR, argv[1], argv[2], Arguments); 
        } else if (_tcscmp("-e", argv[0]) == 0) { //invoke method
            if (argc > 4) Arguments = Linearize(argc-4,argv+4);
            r = pTheCLR->v->InvokeMethod(pTheCLR, argv[1], argv[2], argv[3], Arguments); 
        }
    } else {
        /* Load it
         */
        r = pTheCLR->v->LoadImage(pTheCLR, argv[0], Arguments,
			      LOAD_IMAGE_CREATE_THREAD | LOAD_IMAGE_CALL_ENTRY_POINT,
			      &pMod);
    }
    /* If it didnt work say so
     */
    if (FAILED(r)) {
        puts(argv[0]);
        puts(eLoad);
        goto Out;
    }

    /* We never wait for a CLR process.
     */
/*     if (Wait) */
/*         r = pMod->v->WaitFor(pMod); */

    /* Dont leak refs..
     */
    if (pMod)
        pMod->v->Release(pMod);

    /* Done, free resources and go home
     */
 Out:
    if (Arguments)
        free(Arguments);
    return r;
}
#endif /* _CLR */

/* ECHO
 * Evaluates arguments and echoes them
 */
PRIVATE int doEcho(CONTEXT *pC, int argc, _TCHAR **argv)
{
    /* Echo arguments one by one
     */
    while (argc-- > 0) {
        puts(*(argv++));
        if (argc) 
            puts(SPACE);
    }
    puts(CRLF);
    return rOK;
}

/* QUIT
 * Terminates the program.
 * Returns its special return value.
 */
PRIVATE int doQuit(CONTEXT *pC, int argc, _TCHAR **argv)
{
    return rQUIT;
}

/* HELP
 * Print some hopefully useful info about a command
 * Called without arguments prints them all.
 */
PRIVATE int doHelp(CONTEXT *pC, int argc, _TCHAR **argv)
{
    int i;
    BUILTIN *a;

    /* Something in particular ?
     */
    if (argc > 0) {

        /* Yes, find what and print the description
         */
        a = findcmd(argv[0]);
        if (a) puts(a->description);

        else {
            /* Dont know much about that
             */
            puts(noHelp);
            puts(argv[0]);
        }
        /* Terminate output
         */
        goto crlf;
    }

    /* Tell all
     */
    for (i = 0; i < Nbuiltins; i++) {
        a = builtins+i;
        puts(a->name);
        puts(COLON);
        puts(a->description);
        puts(CRLF);
    }

    /* Say what we do ifnot
     */
    puts(hElse);
 crlf:
    puts(CRLF);
    return rOK;
}

/* Ancillary for the next few: type/more/source/.. a file
 * Open a file, read it in and hand each line to the ACTION function.
 */
PRIVATE int WalkOneFile(CONTEXT *pC, _TCHAR *filename, EOF_MODE Eof,
                        COMMAND *action, void * sub_context)
{
    _TCHAR *buffer;
    PINAMESPACE pNs = pC->pNs;
    PIFILE pF;
    SCODE r, ret = rOK;
    UINT64 Position;
    UINT nBytesRead = 0, nChars, nCharsLeft;
    _TCHAR EofChar;
    UINT i;


    /* See if we can find it
     */
    r = BindToFile(pNs,filename,&pF, NAME_SPACE_READ, TRUE);
    if (FAILED(r)) {
        return r;
    }

    /* Poifect, proceed to work on it
     * Get a buffer to work with
     */
    buffer = mymalloc(MAX_PATH);
    if (buffer == NULL) {
        puts(eMemory);
        return rOK;
    }

    /* Reuse Eof as an internal flag that we did see an EOF char
     */
    switch (Eof) {
    case CtlD: EofChar = CD; break;
    case CtlZ: EofChar = CZ; break;
    default:
        EofChar = 0;
    }
    Eof = NoEof;

    /* Start from the beginning, and loop while we find some
     */
    Uint32ToUint64(Position, 0);
    nChars = 0;
    for (; Eof == NoEof;) {

        /* Get a bufferfull
         */
        r = pF->v->ReadAt(pF,
                          Position,
                          (BYTE*)(buffer + nChars),
                          MAX_PATH - (nChars * sizeof(_TCHAR)),
                          &nBytesRead);

        /* EOF ?
         */
        if ( FAILED(r) || ((nBytesRead + nChars) == 0) )
            break;

        /* More on EOF 
         */
        if (EofChar != 0) {
            nCharsLeft = nBytesRead / sizeof(_TCHAR);
            for (i = nChars; i < (nChars + nCharsLeft); i++)
                if (buffer[i] == EofChar) {
                    /* Found an EOF, stop here
                     */
                    Eof = CtlD;
                    break;
                }
            nChars = i;
        } else
            nChars += nBytesRead / sizeof(_TCHAR);

        /* Pass the exact number of chars to parse
         */
        nCharsLeft = nChars;
        ret = action(sub_context,&buffer,(int*)&nCharsLeft);
        if (FAILED(ret))
            break;

        /* Did the action consume all characters ?
         * If not, move the leftovers upfront before we refill.
         */
        if (nCharsLeft) {
            memcpy(buffer,
                   buffer+(nChars-nCharsLeft),
                   nCharsLeft*sizeof(_TCHAR));
//            buffer[nCharsLeft] = EOS;
        }
        nChars = nCharsLeft;

        /* Next buffer's file position
         */
        Position = Uint64AddUint32(Position, nBytesRead);
    }

    /* Release file
     */
    (void) pF->v->Release(pF);

    /* Free resources and go home
     */
    free(buffer);
    return ret;
}

/* Ancillary: loop over a file list
 * BUGBUG: this should be part of eval
 */
PRIVATE int WalkFiles(CONTEXT *pC, int argc, _TCHAR **argv, EOF_MODE Eof,
                      COMMAND *action, void * sub_context)
{
    int r, i;

    r = rOK;
    for (i = 0; (i < argc) && (r == rOK); i++) {
        r = WalkOneFile(pC, argv[i], Eof, action, sub_context);
    }
    return r;
}

/* TYPE
 * Shows the content of an IFile.
 */
PRIVATE int doType(CONTEXT *pC, int argc, _TCHAR **argv)
{
    EOF_MODE Eof = EofOption(&argc, &argv);

    /* We need an argument
     * BUGBUG What about using stdin if not provided ?
     */
    if (argc <= 0)
        return rOK;

    return WalkFiles(pC, argc, argv, Eof, type, NULL);
}

/* TYPE
 * Shows the content of an IFile.
 */
PRIVATE int doMore(CONTEXT *pC, int argc, _TCHAR **argv)
{
    UINT nLines = 0;
    EOF_MODE Eof = EofOption(&argc, &argv);

    /* We need an argument
     * BUGBUG What about using stdin if not provided ?
     */
    if (argc <= 0)
        return rOK;

    return WalkFiles(pC, argc, argv, Eof, more, &nLines);
}

/* PWD
 * Print working directory.
 * Silly for now, but if/when we can mess with the prompt..
 */
PRIVATE int doPwd(CONTEXT *pC, int argc, _TCHAR **argv)
{
    puts(STR_PATH_SEPARATOR);
    if (pC->curdir != NULL)
        puts(pC->curdir);
    puts(CRLF);
    return rOK;
}

/* LINK
 * ln [-r|w] <src> <dst>
 */
PRIVATE int doLink(CONTEXT *pC, int argc, _TCHAR **argv)
{
    SCODE sc = S_OK;
    PINAMESPACE ns = pC->pNs;
    PIUNKNOWN pUnk;
    NAME_SPACE_FLAGS flags;

    if (argc != 2 && argc != 3) {
        puts(eArgs);
        sc = E_INVALID_PARAMETER;
        goto out;
    }

    if (argc == 2) {
        /* default */
        flags = NAME_SPACE_READ | NAME_SPACE_WRITE;
    } else if (argv[0][0] == '-') {
        switch(argv[0][1]) {
        case 'r':
            flags = NAME_SPACE_READ;
            argv++;
            break;
        case 'w':
            flags = NAME_SPACE_WRITE;
            argv++;
            break;
        case 'c':
            flags = NAME_SPACE_READ | NAME_SPACE_WRITE | NAME_SPACE_CREATE;
            argv++;
            break;
        default:
            sc = E_INVALID_PARAMETER; 
            goto out;
        }
    } else {
        sc = E_INVALID_PARAMETER; 
        goto out;
    }

    sc = ns->v->Bind(ns, argv[0], flags, &pUnk);

    if (FAILED(sc)) {
        puts(argv[0]);
        puts(eFound);
        goto out;
    }

    sc = ns->v->Register(ns, argv[1], pUnk, 0, NULL);
    pUnk->v->Release(pUnk);
    if (FAILED(sc))
        puts(eReg);
out:
    return sc;
}

/* DEL
 * rm <file>
 */
PRIVATE int doDel(CONTEXT *pC, int argc, _TCHAR **argv)
{
    SCODE sc = S_OK;
    PINAMESPACE ns = pC->pNs;

    if (argc != 1) {
        puts(eArgs);
        return E_INVALID_PARAMETER;
    }

    sc = ns->v->Unregister(ns, argv[0]);
    if (FAILED(sc)) {
        puts(eDeReg);
        return sc;
    }

    return sc;
}

/* SOURCE
 * source <file>
 * Execute a script from a given file.
 * BUGBUG Make it into something that takes an IFile as input,
 * BUGBUG so that we can do web-scripting and typeups.
 */
PRIVATE int doSource(CONTEXT *pC, int argc, _TCHAR **argv)
{
    EOF_MODE Eof = EofOption(&argc, &argv);

    if (argc == 0) {
        puts(eArgs);
        return E_INVALID_PARAMETER;
    }

    return WalkFiles(pC, argc, argv, Eof, command, pC);
}

/* COPY
 * copy <source> <destination>
 */
PRIVATE int doCopy(CONTEXT *pC, int argc, _TCHAR **argv)
{
    SCODE r;
    _TCHAR *DstName;
    COPY_CONTEXT cc;
    EOF_MODE Eof = EofOption(&argc, &argv);

    /* Check arguments
     */
    if (argc < 2) {
        puts(eArgs);
        return E_INVALID_PARAMETER;
    }

    /* Destination is last
     */
    argc--;
    DstName = argv[argc];
    r = BindToFile(pC->pNs,DstName,&cc.Destination,NAME_SPACE_WRITE, FALSE);
    if (FAILED(r)) {
        /* We might have to create it
         */
        r = BindToFile(pC->pNs,DstName,&cc.Destination,
                       NAME_SPACE_WRITE|NAME_SPACE_CREATE, TRUE);
        if (FAILED(r)) {
            puts(DstName);
            puts(eNocreate);
            return r;
        }
    }

    /* Drop it from the list
     */
    argv[argc] = NULL; /* sanity */

    /* Do the copying
     */
    Uint32ToUint64(cc.CurrentOffset,0);
    r = WalkFiles(pC, argc, argv, Eof, copy, &cc);
    
    /* Put the name in and close the file object
     */
    argv[argc] = DstName;
    (void) cc.Destination->v->Release(cc.Destination);

    return r;
}

/* DEBUG
 * debug <r8,r16,r32,w8,w16,w32> <hex-address> [<hex-value>]
 */
PRIVATE int doDebug(CONTEXT *pC, int argc, _TCHAR **argv)
{
    ADDRESS Where;
    UINT32 Value;
    _TCHAR Number[NTOA_MAX_DIGITS];

    /* Check arguments
     */
    if (argc < 2) {
        puts(eArgs);
        return E_INVALID_PARAMETER;
    }
    
    Where = _ttoh(argv[1]);

    /* Read/write
     */
    switch (argv[0][0]) {

    case 'r':
    case 'R':
        if (argv[0][1] == '8') {
            Value = *(UINT8*) Where;
        } else
        if (argv[0][1] == '1') {
            Value = *(UINT16*) Where;
        } else {
            Value = *(UINT32*) Where;
        }

        (void) ntoa(Number,_T("%x"),Value);
        puts(Number);
        puts(CRLF);
        break;

    case 'w':
    case 'W':
        if (argc < 3) {
            puts(eArgs);
            return E_INVALID_PARAMETER;
        }

        if (argv[0][1] == '8') {
            Value = (UINT8) _ttoh(argv[2]);
            *((UINT8*)Where) = (UINT8) Value;
        } else
        if (argv[0][1] == '1') {
            Value = (UINT16) _ttoh(argv[2]);
            *((UINT16*)Where) = (UINT8) Value;
        } else {
            Value = _ttoh(argv[2]);
            *((UINT32*)Where) = (UINT32) Value;
        }
        break;

    default:
        puts(eDebug);
        return E_INVALID_PARAMETER;
    }

    return rOK;

}

/* SLEEP
 * sleep <seconds>
 */
PRIVATE int doSleep(CONTEXT *pC, int argc, _TCHAR **argv)
{
    UINT nSeconds;
    TIME HowLong;

    /* Check arguments
     */
    if (argc < 1) {
        puts(eArgs);
        return E_INVALID_PARAMETER;
    }
    
    nSeconds = _ttoi(argv[0]);

    Int32ToInt64(HowLong, TIME_RELATIVE(TIME_SECONDS(1)));
    SleepUntil(Int64TimesInt32(HowLong, nSeconds));

    return rOK;
}

/* MEMORY
 * memory
 */
PRIVATE int doMemory(CONTEXT *pC, int argc, _TCHAR **argv)
{
    struct MACHINE_INFO mi;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();
    PIHEAP pHeap = CurrentHeap();
    int i;
#define _nValues 6
    ADDR_SIZE Values[_nValues];//Reserve,Commit,Used,MaxUsed;
    _TCHAR Number[NTOA_MAX_DIGITS];

    memset(Values,0,sizeof Values);

    /* See how much memory we have
     */
    sc = pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi );
    if (FAILED(sc)) {
        puts(eInternal);
        return sc;
    }

    Values[0] = mi.TotalPhysicalMemory;
    Values[1] = mi.AvailablePhysicalMemory;

    /* The heap now
     */
    sc = pHeap->v->Status(pHeap,
                          Values+2, /*Reserve*/
                          Values+3, /*Commit*/
                          Values+4, /*Used*/
                          Values+5);/*MaxUsed*/
    if (FAILED(sc)) {
        puts(eInternal);
        return sc;
    }

    /* Say it, without printf.
     */
    puts(sMemBanner);
    for (i = 0; i < _nValues; i++) {
        ntoa(Number,_T("%8u"),Values[i]);
        puts(Number);puts(SPACE);
    }
    puts(CRLF);

    /* If we are using a debug heap, make extra arg dump all allocations */
    if (argc > 0)
        pHeap->v->Status(pHeap, NULL, NULL, NULL, NULL);

    return rOK;
#undef _nValues
}

#if 0
/* NOTHING
 * Null command.
 */
PRIVATE int doNothing(CONTEXT *pC, int argc, _TCHAR **argv)
{
    return rOK;
}
#endif

/*+++++++++++++++++
 * Main loop
 */
PRIVATE void prompt(CONTEXT *pC)
{
    if (pC->curdir != NULL)
        puts(pC->curdir);
    puts(Prompt);
}

PRIVATE SCODE TzkMain(PIPROGRAM iThis, const _TCHAR *CommandLine)
{
    _TCHAR ch;
    BOOL rd = TRUE;
    int r = rOK;
    _TCHAR *cmd;
    int cmdlen;
    LINE *history = NULL;
    int histix;
    CONTEXT root;
    BOOL FreeCmd = FALSE;

    /* Setup root context
     */
    root.curdir = NULL;
    root.pNs = CurrentNameSpace();

    /* See if we got any command line stuff
     * If not, see if there is a boot script we should execute
     * BUGBUG Questionable weather this should be done before/after/instead
     * of the command line stuff.  For now do it instead, and a command
     * line of " " will therefore disable it.  Yes we should have flags.
     */
    if ((CommandLine == NULL) || (CommandLine[0] == EOS)) {
        cmd = BootScript;
    } else {
        /* The tokenizer writes into the string */
        cmd = _tcsdup(CommandLine);
        if (!CommandLine) {
            puts(_T("tzk: strdup failed\r\n"));
            return E_NOT_ENOUGH_MEMORY;
        }
        FreeCmd = TRUE;
    }

    cmdlen = _tcslen(cmd);
    r = command(&root,&cmd,&cmdlen);
    if ((r == rQUIT) || (cmd != BootScript))
        goto Out;

    /* Allocate&zero our buffer.
     * Dont put stuff on the stack, minimize its use
     */
    history = calloc(MAX_HIST,sizeof(LINE));
    if (history == NULL) {
        r = rQUIT;
        goto Out;
    }
    histix = 0;
    cmd = history[histix].chars;
    cmdlen = 0;

    /* Forever loop until we either get an EOF, or the user
     * types an end-of-file characters (^Z for Windows ^D Unix)
     */

    prompt(&root);
    do {
        ch = getchar();

        switch (ch) {

        case DEL:
            /* Edit (ahem..)
             * BUGBUG some do '\r'<cmd>
             */
            if (cmdlen <= 0)
                putchar(BEL);
#if (LINE_EDIT==0)
            else {
                cmdlen--;
                puts(OVWR);
            }
#endif
            break;

        case KILL:
            for (; cmdlen; cmdlen--)
                puts(OVWR);
            break;

        case NEXT:
        case PREV:
            /* save current, since we work inplace
             */
            if (cmdlen)
                history[histix].len = cmdlen;
            if (ch == PREV) {
                if (--histix < 0) histix = MAX_HIST-1;
            } else {
                if (++histix == MAX_HIST) histix = 0;
            }
            /* Wipe out current line, switch, retype
             */
            //printf("[%d %d %d]\n",histix,cmdlen,history[histix].len);
            while (cmdlen-- > 0)
                puts(OVWR);
            cmd = history[histix].chars;
            cmdlen = history[histix].len;
            cmd[cmdlen] = 0;
            puts(cmd);
            break;

        case EOT:
        case CZ:
        case CD:
            /* End. Still must exec what was typed so far
             */
            rd = FALSE;
            /* fallthrough */

        case CR:
        case LF:
            /* EOL. Just do it.
             */
#if (LINE_EDIT==0)
            puts(CRLF);
#endif
            cmd[cmdlen] = 0;
            history[histix].len = cmdlen;
            while ((r = command(&root,&cmd,&cmdlen)) > 0)
                ;
            if (r == rQUIT) {
                rd = FALSE;
                /* exit status */
                r = (r == rQUIT) ? rOK : -r;
            }
            /* Still going ?
             */
            if (rd) {
                /* Prompt & history
                 */
                prompt(&root);
                if (++histix == MAX_HIST) histix = 0;
                cmd = history[histix].chars;
                cmdlen = 0;
            }
            break;

        case TAB: /*later: auto-completion */
#if (LINE_EDIT==0)
            putchar(ch);
#endif
            ch = SP;
            /* fallthrough*/

        default:
            /* Regular char. Make sure not too many.
             */
            if (cmdlen < MAX_LINE)
                cmd[cmdlen++] = ch;
#if (LINE_EDIT==0)
            else
                /* Try to warn
                 */
                ch = BEL;
            putchar(ch);
#endif
            break;
        }
    } while (rd);

 Out:
    /* Done, free up memory and go home.
     */
    if (history)
        free(history);
    if (FreeCmd)
        free(cmd);
    return r;
}

#include <base/cobdesc.h>
typedef struct _DefCob Tzk;
#define TzkDestroy NULL
#include <s_tzk_v.c>
