/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* TODO
 * - suballocs
 */

/*#define TRACE_USER_LEAKS(x) x */
#define TRACE_USER_LEAKS(x)

#ifndef IN_WINDOWS
#define IN_WINDOWS 0
#endif

#if IN_WINDOWS

#define _DEBUG 1

/* Precise types
 */
typedef void * PTR;
typedef int INT32; /* this better be a 32bit box */
typedef unsigned int UINT32;
typedef unsigned char UINT8;

typedef UINT32 *PUINT32;

/* Generic types, assume again a 32bit box
 */
typedef UINT8 BYTE;
typedef INT32 INT;
typedef UINT32 UINT; 
typedef UINT32 ADDRESS;
typedef UINT32 ADDR_SIZE;
typedef UINT32 BOOL;
#define TRUE 1
#define FALSE 0
typedef INT32 SCODE;
#define NO_ERROR 0

typedef BYTE *PBYTE;

/* Other MMLite structs&defs
 */
#define _DCACHELINESIZE 64

typedef struct {
    UINT s[1];
} MUTEX;

/* Linkage bull and random stuff
 */
#define MCT
#define UnusedParameter(x) x = x
#define NULL ((void *)0)
#define RESTRICT

extern void * memset(void *, int, ADDR_SIZE);
#if _DEBUG
extern int printf(const char *, ...);
#define __assert_print(_m_, _f_, _l_) printf("Assertion `%s' failed at %s:%u\n", _m_, _f_, _l_);
#define assert(_x_) { if (!(_x_)) {__assert_print(#_x_, __FILE__, __LINE__)} }
#else
#define assert(x)
#endif

/* nullifications */
#define BaseSetLastError(x)
#define Mutex_Init(x)
#define Mutex_Lock(x)
#define Mutex_Unlock(x)
#define AtomicCmpAndSwap(_p_,_o_,_n_) \
    ((*(_p_) == (_o_)) ? ((*(_p_) = (_n_)),TRUE) : (FALSE))


/*
 * Ignoring the IUnknown, this is the IHeap interface definition
 */

#define HEAP_NO_SERIALIZE   ( 0x1 )
#define HEAP_GROWABLE   ( 0x2 )
#define HEAP_NO_COPY (0x4)
#define HEAP_ZERO_MEMORY    ( 0x8 )
#define HEAP_REALLOC_IN_PLACE_ONLY  ( 0x10 )
#define HEAP_CACHE_ALIGN ( 0x100 )

typedef struct IHeap IHeap;
typedef IHeap *PIHEAP;

typedef struct IHeapVtbl {

    PTR ( MCT *Alloc )(
            IHeap *This,
            UINT Flags,
            UINT Size,
            UINT Alignment);

    PTR ( MCT *ReAlloc )(
            IHeap *This,
            UINT Flags,
            PTR pMem,
            UINT NewSize,
            UINT Alignment);

    SCODE ( MCT *Free )(
            IHeap *This,
            UINT Flags,
            PTR pMem);

    ADDR_SIZE ( MCT *Size )(
            IHeap *This,
            UINT Flags,
            PTR pMem);

    SCODE ( MCT *Validate )(
            IHeap *This,
            UINT Flags,
            PTR pMem);

    PTR ( MCT *Extract )(
            IHeap *This,
            UINT Flags,
            PTR pMem,
            UINT Size);

    SCODE ( MCT *Status )(
            IHeap *This,
            ADDR_SIZE *pReserve,
            ADDR_SIZE *pCommit,
            ADDR_SIZE *pUsed,
            ADDR_SIZE *pMaxUsed);

    } IHeapVtbl;

struct IHeap {
    const struct IHeapVtbl *v;
};

#define CHECK_DELAYED_FREE()

#else /* !IN_WINDOWS */

#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <stdio.h>
#include <assert.h>

extern PTR *ppDelayedFree;
#define CHECK_DELAYED_FREE() {if (*ppDelayedFree) ppDelayedFree = CheckDelayedFree();}

#endif

#include "heap.h"
#include <s_BitMapHeap_v.c>
#include <s_BitMapHeapFactory_v.c>

/************************************************************************
 * So here is our code
 */

#define BITS_PER_BYTE 8
#define BITS_PER_WORD   (sizeof(UINT)*BITS_PER_BYTE)

#define ST_SUBALLOC  0
#define ST_FREE      1
#define ST_END       2
#define ST_CONTINUE  3

#define ST_MASK      0x3
#define ST_SIZE      2
#define ST_SHIFT     1

#if _UINTSIZE==16
#define IsAllBusy(x) (((x) & 0xAAAA) == 0xAAAA)
#define AllFree      0x5555
#define AllContinue  0xffff
#define AllEnd       0xAAAA
#elif _UINTSIZE==32
#define IsAllBusy(x) (((x) & 0xAAAAAAAA) == 0xAAAAAAAA)
#define AllFree      0x55555555
#define AllContinue  0xffffffff
#define AllEnd       0xAAAAAAAA
#elif _UINTSIZE==64
#define IsAllBusy(x) (((x) & 0xAAAAAAAAAAAAAAAA) == 0xAAAAAAAAAAAAAAAA)
#define AllFree      0x5555555555555555
#define AllContinue  0xffffffffffffffff
#define AllEnd       0xAAAAAAAAAAAAAAAA
#endif

#define MINBLOCK_SIZE 16 /* bytes */
#define STATE_PER_MINBLOCK ((BITS_PER_BYTE / ST_SIZE) * MINBLOCK_SIZE) /* 64 */
#define CHUNK_SIZE (STATE_PER_MINBLOCK * MINBLOCK_SIZE) /* 1024 */

#define BLOCKS_PER_WORD (BITS_PER_WORD >> ST_SHIFT)
#define INVALID_BLOCKNUM ((UINT)(~0))

typedef struct _HPINFO {
    const struct IHeapVtbl *v;
    UINT Refs;
    MUTEX Lock;
    PIHEAP Parent;
    const PUINT pBitMap;
    const UINT BitMapSize; /* in words */
    const PBYTE pMemStart;
    const UINT Flags;
    struct {
        UINT FirstFree;
#define BAD_HINT 0x80000000
        UINT LastUsed;
        UINT Suballoc;
    } Hints;
    ADDR_SIZE ReservedMemory;
    ADDR_SIZE UsedMemory;
    ADDR_SIZE MaxUsedMemory;
} HPINFO;


/* Forward decls */
SCODE MCT HeapFree (IHeap *pThis, UINT Flags, PTR pMem);

#if _DEBUG
void HeapDumpState( PIHEAP pThis )
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT *pBitMap = pHp->pBitMap, v;
    UINT i, j, s, prev, max;
    static char *StateName[5] = {
         "Suballoc", "Free", "BlockEnds", "Used", "Start"
    };

    if (pBitMap == NULL) {
        printf("Heap not initialized\n");
        return;
    }

    printf("Heap @%x, bitmap @%x length %d words, memory @%x\n",
           (ADDRESS) pThis, (ADDRESS) pBitMap, pHp->BitMapSize,
           (ADDRESS) pHp->pMemStart);
    printf("Hints: %sffree %d lastu %d suballoc %d\n",
           (pHp->Hints.FirstFree & BAD_HINT) ? "?" : "",
           pHp->Hints.FirstFree & ~BAD_HINT,
           pHp->Hints.LastUsed, pHp->Hints.Suballoc);

    prev = 4;
    printf(" Block [0");
    max = pHp->Hints.LastUsed / BLOCKS_PER_WORD + 1;
    for (i = 0; i < max; i++) {
        v = pBitMap[i];

        for (j = 0; j < BLOCKS_PER_WORD; j++) {
            s = v & ST_MASK;
            v >>= ST_SIZE;
            if (s != prev) {
#define current (i * BLOCKS_PER_WORD + j)
                printf(" .. %d) %s", current, StateName[prev]);
                printf("\n Block [%d", current);
                prev = s;
#undef current
            }
        }
    }
    printf(" .. %d) %s\n", i * BLOCKS_PER_WORD, StateName[prev]);
    if (max != pHp->BitMapSize)
        printf(" Block [%d .. %d) %s (unused)\n",
               i * BLOCKS_PER_WORD, pHp->BitMapSize * BLOCKS_PER_WORD,
               StateName[ST_FREE]);
}
#endif


/* Search for NBLOCKS contiguous blocks.
 * If none found returns INVALID_BLOCKNUM.
 * If found, allocates them and returns the
 * index of the first one.
 */
static UINT HpAllocateBlocks( HPINFO * RESTRICT pHp, UINT nBlocks)
{
    UINT nBlocksSoFar, i, FirstBlockFree;
    UINT * RESTRICT pBitMap;
    UINT v;
    INT BitPos;
    BOOL BadHint;
    UINT EndBlk;

    assert(nBlocks > 0);

    Mutex_Lock(&pHp->Lock);

    nBlocksSoFar = 0;
    FirstBlockFree = 0;
    pBitMap = pHp->pBitMap;
    EndBlk = pHp->BitMapSize;

    i = (pHp->Hints.FirstFree & ~BAD_HINT) / BLOCKS_PER_WORD;
    BadHint = (pHp->Hints.FirstFree & BAD_HINT) != 0;

    for (; i < EndBlk; i++) {

        v = pBitMap[i];

        /*
         * Bitscan this one word
         */
        if (!IsAllBusy(v)) {
            /*
             * keep counting beans here
             */
            BitPos = 0;

            if (nBlocksSoFar) {
CountBeans:
                if ((v == AllFree) &&
                    ((nBlocks - nBlocksSoFar) >= BLOCKS_PER_WORD)) {
                    /* Large allocs get here
                     */
                    nBlocksSoFar += BLOCKS_PER_WORD;
                    BitPos = BITS_PER_WORD-ST_SIZE;
                    if (nBlocksSoFar == nBlocks)
                        goto FoundEm;
                    continue;
                }
                while ((v & ST_MASK) == ST_FREE) {
                    /*
                     * got enough beans ?
                     */
                    if (++nBlocksSoFar == nBlocks)
                        goto FoundEm;
                    BitPos += ST_SIZE;
                    v >>= ST_SIZE;
                }
                /* if we are being lucky, roll again */
                if (BitPos == BITS_PER_WORD)
                    continue;
            }

            /*
             * search for beans here
             */
            nBlocksSoFar = 0;

            while (((v & ST_MASK) != ST_FREE) && (BitPos < BITS_PER_WORD)) {
                BitPos += ST_SIZE;
                v >>= ST_SIZE;
            }
            if ((v & ST_MASK) == ST_FREE) {
                FirstBlockFree = ((i * BITS_PER_WORD) + BitPos) >> ST_SHIFT;
                if (BadHint) {
                    BadHint = FALSE;
                    pHp->Hints.FirstFree = FirstBlockFree;                  
                }
                if (nBlocks > 1)
                    goto CountBeans;
                goto FoundEm;
            }
        }
        /*
         * No luck
         */
        nBlocksSoFar = 0;
    }
    /* Bad luck
     */
    FirstBlockFree = INVALID_BLOCKNUM;
    goto GetOut;

    /* We do have enough blocks, mark them busy.
     */
 FoundEm:
    pHp->UsedMemory += nBlocks * MINBLOCK_SIZE;
    if (pHp->MaxUsedMemory < pHp->UsedMemory)
        pHp->MaxUsedMemory = pHp->UsedMemory;

    if ((pHp->Hints.FirstFree & ~BAD_HINT) == FirstBlockFree)
        /* incorrect, but good 'nuf */
        pHp->Hints.FirstFree = (FirstBlockFree + nBlocks) | BAD_HINT;

    if (pHp->Hints.LastUsed < FirstBlockFree + nBlocks)
        pHp->Hints.LastUsed = FirstBlockFree + nBlocks;

    /* Do it in reverse (last cache line is hot),
     * first mark the last block...
     */
    v = pBitMap[i];
    v = (v & ~(ST_MASK << BitPos)) | (ST_END << BitPos);
    pBitMap[i] = v;

    /* ...Then (if any) mark the preceeding ones as to-be-continued
     */
    for (; --nBlocks > 0;) {
        BitPos -= ST_SIZE;
        if (BitPos < 0) {
            BitPos = BITS_PER_WORD - ST_SIZE;
            i--;
            while (nBlocks >= BLOCKS_PER_WORD) {
                nBlocks -= BLOCKS_PER_WORD;
                pBitMap[i] = AllContinue;
                i--;
                if (nBlocks == 0)
                    goto GetOut;
            }
        }
        v = pBitMap[i];
        /* nb: this knows the encoding */
        v = (v /*& ~(ST_MASK << BitPos)*/) | (ST_CONTINUE << BitPos);
        pBitMap[i] = v;
    }
    
 GetOut:
    Mutex_Unlock(&pHp->Lock);
    return FirstBlockFree;
}

/* Search for NBLOCKS contiguous and aligned blocks.
 * If none found returns INVALID_BLOCKNUM.
 * If found, allocates them and returns the
 * index of the first one.
 */
static UINT HpAllocateAlignedBlocks( HPINFO * RESTRICT pHp, UINT nBlocks, UINT Alignment)
{
    UINT nBlocksSoFar, i, j, FirstBlockFree;
    UINT * RESTRICT pBitMap;
    UINT v;
    INT BitPos;
    UINT EndBlk;
    UINT BaseI, BaseB, DeltaI, DeltaB;

    /* What's the first acceptable pointer into the heap ?
     * Find its "word" and "bit" indexes.
     */
    BaseB = ((ADDRESS)pHp->pMemStart & (Alignment - 1));
    if (BaseB) BaseB = (Alignment - BaseB) / MINBLOCK_SIZE;
    BaseI = BaseB / BLOCKS_PER_WORD;
    BaseB = BaseB & (BLOCKS_PER_WORD - 1);

    /* What's the distance to the next acceptable pointer ?
     * Find its "word" and "bit" deltas.
     */
    assert(Alignment > MINBLOCK_SIZE);
    DeltaB = Alignment / MINBLOCK_SIZE;
    DeltaI = DeltaB / BLOCKS_PER_WORD;
    if (DeltaI == 0) DeltaI = 1;
    DeltaB = DeltaB & (BLOCKS_PER_WORD - 1);
    if (DeltaB == 0) DeltaB = BLOCKS_PER_WORD;

    assert(nBlocks > 0);

    nBlocksSoFar = 0;
    FirstBlockFree = 0;
    pBitMap = pHp->pBitMap;
    EndBlk = pHp->BitMapSize;

    Mutex_Lock(&pHp->Lock);

    /* See where we can start our search
     */
    i = (pHp->Hints.FirstFree & ~BAD_HINT) / BLOCKS_PER_WORD;
    if (i <= BaseI)
        i = BaseI;
    else {
        /* Our Deltas are of the 2^N form.
         * Instead of expensively dividing, use shifts.
         */
        UINT32 ShiftI = (DeltaI & -(INT32)DeltaI);
        ShiftI = ((ShiftI & 0xFFFF0000)!=0) << 4
               | ((ShiftI & 0xFF00FF00)!=0) << 3
               | ((ShiftI & 0xF0F0F0F0)!=0) << 2
               | ((ShiftI & 0xCCCCCCCC)!=0) << 1
               | ((ShiftI & 0xAAAAAAAA)!=0);

        i = (i >> ShiftI) << ShiftI; /* truncate (should round instead) */
    }

    for (; i < EndBlk;) {

        v = pBitMap[i];

        if (!IsAllBusy(v)) {

            BitPos = ST_SIZE*BaseB;
            v >>= BitPos;
            nBlocksSoFar = 0;

            while (((v & ST_MASK) != ST_FREE) && (BitPos < BITS_PER_WORD)) {
                BitPos += ST_SIZE*DeltaB;
                v >>= ST_SIZE*DeltaB;
            }
            if ((v & ST_MASK) == ST_FREE) {
                FirstBlockFree = ((i * BITS_PER_WORD) + BitPos) >> ST_SHIFT;
                j = 0;
                if (nBlocks > 1)
                    goto CountBeans;
                goto FoundEm;
            }
        }
        /* bad luck */
        i += DeltaI;
        continue;

CountBeans:
        do {
            if (++nBlocksSoFar == nBlocks) {
                i += j;
                goto FoundEm;
            }
            BitPos += ST_SIZE;
            v >>= ST_SIZE;
        } while ((v & ST_MASK) == ST_FREE);

        if (BitPos == BITS_PER_WORD) {
            j++;
            BitPos = 0;
            v = pBitMap[i+j];
            if ((v & ST_MASK) == ST_FREE)
              goto CountBeans;
        }
        j = j & ~(DeltaI - 1);/* trunc */
        i += j + DeltaI;
    }
    /* Bad luck
     */
    FirstBlockFree = INVALID_BLOCKNUM;
    goto GetOut;

    /* We do have enough blocks, mark them busy.
     */
 FoundEm:
    pHp->UsedMemory += nBlocks * MINBLOCK_SIZE;
    if (pHp->MaxUsedMemory < pHp->UsedMemory)
        pHp->MaxUsedMemory = pHp->UsedMemory;

    if ((pHp->Hints.FirstFree & ~BAD_HINT) == FirstBlockFree)
        /* incorrect, but good 'nuf */
        pHp->Hints.FirstFree = (FirstBlockFree + nBlocks) | BAD_HINT;

    if (pHp->Hints.LastUsed < FirstBlockFree + nBlocks)
        pHp->Hints.LastUsed = FirstBlockFree + nBlocks;

    /* Do it in reverse (last cache line is hot),
     * first mark the last block...
     */
    v = pBitMap[i];
    v = (v & ~(ST_MASK << BitPos)) | (ST_END << BitPos);
    pBitMap[i] = v;

    /* ...Then (if any) mark the preceeding ones as to-be-continued
     */
    for (; --nBlocks > 0;) {
        BitPos -= ST_SIZE;
        if (BitPos < 0) {
            BitPos = BITS_PER_WORD - ST_SIZE;
            i--;
        }
        v = pBitMap[i];
        /* nb: this knows the encoding */
        v = (v /*& ~(ST_MASK << BitPos)*/) | (ST_CONTINUE << BitPos);
        pBitMap[i] = v;
    }
    
 GetOut:
    Mutex_Unlock(&pHp->Lock);
    return FirstBlockFree;
}

/* Marks NBLOCKS contiguous blocks as used.
 * Start at STARTBLOCK (inclusive).
 * If already in use returns FALSE.
 * Otherwise, allocates them and returns TRUE.
 */
static BOOL HpStealBlocks( HPINFO *pHp, UINT StartBlock, UINT nBlocks)
{
    UINT nBlocksSoFar, i;
    UINT *pBitMap = pHp->pBitMap, v;
    INT BitPos;

    assert(nBlocks > 0);

    i = StartBlock / BLOCKS_PER_WORD;
    BitPos = (StartBlock & (BLOCKS_PER_WORD - 1)) * ST_SIZE;

    nBlocksSoFar = 0;

    Mutex_Lock(&pHp->Lock);

    /* Forward-scan to make sure they are all free
     */
    for (; i < pHp->BitMapSize; i++) {

        v = pBitMap[i];

        /*
         * Bitscan this one word
         */
        v >>= BitPos;

        while ((v & ST_MASK) == ST_FREE) {
            /*
             * got enough beans ?
             */
            if (++nBlocksSoFar == nBlocks)
                goto FoundEm;
            BitPos += ST_SIZE;
            v >>= ST_SIZE;
        }

        /*
         * Did we hit a non-free block ?
         */
        if (BitPos != BITS_PER_WORD)
            break;

        /*
         * All is right, keep going.
         */
        BitPos = 0;
    }

    /* Bad luck
     */
    Mutex_Unlock(&pHp->Lock);
    return FALSE;

    /* We do have enough blocks, mark them busy.
     */
 FoundEm:
    pHp->UsedMemory += nBlocks * MINBLOCK_SIZE;
    if (pHp->MaxUsedMemory < pHp->UsedMemory)
        pHp->MaxUsedMemory = pHp->UsedMemory;

    if ((pHp->Hints.FirstFree & ~BAD_HINT) == StartBlock)
        /* incorrect, but good 'nuf */
        pHp->Hints.FirstFree = (StartBlock + nBlocks) | BAD_HINT;

    if (pHp->Hints.LastUsed < StartBlock + nBlocks)
        pHp->Hints.LastUsed = StartBlock + nBlocks;

    /* Do it in reverse (last cache line is hot),
     * first mark the last block...
     */
    v = pBitMap[i];
    v = (v & ~(ST_MASK << BitPos)) | (ST_END << BitPos);
    pBitMap[i] = v;

    /* ...Then (if any) mark the preceeding ones as to-be-continued
     */
    for (; --nBlocks > 0;) {
        BitPos -= ST_SIZE;
        if (BitPos < 0) {
            BitPos = BITS_PER_WORD - ST_SIZE;
            i--;
        }
        v = pBitMap[i];
        /* nb: this knows the encoding */
        v = (v /*& ~(ST_MASK << BitPos)*/) | (ST_CONTINUE << BitPos);
        pBitMap[i] = v;
    }
    
    Mutex_Unlock(&pHp->Lock);
    return TRUE;
}

static void HpFreeBlocks( HPINFO *pHp, UINT StartBlock)
{
    UINT nBlocks, FirstBlockToFree, i, BitPos;
    UINT *pBitMap = pHp->pBitMap, v, mask;

    /* First pass: make sure we are really given the beginning
     * of a busy block.  This section of code allows users to
     * free memory without knowing where it really started.
     * This can be useful in things like network protocols,
     * e.g. where data is one-scan parsed by independent layers 
     * and the final layer (who frees the buffer) no longer
     * knows where the buffer starts.
     */
    i = 0;
    FirstBlockToFree = StartBlock;
    if (StartBlock > 0) {
#if later
        /* Backup one
         */
        StartBlock -= 1;
#endif
        i = StartBlock & ~(BLOCKS_PER_WORD - 1);
        StartBlock = StartBlock - i;
        i /= BLOCKS_PER_WORD;

#if later
        /* Scan backward until we find a non-continue block.
         */
        v = pBitMap[i];
        ...;
#endif
    }

    /* Now free the blocks with a forward scan.
     */

    Mutex_Lock(&pHp->Lock);

    for (nBlocks = 0; ; ) {
        v = pBitMap[i];
        BitPos = ST_SIZE * StartBlock;
        mask = (0xffffffff >> BitPos) << BitPos;

        if ((v & mask) == (AllContinue & mask)) {

            /* easy case: a used chunk */
            v = (v & ~mask) | (AllFree & mask);
            nBlocks += BLOCKS_PER_WORD - StartBlock;

            /* Done with this block */
            pBitMap[i] = v;
            i++;
            StartBlock = 0;

            continue;

        }

        /* End of block must be in here, somewhere.
         * Therefore this loop does terminate.
         */
        while (((v >> BitPos) & ST_MASK) == ST_CONTINUE) {
            BitPos += ST_SIZE;
            nBlocks++;
        }
        nBlocks++;

        /* BitPos points to where ST_END is */
        assert(((v >> BitPos) & ST_MASK) == ST_END);

        /* Dont touch the high bits (if any) */
        BitPos = BITS_PER_WORD - ST_SIZE - BitPos;
        mask = (mask << BitPos) >> BitPos;

        v = (v & ~mask) | (AllFree & mask);
        pBitMap[i] = v;

        break;
    }

    if ((pHp->Hints.FirstFree &~BAD_HINT) >= FirstBlockToFree)
        pHp->Hints.FirstFree = FirstBlockToFree;
    if (pHp->Hints.LastUsed == FirstBlockToFree + nBlocks)
        pHp->Hints.LastUsed = FirstBlockToFree;
    pHp->UsedMemory -= nBlocks * MINBLOCK_SIZE;

    Mutex_Unlock(&pHp->Lock);
}

/* Lock-free */
static UINT HpBlockOf( HPINFO *pHp, BYTE *pMem)
{
    UINT BlkNum;
    ADDRESS Offset;

    if (pMem < pHp->pMemStart)
        goto InvalExit;

    Offset = pMem - pHp->pMemStart;
    BlkNum = Offset / MINBLOCK_SIZE;

    if (BlkNum <= (pHp->BitMapSize * BLOCKS_PER_WORD))
        return BlkNum;

 InvalExit:
    {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_INVALID_PARAMETER);
    }
    return INVALID_BLOCKNUM;
}

/* Lock-free */
static UINT HpSizeOf( HPINFO *pHp, UINT StartBlock)
{
    UINT *pBitMap = pHp->pBitMap, v;
    UINT i, j, s, n, max;

    max = pHp->Hints.LastUsed / BLOCKS_PER_WORD + 1;

    /* See where we start from
     * xxx should we check this is actually the beginning of the block ?
     */
    i = StartBlock & ~(BLOCKS_PER_WORD - 1);
    StartBlock = StartBlock - i;
    i /= BLOCKS_PER_WORD;

    /* First block cant be free or suballoced
     */
    v = pBitMap[i];
    s = (v & (ST_MASK << (ST_SIZE * StartBlock))) >> (ST_SIZE * StartBlock);
    if ((s != ST_CONTINUE) && (s != ST_END)) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_INVALID_PARAMETER);
        return 0;
    }

    /* How many (more) contiguous blocks do we have
     */
    n = 1;
    for (; i < max; ) {

        for (j = StartBlock; j < BLOCKS_PER_WORD; j++) {

            s = (v & (ST_MASK << (ST_SIZE * j))) >> (ST_SIZE * j);

            /* Should terminate with an ST_END, but play it safe
             */
            if (s != ST_CONTINUE)
                return n;

            n++;
        }

        /* Next word in bitmap
         */
        v = pBitMap[++i];
        StartBlock = 0;
    }

    return n;
}

PTR MCT HeapAlloc (
            IHeap *pThis,
            UINT Flags,
            UINT Size,
            UINT Alignment)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT BlkNum;
    PTR pMem;
    TRACE_USER_LEAKS(UINT tSize = Size;)

    CHECK_DELAYED_FREE();

    /* Alignment issues
     */
    if (Flags & HEAP_CACHE_ALIGN)
    {
        Alignment = _DCACHELINESIZE;
        Size = (Size + _DCACHELINESIZE - 1) & ~(_DCACHELINESIZE - 1);
    }
    else if (Alignment <= MINBLOCK_SIZE)
        Alignment = 0;
    /* Check that it is a power of two */
    else if ((Alignment-1) & Alignment) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_INVALID_PARAMETER);
        return NULL;
    }

    /* Roundup and allocate
     */
    Size = (Size) ? Size : 1;
    Size = (Size + MINBLOCK_SIZE - 1) & ~(MINBLOCK_SIZE - 1);
    BlkNum = (Alignment) ? 
        HpAllocateAlignedBlocks( pHp, Size / MINBLOCK_SIZE, Alignment) :
        HpAllocateBlocks( pHp, Size / MINBLOCK_SIZE);

    if (BlkNum != INVALID_BLOCKNUM) {
        pMem = pHp->pMemStart + (BlkNum * MINBLOCK_SIZE);
        /* Zero if we have to
         */
        if ((Flags | pHp->Flags) & HEAP_ZERO_MEMORY)
            memset(pMem, 0, Size);
        TRACE_USER_LEAKS(printf("a %x %d\n", pMem, tSize);)
        return pMem;
    }
    {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_NOT_ENOUGH_MEMORY);
    }
    return NULL;
}

/* BUGBUG Alignment BUGBUG */
/* BUGBUG Alignment BUGBUG */
/* BUGBUG Alignment BUGBUG */
/* BUGBUG Alignment BUGBUG */
PTR MCT HeapRealloc (
            IHeap *pThis,
            UINT Flags,
            PTR pMem,
            UINT NewSize,
            UINT Alignment)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT BlkNum, EndBlk;
    UINT i, v, t;
    ADDR_SIZE OldSize, Size;
    INT Diff;
    TRACE_USER_LEAKS(UINT tSize = NewSize;)
    TRACE_USER_LEAKS(PTR tpMem = pMem;)

    /* ANSI sez..
     */
    if (pMem == NULL)
        return HeapAlloc(pThis, Flags, NewSize, Alignment);

    if (NewSize == 0)
        return (HeapFree(pThis,0,pMem)) ? NULL : pMem;

    Flags |= pHp->Flags;

    CHECK_DELAYED_FREE();

    /* Alignment issues
     */
    if (Flags & HEAP_CACHE_ALIGN)
    {
        Alignment = _DCACHELINESIZE;
        NewSize = (NewSize + _DCACHELINESIZE - 1) & ~(_DCACHELINESIZE - 1);
    }
    else if (Alignment <= MINBLOCK_SIZE)
        Alignment = 0;
    /* Check that it is a power of two */
    else if ((Alignment-1) & Alignment) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_INVALID_PARAMETER);
        return NULL;
    }

    /* Check its a honest pointer
     */
    BlkNum = HpBlockOf(pHp,(BYTE*)pMem);
    if (BlkNum == INVALID_BLOCKNUM) {
        pMem = NULL;
        goto GetOut;
    }

    /* Note that both OldSize and Size are at least 1 block large.
     */
    OldSize = HpSizeOf(pHp,BlkNum) * MINBLOCK_SIZE;
    Size    = (NewSize + MINBLOCK_SIZE - 1) & ~(MINBLOCK_SIZE - 1);
    Diff    = Size - OldSize;

    /* Three cases here:
     * a- its a shrink, but not worth a penny
     * b- its a big shrink
     * c- we must grow the allocation
     */
    if (Diff == 0)
        /* Case A.  Just return pMem.
         */
        goto GetOut;

    if (Diff < 0) {
        /* Case B.  Free the excess.
         */
        EndBlk = BlkNum + (Size / MINBLOCK_SIZE) - 1;

        /* First should mark EndBlk as ST_END
         */
        i = EndBlk & ~(BLOCKS_PER_WORD - 1);
        BlkNum = EndBlk - i; /* num within word */
        i /= BLOCKS_PER_WORD;
#define mask v
        do {
            mask = ST_MASK << (BlkNum << ST_SHIFT);
            t = pHp->pBitMap[i];
            v = (t & ~mask) | (AllEnd & mask);
#undef mask
        } while (!AtomicCmpAndSwap(&pHp->pBitMap[i], t, v));

        /* Now we can free and return
         */
        HpFreeBlocks(pHp,EndBlk+1);
        goto GetOut;
    }

    /* Case C.  See if we can do it in place.
     */
    EndBlk = BlkNum + (OldSize / MINBLOCK_SIZE) - 1;
    if (HpStealBlocks(pHp, EndBlk + 1, Diff / MINBLOCK_SIZE)) {

        /* Yes can do.
         * Mark EndBlk as ST_CONTINUE to join the two partials.
         */
        i = EndBlk & ~(BLOCKS_PER_WORD - 1);
        BlkNum = EndBlk - i; /* num within word */
        i /= BLOCKS_PER_WORD;
#define mask v
        do {
            mask = ST_MASK << (BlkNum << ST_SHIFT);
            t = pHp->pBitMap[i];
            v = (t & ~mask) | (AllContinue & mask);
#undef mask
        } while (!AtomicCmpAndSwap(&pHp->pBitMap[i], t, v));

        /* Done.
         */
        goto GetOut;
    }

    /* If we cant do it in place allocate a new one.
     * Unless prohibited by the flags.
     */
    if (Flags & HEAP_REALLOC_IN_PLACE_ONLY) {
        pMem = NULL;
        goto GetOut;
    }

    EndBlk = (Alignment) ? 
        HpAllocateAlignedBlocks( pHp, Size / MINBLOCK_SIZE, Alignment) :
        HpAllocateBlocks( pHp, Size / MINBLOCK_SIZE);
    if (EndBlk == INVALID_BLOCKNUM) {
        /* BUGBUG make a last-ditch effort searching back from BlkNum
         */
        goto GetOut;
    }

    /* compilerhelp */
    {
        PTR pTmp = pMem;
        pMem = pHp->pMemStart + (EndBlk * MINBLOCK_SIZE);
        if ((Flags & HEAP_NO_COPY) == 0)
            memcpy(pMem, pTmp, OldSize);
    }

    HpFreeBlocks(pHp,BlkNum);


 GetOut:
    /* xxx handle here all zeroings of trail if Flags & ZERO_MEMORY */
    TRACE_USER_LEAKS(printf("r %x %d %x\n", tpMem, tSize, pMem);)

    return pMem;
}

SCODE MCT HeapFree (
            IHeap *pThis,
            UINT Flags,
            PTR pMem)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT BlkNum;

    UnusedParameter(Flags);

    CHECK_DELAYED_FREE();

    BlkNum = HpBlockOf(pHp,(BYTE*)pMem);
    if (BlkNum == INVALID_BLOCKNUM)
        return FALSE;

    HpFreeBlocks( pHp, BlkNum);

    TRACE_USER_LEAKS(printf("f %x\n", pMem);)
    return S_OK;

}

ADDR_SIZE MCT HeapSize (
            IHeap *pThis,
            UINT Flags,
            PTR pMem)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT BlkNum;

    UnusedParameter(Flags);

    CHECK_DELAYED_FREE();

    BlkNum = HpBlockOf(pHp,(BYTE*)pMem);
    if (BlkNum == INVALID_BLOCKNUM)
        return 0;

    return HpSizeOf(pHp,BlkNum) * MINBLOCK_SIZE;
}

SCODE MCT HeapValidate (
            IHeap *pThis,
            UINT Flags,
            PTR pMem)
{
#if _DEBUG
    /* In debug builds, check bitmap for consistency
     * 
     * A valid bitmap should only contain valid sequences
     * of ST_xxx values:
     * - ST_FREE (a) can be the first block (b) cannot follow ST_CONTINUE
     * - ST_CONTINUE (a) can be first (b) can follow anything
     * - ST_END (a) can be first (b) can follow anything
     * - ST_SUBALLOC -- not implemented yet
     *
     * Given the encoding, for all possible nibbles this gives:
     *      Nibble  Currently  Eventually
     *       0000    invalid
     *       0001    invalid
     *       0010    invalid
     *       0011    invalid
     *       0100    invalid
     *       0101 
     *       0110 
     *       0111    invalid    invalid
     *       1000    invalid
     *       1001 
     *       1010 
     *       1011 
     *       1100    invalid
     *       1101 
     *       1110 
     *       1111 
     */
    HPINFO *pHp = (HPINFO *)pThis;
    UINT *pBitMap = pHp->pBitMap, v;
    UINT i, j, s, prev, max;

    prev = 4;
    max = pHp->BitMapSize;

    Mutex_Lock(&pHp->Lock);
    for (i = 0; i < max; i++) {
        v = pBitMap[i];

        /* trivial optim */
        if (v == AllContinue) {
            prev = ST_CONTINUE;
            continue;
        }

        for (j = 0; j < BLOCKS_PER_WORD; j++) {
            s = v & ST_MASK;
            v >>= ST_SIZE;
            switch (s) {

            case ST_FREE:
                if (prev == ST_CONTINUE)
                    goto BitMapError;
                break;

            case ST_END:
            case ST_CONTINUE:
                break;

                /* case ST_SUBALLOC not implem */
            default:
#define current (i * BLOCKS_PER_WORD + j)
    BitMapError:
#if _DEBUG
                printf("HeapCorrupted! bad bitmap at block %d [%x]\n",
                       current, pBitMap[i]);
#else
                return E_FAIL;
#endif
            }
#undef current
            prev = s;
        }
    }
    Mutex_Unlock(&pHp->Lock);

//    if ((ADDRESS)pMem == 1) HeapDumpState(pThis);
#endif

    /* Either way, we have no use for flags and a valid pointer is..
     */
    UnusedParameter(Flags);

    if ((pMem == NULL) || (HeapSize(pThis, 0, pMem) != 0))
        return S_OK;
    else
        return E_FAIL;
}

PTR MCT HeapExtract (
            IHeap *pThis,
            UINT Flags,
            PTR pMem,
            UINT Size)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT BlkNum;

    UnusedParameter(Flags);

    CHECK_DELAYED_FREE();

    BlkNum = HpBlockOf(pHp,(BYTE*)pMem);
    if (BlkNum == INVALID_BLOCKNUM)
        return NULL;

    /* Roundup SIZE for starting and trailing slack
     */
    Size += ((ADDRESS)pMem & (MINBLOCK_SIZE - 1));
    Size = (Size + MINBLOCK_SIZE - 1) & ~(MINBLOCK_SIZE - 1);

    /* Allocate if we can
     */
    if (! HpStealBlocks(pHp, BlkNum, Size / MINBLOCK_SIZE)) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, E_INVALID_PARAMETER);
        return NULL;
    }

    /* Trim back to the beginning
     */
    pMem = pHp->pMemStart + (BlkNum * MINBLOCK_SIZE);

    /* Zero if we have to
     */
    if ((Flags | pHp->Flags) & HEAP_ZERO_MEMORY)
        memset(pMem, 0, Size);
    return pMem;
}

SCODE MCT HeapStatus (
            IHeap *pThis,
            ADDR_SIZE *pReserve,
            ADDR_SIZE *pCommit,
            ADDR_SIZE *pUsed,
            ADDR_SIZE *pMaxUsed)
{
    HPINFO *pHp = (HPINFO *)pThis;

    CHECK_DELAYED_FREE();

    TRACE_USER_LEAKS(printf("S\n");)

    if (pReserve)
        *pReserve = pHp->ReservedMemory;
    if (pCommit)
        *pCommit  = pHp->ReservedMemory;
    if (pUsed)
        *pUsed    = pHp->UsedMemory;
    if (pMaxUsed) {
        do {
            *pMaxUsed = pHp->MaxUsedMemory;
        } while (!AtomicCmpAndSwap((UINT*) &pHp->MaxUsedMemory,
                                   (UINT) *pMaxUsed,
                                   (UINT) pHp->UsedMemory));
    }

    return NO_ERROR;
}

PRIVATE SCODE HeapCreateHeap(PIHEAP iThis,
                             UINT Flags, UINT InitialSize, UINT MaxSize,
                             PIHEAP *ppHeap)
{
    SCODE sc;
    PIHEAP NewHeap;
    HPINFO *phi;
    PTR Mem = iThis->v->Alloc(iThis, Flags, MaxSize, 0);
    if (!Mem)
        return E_NOT_ENOUGH_MEMORY;

    sc = BitMapHeapFactoryCreate(NULL, 
                                 (ADDRESS) Mem, Flags,
                                 InitialSize, MaxSize,
                                 &NewHeap);
    if (FAILED(sc)) {
        iThis->v->Free(iThis, Flags, Mem);
        return sc;
    }

    phi = (HPINFO*) NewHeap;
    phi->Parent = iThis;
    iThis->v->AddRef(iThis);

    *ppHeap = NewHeap;
    return S_OK;
}

#if !IN_WINDOWS
/*
 * IUnknown methods
 */
UINT MCT HeapAddRef(PIHEAP pThis)
{
    HPINFO *pHp = (HPINFO *)pThis;

    return AtomicInc(&pHp->Refs);
}

UINT MCT HeapRelease(PIHEAP pThis)
{
    HPINFO *pHp = (HPINFO *)pThis;
    UINT newrefs;

    newrefs = AtomicDec(&pHp->Refs);

    if (newrefs == 0) {
        PIHEAP Parent = pHp->Parent;
        Parent->v->Free(Parent,0,pThis);
    }

    return newrefs;
}

SCODE MCT HeapQueryInterface(PIHEAP pThis, REFIID pIid, void* *ppUnk)
{
    return GenericQueryInterface((IUnknown *)pThis,pIid,ppUnk,&IID_IHeap);
}
#endif

const struct IHeapVtbl HeapVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
#if !IN_WINDOWS
    HeapQueryInterface, /* implemented        */
    HeapAddRef,         /* implemented        */
    HeapRelease,        /* implemented        */
#endif
    HeapAlloc,          /* partly implemented */
    HeapRealloc,        /* partly implemented */
    HeapFree,           /* implemented        */
    HeapSize,           /* implemented        */
    HeapValidate,       /* implemented        */
    HeapExtract,        /* implemented        */
    HeapStatus,          /* implemented        */
    HeapCreateHeap
};

SCODE BitMapHeapFactoryCreate(
    PIHEAPFACTORY iThis,
    ADDRESS Mem,
    UINT Flags,
    ADDR_SIZE InitSize,
    ADDR_SIZE MaxSize,
    /*out*/ PIHEAP *ppHeap)
{
    HPINFO *pHp;
    UINT Ovhd;

    /* Memory must be at least word-aligned
     * (else assume machine state is corrupt)
     */
    if (Mem & (sizeof(UINT) - 1))
        return E_INVALID_PARAMETER;

    UnusedParameter(InitSize); /* cannot extend, really */

    /* Carve out our overhead costs from the start of the heap
     * 
     * Notice the bitmap need not cover the memory for overhead,
     * and the equation therefore is
     *   Overhead = sizeof(HPINFO) + ((MaxSize - Overhead) / frac);
     * where FRAC is the overhead per block.
     *
     */
#define fract (MINBLOCK_SIZE * (BITS_PER_BYTE / ST_SIZE))

    Ovhd = ( (fract * sizeof(HPINFO)) + MaxSize ) / ( 1 + fract );

    /* Add into overhead the alignment slop, if any
     */
    Ovhd += (MINBLOCK_SIZE - ((Mem + Ovhd) & (MINBLOCK_SIZE - 1)))
        & (MINBLOCK_SIZE - 1);

    pHp = (HPINFO *) Mem;
    memset(pHp, 0, sizeof *pHp);

    *(UINT*)(&pHp->Flags) = Flags;
    pHp->v = &HeapVtbl;
    pHp->Refs = 1;

    Mutex_Init(&pHp->Lock);

    *(PUINT*)(&pHp->pBitMap) = (UINT *) (Mem + sizeof(HPINFO));
    /* BUGBUG this should be in blocks, not words */
    *(UINT*)(&pHp->BitMapSize) = (MaxSize - Ovhd) / (fract * sizeof(UINT));

    *(PBYTE*)(&pHp->pMemStart) = (BYTE *)(Mem + Ovhd);

    pHp->ReservedMemory = MaxSize;
    pHp->UsedMemory = Ovhd;
    pHp->MaxUsedMemory = pHp->UsedMemory;

    memset(pHp->pBitMap, AllFree, Ovhd - sizeof(HPINFO));

    *ppHeap = (PIHEAP) pHp;
    return S_OK;
#undef fract
}

PRIVATE struct HEAPCOB BitMapHeapCobObject = {
    &BitMapHeapFactoryVtbl,
    1
};

PIUNKNOWN BitMapHeapCobMain(void)
{
    return (PIUNKNOWN) &BitMapHeapCobObject;
}

/* And here is the test code
 */
#if IN_WINDOWS
#include "test.c"
#endif


