/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

PIHEAP
GetHeap(
        _TCHAR *Name,
        ADDRESS MemoryStart,
        UINT Flags,
        UINT InitialSize,
        UINT MaximumSize
        )
{
    SCODE sc;
    PIHEAP Heap, pUnk;
    PIHEAPFACTORY Factory;

    sc = BindToObject(CurrentNameSpace(), Name,
                      NAME_SPACE_READ, &IID_IHeapFactory,
                      (void *) &Factory);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Could not find the %s factory sc=%x\n\r"),
                 Name, sc);
        return NULL;
    }

    sc = Factory->v->Create(Factory,MemoryStart,Flags,InitialSize,MaximumSize,
                            &pUnk);
    (void) Factory->v->Release(Factory);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Constructor failed sc=%x\n\r"), sc);
        return NULL;
    }

    sc = pUnk->v->QueryInterface(pUnk,&IID_IHeap,(void**)&Heap);
    pUnk->v->Release(pUnk);

    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not a heap sc=%x\n\r"), Name, sc);
        return NULL;
    }

    return Heap;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    INT RetVal = EXIT_FAILURE;
    _TCHAR *HeapName, *FactoryName;
    void *Mem;
    UINT Flags, InitialSize, MaximumSize;
    PIHEAP NewHeap;
    PINAMESPACE Ns;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\n\r"), sc);
        goto Fail;
    }

    if (argc < 6) {
        _tprintf(_TEXT("Usage: %s HeapCobName NewHeapName HexFlags StartSize MaximumSize\n\r"),
                 argv[0]);
        goto Out;
    }

    FactoryName = argv[1];
    HeapName = argv[2];
    Flags = _tcstoul(argv[3],NULL,16);
    InitialSize = _tcstoul(argv[4],NULL,10);
    MaximumSize = _tcstoul(argv[5],NULL,10);

    Mem = malloc(MaximumSize);
    if (Mem == NULL) {
        _tprintf(_TEXT("Failed to allocate %d bytes of memory for heap.\n\r"),
                 MaximumSize);
        goto Out;
    }

    _tprintf(_TEXT("Building %s(%x,%x,%d,%d)\n\r"),
             FactoryName,(ADDRESS)Mem,Flags,InitialSize,MaximumSize);

    NewHeap = GetHeap(FactoryName,(ADDRESS)Mem,Flags,InitialSize,MaximumSize);

    _tprintf(_TEXT("Constructor returned x%x\n\r"), (ADDRESS) NewHeap);
    if (NewHeap == NULL)
        goto Out;

    /* Register in namespace and say what
     */
    Ns = CurrentNameSpace(); /* wont fail */
    assert(Ns != NULL);

    sc = Ns->v->Register(Ns, HeapName, (PIUNKNOWN) NewHeap, 0, NULL);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Failed (sc=0x%x) to register HP 0x%x as %s.\n"),
                 sc, (ADDRESS)NewHeap, HeapName);
        NewHeap->v->Release(NewHeap);
        goto Out;
    }
    RetVal = EXIT_SUCCESS;

 Out:
    (void) FreeCmdLine(&argc,&argv);
 Fail:
    exit(RetVal);
    return 0;/*amusme*/
}
