/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * MMLite hardware loader for ACE image format.
 */

/* Always included by someone else
 */

extern BOOL SysaceSendConfig(/*struct _Sac*/ void *Interface, UINT32 *Data, UINT nBytes);

/* By trial-and-error we figured out that the SystemACE chip skips
 * the first 512 bytes of the ACE file, which is this ASCII header:
 */
PRIVATE char AceHeader[] = 
"0000000000000000000000000000000000000000000000000000000000000000\n"
"Automatically generated. DO NOT MODIFY\n"
"Xilinx, Inc.\n"
"xmpmutil(tm) Multi-Function MPM Utility Version 1.01 (c) 2001 Xilinx, Inc.\n";

/* ...and spaces to fill the rest.
 * When fed by the MPU the chip doesnt want to see this header.
 */
#define ACEFILEOFFSET 512
#define ACEROUNDUP 32

/*
 * Load an ACE-formatted program image into the FPGA configuration memory.
 */
SCODE AceLdrLoadImage(UINT32 *Data, UINT32 DataSize, UINT32 Flags)
{
    SCODE sc = E_INVALID_PARAMETER; /* most likely */
    int i;

    UnusedParameter(Flags);

    /* Roundup to 32 */
    DataSize = (DataSize + ACEROUNDUP - 1) & ~(ACEROUNDUP-1);

    /* Quick sanity check
     */
    if (DataSize < ACEFILEOFFSET) {
        DBGME(3,_tprintf(_T("File too small (%d<%d)\n"), DataSize,ACEFILEOFFSET));
        return sc;
    }

    /* Make sure its the right stuff
     */
    i = memcmp(Data,AceHeader,(sizeof AceHeader)-1);
    if (i != 0) {
        DBGME(3,_tprintf(_T("Invalid ACE file header(%d,%d)?\n"),
                         i, sizeof AceHeader));
        goto Exit;
    }

    /* Send it off
     */
    if (!SysaceSendConfig(NULL,
                          (UINT32*) ((char *)Data + ACEFILEOFFSET),
                          DataSize - ACEFILEOFFSET)) {
        DBGME(3,_tprintf(_T("AceLdr: sysace failed.\n"), DataSize));
        goto Exit;
    }

    /* Done ok
     */
    sc = NO_ERROR;
    
Exit:
    DBGME(2,printf("se_ldr loaded sc=%x\n", sc));
    return sc;
}
