/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * I/O interface for the loader (old, defunct).
 */

#if defined(ANGEL_SEMIHOST)

typedef UINT FILE_HANDLE;
#define INVALID_FILE_HANDLE 0

typedef struct _openArgs {
    const _TCHAR * FileName;
    UINT Mode;
    UINT FileNameLen;
} OpenArgs;

typedef struct _closeArgs {
    FILE_HANDLE fh;
} CloseArgs;

typedef struct _readArgs {
    FILE_HANDLE fh;
    BYTE *buffer;
    UINT nBytes;
    UINT Mode;
} ReadArgs;

typedef struct _writeArgs {
    FILE_HANDLE fh;
    BYTE *buffer;
    UINT nBytes;
} WriteArgs;

typedef struct _seekArgs {
    FILE_HANDLE fh;
    UINT offset;
} SeekArgs;

typedef struct _fLenArgs {
    FILE_HANDLE fh;
} FLenArgs;

#define swiOpen     0x1
#define swiClose    0x2
#define swiWrite    0x5
#define swiRead     0x6
#define swiSeek     0xa
#define swiFLen     0xc

FILE_HANDLE __swi(0x123456) _ARMOpen( const UINT swiID, OpenArgs *args );
INT __swi(0x123456)         _ARMClose( const UINT swiID, CloseArgs *args );
INT __swi(0x123456)         _ARMRead( const UINT swiID, ReadArgs *args );
INT __swi(0x123456)         _ARMWrite( const UINT swiID, WriteArgs *args );
UINT __swi(0x123456)        _ARMSeek( const UINT swiID, SeekArgs *args );
UINT __swi(0x123456)        _ARMFLen( const UINT swiID, FLenArgs *args );

/* Since we are not using the C library, we must open the stdio
 * streams ourselves.
 */
FILE_HANDLE CheckStd(FILE_HANDLE f)
{
    static FILE_HANDLE ZeroOneTwo[3];
    static int inited;

    if (!inited) {
        OpenArgs args;

        args.FileName = ":tt";
        args.FileNameLen = 3;

        args.Mode = 0;
        ZeroOneTwo[0] = _ARMOpen(swiOpen, &args);    /* stdin  */
        args.Mode = 4;
        ZeroOneTwo[1] = _ARMOpen(swiOpen, &args);    /* stdout&stderr */
        args.Mode = 5;
        ZeroOneTwo[2] = _ARMOpen(swiOpen, &args);    /* stdout&stderr */

        inited = 1;
    }

    if (f < 3)
        return ZeroOneTwo[f];
    else
        return f;
}

#define _Open open

FILE_HANDLE open( const _TCHAR * FileName, UINT Mode /*, UINT Flag*/)
{
    OpenArgs args;

    CheckStd(0);        // make sure stdio streams are open

    args.FileName = FileName;
    args.Mode = Mode;
    args.FileNameLen = strlen(FileName);

    return _ARMOpen(swiOpen, &args);
}

#define _Close close

INT close( FILE_HANDLE File )
{
    CloseArgs args;

    args.fh = CheckStd(File);

    return _ARMClose( swiClose, &args );
}

#define _Read read

INT read( FILE_HANDLE File, BYTE *Buffer, UINT nBytes)
{
    INT RetVal;
    ReadArgs args;

    args.fh = CheckStd(File);
    args.buffer = Buffer;
    args.nBytes = nBytes;
    args.Mode = 0;
    
    RetVal = _ARMRead(swiRead, &args);

    if (RetVal < 0)
        return RetVal;

    return nBytes - RetVal;
}

#define _Write write

INT write( FILE_HANDLE File, BYTE *Buffer, UINT nBytes )
{
    INT RetVal;
    WriteArgs args;

    args.fh = CheckStd(File);
    args.buffer = Buffer;
    args.nBytes = nBytes;

    RetVal = _ARMWrite(swiWrite, &args);

    if (RetVal < 0)
        return RetVal;

    return nBytes - RetVal;
}

#define _Seek seek

UINT lseek( FILE_HANDLE File, UINT Position, UINT Whence )
{
    SeekArgs args;

    UnusedParameter(whence);        /* BUGBUG */

    args.fh = CheckStd(File);
    args.offset = Position;

    return _ARMSeek(swiSeek, &args);
}

UINT seek( FILE_HANDLE File, UINT Position )
{
    return lseek(File, Position, 0);
}

#define stat filelength
#define _Stat stat

INT filelength( FILE_HANDLE File )
{
    FLenArgs args;

    args.fh = CheckStd(File);

    return _ARMFLen(swiFLen, &args);
}


/* This does not quite belong but its for the loader
 */

typedef struct _cmdLineArgs {
    BYTE *buffer;
    UINT nBytes;
} CmdLineArgs;

#define swiGetCmdLine   0x15
typedef const _TCHAR *CTSTR;

CTSTR __swi(0x123456) _ArmCmdLine(UINT swiID, CmdLineArgs *args);

const _TCHAR * ArmCmdLine(BYTE *Buffer, UINT nBytes)
{
    CmdLineArgs args;

    args.buffer = Buffer;
    args.nBytes = nBytes;

    return _ArmCmdLine(swiGetCmdLine, &args);
}

#endif
