/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*-
 * Copyright (c) 1985, 1989, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * -
 * Portions Copyright (c) 1993 by Digital Equipment Corporation.
 * 
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies, and that
 * the name of Digital Equipment Corporation not be used in advertising or
 * publicity pertaining to distribution of the document or software without
 * specific, written prior permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * -
 * --Copyright--
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)res_init.c  8.1 (Berkeley) 6/7/93";
static char rcsid[] = "$Id: res_init.c,v 4.9.1.1 1993/05/02 22:43:03 vixie Rel $";
#endif /* LIBC_SCCS and not lint */

#include "res.h"

//#include <sys/param.h>
#include <socket.h>
#include <in.h>
#include <inet.h>
#include <nameser.h>
#include <resolv.h>
//#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <if.h>
#include <dhcp/dhcp_var.h>

int WINAPI ioctlsocket (SOCKET s, long cmd, u_long *argp);

/*
 * Resolver state default settings
 */

struct __res_state _res = {
        RES_TIMEOUT,                    /* retransmition time interval */
        4,                              /* number of times to retransmit */
        RES_DEFAULT,                    /* options flags */
        NULL,                           /* number of name servers */
};

/*
 * Set up default settings.  If the configuration file exist, the values
 * there will have precedence.  Otherwise, the server address is set to
 * INADDR_ANY and the default domain name comes from the gethostname().
 *
 * The configuration file should only be used if you want to redefine your
 * domain or run without a server on your machine.
 *
 * Return 0 if completes successfully, -1 on error
 */
int res_init(void)
{
        register FILE *fp;
        register char *cp, **pp;
        register int n;
        char buf[BUFSIZ];
        int nserv = 0;    /* number of nameserver records read from file */
        int haveenv = 0;
        int havesearch = 0;
#if 1
        struct DhcpInfo *DhcpInfo;
        struct ifreq _ifr, *rp = &_ifr;
        char *name = "*0";
        SOCKET Skt;

        strncpy(rp->ifr_name, name, sizeof(rp->ifr_name));

        /* Get info from DHCP record */
        Skt = socket(AF_INET, SOCK_DGRAM, 0);
        if (Skt == -1) {
            printf("socket() failed in res_init!\n");
            return -1;
        }
        n = ioctlsocket(Skt, SIOCGDHCPINFOPTR, (PTR) rp);
        if (n < 0) {
            printf("ioctlsocket(SIOCGDHCPPTR) failed in res_init!\n");
            return -1;
        }
        DhcpInfo = *(PTR*) rp;

        /* We take a ref to the internal thing here, ok, since never deleted */
        _res.nsaddr_list = DhcpInfo->nsaddr_list;
        _res.pNscount = &DhcpInfo->nscount;
        assert(*_res.pNscount >= 1);
        assert(*_res.pNscount <= MAXNS);

        /* Get the domain from DHCP if available */
        if (DhcpInfo->DomainName) {
            strncpy(_res.defdname, DhcpInfo->DomainName, MAXDNAME);
            _res.defdname[MAXDNAME-1] = '\0';
        }

#else
#if 1 /* UPnP */
        /* Start with a multicast address. XXX should we do loopback first? */
#ifdef notyet                   /* XXX XXX XXX */
        _res.nsaddr.sin_addr.s_addr = inet_addr(NAMESERVER_MULTICAST_ADDRESS);
#else
        _res.nsaddr.sin_addr.s_addr = inet_addr("10.10.10.1");
#endif
#elif defined(USELOOPBACK)
        _res.nsaddr.sin_addr = inet_makeaddr(IN_LOOPBACKNET, 1);
#else
        _res.nsaddr.sin_addr.s_addr = INADDR_ANY;
#endif
        _res.nsaddr.sin_family = AF_INET;
        _res.nsaddr.sin_port = htons(NAMESERVER_PORT);
        _res.nscount = 1;
        _res.pfcode = 0;
#endif

#if 0
        /* Allow user to override the local domain definition */
        if ((cp = getenv("LOCALDOMAIN")) != NULL) {
                (void)strncpy(_res.defdname, cp, sizeof(_res.defdname));
                if ((cp = strpbrk(_res.defdname, " \t\n")) != NULL)
                        *cp = '\0';
                haveenv++;
        }
#endif

#ifndef h8
        if ((fp = fopen(_PATH_RESCONF, "r")) != NULL) {
            /* read the config file */
            while (fgets(buf, sizeof(buf), fp) != NULL) {
                /* skip comments */
                if ((*buf == ';') || (*buf == '#'))
                        continue;
                /* read default domain name */
                if (!strncmp(buf, "domain", sizeof("domain") - 1)) {
                    if (haveenv)        /* skip if have from environ */
                            continue;
                    cp = buf + sizeof("domain") - 1;
                    while (*cp == ' ' || *cp == '\t')
                            cp++;
                    if ((*cp == '\0') || (*cp == '\n'))
                            continue;
                    (void)strncpy(_res.defdname, cp,
                                  sizeof(_res.defdname) - 1);
                    if ((cp = strpbrk(_res.defdname, " \t\n")) != NULL)
                            *cp = '\0';
                    havesearch = 0;
                    continue;
                }
                /* set search list */
                if (!strncmp(buf, "search", sizeof("search") - 1)) {
                    if (haveenv)        /* skip if have from environ */
                            continue;
                    cp = buf + sizeof("search") - 1;
                    while (*cp == ' ' || *cp == '\t')
                            cp++;
                    if ((*cp == '\0') || (*cp == '\n'))
                            continue;
                    (void)strncpy(_res.defdname, cp,
                                  sizeof(_res.defdname) - 1);
                    if ((cp = strchr(_res.defdname, '\n')) != NULL)
                            *cp = '\0';
                    /*
                     * Set search list to be blank-separated strings
                     * on rest of line.
                     */
                    cp = _res.defdname;
                    pp = _res.dnsrch;
                    *pp++ = cp;
                    for (n = 0; *cp && pp < _res.dnsrch + MAXDNSRCH; cp++) {
                            if (*cp == ' ' || *cp == '\t') {
                                    *cp = 0;
                                    n = 1;
                            } else if (n) {
                                    *pp++ = cp;
                                    n = 0;
                            }
                    }
                    /* null terminate last domain if there are excess */
                    while (*cp != '\0' && *cp != ' ' && *cp != '\t')
                            cp++;
                    *cp = '\0';
                    *pp++ = 0;
                    havesearch = 1;
                    continue;
                }
                /* read nameservers to query */
                if (!strncmp(buf, "nameserver", sizeof("nameserver") - 1) &&
                   nserv < MAXNS) {
                   struct in_addr a;

                    cp = buf + sizeof("nameserver") - 1;
                    while (*cp == ' ' || *cp == '\t')
                        cp++;
                    if ((*cp != '\0') && (*cp != '\n') && inet_aton(cp, &a)) {
                        _res.nsaddr_list[nserv].sin_addr = a;
                        _res.nsaddr_list[nserv].sin_family = AF_INET;
                        _res.nsaddr_list[nserv].sin_port =
                                htons(NAMESERVER_PORT);
                        nserv++;
                    }
                    continue;
                }
            }
            if (nserv > 1) 
                *_res.pNscount = nserv;
            (void) fclose(fp);
        }
#endif /* !h8 */

/* If we don't have a domain name by now, then DHCP structs haven't
 * been filled and we haven't found one in a file either. Then it
 * is better to keep it as null because of the following problem.
 * - We are calling gethostname() to get our own name and hopefully
 *   the domain name from that
 * - gethostname() tries DHCP info first. But if that were to work,
 *   then we would have got our domain name above anyway. Otherwise,
 *   it gets its own ip address and tries to get its name from DNS
 *   by calling gethostbyaddr()
 * - But, gethostbyaddr() needs to call res_query() to send a query
 *   and hence calls res_init(), completing a whole loop !!
 */
/*      if (_res.defdname[0] == 0) {
            if (gethostname(buf, sizeof(_res.defdname)) == 0) {
                cp = strchr(buf, '.');
                if (cp)
                    (void)strcpy(_res.defdname, cp + 1);
            }
        }
*/
        /* find components of local domain that might be searched */
        if (havesearch == 0) {
                pp = _res.dnsrch;
                *pp++ = _res.defdname;
                for (cp = _res.defdname, n = 0; *cp; cp++)
                        if (*cp == '.')
                                n++;
                cp = _res.defdname;
                for (; n >= LOCALDOMAINPARTS && pp < _res.dnsrch + MAXDFLSRCH;
                    n--) {
                        cp = strchr(cp, '.');
                        *pp++ = ++cp;
                }
                *pp++ = 0;
        }
        _res.options |= RES_INIT;
        closesocket(Skt);
        return (0);
}
