/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*-
 * Copyright (c) 1985, 1989, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * -
 * Portions Copyright (c) 1993 by Digital Equipment Corporation.
 * 
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies, and that
 * the name of Digital Equipment Corporation not be used in advertising or
 * publicity pertaining to distribution of the document or software without
 * specific, written prior permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * -
 * --Copyright--
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)res_send.c  8.1 (Berkeley) 6/4/93";
static char rcsid[] = "$Id: res_send.c,v 4.9.1.1 1993/05/02 22:43:03 vixie Rel $";
#endif /* LIBC_SCCS and not lint */

/*
 * Send query to name server and wait for reply.
 */

#include "res.h"

#include <socket.h>
#include <in.h>
#include <nameser.h>
#include <inet.h>
#include <stdio.h>
#include <resolv.h>
#include <string.h>

static int s = INVALID_SOCKET;  /* socket used for communications */
static struct sockaddr no_addr;

int res_send(
        const char *buf,
        int buflen,
        char *answer,
        int anslen)
{
        register int n;
        int try, v_circuit, resplen;
        UINT ns;
        int gotsomewhere = 0, connected = 0;
        u_short id;
        fd_set dsmask;
        struct timeval timeout;
        HEADER *hp = (HEADER *) buf;
        HEADER *anhp = (HEADER *) answer;
        u_int badns;            /* XXX NSMAX can't exceed #/bits per this */
        int terrno = ETIMEDOUT;

        /* Added for multicast */
        int is_multicast_addr = 0;
        struct sockaddr_in cliAddr;
        unsigned char ttl = 3;
        /* End multicast additions */

#ifdef DEBUG
        if ((_res.options & RES_DEBUG) || (_res.pfcode & RES_PRF_QUERY)) {
                printf(";; res_send()\n");
                __p_query(buf);
        }
#endif
        if (!(_res.options & RES_INIT))
                if (res_init() == -1) {
                        return(-1);
                }
        v_circuit = (_res.options & RES_USEVC) || buflen > PACKETSZ;
        id = hp->id;
        badns = 0;
        /*
         * Send request, RETRY times, or until successful
         */
        for (try = 0; try < _res.retry; try++) {
            for (ns = 0; ns < *_res.pNscount; ns++) {

                is_multicast_addr = IN_MULTICAST(ntohl(_res.nsaddr_list[ns].sin_addr.s_addr));
                      
                if (badns & (1<<ns))
                        continue;
#ifdef DEBUG
                if (_res.options & RES_DEBUG)
                        printf(";; Querying server (# %d) address = %s\n",
                               ns+1,
                               inet_ntoa(_res.nsaddr_list[ns].sin_addr));
#endif
                {
                        /*
                         * Use datagrams.
                         */
                        if (s == INVALID_SOCKET) {
                                s = socket(AF_INET, SOCK_DGRAM, 0);
                                if (s == INVALID_SOCKET) {
                                        terrno = errno;
#ifdef DEBUG
                                        if (_res.options & RES_DEBUG)
                                            perror("socket (dg) failed");
#endif
                                        continue;
                                }

                                /* New code added for multicast DNS query */
                                if(is_multicast_addr) {

                                    // Set the outgoing interface to DEFAULT
/*                                    if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF, &_res.nsaddr_list[ns], 
                                        sizeof(struct in_addr))<0) {
                                        printf("res_send.c : multicast setsockopt error\n");
                                        continue;
                                    }
*/    
                                    // Set multicast packet TTL to 3; default TTL is 1
                                    if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL, (void *)&ttl, sizeof(ttl))<0) {
                                        printf("res_send.c : cannot set ttl = %d for multicast\n",ttl);
                                        continue;
                                    }
                                    /* bind any port number */
                                    cliAddr.sin_family = AF_INET;
                                    cliAddr.sin_addr.s_addr = htonl(INADDR_ANY);
                                    cliAddr.sin_port = htons(0);
                                    if(bind(s,(struct sockaddr *) &cliAddr,sizeof(cliAddr))<0) {
                                        printf("bind");
                                        continue;
                                    }
                                }
                                /* New code ends */
                        }
                        /*
                         * I'm tired of answering this question, so:
                         * On a 4.3BSD+ machine (client and server,
                         * actually), sending to a nameserver datagram
                         * port with no nameserver will cause an
                         * ICMP port unreachable message to be returned.
                         * If our datagram socket is "connected" to the
                         * server, we get an ECONNREFUSED error on the next
                         * socket operation, and select returns if the
                         * error message is received.  We can thus detect
                         * the absence of a nameserver without timing out.
                         * If we have sent queries to at least two servers,
                         * however, we don't want to remain connected,
                         * as we wish to receive answers from the first
                         * server to respond.
                         */
                        if ((*_res.pNscount == 1 || (try == 0 && ns == 0)) &&
                            (!is_multicast_addr)) /* For multicast never use connect */
                        {
    
                                /*
                                 * Don't use connect if we might
                                 * still receive a response
                                 * from another server.
                                 */
                                if (connected == 0) {
                                        if (connect(s,
                                            (struct sockaddr *)
                                            &_res.nsaddr_list[ns],
                                            sizeof(struct sockaddr)) < 0) {
#ifdef DEBUG
                                                if (_res.options & RES_DEBUG)
                                                        perror("connect");
#endif
                                                continue;
                                        }
                                        connected = 1;
                                }
                                if (send(s, buf, buflen, 0) != buflen) {
#ifdef DEBUG
                                    if (_res.options & RES_DEBUG)
                                            perror("send");
#endif  
                                    continue;
                                }
                        } else {
                                /*
                                 * Disconnect if we want to listen
                                 * for responses from more than one server.
                                 */
                                if (connected) {
                                        (void) connect(s, &no_addr,
                                            sizeof(no_addr));
                                        connected = 0;
                                }
                                if (sendto(s, buf, buflen, 0,
                                    (struct sockaddr *)&_res.nsaddr_list[ns],
                                    sizeof(struct sockaddr)) != buflen) {
#ifdef DEBUG
                                        if (_res.options & RES_DEBUG)
                                                perror("sendto");
#endif
                                        continue;
                                }
                        }

                        /*
                         * Wait for reply
                         */
                        timeout.tv_sec = _res.retrans << try;
                        if (try > 0)
                                timeout.tv_sec /= *_res.pNscount;
                        if (timeout.tv_sec <= 0)
                                timeout.tv_sec = 1;
                        timeout.tv_usec = 0;
wait:
                        FD_ZERO(&dsmask);
                        FD_SET(s, &dsmask);
                        n = select(1, &dsmask, (fd_set *)NULL,
                                (fd_set *)NULL, &timeout);
                        if (n < 0) {
#ifdef DEBUG
                                if (_res.options & RES_DEBUG)
                                        perror("select");
#endif
                                continue;
                        }
                        if (n == 0) {
                                /*
                                 * timeout
                                 */
#ifdef DEBUG
                                if (_res.options & RES_DEBUG)
                                        printf(";; timeout\n");
#endif
                                gotsomewhere = 1;
                                continue;
                        }
                        if ((resplen = recv(s, answer, anslen, 0)) <= 0) {
#ifdef DEBUG
                                if (_res.options & RES_DEBUG)
                                        perror("recvfrom");
#endif
                                continue;
                        }
                        gotsomewhere = 1;
                        if (id != anhp->id) {
                                /*
                                 * response from old query, ignore it
                                 */
#ifdef DEBUG
                                if ((_res.options & RES_DEBUG) ||
                                    (_res.pfcode & RES_PRF_REPLY)) {
                                        printf(";; old answer:\n");
                                        __p_query(answer);
                                }
#endif
                                goto wait;
                        }
                        if (anhp->rcode == SERVFAIL || anhp->rcode == NOTIMP ||
                            anhp->rcode == REFUSED) {
#ifdef DEBUG
                                if (_res.options & RES_DEBUG) {
                                        printf("server rejected query:\n");
                                        __p_query(answer);
                                }
#endif
                                badns |= (1<<ns);
                                continue;
                        }
#if 0
                        if (!(_res.options & RES_IGNTC) && anhp->tc) {
                                /*
                                 * get rest of answer;
                                 * use TCP with same server.
                                 */
#ifdef DEBUG
                                if (_res.options & RES_DEBUG)
                                        printf(";; truncated answer\n");
#endif
                                (void) closesocket(s);
                                s = INVALID_SOCKET;
                                v_circuit = 1;
                                goto usevc;
                        }
#endif
                }
#ifdef DEBUG
                if (_res.options & RES_DEBUG)
                        printf(";; got answer:\n");
                if ((_res.options & RES_DEBUG) ||
                    (_res.pfcode & RES_PRF_REPLY))
                        __p_query(answer);
#endif
                /*
                 * If using virtual circuits, we assume that the first server
                 * is preferred * over the rest (i.e. it is on the local
                 * machine) and only keep that one open.
                 * If we have temporarily opened a virtual circuit,
                 * or if we haven't been asked to keep a socket open,
                 * close the socket.
                 */
                if ((v_circuit &&
                    ((_res.options & RES_USEVC) == 0 || ns != 0)) ||
                    (_res.options & RES_STAYOPEN) == 0) {
                        (void) closesocket(s);
                        s = INVALID_SOCKET;
                }
                return (resplen);
           }
        }
        if (s >= 0) {
                (void) closesocket(s);
                s = INVALID_SOCKET;
        }
        if (v_circuit == 0) {
                if (gotsomewhere == 0)
                        errno = ECONNREFUSED;   /* no nameservers found */
                else
                        errno = ETIMEDOUT;      /* no answer obtained */
        } else {
                errno = terrno;
        }
        return (-1);
}

/*
 * This routine is for closing the socket if a virtual circuit is used and
 * the program wants to close it.  This provides support for endhostent()
 * which expects to close the socket.
 *
 * This routine is not expected to be user visible.
 */
void _res_close()
{
        if (s != INVALID_SOCKET) {
                (void) closesocket(s);
                s = INVALID_SOCKET;
        }
}
