/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>

#if !defined(DebugBreak)
#define DebugBreak() __asm { __asm int 3 __asm }
#endif
#define __Assert_print(_m_, _f_, _l_) \
     printf("Assertion `%s' failed at %s:%u\n", _m_, _f_, _l_)
#define assert(_x_) \
     { if (!(_x_)) {__Assert_print(#_x_, __FILE__, __LINE__);DebugBreak(); }}


typedef unsigned char BYTE;
typedef unsigned short UINT16;
#define u_short UINT16
typedef int INT32;
typedef unsigned int UINT32;
typedef unsigned int UINT;
typedef int BOOL;
#define TRUE 1
#define FALSE 0

/* pseudo data structs, macros, code
 */
struct mbuf {
  struct mbuf *m_next;
  struct mbuf *m_nextpkt;
  BYTE m_data[100];
};
#define mtod(m,t) ((t)((m)->m_data))

struct mbuf *
m_get(void)
{
    struct mbuf *m = calloc(1,sizeof *m);
    printf(" mget()->%x\n",m);
    return m;
}

struct mbuf *
m_free(struct mbuf *m)
{
    struct mbuf *n = m->m_next;
    printf(" m_free(%x)\n",m);
    free(m);
    return n;    
}

void m_freem(struct mbuf *m)
{
    for (;m;m = m_free(m))
        ;
}

void m_cat(struct mbuf *m, struct mbuf *n)
{
    while (m->m_next)
        m = m->m_next;
    m->m_next = n;
}

#define m_adj(m,d) {printf("Would shrink %x by %d\n",m,d);}

struct ip {
    UINT ip_off;
    UINT ip_len;
    UINT ip_tos;
};

struct ipq {
    struct mbuf *ipq_first;
    struct mbuf *ipq_last;
};
#define MYQUEUE struct ipq
struct ipq *ipq = NULL; /* not really a Q */
#undef insque
#define insque(f,q) {ipq = f;printf("Inserting fragQ @%x\n",f);}
#undef remque
#define remque(f) {printf("Freeing fragQ @%x\n",f);}

#define bsd_malloc(s,x,y) malloc(s)
#define bsd_free(p,x) free(p)

void ip_freef(struct ipq *fp)
{
    struct mbuf *q, *p;

    for (q = fp->ipq_first; q != NULL; q = p) {
        p = q->m_nextpkt;
        m_freem(q);
    }

    remque((MYQUEUE*)fp);
    (void) bsd_free(fp, M_FTABLE);
}

/* test scaffolding
 */
typedef enum {
    aComplete, aDrop, aNew, aInsertMore, aInsertMissing
} ACTION;

ACTION ip_ass(struct mbuf *m, struct ipq *fp);


typedef struct TEST *PTEST;
struct TEST {
    BOOL NewFrag;
    BOOL LastFrag;
    UINT Offset;
    UINT Length;
    ACTION ExpectedAction;
    char *Description;
} AllTests[] = {
    /* Seq */
     { TRUE,  FALSE, 1500*0, 1500, aNew, "sequential 1-2" }
    ,{ FALSE, TRUE,  1500*1, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "sequential 1-2-3" }
    ,{ FALSE, FALSE, 1500*1, 1500, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*2, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "sequential 1-2..5" }
    ,{ FALSE, FALSE, 1500*1, 1500, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*2, 1500, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*3, 1500, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*4, 1500, aComplete,  }


    /* swaps */
    ,{ TRUE,  TRUE,  1500*1, 1500, aNew, "swap 2-1" }
    ,{ FALSE, FALSE, 1500*0, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*1, 1500, aNew, "swap 2-1-3"  }
    ,{ FALSE, FALSE, 1500*0, 1500, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*2, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "swap 1-3-2" }
    ,{ FALSE, TRUE,  1500*2, 1500, aInsertMissing,  }
    ,{ FALSE, FALSE, 1500*1, 1500, aComplete,  }

    ,{ TRUE,  TRUE,  1500*2, 1500, aNew, "swap 3-2-1" }
    ,{ FALSE, FALSE, 1500*1, 1500, aInsertMissing,  }
    ,{ FALSE, FALSE, 1500*0, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*1, 1500, aNew, "swap 2-3-1-5-4" }
    ,{ FALSE, FALSE, 1500*2, 1500, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*0, 1500, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*4, 1500, aInsertMissing,  }
    ,{ FALSE, FALSE, 1500*3, 1500, aComplete,  }


    /* dropped packets */
    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "Simple drop 1-x-3" }
    ,{ FALSE, TRUE,  1500*2, 1500, aInsertMissing,  }

    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "Double drop 1-x-3-x-5" }
    ,{ FALSE, FALSE, 1500*2, 1500, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*4, 1500, aInsertMissing,  }


    /* overlaps */
    ,{ TRUE,  FALSE, 1500*0, 1500*2, aNew, "1 fully covers 2" }
    ,{ FALSE, FALSE, 1500*1, 1500, aDrop,  }
    ,{ FALSE, TRUE,  1500*2, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500*2, aNew, "1 fully covers 2 (+3)" }
    ,{ FALSE, FALSE, 1500*2, 1500, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*1, 1500, aDrop,  }
    ,{ FALSE, TRUE,  1500*3, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500*2, aNew, "1 partially covers 2" }
    ,{ FALSE, FALSE, 1500*1, 1500*2, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*3, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500, aNew, "2 overlaps 3" }
    ,{ FALSE, FALSE, 1500*2, 1500*2, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*1, 1500*2, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*4, 1500, aComplete,  }

    ,{ TRUE,  FALSE, 1500*0, 1500*2, aNew, "2 overlaps 1 and 3" }
    ,{ FALSE, FALSE, 1500*2, 1500*2, aInsertMore,  }
    ,{ FALSE, FALSE, 1500*1, 1500*2, aInsertMore,  }
    ,{ FALSE, TRUE,  1500*4, 1500, aComplete,  }

};
#define nTESTS (sizeof(AllTests) / sizeof(AllTests[0]))

int main(int argc, char **argv)
{
    PTEST t = AllTests;
    int i;
    ACTION ret;
    struct mbuf *m;
    struct ip *ip;

    for (i = 0; i < nTESTS; i++, t++) {

        if (t->NewFrag && ipq) {
            ip_freef(ipq);
            printf("\n");
            ipq = NULL;
        }

        if (t->Description)
            printf("=== %s ===\n",t->Description);

        m = m_get();
        ip = mtod(m,struct ip*);
        ip->ip_off = t->Offset;
        ip->ip_len = t->Length;
        ip->ip_tos = (t->LastFrag) ? 0 : 1;

        ret = ip_ass(m, ipq);

        if (ret != t->ExpectedAction) {
            printf("Testing failed at step %d, ret = %d != %d\n",
                   i, ret, t->ExpectedAction);
            break;
        }

        if (ret == aComplete)
            printf("Frag %x Complete\n",ipq);

    }
    if (i == nTESTS)
        printf("Testing ok.\n");
    if (ipq)
        ip_freef(ipq);
    return 0;
}

/* ze culprit
 */
ACTION
ip_ass(struct mbuf *m, struct ipq *fp)
{
    struct ip *ip = mtod(m, struct ip*);
    struct ip *q;
    struct mbuf *n, *prev, *nxt;
    INT32 i;
    UINT32 next_offset;

    /* Terminate chain
     */
    m->m_nextpkt = NULL;

    /*
     * If first fragment to arrive, create a reassembly queue.
     */
    if (fp == 0) {
        if ((fp = bsd_malloc(sizeof(*fp), M_FTABLE, M_NOWAIT)) == NULL)
            goto dropfrag;
#if 0
        fp->ipq_ttl = IPFRAGTTL;
        fp->ipq_p = ip->ip_p;
        fp->ipq_id = ip->ip_id;
        fp->ipq_src = ((struct ip *)ip)->ip_src;
        fp->ipq_dst = ((struct ip *)ip)->ip_dst;
#endif
        insque((MYQUEUE*)fp, (MYQUEUE*)&ipq);
        fp->ipq_first = fp->ipq_last = m;
        /* recheck: should we check we are done ? */
        return aNew;
    }

    assert(fp->ipq_first != NULL);
    assert(fp->ipq_last != NULL);

    /* A new fragment for this queue.
     */
    nxt = NULL;
    prev = fp->ipq_last;
    q = mtod(prev,struct ip *);
    if ((q->ip_off + q->ip_len > ip->ip_off)) {

        /* Find the first segment 'nxt' that starts after this one.
         * Also, the last segment 'prev' which starts not-after this one.
         * If they exist.
         */
        prev = nxt = NULL;
        n = fp->ipq_first;
        for (;n != NULL; ) {
            q = mtod(n,struct ip*);
            if (q->ip_off > ip->ip_off) {
                nxt = n;
                break;
            }
            prev = n;
            n = n->m_nextpkt;
        }

        /* If there is a preceding segment, it may provide some of
         * our data already.  If so, drop the excess data from that
         * segment.  If it provides all of our data, drop us.
         */
        if (prev != NULL) {
            q = mtod(prev,struct ip*);
            i = (INT32)(q->ip_off + q->ip_len)
                 - ip->ip_off;
            if (i > 0) {
                if (i >= ip->ip_len)
                    goto dropfrag;
                m_adj(prev,-i);
                q->ip_len -= (UINT16) i;
            }
            /* There is a prev, ipq_first wont change
             */
            prev->m_nextpkt = m;
        } else {
            /* There is no prev, so we are first (and we cant be last) */
            fp->ipq_first = m;
        }

        /* While we overlap succeeding segments trim us or,
         * if they are completely covered, dequeue them.
         */
        if (nxt != NULL) {
            q = mtod(nxt,struct ip*);
            i = (INT32)(ip->ip_off + ip->ip_len) - q->ip_off;
            while (i > 0) {
                /* partial overlap ? trim & stop
                 */
                if (i < q->ip_len) {
                    m_adj(m,-i);
                    ip->ip_len -= (UINT16) i;
                    break;
                }
                /* full overlap ? drop & continue
                 */
                n = nxt;
                nxt = n->m_nextpkt;
                if (fp->ipq_last == n) fp->ipq_last = nxt;
                m_freem(n);
                if (nxt == NULL)
                    break;
                q = mtod(nxt,struct ip*);
                i = (INT32)(ip->ip_off + ip->ip_len) - q->ip_off;
            }
        }
        m->m_nextpkt = nxt;
        if (nxt == NULL) {
            /* there is no next, so we are last */
            fp->ipq_last = m;
        }

    } else {
        /* common case: next frag in order */
        /* also, the one after a missed one */
        prev->m_nextpkt = m;
    }

    /*
     * Stick new segment in its place;
     * check for complete reassembly.
     * 'prev' points to the frag that preceeds m, if any.
     */
    if (fp->ipq_last == prev)
        fp->ipq_last = m;

    /* Have we got all the fragments yet?
     */
    q = mtod(fp->ipq_last,struct ip *);
    if (q->ip_tos & 1)
        return aInsertMore;
    next_offset = 0;
    for (n = fp->ipq_first; n != NULL; n = n->m_nextpkt) {
        q = mtod(n,struct ip *);
        if (q->ip_off != next_offset)
            return (aInsertMissing);
        next_offset += q->ip_len;
    }

    /* Reassembly is complete; concatenate mbuf chains into one.
     */
    m = fp->ipq_first;
    n = m->m_next;
    m->m_next = 0;
    m_cat(m, n); /* a close friend of m_pullup */
    for (n = m->m_nextpkt; n != NULL; n = n->m_nextpkt) {
#if 0
        /* drop ip header now */
        q = mtod(n,struct ip *);
        i = q->ip_hl << 2;
        n->m_data += i;
        n->m_len -= i;
#endif
        m_cat(m, n);
    }
    m->m_nextpkt = NULL;

    /* Create header for new ip packet by
     * modifying header of first packet;
     * dequeue and discard fragment reassembly header.
     */
    ip = mtod(m,struct ip *);
    ip->ip_len = (u_short)next_offset;
#if 0
    ip->ip_tos &= ~1;
    ip->ip_src = fp->ipq_src;
    ip->ip_dst = fp->ipq_dst;

    remque((MYQUEUE*)fp);
    (void) bsd_free(fp, M_FTABLE);

    m->m_flags |= M_PKTHDR;
    m->m_pkthdr.len = next_offset + (ip->ip_hl << 2);
#endif
    return aComplete;

dropfrag:
#if 0
    ipstat.ips_fragdropped++;
#endif
    m_freem(m);
    return (aDrop);
}

