/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1991, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)in_pcb.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <in_var.h>
#include <ip_var.h>

#include <assert.h>

INT in_pcbbindNEW(struct inpcb *inp, struct sockaddr_in *Name);
INT in_pcbconnectNEW(struct inpcb *inp, struct sockaddr_in *Name);
void in_setpeeraddrNEW(struct inpcb *inp, struct sockaddr_in *Name);
void in_setsockaddrNEW(struct inpcb *inp, struct sockaddr_in *Name);

struct  in_addr zeroin_addr;

INT
in_pcballoc(struct socket *so, struct inpcb *head)
{
        struct inpcb *inp;

        MALLOC(inp, struct inpcb *, sizeof(*inp), M_PCB, M_WAITOK);
        if (inp == NULL)
                return (ENOBUFS);
        bzero((caddr_t)inp, sizeof(*inp));
        inp->inp_head = head;
        inp->inp_socket = so;
        insque((MYQUEUE*)inp, (MYQUEUE*)head);
        so->so_pcb = (caddr_t)inp;
        return (0);
}

INT
in_pcbbind(struct inpcb *inp, struct mbuf *nam)
{
    struct sockaddr_in *sin;

    if (nam)
        sin = mtod(nam, struct sockaddr_in *);
    else
        sin = NULL;

    return in_pcbbindNEW(inp, sin);
}

INT
in_pcbbindNEW(struct inpcb *inp, struct sockaddr_in *sin)
{
        struct socket *so = inp->inp_socket;
        struct inpcb *head = inp->inp_head;
        u_short lport = 0;
        int wild = 0, reuseport = (so->so_options & SO_REUSEPORT);

        if (_in_ifaddr == 0)
                return (EADDRNOTAVAIL);
        if (inp->inp_lport || inp->inp_laddr.s_addr != INADDR_ANY)
                return (EINVAL);
        if ((so->so_options & (SO_REUSEADDR|SO_REUSEPORT)) == 0 &&
            ((so->so_proto->pr_flags & PR_CONNREQUIRED) == 0 ||
             (so->so_options & SO_ACCEPTCONN) == 0))
                wild = INPLOOKUP_WILDCARD;
        if (sin) {
#ifdef notdef
                /*
                 * We should check the family, but old programs
                 * incorrectly fail to initialize it.
                 */
                if (sin->sin_family != AF_INET)
                        return (EAFNOSUPPORT);
#endif
                lport = sin->sin_port;
                if (IN_MULTICAST(ntohl(sin->sin_addr.s_addr))) {
                        /*
                         * Treat SO_REUSEADDR as SO_REUSEPORT for multicast;
                         * allow complete duplication of binding if
                         * SO_REUSEPORT is set, or if SO_REUSEADDR is set
                         * and a multicast address is bound on both
                         * new and duplicated sockets.
                         */
                        if (so->so_options & SO_REUSEADDR)
                                reuseport = SO_REUSEADDR|SO_REUSEPORT;
                } else if (sin->sin_addr.s_addr != INADDR_ANY) {
                        sin->sin_port = 0;              /* yech... */
                        if (ifa_ifwithaddr((struct sockaddr *)sin) == 0)
                                return (EADDRNOTAVAIL);
                }
                if (lport) {
                        struct inpcb *t;
#ifdef SECURITY
                        struct proc *p = curproc;               /* XXX */
                        int error;

                        /* GROSS */
                        if (ntohs(lport) < IPPORT_RESERVED &&
                            (error = suser(p->p_ucred, &p->p_acflag)))
                                return (error);
#endif
                        t = in_pcblookup(head, zeroin_addr, 0,
                            sin->sin_addr, lport, wild);
                        if (t && (reuseport & t->inp_socket->so_options) == 0)
                                return (EADDRINUSE);
                }
                inp->inp_laddr = sin->sin_addr;
        }
        if (lport == 0)
                do {
                        if (head->inp_lport++ < IPPORT_RESERVED ||
                            head->inp_lport > IPPORT_USERRESERVED)
                                head->inp_lport = IPPORT_RESERVED;
                        lport = htons(head->inp_lport);
                } while (in_pcblookup(head,
                            zeroin_addr, 0, inp->inp_laddr, lport, wild));
        inp->inp_lport = lport;
        return (0);
}

/*
 * Connect from a socket to a specified address.
 * Both address and port must be specified in argument sin.
 * If don't have a local address for this socket yet,
 * then pick one.
 */
INT
in_pcbconnect(struct inpcb *inp, struct mbuf *nam)
{
    if (nam->m_len != sizeof (struct sockaddr_in))
        return (EINVAL);
    return in_pcbconnectNEW(inp, mtod(nam, struct sockaddr_in *));
}

INT
in_pcbconnectNEW(struct inpcb *inp, struct sockaddr_in *sin)
{
        struct in_ifaddr *ia;
        struct sockaddr_in *ifaddr = NULL;
        struct route_in *rti = &inp->inp_route;

        if (sin->sin_family != AF_INET)
                return (EAFNOSUPPORT);
        if (sin->sin_port == 0)
                return (EADDRNOTAVAIL);
        if (_in_ifaddr) {
                /*
                 * If the destination address is INADDR_ANY,
                 * use the primary local address.
                 * If the supplied address is INADDR_BROADCAST,
                 * and the primary interface supports broadcast,
                 * choose the broadcast address for that interface.
                 */
#define satosin(sa)     ((struct sockaddr_in *)(sa))
#define sintosa(sin)    ((struct sockaddr *)(sin))
#define ifatoia(ifa)    ((struct in_ifaddr *)(ifa))
                if (sin->sin_addr.s_addr == INADDR_ANY)
                    sin->sin_addr = IA_SIN(_in_ifaddr)->sin_addr;
                else if (sin->sin_addr.s_addr == (u_long)INADDR_BROADCAST &&
                  (_in_ifaddr->ia_ifp->if_flags & IFF_BROADCAST))
                    sin->sin_addr = satosin(&_in_ifaddr->ia_broadaddr)->sin_addr;
        }

        /*
         * Route if we don't have an existing route already.
         * This differs from the old Berkeley code in the sence that
         * we should always have a known route after this point instead
         * of deferring the routing lookup to ip_output() in some
         * cases.
         *
         * XXX: this should honor SO_DONTROUTE, but since Windows TCP/IP
         * doesn't honor it anyways, I'll just claim we're compatible
         * with Windows on that front.
         */
        if (rti->rt_valid == 0) {
                if (rt_ifaout(sin->sin_addr, &rti->rt_ifa, &rti->rt_gw))
                        return EHOSTUNREACH;

                /*
                 * If we found a route, use the address
                 * corresponding to the outgoing interface
                 * unless it is the loopback (in case a route
                 * to our address on another net goes to loopback) ...
                 */
                ia = NULL;
                if (!(rti->rt_ifa->ifa_ifp->if_flags & IFF_LOOPBACK))
                        ia = ifatoia(rti->rt_ifa);

                if (ia == 0) {
                        u_short fport = sin->sin_port;

                        sin->sin_port = 0;
                        ia = ifatoia(ifa_ifwithdstaddr(sintosa(sin)));
                        if (ia == 0)
                                ia = ifatoia(ifa_ifwithnet(sintosa(sin)));
                        sin->sin_port = fport;
                        if (ia == 0)
                                ia = _in_ifaddr;
                        if (ia == 0)
                                return (EADDRNOTAVAIL);
                }

                /*
                 * ... but check if routing matches the interface the
                 * socket was bound to, INADDR_ANY and multicast act
                 * as an allow-all.. XXX: I don't like special-casing
                 */
                if (inp->inp_laddr.s_addr != INADDR_ANY
                    && !IN_MULTICAST(ntohl(inp->inp_laddr.s_addr))) {
                        if (ia->ia_addr.sin_addr.s_addr
                            != inp->inp_laddr.s_addr)
                                return EHOSTUNREACH;
                }

                /*
                 * If the destination address is multicast and an outgoing
                 * interface has been set as a multicast option, use the
                 * address of that interface as our source address.
                 */
                if (IN_MULTICAST(ntohl(sin->sin_addr.s_addr)) &&
                    inp->inp_moptions != NULL) {
                        struct ip_moptions *imo;
                        struct ifnet *ifp;

                        imo = inp->inp_moptions;
                        if (imo->imo_multicast_ifp != NULL) {
                                ifp = imo->imo_multicast_ifp;
                                for (ia = _in_ifaddr; ia; ia = ia->ia_next)
                                        if (ia->ia_ifp == ifp)
                                                break;
                                if (ia == 0)
                                        return (EADDRNOTAVAIL);
                        }
                }
                ifaddr = (struct sockaddr_in *)&ia->ia_addr;
                rti->rt_valid = 1;
        } else {
                ifaddr = (struct sockaddr_in *)ifatoia(rti->rt_ifa);
        }

        if (in_pcblookup(inp->inp_head,
                         sin->sin_addr,
                         sin->sin_port,
                         inp->inp_laddr.s_addr ? inp->inp_laddr : ifaddr->sin_addr,
                         inp->inp_lport,
                         0))
                return (EADDRINUSE);
        if (inp->inp_laddr.s_addr == INADDR_ANY) {
                if (inp->inp_lport == 0)
                        (void)in_pcbbind(inp, (struct mbuf *)0);
                inp->inp_laddr = ifaddr->sin_addr;
        }
        inp->inp_faddr = sin->sin_addr;
        inp->inp_fport = sin->sin_port;

        assert(rti->rt_valid);

        return (0);
}

void
in_pcbdisconnect(struct inpcb *inp)
{
        inp->inp_faddr.s_addr = INADDR_ANY;
        inp->inp_fport = 0;
        if (inp->inp_socket->so_state & SS_NOFDREF)
                in_pcbdetach(inp);

        /*
         * XXX: we could leave this lingering, but then route_in would
         * need the target address.. which wouldn't be all that difficult,
         * but currently it just isn't in there
         */
        inp->inp_route.rt_valid = 0;
}

void
in_pcbdetach(struct inpcb *inp)
{
        struct socket *so = inp->inp_socket;

        so->so_pcb = 0;
        sofree(so);
        if (inp->inp_options)
                (void)m_free(inp->inp_options);
        ip_freemoptions(inp->inp_moptions);
        remque((MYQUEUE*)inp);
        FREE(inp, M_PCB);
}

void
in_setsockaddr(struct inpcb *inp, struct mbuf *nam)
{
    nam->m_len = sizeof (struct sockaddr_in);
    in_setsockaddrNEW(inp, mtod(nam, struct sockaddr_in *));
}

void
in_setsockaddrNEW(struct inpcb *inp, struct sockaddr_in *sin)
{
        bzero((caddr_t)sin, sizeof (*sin));
        sin->sin_family = AF_INET;
        sin->sin_len = sizeof(*sin);
        sin->sin_port = inp->inp_lport;
        sin->sin_addr = inp->inp_laddr;
}

void
in_setpeeraddr(struct inpcb *inp, struct mbuf *nam)
{
    nam->m_len = sizeof (struct sockaddr_in);
    in_setpeeraddrNEW(inp, mtod(nam, struct sockaddr_in *));
}

void
in_setpeeraddrNEW(struct inpcb *inp, struct sockaddr_in *sin)
{
        bzero((caddr_t)sin, sizeof (*sin));
        sin->sin_family = AF_INET;
        sin->sin_len = sizeof(*sin);
        sin->sin_port = inp->inp_fport;
        sin->sin_addr = inp->inp_faddr;
}

/*
 * Pass some notification to all connections of a protocol
 * associated with address dst.  The local address and/or port numbers
 * may be specified to limit the search.  The "usual action" will be
 * taken, depending on the ctlinput cmd.  The caller must filter any
 * cmds that are uninteresting (e.g., no error in the map).
 * Call the protocol specific routine (if any) to report
 * any errors for each matching socket.
 *
 * Must be called at splnet.
 */
void
in_pcbnotify(
        struct inpcb *head,
        struct sockaddr *dst,
        u_int fport_arg, 
        struct in_addr laddr,
        u_int lport_arg,
        INT cmd,
        void (*notify) (struct inpcb *, INT) )
{
        extern u_char inetctlerrmap[];
        struct inpcb *inp, *oinp;
        struct in_addr faddr;
        u_short fport = (u_short)fport_arg, lport = (u_short)lport_arg;
        int errno;

        if ((unsigned)cmd > PRC_NCMDS || dst->sa_family != AF_INET)
                return;
        faddr = ((struct sockaddr_in *)dst)->sin_addr;
        if (faddr.s_addr == INADDR_ANY)
                return;

        /*
         * Redirects go to all references to the destination,
         * and use in_rtchange to invalidate the route cache.
         * Dead host indications: notify all references to the destination.
         * Otherwise, if we have knowledge of the local port and address,
         * deliver only to that socket.
         */
        if (PRC_IS_REDIRECT(cmd) || cmd == PRC_HOSTDEAD) {
                fport = 0;
                lport = 0;
                laddr.s_addr = 0;
                if (cmd != PRC_HOSTDEAD)
                        notify = in_rtchange;
        }
        errno = inetctlerrmap[cmd];
        for (inp = head->inp_next; inp != head;) {
                if (inp->inp_faddr.s_addr != faddr.s_addr ||
                    inp->inp_socket == 0 ||
                    (lport && inp->inp_lport != lport) ||
                    (laddr.s_addr && inp->inp_laddr.s_addr != laddr.s_addr) ||
                    (fport && inp->inp_fport != fport)) {
                        inp = inp->inp_next;
                        continue;
                }
                oinp = inp;
                inp = inp->inp_next;
                if (notify)
                        (*notify)(oinp, errno);
        }
}

/*
 * Check for alternatives when higher level complains
 * about service problems.  For now, invalidate cached
 * routing information.
 */
void
in_losing(struct inpcb *inp)
{

        inp->inp_route.rt_valid = 0;
}

/*
 * After a routing change, flush old routing
 * and allocate a (hopefully) better one.
 */
void
in_rtchange(struct inpcb *inp, INT errno)
{

        inp->inp_route.rt_valid = 0;
}

struct inpcb *
in_pcblookup(
        struct inpcb *head,
        struct in_addr faddr,
        u_int fport_arg,
        struct in_addr laddr,
        u_int lport_arg,
        INT flags)
{
        struct inpcb *inp, *match = 0;
        int matchwild = 3, wildcard;
        u_short fport = (u_short)fport_arg, lport = (u_short)lport_arg;

        for (inp = head->inp_next; inp != head; inp = inp->inp_next) {
                if (inp->inp_lport != lport)
                        continue;
                wildcard = 0;
                if (inp->inp_laddr.s_addr != INADDR_ANY) {
                        if (laddr.s_addr == INADDR_ANY)
                                wildcard++;
                        else if (inp->inp_laddr.s_addr != laddr.s_addr)
                                continue;
                } else {
                        if (laddr.s_addr != INADDR_ANY)
                                wildcard++;
                }
                if (inp->inp_faddr.s_addr != INADDR_ANY) {
                        if (faddr.s_addr == INADDR_ANY)
                                wildcard++;
                        else if (inp->inp_faddr.s_addr != faddr.s_addr ||
                            inp->inp_fport != fport)
                                continue;
                } else {
                        if (faddr.s_addr != INADDR_ANY)
                                wildcard++;
                }
                if (wildcard && (flags & INPLOOKUP_WILDCARD) == 0)
                        continue;
                if (wildcard < matchwild) {
                        match = inp;
                        matchwild = wildcard;
                        if (matchwild == 0)
                                break;
                }
        }
        return (match);
}
