/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)in_proto.c  8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <domain.h>

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <ip_var.h>
#include <ip_icmp.h>
#include <in_pcb.h>
#include <igmp_var.h>
#include <tcp.h>
#include <tcp_fsm.h>
#include <tcp_seq.h>
#include <tcp_tim.h>
#include <tcp_var.h>
#include <tcpip.h>
#include <udp.h>
#include <udp_var.h>
#if MONARCH_ADHOC
#include <dsr/ip6_opts.h>
#include <dsr/ip6.h>
#include <dsr/dsr_var.h>
#endif /* MONARCH_ADHOC */

/*
 * TCP/IP protocol family: IP, ICMP, UDP, TCP.
 */

#ifdef NSIP
void    idpip_input(), nsip_ctlinput();
#endif

#ifdef TPIP
void    tpip_input(), tpip_ctlinput(), tp_ctloutput();
int     tp_init(), tp_slowtimo(), tp_drain(), tp_usrreq();
#endif

extern  struct domain inetdomain;

#if 1
#define ip_sysctl 0
#define icmp_sysctl 0
#define udp_sysctl 0
#endif

#ifdef _MSC_VER
#pragma warning(disable:4028)   /* formal parameter different */
#endif

const struct protosw inetsw[] = {
{ 0,            &inetdomain,    0,              0,
  0, (INT(*)())ip_output,      0,              0,
  0,
  ip_init,      0,              ip_slowtimo,    ip_drain,       ip_sysctl
},
#ifdef __WANTS_UDP
{ SOCK_DGRAM,   &inetdomain,    IPPROTO_UDP,    PR_ATOMIC|PR_ADDR,
  (void(*)())udp_input, 0,      udp_ctlinput,   ip_ctloutput,
  udp_usrreq,
  udp_init,     0,              0,              0,              udp_sysctl
},
#endif
#ifdef __WANTS_TCP
{ SOCK_STREAM,  &inetdomain,    IPPROTO_TCP,    PR_CONNREQUIRED|PR_WANTRCVD,
  (void(*)())tcp_input, 0,      tcp_ctlinput,   tcp_ctloutput,
  tcp_usrreq,
  tcp_init,     tcp_fasttimo,   tcp_slowtimo,   tcp_drain,
},
#endif
{ SOCK_RAW,     &inetdomain,    IPPROTO_RAW,    PR_ATOMIC|PR_ADDR,
  (void(*)())rip_input,(INT(*)())rip_output, 0,rip_ctloutput,
  rip_usrreq,
  0,            0,              0,              0,
},
{ SOCK_RAW,     &inetdomain,    IPPROTO_ICMP,   PR_ATOMIC|PR_ADDR,
  (void(*)())icmp_input, (INT(*)())rip_output, 0, rip_ctloutput,
  rip_usrreq,
  0,            0,              0,              0,              icmp_sysctl
},
#ifdef __WANTS_IGMP
{ SOCK_RAW,     &inetdomain,    IPPROTO_IGMP,   PR_ATOMIC|PR_ADDR,
  (void(*)())igmp_input, (INT(*)())rip_output, 0, rip_ctloutput,
  rip_usrreq,
  igmp_init,    igmp_fasttimo,  0,              0,
},
#endif
#ifdef TPIP
{ SOCK_SEQPACKET,&inetdomain,   IPPROTO_TP,     PR_CONNREQUIRED|PR_WANTRCVD,
  (void(*)())tpip_input, 0,     tpip_ctlinput,  tp_ctloutput,
  tp_usrreq,
  tp_init,      0,              tp_slowtimo,    tp_drain,
},
#endif
#if MONARCH_ADHOC
{ 0,            &inetdomain,    IP6_NHDR_HOP,   0,
  hop_input,    0,              0, /* opt6_ctlinput */  0,
  rip_usrreq,
  opt6_init,    0,              0,              0,
  0 /*&rip_usrreqs*/ /* XXX */
}, 
{ 0,            &inetdomain,    IP6_NHDR_RT,    0,
  rt_input,     0,              0,      0,
  rip_usrreq,
  dsr_init, 0,      0,      0,
  0 /*&rip_usrreqs*/ /* XXX */
},
{ SOCK_RAW,     &inetdomain,    IP6_NHDR_DOPT,  PR_ATOMIC|PR_ADDR,
  dopt_input,   0,              0,      rip_ctloutput,
  rip_usrreq,
  0,            0,              0,              0,
  0 /*&rip_usrreqs*/
},
{ SOCK_RAW,     &inetdomain,    IP6_NHDR_NONH,  PR_ATOMIC|PR_ADDR,
  end_input,    0,              0,      rip_ctloutput, 
  rip_usrreq,
  0,            0,              0,              0,
  0 /*&rip_usrreqs*/
}, 
#endif /* MONARCH_ADHOC */
        /* raw wildcard */
{ SOCK_RAW,     &inetdomain,    0,              PR_ATOMIC|PR_ADDR,
  (void(*)())rip_input, (INT(*)())rip_output, 0, rip_ctloutput,
  rip_usrreq,
  rip_init,     0,              0,              0,
},
};

struct domain inetdomain =
    { AF_INET, "internet", 0, 0, 0, 
      inetsw, &inetsw[sizeof(inetsw)/sizeof(inetsw[0])], 0 };
