/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <sock.h>
#include <if.h>

/********************** CNetDriverCallback **********************/

typedef struct _NETDRIVERCALLBACK {
  struct INetDriverCallbackVtbl *v;
  UINT RefCnt;
} *PNETDRIVERCALLBACK;

#define pPNETDRIVERCALLBACK(_pThis_) ((PNETDRIVERCALLBACK) (_pThis_))

static SCODE NDCQueryInterface (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((PIUNKNOWN) pThis, iid, ppObject,
                                 &IID_INetDriverCallback);

}

static UINT NDCAddRef (
    /*in*/ PINETDRIVERCALLBACK pThis)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);

    return AtomicInc(&x->RefCnt);
}

static UINT NDCRelease (
    /*in*/ PINETDRIVERCALLBACK pThis)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UINT Cnt = AtomicDec(&x->RefCnt);

    /* If (Cnt == 0) ... */

    return Cnt;
}

static SCODE NDCIfAttach (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct ifnet *ifp )
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UnusedParameter(x);

    if_attach(ifp);
    return S_OK;
}

static SCODE NDCMfreem (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct mbuf * m )
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UnusedParameter(x);

    m_freem(m);

    return S_OK;
}

static SCODE NDCMclget (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct mbuf * m,
    /*in*/ INT how)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UnusedParameter(x);

    MCLGET(m, how);

    if (m)
      return S_OK;
    else
      return E_NOT_ENOUGH_MEMORY;
}

static SCODE NDCMget (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct mbuf * m,
    /*in*/ INT how,
    /*in*/ INT type)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);

    MGET(m, how, type);

    if (m)
      return S_OK;
    else
      return E_NOT_ENOUGH_MEMORY;
}

static SCODE NDCMgethdr (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ INT how,
    /*in*/ INT type,
    /*out*/ struct mbuf ** pM)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    struct mbuf *m;
    UnusedParameter(x);

    m = m_gethdr(how, type);

    *pM = m;

    if (m)
      return S_OK;
    else
      return E_NOT_ENOUGH_MEMORY;
}

static SCODE NDCEtherInput (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct ifnet * ifp,
    /*in*/ struct ether_header * eh,
    /*in*/ struct mbuf * m)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UnusedParameter(x);

    ether_input(ifp, eh, m);

    return S_OK;
}

static SCODE NDCEtherOutput (
    /*in*/ PINETDRIVERCALLBACK pThis,
    /*in*/ struct ifnet * ifp,
    /*in*/ struct mbuf * m0,
    /*in*/ struct sockaddr * dst,
    /*in*/ struct rtentry * rt0)
{
    SCODE sc;
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);
    UnusedParameter(x);

    sc = ether_output(ifp, m0, dst, rt0);

    return sc;                  /* XXX error code domain? */
}

static SCODE NDCNetLock (
    /*in*/ PINETDRIVERCALLBACK pThis)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);

    UnusedParameter(x);

    NetLock();

    return S_OK;
}

static SCODE NDCNetUnlock (
    /*in*/ PINETDRIVERCALLBACK pThis)
{
    PNETDRIVERCALLBACK x = pPNETDRIVERCALLBACK(pThis);

    UnusedParameter(x);

    NetUnlock();

    return S_OK;
}

#include "s-CNetDriverCallback-v.c"

struct _NETDRIVERCALLBACK TheNetDriverCallBack = {
  &CNetDriverCallbackVtbl,
  1
};

SCODE InitNetDriverCallBack()
{
  SCODE sc;
  PINAMESPACE ns = CurrentNameSpace();

  sc = ns->v->Register(ns, "NetDriverCallBack",
                       (PIUNKNOWN) &TheNetDriverCallBack,
                       0, NULL);
  return sc;
}
