/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1980, 1986, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)raw_cb.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#ifdef RAW_SOCKETS
#include <domain.h>

#include <if.h>
#include <route.h>
#include <raw_cb.h>

#include <in.h>


/*
 * Routines to manage the raw protocol control blocks. 
 *
 * TODO:
 *      hash lookups by protocol family/protocol + address family
 *      take care of unique address problems per AF?
 *      redo address binding to allow wildcards
 */

u_long  raw_sendspace = RAWSNDQ;
u_long  raw_recvspace = RAWRCVQ;

/*
 * Allocate a control block and a nominal amount
 * of buffer space for the socket.
 */
INT
raw_attach(struct socket *so, INT proto)
{
        struct rawcb *rp = sotorawcb(so);
        int error;

        /*
         * It is assumed that raw_attach is called
         * after space has been allocated for the
         * rawcb.
         */
        if (rp == 0)
                return (ENOBUFS);
        error = soreserve(so, raw_sendspace, raw_recvspace);
        if (error)
                return (error);
        rp->rcb_socket = so;
        rp->rcb_proto.sp_family = (u_short)so->so_proto->pr_domain->dom_family;
        rp->rcb_proto.sp_protocol = (u_short)proto;
        insque((MYQUEUE*)rp, (MYQUEUE*)&rawcb);
        return (0);
}

/*
 * Detach the raw connection block and discard
 * socket resources.
 */
void
raw_detach(struct rawcb *rp)
{
        struct socket *so = rp->rcb_socket;

        so->so_pcb = 0;
        sofree(so);
        remque((MYQUEUE*)rp);
#ifdef notdef
        if (rp->rcb_laddr)
                bsd_free(rp->rcb_laddr, M_TEMP);
        rp->rcb_laddr = 0;
#endif
        bsd_free((caddr_t)(rp), M_PCB);
}

/*
 * Disconnect and possibly release resources.
 */
void
raw_disconnect(struct rawcb *rp)
{

#ifdef notdef
        if (rp->rcb_faddr)
                bsd_free(rp->rcb_faddr, M_TEMP);
        rp->rcb_faddr = 0;
#endif
        if (rp->rcb_socket->so_state & SS_NOFDREF)
                raw_detach(rp);
}

#ifdef notdef
int
raw_bind(struct socket *so, struct mbuf *nam)
{
        struct sockaddr *addr = mtod(nam, struct sockaddr *);
        struct rawcb *rp;
        struct sockaddr *n;

        if (ifnet == 0)
                return (EADDRNOTAVAIL);
        n = bsd_malloc(nam->m_len, M_TEMP, M_WAITOK);
        if (n == NULL)
            return ENOBUFS;
        rp = sotorawcb(so);
        m_copydata(nam,0,nam->m_len,n);
        rp->rcb_laddr = n;
        return (0);
}
#endif

#endif /* RAW_SOCKETS */
