/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <sock.h>
#include <diagnostics.h>
#include <domain.h>

#include <if.h>
#include <route.h>
#include <raw_cb.h>

#include <in.h>
#include <in_var.h>

#include <netif.h>

static struct gateway mygw;

#define INADDR2SIN(a, b)                                                \
do {                                                                    \
    memset(a, 0, sizeof(*(a)));                                         \
    (a)->sin_family = AF_INET;                                          \
    (a)->sin_len = sizeof(struct sockaddr_in);                          \
    (a)->sin_addr = b;                                                  \
} while (/*CONSTCOND*/0)

INT
rt_addgw(struct in_addr gwaddr)
{
    struct ifaddr *ifa;

    mygw.gw_valid = 0;
    INADDR2SIN(&mygw.gw_gw, gwaddr);

    /*
     * First check for an interface which sits on the same network
     * as the alleged gateway.
     */
    ifa = ifa_ifwithnet((struct sockaddr *)&mygw.gw_gw);

    /*
     * In case that's a no go, check for a point-to-point interface
     * with the correct destination address.  If found, set that as
     * the gateway.
     */
    if (!ifa)
        ifa = ifa_ifwithdstaddr((struct sockaddr *)&mygw.gw_gw);
     
    /*
     * return to sender, address unknown 
     * no such number, no such zone 
     */
    if (ifa == NULL)
        return 1;

    mygw.gw_ifa = ifa;
    mygw.gw_valid = 1;

    return S_OK;
}

void
rt_delgw()
{

    mygw.gw_valid = 0;
}

struct ifaddr *
rt_getgw()
{

    return mygw.gw_ifa;
}

INT
rt_ifaout(struct in_addr dest, struct ifaddr **ifap,
    struct sockaddr_in *sin_nexthop)
{
    extern struct ifnet *lo0ifp; /* XXX headers */
    extern struct ifaddr *ifa_multiaddr; /* XXX headers */
    struct sockaddr_in sin;
    struct ifaddr *ifa;

    /*
     * First deal with multicast.  If the target address is in the
     * IP range, send to the multicast address on  the default
     * Multicast interface.
     */
    if (IN_MULTICAST(ntohl(dest.s_addr)) && ifa_multiaddr) {
        ifa = ifa_multiaddr;
        goto done;
    }

    INADDR2SIN(&sin, dest);

    /*
     * Next see if we have an interface configured with the
     * destination address and thread that through the loopback
     * interface.  This matches the original BSD routing behaviour.
     */
    ifa = ifa_ifwithaddr((struct sockaddr *)&sin);
    if (ifa) {
        ifa = lo0ifp->if_addrlist;
    } else {
        /* try point-to-point and netmask matches */
        ifa = ifa_ifwithdstaddr((struct sockaddr *)&sin);
        if (!ifa)
            ifa = ifa_ifwithnet((struct sockaddr *)&sin);
    }

 done:
    if (ifa) {
        if (ifap)
            *ifap = ifa;
        if (sin_nexthop)
            INADDR2SIN(sin_nexthop, dest);
        return S_OK;
    }

    if (mygw.gw_valid) {
        if (ifap)
            *ifap = mygw.gw_ifa;
        if (sin_nexthop)
            *sin_nexthop = mygw.gw_gw;
        return S_OK;
    }

    return !S_OK;
}

/*
 * dummy dummy dummy
 */
INT
rtioctl(INT req, PTR data)
{

    return EOPNOTSUPP;
}
