/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)uipc_domain.c       8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <domain.h>

const struct protosw *
pffindtype(int family, int type)
{
        const struct domain *dp;
        const struct protosw *pr;

        for (dp = domains; dp; dp = dp->dom_next)
                if (dp->dom_family == family)
                        goto found;
        return (0);
found:
        for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                if (pr->pr_type && pr->pr_type == type)
                        return (pr);
        return (0);
}

const struct protosw *
pffindproto(int family, int protocol, int type)
{
        const struct domain *dp;
        const struct protosw *pr;
        const struct protosw *maybe = NULL;

        if (family == 0)
                return (0);
        for (dp = domains; dp; dp = dp->dom_next)
                if (dp->dom_family == family)
                        goto found;
        return (0);
found:
        for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++) {
                if ((pr->pr_protocol == protocol) && (pr->pr_type == type))
                        return (pr);

                if (type == SOCK_RAW && pr->pr_type == SOCK_RAW &&
                    pr->pr_protocol == 0 && maybe == (const struct protosw *)0)
                        maybe = pr;
        }
        return (maybe);
}

#if 0
net_sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp,
           size_t newlen, struct proc *p)
{
        const struct domain *dp;
        const struct protosw *pr;
        int family, protocol;

        /*
         * All sysctl names at this level are nonterminal;
         * next two components are protocol family and protocol number,
         * then at least one addition component.
         */
        if (namelen < 3)
                return (EISDIR);                /* overloaded */
        family = name[0];
        protocol = name[1];

        if (family == 0)
                return (0);
        for (dp = domains; dp; dp = dp->dom_next)
                if (dp->dom_family == family)
                        goto found;
        return (ENOPROTOOPT);
found:
        for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                if (pr->pr_protocol == protocol && pr->pr_sysctl)
                        return ((*pr->pr_sysctl)(name + 2, namelen - 2,
                            oldp, oldlenp, newp, newlen));
        return (ENOPROTOOPT);
}
#endif

void pfctlinput(INT cmd, struct sockaddr *sa)
{
        const struct domain *dp;
        const struct protosw *pr;

        for (dp = domains; dp; dp = dp->dom_next)
                for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                        if (pr->pr_ctlinput)
                                (*pr->pr_ctlinput)(cmd, sa, (caddr_t) 0);
}

void pfslowtimo(void)
{
        const struct domain *dp;
        const struct protosw *pr;

        for (dp = domains; dp; dp = dp->dom_next)
                for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                        if (pr->pr_slowtimo)
                                (*pr->pr_slowtimo)();
#if 0
        timeout(pfslowtimo, (caddr_t)0, hz/2);
#endif
}

void pffasttimo(void)
{
        const struct domain *dp;
        const struct protosw *pr;

        for (dp = domains; dp; dp = dp->dom_next)
                for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                        if (pr->pr_fasttimo)
                                (*pr->pr_fasttimo)();
#if 0
        timeout(pffasttimo, (caddr_t)0, hz/5);
#endif
}


#define ADDDOMAIN(x)    { \
        extern struct domain __CONCAT(x,domain); \
        __CONCAT(x,domain.dom_next) = domains; \
        domains = &__CONCAT(x,domain); \
}

void StartTimers(void); /* in syncprim.c */

void domaininit()
{
        const struct domain *dp;
        const struct protosw *pr;

#ifdef INET
        ADDDOMAIN(inet);
#endif
#if MONARCH_ADHOC
        ADDDOMAIN(adhoc);
#endif

        for (dp = domains; dp; dp = dp->dom_next) {
                if (dp->dom_init)
                        (*dp->dom_init)();
                for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
                        if (pr->pr_init)
                                (*pr->pr_init)();
        }

if (max_linkhdr < 16)           /* XXX */
max_linkhdr = 16;
        max_hdr = max_linkhdr + max_protohdr;
        max_datalen = MHLEN - max_hdr;
#if 0
        pffasttimo();
        pfslowtimo();
#else
        StartTimers();
#endif
}

