/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)tcp_debug.c    8.1 (Berkeley) 6/10/93
 */
/* max verbose with
   #define TCPDEBUG 2
   #define PRUREQUESTS
 */

#include <sock.h>
#ifdef __WANTS_TCP
#ifdef TCPDEBUG
/* load symbolic names */
#define PRUREQUESTS
#define TCPSTATES
#define TCPTIMERS
#define TANAMES

#include <route.h>
#include <if.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <ip_var.h>
#include <tcp.h>
#include <tcp_fsm.h>
#include <tcp_seq.h>
#include <tcp_tim.h>
#include <tcp_var.h>
#include <tcpip.h>
#include <tcp_dbg.h>

INT tcpconsdebug = 0;
struct tcp_debug tcp_debug[TCP_NDEBUG];
INT tcp_debx;

/*
 * Tcp debug routines
 */
void
tcp_trace(INT act, INT ostate, struct tcpcb *tp, struct tcpiphdr *ti, INT req)
{
#if (TCPDEBUG>1)
    tcp_seq seq, ack;
    INT len, flags;
#endif
    struct tcp_debug *td = &tcp_debug[tcp_debx++];

    if (tcp_debx == TCP_NDEBUG)
        tcp_debx = 0;
    td->td_time = iptime();
    td->td_act = (short) act;
    td->td_ostate = (short) ostate;
    td->td_tcb = (caddr_t)tp;
    if (tp)
        td->td_cb = *tp;
    else
        bzero((caddr_t)&td->td_cb, sizeof (*tp));
    if (ti)
        td->td_ti = *ti;
    else
        bzero((caddr_t)&td->td_ti, sizeof (*ti));
    td->td_req = (short) req;
#if (TCPDEBUG>1)
    if (tcpconsdebug == 0)
        return;
    if (tp)
        printf(_T("%x %hs:"), tp, tcpstates[ostate]);
    else
        printf(_T("???????? "));
    printf(_T("%hs "), tanames[act]);
    switch (act) {

    case TA_INPUT:
    case TA_OUTPUT:
    case TA_DROP:
        if (ti == 0)
            break;
        seq = ti->ti_seq;
        ack = ti->ti_ack;
        len = ti->ti_len;
        if (act == TA_OUTPUT) {
            seq = ntohl(seq);
            ack = ntohl(ack);
            len = ntohs((UINT16)len);
        }
        if (act == TA_OUTPUT)
            len -= sizeof (struct tcphdr);
        if (len)
            printf(_T("[%x..%x)"), seq, seq+len);
        else
            printf(_T("%x"), seq);
        printf(_T("@%x, urp=%x"), ack, ti->ti_urp);
        flags = ti->ti_flags;
        if (flags) {
#ifndef lint
            char *cp = "<";
#define pf(f) { if (ti->ti_flags&TH_ ## f) { printf(_T("%hs%hs"), cp, "f"); cp = ","; } }
            pf(SYN); pf(ACK); pf(FIN); pf(RST); pf(PUSH); pf(URG);
#endif
            printf(_T(">"));
        }
        break;

    case TA_USER:
        printf(_T("%hs"), prurequests[req&0xff]);
        if ((req & 0xff) == PRU_SLOWTIMO)
            printf(_T("<%hs>"), tcptimers[req>>8]);
        break;
    }
    if (tp)
        printf(_T(" -> %hs"), tcpstates[tp->t_state]);
    /* print out internal state of tp !?! */
    printf(_T("\n"));
    if (tp == 0)
        return;
    printf(_T("\trcv_(nxt,wnd,up) (%x,%x,%x) snd_(una,nxt,max) (%x,%x,%x)\n"),
        tp->rcv_nxt, tp->rcv_wnd, tp->rcv_up, tp->snd_una, tp->snd_nxt,
        tp->snd_max);
    printf(_T("\tsnd_(wl1,wl2,wnd) (%x,%x,%x)\n"),
        tp->snd_wl1, tp->snd_wl2, tp->snd_wnd);
#endif /* TCPDEBUG>1 */
}

#endif /* TCPDEBUG */
#endif /* __WANTS_TCP */
