/* Copyright (c) Microsoft Corporation. All rights reserved. */

CXTINFO *_Reschedule( CXTINFO *pState )
{
    PITHREAD NextThread, ThisThread;

    /* Return immediately if we cannot reschedule.
     */
    if (Cbs.PreemptionDisableCount) {
        LogIt(2,(UINT)pState);
        RescheduleNeeded(); /* later */
        /*printf("Hit upon preemption disabled..\n");*/
        return pState;
    }

    /* Remember where we stored our state
     * BUGBUG can this be done with interrupts on ?
     */
    ThisThread = CurrentThread();

    if (Cbs.FirstThreadBooted)
        pTH(ThisThread)->CxtInfo = pState;
    Cbs.FirstThreadBooted = TRUE;

    /* Rescheduling path cannot be reentrant.
     */
    Cbs.PreemptionDisableCount++;

    /* Interrupts ok now
     */
    ENABLE_INTERRUPTS();

    /* Invoke scheduler to find who should run next
     */
    NextThread = CbsReschedule(pTheScheduler,ThisThread);

    /* Decided who should run next. No more interrupts.
     */
    DISABLE_INTERRUPTS();
    Cbs.PreemptionDisableCount--;

    assert(!Cbs.ContextSwitchNeeded);

    return pTH(NextThread)->CxtInfo;
}

/* xx in machdep.h ? */
void LoadThreadContext(PITHREAD NextThread)
{
    HWLoadThreadContext(pTH(NextThread)->CxtInfo);
}

