/* Copyright (c) Microsoft Corporation. All rights reserved. */
PCXTINFO _Reschedule( PCXTINFO pState )
{
    PITHREAD NextThread, ThisThread;

    /* Return immediately if we cannot reschedule.
     */
    if (Cbs.PreemptionDisableCount) {
        LogIt(2,(UINT)pState);
        RescheduleNeeded(); /* later */
        //putchar('$');
        /*printf("_Reschedule: Hit upon preemption disabled..\n");*/
        return pState;
    }

    /* Remember where we stored our state
     */
    ThisThread = CurrentThread();

    if (Cbs.FirstThreadBooted)
        pTH(ThisThread)->CxtInfo = pState;
    Cbs.FirstThreadBooted = TRUE;

    /* Rescheduling path cannot be reentrant.
     */
    Cbs.PreemptionDisableCount++;

    /* Interrupts ok now
     */
    ENABLE_INTERRUPTS();

    /* Invoke scheduler to find who should run next
     */
    NextThread = CbsReschedule(pTheScheduler,ThisThread);

    /* Decided who should run next. No more interrupts.
     */
    DISABLE_INTERRUPTS();

    Cbs.PreemptionDisableCount--;

    return pTH(NextThread)->CxtInfo;
}

/*
 * Voluntary context-switch code.
 * Only must save enough state for a function return.
 * But reload might be from interrupt code above, hence frame.
 */
void Reschedule(void)
{
    CXTINFO Cxt, *pCxt;
    INT x;

    //assert(!INTERRUPTS_ARE_DISABLED());
    DISABLE_INTERRUPTS();
    x = HWSaveThreadContext(&Cxt);
    if (x == 0) {
        pCxt = _Reschedule(&Cxt);
        HWLoadThreadContext(pCxt);
    }
    /*putchar('&');*/
}

void LoadThreadContext(PITHREAD NextThread)
{
    HWLoadThreadContext(pTH(NextThread)->CxtInfo);
}
