/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base/thread.h>
#include <base/schedul.h>
#include <schedulif.h>

/* MMLITE trivial round-robin scheduler object (RRS)
 * This code intended solely for illustration purposes.
 */

typedef struct _RRS {
    CONST_VTBL struct ISchedulerVtbl *v;
    PLISTNODE ReadyQueue;
    BOOL PreemptionEnabled;    
    BOOL PreemptionDisableCount;
    TIME IdleTime;
    BOOL ContextSwitchNeeded;
    BOOL FirstThreadBooted;
} RRS, *PRRS;

static RRS Rrs;

#define pRRS(_x_) ((PRRS)_x_)

struct IThread * MCT RrsReschedule( IScheduler *This, IThread *pCurrent);
#define CbsReschedule RrsReschedule /* machdep happy */
#define RescheduleNeeded() {Rrs.ContextSwitchNeeded = TRUE;}
#define Cbs Rrs
#define LogIt(x,y)
#include "_schedul.c"

struct IThread * MCT RrsReschedule( IScheduler *This, IThread *pCurrent)
{
    IThread *r;
    TIME TimeSlice;

    /* If noone else wants the processor, just return
     */
    if (pRRS(This)->ReadyQueue != NULL) {

        PLISTNODE pHead;

        /* Who is the next thread to run
         */
        r = GetThreadFrom( pRRS(This)->ReadyQueue, Queue);

        /* Take it out of the queue. Returns NULL if queue empty.
         */
        pHead = ListRemove( &pTH(r)->Queue );

        /* Bootstrap problem. Sigh.
         */
        if (r != pCurrent) {

            /* Insert current thread at end, unless list is empty.
             */
            if (pHead)
                ListInsert( pHead->pPrev, &pTH(pCurrent)->Queue );
            else
                pHead = &pTH(pCurrent)->Queue;
        }

        /* This thread (if any) will run next time around.
         */
        pRRS(This)->ReadyQueue = pHead;

    } else
        r = pCurrent;

    ThreadSetCurrent( r );

    Int32ToInt64(TimeSlice,TIME_MILLIS(1));
    SetNextInterrupt(TimeSlice, NULL);

    return r;
}

void MCT RrsAddThread( IScheduler *This, IThread *Thread)
{
    pRRS(This)->ReadyQueue = ListInsert( pRRS(This)->ReadyQueue, &pTH(Thread)->Queue );
}

void MCT RrsRemoveThread( IScheduler *This, IThread *Thread)
{
    if (pTH(Thread)->QueueType != QUEUE_READY
        && pTH(Thread)->State != THD_READY)
      printf("RrsRemoveThread: th=%x qt=%x state=%x\n",
             (ADDRESS) Thread, pTH(Thread)->QueueType, pTH(Thread)->State);

    pRRS(This)->ReadyQueue = ListRemove( &pTH(Thread)->Queue );
}

BOOL MCT RrsDisablePreemption( IScheduler *This, BOOL Recursive)
{
    BOOL WhatItWas = pRRS(This)->PreemptionEnabled;
    UnusedParameter(Recursive);
    pRRS(This)->PreemptionEnabled = FALSE;
    return WhatItWas;
}

BOOL MCT RrsEnablePreemption( IScheduler *This, BOOL Recursive)
{
    BOOL WhatItWas = pRRS(This)->PreemptionEnabled;
    UnusedParameter(Recursive);
    pRRS(This)->PreemptionEnabled = TRUE;
    return WhatItWas;
}


BOOL MCT RrsPreemptionIsEnabled( IScheduler *This)
{
    return pRRS(This)->PreemptionEnabled;
}


void MCT RrsWakeupOnMutex(
                IScheduler *This,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
    UnusedParameter(pMutex);
    UnusedParameter(Owner);
    DebugBreak();
}

void MCT RrsWaitForMutex(
                IScheduler *This,
                IThread *Thread,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
    UnusedParameter(pMutex);
    UnusedParameter(Owner);
    UnusedParameter(Thread);
    DebugBreak();
}

TIME MCT RrsGetIdleTime(IScheduler *This)
{
    return pRRS(This)->IdleTime;
}

static CONST_VTBL struct ISchedulerVtbl RrsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    RrsReschedule,
    RrsAddThread,
    RrsRemoveThread,
    RrsDisablePreemption,
    RrsEnablePreemption,
    RrsPreemptionIsEnabled,
    RrsWaitForMutex,
    RrsWakeupOnMutex,
    NULL,
    NULL,
    NULL,
    RrsGetIdleTime
};

static RRS Rrs = {
    &RrsVtbl,
    NULL,
    TRUE,
};

IScheduler *SchedulerCreate(void)
{
    Rrs.ReadyQueue = NULL;
    Rrs.PreemptionEnabled = TRUE;
    Int32ToInt64(Rrs.IdleTime,0);
    return (IScheduler *) &Rrs;
}

SCODE SleepUntil(TIME Until)
{
    UnusedParameter(Until);
    return E_NOT_IMPLEMENTED;
}

PITHREAD ThreadConditionCancel(PITHREAD pThd, PITHREAD pThdCancel)
{
    UnusedParameter(pThd);
    UnusedParameter(pThdCancel);
    DebugBreak();
    return pThd;
}
