/* sched-multi.h
 * Kernel scheduler mechanisms
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#ifndef __sched_multi_h__
#define __sched_multi_h__

#include "sched.h"
#include <util/list-single.h>

#define SCHED_MULTI_STRUCT                                                    \
  SCHED_STRUCT;                                                               \
  UINT ActCount

#define /* void */                                                            \
SCHED_MULTI_CREATE()                                                          \
  SCHED_CREATE();                                                             \
  TheScheduler.ActCount = 0

PRIVATE SCODE MCT
Sched_AddActivity(
     PISCHEDULER This,
     PIACTIVITY Activity,
     PTIME_CONSTRAINT pConstraint
     );

PRIVATE void MCT
Sched_RemoveActivity(
     PISCHEDULER This,
     PIACTIVITY Activity
     );

PRIVATE void MCT
Sched_WakeUpOnMutex(
     PISCHEDULER This,
     PMUTEX pMutex
     );

PRIVATE void MCT Sched_WaitForMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    );

PRIVATE void MCT Sched_WaitForCondition(
    PISCHEDULER This,
    PCONDITION pCondition,
    PTIME pTimeOut
    );

PRIVATE void MCT Sched_WakeUpOnCondition(
    PISCHEDULER This,
    PCONDITION pCond,
    BOOL fAll,
    PTIME_CONSTRAINT ptcNew
    );

PRIVATE SCODE MCT
Sched_WaitForConditionWithConstraint(
    PISCHEDULER pThis,
    PCONDITION pCond,
    PMUTEX pMutex,
    TIME *pTimeout,
    PTIME_CONSTRAINT pConstraint,
    PTIME pTimeTaken
    );

PRIVATE void
Sched_ConditionWakeupTransform(
    PTR pConstraintWaiter,
    PTR ptcNew
    );

/* PRIVATE INLINE void */
/* Sched_MutexRelock( */
/*     PMX pMutex, */
/*     PACTINFO pActWaiter */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_MutexAddWaiter( */
/*     PMX pMutex, */
/*     PACTINFO pActWaiter */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_MutexRemoveWaiter( */
/*     PMX pMutex, */
/*     PACTINFO pActWaiter */
/*     ); */


/* PRIVATE INLINE void */
/* Sched_ConditionWaitWithConstraintBefore( */
/*     PACTINFO pActWaiter, */
/*     PTIME_CONSTRAINT pConstraint */
/*     ); */

/* PRIVATE INLINE SCODE */
/* Sched_ConditionWaitWithConstraintAfter( */
/*     PCOND pCond, */
/*     PACTINFO pActWaiter, */
/*     PTIME pTimeTaken */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_ConditionAddWaiter( */
/*     PCOND pCond, */
/*     PACTINFO pActWaiter, */
/*     PTIME pTimeOut */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_ConditionRemoveWaiter( */
/*     PCOND pCond, */
/*     PACTINFO pActWaiter */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_ReadyActivity( */
/*     PACTINFO pAct, */
/*     BOOL PreemptNow, */
/*     PTIME_CONSTRAINT ptcNew */
/*     ); */

/* PRIVATE INLINE void */
/* Sched_SleepActivity( */
/*     PACTINFO pAct */
/*     ); */

#define /* UINT */                                                            \
Sched_GetActCount()                                                           \
  TheScheduler.ActCount

#define /* void */                                                            \
Sched_InitActCount()                                                          \
  Sched_GetActCount() = 0

#define /* void */                                                            \
Sched_IncActCount()                                                           \
  AtomicInc(&Sched_GetActCount())

#define /* void */                                                            \
Sched_DecActCount()                                                           \
  AtomicDec(&Sched_GetActCount())

#endif /* __sched_multi_h__ */
