/* sched.c
 * Generic scheduler mechanisms. You should include this in a specific
 * scheduler policy module (round-robin, constraint-based, etc.)
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#ifndef __sched_c__
#define __sched_c__

#include <mmlite.h>
#include <assert.h>

/* ===========================================================================
 * Interface method implementations
 */


PRIVATE SCODE MCT 
Sched_QueryInterface( 
    PISCHEDULER pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    )
{
  return GenericQueryInterface((IUnknown *)pThis, pIid, ppUnk, &IID_IScheduler);
}
        
PRIVATE UINT MCT 
Sched_AddRef(
    PISCHEDULER pThis
    )
{
  UINT RefCnt;
  RefCnt = AtomicInc(&TheScheduler.RefCount);
  return RefCnt;
}
        
PRIVATE UINT MCT 
Sched_Release (
    PISCHEDULER pThis
    )
{
  UINT RefCnt;
  RefCnt = AtomicDec(&TheScheduler.RefCount);
  return RefCnt;
}

PRIVATE TIME MCT 
Sched_GetIdleTime(
    PISCHEDULER This
    )
{
  return (pSCHED(This)->IdleTime);
}

/* Disables preemption, or if recursion (nested disables) is TRUE,
 * increments the preemption disable count.
 * Returns TRUE if preemption was already disabled, FALSE otherwise. */
PRIVATE BOOL MCT 
Sched_DisablePreemption(
    PISCHEDULER This,
    BOOL Recursive
    )
{
  UINT currentDisableCount;

  currentDisableCount = AtomicInc(&(pSCHED(This)->PreemptionDisableCount));
  assert(Recursive || (currentDisableCount == 1));
  return (currentDisableCount != 1);
}

/* Enables preemption, or if recursion (nested disables) is TRUE,
 * decrements the preemption disable count.
 * Returns TRUE if preemption is now enabled, FALSE otherwise.
 */
PRIVATE BOOL MCT 
Sched_EnablePreemption(
    PISCHEDULER This,
    BOOL Recursive
    )
{
  UINT currentDisableCount;

  currentDisableCount = AtomicDec(&(pSCHED(This)->PreemptionDisableCount));

  assert(Recursive || (currentDisableCount == 0));

  if ((currentDisableCount == 0) && pSCHED(This)->ContextSwitchNeeded) {
    Reschedule();
  }

  return (currentDisableCount == 0);
}

PRIVATE BOOL MCT 
Sched_PreemptionIsEnabled(
    PISCHEDULER This
    )
{
  return (pSCHED(This)->PreemptionDisableCount == 0);
}

PRIVATE SCODE MCT
Sched_CreateConstraint(
    PISCHEDULER pThis,
    TIME pStartTime,
    TIME pPeriod,
    TIME pEstimate,
    UINT Recurrence,
    BOOL Hardness,
    BOOL AutoAdjust,
    BOOL Infinite,
    PIUNKNOWN* ppIUnknown
    )
{
  return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE MCT
Sched_SetConstraint(
    PISCHEDULER pThis,
    PTIME_CONSTRAINT pConstraint,
    PTIME pTimeTaken
    )
{
  return E_NOT_IMPLEMENTED;
}

/* Changes the given timeout to be absolute and in the present or future. */
PRIVATE INLINE void 
Sched_NormalizeTimeout(
    PTIME tTimeout
    )
{
  TIME tNow = CurrentTime();
  
  /* if Timeout is relative to current time, make it absolute.
   * This guarantees a greater-or-equal time than now. */
  if (TimeIsNotPositive(*tTimeout)) {
    INT64SUB(*tTimeout, tNow, *tTimeout);
  }
  /* if the Timeout time has already occurred, use current time */
  else if (TimeLess(*tTimeout, tNow)) {
    *tTimeout = tNow;
  }
}


#endif /* __sched_c__ */
