/* sched.h
 * Header file for kernel scheduler mechanism.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#ifndef __sched_h__
#define __sched_h__

#define OPTIMIZE      /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <assert.h>
#include <diagnostics.h>
#include <base/condition2.h>
#include <base/schedul.h>
#include <base/mutex2.h>
#include <base/activity.h>
#include <base/workitem.h>
#include <base/thread2.h>
#include <base/tconstraint2.h>

#if 0
/* Use with ListTransform to print out constraints in queues for debugging */
void Transform(PTR pConstraint) {
    printf("0x%x %ld\n", KConstraintGetSchedObject(pConstraint),
	   KConstraintGetRestartBy(pConstraint));
}

/* Use with ListTransformWithArg to prevent duplicates in a linked list */
void TransformEx(PTR pConstraint, PTR pAct) {
    assert(KConstraintGetSchedObject(pConstraint) != pAct);
}
#endif

/* MMLite Constraint-based scheduler object (Sched) */
#define SCHED_STRUCT                                                          \
  const struct ISchedulerVtbl *v;                                             \
  UINT PreemptionDisableCount;                                                \
  UINT RefCount;                                                              \
  TIME IdleTime;                                                              \
  BOOL ContextSwitchNeeded;                                                   \
  BOOL FirstThreadBooted

/* Constructor to override this method in the policy file but call this one */
#define /* void */                                                            \
SCHED_CREATE()                                                                \
  TheScheduler.PreemptionDisableCount = 0;                                    \
  TheScheduler.RefCount = 0;                                                  \
  TheScheduler.ContextSwitchNeeded = FALSE;                                   \
  TheScheduler.FirstThreadBooted = FALSE;                                     \
  Int32ToInt64(TheScheduler.IdleTime, 0)

#define RescheduleNeeded() TheScheduler.ContextSwitchNeeded = TRUE
#define RescheduleNotNeeded() TheScheduler.ContextSwitchNeeded = FALSE


/* ===========================================================================
 * Scheduler functions that implement exported interface methods.
 */

PRIVATE SCODE MCT Sched_QueryInterface( 
     PISCHEDULER pThis,
     /* [in] */ REFIID pIid,
     /* [out] */ void **ppUnk
     );

PRIVATE UINT MCT Sched_AddRef(
     PISCHEDULER pThis
     );

PRIVATE UINT MCT Sched_Release (
     PISCHEDULER pThis
     );

PRIVATE PIACTIVITY MCT Sched_Reschedule(
    PISCHEDULER This,
    PACTINFO pActCurrent
    );

PRIVATE BOOL MCT Sched_DisablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     );

PRIVATE BOOL MCT Sched_EnablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     );

PRIVATE BOOL MCT Sched_PreemptionIsEnabled(
     PISCHEDULER This
     );

PRIVATE SCODE MCT
Sched_CreateConstraint(
    PISCHEDULER pThis,
    TIME pStartTime,
    TIME pPeriod,
    TIME pEstimate,
    UINT Recurrence,
    BOOL Hardness,
    BOOL AutoAdjust,
    BOOL Infinite,
    PIUNKNOWN* ppIUnknown
    );

PRIVATE SCODE MCT
Sched_SetConstraint(
    PISCHEDULER pThis,
    PTIME_CONSTRAINT pConstraint,
    PTIME pTimeTaken
    );

PRIVATE TIME MCT Sched_GetIdleTime(
    PISCHEDULER This
    );

PRIVATE INLINE void 
Sched_NormalizeTimeout(
    PTIME tTimeout
    );

#endif /* __sched_h__ */
