/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* These two constants control the max and min
 * size of a buffer cache block.  The first is the
 * amount of virtual memory allocated to a block,
 * the second the minimum amount of committed memory.
 */
#define MIN_BCACHE_BLOCK_SIZE 512

/* Lookup a block for reading
 */
extern SCODE BcacheLookupRead( PIFILE Device, DISK_ADDRESS Block, UINT BlkSize,
                               PTR *ppRet);
/* Lookup a block for writing
 */
extern SCODE BcacheLookupWrite( PIFILE Device, DISK_ADDRESS Block,
                                UINT BlkSize, PTR *ppRet);
/* Get buffer status bits
 */
extern UINT16 BcacheGetBufferState( PTR pBuf);

#define BLK_DIRTY 1
#define BLK_INVALID_DATA 2

/* Modify buffer status bits, add to them
 */
extern SCODE BcacheAddToBufferState( PTR pBuf, UINT16 Bits);

/* Modify buffer status bits, clear some of them
 */
extern SCODE BcacheRemoveFromBufferState( PTR pBuf, UINT16 Bits);

/* Get a pointer to the buffer's data
 */
extern BYTE *BcacheGetBufferData( PTR pBuf);

/* Which block is this a buffer for
 */
extern DISK_ADDRESS BcacheGetBufferBlockNo( PTR pBuf);

/* How big is the buffer
 */
extern UINT BcacheGetBufferSize( PTR pBuf);

/* Release a buffer. If dirty, ignore.
 */
extern void BcacheFlushBuffer( PTR pBuf, BOOL RidOfIt );

/* Release a buffer. If dirty, write back.
 */
extern void BcacheSpillBuffer( PTR pBuf, BOOL RidOfIt );

/* Release all buffer with a given tag (device)
 * Ignore if dirty.
 */
extern void BcacheFlushTag( UINT Tag );

/* Release all buffer with a given tag (device)
 * Write back if dirty.
 */
extern void BcacheSpillTag( UINT Tag );

/* Initialization, can be called multiple times.
 */
extern void BcacheInit(void);

/* Finalization, can be called multiple times.
 */
extern void BcacheDestroy(void);

/* Print interesting statistics.
 */
extern void BcacheStats(void);
