/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    fat16.c
 *
 * Purpose:
 *
 *    Constructor for a FAT12/16 filesystem driver
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <tchar.h>
//#define _DEBUG 1
#define DBGLVL 0
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>

extern PINAMESPACE RTLCALLTYPE InitFatFileSystem(const _TCHAR *DeviceName);

/* ** 
*/
SCODE __cdecl Fat16Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * DeviceName,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("Fat16Constructor(%x %s %x %x %d)\n"),
                     (ADDRESS) pThis, Name, (ADDRESS) ppUnk, DeviceName));

    pUnk = (PIUNKNOWN) InitFatFileSystem(DeviceName);

    DBGME(0,_tprintf(_T("Fat16Constructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("Fat16Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Fat16DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Fat16Constructor
};

static struct _DRIVERCOB Fat16DriverCobObject = { &Fat16DriverCobVtable, 1 };

PIUNKNOWN Fat16CobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Fat16DriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &Fat16DriverCobObject;
}
