/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <diagnostics.h>

extern BOOL IsLongTimeout;

/* BUGBUG per link ! */
struct _LINK_ERRORS {
    UINT LongTimeout;
    UINT BeforeSTX;
    UINT Retransmit;
    UINT FirstByteSum;
    UINT SecondByteSum;
    UINT OutOfSequence;
    UINT BadLength;
    UINT Reset;
} LinkErrors = {0,};
void PrintLinkErrors(struct _LINK_ERRORS *Errors)
{
#ifdef _DEBUG
    _tprintf(_T("{LE"));
    if (Errors->LongTimeout)
        _tprintf(_T(" timeo %d"), Errors->LongTimeout);
    if (Errors->BeforeSTX)
        _tprintf(_T(" stx %d"), Errors->BeforeSTX);
    if (Errors->Retransmit)
        _tprintf(_T(" rep %d"), Errors->Retransmit);
    if (Errors->FirstByteSum)
        _tprintf(_T(" 1st %d"), Errors->FirstByteSum);
    if (Errors->SecondByteSum)
        _tprintf(_T(" 2nd %d"), Errors->SecondByteSum);
    if (Errors->OutOfSequence)
        _tprintf(_T(" seq %d"), Errors->OutOfSequence);
    if (Errors->BadLength)
        _tprintf(_T(" len %d"), Errors->BadLength);
    if (Errors->Reset)
        _tprintf(_T(" rst %d"), Errors->Reset);
    _tprintf(_T("}\n"));
#endif
}

#define EXPAND_IN_PLACE 1

/* Put a packet out, obey protocol
 */
UINT PutPacket(LINK_HANDLE hLine, struct packet *Pkt, UINT BufferSize)
{
    UINT Size = Pkt->packet_length;
    UINT8 *pNext, Byte;
    UINT16 Sum;
    UINT nn = 0;

#if EXPAND_IN_PLACE
    /* Make a pass over data to (a) compute xsum (b) check we'll fit in
     */
    UINT nEscapes = 0;
    UINT8 *pSrc;

    pNext = (UINT8*)Pkt;
    Sum = 0;

    for (nn = 0; nn < Size; nn++) {
        Byte = *pNext++;
        Sum += Byte;
        if ((Byte == STH_CHAR) ||
            (Byte == EOT_CHAR) || 
            (Byte == ESC_CHAR)) {
            nEscapes++;
        }
        
    }

    /* Fix checksum on account of STX chars */
    if ((Sum & 0xff) == STH_CHAR)
        Sum++;
    if ((Sum >> 8) == STH_CHAR)
        Sum += (1 << 8);/* this must not rollover */

    /* Do we fit in ? Ifnot take the slow path
     */
    nn += 4 + nEscapes;
    if (nn > BufferSize)
        goto SlowPath;

    /* Expand, working backwards
     */
    pSrc  = ((UINT8*)Pkt) + Size - 1;
    pNext = ((UINT8*)Pkt) + nn - 1;

    /* Trailer */
    *pNext-- = (UINT8)(Sum>>8);
    *pNext-- = (UINT8)Sum;
    *pNext-- = EOT_CHAR;

    /* Data proper */
    while (Size-- > 0) {
        Byte = *pSrc--;
        *pNext-- = Byte;
        if ((Byte == STH_CHAR) ||
            (Byte == EOT_CHAR) ||
            (Byte == ESC_CHAR)) {
            *pNext-- = ESC_CHAR;
        }
    }

    /* Header */
    *pNext = STH_CHAR;

    /* Send them out
     */
    Petabytes(hLine,pNext,nn);

    /* Done we are
     */
    goto Done;

 SlowPath:
    Petabyte(hLine,STH_CHAR);
    nn = 1;

    pNext = (UINT8*)Pkt;
    for (;Size-- > 0;) {

        Byte = *pNext++;
        if ((Byte == STH_CHAR) ||
            (Byte == EOT_CHAR) || 
            (Byte == ESC_CHAR)) {
            Petabyte(hLine,ESC_CHAR);
            nn++;
        }
        Petabyte(hLine,Byte);
        nn++;
    }

    Petabyte(hLine,EOT_CHAR);
    nn++;

    Petabyte(hLine,(UINT8)Sum);
    nn++;
    Petabyte(hLine,(UINT8)(Sum>>8));
    nn++;
 Done:
#else

    pNext = (UINT8*)Pkt;
    Sum = 0;

    Petabyte(hLine,STH_CHAR);
    nn++;

    for (;Size-- > 0;) {

        Byte = *pNext++;
        Sum += Byte;
        if ((Byte == STH_CHAR) ||
            (Byte == EOT_CHAR) || 
            (Byte == ESC_CHAR)) {
            Petabyte(hLine,ESC_CHAR);
            nn++;
        }
        Petabyte(hLine,Byte);
        nn++;
    }

    Petabyte(hLine,EOT_CHAR);
    nn++;

    /* Fix checksum on account of STX chars */
    if ((Sum & 0xff) == STH_CHAR)
        Sum++;
    if ((Sum >> 8) == STH_CHAR)
        Sum += (1 << 8);/* this must not rollover */

    Petabyte(hLine,(UINT8)Sum);
    nn++;
    Petabyte(hLine,(UINT8)(Sum>>8));
    nn++;
#endif
    
    /* In case there is buffering down below
     */
    Pushout(hLine,nn);

#if !defined(__HOSTFS_CLIENT)
    DPRINT(("(%d)\n",nn));
#endif
    /* BUGBUG xmit error checking */
    return nn;
}

/* Pull a packet in, obey protocol
 */
UINT GetPacket(LINK_HANDLE hLine, struct packet *Pkt, UINT8 ExpectedSeqNo)
{
    UINT8 *pNext, Byte;
    UINT State;
    UINT16 Sum;
    BOOL Escape = FALSE;
    SCODE sc;

    LongTimeout(hLine);
 StartOfPacket:
    State = 0;
    Sum = 0;
    pNext = (UINT8*)Pkt;
    for (;;) {

        /* Look for input
         */
        sc = Getabyte(hLine,&Byte);

        /* On timeouts, only give up after a long wait
         */
        if (FAILED(sc)) {
            if (IsLongTimeout) {
                LinkErrors.LongTimeout++;
                return 0;
            }
            LongTimeout(hLine);
            continue;
        }

        /* Got a valid byte, see whatnot
         */
        switch (State) {

        case 0: /* Wait for STX */

            if ((Byte == STH_CHAR) && !Escape) {
                State = 1;/* get data now */
                ShortTimeout(hLine);

            /* else ignore till we get it */
            } else {
                Escape = (Byte == ESC_CHAR);
                LinkErrors.BeforeSTX++;

                DBGME(0,_tprintf(_T("SYN: x%x "),Byte));
            }

            break;

        case 1: /* Pull data in, end at EOT */

            if ((Byte == EOT_CHAR) && !Escape) {
                State++; /* done with data */
                /* Fix expected checksum on account of STX chars */
                if ((Sum & 0xff) == STH_CHAR)
                    Sum++;
                if ((Sum >> 8) == STH_CHAR)
                    Sum += (1 << 8);/* this must not rollover */

            } else {
                /* A good char, unless.. */
                if (!Escape && (Byte == STH_CHAR)) {
                    LinkErrors.Retransmit++;
                    DBGME(0,_tprintf(_T("Retransmission detected..\n")));
                    goto StartOfPacket;
                }
                if (Escape || (Byte != ESC_CHAR)) {
                    *pNext++ = Byte;
                    Sum += Byte;
                    Escape = FALSE;
                } else {
                    Escape = TRUE;
                }
            }
            break;

        case 2: /* get and verify checksum, first byte */

            if ((Sum & 0xff) == Byte)
                State++;
            else {
                LinkErrors.FirstByteSum++;
                DBGME(0,_tprintf(_T("Bad pkt checksum, 1st byte %x != %x [%x]\n"),
                       Byte, Sum & 0xff, Sum>>8));
#if defined(__HOSTFS_CLIENT)
                /* We saw the EOT, we expect no more than 2 bytes after that.
                 * ==> this one is the second checksum byte, then no more.
                 */
                return 0;
#else
                /* wait for retransmit
                 */
                LongTimeout(hLine);
                goto StartOfPacket;
#endif
            }
            break;

        case 3: /* get and verify checksum, second byte */

            if ((Sum >> 8) != Byte) {
                LinkErrors.SecondByteSum++;
                DBGME(0,_tprintf(_T("Bad pkt checksum, 2nd byte %x != %x\n"),
                       Byte, Sum & 0xff));
#if defined(__HOSTFS_CLIENT)
                /* We saw the EOT, and the first good checksum byte.
                 * ==> this one got garbled, and no more is coming.
                 */
                return 0;
#else
                /* wait for retransmit.
                 */
                LongTimeout(hLine);
                goto StartOfPacket;
#endif
            }

            /* NB: We cannot detect missing zeroes with the check-sum,
             * NB: should have shifted also for proper polinomial sum.
             * We'll check the number of bytes instead.
             */
            Sum = (UINT16)(pNext - (UINT8*)Pkt);
#if !defined(__HOSTFS_CLIENT)
            /* Client gets it in native byteorder, server maybe not
             */
            if (Sum != ByteSwap16(Pkt->packet_length))
#else
            if (Sum != Pkt->packet_length)
#endif
            {
                /* BUGBUG Server should send a NACK ! */
                LinkErrors.BadLength++;
                return 0;
            }

            /* Verify packet flow logic, check reconnection/target restart
             */
            if ((Pkt->seqno == ExpectedSeqNo) ||
                (Pkt->packet_type == PACKET_TYPE_RESET)) {
                /* Done */
                return pNext - (UINT8*)Pkt;
            }

            LinkErrors.OutOfSequence++;
#if !defined(__HOSTFS_CLIENT)
            _tprintf(_T("ORD: %d != %d\n"),Pkt->seqno, ExpectedSeqNo);
#endif

            /* Send a NACK and repeat
             */
            Pkt->packet_type = PACKET_TYPE_RESET;
            Pkt->seqno = ExpectedSeqNo;
            Pkt->sc = E_RESTART;

            /* Send back, fix length */
            Pkt->packet_length = MIN_PACKET_SIZE;
            (void) PutPacket(hLine, Pkt, pNext - (UINT8*)Pkt);

            LongTimeout(hLine);
            goto StartOfPacket;
        }
    }
    return 0;/*amusme*/
}

/* If the connection is reliable and b/w not an issue we can go naked
 */

/* Put a packet out
 */
UINT PutPacketRaw(LINK_HANDLE hLine, struct packet *Pkt, UINT BufferSize)
{
    UnusedParameter(BufferSize);
    Petabytes(hLine,(UINT8*)Pkt,Pkt->packet_length);
    return Pkt->packet_length;
}

/* Pull a packet in
 */
UINT GetPacketRaw(LINK_HANDLE hLine, struct packet *Pkt, UINT8 ExpectedSeqNo)
{
    UINT8 *pNext, Byte;
    UINT State, Length;
    SCODE sc;

    LongTimeout(hLine);
 StartOfPacket:
    State = 0;
    pNext = (UINT8*)Pkt;
    for (;;) {

        /* Look for input
         */
        sc = Getabyte(hLine,&Byte);

        /* On timeouts, only give up after a long wait
         */
        if (FAILED(sc)) {
            if (IsLongTimeout) {
                LinkErrors.LongTimeout++;
                return 0;
            }
            LongTimeout(hLine);
            continue;
        }

        /* Got a valid byte, see whatnot
         */
        ShortTimeout(hLine);
        *pNext++ = Byte;
        Length = pNext - (UINT8*)Pkt;

        switch (State) {

        case 0: /* Wait for packet_length to be known */

            if (Length >= 4)
                State = 1;/* get the rest of the data now */

            break;

        case 1: /* Pull data in, only stop at length reached */

#if !defined(__HOSTFS_CLIENT)
            /* Client gets it in native byteorder, server maybe not
             */
            if (Length < ByteSwap16(Pkt->packet_length))
                break;
#else
            if (Length < Pkt->packet_length)
                break;
#endif

            /* Verify packet flow logic, check reconnection/target restart
             */
            if ((Pkt->seqno == ExpectedSeqNo) ||
                (Pkt->packet_type == PACKET_TYPE_RESET)) {
                /* Done */
                return Length;
            }

            LinkErrors.OutOfSequence++;
#if !defined(__HOSTFS_CLIENT)
            _tprintf(_T("ORD: %d != %d\n"),Pkt->seqno, ExpectedSeqNo);
#endif

            /* Send a NACK and repeat
             */
            Pkt->packet_type = PACKET_TYPE_RESET;
            Pkt->seqno = ExpectedSeqNo;
            Pkt->sc = E_RESTART;

            /* Send back, fix length */
            Pkt->packet_length = MIN_PACKET_SIZE;
            (void) PutPacketRaw(hLine, Pkt, Length);

            LongTimeout(hLine);
            goto StartOfPacket;
        }
    }
    return 0;/*amusme*/
}
