/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Microsoft MMLITE June 1995 --jvh
 * NT server for remote IFile using serial lines or sockets.
 */

#include <import_windows.h>
#include <winsock.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <tchar.h>
#include <wchar.h>

#define __HOSTFS_SERVER 1
#include "hostfsp_old.h"

#if BYTESWAPEM
UINT16 ByteSwap16(UINT16 Value)
{
    Value
        = ((Value & 0x00ff) << 8)
        | ((Value & 0xff00) >> 8);

    return Value;
}

UINT32 ByteSwap32(UINT32 Value)
{
    Value
        = ((Value & 0x000000ff) << 24)
        | ((Value & 0x0000ff00) << 8)
        | ((Value & 0x00ff0000) >> 8)
        | ((Value & 0xff000000) >> 24);

    return Value;
}

UINT64 ByteSwap64(UINT64 Value, UINT64 *pRetValue)
{
    UINT32 l0, l1, h;

    l0 = ((LARGE_INTEGER *) &Value)->u.LowPart;
    h = ((LARGE_INTEGER *) &Value)->u.HighPart;

    l1 = ((h & 0x000000ff) << 24)
        | ((h & 0x0000ff00) << 8)
        | ((h & 0x00ff0000) >> 8)
        | ((h & 0xff000000) >> 24);

    h = ((l0 & 0x000000ff) << 24)
        | ((l0 & 0x0000ff00) << 8)
        | ((l0 & 0x00ff0000) >> 8)
        | ((l0 & 0xff000000) >> 24);

    ((LARGE_INTEGER *) &Value)->u.LowPart = l1;
    ((LARGE_INTEGER *) &Value)->u.HighPart = h;

    if (pRetValue)
        *pRetValue = Value;
    return Value;
}
#endif

/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 */
PRIVATE void StrNCpyTcharToWchar(wchar_t *pDst, const _TCHAR *pSrc, UINT Count)
{
    wchar_t c;

    while (--Count != 0) {
        c = (wchar_t) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst = 0;
}

/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 * COUNT should be > 1
 * Returns the number of bytes copied, including EOS.
 */
PRIVATE UINT StrNCpyWcharToTchar(_TCHAR *pDst, const wchar_t *pSrc, UINT Count)
{
    _TCHAR c;
    _TCHAR *pDstStart = pDst;

    while (--Count != 0) {
        c = (_TCHAR) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst++ = _EOS;
    return (pDst - pDstStart);
}


UINT hostfsd_debug_level = 0;

/* An object for NT files/dirs
 */
typedef struct _NTFILE {
    const struct IFileVtbl *v;
    UINT RefCnt;
    HANDLE Handle;
} *PNTFILE;
#define pNF(_ifile_) ((PNTFILE)(_ifile_))
#define iNF(_file_) ((PIFILE)(_file_))

typedef struct _NTNS {
    const struct INameSpaceVtbl *v;
    UINT RefCnt;
    HANDLE Handle;
} *PNTNS;
#define pNS(_ins_) ((PNTNS)(_ins_))
#define iNS(_ns_) ((PINAMESPACE)(_ns_))


INLINE PIFILE get_FILE(UINT32 client_says, PCONNECTION Conn)
{
    UnusedParameter(Conn);
    return (PIFILE) client_says;
}

INLINE UINT32 put_FILE(PIFILE file, PCONNECTION Conn, /* tmp in */ wchar_t *name )
{
    return (UINT32) file;
}

INLINE PINAMESPACE get_NS(UINT32 client_says, PCONNECTION Conn)
{
    UnusedParameter(Conn);
    return (PINAMESPACE) client_says;
}

INLINE UINT32 put_NS(PINAMESPACE ns, PCONNECTION Conn, /* tmp in */ wchar_t *name )
{
    return (UINT32) ns;
}

SCODE error_code_NT2MMLITE(SCODE sc, char *where)
{
    /* NT doesn't seem to be very consistent in what error codes it returns */
    if (sc > 0) {
        /* Assume this is a Win32 error. */
        sc = STATUS_WIN32_ERROR(sc);
    }
    return sc;
}


extern const struct IFileVtbl NtFileVtbl; /* forward */
extern const struct INameSpaceVtbl NtNsVtbl; /* ditto */

static PIFILE AllocNtFile(HANDLE handle)
{
    PNTFILE File;

    File = (PNTFILE) malloc(sizeof(*File));
    if (!File)
        return NULL;

    File->v = &NtFileVtbl;
    File->RefCnt = 1;
    File->Handle = handle;

    return iNF(File);
}

static PINAMESPACE AllocNtNs()
{
    PNTNS Ns;

    Ns = (PNTNS) malloc(sizeof(*Ns));
    if (!Ns)
        return NULL;

    Ns->v = &NtNsVtbl;
    Ns->RefCnt = 1;
    Ns->Handle = INVALID_HANDLE_VALUE; /* no such thing for dirs */

    return iNS(Ns);
}

PRIVATE UINT MCT NtFileAddRef(PIFILE This)
{
    DPRINT(("NtFileADDREF"));
    return AtomicInc(&pNF(This)->RefCnt);
}

PRIVATE UINT MCT NtNsAddRef(PINAMESPACE This)
{
    DPRINT(("NSaddref"));
    return AtomicInc(&pNS(This)->RefCnt);
}

PRIVATE SCODE MCT NtFileQueryInterface(PIFILE This, REFIID pIid, void* *ppNew)
{
#if 0
    if (IidCmp(&IID_IFile, pIid) || IidCmp(&IID_IUnknown, pIid)) {
        DPRINT(("NtFileQIok"));
        NtFileAddRef(This);
        *ppNew = This;
        return NO_ERROR;
    }
    DPRINT(("NtFileQIfail"));
    *ppNew = NULL;
#endif
    return E_NO_INTERFACE;
}

PRIVATE SCODE MCT NtNsQueryInterface(PINAMESPACE This, REFIID pIid, void* *ppNew)
{
#if 0
    if (IidCmp(&IID_INameSpace, pIid) || IidCmp(&IID_IUnknown, pIid)) {
        DPRINT(("NSqiok"));
        NetFileNSAddRef((PINAMESPACE) This);
        *ppNew = This;
        return NO_ERROR;
    }
    DPRINT(("NSqifail"));
    *ppNew = NULL;
#endif
    return E_NO_INTERFACE;
}

PRIVATE UINT MCT NtFileRelease(PIFILE pThis)
{
    SCODE sc = SUCCESS;
    PNTFILE File = pNF(pThis);
    INT Refs;
    BOOL ok;
    UINT State;

    Refs = AtomicDec(&File->RefCnt);
    DPRINT(("NtFileRELE%d", Refs));
    if (Refs == 0) {
        State = HostFsd_NTBegin();
        ok = CloseHandle(File->Handle);
        if (!ok) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NtFileRelease(%x) failed sc=%x Handle=%x\n",
                   File, sc, File->Handle));
        } else {
            HostFsd_NTEnd(State);
        }
        free(File);
    }

    return Refs;
}

PRIVATE UINT MCT NtNsRelease(PINAMESPACE pThis)
{
    PNTNS Ns = pNS(pThis);
    INT Refs;

    Refs = AtomicDec(&Ns->RefCnt);
    DPRINT(("NtNsRELE%d", Refs));
    if (Refs == 0) {
        free(Ns);
    }

    return Refs;
}

PRIVATE SCODE MCT NtFileGetSize(PIFILE This, PUINT64 pSize);

PRIVATE SCODE MCT
NtFileReadAt(PIFILE This, UINT64 Position, 
             BYTE *Buffer, UINT ByteCount, UINT * pSizeRead)
{
    PNTFILE File = pNF(This);
    SCODE sc = SUCCESS, sc1;
    OVERLAPPED ol;              /* NT type */
    UINT64 Size64;
    BOOL r;
    UINT State;

    *(UINT64*)&ol.Offset = Position;
    ol.hEvent = NULL;

    DPRINT(("read pos %I64x (x%x.%x) count x%x file %x\n",
            Position, ol.OffsetHigh, ol.Offset, ByteCount, (UINT32) File));

    State = HostFsd_NTBegin();
    r = ReadFile(File->Handle, Buffer, ByteCount, pSizeRead, &ol);
    if (r == FALSE)
        sc = GetLastError();
    HostFsd_NTEnd(State);
        
    if ((sc == SUCCESS) && (*pSizeRead < ByteCount))
        sc = S_FALSE;
    else
    if (sc == ERROR_HANDLE_EOF) {
        /* at or past EOF ? we need to know */
        sc1 = NtFileGetSize(This, &Size64);
        sc = S_FALSE;
        if (!FAILED(sc1) && (Position > Size64))
            sc = E_INVALID_PARAMETER;
    }
    else
        sc = error_code_NT2MMLITE(sc, "NtFileReadAt");

    DPRINT(("read(x%x) => x%x  sc=x%x\n",
            (UINT32) File, *pSizeRead, sc));

    return sc;
}

PRIVATE SCODE MCT
NtFileWriteAt(PIFILE This, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, UINT *pSizeWritten)
{
    SCODE sc = SUCCESS;
    PNTFILE File = pNF(This);
    OVERLAPPED ol;              /* NT type */
    UINT State;

    *(UINT64*)&ol.Offset = Position;
    ol.hEvent = NULL;

    DPRINT(("write pos %I64x (x%x.%x) count x%x file %x\n",
            Position, ol.OffsetHigh, ol.Offset, ByteCount, (UINT32) File));

    State = HostFsd_NTBegin();
    if (WriteFile(File->Handle, Buffer, ByteCount, pSizeWritten, &ol) ==FALSE)
        sc = GetLastError();
    HostFsd_NTEnd(State);
        
    DPRINT(("write(x%x) => x%x  sc=x%x\n",
            (UINT32) File, *pSizeWritten, sc));

    return error_code_NT2MMLITE(sc, "NtFileWriteAt");
}

/* Get/Set a file's size
 */
PRIVATE SCODE MCT NtFileGetSize(PIFILE This, PUINT64 pSize)
{
    SCODE sc = SUCCESS;
    PNTFILE File = pNF(This);
    DWORD sizeLow, sizeHigh;
    UINT State;

    sizeHigh = -1;
    State = HostFsd_NTBegin();
    sizeLow = GetFileSize(File->Handle, &sizeHigh);
    if (sizeLow == 0xffffffff) {
        sc = GetLastError();
        HostFsd_NTEnd(State);
    } else {
        HostFsd_NTEnd(State);
        *pSize = (UINT64)(sizeLow) | (((UINT64)sizeHigh) << 32);
    }
    return error_code_NT2MMLITE(sc, "NtFileGetSize");
}

PRIVATE SCODE MCT NtFileSetSize(PIFILE This, UINT64 Size)
{
    SCODE sc = SUCCESS;
    PNTFILE File = pNF(This);
    UINT State;

    /* We do not truncate files (should ?)
     */
    sc = STATUS_WIN32_ERROR(  87);

    /* Cache probe ?
     */
    if (Size == (UINT64)(~0)) {
        EPRINT(("simfiled: Cache probe unimplemented\n"));
#if 0
        if (Conn->CacheTimestamp != CacheTimestamp) {
            sc = /*E_MEDIA_CHANGED*/ STATUS_WIN32_ERROR(1110);
            Conn->CacheTimestamp = CacheTimestamp;
        }
#endif
    } else {
        State = HostFsd_NTBegin();
        /* First seek to the right position and then set the size */
        if(SetFilePointer(File->Handle, (UINT32) Size, NULL, FILE_BEGIN)
           == 0xffffffff)
            sc = GetLastError();
        else if (!SetEndOfFile(File->Handle))
            sc = GetLastError();
        else
            sc = S_OK;
        HostFsd_NTEnd(State);
        sc = error_code_NT2MMLITE(sc, "NtFileSetSize");
    }
    return sc;
}

PRIVATE SCODE MCT Unimp()
{
    return E_NOT_IMPLEMENTED;
}

const struct INameSpaceVtbl NtNsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NtNsQueryInterface,
    NtNsAddRef,
    NtNsRelease,
    (_INameSpace_Register_METHOD_TYPE)Unimp,
    (_INameSpace_Unregister_METHOD_TYPE)Unimp,
    (_INameSpace_Bind_METHOD_TYPE)Unimp,
    (_INameSpace_FindNext_METHOD_TYPE)Unimp,
    (_INameSpace_GetCapabilities_METHOD_TYPE)Unimp
};

const struct IFileVtbl NtFileVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NtFileQueryInterface, 
    NtFileAddRef, 
    NtFileRelease, 
    NtFileReadAt, 
    NtFileWriteAt, 
    NtFileSetSize, 
    NtFileGetSize
};


/*------------------ "methods" ------------------------------*/
#if BYTESWAPEM
void ByteSwapUnicodeString(wchar_t *s)
{
    for (; *s != 0; s++)
        *s = ByteSwap16(*s);
}
#else
#define ByteSwapUnicodeString(_s_)
#endif

/* By default writable when used in NTU but not standalone */
BOOL hostfsd_may_write = TRUE;

void do_open(struct packet *p, PCONNECTION Conn)
{
    struct open_req *req = &p->u.open_req;
    struct open_rep *rep = &p->u.open_rep;
    SCODE sc = SUCCESS;
    UINT flags = ByteSwap32(req->flags);
    UINT openflags = 0, createflags = 0, Handle;
    wchar_t *filename;
    wchar_t namebuf[MAX_PATH];
    STRING_ESCAPE *Check;
    BOOL isdir = FALSE;
    HANDLE file = NULL;
    UINT State;

    Check = (STRING_ESCAPE*)req->path;
    if (Check->Escape == UNICODE_NOT_A_CHARACTER) {
        DPRINT(("ASCII "));
        filename = namebuf;
        StrNCpyTcharToWchar(filename,(_TCHAR*)Check->String,MAX_PATH);
    } else  {
        filename = (wchar_t*)req->path;/* done inplace */
        ByteSwapUnicodeString(filename);
    }

    if (flags & NAME_SPACE_FAILIFEXIST)         /* 0x20 */
        createflags = CREATE_NEW;
    else if (flags & NAME_SPACE_CREATE)         /* 0x10 */
        createflags = OPEN_ALWAYS;
    else
        createflags = OPEN_EXISTING;

    if (flags & NAME_SPACE_READ)                /* 1 */
        openflags |= GENERIC_READ;
    if (flags & NAME_SPACE_WRITE)               /* 2 */
        openflags |= GENERIC_WRITE;

    if (flags & NAME_SPACE_EXTENSION) {         /* mkdir */
        isdir = TRUE;
        if (!(openflags & GENERIC_WRITE)) {
            DPRINT(("Can't create directory without NAME_SPACE_WRITE\n"));
            sc = STATUS_WIN32_ERROR(5);
            goto replystuff;
        }
    }

    if (!hostfsd_may_write
        && (openflags != GENERIC_READ || createflags != OPEN_EXISTING))
    {
        DPRINT(("Rejected \"%ls\" flags=x%x because server is Read Only\n",
                   filename, flags));
        sc = E_ACCESS_DENIED;
        goto replystuff;
    }

    /* Check for \\.\ escape */
    if (wcsncmp(filename, L"....", 4) == 0) {
        filename[0] = filename[1] = filename[3] = '\\';
        filename[2] = '.';
    }

    Handle = 0;
    if (isdir) {
        State = HostFsd_NTBegin();
        /* In the end it might be better to have a separate RPC for mkdir */
        if (CreateDirectoryW(filename, NULL) == FALSE) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NOT! "));
        } else {
            HostFsd_NTEnd(State);
            Handle = put_NS(AllocNtNs(), Conn, filename);
        }
    } else {
        State = HostFsd_NTBegin();
        file = CreateFileW(filename,
                           openflags,
                           FILE_SHARE_READ|FILE_SHARE_WRITE,
                           NULL,
                           createflags,
                           FILE_ATTRIBUTE_NORMAL,
                           NULL);
        if (file == INVALID_HANDLE_VALUE) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NOT! "));
        } else {
            HostFsd_NTEnd(State);
            Handle = put_FILE(AllocNtFile(file), Conn, filename);
        }
    }
    EPRINT(("Opened \"%ls\" %s as file x%x handle x%x sc=x%x flags=x%x\n",
               filename,
               (openflags & GENERIC_WRITE)
               ? (isdir ? "New directory" : "WRITABLE")
               : "read only",
               (UINT) Handle, file, sc, flags));
 replystuff:
    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_OPEN_REP);
    p->sc = error_code_NT2MMLITE(sc, "do_open");
    rep->handle = Handle;
    packet_prepare(p, sizeof(*rep));
}

void do_delete(struct packet *p, PCONNECTION Conn)
{
    struct delete_req *req = &p->u.open_req;/*overlap, same reqs */
    BOOL r;
    SCODE sc = SUCCESS;
    wchar_t *objectname;
    wchar_t namebuf[MAX_PATH];
    STRING_ESCAPE *Check;
    UINT State;

    Check = (STRING_ESCAPE*)req->path;
    if (Check->Escape == UNICODE_NOT_A_CHARACTER) {
        DPRINT(("ASCII "));
        objectname = namebuf;
        StrNCpyTcharToWchar(objectname,(_TCHAR*)Check->String,MAX_PATH);
    } else  {
        objectname = (wchar_t*)req->path;/* done inplace */
        ByteSwapUnicodeString(objectname);
    }

    if (!hostfsd_may_write) {
        r = FALSE;
        sc = E_ACCESS_DENIED;
        goto Done;
    }

    /* try first to delete as file, then as dir.
     * yes, this mangles error reporting a bit.
     */

    State = HostFsd_NTBegin();
    r = DeleteFileW(objectname);
    if (!r) {
        sc = GetLastError();
        if (sc == ERROR_ACCESS_DENIED) {
            r = RemoveDirectoryW(objectname);
            if (!r)
                sc = GetLastError();
            else
                sc = SUCCESS;
        }
    }
    HostFsd_NTEnd(State);

 Done:
    EPRINT(("%sDeleted \"%ls\" sc=x%x\n",
            (r) ? "" : "NOT! ",
            objectname,
            sc));

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_DELETE_REP);
    p->sc = error_code_NT2MMLITE(sc, "do_delete");
    packet_prepare(p, 0);
}

void do_close(struct packet *p, PCONNECTION Conn)
{
    struct close_req *req = &p->u.close_req;
    PIFILE nf;
    SCODE sc = SUCCESS;

    DPRINT(("do_close handle x%x\n", req->handle));
    nf = get_FILE(req->handle, Conn);

    NtFileRelease(nf);

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_CLOSE_REP);
    p->sc = error_code_NT2MMLITE(sc, "do_close");
    packet_prepare(p, 0);
}

void do_getsize(struct packet *p, PCONNECTION Conn)
{
    struct getsize_req *req = &p->u.getsize_req;
    struct getsize_rep *rep = &p->u.getsize_rep;
    PIFILE File;
    UINT64 Size64;

    File = get_FILE(req->handle, Conn);
    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_GETSIZE_REP);

    p->sc = NtFileGetSize(File, &Size64);

    DPRINT(("getsize(x%x) => %lx (x%x.%x)  sc=x%x\n", (UINT32)File, Size64, Size64, p->sc));

    /* compilerbug->byref */
    ByteSwap64(Size64,&rep->size);

    packet_prepare(p, sizeof(*rep));
}

/* Partial implementation: if size is set to -1, check if floppy changed */
void do_setsize(struct packet *p, PCONNECTION Conn)
{
    struct setsize_req *req = &p->u.setsize_req;
    HANDLE file;
    SCODE sc = SUCCESS;

    file = pNF(get_FILE(req->handle, Conn))->Handle;

    /* We do not truncate files (should ?)
     */
    sc = STATUS_WIN32_ERROR(  87);

    /* Cache probe ?
     */
    if (req->size == (UINT64)(~0)) {
        EPRINT(("simfiled: Cache probe unimplemented\n"));
#if 0
        if (Conn->CacheTimestamp != CacheTimestamp) {
            sc = /*E_MEDIA_CHANGED*/ STATUS_WIN32_ERROR(1110);
            Conn->CacheTimestamp = CacheTimestamp;
        }
#endif
    } else {
        UINT State = HostFsd_NTBegin();
        if(SetFilePointer(file, (UINT32) req->size, NULL, FILE_BEGIN)
           == 0xffffffff)
            sc = GetLastError();
        else if (!SetEndOfFile(file))
            sc = GetLastError();
        else
            sc = S_OK;
        HostFsd_NTEnd(State);
    }
    
    DPRINT(("setsize(x%x) => x%x  sc=x%x\n", (UINT32) file, req->size, sc));
    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_SETSIZE_REP);
    p->sc = error_code_NT2MMLITE(sc, "do_setsize");
    packet_prepare(p, 0);
}

void do_read(struct packet *p, PCONNECTION Conn)
{
    struct read_req *req = &p->u.read_req;
    struct read_rep *rep = &p->u.read_rep;
    PIFILE File;
    SCODE sc = SUCCESS;
    UINT count, n;
    UINT64 Pos;

    File = get_FILE(req->handle, Conn);
    count = ByteSwap32(req->count);
    ByteSwap64(req->position, &Pos);

    if (count > MAX_READ_SIZE)
        count = MAX_READ_SIZE;

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_READ_REP);

    n = 0;
    p->sc = NtFileReadAt(File, Pos, rep->data, count, &n);

    rep->nread = ByteSwap32(n);
    packet_prepare(p, sizeof(*rep) - (MAX_READ_SIZE - n));
}

void do_write(struct packet *p, PCONNECTION Conn)
{
    struct write_req *req = &p->u.write_req;
    struct write_rep *rep = &p->u.write_rep;
    PIFILE File;
    SCODE sc = SUCCESS;
    UINT count, n;
    UINT64 Pos;

    File = get_FILE(req->handle, Conn);
    count = ByteSwap32(req->count);
    ByteSwap64(req->position, &Pos);

    sc = NtFileWriteAt(File, Pos, req->data, count, &n);

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_WRITE_REP);
    p->sc = sc;
    rep->nwritten = ByteSwap32(n);
    packet_prepare(p, sizeof(*rep));
}

/* 
 */
SCODE findnext_helper(const wchar_t *prefix, const wchar_t *previous,
                                                      WIN32_FIND_DATAW *search)
{
    wchar_t path[MAX_PATH * 2];
    HANDLE dir;
    SCODE sc = SUCCESS;
    UINT State;

    wcscpy(path,prefix);
    if (prefix[0] != L'\0')
        wcscat(path,L"\\");
    wcscat(path,L"*");

    /* Assume error */
    sc = error_code_NT2MMLITE(ERROR_FILE_NOT_FOUND,"do_findnext");

    State = HostFsd_NTBegin();
    dir = FindFirstFileW(path,search);
    DPRINT(("findnext_helper FFW %s->%x %x\n", path, dir, GetLastError()));

    if (dir == INVALID_HANDLE_VALUE) {
        HostFsd_NTEnd(State);
        return sc;
    }

    if (previous[0] == L'\0')  /* This is a FindFirst */
        sc = S_OK;
    else {
        do {
            if (wcscmp(previous,search->cFileName) == 0) { /* Found previous */
                WIN32_FIND_DATAW Search2;
                if (FindNextFileW(dir,search)) { /* Found next */
                    /* FindNext is supposed to return S_FALSE when it returns the last entry
                     * See if this is the last one and if so return S_FALSE + name
                     */
                    Search2 = *search; /* make copy so as to not mess up returned values */
                    if (FindNextFileW(dir, &Search2))
                        sc = S_OK;
                    else
                        sc = S_FALSE;
                } else {                          /* Previous was last */
                    /* Search beyond end. Say this is last but return empty string */
                    search->cFileName[0] = L'\0';
                    sc = S_FALSE;
                }
                break;
            }
        } while (FindNextFileW(dir,search));
    }

    FindClose(dir);
    HostFsd_NTEnd(State);
    return sc;
}

void do_findnext(struct packet *p, PCONNECTION Conn)
{
    struct findnext_req *req = &p->u.findnext_req;
    struct findnext_rep *rep = &p->u.findnext_rep;
    SCODE sc = SUCCESS;
    WIN32_FIND_DATAW SearchData;
    wchar_t namebuf1[MAX_PATH], namebuf2[MAX_PATH];
    wchar_t *prefix, *previous;
    STRING_ESCAPE *Check;
    BOOL IsAscii;
    UINT len; 

    req->offset_of_previous = ByteSwap16(req->offset_of_previous);

    /* If the first string is ASCII then the second is also
     */
    Check = (STRING_ESCAPE*)req->prefix;
    IsAscii = (Check->Escape == UNICODE_NOT_A_CHARACTER);
    if (IsAscii) {
        DPRINT(("ASCII "));
        prefix = namebuf1;
        StrNCpyTcharToWchar(prefix, (_TCHAR*) Check->String,MAX_PATH);
        previous = namebuf2;
        StrNCpyTcharToWchar(previous,
                            (_TCHAR*)Check->String+req->offset_of_previous,
                            MAX_PATH);
    } else  {
        prefix = (wchar_t*)req->prefix;/* done inplace */
        ByteSwapUnicodeString(prefix);
        previous = &prefix[req->offset_of_previous];
        ByteSwapUnicodeString(previous);
    }

    sc = findnext_helper(prefix,previous,&SearchData);
    DPRINT(("do_findnext(\"%ls\", \"%ls\") -> %x \"%ls\"\n",
      prefix, previous, sc, SearchData.cFileName));

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_FINDNEXT_REP);

    /* Copy path to reply msg. */
    if (IsAscii) {
        /* Set Escape, make sure answer doesnt overflow, xlate
         */
        Check = (STRING_ESCAPE*)rep->name;
        Check->Escape = (_TCHAR) UNICODE_NOT_A_CHARACTER;
        len = StrNCpyWcharToTchar((_TCHAR*)Check->String,SearchData.cFileName,MAX_PATH);
        len += offsetof(STRING_ESCAPE,String);
    } else {
        /* Make sure no overflow */
        len = (wcslen(SearchData.cFileName) + 1) * sizeof(wchar_t);
        if (len > MAX_PATH)
            len = MAX_PATH;
        memcpy(prefix,SearchData.cFileName,len);
        prefix[MAX_PATH-1] = _EOS;
        ByteSwapUnicodeString(prefix);
    }

    p->sc = sc;
    packet_prepare(p, len);
}

void do_reset(struct packet *p, PCONNECTION Conn)
{
    EPRINT(("Client resets connection (%d -> %d+1)\n",
           Conn->SequenceNumber, p->seqno));
    
    Conn->SequenceNumber = p->seqno; /* expected */

    /* Reuse packet for reply */
    packet_init(p, PACKET_TYPE_RESET);
    
    packet_prepare(p,0);
}

/* 
 * Use one thread per socket.  One socket has only one transaction
 * going on at a time.
 * Windowing etc. is achieved by using multiple sockets.
 */

UINT HostFile_Server(void *handle, struct packet *p, UINT Size)
{
    PCONNECTION Conn = (PCONNECTION)handle;
    UINT16 l;

    p->sc = ByteSwap32(p->sc);

    switch (p->packet_type) {
    case PACKET_TYPE_RESET:
        do_reset(p, Conn);
        break;
    case PACKET_TYPE_OPEN_REQ:
        do_open(p, Conn);
        break;
    case PACKET_TYPE_CLOSE_REQ:
        do_close(p, Conn);
        break;
    case PACKET_TYPE_GETSIZE_REQ:
        do_getsize(p, Conn);
        break;
    case PACKET_TYPE_SETSIZE_REQ:
        do_setsize(p, Conn);
        break;
    case PACKET_TYPE_READ_REQ:
        do_read(p, Conn);
        break;
    case PACKET_TYPE_WRITE_REQ:
        do_write(p, Conn);
        break;
    case PACKET_TYPE_FINDNEXT_REQ:
        do_findnext(p, Conn);
        break;
    case PACKET_TYPE_DELETE_REQ:
        do_delete(p, Conn);
        break;
    default:
        EPRINT(("Unknown packet type x%x\n", p->packet_type));
        p->sc = 0x88888888; 
    }

    l = p->packet_length;
    p->packet_length = ByteSwap16(p->packet_length);
    p->sc = ByteSwap32(p->sc);

    return l;
}
