/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <import_windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#define MYSIZE (64*1024)
#define TRACE  printf

BOOL NextRequest(HANDLE h, char *lpBuffer, DWORD *lpSize, LPOVERLAPPED lpOver);
BOOL NextReply(HANDLE h, char *lpBuffer, DWORD *lpSize, LPOVERLAPPED lpOver);

void DoClient(HANDLE h)
{
    OVERLAPPED ReadOver, WriteOver;
    char *lpReadBuffer, *lpWriteBuffer;
    DWORD BytesRead, BytesWritten, err, OpsPosted;
    char *op;
    BOOL r, GotReply;
    HANDLE Events[2];
#define evRead 0
#define evWrite 1

    memset(&ReadOver,0,sizeof ReadOver);
    ReadOver.hEvent = 
        Events[evRead] = 
            CreateEvent(NULL,FALSE,FALSE,NULL);

    memset(&WriteOver,0,sizeof WriteOver);
    WriteOver.hEvent = 
        Events[evWrite] =
            CreateEvent(NULL,FALSE,FALSE,NULL);

    lpReadBuffer = VirtualAlloc(NULL,
                            MYSIZE,
                            MEM_COMMIT,
                            PAGE_READWRITE);

    lpWriteBuffer = VirtualAlloc(NULL,
                            MYSIZE,
                            MEM_COMMIT,
                            PAGE_READWRITE);

 AllOverAgain:
    op = "ConnectNamedPipe";
    r = ConnectNamedPipe(h,
                         &ReadOver);
    if (r) goto Done; // should fail!

    TRACE("Pipe not yet connected ...\n");

    op = "GetOverlappedResult";
    r = GetOverlappedResult(h,
                            &ReadOver,
                            &BytesRead,
                            TRUE);
    if (!r) goto Done;

    TRACE("Pipe connected ...\n");

    OpsPosted = 0;
    GotReply = TRUE;
    for (;;) {
        if ( ! (OpsPosted & (1 << evRead))) {
            op = "ReadFile";
            r = ReadFile(h,
                         lpReadBuffer,
                         MYSIZE,
                         &BytesRead,
                         &ReadOver);
            if (r) goto GotSomeRead;

            err = GetLastError();
            switch (err) {
            case ERROR_HANDLE_EOF:
            case ERROR_BROKEN_PIPE:
                TRACE("Client is gone..");

                op = "DisconnectNamedPipe";
                r = DisconnectNamedPipe(h);
                if (!r) goto Done;

            case ERROR_PIPE_LISTENING:
                // get back to the beginning..
                    TRACE("Waiting for new connection...\n");
                goto AllOverAgain;
                break;

            case ERROR_IO_PENDING:
                break;

            default: goto Done;
            }
            OpsPosted |= (1 << evRead);
        }

        if ( GotReply && ! (OpsPosted & (1 << evWrite))) {

            GotReply = FALSE;
            op = "WriteFile";
            r = NextRequest(h,lpWriteBuffer,&BytesWritten,&WriteOver);

            if (r) goto GotSomeWritten;

            err = GetLastError();
            switch (err) {
            case ERROR_HANDLE_EOF:
            case ERROR_BROKEN_PIPE:
                TRACE("Client is gone..");

                op = "DisconnectNamedPipe";
                r = DisconnectNamedPipe(h);
                if (!r) goto Done;

            case ERROR_PIPE_LISTENING:
                // get back to the beginning..
                TRACE("Waiting for new connection...\n");
                goto AllOverAgain;
                break;

            case ERROR_IO_PENDING:
                break;

            default: goto Done;
            }
            OpsPosted |= (1 << evWrite);
        }

        op = "WaitForMultiple";
        err = WaitForMultipleObjects(2,Events,FALSE,INFINITE);
        if (err == WAIT_FAILED) goto Done;

        if (err == evRead) {
            op = "GetOverlappedResult(r)";
            if (GetOverlappedResult(h,
                                    &ReadOver,
                                    &BytesRead,
                                    TRUE)) {        /* wait */
            GotSomeRead:            
                OpsPosted &= ~(1 << evRead);
#if 0
                printf("Clientgot(%d)\n", BytesRead);
#endif
                GotReply = NextReply(h,lpReadBuffer,&BytesRead,&ReadOver);
            } else {
                err = GetLastError();
                if (err == ERROR_BROKEN_PIPE) {
                    // there might still be data pending ?
                } else
                    break;
            }
        } else if (err == evWrite) {
            op = "GetOverlappedResult(w)";
            if (GetOverlappedResult(h,
                                    &WriteOver,
                                    &BytesWritten,
                                    TRUE)) {        /* wait */
            GotSomeWritten:
                OpsPosted &= ~(1 << evWrite);
                printf("\nWrite completed..\n");
            } else {
                err = GetLastError();
                if (err == ERROR_BROKEN_PIPE) {
                    // there might still be data pending ?
                } else
                    break;
            }
        }
    }

 Done:
    printf("Last %s failed with err = %d\n", op, GetLastError());

    VirtualFree(lpReadBuffer,MYSIZE,MEM_DECOMMIT);
    CloseHandle(ReadOver.hEvent);
    CloseHandle(WriteOver.hEvent);
                      
}

int _tmain(int argc, _TCHAR **argv)
{
    HANDLE np;
    _TCHAR PipeName[256];
#define TRYSEC 1
#if TRYSEC
    SECURITY_ATTRIBUTES saPipe;
#endif
    LPSECURITY_ATTRIBUTES lpSec;

    if (argc < 2) {
        printf("Usage: %s <NamedPipeName>\n", argv[0]);
        exit(0);
    }

    _stprintf(PipeName, _T("\\\\.\\pipe\\%s"), argv[1]);
    TRACE("pipe is %s\n", PipeName);

#if TRYSEC
    saPipe.lpSecurityDescriptor =
        (PSECURITY_DESCRIPTOR) malloc(SECURITY_DESCRIPTOR_MIN_LENGTH);
    InitializeSecurityDescriptor(saPipe.lpSecurityDescriptor,
                                 SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl(saPipe.lpSecurityDescriptor,
                              TRUE, (PACL) NULL, FALSE);
    saPipe.nLength = sizeof(saPipe);
    saPipe.bInheritHandle = TRUE;
    lpSec = &saPipe;
#else
    lpSec = NULL;
#endif
    np = CreateNamedPipe(PipeName,
                         PIPE_ACCESS_DUPLEX |
                         FILE_FLAG_OVERLAPPED,
                         PIPE_TYPE_BYTE |
                         PIPE_READMODE_BYTE |
                         PIPE_WAIT,
                         PIPE_UNLIMITED_INSTANCES,
                         MYSIZE,     // output buffer size
                         MYSIZE,     // input buffer size
                         INFINITE,   // default timeout
                         lpSec       // security attrib
                         );
    if (np == INVALID_HANDLE_VALUE) {
        printf("CreateNamedPipe(%s) failed with err = %d\n", 
               PipeName, GetLastError());
        exit(0);
    }

    DoClient(np);

    TRACE("done.\n");
    CloseHandle(np);
    return 0;
}

/* This is the part that really tests things
 */
#include <mmlite.h>
#include "hostfsp_old.h"
/* ------------ Loose ends from netmisc.c ----------------- */
#define DPRINT(x) printf x

PRIVATE UINT StrLenWchar(const wchar_t *pStr)
{
    UINT Count = 0;

    while (*pStr++ != _EOS)
        Count++;

    return Count;
}

/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 */
PRIVATE void StrNCpyWcharToTchar(char *pDst, const wchar_t *pSrc, UINT Count)
{
    char c;

    while (--Count != 0) {
        c = (char) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst = _EOS;
}

/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 */
PRIVATE void StrNCpyTcharToWchar(wchar_t *pDst, const _TCHAR *pSrc, UINT Count)
{
    wchar_t c;

    while (--Count != 0) {
        c = (wchar_t) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst = 0;
}

PRIVATE SCODE ComposePathToTchar(_TCHAR *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name)
{
    _TCHAR *s;
    UINT len;

    len = _tcslen(Prefix) + _tcslen(Name) + 2;
    if (len > MaxLen) {
        /* So what should we do? */
        DPRINT(("ComposePath \"%s\" \"%s\" too long %x > %x\n",
                 Prefix, Name, len, MaxLen));
        return E_INVALID_PARAMETER; /* reject it altogether */
    }

    /* Combine Prefix and Name  "Prefix/Name" or "Prefix" or "Name" or "" */
    _tcscpy(NewPath, Prefix);
    if (Prefix[0] != _EOS && Name[0] != _EOS)
        _tcscat(NewPath, _T("\\")); /* NT separator */
    _tcscat(NewPath, Name);

    /* Remove trailing slash */
    len = _tcslen(NewPath);
    if (len > 0 && IS_PATH_SEPARATOR(NewPath[len - 1]))
        NewPath[len - 1] = _EOS;

    /* Convert slashes to backslashes for NT */
    for (s = NewPath; s[0] != _EOS; s++)
        if (IS_PATH_SEPARATOR(s[0]))
            s[0] = _T('\\');    /* NT separator */
    DPRINT(("ComposePath \"%s\" \"%s\" -> \"%s\" sizoef(*p)=%x\n", Prefix, Name, NewPath, sizeof(struct packet)));
    return NO_ERROR;
}

PRIVATE SCODE ComposePathToWchar(wchar_t *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name)
{
    _TCHAR Buffer[MAX_PATH * sizeof(char)];
    SCODE sc;

    DPRINT(("ToWchar"));
    sc = ComposePathToTchar(Buffer, MAX_PATH, Prefix, Name);
    if (FAILED(sc))
        return sc;

    StrNCpyTcharToWchar(NewPath, Buffer, MaxLen);
    return sc;
}

int step = 0;
int enbr = 0;

BOOL NextRequest(HANDLE h, char *lpBuffer, DWORD *lpSize, LPOVERLAPPED lpOver)
{
    BOOL r;
    UINT8 *pNext, Byte, *pSrc;
    UINT16 Sum;
    int i;
    STRING_ESCAPE *As;
    struct packet Pkt;
    
    pNext = (UINT8*)lpBuffer;
    *pNext++ = STH_CHAR;

    switch (step) {
    case 0:
        /* Connect */
        Pkt.packet_type = PACKET_TYPE_RESET;
        Pkt.seqno = 0;
        Pkt.packet_length = 8;
        Pkt.sc = 0;
        enbr = 13;
        break;
    case 1:
        /* Open using a UNICODE path */
        packet_init(&Pkt, PACKET_TYPE_OPEN_REQ);
        Pkt.u.open_req.flags = NAME_SPACE_READ;
        ComposePathToWchar((wchar_t*)Pkt.u.open_req.path, MAX_PATH,
                           _T("./"), _T("test.dat"));
        packet_prepare(&Pkt, sizeof(Pkt.u.open_req));
        enbr = 16;
        break;
    case 2:
        /* Open using an ASCII path */
        packet_init(&Pkt, PACKET_TYPE_OPEN_REQ);
        Pkt.u.open_req.flags = NAME_SPACE_READ;
        As = (STRING_ESCAPE*)Pkt.u.open_req.path;
        As->Escape = (_TCHAR) UNICODE_NOT_A_CHARACTER;
        ComposePathToTchar((_TCHAR*) As->String, MAX_PATH, _T("./"), _T("test.dat"));
        packet_prepare(&Pkt, sizeof(Pkt.u.open_req));
        enbr = 16;
        break;
    default:
        printf("Done testing, PASS.\n");
Sleep(10000);
        exit(0);
    }
    step++;

    Sum = 0;
    pSrc = (UINT8*)&Pkt;
    for (i = 0; i < Pkt.packet_length; i++) {
        Byte = *pSrc++;
        Sum += Byte;
        if ((Byte == STH_CHAR) ||
            (Byte == EOT_CHAR) || 
            (Byte == ESC_CHAR))
            *pNext++ = ESC_CHAR;
        *pNext++ = Byte;
    }

    *pNext++ = EOT_CHAR;
    /* Fix checksum on account of STX chars */
    if ((Sum & 0xff) == STH_CHAR)
        Sum++;
    if ((Sum >> 8) == STH_CHAR)
        Sum += (1 << 8);/* this must not rollover */

    *pNext++ = (Sum&0xff);
    *pNext++ = (Sum>>8);

printf("Writing %d bytes\n",pNext - lpBuffer);
if (0) {
    UINT8* p = lpBuffer;
    for (;p < pNext;p++) printf("%x ",*p);
    printf("\n");
}
    r = WriteFile(h,
                  lpBuffer,
                  pNext - lpBuffer,
                  lpSize,
                  lpOver);
    return r;
}

BOOL NextReply(HANDLE h, char *lpBuffer, DWORD *lpSize, LPOVERLAPPED lpOver)
{
    int i = *lpSize;
    UINT8* p = (UINT8*)lpBuffer;

    enbr -= i;
    while (i-- > 0) printf("%x ",*p++);
    printf("\n");
    return (enbr == 0);
}
