/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Microsoft MMLITE June 1995 --jvh
 * NT server for remote IFile using serial lines or sockets.
 */

#include <import_windows.h>
#include <winsock.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <tchar.h>
#include <wchar.h>

#define __HOSTFS_SERVER 1
#include "hostfsp_old.h"

#ifdef InterlockedCompareExchangePointer
/* Resolve version differences between various winbase.h */
#define AtomicCmpAndSwap(Where,OldValue,NewValue) \
    ((PVOID)OldValue == (PVOID)InterlockedCompareExchange((PVOID)Where,NewValue,OldValue))
#else
#define AtomicCmpAndSwap(Where,OldValue,NewValue) \
    ((PVOID)OldValue == (PVOID)InterlockedCompareExchange((PVOID)Where,(PVOID)NewValue,(PVOID)OldValue))
#endif

UINT AtomicDec(PUINT pCount)
{
    UINT OldCount, NewCount;

    for (;;) {
        OldCount = *pCount;
        NewCount = OldCount - 1;
        if (AtomicCmpAndSwap(pCount, OldCount, NewCount)) {
            return NewCount;
        }
    }
}
UINT AtomicInc(PUINT pCount)
{
    UINT OldCount, NewCount;

    for (;;) {
        OldCount = *pCount;
        NewCount = OldCount + 1;
        if (AtomicCmpAndSwap(pCount, OldCount, NewCount)) {
            return NewCount;
        }
    }
}

static UINT CacheTimestamp = 0;

/* Read timeouts
 */
BOOL IsLongTimeout = FALSE;

/* Customize generic packet I/O routines
 */
#define LINK_HANDLE PCONNECTION
#define Getabyte(_l_,_p_) (_l_)->v->Getabyte(_l_,_p_)
#define Petabyte(_l_,_c_) (_l_)->v->Petabyte(_l_,_c_)
#define Petabytes(_l_,_b_,_n_) (_l_)->v->Petabytes(_l_,_b_,_n_)
#define LongTimeout(_l_) (_l_)->v->LongTimeout(_l_)
#define ShortTimeout(_l_) (_l_)->v->ShortTimeout(_l_)
#define Pushout(_l_,_n_) (_l_)->v->Pushout(_l_,_n_)

#include "getput_old.c"

/* We can speak many languages..
 */
int error = 0;
#include "p_serial.c"
#include "p_udp.c"
#include "p_tcp.c"
#include "p_usb.c"

struct {
    _TCHAR *Name;
    PCONNECTION (*Constructor)(_TCHAR *Arg1, UINT Arg2, UINT Arg3);
} Protocols[] = {
  /* Sockets: UDP */
  { _T("UDP"), UdpNew },
  /* Sockets: TCP */
  { _T("TCP"), TcpNew },
  /* Special buffering: USB */
  { _T("USB"), UsbNew },
  /* Defaul: serial lines */
  { NULL, SerialNew }
};
#define nProtocols (sizeof Protocols / sizeof Protocols[0])

char *Usage = "%s [-writable] [-b<BAUDRATE>] [-p<PORTNUMBER>] [-raw] [-verbose] [UDP|TCP|USB|COM1|COM2|...]";

/* Aligned buffer for messaging */
UINT32 MessageBuffer[MAX_PACKET_SIZE/sizeof(UINT32)];

int _tmain(int argc, _TCHAR **argv)
{
    _TCHAR *Line, *pgmname = argv[0];
    PCONNECTION hConn;
    struct packet *Msg = (struct packet *)MessageBuffer;
    UINT nBytes;
    UINT BaudRate, PortNumber;
    BOOL fRaw = FALSE;
    int i;

    Line = _T("COM1");
    BaudRate = 38400;
    PortNumber = 4444;
    hostfsd_may_write = FALSE;  /* set default for standalone server */

    /* Parse args, if any
     */
    while (argc > 1) {
        if (argv[1][0] == '-') {
            switch (argv[1][1]) {
            case 'w':
            case 'W':
                hostfsd_may_write = TRUE;
                printf("Will allow changes to the local filesystem\n");
                break;
            case 'b':
            case 'B':
                BaudRate = _ttoi(argv[1]+2);
                printf("Baudrate will be %d baud\n", BaudRate);
                break;                
            case 'p':
            case 'P':
                PortNumber = _ttoi(argv[1]+2);
                printf("Will use port no. %d\n", PortNumber);
                break;                
            case 'r':
            case 'R':
                fRaw = TRUE;
                printf("Will use raw packet transport.\n");
                break;
            case 'V': hostfsd_debug_level++;/*fallthrough*/
            case 'v': hostfsd_debug_level++;
                printf("debug_level = %d\n", hostfsd_debug_level);
                break;
            default:
                printf(Usage,pgmname);
                exit(-1);
            }
        } else {
            Line = argv[1];
        }
        argc--,argv++;
    }

    /* Open it
     */
    hConn = NULL;
    Protocols[nProtocols-1].Name = Line;/*default, will match */
    for (i = 0; i < nProtocols; i++) {
        if (_tcsicmp(Line,Protocols[i].Name) == 0) {
            hConn = Protocols[i].Constructor(Line,BaudRate,PortNumber);
            break;
        }
    }

    if (hConn == NULL)
        return error;

    /* Start by resetting the connection..
     */
    packet_init(Msg,PACKET_TYPE_RESET);
    packet_prepare(Msg,0);
    Msg->seqno = hConn->SequenceNumber;
    nBytes = Msg->packet_length;
    Msg->packet_length = ByteSwap16(Msg->packet_length);
    Msg->sc = ByteSwap32(Msg->sc);

    i = (fRaw? PutPacketRaw : PutPacket)( hConn, Msg, MAX_PACKET_SIZE);
    DPRINT(("Init::Wrote %d bytes (%d) to '%s'\n",nBytes,i,Line));

    /* Forever ping pong
     */
    for (;;) {

        /* Get next packet, check serial number is in order
         */
        nBytes = (fRaw ? GetPacketRaw : GetPacket)(hConn,Msg,hConn->SequenceNumber);
        DPRINT((" Read %d bytes from '%s'\n",nBytes,Line));

        if (nBytes == 0) {
            printf("Input error: ");
            PrintLinkErrors(&LinkErrors);
            continue;
        }

        /* Process request
         */
        nBytes = HostFile_Server( hConn, Msg, MAX_PACKET_SIZE);

        /* Set packet's seqno, update expected seqno.
         */
        Msg->seqno = ++hConn->SequenceNumber;

        /* Send reply
         */
        i = (fRaw ? PutPacketRaw : PutPacket)( hConn, Msg, MAX_PACKET_SIZE);
        if (i) {
            DPRINT(("Wrote %d bytes (%d) to '%s'\n",nBytes,i,Line));
        } else
            goto Done;
    }

Done:
    hConn->v->Close(hConn);
    return error;
}

/* The exe is not run under the simulator so no action needed */
UINT HostFsd_NTBegin()
{
    return 0;
}
void HostFsd_NTEnd(UINT State)
{
    UnusedParameter(State);
}
