/* Copyright (c) Microsoft Corporation. All rights reserved. */

typedef struct _MY_TCP_STATE {
  CONNECTION MiState;
  SOCKET hSocket;
  SOCKET hListenSocket;
  SOCKADDR_IN MyPort;/*doc*/
  SOCKADDR_IN HisPort;/*doc*/
  int State;
#define UNCONNECTED 0
#define CONNECTED 1
} MY_TCP_STATE, *PMY_TCP_STATE;

/* Ancillaries
 */
BOOL MakeConnected(PMY_TCP_STATE Tcp)
{
    int len = sizeof Tcp->HisPort;
    Tcp->hSocket = accept(Tcp->hListenSocket,
                          (struct sockaddr *)&Tcp->HisPort,
                          &len);
    if (Tcp->hSocket == INVALID_SOCKET) {
        len = WSAGetLastError();
        printf("Accept failed (x%x)\n",len);
        return FALSE;
    }

    printf("Connected with %s port %d\n",
           inet_ntoa(Tcp->HisPort.sin_addr),
           htons(Tcp->HisPort.sin_port));
    Tcp->State = CONNECTED;
    return TRUE;
}

/* Since TCP is reliable, there is no need for timeouts.
 * Or so one thinks.
 */
void TcpShortTimeout(PCONNECTION pThis)
{
    IsLongTimeout = FALSE;
}

void TcpLongTimeout(PCONNECTION pThis)
{
    IsLongTimeout = TRUE;
}

SCODE TcpGetabyte(PCONNECTION pThis, UINT8 *pByte)
{
    PMY_TCP_STATE Tcp = (PMY_TCP_STATE)pThis;
    int Count;

    while (Tcp->State != CONNECTED)
        MakeConnected(Tcp);

    Count = recv(Tcp->hSocket,pByte,1,0);
    if (Count != 1) {
        error = WSAGetLastError();
        printf("Read failed (x%x) disconnecting.\n",error);
        closesocket(Tcp->hSocket);
        Tcp->hSocket = INVALID_SOCKET;
        Tcp->State = UNCONNECTED;
        return E_TIMED_OUT;
    }

    NOISE((" %x",*pByte));
    return NO_ERROR;
}

SCODE TcpPetabyte(PCONNECTION pThis, UINT8 Byte)
{
    PMY_TCP_STATE Tcp = (PMY_TCP_STATE)pThis;
    int Count;
    
#if 1
    while (Tcp->State != CONNECTED)
        MakeConnected(Tcp);
#endif

    NOISE(("%x ",Byte));

    /* Assume NAGLE */
    Count = send(Tcp->hSocket,&Byte,1,0);
    if (Count != 1) {
        EPRINT(("send failed (x%x)\n",WSAGetLastError()));
    }
    return NO_ERROR;
}

SCODE TcpPetabytes(PCONNECTION pThis, UINT8 *pBytes, UINT nBytes)
{
    PMY_TCP_STATE Tcp = (PMY_TCP_STATE)pThis;
    int Count;
    
#if 1
    while (Tcp->State != CONNECTED)
        MakeConnected(Tcp);
#endif

    if (hostfsd_debug_level) {
        for (Count = 0; Count < (int)nBytes; Count++)
            NOISE(("%x ",pBytes[Count]));
    }

    /* Assume NAGLE */
    Count = send(Tcp->hSocket,pBytes,nBytes,0);
    if (Count != (int) nBytes) {
        EPRINT(("send failed (x%x)\n",WSAGetLastError()));
    }
    return NO_ERROR;
}

SCODE TcpPushout(PCONNECTION pThis, UINT Count)
{
    /* nothing to do */
    return NO_ERROR;
}

void TcpClose(PCONNECTION pThis)
{
    PMY_TCP_STATE Tcp = (PMY_TCP_STATE)pThis;
    closesocket(Tcp->hSocket);
    Tcp->hSocket = INVALID_SOCKET;
    closesocket(Tcp->hListenSocket);
    Tcp->hListenSocket = INVALID_SOCKET;
}

const struct IConnectionVtbl TcpVtable = {
    TcpShortTimeout,
    TcpLongTimeout,
    TcpGetabyte,
    TcpPetabyte,
    TcpPetabytes,
    TcpPushout,
    TcpClose
};

CONNECTION *TcpNew(_TCHAR *Arg1, UINT Arg2, UINT Arg3)
{
    SOCKET hSocket;
    MY_TCP_STATE *Ret;
    WSADATA WSAData;
    SOCKADDR_IN Addr;

    memset(&WSAData,0,sizeof WSAData);
    error = WSAStartup(MAKEWORD(2,0),&WSAData);
    if (error) {
        printf("WSAStartup failed with x%x\n",error);
        return NULL;
    }

    hSocket = socket( AF_INET, SOCK_STREAM, 0);
    if (hSocket == INVALID_SOCKET) {
        error = -1;
        printf("Could not create socket (x%x)\n",WSAGetLastError());
        return NULL;
    }

    memset(&Addr,0,sizeof Addr);
    Addr.sin_port = htons((short)Arg3);
    Addr.sin_family = AF_INET;

    error = bind(hSocket, (PSOCKADDR)&Addr, sizeof Addr);
    if (error) {
        printf("Could not bind socket to port %d (x%x)\n",
               Arg3, WSAGetLastError());
        goto Bad;
    }

    error = listen(hSocket,1);
    if (error) {
        printf("Could not listen socket, port %d (x%x)\n",
               Arg3, WSAGetLastError());
        goto Bad;
    }

    Ret = malloc(sizeof(MY_TCP_STATE));
    if (Ret == NULL) {
        error = -4;
        printf("Out of memory?!? (x%x)\n",GetLastError());
        goto Bad;
    }

    memset(Ret,0, sizeof *Ret);
    Ret->MiState.v = &TcpVtable;
    Ret->MiState.SequenceNumber = 0;/*doc*/
    Ret->hListenSocket = hSocket;
    Ret->hSocket = INVALID_SOCKET;
    Ret->MyPort = Addr;

    return &Ret->MiState;
 Bad:
    closesocket(hSocket);
    return NULL;
}

