/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* UNTESTED !! */

#define MAX_PKTSIZE (2*1024) /* its more like 500 or so */

typedef struct _MY_UDP_STATE {
  CONNECTION MiState;
  SOCKET hSocket;
  SOCKADDR_IN MyPort;
  SOCKADDR_IN HisPort;
  UINT InBufferSize;
  UINT InBufferPtr;
  BYTE InBuffer[MAX_PKTSIZE];
  UINT OutBufferPtr;
  BYTE OutBuffer[MAX_PKTSIZE];
} MY_UDP_STATE, *PMY_UDP_STATE;

/* We dont need timeouts because we use full packets
 */
void UdpShortTimeout(PCONNECTION pThis)
{
    IsLongTimeout = FALSE;
}

void UdpLongTimeout(PCONNECTION pThis)
{
    IsLongTimeout = TRUE;
}

SCODE UdpGetabyte(PCONNECTION pThis, UINT8 *pByte)
{
    PMY_UDP_STATE Udp = (PMY_UDP_STATE)pThis;
    int Count, FromLen;

    if (Udp->InBufferSize == 0) {
        FromLen = sizeof(SOCKADDR_IN);
        Count = recvfrom(Udp->hSocket,
                         Udp->InBuffer,
                         MAX_PKTSIZE,
                         0,
                         (SOCKADDR*)&Udp->HisPort,
                         &FromLen);
        if (Count == SOCKET_ERROR) {
            error = WSAGetLastError();
            if (error == WSAETIMEDOUT)
                return E_TIMED_OUT;
            printf("recvfrom failed (x%x)\n",error);
            exit(1);/* aurghfixme*/
        }

        DPRINT(("got %d bytes from x%x x%x\n", Count,
                Udp->HisPort.sin_addr, Udp->HisPort.sin_port));

        Udp->InBufferSize = Count;
        Udp->InBufferPtr = 0;
    }

    /* Getabyte, last byte ?
     */
    *pByte = Udp->InBuffer[Udp->InBufferPtr++];
    if (Udp->InBufferPtr == Udp->InBufferSize)
        Udp->InBufferSize = 0;

    NOISE((" %x",*pByte));
    return NO_ERROR;
}

SCODE UdpPetabyte(PCONNECTION pThis, UINT8 Byte)
{
    PMY_UDP_STATE Udp = (PMY_UDP_STATE)pThis;

    if (Udp->OutBufferPtr >= MAX_PKTSIZE) {
        EPRINT(("Xmit buffer overflow, wrapping!\n"));
        Udp->OutBufferPtr = 0;
        return E_NOT_ENOUGH_MEMORY;
    }

    Udp->OutBuffer[Udp->OutBufferPtr++] = Byte;
    NOISE(("%x ",Byte));

    return NO_ERROR;
}

SCODE UdpPetabytes(PCONNECTION pThis, UINT8 *pBytes, UINT nBytes)
{
    SCODE sc = NO_ERROR;
    while (nBytes-- > 0) {
        sc = UdpPetabyte(pThis,*pBytes++);
        if (FAILED(sc))
            break;
    }
    return sc;
}

SCODE UdpPushout(PCONNECTION pThis, UINT PacketSize)
{
    PMY_UDP_STATE Udp = (PMY_UDP_STATE)pThis;
    int Count;

    if (Udp->OutBufferPtr != PacketSize) {
        EPRINT(("Inconsistent xmit sizes %d != %d\n",
                Udp->OutBufferPtr,PacketSize));
        /* use what we know */
        PacketSize = Udp->OutBufferPtr;
    }

    if (PacketSize == 0) {
        EPRINT(("Zero xmit !?!?\n"));
        return NO_ERROR;
    }

    /* BUGBUG methinks we start first sending a reset pkt ? wont work */

    error = NO_ERROR;
    Count = sendto(Udp->hSocket,
                   Udp->OutBuffer,
                   Udp->OutBufferPtr,
                   0,
                   (SOCKADDR*)&Udp->HisPort,
                   sizeof Udp->HisPort);
    if ((UINT)Count != Udp->OutBufferPtr) {
        error = WSAGetLastError();
        EPRINT(("sendto says %d (!= %d) err = x%x\n",
                Count,Udp->OutBufferPtr,error));
    }

    /* Always consume buffer, regardless
     */
    Udp->OutBufferPtr = 0;
    return error;
}

void UdpClose(PCONNECTION pThis)
{
    PMY_UDP_STATE Udp = (PMY_UDP_STATE)pThis;
    closesocket(Udp->hSocket);
    Udp->hSocket = INVALID_SOCKET;
}

const struct IConnectionVtbl UdpVtable = {
    UdpShortTimeout,
    UdpLongTimeout,
    UdpGetabyte,
    UdpPetabyte,
    UdpPetabytes,
    UdpPushout,
    UdpClose
};

CONNECTION *UdpNew(_TCHAR *Arg1, UINT Arg2, UINT Arg3)
{
    SOCKET hSocket;
    MY_UDP_STATE *Ret;
    WSADATA WSAData;
    SOCKADDR_IN Addr;

    memset(&WSAData,0,sizeof WSAData);
    error = WSAStartup(MAKEWORD(2,0),&WSAData);
    if (error) {
        printf("WSAStartup failed with x%x\n",error);
        return NULL;
    }

    hSocket = socket( AF_INET, SOCK_DGRAM, 0);
    if (hSocket == INVALID_SOCKET) {
        error = -1;
        printf("Could not create socket (x%x)\n",WSAGetLastError());
        return NULL;
    }

    memset(&Addr,0,sizeof Addr);
    Addr.sin_port = htons((short)Arg3);
    Addr.sin_family = AF_INET;

    error = bind(hSocket, (PSOCKADDR)&Addr, sizeof Addr);
    if (error) {
        printf("Could not bind socket to port %d (x%x)\n",
               Arg3, WSAGetLastError());
        goto Bad;
    }

    Ret = malloc(sizeof(MY_UDP_STATE));
    if (Ret == NULL) {
        error = -4;
        printf("Out of memory?!? (x%x)\n",GetLastError());
        goto Bad;
    }

    memset(Ret,0, sizeof *Ret);
    Ret->MiState.v = &UdpVtable;
    Ret->MiState.SequenceNumber = 0;/*doc*/
    Ret->hSocket = hSocket;
    Ret->MyPort = Addr;

    return &Ret->MiState;
 Bad:
    closesocket(hSocket);
    return NULL;
}

