/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <diagnostics.h>

#include <drivers/hostfsp.h>
 
int _tmain(INT argc, _TCHAR** argv)
{
    SCODE sc;
    PINAMESPACE pNs, pCur;
    PIMODULE pMod;
    _TCHAR *FsName = NULL;
    _TCHAR *LinkName = NULL;
    int error = 0;

    /* Parse and verify arguments
     */
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc) || (argc < 2)) {
        _tprintf(_TEXT(
          "Usage: %s NameForHostFs [LinkName]\n"),
                 argv[0]);
        error = -1;
        goto Done;
    }

    FsName = argv[1];

    if (argc > 2)
        LinkName = argv[2];

    DBGME(1,_tprintf(_TEXT(
         "Creating HostFs namespace as %s using %s as link\n"), 
                     FsName, LinkName));

    /* Go do it and say what
     */
    pNs = InitHostFileSystem(LinkName);
    DBGME(0,_tprintf(_TEXT("IHFS returns 0x%x...\n"), (ADDRESS)pNs));

    if (pNs == NULL) {
        _tprintf(_TEXT("That didnt work, sorry.\n"));
        error = -1;
        goto Done;
    }

    /* Register in namespace and say what
     */
    pCur = CurrentNameSpace(); /* wont fail */
    assert(pCur != NULL);

    sc = pCur->v->Register(pCur, FsName, (PIUNKNOWN) pNs, 0, NULL);

    /* Release this regardless
     */
    pNs->v->Release(pNs);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Failed (sc=0x%x) to register NS 0x%x as %s.\n"),
                 sc, (ADDRESS)pNs, FsName);
        error = -1;
        goto Done;
    }
    
    /* Register ourselves in namespace, so code wont disappear.
     */
    sc = GetModuleContainsAddr((ADDRESS)_tmain, &pMod);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Donno who I am ?!?!? (sc=0x%x).\n"), sc);
        error = -1;
        goto Done;
    }

    sc = pCur->v->Register(pCur, _T("hostfstest.exe"), (PIUNKNOWN) pMod, 0, NULL
);

    /* If it worked then ok, else leak a ref
     */
    if (!FAILED(sc))
        pMod->v->Release(pMod);

 Done:
    /* Free args
     */
    (void) FreeCmdLine(&argc,&argv);

    /* Done.
     */
    if (error == 0)
        DBGME(0,_tprintf(_TEXT("Success. Dont forget to delete it :-)\n")));
    return error;
}
