/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Example of polling serial line code, the Atmel m63200 part
 */
#include "at91m63200.h"
#define Eb63Clock (25*1000*1000)

/* *** UsartInit
 *
 * Initializes a COM port for low-level (polling) debugging use.
 *
 */
void UsartInit(USART *Port, UINT Baud, UINT8 PortNo)
{
    UINT32 Divisor;

    /* Enable clock to the part */
    ThePmc->PerClockEnable = PMCPC_USART0 << PortNo;

    /* disable unused stuff */
    Port->RxTimeout =
    Port->TxTimeout =
    Port->RxPointer =
    Port->TxPointer =
    Port->RxCounter =
    Port->TxCounter = 0;

    /* Better be slower than faster, roundup 
     * NB: ASYNC, else no x16;
     */
    Divisor = (Eb63Clock/8)/Baud;
    Divisor = (Divisor/2) + (Divisor&1);
    Port->Baud = Divisor;

    /* 8 bit, no parity, 1 stop */
    Port->Mode = USM_CLK_MCKI | USM_BPC_8 | USM_NONE | USM_1STOP;

    /* disable all interrupts */
    Port->IntrDisable = ~0;

    /* Enable rx/tx and clear any leftover status bits */
    Port->Control = USC_RXEN | USC_TXEN | USC_RSTSTA | USC_STPBRK;

    /* Turn on external pins */
    PioA->Disable = ((PIOA_TXD0|PIOA_RXD0) << (PortNo*3));
}

/* *** UsartPut
 *
 * Write a byte to the specified port.
 */
void UsartPut(USART *Port, UINT8 Byte)
{
    /* wait until we can transmit */
    while ((Port->ChannelStatus & USI_TXEMPTY) == 0)
        continue;

    Port->TxData = Byte;
}

/* *** UsartGet
 *
 * Read a byte from the specified port.
 * Returns FALSE if there was no byte ready
 * to be read, or in case of error.
 */
BOOL UsartGet(USART *Port, UINT8 *pByte)
{
    UINT32 Status = Port->ChannelStatus;

    /* See if anything available */
    if (Status & USI_RXRDY) {
        *pByte = (UINT8) Port->RxData;

        /* Check for errors */
        if (Status & (USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE)) {
            if ((Status & USI_OVRE) == 0)   /* Overrun? */
                return FALSE;           /*  no, parity or framing error */
        }

        return TRUE;
    }

    return FALSE;
}

static BOOL dxNeedsInit = TRUE;
static void initme(void)
{
    if (dxNeedsInit) {
        dxNeedsInit = FALSE;
        UsartInit(Usart1, 38400, 1);
    }
}

void SleepItOff(int MilliSeconds)
{
    TIME t;

    Int64FromHighAndLow(t,-1,TIME_RELATIVE(TIME_MILLIS(MilliSeconds)));
    SleepUntil(t);
}


TIME Timeo;

void LongTimeout(PIFILE Link)
{
    Int64FromHighAndLow(Timeo,0,TIME_SECONDS(2));
    IsLongTimeout = TRUE;
}
void ShortTimeout(PIFILE Link)
{
    Int64FromHighAndLow(Timeo,0,TIME_MILLIS(100));
    IsLongTimeout = FALSE;
}



SCODE Getabyte(PIFILE Link, UINT8 *pByte)
{
    int i;
    TIME t0, t;

    initme();
    t0 = CurrentTime();
    for (;;) {
        for (i = 0; i < 100; i++)
            if (UsartGet(Usart1,pByte))
                return NO_ERROR;
        t = CurrentTime();
        t = TimeSubtract(t,t0);
        if (!TimeLess(t,Timeo))
            break;
    }
    return E_TIMED_OUT;
}

SCODE Petabyte(PIFILE Link, UINT8 Byte)
{
    initme();
    UsartPut(Usart1,Byte);
    return NO_ERROR;
}

SCODE Petabytes(PIFILE Link, UINT8 *pBytes, UINT nBytes)
{
    while (nBytes-- > 0)
        Petabyte(Link,*pBytes++);
    return NO_ERROR;
}


