/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <diagnostics.h>

extern PINAMESPACE RomFsNew( PTR BaseOfMemory, UINT Size);

static _TCHAR TestName[] = _TEXT("FlashTest");

int _tmain(INT argc, _TCHAR** argv)
{
    SCODE sc;
    ADDRESS RomfsStart;
    UINT RomfsSize;
    PINAMESPACE pNs, pCur;
    PIMODULE pMod;
    _TCHAR *FsName = TestName;

    /* Parse and verify arguments
     */
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc) || (argc < 3)) {
        _tprintf(_TEXT(
          "Usage: %s <RomImageAddressInHex> <SizeInDecimalBytes> [FsName]\n"),
                 argv[0]);
        return -1;
    }

    /* Convert and echo args
     */
    RomfsStart = _tcstoul(argv[1],NULL,16);
    RomfsSize  = _tcstoul(argv[2],NULL,10);
    if (argc > 3)
        FsName = argv[3];

    DBGME(1,_tprintf(_TEXT(
         "Creating a RomFs namespace %s from %d bytes at 0x%x..\n"),
             FsName, RomfsSize, RomfsStart));

    /* Go do it and say what
     */
    pNs = RomFsNew( (PTR)RomfsStart, RomfsSize);
    DBGME(0,_tprintf(_TEXT("RomFsNew returns 0x%x...\n"), (ADDRESS)pNs));

    if (pNs == NULL) {
        _tprintf(_TEXT("That didnt work, sorry.\n"));
        return -1;
    }

    /* Register in namespace and say what
     */
    pCur = CurrentNameSpace(); /* wont fail */
    assert(pCur != NULL);

    sc = pCur->v->Register(pCur, FsName, (PIUNKNOWN) pNs, 0, NULL);

    /* Release this regardless
     */
    pNs->v->Release(pNs);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Failed (sc=0x%x) to register NS 0x%x as %s.\n"),
                 sc, (ADDRESS)pNs, FsName);
        return -1;
    }
    
    /* Register ourselves in namespace, so code wont disappear.
     */
    sc = GetModuleContainsAddr((ADDRESS)_tmain, &pMod);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Donno who I am ?!?!? (sc=0x%x).\n"), sc);
        return -1;        
    }

    sc = pCur->v->Register(pCur, _T("romfstest.exe"), (PIUNKNOWN) pMod, 0, NULL);

    /* If it worked then ok, else leak a ref
     */
    if (!FAILED(sc))
        pMod->v->Release(pMod);

    /* Free args
     */
    (void) FreeCmdLine(&argc,&argv);

    /* Done.
     */
    DBGME(0,_tprintf(_TEXT("Success. Dont forget to delete it :-)\n")));
    return 0;
}
