/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <diagnostics.h>

extern PINAMESPACE SromFsNew( PIFILE Storage);

static _TCHAR TestName[] = _TEXT("SerialFlashTest");

int _tmain(INT argc, _TCHAR** argv)
{
    SCODE sc;
    PIFILE File;
    PINAMESPACE pNs, pCur;
    PIMODULE pMod;
    _TCHAR *FsName = TestName, *FlashName;

    /* Parse and verify arguments
     */
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc) || (argc < 2)) {
        _tprintf(_T(
          "Usage: %s <NameOfSflashIFile> [NewFsName]\n"),
                 argv[0]);
        goto ErrorExit;
    }

    /* Get current namespace to look for storage file
     */
    pCur = CurrentNameSpace(); /* wont fail */
    assert(pCur != NULL);

    /* Convert, lookup and echo args
     */
    FlashName = argv[1];
    if (argc > 2)
        FsName = argv[2];

    sc = BindToObject(pCur, FlashName, NAME_SPACE_READ, &IID_IFile, 
                      (void**) &File);
    if (FAILED(sc)) {
        _tprintf(_T("Could not BindToIFile %s, sc=%x\n"), FlashName, sc);
        goto ErrorExit;
    }

    DBGME(1,_tprintf(_T(
         "Creating a RomFs namespace %s on top of %s..\n"),
             FsName, FlashName));

    /* Go do it and say what
     */
    pNs = SromFsNew( File );
    DBGME(0,_tprintf(_T("RomFsNew returns 0x%x...\n"), (ADDRESS)pNs));

    /* Release File regardless. If it worked it kept its own ref
     */
    File->v->Release(File);

    /* Check
     */
    if (pNs == NULL) {
        _tprintf(_T("That didnt work, sorry.\n"));
        goto ErrorExit;
    }

    /* Register in namespace and say what
     */
    sc = pCur->v->Register(pCur, FsName, (PIUNKNOWN) pNs, 0, NULL);

    /* Release this regardless
     */
    pNs->v->Release(pNs);

    if (FAILED(sc)) {
        _tprintf(_T("Failed (sc=0x%x) to register NS 0x%x as %s.\n"),
                 sc, (ADDRESS)pNs, FsName);
        goto ErrorExit;
    }
    
    /* Register ourselves in namespace, so code wont disappear.
     */
    sc = GetModuleContainsAddr((ADDRESS)_tmain, &pMod);
    if (FAILED(sc)) {
        _tprintf(_T("Donno who I am ?!?!? (sc=0x%x).\n"), sc);
        goto ErrorExit;
    }

    sc = pCur->v->Register(pCur, _T("sromfstest.exe"), (PIUNKNOWN) pMod, 0, NULL);
    /* If it worked then ok, else leak a ref
     */
    if (!FAILED(sc))
        pMod->v->Release(pMod);

    /* Free args
     */
    (void) FreeCmdLine(&argc,&argv);

    /* Done.
     */
    DBGME(0,_tprintf(_T("Success. Dont forget to delete it :-)\n")));
    return 0;
 ErrorExit:
    (void) FreeCmdLine(&argc,&argv);
    return -1;
}
