/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

#include "tmpfs.h"

#define PRIVATE static

typedef struct _FACTORY {
    struct INameSpaceVtbl *v;
    UINT RefCnt;
} *PFACTORY;

PRIVATE struct INameSpaceVtbl TheNameVtbl;

PRIVATE struct _FACTORY _TheFactory = {
    &TheNameVtbl,
    1
};

#define pSN(_i_) ((PFACTORY)(_i_))
#define iSN(_p_) ((PINAMESPACE)(_p_))

/* Make our single instance visible to code in zerofile.c etc. */
const PINAMESPACE pTmpFsFactory = iSN(&_TheFactory);

PRIVATE SCODE MCT SNQueryInterface(PINAMESPACE pThis, REFIID Iid, void **ppObj)
{
    SCODE sc;
    sc = GenericQueryInterface((PIUNKNOWN) pThis, Iid, ppObj, &IID_INameSpace);

    if (!FAILED(sc))
        DPRINT(("SNQueryInterface ==> NameSpace\n"));
    else
        DPRINT(("SNQueryInterface ==> Fail\n"));

    return sc;
}

PRIVATE UINT MCT SNAddRef(PINAMESPACE pThis)
{
    UINT RefCnt = AtomicInc(&pSN(pThis)->RefCnt);
    DPRINT(("SNAddRef x%x ==> %d\n", pThis, RefCnt));
    return RefCnt;
}

PRIVATE UINT MCT SNRelease(PINAMESPACE pThis)
{
    PFACTORY sn = pSN(pThis);
    UINT RefCnt = AtomicDec(&sn->RefCnt);

    DPRINT(("SNRelease x%x ==> %d\n", pThis, RefCnt));
    if (RefCnt == 0) {
        /* Nothing to finalize or deallocate in this implementation */
    }
    return RefCnt;
}

PRIVATE SCODE SNBind(PINAMESPACE pThis, const _TCHAR * pName,
                     NAME_SPACE_FLAGS Flags, PIUNKNOWN *ppObject)
{
    UnusedParameter(pThis);
    DPRINT(("SNBind(%s)\n", pName));

    if (!_tcscmp(pName, _T("file"))) {
        /* Implemented in memfile.c */
        return CreateMemFile((PIFILE *) ppObject);
    } else {
        return E_NAME_NOT_FOUND;
    }
}

/* Default method */
PRIVATE SCODE Unimp(void)
{
    return E_NOT_IMPLEMENTED;
}

/* The NameSpace method table.
 * Some methods don't do anything and use Unimp()
 */
PRIVATE struct INameSpaceVtbl TheNameVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    SNQueryInterface,
    SNAddRef,
    SNRelease,
    (_INameSpace_Register_METHOD_TYPE) Unimp,
    (_INameSpace_Unregister_METHOD_TYPE) Unimp,
    SNBind,
    (_INameSpace_FindNext_METHOD_TYPE) Unimp,
    (_INameSpace_GetCapabilities_METHOD_TYPE) Unimp
};

/* entry point */
PIUNKNOWN TmpFsCobMain(void)
{
    DPRINT(("TmpFsCobMain\n"));
    return (PIUNKNOWN) iSN(&_TheFactory);
}
