/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <base/cobdesc.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>
#include <util/string.h>
#include <sampleif.h>

#ifndef _MINIMIZE
#define __WANTS_TRUST 1
#endif

typedef struct _PropertyClass PropertyClass;
typedef const struct _PropertyClass *CPPropertyClass;

/* Current Request Object which is used for the Access check*/
typedef struct _CurrentRequest
{
    PICIMRESOURCE Client;
    PICIMRESOURCE Context;
    UINT Mode;

} CurrentRequest;


struct _PropertyClass {
    const _TCHAR *ResourceURI;  /* long form */
    const _TCHAR *ResourceName; /* short form */
    UINT DataSize;              /* size of resource element data */
    const _TCHAR *EncodingType_in;
    const _TCHAR *EncodingType_out;
    const PropertyClass *super;
    SCODE (*Get)(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest);
    SCODE (*Put)(const PropertyClass *Class, PTR Value);
    SCODE (*Init)(const PropertyClass *Class, PTR Value, PIHEAP Heap, PICIMRESOURCE Resource);
    SCODE (*Finalize)(const PropertyClass *Class, PTR Value, PIHEAP Heap);
    SCODE (*Copy)(const PropertyClass *Class, PTR Dst,
                  const void *Src, PIHEAP Heap);
    SCODE (*Match)(const PropertyClass *Class, PTR Value,
                   PSelector Selector);
    SCODE (*Match2)(const PropertyClass *Class, PTR Value, const void *Value2, const UINT32 *FieldSel);

    /* Update the EPR to current values or 
     * query the sel count and size needed for non-string arguments */
    SCODE (*UpdateSelector)(const PropertyClass *Class,
           PTR Value, PSelector Selector,
           UINT *pCount, UINT *pSize);
    /* Gets the value of the specified field. In case of TrustScript it actually evaluates the policies*/
    UINT (*GetValue)(const PropertyClass *Class, PTR Value,CurrentRequest *currentRequest,const _TCHAR *Field);
   SCODE (*Start)(const PropertyClass *Class, PICIMFACTORY Factory);
   SCODE (*Apply)(const PropertyClass *Class, PICIMRESOURCE AppliedResource, PICIMRESOURCE Resource);
};

/* Lock is held. */

/* Looks for a particular resource*/
PICIMRESOURCE CimResourceLookup(
    PICIMRESOURCE iThis,
    const _TCHAR *ResourceURI,
    PSelector SelectorSet);

/*Looks for a particular resource in the give directory (Resource URI), and if not found looks into the shadow parent and so on*/
PICIMRESOURCE CimResourceShadowLookup(
    PICIMRESOURCE iThis,
    const _TCHAR *ResourceURI,
    PSelector SelectorSet);

/*Class GetValue Method of the property class*/
UINT CimResourceGetValue(
    PICIMRESOURCE iThis, CurrentRequest *currentRequest,const _TCHAR *Field);

/*Looks for the directory*/
PICIMRESOURCE CimDirectoryLookup(
    PICIMRESOURCE iThis, const _TCHAR *To);

PICIMRESOURCE ReturnShadowParent(PTR Value);


/* Have been tranferred from cimdb.c to this file, because these structs have to be used in several files*/

struct _CimFactory;
struct _CimResource;


typedef struct _CimEnum {
    struct CUnknown super;

    struct _CimResource *Owner;
    void *Current;              /* last one processed */
    PIHEAP Heap;

    _TCHAR *name;
    RESOURCE_FLAGS Flags;
    PDataItem Query;
    
    ListNodeDeclare(_CimEnum, ListNode);
#define cenoff offsetof(struct _CimEnum, ListNode)
} CimEnum;

typedef struct _CimResource {
    struct CUnknown super;
    
    struct _CimResource *Parent;
    const PropertyClass *PropertyClass;
    PIHEAP Heap;

    /* Live objects in a directory */
    struct _CimResource *Resources;
    /* Active enumerators in directory */
    struct _CimEnum *Enumerators;

 /*   struct _CimResource *ShadowParent;
    struct _CimResource *ShadowResources;
  */  

    /* To add this to a list */
    ListNodeDeclare(_CimResource, ListNode);
#define croff offsetof(struct _CimResource, ListNode)

// XXX Should be in DemoServiceInfoType     
    ListNodeDeclare(_CimResource, ShadowListNode);
#define shadowoff offsetof(struct _CimResource, ShadowListNode)


    /* EPR temporarily cached here */
    PEndpointReference_QUIET Epr;

    /* Actual contents */
    void *Value;
} CimResource;

typedef struct _CimFactory {
    struct _DefCob super;

    /* House keeping */
    MUTEX mx;
    CONDITION cd;

    /* Cached value of wsa:To of service */
    EndpointReference_QUIET Epr;
    _TCHAR *EprAddress;
    BOOL Started;

    struct _CimResource *Root;
} CimFactory;

/* Compare, copy, etc. utility macros for value class use */
#define S_MATCH2(_name_, _ix_) \
    if (!FieldSel || (FieldSel[_ix_/32] & (1 << _ix_%32))) { \
        CHECKSETSC(val->_name_ && ref->_name_, E_NAME_NOT_FOUND, "field present"); \
        CHECKSETSC(!_tcscmp(val->_name_, ref->_name_), E_FAIL, "compare required field " # _name_); \
    }
#define N_MATCH2(_name_, _ix_) \
    if (!FieldSel || (FieldSel[_ix_/32] & (1 << _ix_%32))) { \
        CHECKSETSC(val->_name_ == ref->_name_, E_FAIL, "compare required field " # _name_); \
    }

#define S_DEL(_f_) if (val->_f_) {Heap->v->Free(Heap, 0, val->_f_); val->_f_ = NULL;}

#define S_COPY(_f_) if (dst->_f_) {Heap->v->Free(Heap, 0, dst->_f_); dst->_f_ = NULL;} if (!src->_f_ || !*src->_f_) {dst->_f_ = NULL;} else {tmp = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:string"), (_tcslen(src->_f_)+1)*sizeof(_TCHAR), 0); CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "alloc string"); _tcscpy(tmp, src->_f_); dst->_f_ = tmp;}

#define N_COPY(_f_) dst->_f_ = src->_f_