/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include "demoserviceinfotype.h"
#include <stdlib.h>
#include<wsif.h>
#include<wsifhelp.h>


#define MAX_ARGUMENTS 10


/******** Temporary Functions***********/
PICIMRESOURCE ReturnShadowParent(PTR Value)
{
     DemoServiceInfoType *dsit  = Value;
     if(dsit)
        return (PICIMRESOURCE)(dsit->ShadowParent);
     else
         return NULL;
}

PRIVATE PICIMRESOURCE ReturnRoot(PICIMRESOURCE Resource)
{
    CimResource *root = NULL;
    CimResource *resource = (CimResource *)Resource;
    
    while(resource)
    {
        root = resource;
        resource = resource->Parent;
    }
    return (PICIMRESOURCE) root;

}



typedef struct GetServiceTemplate
{
    PICIMRESOURCE Current;
    DemoServiceInfoType *dsit; 
}GetServiceTemplate;


PRIVATE INT UpdateShadowParent(CimResource *resource, CimResource* newParent)
{
    DemoServiceInfoType *temp = resource->Value;
    temp->ShadowParent = newParent;
    return 0;
}

PRIVATE INT GetCurrentService(CimResource *Resource, GetServiceTemplate *gsit)
{
    if (!_tcscmp(((Resource->PropertyClass)->ResourceName),_T("DemoServiceInfoType")))
    {
        DemoServiceInfoType *temp = Resource->Value;
        if(!_tcscmp(  (gsit->dsit)->super.Name,     temp->super.Name  ))
            gsit->Current = (PICIMRESOURCE)Resource;
    }
    return 0;
}
PRIVATE SCODE DemoServiceInfoType_Copy(const PropertyClass *Class,
                                     PTR _Dst, const void *_Src, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(DemoServiceInfoType, Copy);
    DemoServiceInfoType *dsit = _Dst;
    const DemoServiceInfoType *src = _Src;
    sc = S_OK;

    dsit->ShadowParent = src->ShadowParent;
    dsit->ShadowResources = NULL;


METHOD_CLEANUP;
    return sc;
}


PRIVATE SCODE DemoServiceInfoType_Finalize(const PropertyClass *Class,
                                         PTR Value, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(DemoServiceInfoType, Finalize);
    DemoServiceInfoType *dsit = Value;
    
    CimResource *sparent = dsit->ShadowParent;
    CimResource *current;
    
    if(sparent)
    {
        DemoServiceInfoType *dsit1 = sparent->Value;
        GetServiceTemplate gst;
        gst.dsit = dsit;
        gst.Current = NULL;
        ListWalkT(CimResource, shadowoff, GetServiceTemplate,
                      &dsit1->ShadowResources, LIST_WALK_ITERATE,
                      GetCurrentService, &gst);
    
        
        current = (CimResource *) gst.Current;
        ListRemoveItem(&dsit1->ShadowResources, shadowoff, current);
           
    } 
    if(dsit->ShadowResources)
    {
        ListWalkT(CimResource, shadowoff, CimResource,
                  &dsit->ShadowResources, LIST_WALK_ITERATE,
                  UpdateShadowParent, sparent);

    }
    sc = S_OK;
    METHOD_CLEANUP;
   
    return sc;
}

PRIVATE SCODE DemoServiceInfoType_UpdateSelector(const PropertyClass *Class,
           PTR Value, PSelector Selector,
           UINT *pCount, UINT *pSize)
{
    METHOD_DECLARE_NOTHIS(DemoServiceInfoType, Match2);
    DemoServiceInfoType *dsit = Value;

    if (pCount)
        *pCount = 1;
    if (pSize)
        *pSize = 0;
    if (Selector) {
        Selector->Name = _T("Name");
        Selector->Value = dsit->super.Name;
    }

    sc = S_OK;
    METHOD_CLEANUP;
    return sc;
}


//PRIVATE _TCHAR* ServiceInfoType_ReturnShadowParentUrl(const PropertyClass *Class, PTR Value)
//{
//    
//    struct ServiceInfoType *sit = Value;
//    return sit->PartnerList[0].ServiceDirectory;
//    
//}

/*********** ServiceInfoType class ******************/


PRIVATE SCODE DemoServiceInfoType_Get(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest)
{
   
       return S_OK;
}

PRIVATE SCODE DemoServiceInfoType_Put(const PropertyClass *Class, PTR Value)
{
        return S_OK;
}

/* Initialize private fields */
PRIVATE SCODE DemoServiceInfoType_Init(const PropertyClass *Class, PTR Value,
                                /*unused*/ PIHEAP Heap,PICIMRESOURCE Resource)
{
    DemoServiceInfoType *dsit = Value;
    PICIMRESOURCE root = ReturnRoot(Resource);
    PICIMRESOURCE shadowParent = NULL;
    shadowParent = CimDirectoryLookup(root,(dsit->super).PartnerList[0].ServiceDirectory);

    //Does checking to avod cycle. Shadow parent must not be itself.
    if(shadowParent != Resource)
    {
        dsit->ShadowParent = (CimResource *)shadowParent;
        if(dsit->ShadowParent)
        {
            DemoServiceInfoType *dsit1 = (dsit->ShadowParent)->Value;
            ListAddLast(&dsit1->ShadowResources, shadowoff, (CimResource *)Resource);
        }
    }
    else
        dsit->ShadowParent = NULL;
    return S_OK;
     
}

extern const PropertyClass ServiceInfoType_Prop;

const PropertyClass DemoServiceInfoType_Prop = {
    _T("dssp:DemoServiceInfoType"), 
    _T("DemoServiceInfoType"),
    sizeof(DemoServiceInfoType),
    _T("dssp:ServiceInfoType"),
    _T("dssp:ServiceInfoType"),
    &ServiceInfoType_Prop,
    DemoServiceInfoType_Get, 
    DemoServiceInfoType_Put, 
    DemoServiceInfoType_Init,
    DemoServiceInfoType_Finalize,
    DemoServiceInfoType_Copy,
    NULL,
    NULL,
    DemoServiceInfoType_UpdateSelector,

    
    NULL, 
    NULL
};

