/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <arm/at91m63200.h>
 
#define TIME_HUNDRED_NANOS(micros) (((micros) * (1)))

UINT BytesToHalfword(UINT8 byte_H, UINT8 byte_L);

/* Gyro device addresses */
UINT8 GYRO_WRITE  = 0xAE; //{1,0,1,0,1,1,1,0}
UINT8 GYRO_READ   = 0xAF; //{1,0,1,0,1,1,1,1}
UINT8 EEPROM_READ = 0xA1; //{1,0,1,0,0,0,0,1}
UINT8 EEPROM_WRITE = 0xA0; //{1,0,1,0,0,0,0,0}

/* Globals */
UINT8 EEPROM_stor[16];  //array to hold EEPROM memory
UINT8 STATUS = 49;      //{0,1,0,0,1,0,0,1}

//sub-functions
 void START();
 void STOP();
 void Validate_gyro_Output();
 void Receive_ACK();
 void Send_Bit(BOOL bit);
 BOOL Receive_Bit();
 void Send_Byte(UINT8 byte);
 UINT8 Receive_Byte();

//main functions
 void Byte_Write(UINT8 dev_add, UINT8 word_add, UINT8 data);
 void Page_Write(UINT8 dev_add, UINT8 word_add, UINT8 data[16]);
 void Current_Read(UINT8 dev, PUINT8 data);
 void Random_Read(UINT8 dev_w, UINT8 dev_r, UINT8 mem, PUINT8 data);
 void Dummy_Write(UINT8 dev_add, UINT8 word_add);

//program functions
void Power_Up();
void Start_Up();
void Poll_Data(PUINT8 HA, PUINT8 LA, PUINT8 HB, PUINT8 LB);
void gyro_init_eb63pins();
void wait(int delay);


void gyro_power_start(PUINT8 OFFSET_HA, PUINT8 OFFSET_LA, PUINT8 OFFSET_HB, PUINT8 OFFSET_LB)
{
    gyro_init_eb63pins();
    Power_Up();
    Start_Up();
	
    wait(2000000);
    Poll_Data(OFFSET_HA, OFFSET_LA, OFFSET_HB, OFFSET_LB);
}

INT gyro_poll(PUINT8 HIGH_A, PUINT8 LOW_A, PUINT8 HIGH_B, PUINT8 LOW_B)
{			
    Poll_Data(HIGH_A, LOW_A, HIGH_B, LOW_B);
    return 0;
}

void gyro_init_eb63pins()
{			
    //OUTPUT SETUP
    PioB->Enable = PIOB_PB6;
    PioB->OutEnable = PIOB_PB6;
    PioB->FilterEnable = PIOB_PB6;
    PioB->IntrDisable = PIOB_PB6;
    PioB->SetData = PIOB_PB6;

    //INPUT SETUP
    //PB7: SDA pin...is bi-directional
    PioB->Enable = PIOB_PB7 ;
    PioB->FilterEnable = PIOB_PB7;
    PioB->IntrDisable = PIOB_PB7;
    PioB->SetData = PIOB_PB7;
    PioB->OutDisable = PIOB_PB7;
}

//**********************************************


  void wait(int delay){
	TIME wt1;
    Int32ToInt64(wt1, TIME_RELATIVE(TIME_MICROS(delay)));
	SleepUntil(wt1);  // put the thread to sleep until wt1
  }
   
/////////////////////////////////////////////
///////////PROGRAM FUNCTIONS/////////////////
/////////////////////////////////////////////
 

void Power_Up()
{
    
    UINT8 temp = 0;
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x00,&EEPROM_stor[0]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x01,&EEPROM_stor[1]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x02,&EEPROM_stor[2]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x03,&EEPROM_stor[3]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x04,&EEPROM_stor[4]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x05,&EEPROM_stor[5]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x06,&EEPROM_stor[6]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x07,&EEPROM_stor[7]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x08,&EEPROM_stor[8]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x09,&EEPROM_stor[9]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0A,&EEPROM_stor[10]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0B,&EEPROM_stor[11]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0C,&EEPROM_stor[12]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0D,&EEPROM_stor[13]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0E,&EEPROM_stor[14]);
    Random_Read(EEPROM_WRITE,EEPROM_READ,0x0F,&EEPROM_stor[15]);  

    //Write page to gyro Function memory
    Page_Write(GYRO_WRITE, 0x00, EEPROM_stor);

    wait(25);

    /* XXX temp = {0,1,0,0,1,EEPROM_stor[15][5],EEPROM_stor[15][6],EEPROM_stor[15][7]} */
    temp = EEPROM_stor[15] & 0x07;
    temp |= 0x48;
	
    wait(25);
	
    //set INIT bit to HIGH
    Byte_Write(GYRO_WRITE, 0x0F, temp);

    wait(20);
}

void Start_Up()
{
    
    UINT8 temp = 0;
    //STATUS = {0,1,0,0,1,0,0,1}	

    wait(25);	
	
    //set INIT bit to HIGH
    Byte_Write(GYRO_WRITE, 0x0F, STATUS);
	
    wait(20);
	
    //while(!GRNR);
    while((temp & 0x02) != 0)
    {
	    Random_Read(GYRO_WRITE, GYRO_READ, 0x08, &temp);
	    wait(5);
    }
	
    wait(20);
    /* gyro is now ready */
}


/* Poll both axes of the gyro. */
void Poll_Data(PUINT8 HA, PUINT8 LA, PUINT8 HB, PUINT8 LB)
{	
    Random_Read(GYRO_WRITE,GYRO_READ, 0x00, HA);
    Current_Read(GYRO_READ, LA);
    Current_Read(GYRO_READ, HB);
    Current_Read(GYRO_READ, LB);
}


/////////////////////////////////////////////
//////////////READ FUNCTIONS/////////////////
/////////////////////////////////////////////  
  
/* Current_Read */
void Current_Read(UINT8 dev, PUINT8 pdata)
{
    START();
	
    //send device address	
    Send_Byte(dev);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //read in data    	
    *pdata = Receive_Byte();

    //Stop
    STOP();

}//end current read


/* Random_Read */
void Random_Read(UINT8 dev_w, UINT8 dev_r, UINT8 mem, PUINT8 pdata)
{
    //Start
    START();

    //send device address	
    Send_Byte(dev_w);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //SDA->Output_Enable();
    PioB->OutEnable = PIOB_PB7;

    //send memory address
    Send_Byte(mem);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();
	
    //SDA->Output_Enable();
    PioB->OutEnable = PIOB_PB7;
	
    //Current_Read(dev_r, data);

    START();
	
    //send device address
    Send_Byte(dev_r);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //read in data    	
    *pdata = Receive_Byte();

    //Stop
    STOP();


}//end random read


/////////////////////////////////////////////
//////////////WRITE FUNCTIONS////////////////
/////////////////////////////////////////////  


//this function writes a byte to the gyro Function 
//for INIT or PWRxxx

//SCL = PIOB_PB6
//SDA = PIOB_PB7


void Dummy_Write(UINT8 dev_add, UINT8 word_add)
{
    //Start
    START();

    //send device address	
    Send_Byte(dev_add);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //SDA->Output_Enable();
    PioB->OutEnable = PIOB_PB7;

    //send memory address
    Send_Byte(word_add);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();
}



void Byte_Write(UINT8 dev_add, UINT8 word_add, UINT8 data)
{
    //Start
    START();

    //send device address	
    Send_Byte(dev_add);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //SDA->Output_Enable();
    PioB->OutEnable = PIOB_PB7;

    //send memory address
    Send_Byte(word_add);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //SDA->Output_Enable();
    PioB->OutEnable = PIOB_PB7;

    //send data
    Send_Byte(data);

    //poll ACK
    Validate_gyro_Output();
    Receive_ACK();

    //STOP signal
    STOP();    	

}//end Byte_Write




 void Page_Write(UINT8 dev_add, UINT8 word_add, UINT8 data[16])
{
    int i;

	//Start
	START();

	//send device address	
    Send_Byte(dev_add);

	//poll ACK
	Validate_gyro_Output();
	Receive_ACK();
	//SDA->Output_Enable();
	PioB->OutEnable = PIOB_PB7;

	//send memory address	
    Send_Byte(word_add);

	//poll ACK
	Validate_gyro_Output();

	Receive_ACK();

	for(i=0;i<16;i++)
	{	
		//SDA->Output_Enable();
		PioB->OutEnable = PIOB_PB7;

		//send data
	    Send_Byte(data[i]);

		//poll ACK
		Validate_gyro_Output();
		Receive_ACK();
	}//endfor
	

	//Stop
	STOP();
}


/////////////////////////////////////////////
///////////////SUB FUNCTIONS/////////////////
/////////////////////////////////////////////  

//these are called by the main functions
//SCL = PIOB_PB6   ---clock
//SDA = PIOB_PB7  ---data

/* Send Byte */
void Send_Byte(UINT8 byte) {
    int i;
    BOOL bit = FALSE;

    for(i=0; i<8; i++)
    {
	    switch(i)
        {
            case 0: 
                if((byte & 0x80) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 1: 
                if((byte & 0x40) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 2: 
                if((byte & 0x20) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 3: 
                if((byte & 0x10) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 4: 
                if((byte & 0x08) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 5: 
                if((byte & 0x04) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 6: 
                if((byte & 0x02) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
            case 7: 
                if((byte & 0x01) == 0)
                    bit = FALSE;
                else
                    bit = TRUE;        
                break;
        }//endswitch

        Send_Bit(bit);

    }//endfor
}

/* Receive Byte*/
UINT8 Receive_Byte() {
    int i;
    UINT8 data = 0;

    for(i=0; i<8; i++)
    {    
        //initialize to zero        
        switch(i)
        {
            case 0: 
                if(Receive_Bit())
                    data |= 0x80;   
                break;
            case 1: 
                if(Receive_Bit())
                    data |= 0x40;   
                break;
            case 2: 
                if(Receive_Bit())
                    data |= 0x20;   
                break;
            case 3: 
                if(Receive_Bit())
                    data |= 0x10;   
                break;
            case 4: 
                if(Receive_Bit())
                    data |= 0x08;   
                break;
            case 5: 
                if(Receive_Bit())
                    data |= 0x04;   
                break;
            case 6: 
                if(Receive_Bit())
                    data |= 0x02;   
                break;
            case 7: 
                if(Receive_Bit())
                    data |= 0x01;   
                break;
        }//endswitch
    }//endreaddata

    return data;

}


/* Send Bit */
void Send_Bit(BOOL bit)
{
    int stat, SDA;
    //SCL = LOW;
    PioB->ClearData = PIOB_PB6;
    wait(1);
    //SDA = array[i];
    //if(array[i] != 0)
    if(bit)
	    PioB->SetData = PIOB_PB7;
    else
	    PioB->ClearData = PIOB_PB7;

    stat = PioB->PinDataStatus;		
    SDA = stat & PIOB_PB7;
	
    wait(1);
    //SCL = HIGH;
    PioB->SetData = PIOB_PB6;
    wait(1);

}//end Send_Bit

/* Receieve Bit */
BOOL Receive_Bit()
{
    int stat;
    //SCL = LOW;
    PioB->ClearData = PIOB_PB6;
    wait(4);
    //SCL = HIGH;
    PioB->SetData = PIOB_PB6;
    //wait(1);
    //array[i] = SDA;
	
    stat = PioB->PinDataStatus;		
    
    wait(1);

    //array[i] = stat & PIOB_PB7;
    if((stat & PIOB_PB7) == 0)
        return FALSE;
    else
        return TRUE;

}//end Receive_Bit


 void Receive_ACK()
{
    //SCL = LOW;
	//PioB->ClearData = PIOB_PB6;
	//wait(t_f);
	wait(1);
	PioB->SetData = PIOB_PB6;
	wait(1);
}

/* Validate gyro output */
void Validate_gyro_Output()
{
	PioB->ClearData = PIOB_PB6;
    wait(1);
	PioB->OutDisable = PIOB_PB7;
}


/* Start command */
void START() {
    PioB->OutEnable = PIOB_PB7;
	wait(1);
	//SDA = HIGH;  //should be already
	PioB->ClearData = PIOB_PB6;
	wait(1);
	PioB->SetData = PIOB_PB7;
    wait(1);
	//SCL = HIGH;
	PioB->SetData = PIOB_PB6;
	wait(1);
	//SDA = LOW;
	PioB->ClearData = PIOB_PB7;
	wait(1);
}

/* Stop command */
void STOP() {

	PioB->OutEnable = PIOB_PB7;

	PioB->ClearData = PIOB_PB6;
	wait(1);
	//SDA = LOW;
	PioB->ClearData = PIOB_PB7;wait(1);
	//SCL = HIGH;
	PioB->SetData = PIOB_PB6;
	wait(1);
	//SDA = HIGH;
	PioB->SetData = PIOB_PB7;
}

UINT BytesToHalfword(UINT8 byte_H, UINT8 byte_L)
{
	INT decimal = 0;
	
	//conversion code
    decimal |= byte_H;
    decimal = decimal << 8;
    decimal |= byte_L;

	//return decimal value
	return decimal;	
}
