/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <dbres.h>
#include <stdlib.h>
#include<wsif.h>
#include<wsifhelp.h>

#define aOff offsetof(struct Alias, ListNode_Alias)
#define pOff offsetof(struct Partner, ListNode_Partner)
#define MAX_ARGUMENTS 10


PTR Alias_Copy(PTR a, PTR b)
{
    PAlias pos, src = a;
    PIHEAP Heap = b;

    UINT size = (_tcslen(src->Value) + 1) * sizeof(_TCHAR) + sizeof(*pos);
    pos = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:alias") | HEAP_ZERO_MEMORY, size, 0);
    if (!pos)
        return NULL;

    pos->Value = (PTR) (pos + 1);
    _tcscpy(pos->Value, src->Value);
    ListNodeInit(aOff, pos);
    return pos;
} 

INT Alias_Delete(PTR a, PTR b)
{
    PIHEAP Heap = b;
    Heap->v->Free(Heap, 0, a);
    return 0;
} 

PTR Partner_Copy(PTR a, PTR b)
{
    PPartner pos, src = a;
    PIHEAP Heap = b;

    UINT size = (_tcslen(src->Name) + 1) * sizeof(_TCHAR)
                +(_tcslen( src->ServiceDirectory) + 1) *sizeof(_TCHAR) 
                + sizeof(*pos);
    pos = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:partner") | HEAP_ZERO_MEMORY, size, 0);
    if (!pos)
        return NULL;

    pos->Name = (PTR) (pos+1);
    _tcscpy(pos->Name, src->Name);
    pos->ServiceDirectory = (PTR) (pos->Name +((_tcslen(src->Name) + 1) * sizeof(_TCHAR)));
    _tcscpy(pos->ServiceDirectory, src->ServiceDirectory);
    ListNodeInit(pOff, pos);
    return pos;
} 

INT Partner_Delete(PTR a, PTR b)
{
    PIHEAP Heap = b;
    
    Heap->v->Free(Heap, 0, a);
    return 0;
} 


PRIVATE SCODE ServiceInfoType_Copy(const PropertyClass *Class,
                                     PTR _Dst, const void *_Src, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(ServiceInfoType, Copy);
    struct ServiceInfoType *sit = _Dst;
    const struct ServiceInfoType *src = _Src;
    _TCHAR *tmp;
    sc = S_OK;

    /* Copy all the fields. */

#define S(f) if (sit->f) {Heap->v->Free(Heap, 0, sit->f); sit->f = NULL;} if (!src->f || !*src->f) {sit->f = NULL;} else {tmp = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:string"), (_tcslen(src->f)+1)*sizeof(_TCHAR), 0); CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "alloc string"); _tcscpy(tmp, src->f); sit->f = tmp;}

    S(Contract);
    S(Name);

    if (sit->AliasList) {
        ListDelete(&sit->AliasList, aOff,
                   Alias_Delete,
                   Heap);
    }
    if (src->AliasList) {
        tmp = ListCopy(&src->AliasList, aOff,
                       Alias_Copy,
                       Alias_Delete,
                       Heap);
        CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "copy AliasList");
        sit->AliasList = (PTR) tmp;
    } else {
        sit->AliasList = NULL;
    }

    if (sit->PartnerList) {
        ListDelete(&sit->PartnerList, pOff,
                   Partner_Delete,
                   Heap);
    }
    if (src->PartnerList) {
        tmp = ListCopy(&src->PartnerList, pOff,
                       Partner_Copy,
                       Partner_Delete,
                       Heap);
        CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "copy PartnerList");
        sit->PartnerList = (PTR) tmp;
    } else {
        sit->PartnerList = NULL;
    }

#undef S
METHOD_CLEANUP;
    return sc;
}


PRIVATE SCODE ServiceInfoType_Finalize(const PropertyClass *Class,
                                         PTR Value, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(ServiceInfoType, Finalize);
    struct ServiceInfoType *sit = Value;

    sc = S_OK;

#define S(f) if (sit->f) {Heap->v->Free(Heap, 0, sit->f); sit->f = NULL;}

    S(Contract);
    S(Name);
    if (sit->AliasList) {
        ListDelete(&sit->AliasList, aOff, Alias_Delete, Heap);
    }
    if (sit->PartnerList) {
        ListDelete(&sit->PartnerList, pOff, Partner_Delete, Heap);
    }
#undef S
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ServiceInfoType_Match(const PropertyClass *Class, PTR Value,
                                PSelector Selector)
{
    METHOD_DECLARE_NOTHIS(ServiceInfoType, Match);
    struct ServiceInfoType *sit = Value;

#define S(f) if (sit->f && !_tcscmp(_T(# f), Selector->Name)) return _tcscmp(Selector->Value, sit->f) ? E_FAIL : S_OK

    S(Contract);
    S(Name);
#undef S

    sc = E_NAME_NOT_FOUND;
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ServiceInfoType_Match2(const PropertyClass *Class, PTR Value,
                                const void *Value2, const UINT32 *FieldSel)
{
    METHOD_DECLARE_NOTHIS(ServiceInfoType, Match2);
    struct ServiceInfoType *sit = Value;
    struct ServiceInfoType *sit2 = (void *)Value2;

#define S(f) if (sit->f && sit2->f) return _tcscmp(sit->f, sit2->f) ? E_FAIL : S_OK
    S(Name);
#undef S

    sc = E_NAME_NOT_FOUND;
    METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ServiceInfoType_UpdateSelector(const PropertyClass *Class,
           PTR Value, PSelector Selector,
           UINT *pCount, UINT *pSize)
{
    METHOD_DECLARE_NOTHIS(ServiceInfoType, UpdateSelector);
    struct ServiceInfoType *sit = Value;

    if (pCount)
        *pCount = 1;
    if (pSize)
        *pSize = 0;
    if (Selector) {
        Selector->Name = _T("Name");
        Selector->Value = sit->Name;
    }

    sc = S_OK;
    METHOD_CLEANUP;
    return sc;
}

//PRIVATE _TCHAR* ServiceInfoType_ReturnShadowParentUrl(const PropertyClass *Class, PTR Value)
//{
//    
//    struct ServiceInfoType *sit = Value;
//    return sit->PartnerList[0].ServiceDirectory;
//    
//}

/*********** ServiceInfoType class ******************/


PRIVATE SCODE ServiceInfoType_Get(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest)
{
   
       return S_OK;
}

PRIVATE SCODE ServiceInfoType_Put(const PropertyClass *Class, PTR Value)
{
        return S_OK;
}

/* Initialize private fields */
PRIVATE SCODE ServiceInfoType_Init(const PropertyClass *Class, PTR Value,
                                /*unused*/ PIHEAP Heap, PICIMRESOURCE Resource )
{
    return S_OK;
}


const PropertyClass ServiceInfoType_Prop = {
    _T("dssp:ServiceInfoType"), 
    _T("ServiceInfoType"),
    sizeof(struct ServiceInfoType),
    _T("dssp:ServiceInfoType"),
    _T("dssp:ServiceInfoType"),
    NULL,
    ServiceInfoType_Get, 
    ServiceInfoType_Put, 
    ServiceInfoType_Init,
    ServiceInfoType_Finalize,
    ServiceInfoType_Copy,
    ServiceInfoType_Match,
    ServiceInfoType_Match2,
    ServiceInfoType_UpdateSelector,
    NULL, 
    NULL,
    NULL
};

