/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <dbres.h>
#include <stdlib.h>
#include <wsif.h>

PRIVATE SCODE TrustEvidence_Copy(const PropertyClass *Class,
                               PTR _Dst, const void *_Src, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(TrustEvidence, Copy);
    struct TrustEvidence *te = _Dst;
    const struct TrustEvidence *src = _Src;
    _TCHAR *tmp;

    sc = S_OK;
    CHECKSETSC(src->Name, E_INVALID_PARAMETER, "src->Name");

#define S(f) if (te->f) {Heap->v->Free(Heap, 0, te->f); te->f = NULL;} if (!src->f || !*src->f) {te->f = NULL;} else {tmp = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:string"), (_tcslen(src->f)+1)*sizeof(_TCHAR), 0); CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "alloc string"); _tcscpy(tmp, src->f); te->f = tmp;}

    /* Copy all the fields. */
    S(Name);
    S(RequiredEvidenceUrl);
    S(Url);

    te->Value = src->Value;
    te->Threshold = src->Threshold;
    
#undef S
   
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustEvidence_Finalize(const PropertyClass *Class,
                                         PTR Value, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(TrustEvidence, Finalize);
    struct TrustEvidence *te = Value;

    sc = S_OK;
#define S(f) if (te->f) {Heap->v->Free(Heap, 0, te->f); te->f = NULL;}
    
    S(Name);
    S(Url);
    S(RequiredEvidenceUrl);
#undef S
    
    METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustEvidence_Match(const PropertyClass *Class, PTR Value,
                                PSelector Selector)
{
    METHOD_DECLARE_NOTHIS(TrustEvidence, Match);
    struct TrustEvidence *te = Value;

#define S(f) if (te->f && !_tcscmp(_T(# f), Selector->Name)) return _tcscmp(Selector->Value, te->f) ? E_FAIL : S_OK
#define SPrefix(f) if (te->f && !_tcscmpprefix(_T(# f), Selector->Name)) return _tcscmpprefix(Selector->Value, te->f) ? E_FAIL : S_OK
#define N(f) if (!_tcscmp(_T(# f), Selector->Name)) return (_ttoi(Selector->Value) - te->f) ? E_FAIL : S_OK

    SPrefix(Name);
    S(Url);
    N(Value);
    N(Threshold);

#undef S
#undef SPrefix
#undef N
    sc = E_NAME_NOT_FOUND;
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustEvidence_Match2(const PropertyClass *Class, PTR Value,
                                const void *Value2, const UINT32 *FieldSel)
{
    METHOD_DECLARE_NOTHIS(TrustEvidence, Match2);
    struct TrustEvidence *te = Value;
    struct TrustEvidence *te2 = (void *)Value2;

#define S(f) if (te->f && te2->f) return _tcscmp(te->f, te2->f) ? E_FAIL : S_OK
#define SPrefix(f) if (te->f && te2->f) return _tcscmpprefix(te->f, te2->f) ? E_FAIL : S_OK
    SPrefix(Name);

#undef SPrefix
#undef S
    sc = E_NAME_NOT_FOUND;
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustEvidence_UpdateSelector(const PropertyClass *Class,
           PTR Value, PSelector Selector,
           UINT *pCount, UINT *pSize)
{
    METHOD_DECLARE_NOTHIS(TrustEvidence, UpdateSelector);
    struct TrustEvidence *te = Value;

    if (pCount)
        *pCount = 1;
    if (pSize)
        *pSize = 0;
    if (Selector) {
        Selector->Name = _T("Name");
        Selector->Value = te->Name;
    }
    sc = S_OK;
    METHOD_CLEANUP;
    return sc;
}

/*********** TrustEvidence class ******************/

PRIVATE SCODE TrustEvidence_Get(const PropertyClass *Class, PTR Value,CurrentRequest *currentRequest)
{
        
        struct TrustEvidence *te = Value;
        if(te->Url){
          PICIMRESOURCE Resource = NULL;
          PICIMRESOURCE Context = currentRequest->Context;
          struct Selector  selector;
          selector.Name = _T("Name");
          selector.Value = te->Url;
          selector.Type = NULL;
          selector.ListNode_Selector.pNext = &selector;
          selector.ListNode_Selector.pPrev = &selector;
          Resource = CimResourceShadowLookup(Context, _T("TrustScript"), &selector);
          if(Resource)
          {
            te->Value = CimResourceGetValue(Resource, currentRequest,_tcsreturnfield(te->Url));
            te->RequiredEvidenceUrl = ((TrustScript *)((CimResource *)Resource)->Value)->RequiredEvidenceUrl;
          }
          
        }
        return S_OK;
}

PRIVATE SCODE TrustEvidence_Put(const PropertyClass *Class, PTR Value)
{
        return S_OK;
}

/* Initialize private fields */
PRIVATE SCODE TrustEvidence_Init(const PropertyClass *Class, PTR Value,
                                /*unused*/ PIHEAP Heap,PICIMRESOURCE Resource)
{
    return S_OK;
}
PRIVATE UINT TrustEvidence_GetValue(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest,const _TCHAR *Field)
{
     struct TrustEvidence *te = Value;

     if(Field)
     {
        if(!_tcscmp(Field,_T("Value")))
            return te->Value;
        else if(!_tcscmp(Field,_T("Threshold")))
            return te->Threshold;
     }
    return te->Value;
}
const PropertyClass TrustEvidence_Prop = {
    _T("http://tempuri.org/trust/TrustEvidence"), 
    _T("TrustEvidence"),
    sizeof(struct TrustEvidence),
    _T("tpolicy:TrustEvidence"),
    _T("tpolicy:TrustEvidence"),
    NULL,
    TrustEvidence_Get, 
    TrustEvidence_Put, 
    TrustEvidence_Init,
    TrustEvidence_Finalize,
    TrustEvidence_Copy,
    TrustEvidence_Match,
    TrustEvidence_Match2,
    TrustEvidence_UpdateSelector,
    TrustEvidence_GetValue,
   
    NULL
};