/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <dbres.h>
#include <stdlib.h>
#include<wsif.h>

#define tsOff offsetof(struct TrustScript, ListNode)
#define croff offsetof(struct _CimResource, ListNode)
#define MAX_ARGUMENTS 10



INT TrustScript_Delete(PTR a, PTR b)
{
    TrustScript *ts = a;
    PIHEAP Heap = b;
    if (ts->ArgList) {
        ListDelete(&ts->ArgList, tsOff, TrustScript_Delete, Heap);
    }
    if (ts->Name)
        Heap->v->Free(Heap, 0, ts->Name);
    if (ts->RequiredEvidenceUrl)
        Heap->v->Free(Heap, 0, ts->RequiredEvidenceUrl);
   
    Heap->v->Free(Heap, 0, a);
    return 0;
} 

PRIVATE SCODE TrustScript_Copy(const PropertyClass *Class,
                               PTR _Dst, const void *_Src, PIHEAP Heap);

PTR TrustScript_CopySub(PTR a, PTR b)
{
    TrustScript *ts, *src = a;
    PIHEAP Heap = b;
    SCODE sc;

    UINT size = sizeof(*ts);
    ts = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:ts") | HEAP_ZERO_MEMORY, size, 0);
    if (!ts)
        return NULL;
    sc = TrustScript_Copy(NULL, ts, src, Heap);
    return ts;
} 

PRIVATE SCODE TrustScript_Copy(const PropertyClass *Class,
                               PTR _Dst, const void *_Src, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(TrustScript, Copy);
    struct TrustScript *ts = _Dst;
    const struct TrustScript *src = _Src;
    _TCHAR *tmp;
    sc = S_OK;

    CHECKSETSC(src->Name, E_INVALID_PARAMETER, "src->Name");
    
    /* Copy all the fields. */

#define S(f) if (ts->f) {Heap->v->Free(Heap, 0, ts->f); ts->f = NULL;} if (!src->f || !*src->f) {ts->f = NULL;} else {tmp = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cimdb:string"), (_tcslen(src->f)+1)*sizeof(_TCHAR), 0); CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "alloc string"); _tcscpy(tmp, src->f); ts->f = tmp;}

    S(Name);

    ts->RequiredEvidenceUrl = NULL; // XXX Used to propagate value from evidence only (maybe use CurrentRequest instead?)
    ts->Weight = src->Weight;

    ListNodeInit(tsOff, ts);
#if 1
    if (ts->ArgList) {
        ListDelete(&ts->ArgList, tsOff,
                   TrustScript_Delete,
                   Heap);
    }
#endif
    if (src->ArgList) {
        tmp = ListCopy(&src->ArgList, tsOff,
                       TrustScript_CopySub,
                       TrustScript_Delete,
                       Heap);
        CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "copy ts->ArgList");
        ts->ArgList = (PTR) tmp;
    } else {
        ts->ArgList = NULL;
    }    
#undef S

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustScript_Finalize(const PropertyClass *Class,
                                         PTR Value, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(TrustScript, Finalize);
    struct TrustScript *ts = Value;

    sc = S_OK;

    if (ts->Name)
        Heap->v->Free(Heap, 0, ts->Name);
   
    if (ts->ArgList) {
        ListDelete(&ts->ArgList, tsOff, TrustScript_Delete, Heap);
    }

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustScript_Match(const PropertyClass *Class, PTR Value,
                                PSelector Selector)
{
    METHOD_DECLARE_NOTHIS(TrustScript, Match);
    struct TrustScript *ts = Value;

#define S(f) if (ts->f && !_tcscmp(_T(# f), Selector->Name)) return _tcscmp(Selector->Value, ts->f) ? E_FAIL : S_OK
#define SPrefix(f) if (ts->f && !_tcscmpprefix(_T(# f), Selector->Name)) return _tcscmpprefix(Selector->Value, ts->f) ? E_FAIL : S_OK
#define N(f) if (!_tcscmp(_T(# f), Selector->Name)) return (_ttoi(Selector->Value) - ts->f) ? E_FAIL : S_OK

    SPrefix(Name);
    N(Weight);
   
#undef SPrefix
#undef S
#undef N
    sc = E_NAME_NOT_FOUND;
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustScript_Match2(const PropertyClass *Class, PTR Value,
                                const void *Value2, const UINT32 *FieldSel)
{
    METHOD_DECLARE_NOTHIS(TrustScript, Match2);
    struct TrustScript *ts = Value;
    struct TrustScript *ts2 = (void *)Value2;

#define S(f) if (ts->f && ts2->f) return _tcscmp(ts->f, ts2->f) ? E_FAIL : S_OK
#define SPrefix(f) if (ts->f && ts2->f) return _tcscmpprefix(ts->f, ts2->f) ? E_FAIL : S_OK

    SPrefix(Name);

#undef SPrefix
#undef S
    sc = E_NAME_NOT_FOUND;
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE TrustScript_UpdateSelector(const PropertyClass *Class,
           PTR Value, PSelector Selector,
           UINT *pCount, UINT *pSize)
{
    METHOD_DECLARE_NOTHIS(TrustScript, Match2);
    struct TrustScript *ts = Value;

    if (pCount)
        *pCount = 1;
    if (pSize)
        *pSize = 0;
    if (Selector) {
        Selector->Name = _T("Name");
        Selector->Value = ts->Name;
    }

    sc = S_OK;
METHOD_CLEANUP;
    return sc;
}

/*********** TrustScript class ******************/


PRIVATE SCODE TrustScript_Get(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest)
{
   
       return S_OK;
}

PRIVATE SCODE TrustScript_Put(const PropertyClass *Class, PTR Value)
{
        return S_OK;
}

/* Initialize private fields */
PRIVATE SCODE TrustScript_Init(const PropertyClass *Class, PTR Value,
                                /*unused*/ PIHEAP Heap,PICIMRESOURCE Resource)
{
    return S_OK;
}

PRIVATE UINT TrustScript_GetValue(const PropertyClass *Class, PTR Value,CurrentRequest *currentRequest,const _TCHAR *Field )
{
          
          struct TrustScript *ts = Value;
          UINT count ;
          UINT tempValue = 1;
          UINT values[MAX_ARGUMENTS];
            
          count = 0;
          if(ts->ArgList)
          {
               PLIST_HEAD Head;
               PTR pHead = &ts->ArgList;
               PTR pItem, Next;
              
               Head = pHead;
               pItem = *Head;
        
               if(!_tcscmp(ts->Name, _T("Select")))
               {
                    if (pItem) do {
                        if(count == currentRequest->Mode)
                        {
                            PICIMRESOURCE Resource = NULL;
                            PICIMRESOURCE Context = currentRequest->Context;
                            struct Selector  selector;
                            selector.Name = _T("Name");
                            selector.Value = ((struct TrustScript * )pItem)->Name;
                            selector.Type = NULL;
                            selector.ListNode_Selector.pNext = &selector;
                            selector.ListNode_Selector.pPrev = &selector;
                            Resource = CimResourceShadowLookup(Context, _T("TrustScript"), &selector);
                            if(Resource)
                            {
                                tempValue = CimResourceGetValue(Resource, currentRequest,_tcsreturnfield((const _TCHAR* )pItem));
                                ts->RequiredEvidenceUrl = ((TrustScript *)((CimResource *)Resource)->Value)->RequiredEvidenceUrl;
                            }
                            break;
                        }
                        Next = ListNodeNext(tsOff, pItem);
                        count++;
                        pItem = Next;
                   } while (*Head && pItem != *Head);
               }
               else
               {
                   if (pItem) do {
                        Next = ListNodeNext(tsOff, pItem);
                        values[count++] = TrustScript_GetValue(Class,Next,currentRequest,Field);
                        if(values[count-1]!=1)
                            ts->RequiredEvidenceUrl = ((TrustScript *)Next)->RequiredEvidenceUrl;
                        pItem = Next;
                   } while (*Head && pItem != *Head);
                   if(count == 1)
                   {
                        tempValue = values[0];
                   }else if (!_tcscmp(ts->Name, _T("Min")))
                   {
                      UINT i;
                      UINT min = values[0];
                      for(i = 0; i < count; i++)
                          if(values[i] <  min)
                              min = values[i];
                      tempValue = min;
                   }else if (!_tcscmp(ts->Name, _T("Max")))
                   {
                      UINT i;
                      UINT max = values[0];
                      for(i = 0; i < count; i++)
                          if(values[i] >  max)
                              max = values[i];
                      tempValue = max;
                   }else if (!_tcscmp(ts->Name, _T("Average")))
                   {
                      UINT i;
                      UINT avg = 0;
                      for(i = 0; i < count; i++)
                          avg = avg + values[i];
                      tempValue = avg/count;
                   }else if (!_tcscmp(ts->Name, _T("Multiply")))
                   {
                      UINT i;
                      UINT pdt = 1;
                      for(i = 0; i < count; i++)
                          pdt = pdt * values[i];
                      tempValue = pdt;
                   }else if (!_tcscmp(ts->Name, _T("Or")))
                   {
                      UINT i;
                      UINT flag = 0;
                      for(i = 0; i < count; i++)
                          if (values[i]>0)
                            flag++;
                      if(flag>0)
                          flag = 1;
                      tempValue = flag;
                   }else if (!_tcscmp(ts->Name, _T("And")))
                   {
                      UINT i;
                      tempValue = 1;
                      for(i = 0; i < count; i++)
                      {
                          if(values[i]==0)
                               tempValue = 0;
                      }
                   }
               }
                
            }
        else {
          PICIMRESOURCE Resource = NULL;
          PICIMRESOURCE Client = currentRequest->Client;
          struct Selector  selector;
          selector.Name = _T("Name");
          selector.Value = ts->Name;
          selector.Type = NULL;
          selector.ListNode_Selector.pNext = &selector;
          selector.ListNode_Selector.pPrev = &selector;
          Resource = CimResourceShadowLookup(Client, _T("TrustEvidence"), &selector);
          if(Resource)
          {
            tempValue = CimResourceGetValue(Resource, currentRequest,_tcsreturnfield(ts->Name));
            ts->RequiredEvidenceUrl = ((TrustEvidence *)((CimResource *)Resource)->Value)->RequiredEvidenceUrl;
          }
       }
       if(Field)
       {
           if(!_tcscmp(Field,_T("Weight")))
               return ts->Weight;
       }
       return tempValue;
}

const PropertyClass TrustScript_Prop = {
    _T("http://tempuri.org/trust/TrustScript"), 
    _T("TrustScript"),
    sizeof(struct TrustScript),
    _T("tpolicy:TrustScript"),
    _T("tpolicy:TrustScript"),
    NULL,
    TrustScript_Get, 
    TrustScript_Put, 
    TrustScript_Init,
    TrustScript_Finalize,
    TrustScript_Copy,
    TrustScript_Match,TrustScript_Match2,
    TrustScript_UpdateSelector,
    TrustScript_GetValue, 
  
    NULL
};

