/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <mmlite.h>
#include "convert.h"
#include <mmmacros.h>

/* Generate random UINT32 */
PRIVATE UINT32 RandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}

PRIVATE SCODE test1(const _TCHAR *p, const _TCHAR *p2)
{
    METHOD_DECLARE_NOTHIS(datetest.c, test1);
    TIME t, t2;
    _TCHAR buf[100];
    sc = TimeParse(p, &t);
    CHECKSC("TimeParse1");
    TimeFormat(t, buf, 99);
    printf ("Time %s -> x%016llx\n", p, t);
    if (_tcscmp(buf, p2 ? p2 : p))
        CHECKSETSC(FALSE, E_FAIL, (PTR/*XXX*/) buf);

    sc = TimeParse(p, &t2);
    CHECKSC("TimeParse2");
    CHECKSETSC(Int64Equal(t, t2), E_FAIL, "==");
    METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE test2(TIME t)
{
    METHOD_DECLARE_NOTHIS(datetest.c, test2);
    TIME t2;
    _TCHAR buf[100], buf2[100];

    TimeFormat(t, buf, 99);
    sc = TimeParse(buf, &t2);
    CHECKSC("TimeParse3");
    CHECKSETSC(Int64Equal(t, t2), E_FAIL, "==");

    TimeFormat(t2, buf2, 99);
    CHECKSETSC(!_tcscmp(buf, buf2), E_FAIL, (PTR/*XXX*/) buf);

    sc = TimeParse(buf2, &t2);
    CHECKSC("TimeParse4");
    CHECKSETSC(Int64Equal(t, t2), E_FAIL, "===");

    METHOD_CLEANUP;
    return sc;
}

#define T sc = test1(p, NULL); CHECKSC((PTR/*XXX*/) (p));
#define T2(_cmp_) sc = test1(p, (_cmp_)); CHECKSC((PTR/*XXX*/) (p));
#define U sc = test2(t); CHECKSC("test2");

int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(datetest, _tmain);
    _TCHAR buf[100], *p = 0;
    TIME t;
    INT i;
    sc = S_OK;
    ParseCmdLine(&argc, &argv);
    if (argc > 1 && argv[1][0] == _T('2')) {
        TIME now = CurrentTime();
        PICONVERT convert;
        sc = TimeParse(argv[1], &t);
        sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                          NAME_SPACE_READ, &IID_IConvert,
                          (void**) &convert);
        t = Int64Subtract(t, now);
        convert->v->SetTimeOffset(convert, t, FALSE);
    } else if (argc > 1)  {
        p = _T("P10.0100000S"); T;
        p = _T("P0.0000000S"); T;
        p = _T("P99999.9999999S"); T;

        p = _T("1600-12-31T16:04:02.7541644-08:00"); T2(_T("1601-01-01T00:04:02.7541644Z"));
        p = _T("2004-07-17T23:20:03.0400000Z"); T;
        p = _T("2002-02-26T18:39:30.0000700Z"); T;
        p = _T("2002-10-26T18:39:30.0000700Z"); T;
        p = _T("2003-07-18T01:00:00.0000000Z"); T;
        p = _T("2003-12-31T01:00:00.0000000Z"); T;
        p = _T("2002-09-26T18:39:30.0000700Z"); T;
        p = _T("2000-02-29T00:00:00.0000700Z"); T;
        p = _T("5500-12-31T23:59:59.9999999Z"); T;
        p = _T("1900-01-01T00:00:00.0000000Z"); T;
        p = _T("1968-01-20T03:14:08.0000000Z"); T;
        p = _T("2005-01-01T00:00:00.0000000Z"); T;

        t = TIME_ORIGIN; U;
        t = TIME_FOREVER; U;
        Int64FromHighAndLow(t, 0x3d3953ea, 0x0c0d3e84); U;
        Int64FromHighAndLow(t, 0xf019f657, 0xd0171f75); U;

        for (i = 0; i < 100000; i++) {
            Int64FromHighAndLow(t, RandUint32(), RandUint32());
            U;
        }
    }

    t = CurrentTime();
    TimeFormat(t, buf, 99);
    printf("Test passed! Now=%s t=%lld\n", buf, t);
        
    METHOD_CLEANUP;
    FreeCmdLine(&argc,&argv);
    if (FAILED(sc))
        printf("datetest: last t=x%016llx = %lld\n", t, t);
    return sc;
}

