/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <mmlite.h>
#include "convert.h"

PRIVATE INT8 days[2][12] = {{31,28,31,30,31,30, 31,31,30,31,30,31},
                            {31,29,31,30,31,30, 31,31,30,31,30,31}};
PRIVATE char *DayName[7] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

PRIVATE BOOL IsLeap(UINT y)
{
    BOOL is = (!(y%4) && ((y%100) || !(y%400)));
    //DBG(("IsLeap(%d) => %d\n", y, is));
    return is;
}

SCODE TimeParse(const _TCHAR *buf, TIME *tret)
{
    TIME t, tmp;
    INT leaps, f,s,min,h,d,mo,y, r;
    INT tzh, tzmin;
    BOOL relative = FALSE;
    const _TCHAR *p;

    if (!buf || !*buf) {
        *tret = TIME_ORIGIN;
        return S_OK;
    }

    if (buf[0] == 'P') {
        buf++;
        if (buf[0] == 'T')
            buf++;

        relative = TRUE;
        /* Only allow P1000.100S or PT1000.100S for now */
        f = 0;
        t = TIME_ORIGIN;

        tmp = _tcstoi64(buf, (_TCHAR **) &buf, 10);
        if (*buf == 'W') {
            tmp = Int64TimesInt32(tmp, 6048000);
            tmp = Int64TimesInt32(tmp, 1000000);
            t = Int64Add(t, tmp);
            buf++;
            tmp = _tcstoi64(buf, (_TCHAR **) &buf, 10);
        }
        if (*buf == 'D') {
            tmp = Int64TimesInt32(tmp, 864000000);
            tmp = Int64TimesInt32(tmp, 1000);
            t = Int64Add(t, tmp);
            buf++;
            tmp = _tcstoi64(buf, (_TCHAR **) &buf, 10);
        }
        if (*buf == 'H') {
            tmp = Int64TimesInt32(tmp, 3600000000);
            tmp = Int64TimesInt32(tmp, 10);
            t = Int64Add(t, tmp);
            buf++;
            tmp = _tcstoi64(buf, (_TCHAR **) &buf, 10);
        }
        if (*buf == 'M') {
            tmp = Int64TimesInt32(tmp, 600000000);
            t = Int64Add(t, tmp);
            buf++;
            tmp = _tcstoi64(buf, (_TCHAR **) &buf, 10);
        }
        if (*buf == '.') {
            INT e = 10000000;
            for (buf++; *buf >= '0' && *buf <= '9'; buf++)
                e /= 10;

            if (*buf != 'S')
                return E_INVALID_PARAMETER;

            while (*--buf != '.') {
                f += (*buf - '0') * e;
                e *= 10;
            }
        } else if (*buf != 'S') /* trailing junk */
            return E_INVALID_PARAMETER;

        tmp = Int64TimesInt32(tmp, 10000000);
        t = Int64Add(t, tmp);
        t = Int64AddInt32(t, f);
        t = Int64Subtract(TIME_ORIGIN, t);
    } else {
        r = _stscanf(buf,_T("%d-%02d-%02dT%02d:%02d:%02d.%dZ"),
                     &y,&mo,&d,&h,&min,&s,&f);
        if (y < 1600) {
            *tret = TIME_ORIGIN;
            return S_OK;
        }
        if (r != 7 || mo < 1 || d < 1 || h < 0 || min < 0 || s < 0 || f < 0)
            return E_INVALID_PARAMETER;

        /* Figure out after decimal point if applicable */
        f = 0;
        p = _tcspbrk(buf, _T("."));
        if (p) {
            INT e = 10000000;
            buf = p;
            for (buf++; *buf >= '0' && *buf <= '9'; buf++)
                e /= 10;

            while (*--buf != '.') {
                f += (*buf - '0') * e;
                e *= 10;
            }
        }
        /* Time zone */
        p = _tcspbrk(buf, _T("+-Z"));
        if (!p)
            return E_INVALID_PARAMETER;

        tzh = tzmin = 0;
        if (*p != 'Z') {
            r = _stscanf(p, _T("%d:%d"), &tzh, &tzmin);
            if (r != 2 || tzh > 23 || tzh < -23 || tzmin < 0 || tzmin > 59)
                return E_INVALID_PARAMETER;

            if (tzh < 0)
                tzmin = -tzmin;
        }

        /* Deal with straddle just before new year 1601 */
        if (y < 1601) {
            if (d!=31 || mo!=12 || y!=1600 || (h*60+min)-(tzh*60+tzmin) < 0)
                return E_INVALID_PARAMETER;
            d=0; mo=1; y=1601;
        }

        /* Make them zero based */
        d--;
        mo--;
        y -= 1601;

        /* Days within this year */
        leaps = IsLeap(y+1601);
        //printf("AAA d=%d mo=%d y=%d leap=%d h=%d tzh=%d tzmin=%d\n", d,mo,y,leaps,h,tzh,tzmin);
        for (mo--; mo >= 0; mo--) /* previous months */
            d += days[leaps][mo];

        leaps = y / 4 - y / 100 + y / 400; /* not counting this year */
        d += 365*y + leaps;
        s = s + 60*(min-tzmin) + 3600*(h-tzh);
        //printf("BBB d=%d s=%d f=%d\n", d, s, f);

        Int32ToInt64(t, d);
        t = Int64TimesInt32(t, 86400);
        t = Int64AddInt32(t, s);
        t = Int64TimesInt32(t, 10000000);
        t = Int64AddInt32(t, f);
		if (y > 10000000) /* avoid overflow with date conversion */
			t = TIME_FOREVER;
		else if (Int64IsNegative(t))  /* underflowed straddle to 1601 */
            return E_INVALID_PARAMETER;

        //printf("day=%d leaps=%d y=%d  f=%d r=%d\n", d, leaps, y, f, r);
    }

    *tret = t;
    return S_OK;
}

void TimeFormat(TIME t, _TCHAR *buf, UINT Len)
{
    TIME t2;
    INT leaps, f,s,min,h,d,mo,y, wd, xxx;
    BOOL relative = TimeIsRelative(t) || Int64Equal(t, TIME_ORIGIN);

    if (relative)
        t = Int64Subtract(TIME_ORIGIN, t);

    /* First deal with fractions */
    t2 = Int64ModulusInt32(t,  10000000);
    t = Int64DividedByInt32(t, 10000000);
    f = Int64ToInt32(t2);

    if (relative) {
        _stprintf(buf, _T("PT%lld.%07dS"), t,f);
        return;
    }

    /* seconds */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    s = Int64ToInt32(t2);

    /* minutes */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    min = Int64ToInt32(t2);

    /* hours */
    t2 = Int64ModulusInt32(t, 24);
    t = Int64DividedByInt32(t, 24);
    h = Int64ToInt32(t2);
    
#if !defined(arm)               /* arm_arm generates call to __rt_divtest */
    /* days since Jan 1, 1601, Gregorian calendar. */
    xxx = wd = Int64ToInt32(t);
    wd %= 7;
    wd++;
#endif

    /* figure out rough year, then adjust with leap days */
    t2 = Int64ModulusInt32(t, 365);
    t = Int64DividedByInt32(t, 365);
    d = Int64ToInt32(t2);

    y = Int64ToInt32(t);        /* from 1601 */

    leaps = y / 4 - y / 100 + y / 400; /* from 1600, does not include current year */

    d -= leaps;

    //printf("y=%d leaps=%d now=%d d=%d wd=%d [%s] totaldays=%d\n",
             //y, leaps, (y%4 == 0), d, wd, DayName[wd-1], xxx);

    while (d < 0) {
      /* if we are millenia in the future there might be more bogus leaps
       * than one year worth */
        leaps = IsLeap(y+1601-1);       /* last year */
        d += (leaps ? 366 : 365);
        y--;
    }
    leaps = IsLeap(y+1601);       /* this year */

    for (mo = 0; mo < 12; ) {
        if (d < days[leaps][mo])
            break;
        d -= days[leaps][mo];
        mo++;
    }
    /* make them one-based */
    mo++;
    d++;
    y += 1601;

    /* Time zone: UTC (boot time?) */
    _stprintf(buf,_T("%04d-%02d-%02dT%02d:%02d:%02d.%07dZ"), y,mo,d,h,min,s,f);
}
