/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Internal state of deserializer -jvh 2003 */

#ifndef __DESERIALIZER_H__
#define __DESERIALIZER_H__ 1

#include <util/list-double2.h>

extern int __cdecl _ttoin(const _TCHAR *String, UINT Length);

enum deser_state {
    DS_EMPTY,                   /* 0 */
    DS_ENVELOPE,                /* 1 */
    DS_HEADER,                  /* 2 */
    DS_BODY,                    /* 3 */
    DS_METHOD,                  /* 4 */
    DS_STRUCT,                  /* 5 */
    DS_ARG,                     /* 6 */
    DS_ARG_RECEIVED,            /* 7 */
    DS_ARRAY,                   /* 8 */
    DS_ARRAY_ITEM,              /* 9 */
    DS_ARRAY_ITEM_RECEIVED,     /* 10 */
    DS_LIST                     /* 11 */
};

/* Saved states */
#define DSS_HEADER       0x01   /* else in body */
#define DSS_STRUCT       0x02   /* else method */
#define DSS_LIST         0x04   /* inside list */
#define DSS_NESTED       0x08   /* some state stored in href */
#define DSS_NAMELESSARG  0x10   /* processing inside nameless listhead */
#define DSS_TO_NAMELESS  0x20   /* called nameless listhead */

typedef struct _Deserializer Deserializer, *PDESERIALIZER;
struct _Deserializer {
    struct ISaxHandlerVtbl const *v;
    UINT   Refs;

    UINT       isResponse;
    struct TypeInfo *ptypeinfo;
    PICONTINUATION   prpcinfo;
    PCONTINFO   pCont;           /* continuation when client */
    PIHEAP Heap;                /* Heap to use for temporary allocations */
    BOOL ultimateReceiver; /* The role of this node as indicated by <Path> */
    enum _ROLEACTION {RA_doit, RA_forward, RA_skip} RoleAction;
    UINT skiptoxmldepth;
    UINT xmldepth;
    UINT SaxState;

    PHREF       hreflist; /* keep the IDs and the address in stack for input arrays and structs */

    struct _CurrentState {
        UINT State;
        UINT SavedState;
        CPMETHOD_DESCR metd;
        CPARGUMENT_DESCR argd;
        DESCR_TYPE     arraytype;
        UINT           arraysize;
        UINT           arrayindex;
        PHREF          href;
#define MAX_SELECTOR 8
        UINT32 _SeenItems[MAX_SELECTOR]; /* XXX 256 arguments will have to do for now */
        UINT32 *SeenItems;
#define MAX_STRUCT_NESTING 8

        /* HREF for current nested list and struct */
        struct HREF listhref;
    } _current[MAX_STRUCT_NESTING], *current;

    const _TCHAR* reqLocalURL;
    BOOL ShouldFreeURL;         /* TRUE: Free above. FALSE: don't */
    const _TCHAR *UriQuery;

    PISOAPSERIALIZER Serializer; /* Where the output message goes */
    /* State saved until prpcinfo created */
    CPXMLNAMESPACE envelopeNS;
    PXMLNAMESPACE namespaces;

#define MAXNSCOUNT 100
    UINT ActiveNsCount;
    struct _XMLNAMESPACEREF ActiveNamespaces[MAXNSCOUNT];
};

#define pDeserializer(_i_) ((PDESERIALIZER)(_i_))
#define iDeserializer(_p_) ((PISAXHANDLER)(_p_))
#define pWsRouting(_i_) ((PDESERIALIZER)((_i_) ? (char*)(_i_) - offsetof(struct _PDESERIALIZER, WsRouting_v) : NULL))
#define iWsRouting(_p_) ((PIWSROUTING)((_p_) ? &(_p_)->WsRouting_v : NULL))

/* WS-Routing */
SCODE isLocalCall(PDESERIALIZER Handler);
SCODE ForwardMsg(PDESERIALIZER Handler, PIBUFFER *pContent);
#endif 
