/* Copyright (c) Microsoft Corporation. All rights reserved. */
//#define _DEBUG 1
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <netif.h>              /* for IBuffer */
#include "convert.h"
#include "tokens.h"

#include "xns.h"                /* for GetBottom */

#include <s_xsi_t.h>
#include <s_xsd_t.h>
#include <s_soap11_t.h>
#include <s_soapenc_t.h>

#define MAXDATELENGTH 30

PRIVATE SCODE Start(PISOAPSERIALIZER pThis,
                    PSAXATTRIBUTE Elem,
                    PSAXATTRIBUTE Attr, UINT Count,
                    CPXMLNAMESPACE EncodingNS)
{
    /* Output encodingstyle when soap11 */
    if (!_tcscmp(Elem->Xns->QName,
                 _T("http://schemas.xmlsoap.org/soap/envelope/")))
    {
        Attr[Count].Prefix = Elem->Prefix;
        Attr[Count].Xns = Elem->Xns;
        Attr[Count].Token = TOKEN_soap11_encodingStyle;
        Attr[Count].Name = _T("encodingStyle");
        Attr[Count].Value = EncodingNS->QName;
        Count++;
    }

    return pThis->v->StartElement(pThis, Elem, Attr, Count);
}

PRIVATE SCODE End(PISOAPSERIALIZER pThis, PSAXATTRIBUTE Elem)
{
    return pThis->v->EndElement(pThis, Elem);
}

PRIVATE SCODE SoapSerializerQueryInterface(
    /*in*/ PISOAPSERIALIZER pThis, 
    /*in*/ REFIID Iid,
    /*out*/void **Unk)
{
    SCODE sc;
    sc = GenericQueryInterface((PIUNKNOWN) pThis, Iid, Unk, &IID_ISoapSerializer);
    if (sc == E_NO_INTERFACE)
        sc = GenericQueryInterface((PIUNKNOWN) pThis, Iid, Unk, &IID_ISaxHandler);
    return sc;
}

PRIVATE SCODE SoapSerializerGetNamespace (
    PISOAPSERIALIZER pThis,
    const _TCHAR * Prefix,
    /*out*/ CPXMLNAMESPACE * pNamespace)
{
    CPXMLNAMESPACE Xns;

    Xns = GetNamespace(Prefix, 0);
    if (!Xns)
        return E_NAME_NOT_FOUND;

    *pNamespace = Xns;
    return S_OK;
}

PRIVATE SCODE SoapSerializerStartEnvelope (
    PISOAPSERIALIZER pThis,
    PICONTINUATION iCont,
    CPXMLNAMESPACE Prefix,
    void *pNS,
    CPXMLNAMESPACE EncodingNS)
{
    PCONTINFO cont = pCONT(iCont);
    CPXMLNAMESPACE methodNS = cont->methodNS;
#define MAXATTRIBUTES 20
    SAXATTRIBUTE Attributes[MAXATTRIBUTES], *Attr, _elem, *Elem = &_elem;

    UINT Count = 0;
    BOOL xsiSeen = FALSE;
    BOOL xsdSeen = FALSE;
    BOOL soapencSeen = FALSE;
    BOOL soapSeen = FALSE;
    BOOL wsaSeen = FALSE;
#ifndef _MINIMIZE
    BOOL wsmanSeen = FALSE;
#endif
    BOOL methodNSSeen = FALSE;
    CPXMLNAMESPACE pns;

    memset(Elem, 0, sizeof(*Elem));
    Attr = &Attributes[0];

    /* <?xml version="1.0" ?> */
    Attr->Xns = XMLNS_EMPTY;
    Attr->Prefix = _T("");
    Attr->Name = _T("version");
    Attr->Token = TOKEN_invalid;
    Attr->Value = _T("1.0");
    Elem->Name = _T("xml");
    Elem->Token = TOKEN_invalid;
    pThis->v->StartInstruction(pThis, Elem, Attributes, 1);
    pThis->v->EndInstruction(pThis, Elem);

    /* <Envelope namespaces... */
    for (pns = (CPXMLNAMESPACE) pNS; pns; pns = pns->next) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        /* XXX Assume that most namespaces are garbage */
        if (pns->nstype == NSTYPE_UNRESOLVED)
            continue;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = pns->name;
        Attr->Token = TOKEN_invalid;
        Attr->Value = pns->QName;
        Count++, Attr++;

        if (!_tcscmp(pns->name, _T("xsi")))
            xsiSeen = TRUE;
        if (!_tcscmp(pns->name, _T("xsd")))
            xsdSeen = TRUE;
        if (!_tcscmp(pns->name, _T("soapenc")))
            soapencSeen = TRUE;
        if (!_tcscmp(pns->name, Prefix->name ? Prefix->name : _T("soap")))
            soapSeen = TRUE;
        if (!_tcscmp(pns->name, _T("wsa")))
            wsaSeen = TRUE;
#ifndef _MINIMIZE
        if (!_tcscmp(pns->name, _T("wsman")))
            wsmanSeen = TRUE;
#endif
        if (pns == methodNS)
            methodNSSeen = TRUE;
    }

    /* Always define xsd,xsi,soapenc namespaces so that types can be tagged. */
    if (!xsdSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = _T("xsd");
        Attr->Token = TOKEN_invalid;
        Attr->Value = _T("http://www.w3.org/2001/XMLSchema");
        Count++;
        Attr++;
    }
    if (!xsiSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = _T("xsi");
        Attr->Token = TOKEN_invalid;
        Attr->Value = _T("http://www.w3.org/2001/XMLSchema-instance");
        Count++;
        Attr++;
    }

#if 0
    if (!soapencSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = _T("soapenc");
        Attr->Token = TOKEN_invalid;
        Attr->Value = _T("http://schemas.xmlsoap.org/soap/encoding/");
        Count++;
        Attr++;
    }
#endif
#if 1                           /* XXX for Indigo client */
    if (!soapSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = Prefix->name;
        Attr->Token = TOKEN_invalid;
        Attr->Value = Prefix->QName;
        Count++;
        Attr++;
    }
#endif
    if (!wsaSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = _T("wsa");
        Attr->Token = TOKEN_invalid;
        Attr->Value = _T("http://schemas.xmlsoap.org/ws/2004/08/addressing");
        Count++;
        Attr++;
    }
#ifndef _MINIMIZE
    if (!wsmanSeen) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = _T("wsman");
        Attr->Token = TOKEN_invalid;
        Attr->Value = _T("http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd");
        Count++;
        Attr++;
    }
#endif
    if (!methodNSSeen && methodNS != Prefix) {
        if (Count >= MAXATTRIBUTES)
            return E_OUT_OF_STRUCTURES;

        Attr->Xns = XMLNS_DECLARE;
        Attr->Prefix = _T("xmlns");
        Attr->Name = methodNS->name;
        Attr->Token = TOKEN_invalid;
        Attr->Value = methodNS->QName;
        Count++;
        Attr++;
    }

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Envelope");
    return Start(pThis, Elem, Attributes, Count, EncodingNS);
}

PRIVATE SCODE SoapSerializerEndEnvelope (
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix)
{
    SAXATTRIBUTE _elem, *Elem = &_elem;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Envelope");
    return End(pThis, Elem);
}

PRIVATE SCODE SoapSerializerStartBody (
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix,
    CPXMLNAMESPACE EncodingNS)
{
    SAXATTRIBUTE _elem, *Elem = &_elem, Attr;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Body");
    return Start(pThis, Elem, &Attr, 0, EncodingNS);
}

PRIVATE SCODE SoapSerializerEndBody (    
     PISOAPSERIALIZER pThis,
     CPXMLNAMESPACE Prefix)
{
    SAXATTRIBUTE _elem, *Elem = &_elem;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Body");
    return End(pThis, Elem);
}

PRIVATE SCODE SoapSerializerStartHeader (
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix,
    CPXMLNAMESPACE EncodingNS)
{
    SAXATTRIBUTE _elem, *Elem = &_elem, Attr;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Header");
    return Start(pThis, Elem, &Attr, 0, EncodingNS);
}

PRIVATE SCODE SoapSerializerEndHeader (    
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix)
{
    SAXATTRIBUTE _elem, *Elem = &_elem;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix;
    Elem->Token = TOKEN_invalid; Elem->Name = _T("Header");
    return End(pThis, Elem);
}

#ifndef _MINIMIZE
PRIVATE SCODE SoapSerializerStartFault (
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix,
    CPXMLNAMESPACE encodingNS,
    const _TCHAR* FaultCode,
    const _TCHAR* FaultString,
    const _TCHAR* FaultActor)
{
    SAXATTRIBUTE _elem, *Elem = &_elem, Attr;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix; Elem->Token = TOKEN_soap11_Fault; Elem->Name = _T("Fault");
    Start(pThis, Elem, &Attr, 0, encodingNS);
    Elem->Xns = Prefix;
    Elem->Token = TOKEN_soap11_faultcode; Elem->Name = _T("faultcode");
    pThis->v->StartElement(pThis, Elem, NULL, 0);

    if(Prefix && _tcslen(Prefix->name)) {
        pThis->v->Text(pThis, Prefix->name, _tcslen(Prefix->name), NULL, 0);
        pThis->v->Text(pThis, _T(":"), 1, NULL, 0);
    }
    pThis->v->Text(pThis, FaultCode, _tcslen(FaultCode), NULL, 0);

    pThis->v->EndElement(pThis, Elem);

    if (FaultString) {
        Elem->Xns = encodingNS; Elem->Prefix = _T("");
        Elem->Token = TOKEN_soap11_faultstring; Elem->Name = _T("faultstring");
        pThis->v->StartElement(pThis, Elem, NULL, 0);
        pThis->v->Text(pThis, FaultCode, _tcslen(FaultString), NULL, 0);
        pThis->v->EndElement(pThis, Elem);
    }    
    return S_OK;
}

PRIVATE SCODE SoapSerializerEndFault (
    PISOAPSERIALIZER pThis,
    CPXMLNAMESPACE Prefix)
{
    SAXATTRIBUTE _elem, *Elem = &_elem;
    memset(Elem, 0, sizeof(*Elem));

    Elem->Xns = Prefix; Elem->Token = TOKEN_soap11_Fault; Elem->Name = _T("Fault");
    return End(pThis, Elem);
}
#else /*!_MINIMIZE*/
#define SoapSerializerStartFault (_ISoapSerializer_StartFault_METHOD_TYPE) UNIMPLEMENTEDMethodSCODE
#define SoapSerializerEndFault (_ISoapSerializer_EndFault_METHOD_TYPE) UNIMPLEMENTEDMethodSCODE
#endif

#define rloff offsetof(struct reservation, ListNode)

void SoapTimeFormat(TIME t, _TCHAR *buf, UINT Len)
{
    if (TimeIsPositive(t)) {
        t = Int64Add(t, GetSoapTimeOffset());
        /* Make sure we don't under/overflow */
        if (Int64IsNegative(t)) {
            if (Int64IsNegative(GetSoapTimeOffset()))
                t = TIME_ORIGIN;
            else
                t = TIME_FOREVER;
        }
    }
    TimeFormat(t, buf, Len);
}

/* Attr points to Attribute list. Count is number of attributes so far. */
PRIVATE SCODE SoapSerializerSimpleType(PISOAPSERIALIZER serializer,
        UINT /*DESCR_TYPE*/ argtype, UINT /*ARGUMENT_DESCR_FLAGS*/ Flags,
        PTR where, UINT Length,
        PSAXATTRIBUTE Elem,
        PSAXATTRIBUTE Attr, UINT Count,
        _TCHAR *intstr, UINT intstrsize)
{
    _TCHAR *value, *xsd;
    SCODE sc;
    UINT xt = TOKEN_invalid, len;
    PIBUFFER im = NULL;

    assert (intstrsize >= MAXDATELENGTH);

    switch (argtype) {
    case ARG_INT:
        value = _uitotcs(*(INT*)where, 10, intstr + 23);
        xsd = _T("xsd:int");
        break; 

    case ARG_FLOAT: {
        UINT* addr = where;
        if (*addr == INT_MAX)
            value = _T("NaN");
        else if (*addr == INT_MAX-1)
            value = _T("INF");
        else if (*addr == INT_MIN)
            value = _T("-INF");
        else {
            value = _uitotcs(*(INT*)where, 10, intstr + 23);
            _tcscat(value, _T(".0")); /* floats get _T("truncated") */
        }
        xsd = _T("xsd:float");
        xt = TOKEN_xsd_float;
        break;
    }
    case ARG_INT64:
        /* There are 20 digits in a 64 bit decimal number */
        value = _i64totcs(*(INT64*)where, 10, intstr + 23);
        xsd = _T("xsd:decimal");
        xt = TOKEN_xsd_decimal;
        break;

    case ARG_BOOL: {
        BOOL* addr = where;
        if (*addr)
            value = _T("true");
        else
            value = _T("false");
        xsd = _T("xsd:boolean");
        xt = TOKEN_xsd_boolean;
        break;
    }
    case ARG_QNAME:
    case ARG_STRING:
        value = (_TCHAR*) where;
        if (!value)
            value = _T("");
        xsd = _T("xsd:string");
        xt = TOKEN_xsd_string;
        break;

    case ARG_IBUFFER:
        im = *(PIBUFFER *) where;
        value = NULL;
        xsd = _T("xsd:string");
        xt = TOKEN_xsd_string;
        break;

    case ARG_URI:
        value = (_TCHAR*) where;
        if (!value)
            value = _T("");
        xsd = _T("xsd:anyURI");
        xt = TOKEN_xsd_anyURI;
        break;
                                  
    case ARG_DATETIME: {
        TIME *addr = where;
        value = intstr;
        SoapTimeFormat(*addr, intstr, intstrsize-1);
        if (TimeIsRelative(*addr)) {
            xsd = _T("xsd:duration");
            xt = TOKEN_xsd_duration;
        } else {
            xsd = _T("xsd:dateTime");
            xt = TOKEN_xsd_dateTime;
        }
        //DBGT((_T("SerializeSimpleType(%016llx) -> %s:%s\n"),
        //      *addr, xsd, intstr));
        break;
    }
    case ARG_CONTINUATION:
    case ARG_LISTNODE:
        /* skip it */
        return S_OK;

    default: /* given as pointers */
        value = where;
        xsd = _T("xsd:BOGUS");
        DebugBreak();
    }

    if (Elem && Attr) {
#if 0
        Attr[Count].Xns = GetNamespace(_T("xsi"), 3);
        Attr[Count].Prefix = _T("xsi");
        Attr[Count].Token = TOKEN_xsi_type;
        Attr[Count].Name = _T("type");
        Attr[Count].Value = xsd;
        Count++;
#endif
    } else if (Attr) {
        Attr[Count].Value = value;
    }

    sc = S_OK;
    if (Elem)
        serializer->v->StartElement(serializer, Elem, Attr, Count);
    if (im) {
        sc = serializer->v->Text(serializer, NULL, 0, im, Flags);
        im->v->Release(im);
        /* XXX if _OUT, may need to zero value now (MissingArg covers?) */
    } else if ((Elem || !Attr) && value) {
        /* The length is min(strlen, length-is) */
        len = _tcslen(value);
        if (len > Length)
            len = Length;
        sc = serializer->v->Text(serializer, value, len, NULL, 0);
    }
    if (Elem)
        sc = serializer->v->EndElement(serializer, Elem);
    return sc;
}

PRIVATE SCODE SoapSerializerBlob(PISOAPSERIALIZER serializer,
                                 PTR where, UINT total,
                                 PSAXATTRIBUTE Elem)
{
    _TCHAR buf[80];
    UINT size, count;
    SCODE sc = S_OK;
    BYTE *p = where;

    if (Elem)
        sc = serializer->v->StartElement(serializer, Elem, NULL, 0);
    while (total > 0 && !FAILED(sc)) {
        count = total;
        if (count > 48)
            count = 48;
        sc = base64Coder(NULL, CYPHER_ENCRYPT,
                         p, count,
                         NULL,
                         (PTR) buf, 80,
                         &size);
        assert(!FAILED(sc));
        sc = serializer->v->Text(serializer, buf, size, NULL, 0);
        total -= count;
        p += count;
    }

    if (Elem)
        sc = serializer->v->EndElement(serializer, Elem);
    return sc;
}

PRIVATE SCODE SoapSerializerComplexRef(PISOAPSERIALIZER serializer,
                                       PSAXATTRIBUTE Elem, UINT id,
                                       CPXMLNAMESPACE encodingNS)
{
    _TCHAR buf[15];
    SAXATTRIBUTE Attributes[1];

    _stprintf(buf, _T("#id%d"), id);

    Attributes[0].Prefix = _T("");
    Attributes[0].Xns = encodingNS;
    Attributes[0].Token = TOKEN_soapenc_href;
    Attributes[0].Name = _T("href");
    Attributes[0].Value = buf;
    serializer->v->StartElement(serializer, Elem, Attributes, 1);
    return serializer->v->EndElement(serializer, Elem);
}

void DestroyHRefFromList(PHREF *pphref, PHREF del, PIHEAP Heap)
{
    PHREF phref, tmp;
    for (phref  = *pphref; phref; phref = tmp) {
        tmp = phref->next;
        if (tmp == del)
            phref->next = tmp->next;
        if (phref == del) {
            if (del == *pphref)
                *pphref = tmp;
            Heap->v->Free(Heap, 0, phref);
            return;
        }
    }
}

PRIVATE SCODE ItemHelper(PISOAPSERIALIZER serializer,
                         PHREF phref, PUINT phrefn,
                         CPARGUMENT_DESCR itemd, PTR a,
                         SAXATTRIBUTE *Elem, SAXATTRIBUTE *Attrib,
                         _TCHAR *intstr, CPXMLNAMESPACE encodingNS)
{
    DESCR_TYPE Type = itemd->descr_type;
    CPMETHOD_DESCR structd = NULL;
    CPXMLNAMESPACE Xns = NULL;
    UINT32 *FieldSelector = NULL;
    UINT Length = UINT_MAX;

    /* Fetch field selector if specified */
    if (itemd->SelectorIsArg) {
        DebugBreak();
        FieldSelector = *(UINT32**)((PBYTE) a +
                                    (itemd+itemd->SelectorIsArg)->FrameOffset);
        assert(itemd->FrameOffset == 0); /* XXX tmp, check above */
    }

    /* Fetch length-is if specified */
    if (itemd->LengthIsArg) {
        DebugBreak();
        Length = *(UINT*)((PBYTE) a
                          + (itemd+itemd->LengthIsArg)->FrameOffset
                          - itemd->FrameOffset);
    }

    /* Fetch type descriptor if specified */
    if (itemd->TypeIsArg) {
        const _TCHAR *typestring;
        UINT len;

        assert(itemd->descr_type == ARG_ANYTYPE);

        typestring = *(const _TCHAR **)((PBYTE)a + (itemd+itemd->TypeIsArg)->FrameOffset); /* XXX - itemd->FrameOffset? */
        assert(itemd->FrameOffset == 0); /* XXX tmp, check above */
        if (typestring) {
            len = _tcscspn(typestring, _T(":"));
            if (typestring[len] != _T(':') || typestring[len+1] == _T('\0'))
                return E_INVALID_PARAMETER; /* XXX could try SoapAction style instead */
            Xns = GetNamespace(typestring, len);
            if (!Xns)
                return E_NAME_NOT_FOUND;
            structd = FindMethod(Xns, typestring + len + 1, FALSE);
            if (!structd)
                return E_NAME_NOT_FOUND;
            Type = ARG_STRUCT;
        } else {
            Type = ARG_STRING;
        }
    } else {
        assert(itemd->descr_type != ARG_ANYTYPE);
    }

    if (itemd->descr_type == ARG_LISTHEAD) {
        PTR node = *(PTR*)a;
        if (node) {
            /* In case wrapper element is desired: output it now */
            if ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0)
                serializer->v->StartElement(serializer, Elem, NULL, 0);
            do {
                /* serialize node; use tmp href as ::Complex expects it */
                PHREF phref2 = MakeHref(&phref, CurrentHeap(),
                                        NULL, 0, node,
                                        NULL, 0, NULL, itemd, phref, 0, 0);
                phref2->Xns = phref->Xns;
                if (itemd->_structXns)
                    phref2->Xns = (void*) itemd->_structXns;
                phref->FieldSelector = FieldSelector;
                serializer->v->Complex(serializer, NULL, phref2, phrefn, 0);
                DestroyHRefFromList(&phref, phref2, CurrentHeap());
                node = ListNodeNext(itemd->_ListNodeOffset, node);
            } while (node != *(PTR*)a);
            if ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0)
                serializer->v->EndElement(serializer, Elem);
        }
        return S_OK;
    } else if ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY) ||
               Type == ARG_STRUCT)
    {
        PHREF pnewhref;
        /* Nested, thus there is no pointer, the subelement begins
         * immediately at the element.
         */
        /* NEW: BYREF tells whether this is a struct pointer or inline */
        if (itemd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF)
            a = *(PTR*) a;

        if (a == NULL) {
            if ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL) == 0) {
                DBG(("ItemHelper: non-optional field %s in NULL\n",
                     itemd->name));
                return E_INVALID_PARAMETER;
            }
            /* Skip this item */
            return S_OK;
        }

        pnewhref = MakeHref(&phref->next, CurrentHeap(),
                            NULL, 0, a,
                            NULL, 0, NULL, itemd, phref, 0, phrefn ? *phrefn : 0);
        phref->FieldSelector = FieldSelector;
        if (structd) {
            pnewhref->Xns = Xns;
            pnewhref->structd = structd;
            pnewhref->type = Type;
        } else if (itemd->_structXns) {
            pnewhref->Xns = (void*) itemd->_structXns;
        } else {
            pnewhref->Xns = phref->Xns;
        }
        if (Type != ARG_STRUCT) {
            /* nested arrays must be fixed size */
            assert(!itemd->SizeIsArg && itemd->ArraySize > 0);
            pnewhref->ArraySize = itemd->ArraySize;
        }
        /* Can only marshal ref into text or attr */
        structd = structd ? structd : itemd->_structd;
        if (Elem && !(structd && (structd->any.Flags & ENTRY_FLAGS_NAMELESS))
            && phrefn)
        {
            return serializer->v->ComplexRef(serializer, Elem,
                                             (*phrefn)++, encodingNS);
        } else {
            SCODE sc;
            if (Elem)
                serializer->v->StartElement(serializer, Elem, NULL, 0);
            sc = serializer->v->Complex(serializer, NULL, pnewhref, phrefn, 0);
            if (Elem)
                serializer->v->EndElement(serializer, Elem);
            DestroyHRefFromList(&phref, pnewhref, CurrentHeap());
            return sc;
        }
    } else {
        if (itemd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF) {
            if (!a) {
                /* Optional and NULL, skip */
                if (itemd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL)
                    return S_OK;
                else
                    return E_INVALID_PARAMETER;
            }
            a = *(PTR*) a;
            if (!a)
                return S_FALSE;
        }
        /* Output simple type unless fieldselector says not to (=0) */
        if (FieldSelector && !(FieldSelector[0] & 1))
            return S_OK;
        return serializer->v->SimpleType(serializer,
                                         itemd->descr_type, itemd->_Flags,
                                         a, Length,
                                         Elem, Attrib, 0,
                                         intstr, MAXDATELENGTH);
    }
}

#ifdef __ADSPBLACKFIN__
/* bf compiler chokes when optimizing... */
#pragma optimize_off
#endif
PRIVATE SCODE SoapSerializerComplex(PISOAPSERIALIZER serializer,
                                    CPXMLNAMESPACE encodingNS,
                                    PHREF phref, PUINT phrefn, UINT32 RetVal)
{
#define MAXATTRIBUTES 20
    SAXATTRIBUTE Attributes[MAXATTRIBUTES], _elem, *Elem = &_elem;
    _TCHAR _intstr[MAXDATELENGTH], *intstr = _intstr;
    _TCHAR buf[150];

    PUINT a;
    UINT j, size, Count = 0;
    PSAXATTRIBUTE Attr = &Attributes[0];
    SCODE sc = S_OK;

    memset(Elem, 0, sizeof(*Elem));

    /* Get pointer to data */
    a = phref->addr;
    intstr = _intstr;
    if (phref->ArraySize > 0) { /* This is an array */
        assert(phref->flags & ARGUMENT_DESCR_FLAGS_ARRAY);
        size = phref->ArraySize;
        if (phref->flags & ARGUMENT_DESCR_FLAGS_RETURN)
            a = (PUINT)RetVal;

        if (phref->ID) {
            Attr->Xns = encodingNS; Attr->Prefix = _T("");
            Attr->Token = TOKEN_soapenc_id; Attr->Name = _T("id");
            Attr->Value = phref->ID;
            Attr++; Count++;
        }

        if (phref->structd) {
            /* struct array. Need to define q1 namespace.
             * XXX URL should not be hard coded.
             */
            Attr->Xns = XMLNS_DECLARE;
            Attr->Prefix = _T("xmlns");
            Attr->Name = _T("q1");
            Attr->Value = _T("http://soapinterop.org/xsd");
            Attr++; Count++;

            /* Use + 50 here so as not to overwrite id above */
            _stprintf(buf + 50, _T("q1:%s[%d]"), phref->structd->name, size);
        } else {
            _stprintf(buf + 50, _T("xsd:%s[%d]"),
                    (phref->type == ARG_INT) ? _T("int")
                    : (phref->type == ARG_FLOAT) ? _T("float")
                    : (phref->type == ARG_STRING) ? _T("string")
                    : _T("BOGUS"),
                    size);
        }
        Attr->Xns = encodingNS; Attr->Prefix = _T("soapenc");
        Attr->Token = TOKEN_soapenc_arrayType; Attr->Name = _T("arrayType");
        Attr->Value = buf + 50;
        Attr++; Count++;
        Elem->Xns = encodingNS; Elem->Prefix = _T("soapenc");
        Elem->Token = TOKEN_soapenc_Array; Elem->Name = _T("Array");
        serializer->v->StartElement(serializer, Elem, Attributes, Count);
        Count = 0;

        for (j = 0; j < size; j++) {
            Elem->Xns = encodingNS; Elem->Prefix = _T("");
            Elem->Token = TOKEN_soapenc_Item; Elem->Name = _T("Item");
            if (phref->type == ARG_STRUCT) {
                PHREF pnewhref;
                assert(phref->type == ARG_STRUCT);
                pnewhref = MakeHref(&phref->next, CurrentHeap(),
                                    NULL, j, a, /* remember the arg value */
                                    NULL, 0, NULL, NULL, phref, 0, *phrefn);
                pnewhref->FieldSelector = phref->FieldSelector;
                serializer->v->ComplexRef(serializer, Elem,
                                          (*phrefn)++, encodingNS);
            } else {
                PTR aa = a + j;
                if (phref->type == ARG_STRING
                    || phref->type == ARG_URI)
                    aa = *(_TCHAR**) aa;

                serializer->v->SimpleType(serializer,
                                          phref->type, phref->flags,
                                          aa, UINT_MAX,
                                          Elem, NULL, 0,
                                          intstr, MAXDATELENGTH);
            }
        }
        Elem->Xns = encodingNS; Elem->Prefix = _T("soapenc");
        Elem->Token = TOKEN_soapenc_Array; Elem->Name = _T("Array");
        sc = serializer->v->EndElement(serializer, Elem);
    } else { /* struct */
        CPMETHOD_DESCR structd = phref->structd;
        _TCHAR *bigbuf = NULL, *bb = NULL;

        assert(phref->type == ARG_STRUCT
               || phref->type == ARG_LISTHEAD);
        assert(structd != NULL);

        if (phref->flags & ARGUMENT_DESCR_FLAGS_RETURN)
            a = (PUINT)RetVal;

        /* In case this is an array entry, we need to do some scaling */
        a = (PUINT) ((char*)a + phref->ArrayIndex * structd->_FrameSize);

        Count = 0;
        if (phref->ID) {
            Attr = &Attributes[0];
            Attr->Xns = encodingNS; Attr->Prefix = _T("");
            Attr->Token = TOKEN_soapenc_id; Attr->Name = _T("id");
            Attr->Value = phref->ID;
            Attr++; Count++;
        }

        if (!phref->Xns) {      /* XXXXXXXXXX || 1 for asp+ compat */
            static int qix = 1;
            qix++;
            _stprintf(buf + 50, _T("q%d:%s"), qix, structd->name);
            _stprintf(buf + 100, _T("q%d"), qix);
#if 0
            Attr->Xns = GetNamespace(_T("xsi"), 3); Attr->Prefix = _T("xsi");
            Attr->Token = TOKEN_xsi_type; Attr->Name = _T("type");
            Attr->Value = buf + 50;
            Attr++; Count++;
#endif

            /* Need to define q2 namespace.
             * XXX URL should not be hard coded.
             * XXX define xsi namespace as part of interface.
             */
            Attr->Xns = XMLNS_DECLARE;
            Attr->Prefix = _T("xmlns");
            Attr->Name = buf + 100;
            if (phref->Xns) {
                Attr->Value = phref->Xns->QName;
                if (phref->Xns->nstype != NSTYPE_STRUCT) {
                    /* Search for designated struct namespace */
                    CPXMLNAMESPACE Xns = phref->Xns;
                    CPINTERFACE_DESCR ifd = Xns->ifd;
                    Xns = Xns->ifd->iface_extras->namespaces;
                    for (; Xns && Xns->ifd == ifd; Xns = Xns->next)
                        if (Xns->nstype == NSTYPE_STRUCT) {
                            Attr->Value = Xns->QName;
                            //Attr->Name = Xns->name;
                        }
                }
                //_stprintf(buf + 100, _T("%sxsd"), phref->Xns->QName);
                //Attr->Value = buf + 100;
            } else {
                Attr->Value = _T("http://soapinterop.org/xsd");
            }
            Attr++; Count++;
        } else if (phref->Xns->QName) {
            /* Output namespace declaration if specified */
            Attr->Xns = XMLNS_DECLARE;
            Attr->Prefix = _T("xmlns");
            Attr->Name = phref->Xns->name;
            Attr->Value = phref->Xns->QName;
            Attr++; Count++;
            buf[100] = _EOS;
        }

        /* Output the struct fields @attribute */
        for (j = 0; j < structd->table_size; j++) {
            CPARGUMENT_DESCR itemd = &structd->ptable[j];
            if (itemd->descr_type == ARG_LISTNODE
                || itemd->descr_type == ARG_CONTINUATION)
                continue;
            if (itemd->_Flags & (ARGUMENT_DESCR_FLAGS_MODIFIER_OF|ARGUMENT_DESCR_FLAGS_NONE))
                continue;

            /* If one of the arguments is a lone QName, output XMLNS
             * in case it isn't yet defined.
             */
            if (itemd->descr_type == ARG_QNAME && a) do {
                /* Is it a specific NS? */
                UINT len;
                CPXMLNAMESPACE Xns;
                const _TCHAR *aa = *(PTR*) ((char*) a + itemd->FrameOffset);

                len  = _tcscspn(aa, _T(":"));
                if (aa[len] != _T(':'))
                    break;
                /* Same as parent or unknown? */
                Xns = GetNamespace(aa, len);
                if (!Xns || Xns == phref->Xns)
                    break;
                Attr->Xns = Xns;
                Attr->Prefix = _T("xmlns");
                Attr->Name = Xns->name;
                Attr->Value = Xns->QName;
                Attr++; Count++;
            } while(FALSE);

            if (!(itemd->_Flags & ARGUMENT_DESCR_FLAGS_ATTRIBUTE))
                continue;

            if (phref->FieldSelector
                && (phref->FieldSelector[j/32] & (1 << j%32)) == 0)
                continue;

            if (!bigbuf) {
                /* Allocate space for temporary attr value storage */
                bigbuf = malloc(MAXATTRIBUTES * MAXDATELENGTH *sizeof(_TCHAR));
                bb = bigbuf;
            }
            assert(bigbuf);
            if (!bigbuf)
                return E_NOT_ENOUGH_MEMORY;

            Attr->Prefix = buf + 100;
            Attr->Xns = phref->Xns;
            Attr->Name = itemd->name;
            Attr->Token = itemd->NameToken;
            sc = ItemHelper(serializer, phref, phrefn, itemd,
                            (char*)a + itemd->FrameOffset,
                            NULL, Attr, bb, encodingNS);
            if (sc == S_OK) {
                Attr++; Count++;
                bb += MAXDATELENGTH;
            }

            assert(Count < MAXATTRIBUTES);
        }

        if ((structd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0) {
            Elem->Xns = phref->Xns; //Elem->Prefix = buf + 100; XXXXXXXXX;
            Elem->Prefix = phref->Xns->name; /* XXXXXXXXX */
            Elem->Token = structd->NameToken; Elem->Name = structd->name,
            serializer->v->StartElement(serializer, Elem, Attributes, Count);
        }
        if (bigbuf) {
            /* Since it was output it is no longer needed */
            free(bigbuf);
            bigbuf = NULL;
        }

        /* Output the struct fields <element> */
        for (j = 0; j < structd->table_size; j++) {
            CPARGUMENT_DESCR itemd = &structd->ptable[j];
            if (itemd->_Flags
                & (ARGUMENT_DESCR_FLAGS_ATTRIBUTE|ARGUMENT_DESCR_FLAGS_HEADER
                   |ARGUMENT_DESCR_FLAGS_MODIFIER_OF|ARGUMENT_DESCR_FLAGS_NONE))
            continue;

            if (phref->FieldSelector
                && (phref->FieldSelector[j/32] & (1 << j%32)) == 0)
                continue;

            Elem->Xns = phref->Xns;
            Elem->Prefix = phref->Xns->name; /* XXXXXXXXX */
            Elem->Token = itemd->NameToken; Elem->Name = itemd->name;

            if (itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) {
                ItemHelper(serializer, phref, phrefn, itemd,
                           (char*)a + itemd->FrameOffset,
                           NULL, NULL, buf, encodingNS);
                if (structd->any.Flags & ENTRY_FLAGS_NAMELESS) {
                    /* For simple type lists output trailing (separating) space.
                     * XXX This actually outputs generously, not only in lists...
                     */
                    sc = serializer->v->Text(serializer, _T(" "), 1, NULL, 0);
                }
            } else {
                ItemHelper(serializer, phref, phrefn, itemd,
                           (char*)a + itemd->FrameOffset,
                           Elem, Attributes, buf, encodingNS);
            }
        }

        if ((structd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0) {
            Elem->Xns = phref->Xns; //Elem->Prefix = buf + 100; XXXXXXX;
            Elem->Prefix = phref->Xns->name; /* XXXXXXXXX */
            Elem->Token = structd->NameToken; Elem->Name = structd->name,
            sc = serializer->v->EndElement(serializer, Elem);
        }
    }
    return sc;
}

PRIVATE SCODE MethodHelper(PISOAPSERIALIZER serializer,
                           PCONTINFO cont, BOOL IsResponse,
                           PHREF *pphref, PUINT phrefn,
                           CPXMLNAMESPACE Xns, CPARGUMENT_DESCR argd,
                           PBYTE Frame, PUINT a,
                           SAXATTRIBUTE *Elem, SAXATTRIBUTE *Attrib)
{
    _TCHAR _intstr[MAXDATELENGTH], *intstr = _intstr;
    PHREF phref = NULL;
    DESCR_TYPE Type = argd->descr_type;
    UINT32 *FieldSelector = NULL;

    CPMETHOD_DESCR structd = NULL;

    /* Fetch field selector if specified */
    if (argd->SelectorIsArg) {
        FieldSelector = *(UINT32**)(Frame
                                    + (argd+argd->SelectorIsArg)->FrameOffset);
    }

    /* Fetch type descriptor if specified */
    if (argd->TypeIsArg) {
        const _TCHAR *typestring;
        UINT len;

        typestring = *(const _TCHAR **)(Frame + (argd+argd->TypeIsArg)->FrameOffset);
        if (typestring) {
            len = _tcscspn(typestring, _T(":"));
            if (typestring[len] != _T(':'))
                return E_INVALID_PARAMETER; /* XXX could try SoapAction style instead */
            Xns = GetNamespace(typestring, len);
            if (!Xns)
                return E_NAME_NOT_FOUND;
            structd = FindMethod(Xns, typestring + len + 1, FALSE);
            if (!structd)
                return E_NAME_NOT_FOUND;
            Type = ARG_STRUCT;
        } else {
            Type = ARG_STRING;
        }
    }

    if (argd->descr_type == ARG_LISTHEAD) {
        PTR node, head;
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT)
            node = (PTR) *a;
        else
            node = a;
        head = node;
        if (node) {
            /* In case wrapper element is desired: output it now */
            if ((argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0) {
                Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
                serializer->v->StartElement(serializer, Elem, NULL, 0);
            }

            do {
                /* serialize node; */
                phref = MakeHref(pphref, CurrentHeap(),
                                 NULL, 0, node,
                                 NULL, 0, NULL, argd, NULL, 0, 0);
                phref->Xns = Xns;
                phref->FieldSelector = FieldSelector;
                if (structd)
                    phref->structd = structd;
                serializer->v->Complex(serializer, cont->encodingNS, phref,
                                       phrefn, cont->RetVal);
                DestroyHRefFromList(pphref, phref, CurrentHeap());
                node = ListNodeNext(argd->_ListNodeOffset, node);
            } while (node != head);
            if ((argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0)
                serializer->v->EndElement(serializer, Elem);
        }
        return S_OK;
    } else if (argd->descr_type == ARG_LISTNODE
               || argd->descr_type == ARG_CONTINUATION)
    {
        /* skip */
        return S_OK;
    } else if (argd->descr_type == ARG_BLOB) {
        UINT total;

        if (!IsResponse) {
            if (argd->LengthIsArg) {
                total = *(PUINT)*(ADDRESS*)(Frame + (argd+argd->LengthIsArg)->FrameOffset);
            } else if (argd->SizeIsArg) {
                total = *(PUINT)(Frame + (argd+argd->SizeIsArg)->FrameOffset);
            } else {
                assert(argd->ArraySize);
                total = argd->ArraySize;
            }
        } else if (argd->LengthIsArg) {
            /* Pick up length from save area right next to item */
            //total = cont->pSaveArea[argd->SaveIndex + 1];
            total = *(PUINT)*(ADDRESS*)(Frame + (argd+argd->LengthIsArg)->FrameOffset);
        } else {
            /* Fixed size blob. Use value from typeinfo. */
            total = argd->ArraySize;
        }

        //Elem->Token = argd->Token; Elem->Name = argd->name;
        Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
        return serializer->v->Blob(serializer, (PTR) a, total, Elem);
    } else if ((argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY)
               || Type == ARG_STRUCT) {
        /* Arrays and structs are marshaled as HREFs */
        /* Fish out array size (0 for flat struct) */
        UINT Size = 0;

        /* Check for NULL pointer */
        if (a == NULL) {
            assert(!IsResponse); /* XXX verify for response case */
            if (argd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL)
                return S_OK;
            else
                return E_INVALID_PARAMETER;
        }
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY) {
            if (IsResponse && argd->LengthIsArg)
                //Size = cont->pSaveArea[argd->SaveIndex + 1];
                Size = *(PUINT)*(ADDRESS*)(Frame
                                     + (argd+argd->LengthIsArg)->FrameOffset);
            else if (argd->LengthIsArg)
                Size = *(PUINT)*(ADDRESS*)(Frame
                                     + (argd+argd->LengthIsArg)->FrameOffset);
            else if (!IsResponse && argd->SizeIsArg)
                Size = *(PUINT)(Frame+(argd+argd->SizeIsArg)->FrameOffset);
            else
                Size = argd->ArraySize;
        }
        phref = MakeHref(pphref, CurrentHeap(),
                         NULL, 0, a, /* remember the arg value */
                         NULL, 0, NULL, argd, NULL, 0,
                         phrefn ? *phrefn : 0);
        phref->ArraySize = Size; Size = 0;
        phref->FieldSelector = FieldSelector;
        if (structd) {
            phref->Xns = Xns;
            phref->structd = structd;
            phref->type = Type;
        } else if (argd->_structXns) {
            phref->Xns = (void*) argd->_structXns;
        } else {
            phref->Xns = cont->methodNS;
        }
        structd = structd ? structd : argd->_structd;
        if (Elem && !(structd && (structd->any.Flags & ENTRY_FLAGS_NAMELESS))){
            /* Array or named struct */
            //Elem->Token = argd->Token; Elem->Name = argd->name;
            Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
            return serializer->v->ComplexRef(serializer, Elem,
                                             (*phrefn)++, cont->encodingNS);
        } else {
            /* If nameless struct, it has to be inline */
            if (Elem) {
#if 0
                if (structd) {
                    Elem->Token = TOKEN_invalid;
                    Elem->Name = structd->name;
                }
#else
                Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
#endif
                serializer->v->StartElement(serializer, Elem, NULL, 0);
            }
            serializer->v->Complex(serializer, cont->encodingNS, phref,
                                   phrefn, cont->RetVal);
            DestroyHRefFromList(pphref, phref, CurrentHeap());
            if (Elem)
                serializer->v->EndElement(serializer, Elem);
        }
        return S_OK;
    } else if (argd->descr_type == ARG_BOOL
               && (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT))
    {
        /* Special use for BOOL serialize-as="text()"
         * print empty elem (TRUE) or nothing (FALSE)
         */
        BOOL *addr = a;
        if (!*addr)
            return S_OK;
        Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
Elem->Prefix = Xns->name;
        serializer->v->StartElement(serializer, Elem, Attrib, 0);
        return serializer->v->EndElement(serializer, Elem);
    } else if ((argd->descr_type == ARG_STRING
                || argd->descr_type == ARG_QNAME)
               && (argd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL)
               && a == NULL)
    {
        /* Skip optional string arguments that are NULL */
        return S_OK;
    } else {
        if (Elem) {
            //Elem->Token = argd->Token; Elem->Name = argd->name;
            Elem->Token = TOKEN_invalid; Elem->Name = argd->name;
Elem->Prefix = Xns->name;
        }
        /* Output simple type unless fieldselector says not to (=0) */
        if (FieldSelector && !(FieldSelector[0] & 1))
            return S_OK;
        return serializer->v->SimpleType(serializer,
                                         Type, argd->_Flags, a, UINT_MAX,
                                         Elem, Attrib, 0,
                                         intstr, MAXDATELENGTH);
    }
}

PRIVATE PUINT GetArgAddr(PCONTINFO cont, BOOL IsResponse, CPARGUMENT_DESCR argd)
{
    PUINT a;
    if (argd->_Flags & ARGUMENT_DESCR_FLAGS_RETURN) {
        if (!IsResponse & (argd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF))
            return NULL;

        /* Fetch data from return value.
         * Pointer values need to be fetched.
         */
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF)
            a = (PUINT)cont->RetVal;
        else
            a = (PUINT)&cont->RetVal;
#if 0
    } else if (IsResponse) {
        //assert(argd->SaveIndex);
        //assert(cont->pSaveArea);
        a = (PUINT) cont->pSaveArea[argd->FrameOffset];
#endif
    } else if (IsResponse) {
        ADDRESS *ap;
        ap = (ADDRESS*)cont->Frame;
        ap = (ADDRESS*) ((char *)ap + argd->FrameOffset);
        a = (PUINT)*ap;
    } else {
        ADDRESS *ap;
        ap = (ADDRESS*)cont->Frame;
        ap = (ADDRESS*) ((char *)ap + argd->FrameOffset);
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF)
            a = (PUINT)*ap;
        else
            a = (PUINT)ap;
    }
    return a;
}

void DestroyHReflist(PHREF *pphref, PIHEAP Heap); /* XXX, declare in header */


/* Output method request/reply and collect any open refs into pphref */
PRIVATE SCODE SoapSerializerMethod(PISOAPSERIALIZER serializer,
                                   PICONTINUATION iCont, BOOL IsResponse,
                                   PHREF *pphref, PUINT phrefn)
{
    METHOD_DECLARE_NOTHIS(SoapSerializer, Method);
    PHREF default_phref = NULL;
    PCONTINFO cont = pCONT(iCont);
    CPXMLNAMESPACE methodNS = cont->methodNS;
    _TCHAR _intstr2[MAXDATELENGTH], *intstr2 = _intstr2;
    UINT i, Count;
    CPMETHOD_DESCR metd = cont->metd;
    CPARGUMENT_DESCR argd;
#define MAXATTRIBUTES 20
    SAXATTRIBUTE Attributes[MAXATTRIBUTES], *Attrib = &Attributes[0];
    SAXATTRIBUTE _elem, *Elem = &_elem;
    PUINT a;

    sc = S_OK;
    if (!pphref)
        pphref = &default_phref;
    if ((metd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0) {
        memset(Elem, 0, sizeof(*Elem));
        Elem->Token = TOKEN_invalid;

        Count = 0;
        if (methodNS->QName
#ifndef _MINIMIZE
            && (_tcscmp(methodNS->QName, _T("http://schemas.xmlsoap.org/ws/2005/06/management"))
                || _tcscmp(methodNS->name, _T("wsman")))
#endif
                && (_tcscmp(methodNS->QName, _T("http://schemas.xmlsoap.org/ws/2004/08/addressing"))
                || _tcscmp(methodNS->name, _T("wsa")))
            ) {
            /* Output namespace declaration if specified */
            Attrib->Xns = XMLNS_DECLARE;
            Attrib->Prefix = _T("xmlns");
            Attrib->Name = methodNS->name;
            Attrib->Value = methodNS->QName;
            Count++; Attrib++;
        }

        /* Deal with serialize="@" */
        for (i = 0; i < metd->table_size; i++) {
            argd = &metd->ptable[i];
            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_ATTRIBUTE))
                continue;
            if (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)
                continue;

            if (IsResponse) {
                if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                    continue;
            } else {
                if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))
                    continue;
            }
            assert(!(argd->_Flags & ARGUMENT_DESCR_FLAGS_RETURN));
            assert(argd->descr_type != ARG_BLOB);
            assert(!(argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY));
            assert(argd->descr_type != ARG_STRUCT);

            a = GetArgAddr(cont, IsResponse, argd);
            Attrib->Xns = methodNS;
            Attrib->Prefix = methodNS->name; 
            Attrib->Token = TOKEN_invalid;
            Attrib->Name = argd->name;
            assert(intstr2);
            serializer->v->SimpleType(serializer, argd->_Flags,
                                      argd->descr_type, a, UINT_MAX,
                                      NULL, Attrib, 0, intstr2, MAXDATELENGTH);
            /* XXX Only can handle *one* integer as intstr gets recycled! */
            intstr2 = NULL;
            Count++; Attrib++;
        }

        Elem->Prefix = methodNS->name; Elem->Xns = methodNS;
        if (IsResponse) {
            Elem->Token = metd->ResponseNameToken;
        } else {
            Elem->Token = metd->NameToken;
        }
        Elem->Name = NULL;
        BResolveName(Elem);
        serializer->v->StartElement(serializer, Elem, Attributes, Count);
    }

    for (i = 0; i < metd->table_size; i++) {
        argd = &metd->ptable[i];

        /* Already handled serialize="@" args */
        if (argd->_Flags & (ARGUMENT_DESCR_FLAGS_ATTRIBUTE
                            |ARGUMENT_DESCR_FLAGS_HEADER))
            continue;

        if (argd->_Flags & (ARGUMENT_DESCR_FLAGS_MODIFIER_OF|ARGUMENT_DESCR_FLAGS_NONE))
            continue;

        if (IsResponse) {
            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                continue;
        } else {
            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))
                continue;
        }
        
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_RETURN) {
            /* Disallow returning types that need storage at client */
            if (!IsResponse & (argd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF))
                CHECKSETSC(FALSE, E_INVALID_PARAMETER, "bad return type");

            /* If the method returns SCODE check for special return code */
            if (IsResponse && cont->RetVal == S_NO_RESPONSE) {
                sc = S_NO_RESPONSE;
                goto Out;
            }
        }

        a = GetArgAddr(cont, IsResponse, argd);
        if ((argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)
            && argd->descr_type != ARG_BOOL)
        {
            assert (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY));
            assert(!(argd->_Flags & ARGUMENT_DESCR_FLAGS_RETURN));

            Elem->Prefix = _T("");
            if (metd->any.Flags & ENTRY_FLAGS_NAMELESS)
                phrefn = NULL;      /* no @id */
            MethodHelper(serializer, cont, IsResponse, pphref, phrefn,
                         methodNS, argd, cont->Frame, a, NULL, NULL);
            continue;
        }

        a = GetArgAddr(cont, IsResponse, argd);
        /* Normally args are unqualified. Except nameless methods (no scope) */
        //if (metd->any.Flags & ENTRY_FLAGS_NAMELESS)
            Elem->Prefix = methodNS->name;
        //else
        //    Elem->Prefix = _T("");

        sc = MethodHelper(serializer, cont, IsResponse, pphref, phrefn,
                          methodNS, argd, cont->Frame, a, Elem, Attrib);
    }

    if ((metd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0) {
        Elem->Prefix = methodNS->name; Elem->Xns = methodNS;
        if (IsResponse) {
            Elem->Token = metd->ResponseNameToken;
        } else {
            Elem->Token = metd->NameToken;
        }
        Elem->Name = NULL;
        BResolveName(Elem);
        sc = serializer->v->EndElement(serializer, Elem);
    }

METHOD_CLEANUP;
    if (pphref == &default_phref)
        DestroyHReflist(pphref, CurrentHeap());
    return sc;
}

PRIVATE SCODE MCT SoapSerializerStartMessage(
        PISOAPSERIALIZER serializer,
        PICONTINUATION iCont, BOOL isResponse,
        const void *metd_)
{
    METHOD_DECLARE_NOTHIS(SoapSerializer, StartMessage);
    CPMETHOD_DESCR metd = metd_;
    PCONTINFO cont = pCONT(iCont);
    PWSCONTEXT pCtx = (PWSCONTEXT) cont->pContext;
    CPXMLNAMESPACE envelopeNS = cont->envelopeNS;
    CPXMLNAMESPACE encodingNS = cont->encodingNS;
    PXMLNAMESPACE namespaces = cont->namespaces;
    BOOL ContSerIsNull = FALSE;
    _TCHAR *tmp;
    const _TCHAR *address;
    UINT i;
    BOOL wsaToArg = FALSE;
    PISOAPFACTORY soapfactory = NULL;

    sc = serializer->v->StartDocument(serializer);
    CHECKSC("StartDocument");

    /* <Envelope> */
    sc = serializer->v->StartEnvelope(serializer, iCont,
                                      envelopeNS, namespaces, encodingNS);

    CHECKSC("StartEnvelope");

    /* <Header/> */
    if (pCtx) {
        static PIWSADDRESSING wsapxy = NULL;
#ifndef _MINIMIZE
        static PIWSEVENTING wsepxy = NULL;
        static PIWSMANAGEMENT wsmanpxy = NULL;
        static PIWSRESERVATION respxy = NULL;
#endif
        static PIDSSP dssppxy = NULL;
        if (!wsapxy
#ifndef _MINIMIZE
            || !wsepxy || !wsmanpxy || !respxy
#endif
            || !dssppxy) {
            sc = BindToObject(CurrentNameSpace(),
                              _T("COB/soap.cob"), NAME_SPACE_READ,
                              &IID_ISoapFactory, (void **) &soapfactory); 
            CHECKSC("bind to soap factory");

            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              0, NULL, _T("wsa"), 0,
                                              (PIUNKNOWN*) &wsapxy);
            CHECKSC("create wsa proxy");

#ifndef _MINIMIZE
            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              0, NULL, _T("wse"), 0,
                                              (PIUNKNOWN*) &wsepxy);
            CHECKSC("create wse proxy");

            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              0, NULL, _T("wsman"),0,
                                              (PIUNKNOWN*) &wsmanpxy);
            CHECKSC("create wsman proxy");

            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              0, NULL, _T("res"), 0,
                                              (PIUNKNOWN*) &respxy);
            CHECKSC("create res proxy");
#endif
            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              0, NULL, _T("dssp"), 0,
                                              (PIUNKNOWN*) &dssppxy);
            CHECKSC("create dssp proxy");
        }

        if (!cont->Serializer) {
            /* Likely from SoapFault path, put it in for a bit */
            cont->Serializer = serializer;
            ContSerIsNull = TRUE;
        }
        serializer->v->StartHeader(serializer, envelopeNS, encodingNS);
        assert(cont);

        /* See if there are @wsa:To args within the method.
         * If so, they get stuffed into the URI.
         * These are indicated by both TEXT and ATTRIBUTE flags being set.
         */
        tmp = NULL;
        address = NULL;
        if (isResponse) {
            if (pCtx->ReplyTo)
                address = pCtx->ReplyTo->Address;
        } else {
            if (pCtx->To)
                address = pCtx->To->Address;
        }

        if (metd && address) {
            _TCHAR *tmp2;
            const _TCHAR *name;
            tmp = NULL;
            for (i = 0; i < metd->table_size; i++) {
                CPARGUMENT_DESCR argd = &metd->ptable[i];
                if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_ATTRIBUTE)
                    || !(argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT))
                    continue;

                if (isResponse) {
                    if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                        continue;
                } else {
                    if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))
                        continue;
                }
                assert(!(argd->_Flags & ARGUMENT_DESCR_FLAGS_RETURN));
                assert(argd->descr_type == ARG_STRING);
                assert(!(argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY));

                name = (_TCHAR*) GetArgAddr(cont, isResponse, argd);

                if (wsaToArg) {
                    /* In case multiple args to stuff here. Inefficient code */
                    tmp2 = malloc((_tcslen(tmp) + _tcslen(argd->name) + 3)
                                  * sizeof(_TCHAR));
                    _stprintf(tmp2, _T("%s?%s=%s"), tmp, argd->name, name);
                    free(tmp);
                    tmp = tmp2;
                } else {
                    tmp = malloc((_tcslen(address) + _tcslen(name)
                                  + _tcslen(argd->name) + 3) * sizeof(_TCHAR));
                    _stprintf(tmp, _T("%s?%s=%s"), address, argd->name, name);
                }
                address = tmp;
                wsaToArg = TRUE;
            }
            /* Save for http POST generation */
            if (cont->WsaTo)
                free(cont->WsaTo);
            cont->WsaTo = tmp;  /* may be NULL */
            tmp = NULL;
                
        }

        if (address)
            sc = wsapxy->v->To(wsapxy, iCONT(cont), address);
        if (tmp)
            free(tmp);
        if (address)
            CHECKSC("To");

        /* If necessary, output wsa:Action. For now, only when it is not
         * redundant. Some specs might want it always/more often.
         * wsman interop reuqires it always...
         */
        if (metd /* && (metd->any.Flags & ENTRY_FLAGS_NAMELESS) */) {
            /* <wsa:Action>QName/methodName</> */
            const _TCHAR *name;
            const _TCHAR *QName = cont->methodNS->QName;
            _TCHAR separator = _T('/');
            UINT len = cont->methodNS->uri_length;
            /* Special hack for DSSP: use colon instead of slash as separator */
            if (isResponse != 2 && !_tcscmp(cont->methodNS->name, _T("dssp")))
                separator = _T(':');

            if (isResponse == 1) { /* Except fault */
                /* Resolve token to string. XXX use helper rather than
                 * digging (should not include xns.h) XXX
                 */
                PSTR_ENTRY str;
                str = (PSTR_ENTRY) GetBottom((CPANY_ENTRY)
                                             cont->methodNS->entries,
                                             metd->ResponseNameToken,
                                             FALSE, ENTRY_TYPE_STRING);
                assert(str);
                name = str->Str;
            } else if (isResponse == 2) { /* XXX special hack for interop */
                QName = _T("http://schemas.xmlsoap.org/ws/2004/08/addressing");
                len = _tcslen(QName);
                name = _T("fault");
                
            } else {
                name = metd->name;
                //len = _tcslen(name);
            }
            tmp = malloc((_tcslen(QName) + _tcslen(name) + 2)
                         * sizeof(_TCHAR));
            CHECKSETSC(tmp, E_NOT_ENOUGH_MEMORY, "malloc Action value");
            /* cap length with uri_length, settable through @uri-skip-suffix */
            _stprintf(tmp, _T("%.*s%c%s"), len, QName, separator, name);
            assert(cont->SoapAction == NULL);

            cont->SoapAction = tmp;

            sc = wsapxy->v->Action(wsapxy, iCONT(cont), tmp);
            tmp = NULL;         /* freed by cont->Release */
            CHECKSC("Action");
        }

        if (pCtx->RelatesTo) {
            sc = wsapxy->v->RelatesTo(wsapxy, iCONT(cont), pCtx->RelatesTo);
            CHECKSC("RelatesTo");
        }

        if (!isResponse && pCtx->ReplyTo) {
            sc = wsapxy->v->ReplyTo(wsapxy,iCONT(cont),pCtx->ReplyTo->Address);
            CHECKSC("ReplyTo");
        }

        if (pCtx->MessageId) {
            sc = wsapxy->v->MessageID(wsapxy, iCONT(cont), pCtx->MessageId);
            CHECKSC("MessageId");
        }

#ifndef _MINIMIZE
        if (pCtx->ResourceURI) {
            sc = wsmanpxy->v->ResourceURI(wsmanpxy, iCONT(cont),
                                          pCtx->ResourceURI);
            CHECKSC("wsman:ResourceURI");
        }
        if (pCtx->MaxEnvelopeSizeValue) {
            sc = wsmanpxy->v->MaxEnvelopeSize(wsmanpxy, iCONT(cont),
                                              NULL,
                                              pCtx->MaxEnvelopeSizeValue);
            CHECKSC("wsman:MaxEnvelopeSize");
        }
        if (!Int64IsZero(pCtx->OperationTimeout)) {
            sc = wsmanpxy->v->OperationTimeout(wsmanpxy, iCONT(cont),
                                               pCtx->OperationTimeout);
            CHECKSC("wsman:OperationTimeout");
        }
        if (pCtx->SelectorSet) {
            sc = wsmanpxy->v->SelectorSet(wsmanpxy, iCONT(cont), (PSelector) pCtx->SelectorSet);
            CHECKSC("wsman:SelectorSet");
        }

        if (pCtx->Reservation) {
            sc = respxy->v->task(respxy, iCONT(cont), pCtx->TaskName, pCtx->Reservation, pCtx->FeedbackURL);
            CHECKSC("task");
        }
        if (pCtx->taskName) {
            sc = respxy->v->taskName(respxy, iCONT(cont), pCtx->taskName);
            CHECKSC("taskName");
            sc = respxy->v->sendTime(respxy, iCONT(cont), CurrentTime());
            CHECKSC("sendTime");
        }

#endif
        sc = dssppxy->v->Timestamp(dssppxy, iCONT(cont), CurrentTime());
        CHECKSC("Timestamp");
        if (pCtx->ForwardedAction) {
            sc = dssppxy->v->NotificationHeader(dssppxy, iCONT(cont),
                pCtx->ForwardedAction);
            CHECKSC("Timestamp");
        }

        sc = serializer->v->EndHeader(serializer, envelopeNS);
        CHECKSC("EndHeader");
    } else {
        sc = serializer->v->StartHeader(serializer, envelopeNS, encodingNS);
        CHECKSC("StartHeader");
        sc = serializer->v->EndHeader(serializer, envelopeNS);
        CHECKSC("EndHeader");
    }
    /* <Body> */
    sc = serializer->v->StartBody(serializer, envelopeNS, encodingNS);
    CHECKSC("StartBody");
METHOD_CLEANUP;
    if (ContSerIsNull)          /* put it back the way it was */
        cont->Serializer = NULL;
    RELEASE(soapfactory);
    return sc;
}

PRIVATE SCODE MCT SoapSerializerEndMessage(
        PISOAPSERIALIZER serializer,
        CPXMLNAMESPACE envelopeNS)
{
    /* </Body> */
    serializer->v->EndBody(serializer, envelopeNS);
    /* </Envelope>*/
    serializer->v->EndEnvelope(serializer, envelopeNS);
    return serializer->v->EndDocument(serializer);
}

PRIVATE SCODE MCT SoapSerializerMethodMessage(
        PISOAPSERIALIZER serializer,
        PICONTINUATION iCont,
        BOOL isResponse)
{
    PHREF phref = NULL, tmp;
    UINT hrefn = 1;
    SCODE sc;
    PCONTINFO cont = pCONT(iCont);

    if (!cont->MsgStarted) {
        sc = serializer->v->StartMessage(serializer, iCONT(cont), isResponse,
                                         cont->metd);
        if (FAILED(sc))
            return sc;
    }

 sc = serializer->v->Method(serializer,
                               iCONT(cont), isResponse,
                               &phref, &hrefn);
    if (FAILED(sc) || sc == S_NO_RESPONSE)
        return sc;
    /* Construct the href's contents */
    while (phref) {
        sc = serializer->v->Complex(serializer, cont->encodingNS,
                                    phref, &hrefn, cont->RetVal);
        tmp = phref;
        phref = phref->next;
        DestroyHRef(tmp, CurrentHeap());
    }
    return serializer->v->EndMessage(serializer, cont->envelopeNS);
}

/* XXX Merge the fault into regular method serialization */
#include "web/proxy.h"              /* XXX */

PRIVATE SCODE MCT SoapSerializerFaultMessage(
        PISOAPSERIALIZER serializer,
        PICONTINUATION iCont, 
        SCODE err,
        PCode FaultCode,
        const _TCHAR *faultstring)
{
    METHOD_DECLARE_NOTHIS(SoapSerializer, FaultMessage);
    PCONTINFO cont = pCONT(iCont);
    CPXMLNAMESPACE envelopeNS = cont->envelopeNS, encodingNS = cont->encodingNS;
    SAXATTRIBUTE _elem, *Elem = &_elem;
    _TCHAR msg[50];

    memset(Elem, 0, sizeof(*Elem));
    _stprintf(msg, _T("Message processing failed, sc=0x%x"), err);

    if (!_tcscmp(envelopeNS->QName,
                 _T("http://www.w3.org/2003/05/soap-envelope")))
    {
        CPMETHOD_DESCR metd = NULL;
        CPXMLNAMESPACE Xns;
        BOOL SerNull = FALSE;
        struct Code subcode = {0,};
        struct Code code = {0, };
        struct Text text = {0, };

        static PISOAP12 soap12pxy = NULL;

        switch (err) {
            /* XXX incomplete, unnecessary and duplicate strings, prefixes */
#ifndef _MINIMIZE
        case E_WSMAN_INVALID_SELECTORS:
            code.Value = _T("soap:Sender");
            subcode.Value = _T("wsman:InvalidSelectors");
            break;
        case E_WSEN_INVALID_ENUMERATION_CONTEXT:
            code.Value = _T("soap:Receiver");
            subcode.Value = _T("wsen:InvalidEnumerationContext");
            break;
#endif
        case E_WSA_DESTINATION_UNREACHABLE:
        default:
            code.Value = _T("soap:Receiver");
            subcode.Value = _T("wsa:DestinationUnreachable");
            break;
#ifndef _MINIMIZE
        case E_WSMAN_ENCODING_LIMIT:
            code.Value = _T("soap:Sender");
            subcode.Value = _T("wsman:EncodingLimit");
            break;
        case E_WSMAN_TIMED_OUT:
            code.Value = _T("soap:Receiver");
            subcode.Value = _T("wsman:TimedOut");
            break;
#endif
        }

        code.Subcode = &subcode;
        text.XXlang = _T("en");
        text.Value = faultstring ? (_TCHAR*) faultstring : msg;

        ListNodeInit(offsetof(struct Text, ListNode), &text);

        if (!soap12pxy) {
            PISOAPFACTORY soapfactory;
            sc = BindToObject(CurrentNameSpace(),
                              _T("COB/soap.cob"), NAME_SPACE_READ,
                              &IID_ISoapFactory, (void **) &soapfactory); 
            CHECKSC("bind to soap factory");

            sc = soapfactory->v->CreateProxy3(soapfactory,
                                              NULL, NULL, _T("soap"), 0,
                                              (PIUNKNOWN*) &soap12pxy);
            RELEASE(soapfactory);
            CHECKSC("create soap12 fault handler proxy");
        }

        Xns = ((PPROXY) soap12pxy)->NameSpace;
        metd = FindMethod(Xns, _T("Fault"), TRUE);
        cont->methodNS = Xns;
        cont->metd = metd;
        sc = serializer->v->StartMessage(serializer, iCont, 2, metd);
        if (FAILED(sc))
            return sc;

        if (cont->Serializer == NULL) {
            /* XXX What a mess */
            cont->Serializer = serializer;
            SerNull = TRUE;
        }
        sc = soap12pxy->v->Fault(soap12pxy, iCont,
                                 FaultCode ? FaultCode : &code,
                                 &text,
                                 NULL, NULL, NULL);
        if (SerNull)
            cont->Serializer = NULL;
        CHECKSC("generate fault message");
    } else {
        sc = serializer->v->StartMessage(serializer, iCont, TRUE, NULL);
        if (FAILED(sc))
            return sc;

        /* <fault> */
        serializer->v->StartFault(serializer, envelopeNS, encodingNS,
                                  _T("Client"), NULL, NULL);
        Elem->Xns = envelopeNS; Elem->Prefix = _T("");
        /* Elem->Xns = envelopeNS; chokes SOAP Toolkit 3.0 */
        Elem->Token = TOKEN_soap11_faultstring; Elem->Name = _T("faultstring");
        serializer->v->StartElement(serializer, Elem, NULL, 0);
        serializer->v->Text(serializer,
                            faultstring ? faultstring : msg,
                            faultstring ? _tcslen(faultstring) : _tcslen(msg),
                            NULL, 0);
        serializer->v->EndElement(serializer, Elem);
        serializer->v->EndFault(serializer, envelopeNS);
    }
    sc = serializer->v->EndMessage(serializer, envelopeNS);
METHOD_CLEANUP;
    return sc;
}

#include <s_CSoapSerializer_v.c>
#include <s_CSoapSerializer2_v.c>

#define use_bax 0

#if use_bax
SCODE MCT BaxCreatePrinter(PISAXPARSER Parser, PIFILE OutputFile,
                           UINT IndentLevel, PISAXHANDLER *pPrinter);
#endif

SCODE MCT CreateSerializer (
    PIFILE OutputFile,
    PARSE_URL_FLAGS Flags,
    /*out*/ PISOAPSERIALIZER* serializer)
{
    PISAXHANDLER saxprinter;
    SCODE sc;
    UINT i;
    PISAXPARSER saxparser;
    ADDRESS *a;

    sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"), NAME_SPACE_READ,
                      &IID_ISaxParser, (void **) &saxparser);
    if (FAILED(sc))
        return sc;

#if use_bax
    if (Flags & PARSE_URL_FLAGS_XTE) {
        DBG(("Creating XTE printer\n"));
        a = (ADDRESS *) &CSoapSerializer2Vtbl;
        sc = BaxCreatePrinter(NULL, OutputFile, 0, &saxprinter);
    } else
#endif
    {
        a = (ADDRESS *) &CSoapSerializerVtbl;
        sc = saxparser->v->CreatePrinter(saxparser, OutputFile, 0, &saxprinter);
    }
    if (FAILED(sc))
        goto Out;

    /* Subclass original v-table for our purposes */
    for (i = 0; i < sizeof(CSoapSerializerVtbl)/sizeof(ADDRESS); i++)
        if (a[i] == 0)
            a[i] = ((ADDRESS*)saxprinter->v)[i];
    /* We highjack the saxprinter and turn it into a soapserializer.
     * ok because there is no private state.
     */
    saxprinter->v = (void*) a;

    *serializer = (PISOAPSERIALIZER) saxprinter;

 Out:
    saxparser->v->Release(saxparser);
    return sc;
}
