/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>

#include <netif.h>
#include "convert.h"
#include "deserializer.h"

/* SOAP box */
struct SOAPBOX {
    struct IHttpProcessVtbl const *     v;
    UINT                                Refs;
    //    struct TypeInfo *                   pTypeInfo;
};
typedef struct SOAPBOX      SOAPBOX,        *PSOAPBOX;

CPXMLNAMESPACE GetNamespace(const _TCHAR *prefix, UINT len /* prefix length */)
{
    CPXMLNAMESPACE ns;
    struct TypeInfo *ti;

    SCODE sc = GetTypeInfo(&ti);
    assert(SUCCEEDED(sc));
    if (FAILED(sc))
        return NULL;

    if (len == 0)
        len = _tcslen(prefix);

#if 1
    ns = FindXnsByPrefix(pTheNamespaceNamespace, prefix, len);
    return ns;
#else
    /* First resolve local name to QName */
    for (ns = ti->namespaces; ns; ns = ns->next)
        if (!_tcsncmp(ns->name, prefix, len) && ns->name[len] == _EOS)
            return ns;
    return NULL;
#endif
}

#if 0
#define use_bax 0

#if use_bax
SCODE XBaxParse(PITOKENIZER Tokenizer, PISAXHANDLER Handler);
#else
#define XBaxParse(Tokenizer, Handler) E_NOT_IMPLEMENTED
#endif

PRIVATE SCODE SOAPHTTPProcessPost(PIHTTPPROCESS iThis, 
                                  UINT ContentLength,
                                  PARSE_URL_FLAGS Flags,
                                  PITOKENIZER Tokenizer,
                                  const _TCHAR *LocalUrl,
                                  PIBUFFER *pContent,
                                  PIUNKNOWN *pDeserializer,
                                  UINT *ParsingState,
                                  PIHTTPCONNECTOR Request)
{
    METHOD_DECLARE_THIS("SOAPHTTPProcessPost", PSOAPBOX);
    PCONTINFO cont = NULL;
    PDESERIALIZER Handler = NULL;
    PISAXPARSER SAXParser = NULL;
    PISOAPSERIALIZER Serializer = NULL;
    PISOAPFACTORY SoapFactory = NULL;
    PIHEAP Heap = _GetHeap(_T("Handler"), 4000);
    
    TRACE("SOAPHTTPProcessPost");
    assert(Tokenizer != NULL);

#if 1
    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"),
                      NAME_SPACE_READ, &IID_ISoapFactory,
                      (void**) &SoapFactory);
    CHECKSC2("bind to soap.cob");

    sc = SoapFactory->v->ProcessRequest(SoapFactory,
                                        ContentLength, Flags, Tokenizer,
                                        LocalUrl, pContent,
                                        (PISAXHANDLER *) pDeserializer);
    CHECKSC2("ProcessRequest");
    Handler = (PDESERIALIZER) *pDeserializer;
#else
    if (!*pDeserializer) {
        /* Fetch the Typeinfo from the input interface definition file */
        sc = GetTypeInfo(&This->pTypeInfo);
        CHECKSC("GetTypeInfo");

        /*** Failures from here on produce SoapFaults (Out2) ***/

        sc = CreateDeserializer((PISAXHANDLER*) &Handler, This->pTypeInfo,
                                LocalUrl, Heap, NULL);
        CHECKSC2("CreateDeserializer");
        /* XXX does not take ref on serializer. RESTARTABLE XXX */

        *pDeserializer = (PIUNKNOWN) Handler;
    } else {
        //assert(FALSE);
        /* Restart: reinit Hanlder, Serializer */
        Handler = (PDESERIALIZER) *pDeserializer; /* XXX type check */
    }

    if (Flags & PARSE_URL_FLAGS_XTE) {
        sc = XBaxParse(Tokenizer, (PISAXHANDLER) Handler);
    } else {
        /* Construct the SAX parser */
        sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"),
                          NAME_SPACE_READ, &IID_ISaxParser,
                          (void**) &SAXParser);
        CHECKSC2("bind to sax.cob");

        sc = SAXParser->v->Parse(SAXParser, Tokenizer, (PISAXHANDLER) Handler,
                                 &Handler->SaxState);
        RELEASE_NULL(SAXParser);
        if (sc == E_WOULD_BLOCK) {
            Handler = NULL;
            goto Out;
        }
    }
#endif

    cont = Handler->prpcinfo;

    CHECKSC2("Request message parsing");
    *ParsingState = 1;

    sc = CreateSerializer((PIFILE) *pContent, Flags, &Serializer);
    CHECKSC("CreateSerializer");

    sc = Serializer->v->Reset(Serializer, NULL);
    CHECKSC("truncate repsonse file");

    if (Handler->ultimateReceiver == FALSE) { /* routing */
        assert(FALSE);
        sc = E_NOT_IMPLEMENTED;
        //sc = ForwardMsg(Handler, pContent);
    } else { /* local call */
        CHECKSETSC2(cont, E_UNEXPECTED_ERROR, "exists(cont)");

        assert(cont->Method && cont->Frame && Serializer
               && !cont->Serializer && !cont->Request && !cont->namespaces);

        cont->Serializer = Serializer;
        Serializer = NULL;

        cont->Request = Request;
        if (Request)
            Request->v->AddRef(Request);

        /* Save into cont what we need later */
        /* Now done here instead of deserializer.c after CreateAppStack */
        cont->namespaces = Handler->namespaces;
        Handler->namespaces = NULL;

        if (Handler->envelopeNS)
            Handler->prpcinfo->envelopeNS = Handler->envelopeNS;

        CHECK_HEAP();
        sc = cont->v->Execute(iCONT(cont), FALSE, NULL);
        CHECK_HEAP();
        if (Request && FAILED(sc)) {
            /* Drop ref to request we took above */
            RELEASE(Request);
            cont->Request = NULL;
        }
        CHECKSC2("cont->Execute");
    }

    if (cont && cont->metd
        && cont->metd->ResponseNameToken == TOKEN_invalid)
        sc = S_NO_RESPONSE;
        
METHOD_CLEANUP2;
    assert(sc != S_FALSE);
    /* When UDP, don't send fault message unless we know at least the method
     * and that that method wanted a reply.
     */
    if (FAILED(sc) && (Flags & PARSE_URL_FLAGS_UDP)
        && (!cont || !cont->metd || (cont->metd->ResponseNameToken
                                     == TOKEN_invalid)))
        sc = S_NO_RESPONSE;

    /* Only generate Fault message when continuation present.
     * In certain failure cases there will only be an HTTP error or
     * connection closure / drop.
     */
    if (FAILED(sc) && cont) {
        SCODE sc2 = sc;
        CHECK_HEAP();
        if (!Serializer)
            Serializer = cont->Serializer;

        if (!Serializer) {
            sc = CreateSerializer((PIFILE) *pContent, Flags, &Serializer);
            CHECKSC("CreateSerializer");
        }

        Serializer->v->Reset(Serializer, NULL);
        CHECKSC("truncate response file");

        sc = Serializer->v->FaultMessage(Serializer, iCONT(cont), sc2, NULL);
        CHECKSC("generate FaultMessage");
    }

METHOD_CLEANUP;
    CHECK_HEAP();
    RELEASE(SAXParser);
    if (Handler) {
        *pDeserializer = NULL;
        RELEASE((PISAXHANDLER) Handler); /* releases cont */
    }

    if (Request && sc != S_NO_RESPONSE && sc != E_WOULD_BLOCK) {
        PIFILE File = NULL;
        SCODE sc2 = Serializer->v->GetContent(Serializer, &File);
        if (FAILED(sc2))
            return sc2;
        sc = Request->v->SendMessage(Request,
                                     (SUCCEEDED(sc)
                                      ? TOKEN_OK : TOKEN_SOAP_PARSING_ERROR),
                                     NULL, /* SoapAction */
                                     File,
                                     NULL);
        File->v->Release(File);
        RELEASE(Serializer);
        return S_NO_RESPONSE;
    }
    RELEASE(Serializer);
    RELEASE(SoapFactory);
    return sc;
}

PRIVATE SCODE SOAPHTTPProcessHead(PIHTTPPROCESS Http, PIHTTPCONNECTOR Request)
{
    TRACE("SOAPHTTPProcessHead");
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE SOAPHTTPProcessPut(PIHTTPPROCESS Http, PIHTTPCONNECTOR Request)
{
    TRACE("SOAPHTTPProcessPut");
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE SOAPHTTPProcessDelete(PIHTTPPROCESS Http, PIHTTPCONNECTOR Request)
{
    TRACE("SOAPHTTPProcessDelete");
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE SOAPHTTPProcessOptions(PIHTTPPROCESS Http, PIHTTPCONNECTOR Request)
{
    TRACE("SOAPHTTPProcessOptions");
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE SOAPBoxDestroy(PIHTTPPROCESS Box)
{
    PSOAPBOX    This = (PSOAPBOX) Box;
    PIHEAP Heap = CurrentHeap();

    TRACE("SOAPFactoryDestroyBox");

    Heap->v->Free(Heap, 0, This);
    return S_OK;
}

/* HTTP processing implementation */
PRIVATE SCODE SOAPHTTPProcessQueryInterface(PIHTTPPROCESS Box, REFIID IidQuery,
                                            void** Object)
{
    TRACE("SOAPHTTPProcessQueryInterface");

    return GenericQueryInterface((PIUNKNOWN) Box, IidQuery,
                                 Object, &IID_IHttpProcess);
}

PRIVATE UINT SOAPHTTPProcessAddRef(PIHTTPPROCESS Box)
{
    PSOAPBOX    This = (PSOAPBOX) Box;
    UINT        Refs;

    TRACE("SOAPBoxAddRef");
    Refs = AtomicInc(&This->Refs);
    return Refs;
}

PRIVATE UINT SOAPHTTPProcessRelease(PIHTTPPROCESS Box)
{
    PSOAPBOX This = (PSOAPBOX) Box;
    UINT Refs;

    TRACE("SOAPBoxRelease");

    Refs = AtomicDec(&This->Refs);
    if (Refs > 0) return Refs;

    SOAPBoxDestroy(Box);
    return 0;
}

PRIVATE SCODE SOAPHTTPProcessGet(PIHTTPPROCESS Process, PIHTTPCONNECTOR Request,
                   /*out*/ PIFILE *pResponseFile)
{
    TRACE("SOAPHTTPProcessGet");
    return E_NOT_IMPLEMENTED;
}

#include <s_soapbox_v.c>

PIHTTPPROCESS SOAPBoxCreate(void)
{
    PSOAPBOX    This = NULL;
    PIHEAP      Heap = CurrentHeap();

    TRACE("SOAPFactoryCreateBox");

    This = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("SoapBox"), sizeof(SOAPBOX),0);
    if (This == NULL) return NULL;

    This->v     = &SOAPBoxVtbl;
    This->Refs      = 1;

    return (PIHTTPPROCESS) This;
}
#endif
