/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Helper for initializing AES and running it on IBuffers */
#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <tchar.h>
#include <assert.h>
#include <crypto.h>
#include "ssoap.h"
#include <netif.h>

PICYPHER GetCypher(_TCHAR *AesName, UINT KeyStrength);

SCODE MCT AESBuffer(UINT  Strength, const BYTE *Key,
                    PIBUFFER File, CYPHER_OPS Operation)
{
    PICYPHER Cypher;
    SCODE sc = S_OK;
    UINT KeyLen;

    /* Lookup the 256bit AES cypher 
     */
    DBG(("Looking for cypher %s (%d bits)\n\r", "COB/aes.cob", Strength));
    Cypher = GetCypher(_T("aes"), Strength);
    if (!Cypher)
        Cypher = GetCypher(_T("COB/aes.cob"), Strength);
    if (Cypher == NULL)
        return E_FAIL;

    /* Set the desired key
     */
    KeyLen = Strength/8;
    DBG(("Setting key (length=%d)\n\r", KeyLen));
    sc = Cypher->v->SetKey(Cypher,
                           Key,
                           KeyLen,
                           CYPHER_ENCRYPT|CYPHER_DECRYPT);
    if (FAILED(sc)) {
        DBG(("SetKey failed sc=%x\n\r", sc));
        goto Out;
    }

    sc = File->v->ApplyCypher(File, Cypher, Operation);

    /* Done
     */
 Out:
    Cypher->v->Release(Cypher);
    return sc;
}

