/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>

#ifdef _UNICODE   /* This goes into tchar.h */
#define _i64totcs _i64towcs
#define _uitotcs _uitowcs
#else
#define _i64totcs _i64tostr
#define _uitotcs _uitostr
#endif

/* Converts the digits of the given Val argument to a null-terminated _TCHAR
 * string and stores the result in *pBuf.  Note that pBuf points to the END
 * of the buffer on entrance.  Return the pointer to the begining of string.
 */
_TCHAR * _uitotcs(UINT Val, UINT Base, _TCHAR * pBuf)
{
    UINT Digit;

    *pBuf = _EOS;            /* pBuf points to the end of buffer */

    do {
    /* NOTE: could calculate both at once: div(Val, Base) */
        Digit = Val % Base;
        Val = Val / Base;

        --pBuf;
        if (Digit > 9)
            *pBuf = (char)(Digit - 10 + 'a');
        else
            *pBuf = (char)(Digit + '0');
    } while (Val > 0);

    return pBuf;
}

/* Ditto for 64bit values.
 * Converts the digits of the given Val argument to a null-terminated _TCHAR
 * string and stores the result in *pBuf.  Note that pBuf points to the END
 * of the buffer on entrance.  Return the pointer to the begining of string.
 */
_TCHAR * _i64totcs(INT64 Val, UINT Base, _TCHAR * pBuf)
{
    UINT Digit;
    INT64 q, r;

    *pBuf = _EOS;                       /* pBuf points to the end of buffer */

    do {
            q = Int64DividedByInt32(Val,Base);
            r = Int64TimesInt32(q,Base);
            r = Int64Subtract(Val,r);
            Digit = Int64ToInt32(r);        /* Val % Base; at long last */

            Val = q;                        /* Val /= Base; */
        --pBuf;

        if (Digit > 9)
            *pBuf = (_TCHAR)(Digit - 10 + _TEXT('a'));
        else
            *pBuf = (_TCHAR)(Digit + _TEXT('0'));

    } while (!Int64IsZero(Val));

    return pBuf;
}
