/* Copyright (c) Microsoft Corporation. All rights reserved. */

#ifndef __HTTP_SERVER__
#define __HTTP_SERVER__ 1

#include <httputils.h>
#include "tcpserver.h"

/* CUnknown with a different v-table */
struct _HttpProcessNS {
    struct IHttpProcessVtbl const * v;
    UINT Refs;
};
struct _HttpProcessFile {
    struct IHttpProcessVtbl const * v;
    UINT Refs;
};
struct _HttpProcessSoap {
    struct IHttpProcessVtbl const * v;
    UINT Refs;
};

extern const struct _HttpProcessNS HttpProcessNS;
extern const struct _HttpProcessFile HttpProcessFile;
extern const struct _HttpProcessSoap HttpProcessSoap;

/* HTTP requests */
typedef struct HTTPHEADER           HTTPHEADER,         * PHTTPHEADER;
/* HTTPHEADERFIELD is public and accessible to proand defined in httpprocessif.h */
#define NUM_FIELDS_IN_BLOCK 10

struct HTTPHEADER {
    struct HTTPHEADERFIELD  Fields[NUM_FIELDS_IN_BLOCK];
    PHTTPHEADER             More;
};

typedef struct _HttpConnector HttpConnector;
struct _HttpConnector {
    struct IHttpConnectorVtbl const * v;
    UINT                    Refs;
#define HTTP_REQUEST_FLAGS_SOAP PARSE_URL_FLAGS_SOAP
#define HTTP_REQUEST_FLAGS_AES_SOAP PARSE_URL_FLAGS_AES
    MUTEX mx; /* Delay Request reuse until done w/ prev req (other thread) */
    CONDITION cd;
    enum HTTP_PARSING_STATE {
        HS_START = 0, HS_RUNNING = 1,
        HS_START2, HS_LINE1, HS_HEADERS, HS_END_HEAD, HS_END_HEAD2, HS_BODY,
        HS_DISPATCH, HS_XML,
    } State;
    enum HTTP_PROTOCOL_STATE {
        HPS_START = 0, HPS_DEAD
    } ProtocolState;

    /* Zeroed ... */
    PARSE_URL_FLAGS Flags;      /* some bits preserved, however */
    INT                     Method;             /* HTTP request line */
    _TCHAR *                RequestURI;
    UINT                    RequestURILength;
    INT                     HTTPVersion;
    struct HTTPHEADER       Headers;            /* Header collection */
    PIUNKNOWN               Target;             /* Request target object */

    /* Response */
    UINT                StatusCode;
    /* ... Zeroed */

    /* HTTPSession -> merged structs */
    PILISTENER              Server;
    BOOL                    isServer;
    PITOKENIZER             Tokenizer;
    BOOL                    KeepAlive;
    BOOL                    Chunked;
    BOOL                    ContentLength;
    PIHEAP Heap;
    PIENDPOINT              Endpt;

    /* Dispatcher / parser restart state */
    PIFILE ResponseFile;
    PISAXHANDLER Deserializer;

    /* SoapConnector -> merged structs */

    _TCHAR *Hostname; /* or IP address */
    struct sockaddr_in sockaddr;
    UINT sa_len;

    const char *BasicKey;       /* RFC 2617 basic authentication. UNSECURE. */
    const BYTE *Key;            /* AES key */
    UINT KeyStrength;

    PICONTINUATION cont;        /* client only */
    UINT SaxState;

    /* Leftovers from old UdpConnector */
    CERTIFICATE         ServerCert;
    SESSIONKEY          SKey;
    BOOL                isSKeyValid;

};

SCODE MCT   HTTPSessionCreate(PILISTENER, PIENDPOINT,
                              PARSE_URL_FLAGS, HttpConnector**);

BOOL HTTPSessionIsKeepAlive(HttpConnector *);
SCODE MCT   HTTPRequestParse(PIHTTPCONNECTOR);
SCODE MCT   HTTPRequestDispatch(PIHTTPCONNECTOR, BOOL isResponse);

PITOKENIZER MCT HTTPRequestGetContent(PIHTTPCONNECTOR Request);
SCODE MCT HTTPRequestGetHeaderField(PIHTTPCONNECTOR Request, 
                                    INT /*HTTPTOKEN*/ FieldToken, 
                                    PHTTPHEADERFIELD* Field);
SCODE MCT HTTPRequestGetRequestURI(PIHTTPCONNECTOR Request,
                                   PBYTE* URI,
                                   INT*   Length);

#define FIRST_PART_OF_REST_OF_HEADER "Content-Length: "
//#define SECOND_PART_OF_REST_OF_HEADER "\r\nContent-Type: text/xml; charset=\"UTF-8\"\r\n"
#define XML_HEADER "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>"
#define EXPIRY_HEADER "Expiry : -1;"

#endif /* __HTTP_SERVER__ */
