/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements Marionette concrete class in component marionette */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h> /* for offsetof() */
#include <mmlite.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <sampleif.h>
#include <soapif.h>
#include <base/cobdesc.h>

typedef struct _Marionette {
    //CONST_VTBL struct IRobotVtbl *v;
    struct _DefCob super;

    CONST_VTBL struct IProgramVtbl *marcv;
    /* Add object state here */
} Marionette;

#include "s_Marionette_v.c"

/****************** COM scaffolding ***************/

/* Convert between interface and instance pointer and vice versa */
#define pmars(_i_) ((Marionette *)((_i_)))
#define imars(_p_) ((PIROBOT)((_p_)))
#define pmarc(_i_) ((Marionette *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _Marionette, marcv) : 0))
#define imarc(_p_) ((PIPROGRAM)((_p_) ? &(_p_)->marcv : NULL))

#define MovementOff offsetof(struct Movement, ListNode)

PRIVATE SCODE MarionetteDestroy(IUnknown *iThis)
{
    Marionette *This = pmars(iThis);
    PIHEAP pHeap = CurrentHeap();
    /* Finalize object */

    return pHeap->v->Free(pHeap, 0, (void *) This);
}

#if 0
PRIVATE SCODE marsQueryInterface (
    PIROBOT iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                               &IID_IRobot);
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IProgram, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) imarc(pmars(iThis));
        sc = S_OK;
    }

    return sc;
}

PRIVATE UINT marsAddRef (
    PIROBOT iThis)
{
    return AtomicInc(&pmars(iThis)->RefCnt);
}

PRIVATE UINT marsRelease (
    PIROBOT iThis)
{
    Marionette *This = pmars(iThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    if (Refs == 0)
        MarionetteDestroy((IUnknown *) This);

    return Refs;
}
#endif

PRIVATE SCODE marcQueryInterface (
    PIPROGRAM iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    IUnknown *unk = (IUnknown *) pmarc(iThis);

    return unk->v->QueryInterface(unk, iid, ppObject);
}

PRIVATE UINT marcAddRef (
    PIPROGRAM iThis)
{
    IUnknown *unk = (IUnknown *) pmarc(iThis);
    return unk->v->AddRef(unk);
}

PRIVATE UINT marcRelease (
    PIPROGRAM iThis)
{
    IUnknown *unk = (IUnknown *) pmarc(iThis);
    return unk->v->Release(unk);
}

/****************** Server methods ***************/

/* METHOD: make it start */
PRIVATE SCODE marsStart (
    PIROBOT iThis)
{
    Marionette *This = pmars(iThis);
    UINT16 *enable = (void*) 0x30000008;

    UnusedParameter(This);

    /* Make it actually start */
    *enable = 1;

    return S_OK;
}

/* METHOD: make it stop */
PRIVATE SCODE marsStop (
    PIROBOT iThis)
{
    Marionette *This = pmars(iThis);
    UINT16 *enable = (void*) 0x30000008;

    UnusedParameter(This);

    /* Make it actually stop */
    *enable = 0;

    return S_OK;
}

/* HELPER: execute one move */
PRIVATE INT MovementExecute(PTR pItem, PTR pArg)
{
    METHOD_DECLARE_NOTHIS(Movement, Execute);
    PMovement mov = pItem;
    Marionette *This = pArg;    /* get i/o address etc. from here */

    UINT16 *beat = (void*) 0x30000000;
    UINT16 *cmd = (void*) 0x30000004;

    sc = S_OK;

    /* Wait until correct time */
    if (!Int64IsZero(mov->When)) {
        sc = SleepUntil(mov->When);
        CHECKSC("SleepUntil");
    }

#if defined(arm) || defined(mips)
    if (mov->Function <= 128)
        *cmd = mov->Function; // 0..128; def=0;
    if (mov->Amount <= 10)
        *beat = mov->Amount; // 0..10; def=5;
#else
    /* ADD HERE: Make it actually move */
#endif
    printf("Move f=%d a=%d s=%d\n", mov->Function, mov->Amount, mov->Speed);

    METHOD_CLEANUP;
    return sc;
}

/* METHOD: run a script of movements */
PRIVATE SCODE marsRun (
    PIROBOT iThis,
    PMovement Script)
{
    Marionette *This = pmars(iThis);

    return (SCODE) ListWalk(&Script, MovementOff, LIST_WALK_SCODE,
                            MovementExecute, This);
}

/****************** Client methods ***************/

void GloveLoop(THREAD_ARGUMENT arg)
{
    METHOD_DECLARE_NOTHIS(Glove, Loop);
    PIROBOT robot = (void*) arg;
    TIME time = CurrentTime();
    TIME delta = Int64Initializer(0, TIME_SECONDS(2));
    UINT ex, sign;
    struct Movement mov;
    PIGLOVE glove;

printf("GloveLoop irobot=%x\n", robot);

    sc = BindToObject(CurrentNameSpace(),
                      _T("COB/glove.cob"),NAME_SPACE_READ,&IID_IGlove,
                      (void**)&glove);
    CHECKSC("get glove.cob");

    for (;;) {
        /* Initialize singleton movement list */
        memset(&mov, 0, sizeof(mov));
        ListNodeInit(MovementOff, &mov);

        time = Int64Add(time, delta);
        SleepUntil(time);
        
        glove->v->Getglove(glove, &sign, &ex);
        if (!ex) {
            printf("Glove says exit, bye...\n");
            break;
        }

printf("Got %d from glove\n", sign);

        if (sign != mov.Function) {
            mov.Function = sign;
            mov.Amount = ~0;
            
            sc = robot->v->Run(robot, &mov);
            CHECKSC("robot->Run");
        }
    }

METHOD_CLEANUP;
    RELEASE(robot);
}

void PitchLoop(THREAD_ARGUMENT arg)
{
    METHOD_DECLARE_NOTHIS(Pitch, Loop);
    PIROBOT robot = (void*) arg;
    UINT8 value, prev = 0;
    UINT count;
    struct Movement mov;
    PIFILE file;

printf("PitchLoop irobot=%x\n", robot);
    sc = BindToObject(CurrentNameSpace(),
                      _T("pitch"),NAME_SPACE_READ,&IID_IFile,
                      (void**)&file);
    CHECKSC("get pitch file");

    for (;;) {
        /* Initialize singleton movement list */
        memset(&mov, 0, sizeof(mov));
        ListNodeInit(MovementOff, &mov);

        sc = file->v->ReadAt(file, UINT64_ZERO, &value, 1, &count);
        CHECKSC("read pitch");
        if (sc == S_FALSE && count == 0) {
printf("PitchLoop EOF, bye...\n");
            break;
        }
        if (count == 0)
            continue;

        value -= '0';

        mov.Function = ~0;
        mov.Amount = value;

        if (prev != value) {
            sc = robot->v->Run(robot, &mov);
            CHECKSC("robot->Run");
        }
        prev = value;
    }

METHOD_CLEANUP;
    RELEASE(robot);
}

/* Simple main function, called from command line */
PRIVATE SCODE marcMain (
    PIPROGRAM iThis,
    const _TCHAR * cmdline)
{
    METHOD_DECLARE_MULTI(Marionette, marc, Main);
    PIROBOT robot = NULL;
    PISOAPFACTORY soapfactory = NULL;
    const _TCHAR *query, *frag;
    UINT qlen;
    struct Movement mov;

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    CHECKSC("get soap factory");

    sc = soapfactory->v->CreateProxy(soapfactory, NULL, cmdline,
                                     _T("irobot"),
                                     0,
                                     (PIUNKNOWN*) &robot);
    CHECKSC("CreateProxy3 irobot");

    sc = soapfactory->v->ParseURL(soapfactory, cmdline,
                                  NULL, NULL, NULL, NULL, NULL, NULL,
                                  &query, &qlen, &frag);
    CHECKSC("ParseURL cmdline");

    /* USAGE: marionette.cob http://1.2.3.4/COB/marionette.cob?AMOUNT#WHAT
     * e.g. marionette.cob http://127.0.0.1:4000/COB/marionette.cob?3#44
     */
    if (!_tcscmp(frag, _T("start"))) {
        sc = robot->v->Start(robot);
        CHECKSC("start");
    } else if (!_tcscmp(frag, _T("stop"))) {
        sc = robot->v->Stop(robot);
        CHECKSC("stop");
    } else if (!_tcscmp(frag, _T("glove"))) {
        PIPROCESS prc = CurrentProcess();
        sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                                  GloveLoop, (THREAD_ARGUMENT) robot,
                                  0, NULL, NULL);
        CHECKSC("CreateThread GloveLoop");
        robot = NULL;
    } else if (!_tcscmp(frag, _T("pitch"))) {
        PIPROCESS prc = CurrentProcess();
        sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                                  PitchLoop, (THREAD_ARGUMENT) robot,
                                  0, NULL, NULL);
        CHECKSC("CreateThread PitchLoop");
        robot = NULL;
    } else {
        /* Initialize singleton movement list */
        memset(&mov, 0, sizeof(mov));
        ListNodeInit(MovementOff, &mov);

        /* URL fragment is integer, what to move */
        mov.Function = _ttoi(frag);
        /* URL query is amount (this is pretty simplistic parsing) */
        mov.Amount = _ttoi(query);

        sc = robot->v->Run(robot, &mov);
    }

METHOD_CLEANUP;
    RELEASE(robot);
    RELEASE(soapfactory);

    return sc;
}
