/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <crypto.h>
#include <mmmacros.h>
#include <base/cobdesc.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <sampleif.h>
#include <framebuffer.h>
#include <tabletService.h>


typedef struct _tabletServiceImpl{
	struct _DefCob super;
    UINT Position;
} tablet_Service, tabletServiceImpl, *Ptablet_Service;


#define tabletServiceImplDestroy NULL


#include <s_tabletServiceImpl_v.c>

PRIVATE PIFRAMEBUFFER pDispl = NULL;
PRIVATE UINT8 *FrameBufferBase;
PRIVATE BOOL bDispInitOk = FALSE;


int absolute(int iPar)
{
    if(iPar < 0)
        return -iPar;
    else 
        return iPar;
}

SCODE draw(int iWidthScale, int iHeightScale, PtabletStroke strokeList)
{
    int iV = 1;
    int iXN = 0;
    int iYN = 0;
    int iBuffIndex = 0;
    int iVX = 0;
    int iVY = 0;
    int iRatioX = 0;
    int iRatioY = 0;
    int itRatioX = 0;
    int itRatioY = 0;
    int iIterations= 0 ;
    int j = 0;
    int iX = 0;
    int iY = 0;
    
    BYTE* pBuff = FrameBufferBase;
    BYTE bYdirPos = 0;
    PtabletStroke tempStroke = NULL;
    struct point* pPoint = NULL;
    struct point* pOrigPoint = NULL;
    
    if(!strokeList || !strokeList->pointList)
        return E_FAIL;//what are the valid error codes?
    tempStroke = strokeList;

    //clear the screen
    iBuffIndex = 0;
    for (j = 0; j < 64000; j++)
        pBuff[iBuffIndex++] = 0x00;

    while(1)
    {//loop for the strokes
    
        pPoint = tempStroke->pointList;
        pOrigPoint = pPoint;
        iX = pPoint->x / iWidthScale;
        iY = pPoint->y / iHeightScale;
        //for each increment in the Y direction the buffer index must be incremented by 320
        //for each increment in the X direction the buffer index must be incremented by 1
        iBuffIndex = iX + iY * 320;
        pBuff[iBuffIndex] = 0xff;

        while(1)
        {//loop for the points in one stroke
            pPoint = pPoint->ListNode.pNext;
            if(pOrigPoint == pPoint)
                break;
            //calculate the coordinates in the screen coordinate systems
            iXN = pPoint->x / iWidthScale;
            iYN = pPoint->y / iHeightScale;
           
            //compute abs of the coordinates of the vector from the previous point to the current 
            //point
            iVX = absolute(iX - iXN) ;
            iVY = absolute(iY - iYN) ;
            //calculate ratio between the change in coordinates in relation to Y for the X 
            //coordinate
            if(iVX)
            {
                itRatioX = iRatioX = iVY / iVX;
                if((iVY - iRatioX * iVX) > (iVX / 2))
                {
                    itRatioX++;
                    iRatioX++;
                }
            }
            else
                itRatioX = iRatioX = 0;
            //calculate ratio between the change in coordinates in relation to X for the Y
            //coordinate
            if(iVY)
            {
                itRatioY = iRatioY = iVX / iVY;
                if((iVX - iRatioY * iVY) > (iVY / 2))
                {
                    itRatioY++;
                    iRatioY++;
                }
            }
            else
                itRatioY = iRatioY = 0;
                   
            while(iVY || iVX)
            {//loop for drawing the points in one line
                if(!itRatioX && iVX)
                {//x can be updated                    
                    if(iX < iXN)//moving in positive direction on X
                        iBuffIndex++;
                    else
                        iBuffIndex--;
                    iVX --;
                    itRatioX = iRatioX;
                }
                else//decrement counter 
                    itRatioX --;

                if(!itRatioY && iVY)
                {//y can be updated
                    if(iY < iYN)
                        iBuffIndex += 320;
                    else
                        iBuffIndex -= 320;
                    iVY --;
                    itRatioY = iRatioY;
                }
                else
                    itRatioY --;
                //draw the pixel
                pBuff[iBuffIndex] = 0xff;
            }//end of the line drawing loop
            iX = iXN;
            iY = iYN;
        }//end of the point loop
        tempStroke = tempStroke->ListNode.pNext;
        if(tempStroke == strokeList)
            break;
    }//end of the stroke loop    
    return S_OK;
}



SCODE InitScreen(void)
{
    METHOD_DECLARE_NOTHIS(tabletService, ScreenInit);
    UINT Size;

    sc = BindToObject(CurrentNameSpace(), _T("display"),
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IFrameBuffer,
                      (void**)&pDispl);
    if(FAILED(sc))
        return sc;

    CHECKSC("Bind(display)");

    // Set the screen mode
    sc = pDispl->v->SetMode(pDispl,FB_MODE_320x200_in_320x400_8bpp_PALETTE);
    if(FAILED(sc))
        return sc;

    // Map the screen
    Size = 0;
    sc = pDispl->v->Map(pDispl,(PTR*)&FrameBufferBase,&Size);
    if(FAILED(sc))
        return sc;

    memset(FrameBufferBase,0,Size);
    METHOD_CLEANUP;
    return sc;
}

void TerminateScreen(void)
{
    if (pDispl) {
        pDispl->v->Unmap(pDispl,FrameBufferBase);
        pDispl->v->Release(pDispl);
        pDispl = NULL;
        FrameBufferBase = NULL;
    }
}

PRIVATE SCODE MCT tablet_tabletInfo(
							  /*in*/ PITABLET iThis,
							  /*in*/ INT width,
							  /*in*/ INT height,
							  /*in*/ PtabletStroke strokeList)
{
    int iWidthCorrection;
	int iHeightCorrection;

    SCODE sc = E_FAIL;

    if(!bDispInitOk)
    {
        sc = InitScreen();
        if(FAILED(sc))
        {
            _tprintf(_T("Could not initialize display (%x)\r\n"), sc);
            return sc;
        }
        else
            bDispInitOk = TRUE;
    }
    
    iWidthCorrection = 40;
    iHeightCorrection = 40;
    sc = draw(iWidthCorrection, iHeightCorrection, strokeList);
    if(FAILED(sc))
    {
        _tprintf(_T("Could not draw to display(%x)\r\n"), sc);
        return sc;
    }

    //TerminateScreen();

	return S_OK;
}



