/* Copyright (c) Microsoft Corporation. All rights reserved. */

#ifndef __BUILDER__
#define __BUILDER__

/* Implements ReservationBuilder concrete class */
#include <mmlite.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>

typedef struct _ReservationBuilder {
    /* Housekeeping */
    CONST_VTBL struct IReservationBuilderVtbl *v;
    UINT RefCnt;
    /* Reservation data */
    struct reservation *Reservation;
    _TCHAR *Host;       /* Host URI */
    _TCHAR *FullName;   /* Task name, reservation name, suffix */
    /* Dispatch method and data */
    SCODE (*DispatchCallback)(PTR Data, PTR Builder);
    PTR DispatchData;
    /* The module of DispatchFunction, keep ref so code won't get unloaded */
    PIMODULE DispatchModule;
    /* The task the reservation is part of */
    PITASKBUILDER Task;
    /* Other reservations owned by same task */
    ListNodeDeclare(_ReservationBuilder, ListNode);
    /* Adaptivity parameters */
    _TCHAR *DataCollection;
    UINT Confidence; /* 0 means not adaptive */
    SCODE (*AdaptationCallback)
        (PTR pArg, 
         PIPLANNER Planner, 
         struct _ReservationBuilder *Builder, 
         PTR Data, 
         PISTATISTICS Statistics, 
         PBOOL NothingAdapts);
    PTR AdaptationData;
} ReservationBuilder, *PRESERVATIONBUILDER;
    
/* Convert between interface and instance pointer and vice versa */
#define pReservationBuilder(_i_) ((PRESERVATIONBUILDER)((_i_)))
#define iReservationBuilder(_p_) ((PIRESERVATIONBUILDER)((_p_)))
    
#define rbloff offsetof(struct _ReservationBuilder, ListNode)    
#define rloff offsetof(struct reservation, ListNode)    
#define trloff offsetof(struct trigger, ListNode)
    
typedef struct _TaskBuilder {
    /* Housekeeping */
    CONST_VTBL struct ITaskBuilderVtbl *v;
    UINT RefCnt;
    /* Task data */
    _TCHAR *TaskName;
    /* The planning library the task belongs to */
    PIPLANNER Planner;
    /* The instigator that created the task */
    PIINSTIGATOR Instigator;
    /* List of reservations in this task */
    PRESERVATIONBUILDER Reservations; 
    /* Other tasks owned by same planner */
    ListNodeDeclare(_TaskBuilder, ListNode);
} TaskBuilder, *PTASKBUILDER;
    
/* Convert between interface and instance pointer and vice versa */
#define pTaskBuilder(_i_) ((PTASKBUILDER)((_i_)))
#define iTaskBuilder(_p_) ((PITASKBUILDER)((_p_)))

#define tbloff offsetof(TaskBuilder, ListNode)    

SCODE TaskBuilderCreate(
    PIPLANNER Planner, PIINSTIGATOR Instigator, PITASKBUILDER *Task);

INT TaskBuilderDelete(PTR pItem, PTR pArg);

#endif /* __BUILDER__ */
