/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>
#include <sampleif.h>
#include <web/proxy.h>
#include <web/udpsoapserv.h>
#include <arm/at91m63200.h>

/* Localhost settings */
#define HOST_URI _T("http://127.0.0.1/COB/sensor.cob")
#define FEEDBACK_URI _T("http://127.0.0.1/COB/wsres.cob")

/* miura eb63 board */
// #define HOST_URI _T("http://172.31.42.132/COB/sensor.cob")
/* jvh11 */
//#define HOST_URI _T("http://172.31.42.6/COB/sensor.cob")
//#define FEEDBACK_URI _T("http://172.31.42.6/COB/wsres.cob")

#define RES_DEADLINE    CurrentTime()
#define RES_DEADLINE_OFF TIME_RELATIVE(TIME_MILLIS(1000))
#define RES_DURATION    TIME_RELATIVE(TIME_MILLIS(200))
#define RES_TOLERANCE   TIME_RELATIVE(TIME_MILLIS(20))
#define RES_PERIOD      TIME_RELATIVE(TIME_SECONDS(5))
#define RES_REPEATS     100
#define RES_CONFIDENCE  95

#define WORKLOAD_STRING _T("10000")

typedef struct _Instigator {
    /* Housekeeping */
    CONST_VTBL struct IInstigatorVtbl *v;   
    UINT RefCnt;
    /* Instigated task */
    PITASKBUILDER Task;
    /* Dispatch data */
    _TCHAR *HostUri;
    _TCHAR *FeedbackUri;

    PISOAPFACTORY soap;
    PWSCONTEXT ctx;
    PISENSOR pxy;

} Instigator, *PINSTIGATOR;

#define pInstigator(_i_) ((PINSTIGATOR)((_i_)))
#define iInstigator(_p_) ((PIINSTIGATOR)((_p_)))

PRIVATE SCODE MCT Dispatch(
    PTR Arg, PIRESERVATIONBUILDER Reservation)
{
    PIINSTIGATOR iThis = (PIINSTIGATOR) Arg;
    METHOD_DECLARE(Instigator, Dispatch);
/*
    PISOAPFACTORY soap = NULL;
    PWSCONTEXT ctx = NULL;
    PISENSOR pxy = NULL;
*/
    if (!This->ctx) {
        sc = Reservation->v->CreateContext(Reservation, FEEDBACK_URI,
                                           &This->ctx);
        CHECKSC("create context object");
    } else {
        sc = Reservation->v->ResetContext(Reservation, This->ctx);
        CHECKSC("reset context object");
    }

    if (!This->soap) {
        sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                        &IID_ISoapFactory, (void **) &This->soap);
        CHECKSC("bind to SOAP COB");
    }

    if (!This->pxy) {
        sc = This->soap->v->CreateProxy3(This->soap, NULL,
                                         This->ctx, _T("sensor"),
                                         0,
                                         (PIUNKNOWN*) &This->pxy);
        CHECKSC("create proxy");
        printf("planexp1.c:Dispatch: Created proxy: pxy=%x sc=%x\n", (UINT) This->pxy, sc);
    }

    sc = This->pxy->v->Main(This->pxy, WORKLOAD_STRING);
    CHECKSC("Main");
METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE InstigatorQueryInterface (
    PIINSTIGATOR pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_IInstigator);
}
    
PRIVATE UINT InstigatorAddRef (
    PIINSTIGATOR pThis)
{
    return AtomicInc(&pInstigator(pThis)->RefCnt);
}
    
PRIVATE UINT InstigatorRelease (
    PIINSTIGATOR pThis)
{
    PINSTIGATOR This = pInstigator(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    printf("tplanner.c:InstigatorRelease this=0x%x refs=%d\n", 
        This, This->RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */
        if (This->HostUri)
            free(This->HostUri);
        if (This->FeedbackUri) 
            free(This->FeedbackUri);
        if (This->pxy)
            This->pxy->v->Release(This->pxy);
        if (This->ctx) {

            // XXX Free internal strings, allocated in create/reset context

            free(This->ctx);
        }
        if (This->soap)
            This->soap->v->Release(This->soap);
        This->Task = NULL;
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}
    
PRIVATE SCODE InstigatorInitialize(
    PIINSTIGATOR iThis, INT argc, _TCHAR **argv)
{
    METHOD_DECLARE(Instigator, Initialize);

    This->HostUri = _tcsdup(HOST_URI);
    This->FeedbackUri = _tcsdup(FEEDBACK_URI);

    printf("planexp1.c:Initialize: host=%s feedback=%s\n", 
        This->HostUri, This->FeedbackUri);

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE InstigatorDefineTask(PIINSTIGATOR iThis)
{
    METHOD_DECLARE(Instigator, DefineTask);
    PIRESERVATIONBUILDER Reservation;
    _TCHAR *ReservationName;
    TIME Period, dl;

    sc = This->Task->v->CreateReservation(This->Task, 
        This->HostUri, This->FeedbackUri, 
        _T("planexp1-reserve"),
        &Reservation);
    CHECKSC("create reservation");

    dl = RES_DEADLINE;
    dl = Int64SubtractInt32(dl, RES_DEADLINE_OFF);
    sc = Reservation->v->SetParameters(Reservation, 
                                       dl, RES_DURATION, RES_TOLERANCE);
    CHECKSC("set reservation parameters");

    Int32ToInt64(Period, RES_PERIOD);
    sc = Reservation->v->SetRecurrence(Reservation, 
        Period, RES_REPEATS);
    CHECKSC("set reservation trigger");

    ReservationName = (_TCHAR *) Reservation->v->GetFullName(Reservation), 
    sc = Reservation->v->SetAdaptivity(
        Reservation, ReservationName, RES_CONFIDENCE, NULL, NULL);
    free(ReservationName);
    CHECKSC("set reservation adaptivity");

    sc = Reservation->v->SetDispatch(Reservation, 
        (PTR) Dispatch, (PTR) This);
    CHECKSC("set dispatch function and data");

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE InstigatorRun(PIINSTIGATOR iThis)
{
    METHOD_DECLARE(Instigator, Run);
    PIPLANNER Planner;

    sc = BindToObject(CurrentNameSpace(), _T("COB/planner.cob"), NAME_SPACE_READ,
                      &IID_IPlanner, (void **) &Planner);
    CHECKSC("create planner");

    sc = Planner->v->CreateTask(Planner, iThis, &This->Task);
    CHECKSC("create task");

    sc = iThis->v->DefineTask(iThis);
    CHECKSC("define task");

    sc = This->Task->v->Activate(This->Task, iThis);
    CHECKSC("activate task");

METHOD_CLEANUP;
    return FAILED(sc) ? -1 : 0;
}
    
const struct IInstigatorVtbl InstigatorVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    InstigatorQueryInterface,
    InstigatorAddRef,
    InstigatorRelease,
    InstigatorInitialize,
    InstigatorDefineTask,
    InstigatorRun
};    
    
PRIVATE PIINSTIGATOR InstigatorCreate(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(planexp1.c:Instigator, Create);
    PIHEAP Heap = CurrentHeap();
    PIINSTIGATOR Instigator;

    Instigator = (PIINSTIGATOR) Heap->v->Alloc(Heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("planexp1.c:InstigatorCreate"),
        sizeof(Instigator), 0);
    if (!Instigator)
        return NULL;

    Instigator->v = &InstigatorVtbl;
    pInstigator(Instigator)->RefCnt = 1;

    sc = Instigator->v->Initialize(Instigator, argc, argv); 
    CHECKSC("initialize instigator");

METHOD_CLEANUP;
    if (FAILED(sc) && Instigator) {
        Instigator->v->Release(Instigator);
        Instigator = NULL;
    }
    return Instigator;
}
    
int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(planexp1.c, _tmain);

    PIINSTIGATOR Instigator = InstigatorCreate(argc, argv);
    CHECKSETSC(Instigator, E_FAIL, "create instigator");

    sc = Instigator->v->Run(Instigator);
    CHECKSC("run instigator");

METHOD_CLEANUP;
    if (Instigator)
        Instigator->v->Release(Instigator);

    return FAILED(sc) ? -1 : 0;
}
    
