/* Copyright (c) Microsoft Corporation. All rights reserved. */

#ifndef __PLANNER__
#define __PLANNER__

#include <mmlite.h>

/* Maintainer wake-up interval in seconds */
#define DEFAULT_MAINTAINER_PERIOD 3

/* The greatest number of reservation occurrences 
   that the planner allows */
#define DEFAULT_TRIGGER_MAX_COUNT 2000

/* Maximum number of repeats allowed for a trigger. 
   A trigger with a larger maxCount is reset to this 
   maximum value, but the maintainer data for that 
   reservation stores the original maxCount, and the 
   maintainer thread renews the reservation until it 
   has been claimed as often as the original maxCount
   indicates */
#define DEFAULT_TRIGGER_LEASE_COUNT 10

/* Affects how early the maintainer attempts to renew
   a triggered reservation. Also is a lower bound for 
   reservation periods */
#define PLANNER_RENEWAL_SLACK TIME_RELATIVE(TIME_SECONDS(3))

/* The period between garbage collection sweeps through
   the task list, measured in seconds */
#define HOUSE_KEEPER_PERIOD 20

/* The period between feedback collection sweeps through
   the live reservations list, measured in seconds */
#define FEEDBACK_PERIOD 8

/* Rename to LIVELISTENTRY */
typedef struct MAINTAINERDATA {
    PRESERVATIONBUILDER Reservation;
    UINT OriginalMaxCount;
    INT RemainingCount;
    TIME LastDispatchedAt; 
    TIME RenewReservationBy;
    TIME StatisticalPrediction;
    ListNodeDeclare(MAINTAINERDATA, ListNode);
} MAINTAINERDATA, *PMAINTAINERDATA;

#define mdloff offsetof(MAINTAINERDATA, ListNode)

typedef struct _Planner {
    /* Housekeeping */
    CONST_VTBL struct IPlannerVtbl *v;
    UINT RefCnt;
    MUTEX Mx;
    CONDITION Cnd1;
    CONDITION Cnd2;
    BOOL StopAllThreads;
    BOOL FeedbackThreadIsRunning;
    BOOL HouseKeeperThreadIsRunning;
    BOOL MaintainerThreadIsRunning;
    /* List of tasks owned by the planner */
    PTASKBUILDER Tasks;
    /* List of live (in need of renewal) reservations */
    PMAINTAINERDATA MaintainerData;
    /* Reservation feedback source */
    PISTATISTICS StatisticsCob;
} Planner, *PPLANNER;
    
/* Convert between interface and instance pointer and vice versa */
#define pPlanner(_i_) ((PPLANNER)((_i_)))
#define iPlanner(_p_) ((PIPLANNER)((_p_)))

#endif /* __PLANNER__ */
