/* Copyright (c) Microsoft Corporation. All rights reserved. */

#ifndef __STATISTICS__
#define __STATISTICS__

#include <base/cobdesc.h>

/* The individual samples */
typedef struct SAMPLE {
#if _DEBUG
    /* Reservation name */
    _TCHAR *Name;
    /* Sequence number */
    UINT Number;
#endif
    /* CPU cycles */
    TIME Requested;
    TIME Consumed;
} SAMPLE, *PSAMPLE;

/* The size of the sliding window, 
   e.g. number of samples to keep for each entity */
#define WINDOW_SIZE 32

/* The sampling data about a given entity */
typedef struct DATA {
    /* The sliding window of most recently arrived samples */
    SAMPLE Window[WINDOW_SIZE];
    /* False until WINDOW_SIZE samples have been received */
    BOOL IsWindowFull;
    /* The most recently arrived sample. 
       The next sample will be written in the next slot
       (overflow wraps around).
       NULL if no samples have arrived */
    PSAMPLE LatestSample;

} DATA, *PDATA;

/* The maximal number of entities to collect samples for */
#define NUM_SAMPLE_POOLS 3

/* COB facade object */
typedef struct STATISTICS {
    /* Housekeeping */
    //struct IStatisticsVtbl const *v;
    struct _DefCob super;

    MUTEX Mx;
    PIFILE Log;
    /* Data */
    _TCHAR *Names[NUM_SAMPLE_POOLS];
    DATA Data[NUM_SAMPLE_POOLS];
} Statistics, *PSTATISTICS;

#define pStatistics(_x_) ((PSTATISTICS)_x_)

#endif /* __STATISTICS__*/
