/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include "int64sqrt.c"

int test64(INT64 input, INT64 root) {
    INT64 result = Int64SquareRoot(input);
    int r;
    r = !Int64Equal(result, root);

    if (r)
        printf("int64sqrt.c: Int64SquareRoot input=%lld result=%lld correct=%lld\n", 
               input, result, root);

    return r;
}

int utest64(UINT64 input, UINT64 root) {
    UINT64 result = Uint64SquareRoot(input);
    int r;
    r = !Int64Equal(Uint64ToInt64(result), Uint64ToInt64(root));

    if (r)
        printf("int64sqrt.c: Int64SquareRoot input=%llu result=%llu correct=%llu\n", 
               input, result, root);

    return r;
}

int test(int input, int root) {
    INT64 x, y;
    Int32ToInt64(x, input);
    Int32ToInt64(y, root);
    return test64(x, y);
}

int utest(unsigned int input, unsigned int root) {
    UINT64 x, y;
    Uint32ToUint64(x, input);
    Uint32ToUint64(y, root);
    return utest64(x, y);
}

int ntest(unsigned int input, unsigned int n, unsigned int expect)
{
    unsigned int result = UintNRoot(input, n);

    if (result != expect)
        printf("int64sqrt.c: IntNRoot input=%u n=%d result=%u expect=%u\n", 
               input, n, result, expect);

    return result != expect;
}

int _tmain(INT argc, _TCHAR *argv[])
{
    int i = 0;
    INT64 input, root;

    i += ntest(27, 3, 3);
    i += ntest(1000000, 3, 100);
    i += ntest(1000000, 4, 32);

    i += test(-1, -1);
    i += test(-10, -3);
    i += test(4, 2);
    i += test(16, 4);
    i += test(9, 3);
    i += test(144, 12);
    i += test(169, 13);
    i += test(166, 12);
    i += test(172, 13);

    i += test(1 << 24, 4096);
    i += test((1 << 24)+1, 4096);
    i += test((1 << 24)+2, 4096);
    i += test((1 << 24)+3, 4096);
    i += test((1 << 24)+4, 4096);
    i += test((1 << 24)+5, 4096);
    i += test((1 << 24)+6, 4096);
    i += test((1 << 24)+7, 4096);
    i += test((1 << 24)+8, 4096);
    i += test(~0, -1);
    i += utest(~0, 0x10000);

    Int64FromHighAndLow(input, 1 << 16, 0); /* 2^48 */
    Int64FromHighAndLow(root, 0, 1 << 24);  /* 2^24 */
    i += test64(input, root);

    Int64FromHighAndLow(input, 1 << 8, 0);  /* 2^40 */
    Int64FromHighAndLow(root, 0, 1 << 20);  /* 2^20 */
    i += test64(input, root);

    if (i)
        printf("int64sqrt.c: %d tests failed\n", i);
    else
        printf("int64sqrt.c: All tests ran successfully\n");

    return 0;
}
