/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <tokenizerif.h>

int _tmain(int argc, _TCHAR** argv)
{
    SCODE           sc;
    PINAMESPACE     NameSpace       = CurrentNameSpace();
    PIUNKNOWN       Unk;

    _TCHAR          InputFileName[]  = _T("fs/http_request.txt");
    PIFILE          InputFile;

    _TCHAR          OutputFileName[] = _T("fs/Output_Request.xml");
    PIFILE          OutputFile;

    PITOKENIZER     Tokenizer;
    PISAXPARSER     SAXParser;
    PISAXHANDLER  SAXCallbacks = NULL;
    UINT Depth = 0;

    DBG(("Testing SAX parser\n"));
    DBG(("Echoing file %s\n", InputFileName));

    /* Open a file for input*/
    DBG(("Opening file for echoing\n"));
    {
        sc = NameSpace->v->Bind(NameSpace, InputFileName, 0, &Unk);
        if (FAILED(sc)) return sc;

        DBG(("Bound input file\n"));

        sc = Unk->v->QueryInterface(Unk, &IID_IFile, (void**) &InputFile);
        Unk->v->Release(Unk);
        if (FAILED(sc)) return sc;

        DBG(("Queried for input file interface\n"));
    }

    /* Open a file for output*/
    DBG(("Opening file for output\n"));
    {
        sc = NameSpace->v->Bind(NameSpace, OutputFileName, NAME_SPACE_CREATE | NAME_SPACE_WRITE, &Unk);
        if (FAILED(sc)) return sc;

        DBG(("Bound output file\n"));

        sc = Unk->v->QueryInterface(Unk, &IID_IFile, (void**) &OutputFile);
        Unk->v->Release(Unk);
        if (FAILED(sc)) return sc;

        DBG(("Queried for output file interface\n"));
    }
    

    /* Prepare a tokenizer */
    DBG(("Preparing file tokenizer\n"));
    {
        PITOKENIZERFACTORY Factory;

        sc = BindToObject(NameSpace, _T("COB/tokenizer.cob"), 0,
                          &IID_ITokenizerFactory, (void**) &Factory);
        if (FAILED(sc)) goto err_tok_query;

        DBG(("Queried for tokenizer factory interface\n"));

        sc = Factory->v->Create(Factory, (PIUNKNOWN) InputFile, &Tokenizer);
        Factory->v->Release(Factory);
        if (FAILED(sc)) goto err_tok_create;

        DBG(("Constructed tokenizer instance\n"));
    }

 
    DBG(("Constructing SAX parser\n"));
    {
        sc = NameSpace->v->Bind(NameSpace, _T("COB/sax.cob"), 0, &Unk);
        if (FAILED(sc)) goto err_parser_bind;

        DBG(("Bound SAX parser COB\n"));

        sc = Unk->v->QueryInterface(Unk, &IID_ISaxParser, (void**) &SAXParser);
        Unk->v->Release(Unk);
        if (FAILED(sc)) goto err_parser_query;

        DBG(("Queried for SAX parser interface\n"));
    }

    /*Construct the Interface for the Printer*/
    SAXParser->v->CreatePrinter(SAXParser, OutputFile, 2, &SAXCallbacks);

    /* Parse and echo */
    DBG(("Parsing and echoing\n"));

    sc = SAXParser->v->Parse(SAXParser, Tokenizer, SAXCallbacks, &Depth);
    if (FAILED(sc)) goto err_parse;

    DBG(("All done\n"));
    sc = S_OK;

err_parse:
    SAXParser->v->Release(SAXParser);
err_parser_query:
    DBG(("Jags Comment :: Parser Query Error \n"));
err_parser_bind:
    SAXCallbacks->v->Release(SAXCallbacks);
    Tokenizer->v->Release(Tokenizer);
err_tok_create:
err_tok_query:
    InputFile->v->Release(InputFile);
    OutputFile->v->Release(OutputFile);
    return sc;
}
