/* Copyright (c) Microsoft Corporation. All rights reserved. */


#include "instigator.h"


PRIVATE SCODE MCT
SensorDemoAdaptConsumerReservation(
    PINSTIGATOR This, 
    TIME ProducerDeadline, 
    PIPLANNER Planner, 
    PISTATISTICS Statistics, 
    PBOOL NothingAdapts)
{
    METHOD_DECLARE_NOTHIS(SensorDemo, AdaptConsumerDuration);
    PIRESERVATIONBUILDER Reservation = NULL;
    PMAINTAINERDATA ConsumerData;
    TIME TransmissionTime = TIME_ORIGIN;
    TIME ConsumerDeadline = TIME_ORIGIN;
    TIME DeadlineOffset = TIME_ORIGIN;

    /* Access the consumer reservation and planner data */

    sc = Planner->v->
        GetPlannerDataByFullName(
            Planner, 
            This->ConsumerReservationFullName, 
            &Reservation, 
            (PTR*) &ConsumerData);
    CHECKSC("get planner data");

    /* Adapt the consumer reservation duration */

    sc = Planner->v->
        AdaptReservation(Planner, (PTR) ConsumerData, Reservation, NothingAdapts);
    CHECKSC("adapt consumer duration");

    /* Look up the estimated producer-consumer transmission time */
    sc = Statistics->v->GetPredictionFromNormalDistribution(
        Statistics, 
        TRANSMISSION_TIME_SAMPLE_POOL, 
        TRANSMISSION_TIME_CONFIDENCE, 
        &TransmissionTime);
    CHECKSC("-- no feedback");

    /* Adapt the consumer reservation deadline */

    /* Adding two relative times gives a relative time */
    DeadlineOffset = Int64Add(
        TransmissionTime, 
        ConsumerData->Reservation->Reservation->duration);

    /* Bump the producer deadline forward by subtracting a relative time */
    ConsumerData->Reservation->Reservation->deadline = 
        Int64Subtract(ProducerDeadline, DeadlineOffset);

METHOD_CLEANUP;
    return sc;
}


SCODE MCT 
SensorDemoAdaptation(
    PTR pArg, 
    PIPLANNER Planner, 
    PRESERVATIONBUILDER Reservation, 
    PMAINTAINERDATA PlannerData, 
    PISTATISTICS Statistics, 
    PBOOL NothingAdapts)
{
    METHOD_DECLARE_NOTHIS(SensorDemo, Adaptation);
    PINSTIGATOR This = (PINSTIGATOR) pArg;

    /* This method is responsible for adapting producer and consumer 
       reservation durations, and the consumer reservation deadline */

    do {

        /* The consumer reservation can't be adapted without knowing 
        the producer reservation deadline. Postpone adapting the 
        consumer until ready with the producer: */

        if (0 != _tcscmp(PRODUCER_RESERVATION_NAME, Reservation->Reservation->name)) {
            sc = S_OK;
            break;
        }

        /* The PlannerData refers to the producer reservation */

        /* Adapt the producer reservation duration */

        sc = Planner->v->
            AdaptReservation(
                Planner, 
                (PTR) PlannerData, 
                iReservationBuilder(Reservation), 
                NothingAdapts);
        CHECKSC("adapt producer duration");

        /* Adapt the consumer reservation */

        sc = SensorDemoAdaptConsumerReservation(
            This, 
            PlannerData->Reservation->Reservation->deadline, 
            Planner,
            Statistics, 
            NothingAdapts);
        CHECKSC("adapt consumer reservation");

    } while (0);

METHOD_CLEANUP;
    return sc;
}


