/* Copyright (c) Microsoft Corporation. All rights reserved. */


#include "instigator.h"


PRIVATE SCODE MCT
SensorDemoAdaptProducerReservation(
    PINSTIGATOR This, 
    TIME ConsumerDeadline, 
    PIPLANNER Planner, 
    PISTATISTICS Statistics, 
    PBOOL NothingAdapts)
{
    METHOD_DECLARE_NOTHIS(sensor.c, AdaptProducerDuration);
    PIRESERVATIONBUILDER Reservation = NULL;
    PMAINTAINERDATA ProducerData;
    TIME TransmissionTime = TIME_ORIGIN;
    TIME ProducerDeadline = TIME_ORIGIN;
    TIME DeadlineOffset = TIME_ORIGIN;

    /* Access the producer reservation and planner data */

    sc = Planner->v->
        GetPlannerDataByFullName(
            Planner, 
            This->ProducerReservationFullName, 
            &Reservation, 
            (PTR*) &ProducerData);
    CHECKSC("get planner data");

    /* Adapt the consumer reservation duration */

    sc = Planner->v->
        AdaptReservation(Planner, (PTR) ProducerData, Reservation, NothingAdapts);
    CHECKSC("adapt producer duration");

    /* Look up the estimated producer-consumer transmission time */
    sc = Statistics->v->GetPredictionFromNormalDistribution(
        Statistics, 
        TRANSMISSION_TIME_SAMPLE_POOL, 
        TRANSMISSION_TIME_CONFIDENCE, 
        &TransmissionTime);
    if (FAILED(sc)) {
        printf("sensormain.c:SensorDemoAdaptProducerReservation: no feedback, sc=%x\n", sc);
        goto Out;
    }

    /* Adapt the producer reservation deadline */

    /* Adding two relative times gives a relative time */
    DeadlineOffset = Int64Add(
        TransmissionTime, 
        ProducerData->Reservation->Reservation->duration);

    /* Bump the producer deadline backwards by adding a relative time */
    ProducerData->Reservation->Reservation->deadline = 
        Int64Add(ConsumerDeadline, DeadlineOffset);

METHOD_CLEANUP;
    return sc;
}


SCODE MCT 
SensorDemoAdaptation(
    PTR pArg, 
    PIPLANNER Planner, 
    PRESERVATIONBUILDER Reservation, 
    PMAINTAINERDATA PlannerData, 
    PISTATISTICS Statistics, 
    PBOOL NothingAdapts)
{
    METHOD_DECLARE_NOTHIS(sensor.c, SensorDemoAdaptation);
    PINSTIGATOR This = (PINSTIGATOR) pArg;

    /* This method is responsible for adapting producer and consumer 
       reservation durations, and the producer reservation deadline */

    do {

        /* The producer reservation can't be adapted without knowing 
        the consumer reservation deadline. Postpone adapting the 
        producer until ready with the consumer: */

        if (0 != _tcscmp(CONSUMER_RESERVATION_NAME, Reservation->Reservation->name)) {
            sc = S_OK;
            break;
        }

        /* The PlannerData refers to the consumer reservation */

        /* Adapt the consumer reservation duration */

        sc = Planner->v->
            AdaptReservation(
                Planner, 
                (PTR) PlannerData, 
                iReservationBuilder(Reservation), 
                NothingAdapts);
        CHECKSC("adapt consumer duration");

        /* Adapt the producer reservation */

        sc = SensorDemoAdaptProducerReservation(
            This, 
            PlannerData->Reservation->Reservation->deadline, 
            Planner,
            Statistics, 
            NothingAdapts);
        CHECKSC("adapt producer reservation");

    } while (0);

METHOD_CLEANUP;
    return sc;
}


