/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Continuation concrete class -jvh 2004 */

#include <stdio.h>
#include <mmlite.h>
#include <ctype.h>
#include "soap.h"

#include <httpif.h>             /* for TOKEN_SOAP_* */
#include <httputils.h>

#if defined(_CLR)
#include <clrif.h>
#include <clr/clr.h>
#endif

#ifdef _DEBUG
#ifndef i386 /* XXX only for winbig, probe for thread implementation? */
//#define STACK_STATS
#endif
#endif

typedef struct CONTINFO Cont;

#define USE_TMP_HEAP 1

UINT32 H8CalcRegArgCount(CPMETHOD_DESCR pDescr);

#define STACK_OVERFLOW_SIGNATURE 0xbadc0ffe

#if defined(STACK_STATS)

PRIVATE void InitAppStackStats(PCONTINFO cont)
{
    int c;
    PUINT32 p = (PUINT32) cont->StackBase;

    /* If winbig is getting an odd exception here then ignore it (hit continue)
     * This can happen if the stack is large.
     * Since this is an actual thread stack, Windows apparently figures
     * your code is broken when the stack is written too far outside the
     * current $sp.
     */
    for (c = cont->StackSize; c > 0; c -= sizeof(UINT32))
        *p++ = STACK_OVERFLOW_SIGNATURE;
}

PRIVATE INT CheckAppStackUsage(PCONTINFO cont)
{
    INT c = -1, extra = 0;
    PUINT pStk;
    PTR start;
    start = cont->StackBase;

    if (cont) {
        pStk = (PUINT) start;
        extra = (ADDRESS) start - (ADDRESS) cont->StackBase;

        for (c = cont->StackSize - extra; c > 0; c -= sizeof *pStk)
            if (STACK_OVERFLOW_SIGNATURE != *pStk++)
                break;

        printf("use=%d extra=%d\n", c, extra);
    }

    return c + extra;
}

#endif

PRIVATE PTR ContAlloc(PICONTINUATION iThis, HEAP_FLAGS Flags,
                      UINT Size, UINT Align)
{
    PCONTINFO cont = pCONT(iThis);
#if USE_TMP_HEAP
    return cont->TmpHeap->v->Alloc(cont->TmpHeap, Flags, Size, Align);
#else
    PTR p = cont->ArgSpace;
    assert(p);

    /* Make sure the addresses returned are aligned by at least 4 since
     * the arm doesn't work otherwise.
     */
    Size = __ROUND(Size, 4);
    assert((ADDRESS)p == __ROUND((ADDRESS)p, 4));

    cont->ArgSpaceAvail -= Size;
    assert(cont->ArgSpaceAvail > 0);
    if (cont->ArgSpaceAvail < 0)
        return NULL;

    cont->ArgSpaceLast = p;
    cont->ArgSpace += Size;


    memset(p, 0, Size);

    return p;
#endif
}

PRIVATE PTR ContRealloc(PICONTINUATION iThis, UINT Flags, PTR old,
                        UINT NewSize, UINT Align)
{
    PCONTINFO cont = pCONT(iThis);
#if USE_TMP_HEAP
    return cont->TmpHeap->v->Realloc(cont->TmpHeap,
                                     Flags | HEAP_REALLOC_IN_PLACE_ONLY,
                                     old, NewSize, Align);
#else
    if (old != cont->ArgSpaceLast)
        return NULL;

    cont->ArgSpaceAvail += cont->ArgSpace - (PBYTE)old - NewSize;
    if (cont->ArgSpaceAvail <= 0) {
        DBG(("a=%x s=%x o=%x n=%x", cont->ArgSpaceAvail, cont->ArgSpace, (UINT) old, NewSize));
        assert(FALSE);
    }
    if (cont->ArgSpaceAvail < 0)
        return NULL;

    cont->ArgSpace = (PBYTE) old + NewSize;

    return old;
#endif
}

PRIVATE SCODE ContSetBuffer(PICONTINUATION iThis, PTR Where, UINT sizeis)
{
    METHOD_DECLARE_NOTHIS(ActivationSpace, SetBuffer);
    PCONTINFO cont = pCONT(iThis);
    PIHEAPFACTORY HeapFactory = NULL;

    assert(cont->ArgSpace == NULL);
    assert(cont->ArgSpaceAvail == 0);
    assert(cont->TmpHeap == NULL);

    /* Always set ArgSpace for the benefit of AllocComplex
     * that needs to figure out whether the arg is *the* ArgSpace
     */
    cont->ArgSpace = Where;
#if USE_TMP_HEAP
    sc = BindToObject(CurrentNameSpace(), _T("COB/tmp_heap.cob"),
                      NAME_SPACE_READ, &IID_IHeapFactory,
                      (void **) &HeapFactory);
    CHECKSC("bind tmp_heap.cob");
    
    sc = HeapFactory->v->Create(HeapFactory,
                                (ADDRESS) Where, 0, sizeis, sizeis,
                                &cont->TmpHeap);
    HeapFactory->v->Release(HeapFactory);
    CHECKSC("create tmp heap");
#else
    sc = S_OK;
    cont->ArgSpaceAvail = sizeis;
#endif

METHOD_CLEANUP;
    return sc;
}

/* This function is called by the activation thread when the call has returned
 */
PRIVATE void FinishSoapActivation(PCONTINFO cont)
{
    METHOD_DECLARE_NOTHIS(_, FinishSoapActivation);
    PIHTTPCONNECTOR Request = cont->Request;
    PIHEAP Heap = CurrentHeap();
    const _TCHAR *resname;

    CHECK_HEAP();
    sc = S_OK;

    if ((cont->Flags & CONT_FLAGS_LOCAL) == 0) {
        sc = cont->v->SendMessage(iCONT(cont), NULL, NULL,NULL);
    }
    RELEASE_NULL(cont->Serializer);
    if (Request) {
        Request->v->Done(Request);
        Request->v->Release(Request);
        cont->Request = NULL;
    }
    if (cont->SoapAction) {
        free(cont->SoapAction);
        cont->SoapAction = NULL;
    }

    /* Handle triggers if any */
    resname = cont->resname;
    if (!resname && cont->pContext) {
        PWSCONTEXT pCtx = cont->pContext;
        resname = pCtx->ReservationName;
        if (!resname)
            resname = pCtx->taskName;
    }
    if (resname) {
        PIWSRESERVATION zz;
        /* Zero some fields just in case we try to use them later */

        DBG(("FinishSoapActivation: will trigger from res=%s\n", resname));
        sc = BindToObject(CurrentNameSpace(), _T("COB/wsres.cob"),
                          NAME_SPACE_READ, &IID_IWsReservation,
                          (void**) &zz);
        CHECKSC("bind(COB/wsres.cob)");
        sc = zz->v->Done(zz, iCONT(cont), resname, S_OK/*XXX*/);
        CHECKSC("res done");
    }

METHOD_CLEANUP;

    /* Signal that continuation is finished. This is actually relevant
     * only in the local case. Consumes ref.
     */
    cont->v->Complete(iCONT(cont));

    /* Hmm, what about multi-cont that want to use c/c ? XXX */
    CurrentThread()->v->
        SetValue(CurrentThread(),THREAD_VALUE_RPC_CONTINUATION, (ADDRESS)NULL);

    CHECK_HEAP();
}

PRIVATE void _FinishSoapActivation(INT32 retval)
{
    PITHREAD thd = CurrentThread();
    PCONTINFO cont;

    cont = (PCONTINFO) thd->v->GetValue(thd, THREAD_VALUE_RPC_CONTINUATION);
    cont->RetVal = retval;
    FinishSoapActivation(cont);
    ThreadExit();
}

#ifdef i386
PRIVATE void __declspec(naked) _SoapActivation(void)
{
    /* Stuff return value into argument */
    __asm push eax;
    /* Push link */
    __asm push offset ThreadExit;
    __asm jmp _FinishSoapActivation;
}
#elif arm
/* Both arg1 and retval are in r0 so there is really nothing to be done here */
#define _SoapActivation _FinishSoapActivation
#else
void _SoapActivation(INT32 retval)
{
    DBG(("_SoapActivation retval=x%x\n", retval));
    _FinishSoapActivation(retval);
    ThreadExit();
}
#endif

PRIVATE SCODE ContQueryInterface (
    PICONTINUATION iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                                 &IID_IContinuation);
}

PRIVATE UINT ContAddRef (
    PICONTINUATION iThis)
{
    PCONTINFO This = pCONT(iThis);
    UINT Refs = AtomicInc(&This->RefCnt);
    return Refs;
}

/* Replicate here for linkage reasons -> ws.c */
/* Assumes the struct and string were allocated in one call */
PRIVATE INT WsEndpointDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    pHeap->v->Free(pHeap, 0, pItem);
    return 0;
}
PRIVATE void WsContextDestroy(PWSCONTEXT pCtx)
{
    PIHEAP Heap = CurrentHeap();

    if (pCtx->To)
        WsEndpointDelete(pCtx->To, Heap);
    if (pCtx->Action)
        Heap->v->Free(Heap, 0, pCtx->Action);
    if (pCtx->ReplyTo)
        WsEndpointDelete(pCtx->ReplyTo, Heap);
    if (pCtx->RelatesTo)
        Heap->v->Free(Heap, 0, pCtx->RelatesTo);
    if (pCtx->taskName)
        Heap->v->Free(Heap, 0, pCtx->taskName);
    if (pCtx->ReservationName)
        Heap->v->Free(Heap, 0, pCtx->ReservationName);
    if (pCtx->TaskName)
        Heap->v->Free(Heap, 0, pCtx->TaskName);
    if (pCtx->MessageId)
        Heap->v->Free(Heap, 0, pCtx->MessageId);
    Heap->v->Free(Heap, 0, pCtx);
}

PRIVATE UINT ContRelease (
    PICONTINUATION iThis)
{
    PIHEAP Heap = CurrentHeap();
    PCONTINFO This = pCONT(iThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    PICONVERT convert = NULL;

    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();

        (void) BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                            NAME_SPACE_READ, &IID_IConvert,
                            (void**) &convert);

        /* Finalize object */
#if defined(STACK_STATS)
        INT use = CheckAppStackUsage(This);
        printf("RPCINFO %s stack usage %d / %d \n",
               This->metd ? This->metd->name : "<header>",
               use, This->StackSize);
#endif

        /* Release the last object that was invoked. */
        RELEASE(This->pThis);
        RELEASE(This->Thread);/* second ref when native, only one in winshim */
        RELEASE(This->pConnector);
        RELEASE(This->TmpHeap);

        if (convert)
            convert->v->DestroyNamespaces(convert, &This->namespaces, Heap);
        RELEASE(convert);

        if (This->SoapAction)
            free(This->SoapAction);

        if (This->WsaTo)
            free(This->WsaTo);

        if (!(This->Flags & CONT_FLAGS_CONST)) {
            if (This->pContext && !(This->Flags & CONT_FLAGS_CONST_CONTEXT))
                WsContextDestroy(This->pContext);
            pHeap->v->Free(pHeap, 0, (void *) This);
        }
        CHECK_HEAP();
    }
    return Refs;
}

#include <web/proxy.h>

PRIVATE SCODE ContWait (
    PICONTINUATION iThis)
{
    METHOD_DECLARE_NOTHIS(Cont, Wait);
    PCONTINFO This = pCONT(iThis);
    BOOL async = FALSE;

    sc = S_OK;

    if (This->metd->ResponseNameToken == TOKEN_invalid)
        async = TRUE;

    if (!async) {
#if !USE_SERVER_FOR_RESPONSE
        if (This->Flags & CONT_FLAGS_LOCAL) {
#endif
            Mutex_Lock(&This->mx);
            while (This->State != CONT_READY && SUCCEEDED(sc))
                sc = Condition_Wait(&This->cd, &This->mx, NULL);
            Mutex_Unlock(&This->mx);
            CHECKSC("Condition_Wait");
#if !USE_SERVER_FOR_RESPONSE
        } else {
            /* ReceiveMessage consumes ref on continuation - usually desirable
             * but here we want to preserve symmetry for caller,
             * so take one extra here.
             */
            iThis->v->AddRef(iThis);

            assert (This->pConnector);

            do {
                /* Keep receiving until enough data received */
                sc = This->pConnector->v->ReceiveMessage(This->pConnector);
            } while (sc == E_RESTART || sc == E_WOULD_BLOCK);
            if (This->SoapAction) {
                free(This->SoapAction);
                This->SoapAction = NULL;
            }
#if 0
            if ((UrlFlags & PARSE_URL_FLAGS_UDP) == 0
                && (FAILED(sc)
                    || sc == S_FALSE
                    || !HTTPSessionIsKeepAlive(This->pConnector)))
            {
                /* Close socket upon failure / single use socket */
                This->pConnector->ProtocolState = HPS_DEAD;
            }
#endif
            This->pConnector->v->Done(This->pConnector);

            CHECKSC("ReceiveMessage");
        }
#endif
        sc = This->RetVal;
    }

METHOD_CLEANUP;

    return sc;
}

PRIVATE SCODE ContComplete (
    PICONTINUATION iThis)
{
    PCONTINFO This = pCONT(iThis);
    UINT refs;

    /* XXX This assumes lock is used to protect against termination! */
    Mutex_Lock(&This->mx);
    //assert(This->State == CONT_RPC); /* XXX tmp, not true for LOCAL */
    This->State = CONT_READY;    
    Condition_Broadcast(&This->cd);
    refs = iThis->v->Release(iThis);
    if (refs > 0)
        Mutex_Unlock(&This->mx);

    return S_OK;
}

PRIVATE SCODE ContSetMethod(PICONTINUATION iThis,
                    CPXMLNAMESPACE Xns, const void *xmetd,
                    const _TCHAR *objname, PIUNKNOWN pObj) /* one of these */
{
    PCONTINFO cont = pCONT(iThis);
    CPMETHOD_DESCR metd = xmetd;
    SCODE sc;
    PIUNKNOWN Object;
    REFIID iid = Xns->ifd->iface_extras->iid;
#ifdef _CLR
    PINAMESPACE pTheNameSpace;
#endif

    if (cont->Frame && cont->FrameSize != 0
        && cont->FrameSize < metd->_FrameSize)
    {
        /* Resize frame to a bigger one when necessary.
         * The frames get recycled on the server.
         * This is never done on the client, instead cont->FrameSize == 0.
         *
         * Simply drop the old one, the data was allocated from tmp heap
         */
        cont->Frame = NULL;
    }

    if (!cont->Frame) {
        cont->Frame = ContAlloc(iCONT(cont), HEAP_ZERO_MEMORY,
                                metd->_FrameSize, 0);
        if (!cont->Frame)
            return E_NOT_ENOUGH_MEMORY;
        cont->FrameSize = metd->_FrameSize;
    } else if (cont->FrameSize) {
        memset(cont->Frame, 0, metd->_FrameSize);
    }

    if (pObj) {
        sc = pObj->v->QueryInterface(pObj, iid, (void**) &Object);
    } else {
#ifdef _CLR
        if (_tcsncmp("CLR", objname, 3) == 0) {
            cont->isCLR = 1;
            cont->CLRObjectName = (_TCHAR *) (objname + 4); // Assuming that objname is not freed before cont
            pTheNameSpace = CurrentNameSpace();
            sc = pTheNameSpace->v->Bind(pTheNameSpace, CLR_NAME, 0, (PIUNKNOWN*)&Object);
        } else {
            cont->isCLR = 0;
            sc = BindToObject(CurrentNameSpace(), objname, NAME_SPACE_READ,
                          iid, (void**) &Object);
        }
#else
        /* Special for DSSP: load dssp.cob and give URL as argument to it. */
        if (!_tcscmp(Xns->QName,
                     _T("http://schemas.microsoft.com/xw/2004/10/dssp.html")))
            objname = _T("COB/dssp.cob");
        if (!_tcscmp(Xns->QName,
                     _T("http://schemas.xmlsoap.org/ws/2004/09/transfer"))
            || !_tcscmp(Xns->QName,
                     _T("http://schemas.xmlsoap.org/ws/2004/09/enumeration")))
            objname = _T("COB/wsman.cob");
        sc = BindToObject(CurrentNameSpace(), objname, NAME_SPACE_READ,
                          iid, (void**) &Object);
#endif
    }
    if (FAILED(sc)) {
        DBGT((_T("appstack.c: server couldn't bind to %s sc=%x\n"),
              objname ? objname : _T("<this>"), sc));
        return sc;
    }

    cont->metd = metd;

    /* Start from QI to account for possible g++ fillers */
#ifdef _CLR
    if (!cont->isCLR) 
        cont->Method = ((void**)&Object->v->QueryInterface)[metd->_Ordinal];
#else
    cont->Method = ((void**)&Object->v->QueryInterface)[metd->_Ordinal];
#endif
    assert(!cont->pThis);
    cont->pThis = Object;
    *(PTR*)cont->Frame = Object;

    DBG(("cont.c: ContSetMethod bound to object %x [%s] method %x [%s]\n",
         (ADDRESS) Object, objname ? objname : _T("<this>"),
         (UINT) cont->Method, metd->name));

    return S_OK;
}

PRIVATE void FixWsaHeadersForReply(PCONTINFO This)
{
    PWSCONTEXT ctx = This->pContext;
    TIME t = CurrentTime();
    INT tt = Int64ToInt32(t);

    if (!ctx)
        return;

    /* These should not be sent back in the reply message */
    ctx->MaxEnvelopeSizeValue = 0;
    ctx->OperationTimeout = TIME_ORIGIN;
    ctx->SelectorSet = NULL;
    ctx->ResourceURI = NULL;

    /* Fix MessageId and RelatesTo for reply */
    if (ctx->MessageId) {
        assert(!ctx->RelatesTo);
        assert(_tcslen(ctx->MessageId) > 10);
        //assert(_istxdigit(ctx->MessageId[9]));
        ctx->RelatesTo = ctx->MessageId;
        ctx->MessageId = _tcsdup(ctx->RelatesTo);
        /* Modify MessageId: XXX not unique */
        ctx->MessageId[9]++;
        if (!_istxdigit(ctx->MessageId[9]))
            ctx->MessageId[9] = _T('0');
    }
}

PRIVATE SCODE ContSendMessage (
    PICONTINUATION iThis,
    PIHTTPCONNECTOR pConn,
    PCode FaultCode,
    const _TCHAR *FaultText)
{
    METHOD_DECLARE(Cont, SendMessage);
    PISOAPSERIALIZER ser = NULL;
    PIFILE file = NULL;
    BOOL isResponse = !(This->Flags & CONT_FLAGS_CLIENT);
    UINT HttpStatus = TOKEN_OK;
    BOOL noreply = FALSE;
    PIHTTPCONNECTOR Request = This->Request;
    PICONVERT convert = NULL;

    if (This->metd->ResponseNameToken == TOKEN_invalid)
        noreply = TRUE;

    if (!isResponse || (!noreply && Request)) {
        sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                          NAME_SPACE_READ, &IID_IConvert,
                          (void**) &convert);
        CHECKSC("bind to soap.cob");
    }

    /* Serialize */
    if (isResponse) {
        sc = S_OK;
        if (This->metd->any.Flags & ENTRY_FLAGS_METHOD_SCODE) {
            sc = (SCODE) This->RetVal;
            if (sc == S_NO_RESPONSE)
                noreply = TRUE;
        }
        /* If this is a multi-continuation, the second round will see
         * a NULL Request. I.e. a synchronous message becomes async after
         * one response.
         */
        if (!noreply && Request) {
            ser = This->Serializer;
            if (!ser) {
                SCODE sc2 = sc; /* save value from app during creation... */
                sc = Request->v->AllocateBuffer(Request, (PIBUFFER*) &file);
                CHECKSC("AllocateBuffer");

                sc = convert->v->CreateSerializer(convert,
                                                  file, This->Flags, &ser);
                CHECKSC("CreateSerializer");
                RELEASE(file);
                This->Serializer = ser;
                sc = sc2;
            }
            FixWsaHeadersForReply(This);
            if (SUCCEEDED(sc) && !FaultCode)
                sc = ser->v->MethodMessage(ser, iThis, isResponse);
            if (FAILED(sc) || FaultCode) {
                sc = ser->v->FaultMessage(ser, iThis, sc, FaultCode,FaultText);
                HttpStatus = TOKEN_SOAP_PARSING_ERROR;
            }

            sc = ser->v->GetContent(ser, &file);
            ser = NULL;         /* caller deletes */
            CHECKSC("Serializer->GetContent");
            Request->v->SendMessage(Request, HttpStatus,
                                    NULL, /* SoapAction */
                                    file,
                                    NULL);

            if (This->SoapAction) {
                free(This->SoapAction);
                This->SoapAction = NULL;
            }
        }
    } else {
sc = pConn->v->AllocateBuffer(pConn, (PIBUFFER*) &file);
        CHECKSC("AllocateBuffer");

        /* XXX Verify that Flags are properly calculated for this call
           XXX pContext->To->Address --> PARSE_URL_FLAGS_XTE */
        sc = convert->v->CreateSerializer(convert, file, This->Flags, &ser);
        CHECKSC("CreateSerializer");
        This->Serializer = ser;

        sc = ser->v->MethodMessage(ser, iThis, isResponse);
        CHECKSC("MethodMessage");

#if USE_SERVER_FOR_RESPONSE
        if (!noreply) {
            Mutex_Lock(&This->mx);  /* not necessary but seems simpler w/ */
            assert(This->State == CONT_INIT);
            This->State = CONT_RPC;
            sc = pConn->v->Listen(pConn, NULL);
            Mutex_Unlock(&This->mx);
            CHECKSC("Listen for response");
        }
#endif

        sc = pConn->v->SendMessage(pConn,
                                   TOKEN_POST,
                 This->SoapAction ? This->SoapAction : This->methodNS->QName,
                                   file,
                                   noreply ? NULL : iThis);
    }
    CHECKSC("Connector.SendMessage");

METHOD_CLEANUP;
    RELEASE(file);
    RELEASE(ser);
    RELEASE(convert);

    return sc;
}

PRIVATE SCODE ContExecute (
    PICONTINUATION iThis,
    BOOL sync,
    void *pRes)
{
    PCONTINFO cont = pCONT(iThis);
    SCODE sc;
    PITHREAD pthd;
    TIME_CONSTRAINT _cval, *tc = NULL;
    struct reservation _res, *res = pRes;
    UINT RegArgCount = 4;   /* XXX ARM */
#ifdef _CLR
    PICLR pCLR;
#endif
#if defined(h8) && defined(__GNUC__)
    RegArgCount = H8CalcRegArgCount(cont->metd);
#endif

    if (!sync && !res) {
        PIWSRESERVATION zz;
        const _TCHAR *resname = cont->resname;

        if (!resname && cont->pContext) {
            PWSCONTEXT pCtx = cont->pContext;
            resname = pCtx->ReservationName;
            if (!resname)
                resname = pCtx->taskName;
        }
        if (resname) {
            sc = BindToObject(CurrentNameSpace(), _T("COB/wsres.cob"),
                              NAME_SPACE_READ, &IID_IWsReservation,
                              (void**) &zz);
            assert(SUCCEEDED(sc));
            sc = zz->v->Match(zz, iCONT(cont), resname, &_res);
            if (sc == E_OBJECT_NOT_INITIALIZED) {
                DBG(("ContExecute reservation '%s' not triggered\n", resname));
                return sc;      /* ContExecute will get called again */
            } else if (sc == E_ALREADY_EXISTS) {
                DBG(("ContExecute res '%s' already has continuation %s\n",
                     resname, "dropping new message"));
                return sc;
            } else if (FAILED(sc)) {
                /* XXX should we fail or execute? FAILURE-MODE? */
                DBG(("ContExecute could not find reservation '%s' -> %s\n",
                     resname, "running as time sharing thread\n"));
            } else {
                res = &_res;
            }
        }
    }

#ifdef _CLR
    if (cont->isCLR) {
        pCLR = (PICLR) cont->pThis;
        sc = pCLR->v->InvokeMethod2((PICLR) cont->pThis, cont);
        cont->v->AddRef(iCONT(cont));  /* FinishSoapActivation consumes ref */
        FinishSoapActivation(cont);
        return S_NO_RESPONSE;
    }
#endif

    /* Create thread only once and recycle it later */
    if (!cont->Thread) {
        PIPROCESS prc = CurrentProcess();
        CPMETHOD_DESCR metd = cont->metd;
        UINT stack_size = metd->_stacksize + 0x1000;
        PTR stack;

        if (metd->_stacksize == 0)
            stack_size += 0x1000;

        sc = prc->v->CreateThread(prc, 0,
                                  (THREAD_FUNCTION) NULL, NULL,
                                  stack_size+16 + sizeof(UINT),
                                  &stack, &pthd);
        if (FAILED(sc)) {
            DBG(("ContExecute: Could not create sc=%x\n", sc));
            return sc;
        }

        /* What we got back was a pointer to the current stack.
         * Back-calculate the start. There is an annoying adjustment in XXX
         * thread.c so thus we need the + sizeof(UINT). XXX
         * It is needed by the stack overflow checking code. XXX
         */

        CurrentHeap()->v->Validate(CurrentHeap(), 0, 0);
        cont->Thread = pthd;
        pthd->v->SetValue(pthd, THREAD_VALUE_RPC_CONTINUATION, (ADDRESS) cont);
        cont->StackSize = stack_size;
        /* Initialize the stack */
        cont->StackBase = (PBYTE)stack - stack_size - 16 + sizeof(UINT);

        /* Base points *past* the last byte (-16 is arm reg save area) */
        cont->StackCurrent = cont->StackBase + cont->StackSize - 16;

#if defined(STACK_STATS)
        InitAppStackStats(cont);
#endif
        DBG(("CONT: x%08x StackBase=x%08x...x%08x\n",
             (UINT) cont, (UINT) cont->StackBase, (UINT) cont->StackCurrent));

#if defined(CHECK_STACK_OVERFLOW)
        /* SaveCount is one based. Put signature into first slot. */
        *((PUINT32) cont->StackCurrent) = STACK_OVERFLOW_SIGNATURE;
#endif

        cont->StackCurrent -= metd->_FrameSize;
        if (metd->_FrameSize < 16)
            cont->StackCurrent -= 16 - metd->_FrameSize;
    }

    if (cont->Frame != cont->StackCurrent) {
        UINT size = cont->metd->_FrameSize;
#if 1 /* reinit some stack state */
        /* Base points *past* the last byte (-16 is arm reg save area) */
        cont->StackCurrent = cont->StackBase + cont->StackSize - 16;
        cont->StackCurrent -= size;
        if (size < 16)
            cont->StackCurrent -= 16 - size;
#endif
        DBG(("copying frame %x -> %x size=%d\n",
             (ADDRESS) cont->Frame, (ADDRESS) cont->StackCurrent, size));
        memcpy(cont->StackCurrent, cont->Frame, size);
    }

    if (sync) {
        sc = strump((ADDRESS) cont->StackCurrent, (ADDRESS) cont->Method,
#if defined(h8) && defined(__RENESAS__)
                    (UINT)cont->metd
#else
                    RegArgCount
#endif
                    );
        RELEASE_NULL(cont->pThis);
        return sc;
    }

    if (res) {
        tc = &_cval;
        tc->Start = res->Start;
        /* cb_sched wants estimate in positive units */
        tc->Estimate = Int64Subtract(TIME_ORIGIN, res->duration);
        tc->Criticality = _NONCRITICAL;
        tc->TimeTaken = NULL;  /* XXX Point to TBA PCONTINFO field */
    }

    DBG(("ContExecute(%x [%s] seqno=%d)\n",
         (ADDRESS) cont, cont->metd->name, res?res->seqno:0));

    pthd = cont->Thread;
    if (pthd) {
        PBYTE stack = cont->StackCurrent;
        /* Make sure the thread doesn't get recycled after it is started */
        cont->Thread = NULL;
        cont->StackCurrent = NULL;

        cont->v->AddRef(iCONT(cont));  /* SoapActivation uses ref */

        sc = pthd->v->StartWithConstraint(pthd, tc,
                                          (THREAD_FUNCTION) cont->Method,
                                          (THREAD_ARGUMENT) _SoapActivation,
                                          stack,
#if defined(h8) && defined(__RENESAS__)
                                          (UINT)cont->metd
#else
                                          RegArgCount
#endif
                                          );
        RELEASE(pthd);
        assert(SUCCEEDED(sc));
        if (FAILED(sc)) {
            cont->v->Release(iCONT(cont));  /* undo above AddRef */
            return sc;
        }

    } else {
        cont->RetVal = strump((ADDRESS) cont->StackCurrent,
                              (ADDRESS) cont->Method,
#if defined(h8) && defined(__RENESAS__)
                              (UINT)cont->metd
#else
                              RegArgCount
#endif
                              );
        FinishSoapActivation(cont);
    }

    return S_NO_RESPONSE;
}

#include "s_CContinuation_v.c"

SCODE InitContinuation(PCONTINFO cont, CONT_FLAGS Flags, const _TCHAR *EnvName)
{
    METHOD_DECLARE_NOTHIS(cont.c, Continuation);
    PICONVERT convert = NULL;

    sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                      NAME_SPACE_READ, &IID_IConvert,
                      (void**) &convert);
    CHECKSC("bind to convert.cob");

    memset(cont, 0, sizeof(*cont));
    cont->RetVal = S_OK;
    sc = convert->v->GetNamespace(convert, _T("soapenc"),7, &cont->encodingNS);
    CHECKSC("GetNamespace soapenc");

    if (EnvName) {
        sc = convert->v->GetNamespace(convert, EnvName, 0, &cont->envelopeNS);
        CHECKSC("GetNamespace");
    }

    cont->v = &CContinuationVtbl;
    cont->RefCnt = 1;
    ListNodeInit(contloff, cont);
    cont->Flags = Flags;

    Condition_Init(&cont->cd);
    Mutex_Init(&cont->mx);
    cont->State = CONT_INIT;

    /* Save start time for message delay calculation */
    cont->BeginTime = CurrentTime();

METHOD_CLEANUP;
    RELEASE(convert);
    return sc;
}

SCODE CreateContinuation(CONT_FLAGS Flags, PICONTINUATION *pCont)
{
    METHOD_DECLARE_NOTHIS(_, CreateContinuation);
    PIHEAP Heap = CurrentHeap();
    PCONTINFO cont = NULL ;
    UINT size = sizeof(struct CONTINFO);

#if USE_TMP_HEAP
#else
    if ((Flags & CONT_FLAGS_CLIENT) == 0)
        size += 4096;;          /* XXX see below */
#endif
    cont = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("cont"), size, 0);
    CHECKSETSC(cont, E_NOT_ENOUGH_MEMORY, "Alloc cont");

    sc = InitContinuation(cont, Flags, _T("soap"));
    CHECKSC("InitContinuation");

    if ((Flags & CONT_FLAGS_CLIENT) == 0) {
#if USE_TMP_HEAP
        PIHEAPFACTORY HeapFactory;
        sc = BindToObject(CurrentNameSpace(), _T("COB/tmp_heap.cob"),
                          NAME_SPACE_READ, &IID_IHeapFactory,
                          (void **) &HeapFactory);
        CHECKSC("bind tmp_heap.cob");

        sc = HeapFactory->v->CreateNested(HeapFactory, Heap,
                                          0, 0, 4096, &cont->TmpHeap);
        RELEASE(HeapFactory);
        CHECKSC("create tmp heap");
#else
        /* Allocate the temporary heap. 
         * XXX It is not necessary to allocate so much heap space. 
         * XXX A proper temporary heap would be better 
         * XXX (and could grow on-demand).
         * XXX BUGBUG Make sure the ArgSpace is deallocated!
         * NOTE: The ArgSpace is now deallocated only at the end of the message
         */
        cont->ArgSpace = (PTR) (cont + 1);
        cont->ArgSpaceAvail = size - sizeof(*cont);
#endif
    }

METHOD_CLEANUP;
    if (FAILED(sc) && cont) {
        Heap->v->Free(Heap, 0, cont);
    }
    *pCont = iCONT(cont);
    return sc;
}
