/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* This file generates arm proxy call code at runtime. --jvh 2001
 */
#include <mmlite.h>
#include <web/proxy.h>
#include <assert.h>
#include <stdio.h>

#define DPRINT(x)
// printf x

extern void ProxyCommonTemplate(void);
extern void ProxyCommonTemplateEnd(void);
extern void ProxyTemplate(void);
extern void ProxyTemplateEnd(void);

/* XXX it would be more resilient to add labels into assy */
#define DESCR_OFF 8
#define BRANCH_OFF 4
#define INTERP_OFF 36
#define SLACK 1000

/* Save latest copy of ProxyCommonTemplate */
static ADDRESS ProxyCommonRAM = 0;

UINT ProxyCodeInit(PUINT8 Where, CPMETHOD_DESCR pDescr,
         INT32 (__fastcall *Interpreter)(CPMETHOD_DESCR pDescr, PUINT32 pArgs))
{
    UINT ProxySize = (ADDRESS) ProxyTemplateEnd - (ADDRESS) ProxyTemplate;
    UINT CommonSize = (ADDRESS) ProxyCommonTemplateEnd - (ADDRESS) ProxyCommonTemplate;
    INT Diff;
    BOOL NeedsRAMCopy = FALSE;
    ADDRESS Common = (ADDRESS) ProxyCommonTemplate;
    ADDRESS Target;

    /* Figure out which ProxyCommon to use (some slack for code size).
     * 26bit offset, if too far, make copy.
     * Also if compiled template has wrong interpreter (but assume callers
     * don't change their minds all the time).
     */
    Diff = (INT)ProxyCommonTemplate - (INT)Where;
    if (((Diff<0) ? (SLACK - Diff > 0x04000000) : (Diff + SLACK > 0x04000000))
        || *(PUINT32)&((PUINT8)ProxyCommonTemplate)[INTERP_OFF] != (UINT32) Interpreter ) {
        /* Needs to use RAM version */
        Common = ProxyCommonRAM;
        if (ProxyCommonRAM == 0) {
            NeedsRAMCopy = TRUE;
        } else {
            Diff = (INT)ProxyCommonRAM - (INT)Where;
            if (((Diff < 0) ? (SLACK - Diff > 0x04000000) : (Diff + SLACK > 0x04000000))
                || *(PUINT32)&((PUINT8)ProxyCommonRAM)[INTERP_OFF] != (UINT32) Interpreter ) {
                NeedsRAMCopy = TRUE;
                /* XXX the size is overestimated on round1 as ProxyCommonRAM is
                 * XXX not updated */
            }
        }
    }

    DPRINT(("ProxyCodeInit templ=%x where=%x desc=%x interp=%x RAMCopy=%d\r\n", (UINT) ProxyTemplate, (UINT) Where, (UINT) pDescr, (UINT) Interpreter, NeedsRAMCopy));

    if (Where == NULL) {
        /* Just return approximate code size for allocation purposes */
        if (NeedsRAMCopy)
            return ProxySize + CommonSize;
        else
            return ProxySize;
    }

    if (NeedsRAMCopy) {
        memcpy((void*) Where, (void*) ProxyCommonTemplate, CommonSize);
        *(PUINT32)&Where[INTERP_OFF] = (UINT32) Interpreter;
        ProxyCommonRAM = (ADDRESS) Where;
        Common = (ADDRESS) Where;
        Where += CommonSize;
    }

    memcpy((void*) Where, (void*) ProxyTemplate, ProxySize);
    *(PUINT32)&Where[DESCR_OFF] = (UINT) pDescr;

    /* Adjust branch offset
     * Branches are expressed as word offsets.
     * pc points 8 ahead of branch instruction.
     */
    Target = ((INT)Common - (INT)&Where[BRANCH_OFF] - 8) >> 2;
    DPRINT(("OFFSET=%08x (orig=%08x) INTERP=%08x delta=%08x", *(PUINT32)&Where[BRANCH_OFF], (INT) Interpreter - (INT) ProxyTemplate, *(PUINT32)&Where[INTERP_OFF], Target));
    *(PUINT32)&Where[BRANCH_OFF] = (Target & 0x00ffffff) | 0xea000000; /* b Target */ 
DPRINT(("==> %08x (%08x)\r\n", *(PUINT32)&Where[BRANCH_OFF], *(PUINT32)&Where[INTERP_OFF]));

 if (NeedsRAMCopy)
     return ProxySize + CommonSize;
 else
     return ProxySize;
}
