/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* This file generates arm proxy call code at runtime. --jvh 2001
 */
/*
   H8,GNUC version -- takeshim 2005 
                      actual target device is AKI-H8-LAN 3052F.

   H8S,RENESAS HEW version -- takeshim 2006
                              actual target device is H8S 2068F OPMA mCard.

 */
#include <mmlite.h>
#include "proxy.h"

#define DPRINT(x)

#ifdef __GNUC__

#define MAX_REGARG_COUNT 3

const UINT8 ProxyTemplate[] = {
    0x7A, 0x03, 00, 00, 00, 00,   // mov immediate(AMethodDescriptor), er3
    0x5A, 00, 00, 00              // jmp ProxyCommonTemplate
};

extern void ProxyCommonTemplate1(void);
extern void ProxyCommonTemplate2(void);
extern void ProxyCommonTemplate3(void);

extern void HeaderProxyCommonTemplate1(void);
extern void HeaderProxyCommonTemplate2(void);
extern void HeaderProxyCommonTemplate3(void);

#else

// __RENESAS__

#define MAX_REGARG_COUNT 2

const UINT8 ProxyTemplate[] = {
    0x01, 00, 0x6D, 0xF0,         // push er0
    0x7A, 0x00, 00, 00, 00, 00,   // mov immediate(AMethodDescriptor), er0
    0x5A, 00, 00, 00              // jmp ProxyCommonTemplate
};

extern void ProxyCommonTemplate(void);
extern void HeaderProxyCommonTemplate(void);

#endif

extern INT32 __fastcall ProxyInterpreter(CPMETHOD_DESCR metd, PUINT32 pStack);
extern INT32 __fastcall HeaderProxyInterpreter(CPMETHOD_DESCR metd, PUINT32 pStack);
  
UINT32 H8CalcRegArgCount(CPMETHOD_DESCR pDescr);

UINT ProxyCodeInit(PUINT8 Where, CPMETHOD_DESCR pDescr,
         INT32 (__fastcall *Interpreter)(CPMETHOD_DESCR pDescr, PUINT32 pArgs))
{
	UINT32 RegArgCount;
	void (*ProxyTemplateAddr)(void);
	int i;

	if (Interpreter != ProxyInterpreter && Interpreter != HeaderProxyInterpreter) {
		return 0;
	}

	if (Where == NULL) {
		return sizeof(ProxyTemplate);
	}

#ifdef __GNUC__
	RegArgCount = H8CalcRegArgCount(pDescr);

	memcpy(Where, ProxyTemplate, sizeof(ProxyTemplate));
	
    *(PUINT32)&Where[2] = (UINT32)pDescr;
	
	if (Interpreter == ProxyInterpreter) {
		if (RegArgCount == 1) {
			ProxyTemplateAddr = ProxyCommonTemplate1;
		} else if (RegArgCount == 2) {
			ProxyTemplateAddr = ProxyCommonTemplate2;
		} else {
			ProxyTemplateAddr = ProxyCommonTemplate3;
		}
	} else {
		if (RegArgCount == 1) {
			ProxyTemplateAddr = HeaderProxyCommonTemplate1;
		} else if (RegArgCount == 2) {
			ProxyTemplateAddr = HeaderProxyCommonTemplate2;
		} else {
			ProxyTemplateAddr = HeaderProxyCommonTemplate3;
		}
	}
	
    *(PUINT32)&Where[6] |= (UINT32)ProxyTemplateAddr;
	return sizeof(ProxyTemplate);
#else
	// __RENESAS__
	memcpy(Where, ProxyTemplate, sizeof(ProxyTemplate));
	
    *(PUINT32)&Where[6] = (UINT32)pDescr;
	
	if (Interpreter == ProxyInterpreter) {
		ProxyTemplateAddr = ProxyCommonTemplate;
	} else {
		ProxyTemplateAddr = HeaderProxyCommonTemplate;
	}
	
    *(PUINT32)&Where[10] |= (UINT32)ProxyTemplateAddr;
	return sizeof(ProxyTemplate);
#endif
}

#ifdef __GNUC__
/* 
    Calcuate number of arguments to be passed in register.
    At most, ER0, ER1, ER2 would be used for arguments passing.

    NOTE:
    This is h8300-hms-gcc compiler specific.
    Renesus compilter has different convention.
*/
UINT32 H8CalcRegArgCount(CPMETHOD_DESCR pDescr)
{
	UINT32 RegArgCount = 1; /* Initial Value is 1 for "THIS" Pointer */

	CPMETHOD_DESCR pArgTable = pDescr->ptable;
	CPMETHOD_DESCR pArgTableEnd = pArgTable + pDescr->table_size;
	
	for(; pArgTable < pArgTableEnd; pArgTable++) {
		if (pArgTable->_Flags & ARGUMENT_DESCR_FLAGS_RETURN) {
			// ignoring return value
			continue;
		}

		if ((pArgTable->_Flags & ARGUMENT_DESCR_FLAGS_BYREF) ||
			pArgTable->descr_type == ARG_INT ||
			pArgTable->descr_type == ARG_CONTINUATION) {
			if (++RegArgCount >= MAX_REGARG_COUNT) {
				// exit out from the loop.
				break;
			}
		} else {
			// exit out from the loop.
			break;
		}
	}

	return RegArgCount;
}
#endif

#ifdef __RENESAS__
//
// New Code For Renesas HEW Compiler
//

extern UINT32 strump_asm(ADDRESS newsp, ADDRESS newpc, UINT32 regarg0, UINT32 regarg1);
extern UINT32 H8SetRegArgs(ADDRESS sp, UINT32 regargs[], CPMETHOD_DESCR pDescr);	// in base/md/h8/_thread.c

UINT32 strump(ADDRESS newsp, ADDRESS newpc, CPMETHOD_DESCR desc)
{
	UINT32 regargs[MAX_REGARG_COUNT];

	newsp = H8SetRegArgs(newsp, regargs, desc);

	return strump_asm(newsp, newpc, regargs[0], regargs[1]);
}

//
// Called by ProxyTemplate
//
// RegArg -> Stack
//
// NOTE: Assuming that interger arguments are 32-bit.
//
UINT32 H8SetRegArgs2(CPMETHOD_DESCR pDescr, ADDRESS sp)
{
	CPMETHOD_DESCR pArgTable = pDescr->ptable;
	CPMETHOD_DESCR pArgTableEnd = pArgTable + pDescr->table_size;
	UINT32 regarg1, regarg2;
	int RegArgCount;
	
	regarg1 = ((UINT32 *)sp)[1];
	regarg2 = ((UINT32 *)sp)[0];
	
	//
	// swap retAddress and arg1 (This pointer)
	//
	((UINT32 *)sp)[1] = ((UINT32 *)sp)[2];
	((UINT32 *)sp)[2] = regarg1;
	RegArgCount = 1;
	sp += sizeof(UINT32);
	
	//
	// look for 2nd regarg
	//
	for(; pArgTable < pArgTableEnd && RegArgCount < MAX_REGARG_COUNT; pArgTable++) {

		if (pArgTable->_Flags & ARGUMENT_DESCR_FLAGS_RETURN) {
			// ignoring return value
			continue;
		}

		if ((pArgTable->_Flags & ARGUMENT_DESCR_FLAGS_BYREF) ||
			pArgTable->descr_type == ARG_INT ||
			pArgTable->descr_type == ARG_CONTINUATION) {
			
			// found 2nd RegArg
			RegArgCount++;
			sp -= sizeof(UINT32);
			
			((UINT32 *)sp)[0] = ((UINT32 *)sp)[1];	// retAddress
			((UINT32 *)sp)[1] = ((UINT32 *)sp)[2];	// arg1
			
			if (pArgTable->FrameOffset == sizeof(UINT32)) {
				((UINT32 *)sp)[2] = regarg2;	// arg2
			} else {
				// TODO
			}
		}
	}
	
	// Adjusted New SP
	return (UINT32)sp;
}

#endif
