; Copyright (c) Microsoft Corporation. All rights reserved.
;
; H8 Version for Renesas Assembler
;
    .SECTION    P,CODE,ALIGN=2
    
    .IMPORT     _ProxyInterpreter
    .IMPORT     _HeaderProxyInterpreter
    
;;;;;; strump
    .EXPORT     _strump_asm

;;;;;; more HEW proxy template
    .EXPORT     _ProxyCommonTemplate
    .EXPORT     _HeaderProxyCommonTemplate
    .IMPORT     _H8SetRegArgs2

;
; GNU Compiler use ER0, ER1 and ER2 for register arguments.
; ER5 and ER6 must be saved.
;

;
; Renesas Compiler uses ER0 and ER1 for register arguments.
; ER2...ER6 must be saved.
;

;
; UINT32 strump_asm(ADDRESS newsp, ADDRESS newpc, UINT32 regarg0, UINT32 regarg1);
;
_strump_asm:
    push er6
    push er5
    mov.l sp, er6

    ;
    ; switch to the new stack
    ;
    mov.l er0, sp

    ;
    ; avoid clobbering the pc argument
    ;
    mov.l er1, er5  ;; newpc
    
    ;
    ; copy reg args to er0, er1
    ;
    mov.l @(20,er6), er0
    mov.l @(16,er6), er1

    ;
    ; execute RPC
    ;
    jsr @er5

    ;
    ; RPC return value now in er0
    ;

    ;
    ; switch to the old stack
    ;
    mov.l er6, sp
    pop er5
    pop er6
    rts


;
;  _ProxyCommonTemplate
;
;  When Called:
;
;    ER0 -- Method Descriptor
;    ER1 -- 2nd regarg
;
;             +-----------+
;    SP ----> | Saved ER0 | 
;             +-----------+
;             |     PC    | 
;             +-----------| 
;             |    args   | 
;             |-----------|
;             |    ....   | 
;
;
;  Before Calling _H8SetRegArgs2:
;
;    ER0 -- Method Descriptor
;    ER1 -- SP pointing to Saved ER1
;
;             +-----------+
;    SP ----> | Saved MD  | 
;             +-----------+
;             | Saved PC  | 
;             +-----------+
;    ER1 ---> | Saved ER1 | 
;             +-----------+
;             | Saved ER0 | 
;             +-----------+
;             |     PC    | 
;             +-----------| 
;             |    args   | 
;             |-----------|
;             |    ....   | 
;
;
;  Before Jumping to Proxy Interpreter:
;
;    ER0 -- Method Descriptor
;    ER1 -- SP
;
;    [if regarg count is 1]
;
;             +-----------+
;    SP ----> | Saved PC  |
;             +-----------+
;             | not used  |
;             +-----------+
;    ER1 ---> |     PC    |
;             +-----------+
;             | 1st regarg| ("This" pointer)
;             +-----------+
;             |    args   |
;             +-----------+
;             |    ....   |
;
;
;    [if regarg count is 2]

;             +-----------+
;    SP ----> | Saved PC  |
;             +-----------+
;    ER1 ---> |     PC    |
;             +-----------+
;             | 1st regarg| ("This" pointer)
;             +-----------+
;             | 2nd regarg|    2nd regarg could be in middle of stack args.
;             |    args   |
;             +-----------+
;             |    ....   |
;
_ProxyCommonTemplate:
    push er1				;; save er1
    mov.l @(8,sp),er1
    push er1
    mov.l sp, er1			;; arg2 of H8SetRegArg2 is sp
    add.l #4,er1
    push er0				;; save MD
    jsr @_H8SetRegArgs2

    mov.l er0, er1    		;; set new sp to arg2 of ProxyInterpreter
    pop er0					;; restore MD for arg1
    
    jsr   @_ProxyInterpreter
    pop   er1             ;; pop PC
    add.l #12,sp          ;; remove arg1,arg2,PC
    push  er1             ;; push PC
    rts

;
; Same as ProxyCommonTemplate but calls HeaderProxyInterpreter
;
_HeaderProxyCommonTemplate:
    push er1				;; save er1
    mov.l @(8,sp),er1
    push er1
    mov.l sp, er1			;; arg2 of H8SetRegArg2 is sp
    add.l #4,er1
    push er0				;; save MD
    jsr @_H8SetRegArgs2

    mov.l er0, er1    		;; set new sp to arg2 of ProxyInterpreter
    pop er0					;; restore MD for arg1
    
    jsr   @_HeaderProxyInterpreter
    pop   er1             ;; pop PC
    add.l #12,sp          ;; remove arg1,arg2,PC
    push  er1             ;; push PC
    rts

    .END
