/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include "proxy.h"

#define DPRINT(x)

/* This file generates i386 proxy call code at runtime. --jvh 11/98
 * It uses one of a number of possible templates and fills in the missing
 * pieces. The generated code then calls a generic ProxyInterpreter.
 * This implementation generates compact code that delegates to generic code.
 * It would be feasible to generate custom code to marshal stuff according
 * to the method descriptor but I doubt that would win too much as long as
 * the generic interpreter performs ok.
 * We use __fastcall for the generic code in order to simplify stack management
 * and avoid unnecessary work in pushing and popping.
 */
UINT8 CdeclProxyTemplate[12] = {
    0xb9, 00,00,00,00,      /* mov ecx, immediate(AMethodDescriptor) */
    0x8b,0xd4,              /* mov edx,esp */
    0xe9, 00,00,00,00       /* jmp ProxyInterpreter */
};

UINT8 CdeclNearProxyTemplate[9] = {
    0xb9, 00,00,00,00,      /* mov ecx, immediate(AMethodDescriptor) */
    0x8b,0xd4,              /* mov edx,esp */
    0xeb, 00                /* jmp NEAR ProxyInterpreter */
};

UINT8 StdProxyTemplate[16] = {
    0xb9, 00,00,00,00,      /* mov ecx, imm(AMethodDescriptor) */
    0x8b,0xd4,              /* mov edx,esp */
    0xe8, 00,00,00,00,      /* call ProxyInterpreter */
    0xc2, 04,00,            /* ret 4 */
    0xcc                    /* pad with breakpoint to keep alignment */
};

UINT ProxyCodeInit(PUINT8 Where, CPMETHOD_DESCR pDescr,
         INT32 (__fastcall *Interpreter)(CPMETHOD_DESCR pDescr, PUINT32 pArgs))
{
    if (Where == NULL) {
        /* Just return approximate code size for allocation purposes */
        if (pDescr->_StupidCallAdjust)
            return 16;
        else
            return 12;
    }
    /* Decide which code to use:
     * Cdecl near is most efficient and compact but is in practice not
     * applicable because the offset is just 8 bits.
     * stdcall requires a stack adjustment in the end so we have to return
     * through an adjuster thunk (3 bytes of code).  We accomplish this with
     * a call where cdecl gets by with a jmp (cdecl returns straight to app)
     */
    if (pDescr->_StupidCallAdjust == 0) {
        INT Diff = (INT) Interpreter - (INT) &Where[9];
        if (Diff < 128 && Diff > -128) {
            DPRINT(("ProxyCodeInit Cdecl NEAR %x %x %x\n",
                    Where, pDescr, Interpreter));
            memcpy(Where,
                   CdeclNearProxyTemplate,
                   sizeof(CdeclNearProxyTemplate));
            *(PUINT32)&Where[1] = (UINT) pDescr;
            *(PUINT8)&Where[8] = (UINT8) Diff;
            return sizeof(CdeclNearProxyTemplate);
        } else {
            DPRINT(("ProxyCodeInit Cdecl %x %x %x\n",
                    Where, pDescr, Interpreter));
            memcpy(Where, CdeclProxyTemplate, sizeof(CdeclProxyTemplate));
            *(PUINT32)&Where[1] = (UINT) pDescr;
            *(PUINT32)&Where[8] = (UINT) Interpreter - (UINT) &Where[12];
            return sizeof(CdeclProxyTemplate);
        }
    } else {
        DPRINT(("ProxyCodeInit StdCall %x %x %x\n",
                Where, pDescr, Interpreter));
        memcpy(Where, StdProxyTemplate, sizeof(StdProxyTemplate));
        *(PUINT32)&Where[1] = (UINT32) pDescr;
        *(PUINT32)&Where[8] = (UINT32) Interpreter - (UINT) &Where[12];
        *(PUINT16)&Where[13] = (UINT16) pDescr->_StupidCallAdjust;
        return sizeof(StdProxyTemplate);
    }
}
