/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Automatic proxy generation --jvh */
//#define DBGLVL 0

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <netif.h>
#include <httpif.h>
#include <web/proxy.h>
#include <soap.h>

PRIVATE SCODE GetConnector(const _TCHAR *Url, CPXMLNAMESPACE SoapAction,
                           PIHTTPCONNECTOR *pps)
{
    METHOD_DECLARE_NOTHIS(proxy.c, GetConnector);
    PIWEBPROTOCOLFACTORY factory;

    sc = BindToObject(CurrentNameSpace(), _T("COB/http.cob"), NAME_SPACE_READ,
                      &IID_IWebProtocolFactory, (void **) &factory);
    CHECKSC("bind to COB/http.cob");

    sc = factory->v->CreateConnector(factory, Url, PARSE_URL_FLAGS_SOAP, pps);
    CHECKSC("CreateConnector");

METHOD_CLEANUP;
    return sc;
}

PRIVATE void DestroyProxy(PPROXY pxy);

PRIVATE SCODE ProxyQueryInterface(PPROXY pxy, REFIID IidQuery, void** Object)
{
    CPINTERFACE_DESCR ifd;

    TRACE("ProxyQueryInterface");

    for (ifd = pxy->NameSpace->ifd; ifd; ifd = ifd->iface_extras->SuperClass)
        if (UuidCmp(IidQuery, ifd->iface_extras->iid)) {
            AtomicInc(&pxy->RefCount);
            *Object = pxy;
            return S_OK;
        }

    if (UuidCmp(IidQuery, &IID_IHttpConnector)) {
        pxy->pConnector->v->AddRef(pxy->pConnector);
        *Object = pxy->pConnector;
        return S_OK;
    }

    return GenericQueryInterface((PIUNKNOWN) pxy, 
                                 IidQuery, 
                                 Object, 
                                 &IID_IUnknown);
}

PRIVATE UINT ProxyAddRef(PPROXY pxy)
{
    UINT Refs = AtomicInc(&pxy->RefCount);
    DBG(("ProxyAddRef refs=%d\n", Refs));

    return Refs;
}

PRIVATE UINT ProxyRelease(PPROXY pxy)
{
    UINT Refs = AtomicDec(&pxy->RefCount);
    DBG(("ProxyRelease refs=%d\n", Refs));

    if (Refs > 0) return Refs;

    DestroyProxy(pxy);
    return 0;
}


/* Common code */
PRIVATE SCODE ProxyStart(PCONTINFO pcont,
                         CPXMLNAMESPACE Xns,
                         CPMETHOD_DESCR metd,
                         PIHTTPCONNECTOR pConn,
                         PWSCONTEXT pContext,
                         _TCHAR *Uri,
                         CONT_FLAGS Flags,
                         const void *Frame)
{
    METHOD_DECLARE_NOTHIS(Proxy, Start);

    PISOAPSERIALIZER ser = NULL;
    BOOL async = FALSE;
    PIFILE file = NULL;

    if (metd->ResponseNameToken == TOKEN_invalid)
        async = TRUE;

    sc = InitContinuation(pcont, CONT_FLAGS_CLIENT | Flags,
                          (Flags & CONT_FLAGS_SOAP_1_1) ? _T("soap11") : _T("soap"));
    CHECKSC("InitContinuation");

    pcont->Uri = Uri;
    pcont->Frame = (PTR) Frame;
    pcont->metd = metd;
    pcont->methodNS = Xns;
    if (pConn && !async) {
        //ContinuationList = ListInsert(ContinuationList, &pcont->Link);
        pConn->v->AddRef(pConn);
        pcont->pConnector = pConn;
    }

    /* CONT_FLAGS_CONST_CONTEXT controls deletion of this item */
    pcont->pContext = pContext;

    if (Flags & CONT_FLAGS_LOCAL) {
        PARSE_URL_FLAGS Flags;
        const _TCHAR *path;
        UINT pathlen;           /* XXX ignored */
        PISOAPFACTORY soapfactory;

        sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"),
                          NAME_SPACE_READ, &IID_ISoapFactory,
                          (void**) &soapfactory);
        CHECKSC("bind to soap.cob");

        sc = soapfactory->v->ParseURL(soapfactory, pContext->To->Address,
                                      &Flags, NULL, NULL, NULL,
                                      &path, &pathlen,
                                      NULL,NULL,NULL);
        soapfactory->v->Release(soapfactory);
        CHECKSC("ParseURL");

        sc = pcont->v->SetMethod(iCONT(pcont), Xns, metd,
                                 path, NULL);
        CHECKSC("SetMethod");

        sc = pcont->v->Execute(iCONT(pcont), FALSE, NULL);
        CHECKSC("Execute");
    } else {
        sc = pcont->v->SendMessage(iCONT(pcont), pConn, NULL,NULL);
        CHECKSC("SendMessage");
    }

METHOD_CLEANUP;
    RELEASE(file);
    RELEASE(ser);

    return sc;
}

PRIVATE const _TCHAR *ContinuationUriPrefix = _T("http://tempuri.org/continuation/");

/* Normal RPC */
/* PRIVATE should be but gcc/arm has trouble? */
INT32 __fastcall ProxyInterpreter(CPMETHOD_DESCR metd, PUINT32 pStack)
{
    METHOD_DECLARE_NOTHIS(Proxy, Interpreter);
    PUINT32 pArgs = pStack + 1; /* zap return address */

    PPROXY px = pPX(pArgs[0]);
    struct CONTINFO _continuation;
    PCONTINFO pcont = &_continuation;
    _TCHAR Uri[50];               /* size for prefix + integer */
    static int cycle = 100;
    UINT id = cycle++;
#ifdef MEASURE
    TIME t0, t1;    /* For measuring processing latency (RTSOAP) */
#endif

    _stprintf(Uri, _T("%s0x%08x"), ContinuationUriPrefix, id);

    DBGME(0,printf("Proxy metd=%x [%s] px=%x stk=%x *%x pArgs=%x a0=%x a1=%x a2=%x a3=%x a4=%x a5=%x\r\n",
                   (UINT)metd, metd->name, (UINT) px, (UINT) pStack, *pStack,
                   (UINT)pArgs, pArgs[0], pArgs[1], pArgs[2], pArgs[3],
                   pArgs[4], pArgs[5]));

#ifdef MEASURE
    /* Latency measurement set-up */
    /* Read the current time to set up for client-side measurement */
    t0 = CurrentTime();
#endif

    sc = ProxyStart(pcont, px->NameSpace, metd, px->pConnector, px->pContext,
                    Uri, CONT_FLAGS_CONST | px->Flags, pArgs);
    CHECKSC("ProxyStart");

    sc = pcont->v->Wait(iCONT(pcont));

    /* Finalize continuation */
    pcont->v->Release(iCONT(pcont));

#ifdef MEASURE
    /* Read time again to finish latency measurement */
    t1 = CurrentTime();
    t0 = Int64Subtract(t1, t0); 
    PrintTime(_T("proxy.c:ProxyInterpreter: CPU time:"), t0);
#endif

METHOD_CLEANUP;
    return sc;
}

/* RPC with async API call pattern. Returns future (continuation) */
/* PRIVATE should be but gcc/arm has trouble? */
INT32 __fastcall FutureProxyInterpreter(CPMETHOD_DESCR metd, PUINT32 pStack)
{
    PUINT32 pArgs = pStack + 1; /* zap return address */
    BYTE *frame;
    METHOD_DECLARE_NOTHIS(Proxy, FutureInterpreter);

    PPROXY px = pPX(pArgs[0]);
    PIFILE file = NULL;
    PISOAPSERIALIZER ser = NULL;
    PCONTINFO pcont = NULL;
    _TCHAR *Uri;
    static int cycle = 100;
    UINT id = cycle++;
    PIHTTPCONNECTOR conn = px->pConnector;
    PIHEAP Heap = CurrentHeap();

    pcont = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, sizeof(*pcont)
                           + metd->_FrameSize
                           + 50 * sizeof(_TCHAR), 8);
    CHECKSETSC(pcont, E_NOT_ENOUGH_MEMORY, "Alloc cont");

    /* Make more resilient copy of frame */
    frame = (void *) (pcont + 1);
    memcpy(frame, pArgs, metd->_FrameSize);

    /* URI for wsa:MessageID */
    Uri = (_TCHAR*) (frame + metd->_FrameSize);
    _stprintf(Uri, _T("%s0x%08x"), ContinuationUriPrefix, id);

    sc = ProxyStart(pcont, px->NameSpace, metd, conn, px->pContext, Uri,
                    px->Flags, frame);
    CHECKSC("ProxyStart");

METHOD_CLEANUP;
    /* So far we pretend success, errors will be picked up later */
    if (FAILED(sc) && pcont)
        pcont->RetVal = sc;
    RELEASE(file);
    RELEASE(ser);

    return (INT32) pcont;
}

/* This is a no-return proxy
 * that is called on top of an existing continuation  */
/* PRIVATE should be but gcc/arm has trouble? */
INT32 __fastcall HeaderProxyInterpreter(CPMETHOD_DESCR metd, PUINT32 pStack)
{
    PUINT32 pArgs = pStack + 1; /* zap return address */
    PPROXY px = pPX(pArgs[0]);
    PCONTINFO cont = (PCONTINFO) pArgs[1];
    PISOAPSERIALIZER ser;
    PTR SavedFrame;
    CPARGUMENT_DESCR SavedMetd;
    CPXMLNAMESPACE SavedXns;
    SCODE sc;
    UINT hrefn = 1000000; /* can there be a conflict between diff headers? */
    PHREF hreflist = NULL;

    //DBGME(0,printf("ProxyHeader metd=%x px=%x cont=%x\n",
    //               (ADDRESS) metd, (ADDRESS) px, (ADDRESS) cont));

    SavedFrame = cont->Frame;
    SavedMetd = cont->metd;
    SavedXns = cont->methodNS;

    /* Make sure cont makes sense */
    assert(metd->table_size >= 2
           && metd->ptable[0].descr_type == ARG_CONTINUATION);
    assert(cont->v);

    ser = cont->Serializer;
#if 1
    DBGME(0,printf("ProxyHeader metd=%x %x[%s] px=%x stk=%x *%x pArgs=%x a0=%x a1=%x a2=%x a3=%x a4=%x a5=%x\r\n",
                   (UINT)metd, metd->name, metd->name,
                   (UINT) px, (UINT) pStack, *pStack,
                   (UINT)pArgs, pArgs[0], pArgs[1], pArgs[2], pArgs[3],
                   pArgs[4], pArgs[5]));
#endif
    /* Initialize continuation (on stack) */
    cont->Frame = (PTR) pArgs; /* deserializer expects it here */
    cont->metd = metd;
    cont->methodNS = px->NameSpace;

    /* Serialize */
    sc = ser->v->Method(ser, iCONT(cont), FALSE, NULL/*&hreflist*/, &hrefn);
    /* XXX if this is server, we need to output the hrefs,
     * I suppose into the header?
     */
    //DestroyHReflist(&hreflist, CurrentHeap());

    cont->Frame = SavedFrame;
    cont->metd = SavedMetd;
    cont->methodNS = SavedXns;
    return sc;
}

SCODE CreateFuture(PIHTTPCONNECTOR pConn,
                   PWSCONTEXT pContext,
                   CONT_FLAGS Flags,
                   const void * Frame,
                   const _TCHAR *FrameType,
                   /*out*/ PICONTINUATION *ppCont)
{
    METHOD_DECLARE_NOTHIS(proxy.c, CreateFuture);
    CPMETHOD_DESCR metd;
    UINT len;
    CPXMLNAMESPACE Xns;
    BOOL async = FALSE;
    PIHEAP Heap = CurrentHeap();
    PCONTINFO pcont = NULL;
    PICONVERT convert = NULL;

    CHECKSETSC(FrameType && Frame && ppCont,
               E_INVALID_PARAMETER, "param sanity");

    sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                      NAME_SPACE_READ, &IID_IConvert,
                      (void**) &convert);
    CHECKSC("bind to convert.cob");

    len = _tcscspn(FrameType, _T(":"));
    if (FrameType[len] != _T(':') || FrameType[len+1] == _T('\0'))
        return E_INVALID_PARAMETER;/* XXX could try SoapAction style instead */
    sc = convert->v->GetNamespace(convert, FrameType, len, &Xns);
    CHECKSC("GetNamespace");

    sc = convert->v->FindMethod(convert, Xns, FrameType + len + 1, TRUE,
                                (void **) &metd);
    CHECKSC("FindMethod");

    pcont = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, sizeof(*pcont), 0);
    CHECKSETSC(pcont, E_NOT_ENOUGH_MEMORY, "Alloc cont");

    *ppCont = iCONT(pcont);

    sc = ProxyStart(pcont, Xns, metd, pConn, pContext, NULL, 
                    Flags, Frame);
    CHECKSC("ProxyStart");

METHOD_CLEANUP;
    /* So far we pretend success, errors will be picked up later */
    if (FAILED(sc) && pcont)
        pcont->RetVal = sc;

    RELEASE(convert);
    return sc;
}

SCODE CreateCLRContinuation(
                   PWSCONTEXT pContext,
                   CONT_FLAGS Flags,
                   const _TCHAR *FrameType,
                   /*out*/ PICONTINUATION *ppCont
                   )
{
    METHOD_DECLARE_NOTHIS(proxy.c, CreateCLRContinuation);
    CPMETHOD_DESCR metd;
    UINT len;
    CPXMLNAMESPACE Xns;
    PIHEAP Heap = CurrentHeap();
    PCONTINFO pcont = NULL;
    PICONVERT convert = NULL;

    CHECKSETSC(FrameType && pContext && ppCont,
               E_INVALID_PARAMETER, "param sanity");

    sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                      NAME_SPACE_READ, &IID_IConvert,
                      (void**) &convert);
    CHECKSC("bind to convert.cob");

    len = _tcscspn(FrameType, _T(":"));
    if (FrameType[len] != _T(':') || FrameType[len+1] == _T('\0'))
        return E_INVALID_PARAMETER;/* XXX could try SoapAction style instead */

    sc = convert->v->GetNamespace(convert, FrameType, len, &Xns);
    CHECKSC("GetNamespace");

    sc = convert->v->FindMethod(convert, Xns, FrameType + len + 1, TRUE,
                                (void **) &metd);
    CHECKSC("FindMethod");

//    CreateContinuation(0, &pcont);

    pcont = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, sizeof(*pcont), 0);
    CHECKSETSC(pcont, E_NOT_ENOUGH_MEMORY, "Alloc cont");

    InitContinuation(pcont, CONT_FLAGS_CLIENT | Flags, 
        (Flags & CONT_FLAGS_SOAP_1_1) ? _T("soap11") : _T("soap"));
    
    {
        PIHEAPFACTORY HeapFactory;
        
        sc = BindToObject(CurrentNameSpace(), _T("COB/tmp_heap.cob"),
                          NAME_SPACE_READ, &IID_IHeapFactory,
                          (void **) &HeapFactory);
        CHECKSC("bind tmp_heap.cob");

        sc = HeapFactory->v->CreateNested(HeapFactory, Heap,
                                          0, 0, 4096, &pcont->TmpHeap);
        RELEASE(HeapFactory);
        CHECKSC("create tmp heap");
    }

    pcont->Uri = pContext->To->Address;
    pcont->metd = metd;
    pcont->methodNS = Xns;
    pcont->pContext = pContext;
    if (!(Flags & CONT_FLAGS_LOCAL)) 
        GetConnector(pcont->Uri, Xns, &pcont->pConnector);

    *ppCont = iCONT(pcont);

METHOD_CLEANUP;
    /* So far we pretend success, errors will be picked up later */
    if (FAILED(sc) && pcont)
        pcont->RetVal = sc;

    RELEASE(convert);
    return sc;
}


/* This function creates a vtable and code that jumps to ProxyInterpreter */
PRIVATE ADDRESS *ProxyMakeVTable(CPINTERFACE_DESCR pInterface,
          INT32 (__fastcall *Interpreter)(CPMETHOD_DESCR metd, PUINT32 pStack))
{
    INT i, mcount, size, methodcount;
    PIHEAP heap = CurrentHeap();
    ADDRESS *vtab;
    UINT8 *CodeBuf;
    UINT iunkslots;
    CPMETHOD_DESCR metd;
    CPINTERFACE_DESCR ifd;
#ifdef __GNUC__
    /* g++ expects two dummy words at the beginning of a v-table */
    UINT vtabfillerslots = 2;
#else
    UINT vtabfillerslots = 0;
#endif

    /* XXX deal with missing/present IUnknown and assume it should be there */
    if (!_tcscmp(pInterface->ptable[0].name, _T("QueryInterface")))
        iunkslots = 0;
    else
        iunkslots = 3;

    /* Start by calculating size needed for code & v-table */
    methodcount = size = 0;
    for (ifd = pInterface; ifd; ifd = ifd->iface_extras->SuperClass)
        for (i = ifd->table_size-1; i >= 0; i--) {
            metd = (CPMETHOD_DESCR) &ifd->ptable[i];
            if (metd->any.Type != ENTRY_TYPE_METHOD
                || metd->descr_type != DESCR_TYPE_METHOD)
                continue;
            size += ProxyCodeInit(NULL, metd, Interpreter);
            methodcount++;
        }
    /* Add vtable size */
    size += (methodcount + vtabfillerslots + iunkslots) * sizeof(ADDRESS);
    /* If inter-procedure alignment is required the MD code should
     * account for it in the returned size.
     */

    /* Allocate */
    vtab = (ADDRESS*) heap->v->Alloc(heap,
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("pxy-v"),
        size, sizeof(ADDRESS));
    if (!vtab)
        return NULL;

    DBG(("ProxyMakeVTable: ProxyInterpreter=%x &DescrTable[1]=%x\n",
        (ADDRESS) Interpreter, (UINT) &vtab[1]));

    /* put stub code right after vtable */
    CodeBuf = (UINT8*) &vtab[methodcount + vtabfillerslots + iunkslots];

    /* Fill em in (vtab + code), go backwards so as to get inherited first */
    size = 0;
    mcount = methodcount-1;
    for (ifd = pInterface; ifd; ifd = ifd->iface_extras->SuperClass)
        for (i = ifd->table_size-1; i >= 0; i--) {
            metd = (CPMETHOD_DESCR) &ifd->ptable[i];
            if (metd->any.Type != ENTRY_TYPE_METHOD
                || metd->descr_type != DESCR_TYPE_METHOD)
                continue;
            vtab[mcount + vtabfillerslots + iunkslots] = (ADDRESS) &CodeBuf[size];
            size += ProxyCodeInit(&CodeBuf[size], metd, Interpreter);
DBG(("Proxy[%d] metd=%x %x[%s]\n", i, metd, metd->name, metd->name));
            mcount--;
        }
    assert(mcount == -1);

    /* Put in proxy IUnknown method */
    vtab[vtabfillerslots    ] = (ADDRESS) ProxyQueryInterface;
    vtab[vtabfillerslots + 1] = (ADDRESS) ProxyAddRef;
    vtab[vtabfillerslots + 2] = (ADDRESS) ProxyRelease;

    DBG(("Used %d bytes of memory for %d proxy methods and vtable\n", 
         size, methodcount));

    return vtab;
}

#define vloff offsetof(struct via, ListNode)
#define rloff offsetof(struct reservation, ListNode)

PRIVATE SCODE MakeProxyFromV(PIHTTPCONNECTOR pConn, PWSCONTEXT pCtx, 
                             ADDRESS *VTable, CPXMLNAMESPACE ns,
                             CONT_FLAGS Flags,
                             PIUNKNOWN *ppp)
{
    PIHEAP pHeap = CurrentHeap();
    PPROXY This = (PPROXY) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("proxy"), 
              sizeof(*This), 0);
    SCODE sc = E_NOT_ENOUGH_MEMORY;
    if (!This) return sc;

    /* Allocated in CreateProxyX, so direct assignment is OK */
    This->pContext = pCtx;

    if (pConn != NULL) {
        This->pConnector = pConn;
        pConn->v->AddRef(pConn);
    } else if (pCtx && (Flags & CONT_FLAGS_LOCAL) == 0) {
        sc = GetConnector(This->pContext->To->Address, ns, &This->pConnector);
        if (FAILED(sc))
            goto Error;
    }

    This->v = VTable;
    This->RefCount = 1;
    This->NameSpace = ns;
    This->Flags = Flags;

    *ppp = (PIUNKNOWN) This;
    return S_OK;

 Error:
    pHeap->v->Free(pHeap, 0, (PTR) This);
    return sc;
}

/* Later... reuse vtables (e.g. register in namespace) */
PRIVATE SCODE MakeProxy(PIHTTPCONNECTOR pConn, PWSCONTEXT pCtx, 
                        CPINTERFACE_DESCR ifd, CPXMLNAMESPACE ns,
                        CONT_FLAGS Flags,
                        PIUNKNOWN *ppp)
{
    SCODE sc;
    ADDRESS *vtab = ProxyMakeVTable(ifd, pCtx ? ((Flags & CONT_FLAGS_FUTURE)
                                                 ? FutureProxyInterpreter
                                                 : ProxyInterpreter)
                                              : HeaderProxyInterpreter);
    if (!vtab)
        return E_FAIL;
    sc = MakeProxyFromV(pConn, pCtx, vtab, ns, Flags, ppp);
    if (FAILED(sc)) 
        free(vtab); /* XXX Later... cache v-table (maybe) */
    return sc;
}

PRIVATE void DestroyProxy(PPROXY pxy)
{
    RELEASE(pxy->pConnector);
    if (pxy->Flags & CONT_FLAGS_SHOULD_FREE_CONTEXT)
        free(pxy->pContext);
    free(pxy->v);                /* XXX Later... cache v-table */
    free(pxy);
}

/* Allocates and initializes a WSCONTEXT from a To address parameter.
 * Only the WSCONTEXT pointer needs to be freed.
 * (The string copy is co-allocated with the structure.)
 */
SCODE CreateWsContextFromUri(const _TCHAR *Uri, PWSCONTEXT *ppCtx)
{
    PIHEAP Heap = CurrentHeap();
    /* Context and endpoint structures plus string buffer size */
    UINT BufSz = sizeof(WSCONTEXT) 
               + sizeof(EndpointReference) 
               + (_tcslen(Uri) + 1) * sizeof(_TCHAR);

    /* Allocate structures and string */
    PWSCONTEXT pCtx = Heap->v->Alloc(Heap,
                            HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("pxywscontext"),
                            BufSz, 0);
    if (0 == pCtx) return E_NOT_ENOUGH_MEMORY;
    memset(pCtx, 0, BufSz);

    /* Lay the string next to the structure */
    pCtx->To = (PEndpointReference) (pCtx + 1);
    pCtx->To->Address = (_TCHAR *) (pCtx->To + 1);
    _tcscpy(pCtx->To->Address, Uri);

    /* Done */
    *ppCtx = pCtx;
    return S_OK;
}

SCODE CreateProxy3(PIHTTPCONNECTOR pConn, PWSCONTEXT pCtx, 
                   const _TCHAR *XnsName,
                   CONT_FLAGS Flags,
                   PIUNKNOWN *ppProxy)
{
    METHOD_DECLARE_NOTHIS(proxy.c, CreateProxy3);
    CPINTERFACE_DESCR ifd;
    CPXMLNAMESPACE Xns;
    PICONVERT convert = NULL;

    sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                      NAME_SPACE_READ, &IID_IConvert,
                      (void**) &convert);
    CHECKSC("bind to convert.cob");

    sc = convert->v->GetNamespace(convert, XnsName, 0, &Xns);
    CHECKSC("GetNamespace");

    ifd = Xns->ifd;
    CHECKSETSC(ifd, E_NO_INTERFACE, "Xns->ifd");
    if (!ifd) {
        /* Search for class in typeinfo and call CreateProxy2 */
        return E_NO_INTERFACE;
    }

    sc = MakeProxy(pConn, pCtx, ifd, Xns, Flags, ppProxy);
    CHECKSC("MakeProxy");

METHOD_CLEANUP;
    RELEASE(convert);
    return sc;
}

SCODE CreateProxy2(PIHTTPCONNECTOR pConn, const _TCHAR *Uri, 
                   CPINTERFACE_DESCR ifd, CPXMLNAMESPACE ns, 
                   PIUNKNOWN *ppProxy)
{
    SCODE sc;
    PWSCONTEXT pCtx = NULL;

    sc = CreateWsContextFromUri(Uri, &pCtx);
    if (FAILED(sc)) return sc;

    sc = MakeProxy(pConn, pCtx, ifd, ns, 0, ppProxy);
    if (FAILED(sc))
        free(pCtx);
    else
        ((PPROXY) *ppProxy)->Flags |= CONT_FLAGS_SHOULD_FREE_CONTEXT;

    return sc;
}

SCODE CreateProxy(PIHTTPCONNECTOR pConn, const _TCHAR *Uri,
                  const _TCHAR *XnsName,
                  CONT_FLAGS Flags,
                  PIUNKNOWN *ppProxy)
{
    SCODE sc;
    PWSCONTEXT pCtx = NULL;

    sc = CreateWsContextFromUri(Uri, &pCtx);
    if (FAILED(sc)) return sc;

    sc = CreateProxy3(pConn, pCtx, XnsName, Flags, ppProxy);
    if (FAILED(sc))
        free(pCtx);
    else
        ((PPROXY) *ppProxy)->Flags |= CONT_FLAGS_SHOULD_FREE_CONTEXT;

    return sc;
}
