/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>
#include <soapif.h>
#include <web/rpc.h>
#include <web/proxy.h>
#include <web/udpsoapserv.h>
#include <assert.h>

#if _MINIMIZE
int _tmain(INT argc, _TCHAR *argv[])
{
    return 1;
}
#else
/* TEST INTERFACE */
typedef struct ITest ITest;
typedef struct ITestVtbl {
    __VTABLE_COMPATIBILITY_FILLER_DECLARATION

    SCODE ( __cdecl *QueryInterface )(
        ITest *This,
        /* [in] */ REFIID Iid,
        /* [out] */ void **ppObject);
    UINT ( MCT *AddRef )(
        ITest *This);
    UINT ( MCT *Release )(
        ITest *This);
    SCODE ( __stdcall *StdCallQI )(
        ITest *This,
        /* [in] */ REFIID Iid,
        /* [out] */ void **ppObject);
    UINT ( MCT *Add )(
        ITest *This,
        UINT a,
        UINT b);
    void ( MCT *Add3 )(
        ITest *This,
        UINT a,
        UINT b,
        PUINT rv /*out*/);
} ITestVtbl;

typedef struct ITest {
    CONST_VTBL struct ITestVtbl *v;
} *PITEST;

/* quick inplace typeinfo */
#define I ARGUMENT_DESCR_FLAGS_IN
#define O ARGUMENT_DESCR_FLAGS_OUT
#define R ARGUMENT_DESCR_FLAGS_RETURN

#define ME {ENTRY_TYPE_METHOD, 0, 1}
#define AE {ENTRY_TYPE_ARG, 0, 1}
#define IE {ENTRY_TYPE_INTERFACE, 0, 1}

struct _METHOD_DESCR qiarg[] = {
    {AE, "Iid", NULL, I,  7,ARG_GUID, 4,}, /* iid */
    {AE, "pObj", NULL, O, 8,ARG_URI, 8,}, /* object */
    {AE, "sc", NULL, O|R, 9,ARG_INT, } /* sc return */
};

struct _METHOD_DESCR addrefarg[] = {
    {ME, "Count", NULL, O|R, 10,ARG_INT,} /* int return */
};

struct _METHOD_DESCR addarg[] = {
    {AE, "A",      NULL, I,   11,ARG_INT, 4,}, /* add arg */
    {AE, "B",      NULL, I,   12,ARG_INT, 8,}, /* add arg */
    {AE, "Result", NULL, O|R, 13,ARG_INT, } /* add result */
};

struct _METHOD_DESCR add3arg[] = {
    {AE, "A",      NULL, I, 14,ARG_INT, 4,}, /* add arg */
    {AE, "B",      NULL, I, 15,ARG_INT, 8,}, /* add arg */
    {AE, "Result", NULL, O, 16,ARG_INT, 12,} /* add result */
};

#define S0 0
#define S(x) x
#define M DESCR_TYPE_METHOD
#define V (void*)
struct _METHOD_DESCR testmethods[] = {
    {ME,"QueryInterface",qiarg,    3,1,M,12,4096,0,S0,   0,V "QIResponse"    },
    {ME,"AddRef",        addrefarg,1,2,M,4, 4096,0,S0,   1,V "AddRefResponse"},
    {ME,"Release",       addrefarg,0,3,M,4, 4096,0,S0,   2,V "ReleaseResponse"},
    {ME,"QIstd",         qiarg,    3,4,M,12,4096,0,S(12),3,V "QIstdResponse" },
    {ME,"Add",           addarg,   3,5,M,12,4096,0,S0,   4,V "AddResponse"   },
    {ME,"Add",           add3arg,  3,6,M,16,4096,0,S0,   5,V "AddResponse"   }
};
#undef V
#undef M
#undef S
#undef S0

struct _METHOD_DESCR testif[1] = {
    {IE,"Calc",testmethods,6, 0, DESCR_TYPE_INTERFACE, 0,}
};

const struct XMLNAMESPACE calc_testns = {
    "tns", 
    "http://tempuri.org/Calc/message/"/*"http://tempuri.org/AddrBook/message/"*/,
    NSTYPE_UNRESOLVED,
    NS_Sample_AddrBook, 
    0, 
    testif,
};

const struct XMLNAMESPACE stk_testns = {
    "tns", "http://tempuri.org/AddrBook/message/", NSTYPE_UNRESOLVED,
    NS_Sample_AddrBook, 0, testif,
};

int _tmain(INT argc, _TCHAR *argv[])
{
    PITEST pxy;
    PPROXY calc;
    UINT i;
    SCODE sc;
    char *target = MulticastIPAddr1, *contentType = "x-udp-soap";
    char *path = "/COB/calc.cob";
    char to[80];
    CPXMLNAMESPACE testns = &calc_testns;

    PISOAPFACTORY soapfactory;

    ParseCmdLine(&argc,&argv);
    if (argc > 4) {
        _tprintf(_T("Usage: proxytest.exe [<target IP address> [<content type (http|udp|xte)>][stk]]\n"));
        return 0;
    } else if (argc > 1) {
#ifdef _UNICODE
        target = _wcstos(argv[1]);
        if (argc >= 3)
            contentType = _wcstos(argv[2]);
#else
        target = argv[1];
        if (argc >= 3)
            contentType = argv[2];
#endif
        if (argc >= 4) {
            path = "/MSSoapSamples30/Calc/Service/SrSz/AspJs/Calc.asp";
            testns = &stk_testns;
        }
    }
    strcpy(to, contentType);
    strcat(to, "://");
    strcat(to, target);
    strcat(to, path);

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISOAPFACTORY, (void **) &soapfactory);
    assert(sc == S_OK);

    sc = soapfactory->v->CreateProxy2(soapfactory, NULL, to, testif, testns, (PIUNKNOWN*) &pxy);
    if (FAILED(sc))
        return sc;
    
    calc = (PPROXY)pxy;
    sc = calc->pConnector->v->SetKey(calc->pConnector);
    if (FAILED(sc))
        goto Out;

#if 0
    {
        PTR p;
        PUINT z;
        printf("ProxyTest2 sc=%x pxy=%x iunk=%x &p=%x\n",
               sc, (UINT) pxy, (UINT) &IID_IUnknown, (UINT) &p);
        z = (PUINT) pxy->v->Add;
        printf("Add=%08x [%08x %08x %08x %08x %08x]\r\n", pxy->v->Add,
               z[0], z[1], z[2], z[3], z[4]);
    }
#endif
    /* Call the proxy */
    i = pxy->v->Add(pxy, 14, 27);
    printf("ProxyTestAdd x%x (%d). Should be 41.\n", i, i);
    assert(i == 41);

    sc = calc->pConnector->v->ClearKey(calc->pConnector);
    if (FAILED(sc))
        goto Out;

    /* Call the proxy */
    i = -5;
    pxy->v->Add3(pxy, 13, 29, &i);
    printf("ProxyTestAdd3 x%x (%d). Should be 42.\n", i, i);
    assert(i == 42);

//DebugBreak();

#if 0
    i = pxy->v->AddRef(pxy);
    printf("ProxyTestAddRef %x\n", i);

    i = pxy->v->StdCallQI(pxy, &IID_IUnknown, &p);
    printf("ProxyTestQIstd %x\n", i);

    i = pxy->v->QueryInterface(pxy, &IID_IUnknown, &p);
    printf("ProxyTestQI %x\n", i);
#endif
Out:
    pxy->v->Release(pxy);
    soapfactory->v->Release(soapfactory);

    printf("done\n");

#ifdef _UNICODE
    if (argc == 2)
        free(target);
    if (argc == 3)
        free(contentType);
#endif
    FreeCmdLine(&argc,&argv);

    return 0;
}
#endif
