/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <mmlite.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <sampleif.h>
#include <tchar.h>
#include <web/proxy.h>
#include <web/udpsoapserv.h>
#include <wsif.h>

_TCHAR ss1[20], ss2[20], ss3[20], ss4[20], ss5[20], ss6[20];

void TimeFormat(TIME t, _TCHAR *buf, UINT Len);

#define CHECK(_c_) \
    MACRO_BEGIN \
    assert(_c_); \
    if (!(_c_)) goto Out; \
    MACRO_END


/* XXX Next three functions duplicated from sensormain.c, 
   should be moved into some library... */

PRIVATE UINT32 RandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}

_TCHAR *CreateUniqueTaskName(void)
{
    TIME current = CurrentTime();
#define BUFLEN 256
    _TCHAR buf[BUFLEN];
    UINT i;
    TimeFormat(current, buf, BUFLEN);
    for (i = 0; i < BUFLEN; ++i)
        if ('\0' == buf[i])
            break;
    if (10 > BUFLEN - i)
        return NULL;
    _stprintf(buf + i, _T("%d"), RandUint32());
    return _tcsdup(buf); /* XXX LEAK Look out for memory leaks here... */
}

_TCHAR *CreateReservationName(
    const _TCHAR *taskName, 
    const _TCHAR *resName, 
    const _TCHAR *Suffix)
{
    UINT tlen = _tcslen(taskName);
    UINT rlen = _tcslen(resName) + 1;
    UINT slen = (NULL != Suffix) ? _tcslen(Suffix) : 0;
    UINT len = tlen + rlen + slen + 1;
    _TCHAR *buf = (_TCHAR *) malloc(len * sizeof(_TCHAR));
    if (!buf)
        return NULL;
    _tcscpy(buf, taskName);
    _tcscat(buf, _T("/"));
    _tcscat(buf, resName);
    if (Suffix)
        _tcscat(buf, Suffix);
    return buf;
}

int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(proxytest2, main);
    PISAMPLECALCULATOR pxy = NULL, pxyB = NULL;
    PTR p;
    UINT i1;
#if !_MINIMIZE
    UINT i2;
    UINT loop;
    INT64 ll1, ll2;
    _TCHAR *s1;
    TIME t1, t2;
    _TCHAR sbuf[100];
    UINT ia[] = {13,17,19,23}, ia2[] = {130,170,190,230}, oa[5];
    _TCHAR *sia[] = {_T("foo"), _T("bar")}, *sia2[] = {_T("FOO"), _T("BAR")};
    _TCHAR *soa[6] = {ss1,ss2,ss3,ss4,ss5,ss6};
    struct SOAPStruct ti = {42, _T("zappa"), 144}, to;
    struct SOAPArrayStruct tai = {_T("qwerty"), 42, 17,
                                     {_T("NestedFoo"), _T("NestedBar")}}, tao;
    struct SOAPStructStruct tti = {_T("qaz"), 42, 17,
                                      {4242, _T("nested"), 144144}}, tto;
    struct SOAPStruct ati[2] = {{111, _T("foo"), 11},
                                {222, _T("bar"), 22}}, ato[3];
#endif
    _TCHAR *url = _T("http://172.31.41.15/asmx/simple.asmx");
    //char *url = _T("x-udp-xte-soap://") _T(MulticastIPAddr1) _T("/COB/calc.cob");
    _TCHAR *urlB = _T("http://172.31.41.15/asmx/simpleB.asmx");
    _TCHAR *urlUdp = _T("x-udp-soap://") _T(MulticastIPAddr1) _T("/COB/calc.cob");
    _TCHAR *urlUdpAes = _T("x-udp-aes-soap://") _T(MulticastIPAddr1) _T("/COB/calc.cob");
    _TCHAR *urlLocal = _T("http://127.0.0.1/COB/calc.cob");
    _TCHAR *urlLocal2 = _T("http://127.0.0.1:4000/COB/calc.cob");
    _TCHAR *urlLocal3 = _T("http://127.0.0.1/COB/sensor.cob");
    BOOL DoLocals = FALSE;
    BOOL DoTime = FALSE;
    BOOL DoKey = FALSE;
    BOOL DoMain = FALSE;
    BOOL DoReservations = FALSE;
    UINT loopcount = 1;
    CONT_FLAGS ProxyFlags = CONT_FLAGS_CONST_CONTEXT;
    BOOL DoNoProxy = FALSE;

    PISOAPFACTORY soapfactory = NULL;
    PICONVERT convert = NULL;

    /*** Command line ***/
    ParseCmdLine(&argc,&argv);
#if !_MINIMIZE
    if (argc >= 2 && !_tcscmp(argv[1], _T("-t"))) { /* time calibration */
        argc--;
        argv++;
        DoTime = TRUE;
    }
#endif

    if (argc >= 2 && !_tcscmp(argv[1], _T("-w"))) { /* reservations */
        argc--;
        argv++;
        DoReservations = TRUE;
    }

    if (argc >= 2 && !_tcscmp(argv[1], _T("-f"))) { /* futures */
        argc--;
        argv++;
        ProxyFlags |= CONT_FLAGS_FUTURE;
    }

    if (argc >= 2 && !_tcscmp(argv[1], _T("-n"))) { /* no RPC */
        argc--;
        argv++;
        ProxyFlags |= CONT_FLAGS_LOCAL;
    }

    if (argc >= 2 && !_tcscmp(argv[1], _T("-p"))) { /* no RPC */
        argc--;
        argv++;
        DoNoProxy = TRUE;
    }

    if (argc >= 2 && !_tcscmp(argv[1], _T("-k"))) { /* time calibration */
        argc--;
        argv++;
        DoKey = TRUE;
    }

    if (argc >= 2 && !_tcscmp(argv[1], _T("-m"))) { /* Main() */
        argc--;
        argv++;
        DoMain = TRUE;
    }
    if (argc >= 2 && !_tcscmp(argv[1], _T("-m2"))) { /* Main() different ns */
        argc--;
        argv++;
        DoMain = 2;
    }
    if (argc >= 2 && !_tcscmp(argv[1], _T("-l"))) {
        argc--;
        argv++;
        DoLocals = TRUE;
    }
    if (argc >= 2 && !_tcscmp(argv[1], _T("-o"))) {
        argc--;
        argv++;
        ProxyFlags |= CONT_FLAGS_SOAP_1_1;
    }
    if (argc >= 2 && argv[1][0] == _T('-') && _istdigit(argv[1][1])) {
        loopcount = _ttoi(argv[1]+1);
        argc--;
        argv++;
    }

    if (argc > 3) {
        printf("Usage: proxytest2.exe <url> [<urlB>]\n");
        return 1;
    }

    if (argc >= 2) {
        if (!_tcscmp(argv[1], _T("l"))) { /* l for local */
            urlB = url = urlLocal;
            DoLocals = TRUE;
        } else if (!_tcscmp(argv[1], _T("l2"))) { /* l for local */
            urlB = url = urlLocal2;
            DoLocals = TRUE;
        } else if (!_tcscmp(argv[1], _T("l3"))) { /* l for local */
            urlB = url = urlLocal3;
            DoLocals = TRUE;
        } else if (!_tcscmp(argv[1], _T("u"))) { /* l for local */
            urlB = url = urlUdp;
            DoLocals = TRUE;
        } else if (!_tcscmp(argv[1], _T("a"))) { /* l for local */
            urlB = url = urlUdpAes;
            DoLocals = TRUE;
        } else {
            urlB = url = argv[1];
        }
    }
    if (argc >= 3) {
        urlB = argv[2];
    }

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    CHECKSC("bind to soap.cob");

    if (DoReservations) {

#define rloff offsetof(struct reservation, ListNode)    
#define trloff offsetof(struct trigger, ListNode)

        WSCONTEXT Ctx; PWSCONTEXT pCtx = &Ctx;
        EndpointReference To; PEndpointReference pTo = &To;
        struct reservation rsrv;
        struct trigger trigger;

        TIME deadline, Duration, Tolerance;
        TIME Wait, ProDelta;
        
        _TCHAR *TaskName = CreateUniqueTaskName();
        _TCHAR *RsrvName= _T("sensnet-producer-reserve");
        _TCHAR *FullName= CreateReservationName(TaskName, RsrvName, _T("/*"));

        /* Reservation times */
        Int32ToInt64(Wait, TIME_RELATIVE(TIME_SECONDS(10)));
        deadline = Int64Subtract(CurrentTime(), Wait); /* Will be updated with delta... */
        Int32ToInt64(Duration, TIME_RELATIVE(TIME_MILLIS(20)));
        Int32ToInt64(Tolerance, TIME_RELATIVE(TIME_MILLIS(5)));
        PrintTime("Current time", CurrentTime());
        PrintTime("deadline", deadline);

        /* Reservation time deltas, e.g. repeat periods */
        Int32ToInt64(ProDelta, TIME_RELATIVE(TIME_SECONDS(10)));

        /* Addressing */
        memset(pCtx, 0, sizeof(WSCONTEXT));
        memset(pTo, 0, sizeof(*pTo));
        pCtx->To = pTo;
        pTo->Address = url;
        pCtx->Reservation = &rsrv;
        pCtx->taskName = FullName;
        pCtx->TaskName = TaskName;

        /* Trigger */
        memset(&trigger, 0, sizeof trigger);
        trigger.maxCount = 1;
        trigger.offset = ProDelta;
        ListNodeInit(trloff, (&trigger));
        trigger.what = RsrvName;

        /* Reservations - reused and updated. */
        memset(&rsrv, 0, sizeof(struct reservation));
        rsrv.name = RsrvName;
        rsrv.deadline = deadline;
        rsrv.duration = Duration;
        rsrv.tolerance = Tolerance;
        rsrv.contact = urlB;
        rsrv.TriggerList = &trigger;
        ListNodeInit(rloff, (&rsrv));

        /* Proxy construction */
        assert(!pxy);
        sc = soapfactory->v->
            CreateProxy3(soapfactory, NULL, pCtx, _T("tns"),
                         ProxyFlags,
                         (PIUNKNOWN*) &pxy);
        if (FAILED(sc)) {
            printf("proxytest2.exe: DoReservations: No proxy\n");
            goto DoResDone;
        }

        printf("proxytest2.exe: DoReservations: Created proxy: pxy=%x sc=%x\n", (UINT) pxy, sc);

        /* Do something... */
        {
            printf("proxytest2.exe: DoReservations: Invoking proxy\n");

            sc = pxy->v->AsyncTest(pxy, _T("foo"));

            printf("proxytest2.exe: DoReservations: Proxy returned sc=%x\n", sc);
        }

DoResDone:
        if (TaskName)
            free(TaskName);
        if (FullName)
            free(FullName);

        goto Out;
    }

    if (DoMain) {
        PIPROGRAM pgm;
        sc = soapfactory->v->CreateProxy(soapfactory, NULL, url,
                                      (DoMain==2) ? _T("sensor") : _T("prog"),
                                         ProxyFlags,
                                         (PIUNKNOWN*) &pgm);
        if (FAILED(sc)) {
            printf("CreateProxy %s failed sc=%x\n", url, sc);
            goto Out;
        }
        printf("proxytest2.exe: DoMain: Calling Main with arg=(%s)\n",
               urlB);

        sc = pgm->v->Main(pgm, urlB);

        printf("proxytest2.exe: DoMain: Proxy returned sc=%x\n", sc);
        pgm->v->Release(pgm);
        goto Out;
    }

    assert(!pxy);
    sc = soapfactory->v->CreateProxy(soapfactory, NULL, url, _T("tns"),
                                     ProxyFlags,
                                     (PIUNKNOWN*) &pxy);
    if (FAILED(sc)) {
        printf("CreateProxy %s failed sc=%x\n", url, sc);
        goto Out;
    }

    printf("ProxyTest2 Created proxy: pxy=%x iunk=%x &p=%x sc=%x\n",
           (UINT) pxy, (UINT) &IID_IUnknown, (UINT) &p, sc);

    if (url == urlB) {
        /* Same URL, no need for a separate proxy */
        assert(!pxyB);
        pxy->v->AddRef(pxy);
        pxyB = pxy;
    } else {
        assert(!pxyB);
        sc = soapfactory->v->CreateProxy(soapfactory, NULL, urlB, _T("tns"),
                                         ProxyFlags,
                                         (PIUNKNOWN*) &pxyB);
        if (FAILED(sc)) {
            printf("CreateProxy %s failed sc=%x\n", urlB, sc);
            goto Out;
        }
    }
    if (DoNoProxy) {
        PICONTINUATION fut;
        _PACKED struct _ISampleCalculator_Subtract_METHOD_FRAME frame1;
#if !_MINIMIZE
        _PACKED struct _ISampleCalculator_echoBase64_METHOD_FRAME frame2;
#endif
        WSCONTEXT Ctx; PWSCONTEXT pCtx = &Ctx;
        EndpointReference To; PEndpointReference pTo = &To;

        /* Addressing */
        memset(pCtx, 0, sizeof(WSCONTEXT));
        memset(pTo, 0, sizeof(*pTo));
        pCtx->To = pTo;
        pTo->Address = url;

        frame1.pThis = (ADDRESS) pxy;
        frame1._A_ = 42;
        frame1._B_ = 7;
        i1 = 5;
        sc = soapfactory->v->CreateFuture(soapfactory,
                                          NULL,
                                          pCtx,
                                          ProxyFlags,
                                          (void *) &frame1,
                                          _T("tns:Add"),
                                          &fut);
        printf("Created inline future add %x sc=%x\n", fut, sc);
        CHECK(SUCCEEDED(sc));
        CHECK(fut != NULL);
        i1 = fut->v->Wait(fut);
        printf("ProxyTest2 Add(42+7)=%d. Should be %d.\n", i1, 49);
        CHECK(i1 == 49);
        RELEASE(fut);

#if !_MINIMIZE
        s1 = _T("foo");
        frame2.pThis = (ADDRESS) pxy;
        frame2._inputBase64_ = s1;
        frame2._inputDataSize_ = _tcslen(s1) * sizeof(_TCHAR);
        frame2._return_ = sbuf;
        frame2._outputDataSize_ = 99;
        frame2._outputDataLength_ = &i1;

        sc = soapfactory->v->CreateFuture(soapfactory,
                                          NULL,
                                          pCtx,
                                          ProxyFlags,
                                          (void *) &frame2,
                                          _T("tns:echoBase64"),
                                          &fut);
        printf("Created inline future echoBase64 %x\n", fut);
        CHECK(SUCCEEDED(sc));
        CHECK(fut != NULL);
        sc = fut->v->Wait(fut);
        printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
        CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
        CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));
        RELEASE(fut);
#endif
        goto Out;
    }

    if (ProxyFlags & CONT_FLAGS_FUTURE) {
        PICONTINUATION fut;
        i1 = 5;
        fut = (PICONTINUATION) pxy->v->Add(pxy, 42, 7);
        printf("Created future add %x\n", fut);
        CHECK(fut != NULL);
        i1 = fut->v->Wait(fut);
        printf("ProxyTest2 Add(42+7)=%d. Should be %d.\n", i1, 49);
        CHECK(i1 == 49);
        RELEASE(fut);

#if !_MINIMIZE
        s1 = _T("foo");
        fut = (PICONTINUATION) pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
        printf("Created future echoBase64 %x\n", fut);
        CHECK(fut != NULL);
        sc = fut->v->Wait(fut);
        printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
        CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
        CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));
        RELEASE(fut);
#endif
        goto Out;
    }

#if !_MINIMIZE
    if (DoTime) {
        INT i;
        TIME his, hisoff, myoff, my, cur, off, jitter, prevjitter, j, o;
        TIME prevadjust, minerr;
        BOOL neg = FALSE;

        myoff = TIME_ORIGIN;
        prevadjust = prevjitter = TIME_FOREVER;
        Int32ToInt64(minerr, TIME_MILLIS(1));
        for (i = 0; i < 5; i++) {
            cur = CurrentTime();
            my = Int64Add(cur, myoff);
#if 0
            /* Make it relative so current SoapTimeOffset doesn't affect it */
            my = Int64Subtract(TIME_ORIGIN, my);
            sc = pxy->v->QueryTime(pxy, my, &his, &hisoff, &neg);
#else
            sc = pxy->v->QueryTime(pxy, my, &his, &hisoff);
#endif
            if (FAILED(sc)) {
                printf("proxytest2: SetTime failed sc=%x\n", sc);
                goto Out;
            }
            off = Int64Subtract(his, my);
            jitter = Int64Subtract(off, hisoff);
#if 0
            printf("DoTime:  my=%lld  myoff=%lld cur=%016llx sc=%x sahead=%x\n"
                   "DoTime: his=%lld hisoff=%lld off=%lld jitter=%lld\n",
                   my, myoff, cur, sc, neg,
                   his, hisoff, off, jitter);
#else
            printf("DoTime:  my=%016llx  myoff=%lld cur=%016llx sc=%x \n"
                   "DoTime: his=%016llx hisoff=%lld off=%lld jitter=%lld\n",
                   my, myoff, cur, sc, his, hisoff, off, jitter);

#endif

            j = Int64IsNegative(jitter) ? Int64Subtract(TIME_ORIGIN, jitter)
                                       : jitter;
            o = Int64IsNegative(off) ? Int64Subtract(TIME_ORIGIN, off)
                                     : off;
            if ((!TimeLess(j, o) && TimeLess(prevjitter, j))
                || TimeLess(o, minerr)) {
                printf("Adjust amount below precision...\n");
                continue;
            }
            if (TimeLess(j, prevjitter))
                prevjitter = j;

            /* Add improvement, reduced by jitter, to myoff */
            if (TimeLess(j, o)) {
                if (Int64IsNegative(off)) {
                    off = Int64Add(off, j);
                } else {
                    off = Int64Subtract(off, j);
                }
            }
            /* Reduce |adjustment| to prevadjust */
            if (Int64IsNegative(off)) {
                TIME o2 = Int64Subtract(TIME_ORIGIN, off);
                if (TimeLess(o2, prevadjust))
                    prevadjust = o2;
                else
                    off = Int64Subtract(TIME_ORIGIN, prevadjust);
            } else {
                if (TimeLess(off, prevadjust))
                    prevadjust = off;
                else
                    off = prevadjust;
            }

            myoff = Int64Add(myoff, off);
            printf("j=%lld o=%lld ADJUST=%lld => %lld\n", j, o, off, myoff);
        }
        if (!Int64IsNegative(myoff)) {
            myoff = Int64Subtract(TIME_ORIGIN, myoff);
            neg = TRUE;
        }

        sc = BindToObject(CurrentNameSpace(),
                          _T("COB/convert.cob"), NAME_SPACE_READ,
                          &IID_IConvert, (void **) &convert);
        CHECKSC("bind to soap.cob");

        convert->v->SetTimeOffset(convert, myoff, neg);

        {
            _TCHAR buf[50];
            TimeFormat(myoff, buf, sizeof(buf)-1);
            printf("Got offset %s from server neg=%x\n", buf, neg);

            cur = CurrentTime();

            off = myoff;
            if (neg) {
                off = Int64Subtract(TIME_ORIGIN, off);
#if 0
                TimeFormat(off, buf, sizeof(buf)-1);
#else
                his = Int64Add(cur, off);
                TimeFormat(off, buf, sizeof(buf)-1);
            } else {
                off = Int64Subtract(TIME_ORIGIN, off);
                TimeFormat(off, buf, sizeof(buf)-1);
                his = Int64Subtract(cur, off);
#endif
            }
            printf("Corresponding to t=%s\n", buf);
#if 0
            his = Int64Add(cur, off);
#endif
            /* Make sure we don't under/overflow */
            if (Int64IsNegative(his)) {
                if (Int64IsNegative(off))
                    his = TIME_ORIGIN;
                else
                    his = TIME_FOREVER;
            }
            
            TimeFormat(his, buf, sizeof(buf)-1);
            printf("That means now is %s\n", buf);
        }
        goto Out;
    }
#endif

#if !_MINIMIZE
    if (DoKey && 0) {
        sc = ((PPROXY)pxy)->pConnector->v->SetKey(((PPROXY)pxy)->pConnector);
        printf("SerKey returned sc=%x\n", sc);
        if (FAILED(sc))
            goto Out;
    }

    printf("proxytest2 main: Starting tests loops=%d\n", loopcount);

    for (loop = 0; loop < loopcount; loop++) {
    if (DoLocals) {
#if 0
        _TCHAR sbuf2[100];
        _TCHAR sbuf3[100] = _T("12345678901234567890");
        UINT l2, l3 = 20;

        i1 = pxy->v->MarshalTest(pxy,
                                 _T("Foo"), &i2, &ll1,
                                 sbuf, sizeof(sbuf)/sizeof(_TCHAR),
                                 "bar", 4,
                                 sbuf2, sizeof(sbuf2), &l2,
                                 sbuf3, sizeof(sbuf3), &l3);
        printf("MarshalTest: sbuf='%s'\n", sbuf);
#endif
        i1 = 5;
        i1 = pxy->v->Subtract(pxy, 42, 7);
        printf("ProxyTest2 Subtract(42-7)=%d. Should be %d.\n", i1, 35);
        CHECK(i1 == 35);

        i1 = 5;
        i1 = pxy->v->Add(pxy, 42, 7);
        printf("ProxyTest2 Add(42+7)=%d. Should be %d.\n", i1, 49);
        CHECK(i1 == 49);

        sc = pxy->v->AsyncTest(pxy, _T("foo"));
        CHECK(SUCCEEDED(sc));

        i1 = 5;
        i1 = pxy->v->Multiply(pxy, 42, 7);
        printf("ProxyTest2 Multiply(42*7)=%d. Should be %d.\n", i1, 294);
        CHECK(i1 == 294);
    }

    /* SIMPLE TYPES */
    i1 = 42;
    i2 = pxy->v->echoInteger(pxy, i1);
    printf("ProxyTest2 echoInteger %d. Should be %d.\n", i2, i1);
    CHECK(i1 == i2);

    i1 = 43;
    pxy->v->echoFloat(pxy, i1, &i2);
    printf("ProxyTest2 echoFloat %d. Should be %d.\n", i2, i1);
    CHECK(i1 == i2);

    Int32ToInt64(ll1, 42);
    pxy->v->echoDecimal(pxy, ll1, &ll2);
    printf("ProxyTest2 echoDecimal %lld. Should be %lld.\n", ll2, ll1);
    CHECK(Int64Equal(ll1,ll2));

    pxy->v->echoVoid(pxy);
    printf("ProxyTest2 echoVoid.\n");

    i1 = TRUE;
    pxy->v->echoBoolean(pxy, i1, &i2);
    printf("ProxyTest2 echoBoolean %d. Should be %d.\n", i2, i1);
    CHECK(i1 == i2);

    i1 = FALSE;
    pxy->v->echoBoolean(pxy, i1, &i2);
    printf("ProxyTest2 echoBoolean %d. Should be %d.\n", i2, i1);
    CHECK(i1 == i2);

    s1 = _T("foo");
    pxy->v->echoString(pxy, s1, sbuf, 10);
    printf("ProxyTest2 echoString [%s]. Should be [%s].\n",sbuf,s1);
    CHECK(!_tcscmp(sbuf, s1));

    t1 = CurrentTime();
    pxy->v->echoDate(pxy, t1, &t2);
    printf("ProxyTest2 echoDate [%016llx]. Should be [%016llx].\n",t2,t1);
    CHECK(Int64Equal(t1,t2));

#if 0
    s1 = "P10S";
    pxy->v->echoDate(pxy, s1, sbuf, 10);
    printf("ProxyTest2 echoDate period [%s]. Should be [%s].\n",sbuf,s1);
    CHECK(!_tcscmp(sbuf, s1));
#endif

    /* BASE64 */
    s1 = _T("");
    i1 = 0; /* XXX BUGBUG not properly zeroed by deserializer */
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    s1 = _T("!");
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    s1 = _T("!!");
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    s1 = _T("");
    i1 = 0; /* XXX BUGBUG not properly zeroed by deserializer */
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    s1 = _T("foo");
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    s1 = _T("foo!");
    sc = pxy->v->echoBase64(pxy, s1,_tcslen(s1)*sizeof(_TCHAR), sbuf, 99, &i1);
    printf("ProxyTest2 echoBase64 [%.*s]%d. Should be [%s] sc=%x\n",(int)i1,sbuf,i1,s1,sc);
    CHECK(i1 == _tcslen(s1) * sizeof(_TCHAR));
    CHECK(!_tcsncmp(sbuf, s1, i1 / sizeof(_TCHAR)));

    /* ARRAYS */
    memset(&oa, 0xcc, sizeof(oa));
    pxy->v->echoIntegerArray(pxy, ia, 4, oa, 5, &i1);
    printf("ProxyTest2 echoIntegerArray [%x %x %x %x]%d should be [%x %x %x %x]4\n", oa[0], oa[1], oa[2], oa[3], i1, ia[0], ia[1], ia[2], ia[3], 4);
    CHECK(i1 == 4);
    CHECK(ia[0] == oa[0]);
    CHECK(ia[1] == oa[1]);
    CHECK(ia[2] == oa[2]);
    CHECK(ia[3] == oa[3]);

    memset(&oa, 0xcc, sizeof(oa));
    pxy->v->echoFloatArray(pxy, ia2, 3, oa, 5, &i1);
    printf("ProxyTest2 echoFloatArray\n");
    CHECK(i1 == 3);
    CHECK(ia2[0] == oa[0]);
    CHECK(ia2[1] == oa[1]);
    CHECK(ia2[2] == oa[2]);

    pxy->v->echoStringArray(pxy, sia, 2, soa, 5, &i1);
    printf("ProxyTest2 echoStringArray [%s %s]\n", soa[0], soa[1]);
    CHECK(i1 == 2);
    CHECK(!_tcscmp(sia[0],soa[0]));
    CHECK(!_tcscmp(sia[1],soa[1]));

    pxyB->v->echo2DStringArray(pxyB, sia2, 2, soa, 5, &i1);
    printf("ProxyTest2 echo2DStringArray [%s %s]\n", soa[0], soa[1]);
    CHECK(i1 == 2);
    CHECK(!_tcscmp(sia2[0],soa[0]));
    CHECK(!_tcscmp(sia2[1],soa[1]));

    /* STRUCTS */
    memset(&to, 0xcc, sizeof(to));
    to.varString = ss1;
    pxy->v->echoStruct(pxy, &ti, &to);
    printf("ProxyTest2 echoStruct\n");
    CHECK(ti.varInt == to.varInt);
    CHECK(ti.varFloat == to.varFloat);
    CHECK(!_tcscmp(ti.varString,to.varString));
    
    memset(&tto, 0xcc, sizeof(tto));
    tto.varString = ss1;
    tto.varStruct.varString = ss2;
    pxyB->v->echoNestedStruct(pxyB, &tti, &tto);
    printf("ProxyTest2 echoNestedStruct\n");
    CHECK(tti.varInt == tto.varInt);
    CHECK(tti.varFloat == tto.varFloat);
    CHECK(!_tcscmp(tti.varString,tto.varString));
    CHECK(tti.varStruct.varInt == tto.varStruct.varInt);
    CHECK(tti.varStruct.varFloat == tto.varStruct.varFloat);
    CHECK(!_tcscmp(tti.varStruct.varString,tto.varStruct.varString));
    
    memset(&tao, 0xcc, sizeof(tao));
    tao.varString = ss1;
    tao.varArray[0] = ss2;
    tao.varArray[1] = ss3;
    pxyB->v->echoNestedArray(pxyB, &tai, &tao);
    printf("ProxyTest2 echoNestedArray\n");
    CHECK(tai.varInt == tao.varInt);
    CHECK(tai.varFloat == tao.varFloat);
    CHECK(!_tcscmp(tai.varString,tao.varString));
    CHECK(!_tcscmp(tai.varArray[0],tao.varArray[0]));
    CHECK(!_tcscmp(tai.varArray[1],tao.varArray[1]));

    pxyB->v->echoStructAsSimpleTypes(pxyB, &ti, sbuf,10, &i1, &i2);
    printf("ProxyTest2 echoStructAsSimpleTypes\n");
    CHECK(ti.varInt == i1);
    CHECK(ti.varFloat == i2);
    CHECK(!_tcscmp(ti.varString,sbuf));

    memset(&to, 0xcc, sizeof(to));
    to.varString = ss1;
    s1 = _T("foo"); i1 = 42; i2 = 77;
    pxyB->v->echoSimpleTypesAsStruct(pxyB, s1, i1, i2, &to);
    printf("ProxyTest2 echoSimpleTypesAsStruct\n");
    CHECK(to.varInt == i1);
    CHECK(to.varFloat == i2);
    CHECK(!_tcscmp(to.varString,s1));

    memset(&ato, 0xcc, sizeof(ato));
#if 0
    ato[0].varString = ss1;
    ato[1].varString = ss2;
    pxy->v->echoStructArray(pxy, ati,2, ato,2,&i1);
    printf("ProxyTest2 echoStructArray\n");
    CHECK(i1 == 2);
    CHECK(ati[0].varInt == ato[0].varInt);
    CHECK(ati[0].varFloat == ato[0].varFloat);
    CHECK(!_tcscmp(ati[0].varString,ato[0].varString));
    CHECK(ati[1].varInt == ato[1].varInt);
    CHECK(ati[1].varFloat == ato[1].varFloat);
    CHECK(!_tcscmp(ati[1].varString,ato[1].varString));
#endif
    }
    if (DoKey && 0) {
        sc = ((PPROXY)pxy)->pConnector->v->ClearKey(((PPROXY)pxy)->pConnector);
        printf("SerKey returned sc=%x\n", sc);
    }
#endif /* !_MINIMIZE */
    printf("proyxtest2: all tests completed OK\n");
    goto Out;

METHOD_CLEANUP;
    RELEASE(pxy);
    RELEASE(pxyB);
    RELEASE(soapfactory);
    RELEASE(convert);

    printf("done\n");

    FreeCmdLine(&argc,&argv);
    return 0;
}
