/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <sampleif.h>
#include <web/proxy.h>

#define MAX_URL_LEN     80

int _tmain(INT argc, _TCHAR *argv[])
{
#if !defined(_UNICODE)          /* XXX fix later */
    PISAMPLECALCULATOR pxy = NULL;
    PISOAPFACTORY soapfactory = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    PINAMERANGES nr = (PINAMERANGES)CurrentNameSpace();
    SCODE sc;
    char ServerURL[MAX_URL_LEN + 1] = "x-udp-aes-soap://";
    char *url = "x-udp-soap://224.1.2.1/COB/calc.cob"; //multicast for service discovery
    UINT size = 0;
    _TCHAR *ServiceURL, *ServiceName;

    ParseCmdLine(&argc,&argv);
    if (argc == 1) {
        ServiceURL = _T("x-udp-aes-soap://169.254.0.1/COB/calc.cob");
    } else if (argc == 2) {
#ifdef _UNICODE
        ServiceName = _wcstos(argv[1]);
#else
        ServiceName = argv[1];
#endif
    } else if (argc == 3) {
#ifdef _UNICODE
        ServiceURL = _wcstos(argv[1]);
#else
        ServiceURL = argv[1];
#endif
    } else {
        _tprintf(_T("Usage: proxytest3.exe <service name>\nor\nproxytest3.exe <service URL> -u\n"));
        return 0;
    }
    sc = BindToObject(ns, _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    if (FAILED(sc)) {
        printf("Could not bind to soap.cob sc=%x\n", sc);
        goto Out;
    }

    if (argc == 2) {
        sc = nr->v->LookupRange(nr, ServiceName, &IID_IUnknown, (ADDRESS*)&ServiceURL, &size);
        if (FAILED(sc)) {
            /* Not cached then ... */
            sc = soapfactory->v->CreateProxy(soapfactory, NULL,
                                             url, "tns",
                                             FALSE,
                                             (PIUNKNOWN*) &pxy);
            if (FAILED(sc)) {
                printf("CreateProxy %s failed sc=%x\n", url, sc);
                goto Out;
            }

            /* Call the proxy */
            printf("Discover service %s...\n",ServiceName);
            sc = pxy->v->DiscoverService(pxy, ServiceName, ServerURL + strlen(ServerURL), MAX_URL_LEN);
            printf("got %s sc=%x\n", ServerURL, sc);
            if (FAILED(sc)) 
                goto Out;
            pxy->v->Release(pxy);
            
            ServiceURL = _strdup(ServerURL);
            sc = nr->v->RegisterRange(nr, ServiceName, (PIUNKNOWN)nr, 
                                    NAME_SPACE_READ, (ADDRESS)ServiceURL, 
                                    strlen(ServiceURL), &IID_IUnknown);
            if (FAILED(sc)) 
                goto Out;
        }
    }
    sc = soapfactory->v->CreateProxy(soapfactory, NULL,
                                     ServiceURL, "tns",
                                     FALSE,
                                     (PIUNKNOWN*) &pxy);
    if (FAILED(sc)) {
        printf("CreateProxy %s failed sc=%x\n", url, sc);
        goto Out;
    }
    
    sc = ((PPROXY)pxy)->pConnector->v->SetKey(((PPROXY)pxy)->pConnector);
    if (FAILED(sc)) {
        if (argc == 2) {
            sc = nr->v->Unregister(nr, ServiceName);
            assert(sc==S_OK);
            free(ServiceURL);
        }
        goto Out;
    }

#if !_MINIMIZE
    {
    UINT sum;
    sum = pxy->v->Add(pxy, 12, 20);
    printf("SOAP request 1: 12 + 20 = %d (should be 32)\n", sum);

    sum = pxy->v->Add(pxy, 22, 20);
    printf("SOAP request 2: 22 + 20 = %d (should be 42)\n", sum);
    }
#endif

    sc = ((PPROXY)pxy)->pConnector->v->ClearKey(((PPROXY)pxy)->pConnector);
Out:
    if (pxy)
        pxy->v->Release(pxy);
    if (soapfactory)
        soapfactory->v->Release(soapfactory);

    printf("Done\n");
#ifdef _UNICODE
    if (argc == 2)
        free(ServiceName);
    else if (argc == 3) {
        free(ServiceURL);
#endif
    FreeCmdLine(&argc,&argv);
#endif
    return 0;
}
