/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include "ssoap.h"
#include <sampleif.h>
#include <web/udpsoapserv.h>

extern SCODE VerifyCertificate(PICYPHER Cypher, 
                       /*in*/BYTE *SignedCert,
                       /*out*/BYTE *UnsignedCert,
                       /*in*/PBSAFE_KEY PubKey);

int _tmain(INT argc, _TCHAR *argv[])
{
    PISAMPLECALCULATOR pxy = NULL;
    PISOAPFACTORY soapfactory = NULL;
    _TCHAR *target = _T(MulticastIPAddr2), *arg;
    _TCHAR url[80];
    SCODE sc;
    BOOL doRevoke = FALSE;
    
    ParseCmdLine(&argc,&argv);

    if (argc < 1 || argc > 2) {
        _tprintf(_T("Usage: sectest.exe [-r]\n"));
        goto Out;
    }
    if (argc == 2) {
        arg = argv[1];
        if (arg[0] == '-' && arg[1] == 'r')
            doRevoke = TRUE;
    }
    _tcscpy(url, _T("x-udp-soap://"));
    _tcscat(url, target);
    _tcscat(url, _T("/COB/calc.cob"));

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    if (FAILED(sc)) {
        printf("Could not bind to soap.cob sc=%x\n", sc);
        goto Out;
    }

    //printf("Create a proxy url=%s\n",url);
    sc = soapfactory->v->CreateProxy(soapfactory, NULL,
                                     url, _T("tns"),
                                     FALSE,
                                     (PIUNKNOWN*) &pxy);
    if (FAILED(sc)) {
        printf("CreateProxy %s failed sc=%x\n", url, sc);
        goto Out;
    }
#if 1
    {
        /* Contruct the URL of the seal's soap server */
        _tcscpy(url, _T("x-udp-soap://"));
        _tcscat(url, _T(MulticastIPAddr2));
        _tcscat(url, _T("/COB/calc.cob"));
        
        if (doRevoke) {
            sc = pxy->v->Revoke(pxy, url);
            if (sc == E_OBJECT_NOT_INITIALIZED)
                printf("sectest.c: Have no owner yet!\n");/* sc=x%x\n",sc); */
            else if (FAILED(sc))
                printf("sectest.c: Revoke failed! sc=x%x\n",sc);
            //else 
              //  printf("sectest.c: Revoke is done! sc=x%x\n",sc);
        } else {
            sc = pxy->v->InitOwner(pxy, url);
            if (sc == E_ALREADY_EXISTS)
                printf("sectest.c: Already have an owner! sc=x%x\n",sc);
            else if (FAILED(sc))
                printf("sectest.c: InitOwner failed! sc=x%x\n",sc);
            //else 
              //  printf("sectest.c: InitOwner is done! sc=x%x\n",sc);
        }
    }
#else
    {
        SELFINFO Self;
        PICYPHER Cypher;
        UINT i;
        BYTE signedcert[CERTSIZE]; //CERTSIZE is the size of the certificate
        CERTIFICATE unsignedcert;
        struct {
            PUB_KEY Key;
            UINT32 pad[2];
        } Pub_key;
  
        /* Test GetPubKey */
        memset(&Pub_key,0,sizeof(Pub_key));
        sc = pxy->v->GetPubKey(pxy, (UINT32*)&Pub_key.Key);
        if (FAILED(sc))
            printf("GetPubKey failed 0x%x\n", sc);
        printf("Got the pub key: ");
        for(i=0;i<sizeof(Pub_key)/4;i++)
            printf(" x%x",((UINT32*)&Pub_key)[i]);
        printf("\n");

        /* Test GetCertificate */
        sc = pxy->v->GetCertificate(pxy, signedcert);
        if (FAILED(sc))
            printf("GetCertificate failed 0x%x\n", sc);
        printf("Got cert = ");
        for (i=0;i<CERTSIZE;i++)
            printf("%x ", signedcert[i]);
  
        /* Initialize the rsa cypher */
        Cypher = GetCypher("COB/rsa.cob",RSAKEYSTRENGTH);
        if (Cypher == NULL) {
            DBG(("Failed in creating cypher COB/rsa.cob\n"));
            goto ErrorOut;
        }

        /* Test SignCertificate */
        /* Initialize a certificate */
        memset(&Self, 0, sizeof(SELFINFO));
        sc = Cypher->v->GenKey(Cypher, RSAKEYSTRENGTH,
                               (void*)&Self.PriKey, &Self.PriKeySize,
                               (void*)&Self.PubKey, &Self.PubKeySize);
        if (FAILED(sc)) {
            printf("GenRsaKeyPair failed\n");
            goto ErrorOut;
        }

        _tcscpy((char*)Self.Cert.ID, "seal");
        _tcscpy((char*)Self.Cert.Name, "seal");
        memcpy(&Self.Cert.PubKey, Self.PubKey, sizeof(PUB_KEY));
        sc = pxy->v->SignCertificate(pxy, &Self.Cert, Self.SignedCert);
        if (FAILED(sc)) {
            printf("SignCertificate failed 0x%x\n", sc);
            goto ErrorOut;
        }
        
        sc = VerifyCertificate(Cypher, (BYTE*)Self.SignedCert, 
                             (BYTE*)&unsignedcert, (PBSAFE_KEY)&Pub_key);
        if (FAILED(sc)) {
            printf("VerifyCertificate failed 0x%x\n", sc);
            goto ErrorOut;
        }
        assert(!memcmp(&Self.Cert, &unsignedcert, sizeof(CERTIFICATE)));

ErrorOut:
        if (Self.PriKey)
            free(Self.PriKey);
        if (Self.PubKey)
            free(Self.PubKey);
    }
#endif

Out:
    if (pxy)
        pxy->v->Release(pxy);
    if (soapfactory)
        soapfactory->v->Release(soapfactory);
    //printf("done\n");

    FreeCmdLine(&argc,&argv);
    return 0;
}
