/* Copyright (c) Microsoft Corporation. All rights reserved. */
#ifndef __SOAP__
#define __SOAP__ 1

#define USE_SERVER_FOR_RESPONSE 0

#include <mmlite.h>
#include <crypto.h>
#include <heaps/dbg_heap.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <web/cont.h>
#include <web/typeinfo.h>
 
struct HREF {
    _TCHAR *ID;
    void *addr;             /* reference */
    char *data;             /* actual data */
    PUINT pStoredSize;

    /* Saved arg info */
    CPXMLNAMESPACE Xns;         /* namespace of output struct */
    CPARGUMENT_DESCR argd;      /* to be deleted */
    CPMETHOD_DESCR structd;
    DESCR_TYPE type;
    ARGUMENT_DESCR_FLAGS flags; /* original in/out/inout | current arg/field */
    UINT ListNodeOffset;        /* when list (combine with ArrayXXX?) */
    UINT32 *FieldSelector;      /* Selectively output fields */
    UINT xmldepth;              /* Where this href was created, for nested lists */

    /* For message generator in soap.c */
    INT ArrayIndex;
    UINT ArraySize;

    PHREF next;
};

SCODE InitTypeInfo(struct TypeInfo **pTypeInfo);
SCODE DestroyTypeInfo(struct TypeInfo *pTypeInfo);
SCODE GetTypeInfo(struct TypeInfo **pRetval);

PIUNKNOWN GetPeerState(void);

#ifdef _UNICODE   /* This goes into tchar.h */
#define _i64totcs _i64towcs
#define _uitotcs _uitowcs
#else
#define _i64totcs _i64tostr
#define _uitotcs _uitostr
#endif

INT32         strump(ADDRESS, ADDRESS, UINT RegArgCount);
char *       itoa(UINT i, char* buf);
/*PRIVATE*/ _TCHAR * _uitotcs(UINT Val, UINT Base, _TCHAR * pBuf);
/*PRIVATE*/ _TCHAR * _i64totcs(INT64 Val, UINT Base, _TCHAR * pBuf);

SCODE MCT CreateSerializer (PIFILE, PARSE_URL_FLAGS, PISOAPSERIALIZER*);


SCODE base64Coder(PTR pThis,
    /*in*/ CYPHER_OPS Operation,
    /*in*/ const BYTE * pInBuffer,
    /*in*/ UINT InSize,
    /*out*/ UINT * pSizeConsumed,
    /*out*/ BYTE * pOutBuffer,
    /*in*/ UINT OutSize,
    /*out*/ UINT * pSizeProduced);

SCODE TimeParse(const _TCHAR *buf, TIME *tret);
void TimeFormat(TIME t, _TCHAR *buf, UINT Len);

SCODE CreateProxy(PIHTTPCONNECTOR pConn, const _TCHAR *Uri, const _TCHAR *XnsName, BOOL Future, PIUNKNOWN *ppProxy);
SCODE CreateProxy2(PIHTTPCONNECTOR pConn, const _TCHAR *Uri, CPINTERFACE_DESCR, CPXMLNAMESPACE ns, PIUNKNOWN *ppProxy);
SCODE CreateProxy3(PIHTTPCONNECTOR pConn, PWSCONTEXT pContext, const _TCHAR *XnsName, BOOL Future, PIUNKNOWN *ppProxy);

SCODE CreateFuture(PIHTTPCONNECTOR pConn,
                   PWSCONTEXT pContext,
                   CONT_FLAGS Flags,
                   const void * Frame,
                   const _TCHAR *FrameType,
                   /*out*/ PICONTINUATION *ppCont);

SCODE CreateCLRContinuation(PWSCONTEXT pContext,
                   CONT_FLAGS Flags,
                   const _TCHAR *FrameType,
                   /*out*/ PICONTINUATION *ppCont
                   );


PTR CopyReservation(PTR pItem, PTR pArg);
INT DeleteReservation(PTR via, PTR pArg);

CPXMLNAMESPACE GetNamespace(const _TCHAR *prefix, UINT len);
CPMETHOD_DESCR FindMethod(CPXMLNAMESPACE Xns, const _TCHAR *Name, BOOL IsM);


SCODE InitContinuation(PCONTINFO cont, CONT_FLAGS Flags, const _TCHAR *EnvName);
SCODE CreateContinuation(CONT_FLAGS Flags, PICONTINUATION *pCont);
PTR ActivationSpaceAlloc(PCONTINFO cont, UINT Size);
PTR ActivationSpaceRealloc(PCONTINFO cont, PTR old, UINT NewSize);
SCODE ActivationSpaceSetBuffer(PCONTINFO cont, PTR Where, UINT sizeis);
void DestroyNameSpaces(PXMLNAMESPACE *ppns, PIHEAP Heap);
void DestroyRouting(PTR pvia, PIHEAP Heap);

PHREF MakeHref(PHREF *pphref, PIHEAP Heap,
               const _TCHAR *value, UINT valueLength,
               PTR addr, PTR data, UINT xmldepth, PUINT pStoredSize,
               CPARGUMENT_DESCR argd, PHREF parent,
               ARGUMENT_DESCR_FLAGS Flags,
               UINT id);
void DestroyHRef(PHREF phref, PIHEAP Heap);

PIHTTPPROCESS SOAPBoxCreate(void);

TIME GetSoapTimeOffset(void);

SCODE ServeSoapRequest(UINT Flags, PIBUFFER Inout, /*inout*/PIBUFFER *pOutput,
                       void *pSKey, PIHTTPCONNECTOR Request);

SCODE MCT CreateDeserializer(PISAXHANDLER *SaxHandler, struct TypeInfo *ptypeinfo,
                             const _TCHAR * reqLocalURL,
                             PIHEAP Heap,
                             PICONTINUATION cont); /* NULL for server */

void BaxTest(void); /* XXX temp */
SCODE BResolveName(/* inout */ PSAXATTRIBUTE Entry);

/* Exported for the benefit of bax.c */
SCODE RStartElement(PISAXHANDLER iThis, CPSAXATTRIBUTE Elem,
                    CPSAXATTRIBUTE Attr, UINT AttrCount);
SCODE MCT REndElement(PISAXHANDLER This, CPSAXATTRIBUTE Elem);
CPXMLNAMESPACE LookupNsQName(PISAXHANDLER iThis, PSTR_ENTRY QName);
CPXMLNAMESPACE FindXnsByPrefix(CPNS_TABLE nst, const _TCHAR *value, UINT len);

#endif /* __SOAP__ */
