/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <base/cobdesc.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <netif.h>
#include <tokenizerif.h>
#include "soap.h"

#include <web/udpsoapserv.h>

typedef struct _SoapFactory {
    struct _DefCob super;

    /* The is the offset from the local system clock to the
       authoritative wall clock (in UTC). A forward offset is 
       a positive number and indicates that the local clock is
       behind the wall clock */
    INT64 TimeOffset; 
} SoapFactory, *PSOAPFACTORY;

/* XXX should be generated and may be wrong size XXX */
extern const CLASSDESC(CSoapSerializer, 3);
extern const CLASSDESC(CSoapSerializer2, 3);
extern const CLASSDESC(Deserializer, 3);
/*extern*/ const CLASSDESC(CContinuation, 3);
#define _MANUAL_COBMAIN

#include <s_SoapFactory_v.c>

#if 0
extern const struct XMLNAMESPACE IWsAddressing_ifd_namespace_wsa;

PRIVATE SCODE SOAPFactoryCreateSerializer (
    PISOAPFACTORY Factory,
    PIFILE Output_XML_File,
    PARSE_URL_FLAGS Flags,
    /*out*/ PISOAPSERIALIZER* serializer)
{
    return CreateSerializer(Output_XML_File, Flags, serializer);
}

PRIVATE SCODE SOAPFactoryCreateDeserializer(PISOAPFACTORY Factory,
                                        const _TCHAR * reqUri, UINT reqUriLength,
                                        PICONTINUATION pCont, /* NULL for server */
                                        PISAXHANDLER *ppHandler)
{
    struct TypeInfo *ptypeinfo;
    SCODE sc = GetTypeInfo(&ptypeinfo);
    PIHEAP Heap = _GetHeap(_T("Deserializer"), 4000);
    if (FAILED(sc)) {
        DBG(("SOAPFactoryCreateDeserializ: InitTypeInfo fail sc=%x\n",sc));
        return sc;
    }

    return CreateDeserializer(ppHandler, ptypeinfo,
                              reqUri, Heap, pCont);
}
#endif

/* If pConn is not NULL, then use it as the ISOAPCONNECTOR
 * instead of creating a new ISOAPCONNECTOR.
 */
PRIVATE SCODE SOAPFactoryCreateProxy(PISOAPFACTORY Factory,
                                     PIHTTPCONNECTOR pConn,
                                     const _TCHAR *Uri,
                                     const _TCHAR *XnsName,
                                     BOOL Future,
                                     PIUNKNOWN *ppProxy)
{
    return CreateProxy(pConn, Uri, XnsName, Future, ppProxy);
}

PRIVATE SCODE SOAPFactoryCreateProxy2(PISOAPFACTORY Factory,
                                      PIHTTPCONNECTOR pConn,
                                      const _TCHAR *Uri,
                                      PTR Iface,
                                      CPXMLNAMESPACE ns, 
                                      PIUNKNOWN *ppProxy)
{
    return CreateProxy2(pConn, Uri, (PINTERFACE_DESCR)Iface, ns, ppProxy);
}

PRIVATE SCODE SOAPFactoryCreateProxy3(PISOAPFACTORY Factory, 
                                      PIHTTPCONNECTOR pConn, 
                                      PTR pContext,
                                      const _TCHAR *XnsName, 
                                      BOOL Future,
                                      PIUNKNOWN *ppProxy)
{
    return CreateProxy3(pConn, pContext, XnsName, Future, ppProxy);
}

PRIVATE SCODE SOAPFactoryCreateFuture(PISOAPFACTORY Factory,
                                      PIHTTPCONNECTOR pConn,
                                      PTR pContext,
                                      CONT_FLAGS Flags,
                                      const void * Frame,
                                      const _TCHAR *FrameType,
                                      /*out*/ PICONTINUATION *ppCont)
{
    return CreateFuture(pConn, pContext, Flags, Frame, FrameType, ppCont);
}

PRIVATE SCODE SOAPFactoryCreateCLRContinuation(PISOAPFACTORY Factory,
                                      PTR pContext,
                                      CONT_FLAGS Flags,
                                      const _TCHAR *FrameType,
                                      /*out*/ PICONTINUATION *ppCont
                                      )
{
    return CreateCLRContinuation(pContext, Flags, FrameType, ppCont);
}

SCODE SOAPFactoryParseURL(
        PISOAPFACTORY Factory,
        const _TCHAR *Url,
        /*out*/ UINT *pFlags,
        /*out*/ const _TCHAR **pHost,
        /*out*/ UINT *pHostLen,
        /*out*/ UINT16 *pPort,
        /*out*/ const _TCHAR **pPath,
        /*out*/ UINT *pPathLen,
        /*out*/ const _TCHAR **pQuery,
        /*out*/ UINT *pQueryLen,
        /*out*/ const _TCHAR **pFrag)
{
    UINT Flags = 0, len;
    UINT16 port;
    BOOL IsFile = FALSE;
    const _TCHAR *host = _T("");

  /* This assertion has been commented because in wsman.c only Local Path (pPath) was required */
  //  assert(!pPath || pPathLen); /* tmp: detect obsolete use */

    while (Url[0] == _T(' '))
        Url++;

    /* Parse type string:
     * http|soap.udp|xml/soap|xml/text|x-[udp-|tcp-|http-|file-][xte-][aes-](dime|soap)
     */
    if (!_tcsncmp(Url, _T("x-"), 2)) {
        Url += 2;

        if (!_tcsncmp(Url, _T("udp-pm"), 6)) {
            /* x-udp-pm shorthand for public multicast soap */
            Url += 6;
            Flags |= (PARSE_URL_FLAGS_UDP | PARSE_URL_FLAGS_SOAP
                      | PARSE_URL_FLAGS_AES);
            host = _T(MulticastIPAddr1);
        } else if (!_tcsncmp(Url, _T("udp-sm"), 6)) {
            /* x-udp-pm shorthand for secret multicast soap */
            Url += 6;
            Flags |= PARSE_URL_FLAGS_UDP | PARSE_URL_FLAGS_SOAP;
            host = _T(MulticastIPAddr2);
        } else if (!_tcsncmp(Url, _T("udp-"), 4)) {
            Url += 4;
            Flags |= PARSE_URL_FLAGS_UDP;
        } else if (!_tcsncmp(Url, _T("tcp-"), 4)) {
            Url += 4;
        } else if (!_tcsncmp(Url, _T("http-"), 5)) {
            Url += 5;
            Flags |= PARSE_URL_FLAGS_HTTP;
        } else if (!_tcsncmp(Url, _T("file-"), 5)) {
            Url += 5;
            //Flags |= PARSE_URL_FLAGS_FILE;
            IsFile = TRUE;
        }

        if (!_tcsncmp(Url, _T("xte-"), 4)) {
            Url += 4;
            Flags |= PARSE_URL_FLAGS_XTE;
        }

        if (!_tcsncmp(Url, _T("aes-"), 4)) {
            Url += 4;
            Flags |= PARSE_URL_FLAGS_AES;
        }

        if (!_tcsncmp(Url, _T("dime"), 4)) {
            Url += 4;
            Flags |= PARSE_URL_FLAGS_DIME;
        } else if (!_tcsncmp(Url, _T("soap"), 4)) {
            Url += 4;
            Flags |= PARSE_URL_FLAGS_SOAP;
        } else if (!(Flags & PARSE_URL_FLAGS_SOAP))
            return E_INVALID_PARAMETER;

    } else if (!_tcsncmp(Url, _T("http"), 4)) {
        Url += 4;
        Flags |= PARSE_URL_FLAGS_HTTP;
    } else if (!_tcsncmp(Url, _T("xml/soap"), 8) || !_tcsncmp(Url, _T("text/xml"), 8)) {
        Url += 8;
        Flags |= PARSE_URL_FLAGS_HTTP | PARSE_URL_FLAGS_SOAP;
        port = 80;
    } else if (!_tcsncmp(Url, _T("soap.udp"), 8)) {
        Url += 8;
        port = SOAP_SERVER_PORT_UDP;
        Flags |= PARSE_URL_FLAGS_UDP | PARSE_URL_FLAGS_SOAP;
    } else if (!_tcsncmp(Url, _T("uuid"), 4)) {
        Url += 4;
        Flags |= PARSE_URL_FLAGS_UUID | PARSE_URL_FLAGS_SOAP;
        IsFile = TRUE;
    } else
        return E_INVALID_PARAMETER;

    /* Figure out which port to use */
    if (Flags & PARSE_URL_FLAGS_DIME)
        port = SOAP_SERVER_PORT_UDP_DIME;
    else if (Flags & PARSE_URL_FLAGS_UDP) {
        port = SOAP_SERVER_PORT_UDP;
        if (Flags & PARSE_URL_FLAGS_XTE)
            port = SOAP_SERVER_PORT_UDP_XTE;
        if (Flags & PARSE_URL_FLAGS_AES) {
            port = SOAP_SERVER_PORT_UDP_AES;
            if (Flags & PARSE_URL_FLAGS_XTE)
                port = SOAP_SERVER_PORT_UDP_XTE_AES;
        }
    } else if (Flags & PARSE_URL_FLAGS_HTTP) {
        port = 80;
    } else {
        port = 9020;
        //port = SOAP_SERVER_PORT_TCP;
    }

    if ((Flags & PARSE_URL_FLAGS_UUID) && !_tcsncmp(Url, _T(":"), 1)) {
        Url += 1;
    } else if (!_tcsncmp(Url, _T("://"), 3)) {
        Url += 3;
    } else if (!*Url) {
        if (pFlags)
            *pFlags = Flags;
        if (pHost)
            *pHost = host;
        if (pHostLen)
            *pHostLen = _tcslen(host);
        if (pPort)
            *pPort = port;
        if (pPath)
            *pPath = _T("");
        if (pPathLen)
            *pPathLen = 0;
        if (pQuery)
            *pQuery = _T("");
        if (pQueryLen)
            *pQueryLen = 0;
        if (pFrag)
            *pFrag = _T("");
        return S_OK;
    } else
        return E_INVALID_PARAMETER;

    if (IsFile) {
        if (pHost)
            *pHost = NULL;
        if (pHostLen)
            *pHostLen = 0;
        if (pPort)
            *pPort = 0;
    } else {
        if (pHost)
            *pHost = Url;
        len = _tcscspn(Url, _T(":/"));
        if (pHostLen)
            *pHostLen = len;
        Url += len;
        if (*Url == ':') {
            Url++;
			if (isdigit(*(Url+1))) {
				port = _ttoi(Url);
				Url += _tcsspn(Url, _T("0123456789"));
			}
        }
        if (*Url) {               /* skip slash */
            if (*Url != _T('/'))
                return E_INVALID_PARAMETER;
            Url++;
        }

        if (pPort)
            *pPort = port;
    }

    if (pFlags)
        *pFlags = Flags;

    if (pPath)
        *pPath = Url;

    if (pPathLen)
        *pPathLen = 0;
    if (pQuery)
        *pQuery = _T("");
    if (pQueryLen)
        *pQueryLen = 0;
    if (pFrag)
        *pFrag = _T("");

    if (!Url || !Url)
        return S_OK;

    len = _tcscspn(Url, _T("?#"));
    if (pPathLen)
        *pPathLen = len;

    Url += len;
    if (!*Url)                  /* no QUERY or FRAG */
        return S_OK;

    if (*Url == _T('?')) {
        Url++;
        if (pQuery)
            *pQuery = Url;
        len = _tcscspn(Url, _T("#"));
        if (pQueryLen)
            *pQueryLen = len;
        Url += len;
        if (!*Url)              /* no FRAG */
            return S_OK;
    }
    Url++;

    if (pFrag)
        *pFrag = Url;

    return S_OK;
}

PRIVATE SCODE SOAPFactorySetTimeOffset(PISOAPFACTORY Factory, 
                                       TIME Off,
                                       BOOL AddWhenSerializing)
{
    PSOAPFACTORY This = (PSOAPFACTORY) Factory;

#if 0 /* Fixing this will likely break any callers */
    if (!AddWhenSerializing)
        Off = Int64Subtract(TIME_ORIGIN, Off);
    if (!TimeIsNotPositive(Off))
        return E_INVALID_PARAMETER;
#endif

    This->TimeOffset = Off;
    return S_OK;
}

#if 0
#define use_bax 0

#if use_bax
SCODE XBaxParse(PITOKENIZER Tokenizer, PISAXHANDLER Handler);
#else
#define XBaxParse(Tokenizer, Handler) E_NOT_IMPLEMENTED
#endif

PRIVATE SCODE SOAPFactoryProcessRequest (
    PISOAPFACTORY iThis,
    PARSE_URL_FLAGS Flags,
    PITOKENIZER Tokenizer,
    const _TCHAR *LocalUrl,
    PISAXHANDLER *pDeserializer, /* inout for restartability */
    PICONTINUATION *pCont)
{
    METHOD_DECLARE(SoapFactory, ProcessRequest);
    PISAXPARSER SAXParser = NULL;
    PIHEAP Heap = _GetHeap(_T("Handler"), 4000);
    PCONTINFO cont = NULL;
    struct TypeInfo *pTypeInfo = NULL;
    PDESERIALIZER Handler = NULL;

    TRACE("SOAPHTTPProcessPost");
    assert(Tokenizer != NULL);

    if (!*pDeserializer) {
        /* Fetch the Typeinfo from the input interface definition file */
        sc = GetTypeInfo(&pTypeInfo);
        CHECKSC("GetTypeInfo");

        sc = CreateDeserializer(pDeserializer, pTypeInfo,
                                NULL, Heap, NULL);
        CHECKSC("CreateDeserializer");
    }

    if (Flags & PARSE_URL_FLAGS_XTE) {
        sc = XBaxParse(Tokenizer, *pDeserializer);
    } else {
        /* Construct the SAX parser */
        sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"),NAME_SPACE_READ,
                          &IID_ISaxParser, (void**) &SAXParser);
        CHECKSC("bind to sax.cob");

        sc = SAXParser->v->Parse(SAXParser, Tokenizer, *pDeserializer,
                                 &((PDESERIALIZER)*pDeserializer)->SaxState);
    }

    CHECKSC("Request message parsing");

    Handler = (PDESERIALIZER) *pDeserializer;
    if (Handler) {
        cont = pCONT(Handler->prpcinfo);
        CHECKSETSC(cont, E_UNEXPECTED_ERROR, "exists(cont)");

        assert(cont->Method && cont->Frame
               && !cont->Request && !cont->namespaces);

        /* Save into cont what we need later */
        /* Now done here instead of deserializer.c after CreateAppStack */
        cont->namespaces = Handler->namespaces;
        Handler->namespaces = NULL;

        if (Handler->envelopeNS)
            cont->envelopeNS = Handler->envelopeNS;

        assert (Handler->ultimateReceiver == TRUE); /* routing disabled */
        *pCont = (PICONTINUATION) cont;
        cont->v->AddRef(iCONT(cont)); /* XXX might as well zap from Handler? */
    }

METHOD_CLEANUP;
    CHECK_HEAP();
    RELEASE(SAXParser);
    return sc;
}

PRIVATE SCODE SOAPFactoryProcessRequestFile (
    PISOAPFACTORY iThis,
    PARSE_URL_FLAGS Flags,
    PIFILE InputFile,
    PIFILE OutputFile,
    const _TCHAR *LocalUrl)
{
    METHOD_DECLARE(SoapFactory, ProcessRequestFile);
    PITOKENIZER Tokenizer = NULL;
    PITOKENIZERFACTORY TokenizerFactory = NULL;
    PIUNKNOWN Deserializer = NULL;
    PCONTINFO cont = NULL;

#if 0
    if (Flags & PARSE_URL_FLAGS_AES) {
        DBG(("soapreq: Decrypt the file...\n"));
        sc = AESBuffer(pSKey->KeyStrength, pSKey->Key,
                       Input, CYPHER_DECRYPT);
        CHECKSC("AESBuffer");
    }
#endif

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &TokenizerFactory);
    CHECKSC("bind to tokenizer.cob");

    sc = TokenizerFactory->v->Create(TokenizerFactory, (PIUNKNOWN) InputFile,
                                     &Tokenizer);
    CHECKSC("Create tokenizer");

    sc = SOAPFactoryProcessRequest(iThis, Flags, Tokenizer, LocalUrl,
                                   (PISAXHANDLER *) &Deserializer,
                                   (PICONTINUATION*) &cont);
    CHECKSC("ProcessRequest");

    sc = cont->v->Execute(iCONT(cont), FALSE, NULL);
    assert(sc == S_NO_RESPONSE || FAILED(sc));
    CHECKSC("cont->Execute");

    /* XXX If response is requested, create request obj, prepare to execute, then wait */

METHOD_CLEANUP;
    RELEASE(Tokenizer);
    RELEASE(TokenizerFactory);
    RELEASE(Deserializer);
    return sc;
}

PRIVATE SCODE SOAPFactoryProcessRequestString (
    PISOAPFACTORY iThis,
    PARSE_URL_FLAGS Flags,
    const _TCHAR *InputData,
    const _TCHAR *LocalUrl)
{
    METHOD_DECLARE(SoapFactory, ProcessRequestString);
    PIENDPOINTFACTORY netfactory = NULL;
    PIBUFFER Input = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    UINT InputDataSize = _tcslen(InputData) * sizeof(_TCHAR);

    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &netfactory);
    CHECKSC("load protocol.cob");

    /* Ask network stack to create "temp file" */
    sc = netfactory->v->AllocateBuffer(netfactory, InputDataSize, NULL,
                                       &Input);
    CHECKSC("AllocateBuffer");

    /* Copy plain buffer into it */
    sc = Input->v->WriteAt(Input, UINT64_ZERO,
                           (BYTE*) InputData, InputDataSize,
                           &InputDataSize);
    CHECKSC("copy plain buffer to IBuffer");

    sc = SOAPFactoryProcessRequestFile(iThis, Flags, (PIFILE) Input,
                                       NULL, LocalUrl);
    CHECKSC("ProcessRequestFile");

METHOD_CLEANUP; 
    RELEASE(netfactory);
    RELEASE(Input);
    return sc;
}

PRIVATE SCODE SOAPFactoryProcessData(
    PISOAPFACTORY iThis,
    PARSE_URL_FLAGS Flags,
    PITOKENIZER Tokenizer,
    PISAXHANDLER *pDeserializer,
    PIHEAP Heap,
    PAnyItem Item)
{
    METHOD_DECLARE(SoapFactory, ProcessData);
    struct TypeInfo *pTypeInfo = NULL;
    PDESERIALIZER Handler = NULL;
    PISAXPARSER SAXParser = NULL;
    CPMETHOD_DESCR metd;
    CPARGUMENT_DESCR argd;
    UINT i;

    if (!*pDeserializer) {
        /* Fetch the Typeinfo from the input interface definition file */
        sc = GetTypeInfo(&pTypeInfo);
        CHECKSC("GetTypeInfo");

        sc = CreateDeserializer(pDeserializer, pTypeInfo,
                                NULL, Heap, NULL);
        CHECKSC("CreateDeserializer");

        Handler = (PDESERIALIZER) *pDeserializer;
        Handler->current->State = DS_BODY;

        metd = FindMethod(&IWsAddressing_ifd_namespace_wsa,
                          _T("AnyItem"), FALSE);
        Handler->current->SeenItems = Handler->current->_SeenItems;

        for (i = 0; i < metd->table_size; i++) {
            argd = &metd->ptable[i];
            if (argd->descr_type == ARG_ANYTYPE)
                break;
        }

        Handler->current->href = MakeHref(&Handler->hreflist, Handler->Heap,
                                          _T("1"), 1,
                                          Item, NULL, NULL,
                                          argd, NULL, 0, 0);

#define STARTED     0x40000000
        Handler->SaxState = STARTED;

        /* Also make the current list href point to the same */
        Handler->current->listhref.ID = _T("#list7");
        Handler->current->listhref.addr = Item;
        Handler->current->listhref.argd = argd;

        /* XXX Undo temporary heap that was created and replace with given */
        {
            PCONTINFO cont = pCONT(Handler->prpcinfo);
            RELEASE(cont->TmpHeap);
            cont->TmpHeap = Heap;
            if (Heap)
                Heap->v->AddRef(Heap);
        }
        
    }

    if (Flags & PARSE_URL_FLAGS_XTE) {
        sc = XBaxParse(Tokenizer, *pDeserializer);
    } else {
        /* Construct the SAX parser */
        sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"),NAME_SPACE_READ,
                          &IID_ISaxParser, (void**) &SAXParser);
        CHECKSC("bind to sax.cob");

        sc = SAXParser->v->Parse(SAXParser, Tokenizer, *pDeserializer,
                                 &((PDESERIALIZER)*pDeserializer)->SaxState);
    }

    CHECKSC("Parse");

METHOD_CLEANUP;
    CHECK_HEAP();
    RELEASE(SAXParser);
    return sc;
}

PRIVATE SCODE SOAPFactoryProcessFragmentFile (
    PISOAPFACTORY iThis,
    PIFILE InputFile,

    PIHEAP Heap,
    PTR *pData,
    _TCHAR *pType)
{
    METHOD_DECLARE(SoapFactory, ProcessFragmentFile);
    PITOKENIZER Tokenizer = NULL;
    PITOKENIZERFACTORY TokenizerFactory = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    PISAXHANDLER Deserializer = NULL;
    struct AnyItem Item;

    memset(&Item, 0, sizeof(Item));

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &TokenizerFactory);
    CHECKSC("bind to tokenizer.cob");

    sc = TokenizerFactory->v->Create(TokenizerFactory, (PIUNKNOWN) InputFile,
                                     &Tokenizer);
    CHECKSC("Create tokenizer");

    sc = SOAPFactoryProcessData(iThis, 0, Tokenizer, &Deserializer,
                                Heap, &Item);
    CHECKSC("ProcessData");

    *pData = Item.Value;
    if (Item.ValueType && pType)
        _tcscpy(pType, Item.ValueType);

    /* Do not leak ValueType string if this was a permanent heap */
    if (Item.ValueType && Heap)
        Heap->v->Free(Heap, 0, Item.ValueType);

METHOD_CLEANUP; 
    CHECK_HEAP();
    RELEASE(Deserializer);
    RELEASE(Tokenizer);
    RELEASE(TokenizerFactory);
    return sc;
}

PRIVATE SCODE SOAPFactorySerializeFragmentFile (
    PISOAPFACTORY iThis,
    PTR Data,
    const _TCHAR *Type,
    PIFILE OutputFile)
{
    METHOD_DECLARE(SoapFactory, SerializeFragmentFile);
    PINAMESPACE ns = CurrentNameSpace();
    struct AnyItem Item;

    memset(&Item, 0, sizeof(Item));
    Item.Value = Data;
    Item.ValueType = (_TCHAR *) Type;
    ListNodeInit(offsetof(struct AnyItem, ListNode), &Item);

    sc = E_NOT_IMPLEMENTED;

METHOD_CLEANUP; 
    CHECK_HEAP();
    return sc;
}

PRIVATE SCODE SOAPFactoryProcessFragment (
    PISOAPFACTORY iThis,
    const _TCHAR *InputData,

    PIHEAP Heap,
    PTR *pData,
    _TCHAR *pType)
{
    METHOD_DECLARE(SoapFactory, ProcessFragment);
    PIENDPOINTFACTORY netfactory = NULL;
    PIBUFFER Input = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    UINT InputDataSize = _tcslen(InputData) * sizeof(_TCHAR);

    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &netfactory);
    CHECKSC("load protocol.cob");

    /* Ask network stack to create "temp file" */
    sc = netfactory->v->AllocateBuffer(netfactory, InputDataSize, NULL,
                                       &Input);
    CHECKSC("AllocateBuffer");

    /* Copy plain buffer into it */
    sc = Input->v->WriteAt(Input, UINT64_ZERO,
                           (BYTE*) InputData, InputDataSize,
                           &InputDataSize);
    CHECKSC("copy plain buffer to IBuffer");

    sc = SOAPFactoryProcessFragmentFile(iThis, (PIFILE) Input, Heap,
                                        pData, pType);
    CHECKSC("ProcessFragmentFile");

METHOD_CLEANUP; 
    CHECK_HEAP();
    RELEASE(netfactory);
    RELEASE(Input);
    return sc;
}
#endif

PRIVATE SCODE SOAPFactoryCreateContinuation (
    PISOAPFACTORY iThis,
    CONT_FLAGS Flags,
    PICONTINUATION *pCont)
{
    METHOD_DECLARE(SoapFactory, CreateContinuation);
    sc = CreateContinuation(Flags, pCont);

METHOD_CLEANUP; 
    return sc;
}

/* SOAP box construction/destruction */
PRIVATE SCODE SoapFactoryDestroy(PIUNKNOWN Factory)
{
    TRACE("SOAPFactoryDestroy");
    /* Nothing to do here */
    return S_OK;
}

PRIVATE SoapFactory *pTheSoapFactory = NULL;

#if 0
TIME GetSoapTimeOffset(void)
{
    return pTheSoapFactory->TimeOffset;
}
#endif

PIUNKNOWN soapCobMain(void)
{
    PIUNKNOWN cob = GenericCobInit((void*) &soap_CobDesc, NULL);
    pTheSoapFactory = (SoapFactory *) cob;
    return cob;
}
