/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Deal with a soap request that is in an IBuffer and return the result
 * in another IBuffer.
 */
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <tokenizerif.h>
#include "ssoap.h"

SCODE ServeSoapRequest(
    UINT Flags,
    PIBUFFER Input,
    PIBUFFER *pOutput,
    void *key,
    PIHTTPCONNECTOR Request)
{
    METHOD_DECLARE_NOTHIS(_, ServeSoapRequest);
    PIHTTPPROCESS Process = NULL;
    PITOKENIZER Tokenizer = NULL;
    PITOKENIZERFACTORY TokenizerFactory = NULL;
    PSESSIONKEY pSKey = key;

    PIUNKNOWN Deserializer = NULL;
    UINT State = 0;

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_IHttpProcess, (void**)&Process);
    CHECKSC("bind to soap.cob");

    if (Flags & PARSE_URL_FLAGS_AES) {
        DBG(("soapreq: Decrypt the file...\n"));
        sc = AESBuffer(pSKey->KeyStrength, pSKey->Key,
                       Input, CYPHER_DECRYPT);
        CHECKSC("AESBuffer");
    }

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &TokenizerFactory);
    CHECKSC("bind to tokenizer.cob");

    sc = TokenizerFactory->v->Create(TokenizerFactory, (PIUNKNOWN) Input,
                                     &Tokenizer);
    CHECKSC("Create tokenizer");

    DBG(("soapreq: Calling post...\n"));
    sc = Process->v->Post(Process, 
                          Flags,
                          Tokenizer,
                          NULL,
                          pOutput,
                          &Deserializer, &State,
                          Request);
    CHECKSC("Post"); /* XXX Should send SoapFault instead */

    if (sc != S_NO_RESPONSE && (Flags & PARSE_URL_FLAGS_AES)) {
        sc = AESBuffer(pSKey->KeyStrength, pSKey->Key,
                       *pOutput, CYPHER_ENCRYPT);
        CHECKSC("AESBuffer encrypt");
    }

METHOD_CLEANUP;
    RELEASE(Process);
    RELEASE(Tokenizer);
    RELEASE(TokenizerFactory);
    RELEASE(Deserializer);

    return sc; /* S_NO_RESPONSE usually */
}
