/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <netif.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <winsock.h>
#include <assert.h>
#include <net/if.h>             /* for struct ifreq */
#include <base/cons.h>
#include <web/udpsoapserv.h>
#include "soap.h"
#include "../http/httpserver.h"

SCODE RunSoapServer(const _TCHAR *Url, BOOL Wait)
{
    SCODE sc;
    PINAMESPACE ns = CurrentNameSpace();
    PIUNKNOWN       Unk;
    PILISTENER      Server      = NULL;
    PISOAPFACTORY   SoapFactory = NULL;
    PIWEBPROTOCOLFACTORY WebFactory = NULL;
    const _TCHAR *cobname;
    PARSE_URL_FLAGS Flags;
    const _TCHAR *host;
    UINT hostlen;

    sc = BindToObject(ns, _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &SoapFactory);
    if (FAILED(sc)) {
        DBG(("soapserver.c: could not load COB soap.cob sc=x%x\n", sc));
        goto Out;
    }

    DBGME(3, _tprintf(_T("RunSoapServer(\"%s\")\n"), Url));
    sc = SoapFactory->v->ParseURL(SoapFactory, Url,
                                  &Flags, &host, &hostlen, NULL,
                                  NULL,NULL,NULL,NULL,NULL);
    if (FAILED(sc))
        goto Out;

    /* Make sure the network stack COB has been loaded */
    sc = ns->v->Bind(ns, _T("COB/protocol.cob"), 0, (PIUNKNOWN*) &Unk);
    /* LEAK protocol stack ref */
    if (FAILED(sc)) {
        DBG(("httptest.c::main: Failed to load network stack\n"));
        goto Out;
    }

    cobname = _T("COB/http.cob");
    //cobname = _T("COB/udpfactory.cob");

    /* Load the UDP/HTTP factory COB and construct a server */
    sc = BindToObject(ns, cobname, NAME_SPACE_READ,
                      &IID_IWebProtocolFactory, (void **) &WebFactory);
    if (FAILED(sc)) {
        DBG(("soapserver.c: could not load COB %s sc=x%x\n", cobname, sc));
        goto Out;
    }

    /* Construct the server instance */
    sc = WebFactory->v->CreateServer(WebFactory, CurrentNameSpace(), Flags,
                                     &Server);
    WebFactory->v->Release(WebFactory); WebFactory = NULL;
    if (FAILED(sc))
        goto Out;

    /* Start the HTTP/UDP server listening for requests... */
    sc = Server->v->Listen(Server, Url);
    if (FAILED(sc))
        goto Out;

    if (Wait) {
        sc = Server->v->Loop(Server);
        if (FAILED(sc)) goto Out;
        /* Should never get here */
        assert(FALSE);
    } else {
        /* Create a thread inside the Server module */
        PIPROCESS prc;
        PITHREAD pthd;
        PIMODULE mod;

        sc = GetModuleContainsAddr((ADDRESS) Server->v->Loop, &mod);
        if (FAILED(sc)) goto Out;
        sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **) &prc);
        mod->v->Release(mod);
        if (FAILED(sc)) goto Out;

        sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                                  (THREAD_FUNCTION) Server->v->Loop,
                                  Server, 0, NULL, &pthd);
        prc->v->Release(prc);
        if (FAILED(sc))
            goto Out;
        pthd->v->Release(pthd);
    }
    return S_OK;

Out:
    if (Server)
        Server->v->Release(Server);
    if (SoapFactory)
        SoapFactory->v->Release(SoapFactory);

    return sc;
}
