/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include "ssoap.h"
#include "soap.h"
#include <netif.h>
#include <httpif.h>

/* Generate random UINT32 */
UINT32 RandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}

int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(soaptest2,_tmain);
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY netfactory = NULL;
    PISOAPFACTORY soapfactory = NULL;
    const _TCHAR *target = _T("x-file-soap://fs/testmsg.xml"), *filename;
    const _TCHAR *outfile = _T("fs/testmsg_out.xml");
    const _TCHAR *resultfile = _T("fs/local_add_lt.txt");
    UINT Flags;
    PIFILE InputFile = NULL;
#if _SAVE_RESULT_XML
    PIFILE OutputFile = NULL;
#endif
    PIBUFFER Input = NULL, Output = NULL;
    UINT size, i, j, size2;
    UINT64 size64;
    TIME t0, t1;
    FILE *fp = NULL;
    SESSIONKEY SKey;
    const char *shouldfree = NULL;

    /*** Command line ***/
    ParseCmdLine(&argc,&argv);
    if (argc > 3) {
        printf("Usage: soaptest2.exe <url> <resultfile> e.g. %s %s\n",
               target, resultfile);
        return 1;
    }
    if (argc >= 2) {
        if (!_tcscmp(argv[1], _T("a"))) {
            target = _T("x-file-aes-soap://fs/testmsg.xml");
        } else {
            target = argv[1];
        }
    }
    if (argc >= 3)
        resultfile = argv[2];

    /*** Create all the objects ***/
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &netfactory);
    CHECKSC("load protocol.cob");

    sc = BindToObject(ns, _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    CHECKSC("load soap.cob");

    sc = soapfactory->v->ParseURL(soapfactory, target, 
                                  &Flags, NULL, NULL, NULL,
                                  (void*) &filename, &size,
                                  NULL,NULL,NULL);
    if (FAILED(sc)) {
        DBG(("soaptest2: Could not parse '%s'\n", target));
        CHECKSC("ParseURL");
    }

    sc = BindToObject(CurrentNameSpace(),
                      filename,
                      NAME_SPACE_WRITE | NAME_SPACE_READ,
                      &IID_IFile, 
                      (void **) &InputFile); 
    if (FAILED(sc)) {
        DBG(("soaptest2: Could not open '%s' for input\n", filename));
        CHECKSC("open file");
    }
    sc = InputFile->v->GetSize(InputFile, &size64);
    CHECKSC("get input file size");
    size = Uint64ToUint32(size64);

    DBG(("soaptest2 opened file '%s'\n", filename));

    /* Ask network stack to create "temp file" */
    sc = netfactory->v->AllocateBuffer(netfactory, size, NULL, &Input);
    CHECKSC("AllocateBuffer");

    sc = netfactory->v->AllocateBuffer(netfactory, 0, NULL, &Output);
    CHECKSC("AllocateBuffer");

    /* Open a file to store the results */
    fp = _tfopen(resultfile, _T("w"));
    CHECKSETSC(fp != NULL, E_FAIL, "open results file");

#define AES_KEY_STRENGTH   128
    SKey.KeyStrength = AES_KEY_STRENGTH;

    /*** Process message in a loop ***/
    for (i = 0; i < 2; i++) {
        sc = Output->v->SetSize(Output, UINT64_ZERO);
        sc = Input->v->SetSize(Input, UINT64_ZERO);
        sc = Input->v->ReadWriteFile(Input, InputFile, FALSE,
                                     UINT64_ZERO, UINT_MAX, size, &size2);
        CHECKSC("AppendFile");

        DBG(("soaptest2 wrote %d bytes\n", size));
        if (Flags & PARSE_URL_FLAGS_AES) {
            /* Generate a 128-bit AES key */
            for (j = 0; j < AES_KEY_STRENGTH/8; j += sizeof(UINT32))
                *(UINT32*)&SKey.Key[j] = RandUint32();
            sc = AESBuffer(SKey.KeyStrength, SKey.Key, Input, CYPHER_ENCRYPT);
            CHECKSC("AES encrypt");
        }

        t0 = CurrentTime();
        /*** Actual work, TIMED ***/
        sc = ServeSoapRequest(Flags, Input, &Output, &SKey, NULL);
        t1 = CurrentTime();
        t0 = Int64Subtract(t1, t0);
        t1 = Int64DividedByInt32(t0, 10000);
        CHECKSC("serve request");

        fprintf(fp, "%lld 00 ns (%lld ms)\n", t0, t1);
        //printf("i=%d %lld 00 ns (%lld ms)\n", i, t0, t1);
    }

#if _SAVE_RESULT_XML
    /* Write last result XML into OutputFile so it can be examined */
    sc = Output->v->GetSize(Output, &size64);
    CHECKSC("get input file size");
    size = Uint64ToUint32(size64);
    DBG(("soaptest2.c: output buffer size is %d\n", size));

    sc = BindToObjectA(CurrentNameSpace(),
                       outfile,
                       NAME_SPACE_WRITE | NAME_SPACE_READ | NAME_SPACE_CREATE,
                       &IID_IFile, 
                       (void **) &OutputFile); 
    if (FAILED(sc)) {
        DBG(("soaptest2: Could not open '%s' for output\n", outfile));
        CHECKSC("open file");
    }

    sc = OutputFile->v->SetSize(OutputFile, UINT64_ZERO);

    sc = Output->v->ReadWriteFile(Output, OutputFile, TRUE,
                                  UINT64_ZERO, 0, UINT_MAX, &size2);
    CHECKSC("SaveFile");
#endif

    /*** Cleanup ***/
METHOD_CLEANUP;
    RELEASE(InputFile);
#if _SAVE_RESULT_XML
    RELEASE(OutputFile);
#endif
    RELEASE(Input);
    RELEASE(Output);
    RELEASE(netfactory);
    RELEASE(soapfactory);
    if (fp)
        fclose(fp);

#if _UNICODE
    if (shouldfree)
        free((void *) shouldfree);
#endif

    FreeCmdLine(&argc,&argv);

    DBG(("soaptest2: Done\n"));
    return 0;
}
