/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define DBGLVL 0
#define _DEBUG 1

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>

#include <netif.h>
#include <winsock.h>
#include <httpif.h>
#include <httputils.h>
#include <soapif.h>
#include <tokenizerif.h>
#include "soap.h"
#include "deserializer.h"
#include <web/udpsoapserv.h>

typedef struct _REGISTRARY {
    UINT vid;
    PIENDPOINT pEndPoint;
    struct _REGISTRARY *next;
} *PREGISTRARY, *PITEM;

PREGISTRARY EndPointPool = NULL;

PRIVATE PITEM Search(PREGISTRARY registrary, UINT vid)
{
    PREGISTRARY reg = registrary;

    if (reg == NULL)
        return NULL;

    while (reg->vid != vid) {
        reg = reg->next;
        if (reg == NULL)
            return NULL;
    }

    return reg;
}

#define vloff offsetof(struct via, ListNode)


SCODE ForwardMsg(PDESERIALIZER Handler, PIBUFFER *pContent)
{
    SCODE sc;
    INT32 result;
    PISOAPCONNECTOR Connector;
    PIWEBPROTOCOLFACTORY pfactory;
    struct CONTINFO cont;
    _TCHAR *cob = _T("undefined");
    const _TCHAR *host = NULL;
    UINT16 port;
    PISOAPSERIALIZER Serializer = Handler->Serializer;
    Pvia via;
    PARSE_URL_FLAGS Flags;
    PIBUFFER im;

    /* We are routing either forwards or backwards */
    assert(Handler->prpcinfo);

    if (Handler->prpcinfo->pFwdVia)
        via = Handler->prpcinfo->pFwdVia;
    else
        via = Handler->prpcinfo->pRevVia;

    /* Complete request message */
    Serializer->v->EndMessage(Serializer, Handler->envelopeNS);

    sc = SOAPFactoryParseURL(NULL, via->target,
                             &Flags, &host, NULL, &port,
                             NULL,NULL,NULL,NULL,NULL);
    if (FAILED(sc))
        return sc;

    if (Flags & PARSE_URL_FLAGS_HTTP)
        cob = _T("COB/http.cob");
    else
        cob = _T("COB/udpfactory.cob");

    sc = BindToObject(CurrentNameSpace(), cob, NAME_SPACE_READ,
                      &IID_IWebProtocolFactory, (void **) &pfactory);
    if (FAILED(sc)) {
        DBG(("httpconnectortest.c: main: Failed to load %s\n", cob));
        return sc;
    }
    sc = pfactory->v->CreateConnector(pfactory, via->target, &Connector);
    pfactory->v->Release(pfactory);
    if (FAILED(sc)) {
        DBG(("Failed to create pconnector cob=%s\n", cob));
        return sc; 
    }

    sc = Connector->v->SetKey(Connector);
    assert(sc == S_OK);

    memset(&cont, 0, sizeof(cont));
    cont.RetVal = 0;
    cont.metd = 0;    /* XXX, use detected typeinfo */

    sc = Connector->v->SendMessage(Connector, NULL, (PIFILE) *pContent,
                                   iCONT(&cont), Serializer);
    if (FAILED(sc))
        goto Out;

    sc = Connector->v->AllocateBuffer(Connector, &im);
    if (FAILED(sc))
        goto Out;

    Serializer->v->Reset(Serializer, (PIFILE) im); /* Set new message */

    {
        UINT Refs;
        Refs = (*pContent)->v->Release(*pContent);
        assert(Refs <= 1);
        *pContent = im;
    }

    if (FAILED(sc)) {
        if (sc == 0x8007274c) {
            DBG(("Times out!\n"));
            sc =  E_TIMED_OUT; 
        }
        goto Out;
    }

    DBG(("RouteMsg part1 Done!\n"));
    sc = Connector->v->ReceiveReply(Connector);
    if (sc != S_OK) {
        (void) Serializer->v->Reset(Serializer, NULL);
        goto Out;
    }
    result = cont.RetVal;

    /* Complete reply message */
    sc = Serializer->v->EndMessage(Serializer, Handler->envelopeNS);

    sc = Connector->v->ClearKey(Connector);
    assert(sc == S_OK);

 Out:
    Connector->v->Release(Connector);
    return sc;
}

/* Copier for via structs. Allocates memory for struct and embedded 
 * strings in a single malloc, requires only single free (see Deleter). 
 */
PTR CopyVia(PTR pItem, PTR pArg)
{
    PIHEAP Heap = pArg;
    Pvia via = pItem, vnew;
    UINT tlen = via->target ? _tcslen(via->target) + 1 : 0;
    UINT rlen = via->reservation ? _tcslen(via->reservation) + 1 : 0;

    vnew = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("soap:via"),
                          sizeof(*vnew) + (tlen + rlen)*sizeof(_TCHAR), 0);
    if (!vnew)
        return NULL;
    vnew->vid = via->vid;
    if (tlen) {
        vnew->target = (_TCHAR*) (vnew + 1);
        _tcscpy(vnew->target, via->target);
    } else {
        vnew->target = NULL;
    }
    if (rlen) {
        vnew->reservation = ((_TCHAR*) (vnew + 1)) + tlen;
        _tcscpy(vnew->reservation, via->reservation);
    } else {
        vnew->reservation = NULL;
    }
    return vnew;
}

/* Deleter for via structs. Assumes memory for embedded strings 
 * either does not need to be freed or was allocated in the same
 * malloc as the struct (see Copier).
 */
INT DeleteVia(PTR via, PTR pArg)
{
    PIHEAP Heap = pArg;
    Heap->v->Free(Heap, 0, via);
    return 0;
}

void DestroyRouting(PTR pvia, PIHEAP Heap)
{
//    if (pvia)
//        ListWalk(&pvia, vloff, 0, DeleteVia, Heap);
}


#if 0
SCODE WsRouting_path(PIWSROUTING iThis,
                     _TCHAR *action, _TCHAR *to,
                     Pvia fwd,
                     Pvia rev,
                     _TCHAR *from, _TCHAR *id, _TCHAR *relatesTo)
{
    PDESERIALIZER This = pWsRouting(iThis);
    PIHEAP Heap = This->Heap;
    SCODE sc;
    Pvia via;
    PITEM pItem;

    if (fwd && !ListNodeIsAlone(vloff, fwd)) {
        UINT vid;

        TRACE("FORWARD INTERMEDIARY");
        vid = VidCounter++;
        if (vid == 0)
            vid = VidCounter++;

        sc = Register(&EndPointPool, NULL, vid, Heap);
        if (FAILED(sc))
            return sc;

        /* do routing */
        via = ListRemoveFirst(&fwd, vloff);

        /* Make a reverse entry with continuation in it (vid) */
        /* Reuse the descriptor for a new reverse entry */
        via->target = NULL;
        via->vid = vid;

        ListAddFirst(&rev, vloff, via);
        goto RouteIt;
    } else if (fwd) {
        _TCHAR *pdest = NULL;

        TRACE("ULTIMATE RECEIVER");
        /* Drop first and only entry */

        via = ListRemoveFirst(&fwd, vloff);
        /* URL required here */
        if (!via->target)
            return E_SCHEMA_CHECK;

        sc = SOAPFactoryParseURL(NULL, via->target,
                                 NULL/*flags*/, NULL/*host*/,
                                 NULL/*hostlen*/, NULL/*port*/,
                                 (const _TCHAR **)&pdest);

        if (SUCCEEDED(sc)) {
            if (This->ShouldFreeURL)
                free((void*) This->reqLocalURL);
            sc = Allocate_And_Copy(Heap, (void*) &This->reqLocalURL, 
                                   pdest, _tcslen(pdest) + 1);
            assert(sc == S_OK); /* XXX */
            This->ShouldFreeURL = TRUE;
            DBG(("WsRouting_path: reqLocalURL = %s\n", This->reqLocalURL));

            if (via->reservation) {
                sc = Allocate_And_Copy(Heap, (void*) &This->resname, 
                                       via->reservation,
                                       _tcslen(via->reservation) + 1);
                assert(sc == S_OK); /* XXX */
            }
        }
#if 0
        sc = This->Serializer->v->StartMessage(This->Serializer,
                                               This->envelopeNS,
                                               GetNamespace("soapenc"),
                                               This->namespaces,
                                               rev,
                                               NULL, 
                                               NULL);
    This->MsgStarted = TRUE;
#endif
    } else if (!rev) {
        /* No reverse path, no forward path. Bad routing. */
        TRACE("wsroute: isLocal no path");
        return E_PATH_NOT_FOUND;
    } else if (!ListNodeIsAlone(vloff, rev)) {
        TRACE("REVERSE INTERMEDIARY");
        via = ListRemoveFirst(&rev, vloff);

        /* lookup the vid in the endpoint pool and set active endpoint.
           If not found, return error */
        assert(via->vid);
        pItem = Search(EndPointPool, via->vid);
        if (pItem) {
            DBG(("wsroute:isLocalCall vid not matched (XXX IGNORED)!\n"));
            Unregister(&EndPointPool, via->vid, Heap);
        }

        goto RouteIt;
    } else {
        TRACE("ORIGINAL SENDER");

        via = ListRemoveFirst(&rev, vloff);
        assert(via->vid);
        pItem = Search(EndPointPool, via->vid);
        if (pItem)
            Unregister(&EndPointPool, via->vid, Heap);
#if 0
        else
            return E_PATH_NOT_FOUND;
#endif
        sc = S_OK;
    }

 SavePath:
    assert(This->pFwdVia == NULL);
    assert(This->pRevVia == NULL);

    This->pFwdVia = ListCopy(&fwd, vloff, CopyVia, DeleteVia, Heap);
    This->pRevVia = ListCopy(&rev, vloff, CopyVia, DeleteVia, Heap);
    return sc;

 RouteIt:
/* Disabled - addressing replaces routing 
    sc = This->Serializer->v->StartMessage(This->Serializer,
                                           This->envelopeNS,
                                           GetNamespace(_T("soapenc")),
                                           This->namespaces,
                                           fwd,
                                           rev, 
                                           NULL);
*/
    This->ultimateReceiver = FALSE;
    goto SavePath;
}
#endif
