/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Given samples, predict required time -jvh 2004 */

#include <stddef.h> /* for offsetof() */
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <mmlite.h>
#include <util/list-double2.h>

UINT ffs(UINT x)
{
    UINT i;
    for (i = 0; i < 32; i++) {
        if (!x)
            return i;
        x >>= 1;
    }
    return 32;
}

UINT isqrt(UINT x)
{
    UINT hi, lo, a, b;
    if (x < 2)
        return x;
    a = ffs(x);
    //printf("ffs=%d\n", a);
    hi = 1 << a/2;
    lo = 1 << (a-1)/2;

    for(;;) {
        a = (hi - lo + 1) / 2;
        if (a == 0)
            return lo;
        b = lo + a;
        b *= b;
        //printf("lo=%d hi=%d x=%d a=%d b=%d\n", lo, hi, x, a, b);
        if (b < x)
            lo += a;
        else
            hi -= a;
    };
}

UINT iavg(UINT *val, UINT Count)
{
    UINT mean = 0, i;

    for (i = 0; i < Count; i++)
        mean += val[i];
    mean += i/2;
    mean /= i;
    return mean;
}

UINT istddev(UINT *val, UINT Count, UINT mean)
{
    UINT i, x, std = 0;

    for (i = 0; i < Count; i++) {
        x = val[i];
        if (x < mean)
            x = mean - x;
        else
            x = x - mean;
        x *= x;
        std += x;
    }
    std += (Count-1)/2;
    std /= (Count-1);
    return isqrt(std);
}

UINT val[] = {10000, 10100, 10200, 9900, 9000, 11000, 20000, 10000};
#define COUNT (sizeof(val)/sizeof(UINT))

int _tmain(int argc, _TCHAR** argv)
{
    UINT mean, std;

    mean = iavg(val, COUNT);
    std = istddev(val, COUNT, mean);
    printf ("mean=%d std=%d\n", mean, std);
}
