/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define _DEBUG 1
#include <stdio.h>
#include <mmlite.h>
#include <assert.h>
#include <string.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <web/tokenizerutils.h>
#include "wsdiscov.h"

#define _MANUAL_COBMAIN 1
#include "s_WsDiscoEnum_v.c"
#include "s_WsDiscovery_v.c"

/* IUnknown */

PRIVATE SCODE MCT WsDiscoveryDestroy(PIUNKNOWN pDisco)
{
    PWSDISCOVERY pThis = pD(pDisco);
    PIHEAP pHeap = CurrentHeap();

    Condition_Destroy(&pThis->cd);
    Mutex_Destroy(&pThis->mx);
    pHeap->v->Free(pHeap, 0, pThis);

    DBG(("WsDiscoveryDestroy %x\n", (UINT) pThis));
    return S_OK;
}


/* Utility functions */

#ifdef _DEBUG
PRIVATE INT NodeEq(PTR p1, PTR p2)
{
    return p1 != p2;
}
#endif

PRIVATE INT EnumBump(PTR _pEnum, PTR _pAd);

PRIVATE INT AddToCache(PWSDISCOVERY This, PWSADVERT *cache, PWSADVERT ad)
{
    PWSADVERT pItem;
    PIHEAP pHeap = CurrentHeap();

    /* check if there is an item already.
     * if so, remove that entry and add an update one.
     */
    Mutex_Lock(&This->mx);
    ListAddLast(cache, aloff, ad);
    pItem = ListWalk(cache, aloff, LIST_WALK_SEARCH, WsAdvertEndpointCmp, ad->Endpoint);
    if (pItem) {
        ListWalk(&This->Enumerators, denoff, LIST_WALK_ITERATE,
                 EnumBump, pItem);
        ListRemoveItem(cache, aloff, pItem);
        WsAdvertDelete(pItem, pHeap);
    }
    Mutex_Unlock(&This->mx);
    return 0;
}

PRIVATE INT RemoveFromCache(PWSDISCOVERY This, PWSADVERT *cache, PEndpointReference ep)
{
    PWSADVERT pItem;
    PIHEAP pHeap = CurrentHeap();

    /* check if there is an item already.
     * if so, remove that entry and add an update one.
     */
    Mutex_Lock(&This->mx);
    pItem = ListWalk(cache, aloff, LIST_WALK_SEARCH, WsAdvertEndpointCmp, ep);
    if (pItem) {
        ListWalk(&This->Enumerators, denoff, LIST_WALK_ITERATE,
                 EnumBump, pItem);
        ListRemoveItem(cache, aloff, pItem);
        WsAdvertDelete(pItem, pHeap);
    }
    Mutex_Unlock(&This->mx);
    return 0;
}

typedef struct AdvertMatchTemplate {
    PQNameItem Types;
    PScopes Scopes;
} AdvertMatchTemplate, *PAdvertMatchTemplate;

PRIVATE INT WsAdvertMatchCmp(PTR ad, PTR amt)
{
    PWSADVERT pAd = (PWSADVERT) ad;
    PAdvertMatchTemplate pAmt = (PAdvertMatchTemplate) amt;
    PQNameItem head, item;
    _TCHAR *value;
    
    /* First, check the types. All the types should match. */
    head = pAmt->Types;
    if (head) {
        item = head;
        do {
            value = item->Value;
            if (!ListWalk(&pAd->Types, qnloff, LIST_WALK_SEARCH, WsQNameItemCmp, value)) 
                return 1;
            item = ListNodeNext(qnloff, item);
        } while (item != head);
    }

    /* Then, check the scopes. 
     * It needs to support various matching rules. 
     * For now, it supports "http://schemas.xmlsoap.org/ws/2005/04/discovery/strcmp0"
     */
/*  we don't need this part for demo... :)
    head = pAmt->Scopes->Items;
    if (head) {
        item = head;
        do {
            value = item->Value;
            if (!ListWalk(&pAd->Scopes->Items, qnloff, LIST_WALK_SEARCH, WsQNameItemCmp, value)) 
                return 1;
            item = ListNodeNext(qnloff, item);
        } while (item != head);
    }
*/
    /* Finally, it passed type and scopes matching */
    return 0;
}

PRIVATE const PWSADVERT LookUpCache(PWSDISCOVERY pThis, PWSADVERT *cache, 
                              PQNameItem Types, PScopes Scopes)
{
    PWSADVERT pAd; 
    AdvertMatchTemplate Amt;
    Amt.Types = Types;
    Amt.Scopes = Scopes;

    Mutex_Lock(&pThis->mx);
    pAd = ListWalk(cache, aloff, LIST_WALK_SEARCH, WsAdvertMatchCmp, &Amt);
    Mutex_Unlock(&pThis->mx);
    return pAd;
}

/*******************************************/

PRIVATE SCODE MCT WsDiscoveryHello(
    PIWSDISCOVERY pDisco, 
    PEndpointReference Endpoint, 
    PQNameItem Types, 
    PScopes Scopes, 
    PStringItem XAddr,
    UINT MetadataVersion)
{
    PWSDISCOVERY pThis = pD(pDisco);
    PWSADVERT pAdvert;

    DBG(("Hello: Got It!\r\n"));
    pAdvert = WsAdvertNew(Endpoint, Types, Scopes, XAddr, MetadataVersion);
    AddToCache(pThis, &pThis->CacheRemote, pAdvert);
    return S_NO_RESPONSE;
}

PRIVATE SCODE MCT WsDiscoveryBye(
    PIWSDISCOVERY pDisco,   
    PEndpointReference Endpoint)
{
    PWSDISCOVERY pThis = pD(pDisco);
 
    DBG(("Bye: Got It!\r\n"));
    RemoveFromCache(pThis, &pThis->CacheRemote, Endpoint);
    return S_NO_RESPONSE;
}

PRIVATE SCODE MCT WsDiscoveryProbe(
    PIWSDISCOVERY iThis,
    PQNameItem Types, 
    PScopes Scopes)
{
    METHOD_DECLARE_NOTHIS(WsDiscovery, Probe);
    PCONTINFO cont;
    PITHREAD thd = CurrentThread();
    PIHTTPCONNECTOR con;
    PISOAPFACTORY pSoapFactory = NULL;
    PIWSDISCOVERY pxy = NULL;
    PWSCONTEXT pCtxIn, pCtxOut = NULL;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz;
    _TCHAR *pMsgId = NULL;
    PWSADVERT pAd = NULL;
    PWSDISCOVERY pThis = pD(iThis);
    PProbeMatch pMatch = NULL;

    DBG(("Probe: Got It!\r\n"));

    pAd = LookUpCache(pThis, &pThis->CacheLocal, Types, Scopes);
    if (!pAd) {
        sc = S_OK;
        goto Out;
    }
    
    pMatch = WsProbeMatchNew(
        pAd->Endpoint,
        pAd->Types,
        pAd->Scopes,
        pAd->XAddrs,
        pAd->MetadataVersion);

    cont = (PCONTINFO) thd->v->GetValue(thd, THREAD_VALUE_RPC_CONTINUATION);
    CHECKSETSC(cont, E_FAIL, "get cont");

    pCtxIn = cont->pContext;
    CHECKSETSC(pCtxIn, E_FAIL, "get Context");

    con = cont->Request;
    CHECKSETSC(con, E_FAIL, "get Request");

    sc = BindToObject(
        CurrentNameSpace(), 
        _T("COB/soap.cob"), 
        NAME_SPACE_READ,
        &IID_ISoapFactory, 
        (void **) &pSoapFactory);
    CHECKSC("get Soapfactory");

    BufSz = sizeof(WSCONTEXT);
    pCtxOut = (PWSCONTEXT) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("WsDiscovery"), BufSz, 0);
    pMsgId = (_TCHAR *) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("WsDiscovery"), 50, 0);
    _tcscpy(pMsgId, _T("uuid:E9E02CCC-5EB2-4A46-0000-000000000000"));
    {
        TIME t = CurrentTime();
        INT tt = Int64ToInt32(t);
        _stprintf(pMsgId + 29, _T("%012u"), tt);
    }
    pCtxOut->MessageId = pMsgId;
    pCtxOut->RelatesTo = pCtxIn->MessageId;
    pCtxOut->To = WsEndpointNew(_T("http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"));
    sc = pSoapFactory->v->CreateProxy3(
        pSoapFactory, 
        con, 
        pCtxOut,
        _T(IWsDiscovery_NSID),
        0,
        (PIUNKNOWN*) &pxy);
    CHECKSC("CreateProxy");
    
    sc = pxy->v->ProbeMatches(pxy, pMatch);
    CHECKSC("ProbeMatches");

METHOD_CLEANUP;
    if (pMatch) 
        WsProbeMatchDelete(pMatch, pHeap);
    if (pCtxOut)
        pHeap->v->Free(pHeap, 0, (PTR) pCtxOut);
    if (pMsgId)
        pHeap->v->Free(pHeap, 0, (PTR) pMsgId);
    if (pxy)
        RELEASE(pxy);
    if (pSoapFactory)
        RELEASE(pSoapFactory);
    return S_NO_RESPONSE;
}

PRIVATE SCODE MCT WsDiscoveryProbeMatches(
    PIWSDISCOVERY iThis,
    PProbeMatch ProbeMatch)
{
    METHOD_DECLARE_NOTHIS(WsDiscovery, ProbeMatches);
    PWSADVERT pAd = NULL;
    PWSDISCOVERY pThis = pD(iThis);

    DBG(("ProbeMatches: Got It!\r\n"));
    sc = S_OK;

    pAd = WsAdvertNew(
        ProbeMatch->EndpointReference,
        ProbeMatch->Types,
        ProbeMatch->Scopes,
        ProbeMatch->XAddrs,
        ProbeMatch->MetadataVersion);

    AddToCache(pThis, &pThis->CacheRemote, pAd);
    Condition_Broadcast(&pThis->cd);
METHOD_CLEANUP;
    return sc;
}


PRIVATE SCODE MCT WsDiscoveryResolve(
    PIWSDISCOVERY pDisco,
    PEndpointReference Endpoint)
{
    DBG(("Resolve: Got It!\r\n"));

    return S_NO_RESPONSE;
}

PRIVATE SCODE MCT WsDiscoveryResolveMatch(
    PIWSDISCOVERY pDisco,
    PEndpointReference Endpoint,
    PQNameItem Types,
    PScopes Scopes,
    PStringItem XAddr,
    UINT MetadataVersion)
{
    DBG(("ResolveMatch: Got It!\r\n"));

    return S_NO_RESPONSE;
}

PRIVATE SCODE MCT WsDiscoveryAppSequence(
    PIWSDISCOVERY pDisco,
    UINT InstanceId,
    const _TCHAR *SequenceId,
    UINT MessageNumber)
{
    /* XXX ignored */
    return S_OK;
}

/* Internal utilities */

#define WsDiscoveryIPv4McastAddr "239.255.255.250:3702"

PRIVATE SCODE MCT WsDiscoveryCreateProxy(
    PIWSDISCOVERY *ppProxy)
{
    SCODE sc = S_OK;
    PISOAPFACTORY pSoapFactory = NULL;
    const _TCHAR *McastChannel = 
        _T("x-udp-soap://") _T(WsDiscoveryIPv4McastAddr) _T("/COB/wsdiscov.cob"); 
    PWSCONTEXT pCtx;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz; 
    PEndpointReference pEp1, pEp2;
    _TCHAR *pMsgId;

    DBG(("wsdiscov.c:WsDiscoveryCreateProxy: mcast=%s\n", McastChannel));

    sc = BindToObject(
        CurrentNameSpace(), 
        _T("COB/soap.cob"), 
        NAME_SPACE_READ,
        &IID_ISoapFactory, 
        (void **) &pSoapFactory);
    if (FAILED(sc)) {
        DBG(("WsDiscoveryCreateProxy: Could not bind to soap.cob sc=%x\n", sc));
        goto Out;
    }

    BufSz = sizeof(WSCONTEXT);
    pCtx = (PWSCONTEXT) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("WsDiscovery"), BufSz, 0);
    if (!pCtx) {
        DBG(("WsDiscoveryCreateProxy: Could not allocate memory for context\n"));
        goto Out;
    }

    pEp1 = WsEndpointNew(McastChannel);
    pEp2 = WsEndpointNew(_T("http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"));

    pMsgId = (_TCHAR *) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("WsDiscovery"), 50, 0);

    _tcscpy(pMsgId, _T("uuid:E9E02CCC-5EB2-4A46-0000-000000000000"));
    {
        TIME t = CurrentTime();
        INT tt = Int64ToInt32(t);
        _stprintf(pMsgId + 29, _T("%012u"), tt);
    }
    pCtx->To = pEp1;
    pCtx->ReplyTo = pEp2;
    pCtx->MessageId = pMsgId;

    sc = pSoapFactory->v->CreateProxy3(
        pSoapFactory, 
        NULL, 
        pCtx,
        _T(IWsDiscovery_NSID),
        0,
        (PIUNKNOWN*) ppProxy);
    if (FAILED(sc)) {
        DBG(("WsDiscoveryCreateProxy: CreateProxy failed sc=%x url=%s ns=%s\n",
             sc, McastChannel, _T(IWsDiscovery_NSID)));
        goto Out;
    }

    WsEndpointDelete(pEp1, pHeap);
    pEp1 = WsEndpointNew(_T("urn:schemas-xmlsoap-org:ws:2005:04:discovery"));
    pCtx->To = pEp1;

Out:
    if (pSoapFactory)
        pSoapFactory->v->Release(pSoapFactory);
    DBG(("wsdisco: Done with proxy creation sc=%x\n", sc));
    return sc;
}

PRIVATE SCODE MCT WsDiscoverySendHello(PWSADVERT pAdvert)
{
    SCODE sc;
    PIWSDISCOVERY pProxy;
    
    DBG(("wsdiscov.c:WsDiscoverySendHello: endpoint=%s\n", 
        pAdvert->Endpoint->Address));

    sc = WsDiscoveryCreateProxy(&pProxy);
    if (!pProxy) return sc;

    sc = pProxy->v->Hello(
        pProxy, 
        pAdvert->Endpoint, 
        pAdvert->Types,
        pAdvert->Scopes,
        pAdvert->XAddrs,
        pAdvert->MetadataVersion);

    RELEASE(pProxy);
    return sc;
}

PRIVATE SCODE MCT WsDiscoverySendBye(PEndpointReference pEp)
{
    SCODE sc;
    PIWSDISCOVERY pProxy;
    
    DBG(("wsdiscov.c:WsDiscoverySendBye: endpoint=%s\n", 
        pEp->Address));

    sc = WsDiscoveryCreateProxy(&pProxy);
    if (!pProxy) return sc;

    sc = pProxy->v->Bye(pProxy, pEp);

    RELEASE(pProxy);
    return sc;
}

PRIVATE SCODE MCT WsDiscoveryUtilQueryInterface(
    PIWSDISCOVERYUTIL pUtil, REFIID Iid, void** Unk)
{
    PIUNKNOWN t = (PIUNKNOWN) pDU(pUtil);
    return t->v->QueryInterface(t, Iid, Unk);
}
        
PRIVATE UINT MCT WsDiscoveryUtilAddRef(PIWSDISCOVERYUTIL pUtil)
{
    PIUNKNOWN t = (PIUNKNOWN) pDU(pUtil);
    return t->v->AddRef(t);
}

PRIVATE UINT MCT WsDiscoveryUtilRelease(PIWSDISCOVERYUTIL pUtil)
{
    PIUNKNOWN t = (PIUNKNOWN) pDU(pUtil);
    return t->v->Release(t);
}

/* IWsDiscoveryUtil - external utilities */

#include <stdlib.h>
#include <netif.h>
#include <winsock.h>
#include <net/ip/in.h>          /* struct ip_mreq */
#include <net/if.h>             /* for struct ifreq */

PRIVATE UINT32 GetInterfaceIp(const _TCHAR *iname, UINT inamelen)
{
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY ef;
    PIENDPOINT ep;
    struct ifreq ifr;
    UINT32 LocalIPAddr = INADDR_ANY;
    SCODE sc;
    
    /* Prepare address and port to listen at */
    memset((void*) &ifr, 0, sizeof(struct ifreq));

    /* Get an endpoint factory to create an endpoint */
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &ef);
    if (FAILED(sc)) goto Out;

    /* Create an endpoint to query for the interface address */
    sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep);
    ef->v->Release(ef);
    if (FAILED(sc)) goto Out;

    /* This is where we specify which interface to use */
#ifdef _UNICODE
{
    char *s = _wcstos(iname);
    if (!s)
        return E_NOT_ENOUGH_MEMORY;
    strncpy(ifr.ifr_name, s, __min(inamelen, sizeof(ifr.ifr_name)));
    free(s);
}
#else
    strncpy(ifr.ifr_name, iname, __min(inamelen, sizeof(ifr.ifr_name)));
#endif
    
    /* Query for the interface address */
    sc = ep->v->Ioctl(ep, SIOCGIFADDR, (BYTE*) &ifr);
    ep->v->Release(ep);
    if (FAILED(sc)) goto Out;

    LocalIPAddr = ((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr;
Out:
    return LocalIPAddr;
}



PRIVATE SCODE MCT WsDiscoveryUtilRegisterAdvert(
    PIWSDISCOVERYUTIL pUtil,
    PWSADVERT pAdvert)
{
    SCODE sc;
    PWSDISCOVERY pThis = pDU(pUtil);
    PWSADVERT pNew;

    pNew = WsAdvertCopy(pAdvert, CurrentHeap());

    /* register the advertisement into the cache for local advert. */

    AddToCache(pThis, &pThis->CacheLocal, pNew);

    sc = S_OK;
    sc = WsDiscoverySendHello(pAdvert);
    return sc;
}

PRIVATE SCODE MCT WsDiscoveryUtilUnregisterAdvert(
    PIWSDISCOVERYUTIL pUtil,
    PEndpointReference pEp)
{
    SCODE sc;
    PWSDISCOVERY pThis = pDU(pUtil);

    /* remove the advertisement from the cache for local advert. */

    RemoveFromCache(pThis, &pThis->CacheLocal, pEp);

    sc = WsDiscoverySendBye(pEp);
    return sc;
}

INT tokenizer(_TCHAR **line, _TCHAR **token)
{
    _TCHAR *pt = NULL;

    if (*line == NULL) {
        *token = NULL;
        return 0;
    }
    *token = *line;
    pt = _tcschr(*line, ' ');
    if (pt == NULL) {
        *line = NULL;
        return 1;
    }
    *line = pt + 1;
    *pt = '\0';
    return 1;
}


PRIVATE SCODE MCT WsDiscoveryUtilSendProbe(
    PIWSDISCOVERYUTIL iThis,
    PQNameItem Types,
    PScopes Scopes)
{
    METHOD_DECLARE_NOTHIS(WsDiscoveryUtil, SendProbe);
    PIWSDISCOVERY pProxy;
    PIHTTPCONNECTOR con;
    TIME timeo;
    
    DBG(("WsDiscoveryUtilSendProbe types=%x scopes=%x\n",
         (UINT) Types, (UINT) Scopes));

    sc = WsDiscoveryCreateProxy(&pProxy);
    CHECKSC("Create Proxy");
    
    sc = pProxy->v->QueryInterface(pProxy, &IID_IHttpConnector, (void**) &con);
    CHECKSC("QI IHttpConnector");

    Int32ToInt64(timeo, TIME_SECONDS(5));

    sc = con->v->Listen(con, &timeo);
    CHECKSC("Listen connector");

    DBG(("sending probe\n"));
    sc = pProxy->v->Probe(pProxy, Types, Scopes);
    DBG(("Probe sent...\n"));

METHOD_CLEANUP;
    RELEASE(pProxy);
    RELEASE(con);
    return sc;
}

PRIVATE SCODE MCT WsDiscoveryUtilLookUp(
    /*in*/  PIWSDISCOVERYUTIL iThis,
    /*in*/  PQNameItem Types,
    /*in*/  PScopes Scopes,
    /*out*/ PWSADVERT *Advert)
{
    METHOD_DECLARE_NOTHIS(WsDiscoveryUtil, LookUp);
    PWSDISCOVERY pThis = pDU(iThis);
    PIHEAP pHeap = CurrentHeap();
    PWSADVERT pAd = NULL;
    TIME timeo;
    INT i;

    *Advert = NULL;
    sc = S_OK;

    DBG(("WsDiscoveryUtilLookUp\n"));
    Int32ToInt64(timeo, TIME_RELATIVE(TIME_SECONDS(2)));
//    timeo = TIME_FOREVER;

    for (i = 0; i < 5; i++) {
        /* Search for cache (local) match */
        pAd = LookUpCache(pThis, &pThis->CacheLocal, Types, Scopes);
        if (pAd) {
            *Advert = WsAdvertCopy(pAd, pHeap);
            goto Out;
        } 
        pAd = LookUpCache(pThis, &pThis->CacheRemote, Types, Scopes);
        if (pAd) {
            *Advert = WsAdvertCopy(pAd, pHeap);
            goto Out;
        } 

        sc = WsDiscoveryUtilSendProbe(iThis, Types, Scopes);
        CHECKSC("SendProbe");
        sc = Condition_Wait(&pThis->cd, NULL, &timeo);
    }
    
METHOD_CLEANUP;
    DBG(("Lookup done sc=%x", sc));
    return sc;
}

PRIVATE SCODE MCT WsDiscoveryUtilLookUp2(
    /*in*/  PIWSDISCOVERYUTIL iThis,
    /*in*/  const _TCHAR * Type,
    /*in*/  UINT bufSz,
    /*out*/ _TCHAR * Buffer)
{
    METHOD_DECLARE_NOTHIS(WsDiscoveryUtil, LookUp2);
    PWSDISCOVERY This = pDU(iThis);
    PWSADVERT pAd = NULL;
    PQNameItem pName = NULL;
    PScopes pScopes = NULL;
    PIHEAP pHeap = CurrentHeap();
    UINT len;

    DBG(("WsDiscoveryUtilLookUp2\n"));

    pName = WsQNameItemNew(Type);
    pScopes = WsScopesNew(NULL, WsQNameItemNew(_T("here")));
    sc = WsDiscoveryUtilLookUp(iThis, pName, pScopes, &pAd);

    if (pAd == NULL) {
        _tcscpy(Buffer, _T(""));
    } else {
        len = _tcslen(pAd->XAddrs->Value);
        if (bufSz < (len + 1)) {
            _tcscpy(Buffer, _T(""));
        } else {
            _tcscpy(Buffer, pAd->XAddrs->Value);
        }
    }

METHOD_CLEANUP;
    if (pName) WsQNameItemDelete(pName, pHeap);
    if (pScopes) WsScopesDelete(pScopes, pHeap);
    if (pAd) WsAdvertDelete(pAd, pHeap);
    return sc;
}

/* XXX BUGBUG Missing code to associate (eventual) resolve matches */
PRIVATE SCODE MCT WsDiscoveryUtilSendResolve(
    PIWSDISCOVERYUTIL pUtil, 
    PEndpointReference pEndpoint) 
{
    SCODE sc = S_OK;

/* XXX: not implemented. BTW, is it necessary? */
/*
    PIWSDISCOVERY pProxy;
    
    DBG(("WsDiscoveryUtilSendResolve endpoint=%s\n", pEndpoint->Address));

    DebugBreak();

    sc = WsDiscoveryCreateProxy(&pProxy);
    if (!pProxy) return sc;

    sc = pProxy->v->Resolve(pProxy, pEndpoint);

    pProxy->v->Release(pProxy);
*/
    return sc;
}

PRIVATE INT Add2StringItemList(PStringItem *ppStr, _TCHAR *str)
{
    PStringItem pStr;
    if (!str) return 1;
    pStr = WsStringItemNew(str);
    if (!*ppStr)
        ListAddLast(ppStr, sloff, pStr);
    else 
        *ppStr = pStr;
    return 0;
}

PRIVATE SCODE MCT WsDiscoveryUtilMain(PIWSDISCOVERYUTIL iThis, const _TCHAR *CommandLine)
{
    METHOD_DECLARE_NOTHIS(WsDiscoveryUtil, Main);
    PIHEAP pHeap = CurrentHeap();

    _TCHAR *cmd, *tmp;
    _TCHAR *token;

/* command line parsing state */
#define STATE_NONE 0
#define STATE_IP 1
#define STATE_IF 2
#define STATE_TYPE 3
#define STATE_QUERY 4
#define STATE_REMOVE 5
    INT state = STATE_NONE;
    INT k = 0;

/* command type */
#define COMMAND_REGISTER 1
#define COMMAND_QUERY 2
#define COMMAND_REMOVE 3
    INT command = COMMAND_REGISTER;

    PStringItem pIP = NULL, pIF = NULL , pType = NULL, pQueryType = NULL, pRemoveType = NULL;
    PStringItem tmpItem1 = NULL, tmpItem2 = NULL;
 
    /* hard-coded table of available services */
    int svcs = 3;
    static const struct ServiceTable {
        _TCHAR *uuid;
        _TCHAR *itype;
        _TCHAR *path;
        UINT version;
    } table[] = 
        { {_T("183a7ace-ea5f-43d9-80cd-d136a30ce38b"), 
         _T("ITablet"),
         _T("COB/tabletservice.cob"),
         1 },
        {_T("183a7ace-ea5f-43d9-80cd-d136a30ce383"),
         _T("IPong"),
         _T("COB/ProjectObj.cob"),
         1 },
        {_T("183a7ace-ea5f-43d9-80cd-d136a30ce387"),
         _T("IGyro"),
         _T("COB/gyro.cob"),
         1 } };

    sc = S_OK;

    DBG(("Disco main %s\n", CommandLine));

    if ((CommandLine == NULL || !*CommandLine)) {
        DBG(("Usage: wsdiscov.cob [-ip ipaddress ...] [-if interface ...] [-type type ...] [-query type] [-remove type ...]"));
        return sc;
    }

    /* The tokenizer writes into the string */
    tmp = cmd = _tcsdup(CommandLine);
    if (!cmd) {
        DBG(("strdup failed"));
        return E_NOT_ENOUGH_MEMORY;
    }
    while (1) {
        if (!tokenizer(&cmd, &token)) break;
        if (token[0] == '-') {
            if (!_tcscmp(_T("ip"), token + 1)) {
                state = STATE_IP;
                continue;
            } else if (!_tcscmp(_T("if"), token + 1)) {
                state = STATE_IF;
                continue;
            } else if (!_tcscmp(_T("type"), token + 1)) {
                state = STATE_TYPE;
                continue;
            } else if (!_tcscmp(_T("query"), token + 1)) {
                state = STATE_QUERY;
                command = COMMAND_QUERY;
                continue;
            } else if (!_tcscmp(_T("remove"), token + 1)) {
                state = STATE_REMOVE;
                command = COMMAND_REMOVE;
                continue;
            } else 
                break;
        } else { 
            switch (state) {
                case STATE_NONE:
                    break;
                case STATE_IP:
                    Add2StringItemList(&pIP, token);
                    break;
                case STATE_IF:
                    Add2StringItemList(&pIF, token);
                   break;
                case STATE_TYPE:
                    Add2StringItemList(&pType, token);
                    break;
                case STATE_QUERY:
                    Add2StringItemList(&pQueryType, token);
                    break;
                case STATE_REMOVE:
                    Add2StringItemList(&pRemoveType, token);
                    break;
            }
        }
    }

    /* if it's a query, then process the query and return */
    if (command == COMMAND_QUERY && pQueryType) {
        tmpItem1 = pQueryType;
        do {
            _TCHAR path[200];
            _TCHAR *type = tmpItem1->Value;
            WsDiscoveryUtilLookUp2(iThis, type, 200, path);
            _tprintf(_T("Address:%s\n"), path);
            tmpItem1 = ListNodeNext(sloff, tmpItem1);
        } while (tmpItem1 != pQueryType);
        goto Out;
    }

    /* if it's remove request, process it and return */
    if (command == COMMAND_REMOVE && pRemoveType) {
        tmpItem1 = pRemoveType;
        do {
            _TCHAR *type = tmpItem1->Value;
            for (k = 0; k < svcs; k++) {
                if (!_tcscmp(type, table[k].itype)) 
                    break;
            }
            if (k < svcs) {
                PEndpointReference pEp = WsEndpointNew(table[k].uuid);
                WsDiscoveryUtilUnregisterAdvert(iThis, pEp);
                WsEndpointDelete(pEp, pHeap);
            }
            tmpItem1 = ListNodeNext(sloff, tmpItem1);
        } while (tmpItem1 != pRemoveType);
        goto Out;
    }

    /* if IP address is given, use the ip address instead of if name */
    if (pIP) {
        /* do nothing */
    } else if (pIF) {
        tmpItem1 = pIF;
        do {
            UINT32 addr;
            UINT8 *ptr;
            _TCHAR buf[17];
            _TCHAR *ifname = tmpItem1->Value;
            addr = GetInterfaceIp(ifname, _tcslen(ifname));
            ptr = (UINT8 *) &addr;
            _stprintf(buf, _T("%d.%d.%d.%d"), ptr[0], ptr[1], ptr[2], ptr[3]);
            Add2StringItemList(&pIP, buf);
            tmpItem1 = ListNodeNext(sloff, tmpItem1);
        } while (tmpItem1 != pIF);
    } else {
        DBG(("WsDiscovery: Cannot find out the host address"));
        goto Out;
    }

    if (!pIP || !pType) return sc;

    /* Now, send advertisements */
    tmpItem1 = pIP;
    do {
        _TCHAR *ipaddr = tmpItem1->Value;
        tmpItem2 = pType;
        do {
            _TCHAR *itype = tmpItem2->Value;
            for (k = 0; k < svcs; k++) {
                if (!_tcscmp(itype, table[k].itype)) 
                    break;
            }
            if (k < svcs) {
                _TCHAR path[300];
                _stprintf(path, _T("http://%s/%s"), ipaddr, table[k].path);
                WsDiscoveryUtilRegisterAdvert(iThis, 
                    WsAdvertNew(WsEndpointNew(table[k].uuid),
                                WsQNameItemNew(table[k].itype),
                                WsScopesNew(NULL, WsQNameItemNew(_T("here"))),
                                WsStringItemNew(path),
                                table[k].version));
            }
            tmpItem2 = ListNodeNext(sloff, tmpItem2);
        } while (tmpItem2 != pType);
        tmpItem1 = ListNodeNext(sloff, tmpItem1);
    } while (tmpItem1 != pIP);

METHOD_CLEANUP;
    if (cmd) 
        pHeap->v->Free(pHeap, 0, cmd);
    if (pIP) 
        ListWalk(&pIP, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    if (pIF) 
        ListWalk(&pIF, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    if (pType) 
        ListWalk(&pType, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    if (pQueryType) 
        ListWalk(&pQueryType, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    if (pRemoveType) 
        ListWalk(&pRemoveType, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    return sc;
}

/* Initialization */

PRIVATE PIUNKNOWN CreateWsDiscovery()
{
    PWSDISCOVERY pDisco;
    PIHEAP pHeap = CurrentHeap();
    
    /* Memory allocation */
    DBG(("wsdiscov.c:CreateWsDiscovery: Allocating memory\n"));
    pDisco = (PWSDISCOVERY) pHeap->v->
        Alloc(pHeap,
              HEAP_DEBUG_NAME("WS-Discovery") | HEAP_ZERO_MEMORY,
              sizeof(WSDISCOVERY), 0);
    if (!pDisco) return NULL;

    /* Initialize structure. GenericCobInit initializes super-class */
    DBG(("wsdiscov.c:CreateWsDiscovery: this=0x%x &this=0x%x vd=0x%x vdu=0x%x\n",
         (UINT) pDisco, (UINT) &pDisco,
         (UINT) &WsDiscoveryWsDiscoveryVtbl.v,
         (UINT) &WsDiscoveryUtilWsDiscoveryVtbl.v));
    //pDisco->super.super.v = &WsDiscoveryWsDiscoveryVtbl.v;
    //pDisco->super.super.RefCnt = 1;
    //pDisco->WsDiscoveryUtilv = &WsDiscoveryUtilWsDiscoveryVtbl.v;

    Mutex_Init(&pDisco->mx);
    Condition_Init(&pDisco->cd);

    DBG(("wsdiscov.c:CreateWsDiscovery: Done\n"));
    return GenericCobInit((void*) &wsdiscov_CobDesc, &pDisco->super);
}

PRIVATE SCODE MCT WsDiscoveryUtilEnumerate(
    /*in*/  PIWSDISCOVERYUTIL iThis,
    /*optional*/ PIHEAP Heap,
    /*in*/  PQNameItem Types,
    /*in*/  PScopes Scopes,
    /*out*/ PIENUM *pEnum)
{
    METHOD_DECLARE_NOTHIS(WsDiscoveryUtil, Enumerate);
    PWSDISCOVERY This = pDU(iThis);
    WsDiscoEnum *Enum;
    
    sc = S_OK;
    *pEnum = NULL;

    if (!Heap)
        Heap = CurrentHeap();

    /* Memory allocation */
    DBG(("wsdiscov.c:WsDiscoveryUtilEnumerate: Allocating memory\n"));
    Enum = (WsDiscoEnum *) Heap->v->
        Alloc(Heap,
              HEAP_DEBUG_NAME("WSDiscoEnum") | HEAP_ZERO_MEMORY,
              sizeof(*Enum), 0);
    CHECKSETSC(Enum, E_NOT_ENOUGH_MEMORY, "alloc enum");

    Enum->LastProbe = CurrentTime();

    sc = WsDiscoveryUtilSendProbe(iThis, Types, Scopes);
    CHECKSC("SendProbe");

    /* Save filter for further use */
    Enum->Types = ListCopy(&Types, qnloff,
                           WsQNameItemCopy, WsQNameItemDelete, Heap);
    Enum->Scopes = ListCopy(&Scopes, scloff,
                            WsScopesCopy, WsScopesDelete, Heap);
    Enum->State = DEN_LOCAL;
    Enum->Heap = Heap;
    Heap->v->AddRef(Heap);

    /* Link new enumerator to disco object */
    Mutex_Lock(&This->mx);
    iThis->v->AddRef(iThis);
    Enum->Owner = This;
    ListAddLast(&This->Enumerators, denoff, Enum);
    Mutex_Unlock(&This->mx);

METHOD_CLEANUP;
    DBG(("Lookup done sc=%x", sc));
    return sc;
}

/* An pAd item was deleted, bump each enumerator off it */
PRIVATE INT EnumBump(PTR _pEnum, PTR _pAd)
{
    /* Lock is held */
    WsDiscoEnum *pEnum = _pEnum;
    PWSADVERT pAd = _pAd;
    PWSDISCOVERY pDisco = pEnum->Owner;

    if (pEnum->Current != pAd) {
        /* nothing */;
    } else if (pEnum->State == DEN_LOCAL) {
        assert(pAd&&ListWalk(&pDisco->CacheLocal,aloff,LIST_WALK_SEARCH,
                             NodeEq,pAd)==pAd);
        if (pAd == pDisco->CacheLocal)
            pEnum->Current = NULL;
        else
            pEnum->Current = ListNodePrev(aloff, pAd);
    } else if (pEnum->State == DEN_REMOTE) {
        assert(pAd&&ListWalk(&pDisco->CacheRemote,aloff,LIST_WALK_SEARCH,
                             NodeEq,pAd)==pAd);
        if (pAd == pDisco->CacheRemote)
            pEnum->Current = NULL;
        else
            pEnum->Current = ListNodePrev(aloff, pAd);
    }

    return 0;
}

PRIVATE SCODE MCT WsDiscoEnumDestroy(PIUNKNOWN iThis)
{
    METHOD_DECLARE(WsDiscoEnum, Destroy);
    PWSDISCOVERY pDisco = This->Owner;
    PIHEAP Heap = This->Heap;
    sc = S_OK;

    DBG(("WsDiscoEnumDestroy %x\n", (UINT) This));

    /* Unlink from owning object */
    Mutex_Lock(&pDisco->mx);
    ListRemoveItem(&pDisco->Enumerators, denoff, This);
    ((PIWSDISCOVERY)pDisco)->v->Release((PIWSDISCOVERY) pDisco);
    Mutex_Unlock(&pDisco->mx);

    /* Delete private data */
    ListDelete(&This->Types, qnloff, WsQNameItemDelete, Heap);
    ListDelete(&This->Scopes, scloff, WsScopesDelete, Heap);

    sc = Heap->v->Free(Heap, 0, This);

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE MCT WsDiscoEnumPull(
    PIENUM iThis,
    /*optional*/ PIHEAP Heap,
    /*optional*/ PTIME pTimeout,
    /*out*/ PTR *pEntry,
    /*out*/ _TCHAR *pType)
{
    METHOD_DECLARE(WsDiscoEnum, Pull);
    PWSDISCOVERY pDisco = This->Owner;
    PWSADVERT pAd;
    TIME timeo;
    AdvertMatchTemplate Amt;

    DBG(("WsDiscoEnumDestroy %x\n", (UINT) This));
    sc = S_OK;
    Amt.Types = This->Types;
    Amt.Scopes = This->Scopes;

    if (!Heap)
        Heap = This->Heap;

    Mutex_Lock(&pDisco->mx);
    for (;;) {
        pAd = NULL;
        if (This->Current)
            pAd = ListNodeNext(aloff, This->Current);
        if (This->State == DEN_LOCAL) {
            if (pAd == pDisco->CacheLocal) {
                /* last or none, local exhausted */
                This->Current = NULL;
                This->State = DEN_REMOTE;
                continue;
            }
            if (!pAd)           /* first */
                pAd = pDisco->CacheLocal;
        } else if (This->State == DEN_REMOTE) {
            if (pAd == pDisco->CacheRemote) {
                /* last or none, remote exhausted, waiting */
                This->Current = NULL;
                if (pTimeout && TimeIsRelative(*pTimeout)) {
                    /* Normalize timeout to absolute for the benefit of loop */
                    timeo = CurrentTime();
                    timeo = Int64Subtract(timeo, *pTimeout);
                    pTimeout = &timeo;
                }

                /* Send probe again XXX finesse this and send less */
                sc = WsDiscoveryUtilSendProbe(iDU(pDisco),
                                              This->Types, This->Scopes);
                CHECKSC("SendProbe");

                sc = Condition_Wait(&pDisco->cd, &pDisco->mx, pTimeout);
                CHECKSC("enum wait");
                continue;
            }

            if (!pAd)           /* first */
                pAd = pDisco->CacheRemote;
        } else {
            CHECKSETSC(FALSE, E_END_OF_DATA, "enum exhausted");
        }
        This->Current = pAd;
        if (WsAdvertMatchCmp(pAd, &Amt) == 0)
            break;              /* match! */
        else
            continue;           /* skip this entry */
    }

    if (pAd)
        pAd = WsAdvertCopy(pAd, Heap);
    
    *pEntry = pAd;
    if (pAd && pType)
        _tcscpy(pType, _T("wsd:WSADVERT"));
    //*pType = QID_ENUM_TYPE_DISCOVERY;

METHOD_CLEANUP;
    Mutex_Unlock(&pDisco->mx);
    return sc;
}

PRIVATE SCODE MCT WsDiscoEnumReleasePullResult(PIENUM iThis,
                                               PIHEAP Heap,
                                               PTR Entry,
                                               const _TCHAR *Type)
{
    METHOD_DECLARE(WsDiscoEnum, ReleasePullResult);
    DBG(("WsDiscoEnumReleasePullResult %x\n", (UINT) iThis));
    CHECKSETSC(!_tcscmp(Type, _T("wsd:WSADVERT")), E_INVALID_PARAMETER,
               "type check");

    if (!Heap)
        Heap = This->Heap;

    WsAdvertDelete(Entry, Heap);

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}

PIUNKNOWN wsdiscovCobMain(void)
{
    return CreateWsDiscovery();
}
