/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* wsman interop test */

//#define _DEBUG 1

#include <stdio.h>
#include <mmlite.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include<util/string.h>
#include <httpif.h>
#include <sampleif.h>           /* for test */
#include <wsifhelp.h>
#ifdef _WITH_PRIVATE_INCLUDES
#include <cimif.h>
#endif
#include <base/cobdesc.h>

#define MEASURE 1
#define MEASUREMENT_COUNT 1000

#ifdef MEASURE
PRIVATE BOOL doMeasure = FALSE;
#endif

SCODE ShowBmp(PIBUFFER im);
SCODE GetFile(const _TCHAR *filename, PIBUFFER *pBuf, /*inout*/ UINT *pSize);

#define CIM_NS _T("http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2")
#define WINRM_NS _T("http://schemas.microsoft.com/wbem/wsman/1")
#define IPMI_NS WINRM_NS _T("/wmi/root/hardware")
#define WSMAN_NS _T("http://schemas.dmtf.org/wbem/wsman/1")

#define AnyItemOff offsetof(struct AnyItem, ListNode)
#define itemsoff offsetof(struct Items, ListNode)
#define EprOff offsetof(struct EndpointReference, ListNode)

typedef struct _Enumerator Enumerator;
struct _PULL_STATE;

struct _Enumerator {
    TIME Expires;
    struct IID Context;
    PIENUM sub;
    ListNodeDeclare(_Enumerator, ListNode);
    //PTR current;
    struct _PULL_STATE *PullState;
#define enoff offsetof(struct _Enumerator, ListNode)
};

/* Create a pseudo-random GUID for use in enumerator */
PRIVATE void SetRandomIid(struct IID *iid)
{
    /* XXX not really random or unique */
    static UINT seqno = 1;
    * (TIME*) iid->Data4 = CurrentTime();
    iid->Data1 = seqno++;
}

PRIVATE Enumerator *EnumeratorNew(PIHEAP Heap)
{
    Enumerator *This;
    This = (void *) Heap->v->
        Alloc(Heap, HEAP_DEBUG_NAME("wsman:Enumerator") | HEAP_ZERO_MEMORY,
              sizeof(*This), 0);
    if (!This)
        return NULL;

    This->Expires = TIME_FOREVER;
    SetRandomIid(&This->Context);

    return This;
}

/* Delete a resource. Use different name to not to confuse with wxf method. */
PRIVATE INT EnumeratorDelete(PTR pItem, PTR pArg) 
{
    Enumerator *This = pItem;
    PIHEAP Heap = pArg;

    RELEASE(This->sub);

    Heap->v->Free(Heap, 0, This);
    return 0;
}

PRIVATE INT EnumeratorMatch(PTR pItem, PTR pArg) 
{
    Enumerator *This = pItem;
    PIID guid = pArg;
    BOOL same = UuidCmp(guid, &This->Context);

    return !same;
}


typedef struct _WsMan {
    //    CONST_VTBL struct IProgramVtbl *progv;
    //    UINT RefCnt;
    struct _DefCob super;

    CONST_VTBL struct IWsTransferVtbl *wxfv;
    CONST_VTBL struct IWsManIdVtbl *wsmidv;
    CONST_VTBL struct IWsEnumerationVtbl *wsenv;
    CONST_VTBL struct IWsEventingVtbl *wsev;
    CONST_VTBL struct ICIM_RecordLogVtbl *cimsv;
#ifdef _WITH_PRIVATE_INCLUDES
    CONST_VTBL struct IWsShellVtbl *wsshv;
#endif

    MUTEX mx;
    PTR ShellList;
    struct _Enumerator *Enumerators;
    PICIMRESOURCE Cim;
} WsMan;

#include "s_WsMan_v.c"

/* Convert between interface and instance pointer and vice versa */
#define pprog(_i_) ((WsMan *)((_i_)))
#define iprog(_p_) ((PIWSTRANSFER)((_p_)))
#define pwxf(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, wxfv) : 0))
#define iwxf(_p_) ((PIWSTRANSFER)((_p_) ? &(_p_)->wxfv : NULL))
#define pwsmid(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, wsmidv) : 0))
#define iwsmid(_p_) ((PIWSMANID)((_p_) ? &(_p_)->wsmidv : NULL))
#define pwsen(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, wsenv) : 0))
#define iwsen(_p_) ((PIWSENUMERATION)((_p_) ? &(_p_)->wsenv : NULL))
#define pwse(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, wsev) : 0))
#define iwse(_p_) ((PIWSEVENTING)((_p_) ? &(_p_)->wsev : NULL))
#define pcims(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, cimsv) : 0))
#define icims(_p_) ((PICIM_RecordLog)((_p_) ? &(_p_)->cimsv : NULL))
#define pwssh(_i_) ((WsMan *)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _WsMan, wsshv) : 0))
#define iwssh(_p_) ((PIWSSHELL)((_p_) ? &(_p_)->wsshv : NULL))

#if 0
PRIVATE SCODE wxfQueryInterface (
    PIWSTRANSFER iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                               &IID_IWsTransfer);
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IWsEnumeration, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iwsen(pwxf(iThis));
        sc = S_OK;
    }

    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IWsEventing, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iwse(pwxf(iThis));
        sc = S_OK;
    }

    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IWsManId, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iwsmid(pwxf(iThis));
        sc = S_OK;
    }

    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_ICIM_RecordLog, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) icims(pwxf(iThis));
        sc = S_OK;
    }

#ifdef _WITH_PRIVATE_INCLUDES
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IWsShell, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iwssh(pwxf(iThis));
        sc = S_OK;
    }
#endif

    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IProgram, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iprog(pwxf(iThis));
        sc = S_OK;
    }

    return sc;
}

PRIVATE UINT wxfAddRef (
    PIWSTRANSFER iThis)
{
    return AtomicInc(&pwxf(iThis)->super.super.RefCnt);
}

PRIVATE UINT wxfRelease (
    PIWSTRANSFER iThis)
{
    WsMan * This = pwxf(iThis);
    UINT Refs = AtomicDec(&This->super.super.RefCnt);
    if (Refs == 0)
        WsManDestroy((PIUNKNOWN) iThis);

    return Refs;
}
#endif

PRIVATE SCODE WsManDestroy(IUnknown *iThis)
{
    METHOD_DECLARE(WsMan, Destroy);
    sc = S_OK;

    RELEASE_NULL(This->Cim);
    ListDelete(&This->Enumerators, enoff, EnumeratorDelete, CurrentHeap());

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE wxfQueryInterface (
    PIWSTRANSFER iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    IUnknown *unk = (IUnknown *) pwxf(iThis);
    return unk->v->QueryInterface(unk, iid, ppObject);
}

PRIVATE UINT wxfAddRef (
    PIWSTRANSFER iThis)
{
    IUnknown *unk = (IUnknown *) pwxf(iThis);
    return unk->v->AddRef(unk);
}

PRIVATE UINT wxfRelease (
    PIWSTRANSFER iThis)
{
    IUnknown *unk = (IUnknown *) pwxf(iThis);
    return unk->v->Release(unk);
}

struct Listener listener = {
    _T("172.30.171.64"),
    443,
    _T("jvh-test"),
    _T(""),
    TRUE,
    _T("wsman"),
    _T("HTTPS"),
    _T("a56e126611d4b8a006f24b44331549379621c53d")
};

struct Roles r1 = {
    _T("Hardware Management Controller"),
    {&r1, &r1}
};

struct Win32_ComputerSystem w32c = {
    2,
    FALSE, TRUE, NULL, NULL, FALSE,

    _T("OK"),
    _T("Invisible Device"),
    1,
    _T("ComputerSystem"),

    0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,
    &r1,
    0,
};

struct CIM_ComputerSystem cimc = {
    _T("BMC 32"),
    _T("BMC 32 - IPMI v1.5 - fw v0.30.0.0.0.0"),
    _T("IPMI Controller 32"),

    _T("ComputerSystem"),
    _T("IPMI Controller 32"),
    &r1,

    _T("Other"),
    _T("Manufacturer ID-Product ID"),
    _T("0-0"),
    14, 
};

struct SelectorSetRef sr = {
    _T("GetSelector")
};

struct SchemaRef scr = {
    _T("http://schemas.dmtf.org/cimv2.9/CIM_Schema.xsd"),
    _T("ac:CIM_Processor")
};

struct Operation op = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/transfer/Get"),
    &sr,
    &scr
};

struct Selector2 s2 = {
    _T("DeviceID"),
    _T("xsd:string"),
    NULL,
    {&s2, &s2}
};

struct SelectorSet2 ss = {
    _T("GetSelector"),
    &s2
};

struct Access2 acc2 = {
    WINRM_NS,
    //&op,
    //&ss
    NULL,
        NULL,
    {&acc2, &acc2}
};

struct Resource cat = {
    _T("wsman:root.cim/CIM_Processor"),
    _T("1.0"),
    _T("Returns version and other information of the processor resource"),
    _T("ACME"),
    &acc2
};

static struct PossibleStates2 pos9[8] = {
    {_T("OK"), {&pos9[1], &pos9[7]}},
    {_T("N/A"), {&pos9[2], &pos9[0]}},
    {_T("at or below lower non-critical threshold"), {&pos9[3], &pos9[1]}},
    {_T("at or below lower critical threshold"), {&pos9[4], &pos9[2]}},
    {_T("at or below lower non-recoverable threshold"), {&pos9[5], &pos9[3]}},
    {_T("at or above upper non-critical threshold"), {&pos9[6], &pos9[4]}},
    {_T("at or above upper critical threshold"), {&pos9[7], &pos9[5]}},
    {_T("at or above upper non-recoverable threshold"), {&pos9[0], &pos9[6]}}
};

static struct CIM_NumericSensor sen9 = {
    _T("Temperature(10.0.32)"),
    _T("PCI-X Slot Temp(10.0.32):Temperature for Other 0"),
    _T("PCI-X Slot Temp(10.0.32)"),

#if 0
    Int64Initializer(0,0), _T("Name"), _T("OK"), _T(""),

    _T("NumericSensor"), /* CreationClassName> */
    _T("10.0.32"),
    _T("ComputerSystem"),
    _T("IPMI Controller 32"),

    2,
    _T(""),
    pos9,
    _T("OK"),

    2,
    -1,
    0,
    240,
    300,
    350,                     /* NormalMax */
    250,
    2000,
    -550,
    0,
    0,
    TRUE,                       /* IsLinear */
    0,
    100, 400,
    -550, 550,
#endif
};

static struct Items item9 = {
    &sen9,
    _T("p:CIM_NumericSensor"),
    {&item9, &item9}
};

static struct PossibleStates2 pos1[7] = {
    {_T("N/A"), {&pos1[1], &pos1[6]}},
    {_T("at or below lower non-critical threshold"), {&pos1[2], &pos1[0]}},
    {_T("at or below lower critical threshold"), {&pos1[3], &pos1[1]}},
    {_T("at or below lower non-recoverable threshold"), {&pos1[4], &pos1[2]}},
    {_T("at or above upper non-critical threshold"), {&pos1[5], &pos1[3]}},
    {_T("at or above upper critical threshold"), {&pos1[6], &pos1[4]}},
    {_T("at or above upper non-recoverable threshold"), {&pos1[0], &pos1[5]}}
};

static struct CIM_NumericSensor sen1 = {
    _T("Temperature(10.0.32)"),
    _T("PCI-X Slot Temp(10.0.32):Temperature for Other 0"),
    _T("PCI-X Slot Temp(10.0.32)"),

#if 0
    Int64Initializer(0,0), _T("Name"), _T("OK"), _T("OK"),

    _T("NumericSensor"), /* CreationClassName> */
    _T("10.0.32"),
    _T("ComputerSystem"),
    _T("IPMI Controller 32"),

    2,
    _T(""),
    pos1,
    _T("OK"),

    2,
    -1,
    0,
    240,
    300,
    350,                     /* NormalMax */
    250,
    2000,
    -550,
    0,
    0,
    TRUE,                       /* IsLinear */
    0,
    20, 400,
    -550, 550,
    -550, -550
#endif
};

struct Auth2 au = {TRUE, FALSE, FALSE, FALSE, {&au,&au}};
struct DefaultPorts dp = {80, 0, {&dp,&dp}};
struct Client cl = {5000, _T("wsman"), TRUE,
                    &au, &dp,
                    _T("*"), {&cl,&cl}};
struct Service sv = {_T("O:NSG:BAD:P(A;;GA;;;BA)S:P(AU;FA;GA;;;WD)(AU;SA;GWGX;;;WD)"),
                     100, 60000, 5, TRUE,
                     &au, &dp,
                     _T("*"), _T(""), {&sv,&sv}};
struct Winrs wr = {TRUE, 900000, 1, 28800000, 2, 1, 2, {&wr,&wr}};
struct Config cfg = {150, 60000, 20, 25,
                     &cl, &sv, &wr};

PRIVATE SCODE cimsQueryInterface (
    PICIM_RecordLog iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return wxfQueryInterface(iwxf(pcims(iThis)), iid, ppObject);
}

PRIVATE UINT cimsAddRef (
    PICIM_RecordLog iThis)
{
    return wxfAddRef(iwxf(pcims(iThis)));
}

PRIVATE UINT cimsRelease (
    PICIM_RecordLog iThis)
{
    return wxfRelease(iwxf(pcims(iThis)));
}

PRIVATE SCODE cimsClearLog (
    PICIM_RecordLog iThis,
    PClearLog_INPUT ClearLog_INPUT,
    /*out*/ PClearLog_OUTPUT clearout)
{
    clearout->ReturnValue = 42;
    return S_OK;
}

#ifdef _WITH_PRIVATE_INCLUDES
#include "cimtest.h"
#endif

static struct Items item1 = {
    &sen1,
    _T("cimns:CIM_NumericSensor"),
    {&item1, &item1}
};

static struct Items itemcs = {
    &cimc,
    _T("p:CIM_ComputerSystem"),
    {&itemcs, &itemcs}
};

static struct _PULL_STATE {
    const _TCHAR *context;
    struct Items *item;
    BOOL End;
} PullState[] = {
    {_T("uuid:B70EB2CB-C545-4B1E-A1E3-F4997542CA91"), &item1, FALSE},
    {_T("uuid:B12202DB-51C1-45A3-8711-03BEA2B24F82"), &item9, TRUE},
    {_T("uuid:70B47D50-7432-437e-9C9D-AC194EA6D073"), NULL, TRUE},
    {_T("uuid:70B47D50-7432-437e-9C9D-000000000000"), &itemcs, TRUE},
    {_T("uuid:70B47D50-7432-437e-9C9D-000000000001"), NULL, TRUE},
    {0,}
};


extern struct Action Act1, Act2, Act3, Act4, Act5;

struct Action Act1 = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/transfer/Get"),
    {&Act2, &Act5}
};

struct Action Act2 = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/transfer/Put"),
    {&Act3, &Act1}
};

struct Action Act3 = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/transfer/Create"),
    {&Act4, &Act2}
};

struct Action Act4 = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete"),
    {&Act5, &Act3}
};

struct Action Act5 = {
    _T("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate"),
    {&Act1, &Act4}
};

extern struct Auth Auth1, Auth2;

struct Auth Auth1 = {
    _T("wsman:auth/basic"),
    {&Auth2, &Auth2}
};

struct Auth Auth2 = {
    _T("wsman:auth/https-mutual"),
    {&Auth1, &Auth1}
};

struct Transport tran1 = {
    _T("http://www.ietf.org/rfc/rfc2818.txt"),
    &Auth1,
    ListNodeInitializer(tran1)
};

struct Access acc = {
    WINRM_NS,
    //&op,
    //&ss
    NULL,
    {&acc, &acc}
};

struct This WsManThis = {
    _T("Microsoft Corporation"),
    _T("OS: 5.2.3790 SP: 1.0 Stack: 1.0"),
    &acc,
    &tran1
};

struct BitmapImage img0 = {
    _T("bmp"), NULL
};

/* XXX dig out ResourceURI from header processor */
#include <web/cont.h>
const PWSCONTEXT GetContext(void)
{
    PITHREAD thd = CurrentThread();
    PCONTINFO cont = (PCONTINFO) thd->v->GetValue(thd,
                                               THREAD_VALUE_RPC_CONTINUATION);
    if (!cont)
        return NULL;
    return (PWSCONTEXT) cont->pContext;
}

PRIVATE SCODE wsmidQueryInterface (
    PIWSMANID iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return wxfQueryInterface(iwxf(pwsmid(iThis)), iid, ppObject);
}

PRIVATE UINT wsmidAddRef (
    PIWSMANID iThis)
{
    return wxfAddRef(iwxf(pwsmid(iThis)));
}

PRIVATE UINT wsmidRelease (
    PIWSMANID iThis)
{
    return wxfRelease(iwxf(pwsmid(iThis)));
}

PRIVATE SCODE wsmidIdentify (
    PIWSMANID iThis,
    _TCHAR *ProtocolVersion,
    UINT ProtocolVersionSize,
    _TCHAR *ProductVendor,
    UINT ProductVendorSize,
    _TCHAR *ProductVersion,
    UINT ProductVersionSize,
    _TCHAR *SpecVersion,
    UINT SpecVersionSize,
    _TCHAR *BuildId,
    UINT BuildIdSize)
{
    WsMan * This = pwsmid(iThis);

    if (ProtocolVersionSize < sizeof(_TCHAR)
        || ProductVendorSize < sizeof(_TCHAR)
        || ProductVersionSize < sizeof(_TCHAR)
        || SpecVersionSize < sizeof(_TCHAR)
        || BuildIdSize < sizeof(_TCHAR))
        return E_INVALID_PARAMETER;

    _tcsncpy(ProtocolVersion, WSMAN_NS _T("/wsman.xsd"), ProtocolVersionSize/sizeof(_TCHAR));
    _tcsncpy(ProductVendor,
             _T("Microsoft Corporation"), ProductVendorSize/sizeof(_TCHAR));
    _tcsncpy(ProductVersion,
             _T("Embedded Toolkit 0.106"), ProductVersionSize/sizeof(_TCHAR));
    _tcsncpy(SpecVersion,
             _T("1.0.0a"), ProductVersionSize/sizeof(_TCHAR));
    _tcsncpy(BuildId,
             _T("106"), BuildIdSize/sizeof(_TCHAR));

    ProtocolVersion[ProtocolVersionSize/sizeof(_TCHAR)] = _EOS;
    ProductVendor[ProductVendorSize/sizeof(_TCHAR)] = _EOS;
    ProductVersion[ProductVersionSize/sizeof(_TCHAR)] = _EOS;
    SpecVersion[ProductVersionSize/sizeof(_TCHAR)] = _EOS;
    BuildId[BuildIdSize/sizeof(_TCHAR)] = _EOS;

    return S_OK;
}

PRIVATE INT SelectorCmpName(PTR s1, PTR arg)
{
    PSelector sel1 = s1;
    _TCHAR *name = arg;

    return _tcscmp(sel1->Name, name);
}
#define Selector_off offsetof(struct Selector, ListNode_Selector)

PRIVATE const _TCHAR *CimBaseName(const _TCHAR *name)
{
    if (!_tcsncmp(name, CIM_NS _T("/"), sizeof(CIM_NS)/sizeof(_TCHAR)))
        return name + sizeof(CIM_NS)/sizeof(_TCHAR);
    if (!_tcsncmp(name, _T("wsman:"), sizeof(_T("wsman"))/sizeof(_TCHAR)))
        return name + sizeof(_T("wsman"))/sizeof(_TCHAR);
    return name;
}

PRIVATE const _TCHAR* GetLocal(const _TCHAR *Url)
{
    PISOAPFACTORY SoapFactory = NULL;
    const _TCHAR* pPath;
    BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &SoapFactory);
    SoapFactory->v->ParseURL(SoapFactory, Url, NULL, 
                                  NULL, NULL, 
                                  NULL, &pPath,NULL, NULL,NULL,NULL);
     RELEASE(SoapFactory); 
    return pPath;
}

PRIVATE const PICIMRESOURCE GetCim(WsMan * This)
{
    if (!This->Cim) {
        PICIMRESOURCE obj;
        SCODE sc = BindToObject(CurrentNameSpace(), _T("COB/cimdb.cob"),
                                NAME_SPACE_READ | NAME_SPACE_WRITE,
                                &IID_ICimResource, (void **) &obj);
        if (FAILED(sc)) {
            DBG(("GetCimFactory failed to bind to COB/cimdb.cob\n"));
        } else {
            /* In case somebody beat us to it, zap the extra */
            if (!AtomicCmpAndSwap((ADDRESS*) &This->Cim,
                                  0, (ADDRESS) obj))
                obj->v->Release(obj);
        }
    }
    return This->Cim;
}

PRIVATE void WsManFault(WsMan *This, const _TCHAR *Text)
{
    METHOD_DECLARE_NOTHIS(WsMan, Fault);
    struct Code subcode = {0,};
    struct Code code = {0, };

    PITHREAD thd = CurrentThread();
    PICONTINUATION cont = (PICONTINUATION) thd->v->GetValue(thd,
                                                THREAD_VALUE_RPC_CONTINUATION);
    CHECKSETSC(cont, E_FAIL, "get current continuation");

    code.Value = _T("soap:Sender");
    subcode.Value = _T("tpolicy:EvidenceRequired");
    code.Subcode = &subcode;

    sc = cont->v->SendMessage(cont, NULL, &code, Text);
    CHECKSC("cont->SendMessage fault message");

METHOD_CLEANUP;
}

PRIVATE SCODE wxfGet (
    PIWSTRANSFER iThis,
    const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet, TIME OperationTimeout,
    void *Value, _TCHAR *pValueType, UINT ValueSize)
{
    METHOD_DECLARE_MULTI(WsMan, wxf, Get);
    SOAPStruct ss = {123, _T("BackTest"), 77777};
    const PWSCONTEXT ctx = GetContext();
    const _TCHAR *local;
   // _TCHAR *FaultMessage1 = NULL ;
    //_TCHAR *FaultMessage ;
    FaultTemplate Fault1;
   
    local = GetLocal(To);
    sc = S_OK;

    CHECKSETSC (ctx && ResourceURI, E_INVALID_PARAMETER, "param check");

    /* Fault generation for test cases:
     * 6.6 If OpTimeout == 1us send timeout fault wsman:TimedOut
     *     Say < 10ms sends a fault.
     *    
     * 6.5 If there are too few selectors (only CreationClassName)
     *     send back wsman:InvalidSelectors
     *
     * 6.4 If MaxEnvelopeSize < 10
     *     send wsman:EncodingLimit
     */
    if (!Int64IsZero(OperationTimeout)) {
        TIME cmp;
        Int32ToInt64(cmp, TIME_RELATIVE(TIME_MILLIS(10)));

        /* The time values are negative */
        CHECKSETSC(!Int64Less(cmp, OperationTimeout),
                   E_WSMAN_TIMED_OUT,
                   "wsmantest/wxfGet: OperationTimeout below minimum");
    }

    if (ctx->MaxEnvelopeSizeValue < 10
        && ctx->MaxEnvelopeSizeValue > 0)
    {
        CHECKSETSC(FALSE, E_WSMAN_ENCODING_LIMIT, "envelop size check");
    }

    if (!_tcscmp(ResourceURI, _T("wsman:system/2005/06/this"))) {
        memcpy(Value, &WsManThis, sizeof(WsManThis));
        _tcscpy(pValueType, _T("wsman:This"));        
#if 0
    } else if (!_tcscmp(ResourceURI,
                        CIM_NS _T("/CIM_NumericSensor"))
               || !_tcscmp(ResourceURI,
                           CIM_NS _T("/cim_numericsensor")))
    {
        struct Selector *sel;

        /* 6.4, 6.5 look at selectors: Find DeviceId */
        CHECKSETSC(SelectorSet, E_WSMAN_INVALID_SELECTORS,
                   "verify there are selector");

        sel = ListWalk(&SelectorSet, Selector_off, LIST_WALK_SEARCH,
                       SelectorCmpName, _T("DeviceID"));
        CHECKSETSC(sel, E_WSMAN_INVALID_SELECTORS, "SelectorCmp search");

        memcpy(Value, &sen9, sizeof(sen9));
        _tcscpy(pValueType, _T("cimns:CIM_NumericSensor"));        
#endif
    } else if (!_tcscmp(ResourceURI, WINRM_NS _T("/config"))) {
        memcpy(Value, &cfg, sizeof(cfg));
        _tcscpy(pValueType, _T("cfg:Config"));        
    } else if (!_tcscmp(ResourceURI, CIM_NS _T("/CIM_ComputerSystem"))
               || !_tcscmp(ResourceURI, IPMI_NS _T("/CIM_ComputerSystem")))
    {
        memcpy(Value, &cimc, sizeof(cimc));
        _tcscpy(pValueType, _T("p:CIM_ComputerSystem"));        
    } else if (!_tcscmp(ResourceURI,
                        CIM_NS _T("/ComputerSystem")))
    {
        memcpy(Value, &w32c, sizeof(w32c));
        _tcscpy(pValueType, _T("p:Win32_ComputerSystem"));        
    } else if (!_tcscmp(ResourceURI,
                 _T("wsman:microsoft.com/wsman/2005/06/config/Listener"))) {
        memcpy(Value, &listener, sizeof(listener));
        _tcscpy(pValueType, _T("cfg2:Listener"));
    } else if (!_tcscmp(ResourceURI,
                        _T("wsman:system/catalog/2005/06/Catalog"))) {
        memcpy(Value, &cat, sizeof(cat));
        _tcscpy(pValueType, _T("cat:Resource"));        
    } else if (!_tcscmp(ResourceURI, _T("TestResource"))) {
        memcpy(Value, &ss, sizeof(ss));
        _tcscpy(pValueType, _T("tnstype:SOAPStruct"));
    } else if (ResourceURI[0] == _T('@')) {
        /* Assume it is a bitmap file */
        PIBUFFER im;
        UINT size;
        SCODE sc = GetFile(ResourceURI + 1, &im, &size);
        CHECKSC("GetFile");

        img0.Content = im;
        memcpy(Value, &img0, sizeof(img0));
        _tcscpy(pValueType, _T("sensor:BitmapImage"));
//    } else if (!_tcscmp(ResourceURI, IPMI_NS _T("/cim_computersyste"))) {
//        CHECKSETSC(FALSE, E_WSA_DESTINATION_UNREACHABLE, "bad resource name");
    } else {
        /* Use ResourceURI first piece to bind to sub-handler.
         * Give remaining ResourceURI as arg.
         */
        const _TCHAR *base = CimBaseName(ResourceURI);
        PICIMRESOURCE sub = GetCim(This);

        CHECKSETSC(sub, E_WSA_DESTINATION_UNREACHABLE,"get resource provider");
        sc = sub->v->GetResource(sub,local, base,
                         SelectorSet, OperationTimeout,
                         Value, pValueType, ValueSize,&Fault1 );
  #if MEASURE
        if (doMeasure) {
            TIME t0,t1,timediff;
            int i;
            t0 = CurrentTime();
            for(i = 0; i < MEASUREMENT_COUNT; i++) {
                sc = sub->v->GetResource(sub,local, base,
                         SelectorSet, OperationTimeout,
                         Value, pValueType, ValueSize,&Fault1 );
            }
            t1=CurrentTime();
            timediff = Int64Subtract(t1, t0);
            printf("Time Taken for CimResourceGetResource: %d\n", timediff);
        }
#endif      
        
        CHECKSC("resource Get");


    }
METHOD_CLEANUP;
    if (sc == E_ACCESS_DENIED) {
        WsManFault(This, Fault1.FaultMessage);
        sc = S_NO_RESPONSE;
    }
    return sc;
}

PRIVATE SCODE wxfPut (
    PIWSTRANSFER iThis,
    const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet,
    const void *NewValue, const _TCHAR *NewValueType, const UINT32 *FieldSel,
    void *Value, _TCHAR *pValueType, UINT ValueSize)
{
    METHOD_DECLARE_MULTI(WsMan, wxf, Put);

    /* Use ResourceURI first piece to bind to sub-handler.
     * Give remaining ResourceURI as arg.
     */
    PICIMRESOURCE sub;
    const _TCHAR *base;
    const _TCHAR *local;
  //  _TCHAR *FaultMessage = NULL ;
    FaultTemplate Fault;

    local = GetLocal(To);
#if 0
    if (!_tcscmp(ResourceURI, CIM_NS _T("/CIM_NumericSensor"))) {
        struct Win32_ComputerSystem *wcs = (struct Win32_ComputerSystem*) NewValue;
        struct SelectorSet *sss = (void*) wcs;

        DBG(("NewValue =%x \n", wcs));
        UnusedParameter(This);

        if (!NewValueType) {
            DBGT((_T("WsTransferPut/NULL\n")));
        } else {
            DBGT((_T("WsTransferPut/type=\"%s\"\n"), NewValueType));
        }

        if (NewValue) {
            memcpy(Value, NewValue, ValueSize);
            _tcscpy(pValueType, NewValueType);
        } else {
            memcpy(Value, &sen9, sizeof(sen9));
            _tcscpy(pValueType, _T("cimns:CIM_NumericSensor"));
        }

        return S_OK;
    }
#endif

    base = CimBaseName(ResourceURI);
    sub = GetCim(This);
    CHECKSETSC(sub, E_WSA_DESTINATION_UNREACHABLE, "get resource provider");

    sc = sub->v->ChangeResource(sub,0,local,  base, SelectorSet,
                             NewValue, NewValueType, FieldSel,
                             Value, pValueType, ValueSize,&Fault);

    
#if MEASURE
        if (doMeasure) {
            TIME t0,t1,timediff;
            int i;
            t0 = CurrentTime();
            for(i = 0; i < MEASUREMENT_COUNT; i++) {
                sc = sub->v->ChangeResource(sub,0,local,  base, SelectorSet,
                             NewValue, NewValueType, FieldSel,
                             Value, pValueType, ValueSize,&Fault);
            }
            t1=CurrentTime();
            timediff = Int64Subtract(t1, t0);
            printf("Time Taken for CimResourcePut: %d\n", timediff);
        }
#endif
   CHECKSC("resource Put");
METHOD_CLEANUP;
  if (sc == E_ACCESS_DENIED) {
        WsManFault(This,  Fault.FaultMessage);
        sc = S_NO_RESPONSE;
    }
    return sc;
}

PRIVATE SCODE wxfDelete (
    PIWSTRANSFER iThis,
    const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet)
{
    METHOD_DECLARE_MULTI(WsMan, wxf, Delete);
    const _TCHAR *base = CimBaseName(ResourceURI);
    PICIMRESOURCE sub = GetCim(This);
    const _TCHAR *local;
  //  _TCHAR *FaultMessage = NULL;
    FaultTemplate Fault;

    local = GetLocal(To);
    CHECKSETSC(sub, E_WSA_DESTINATION_UNREACHABLE,"get resource provider");

    sc = sub->v->ChangeResource(sub, RESOURCE_DELETE,local, base, SelectorSet,
                                NULL, NULL, NULL,
                                NULL, NULL, 0, &Fault);
    CHECKSC("resource Delete");

METHOD_CLEANUP;
  if (sc == E_ACCESS_DENIED) {
        WsManFault(This,  Fault.FaultMessage);
        sc = S_NO_RESPONSE;
    }
    return sc;
}

PRIVATE SCODE wxfCreate (
    PIWSTRANSFER iThis,
    const _TCHAR *To,
    const _TCHAR *ResourceURI,
    const void *InitialValue,
    const _TCHAR *ValueType,
    /*out*/ PEndpointReference_QUIET ResourceCreated)
{
    METHOD_DECLARE_MULTI(WsMan, wxf, Create);
    PICIMRESOURCE sub = NULL;
    PAnyItem Item = NULL;
    PIHEAP Heap = NULL;
    const _TCHAR *base;
    const _TCHAR *local;
  //  _TCHAR *FaultMessage;
    FaultTemplate Fault;

    local = GetLocal(To);

    memset(ResourceCreated, 0, sizeof(*ResourceCreated));

    if (!_tcscmp(ResourceURI, _T("test"))) {
        ResourceCreated->Address = _T("TestCreatedResource");
        goto Out;
    }

    base = CimBaseName(ResourceURI);

#if 0
    sc = BindToObject(CurrentNameSpace(), _T("COB/cimdb.cob"),
                          NAME_SPACE_READ | NAME_SPACE_WRITE,
                          &IID_ICimFactory, (void **) &sub);
    if (SUCCEEDED(sc)) {
        PITHREAD thd = CurrentThread();
        PCONTINFO cont = (PCONTINFO) thd->v->GetValue(thd,
                                               THREAD_VALUE_RPC_CONTINUATION);
        CHECKSETSC(cont, E_FAIL, "get current continuation");
        Heap = cont->TmpHeap;
        CHECKSETSC(Heap, E_FAIL, "get cont heap");

        /* No worries about leaks here */
        Item = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, sizeof(*Item), 0);
        CHECKSETSC(Item, E_FAIL, "alloc ref property SelectorSet");

        ListNodeInit(AnyItemOff, &Item);
        ListAddLast(&ResourceCreated->ReferenceParameters, EprOff, Item);
        Item->ValueType = _T("wsman:SelectorSet");

        /* XXX fixup address */
        ResourceCreated->Address = _T("http://127.0.0.1/wsman");
    
        sc = sub->v->CreateResource(sub, _T("COB/cimdb.cob"),
                                    InitialValue, ValueType,
                                    cont->TmpHeap, TRUE,
                                    (PSelectorSet) &Item->Value);
        CHECKSC("CreateResource");
        goto Out;
    }
#endif
    sub = GetCim(This);
    CHECKSETSC(sub, E_WSA_DESTINATION_UNREACHABLE,"get resource provider");

    
    sc = sub->v->ChangeResource(sub, NAME_SPACE_CREATE | NAME_SPACE_FAILIFEXIST | RESOURCE_RETURN_EPR,local,
                                base, NULL,
                                InitialValue, ValueType, NULL,
                                ResourceCreated, NULL, sizeof(*ResourceCreated), &Fault );
    CHECKSC("sub wxf:Create");

METHOD_CLEANUP;
    if (Heap && Item)
        Heap->v->Free(Heap, 0, Item);
 if (sc == E_ACCESS_DENIED) {
        WsManFault(This, Fault.FaultMessage);
        sc = S_NO_RESPONSE;
    }
    return sc;
}

PRIVATE SCODE wsenQueryInterface (
    PIWSENUMERATION iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return wxfQueryInterface(iwxf(pwsen(iThis)), iid, ppObject);
}

PRIVATE UINT wsenAddRef (
    PIWSENUMERATION iThis)
{
    return wxfAddRef(iwxf(pwsen(iThis)));
}

PRIVATE UINT wsenRelease (
    PIWSENUMERATION iThis)
{
    return wxfRelease(iwxf(pwsen(iThis)));
}

PRIVATE SCODE wsenEnumerate (
    PIWSENUMERATION iThis,
    const _TCHAR *To,
    const _TCHAR *ResourceURI,
    PEnumerationMode EnumerationMode,
    POptimizeEnumeration Optimize,
    /*out,optional*/ TIME *Expires,
    /*out*/ _TCHAR * EnumerationContext,
    UINT EnumerationContextSize)
 {
    METHOD_DECLARE_MULTI(WsMan, wsen, Enumerate);
    Enumerator *Enumerator = NULL;
    RESOURCE_FLAGS Flags = 0;
    const _TCHAR *local;
    //_TCHAR *FaultMessage = NULL ;
    FaultTemplate Fault;

    local = GetLocal(To);
    CHECKSETSC(ResourceURI && EnumerationContext, E_INVALID_PARAMETER,
               "NULL ResourceURI or EnumerationContext");
    CHECKSETSC(EnumerationContextSize > 37 * sizeof(_TCHAR),
               E_BUFFER_TOO_SMALL,
               "make sure GUID fits into EnumerationContext");

    Enumerator = EnumeratorNew(CurrentHeap());
    CHECKSETSC(Enumerator, E_INVALID_PARAMETER, "alloc enumerator");

    if (EnumerationMode && EnumerationMode->Value
        && !_tcscmp(EnumerationMode->Value, _T("EnumerateEPR")))
    {
        Flags |= RESOURCE_RETURN_EPR;
    }
    sc = S_OK;
    if (!_tcscmp(ResourceURI, IPMI_NS _T("/CIM_ComputerSystem"))) {
        Enumerator->PullState = &PullState[3];
    } else if (!_tcscmp(ResourceURI, CIM_NS _T("/CIM_NumericSensor"))) {
        Enumerator->PullState = &PullState[0];
    } else {
        PICIMRESOURCE sub = GetCim(This);
        const _TCHAR *base = CimBaseName(ResourceURI);

        CHECKSETSC(sub, E_WSA_DESTINATION_UNREACHABLE,
                   "Get resource provider");

        sc = sub->v->Enumerate(sub, CurrentHeap(),local, base, Flags,
                               NULL, &Enumerator->sub, &Fault );
        CHECKSC("resource Enumerate");
    }

    FormatIid(EnumerationContext, &Enumerator->Context);

    if (Expires)
        *Expires = Enumerator->Expires;

    Mutex_Lock(&This->mx);
    ListAddLast(&This->Enumerators, enoff, Enumerator);
    Enumerator = NULL;
    Mutex_Unlock(&This->mx);

METHOD_CLEANUP;
    if (Enumerator)
        EnumeratorDelete(Enumerator, CurrentHeap());
    if (sc == E_ACCESS_DENIED) {
        WsManFault(This, Fault.FaultMessage);
        sc = S_NO_RESPONSE;
    }
        return sc;
}

PRIVATE SCODE wsenPull (
    PIWSENUMERATION iThis,
    /*inout*/_TCHAR * EnumerationContext,
    UINT EnumerationContextSize,
    UINT MaxElements,
    void *Buffer, UINT BufferSize,
    /*out*/ PItems * Items,
    /*out*/ BOOL * EndOfSequence)
{
    METHOD_DECLARE_MULTI(WsMan, wsen, Pull);
    struct IID guid;
    Enumerator *Enumerator = NULL;
    PIENUM sub = NULL;
    PIHEAP Heap;
    PITHREAD thd = CurrentThread();
    PCONTINFO cont = (PCONTINFO) thd->v->GetValue(thd,
                                                THREAD_VALUE_RPC_CONTINUATION);
    CHECKSETSC(cont, E_FAIL, "get current continuation");
    Heap = cont->TmpHeap;
    CHECKSETSC(Heap, E_FAIL, "get cont heap");

    CHECKSETSC(EnumerationContext, E_INVALID_PARAMETER,
               "NULL EnumerationContext");
    CHECKSETSC(EnumerationContextSize > 37 * sizeof(_TCHAR),
               E_BUFFER_TOO_SMALL,
               "make sure GUID fits into EnumerationContext");

    sc = ParseIid(EnumerationContext, &guid);
    CHECKSETSC(SUCCEEDED(sc), E_WSEN_INVALID_ENUMERATION_CONTEXT,
               "parse enumerator context");

    Mutex_Lock(&This->mx);

    Enumerator = ListWalk(&This->Enumerators, enoff, LIST_WALK_SEARCH,
                          EnumeratorMatch, &guid);

    if (Enumerator) {
        if (Enumerator->PullState) {
            struct _PULL_STATE *ps = Enumerator->PullState;
            *Items = ps->item;
            *EndOfSequence = ps->End;
            if (!ps->End)
                Enumerator->PullState = ps + 1;
        } else {
            sub = Enumerator->sub;
            sub->v->AddRef(sub);
        }
        // optionally change context
        //SetRandomIid(&Enumerator->Context);
        //FormatIid(EnumerationContext, &Enumerator->Context);
    }
    Mutex_Unlock(&This->mx);

    CHECKSETSC(Enumerator, E_WSEN_INVALID_ENUMERATION_CONTEXT,
               "find enumerator");
    if (sub) {
        PItems it;
        /* XXX later, make this a loop and pull many, until timeout or end */

        /* Get some space for result. XXX 100 chars for type */
        it = Heap->v->Alloc(Heap, 0, sizeof(*it) + 100 * sizeof(_TCHAR), 0);
        CHECKSETSC(it, E_NOT_ENOUGH_MEMORY, "alloc Item from tmp heap");

        it->ValueType = (_TCHAR *)(it + 1);

        sc = sub->v->Pull(sub, Heap, NULL,
                          &it->Value, (_TCHAR*) it->ValueType);
        CHECKSC("sub Pull");
        if (sc == S_FALSE)
            *EndOfSequence = TRUE;
        else
            *EndOfSequence = FALSE;

        ListAddLast(Items, itemsoff, it);

#if 0
        /* Stuff result into item and add it to result list */
        it = Heap->v->ReAlloc(Heap, 0, sizeof(*it) + (_tcslen((_TCHAR*) Buffer)
                                                    + 1) * sizeof(_TCHAR), 0);
        CHECKSETSC(it, E_NOT_ENOUGH_MEMORY, "alloc Item from tmp heap");
        _tcscpy(it->ValueType, (_TCHAR *) Buffer);
#endif
    }

METHOD_CLEANUP;
    RELEASE(sub); 
    if (Enumerator && (FAILED(sc) || *EndOfSequence)) {
        /* When things go bad or we reached the end, get rid of enumerator */
        Mutex_Lock(&This->mx);
        ListRemoveItem(&This->Enumerators, enoff, Enumerator);
        EnumeratorDelete(Enumerator, CurrentHeap());
        Mutex_Unlock(&This->mx);
    }
    return sc;
}

PRIVATE SCODE wsenRenew (
    PIWSENUMERATION iThis,
    /*inout*/ _TCHAR * EnumerationContext,
    UINT EnumerationContextSize,
    /*inout*/ TIME Expires)
{
    WsMan * This = pwsen(iThis);

    UnusedParameter(This);
    UnusedParameter(EnumerationContext);
    UnusedParameter(Expires);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wsenGetStatus (
    PIWSENUMERATION iThis,
    const _TCHAR * EnumerationContext,
    /*out*/ TIME Expires)
{
    WsMan * This = pwsen(iThis);

    DebugBreak();
    UnusedParameter(This);
    UnusedParameter(EnumerationContext);
    UnusedParameter(Expires);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wsenReleaseEnum (
    PIWSENUMERATION iThis,
    const _TCHAR * EnumerationContext)
{
    WsMan * This = pwsen(iThis);

    DebugBreak();
    UnusedParameter(This);
    UnusedParameter(EnumerationContext);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wsenEnumerationEnd (
    PIWSENUMERATION iThis,
    const _TCHAR * EnumerationContext,
    const _TCHAR * Code,
    const _TCHAR * Reason)
{
    WsMan * This = pwsen(iThis);

    DebugBreak();
    UnusedParameter(This);
    UnusedParameter(EnumerationContext);
    UnusedParameter(Code);
    UnusedParameter(Reason);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wseQueryInterface (
    PIWSEVENTING iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return wxfQueryInterface(iwxf(pwse(iThis)), iid, ppObject);
}

PRIVATE UINT wseAddRef (
    PIWSEVENTING iThis)
{
    return wxfAddRef(iwxf(pwse(iThis)));
}

PRIVATE UINT wseRelease (
    PIWSEVENTING iThis)
{
    return wxfRelease(iwxf(pwse(iThis)));
}

PRIVATE SCODE wseIdentifier (
    PIWSEVENTING iThis,
    PICONTINUATION cont,
    const _TCHAR * Value)
{
    WsMan * This = pwse(iThis);

    UnusedParameter(This);
    UnusedParameter(cont);
    UnusedParameter(Value);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wseSubscribe (
    PIWSEVENTING iThis,
    PEndpointReference_QUIET NotifyTo,
    PEndpointReference_QUIET EndTo,
    const _TCHAR *Filter,
    PDelivery DeliveryList,
    PConnectionRetry ConnectionRetry,
    /*out*/ PEndpointReference_QUIET SubscriptionManager,
    /*out*/ _TCHAR * EnumerationContext,
    UINT EnumerationContextSize,
    /*out*/ TIME *Expires)
{
    WsMan * This = pwse(iThis);

    UnusedParameter(This);
    UnusedParameter(EndTo);
    UnusedParameter(DeliveryList);
    UnusedParameter(SubscriptionManager);
    UnusedParameter(EnumerationContext);
    UnusedParameter(Expires);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wseRenew (
    PIWSEVENTING iThis,
    TIME Expires)
{
    WsMan * This = pwse(iThis);

    UnusedParameter(This);
    UnusedParameter(Expires);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wseUnsubscribe (
    PIWSEVENTING iThis,
    /*out*/ _TCHAR * Id,
    UINT IdSize,
    /*out*/ _TCHAR * Code,
    UINT CodeSize,
    /*out*/ PReason ReasonList)
{
    WsMan * This = pwse(iThis);

    UnusedParameter(This);
    UnusedParameter(Id);
    UnusedParameter(Code);
    UnusedParameter(ReasonList);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE wseGetStatus (
    PIWSEVENTING iThis)
{
    WsMan * This = pwse(iThis);

    UnusedParameter(This);

    return E_NOT_IMPLEMENTED;
}


#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <assert.h>
#include <soapif.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <sampleif.h>
#include <wsif.h>

#define ActOff offsetof(struct Action, ListNode)
#define AuthOff offsetof(struct Auth, ListNode)
#define SelOff offsetof(struct Selector, ListNode)

#define SEL static struct Selector

SEL sel62[2] = {
    {_T("CreationClassName"), NULL, _T("ComputerSystem"),
        {&sel62[1], &sel62[1]}},
    {_T("Name"), NULL, _T("IPMI Controller 33"),  {&sel62[0], &sel62[0]}}
};

SEL sel62_B[2] = {
    {_T("CreationClassName"), NULL, _T("ComputerSystem"),
        {&sel62[1], &sel62[1]}},
    {_T("Name"), NULL, _T("EWS test"),  {&sel62[0], &sel62[0]}}
};

SEL sel65[1] = {
    {_T("CreationClassName"), NULL, _T("NumericSensor"),
        {&sel65[0], &sel65[0]}},
};

SEL sel4 = {
    _T("InstanceID"),
    NULL,
    _T("IPMI:IPMI Controller 33 SEL Log"),
    {&sel4, &sel4}
};


SEL sel64[4] = {
    {_T("CreationClassName"), NULL, _T("NumericSensor"),
        {&sel64[1], &sel64[3]}},
    {_T("DeviceID"), NULL, _T("10.0.32"),
        {&sel64[2], &sel64[0]}},
    {_T("SystemCreationClassName"), NULL, _T("ComputerSystem"),
        {&sel64[3], &sel64[1]}},
    {_T("SystemName"), NULL, _T("IPMI Controller 32"), {&sel64[0], &sel64[2]}}
};

SEL sel91[4] = {
    {_T("CreationClassName"), NULL, _T("CIM_NumericSensor"),
        {&sel91[1], &sel91[3]}},
    {_T("DeviceID"), NULL, _T("81.0.32"),
        {&sel91[2], &sel91[0]}},
    {_T("SystemCreationClassName"), NULL, _T("ComputerSystem"),
        {&sel91[3], &sel91[1]}},
    {_T("SystemName"), NULL, _T("IPMI Controller 32"), {&sel91[0], &sel91[2]}}
};

SEL sels0[1] = {
    {_T("DeviceID"), NULL, _T("CPU2_Temp"),
        {&sels0[0], &sels0[0]}},
};

static struct PossibleStates2 pos[11] = {
    {_T("Non-Critical"),       {&pos[1], &pos[10]}},
    {_T("Lower Non-Critical"), {&pos[2], &pos[0]}},
    {_T("Upper Non-Critical"), {&pos[3], &pos[1]}},
    {_T("Critical"),           {&pos[4], &pos[2]}},
    {_T("Lower Critical"),     {&pos[5], &pos[3]}},
    {_T("Upper Critical"),     {&pos[6], &pos[4]}},
    {_T("Fatal"),              {&pos[7], &pos[5]}},
    {_T("Lower Fatal"),        {&pos[8], &pos[6]}},
    {_T("Upper Fatal"),        {&pos[9], &pos[7]}},
    {_T("Normal"),             {&pos[10], &pos[8]}},
    {_T("Unknown"),            {&pos[0], &pos[9]}}
};
struct CIM_NumericSensor demosen0 = {
    _T("[Changed Caption]"), 
    _T("[Changed Description]"), 
    _T("CPU2_Temp"), 
    
    Int64Initializer(0,0), 
    _T("[Changed Name]"), 
    _T("OK"), 
    _T("OK"), 
    
    _T("Enabled"), 
    _T("[Changed OtherEnabledState]"), 
    _T("Not Applicable"), 
    _T("[Changed EnabledDefault]"), 
    Int64Initializer(0,0), 
    
    _T("demo_sensor"), 
    _T("CPU2_Temp"), 
    _T("[Changed OtherIdentifyingInfo]"), 
    _T("[Changed IdentifyingDescriptions]"), 
    
    22, 
    _T("[Changed OtherSensorTypeDescription]"), 
    pos, 
    _T("Normal"), 
    
    4444, 
    6666, 
    42, 
    _T("[Changed SettableThreasholds]"), 
    _T("[Changed SupportedThresholds]"), 
    _T("Tech Ed Demo System"), 
    _T("UnaryComputerSystem")
};

PRIVATE SCODE progMain (
    PIPROGRAM iThis,
    const _TCHAR * cmdline)
{
    METHOD_DECLARE(WsMan, Main);
    PIWSTRANSFER wxf = NULL;
    PIWSENUMERATION wsen = NULL;
    const _TCHAR *url = _T("http://127.0.0.1/wsman");
    _TCHAR *CmdLine2 = NULL, *User = NULL, *Pass = NULL, *Tests = NULL;
    _TCHAR typeis[100], context[100], MessageId[50];
    PISOAPFACTORY soapfactory = NULL;
    TIME OperationTimeout = TIME_ORIGIN;
    PBYTE Buffer;
    PItems Items = NULL;
    BOOL end = FALSE;
    PICIM_RecordLog cims = NULL;
    PIWSMANID wsmid = NULL;
    struct ClearLog_OUTPUT clearout = {0,};
    struct ClearLog_INPUT clearin = {NULL};
#ifdef _WITH_PRIVATE_INCLUDES
    PIWSSHELL wssh;
#endif

    EndpointReference ep = {0,};
    EndpointReference ep2 = {0,};

    WSCONTEXT ctx = {0,};
    UINT32 fieldsel[10] = {~0, ~0, ~0, ~0, };

    struct CIM_NumericSensor nums = {0,};
    struct Selector *sel62xml = NULL;
    struct CIM_NumericSensor *sen9xml = NULL;
    QID qid = 0;

#define BUFFERSIZE 8192
    sc = S_OK;
    Buffer = malloc(BUFFERSIZE);
    if (!Buffer)
        return E_NOT_ENOUGH_MEMORY;

    if (_tcslen(cmdline) > 0 && !_tcsncmp(cmdline, _T("-m"), 2)) {
        doMeasure = !doMeasure;
        //cmdline += 2;
        return S_OK;
    }

    if (_tcslen(cmdline) > 0) {
        if (_tcspbrk(cmdline, _T(" \t"))) {
            /* Parsing required: wsmantest.cob [[-b user pass] url] */
            CmdLine2 = _tcsdup(cmdline);
            User = _tcssplit(CmdLine2, _T(" \t"));
            CHECKSETSC(!_tcscmp(CmdLine2, _T("-b")), E_INVALID_PARAMETER,
                       "recognize commandline arg");
            Pass = _tcssplit(User, _T(" \t"));
            CHECKSETSC(Pass, E_INVALID_PARAMETER, "parse commandline Pass");
            url = _tcssplit(Pass, _T(" \t"));
            CHECKSETSC(url, E_INVALID_PARAMETER, "parse commandline url");
            Tests = _tcssplit((_TCHAR*) url, _T(" \t"));
        } else {
            url = cmdline;
        }
    }

    _tcscpy(MessageId, _T("uuid:E9E02CCC-5EB2-4A46-0000-000000000000"));
    {
        TIME t = CurrentTime();
        INT tt = Int64ToInt32(t);

        _stprintf(MessageId + 29, _T("%012u"), tt);
    }
    ep.Address = (_TCHAR *) url;
    ep2.Address = _T("http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous");

    ctx.To = &ep;
    ctx.ReplyTo = &ep2;
    ctx.MaxEnvelopeSizeValue = 51200;
    Int32ToInt64(OperationTimeout, TIME_RELATIVE(TIME_SECONDS(60)));
    ctx.OperationTimeout = OperationTimeout;
    ctx.MessageId = MessageId;
    _tcscpy(context, _T(""));

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    CHECKSC("get soap factory");

#if 0
    sc = soapfactory->v->ProcessFragment(soapfactory,
                                      _T("<cimns:CIM_NumericSensor xmlns:cimns=\"http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/CIM_NumericSensor\"><cimns:SensorType>0</cimns:SensorType><cimns:BaseUnits>0</cimns:BaseUnits><cimns:UnitModifier>0</cimns:UnitModifier><cimns:RateUnits>0</cimns:RateUnits><cimns:CurrentReading>0</cimns:CurrentReading><cimns:NominalReading>0</cimns:NominalReading><cimns:NormalMax>0</cimns:NormalMax><cimns:NormalMin>0</cimns:NormalMin><cimns:MaxReadable>0</cimns:MaxReadable><cimns:MinReadable>0</cimns:MinReadable><cimns:ResoluAtion>0</cimns:Resolution><cimns:Accuracy>0</cimns:Accuracy><cimns:IsLinear>false</cimns:IsLinear><cimns:Hysteresis>0</cimns:Hysteresis><cimns:LowerThresholdNonCritical>100</cimns:LowerThresholdNonCritical><cimns:UpperThresholdNonCritical>0</cimns:UpperThresholdNonCritical><cimns:LowerThresholdCritical>0</cimns:LowerThresholdCritical><cimns:UpperThresholdCritical>0</cimns:UpperThresholdCritical><cimns:LowerThresholdFatal>0</cimns:LowerThresholdFatal><cimns:UpperThresholdFatal>0</cimns:UpperThresholdFatal></cimns:CIM_NumericSensor>"),
                                      CurrentHeap(),
                                      &sen9xml, &qid);
    CHECKSC("Parse CIM_NumericSensor");
    free(sen9xml);
#endif
#if 0
    sc = soapfactory->v->ProcessData2(soapfactory, 
                                      _T("<wsman:SelectorSet xmlns:wsman=\"http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd\"><wsman:Selector xmlns:wsman=\"http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd\" Name=\"InstanceID\">IPMI:IPMI Controller 33 SEL Log</wsman:Selector></wsman:SelectorSet>"),
                                      CurrentHeap(),
                                      &sel62xml, &qid);
#endif

    if (!Tests || Tests[0] != _T('!')) {
        sc = soapfactory->v->CreateProxy3(soapfactory, NULL,
                                          &ctx, _T("wxf"),
                                          0,
                                          (PIUNKNOWN*) &wxf);
        CHECKSC("CreateProxy3 wxf");
        printf("wstest Created wxf proxy: wxf=%x iunk=%x sc=%x\n",
               (UINT) wxf, (UINT) &IID_IUnknown, sc);
        if (User) {
            PIHTTPCONNECTOR conn;

            sc = wxf->v->QueryInterface(wxf, &IID_IHttpConnector,
                                        (void**) &conn);
            CHECKSC("get connector from wxf proxy");

            sc = conn->v->SetBasicKey(conn, User, Pass);
            CHECKSC("set basic key for wxf proxy");

            conn->v->Release(conn);
        }
    }

    if (Tests && !_tcscmp(Tests, _T("config"))) {
        printf("test wxf:Get wirm/config\n");
        _tcsncpy(MessageId + 24, _T("config"), 6);
        ctx.ResourceURI = WINRM_NS _T("/config"),
        sc = wxf->v->Get(wxf,url,
                         WINRM_NS _T("/config"), NULL, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        CHECKSC("wxf:Get");
        printf("wxf:Get winrm/config typeis=%s buf=%x\n", typeis, (ADDRESS) Buffer);
        /* STATUS: OK */
    }

    if (Tests && !_tcscmp(Tests, _T("create"))) {
        EndpointReference epr = {0,};
        _tcsncpy(MessageId + 24, _T("create"), 6);
        printf("test wxf:Create foo/test\n");
        epr.Address = context;   /* XXX storage for string */
        ctx.ResourceURI = _T("foo/test"),
        sc = wxf->v->Create(wxf, url,
                            _T("foo/test"),
                            &sen1, _T("p:NumericSensor"),
                            (PEndpointReference_QUIET) &epr);
        CHECKSC("wxf:Create");
        printf("wxf:Create typeis=%s epr=%x\n", typeis, (ADDRESS) &epr);
        /* STATUS: OK */
    }

    if (!Tests || !_tcscmp(Tests, _T("6.2"))) {
        ctx.ResourceURI = IPMI_NS _T("/CIM_ComputerSystem");
        _tprintf(_T("scenario 6.2 Get %s"), ctx.ResourceURI);
        _tcsncpy(MessageId + 24, _T("602"), 3);
        ctx.SelectorSet = (PSelector_wsa) sel62;
        sc = wxf->v->Get(wxf,url,
                         IPMI_NS _T("/CIM_ComputerSystem"),
                         sel62, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        CHECKSC("wxf:Get");
        printf("wxf:Get typeis=%s buf=%x\n", typeis, (ADDRESS) Buffer);
        ctx.SelectorSet = NULL;
        /* STATUS: OK */
    }

    if (!Tests || !_tcscmp(Tests, _T("6.3"))) {
        _tcsncpy(MessageId + 24, _T("603"), 3);
        printf("scenario 6.3 Get %s (INVALID)\n", ctx.ResourceURI);
        ctx.ResourceURI = IPMI_NS _T("/cim_computersyste");
        sc = wxf->v->Get(wxf,url,
                         IPMI_NS _T("/cim_computersyste"),
                         sel62, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        printf("wxf:Get sc=%x (SHOULD FAIL, invalid ResourceURI)\n", sc);
        /* STATUS: good, except SoapFault formatted wrong */
    }

    if (!Tests || !_tcscmp(Tests, _T("6.4"))) {
        _tcsncpy(MessageId + 24, _T("604"), 3);
        printf("scenario 6.4 Get http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/CIM_NumericSensor\n");
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        ctx.SelectorSet = (PSelector_wsa) sel64;
        ctx.MaxEnvelopeSizeValue = 7;
        sc = wxf->v->Get(wxf,url,
                         CIM_NS _T("/CIM_NumericSensor"),
                         sel64, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        printf("wxf:Get sc=%x (SHOULD FAIL, response too big)\n", sc);
        ctx.MaxEnvelopeSizeValue = 51200;
        /* STATUS: XXX does not fail */
    }

    if (!Tests || !_tcscmp(Tests, _T("6.5"))) {
        _tcsncpy(MessageId + 24, _T("605"), 3);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        printf("scenario 6.5 Get %s (MISSING SELECTOR)\n", ctx.ResourceURI);
        ctx.SelectorSet = (PSelector_wsa) sel65;
        sc = wxf->v->Get(wxf,url,
                         CIM_NS _T("/CIM_NumericSensor"),
                         sel65, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        printf("wxf:Get sc=%x (SHOULD FAIL, missing selector)\n", sc);
        /* STATUS: XXX untested */
    }

    if (!Tests || !_tcscmp(Tests, _T("6.6"))) {
        _tcsncpy(MessageId + 24, _T("606"), 3);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        printf("scenario 6.6 Get %s (OP TIMEOUT)\n", ctx.ResourceURI);
        Int32ToInt64(OperationTimeout, TIME_RELATIVE(TIME_MICROS(1)));
        ctx.OperationTimeout = OperationTimeout;
        ctx.SelectorSet = (PSelector_wsa) sel62;
        sc = wxf->v->Get(wxf,url,
                         CIM_NS _T("/CIM_NumericSensor"),
                         sel62, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        printf("wxf:Get sc=%x (SHOULD FAIL, op timeout)\n", sc);
        /* STATUS: XXX untested */
        Int32ToInt64(OperationTimeout, TIME_RELATIVE(TIME_SECONDS(60)));
        ctx.OperationTimeout = OperationTimeout;
    }

    if (!Tests || !_tcscmp(Tests, _T("9.1"))) {
        goto s91; s91:
            _tcsncpy(MessageId + 24, _T("901"), 3);
            printf("scenario 9.1: Put NumericSensor\n");
            //nums.LowerThresholdNonCritical = 100;
            nums.CurrentReading = 100;
            ctx.SelectorSet = (PSelector_wsa) sel91;
            memset(typeis, 0, sizeof(typeis));
            memset(fieldsel, 0, sizeof(fieldsel));
            fieldsel[22/32] |= (1 << 22%32);
            ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
            sc = wxf->v->Put(wxf,url,
                             CIM_NS _T("/CIM_NumericSensor"), sel91,
                             &nums, _T("p:NumericSensor"), fieldsel,
                             (void*) Buffer, typeis, BUFFERSIZE);
            CHECKSC("wxf:Put");
            printf("wxf:Put sc=%x typeis=%s\n", sc, typeis);
            /* STATUS: sets too many ints to 0. reply OK */
    }

    if (!Tests || !_tcscmp(Tests, _T("p0"))) {
            _tcsncpy(MessageId + 24, _T("p0"), 3);
            printf("Put CIM_NumericSensor\n");
            ctx.SelectorSet = (PSelector_wsa) sels0;
            memset(typeis, 0, sizeof(typeis));
            ctx.ResourceURI = _T("wsman:demo_sensors");
            sc = wxf->v->Put(wxf,url,
                             _T("wsman:demo_sensors"), sels0,
                             &demosen0, _T("cimns:CIM_NumericSensor"), NULL,
                             (void*) Buffer, typeis, BUFFERSIZE);
            CHECKSC("wxf:Put");
            printf("wxf:Put sc=%x typeis=%s\n", sc, typeis);
    }

    if (!Tests || !_tcscmp(Tests, _T("p0"))) {
            _tcsncpy(MessageId + 24, _T("p0"), 3);
            printf("Put CIM_NumericSensor\n");
            ctx.SelectorSet = (PSelector_wsa) sels0;
            memset(typeis, 0, sizeof(typeis));
            ctx.ResourceURI = _T("wsman:demo_sensors");
            sc = wxf->v->Put(wxf,url,
                             _T("wsman:demo_sensors"), sels0,
                             &demosen0, _T("cimns:CIM_NumericSensor"), NULL,
                             (void*) Buffer, typeis, BUFFERSIZE);
            CHECKSC("wxf:Put");
            printf("wxf:Put sc=%x typeis=%s\n", sc, typeis);
    }

    if (!Tests || !_tcscmp(Tests, _T("pns"))) {
            _tcsncpy(MessageId + 24, _T("pns"), 3);
            printf("Put CIM_NumericSensor\n");
            ctx.SelectorSet = (PSelector_wsa) sels0;
            memset(typeis, 0, sizeof(typeis));
            ctx.ResourceURI = _T("CIM_NumericSensor");
            sc = wxf->v->Put(wxf,url,
                             _T("CIM_NumericSensor"), sels0,
                             &demosen0, _T("cimns:CIM_NumericSensor"), NULL,
                             (void*) Buffer, typeis, BUFFERSIZE);
            CHECKSC("wxf:Put");
            printf("wxf:Put sc=%x typeis=%s\n", sc, typeis);
    }

    if (!Tests || !_tcscmp(Tests, _T("p1"))) {
            _tcsncpy(MessageId + 24, _T("p1"), 3);
            printf("Put CIM_NumericSensor, some fields\n");
            ctx.SelectorSet = (PSelector_wsa) sels0;
            memset(typeis, 0, sizeof(typeis));
            memset(fieldsel, 0, sizeof(fieldsel));
            fieldsel[22/32] |= (1 << 22%32);
            fieldsel[13/32] |= (1 << 13%32);
            fieldsel[2/32] |= (1 << 2%32);
            ctx.ResourceURI = _T("wsman:demo_sensors");
            sc = wxf->v->Put(wxf,url,
                             _T("wsman:demo_sensors"), sels0,
                            &demosen0, _T("cimns:CIM_NumericSensor"), fieldsel,
                             (void*) Buffer, typeis, BUFFERSIZE);
            CHECKSC("wxf:Put");
            printf("wxf:Put sc=%x typeis=%s\n", sc, typeis);
    }

    if (!Tests || !_tcscmp(Tests, _T("g0"))) {
        ctx.ResourceURI = _T("wsman:demo_sensors");
        _tprintf(_T("Get wsman:demo_sensors DeviceID='CPU3_Temp'"),
                 ctx.ResourceURI);
        _tcsncpy(MessageId + 24, _T("g0"), 3);
        ctx.SelectorSet = (PSelector_wsa) sels0;
        sc = wxf->v->Get(wxf,url,
                          _T("wsman:demo_sensors"),
                         sels0, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        CHECKSC("wxf:Get");
        printf("wxf:Get typeis=%s buf=%x\n", typeis, (ADDRESS) Buffer);
        ctx.SelectorSet = NULL;
        /* STATUS: OK */
    }

    if (!Tests || !_tcscmp(Tests, _T("g1"))) {
        ctx.ResourceURI = _T("CIM_NumericSensor");
        _tprintf(_T("Get CIM_NumericSensor DeviceID='CPU3_Temp'"),
                 ctx.ResourceURI);
        _tcsncpy(MessageId + 24, _T("g0"), 3);
        ctx.SelectorSet = (PSelector_wsa) sels0;
        sc = wxf->v->Get(wxf,url,
                          _T("CIM_NumericSensor"),
                         sels0, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        CHECKSC("wxf:Get");
        printf("wxf:Get typeis=%s buf=%x\n", typeis, (ADDRESS) Buffer);
        ctx.SelectorSet = NULL;
        /* STATUS: OK */
    }

    /* Test BMP get */
    if (Tests && Tests[0] == _T('@')) {
        PBitmapImage img;
        printf("Sensor test scenario: get image\n");
        ctx.ResourceURI = Tests;
        sc = wxf->v->Get(wxf,url,
                         Tests, NULL, OperationTimeout,
                         (void*) Buffer, typeis, BUFFERSIZE);
        CHECKSC("wxf:Get image");

        CHECKSETSC(!_tcscmp(typeis, _T("sensor:BitmapImage")),
                   E_INVALID_PARAMETER, "unexpected element type");
        img = (void *) Buffer;
        assert(!_tcscmp(img->Type, _T("bmp")));
        ShowBmp(img->Content);
    }

    RELEASE_NULL(wxf);

/* Separate proxy */
    if (!Tests || !_tcscmp(Tests, _T("6.1"))) {
        sc = soapfactory->v->CreateProxy3(soapfactory, NULL,
                                          &ctx, _T("wsmid"),
                                          0,
                                          (PIUNKNOWN*) &wsmid);
        CHECKSC("CreateProxy3 wsmid");
        printf("wstest Created wsmid proxy: wsmid=%x iunk=%x sc=%x\n",
               (UINT) wsmid, (UINT) &IID_IUnknown, sc);

        if (User) {
            PIHTTPCONNECTOR conn;

            sc = wsmid->v->QueryInterface(wsmid, &IID_IHttpConnector,
                                          (void**)&conn);
            CHECKSC("get connector from wsmid proxy");

            sc = conn->v->SetBasicKey(conn, User, Pass);
            CHECKSC("set basic key for wsmid proxy");

            conn->v->Release(conn);
        }
        _tcsncpy(MessageId + 24, _T("601"), 3);
        printf("scenario 6.1 Identify (http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/cim_recordlog)\n");
        ctx.SelectorSet = NULL;
        ctx.ResourceURI = NULL;
        memset(Buffer, 0, BUFFERSIZE);
        sc = wsmid->v->Identify(wsmid,
                                /* Use quarters of the big buf */
                                (_TCHAR*) Buffer, BUFFERSIZE/4,
                                (_TCHAR*) (Buffer + BUFFERSIZE/4), BUFFERSIZE/4,
                                (_TCHAR*)(Buffer + BUFFERSIZE/2), BUFFERSIZE/4,
                                (_TCHAR*)(Buffer + BUFFERSIZE/2 + BUFFERSIZE/4), BUFFERSIZE/8,
                                (_TCHAR*)(Buffer + BUFFERSIZE/2 + BUFFERSIZE/4 + BUFFERSIZE/8), BUFFERSIZE/8);
        CHECKSC("wsmid->Identify");

        printf("wsmid->Identify sc=%x protocol=%s vendor=%s product=%s\n",
               sc, Buffer, Buffer + BUFFERSIZE/4, Buffer + BUFFERSIZE/2);
        /* STATUS: OK */
        RELEASE_NULL(wsmid);
    }

/* Separate proxy */
    if (!Tests || !_tcscmp(Tests, _T("8.1"))) {
        goto s81; s81:

        sc = soapfactory->v->CreateProxy3(soapfactory, NULL,
                                          &ctx, _T("CIMLogger"),
                                          0,
                                          (PIUNKNOWN*) &cims);
        CHECKSC("CreateProxy3 cims");
        printf("wstest Created cims proxy: cims=%x iunk=%x sc=%x\n",
               (UINT) cims, (UINT) &IID_IUnknown, sc);

        if (User) {
            PIHTTPCONNECTOR conn;

            sc = cims->v->QueryInterface(cims, &IID_IHttpConnector,
                                         (void**)&conn);
            CHECKSC("get connector from cims proxy");

            sc = conn->v->SetBasicKey(conn, User, Pass);
            CHECKSC("set basic key for cims proxy");

            conn->v->Release(conn);
        }
        _tcsncpy(MessageId + 24, _T("801"), 3);
        ctx.ResourceURI = IPMI_NS _T("/RecordLog");
        printf("scenario 8.1 ClearLog (%s)\n", ctx.ResourceURI);
        ctx.SelectorSet = (PSelector_wsa) &sel4;
        sc = cims->v->ClearLog(cims,
                               &clearin,
                               &clearout);
        printf("cims->clearlog sc=%x rv=%d\n", sc, clearout.ReturnValue);
        CHECKSC("cims:clearlog");
        ctx.SelectorSet = NULL;
        /* STATUS: OK */
        RELEASE_NULL(cims);
    }

/* Separate proxy */
    if (!Tests || Tests[0] == _T('7') || Tests[0] == _T('e')) {
        sc = soapfactory->v->CreateProxy3(soapfactory, NULL,
                                          &ctx, _T("wsen"),
                                          0,
                                          (PIUNKNOWN*) &wsen);
        CHECKSC("CreateProxy3 wsen");
        printf("wstest Created wsen proxy: wsen=%x iunk=%x sc=%x\n",
               (UINT) wsen, (UINT) &IID_IUnknown, sc);
        if (User) {
            PIHTTPCONNECTOR conn;

            sc = wsen->v->QueryInterface(wsen, &IID_IHttpConnector,
                                         (void**)&conn);
            CHECKSC("get connector from wsen proxy");

            sc = conn->v->SetBasicKey(conn, User, Pass);
            CHECKSC("set basic key for wsen proxy");

            conn->v->Release(conn);
        }
    }

    if (!Tests || !_tcscmp(Tests, _T("7.1"))) {
        TIME expires;
        _tcsncpy(MessageId + 24, _T("701A"), 4);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        printf("scenario 7.1.a Enumerate %s\n", ctx.ResourceURI);
        _tcscpy(context, _T(""));
        sc = wsen->v->Enumerate(wsen, url,
                                CIM_NS _T("/CIM_NumericSensor"),
                                NULL, NULL, &expires,
                                context, sizeof(context));
        printf("wsen->Enumerate enum context=%s\n", context);
        CHECKSC("wsen->Enumerate");
        /* STATUS: OK */

        printf("scenario 7.1.b Pull\n");
        _tcsncpy(MessageId + 24, _T("701B"), 4);
        memset(Buffer, 0, BUFFERSIZE);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        sc = wsen->v->Pull(wsen, 
                           context, sizeof(context),
                           20,
                           Buffer, BUFFERSIZE,
                           &Items,
                           &end);
        CHECKSC("wsen->Pull");
        printf("wsen->Pull enum context=%s end=%d Items0=%x type0=%s\n",
               context, end, Items, Items ? (char*)Items->ValueType : "???");
        assert(!Items || !_tcscmp(Items->ValueType, _T("cimns:CIM_NumericSensor"))); 
        if (Items && Items->ValueType)
            printf("type=%s\n", Items->ValueType);
        /* STATUS: OK */

        if (!end) {
            printf("scenario 7.1.c Pull\n");
            _tcsncpy(MessageId + 24, _T("701C"), 4); 
           memset(Buffer, 0, BUFFERSIZE);
           ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
            sc = wsen->v->Pull(wsen, 
                               context, sizeof(context),
                               20,
                               Buffer, BUFFERSIZE,
                               &Items,
                               &end);
            printf("wsen->Pull enum context=%s end=%d\n", context, end);
            CHECKSC("wsen->Pull");
            /* STATUS: OK */
        }

        if (!end) {
            printf("scenario 7.1.d Release\n");
            _tcsncpy(MessageId + 24, _T("701D"), 4); 
            ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
            sc = wsen->v->ReleaseEnum(wsen, context);
        }
    }

    if (!Tests || !_tcscmp(Tests, _T("7.2"))) {
        TIME expires;
        _tcsncpy(MessageId + 24, _T("702A"), 4);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        printf("scenario 7.2.a Enumerate OPTIMIZED %s\n", ctx.ResourceURI);
        _tcscpy(context, _T(""));
        sc = wsen->v->Enumerate(wsen,url,
                                CIM_NS _T("/CIM_NumericSensor"),
                                NULL, NULL, &expires,
                                context, sizeof(context));
        printf("wsen->Enumerate enum context=%s\n", context);
        /* STATUS: OK */

        printf("scenario 7.1.b Pull\n");
        _tcsncpy(MessageId + 24, _T("701B"), 4);
        memset(Buffer, 0, BUFFERSIZE);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        sc = wsen->v->Pull(wsen, 
                           context, sizeof(context),
                           20,
                           Buffer, BUFFERSIZE,
                           &Items,
                           &end);
        printf("wsen->Pull enum context=%s end=%d Items0=%x type0=%s\n",
               context, end, Items, Items ? (char*)Items->ValueType : "???");
        assert(!Items || !_tcscmp(Items->ValueType, _T("cimns:CIM_NumericSensor")));
        CHECKSC("wsen->Pull"); 
        if (Items && Items->ValueType)
            printf("type=%s\n", Items->ValueType);
        /* STATUS: OK */

        if (!end) {
            printf("scenario 7.1.c Pull\n");
            _tcsncpy(MessageId + 24, _T("701C"), 4); 
           memset(Buffer, 0, BUFFERSIZE);
           ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
            sc = wsen->v->Pull(wsen, 
                               context, sizeof(context),
                               20,
                               Buffer, BUFFERSIZE,
                               &Items,
                               &end);
            printf("wsen->Pull enum context=%s end=%d\n", context, end);
            CHECKSC("wsen->Pull");
            /* STATUS: OK */
        }

        if (!end) {
            printf("scenario 7.1.d Release\n");
            _tcsncpy(MessageId + 24, _T("701D"), 4); 
            ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
            sc = wsen->v->ReleaseEnum(wsen, context);
        }
    }

    if (!Tests || !_tcscmp(Tests, _T("7.3"))) {
        TIME expires;
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        printf("scenario 7.3.a Enumerate %s\n", ctx.ResourceURI);
        _tcsncpy(MessageId + 24, _T("703A"), 4);
        _tcscpy(context, _T(""));
        sc = wsen->v->Enumerate(wsen, url,
                                CIM_NS _T("/CIM_NumericSensor"),
                                NULL, NULL, &expires,
                                context, sizeof(context));
        printf("wsen->Enumerate enum context=%s\n", context);
        /* STATUS: OK */

        printf("scenario 7.3.b Pull\n");
        _tcsncpy(MessageId + 24, _T("703B"), 4);
        memset(Buffer, 0, BUFFERSIZE);
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        sc = wsen->v->Pull(wsen, 
                           context, sizeof(context),
                           1,
                           Buffer, BUFFERSIZE,
                           &Items,
                           &end);
        printf("wsen->Pull enum context=%s end=%d\n", context, end);
        CHECKSC("wsen->Pull");
        /* STATUS: OK */

        printf("scenario 7.3.c Pull - bad context\n");
        _tcsncpy(MessageId + 24, _T("703C"), 4);
        memset(Buffer, 0, BUFFERSIZE);
        _tcscpy(context, _T("bad"));
        ctx.ResourceURI = CIM_NS _T("/CIM_NumericSensor");
        sc = wsen->v->Pull(wsen, 
                           context, sizeof(context),
                           1,
                           Buffer, BUFFERSIZE,
                           &Items,
                           &end);
        printf("wsen->Pull enum sc=%x context=%s end=%d (SHOULD FAIL)\n",
               sc, context, end);
        /* STATUS: fault malformed */
        sc = S_OK;
    }

    if (!Tests || !_tcscmp(Tests, _T("ed"))) {
        UINT i;
        TIME expires;
        _tcscpy(MessageId + 24, _T("ed0"));
        ctx.ResourceURI = _T("wsman:demo_sensors");
        printf("Enumerate %s\n", ctx.ResourceURI);
        _tcscpy(context, _T(""));
        sc = wsen->v->Enumerate(wsen, url,
                                _T("wsman:demo_sensors"),
                                NULL, NULL, &expires,
                                context, sizeof(context));
        printf("wsen->Enumerate enum context=%s\n", context);
        CHECKSC("wsen->Enumerate");
        /* STATUS: OK */

        end = FALSE;
        for (i = 1; !end; i++) {
            printf("Pull\n");
            _stprintf(MessageId + 24, _T("ed%d"), i);
            memset(Buffer, 0, BUFFERSIZE);
            ctx.ResourceURI = _T("wsman:demo_sensors");

            sc = wsen->v->Pull(wsen, 
                               context, sizeof(context),
                               20,
                               Buffer, BUFFERSIZE,
                               &Items,
                               &end);
            CHECKSC("wsen->Pull");
            printf("wsen->Pull enum context=%s end=%d Items0=%x type0=%s\n",
                   context, end, Items,
                   Items ? (char*)Items->ValueType : "???");
            assert(!Items || !_tcscmp(Items->ValueType,
                                      _T("cimns:CIM_NumericSensor"))); 
            if (Items && Items->ValueType)
                printf("type=%s\n", Items->ValueType);
            /* STATUS: OK */
        }
    }

/* WSSH test: Separate proxy */
#ifdef _WITH_PRIVATE_INCLUDES
    if (Tests && Tests[0] == _T('!')) {
        struct Stream *pstream = NULL;
        UINT val = 0;
        UINT32 DidComplete = 0;
        Tests++;

        sc = soapfactory->v->CreateProxy3(soapfactory, NULL,
                                          &ctx, _T("wssh"),
                                          0,
                                          (PIUNKNOWN*) &wssh);
        CHECKSC("CreateProxy3 wssh");
        printf("wstest Created wssh proxy: pxy=%x iunk=%x sc=%x\n",
               (UINT) wssh, (UINT) &IID_IUnknown, sc);

        if (User) {
            PIHTTPCONNECTOR conn;

            sc = wssh->v->QueryInterface(wssh, &IID_IHttpConnector,
                                         (void**)&conn);
            CHECKSC("get connector from cims proxy");

            sc = conn->v->SetBasicKey(conn, User, Pass);
            CHECKSC("set basic key for wssh proxy");

            conn->v->Release(conn);
        }

        printf("wssh test\n");
        ctx.SelectorSet = NULL;
        sc = wssh->v->OpenShell(wssh, _T("(wssh)"),
                                _T("/"),
                                NULL,
                                context);
        printf("wssh->OpenShell sc=%x context=%s\n", sc, context);
        CHECKSC("wssh:OpenShell");

        sc = wssh->v->Command(wssh, _T("(wssh)"), context,
                              Buffer, BUFFERSIZE,
                              Tests, &pstream, &val, &DidComplete);
        printf("wssh->Command sc=%x stream=%x\n", sc, pstream);
        if (pstream && pstream->Content) {
            PIBUFFER im = pstream->Content;
            _TCHAR *Value;
            UINT ValueLength;
            sc = im->v->GetDataPointer(im, 0, (BYTE**) &Value, &ValueLength);
            DBG(("Got %d bytes of %s buffer >>>%.*s<<<\n",
                 ValueLength, pstream->Name, ValueLength, Value));

            /* XXX walk list */
            RELEASE(pstream->Content);
        }
        CHECKSC("wssh:Command");

        RELEASE_NULL(wssh);
    }
#endif
    printf("Tests completed OK\n");

METHOD_CLEANUP;
    RELEASE(wxf);
    RELEASE(wsen);
    RELEASE(cims);
    RELEASE(wsmid);
    RELEASE(soapfactory);
    if (CmdLine2)
        free(CmdLine2);
    if (Buffer)
        free(Buffer);

    printf("done\n");

    return sc;
}

#if 0
PIUNKNOWN WsManCobMain(void)
{
    PRIVATE struct _WsMan wsmanCob = {&wxfWsManVtbl.v, 1,
                                      &wsmidWsManVtbl.v,
                                      &wsenWsManVtbl.v,
                                      &wseWsManVtbl.v,
                                      &cimsWsManVtbl.v,
#ifdef _WITH_PRIVATE_INCLUDES
                                      &wsshWsManVtbl.v,
#endif
                                      &progWsManVtbl.v,
                                      MUTEX_INITIALIZER, 0, NULL, NULL};
    return (PIUNKNOWN) &wsmanCob;
}
#endif
