/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <import_windows.h>
#include <stdio.h>
#include <mmlite.h>
#include <base/cons.h>
#include <base/loader.h>
#include <tchar.h>
#include <../base/fred.h>

PINAMESPACE InitNtFile();
SCODE InitDebugConsole(void);
void InitFrameBuffer(void);
const char *BootCmdLine(void);
void InitTime(void);

void RunFirstApp(const _TCHAR *Args);

static char *_cmdline = "";
//extern const _TCHAR *FIRST_IMAGE_NAME;
extern const char *FIRST_IMAGE_NAME;/* XXXXXXXXXXXXXXXXXXXXXXX */

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    SCODE sc;
    PINAMESPACE pRoot;
    const _TCHAR *Args;

    UnusedParameter(arg);

    InitTime();

    pRoot = InitNtFile(NULL);

    if (pRoot == NULL) {
            printf("FirstThread: Init File System failed\n");
            goto Out;
    }

    SetImageNameSpace(pRoot);
    pRoot->v->Release(pRoot);   /* ImageNs holds its own ref */

    sc = InitDebugConsole();
    if (FAILED(sc)) {
        printf("Failed to create debug console (sc = x%x)\n", sc);
        goto Out;
    }

#ifdef _UNICODE
    Args = _strtow(BootCmdLine());
    if (!Args) {
        printf("Could not convert args to unicode (malloc)\n");
        goto Out;
    }
#else
    Args = BootCmdLine();
#endif

    FirstApp(Args);

#ifdef _UNICODE
    free((void *) Args);
#endif

 Out:
    BaseDelete();
}

const char *BootCmdLine(void)
{
    return _cmdline;
}

int main(int argc, char *argv[])
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    if (argc > 1)
        FIRST_IMAGE_NAME = argv[1];

    if (argc > 2)
        _cmdline = argv[2];

    if (!argv) {
        char *z;
        z = GetCommandLineA();
        z++; /* skip quote */
        z = _strsplit(z, "\"");
        if (z && *z) {
            z++;         /* skip space */
            if (*z) {    /* guard against trailing space from VS7 */
                FIRST_IMAGE_NAME = z;
                _cmdline = _strsplit(z, " ");
            }
        }
    }

    BaseInit();
    return 1;
}

/* Load state from pCxt (aka longjmp).  Change processor modes if necessary */
__declspec(naked) void HWLoadThreadContext(PCXTINFO pCxt)
{
    __asm {
	mov	esp, [esp+4];	/* fetch argument pCxt */

	pop	ecx;		/* throw away pFpa */
	popad;
	pop	fs;
	pop	gs;
	pop	es;
	pop	ds;
	iretd;
    }
}

PRIVATE HANDLE StdOutHandle;

int main(int argc, char *argv[]);

void mainCRTStartup()
{
    int i;
    StdOutHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    i = main(0, NULL);
}

/* Guard CRT library from "interrupts" */
INT __cdecl NT_printf(const char *format, ...)
{
    va_list arg;
    INT ret;
    static int (*vpr)() = NULL;
    HANDLE h;

    if (vpr == NULL) {
        h = GetModuleHandle(_T("msvcrt.dll"));
        if (h == INVALID_HANDLE_VALUE)
            DebugBreak();
        vpr = (void*) GetProcAddress(h, "vprintf");
        if (vpr == NULL)
            DebugBreak();
    }

    va_start(arg,format);
    ret = (*vpr)( format, arg );
    va_end(arg);

    return ret;
}

void NT_puts(char *s)
{
    DWORD count;
    UINT Len = strlen(s);

    WriteFile(StdOutHandle, s, Len, &count, NULL);
    if (count != Len)
        DebugBreak();
}

void NT_putchar(unsigned char c)
{
    DWORD written;

    WriteFile(StdOutHandle, &c, 1, &written, NULL);
    if (written != 1)
        DebugBreak();
}

int NT_getchar(void)
{
    DWORD count, rv = 0;
    int c;
    BOOL ok;
    HANDLE StdInHandle = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD Input;
    DWORD OldMode = 0, NewMode;

#if 1
    /* Modify console mode */
    if ( GetConsoleMode(StdInHandle, &OldMode) == FALSE ) {
        assert(FALSE);
        return -1;              /* EOF */
    }

    NewMode = OldMode & ~(ENABLE_LINE_INPUT
                          | ENABLE_ECHO_INPUT
                          | ENABLE_WINDOW_INPUT
                          | ENABLE_MOUSE_INPUT);
    NewMode = 0;

    if ( SetConsoleMode(StdInHandle, NewMode) == FALSE ) {
        assert(FALSE);
        return -1;
    }


    for (;;) {
        rv = WaitForSingleObject(StdInHandle, 0);
        if (rv == WAIT_TIMEOUT) {
            c = -2; /* Would block */
            break;
        }
        assert(rv == WAIT_OBJECT_0);

        ok = ReadConsoleInput(StdInHandle, &Input, 1, &count);
        if (!ok || count != 1) {
            assert(FALSE);
            c = -1;
            break;
        }

        if (Input.EventType == KEY_EVENT
            && Input.Event.KeyEvent.bKeyDown == TRUE)
        {
            c = Input.Event.KeyEvent.uChar.AsciiChar;
            break;
        }
    }

    /* Restore mode. Makes control c work again. */
    SetConsoleMode(StdInHandle, OldMode);
#else
    ok = ReadFile(StdInHandle, &c, 1, &count, NULL);
    if (!ok || count == 0)
        return -1; /* EOF */

    if (count != 1)
        DebugBreak();
#endif

    return (int) c;
}

void DCPutChar(UINT8 c)
{
    //UINT32 State;

    //TURN_INTERRUPTS_OFF(State);
    NT_putchar(c);
    //RESTORE_INTERRUPTS(State);
}

/* Caveat emptor: This function will block with interrupts disabled. */
int DCGetChar( BOOL DoEcho )
{
    int c;
    //UINT32 State;
    INT Delay = 1;

    for (;;) {
        //TURN_INTERRUPTS_OFF(State);
        c = NT_getchar();
        //RESTORE_INTERRUPTS(State);

        if (c > 0)
            break;

        if (c == -2) {          /* would block */
            SleepUntil(TIME_RELATIVE(TIME_MILLIS(Delay)));
            Delay += Delay;
            if (Delay > 100)
                Delay = 100;
        }
    }

    if (c == '\r')
        c = '\n';

    if ( DoEcho )
        DCPutChar((UINT8) c);

    return c;
}

int puts(const char *s)
{
    DWORD count;
    UINT Len = strlen(s);

    WriteFile(StdOutHandle, s, Len, &count, NULL);
    if (count != Len) {
        DebugBreak();
        return EOF;
    }
    return 0;
}

PTR pTheScheduler = NULL;
