/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Prints how long Delay() takes in a few cases
 */

#if defined(TUNEDELAY)
#define DELAY(_x_) MyDelay(_x_)
#define MAIN(_c_,_v_) TestResult(_c_,_v_)
#else
#define DELAY(_x_) Delay(_x_)
#define MAIN(_c_,_v_) _tmain(_c_,_v_)
#endif

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

void pdec(int d, int f)
{
    _TCHAR n[32];
    _sntprintf(n,32,_T("%d"),d);
    d = _tcslen(n);
    n[d+1] = 0;
    for (; (f > 0) && (d > 0); f--, d--)
        n[d] = n[d-1];
    n[d] = _T('.');
    _putts(n);
}

BOOL Verbose = FALSE;
UINT TimingOverhead      = 0;

/* Compute & print timing overhead
 */
void ComputeOverhead(void)
{
    TIME Ts, Te;
    UINT Tmin, Tcur, i;
    
    /* Repeat ten times and keep the non-zero minimum
     */
    Tmin = ~0;
    for (i = 0; i < 10; i++) {
        /* Compute the overhead via two consecutive calls.
         * NB: This does not work if the timer's granularity is too coarse.
         */
        Ts = CurrentTime();
        Te = CurrentTime();
        Tcur = Int64ToInt32(TimeSubtract(Te,Ts));


        if (Tcur && Tcur < Tmin)
            Tmin = Tcur;
    }
    TimingOverhead = Tmin;

    _tprintf(_T("Timing delta d="));
    pdec(TimingOverhead,1);
    _tprintf(_T("\r\n"));

}

void TimeIt(int t)
{
    TIME s,e;
    int i;

    _tprintf(_T("Delay(%d):"),t);
    
    for (i = 0; i < 10; i++) {
        _putts(_T(" "));
        s = CurrentTime();
        DELAY(t);
        e = CurrentTime();

        e = TimeSubtract(e,s);
        pdec(Int64ToInt32(e)-TimingOverhead,1);
    }

    /* Now do it again different
     */
    s = CurrentTime();
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    DELAY(t);
    e = CurrentTime();

    e = TimeSubtract(e,s);
    _tprintf(_T(" ["));
    pdec(Int64ToInt32(e)-TimingOverhead,2);
    _tprintf(_T("]\r\n"));
}

/*
 */
int MAIN(int argc, _TCHAR **argv)
{
    int t;

    /* Compute & print timing overhead
     */
    ComputeOverhead();

    /* Print logscale, sortof. No more than a millisec
     */
    for (t = 0; t < 10; t++)
        TimeIt(t);

    for (t = 10; t < 100; t += 10)
        TimeIt(t);

    for (t = 100;t <= 1000; t += 100)
        TimeIt(t);

    return 0;
}


