/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <arm\at91m55800.h>

#define BcdToInt8(_bcd_) (INT8)(((_bcd_) & 0xf) + 10 * (((_bcd_)>>4)&0xf))
#define Int8ToBcd(_int_) (((_int_) % 10) | (((_int_) / 10) << 4))


/* Simple test for the M55800A processor's RTC subsystem.
 */
typedef struct _MY_DATE {
    INT8 Seconds;
    INT8 Minutes;
    INT8 Hours;
    INT8 Month;
    INT16 Year;
    INT8 Day;
    INT8 WeekDay;
} MY_DATE;

BOOL InvalidDate(MY_DATE this)
{
    if ((this.Seconds < 0) ||
        (this.Seconds > 59) ||
        (this.Minutes < 0) ||
        (this.Minutes > 59) ||
        (this.Hours < 0) ||
        (this.Hours > 24) ||
        (this.Day < 0) ||
        (this.Day > 31) ||
        (this.Month < 0) ||
        (this.Month > 12) ||
        (this.Year < 0) ||
        (this.WeekDay < 0) ||
        (this.WeekDay > 7))
        return TRUE;
    if ((this.Hours == 24) &&
        ((this.Minutes > 0) || (this.Seconds > 0)))
        return TRUE;

    /* More calendar checks done by the RTC later
     */
    return FALSE;
}

MY_DATE GetDate(struct _Rtc *this)
{
    MY_DATE r;
    UINT v;

    memset(&r,0,sizeof r);

    if (this->Mode & RTCM_UPDTIM) {
        _tprintf(_T("RTC timer is stopped.\r\n"));
    }

    if (this->Mode & RTCM_UPDCAL) {
        _tprintf(_T("RTC calendar is stopped.\r\n"));
    }

    v = this->Time;
    r.Seconds = BcdToInt8(RTCT_SEC(v));
    r.Minutes = BcdToInt8(RTCT_MIN_G(v));
    r.Hours   = BcdToInt8(RTCT_HR_G(v));
    if (v & RTCT_PM) r.Hours += 12;

    v = this->Calendar;
    r.Day   = BcdToInt8(RTCC_MDAY_G(v));
    r.WeekDay = BcdToInt8(RTCC_WDAY_G(v)); /* not really but.. */
    r.Month = BcdToInt8(RTCC_MNTH_G(v));
    r.Year  = BcdToInt8(RTCC_YEAR_G(v));
    r.Year += 100 * BcdToInt8(RTCC_CENT_G(v));

    return r;
}

BOOL SetDate(struct _Rtc *this, MY_DATE NewDate)
{
    BOOL ok = TRUE;
    UINT vT, vC;

    vT = RTCT_SEC(Int8ToBcd(NewDate.Seconds)) |
         RTCT_MIN(Int8ToBcd(NewDate.Minutes)) |
         RTCT_HR(Int8ToBcd(NewDate.Hours));

    vC = RTCC_CENT(Int8ToBcd(NewDate.Year / 100)) |
         RTCC_YEAR(Int8ToBcd(NewDate.Year % 100)) |
         RTCC_MNTH(Int8ToBcd(NewDate.Month)) |
         RTCC_MDAY(Int8ToBcd(NewDate.Day)) |
         RTCC_WDAY(Int8ToBcd(NewDate.WeekDay));

    this->Mode |= RTCM_UPDTIM | RTCM_UPDCAL;
    while ((this->Status & RTCS_ACK_UPD) == 0) {
        /*BUGBUG*/;
    }
    this->Clear = RTCS_ACK_UPD;

    this->HourMode = RTCHM_24_HOURS;
    this->Time = vT;
    this->Calendar = vC;

    this->Mode &= ~(RTCM_UPDTIM | RTCM_UPDCAL);

    /* How long does this take ? Umbug */
    Delay(10);

    vT = this->ValidEntry & (RTCV_TIME|RTCV_CALENDAR);
    ok = (vT == 0);
    if (!ok) {
        _tprintf(_T("RTC says the %s fields are invalid\r\n"),
                 ((vT & (RTCV_TIME|RTCV_CALENDAR))==(RTCV_TIME|RTCV_CALENDAR))?
                   _T("TIME and CALENDAR") :
                 ((vT & RTCV_TIME) ? _T("TIME") : _T("CALENDAR"))
                 );
    }

    return ok;
}

/* Printing & Parsing
 */
typedef _TCHAR *NAME;
static NAME Months[12] = {
    _T("January"), _T("February"), _T("March"),
    _T("April"), _T("May"), _T("June"), _T("July"),
    _T("August"), _T("September"), _T("October"), _T("November"),
    _T("December")
};
#define MonthName(m) Months[m-1]

int a2m(_TCHAR *Month)
{
    int i;
    for (i = 0; i < 12; i++)
        if (_tcsicmp(Month,Months[i]) == 0)
            return i+1;
    return -1;
}

static NAME Days[7] = {
    _T("Mon"), _T("Tues"), _T("Wednes"), _T("Thurs"), _T("Fri"),
    _T("Satur"), _T("Sun")
};
#define DayName(d) Days[d-1]

int a2d(_TCHAR *Day)
{
    int i;
    i = _tcslen(Day);
    if (i > 3)
        Day[i-3] = 0;
    for (i = 0; i < 7; i++)
        if (_tcsicmp(Day,Days[i]) == 0)
            return i+1;
    return -1;
}

void PrintDate(MY_DATE this)
{
    _tprintf(_T("%s %d, %d (%sday) at %d:%d:%d\r\n"),
             MonthName(this.Month),
             this.Day, this.Year,
             DayName(this.WeekDay),
             this.Hours, this.Minutes, this.Seconds);
}

/* Main driving program
 */
int _tmain (INT argc, _TCHAR **argv)
{
    MY_DATE NewDate;
    int h,m,s;

    _tprintf(_T("RTC says today is: "));
    PrintDate(GetDate(TheRtc));

    /* Should we change the date
     */
    (void) ParseCmdLine(&argc, &argv);
    if (argc > 6) {

        /* init */
        memset(&NewDate,0,sizeof NewDate);

        /* Parse the various fields, we do have enough strings
         */
        NewDate.Month = a2m(argv[1]);
        NewDate.Day = _ttoi(argv[2]);
        NewDate.Year = _ttoi(argv[3]);
        /* Make parens optional around day of week */
        s = _tcslen(argv[4]);
        if (argv[4][s-1] == _T(')')) {
            argv[4][s-1] = 0;
            argv[4]++;
        }
        NewDate.WeekDay = a2d(argv[4]);
        _stscanf(argv[6],_T("%d:%d:%d"), &h, &m, &s);
        NewDate.Hours = h;
        NewDate.Minutes = m;
        NewDate.Seconds = s;

        /* Validate input
         */
        if (InvalidDate(NewDate)) {

            _tprintf(_T("Invalid input, time/date should be valid and formatted like the above!\r\n"));

        } else {

            /* Try to set it, RTC might refuse
             */
            _tprintf(_T("Setting the date to "));
            PrintDate(NewDate);

            if (SetDate(TheRtc,NewDate)) {
                _tprintf(_T("Checking: "));
                PrintDate(GetDate(TheRtc));
            } else {
                _tprintf(_T("Sorry.\r\n"));
            }
        }
    }

    (void) FreeCmdLine(&argc,&argv);

    return 0;    
}





