/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <stdio.h>
#include <arm/at91x40.h>
#include <stdlib.h>
#include <tchar.h>

/* Prints info off the Special Functions unit of the Atmel at91xxx processors
 */
int _tmain(int argc, _TCHAR **argv)
{
    UINT32 id, v, t;
#define nvps(_x_) (((_x_)&1) ? (16 << ((_x_) >> 1)) : 0)
#define  vds(_x_) ((_x_))

    id = TheSf->ChipID;
    id = _bs32(id);
    v = id;

    _tprintf(_TEXT("ChipID      is 0x%x:\n\r"), v);
    if ((v & SFID_FXMASK) != SFID_FX) {
        _tprintf(_TEXT("That aint right!\n\r"));
        goto Done;
    }

    _tprintf(_TEXT("\tVersion: %d\n\r"), v & SFID_VERSION);

    _tprintf(_TEXT("\tArchitecture: AT91x%d%d\n\r"),
             (v & SFID_ARCH) >> 24, ((v & SFID_ARCH) >> 20) & 0xf);

    _tprintf(_TEXT("\tNonvolatile memory:\n\r"));
    t = (v & SFID_NVPTYP) >> 28;
    _tprintf(_TEXT("\t\ttype: %c\n\r"),
             (t == 8) ? _TEXT('R') :
             (t == 1) ? _TEXT('M') :
             (t == 2) ? _TEXT('C') : 
             (t == 3) ? _TEXT('S') : _TEXT('?'));
    _tprintf(_TEXT("\t\tprogram size: %dKB\n\r"),
             nvps( (v & SFID_NVPSIZ) >> 8 ));
    _tprintf(_TEXT("\t\tdata size: %dKB\n\r"),
             nvps( (v & SFID_NVDSIZ) >> 12 ));
 
   _tprintf(_TEXT("\tVolatile memory:\n\r"));
    _tprintf(_TEXT("\t\tdata size: %dKB\n\r"),
             vds( (v & SFID_VDSIZ) >> 16 ));


    if (id & SFID_EXT) {
        v = TheSf->ChipIDExt;
        v = _bs32(v);
        _tprintf(_TEXT("ChipIDExt is 0x%x\n\r"), v);
    }

    v = TheSf->ResetStatus;
        v = _bs32(v);
    _tprintf(_TEXT("ResetStatus is 0x%x (%s reset)\n\r"), v,
             (v == SFRS_EXT) ? _TEXT("external") :
             (v == SFRS_WD) ? _TEXT("watchdog") : _TEXT("unknown"));

    /* This bit only applies to the AR91R40807
     */
    if (id == 0x44080746) {
        v = TheSf->MemoryMode;
        v = _bs32(v);
        _tprintf(_TEXT("MemoryMode  is 0x%x (internal RAM is %sprotected)\n\r"), v,
                 (v & 1) ? _TEXT("not ") : _TEXT(""));
    }

    v = TheSf->ProtectMode;
    v = _bs32(v);
    _tprintf(_TEXT("ProtectMode is 0x%x (AIC is in %s mode)\n\r"), v,
             (v & SFPM_AIC) ? _TEXT("protected") : _TEXT("normal"));

 Done:
    exit(EXIT_SUCCESS);
}
