/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

/* Simple test for simulated framebuffer (e.g. inside a processor simulator)
 * Only works for Armulate and tmsim.
 */

#if defined(arm)
#define FB_BASE 0x80100000
#elif defined(__TCS__)
#define TILER_MMIO_START (0xf0000000)
#define FB_BASE (TILER_MMIO_START+0)
#else
#define FB_BASE 666
#endif
#define FB_PALETTE_LOAD (FB_BASE+0x10000)

struct PAL {
 UINT8 Red;
 UINT8 Blue;
 UINT8 Green;
} Palette[256] = {
 { 0xff, 0xba, 0xd0 },
 { 0, },
};

int _tmain(int argc, _TCHAR **argv)
{
    UINT32 *pFb = (UINT32 *)FB_BASE;
    int i;
    
    UnusedParameter(argc);
    UnusedParameter(argv);

    _tprintf(_TEXT("Testing FrameBuffer at x%x\n"), FB_BASE);

    for (i = 0; i < 256; i++) {
        Palette[i].Red = i;
        Palette[i].Blue = i;
        Palette[i].Green = i;
    }

    _tprintf(_TEXT("Loading palette at x%x from x%x\n"), 
           FB_PALETTE_LOAD, (ADDRESS) Palette);

#if defined(__TCS__)
    /* workaround for...a messy issue */
    memcpy((char *)Palette + 0x800000, Palette, sizeof Palette);
    _tprintf("done copy\n");

    *(UINT32 *)FB_PALETTE_LOAD = (UINT32) Palette + 0x800000;
    _tprintf("done load\n");
#else
    *(UINT32 *)FB_PALETTE_LOAD = (UINT32) Palette;
#endif
    for (i = 0; i < (320*200/4); i++)
        pFb[i] = (UINT32) (i | (i << 16));


    /* dies with data abort if the device is not simulated */

    _tprintf(_TEXT("Test PASSED"));

    exit(EXIT_SUCCESS);
    return 0;
}
