/*
rd.c - Simple utility to read data from a file. 
Calls BindToObject() with flags passed on cmd line. 
Copyright (c) Microsoft Corporation. All rights reserved. 
a-ericf
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"


int _tmain(INT argc, _TCHAR **argv)
{
    UINT64 nPos, nSize; 
    UINT64 nToRead;
    UINT nRead = 0;
    UINT flags = 0;
    SCODE sc; 
    _TCHAR sBuffer[1024]; 
    const _TCHAR * sStartPos = 0;
    IFile * pFile = 0;

    //file size and position values can be 64 bit so conversion is req'd. 
    //owing to lack of _atoi64() function at press time
    Uint32ToUint64(nPos,0);
    Uint32ToUint64(nSize,0);
    Uint32ToUint64(nToRead,0);

    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv);

    if(argc < 2) 
    {
        _tprintf(_T("\n\rUsage:RD.EXE <filename>[flags]")); 
        _tprintf(_T("\n\rFlags are: -r -w -c -x -sr -sw -bogus -startpos <0xoffset> -size <0xsize>")); 
        _tprintf(_T("\n\rMeaning: -r->Read -w->Write -c->Create -x->fail if eXists -sr->Share Read -sw->Share Write -size <hexnum>-><siZe in hex> -startpos <hex starting pos> -bogus->Bogus flag")); 
        TellAboutOptionalBreak();
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    //start setting up call
    //unfortunately we can only get 32 bit args for now
    flags = FetchFlagsFromCommandLine(argc, argv, &nToRead); 
    sStartPos = GetArgFromCommandLine(argc, argv, _T("startpos")); 
    if(sStartPos) 
    {
        _TCHAR * stop = 0; 
        Uint32ToUint64(nPos,_tcstoul(sStartPos, &stop, 16);) 
    }
     
    //bind if possible ... 
    sc = BindToObject(CurrentNameSpace(), argv[1], flags, &IID_IFile, (void **)&pFile);
    
    Outcome(BIND, sc, argv[1], flags, nToRead);
    if(FAILED(sc))
    {
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    //stuff a null in there
    sBuffer[0] = 0;
    sBuffer[sizeof(sBuffer) - 1] = 0;  

    //loop through reading and printing to screen
    for(
        nRead = 0, sc = S_OK; 
        SUCCEEDED(sc) && sc != S_FALSE && Uint64Less(UINT64_ZERO,nToRead); 
        )
    {
        UINT32 ReadNow = (sizeof(sBuffer) - sizeof(sBuffer[0]));

        if ((Uint64High32Bits(nToRead) == 0) &&
            (Uint64Low32Bits(nToRead) < ReadNow))
            ReadNow = Uint64Low32Bits(nToRead);

        sc = pFile->v->ReadAt(
            pFile, 
            nPos, 
            (BYTE*)sBuffer, 
            ReadNow,
            &nRead); 

        _tprintf(sBuffer);
        nPos = Uint64AddUint32(nPos,nRead);
        nToRead = Uint64SubtractUint32(nToRead,nRead);
    }
       
    _tprintf(_T("\n\rRead [%d] bytes (truncated from 64 -> 32 bits)"), Uint64Low32Bits(nPos)); 
    Outcome(READ, sc, argv[1], flags, nToRead);
    pFile->v->Release(pFile); 
    FreeCmdLine(&argc, &argv);

    return SetAndReturn(0); 
}
