/*
* wr.c - Test IFile::WriteAt() interface. 
* Open a file bound with various flags and write at various locations. 
* Copyright (c) Microsoft Corporation. All rights reserved. 
*/

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"

void WRUsage(const _TCHAR *);

_TCHAR sMsg[] = _T("Where do you want to write today?"); 

int _tmain(INT argc, _TCHAR **argv)
{
    UINT64 nPos; 
    UINT64 nToRead, nByteOutcome;
    UINT nRead = 0, nWritten = 0;
    UINT flags = 0;
    SCODE sc; 
    _TCHAR sBuffer[1024]; 
    const _TCHAR *sFileIn = 0,  *sFileOut = 0, *sStartPos = 0;
    IFile * pFileIn = 0, *pFileOut = 0;

    Uint32ToUint64(nPos,0);
    Uint32ToUint64(nToRead,0);
    Uint32ToUint64(nByteOutcome,0);

    ParseCmdLine(&argc, &argv); 

    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    if(argc < 2) 
    {
        WRUsage(argv[0]);
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    flags = FetchFlagsFromCommandLine(argc, argv, &nToRead); 
    sFileIn = GetArgFromCommandLine(argc, argv, _T("infile"));
    if(!sFileIn)
    {
        WRUsage(argv[0]); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    sFileOut = GetArgFromCommandLine(argc, argv, _T("outfile"));
    if(!sFileOut)
    {
        WRUsage(argv[0]); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    sStartPos = GetArgFromCommandLine(argc, argv, _T("startpos")); 
    if(sStartPos) 
    {
        _TCHAR * stop = 0; 
        Uint32ToUint64(nPos,_tcstoul(sStartPos, &stop, 16);) 
    }

    //bind to input file with passed parameters
    sc = BindToObject(CurrentNameSpace(), sFileIn, flags, &IID_IFile, (void **)&pFileIn);
    Outcome(BIND, sc, sFileIn, flags, nToRead);
    if(FAILED(sc))
    {
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    //bind output ... 
    sc = BindToObject(CurrentNameSpace(), sFileOut, flags, &IID_IFile, (void **)&pFileOut);
    Outcome(BIND, sc, sFileOut, flags, nToRead);
    if(FAILED(sc))
    {
        pFileIn->v->Release(pFileIn); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    //stuff a null in there
    sBuffer[0] = 0;
    sBuffer[sizeof(sBuffer) - 1] = 0;  

    //loop through reads and writes ... 
    for(sc = S_OK, nRead = nWritten = 0; 
        Uint64Less(UINT64_ZERO,nToRead); 
        nWritten = 0)
    {
        UINT32 ReadNow = (sizeof(sBuffer) - sizeof(sBuffer[0]));

        if ((Uint64High32Bits(nToRead) == 0) &&
            (Uint64Low32Bits(nToRead) < ReadNow))
            ReadNow = Uint64Low32Bits(nToRead);

        sc = pFileIn->v->ReadAt(
            pFileIn, 
            nPos, 
            (BYTE*)sBuffer, 
            ReadNow,
            &nRead); 
        if(FAILED(sc))
        {
            Outcome(READ, sc, sFileIn, flags, nToRead);
            FreeCmdLine(&argc, &argv); 
            pFileIn->v->Release(pFileIn); 
            pFileOut->v->Release(pFileOut); 
            return SetAndReturn(1);
        }
        Uint32ToUint64(nByteOutcome,nRead);
        if(nRead == 0)
        {
            _tprintf(_T("\n\rNumRead is 0"));
            Outcome(READ, sc, sFileOut, flags, nByteOutcome);
            FreeCmdLine(&argc, &argv); 
            pFileIn->v->Release(pFileIn); 
            pFileOut->v->Release(pFileOut); 
            return SetAndReturn(1);
        }
        sc = pFileOut->v->WriteAt(
            pFileOut, 
            nPos, 
            (BYTE*)sBuffer, 
            min((sizeof(sBuffer) - sizeof(sBuffer[0])), nRead), 
            &nWritten); 
        Uint32ToUint64(nByteOutcome,nWritten);
        if(FAILED(sc))
        {
            Outcome(WRITE, sc, sFileOut, flags, nByteOutcome);
            FreeCmdLine(&argc, &argv); 
            pFileIn->v->Release(pFileIn); 
            pFileOut->v->Release(pFileOut); 
            return SetAndReturn(1);
        }
        if(nRead != nWritten)
        {
            _tprintf(_T("\n\rNumRead[%d] != NumWritten[%d]"), nRead, nWritten);
            Outcome(WRITE, sc, sFileOut, flags, nByteOutcome);
            FreeCmdLine(&argc, &argv); 
            pFileIn->v->Release(pFileIn); 
            pFileOut->v->Release(pFileOut); 
            return SetAndReturn(1);
        }

        nToRead = Uint64SubtractUint32(nToRead,nRead);
        nPos = Uint64AddUint32(nPos,nRead);

    }
        Outcome(WRITE, sc, sFileOut, flags, nByteOutcome);
        FreeCmdLine(&argc, &argv); 
        pFileIn->v->Release(pFileIn); 
        pFileOut->v->Release(pFileOut); 
        return SetAndReturn(0);
}

void WRUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:%s -infile <input file name> -outfile <output file name> [flags]"), s); 
    _tprintf(_T("\n\rFlags are: -r -w -c -x -sr -sw -bogus -size <0xsize> -startpos <0xstartpos>")); 
    _tprintf(_T("\n\rMeaning: -r->Read -w->Write -c->Create -x->fail if eXists -sr->Share Read -sw->Share Write -size <hexnum>-><siZe in hex> -startpos <starting position> -bogus->Bogus flag")); 
}
