/*
* AllMem.c
* Attempts to allocate all available memory (fat chance!), then steps 
* sizes down in small increments until allocation succeeds. 
* Intended as a quickie stress test to ensure heap is not corrupted by failed 
* allocations or by allocation very close to maximum size boundary. 
* Test is expected to bugcheck if it fails, or to trash heap ... does not do 
* much error reporting. 
* Copyright (c) Microsoft Corporation All Rights Reserved
* BUGBUG: Only uses default heap, need to make it use various types of 
* heaps passed in on cmd line. 
* a-ericf
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "heaputils.h"
#include "..\utilities\utility.h"


int _tmain(INT argc, _TCHAR **argv)
{
    ADDR_SIZE reserved0, committed0, used0, maxused0;
    ADDR_SIZE reserved1, committed1, used1, maxused1;
    UINT nBytes;
    IHeap * pHeap = CurrentHeap();
        
    PTR p = 0; 

    TellAboutOptionalBreak(); 
    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv);

    //get amt of memory available
    pHeap->v->Status(pHeap, &reserved0, &committed0, &used0, &maxused0); 
    _tprintf(_T("\n\rMem size, prealloc:[%d]"), reserved0); 

    //starting at max reserved size, step down 8 bytes at a time. 
    //Quit when alloc passes. 
    nBytes = reserved0; 
    for(p = 0; (!p) && nBytes >= 8; nBytes -= 8)
    {
        p = pHeap->v->Alloc(pHeap, 0, nBytes, 0); 
        _puttchar(_T('.'));
    }
//    _tprintf(_T("\n\rMem size returned, postalloc:[%d]"), pHeap->v->Size(pHeap, 0, p)); 
//    _tprintf(_T("\n\rMem size requested:[%d]"), nBytes); 
    pHeap->v->Status(pHeap, &reserved1, &committed1, &used1, &maxused1); 
//    _tprintf(_T("\n\rMemory overhead:[%d], Used [%d]"), reserved0 - nBytes, used1); 

    pHeap->v->Free(pHeap, 0, p); 

    _tprintf(_T("\n\rMem size, postfree:[%d]"), pHeap->v->Size(pHeap, 0, p)); 
    pHeap->v->Status(pHeap, &reserved1, &committed1, &used1, &maxused1); 
    _tprintf(_T("\n\rHeap size, postfree:[%d]"), reserved1); 

    FreeCmdLine(&argc, &argv);
    return SetAndReturn(0);
}

    
    
    




