/*
* BigGulp.c
* Tries to alloc and realloc a huge blob to make sure there is no corruption if
* alloc or realloc fail. 
* Copyright (c) Microsoft Corporation All rights reserved. 
* a-ericf
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"
#include "heaputils.h"

int _tmain(INT argc, _TCHAR **argv)
{
    PTR p = 0, pR = 0;
    ADDR_SIZE reserved, committed, used, maxused;

    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    //validate at first
    if(CurrentHeap()->v->Validate(CurrentHeap(), 0, 0) != S_OK)
    {
        DebugBreak(); 
    }

    else
        _tprintf(_T("\n\rValidate is OK ... "));

    //how much mem we got?
    CurrentHeap()->v->Status(CurrentHeap(), &reserved, &committed, &used, &maxused); 
    _tprintf(_T("\n\rMem size, prealloc:[%d]"), reserved); 

    //alloc all we got times 10
    p = CurrentHeap()->v->Alloc(CurrentHeap(), 0, reserved * 10, 0);

    //expect to fail!
    if(p)
    {
        _tprintf(_T("\n\rGot a pointer!?! Why?"));
        if(CurrentHeap()->v->Validate(CurrentHeap(), 0, p) != S_OK) 
        {
            DebugBreak(); 
        }
        else
            _tprintf(_T("\n\rValidate is OK ... "));

        //free
        CurrentHeap()->v->Free(CurrentHeap(), 0, p);
    }
    //we wanted to fail
    else
        _tprintf(_T("\n\rFailed as expected"));

    //did we thrash the heap?
    if(CurrentHeap()->v->Validate(CurrentHeap(), 0, 0) != S_OK) 
    {
        DebugBreak(); 
    }
    else
    {
        _tprintf(_T("\n\rValidate is OK after a huge alloc ..."));
        CurrentHeap()->v->Free(CurrentHeap(), 0, p);
    }


    //alloc a small hunk. This should work
    p = CurrentHeap()->v->Alloc(CurrentHeap(), 0, reserved / 10, 0);

    if(!p)
        _tprintf(_T("\n\rNo pointer!?! Why?"));
    else
    {
        _tprintf(_T("\n\rPassed as expected"));
        //make sure heap is OK
        if(CurrentHeap()->v->Validate(CurrentHeap(), 0, p) != S_OK) 
        {
            DebugBreak(); 
        }
        else
            _tprintf(_T("\n\rValidate is OK ... "));

        CurrentHeap()->v->Free(CurrentHeap(), 0, p);
    }



    //now for realloc
    CurrentHeap()->v->Status(CurrentHeap(), &reserved, &committed, &used, &maxused); 
    _tprintf(_T("\n\rMem size, pre-realloc:[%d]"), reserved); 

    //get a small hunk ... we are just getting a pointer for reallocation
    p = CurrentHeap()->v->Alloc(CurrentHeap(), 0, 0x100, 0);

    if(!p) {DebugBreak();}

    //BIG GULP
    pR = CurrentHeap()->v->Alloc(CurrentHeap(), 0, reserved * 100, 0);

    //expect pR == NULL here
    if(pR)
    {
        _tprintf(_T("\n\rGot a pointer!?! Why?"));
        if(CurrentHeap()->v->Validate(CurrentHeap(), 0, pR) != S_OK) 
        {
            DebugBreak(); 
        }
        else
            _tprintf(_T("\n\rValidate is OK ... "));

        CurrentHeap()->v->Free(CurrentHeap(), 0, pR);
    }
    else
    {
        _tprintf(_T("\n\rFailed as expected"));
        if(CurrentHeap()->v->Validate(CurrentHeap(), 0, p) != S_OK) 
        {
            DebugBreak(); 
        }

    }

    //did we thrash the heap?
    if(CurrentHeap()->v->Validate(CurrentHeap(), 0, 0) != S_OK) 
    {
        DebugBreak(); 
    }
    else
        _tprintf(_T("\n\rValidate is OK after a huge realloc ..."));


    //did we toast our pointer?
    pR = CurrentHeap()->v->Realloc(CurrentHeap(), 0, p, reserved / 10, 0);

    if(!pR)
    {
        _tprintf(_T("\n\rNo pointer!?! Why?"));
        DebugBreak(); 
    }
    else
    {
        //did we thrash the heap ... 
        _tprintf(_T("\n\rPassed as expected"));
        if(CurrentHeap()->v->Validate(CurrentHeap(), 0, pR) != S_OK) 
        {
            DebugBreak(); 
        }
        else
            _tprintf(_T("\n\rValidate is OK ... "));

        CurrentHeap()->v->Free(CurrentHeap(), 0, pR);
    }
 FreeCmdLine(&argc, &argv);
 return SetAndReturn(0);       
}
    








