/*
* ReallocSize0.c 
* Initially used to test reallocation of size 0 blocks, this is a general purpose
* heap testing tool .... 
* Allows user to try allocation, reallocation, and validation of various size blocks
* with various flags. 
* todo:extend to setting aligments also
* todo:extend to different types of heap
* Copyright (c) Microsoft Corporation All rights reserved
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "heaputils.h"
#include "..\utilities\utility.h"

int _tmain(INT argc, _TCHAR ** argv)
{
    SCODE sc;
    IHeap * pHeap = CurrentHeap(); 
    PTR *pMem, *pReallocMem; 
    _TCHAR * stop;
    const _TCHAR * sArg = 0;
    UINT nAllocSize = 0, nReallocSize = 0, nFlags = 0, iterations = 1;
        
    ParseCmdLine(&argc, &argv);
    OptionalBreak(argc, argv);
    TellAboutOptionalBreak();

    sc = pHeap->v->Validate(pHeap, 0, 0); 
    Outcome(VALIDATE, sc, _T(__FILE__), 0,UINT64_ZERO);
  
    if(argc < 2)
        _tprintf(_T("\n\rUsage:%s [-alloc <allocsize base 10>][-realloc <reallocsize base 10>][-flags <flags in hex>][-iter <iterations base 10>\r\n...Or use blank params for defaults"), argv[0]);
    
    sArg = GetArgFromCommandLine(argc, argv, _T("alloc")); 
    if(sArg)
        nAllocSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("realloc")); 
    if(sArg)
        nReallocSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("flags")); 
    if(sArg)
        nFlags = _tcstoul(sArg, &stop, 16); 

    sArg = GetArgFromCommandLine(argc, argv, _T("iter")); 
    if(sArg)
        iterations = _tcstoul(sArg, &stop, 10); 
    
    //loop through many times ... 
    for( ; iterations; --iterations)
    {

        _tprintf(_T("\n\rIteration %d: Allocating %d, reallocating %d, flags are 0x%x"), iterations, nAllocSize, nReallocSize, nFlags); 
    
        pMem = pHeap->v->Alloc(pHeap, nFlags, nAllocSize, 8); 

        if(!pMem) 
        {
            error(_T("_tmain()"), _T(__FILE__), __LINE__, _T("\n\rInitial alloc failed")); 
            params(pHeap, pMem, nFlags, nAllocSize, 8);
            sc = pHeap->v->Validate(pHeap, 0, 0); 
            Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
            FreeCmdLine(&argc, &argv);
            return SetAndReturn(1);
        }

        sc = pHeap->v->Validate(pHeap, 0, pMem); 
        params(pHeap, pMem, nFlags, nAllocSize, 8);
        Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
      
        pReallocMem = pHeap->v->Realloc(pHeap, nFlags, pMem, nReallocSize, 8); 

        if(!pReallocMem)
        {
            SCODE hr;
            
            /* If the size is 0 we actually requested a free (ISO specs) */
            if (nReallocSize == 0)
                _tprintf(_T("\n\rFree-via-Realloc succeeded"));
            else
            {
                hr = pHeap->v->Validate(pHeap, 0, pMem); 
                Outcome(VALIDATE, hr, _T(__FILE__), 0, UINT64_ZERO);
                _tprintf(_T("\n\rReallocation failed, validate sez %d"), hr);
                /* Make sure the memory is still good before we free it */
                if(SUCCEEDED(hr))
                    pHeap->v->Free(pHeap, 0, pMem);
                else
                {
                    error(_T("_tmain()"), _T(__FILE__), __LINE__, _T("\n\rReallocation failed"));  
                    params(pHeap, pMem, nFlags, nReallocSize, 8);
                }
            }
        }
        else
        {
            _tprintf(_T("\n\rReallocation succeeded")); 
            sc = pHeap->v->Validate(pHeap, 0, pReallocMem); 
            Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
            pHeap->v->Free(pHeap, 0, pReallocMem); 
        }
    }

    _tprintf(_T("\n\rFinishing up...")); 
    sc = pHeap->v->Validate(pHeap, 0, 0); 
    Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
    FreeCmdLine(&argc, &argv);
    return SetAndReturn(0);
}

