/*
Basic testing for Heap::Validate() function
Does various allocs and reallocs and calls Validate() in a loop
Copyright (c) Microsoft Corporation All rights reserved
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "heaputils.h"
#include "..\utilities\utility.h"


int _tmain(INT argc, _TCHAR ** argv)
{
    IHeap * pHeap = CurrentHeap(); 
    PTR pMem, pReallocMem; 
    _TCHAR * stop;
    SCODE sc;
    BOOL b;
    const _TCHAR * sArg = 0;
    UINT32 nAllocSize = 0, nReallocSize = 0, nFlags = 0, iterations = 1;
    UINT64 un64;
     
    ParseCmdLine(&argc, &argv);
    OptionalBreak(argc, argv);
    TellAboutOptionalBreak();
  
    if(argc < 2)
        _tprintf(_T("\n\rUsage:%s [-alloc <allocsize base 10>][-realloc <reallocsize base 10>][-flags <flags in hex>][-iter <iterations base 10>\r\n...Or use blank params for defaults"), argv[0]);
    
    sArg = GetArgFromCommandLine(argc, argv, _T("alloc")); 
    if(sArg)
        nAllocSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("realloc")); 
    if(sArg)
        nReallocSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("flags")); 
    if(sArg)
        nFlags = _tcstoul(sArg, &stop, 16); 

    sArg = GetArgFromCommandLine(argc, argv, _T("iter")); 
    if(sArg)
        iterations = _tcstoul(sArg, &stop, 10); 

    for( ; iterations; --iterations)
    {
        _tprintf(_T("\n\rIteration %d: Allocating %d, reallocating %d, flags are 0x%x"), iterations, nAllocSize, nReallocSize, nFlags); 
    
        pMem = pHeap->v->Alloc(pHeap, nFlags, nAllocSize, 8); 
        
        if(!pMem) 
        {
            _tprintf(_T("\n\rInitial alloc failed")); 
            sc = pHeap->v->Validate(pHeap, 0, pMem);

            Int32ToInt64(un64, nReallocSize);
            Outcome(VALIDATE, sc, _T(__FILE__), nFlags, un64);
            if(FAILED(sc))
            {
                _tprintf(_T("\n\rValidate failed - why?")); 
                DebugBreak(); 
            }

            FreeCmdLine(&argc, &argv);
            return SetAndReturn(1); 
        }

        _tprintf(_T("\n\rInitial alloc succeeded")); 
    
        pReallocMem = pHeap->v->Realloc(pHeap, nFlags, pMem, nReallocSize, 8); 

        if(!pReallocMem)
        {
            _tprintf(_T("\n\rReallocation failed")); 
            sc = pHeap->v->Validate(pHeap, 0, pMem);

            Int32ToInt64(un64, nReallocSize);

            Outcome(VALIDATE, sc, _T(__FILE__), nFlags, un64);
            if(FAILED(sc))
            {
                _tprintf(_T("\n\rValidate failed - why?")); 
                DebugBreak(); 
            }
                
            b = pHeap->v->Free(pHeap, 0, pMem);
            FreeCmdLine(&argc, &argv);
        }
        else
        {
            _tprintf(_T("\n\rReallocation succeeded")); 
            pHeap->v->Free(pHeap, 0, pReallocMem); 
        }
    }
    
    FreeCmdLine(&argc, &argv);
    return SetAndReturn(0);
}

