/* 
* Trivial test for the Debug heap
* Copyright (c) 2002 Microsoft Corp. All rights reserved. 
* This test creates a special debug heap which maintains cookies
* protecting the beginning and end of the allocated block. 
*/

#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <stdlib.h>

#define TS 8

void dump(UINT32 *p, UINT Size); 
int _tmain(int argc, _TCHAR **argv);

int _tmain(int argc, _TCHAR **argv)
{
    PIHEAP Heap;
    PIHEAPFACTORY HeapFactory;
    UINT32 *p;
    UINT i;
    SCODE sc;

    sc = BindToObject(CurrentNameSpace(), _T("COB/dbg_heap.cob"),
                      NAME_SPACE_READ, &IID_IHeapFactory,
                      (void **) &HeapFactory);
    if (FAILED(sc))
        goto Bad;

    sc = HeapFactory->v->CreateNested(HeapFactory, CurrentHeap(),
                                      0,10*1024,TS, &Heap);
    if (FAILED(sc))
        goto Bad;

    _tprintf(_T("CreateDebugHeap got x%x\n\r"),Heap);
    if (Heap == NULL)
        goto Bad;

    i = Heap->v->AddRef(Heap);
    _tprintf(_T("AddRef got x%x\n\r"),i);
    if (i != 2) goto Bad;

    i = Heap->v->Release(Heap);
    _tprintf(_T("Release got x%x\n\r"),i);
    if (i != 1) goto Bad;

    p = Heap->v->Alloc(Heap,HEAP_ZERO_MEMORY,33,0);
    _tprintf(_T("Alloc got x%x\n\r"),(ADDRESS)p);
    if (p == NULL)
        goto Bad;

    dump(p,33);

    i = Heap->v->Free(Heap,0,p);
    _tprintf(_T("Free got x%x\n\r"),i);
    if (i != 1) goto Bad;

    i = Heap->v->Release(Heap);
    Heap = NULL; /* regardless */
    _tprintf(_T("Release got x%x\n\r"),i);
    if (i != 0) goto Bad;

    _tprintf(_T("Trivia test for Debug heap PASSED\n\r"));
    exit(EXIT_SUCCESS);
    return 0;
 Bad:
    if (Heap) {
        i = Heap->v->Release(Heap);
        _tprintf(_T("Release got x%x\n\r"),i);
    }
    _tprintf(_T("Trivia test for Debug heap FAILED\n\r"));
    exit(EXIT_FAILURE);
    return 1;
}


void dump(UINT32 *p, UINT Size)
{
    UINT i;

    /* back up to start of header */
    p -= 2;

    /* roundup to words */
    Size = (Size + 3);

    _tprintf(_T("Block @%x:"),(ADDRESS)p);
    for (i = 0; i < (8+TS+Size)>>2; i++)
        _tprintf(_T(" %x"),p[i]);
    _tprintf(_T("\n\r"));
}
