/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Says how much memory is available.
 * Same as hp_state.c but compiled as C++.
 */

/* Call v-tables the C way */
#define CINTERFACE
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

int _tmain(int argc, _TCHAR **argv)
{
    struct MACHINE_INFO mi;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();
    PIHEAP pHeap = CurrentHeap();
    ADDR_SIZE Reserve,Commit,Used,MaxUsed;

    UnusedParameter(argc);
    UnusedParameter(argv);

    /* See how much free memory we have
     */
    sc = pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi );
    if (FAILED(sc)) {
        _tprintf(_TEXT("GENERIC machine info FAILED (x%x)"), sc);
        exit(EXIT_FAILURE);
    }

    _tprintf(_TEXT("Memory: %d bytes free out of %d total.\n\r"),
             mi.AvailablePhysicalMemory, mi.TotalPhysicalMemory);

    /* The heap now
     */
    Reserve = Commit = Used = MaxUsed = 0;
    sc = pHeap->v->Status(pHeap,&Reserve,&Commit,&Used,&MaxUsed);
    if (FAILED(sc)) {
        _tprintf(_TEXT("IHeap::Status FAILED (x%x) hp=%x"),sc,(ADDRESS)pHeap);
        exit(EXIT_FAILURE);
    }

    _tprintf(_TEXT("Heap: %d reserve %d commit %d used\n\r"), 
             Reserve, Commit, Used);
    _tprintf(_TEXT("Maximum heap memory usage was %d bytes\n\r"), MaxUsed);


    /* Done
     */
    exit(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}
