/* Copyright (c) Microsoft Corporation. All rights reserved. */
#undef PrintTime
#define PrintTime(_desc_,_time_) \
    { \
        _TCHAR buf[50]; \
        TimeFormat(_time_, buf, sizeof(buf)-1); \
        printf(_desc_); \
        printf(": %s\n", buf); \
    }

PRIVATE INT8 days[2][12] = {{31,28,31,30,31,30, 31,31,30,31,30,31},
                            {31,29,31,30,31,30, 31,31,30,31,30,31}};
PRIVATE char *DayName[7] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

PRIVATE BOOL IsLeap(UINT y)
{
    BOOL is = (!(y%4) && ((y%100) || !(y%400)));
    //DBG(("IsLeap(%d) => %d\n", y, is));
    return is;
}

void TimeFormat(TIME t, _TCHAR *buf, UINT Len)
{
    TIME t2;
    INT leaps, f,s,min,h,d,mo,y, wd, xxx;
    BOOL relative = TimeIsRelative(t) || Int64Equal(t, TIME_ORIGIN);

    if (relative)
        t = Int64Subtract(TIME_ORIGIN, t);

    /* First deal with fractions */
    t2 = Int64ModulusInt32(t,  10000000);
    t = Int64DividedByInt32(t, 10000000);
    f = Int64ToInt32(t2);

    if (relative) {
        _stprintf(buf, _T("PT%lld.%07dS"), t,f);
        return;
    }

    /* seconds */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    s = Int64ToInt32(t2);

    /* minutes */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    min = Int64ToInt32(t2);

    /* hours */
    t2 = Int64ModulusInt32(t, 24);
    t = Int64DividedByInt32(t, 24);
    h = Int64ToInt32(t2);
    
#if !defined(arm)               /* arm_arm generates call to __rt_divtest */
    /* days since Jan 1, 1601, Gregorian calendar. */
    xxx = wd = Int64ToInt32(t);
    wd %= 7;
    wd++;
#endif

    /* figure out rough year, then adjust with leap days */
    t2 = Int64ModulusInt32(t, 365);
    t = Int64DividedByInt32(t, 365);
    d = Int64ToInt32(t2);

    y = Int64ToInt32(t);        /* from 1601 */

    leaps = y / 4 - y / 100 + y / 400; /* from 1600, does not include current year */

    d -= leaps;

    //printf("y=%d leaps=%d now=%d d=%d wd=%d [%s] totaldays=%d\n",
             //y, leaps, (y%4 == 0), d, wd, DayName[wd-1], xxx);

    while (d < 0) {
      /* if we are millenia in the future there might be more bogus leaps
       * than one year worth */
        leaps = IsLeap(y+1601-1);       /* last year */
        d += (leaps ? 366 : 365);
        y--;
    }
    leaps = IsLeap(y+1601);       /* this year */

    for (mo = 0; mo < 12; ) {
        if (d < days[leaps][mo])
            break;
        d -= days[leaps][mo];
        mo++;
    }
    /* make them one-based */
    mo++;
    d++;
    y += 1601;

    /* Time zone: UTC (boot time?) */
    _stprintf(buf,_T("%04d-%02d-%02dT%02d:%02d:%02d.%07dZ"), y,mo,d,h,min,s,f);
}

