/*
* Copyright (c) Microsoft All Rights Reserved
* Tests rand() and srand() by using Monte Carlo method to find PI. 
* Returns PI as an integer(!) becuz floats not supported on all platforms
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"

void PIStatus(const _TCHAR * s, UINT nIter, UINT nRadius);
void PIUsage(const _TCHAR * s);

int _tmain(int argc, _TCHAR **argv)
{
    UINT x, y, z; 
    UINT iterations = 100, i; 
    UINT radius = 1024;  /* must be a power of 2 cuz RAND_MAX=2^16 */
    UINT hits = 0, misses = 0; 
    const _TCHAR * sArg = 0; 
    _TCHAR * stop = 0;
    UINT32 pi; 
    UINT64 n64, d64;
    
    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    if(argc < 3)
    {
        PIUsage(argv[0]); 
        _tprintf(_T("\n\rUsing defaults .... ")); 
    }

    else
    {
        sArg = GetArgFromCommandLine(argc, argv, _T("iter")); 
        if(sArg && *sArg)
            iterations = _tcstoul(sArg, &stop, 10); 

        sArg = GetArgFromCommandLine(argc, argv, _T("radius")); 
        if(sArg && *sArg)
            radius = _tcstoul(sArg, &stop, 10); 
    }

    if ((radius > RAND_MAX) ||
        (((RAND_MAX+1) % radius) != 0)) {
        _tprintf(_T("\n\rBad radius %d, should be a divisor of %d (trying anyways)\r\n"),
                 radius, RAND_MAX);
    }

    PIStatus(argv[0], iterations, radius); 

    
    srand(Uint64Low32Bits(CurrentTime()));
    
    for(i = 0, z = radius * radius; i < iterations; ++i)
    {
        x = rand()%radius; x *= x;
        y = rand()%radius; y *= y; 
        if(x + y <= z)
            ++hits; 
        else
            ++misses;
    }

    /* Using Ac = Area of circle = PI * r^2, z = r^2, 
     *       As = Area of square = (2 * r)^2 = 4*r^2
     *       Ac/As = pi/4 --> pi = 4 * Ac/As
     */
    /* Can't use floating point math in ARM so we multiply by large integer
     * and move the decimal point manually
     * Should not overflow, so 64 bit math.
     */
    if(iterations && (hits + misses))
    {
        Uint32ToUint64(n64,hits);
        n64 = Uint64TimesUint32(n64,4 * 10000);

        Uint32ToUint64(d64,hits+misses);

        n64 = Uint64DividedByUint64(n64,d64);

        pi = Uint64ToUint32(n64);

        _tprintf(_T("\n\rPI is [%d.%d] after [%d] hits in [%d] attempts"), 
                 pi / 10000, pi % 10000,
                 hits, hits + misses); 
    }
    else
        _tprintf(_T("\n\rZero attempts means no data ... "));

    _tprintf(_T("\n\r"));

    FreeCmdLine(&argc, &argv); 
    return SetAndReturn(0); 
}

void PIUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:%s -iter <iterations> -radius <radius>, "), s);
    _tprintf(_T("\n\r\twhere <iterations> is number of darts and <radius> is radius of the circle")); 
}

void PIStatus(const _TCHAR * s, UINT nIter, UINT nRadius)
{
    _tprintf(_T("\n\r%s status:iterations [%d], radius[%d]"), s, nIter, nRadius);
}








