/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define DEVICE_ORDER LITTLE_ENDIAN
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <mips/ml40x.h>

#if 0
#define VERBOSE(x) _tprintf x
#else
#define VERBOSE(x)
#endif

#define DEFAULT_SYSACE_ADDRESS IDE_DEFAULT_ADDRESS

void Usage(_TCHAR *MyName)
{
    _tprintf(_T("Usage: %s [-base_address_in_hex] <command>\n"), MyName);
    _tprintf(_T("<command> is one of the following:\n"));
    _tprintf(_T("  dump\t\tShows the register state of the SystemACE\n"));
    _tprintf(_T("  config<N>\tAsks SystemACE to load the configuration file no. N [0-7].\n"));
    _tprintf(_T("  ace <name>\tBlasts the .ACE config file <name> to the SystemACE.\n"));
    _tprintf(_T("The default base address is %8x.\n"), DEFAULT_SYSACE_ADDRESS);
}

typedef _TCHAR *NAME;
typedef struct _REGDESC {
    NAME RegisterName;
    NAME BitNames[32];
} REGDESC, *PREGDESC;

void SysacePrintRegister(const REGDESC *Desc, UINT32 Value)
{
    int i;
    _tprintf(_T("\t%s %x ="), Desc->RegisterName, Value);
    for (i = 31; i >= 0; i--) {
        if (Value & (1 << i))
            _tprintf(_T(" %s"), 
                     (Desc->BitNames[i]) ? Desc->BitNames[i] : _T("?"));
    }
    _tprintf(_T("\n"));            
}

int SysaceDump(struct _Sac *Interface)
{
    int r = EXIT_SUCCESS;

	const REGDESC Control_Names =
        { _T("Control"),
          {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          _T("RST"), //			0x00010000
          _T("BUS8"), //		0x00020000
          _T("BUS16"), //		0x00040000
          _T("BUS32"), //		0x00080000
          _T("IRQ"), //			0x00100000
          _T("BRDY"), //		0x00200000
          _T("IMSK0"), //		0x00400000
          _T("IMSK1"), //		0x00800000
          _T("TD0"), //	    	0x0f000000
          _T("TD1"), //		    0x0f000000
          _T("TD2"), //	    	0x0f000000
          _T("TD3"), //	    	0x0f000000
          _T("BUFW8"), //		0x10000000
          _T("BUFW16"), //		0x20000000
          _T("BUFW32"), //		0x40000000
          _T("DEBUG")} //		0x80000000
        };

	const REGDESC STATUS_Names =
        { _T("STATUS"),
          {_T("CFGLOCK"), //	0x00000001
          _T("MPULOCK"), //		0x00000002
          _T("CFGERROR"), //	0x00000004
          _T("CFCERROR"), //	0x00000008
          _T("CFDETECT"), //	0x00000010
          _T("DATABUFRDY"), //	0x00000020
          _T("DATABUFWRITE"), //0x00000040
          _T("CFGDONE"), //		0x00000080
          _T("RDYFORCFCMD"), //	0x00000100
          _T("CFGMODEPIN"), //	0x00000200
          0,0,0,
          _T("CFGADDRPIN0"), //	0x0000e000
          _T("CFGADDRPIN1"), //	0x0000e000
          _T("CFGADDRPIN2"), //	0x0000e000
          0,
          _T("CFBSY"), //		0x00020000
          _T("CFRDY"), //		0x00040000
          _T("CFDWF"), //		0x00080000
          _T("CFDSC"), //		0x00100000
          _T("CFDRQ"), //		0x00200000
          _T("CFCORR"), //		0x00400000
          _T("CFERR"), //		0x00800000
          0,}
        };

	const REGDESC ERRORREG_Names =
        { _T("ERRORREG"),
          {_T("CARDRESETERR"), //			0x00000001
          _T("CARDRDYERR"), //				0x00000002
          _T("CARDREADERR"), //				0x00000004
          _T("CARDWRITEERR"), //			0x00000008
          _T("SECTORRDYERR"), //			0x00000010
          _T("CFGADDRERR"), //				0x00000020
          _T("CFGFAILED"), //				0x00000040
          _T("CFGREADERR"), //				0x00000080
          _T("CFGINSTRERR"), //				0x00000100
          _T("CFGINITERR"), //				0x00000200
          0,
          _T("CFBBK"), //					0x00000800
          _T("CFUNC"), //					0x00001000
          _T("CFIDNF"), //					0x00002000
          _T("CFABORT"), //					0x00004000
          _T("CFAMNF"), //					0x00008000
           0,}
        };

    const NAME CommandNames[8] =
        { _T("0"), //     				0x0000
          _T("RESETMEMCARD"), //   		0x0100
          _T("IDENTIFYMEMCARD"), //		0x0200
          _T("READMEMCARDDATA"), //		0x0300
          _T("WRITEMEMCARDDATA"), //  	0x0400
          _T("5"), //     				0x0500
          _T("ABORT"), //			    0x0600
          _T("7") //     			    0x0700
        };

	const REGDESC CONTROLREG_Names =
        { _T("CONTROLREG"),
          {_T("FORCELOCKREQ"), //			0x00000001
          _T("LOCKREQ"), //					0x00000002
          _T("FORCECFGADDR"), //			0x00000004
          _T("FORCECFGMODE"), //			0x00000008
          _T("CFGMODE"), //					0x00000010
          _T("CFGSTART"), //				0x00000020
          _T("CFGSEL_MPU"), //  			0x00000040
          _T("CFGRESET"), //				0x00000080
          _T("DATABUFRDYIRQ"), //			0x00000100
          _T("ERRORIRQ"), //				0x00000200
          _T("CFGDONEIRQ"), //				0x00000400
          _T("RESETIRQ"), //				0x00000800
          _T("CFGPROG"), //					0x00001000
          _T("CFGADDR_B0"), //				0x00002000
          _T("CFGADDR_B1"), //				0x00004000
          _T("CFGADDR_B2"), //				0x00008000
          0,}
        };

	const REGDESC FATSTATREG_Names =
        { _T("FATSTATREG"),
          {_T("MBRVALID"), //				0x00000001
          _T("PBRVALID"), //				0x00000002
          _T("MBRFAT12"), //				0x00000004
          _T("PBRFAT12"), //				0x00000008
          _T("MBRFAT16"), //				0x00000010
          _T("PBRFAT16"), //				0x00000020
          _T("CALCFAT12"), //				0x00000040
          _T("CALCFAT16"), //				0x00000080
          0, }
        };

    _tprintf(_T("Dump of Sysace@%x:\n"), (ADDRESS)Interface);
    _tprintf(_T("\tTag %x\n"), Interface->Tag);
    if (Interface->Tag != PMTTAG_SYSTEM_ACE) {
        _tprintf(_T("This is NOT the correct tag! Should be x%x\n"), PMTTAG_SYSTEM_ACE);
        r = EXIT_FAILURE;
    }
    SysacePrintRegister(&Control_Names, Interface->Control);
    _tprintf(_T("\tBUSMODEREG %x\n"), Interface->BUSMODEREG);
    SysacePrintRegister(&STATUS_Names, Interface->STATUS);
    SysacePrintRegister(&ERRORREG_Names, Interface->ERRORREG);
    _tprintf(_T("\tCFGLBAREG %x\n"), Interface->CFGLBAREG);
    _tprintf(_T("\tMPULBAREG %x\n"), Interface->MPULBAREG);
    _tprintf(_T("\tVERSIONREG %x\n"), Interface->VERSIONREG);
    _tprintf(_T("\tSECCNTCMDREG %x = %s cnt=%d\n"), Interface->SECCNTCMDREG,
             CommandNames[(Interface->SECCNTCMDREG >> 8)&7],
             Interface->SECCNTCMDREG & SAC_SECCCNT);
    SysacePrintRegister(&CONTROLREG_Names, Interface->CONTROLREG);
    SysacePrintRegister(&FATSTATREG_Names, Interface->FATSTATREG);

    return (r);
}

int SysaceConfig(struct _Sac *Interface, int FileNumber)
{

    _tprintf(_T("Asking Sysace@%x to load configuration bitfile %d from CFLASH\n"),
             (ADDRESS)Interface, FileNumber);

    if (Interface->Tag != PMTTAG_SYSTEM_ACE) {
        _tprintf(_T("Wrong Tag %x. This is NOT a system ACE, the correct tag is x%x\n"), 
                 Interface->Tag, PMTTAG_SYSTEM_ACE);
        return (EXIT_FAILURE);
    }

    if ((FileNumber < 0) || (FileNumber > 7)) {
        _tprintf(_T("Invalid bitfile number %d, should be between 0 and 7 inclusive.\n"), 
                 FileNumber);
        return (EXIT_FAILURE);
    }

    /* Ok, let's do it
     */
    Interface->CONTROLREG = SAC_FORCELOCKREQ | SAC_LOCKREQ |
            SAC_FORCECFGADDR | SAC_FORCECFGMODE | 
            SAC_CFGMODE | SAC_CFGSTART | SAC_CFGRESET |
            (FileNumber << 13);

    Interface->CONTROLREG = 
            SAC_FORCECFGADDR | SAC_FORCECFGMODE | 
            SAC_CFGMODE | SAC_CFGSTART |
            (FileNumber << 13);

    /* We might have disappeared, but if we are still around...
     */
    return EXIT_SUCCESS;
}

/* Utility for below
 */
BOOL ReadConfigData(_TCHAR *FileName,
                    UINT **pData, UINT *pnBytes)
{
    FILE *File;
    size_t Length, n;
    BOOL r = FALSE;
    UINT *Data = NULL;

    /* Sanity */
    *pData = NULL;
    *pnBytes = 0;

    /* Does it exist */
    File = _tfopen(FileName, _T("rb"));
    if (File == NULL) {
        _tprintf(_T("fopen(%s) failed.\n"), FileName);
        return FALSE;
    }

    /* How big is it */
    Length = fseek(File,0,SEEK_END);
    fseek(File,0,SEEK_SET);

    if (Length == 0) {
        _tprintf(_T("file %s too short (0 bytes?).\n"), FileName);
        goto Out;
    }

    /* Make room for the file in memory.
     * NB: The system ACE chip wants to be fed 32-byte full buffers,
     * but the ACE file is not rounded up correctly. A simple fix is
     * to over-allocate memory and leave the rest as zeroes.
     * NB: Testing shows its ok to fill the last buffer with zeroes.
     */
    n = Length;
#define ACEROUNDUP 32
#ifdef ACEROUNDUP
    n = (Length + ACEROUNDUP - 1) & ~(ACEROUNDUP-1);
#endif
    Data = malloc(n);
    if (Data == NULL) {
        _tprintf(_T("Outof memory for %s (%d bytes).\n"), FileName, n);
        goto Out;
    }
    memset(Data,0,n);

    /* One big read will do */
    _tprintf(_T("Reading file %s (%d bytes)...\n"), FileName, Length);
    n = fread(Data,Length,1,File);
    if (n != 1) {
        _tprintf(_T("Short read for %s (fread got %d).\n"), FileName, n);
        free(Data);
        goto Out;
    }

    /* All is well
     */
    *pData = Data;
    *pnBytes = Length;
    r = TRUE;
 Out:
    /* Done with file */
    fclose(File);
    return r;
}

/* By trial-and-error we figured out that the SystemACE chip skips
 * the first 512 bytes of the ACE file, which is this ASCII header:
 */
char AceHeader[] = 
"0000000000000000000000000000000000000000000000000000000000000000\n"
"Automatically generated. DO NOT MODIFY\n"
"Xilinx, Inc.\n"
"xmpmutil(tm) Multi-Function MPM Utility Version 1.01 (c) 2001 Xilinx, Inc.\n";

/* ...and spaces to fill the rest.
 * When fed by the MPU the chip doesnt want to see this header.
 */
#define ACEFILEOFFSET 512

BOOL SysaceSendConfig(struct _Sac *Interface, UINT32 *Data, UINT nBytes)
{
    UINT i, j, nWords;
    UINT32 CtlWas = Interface->CONTROLREG;
    UINT32 Status;

#ifdef ACEFILEOFFSET
    if (nBytes < ACEFILEOFFSET) {
        _tprintf(_T("File too small (<%d)\n"), ACEFILEOFFSET);
        return FALSE;
    }
    i = memcmp(Data,AceHeader,(sizeof AceHeader)-1);
    if (i != 0) {
        _tprintf(_T("Invalid ACE file header(%d,%d)? Continuing anyways.\n"),
                 i, sizeof AceHeader);
    }
    Data = (UINT*) ((char *)Data + ACEFILEOFFSET);
    nBytes -= ACEFILEOFFSET;
#endif

    VERBOSE((_T(""),SysaceDump(Interface)));

    /* Set the bits but in RESET (pag 49-50 of specs)*/
#define CFGCMD (SAC_FORCELOCKREQ | SAC_LOCKREQ | SAC_CFGSEL | \
                SAC_FORCECFGMODE |/* SAC_CFGMODE |*/ SAC_CFGSTART)

    Interface->CONTROLREG = CFGCMD | SAC_CFGRESET;

    /* Take it out of RESET */
    Interface->CONTROLREG = CFGCMD;

    /* Get the data out to the ACE
     */
#ifdef ACEROUNDUP
    nBytes = (nBytes + ACEROUNDUP - 1) & ~(ACEROUNDUP-1);
#endif
    nWords = (nBytes + 3) / 4;
    _tprintf(_T("Sending bits (as %d words) "), nWords);
    for (i = 0; i < nWords; i += 1/*word*/) {

        /* Stop on errors */
        Status = Interface->ERRORREG;
        if (Status) {
            /* Ooops, get out of here
             */
            _tprintf(_T("Sysace::CFG error (i=%d)\n"), i);
            goto Error;
        }

        /* Make sure the FIFO is ready
         */
        for (j = 0; j < 10; j++) {
            Status = Interface->STATUS;
            if (Status & SAC_DATABUFRDY)
                break;
            Delay(1000);
        }

        /* Got it?
         */
        if (Status & SAC_DATABUFRDY) {
            UINT32 Data32;

            Data32 = Data[i];
            VERBOSE((_T("%x "), Data32));
            Data32 = _bs32(Data32);
            Interface->DATABUFREG[0] = Data32;
        }
        else
        {
            /* Ooops, get out of here
             */
            _tprintf(_T("Sysace::WRITE timeout (i=%d)\n"), i);
            goto Error;
        }
    }
    _tprintf(_T("done ok.\n"));
    VERBOSE((_T(""),SysaceDump(Interface)));

    /* Put it back the way it was (try to.. :-( )*/
    Interface->CONTROLREG = CtlWas;
    return TRUE;

 Error:
    SysaceDump(Interface);
    return (FALSE);
}

/* Meaning, treat the SysACE like an iCAP interface
 */
int SysaceIcap(struct _Sac *Interface, _TCHAR *FileName)
{
    UINT *ConfigData = NULL;
    UINT ConfigBytes = 0;
    int r = EXIT_FAILURE;

    /* Say what, check */
    _tprintf(_T("Sending configuration file %s to Sysace@%x\n"),
             FileName, (ADDRESS)Interface);

    if (Interface->Tag != PMTTAG_SYSTEM_ACE) {
        _tprintf(_T("Wrong Tag %x. This is NOT a system ACE, the correct tag is x%x\n"), 
                 Interface->Tag, PMTTAG_SYSTEM_ACE);
        return (EXIT_FAILURE);
    }

    /* Read file in */
    if (!ReadConfigData(FileName, &ConfigData, &ConfigBytes) ||
        (ConfigData == NULL)) {
        _tprintf(_T("Could not load %s in memory.\n"), FileName);
        return (EXIT_FAILURE);
    }

    /* Send it to the SysACE */
    if (SysaceSendConfig(Interface,ConfigData,ConfigBytes)) {
        _tprintf(_T("File %s configed ok.\n"), FileName);
        r = EXIT_SUCCESS;
    } else {
        _tprintf(_T("Could not send config %s to Sysace@%x.\n"), FileName, (ADDRESS)Interface);
        r = EXIT_FAILURE;
    }

    /* We are done. or Not!
     */
    free(ConfigData);
    return r;
}

int _tmain(INT argc, _TCHAR **argv)
{
    struct _Sac *Interface = (struct _Sac *)DEFAULT_SYSACE_ADDRESS;
    int r = EXIT_FAILURE;
    SCODE sc;
    INT Argc; _TCHAR **Argv;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("No command line args (%x), test FAILED\n"), sc);
        return (EXIT_FAILURE);
    }
    Argc = argc;
    Argv = argv;

 ParseArgs:
    if (Argc < 2) {
        goto Done;
    }

    if (Argv[1][0] == 0) {
        ADDRESS Base = _tcstoul(Argv[1]+1,NULL,16);
        Interface = (struct _Sac *)Base;
        Argc--, Argv++;
        goto ParseArgs;
    }

    /* See what we should do
     */
    if (_tcscmp(Argv[1],_T("dump")) == 0) {
        r = SysaceDump(Interface);
    } else
    if (_tcsncmp(Argv[1],_T("config"), 6) == 0) {
        int nConfig = _ttoi(Argv[1]+6);
        r = SysaceConfig(Interface, nConfig);
    } else
    if (_tcsncmp(Argv[1],_T("ace"), 3) == 0) {
        _TCHAR *FileName = Argv[2];
        if ((Argc < 3) || (FileName == NULL))
            goto Done;
        r = SysaceIcap(Interface, FileName);
    } else {
        goto Done;
    }

 Done:

    if (r == EXIT_SUCCESS)
        _tprintf(_T("Test PASSED ok.\n"));
    else {
        _tprintf(_T("Test FAILED.\n"));
        Usage(argv[0]);
    }

    (void) FreeCmdLine(&argc,&argv);
    return r;
}
