/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Tracks the mouse, stops when all 3 buttons are pressed at once.
 */
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <drivers\mouse_hw.h>
#include <framebuffer.h>

UINT8 *FB = NULL;
UINT32 fW, fWv, fHv;
BOOL fBwFlipped = FALSE;

#define cW 8
#define cH 12
#define O 255
#define B 0

UINT8 PointerBitmap[cW*cH] = {
    B, O, O, O, O, O, O, O,
    B, B, O, O, O, O, O, O,
    B, B, B, O, O, O, O, O,
    B, B, B, B, O, O, O, O,
    B, B, B, B, B, O, O, O,
    B, B, B, B, B, B, O, O,
    B, B, B, B, O, O, O, O,
    B, O, B, B, O, O, O, O,
    O, O, O, B, O, O, O, O,
    O, O, O, B, B, O, O, O,
    O, O, O, B, B, O, O, O,
    O, O, O, O, O, O, O, O
};

/* Obvious way
 */
UINT8 Bg[cW*cH];

void getbg0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            Bg[x + (y * cW)] = FB[(x + x0) + ((y + y0) * fW)];;
}

void setbg0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] = Bg[x + (y * cW)];
}

void paintc0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] = PointerBitmap[x + (y * cW)];
}

/* Using XOR
 */
void getbg1(int x0, int y0) {return;}
#define setbg1 paintc1

void paintc1(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] ^= ~PointerBitmap[x + (y * cW)];
}

void paintc1f(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] ^= PointerBitmap[x + (y * cW)];
}

/* Non-transparent overlay
 */
#define getbg2 getbg0
#define setbg2 setbg0

void paintc2(int x0, int y0)
{
    int x,y;
    UINT8 c;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++) {
            c = PointerBitmap[x + (y * cW)];
            if (c == B)
                FB[(x + x0) + ((y + y0) * fW)] = c;
        }
}

void paintc2f(int x0, int y0)
{
    int x,y;
    UINT8 c;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++) {
            c = PointerBitmap[x + (y * cW)];
            if (c == O)
                FB[(x + x0) + ((y + y0) * fW)] = c;
        }
}



/* Select which is which
 */
typedef struct {
    void (*GetBackground)(int x, int y);
    void (*SetBackground)(int x, int y);
    void (*PaintPointer)(int x, int y);
} PAINTER;

#define nPAINTERS 3
PAINTER Pointers[nPAINTERS] = {
    { getbg0, setbg0, paintc0 },
    { getbg1, setbg1, paintc1 },
    { getbg2, setbg2, paintc2 },
};

PAINTER *Pointer = &Pointers[2];

void FlipBlackAndWhite(void)
{
    int i;
    for (i = 0; i < cW*cH; i++)
        PointerBitmap[i] = (PointerBitmap[i] == B) ? O : B;
    Pointers[1].SetBackground = paintc1f;
    Pointers[1].PaintPointer  = paintc1f;
    Pointers[2].PaintPointer  = paintc2f;
}

/* Framebuffer
 */
SCODE InitializeFrameBuffer(_TCHAR* FbName)
{
    PIFRAMEBUFFER FrameBuffer;
    FRAMEBUFFER_MODE Mode;
    PIXEL_ENCODING Encoding;
    SCODE sc;
    UINT32 Length;
    _TCHAR *Msg;

    Msg = _T("bind to");
    sc = BindToObject(CurrentNameSpace(),
                      FbName,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IFrameBuffer,
                      (void *) &FrameBuffer);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get mode");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_MODE,&Mode);
    if (FAILED(sc)) {
        Msg = _T("set mode");
        sc = FrameBuffer->v->SetMode(FrameBuffer,FB_MODE_640x480_in_768x480_8bpp_PALETTE);
        if (FAILED(sc)) goto Bad;
    }

    FB = NULL;
    Length = 0;
    Msg = _T("map");
    sc = FrameBuffer->v->Map(FrameBuffer,(PTR*)&FB,&Length);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hlen");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_LENGTH,&fW);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_VISIBLE_LENGTH,&fWv);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Vvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_VERTICAL_VISIBLE_LENGTH,&fHv);
    if (FAILED(sc)) goto Bad;

    /* Must check what is white and what is black
     */
    Msg = _T("get Penc");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_PALETTE_ENCODING,&Encoding);
    if (FAILED(sc)) goto Bad;

    if ((Encoding == PIXEL_TYPE_R8G8B8) ||
        (Encoding == PIXEL_TYPE_R8G8B8X8)) {
        UINT8 r8g8b8x8[4];

        Msg = _T("get Pix0");
        sc = FrameBuffer->v->GetPaletteEntry(FrameBuffer,
                                             B,
                                             r8g8b8x8,
                                             sizeof r8g8b8x8,
                                             Encoding);
        if (FAILED(sc)) goto Bad;

        /* We expect 'B' (palette entry #0) to be for Black */
        if ((r8g8b8x8[0] >= 0x0f) &&
            (r8g8b8x8[1] >= 0x0f) &&
            (r8g8b8x8[2] >= 0x0f)) {
            _tprintf(_T("Black&White palette entries are flipped [%x,%x,%x]\n"),
                     r8g8b8x8[0],r8g8b8x8[1],r8g8b8x8[2]);
            fBwFlipped = TRUE;
        }
    }
    /* otherwise we dont know */

    /* BUGBUG strictly speaking we cant do this
     */
    FrameBuffer->v->Release(FrameBuffer);

    return S_OK;
 Bad:
    _tprintf(_T("Could not %s %s sc=%x\n"), Msg, FbName, sc);
    return sc;
}

/* Mouse interface
 */

BOOL RawUsbFormat = FALSE;
typedef struct _USBREPORT {
    UINT8 Buttons;
    INT8 x;
    INT8 y;
    INT8 Wheel;
} USBREPORT, *PUSBREPORT;


SCODE GetReport(PIFILE Mouse,
                MOUSE_INPUT_DATA *Report)
{
    SCODE sc;
    UINT nBytes;
    USBREPORT U;

    nBytes = 0;
    if (RawUsbFormat) {
        sc = Mouse->v->ReadAt(Mouse,
                              UINT64_ZERO,
                              (BYTE*)&U,
                              sizeof U,
                              &nBytes);
        if (FAILED(sc))
            return sc;
        if (nBytes != sizeof U)
            return E_FAIL;
#if _DEBUG
        _tprintf(_T("\rU: %x %d %d     "), U.Buttons, U.x, U.y);
#endif
        Report->LastX += U.x;
        Report->LastY += U.y;
        Report->LastZ = 0;
        Report->RawButtons = U.Buttons;
        Report->ReportTime = CurrentTime();
        return S_OK;
    }

    sc = Mouse->v->ReadAt(Mouse,
                          UINT64_ZERO,
                          (BYTE*)Report,
                          sizeof *Report,
                          &nBytes);
#if _DEBUG
    _tprintf(_T("\rU: %x %d %d     "), 
             Report->RawButtons, Report->LastX, Report->LastY);
#endif
    if (FAILED(sc))
        return sc;
    if (nBytes != sizeof *Report)
        return E_FAIL;
    return S_OK;
}

void THREAD_LINKAGE MouseThread(THREAD_ARGUMENT arg)
{
    PIFILE Mouse = (PIFILE )arg;
    MOUSE_INPUT_DATA R;
    SCODE sc;
    int x, x1, y, y1;

    x = fWv / 2;
    y = fHv / 2;
    Pointer->GetBackground(x,y);
    Pointer->PaintPointer(x,y);

    R.LastX = x;
    R.LastY = y;
    for (;;) {

        /* Wait for report
         */
        sc = GetReport(Mouse,&R);
        if (FAILED(sc))
            break;

        /* 3-button-press is termination
         */
        if (R.RawButtons == 
            (MOUSE_BUTTON_1_DOWN|MOUSE_BUTTON_2_DOWN|MOUSE_BUTTON_3_DOWN))
            break;

        /* Tentatively update pointer position
         */
        x1 = R.LastX;
        y1 = R.LastY;

        /* Box it
         */
        if ((x1 < 0) || (x1 > (int)fWv))
            x1 = x;
        if ((y1 < 0) || (y1 > (int)fHv))
            y1 = y;

        /* Valid move?
         */
        if ((x != x1) || (y != y1)) {
            /* Yes, restore previous background
             */
            Pointer->SetBackground(x,y);
            /* Preserve new bg
             */
            Pointer->GetBackground(x1,y1);
            /* Set pointer at new position
             */
            Pointer->PaintPointer(x1,y1);
            /* Move
             */
            x = x1;
            y = y1;
        }
#if _DEBUG
        else 
            _tprintf(_T("buttons %x position (%d,%d,%d)      \r"),
                     R.RawButtons,
                     R.LastX, R.LastY, R.LastZ);
#endif
    }
    Pointer->SetBackground(x,y);

    _tprintf(_T("MouseThread done.\n"));
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    INT RetVal = EXIT_FAILURE;
    _TCHAR *MouseName = _T("mouse");
    _TCHAR *FbName = _T("display");
    PIFILE Mouse = NULL;
    int c;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("No command line args (%x), test FAILED\n"), sc);
        exit(RetVal);
    }

    if (argc > 1) {
        MouseName = argv[1];
    }

    if (argc > 2) {
        c = _ttoi(argv[2]);
        if ((c >= 0) && (c < nPAINTERS))
            Pointer = &Pointers[c];
    }

    if (argc > 3) {
        FbName = argv[3];
    }

    if (argc > 4) {
        RawUsbFormat = TRUE;
    }

    sc = BindToObject(CurrentNameSpace(), MouseName,
                      NAME_SPACE_READ,
                      &IID_IFile,
                      (void *)&Mouse);
    if (FAILED(sc)) {
        _tprintf(_T("Could not open mouse file %s\n"), MouseName);
        goto Done;
    }

    sc = InitializeFrameBuffer(FbName);
    if (FAILED(sc)) {
        _tprintf(_T("Could not init framebuffer %s\n"), FbName);
        goto Done;
    }

    if (fBwFlipped)
        FlipBlackAndWhite();

    _tprintf(_T("Tracking %s. Press all 3 buttons to terminate\n"),
             MouseName);

    MouseThread((THREAD_ARGUMENT)Mouse);

    RetVal = EXIT_SUCCESS;
Done:
    if (Mouse)
        Mouse->v->Release(Mouse);

    (void) FreeCmdLine(&argc,&argv);
    exit(RetVal);
    return 0;
}

