/*
* FN.C - FindNext api test suite. 
* Takes parameters for function: 
*   - name of "prefix" (quarry) object. Can be NULL or empty
*   - name of "previous" object. Also can be NULL or empty
*   - user can specify NULL or good pointer to UINT to return number of bytes copied. 
*   - user can specify size of output buffer up to a limit.
* Copyright (c) Microsoft Corporation All rights reserved
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"

#define NAMELEN (512)
#ifdef UNICODE
    const   _TCHAR COOKIE       = (_TCHAR)(0xefef);
#else
    const   _TCHAR COOKIE       = (_TCHAR)(0xef);
#endif

void NSStatus(const _TCHAR * sProg, const _TCHAR * sPrev, const _TCHAR * sPref, UINT nBuflen, const UINT * pStrLenOut); 
void FNUsage(const _TCHAR *sProg);

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    const _TCHAR * sArg = 0; 
    _TCHAR * stop = 0;
    _TCHAR sPrefix[NAMELEN] = {0,}, sPrevious[NAMELEN] = {0,}; 
    _TCHAR * sPref = 0, * sPrev = 0;
    _TCHAR * sOutbuf = 0;
    UINT * pStrLen = 0; 
    UINT nBufLenOut = 0, nBufLenIn = 0;
    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak(); 

    if(argc < 2)
        FNUsage(argv[0]);

    //user can specify null or empty or 'real' arg for previous ...
    sArg = GetArgFromCommandLine(argc, argv, _T("prev"));
    if(sArg && *sArg)
        if(!_tcsicmp(sArg, _T("null")))
            sPrev = 0; 
        else
            if(!_tcsicmp(sArg, _T("empty")))
            {
                sPrev = sPrevious; 
                *sPrev = 0; 
            }
            else
            {
                _tcscpy(sPrevious, sArg); 
                sPrev = sPrevious; 
            }
    else
        sPrev = 0; 

    //user can specify null or empty or 'real' arg for prefix ...
    sArg = GetArgFromCommandLine(argc, argv, _T("prefix"));
    if(sArg && *sArg)
        if(!_tcsicmp(sArg, _T("null")))
            sPref = 0; 
        else
            if(!_tcsicmp(sArg, _T("empty")))
            {
                sPref = sPrefix; 
                *sPref = 0; 
            }
            else
            {
                _tcscpy(sPrefix, sArg); 
                sPref = sPrefix; 
            }
    else
        sPref = 0; 

    //user can specify null or empty or 'real' string length return buffer...
    pStrLen = &nBufLenOut;
    sArg = GetArgFromCommandLine(argc, argv, _T("strlen"));
    if(sArg && *sArg)
        if(!_tcsicmp(sArg, _T("null")))
            pStrLen = 0; 
        

    //user can specify size of return buffer, defaults to nice size
    sArg = GetArgFromCommandLine(argc, argv, _T("bufsize"));
    if(sArg && *sArg)
        nBufLenIn = _tcstoul(sArg, &stop, 10);
    else
        nBufLenIn = 256;

    //dynamic allocation of output buffer 'cuz it can change size 
    sOutbuf = CurrentHeap()->v->Alloc(CurrentHeap(), HEAP_ZERO_MEMORY, 2 * nBufLenIn * sizeof(_TCHAR), 0);
    if (!sOutbuf) {
        _tprintf(_T("\n\r\a%s:ERROR:Failed to allocate %d bytes of memory!"),
                      argv[0],2 * nBufLenIn * sizeof(_TCHAR)); 
        return SetAndReturn(1);
    }

    NSStatus(argv[0], sPref, sPrev, nBufLenIn, pStrLen); 
    
    //the 'real' work .. 
    sc = CurrentNameSpace()->v->FindNext(CurrentNameSpace(), sPref, sPrev, sOutbuf, nBufLenIn, pStrLen); 
    
    NSOutcome(FINDNEXT, sc, sPref, sPrev, sOutbuf, nBufLenIn, pStrLen ? *pStrLen : 0);

    if(pStrLen)
    {
        if(*pStrLen > nBufLenIn)
        {
            _tprintf(_T("\n\r\a%s:ERROR:Bytes returned exceeded bytes allocated for return data!"), argv[0]); 
            DebugBreak(); 
        }

    }

    //release outbuf ... 
    if(sOutbuf)
        CurrentHeap()->v->Free(CurrentHeap(), 0, sOutbuf); 

    FreeCmdLine(&argc, &argv); 
    return SetAndReturn(0); 
}

void NSStatus(const _TCHAR * sProg, const _TCHAR * sPref, const _TCHAR * sPrev, UINT nBuflen, const UINT * pStrLenOut)
{
    _tprintf(_T("\n\r%s:Using Prefix:[%s], Previous:[%s], Buflen:[%d], String Length Output Buffer:[%s]"), 
        sProg, sPref ? sPref : _T("<null>"), sPrev ? sPrev : _T("<null>"),  nBuflen, pStrLenOut ? _T("<good>") : _T("<null>")); 
}

void FNUsage(const _TCHAR *sProg)
{
    _tprintf(_T("\n\rUsage:%s [-prefix <prefix | null | empty>][-prev <previous | null | empty>]\n\r\t[-strlen null][-bufsize <size base 10>]"),sProg ); 
    _tprintf(_T("\n\rWhere -prefix is the namespace object to find(can be empty or null, -prev is the previous object(ditto)"));
    _tprintf(_T("\r\n\tusing -strlen null means the API gets a null ptr to return bytes copied, -bufsize is the size of output buffer")); 
}
