/*
* r.c - Simple utility to register an object in a namespace. 
* Calls Bind() and then Register() with flags passed on cmd line. 
* Copyright (c) Microsoft Corporation - All Rights reserved. 
*/

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "..\utilities\utility.h"

void RUsage(const _TCHAR * s);

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc; 
    UINT64 ZERO64 = Int64Initializer(0,0);
    const _TCHAR * sChild = 0;
    const _TCHAR * sParent = 0;
    const _TCHAR * sArg = 0;
    _TCHAR * stop = 0; 
    NAME_SPACE_FLAGS nBindFlags = 0, nRegFlags = 0; 
    IUnknown * pUnk = 0;
    IFile   * pFile = 0; 
    INameSpace * pParent = 0; 
    INameSpace * pCurrentNS = CurrentNameSpace(); 
  
    ParseCmdLine(&argc, &argv); 

    //give user a chance to pull up debugger ... 
    OptionalBreak(argc, argv);
    TellAboutOptionalBreak();

    if(argc < 2)
    {
        RUsage(argv[0]); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1); 
    }

    //--------------------------------------------------------------------
    //-------get args from command line, stuff into local vars ..........
    sParent = GetArgFromCommandLine(argc, argv, _T("parent"));
    if(!sParent)
    {
        RUsage(argv[0]); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    sChild = GetArgFromCommandLine(argc, argv, _T("child"));
    if(!sChild)
    {
        RUsage(argv[0]); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    sArg = GetArgFromCommandLine(argc, argv, _T("bind_flags"));
    if(sArg && *sArg)
        nBindFlags = _tcstoul(sArg, &stop, 16); 


    sArg = GetArgFromCommandLine(argc, argv, _T("reg_flags"));
    if(sArg && *sArg)
        nRegFlags = _tcstoul(sArg, &stop, 16); 


    //---------------------------------------------------------------------
    //---------------the real work begins .....
    
    //bind parent namespace if we can .... 
    sc = pCurrentNS->v->Bind(pCurrentNS, sParent, nBindFlags, &pUnk);
    Outcome(BIND, sc, sParent, nBindFlags, ZERO64); 
    
    if(FAILED(sc))
    {
        _tprintf(_T("\n\rFailed to bind parent namespace, quitting .... ")); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }

    //QI here for namespace ... take ref on pParent if succeeds
    sc = pUnk->v->QueryInterface(pUnk, &IID_INameSpace, (void *)&pParent); 
    Outcome(QUERYINTERFACE, sc, 0, 0, ZERO64); 
    pUnk->v->Release(pUnk); 

    if(FAILED(sc))
    {
        _tprintf(_T("\n\rFailed to get interface for parent namespace, quitting .... ")); 
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1); 
    }
    
    //bind the child object ... takes ref on pUnk if succeeds 
    sc = pParent->v->Bind(pParent, sChild, nBindFlags, &pUnk); 
    Outcome(BIND, sc, sChild, nBindFlags, ZERO64); 
    
    if(FAILED(sc))
    {
        _tprintf(_T("\n\rFailed to bind child namespace, quitting .... ")); 
        pParent->v->Release(pParent);
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    //get IFile interface ..takes ref on pFile if succeeds
    sc = pUnk->v->QueryInterface(pUnk, &IID_IFile, (void *)&pFile);
    Outcome(QUERYINTERFACE, sc, 0, 0, ZERO64); 
    pUnk->v->Release(pUnk);
    if(FAILED(sc))
    {
        _tprintf(_T("\n\rFailed to get interface for child namespace, quitting .... ")); 
        pParent->v->Release(pParent);
        FreeCmdLine(&argc, &argv); 
        return SetAndReturn(1);
    }

    sc = pParent->v->Register(pParent, sChild, (IUnknown *)pFile, nRegFlags, 0); 

    Outcome(REGISTER, sc, sChild, nRegFlags, ZERO64); 

    pFile->v->Release(pFile);
    pParent->v->Release(pParent);

    return SetAndReturn(0);
}

void RUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:%s -parent <parent object> -child <object name> [-bind_flags <hex flags>][-reg_flags <hex flags>]"), s); 
    _tprintf(_T("\n\r\tWhere <parent object> is name of parent, e.g., \'/fs\',"));
    _tprintf(_T("\n\r\t<object name> is the name of the newly bound object, <bind flags> are the flags to use for binding the parent"));
    _tprintf(_T("\n\r\t<reg flags> are the flags used to register the new object"));
}

