/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define _DEBUG 1
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <assert.h>
#include <diagnostics.h>

#include <drivers/drivers.h>

#if 1
/* This test got broken when IMBufFactory was deleted. What does it do? */
int _tmain(INT argc, _TCHAR **argv)
{
    printf("netdrtst.c needs updating\n");
    return 1;
}
#else

SCODE MCT MBufFactoryCreate(PIUNKNOWN* MBufFac);

#define _MAX_ARGS 10 /* totally arbitrary */

#ifdef _MSC_VER
#pragma warning(disable:4054)/* cast from function ptr to data ptr (/Za)*/
#endif

/*
 * Parse arguments to be passed to a new device.
 * Arguments are of three types, identified by
 * their leading char: strings, hex, decimals.
 */
UINT ParseArg( _TCHAR *String)
{
    _TCHAR c;
    UINT Val;

    c = *String;

    /* A hex ?
     */
    if (c == '0' && (String[1] == 'x' || String[1] == 'X')) {
        Val = _ttoh(String+2);
    }

    /* A decimal number ?
     */
    else if (c >= '0' && c <= '9') {
        Val = _ttoi(String);
    }

    /* Must be a string then
     */
    else {
        Val = (UINT) String;
    }

    return Val;
}

/* NB: This one should be used as "netdrvtest cs8900.cob 0x08000300 47"
 */
void cs8900specials(void)
{
typedef struct {
    volatile UINT32 Level;
    volatile UINT32 Direction;
    volatile UINT32 Set;
    volatile UINT32 Clear;
    volatile UINT32 DetectRaising;
    volatile UINT32 DetectFalling;
    volatile UINT32 DetectStatus;
    volatile UINT32 Alternate;    
} GPIO;
#define gpio ((GPIO*)0x90040000)

    printf("Unsleep cs8900 chip\n");
    /* Take the chip out of sleep
     */
#define CS8900_SLEEP_PIN (1<<27)
    gpio->Direction |= CS8900_SLEEP_PIN; /* its an output */
    gpio->Set = (0x5) | CS8900_SLEEP_PIN | (gpio->Level & gpio->Direction);
    _tprintf(_TEXT("GPIO: l=%x d=%x 1=%x 0=%x r=%x f=%x s=%x a=%x\n"), *gpio);
}


EXTERN_C PCNETDRIVER RTLCALLTYPE
Cs8900New(PUINT pPortBase, PUINT pIrq);

int _tmain(INT argc, _TCHAR **argv)
{
/*    UINT PortBase = 0x08000300; */
    UINT Irq = 47; /* 32+(26-11) */
    UINT r, i;
    PCNETDRIVER Dev = NULL;
    TIME Idle;
    _TCHAR *DriverCobName;
    _TCHAR *DriverName;
    UINT Args[_MAX_ARGS];
    PINAMESPACE pNs = CurrentNameSpace();
    SCODE sc;
    PIUNKNOWN pUnk;
    PIDEVICECONSTRUCTOR pDevCons;
    PINETDRIVER pDev;
    PIENDPOINTFACTORY MBufFac;

    /* Check our arguments
     */
    ParseCmdLine(&argc,&argv);
    if (argc < 3) {
        _tprintf(_T("Usage: %s <COB/drivername.cob> <name> [<portbase> <irqnum> ...]\n"),
                 argv[0]);
        
        goto Out;
    }
    if (argc > (_MAX_ARGS+2)) {
        _tprintf(_T("Too many arguments, max is %d\n"), _MAX_ARGS);
        goto Out;
    }

    /* Arguments to driver
     */
    DriverCobName = argv[1];
    DriverName = argv[2];
    for (i = 0; (int) i < (argc - 3); i++)
        Args[i] = ParseArg(argv[i+3]);
    for (; i < _MAX_ARGS; i++)
        Args[i] = 0;

    /* Prepare stuff we need */
     /* load drivers and leak ref */
    sc = pNs->v->Bind(pNs, _T("COB/drivers.cob"), 0, &pUnk);
    assert(!FAILED(sc));
    sc = MBufFactoryCreate((PIUNKNOWN*) &MBufFac);
    assert(!FAILED(sc));
    
    /* Cause the driver to load
     */
    sc = pNs->v->Bind(pNs,DriverCobName,0,&pUnk);
    assert(!FAILED(sc));    /* cob not found */

    sc = pUnk->v->QueryInterface(pUnk, &IID_IDeviceConstructor,
                                 (void **) &pDevCons);
    pUnk->v->Release(pUnk);
    assert(!FAILED(sc));    /* cob entry not a constructor */

    /* Special cases
     */
//    if (_tcscmp(DriverName,_T("cs0")) == 0)
        cs8900specials();

    _tprintf(_TEXT("Calling %sNew @%x\n"),
             DriverName, (ADDRESS) pDevCons->v->Constructor);
    sc = pDevCons->v->Constructor(pDevCons,
                                  DriverName,
                                  Args[0],
                                  Args[1],
                                  &pUnk,
                                  (_TCHAR *) Args[2],
                                  (_TCHAR *) Args[3],
                                  Args[4],
                                  Args[5],
                                  Args[6],
                                  Args[7],
                                  Args[8]);
    pDevCons->v->Release(pDevCons);
    printf("NetDrvTest: Constructor for %s returned sc=%x\n", DriverName, sc);
    if(FAILED(sc))    /* Constructor failed */
        goto Out;

    sc = pUnk->v->QueryInterface(pUnk, &IID_INetDriver, (void **) &pDev);
    pUnk->v->Release(pUnk);
    assert(!FAILED(sc));    /* Constructed object wrong type */

    /* Cast interface to concrete class for this test (generally unsafe) */
    Dev = (PCNETDRIVER) pDev;

    _tprintf(_TEXT("Got @%x\n"), (ADDRESS) Dev);

    if (Dev != NULL) {
        /* Init condition
         */
        Condition_Init(&Dev->IntrPoint);

        /* Install MbufFactory */
        Dev->MBufFact = MBufFac;

        /* Install ISR and enable interrupt pin
         */
        _tprintf(_TEXT("Installing ISR @%x\n"), (ADDRESS) Dev->v->Isr);
        r = AddDevice(Dev,(PTR)Dev->v->Isr,0,Irq,0);
        _tprintf(_TEXT("Got %x\n"),r);

        /* Initialize it
         */
        _tprintf(_TEXT("Calling Reset @%x\n"), (ADDRESS) Dev->v->Initialize);
        r = Dev->v->Initialize(Dev);
        _tprintf(_TEXT("Got %x\n"),r);

        /* Let it work for a while
         */
        Int32ToInt64(Idle,TIME_RELATIVE(TIME_SECONDS(1)));
        for (i = 0; i < 30; ) {
            r = Condition_Wait(&Dev->IntrPoint, NULL, &Idle);

            if (r != E_TIMED_OUT) {
                /* Handle interrupt
                 */
                r = Dev->v->HandleInterrupt(Dev);
#if 0
                /* Print status, just for fun
                 */
                r = Dev->v->Ioctl(Dev,1,NULL);
#endif
            } else {
                _tprintf(_TEXT("Idle.....\n"));
                i++;
            }
        }

        /* Silence it
         */
        _tprintf(_TEXT("Removing ISR @%x\n"), (ADDRESS) Dev->v->Isr);
        r = RemoveDevice(Dev,Irq);
        _tprintf(_TEXT("Got %x\n"),r);
        
        /* Let it die
         */
        Dev->v->Release(Dev);
    }

 Out:
    FreeCmdLine(&argc,&argv);
    return 0;
}

#define STUB(x,n) \
PTR x(UINT32 arg) \
{ \
  _tprintf(_TEXT("%s was called with arg0=%x\n"),n,arg);\
  return NULL; \
}

STUB(ether_output,"ether_output")
STUB(NetLock,"NetLock")
STUB(NetUnlock,"NetUnlock")
STUB(if_attach,"if_attach")


/*=============================================================*/
/*=============================================================*/
/*=============================================================*/
/*=============================================================*/
/*=============================================================*/
#define MSIZE 128
#define MCLBYTES 2048
#define MLEN (MSIZE - sizeof(struct m_hdr))   /* normal data len */
#define MHLEN (MLEN - sizeof(struct pkthdr))  /* data len w/pkthdr */
#define MINCLSIZE (MHLEN + MLEN)   /* smallest amount to put in cluster */

#define mtod(m,t) ((t)((m)->m_data))

/* header at beginning of each mbuf: */
struct m_hdr {
    struct mbuf *mh_next;     /* next buffer in chain */
    struct mbuf *mh_nextpkt;  /* next chain in queue/record */
    INT mh_len;               /* amount of data in this mbuf (in bytes) */
    PBYTE mh_data;            /* location of data */
    UINT16 mh_type;           /* type of data in this mbuf */
    UINT16 mh_flags;          /* flags; see below */
};

/* record/packet header in first mbuf of chain; valid if M_PKTHDR set */
struct pkthdr {
    INT len;              /* total packet length */
    struct ifnet *rcvif;  /* rcv interface */
};

/* description of external storage mapped into mbuf, valid if M_EXT set */
struct m_ext {
    PBYTE ext_buf;       /* start of buffer */
    void (*ext_free)(PBYTE, UINT);  /* free routine if not the usual */
    UINT ext_size;       /* size of buffer, for ext_free */
};

struct mbuf {
    struct m_hdr m_hdr;
    union {
        struct {
            struct pkthdr MH_pkthdr;     /* M_PKTHDR set */
            union {
                struct m_ext MH_ext;     /* M_EXT set */
                char MH_databuf[MHLEN];
            } MH_dat;
        } MH;
        char M_databuf[MLEN];            /* !M_PKTHDR, !M_EXT */
    } M_dat;
};
#define m_next m_hdr.mh_next
#define m_len m_hdr.mh_len
#define m_data m_hdr.mh_data
#define m_type m_hdr.mh_type
#define m_flags m_hdr.mh_flags
#define m_nextpkt m_hdr.mh_nextpkt
#define m_act m_nextpkt
#define m_pkthdr M_dat.MH.MH_pkthdr
#define m_ext M_dat.MH.MH_dat.MH_ext
#define m_pktdat M_dat.MH.MH_dat.MH_databuf
#define m_dat M_dat.M_databuf

/* mbuf flags */
#define M_EXT    0x0001  /* has associated external storage */
#define M_PKTHDR 0x0002  /* start of record */

/* mbuf pkthdr flags, also in m_flags */
#define M_BCAST  0x0100  /* send/received as link-level broadcast */
#define M_MCAST  0x0200  /* send/received as link-level multicast */

/* flags copied when copying m_pkthdr */
#define M_COPYFLAGS (M_PKTHDR|M_EOR|M_BCAST|M_MCAST)

/*
 * Mbuf cluster macros.
 * MCLALLOC(caddr_t p, INT how) allocates an mbuf cluster.
 * MCLGET adds such clusters to a normal mbuf;
 * the flag M_EXT is set upon success.
 * MCLFREE releases a reference to a cluster allocated by MCLALLOC,
 * freeing the cluster if the reference count has reached 0.
 *
 * Normal mbuf clusters are normally treated as character arrays
 * after allocation, but use the first word of the buffer as a free list
 * pointer while on the free list.
 */
union mcluster {
    union mcluster *mcl_next;
    char mcl_buf[MCLBYTES];
};


struct mbuf * RTLCALLTYPE m_get(INT how, INT type)
{
    struct mbuf *m;

    m = malloc(sizeof(struct mbuf));
    if (m) {
        (m)->m_type = (UINT16)(type);
        (m)->m_next = (struct mbuf *)NULL;
        (m)->m_nextpkt = (struct mbuf *)NULL;
        (m)->m_data = (PTR) (m)->m_dat;
        (m)->m_flags = 0;
    }
    return (m);
}

struct mbuf * RTLCALLTYPE m_gethdr(INT how, INT type)
{
    struct mbuf *m;

    m = malloc(sizeof(struct mbuf));
    if (m) {
        (m)->m_type = (UINT16)(type);
        (m)->m_next = (struct mbuf *)NULL;
        (m)->m_nextpkt = (struct mbuf *)NULL;
        (m)->m_data = (PTR) (m)->m_pktdat;
        (m)->m_flags = M_PKTHDR;
    }
    return (m);
}
struct mbuf * RTLCALLTYPE m_clget(struct mbuf *m, INT how)
{
    (m)->m_ext.ext_buf = malloc(MCLBYTES);

    if ((m)->m_ext.ext_buf != NULL) {
        (m)->m_data = (PTR) (m)->m_ext.ext_buf;
        (m)->m_flags |= M_EXT;
        (m)->m_ext.ext_size = MCLBYTES;
    }

    return (m);
}

void RTLCALLTYPE m_freem(struct mbuf *m)
{
    struct mbuf *nn;

    if (m == NULL)
        return;
    do {
        if ((m)->m_flags & M_EXT) {
            free((m)->m_ext.ext_buf);
        }
        (nn) = (m)->m_next;
        free(m);
        m = nn;
    } while (m);
}

/*==========================================*/
/*==========================================*/
/*==========================================*/
/*==========================================*/
/*==========================================*/
struct ether_header {
    UINT8 ether_dhost[6];
    UINT8 ether_shost[6];
    UINT16 ether_type;
};

UINT32 nBroadcasts = 0;

void RTLCALLTYPE
ether_input(struct ifnet *ifp, struct ether_header *eh, struct mbuf *m)
{
    struct _c {
        UINT32 d03;
        UINT16 d45;
        UINT16 s01;
        UINT32 s25;
        UINT16 t;
    } *p = (struct _c *)eh;
    
    if ((p->d03 == 0xffffffff) &&
        (p->d45 == 0xffff)) {
        nBroadcasts++;
    } else {
        printf("[B%d] %08x%04x %04x%08x %04x\n",
               nBroadcasts,
               p-> d03,
               p-> d45,
               p-> s01,
               p-> s25,
               p-> t);
    }

    m_freem(m);
}



typedef struct MBUFFACTORY MBUFFACTORY, *PMBUFFACTORY;


struct MBUFFACTORY {
    const struct IMBufFactoryVtbl*  v;
    UINT                            Refs;
};


SCODE MCT
MBufFactoryQueryInterface(PIMBUFFACTORY This, REFIID IidQuery, void** ppUnk)
{
    return GenericQueryInterface((PIUNKNOWN) This, IidQuery, ppUnk, &IID_IMBufFactory);
}


UINT MCT 
MBufFactoryAddRef(PIMBUFFACTORY This)
{
    return AtomicInc(&((PMBUFFACTORY) This)->Refs);
}


UINT MCT
MBufFactoryRelease(PIMBUFFACTORY This)
{
    UINT Refs = AtomicDec(&((PMBUFFACTORY) This)->Refs);
    if (Refs == 0)
        CurrentHeap()->v->Free(CurrentHeap(), 0, This);
    return Refs;
}


SCODE MCT
MBufFactoryMBFree(PIMBUFFACTORY This, struct mbuf* m)
{
    m_freem(m);
    return S_OK;
}


SCODE MCT
MBufFactoryMBClGet(PIMBUFFACTORY This, struct mbuf* m, INT how)
{
    m = m_clget(m, how);
    if (m == NULL) return E_FAIL;
    return S_OK;
}


SCODE MCT
MBufFactoryMBGet(PIMBUFFACTORY This, INT how, INT type, struct mbuf** m)
{
    *m = m_get(how, type);
    if (*m == NULL) return E_FAIL;
    return S_OK;
}


SCODE MCT
MBufFactoryMBGetHdr(PIMBUFFACTORY This, INT how, INT type, struct mbuf** m)
{
    *m = m_gethdr(how, type);
    if (*m == NULL) return E_FAIL;
    return S_OK;
}


SCODE MCT 
EtherInput(PIMBUFFACTORY pThis, struct ifnet *ifp, struct ether_header *eh, struct mbuf *m)
{
    ether_input(ifp, eh, m);
    return S_OK;
}


SCODE MCT
EtherOutput(PIMBUFFACTORY pThis, struct ifnet *ifp, struct mbuf *m0, struct sockaddr *dst, struct rtentry *rt0)
{
    _tprintf(_TEXT("EtherOutput was called\n"));
    return E_NOT_IMPLEMENTED;
//    return ether_output(ifp, m0, dst, rt0);
}


void MCT 
EFNetLock(PIMBUFFACTORY pThis)
{
    _tprintf(_TEXT("NetLock was called\n"));
    //NetLock();
}


void MCT 
EFNetUnlock(PIMBUFFACTORY pThis)
{
    _tprintf(_TEXT("NetUnlock was called\n"));
    //NetUnlock();
}

SCODE MCT 
IfAttach(PIMBUFFACTORY pThis, struct ifnet *ifp)
{
    _tprintf(_TEXT("if_attach was called\n"));
    return E_NOT_IMPLEMENTED;
//    if_attach(ifp);
}

struct IMBufFactoryVtbl MBufFacVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    MBufFactoryQueryInterface,
    MBufFactoryAddRef,
    MBufFactoryRelease,
    MBufFactoryMBFree,
    MBufFactoryMBClGet,
    MBufFactoryMBGet,
    MBufFactoryMBGetHdr,
    EtherInput,
    EtherOutput,
    EFNetLock,
    EFNetUnlock,
    IfAttach
};


SCODE MCT MBufFactoryCreate(PIUNKNOWN* MBufFac)
{
    PIHEAP Heap = CurrentHeap();
    PMBUFFACTORY MBufFacImpl = NULL;

    if (! Heap) return E_FAIL;

    MBufFacImpl = Heap->v->Alloc(Heap, 0, sizeof(MBUFFACTORY), 0);
    if (! MBufFacImpl) return E_FAIL;

    MBufFacImpl->v = &MBufFacVtbl;
    MBufFacImpl->Refs = 1;

    *MBufFac = (PIUNKNOWN) MBufFacImpl;

    return S_OK;
}

#endif
