/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Regression tester.
 * Runs all tests, in turn.
 * Assumes that if anyone fails it will DebugBreak() itself,
 * this program just runs them tests and thats all.
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#ifdef _DEBUG
#define DPRINT(_s_) printf _s_
#else
#define DPRINT(_s_)
#endif
#define CHECKSC(_sc_) if (FAILED(_sc_)) __asm { int 3 }
#define CHECK(_exp_) if (!(_exp_)) __asm { int 3 }

void ttt(    volatile UINT32 *p, UINT n);

int hpstat()
{
    SCODE sc;
    PIHEAP pHeap = CurrentHeap();
    ADDR_SIZE Reserve,Commit,Used,MaxUsed;

    /* The heap now
     */
    Reserve = Commit = Used = MaxUsed = 0;
    sc = pHeap->v->Status(pHeap,&Reserve,&Commit,&Used,&MaxUsed);
    if (FAILED(sc)) {
        _tprintf(_TEXT("IHeap::Status failed! (x%x) hp=%x"), sc, pHeap);
        DebugBreak();
        return 0;
    }

    _tprintf(_TEXT("PgTets Heap: %d reserve %d commit %d used\n"), 
             Reserve, Commit, Used);
    _tprintf(_TEXT("PgTest: Maximum heap memory usage was %d bytes\n"), MaxUsed);


    /* Done
     */
    return 0;
}

BOOL InUserMode = FALSE;
void FlushCaches(void)
{
    if (InUserMode)
        return;
    __asm {
        wbinvd;
        invd;
    }
}

#pragma warning(disable:4035)   /* no return value */
/*__inline*/ __declspec(naked) __int64 __cdecl ReadPerfC(void)
{
    __asm {
        /* rdtsc; */
        __emit 0x0f;
        __emit 0x31;
        ret;
    }
}

PTR at;

void THREAD_LINKAGE KTestThread(THREAD_ARGUMENT arg)
{
    ttt(at, (int) arg);
    ThreadExit();
}

void RealTest()
{
    ttt((PTR) (1024*1024), 255);
}

void ttt(    volatile UINT32 *p, UINT n)
{
    UINT i, j;
    TIME t0,t1,t2,t3,t4,t5,t6,t7,t8,t9;

    DPRINT(("KTest n=%x p=%x\n", n, p));

    t0 = t1 = t2 = t3 = 0;

    t0 = ReadPerfC();
    t1 = ReadPerfC();
    FlushCaches();
    t2 = ReadPerfC();
    for (i = 0; i < n; i++) {
        p[(i<<10) + (i<<3)];
    }
    t3 = ReadPerfC();
    FlushCaches();
    t4 = ReadPerfC();
    for (i = 0; i < n; i++) {
        p[(i<<10) + (i<<3)];
    }
    t5 = ReadPerfC();
    for (j = 0; j < n/64; j++)
    for (i = 0; i < 64; i++) { // miss in TLB but not in cache
        p[i<<10];
    }
    t6 = ReadPerfC();
    FlushCaches();
    t7 = ReadPerfC();
    for (i = 0; i < n; i++) { // miss in cache but not in TLB
        p[i*(32/4)];
    }
    t8 = ReadPerfC();
    for (i = 0; i < n; i++) { // no misses
        p[i*(32/4)];
    }
    t9 = ReadPerfC();
    printf("KTestTh(%d) %ld %ld %ld %ld %ld %ld %ld %ld %ld\n",
           n, t1-t0, t2-t1, t3-t2, t4-t3, t5-t4,
           t6-t5, t7-t6, t8-t7, t9-t8);
    /* 166 MHz * 6 == 1GHz == 1/ns */
    hpstat();
}

void kt(UINT NPages)
{
    static PIVMVIEW prev = NULL;

    SCODE sc;
    PIPROCESS p = CurrentProcess();
    /*PIVMVIEW vv = NULL;*/
    PITHREAD th;

    /*    vv = VVSetup1(MAP_AT, _PAGE_SIZE * NPAGES); */
    at = (PTR) (1024 * 1024);
#if 0
CurrentHeap()->v->Alloc(CurrentHeap(),
                                       HEAP_ZERO_MEMORY,
                                       4096 * NPages,
                                       4096);
#endif

    sc = p->v->CreateThread(p, 0, KTestThread, (PTR) NPages, 0, NULL, &th);
    DPRINT(("Created thread sc=%x\n"));
    CHECKSC(sc);
#ifdef _DEBUG
    /* allow time for prints */
    SleepUntil(TIME_RELATIVE(600*(TIME)TIME_SECONDS(1)));
#else
    SleepUntil(TIME_RELATIVE(TIME_SECONDS(10)));
#endif

    CurrentHeap()->v->Free(CurrentHeap(), 0, at);
    th->v->Release(th);
    SleepUntil(TIME_RELATIVE(TIME_SECONDS(1)));
}

void KernelTest(void)
{
    kt(10);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    KernelTest();
    
    /* Done
     */
    return 0;
}
