#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "corn.h"
#include "icorn.c"
/*
CORN - Your basic COB. Used to test loading / image binding functions. Need also icorn.c, corn.h
This very simple demo just maintains an internal variable through interface functions. 
*/

struct _ICornVtbl CornObjVtbl; 

/* private data stays private thank you
 */
typedef struct _CORN
{
    struct _ICornVtbl * Vtbl; 
    _TCHAR *sParms;
    UINT refcount; 
    UINT kernels;
}CORN, *PCORN; 

/*
Constructor / initializer for corn COB. 
*/
SCODE SproutCorn(PICORN * ppCorn)
{
    IHeap * pHeap = CurrentHeap(); 
    PCORN pCorn = (CORN *)pHeap->v->Alloc(pHeap, 0, sizeof(CORN), 0); 
    if(!pCorn) return E_NOT_ENOUGH_MEMORY; 
    pCorn->Vtbl = &CornObjVtbl; 

    //DebugBreak(); 

    //here we set refcnt to 1 initially ... seems to be the standard
    pCorn->refcount = 1; 
    pCorn->kernels = 0;
    *ppCorn = (ICORN *)pCorn; 
    return S_OK; 
}

UINT CornAddRef(PICORN pThis)
{
    UINT nNew; 
    //DebugBreak(); 
    nNew = AtomicInc(&((PCORN)pThis)->refcount); 
    return nNew; 
}

UINT CornRelease(PICORN pThis)
{
    PCORN pCorn; 
    UINT refs; 

//    DebugBreak(); 

    pCorn = (PCORN)pThis; 
    refs = AtomicSub(&pCorn->refcount, 1); 
    if(refs == 0)
    {
        IHeap * pHeap = CurrentHeap(); 
        pHeap->v->Free(pHeap, 0, pThis); 
    }

    return refs; 
}
/*
CornQueryInterface()
*/
SCODE CornQI(PICORN pThis, REFIID riid, void **ppNew)
{
    if(UuidCmp(&IID_ICorn, riid) || UuidCmp(&IID_IUnknown, riid))
    {
        *ppNew = (void *)pThis; 
        CornAddRef(pThis); 
        return NO_ERROR; 
    }

    *ppNew = 0; 
    return E_NO_INTERFACE; 
}


/*
* Just some silly functions to test interface pointers. ... should choke if the vtbl 
* got hosed somehow
*/
UINT AddKernels(PICORN pThis, UINT nKernels)
{
    PCORN pCorn = (PCORN)pThis;
    pCorn->kernels += nKernels; 
    return pCorn->kernels; 
}

UINT EatKernels(PICORN pThis, UINT nKernels)
{
    PCORN pCorn = (PCORN)pThis;
    if(nKernels <= pCorn->kernels)
        pCorn->kernels -= nKernels; 
    else
        pCorn->kernels = 0; 

    return pCorn->kernels; 
}

//print parameters .. 
void PrintParms(PICORN pThis)
{
    const _TCHAR * sArgs = ProcessArgs(); 
    _tprintf(_T("\r\nCORN.COB got params of [%s]"), sArgs ? sArgs : _T("null"));
}


struct _ICornVtbl CornObjVtbl = 
{
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    CornQI, 
    CornAddRef, 
    CornRelease,
    AddKernels,
    EatKernels,
    PrintParms
}; 


//entry point ... 
PIUNKNOWN CobMain(void)
{
    ICORN * pCorn;
    SCODE sc; 
 //   DebugBreak(); 
    sc = SproutCorn(&pCorn); 
    if(sc != S_OK)
        return 0; 
    
     _tprintf(_TEXT("Hello, corn cob!"));
     return (PIUNKNOWN)pCorn;
}




