//Copyright (c) Microsoft Corporation All rights reserved
/*
* MI.EXE - Machine Information test program 
* Exercises IProcess->MachineInformation() function
* Cmd line flags: 
*   -flavor <generic | trimedia> default is generic. Chooses type of CPU 
* Prints members of MACHINE_INFO struct. 
* Returns 0 if pass 1 if fail
*/

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "..\utilities\utility.h"

void MIOutcome(SCODE sc, const struct MACHINE_INFO * pMI);
void MIStatus(const _TCHAR * s, const _TCHAR * sProc);
void MIUsage(const _TCHAR * s);

int _tmain(INT argc, _TCHAR ** argv)
{
    SCODE sc; 
    UINT32 nFlavor;
    struct MACHINE_INFO mi = {0,};
    const _TCHAR * sFlavor = 0; 
    
    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv);
    TellAboutOptionalBreak();
    
    if(argc < 2)
        MIUsage(argv[0]);

    sFlavor = GetArgFromCommandLine(argc, argv, _T("flavor")); 
    if(!sFlavor)
        nFlavor = MACHINE_KIND_GENERIC; 
    else
        if(!*sFlavor)
            nFlavor = MACHINE_KIND_GENERIC; 
        else
        {
            if(!_tcsicmp(_T("generic"), sFlavor))
                nFlavor = MACHINE_KIND_GENERIC; 
            else
                if(!_tcsicmp(_T("trimedia"), sFlavor))
                    nFlavor = MACHINE_KIND_TRIMEDIA ; 
                else
                    nFlavor = MACHINE_KIND_GENERIC; 
        }

    MIStatus(argv[0], ProcessorFlavor(nFlavor));
    sc = CurrentProcess()->v->MachineInfo(CurrentProcess(), nFlavor, &mi); 
    MIOutcome(sc, &mi); 
    FreeCmdLine(&argc, &argv);
    if(FAILED(sc))
        return SetAndReturn(1);
    return SetAndReturn(0);
    
}

/*
* Prints MACHINE_INFO and status code information in human friendly format
*/
void MIOutcome(SCODE sc, const struct MACHINE_INFO * pMI)
{
    _tprintf(_T("\n\rIProcess::MachineInfo() returned [0x%.8x : %s]"), sc, ErrorDescription(sc)); 
    _tprintf(_T("\n\rProcessor Kind:[%d:%s]"), pMI->ProcessorKind, ProcessorKind(pMI->ProcessorKind)); 
    _tprintf(_T("\n\rOEM ID:[%d (0x%.8x)]"), pMI->OEMId, pMI->OEMId); 
    _tprintf(_T("\n\rTotal Physical Memory:[%d bytes]"), pMI->TotalPhysicalMemory); 
    _tprintf(_T("\n\rAvailable Physical Memory:[%d (0x%.8x) bytes]"), pMI->AvailablePhysicalMemory, pMI->AvailablePhysicalMemory); 
    _tprintf(_T("\n\rProcessor Speed:[%d Hz]"), Uint64Low32Bits(pMI->ProcessorSpeed)); 
}

void MIUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:[%s] -flavor <generic | trimedia>"), s); 
}

void MIStatus(const _TCHAR * s, const _TCHAR * sProc)
{
    _tprintf(_T("\n\r[%s] Status:Using processor [%s]"), s, sProc); 
}
