/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdlib.h>
#include <stdio.h>
#include <tchar.h>
#include <w32event.h>
#include <queue.h>

#if defined(_DEBUG)
#define DBG(_x_) printf _x_
#else
#define DBG(_x_)
#endif

volatile BOOL KeepGoing = TRUE;
#define MAXNOTHREADS    100
UINT NThreads = 17;
UINT ThreadsLeft;
CONDITION CondMaybeDone;

IThread *Thread[MAXNOTHREADS];

IQueue *pQueue;

typedef struct _TNODE
{
    QUEUENODE Node;
    BOOL OnQueue;
} TNODE;

TNODE TNode[MAXNOTHREADS];

void ThreadFunc(THREAD_ARGUMENT MyId)
{
    TNODE *pNode;
    HANDLE Event;

    printf("Thread %d starting...\n",MyId);

    pNode = &TNode[(int)MyId];

    Event = pQueue->v->GetEventHandle(pQueue);
    while (KeepGoing)
    {
    if (pNode)
    {
        pNode->OnQueue=TRUE;
        pQueue->v->Put(pQueue,(QUEUENODE *)pNode);
    }
    
    pNode = (TNODE *)pQueue->v->Get(pQueue);
    if (pNode)
        pNode->OnQueue=FALSE;
    else
        WaitForSingleObject(Event, ((TIME *)-1) /*INFINITE*/);    /* Sleep until something happens */
    }
    
    printf("%d: Exiting, ThreadsLeft=%d\n",MyId,AtomicDec(&ThreadsLeft));
    Condition_Signal(&CondMaybeDone);
    return;
}

int _tmain(int argc, _TCHAR *argv[])
{
    INT TestTime = 3;
    UINT i;
    SCODE sc;
    PIPROCESS pPrc;
    TNODE *pNode;
    TIME t;
    HANDLE QueueEvent;

    sc = ParseCmdLine(&argc,&argv);

    switch (argc) 
    {
    case 3:
        TestTime = atoi(argv[2]);
    case 2:
        NThreads = atoi(argv[1]);
    case 1:
    case 0:
        break;
    default:
        printf("usage: tcswitch [NThreads [TestTime]]\n");
    }

    (void) FreeCmdLine(&argc,&argv);

    if (NThreads > MAXNOTHREADS) 
    {

        printf("Can't create more than %d threads\n", MAXNOTHREADS);
        exit(0);
    }

    printf("Test %d threads for %d seconds\n", NThreads, TestTime);

    Condition_Init(&CondMaybeDone);

    /* Create all those threads */

    pPrc = CurrentProcess();

    QueueEvent=CreateEvent(FALSE,FALSE);

    if (QueueCreate(QueueEvent,&pQueue)!=S_OK)
    {
        printf("Couldn't create queue\n");
        exit(-1);
    }

    ThreadsLeft=NThreads;
    for (i = 0; i < NThreads; i++) 
    {
        TNode[i].OnQueue=FALSE;
        sc = pPrc->v->CreateThread(pPrc, (THREAD_FUNCTION)ThreadFunc, (THREAD_ARGUMENT)i,
                         0, NULL, &Thread[i]);
        if (FAILED(sc)) 
        {
            printf("tcswitch: error 0x%x creating thread #%d\n", sc, i);
            exit(0);
        }
    }

    printf("All %d threads created; sleeping for %d seconds\n", NThreads, TestTime);

    Int32ToInt64(t,TIME_RELATIVE(TIME_SECONDS(TestTime)));
    SleepUntil(t);

    /* Disable counting, and kill all the threads */

    printf("Waiting for threads to exit\n");
    KeepGoing = FALSE;

    while (ThreadsLeft>0)
    {
        (void)SetEvent(QueueEvent);   /* Kick any sleeping threads */
        Condition_Wait(&CondMaybeDone, NULL, NULL);
        printf("Threads Left = %d\n",ThreadsLeft);
    }

    printf("Releasing threads\n");

    for (i = 0; i < NThreads; i++) 
    {
        if (Thread[i])
        {
            Thread[i]->v->Release(Thread[i]);
        }
    }
    
    printf("Terminated all %d threads\n", NThreads);
    printf("Pulling extra nodes off queue\n");
    while ( (pNode = (TNODE *)pQueue->v->Get(pQueue))!=NULL )
    {
        pNode->OnQueue=FALSE;
    }

    for (i = 0; i < NThreads; i++) 
    {
        printf("i: OnQueue=%d\n",TNode[i].OnQueue);
    }

    Condition_Destroy(&CondMaybeDone);
    pQueue->v->Release(pQueue);
    (void)CloseHandle(QueueEvent);
    return 0;
}
