/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"

#if defined(i386) || defined(arm)
#define TAKES_SOME_EXCEPTIONS 0
#define TAKES_READ_EXCEPTIONS 0
#elif defined(__TCS__)
#define TAKES_SOME_EXCEPTIONS 1
#define TAKES_READ_EXCEPTIONS 0
#elif defined(mips)
#define TAKES_SOME_EXCEPTIONS 1
#define TAKES_READ_EXCEPTIONS 1
#endif

#if !defined(arm)
BOOL testit(BOOL DoReadFirst)
{
    UINT *pUnaligned;
    UINT x[3];

    x[0] = 0x01020304;
    x[1] = 0x05060708;
    x[2] = 0x090a0b0d;

    pUnaligned = (UINT *)((BYTE *)x + 4 + 1);

    /* Unaligned reads
     * NB: Command line lets us skip this and go stright to write test
     */
    if (DoReadFirst) {
        _tprintf(_TEXT("Attempting unaligned read access at address x%x..\n\r"),
                 (ADDRESS) pUnaligned);
        /* On TM1 unaligned reads are silently truncated */
#if TAKES_READ_EXCEPTIONS
        _tprintf(_TEXT("FAILED! Value returned is x%x..\n\r"), * pUnaligned);
        /* will only get here if something is rather broken */
        return FALSE;
#else
        _tprintf(_TEXT("Value returned is x%x..\n\r"), * pUnaligned);
#endif
    }

    /* Unaligned writes
     * NB: If this takes *some* exceptions it oughta take 'em on write
     */
    _tprintf(_TEXT("Attempting unaligned write access at address x%x..\n\r"),
             (ADDRESS) pUnaligned);

    *pUnaligned = 0x090a0b0c;

    /* Shouldnt get here */
    _tprintf(_TEXT("FAILED! Values (around *ptr) are x%x x%x\n\r"),
             x[0], x[1]);
    /* will only get here if something is rather broken */
    return FALSE;
}
#endif

int _tmain(INT argc, _TCHAR **argv)
{
    BOOL bRead;

    ParseCmdLine(&argc,&argv);
    bRead = (argc <= 1);
    FreeCmdLine(&argc,&argv);

#if TAKES_SOME_EXCEPTIONS
    /* We expect the exception to kill this (only) thread,
     * and consequently unload this module.
     * Set the exit code now then.
     */
    {
        PIPROCESS pPrc = CurrentProcess();

        pPrc->v->SetValue(pPrc,PROCESS_VALUE_EXIT_CODE,EXIT_SUCCESS);
    }

    /* Do the thang
     */
    (void) testit(bRead);

    /* So if we ever get back something is screwed
     */
    _tprintf(_TEXT("test FAILED.\n\r"));
    return SetAndReturn(EXIT_FAILURE);
#else
    _tprintf(_TEXT("Unaligned accesses do not trigger exceptions.\n\r"));
    _tprintf(_TEXT("test PASSED.\n\r"));
    return SetAndReturn(EXIT_SUCCESS);
#endif /* none of that on x86 */

}
