/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Regression tester.
 * Runs all tests, in turn.  Counts failures and run time.
 */

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <mmlite.h>
#include <mmmacros.h>

#if arm
#define MY_WAIT_TIME 5
#else
#define MY_WAIT_TIME 1000
#endif

typedef struct {
    const _TCHAR *  ImageName;
    const _TCHAR *  Arguments;
    const _TCHAR *  ScreenResult;
    UINT   Flags;
#define F_DONT_WAIT 0x01
#define F_LEAK_REF  0x02
#define F_DONT_SIM  0x04
#define F_FAILED    ~0    /* collect & print results at end */
} TEST_CASE;

TEST_CASE AllTests[] = {
    /* Say what the machine config looks like,
     * if *anything* goes wrong at least we know where.
     */
    { _TEXT("machinfo.exe"), NULL, NULL },

    /* Do this first, lest we need args later */
    { _TEXT("cmdecho.exe"),
        _TEXT("This is a command line test."),
        _TEXT("cmdecho.exe: This is a command line test.")
    },

    { _TEXT("lsmodule.exe"), NULL, NULL },

    /* Test for exceptions, as we might get some later on.
     */
    { _TEXT("algnflt.exe"), NULL, NULL },

    /* Test for interrupts.
     */
    { _TEXT("tcis.exe"), NULL, NULL },

    { _TEXT("echo.exe"),
        _TEXT("This is another command line test."),
        _TEXT("Got 7 arguments:\n\recho.exe\n\rThis\n\ris\n\ranother\n\rcommand\n\rline\n\rtest.")
    },

    { _TEXT("himom.exe"), _TEXT("Hello MMLite!"), 
        _TEXT("Hi Mom!") },
    { _TEXT("tstshf64.exe"), NULL, NULL },
    { _TEXT("tstxor64.exe"), NULL, NULL },
    { _TEXT("tstads64.exe"), NULL, NULL },
    { _TEXT("tstuads64.exe"), NULL, NULL },
    { _TEXT("tstdiv64.exe"), NULL, NULL },
    { _TEXT("tstudiv64.exe"), NULL, NULL },
    { _TEXT("tstmul64.exe"), NULL, NULL },
    { _TEXT("tstumul64.exe"), NULL, NULL },
    { _TEXT("tstmul64s.exe"), NULL, NULL },
    { _TEXT("tstumul64s.exe"), NULL, NULL },
    { _TEXT("tstmod64.exe"), NULL, NULL },
    { _TEXT("tstumod64.exe"), NULL, NULL },
    { _TEXT("tstmemfn.exe"), NULL, NULL },
    { _TEXT("heaptest.exe"), NULL, NULL },
    { _TEXT("heaptstx.exe"), NULL, NULL },
    { _TEXT("heaptstc.exe"), NULL, NULL },
    { _TEXT("limage.exe"), NULL, NULL },
    { _TEXT("nametest.exe"), NULL, NULL },
    
    { _TEXT("t1.exe"), NULL, NULL },
    { _TEXT("t2.exe"), NULL, NULL },
    { _TEXT("t3.exe"), NULL, NULL },
    { _TEXT("t4.exe"), NULL, NULL },
    { _TEXT("t5.exe"), NULL, NULL },
    { _TEXT("t6.exe"), NULL, NULL },
    { _TEXT("t7.exe"), NULL, NULL },

    /* RAID database tests here, make sure we wont see 'em again
     */
    { _TEXT("raid135.exe"), NULL, NULL },
    { _TEXT("raid164.exe"), NULL, NULL, F_DONT_SIM },
    { _TEXT("raid182.exe"), NULL, NULL },

#if 0
    /* Other test that dont work but we might be told what they do
     */
    { _TEXT("prime_realloc_and_free.exe"), NULL, NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -startpos 0x0 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -w -startpos 0x0 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -w -x -startpos 0x0 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -x -startpos 0x0 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -w -x -startpos 0x0 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -startpos 0x0 -size 0x0"), NULL },

    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -startpos 0x0 -size 0x10"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -w -startpos 0x10 -size 0x0"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -w -x -startpos 0x10 -size 0x10"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -r -x -startpos 0x100000 -size 0x10"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -w -x -startpos 0x100000 -size 0x200000"), NULL },
    { _TEXT("rd.exe"), _T("fs/_tp0.map -startpos 0x100 -size 0x100"), NULL },
    
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -r"), NULL },
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -r -w"), NULL },
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -r -w -x"), NULL },
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -r -x -w"), NULL },
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -w"), NULL },
    { _TEXT("rd.exe"), _T("fs/nosuch.fil -x"), NULL },

    { _TEXT("gs.exe"), _T("fs/_tp0.map -r"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -w -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -x -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -w -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -x -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -x"), NULL },

    { _TEXT("ss.exe"), _T("fs/_tp0.map -r  -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/_tp0.map -r -w -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/_tp0.map -r -w -x -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/_tp0.map -r -x -w -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/_tp0.map -w -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/_tp0.map -x -size 0x100000"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -r -size 0x0"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -r -w -size 0x00"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -r -w -x -size 0x000000"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -r -x -w -size 0x00"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -w -size 0x10"), NULL },
    { _TEXT("ss.exe"), _T("fs/nosuch.fil -x -size 0x10"), NULL },

    { _TEXT("gs.exe"), _T("fs/_tp0.map -r"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -w -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -r -x -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/_tp0.map -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -w -x"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -r -x -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -w"), NULL },
    { _TEXT("gs.exe"), _T("fs/nosuch.fil -x"), NULL },


    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 1 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 1 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 1 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 1 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 0 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 0 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 0 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 0 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 0 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 0 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 1 -broadcast 0 -start 0 -estimate 0 -deadline 0 -threads 20 -endprev 0 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 0 -start 0 -estimate 100 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 0 -start 0 -estimate 100 -deadline 200 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 0 -start 0 -estimate 100 -deadline 0 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 0 -start 0 -estimate 100 -deadline 200 -threads 20 -endprev 1 -record 0"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 1 -start 0 -estimate 100 -deadline 0 -threads 20 -endprev 0 -record 1"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 1 -start 0 -estimate 100 -deadline 200 -threads 20 -endprev 0 -record 1"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 1 -start 0 -estimate 100 -deadline 0 -threads 20 -endprev 0 -record 1"), NULL },
    { _TEXT("cb.exe"), _T("-mutex 0 -broadcast 1 -start 0 -estimate 100 -deadline 200 -threads 20 -endprev 0 -record 1"), NULL },
#endif

    /* Performance tests here, after we checked everything works ok
     */
    { _TEXT("ttcreate.exe"), NULL, NULL },
    { _TEXT("tcswitch.exe"), _TEXT("3 3"), NULL },

/* The following tests are known to have trouble on the simulator.
 * That's ok as they test real-time, that isn't simulated properly.
 * Obviously they should work everywhere else!
 */
    { _TEXT("constrn2.exe"), _TEXT("2 2"), NULL },
    { _TEXT("tmutex.exe"), NULL, NULL },
    { _TEXT("sleeptst.exe"), NULL, NULL },
    { _TEXT("constrnt.exe"), NULL, NULL },
#if defined(i386)
/* These tests use privileged instructions and do not work on the simulator */
    { _TEXT("mmvm.exe"), NULL, NULL, F_LEAK_REF | F_DONT_WAIT | F_DONT_SIM  },
    { _TEXT("syscall.exe"), NULL, NULL, F_LEAK_REF | F_DONT_WAIT| F_DONT_SIM },
    { _TEXT("mmvmtest.exe"), NULL, NULL, F_LEAK_REF | F_DONT_SIM },
#endif
};
#define NTESTS (sizeof(AllTests)/sizeof(AllTests[0]))

BOOL OnSimulator = FALSE;

ADDR_SIZE CheckMem( UINT FreeWas, BOOL IssueWarning)
{
    struct MACHINE_INFO mi;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();

    /* Sometimes we preempt the thread that just terminated
     * the test, before it gets a chance to die and release
     * its stack & descriptor.
     * This mini-sleep should take care of that
     */
    TIME t;
    int i = 0;

 ReSleep:
    Int32ToInt64(t,TIME_RELATIVE(TIME_MICROS(200)));
    SleepUntil(t);
    
    /* See how much free memory we have
     */
    sc = pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi );
    if (FAILED(sc)) {
        _tprintf(_TEXT("GENERIC machine info FAILED (x%x)"), sc);
        exit(EXIT_FAILURE);
    }

    /* Compare with amount before the test was run.
     * FREEWAS is zero the first time around, ignore.
     */
    if (FreeWas && IssueWarning &&
        FreeWas != mi.AvailablePhysicalMemory) {

        /* limage.exe runs himom.exe without waiting..
         * and we also might preempt some terminating
         * thread in the test...so try again 4 times
         */
        if (i++ < 4)
            goto ReSleep;

        /* It really looks like a leak. Say so.
         */
        _tprintf(_TEXT("Possible MEMORY LEAK of %d bytes. (now %d, was %d)\n\r"),
                 FreeWas - mi.AvailablePhysicalMemory,
                 mi.AvailablePhysicalMemory,
                 FreeWas);
    }    

    /* Take a note if we are on a simulator
     */
    OnSimulator = (mi.OEMId == ~0);

    /* Return how much available memory there is now
     */
    return mi.AvailablePhysicalMemory;
}

int _tmain(int argc, _TCHAR **argv)
{
    PIPROCESS pPrc = CurrentProcess(), pSpawn = NULL;
    PIMODULE pMod;
    SCODE sc;
    INT i, nPass, RetVal;
    TEST_CASE *p;
    TIME Ts, Te, TOverall;
    ADDR_SIZE AvailWas, AvailIs;

    UnusedParameter(argc);
    UnusedParameter(argv);

    _tprintf(_TEXT("Running all %d regression tests.\n\r"), (UINT) NTESTS);

    AvailWas = CheckMem(0, FALSE);
    AvailIs = AvailWas;

    nPass = 0;

    TOverall = CurrentTime();

    for (i = 0; i < NTESTS; i++) {
        p = &AllTests[i];

        if ((p->Flags & F_DONT_SIM) && OnSimulator) {
            /* Count as pass but say so
             */
            _tprintf(_TEXT("[SIMULATOR] Skipping test[%d]: %s(%s)...assume a pass\n\r"), i,
                     p->ImageName,
                     p->Arguments);
            nPass++;
            continue;
        }

        _tprintf(_TEXT("Starting test[%d]: %s(%s)...\n\r"), i,
                 p->ImageName,
                 p->Arguments);

        if (p->ScreenResult)
            _tprintf(_TEXT("The expected output is:\n\r%s\n\r"),
                   p->ScreenResult);

        Ts = CurrentTime();
        sc = pPrc->v->LoadImage(pPrc,
                                p->ImageName,
                                p->Arguments,
                                LOAD_IMAGE_CREATE_THREAD|LOAD_IMAGE_CALL_ENTRY_POINT,
                                &pMod);

        if (sc != S_OK) {
            _tprintf(_TEXT("Could not run test %s(%s) sc = x%x\n\r"),
                     p->ImageName,
                     p->Arguments,
                     sc);
            _tprintf(_TEXT("Regression testing utterly FAILED\n\r"));
            exit(EXIT_FAILURE);
        }

        sc = pMod->v->QueryInterface(pMod, &IID_IProcess, (void **) &pSpawn);
        if (sc != S_OK) {
            _tprintf(_TEXT("Could not QI test %s(%s) sc = x%x\n\r"),
                     p->ImageName,
                     p->Arguments,
                     sc);
            _tprintf(_TEXT("Regression testing utterly FAILED\n\r"));
            exit(EXIT_FAILURE);
        }
        RELEASE_NULL(pMod);

        /* Wait for the test to complete, unless.
         * We start off holding one ref.
         */
        if ((p->Flags & F_DONT_WAIT) == 0) {
            pSpawn->v->WaitFor(pSpawn, TIME_FOREVER);

            /* Its done, get the exit code
             */
            RetVal = EXIT_FAILURE;
            RetVal = pSpawn->v->GetValue(pSpawn, PROCESS_VALUE_EXIT_CODE);
        }

        /* Release last reference, unless we have to leak it 
         * If we do have to leak, account for it in AvailXXX
         */
        if (p->Flags & F_LEAK_REF) {

            ADDR_SIZE AvailNow;

            AvailNow = CheckMem(AvailIs, FALSE);
            AvailWas -= AvailIs - AvailNow;
            AvailIs = AvailNow;
        } else {
            RELEASE(pSpawn);
        }

        /* If we are not waiting dont print bogosities
         */
        if (p->Flags & F_DONT_WAIT) {
            /* Count it as pass, even though we dont really know
             */
            nPass++;
            continue;
        }

        Te = TimeSubtract(CurrentTime(),Ts);
        Te = Int64DividedByInt32(Te,10);

        /* Say how long it took.
         */
        _tprintf(_TEXT("Regression test %s(%s) completed in %lld usecs.\n\r"),
                 p->ImageName,
                 p->Arguments,
                 Te);

        AvailIs = CheckMem(AvailIs, TRUE);

        /* Keep tally
         */
        if (RetVal == EXIT_SUCCESS) 
            nPass++;
        else
            p->Flags = F_FAILED;
    }

    TOverall = TimeSubtract(CurrentTime(), TOverall);
    TOverall = Int64DividedByInt32(TOverall,10);
    _tprintf(_TEXT("This took %lld usec total\n\r"), TOverall);


    AvailIs = CheckMem(AvailIs, TRUE);

    if (AvailIs != AvailWas)
        _tprintf(_TEXT("It seems we LEAKED %d bytes of memory (%d - %d)\n\r"),
                 AvailWas - AvailIs, AvailWas, AvailIs);
    {
        PIHEAP pHeap = CurrentHeap();
        ADDR_SIZE MaxUsed = 0;
        (void) pHeap->v->Status(pHeap,NULL,NULL,NULL,&MaxUsed);
        _tprintf(_TEXT("Maximum heap memory usage was %d bytes\n\r"), MaxUsed);
    }

    _tprintf(_TEXT("\n\rRegression testing "));
    if (nPass == NTESTS)
        _tprintf(_TEXT("PASSED\n\r"));
    else {

        _tprintf(_TEXT("found %d FAILED tests of %d total.\n\r\n\r"),
                 NTESTS-nPass,NTESTS);

        _tprintf(_TEXT("The following tests FAILED:\n\r"));
        for (i = 0; i < NTESTS; i++) {
            p = &AllTests[i];
            if (p->Flags == F_FAILED)
                _tprintf(_TEXT("%s "),p->ImageName);
        }
        _tprintf(_TEXT("\n\r"));
    }

    exit ((nPass == NTESTS)?EXIT_SUCCESS:EXIT_FAILURE);
    return 0;/*amusme*/
}







