/* Tests for src/base/condition2.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define SIM 1

#include <stdio.h>
#include <mmlite.h>
#include <base/condition2.h>
#include <base/thread2.h>
#include <base/kconstraint2.h>
#include <util/test.h>
#include <util/fake_sched.h>

PRIVATE CONDITION CD1;
PRIVATE CONDITION CD2;
PRIVATE CONDITION CD3;
PRIVATE PCONDITION pCD1 = &CD1;
PRIVATE PCONDITION pCD2 = &CD2;
PRIVATE PCONDITION pCD3 = &CD3;
PRIVATE PITHREAD pThdCurrent = NULL;
PRIVATE PITHREAD threads[3];
PRIVATE BOOL wait_signals[3];
PRIVATE BOOL end_signals[3];
struct IScheduler* pTheScheduler;
extern PTR DeferredConditionSignals;

BOOL Condition_TestInitializer(void);
BOOL Test_Condition_Init(void);
BOOL Test_Condition_Destroy(void);
BOOL Test_Condition_Signal(void);
BOOL Test_Condition_Broadcast(void);
BOOL Test_Condition_WaitAndBeginConstraint(void);
BOOL Test_Condition_WaiterQueue(void);
BOOL Test_Condition_Next(void);
BOOL Test_Condition_Signalled(void);
BOOL Test_Condition_Queued(void);
BOOL Test_Condition_Push(void);
BOOL Test_Condition_Pop(void);
BOOL Test_Condition_DrainDeferred(void);

void THREAD_LINKAGE ConditionTestThread(THREAD_ARGUMENT arg);
void ClearConditionStack(
    PCOND pCond
    );

int _tmain(void) {
  TestFunctionPair ConditionTests[] = {
    {Condition_TestInitializer,             ""},
    {Test_Condition_Init,                   "Init"},
    {Test_Condition_Destroy,                "Destroy"},
    {Test_Condition_Signal,                 "Signal"},
    {Test_Condition_Broadcast,              "Broadcast"},
    {Test_Condition_WaitAndBeginConstraint, "Wait"},
    {Test_Condition_WaiterQueue,            "WaiterQueue"},
    {Test_Condition_Next,                   "Next"},
    {Test_Condition_Signalled,              "Signalled"},
    {Test_Condition_Queued,                 "Queued"},
    {Test_Condition_Push,                   "Push"},
    {Test_Condition_Pop,                    "Pop"},
    {Test_Condition_DrainDeferred,          "DrainDeferred"}
  };

  pTheNameSpace = CurrentNameSpace();

  if (pTheNameSpace->v->Bind(pTheNameSpace, _TEXT("cb_sched2"),
			     0, (PIUNKNOWN*)&pTheScheduler) != S_OK) {
    printf("Failed to retrieve scheduler.");
    return 1;
  }

  pThdCurrent = CurrentThread();

  RunTestSuite("Condition Test Suite", ConditionTests,
	       GetTestFunctionArraySize(ConditionTests));

  return 0;
}
    
BOOL Condition_TestInitializer(void)
{
  Condition_Init(pCD1);
  Condition_Init(pCD2);
  Condition_Init(pCD3);

  return TRUE;
}

BOOL Test_Condition_Init(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, Condition_WaiterQueueIsEmpty(pCD1),
			     "waiter queue is initially empty");

  success &= compare_integer(NULL, Condition_GetNext(pCD1),
			     "next is initially NULL");

  success &= compare_integer(TRUE, Condition_IsAlone(pCD1),
			     "condition is initially alone");

  return success;
}

BOOL Test_Condition_Destroy(void)
{
  BOOL success = TRUE;

  Condition_Destroy(pCD1);

  success &= compare_integer(TRUE, Condition_WaiterQueueIsEmpty(pCD1),
			     "waiter queue is empty");

  success &= compare_integer(TRUE, Condition_IsBad(pCD1),
			     "destroyed conditions are bad");

  success &= compare_integer(TRUE, Condition_IsSignalled(pCD1),
			     "condition is signalled");

  success &= compare_integer(FALSE, Condition_IsQueued(pCD1),
			     "condition is not queued");

  return success;
}

BOOL Test_Condition_Signal(void)
{
  BOOL success = TRUE;

  threads[0] = NULL;
  threads[1] = NULL;
  threads[2] = NULL;
  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)0);

  while (threads[0] == NULL) {
    MilliSleep(1);
  }
    
  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)1);

  while (threads[1] == NULL) {
    MilliSleep(1);
  }

  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)2);
    
  while (threads[2] == NULL) {
    MilliSleep(1);
  }

  wait_signals[0] = TRUE;

  while (wait_signals[0] == TRUE) {
    MilliSleep(1);
  }

  wait_signals[1] = TRUE;

  while (wait_signals[1] == TRUE) {
    MilliSleep(1);
  }

  wait_signals[2] = TRUE;

  while (wait_signals[2] == TRUE) {
    MilliSleep(1);
  }

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread is waiting for condition");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread is waiting for condition");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread is waiting for condition");

  Condition_Signal(pCD1);

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread has been signalled");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread is waiting for condition");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread is waiting for condition");

  Condition_Signal(pCD1);

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread has been signalled");

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread has been signalled");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread is waiting for condition");

  Condition_Signal(pCD1);

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread has been signalled");

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread has been signalled");

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread has been signalled");

  end_signals[0] = TRUE;

  while (threads[0] != NULL) {
    MilliSleep(1);
  }

  end_signals[1] = TRUE;

  while (threads[1] != NULL) {
    MilliSleep(1);
  }

  end_signals[2] = TRUE;

  while (threads[2] != NULL) {
    MilliSleep(1);
  }

  return success;
}

BOOL Test_Condition_Broadcast(void)
{
  BOOL success = TRUE;

  threads[0] = NULL;
  threads[1] = NULL;
  threads[2] = NULL;
  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)0);

  while (threads[0] == NULL) {
    MilliSleep(1);
  }
    
  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)1);

  while (threads[1] == NULL) {
    MilliSleep(1);
  }

  CreateAThread(ConditionTestThread, (THREAD_ARGUMENT)2);
    
  while (threads[2] == NULL) {
    MilliSleep(1);
  }

  wait_signals[0] = TRUE;

  while (wait_signals[0] == TRUE) {
    MilliSleep(1);
  }

  wait_signals[1] = TRUE;

  while (wait_signals[1] == TRUE) {
    MilliSleep(1);
  }

  wait_signals[2] = TRUE;

  while (wait_signals[2] == TRUE) {
    MilliSleep(1);
  }

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread is waiting for condition");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread is waiting for condition");

  success &= compare_integer(TRUE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread is waiting for condition");
    
  Condition_Broadcast(pCD1);

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[0]),
			     "first thread no longer waits for condition");

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[1]),
			     "second thread no longer waits for condition");

  success &= compare_integer(FALSE, Thread_IsWaitingForCondition(threads[2]),
			     "third thread no longer waits for condition");
  end_signals[0] = TRUE;

  while (threads[0] != NULL) {
    MilliSleep(1);
  }

  end_signals[1] = TRUE;

  while (threads[1] != NULL) {
    MilliSleep(1);
  }

  end_signals[2] = TRUE;

  while (threads[2] != NULL) {
    MilliSleep(1);
  }   

  return success;
}

BOOL Test_Condition_WaitAndBeginConstraint(void)
{
  BOOL success = TRUE;

  return success;
}

BOOL Test_Condition_WaiterQueue(void)
{
  BOOL success = TRUE;

  return success;
}

BOOL Test_Condition_Next(void)
{
  BOOL success = TRUE;

  Condition_SetNext(pCD1, 0x2222);

  compare_integer(0x2220, Condition_GetNext(pCD1),
		  "set next == get next");

  return success;
}

BOOL Test_Condition_Signalled(void)
{
  BOOL success = TRUE;

  Condition_SetSignalled(pCD1);

  success &= compare_integer(TRUE, Condition_IsSignalled(pCD1),
			     "set signalled = get signalled");

  return success;
}

BOOL Test_Condition_Queued(void)
{
  BOOL success = TRUE;

  Condition_SetQueued(pCD1);

  success &= compare_integer(TRUE, Condition_IsQueued(pCD1),
			     "set queued == get queued");

  return success;
}

BOOL Test_Condition_Push(void)
{
  BOOL success = TRUE;

  Condition_Push((PCOND)pCD1);

  success &= compare_integer(pCD1, DeferredConditionSignals,
			     "newly pushed condition is the stack head");

  success &= compare_integer(NULL, Condition_GetNext(pCD1),
			     "stack is still NULL-terminated");

  Condition_Push((PCOND)pCD2);
				   
  success &= compare_integer(pCD2, DeferredConditionSignals,
			     "newly pushed condition is the stack head");

  success &= compare_integer(pCD1, Condition_GetNext(pCD2),
			     "old stack head is next of new stack head");

  success &= compare_integer(NULL, Condition_GetNext(pCD1),
			     "stack is still NULL-terminated");

  Condition_Push((PCOND)pCD3);
				   
  success &= compare_integer(pCD3, DeferredConditionSignals,
			     "newly pushed condition is the stack head");

  success &= compare_integer(pCD2, Condition_GetNext(pCD3),
			     "old stack head is next of new stack head");

  success &= compare_integer(NULL, Condition_GetNext(pCD1),
			     "stack is still NULL-terminated");

  ClearConditionStack(DeferredConditionSignals);

  DeferredConditionSignals = NULL;

  return success;
}

BOOL Test_Condition_Pop(void)
{
  BOOL success = TRUE;

  Condition_Push((PCOND)pCD1);
  Condition_Push((PCOND)pCD2);
  Condition_Push((PCOND)pCD3);

  DeferredConditionSignals = Condition_Pop(DeferredConditionSignals);

  success &= compare_integer(pCD2, DeferredConditionSignals,
			     "new stack head is old stack head's next");

  success &= compare_integer(TRUE, Condition_IsAlone(pCD3),
			     "popped condition is alone");

  DeferredConditionSignals = Condition_Pop(DeferredConditionSignals);

  success &= compare_integer(pCD1, DeferredConditionSignals,
			     "new stack head is old stack head's next");

  success &= compare_integer(TRUE, Condition_IsAlone(pCD2),
			     "popped condition is alone");

  DeferredConditionSignals = Condition_Pop(DeferredConditionSignals);

  success &= compare_integer(NULL, DeferredConditionSignals,
			     "new stack head is old stack head's next");

  success &= compare_integer(TRUE, Condition_IsAlone(pCD1),
			     "popped condition is alone");


  return success;
}

BOOL Test_Condition_DrainDeferred(void)
{
  BOOL success = TRUE;

  return success;
}

void THREAD_LINKAGE ConditionTestThread(THREAD_ARGUMENT arg)
{
  int me = (int)arg;
  end_signals[me] = FALSE;
  wait_signals[me] = FALSE;
  threads[me] = CurrentThread();
  while (!end_signals[me]) {

    while (!wait_signals[me]) {
      if (end_signals[me] == TRUE) {
	goto Exit;
      }
      MilliSleep(1);
    }

    wait_signals[me] = FALSE;
    Condition_Wait(pCD1, NULL, NULL);
  }

 Exit:
  threads[me] = NULL;
}

void ClearConditionStack(
    PCOND pCond
    )
{
  PCOND pCondCurrent;
  PCOND pCondNext;

  pCondCurrent = pCond;

  while (pCondCurrent != NULL) {
    pCondNext = Condition_GetNext(pCondCurrent);
    Condition_SetNext(pCondCurrent, NULL);
    pCondCurrent = pCondNext;
  }
}
