/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#if 0
#define VERBOSE(x) printf x
#else
#define VERBOSE(x)
#endif

SCODE DirCreateFiles(PINAMESPACE pNs, UINT MaxFiles)
{
  PIUNKNOWN pUnk;
  PIFILE pF;
  SCODE sc = NO_ERROR;
  UINT i, nWritten;
#define NAMECH 12
#define NAMESZ (NAMECH*sizeof(_TCHAR))
  _TCHAR FileName[NAMECH];

  /* First pass, create 'em
   */
  memset(FileName,0,NAMESZ);
  _tprintf(_TEXT("Creating %d files....\n\r"), MaxFiles);

  for (i = 0; i < MaxFiles; i++) {
    _sntprintf(FileName,NAMESZ,_TEXT("f%08x"),i);

    /* Create teh file
     */
    VERBOSE((_TEXT("Creating file %s\n"), FileName));
    sc = pNs->v->Bind(pNs, FileName, NAME_SPACE_WRITE | NAME_SPACE_CREATE, &pUnk);
    if (FAILED(sc)) {
      _tprintf(_TEXT("FAILED to create %s (sc=x%x).\n\r"), FileName, sc);
      goto Out;
    }
  
    sc = pUnk->v->QueryInterface(pUnk, &IID_IFile, (void **)&pF);
    pUnk->v->Release(pUnk);
    if (FAILED(sc)) {
      _tprintf(_TEXT("%s not a IFile (sc=x%x).\n\r"), FileName, sc);
      goto Out;
    }
  
    /* Ptu a byte on it
     */
    VERBOSE((_TEXT("Writing to file %s\n"), FileName));
    nWritten = 0;
    sc = pF->v->WriteAt(pF, UINT64_ZERO, (BYTE*) &FileName, NAMESZ, &nWritten);
    /* Close teh file, regardless
     */
    pF->v->Release(pF);

    if (FAILED(sc) || (nWritten != NAMESZ)) {
      _tprintf(_TEXT("FAILED to write %d bytes to %s (sc=x%x, nW=%d).\n\r"),
	       NAMESZ, FileName, sc, nWritten);
      goto Out;
    }

  }
 Out:

  /* Second pass, delete 'em
   */
  if (i < MaxFiles) MaxFiles = i; /* did it break early? */
  _tprintf(_TEXT("Deleting %d files....\n\r"), MaxFiles);
  for (i = 0; i < MaxFiles; i++) {
    _sntprintf(FileName,NAMESZ,_TEXT("f%08x"),i);

    /* Unlink the file
     */
    VERBOSE((_TEXT("Deleting file %s\n"), FileName));
    sc = pNs->v->Unregister(pNs, FileName);
    if (FAILED(sc)) {
      _tprintf(_TEXT("FAILED to Unregister %s (sc=x%x).\n\r"), FileName, sc);
    }
  }

  return sc;
}

SCODE DirTest(_TCHAR *RootName, UINT MaxFiles)
{
  PINAMESPACE pNs = CurrentNameSpace();
  PIUNKNOWN pUnk = NULL;
  SCODE sc;
  _TCHAR *DirName = RootName;
  BOOL RemoveAtEnd = FALSE;

  /* See if it exists first
   */
  sc = pNs->v->Bind(pNs, DirName,
                    NAME_SPACE_WRITE | NAME_SPACE_EXTENSION,
                    &pUnk);
  if (FAILED(sc)) {
      _tprintf(_TEXT("%s not found (sc=x%x) will try to create it...\n\r"), DirName, sc);
      sc = pNs->v->Bind(pNs, DirName,
                    NAME_SPACE_WRITE | NAME_SPACE_CREATE | NAME_SPACE_EXTENSION,
                    &pUnk);
      if (FAILED(sc)) {
          _tprintf(_TEXT("FAILED to bind to %s (sc=x%x).\n\r"), DirName, sc);
          goto Out;
      }
      RemoveAtEnd = TRUE;
  }
  
  sc = pUnk->v->QueryInterface(pUnk, &IID_INameSpace, (void **)&pNs);
  pUnk->v->Release(pUnk);
  if (FAILED(sc)) {
    _tprintf(_TEXT("%s not a INameSpace (sc=x%x).\n\r"), DirName, sc);
    goto Out;
  }
  
  /* Now create all them files */
  sc = DirCreateFiles(pNs, MaxFiles);
  if (FAILED(sc))
    _tprintf(_TEXT("CreateFiles failed with sc=x%x.\n\r"), sc);

  pNs->v->Release(pNs);
 Out:
  if (RemoveAtEnd) {
      pNs = CurrentNameSpace();
      sc = pNs->v->Unregister(pNs, DirName);
      if (FAILED(sc))
          _tprintf(_TEXT("Unregister of %s failed with sc=x%x.\n\r"), DirName, sc);
  }
  if (DirName != RootName)
      free(DirName);
  return sc;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    _TCHAR *RootName = _TEXT("fat/test");
    UINT MaxFiles = 1 + (2048/32); /* overflow one FAT sector worth */
    
    _tprintf(_TEXT("main() @%x\n\r"), (ADDRESS) _tmain);

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x), test FAILED\n\r"), sc);
        return (EXIT_FAILURE);
    }

    if (argc > 1) {
      RootName = argv[1];
    }
    _tprintf(_TEXT("Will run this test in %s\n\r"), RootName);

    if (argc > 2) {
      MaxFiles = _ttoi(argv[2]);
    }
    _tprintf(_TEXT("Will create  %d files per directory.\n\r"), MaxFiles);

    sc = DirTest(RootName,MaxFiles);

    (void) FreeCmdLine(&argc,&argv);

    if (FAILED(sc)) {
      _tprintf(_TEXT("Test FAILED.\n\r"));
      return (EXIT_FAILURE);
    }
    _tprintf(_TEXT("Test PASSED ok.\n\r"));
    return (EXIT_SUCCESS);
}
