/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

/* Create a process, loading a test image.
 * Tests what args the process gets, 
 * if it actually does run to completion,
 * and that it returns the proper exit code.
 */
_TCHAR *Args = _TEXT("Some args for you");
_TCHAR ActualArgs[32] = {0,};

int _tmain(int argc, _TCHAR **argv)
{
    PIPROCESS pPrc = CurrentProcess(), pSpawn = NULL;
    PIMODULE pMod = NULL;
    INT RetVal;
    SCODE sc;

    UnusedParameter(argc);
    UnusedParameter(argv);

    sc = pPrc->v->LoadImage(pPrc,
                            _TEXT("himom.exe"),
                            Args,
                            LOAD_IMAGE_CREATE_THREAD|LOAD_IMAGE_CALL_ENTRY_POINT,
                            &pMod);

    _tprintf(_TEXT("LoadImage got x%x\n\r"), sc);
    if (sc != NO_ERROR) goto Bad;

    /* Check it got the right arguments
     */
    sc = pMod->v->GetArgs(pMod,ActualArgs,sizeof ActualArgs);
    if (sc != NO_ERROR) {
        _tprintf(_TEXT("limage: FAILED GetArgs got x%x\n\r"), sc);
        goto Bad;
    }

    if (_tcscmp(Args,ActualArgs) != 0) {
        _tprintf(_TEXT("limage: FAILED args '%s' != '%s'\n\r"),
                 ActualArgs,Args);
        goto Bad;
    }

    sc = pMod->v->QueryInterface(pMod, &IID_IProcess, (void **) &pSpawn);
    if (sc != NO_ERROR) {
        _tprintf(_TEXT("limage: FAILED QI for IProcess got x%x\n\r"), sc);
        goto Bad;
    }

    pMod->v->Release(pMod);
    pMod = NULL;

    /* Wait till it terminates
     */
    sc = pSpawn->v->WaitFor(pSpawn, TIME_FOREVER);
    if (sc != NO_ERROR) {
        _tprintf(_TEXT("limage: FAILED WaitFor got x%x\n\r"), sc);
        goto Bad;
    }

    /* Check the exit code
     */
    RetVal = pSpawn->v->GetValue(pSpawn, PROCESS_VALUE_EXIT_CODE);

    /* Just let it die.
     * BUGBUG Should we check this is the last ref also ?
     */
    pSpawn->v->Release(pSpawn);

    _tprintf(_TEXT("LoadImage test PASSED\n\r"));
    exit(EXIT_SUCCESS);

 Bad:
    if (pMod)
        pMod->v->Release(pMod);
    if (pSpawn)
        pSpawn->v->Release(pSpawn);

    _tprintf(_TEXT("LoadImage test FAILED\n\r"));
    exit(EXIT_FAILURE);
}
