/* Tests for list and node interface defined in include/base/list.h
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <base/thread2.h>
#include <util/test.h>
#include <util/list-double.h>

/* Global variables for testing */

typedef struct _TestStruct {
  int value;
} TestStruct, *pTestStruct;

#define pTS(_pts_) ((pTestStruct)(_pts_))
#define StructGetValue(_pts_) (pTS(_pts_)->value)
#define StructSetValue(_pts_, _value_) (StructGetValue(_pts_) = (_value_))
#define STRUCT_TRANSFORM_VALUE 22

PRIVATE DNODE nNode1;
PRIVATE DNODE nNode2;
PRIVATE DNODE nNode3;
PRIVATE PDNODE pnNode1 = &nNode1;
PRIVATE PDNODE pnNode2 = &nNode2;
PRIVATE PDNODE pnNode3 = &nNode3;
PRIVATE DLIST lList1;
PRIVATE PDLIST plList1 = &lList1;

PRIVATE PDNODE pnTemp;
PRIVATE PTR pTemp;
PRIVATE PDNODE pnHead;

PRIVATE TestStruct ts1;
PRIVATE TestStruct ts2;
PRIVATE TestStruct ts3;

PRIVATE pTestStruct pts1 = &ts1;
PRIVATE pTestStruct pts2 = &ts2;
PRIVATE pTestStruct pts3 = &ts3;

/* Function prototypes for tests */

BOOL DNode_TestInitializer(void);
BOOL Test_DNode_Init(void);
BOOL Test_DNode_GetNext(void);
BOOL Test_DNode_GetData(void);
BOOL Test_DNode_IsAlone(void);
BOOL Test_DNode_Push(void);
BOOL Test_DNode_Pop(void);
BOOL Test_DNode_InsertAfter(void);
BOOL Test_DNode_RemoveAfter(void);
BOOL Test_DNode_StackRemove(void);
BOOL Test_DNode_StackClear(void);

BOOL DList_TestInitializer(void);
BOOL Test_DList_Init(void);
BOOL Test_DList_Clear(void);
BOOL Test_DList_IsEmpty(void);
BOOL Test_DList_TailAdd(void);
BOOL Test_DList_HeadAdd(void);
BOOL Test_DList_HeadRemove(void);
BOOL Test_DList_TailRemove(void);
BOOL Test_DList_Remove(void);
BOOL Test_DList_FindData(void);
BOOL Test_DList_FindDataAndRemove(void);
BOOL Test_StructGT(void);
BOOL Test_DList_FindDataMax(void);
BOOL Test_DList_RemoveDataMax(void);
BOOL Test_StructTransform(void);
BOOL Test_StructTransformWithArg(void);
BOOL Test_DList_TransformData(void);
BOOL Test_DList_TransformDataWithArg(void);
BOOL Test_DList_InsertSortedDescending(void);

/* Prototypes for auxiliary functions */

BOOL StructGT(
    PTR pts1,
    PTR pts2
    );

void StructTransform(
    PTR pts
    );

void StructTransformWithArg(
    PTR pts,
    PTR ptsArg
    );

int _tmain(void) {
  TestFunctionPair DNodeTestFunctions[] = {
    {DNode_TestInitializer,          ""},
    {Test_DNode_Init,                "DNode_Init"},
    {Test_DNode_GetData,              "DNode_GetData"},
    {Test_DNode_GetNext,              "DNode_GetNext"},
    {Test_DNode_IsAlone,              "DNode_IsAlone"},
    {Test_DNode_Push,                 "DNode_Push"},
    {Test_DNode_Pop,                  "DNode_Pop"},
    {Test_DNode_InsertAfter,          "DNode_InsertAfter"},
    {Test_DNode_RemoveAfter,          "DNode_RemoveAfter"},
    {Test_DNode_StackRemove,          "DNode_StackRemove"},
    {Test_DNode_StackClear,           "DNode_StackClear"}
  };

  TestFunctionPair DListTestFunctions[] = {
    {DList_TestInitializer, ""},
    {Test_DList_Init,                  "DList_Init"},
    {Test_DList_Clear,                  "DList_Clear"},
    {Test_DList_IsEmpty,                "DList_IsEmpty"},
    {Test_DList_TailAdd,                "DList_TailAdd"},
    {Test_DList_HeadAdd,                "DList_HeadAdd"},
    {Test_DList_HeadRemove,             "DList_HeadRemove"},
    {Test_DList_TailAdd,                "DList_TailAdd"},
    {Test_DList_TailRemove,             "DList_TailRemove"},
    {Test_DList_Remove,                 "DList_Remove"},
    {Test_DList_FindData,               "DList_FindData"},
    {Test_DList_FindDataAndRemove,      "DList_FindDataAndRemove"},
    {Test_StructGT,                     "StructGT"},
    {Test_DList_FindDataMax,            "DList_FindDataMax"},
    {Test_DList_RemoveDataMax,          "DList_RemoveDataMax"},
    {Test_StructTransform,              "StructTransform"},
    {Test_StructTransformWithArg,       "StructTransformWithArg"},
    {Test_DList_TransformData,          "DList_TransformData"},
    {Test_DList_TransformDataWithArg,   "DList_TransformDataWithArg"},
    {Test_DList_InsertSortedDescending, "DList_InsertSortedDescending"}
  };

  RunTestSuite("Doubly-Linked Node Test Suite", DNodeTestFunctions,
	       GetTestFunctionArraySize(DNodeTestFunctions));
  RunTestSuite("Doubly-Linked List Test Suite", DListTestFunctions,
	       GetTestFunctionArraySize(DListTestFunctions));

  return 0;
}

BOOL DNode_TestInitializer(void)
{
  DNode_Init(pnNode1, NULL);
  DNode_Init(pnNode2, NULL);
  DNode_Init(pnNode3, NULL);
  pTemp = NULL;
  pnTemp = NULL;
  pnHead = NULL;

  return TRUE;
}

BOOL Test_DNode_Init(void)
{
  BOOL success = TRUE;

  success &= compare_integer(NULL, DNode_GetData(pnNode1),
			     "sets data pointer to NULL");

  success &= compare_integer(pnNode1, DNode_GetNext(pnNode1),
			     "sets next pointer to itself");

  success &= compare_integer(pnNode1, DNode_GetPrev(pnNode1),
			     "sets prev pointer to itself");

  return success;
}

BOOL Test_DNode_GetData(void)
{
  BOOL success = TRUE;

  DNode_SetData(pnNode1, (PTR)0x2222);

  success &= compare_integer(0x2222, DNode_GetData(pnNode1),
			     "set data == get data");

  return success;
}

BOOL Test_DNode_GetNext(void)
{
  BOOL success = TRUE;

  DNode_SetNext(pnNode1, pnNode2);

  success &= compare_integer(pnNode2, DNode_GetNext(pnNode1),
			     "set next == get next");

  return success;
}

BOOL Test_DNode_IsAlone(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode1),
			     "initialized node is alone");

  return success;
}

BOOL Test_DNode_Push(void)
{
  BOOL success = TRUE;

  pnHead = DNode_Push(pnHead, pnNode3);

  success &= compare_integer(pnNode3, pnHead,
			     "new node in empty stack is new head");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "new node in empty stack is alone");

  pnHead = DNode_Push(pnHead, pnNode2);

  success &= compare_integer(pnNode3, DNode_GetNext(pnNode2),
			     "new node in non-empty stack updates next");

  success &= compare_integer(pnNode2, pnHead,
			     "new node in non-empty stack updates head");

  return success;
}

BOOL Test_DNode_Pop(void)
{
  BOOL success = TRUE;
  PDNODE pnOldHead;

  pnHead = DNode_Push(pnHead, pnNode3);
  pnHead = DNode_Push(pnHead, pnNode2);

  pnOldHead = pnHead;
  pnHead = DNode_Pop(pnOldHead);

  success &= compare_integer(pnNode3, pnHead,
			     "popping non-empty stack updates head");

  success &= compare_integer(pnNode2, pnOldHead,
			     "popped node equals last node pushed");

  pnOldHead = pnHead;
  pnHead = DNode_Pop(pnOldHead);

  success &= compare_integer(NULL, pnHead,
			     "popping all nodes results in empty stack");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "popped node is alone");

  return success;
}

BOOL Test_DNode_InsertAfter(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  pnHead = DNode_Push(pnHead, pnNode1);
  pnOldEnd = DNode_GetNext(pnNode1);

  DNode_InsertAfter(pnNode1, pnNode2);

  success &= compare_integer(pnNode2, DNode_GetNext(pnNode1),
			     "updates next correctly");

  success &= compare_integer(pnOldEnd, DNode_GetNext(pnNode2),
			     "stack end is still the same");

  return success;
}

BOOL Test_DNode_RemoveAfter(void)
{
  BOOL success = TRUE;

  pnHead = DNode_Push(pnHead, pnNode3);
  pnHead = DNode_Push(pnHead, pnNode2);
  pnHead = DNode_Push(pnHead, pnNode1);

  DNode_RemoveAfter(pnNode1);

  success &= compare_integer(pnNode3, DNode_GetNext(pnNode1),
			     "updates remaining nodes correctly");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode2),
			     "removed node is alone");

  return success;
}

BOOL Test_DNode_StackRemove(void)
{
  BOOL success = TRUE;

  pnHead = DNode_Push(pnHead, pnNode1);
  pnHead = DNode_Push(pnHead, pnNode2);
  pnHead = DNode_Push(pnHead, pnNode3);

  pnHead = DNode_StackRemove(pnHead, pnNode2);

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode2),
			     "removed node is alone");

  success &= compare_integer(pnNode1, DNode_GetNext(pnNode3),
			     "updates remaining nodes correctly");

  success &= compare_integer(pnNode3, pnHead,
			     "head is still the same");

  pnHead = DNode_StackRemove(pnHead, pnNode3);

  success &= compare_integer(pnNode1, pnHead,
			     "head is updated correctly");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "removed head is now alone");

  return success;
}

BOOL Test_DNode_StackClear(void)
{
  BOOL success = TRUE;

  PDNODE pnHead = NULL;

  pnHead = DNode_Push(pnHead, pnNode1);
  pnHead = DNode_Push(pnHead, pnNode2);
  pnHead = DNode_Push(pnHead, pnNode3);

  pnHead = DNode_StackClear(pnHead);

  success &= compare_integer(NULL, pnHead, "new stack head is NULL");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode1), "pnNode1 is alone");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode2), "pnNode2 is alone");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3), "pnNode3 is alone");

  return success;
}

BOOL DList_TestInitializer(void)
{
  DNode_Init(pnNode1, pts1);
  DNode_Init(pnNode2, pts2);
  DNode_Init(pnNode3, pts3);

  StructSetValue(pts1, 1);
  StructSetValue(pts2, 2);
  StructSetValue(pts3, 3);

  pTemp = NULL;
  pnTemp = NULL;
  pnHead = NULL;
  DList_Init(plList1);

  return TRUE;
}

BOOL Test_DList_Init(void)
{
  BOOL success = TRUE;

  success &= compare_integer(NULL, DList_GetHead(plList1),
			     "sets head pointer to NULL");

  return success;
}

BOOL Test_DList_Clear(void)
{
  BOOL success = TRUE;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  DList_Clear(plList1);

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode1),
			     "node 1 is alone");
    
  success &= compare_integer(TRUE, DNode_IsAlone(pnNode2),
			     "node 2 is alone");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "node 3 is alone");

  return success;
}

BOOL Test_DList_IsEmpty(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, DList_IsEmpty(plList1),
			     "initialized list is empty");

  return success;
}

BOOL Test_DList_TailAdd(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);

  success &= compare_integer(pnNode1, DList_GetHead(plList1),
			     "adding to an empty list updates head");

  success &= compare_integer(pnNode1, DList_GetTail(plList1),
			     "adding to an empty list updates tail");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode1),
			     "new node in empty list is alone");

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));
  DList_TailAdd(plList1, pnNode2);

  success &= compare_integer(pnNode2, DNode_GetNext(pnNode1),
			     "new node updates existing nodes");

  success &= compare_integer(pnNode2, DList_GetTail(plList1),
			     "new node updates list tail");

  success &= compare_integer(pnOldEnd,
			     DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_DList_HeadAdd(void)
{
  BOOL success = TRUE;

  DList_HeadAdd(plList1, pnNode1);

  success &= compare_integer(DList_GetHead(plList1),  pnNode1,
			     "adding to an empty list updates head");

  success &= compare_integer(DList_GetTail(plList1),  pnNode1,
			     "adding to an empty list updates tail");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode1),
			     "new node in list is alone");
    
  DList_HeadAdd(plList1, pnNode3);

  success &= compare_integer(pnNode3, DList_GetHead(plList1),
			     "new node is the head of list");

  success &= compare_integer(pnNode1, DNode_GetNext(pnNode3),
			     "new node's next is the old head");

  success &= compare_integer(pnNode3,
			     DNode_GetNext(DList_GetTail(plList1)),
			     "list is still circular");

  return TRUE;
}

/* Uncommenting these tests just fails an assertion.
 * Change the implemention of ListTailAdd if you would like to
 * handle adding duplicate nodes to a list gracefully instead of failing.
 */
/*      DList_TailAdd(plList1, pnNode2); */

/*      compare_integer(pnNode3, DList_GetTail(plList1), */
/*  		    "ListTailAdd: cannot add existing node to list - tail does not change."); */

/*      compare_integer(pnNode1, DNode_GetNext(pnNode3), */
/*  		    "ListTailAdd: cannot add existing node to list - tail's next does not change."); */

/*      compare_integer(pnNode3, DNode_GetNext(pnNode2), */
/*  		    "ListTailAdd: cannot add existing node to list - existing node stays in its place."); */

BOOL Test_DList_HeadRemove(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));

  pnTemp = DList_HeadRemove(plList1);

  success &= compare_integer(pnNode1, pnTemp,
			     "removes the first node");

  success &= compare_integer(pnNode2, DList_GetHead(plList1),
			     "updates the head");

  success &= compare_integer(pnNode3, DList_GetTail(plList1),
			     "tail does not change.");

  success &= compare_integer(TRUE, DNode_IsAlone(pnTemp),
			     "removed node is alone.");

  success &= compare_integer(pnNode2,
			     DNode_GetNext(DList_GetTail(plList1)),
			     "tail's next is new head");

  DList_HeadRemove(plList1);
  pnTemp = DList_HeadRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "last node removed sequentially is tail");

  success &= compare_integer(TRUE, DList_IsEmpty(plList1),
			     "removing all nodes results in empty list");

  return success;
}

BOOL Test_DList_TailRemove(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));
  pnTemp = DList_TailRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "removed node is old tail");

  success &= compare_integer(pnNode2, DList_GetTail(plList1),
			     "updates tail to next-to-last node");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "removed node is alone");

  DList_TailRemove(plList1);

  success &= compare_integer(pnNode1, DList_GetHead(plList1),
			     "only node in list is the head");

  success &= compare_integer(pnNode1, DList_GetTail(plList1),
			     "only node in list is the tail");

  DList_TailRemove(plList1);

  success &= compare_integer(TRUE, DList_IsEmpty(plList1),
			     "removing only node results in empty list");

  return success;
}

BOOL Test_DList_Remove(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));
  DList_Remove(plList1, pnNode2);

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode2),
			     "removed node is alone");

  success &= compare_integer(pnNode3, DNode_GetNext(pnNode1),
			     "updates remaining nodes correctly");

  success &= compare_integer(pnNode1, DList_GetHead(plList1),
			     "head does not change");

  success &= compare_integer(pnNode3, DList_GetTail(plList1),
			     "tail does not change");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  DList_Remove(plList1, pnNode1);

  success &= compare_integer(pnNode3, DList_GetHead(plList1),
			     "head is updated correctly");

  DList_Remove(plList1, pnNode3);

/*   success &= compare_integer(NULL, DList_GetTail(plList1), */
/* 			     "tail is updated correctly"); */

/*   DList_Remove(plList1, pnNode1); */

/*   success &= compare_integer(TRUE, DList_IsEmpty(plList1), */
/* 			     "removing from empty list has no effect"); */

  return success;
}    

BOOL Test_DList_FindData(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));

  success &= compare_integer(pnNode1, DList_FindData(plList1, pts1),
			     "data that is set is found again");

  success &= compare_integer(NULL, DList_FindData(plList1, NULL),
			     "data that is not in the queue is not found");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(pts1, DNode_GetData(pnNode1),
			     "does not change node data");

  return success;
}

BOOL Test_DList_FindDataAndRemove(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));

  pnTemp = DList_FindDataAndRemove(plList1, pts2);
    
  success &= compare_integer(pnNode2, pnTemp,
			     "duplicates ListFindData");

  success &= compare_integer(TRUE, DNode_IsAlone(pnTemp),
			     "node that is removed is alone");

  success &= compare_integer(pnNode3, DNode_GetNext(pnNode1),
			     "removing node updates remaining nodes correctly");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(pnNode1, DList_GetHead(plList1),
			     "removing node does not change head");

  success &= compare_integer(pnNode3, DList_GetTail(plList1),
			     "removing node does not change tail");

  success &= compare_integer(pts2, DNode_GetData(pnNode2),
			     "does not change node data");

  DList_TailAdd(plList1, pnNode2);

  return success;
}

BOOL Test_StructGT(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, StructGT(pts2, pts1),
			     "2 > 1");

  success &= compare_integer(TRUE, StructGT(pts3, pts2),
			     "3 > 2");

  return success;
}

BOOL Test_DList_FindDataMax(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));
  pTemp = DList_FindDataMax(plList1, StructGT);

  success &= compare_integer(pts3, pTemp,
			     "finds max data in list using comparator");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_DList_RemoveDataMax(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));

  pTemp = DList_RemoveDataMax(plList1, StructGT);

  success &= compare_integer(pts3, pTemp,
			     "duplicates ListFindDataMax");

  success &= compare_integer(TRUE, DNode_IsAlone(pnNode3),
			     "removed node is alone");

  success &= compare_integer(pnNode2, DNode_GetNext(pnNode1),
			     "removing node updates remaining nodes correctly");

  success &= compare_integer(pnNode2, DList_GetTail(plList1),
			     "tail is updated correctly");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  DList_Clear(plList1);
  DList_TailAdd(plList1, pnNode1);
  pTemp = DList_RemoveDataMax(plList1, StructGT);

  success &= compare_integer(pts1, pTemp,
			     "only node in a list is the max");

  success &= compare_integer(TRUE, DList_IsEmpty(plList1),
			     "removing max/only node results in empty list");
    
  return success;
}

BOOL Test_StructTransform(void)
{
  BOOL success = TRUE;

  StructTransform(pts3);

  success &= compare_integer(STRUCT_TRANSFORM_VALUE, StructGetValue(pts3),
			     "pts3 should have value of 22");

  return success;
}

BOOL Test_StructTransformWithArg(void)
{
  BOOL success = TRUE;

  StructTransformWithArg(pts3, pts2);

  success &= compare_integer(StructGetValue(pts2), StructGetValue(pts3),
			     "pts3 value == pts2 value");

  return success;
}

BOOL Test_DList_TransformData(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));
  DList_TransformData(plList1, StructTransform);

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(DNode_GetData(pnNode1)),
			     "correctly transforms node 1");

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(DNode_GetData(pnNode2)),
			     "correctly transforms node 2");

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(DNode_GetData(pnNode3)),
			     "correctly transforms node 3");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_DList_TransformDataWithArg(void)
{
  BOOL success = TRUE;
  PDNODE pnOldEnd;

  TestStruct ts4;
  pTestStruct pts4 = &ts4;

  StructSetValue(pts4, 4);

  DList_TailAdd(plList1, pnNode1);
  DList_TailAdd(plList1, pnNode2);
  DList_TailAdd(plList1, pnNode3);

  pnOldEnd = DNode_GetNext(DList_GetTail(plList1));

  DList_TransformDataWithArg(plList1, StructTransformWithArg, pts4);

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(DNode_GetData(pnNode1)),
			     "correctly transforms node 1");

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(DNode_GetData(pnNode2)),
			     "correctly transforms node 2");

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(DNode_GetData(pnNode3)),
			     "correctly transforms node 3");

  success &= compare_integer(pnOldEnd, DNode_GetNext(DList_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_DList_InsertSortedDescending(void)
{
  BOOL success = TRUE;

  StructSetValue(pts1, 5);
  StructSetValue(pts2, 12);
  StructSetValue(pts3, 4);

  DList_InsertSortedDescending(plList1, StructGT, pnNode1);
  success &= compare_integer(pnNode1, DList_GetHead(plList1),
			     "only node is maximal");
  DList_InsertSortedDescending(plList1, StructGT, pnNode2);
  success &= compare_integer(pnNode2, DList_GetHead(plList1),
			     "maximal node is head");
  DList_InsertSortedDescending(plList1, StructGT, pnNode3);
  success &= compare_integer(pnNode3, DList_GetTail(plList1),
			     "minimal node is tail");

  pnTemp = DList_HeadRemove(plList1);

  success &= compare_integer(pnNode2, pnTemp,
			     "first node is greatest");

  pnTemp = DList_HeadRemove(plList1);

  success &= compare_integer(pnNode1, pnTemp,
			     "second node is second-greatest");

  pnTemp = DList_HeadRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "last node is least");

  return success;
}

BOOL StructGT(
    PTR pts1,
    PTR pts2
    )
{
  return (StructGetValue(pts1) > StructGetValue(pts2));
}

void StructTransform(
    PTR pts
    )
{
  StructSetValue(pts, STRUCT_TRANSFORM_VALUE);
}

void StructTransformWithArg(
    PTR pts,
    PTR ptsArg
    )
{
  StructSetValue(pts, StructGetValue(ptsArg));
}
