/* Tests for list and node interface defined in include/base/list.h
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <base/thread2.h>
#include <util/test.h>
#include <util/list-single.h>

/* Global variables for testing */

typedef struct _TestStruct {
  int value;
} TestStruct, *pTestStruct;

#define pTS(_pts_) ((pTestStruct)(_pts_))
#define StructGetValue(_pts_) (pTS(_pts_)->value)
#define StructSetValue(_pts_, _value_) (StructGetValue(_pts_) = (_value_))
#define STRUCT_TRANSFORM_VALUE 22

PRIVATE NODE nNode1;
PRIVATE NODE nNode2;
PRIVATE NODE nNode3;
PRIVATE PNODE pnNode1 = &nNode1;
PRIVATE PNODE pnNode2 = &nNode2;
PRIVATE PNODE pnNode3 = &nNode3;
PRIVATE LIST lList1;
PRIVATE PLIST plList1 = &lList1;

PRIVATE PNODE pnTemp;
PRIVATE PTR pTemp;
PRIVATE PNODE pnHead;

PRIVATE TestStruct ts1;
PRIVATE TestStruct ts2;
PRIVATE TestStruct ts3;

PRIVATE pTestStruct pts1 = &ts1;
PRIVATE pTestStruct pts2 = &ts2;
PRIVATE pTestStruct pts3 = &ts3;

/* Function prototypes for tests */

BOOL Node_TestInitializer(void);
BOOL Test_Node_Init(void);
BOOL Test_Node_GetNext(void);
BOOL Test_Node_GetData(void);
BOOL Test_Node_IsAlone(void);
BOOL Test_Node_Push(void);
BOOL Test_Node_Pop(void);
BOOL Test_Node_InsertAfter(void);
BOOL Test_Node_RemoveAfter(void);
BOOL Test_Node_StackRemove(void);
BOOL Test_Node_StackClear(void);

BOOL List_TestInitializer(void);
BOOL Test_List_Init(void);
BOOL Test_List_Clear(void);
BOOL Test_List_IsEmpty(void);
BOOL Test_List_TailAdd(void);
BOOL Test_List_HeadAdd(void);
BOOL Test_List_HeadRemove(void);
BOOL Test_List_TailRemove(void);
BOOL Test_List_Remove(void);
BOOL Test_List_FindData(void);
BOOL Test_List_FindDataAndRemove(void);
BOOL Test_StructGT(void);
BOOL Test_List_FindDataMax(void);
BOOL Test_List_RemoveDataMax(void);
BOOL Test_StructTransform(void);
BOOL Test_StructTransformWithArg(void);
BOOL Test_List_TransformData(void);
BOOL Test_List_TransformDataWithArg(void);
BOOL Test_List_InsertSortedDescending(void);

/* Prototypes for auxiliary functions */

BOOL StructGT(
    PTR pts1,
    PTR pts2
    );

void StructTransform(
    PTR pts
    );

void StructTransformWithArg(
    PTR pts,
    PTR ptsArg
    );

int _tmain(void) {
  TestFunctionPair NodeTestFunctions[] = {
    {Node_TestInitializer,          ""},
    {Test_Node_Init,                "Node_Init"},
    {Test_Node_GetData,              "Node_GetData"},
    {Test_Node_GetNext,              "Node_GetNext"},
    {Test_Node_IsAlone,              "Node_IsAlone"},
    {Test_Node_Push,                 "Node_Push"},
    {Test_Node_Pop,                  "Node_Pop"},
    {Test_Node_InsertAfter,          "Node_InsertAfter"},
    {Test_Node_RemoveAfter,          "Node_RemoveAfter"},
    {Test_Node_StackRemove,          "Node_StackRemove"},
    {Test_Node_StackClear,           "Node_StackClear"}
  };

  TestFunctionPair ListTestFunctions[] = {
    {List_TestInitializer, ""},
    {Test_List_Init,                  "List_Init"},
    {Test_List_Clear,                  "List_Clear"},
    {Test_List_IsEmpty,                "List_IsEmpty"},
    {Test_List_TailAdd,                "List_TailAdd"},
    {Test_List_HeadAdd,                "List_HeadAdd"},
    {Test_List_HeadRemove,             "List_HeadRemove"},
    {Test_List_TailAdd,                "List_TailAdd"},
    {Test_List_TailRemove,             "List_TailRemove"},
    {Test_List_Remove,                 "List_Remove"},
    {Test_List_FindData,               "List_FindData"},
    {Test_List_FindDataAndRemove,      "List_FindDataAndRemove"},
    {Test_StructGT,                     "StructGT"},
    {Test_List_FindDataMax,            "List_FindDataMax"},
    {Test_List_RemoveDataMax,          "List_RemoveDataMax"},
    {Test_StructTransform,              "StructTransform"},
    {Test_StructTransformWithArg,       "StructTransformWithArg"},
    {Test_List_TransformData,          "List_TransformData"},
    {Test_List_TransformDataWithArg,   "List_TransformDataWithArg"},
    {Test_List_InsertSortedDescending, "List_InsertSortedDescending"}
  };

  RunTestSuite("Singly-Linked Node Test Suite", NodeTestFunctions,
	       GetTestFunctionArraySize(NodeTestFunctions));
  RunTestSuite("Singly-Linked List Test Suite", ListTestFunctions,
	       GetTestFunctionArraySize(ListTestFunctions));

  return 0;
}

BOOL Node_TestInitializer(void)
{
  Node_Init(pnNode1, NULL);
  Node_Init(pnNode2, NULL);
  Node_Init(pnNode3, NULL);
  pTemp = NULL;
  pnTemp = NULL;
  pnHead = NULL;

  return TRUE;
}

BOOL Test_Node_Init(void)
{
  BOOL success = TRUE;

  success &= compare_integer(NULL, Node_GetData(pnNode1),
			     "sets data pointer to NULL");

  success &= compare_integer(NULL, Node_GetNext(pnNode1),
			     "sets next pointer to NULL");

  return success;
}

BOOL Test_Node_GetData(void)
{
  BOOL success = TRUE;

  Node_SetData(pnNode1, (PTR)0x2222);

  success &= compare_integer(0x2222, Node_GetData(pnNode1),
			     "set data == get data");

  return success;
}

BOOL Test_Node_GetNext(void)
{
  BOOL success = TRUE;

  Node_SetNext(pnNode1, pnNode2);

  success &= compare_integer(pnNode2, Node_GetNext(pnNode1),
			     "set next == get next");

  return success;
}

BOOL Test_Node_IsAlone(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, Node_IsAlone(pnNode1),
			     "initialized node is alone");

  return success;
}

BOOL Test_Node_Push(void)
{
  BOOL success = TRUE;

  pnHead = Node_Push(pnHead, pnNode3);

  success &= compare_integer(pnNode3, pnHead,
			     "new node in empty stack is new head");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "new node in empty stack is alone");

  pnHead = Node_Push(pnHead, pnNode2);

  success &= compare_integer(pnNode3, Node_GetNext(pnNode2),
			     "new node in non-empty stack updates next");

  success &= compare_integer(pnNode2, pnHead,
			     "new node in non-empty stack updates head");

  return success;
}

BOOL Test_Node_Pop(void)
{
  BOOL success = TRUE;

  pnHead = Node_Push(pnHead, pnNode3);
  pnHead = Node_Push(pnHead, pnNode2);

  pnHead = Node_Pop(pnHead, &pnTemp);

  success &= compare_integer(pnNode3, pnHead,
			     "popping non-empty stack updates head");

  success &= compare_integer(pnNode2, pnTemp,
			     "popped node equals last node pushed");

  pnHead = Node_Pop(pnHead, &pnTemp);

  success &= compare_integer(NULL, pnHead,
			     "popping all nodes results in empty stack");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "popped node is alone");

  pnHead = Node_Pop(pnHead, NULL);

  success &= compare_integer(NULL, pnHead,
			     "popping empty stack has no effect");

  return success;
}

BOOL Test_Node_InsertAfter(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  pnHead = Node_Push(pnHead, pnNode1);
  pnOldEnd = Node_GetNext(pnNode1);

  Node_InsertAfter(pnNode1, pnNode2);

  success &= compare_integer(pnNode2, Node_GetNext(pnNode1),
			     "updates next correctly");

  success &= compare_integer(pnOldEnd, Node_GetNext(pnNode2),
			     "stack end is still the same");

  return success;
}

BOOL Test_Node_RemoveAfter(void)
{
  BOOL success = TRUE;

  pnHead = Node_Push(pnHead, pnNode3);
  pnHead = Node_Push(pnHead, pnNode2);
  pnHead = Node_Push(pnHead, pnNode1);

  Node_RemoveAfter(pnNode1);

  success &= compare_integer(pnNode3, Node_GetNext(pnNode1),
			     "updates remaining nodes correctly");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode2),
			     "removed node is alone");

  return success;
}

BOOL Test_Node_StackRemove(void)
{
  BOOL success = TRUE;

  pnHead = Node_Push(pnHead, pnNode1);
  pnHead = Node_Push(pnHead, pnNode2);
  pnHead = Node_Push(pnHead, pnNode3);

  pnHead = Node_StackRemove(pnHead, pnNode2);

  success &= compare_integer(TRUE, Node_IsAlone(pnNode2),
			     "removed node is alone");

  success &= compare_integer(pnNode1, Node_GetNext(pnNode3),
			     "updates remaining nodes correctly");

  success &= compare_integer(pnNode3, pnHead,
			     "head is still the same");

  pnHead = Node_StackRemove(pnHead, pnNode3);

  success &= compare_integer(pnNode1, pnHead,
			     "head is updated correctly");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "removed head is now alone");

  return success;
}

BOOL Test_Node_StackClear(void)
{
  BOOL success = TRUE;

  PNODE pnHead = NULL;

  pnHead = Node_Push(pnHead, pnNode1);
  pnHead = Node_Push(pnHead, pnNode2);
  pnHead = Node_Push(pnHead, pnNode3);

  pnHead = Node_StackClear(pnHead);

  success &= compare_integer(NULL, pnHead, "new stack head is NULL");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode1), "pnNode1 is alone");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode2), "pnNode2 is alone");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3), "pnNode3 is alone");

  return success;
}

BOOL List_TestInitializer(void)
{
  Node_Init(pnNode1, pts1);
  Node_Init(pnNode2, pts2);
  Node_Init(pnNode3, pts3);

  StructSetValue(pts1, 1);
  StructSetValue(pts2, 2);
  StructSetValue(pts3, 3);

  pTemp = NULL;
  pnTemp = NULL;
  pnHead = NULL;
  List_Init(plList1);

  return TRUE;
}

BOOL Test_List_Init(void)
{
  BOOL success = TRUE;

  success &= compare_integer(NULL, List_GetHead(plList1),
			     "sets head pointer to NULL");

  success &= compare_integer(NULL, List_GetTail(plList1),
			     "sets tail pointer to NULL");

  return success;
}

BOOL Test_List_Clear(void)
{
  BOOL success = TRUE;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  List_Clear(plList1);

  success &= compare_integer(TRUE, Node_IsAlone(pnNode1),
			     "node 1 is alone");
    
  success &= compare_integer(TRUE, Node_IsAlone(pnNode2),
			     "node 2 is alone");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "node 3 is alone");

  return success;
}

BOOL Test_List_IsEmpty(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, List_IsEmpty(plList1),
			     "initialized list is empty");

  return success;
}

BOOL Test_List_TailAdd(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);

  success &= compare_integer(pnNode1, List_GetHead(plList1),
			     "adding to an empty list updates head");

  success &= compare_integer(pnNode1, List_GetTail(plList1),
			     "adding to an empty list updates tail");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode1),
			     "new node in empty list is alone");

  pnOldEnd = Node_GetNext(List_GetTail(plList1));
  List_TailAdd(plList1, pnNode2);

  success &= compare_integer(pnNode2, Node_GetNext(pnNode1),
			     "new node updates existing nodes");

  success &= compare_integer(pnNode2, List_GetTail(plList1),
			     "new node updates list tail");

  success &= compare_integer(pnOldEnd,
			     Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_List_HeadAdd(void)
{
  BOOL success = TRUE;

  List_HeadAdd(plList1, pnNode1);

  success &= compare_integer(List_GetHead(plList1),  pnNode1,
			     "adding to an empty list updates head");

  success &= compare_integer(List_GetTail(plList1),  pnNode1,
			     "adding to an empty list updates tail");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode1),
			     "new node in list is alone");
    
  List_HeadAdd(plList1, pnNode3);

  success &= compare_integer(pnNode3, List_GetHead(plList1),
			     "new node is the head of list");

  success &= compare_integer(pnNode1, Node_GetNext(pnNode3),
			     "new node's next is the old head");

  success &= compare_integer(NULL,
			     Node_GetNext(List_GetTail(plList1)),
			     "list is still NULL-terminated");

  return TRUE;
}

/* Uncommenting these tests just fails an assertion.
 * Change the implemention of ListTailAdd if you would like to
 * handle adding duplicate nodes to a list gracefully instead of failing.
 */
/*      List_TailAdd(plList1, pnNode2); */

/*      compare_integer(pnNode3, List_GetTail(plList1), */
/*  		    "ListTailAdd: cannot add existing node to list - tail does not change."); */

/*      compare_integer(pnNode1, Node_GetNext(pnNode3), */
/*  		    "ListTailAdd: cannot add existing node to list - tail's next does not change."); */

/*      compare_integer(pnNode3, Node_GetNext(pnNode2), */
/*  		    "ListTailAdd: cannot add existing node to list - existing node stays in its place."); */

BOOL Test_List_HeadRemove(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));

  pnTemp = List_HeadRemove(plList1);

  success &= compare_integer(pnNode1, pnTemp,
			     "removes the first node");

  success &= compare_integer(pnNode2, List_GetHead(plList1),
			     "updates the head");

  success &= compare_integer(pnNode3, List_GetTail(plList1),
			     "tail does not change.");

  success &= compare_integer(TRUE, Node_IsAlone(pnTemp),
			     "removed node is alone.");

  success &= compare_integer(pnOldEnd,
			     Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  List_HeadRemove(plList1);
  pnTemp = List_HeadRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "last node removed sequentially is tail");

  success &= compare_integer(TRUE, List_IsEmpty(plList1),
			     "removing all nodes results in empty list");

  return success;
}

BOOL Test_List_TailRemove(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));
  pnTemp = List_TailRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "removed node is old tail");

  success &= compare_integer(pnNode2, List_GetTail(plList1),
			     "updates tail to next-to-last node");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "removed node is alone");

  List_TailRemove(plList1);

  success &= compare_integer(pnNode1, List_GetHead(plList1),
			     "only node in list is the head");

  success &= compare_integer(pnNode1, List_GetTail(plList1),
			     "only node in list is the tail");

  List_TailRemove(plList1);

  success &= compare_integer(TRUE, List_IsEmpty(plList1),
			     "removing only node results in empty list");

  return success;
}

BOOL Test_List_Remove(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));
  List_Remove(plList1, pnNode2);

  success &= compare_integer(TRUE, Node_IsAlone(pnNode2),
			     "removed node is alone");

  success &= compare_integer(pnNode3, Node_GetNext(pnNode1),
			     "updates remaining nodes correctly");

  success &= compare_integer(pnNode1, List_GetHead(plList1),
			     "head does not change");

  success &= compare_integer(pnNode3, List_GetTail(plList1),
			     "tail does not change");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  List_Remove(plList1, pnNode1);

  success &= compare_integer(pnNode3, List_GetHead(plList1),
			     "head is updated correctly");

  List_Remove(plList1, pnNode3);

  success &= compare_integer(NULL, List_GetTail(plList1),
			     "tail is updated correctly");

  /* The semantics of List_Remove now require the node to be in the list */
/*   List_Remove(plList1, pnNode1); */

/*   success &= compare_integer(TRUE, List_IsEmpty(plList1), */
/* 			     "removing from empty list has no effect"); */

  return success;
}    

BOOL Test_List_FindData(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));

  success &= compare_integer(pnNode1, List_FindData(plList1, pts1),
			     "data that is set is found again");

  success &= compare_integer(NULL, List_FindData(plList1, NULL),
			     "data that is not in the queue is not found");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(pts1, Node_GetData(pnNode1),
			     "does not change node data");

  return success;
}

BOOL Test_List_FindDataAndRemove(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));

  pnTemp = List_FindDataAndRemove(plList1, pts2);
    
  success &= compare_integer(pnNode2, pnTemp,
			     "duplicates ListFindData");

  success &= compare_integer(TRUE, Node_IsAlone(pnTemp),
			     "node that is removed is alone");

  success &= compare_integer(pnNode3, Node_GetNext(pnNode1),
			     "removing node updates remaining nodes correctly");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  success &= compare_integer(pnNode1, List_GetHead(plList1),
			     "removing node does not change head");

  success &= compare_integer(pnNode3, List_GetTail(plList1),
			     "removing node does not change tail");

  success &= compare_integer(pts2, Node_GetData(pnNode2),
			     "does not change node data");

  List_TailAdd(plList1, pnNode2);

  return success;
}

BOOL Test_StructGT(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, StructGT(pts2, pts1),
			     "2 > 1");

  success &= compare_integer(TRUE, StructGT(pts3, pts2),
			     "3 > 2");

  return success;
}

BOOL Test_List_FindDataMax(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));
  pTemp = List_FindDataMax(plList1, StructGT);

  success &= compare_integer(pts3, pTemp,
			     "finds max data in list using comparator");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_List_RemoveDataMax(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));

  pTemp = List_RemoveDataMax(plList1, StructGT);

  success &= compare_integer(pts3, pTemp,
			     "duplicates ListFindDataMax");

  success &= compare_integer(TRUE, Node_IsAlone(pnNode3),
			     "removed node is alone");

  success &= compare_integer(pnNode2, Node_GetNext(pnNode1),
			     "removing node updates remaining nodes correctly");

  success &= compare_integer(pnNode2, List_GetTail(plList1),
			     "tail is updated correctly");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  List_Clear(plList1);
  List_TailAdd(plList1, pnNode1);
  pTemp = List_RemoveDataMax(plList1, StructGT);

  success &= compare_integer(pts1, pTemp,
			     "only node in a list is the max");

  success &= compare_integer(TRUE, List_IsEmpty(plList1),
			     "removing max/only node results in empty list");
    
  return success;
}

BOOL Test_StructTransform(void)
{
  BOOL success = TRUE;

  StructTransform(pts3);

  success &= compare_integer(STRUCT_TRANSFORM_VALUE, StructGetValue(pts3),
			     "pts3 should have value of 22");

  return success;
}

BOOL Test_StructTransformWithArg(void)
{
  BOOL success = TRUE;

  StructTransformWithArg(pts3, pts2);

  success &= compare_integer(StructGetValue(pts2), StructGetValue(pts3),
			     "pts3 value == pts2 value");

  return success;
}

BOOL Test_List_TransformData(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));
  List_TransformData(plList1, StructTransform);

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(Node_GetData(pnNode1)),
			     "correctly transforms node 1");

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(Node_GetData(pnNode2)),
			     "correctly transforms node 2");

  success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			     StructGetValue(Node_GetData(pnNode3)),
			     "correctly transforms node 3");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_List_TransformDataWithArg(void)
{
  BOOL success = TRUE;
  PNODE pnOldEnd;

  TestStruct ts4;
  pTestStruct pts4 = &ts4;

  StructSetValue(pts4, 4);

  List_TailAdd(plList1, pnNode1);
  List_TailAdd(plList1, pnNode2);
  List_TailAdd(plList1, pnNode3);

  pnOldEnd = Node_GetNext(List_GetTail(plList1));

  List_TransformDataWithArg(plList1, StructTransformWithArg, pts4);

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(Node_GetData(pnNode1)),
			     "correctly transforms node 1");

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(Node_GetData(pnNode2)),
			     "correctly transforms node 2");

  success &= compare_integer(StructGetValue(pts4),
			     StructGetValue(Node_GetData(pnNode3)),
			     "correctly transforms node 3");

  success &= compare_integer(pnOldEnd, Node_GetNext(List_GetTail(plList1)),
			     "tail has same next");

  return success;
}

BOOL Test_List_InsertSortedDescending(void)
{
  BOOL success = TRUE;

  StructSetValue(pts1, 5);
  StructSetValue(pts2, 12);
  StructSetValue(pts3, 4);

  List_InsertSortedDescending(plList1, StructGT, pnNode1);
  success &= compare_integer(pnNode1, List_GetHead(plList1),
			     "only node is maximal");
  List_InsertSortedDescending(plList1, StructGT, pnNode2);
  success &= compare_integer(pnNode2, List_GetHead(plList1),
			     "maximal node is head");
  List_InsertSortedDescending(plList1, StructGT, pnNode3);
  success &= compare_integer(pnNode3, List_GetTail(plList1),
			     "minimal node is tail");

  pnTemp = List_HeadRemove(plList1);

  success &= compare_integer(pnNode2, pnTemp,
			     "first node is greatest");

  pnTemp = List_HeadRemove(plList1);

  success &= compare_integer(pnNode1, pnTemp,
			     "second node is second-greatest");

  pnTemp = List_HeadRemove(plList1);

  success &= compare_integer(pnNode3, pnTemp,
			     "last node is least");

  return success;
}

BOOL StructGT(
    PTR pts1,
    PTR pts2
    )
{
  return (StructGetValue(pts1) > StructGetValue(pts2));
}

void StructTransform(
    PTR pts
    )
{
  StructSetValue(pts, STRUCT_TRANSFORM_VALUE);
}

void StructTransformWithArg(
    PTR pts,
    PTR ptsArg
    )
{
  StructSetValue(pts, StructGetValue(ptsArg));
}
