/* Tests for list and node interface defined in include/base/list.h
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <base/thread2.h>
#include <util/test.h>
#include <util/list2.h>

/* Global variables for testing */

typedef struct _TestStruct {
  int value;
} TestStruct, *pTestStruct;

#define pTS(_pts_) ((pTestStruct)(_pts_))
#define StructGetValue(_pts_) (pTS(_pts_)->value)
#define StructSetValue(_pts_, _value_) (StructGetValue(_pts_) = (_value_))
#define STRUCT_TRANSFORM_VALUE 22

PRIVATE NODE nNode1;
PRIVATE NODE nNode2;
PRIVATE NODE nNode3;
PRIVATE PNODE pnNode1 = &nNode1;
PRIVATE PNODE pnNode2 = &nNode2;
PRIVATE PNODE pnNode3 = &nNode3;
PRIVATE LIST lList1;
PRIVATE PLIST plList1 = &lList1;

PRIVATE PNODE pnTemp;
PRIVATE PTR pTemp;
PRIVATE PNODE pnHead;

PRIVATE TestStruct ts1;
PRIVATE TestStruct ts2;
PRIVATE TestStruct ts3;

PRIVATE pTestStruct pts1 = &ts1;
PRIVATE pTestStruct pts2 = &ts2;
PRIVATE pTestStruct pts3 = &ts3;

/* Function prototypes for tests */

BOOL NodeTestInitializer(void);
BOOL TestNodeInitialize(void);
BOOL TestNodeGetNext(void);
BOOL TestNodeGetData(void);
BOOL TestNodeIsAlone(void);
BOOL TestNodePush(void);
BOOL TestNodePop(void);
BOOL TestNodeInsertAfter(void);
BOOL TestNodeRemoveAfter(void);
BOOL TestNodeStackRemove(void);
BOOL TestNodeStackClear(void);

BOOL ListTestInitializer(void);
BOOL TestListInitialize(void);
BOOL TestListClear(void);
BOOL TestListIsEmpty(void);
BOOL TestListTailAdd(void);
BOOL TestListHeadAdd(void);
BOOL TestListHeadRemove(void);
BOOL TestListTailRemove(void);
BOOL TestListRemove(void);
BOOL TestListFindData(void);
BOOL TestListFindDataAndRemove(void);
BOOL TestStructGT(void);
BOOL TestListFindDataMax(void);
BOOL TestListRemoveDataMax(void);
BOOL TestStructTransform(void);
BOOL TestStructTransformWithArg(void);
BOOL TestListTransformData(void);
BOOL TestListTransformDataWithArg(void);
BOOL TestListInsertSortedDescending(void);

/* Prototypes for auxiliary functions */

BOOL StructGT(
     PTR pts1,
     PTR pts2
     );

void StructTransform(
     PTR pts
     );

void StructTransformWithArg(
     PTR pts,
     PTR ptsArg
     );

int main(void) {
    TestFunctionPair NodeTestFunctions[] = {
      {NodeTestInitializer,            ""},
      {TestNodeInitialize,             "NodeInitialize"},
      {TestNodeGetData,                "NodeGetData"},
      {TestNodeGetNext,                "NodeGetNext"},
      {TestNodeIsAlone,                "NodeIsAlone"},
      {TestNodePush,                   "NodePush"},
      {TestNodePop,                    "NodePop"},
      {TestNodeInsertAfter,            "NodeInsertAfter"},
      {TestNodeRemoveAfter,            "NodeRemoveAfter"},
      {TestNodeStackRemove,            "NodeStackRemove"},
      {TestNodeStackClear,             "NodeStackClear"}
    };

    TestFunctionPair ListTestFunctions[] = {
      {ListTestInitializer, ""},
      {TestListInitialize,             "ListInitialize"},
      {TestListClear,                  "ListClear"},
      {TestListIsEmpty,                "ListIsEmpty"},
      {TestListTailAdd,                "ListTailAdd"},
      {TestListHeadAdd,                "ListHeadAdd"},
      {TestListHeadRemove,             "ListHeadRemove"},
      {TestListTailAdd,                "ListTailAdd"},
      {TestListTailRemove,             "ListTailRemove"},
      {TestListRemove,                 "ListRemove"},
      {TestListFindData,               "ListFindData"},
      {TestListFindDataAndRemove,      "ListFindDataAndRemove"},
      {TestStructGT,                   "StructGT"},
      {TestListFindDataMax,            "ListFindDataMax"},
      {TestListRemoveDataMax,          "ListRemoveDataMax"},
      {TestStructTransform,            "StructTransform"},
      {TestStructTransformWithArg,     "StructTransformWithArg"},
      {TestListTransformData,          "ListTransformData"},
      {TestListTransformDataWithArg,   "ListTransformDataWithArg"},
      {TestListInsertSortedDescending, "ListInsertSortedDescending"}
    };

    RunTestSuite("Singly-Linked Node Test Suite", NodeTestFunctions,
		 GetTestFunctionArraySize(NodeTestFunctions));
    RunTestSuite("Singly-Linked List Test Suite", ListTestFunctions,
		 GetTestFunctionArraySize(ListTestFunctions));

    return 0;
}

BOOL NodeTestInitializer(void)
{
    NodeInitialize(pnNode1, NULL);
    NodeInitialize(pnNode2, NULL);
    NodeInitialize(pnNode3, NULL);
    pTemp = NULL;
    pnTemp = NULL;
    pnHead = NULL;

    return TRUE;
}

BOOL TestNodeInitialize(void)
{
    BOOL success = TRUE;

    success &= compare_integer(NULL, NodeGetData(pnNode1),
			      "sets data pointer to NULL");

    success &= compare_integer(NULL, NodeGetNext(pnNode1),
			       "sets next pointer to NULL");

    return success;
}

BOOL TestNodeGetData(void)
{
    BOOL success = TRUE;

    NodeSetData(pnNode1, (PTR)0x2222);

    success &= compare_integer(0x2222, NodeGetData(pnNode1),
			       "set data == get data");

    return success;
}

BOOL TestNodeGetNext(void)
{
    BOOL success = TRUE;

    NodeSetNext(pnNode1, pnNode2);

    success &= compare_integer(pnNode2, NodeGetNext(pnNode1),
			       "set next == get next");

    return success;
}

BOOL TestNodeIsAlone(void)
{
    BOOL success = TRUE;

    success &= compare_integer(TRUE, NodeIsAlone(pnNode1),
			       "initialized node is alone");

    return success;
}

BOOL TestNodePush(void)
{
    BOOL success = TRUE;

    pnHead = NodePush(pnHead, pnNode3);

    success &= compare_integer(pnNode3, pnHead,
			       "new node in empty stack is new head");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "new node in empty stack is alone");

    pnHead = NodePush(pnHead, pnNode2);

    success &= compare_integer(pnNode3, NodeGetNext(pnNode2),
			       "new node in non-empty stack updates next");

    success &= compare_integer(pnNode2, pnHead,
			       "new node in non-empty stack updates head");

    return success;
}

BOOL TestNodePop(void)
{
    BOOL success = TRUE;

    pnHead = NodePush(pnHead, pnNode3);
    pnHead = NodePush(pnHead, pnNode2);

    pnHead = NodePop(pnHead, &pnTemp);

    success &= compare_integer(pnNode3, pnHead,
			       "popping non-empty stack updates head");

    success &= compare_integer(pnNode2, pnTemp,
			       "popped node equals last node pushed");

    pnHead = NodePop(pnHead, &pnTemp);

    success &= compare_integer(NULL, pnHead,
			       "popping all nodes results in empty stack");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "popped node is alone");

    pnHead = NodePop(pnHead, NULL);

    success &= compare_integer(NULL, pnHead,
			       "popping empty stack has no effect");

    return success;
}

BOOL TestNodeInsertAfter(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    pnHead = NodePush(pnHead, pnNode1);
    pnOldEnd = NodeGetNext(pnNode1);

    NodeInsertAfter(pnNode1, pnNode2);

    success &= compare_integer(pnNode2, NodeGetNext(pnNode1),
			       "updates next correctly");

    success &= compare_integer(pnOldEnd, NodeGetNext(pnNode2),
			       "stack end is still the same");

    return success;
}

BOOL TestNodeRemoveAfter(void)
{
    BOOL success = TRUE;

    pnHead = NodePush(pnHead, pnNode3);
    pnHead = NodePush(pnHead, pnNode2);
    pnHead = NodePush(pnHead, pnNode1);

    NodeRemoveAfter(pnNode1);

    success &= compare_integer(pnNode3, NodeGetNext(pnNode1),
			       "updates remaining nodes correctly");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "removed node is alone");

    return success;
}

BOOL TestNodeStackRemove(void)
{
    BOOL success = TRUE;

    pnHead = NodePush(pnHead, pnNode1);
    pnHead = NodePush(pnHead, pnNode2);
    pnHead = NodePush(pnHead, pnNode3);

    pnHead = NodeStackRemove(pnHead, pnNode2);

    success &= compare_integer(TRUE, NodeIsAlone(pnNode2),
			       "removed node is alone");

    success &= compare_integer(pnNode1, NodeGetNext(pnNode3),
			       "updates remaining nodes correctly");

    success &= compare_integer(pnNode3, pnHead,
			       "head is still the same");

    pnHead = NodeStackRemove(pnHead, pnNode3);

    success &= compare_integer(pnNode1, pnHead,
			       "head is updated correctly");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "removed head is now alone");

    return success;
}

BOOL TestNodeStackClear(void)
{
    BOOL success = TRUE;

    PNODE pnHead = NULL;

    pnHead = NodePush(pnHead, pnNode1);
    pnHead = NodePush(pnHead, pnNode2);
    pnHead = NodePush(pnHead, pnNode3);

    pnHead = NodeStackClear(pnHead);

    success &= compare_integer(NULL, pnHead, "new stack head is NULL");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode1), "pnNode1 is alone");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode2), "pnNode2 is alone");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3), "pnNode3 is alone");

    return success;
}

BOOL ListTestInitializer(void)
{
    NodeInitialize(pnNode1, pts1);
    NodeInitialize(pnNode2, pts2);
    NodeInitialize(pnNode3, pts3);

    StructSetValue(pts1, 1);
    StructSetValue(pts2, 2);
    StructSetValue(pts3, 3);

    pTemp = NULL;
    pnTemp = NULL;
    pnHead = NULL;
    ListInitialize(plList1);

    return TRUE;
}

BOOL TestListInitialize(void)
{
    BOOL success = TRUE;

    success &= compare_integer(NULL, ListGetHead(plList1),
			       "sets head pointer to NULL");

    success &= compare_integer(NULL, ListGetTail(plList1),
			       "sets tail pointer to NULL");

    return success;
}

BOOL TestListClear(void)
{
    BOOL success = TRUE;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    ListClear(plList1);

    success &= compare_integer(TRUE, NodeIsAlone(pnNode1),
			       "node 1 is alone");
    
    success &= compare_integer(TRUE, NodeIsAlone(pnNode2),
			       "node 2 is alone");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "node 3 is alone");

    return success;
}

BOOL TestListIsEmpty(void)
{
    BOOL success = TRUE;

    success &= compare_integer(TRUE, ListIsEmpty(plList1),
			       "initialized list is empty");

    return success;
}

BOOL TestListTailAdd(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);

    success &= compare_integer(pnNode1, ListGetHead(plList1),
			       "adding to an empty list updates head");

    success &= compare_integer(pnNode1, ListGetTail(plList1),
			       "adding to an empty list updates tail");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode1),
			       "new node in empty list is alone");

    pnOldEnd = NodeGetNext(ListGetTail(plList1));
    ListTailAdd(plList1, pnNode2);

    success &= compare_integer(pnNode2, NodeGetNext(pnNode1),
			       "new node updates existing nodes");

    success &= compare_integer(pnNode2, ListGetTail(plList1),
			       "new node updates list tail");

    success &= compare_integer(pnOldEnd,
			       NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    return success;
}

BOOL TestListHeadAdd(void)
{
    BOOL success = TRUE;

    ListHeadAdd(plList1, pnNode1);

    success &= compare_integer(ListGetHead(plList1),  pnNode1,
			       "adding to an empty list updates head");

    success &= compare_integer(ListGetTail(plList1),  pnNode1,
			       "adding to an empty list updates tail");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode1),
			       "new node in list is alone");
    
    ListHeadAdd(plList1, pnNode3);

    success &= compare_integer(pnNode3, ListGetHead(plList1),
			       "new node is the head of list");

    success &= compare_integer(pnNode1, NodeGetNext(pnNode3),
			       "new node's next is the old head");

    success &= compare_integer(NULL,
			       NodeGetNext(ListGetTail(plList1)),
			       "list is still NULL-terminated");

    return TRUE;
}

    /* Uncommenting these tests just fails an assertion.
     * Change the implemention of ListTailAdd if you would like to
     * handle adding duplicate nodes to a list gracefully instead of failing.
     */
/*      ListTailAdd(plList1, pnNode2); */

/*      compare_integer(pnNode3, ListGetTail(plList1), */
/*  		    "ListTailAdd: cannot add existing node to list - tail does not change."); */

/*      compare_integer(pnNode1, NodeGetNext(pnNode3), */
/*  		    "ListTailAdd: cannot add existing node to list - tail's next does not change."); */

/*      compare_integer(pnNode3, NodeGetNext(pnNode2), */
/*  		    "ListTailAdd: cannot add existing node to list - existing node stays in its place."); */

BOOL TestListHeadRemove(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));

    pnTemp = ListHeadRemove(plList1);

    success &= compare_integer(pnNode1, pnTemp,
			       "removes the first node");

    success &= compare_integer(pnNode2, ListGetHead(plList1),
			       "updates the head");

    success &= compare_integer(pnNode3, ListGetTail(plList1),
			       "tail does not change.");

    success &= compare_integer(TRUE, NodeIsAlone(pnTemp),
			       "removed node is alone.");

    success &= compare_integer(pnOldEnd,
			       NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    ListHeadRemove(plList1);
    pnTemp = ListHeadRemove(plList1);

    success &= compare_integer(pnNode3, pnTemp,
			       "last node removed sequentially is tail");

    success &= compare_integer(TRUE, ListIsEmpty(plList1),
			       "removing all nodes results in empty list");

    return success;
}

BOOL TestListTailRemove(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));
    pnTemp = ListTailRemove(plList1);

    success &= compare_integer(pnNode3, pnTemp,
			       "removed node is old tail");

    success &= compare_integer(pnNode2, ListGetTail(plList1),
			       "updates tail to next-to-last node");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "removed node is alone");

    ListTailRemove(plList1);

    success &= compare_integer(pnNode1, ListGetHead(plList1),
			       "only node in list is the head");

    success &= compare_integer(pnNode1, ListGetTail(plList1),
			       "only node in list is the tail");

    ListTailRemove(plList1);

    success &= compare_integer(TRUE, ListIsEmpty(plList1),
			       "removing only node results in empty list");

    return success;
}

BOOL TestListRemove(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));
    ListRemove(plList1, pnNode2);

    success &= compare_integer(TRUE, NodeIsAlone(pnNode2),
			       "removed node is alone");

    success &= compare_integer(pnNode3, NodeGetNext(pnNode1),
			       "updates remaining nodes correctly");

    success &= compare_integer(pnNode1, ListGetHead(plList1),
			       "head does not change");

    success &= compare_integer(pnNode3, ListGetTail(plList1),
			       "tail does not change");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    ListRemove(plList1, pnNode1);

    success &= compare_integer(pnNode3, ListGetHead(plList1),
			       "head is updated correctly");

    ListRemove(plList1, pnNode3);

    success &= compare_integer(NULL, ListGetTail(plList1),
			       "tail is updated correctly");

    ListRemove(plList1, pnNode1);

    success &= compare_integer(TRUE, ListIsEmpty(plList1),
			       "removing from empty list has no effect");

    return success;
}    

BOOL TestListFindData(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));

    success &= compare_integer(pnNode1, ListFindData(plList1, pts1),
			       "data that is set is found again");

    success &= compare_integer(NULL, ListFindData(plList1, NULL),
			       "data that is not in the queue is not found");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    success &= compare_integer(pts1, NodeGetData(pnNode1),
			       "does not change node data");

    return success;
}

BOOL TestListFindDataAndRemove(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));

    pnTemp = ListFindDataAndRemove(plList1, pts2);
    
    success &= compare_integer(pnNode2, pnTemp,
			       "duplicates ListFindData");

    success &= compare_integer(TRUE, NodeIsAlone(pnTemp),
			       "node that is removed is alone");

    success &= compare_integer(pnNode3, NodeGetNext(pnNode1),
			       "removing node updates remaining nodes correctly");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    success &= compare_integer(pnNode1, ListGetHead(plList1),
			       "removing node does not change head");

    success &= compare_integer(pnNode3, ListGetTail(plList1),
			       "removing node does not change tail");

    success &= compare_integer(pts2, NodeGetData(pnNode2),
			       "does not change node data");

    ListTailAdd(plList1, pnNode2);

    return success;
}

BOOL TestStructGT(void)
{
    BOOL success = TRUE;

    success &= compare_integer(TRUE, StructGT(pts2, pts1),
			       "2 > 1");

    success &= compare_integer(TRUE, StructGT(pts3, pts2),
			       "3 > 2");

    return success;
}

BOOL TestListFindDataMax(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));
    pTemp = ListFindDataMax(plList1, StructGT);

    success &= compare_integer(pts3, pTemp,
			       "finds max data in list using comparator");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    return success;
}

BOOL TestListRemoveDataMax(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));

    pTemp = ListRemoveDataMax(plList1, StructGT);

    success &= compare_integer(pts3, pTemp,
			       "duplicates ListFindDataMax");

    success &= compare_integer(TRUE, NodeIsAlone(pnNode3),
			       "removed node is alone");

    success &= compare_integer(pnNode2, NodeGetNext(pnNode1),
			       "removing node updates remaining nodes correctly");

    success &= compare_integer(pnNode2, ListGetTail(plList1),
			       "tail is updated correctly");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    ListClear(plList1);
    ListTailAdd(plList1, pnNode1);
    pTemp = ListRemoveDataMax(plList1, StructGT);

    success &= compare_integer(pts1, pTemp,
			       "only node in a list is the max");

    success &= compare_integer(TRUE, ListIsEmpty(plList1),
			       "removing max/only node results in empty list");
    
    return success;
}

BOOL TestStructTransform(void)
{
    BOOL success = TRUE;

    StructTransform(pts3);

    success &= compare_integer(STRUCT_TRANSFORM_VALUE, StructGetValue(pts3),
			       "pts3 should have value of 22");

    return success;
}

BOOL TestStructTransformWithArg(void)
{
    BOOL success = TRUE;

    StructTransformWithArg(pts3, pts2);

    success &= compare_integer(StructGetValue(pts2), StructGetValue(pts3),
			       "pts3 value == pts2 value");

    return success;
}

BOOL TestListTransformData(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));
    ListTransformData(plList1, StructTransform);

    success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			       StructGetValue(NodeGetData(pnNode1)),
			       "correctly transforms node 1");

    success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			       StructGetValue(NodeGetData(pnNode2)),
			       "correctly transforms node 2");

    success &= compare_integer(STRUCT_TRANSFORM_VALUE,
			       StructGetValue(NodeGetData(pnNode3)),
			       "correctly transforms node 3");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    return success;
}

BOOL TestListTransformDataWithArg(void)
{
    BOOL success = TRUE;
    PNODE pnOldEnd;

    TestStruct ts4;
    pTestStruct pts4 = &ts4;

    StructSetValue(pts4, 4);

    ListTailAdd(plList1, pnNode1);
    ListTailAdd(plList1, pnNode2);
    ListTailAdd(plList1, pnNode3);

    pnOldEnd = NodeGetNext(ListGetTail(plList1));

    ListTransformDataWithArg(plList1, StructTransformWithArg, pts4);

    success &= compare_integer(StructGetValue(pts4),
			       StructGetValue(NodeGetData(pnNode1)),
			       "correctly transforms node 1");

    success &= compare_integer(StructGetValue(pts4),
			       StructGetValue(NodeGetData(pnNode2)),
			       "correctly transforms node 2");

    success &= compare_integer(StructGetValue(pts4),
			       StructGetValue(NodeGetData(pnNode3)),
			       "correctly transforms node 3");

    success &= compare_integer(pnOldEnd, NodeGetNext(ListGetTail(plList1)),
			       "tail has same next");

    return success;
}

BOOL TestListInsertSortedDescending(void)
{
    BOOL success = TRUE;

    StructSetValue(pts1, 5);
    StructSetValue(pts2, 12);
    StructSetValue(pts3, 4);

    ListInsertSortedDescending(plList1, StructGT, pnNode1);
    success &= compare_integer(pnNode1, ListGetHead(plList1),
			       "only node is maximal");
    ListInsertSortedDescending(plList1, StructGT, pnNode2);
    success &= compare_integer(pnNode2, ListGetHead(plList1),
			       "maximal node is head");
    ListInsertSortedDescending(plList1, StructGT, pnNode3);
    success &= compare_integer(pnNode3, ListGetTail(plList1),
			       "minimal node is tail");

    pnTemp = ListHeadRemove(plList1);

    success &= compare_integer(pnNode2, pnTemp,
			       "first node is greatest");

    pnTemp = ListHeadRemove(plList1);

    success &= compare_integer(pnNode1, pnTemp,
			       "second node is second-greatest");

    pnTemp = ListHeadRemove(plList1);

    success &= compare_integer(pnNode3, pnTemp,
			       "last node is least");

    return success;
}

BOOL StructGT(
     PTR pts1,
     PTR pts2
     )
{
    return (StructGetValue(pts1) > StructGetValue(pts2));
}

void StructTransform(
     PTR pts
     )
{
    StructSetValue(pts, STRUCT_TRANSFORM_VALUE);
}

void StructTransformWithArg(
     PTR pts,
     PTR ptsArg
     )
{
    StructSetValue(pts, StructGetValue(ptsArg));
}
