/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <mmlite.h>
#include "../utilities/utility.h"

BOOL PrintModule( PIMODULE pMod, _TCHAR * Title )
{
#define NN 128
    _TCHAR Buffer[NN];
    SCODE sc;
    _TCHAR *pMsg;
    ADDRESS Start;
    UINT nRefs;

    _tprintf(_TEXT("\n\r%sModule @%x:\n\r"), Title, (ADDRESS) pMod);
    
    /* See how many references are outstanding */
    nRefs = pMod->v->AddRef(pMod) - 1;
    pMod->v->Release(pMod);
    _tprintf(_TEXT("  Refs: %d\n\r"), nRefs);

    sc = pMod->v->GetName(pMod, Buffer, NN);
    if (FAILED(sc)) {
        pMsg = _TEXT("GetName failed");
        goto ErrorExit;
    }
    Buffer[NN-1] = 0;
    _tprintf(_TEXT("  Name: %s\n\r"), Buffer);


    sc = pMod->v->GetArgs(pMod, Buffer, NN);
    if (FAILED(sc)) {
        pMsg = _TEXT("GetArgs failed");
        goto ErrorExit;
    }
    Buffer[NN-1] = 0;
    _tprintf(_TEXT("  Arguments: %s\n\r"), Buffer);

#if 0
    sc = pMod->v->RelocateRva(pMod, 0, &Start, &Size);
    if (FAILED(sc)) {
        pMsg = _TEXT("GetLocation failed");
        goto ErrorExit;
    }
    _tprintf(_TEXT("  Address range: [x%x..x%x)\n\r"), Start, Start + Size);
#endif

    sc = pMod->v->GetEntryPoint(pMod, &Start);
    if (FAILED(sc)) {
        pMsg = _TEXT("GetEntryPoint failed");
        goto ErrorExit;
    }
    _tprintf(_TEXT("  EntryPoint: x%x\n\r"), Start);

    /* All is fine */
    return TRUE;
    
 ErrorExit:
    _tprintf(_TEXT("PrintModule FAILED: %s sc = x%x\n\r"), pMsg, sc);
    return FALSE;
}

/* Tests (some of) the IModule support
 */
int _tmain(int argc, _TCHAR **argv)
{
    PIMODULE pMod, pPrev, pCur;
    SCODE sc;
    int nModules;

    UnusedParameter(argc);
    UnusedParameter(argv);

    /* Print the current module first
     */
    sc = GetCurrentModule( &pCur );
    if (FAILED(sc)) {
        _tprintf(_TEXT("GetCurrentModule failed with sc = x%x\n\r"), sc);
        return SetAndReturn(EXIT_FAILURE);
    }
    PrintModule(pCur,_TEXT("Current "));


    /* Now go through all loaded modules
     */
    nModules = 0;
    pMod = pCur;
    for (pPrev = NULL;;) {

        /* Find the next one, if any
         */
        sc = GetModuleNext( pPrev, &pMod );

        /* Let the previous one go, if any
         */
        if (pPrev)
            pPrev->v->Release(pPrev);

        /* End of list ?
         */
        if (sc != S_OK)
            break;
        nModules++;

        /* Nope, print this one
         */
        if (pMod != pCur) {
            if (!PrintModule(pMod, _TEXT("")))
                return SetAndReturn(EXIT_FAILURE);
        }

        /* And move on to the next one.
         */
        pPrev = pMod;
        
    }

    pCur->v->Release(pCur);

    /* There MUST be at leats one other module besides us: the kernel
     */
    if (nModules < 2) {
        _tprintf(_TEXT("test FAILED, we only saw %d modules.\n\r"), nModules);
        return SetAndReturn(EXIT_FAILURE);
    }
    else
        _tprintf(_TEXT("test PASSED.\n\r"));

    return SetAndReturn(EXIT_SUCCESS);
}
