/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

SCODE minfo(PIPROCESS pPrc)
{
    SCODE sc;
#if defined(__TCS__)
    TM_MACHINE_INFO tm;

    sc = pPrc->v->MachineInfo( pPrc,
                               MACHINE_KIND_TRIMEDIA,
                               (PMACHINE_INFO) &tm );
    if (FAILED(sc))
        return sc;
    _tprintf(_TEXT("Trimedia info:\n\r"));
    _tprintf(_TEXT("  MMIOBase x%x\n\r"), tm.MMIOBase);
    _tprintf(_TEXT("  TMMIOBase x%x\n\r"), tm.TMMIOBase);
    _tprintf(_TEXT("  SRAMBase x%x\n\r"), tm.SRAMBase);
    _tprintf(_TEXT("  SRAMSize x%x\n\r"), tm.SRAMSize);
    _tprintf(_TEXT("  RDRAMBase x%x\n\r"), tm.RDRAMBase);
    _tprintf(_TEXT("  RDRAMSize x%x\n\r"), tm.RDRAMSize);
    _tprintf(_TEXT("  Clock Freq %d\n\r"), tm.ClockFreq);
#else
    /* nothing defined for this architecture
     */
    UnusedParameter(pPrc);
    sc = NO_ERROR;
#endif

    return sc;
}

_TCHAR *ProcessorName(PROCESSOR_INFO_FLAVOR Tag)
{
    switch (Tag) {
    case PROCESSOR_KIND_IA32: return _TEXT("IA32");
    case PROCESSOR_KIND_MIPS: return _TEXT("MIPS");
    case PROCESSOR_KIND_MIPS_TIGON: return _TEXT("MIPS_TIGON");
    case PROCESSOR_KIND_PPC: return _TEXT("PPC");
    case PROCESSOR_KIND_H8: return _TEXT("H8");
    case PROCESSOR_KIND_ARM: return _TEXT("ARM");
    case PROCESSOR_KIND_M68K: return _TEXT("M68K");
    case PROCESSOR_KIND_ALPHA: return _TEXT("ALPHA");
    case PROCESSOR_KIND_TRIMEDIA: return _TEXT("TRIMEDIA");
    }
    return _TEXT("Unknown");
}

int _tmain( int argc, _TCHAR **argv)
{
    struct MACHINE_INFO mi;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();
    BOOL Passed = TRUE;
    int i;

    UnusedParameter(argc);
    UnusedParameter(argv);

    sc = pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi );
    if (FAILED(sc)) {
        _tprintf(_TEXT("GENERIC machine info failed! (x%x)"), sc);
        Passed = FALSE;
        goto Out;
    }
    _tprintf(_TEXT("Generic info:\n\r"));
    _tprintf(_TEXT(" ProcessorKind x%x (%s)"),
             mi.ProcessorKind, ProcessorName(mi.ProcessorKind));
    _tprintf(_TEXT(" OEMId x%x%s\n\r"), mi.OEMId,
             (mi.OEMId == ~0) ? " (simulator)" : "");
    _tprintf(_TEXT(" ProcessorSpeed %lld Hz\n\r"), mi.ProcessorSpeed);
    _tprintf(_TEXT(" TotalPhysicalMemory x%x\n\r"), mi.TotalPhysicalMemory);
    _tprintf(_TEXT(" AvailablePhysicalMemory x%x\n\r"), mi.AvailablePhysicalMemory);
    _tprintf(_TEXT(" Reserved fields"));
    for (i = 0; i < 10; i++)
        _tprintf(_TEXT(" x%x"), mi.Reserved[i]);
    _tprintf(_TEXT("\n\r"));


    sc = minfo(pPrc);
    if (FAILED(sc)) {
        _tprintf(_TEXT("machine SPECIFIC info failed! (x%x)"), sc);
        Passed = FALSE;
    }

 Out:
    exit(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}
