/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>

void pdec(int d, int f)
{
    _TCHAR n[32];
    _sntprintf(n,32,_T("%d"),d);
    d = _tcslen(n);
    n[d+1] = 0;
    for (; (f > 0) && (d > 0); f--, d--)
        n[d] = n[d-1];
    n[d] = _T('.');
    _putts(n);
}

BOOL Verbose = FALSE;
UINT TimingOverhead      = 0;

/* Compute & print timing overhead
 */
void ComputeOverhead(void)
{
    TIME Ts, Te;
    UINT PrevCnt[2];

    /* Do it more than once to avoid glitches.
     * A small array of past values avoids oscillating loops
     */
    PrevCnt[1] = 0;
    TimingOverhead = 0;

    do {
        /* This is what we use this time around.
         */
        PrevCnt[0] = PrevCnt[1];
        PrevCnt[1] = TimingOverhead;

        /* Compute the overhead via two consecutive calls.
         * NB: This does not work if the timer's granularity is too coarse.
         */
        Ts = CurrentTime();
        Te = CurrentTime();
        TimingOverhead = Int64ToInt32(TimeSubtract(Te,Ts));

        if (Verbose)
        _tprintf(_T("TimingOverhead: %d %d %d\r\n"),
                 PrevCnt[0], PrevCnt[1], TimingOverhead);

    } while (PrevCnt[1] != TimingOverhead && PrevCnt[0] != TimingOverhead);

    _tprintf(_T("Timing delta d="));
    pdec(TimingOverhead,1);
    _tprintf(_T("\r\n"));

}

UINT TimeIt(int t)
{
    TIME s,e;
    UINT r;

    /* Check the average across 10 runs
     */
    s = CurrentTime();
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    Delay(t);
    e = CurrentTime();

    e = TimeSubtract(e,s);
    r = Int64ToInt32(e)-TimingOverhead;

    //pdec(r,2);

    return r;
}

UINT DelayValues[] = { 10, 30, 50, 100, 200};
#define nDelayValues ((UINT)(sizeof DelayValues / sizeof DelayValues[0]))
int _tmain(int argc, _TCHAR ** argv)
{
    UINT i, t, nFailed = 0;
    UINT p10, m10;

    /* Fine tune
     */
    ComputeOverhead();

    /* Run all tests
     */
    for (i = 0; i < nDelayValues; i++) {

        t = TimeIt(DelayValues[i]);
        
        /* Must be within +-10%
         */
        p10 = ((DelayValues[i] * 10) + DelayValues[i]) * 10;
        m10 = ((DelayValues[i] * 10) - DelayValues[i]) * 10;
        if ((t > p10) || (t < m10)) {
            _tprintf(_T("Delay(%d) took "),DelayValues[i]);
            pdec(t,2);
            _putts(_T(" us, outside the acceptable +-10% range "));
            pdec(m10,2);
            _putts(_T(".."));
            pdec(p10,2);
            _putts(_T("\r\n"));

            nFailed++;
        }
    }

    /* Say what
     */
    if (nFailed) {
        _tprintf(_T("Failed %d tests out of %d total\r\n"),
                 nFailed, nDelayValues);
        goto Fail;
    }

    _tprintf(_T("This test PASSED\n\r"));
    exit(EXIT_SUCCESS);
    return 0;/*amusme*/

 Fail:
    _tprintf(_T("This test FAILED\n\r"));
    exit(EXIT_FAILURE);
    return 0;/*amusme*/
}

