/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * C test program -- 64-bit atoi() functions
 */

#include <mmlite.h>

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

extern int SetAndReturn(int nRet);

/* grumble.. */
typedef struct
    {
    UINT32 LowPart;
    INT32 HighPart;
    }   MY_INT64;

typedef struct
    {
    UINT32 LowPart;
    UINT32 HighPart;
    }   MY_UINT64;

MY_INT64 it[] = {
{ 0x0, 0x0 },
{ 0xffffffff, 0xffffffff },
{ 0x0, 0x80000000 },
{ 0x1, 0x80000000 },
{ 0x1, 0x0 },
{ 0x2, 0x0 },
{ 0x3, 0x0 },
{ 0x4, 0x0 },
{ 0x5, 0x0 },
{ 0x6, 0x0 },
{ 0x7, 0x0 },
{ 0x8, 0x0 },
{ 0x9, 0x0 },
{ 0x10, 0x0 },
{ 0xf0000000, 0x00000000 },
{ 0x0f000000, 0x00000000 },
{ 0x00f00000, 0x00000000 },
{ 0x000f0000, 0x00000000 },
{ 0x0000f000, 0x00000000 },
{ 0x00000f00, 0x00000000 },
{ 0x000000f0, 0x00000000 },
{ 0x0000000f, 0x00000000 },
{ 0x00000000, 0xf0000000 },
{ 0x00000000, 0x0f000000 },
{ 0x00000000, 0x00f00000 },
{ 0x00000000, 0x000f0000 },
{ 0x00000000, 0x0000f000 },
{ 0x00000000, 0x00000f00 },
{ 0x00000000, 0x000000f0 },
{ 0x00000000, 0x0000000f },
{ 0x10000000, 0x00000000 },
{ 0x01000000, 0x00000000 },
{ 0x00100000, 0x00000000 },
{ 0x00010000, 0x00000000 },
{ 0x00001000, 0x00000000 },
{ 0x00000100, 0x00000000 },
{ 0x00000010, 0x00000000 },
{ 0x00000001, 0x00000000 },
{ 0x00000000, 0x10000000 },
{ 0x00000000, 0x01000000 },
{ 0x00000000, 0x00100000 },
{ 0x00000000, 0x00010000 },
{ 0x00000000, 0x00001000 },
{ 0x00000000, 0x00000100 },
{ 0x00000000, 0x00000010 },
{ 0x00000000, 0x00000001 },

/* some random numbers */
{ 0x80000000, 0xffffc88a },
{ 0x649f0000, 0x0 },
{ 0xffff72e7, 0xffffffff },
{ 0x80000000, 0xffffe14d },
{ 0x54000000, 0x10 },
{ 0xfffffe1f, 0xffffffff },
{ 0x80000000, 0xffffc467 },
{ 0x91800000, 0x0 },
{ 0xffff9725, 0xffffffff },
};
#define NITESTS ((UINT) (sizeof it / (sizeof it[0])))
#define NRTESTS 1000

#define error(i,f,a,b,s) \
  _tprintf(_TEXT("/*%d %c*/  {0x%x,0x%x},{0x%x,0x0},%s\n\r"),i,f,a,b,s);

/* No idea what the stat distribution of this is but so what
 */
INT64 rand64(void)
{
    UINT32 Low; INT32 High;
    INT64 ai;

    /* NB: We know that rand() returns 15 good bits */
    Low = rand() | (rand() << 15) | (rand() << 30);
    High = rand() | (rand() << 15) | (rand() << 30);
    Int64FromHighAndLow(ai,High, Low);
    return ai;
}

_TCHAR buf[128];
BOOL Verbose = FALSE;

int _tmain()
{
    int i, n;
    MY_INT64 *pi;
    INT64 ai, bi;

    int nPass, nFail;

    nPass = nFail = 0;
    pi = it;
    for (i = 0; i < NITESTS; ++i) {

        /* Pick next number
         */
        Int64FromHighAndLow(ai, pi->HighPart, pi->LowPart);
        pi++;

        /* To string, signed decimal
         */
        _sntprintf(buf,128,_TEXT("%ld"),ai);
        bi = _tcstoi64(buf,NULL,10);

        if (Verbose)
            _tprintf(_TEXT("%d: {0x%x,0x%x},{0x%x,0x%x},%s\n\r"),i,ai,bi,buf);

        if (!Int64Equal(ai,bi)) {
            error(i,'d',ai,bi,buf);
            nFail++;
        } else
            nPass++;

    }

    /* Now for some randomly generated numbers
     */
    srand(0x1234);
    for (n = i + NRTESTS; i < n; i++) {
        ai = rand64();

        /* To string, signed decimal
         */
        _sntprintf(buf,128,_TEXT("%ld"),ai);
        bi = _tcstoi64(buf,NULL,10);

        if (Verbose)
            _tprintf(_TEXT("%d: {0x%x,0x%x},{0x%x,0x%x},%s\n\r"),i,ai,bi,buf);

        if (!Int64Equal(ai,bi)) {
            error(i,'d',ai,bi,buf);
            nFail++;
        } else
            nPass++;

    }

    _tprintf(_TEXT("signed 64bit conversions to/from strings test "));
    if (nFail == 0) {
        _tprintf(_TEXT("PASSED.\n\r"));

    } else {
        _tprintf(_TEXT("FAILED (%d fail %d pass).\n\r"),
                 nFail, nPass);
    }

    if (nFail)
    _tprintf(_TEXT("Test FAILED, %d errors.\n\r"), nFail);

    return SetAndReturn((nFail == 0) ? EXIT_SUCCESS : EXIT_FAILURE);
}
