/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"
/* Which test(s) should we perform
 */
#define TEST1 1
#define TEST_COND 1
#define TEST_SLEEP 1
#define TEST_CONSTR 1

/* Print names for tests
 */
#define P_TEST_ATOMIC _TEXT("  AtomicXXX functions\n\r")


BOOL Passed = TRUE;
UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}

/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

int nn = 0;
UINT x1 = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST_ATOMIC
               );

    if (me == 0) {

        _tprintf(_TEXT("Initial thread started.\n\r"));
        (void) CreateAThread(FirstThread, NULL);
        (void) CreateAThread(FirstThread, NULL);
    }

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, (ADDRESS) CurrentThread());

    for (i = 0; i < 1000; i++) {
      AtomicInc(&x1);
      AtomicAdd(&x1,20);
      AtomicDec(&x1);
      AtomicInc(&x1);
      AtomicSub(&x1,10);
      AtomicDec(&x1);
      AtomicSub(&x1,10);
    }
    _tprintf(_TEXT("%d found x1 = %d\n\r"), me, x1);

    /* All threads should be done
     */
    MaybeTerminate();

    /* This wasnt a very serious test, all threads will run
     * the sequence above straight in one timeslice..
     */
    Passed = (x1 == 0);

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return(0);
}
