/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"

/* Which test(s) should we perform
 */
#define P_TEST_SLEEP _TEXT("  SleepUntil\n\r")

UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    if (E_NOT_ENOUGH_MEMORY == sc)
        _tprintf(_TEXT("CreateThread return E_NOT_ENOUGH_MEMORY\n\r"));
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

BOOL Passed = TRUE;

int nn = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST_SLEEP
               );

    if (me == 0) {

        _tprintf(_TEXT("Initial thread started.\n\r"));
        (void) CreateAThread(FirstThread, NULL);
        (void) CreateAThread(FirstThread, NULL);
    }

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, (ADDRESS) CurrentThread());

    /* This tests the IdleThread logic too, since all threads sleep
     */
    {
        TIME t, te;
        for (i = 0; i < 10; i++) {
            Int32ToInt64(t,TIME_RELATIVE(10000)); /* 1milli */
            te = CurrentTime();
            SleepUntil(t);
            t = CurrentTime();
            _tprintf(_TEXT("%d slept from %lld to %lld\n\r"),
                     me, te, t);
        }
    }
    /* All threads should be done
     */
    MaybeTerminate();

    /* Test passes iff one-and-only-one thread gets here.
     * Which is guaranteed by MaybeTerminate.
     */
    /*Passed = TRUE;*/

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return(0);
}
