/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <diagnostics.h>
#include "../utilities/utility.h"

/* Which test(s) should we perform
 */
#define P_TEST_COND _TEXT("  Condition variables\n\r")


UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    DBGME(0,_tprintf(_TEXT("Creating a new thread to run @%x(%x).\n\r"),
             (ADDRESS) pStart, (ADDRESS) Arg));
    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

BOOL Passed = TRUE;

CONDITION fooc, bazc;

/* Possibly drop in the debugger if things are iffy
 */
BOOL MaybeDead()
{
    _TCHAR ch = _T('n');
;
    if (Passed) {
        _tprintf(_TEXT("This isnt working, giving up.. "));
        //ch = (_TCHAR) _gettchar();
        ch = 'n';
    }
    if (ch == _TEXT('n')) {
        Passed = FALSE;
        Condition_Broadcast(&fooc);
        Condition_Broadcast(&bazc);
        return TRUE;
    }
    return FALSE;
}

void MilliSleep(int mSecs, int me)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    DBGME(0,_tprintf(_TEXT("%d Sleeping for %d msecs [%ld]\n\r"),
             me, mSecs, t));
    SleepUntil(t);
}

struct {
    BOOL Started;
    BOOL Awoke[2];
} Tstates[3] = {
  { FALSE, },
};

int Step = 0;
int LastWoken = 0;
volatile UINT nHere = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i, k;
    TIME t, t0;

    me = (int) arg;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST_COND
               );

    if (me == 0) {

        /* Say where they are (debug)
         */
        DBGME(0,_tprintf(
           _TEXT("Using condition variables fooc=@%x bazc=@%x\n\r"),
                 (ADDRESS) &fooc, (ADDRESS) &bazc));
        Condition_Init(&fooc);
        Condition_Init(&bazc);

        DBGME(0,_tprintf(_TEXT("Initial thread started.\n\r")));

        /* Create the first thread, give it time to get started.
         */
        (void) CreateAThread(FirstThread, (THREAD_ARGUMENT) 1);
        for (k = 0; ; k++) {
            if (Tstates[1].Started)
                break;
            if (k > 20 && MaybeDead())
                goto Out;
            MilliSleep(1,me);
        }

        /* Ditto, second thread.
         */
        (void) CreateAThread(FirstThread, (THREAD_ARGUMENT) 2);
        for (k = 0; ; k++) {
            if (Tstates[2].Started)
                break;
            if (k > 20 && MaybeDead())
                goto Out;
            MilliSleep(1,me);
        }
    }

    DBGME(0,_tprintf(_TEXT("Thread %d is x%x\n\r"), me, (UINT)CurrentThread()));

    if (me == 0) {

        /* Do this 3 times: the first two times wake one thread,
         * the last time wake them all.
         * The idea is that they take turns waking up, then they
         * both wakeup and we are all happy.
         */
        for (i = 0; i < 3; i++) {

            /* Time to move on to the next step */
            _tprintf(_TEXT("Step %d\n\r"), ++Step);

            if (i == 2) {
                /* We are done, we'll syncup at the barrier down below
                 */
                DBGME(0,_tprintf(
                  _TEXT("Thread %d signalling all other threads.\n\r"), me));
                Condition_Broadcast(&fooc);

            } else {
                /* Let one out
                 */
                DBGME(0,_tprintf(
                  _TEXT("Thread %d signalling 1 other thread.\n\r"), me));
                Condition_Signal(&fooc);

                /* Make sure the right one wakes up: first #1 then #2
                 */
                for (k = 0; ; k++) {
                    if (Tstates[i+1].Awoke[0])
                        break;
                    if (k > 10 && MaybeDead())
                        goto Out;
                    MilliSleep(30,me);
                }
            }

        }
    } else {
        for (i = 0; i < 2; i++) {
            DBGME(0,_tprintf(_TEXT("Thread %d%s into Wait\n\r"), me,
                     (i == 0) ? _TEXT("") : _TEXT(" back")));

            /* Set this after all printouts */
            Tstates[me].Started = TRUE;
            Condition_Wait(&fooc, NULL, NULL);

            Tstates[me].Awoke[i] = TRUE;

            if (Step < 3 && (LastWoken == me)) {
                Passed = FALSE;
                _tprintf(_TEXT("WRONG thread woke up (%d).\n\r"), me);
            }
            LastWoken = me;

            /* Since this is run under timesharing scheduling..
             * makes sure we exhausted our timeslice else we'd
             * be run again next time.
             */
            t0 = CurrentTime();
            Int32ToInt64(t,TIME_MILLIS(20));
            t = TimeAdd(t0,t);
            for (k = 0; ;k++) {
                TIME Now = CurrentTime();
                if (!TimeLess(Now,t))
                    break;
                /* BUGBUG Should be spinning a bit here */
            }

        }
    }

    /* Barrier */
    AtomicInc((PUINT)&nHere);
    for (k = 0; nHere < 3; k++) {
        /* keep  quiet if we can */
        if (k > 0)
            _tprintf(_TEXT("Thread %d waiting for %d more..\n\r"),me,3-nHere);
        MilliSleep(200,me);
        if (k > 10 && MaybeDead())
            goto Out;
    }

    /* Now test a simple timed wait
     */
    {
        Int32ToInt64(t,TIME_RELATIVE(1000*10*10));
        DBGME(1,_tprintf(_TEXT("%d into TimedWait for %ld\n\r"), me, t));
        t0 = CurrentTime();
        Condition_Wait(&bazc, NULL, &t);
        t = CurrentTime();
        DBGME(1,_tprintf(_TEXT("%d was in TimedWait from %ld to %ld\n\r"),
                 me, t0, t));
        t0 = TimeSubtract(t,t0);
        Int32ToInt64(t,(1000*10*10));
        if (TimeLess(t0,t)) {
            _tprintf(_TEXT("%lld [< %lld] wasnt long enough to wait.\n\r"),
                     t0,t);
            Passed = FALSE;
        }
    }


    /* All threads should be done
     * NB: Only the last one gets past this.
     */
 Out:
    MaybeTerminate();

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return(0);
}
