/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"
/* Which test(s) should we perform
 */
#define P_TEST_CONSTR _TEXT("  Time constraints\n\r")

UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

BOOL Passed = TRUE;

int nn = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST_CONSTR
               );

    if (me == 0) {

        _tprintf(_TEXT("Initial thread started.\n\r"));
        (void) CreateAThread(FirstThread, NULL);
        (void) CreateAThread(FirstThread, NULL);
    }

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, (ADDRESS) CurrentThread());

    {
        TIME_CONSTRAINT MyCos;
        TIME t0, t1;
        SCODE sc0, sc1;
        MyCos.Criticality = /*_NON*/_CRITICAL;
        /* start time of now plus "me" milliseconds */
        Int32ToInt64(MyCos.Start,TIME_RELATIVE(me*40000*10));
        /* run time of 4 ms */
        Int32ToInt64(MyCos.Estimate,40000*10);
        /* deadline me*4 ms */
        Int32ToInt64(MyCos.Deadline,TIME_RELATIVE((me+1)*40000*10));

        sc0 = BeginConstraint( FALSE, &MyCos, NULL);
        t0 = CurrentTime();
        /* Printing will take roughly 40 * 1 ms = 40 ms on 9600 bps */
        _tprintf(_TEXT("%d is inside a constraint at %lld\n\r"), me, t0);
        sc1 = EndConstraint(&t1);

        _tprintf(_TEXT("%d time results are %x %x %lld\n\r"),
               me, sc0, sc1, t1);
        if (sc0 != S_OK || sc1 != S_OK)
            Passed = FALSE;
        Delay(3000);
    }

    /* All threads should be done
     */
    MaybeTerminate();

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return(0);
}
