/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"

/* Which test(s) should we perform
 */
#define P_TEST_MUTEX _TEXT("  Mutex object\n\r")


UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

BOOL Passed = TRUE;

MUTEX foom;

int nn = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST_MUTEX
               );

    if (me == 0) {

        Mutex_Init(&foom);

        _tprintf(_TEXT("Initial thread started.\n\r"));
        (void) CreateAThread(FirstThread, NULL);
        (void) CreateAThread(FirstThread, NULL);
    }

    Delay(3000);

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, (ADDRESS) CurrentThread());

#define NLOOPS 12

    for (i = 0;i < NLOOPS; i++) {
        TIME t;

        Mutex_Lock(&foom);

        t = CurrentTime();
        _tprintf(_TEXT("%d at %lld\n\r"), me, t);

        if (!Mutex_Locked(&foom))
            Passed = FALSE;

        if (Mutex_TryLock(&foom))
            Passed = FALSE;

        Delay(1000);

        Mutex_Unlock(&foom);

        Delay(1000);
    }

    /* All threads should be done
     */
    MaybeTerminate();

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    
    return(0);
}

