/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"
/* Which test(s) should we perform
 */
#define P_TEST1 _TEXT("  TimeShare\n\r")

UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

BOOL Passed = TRUE;

int nn = 0;
volatile UINT nHere = 0;

#define NLOOPS 40 /* x3 = 120 */

int WhoRan[(NLOOPS*3) + 1];
int CurPos = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i, prev;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_TEXT("Will perform the following test:\n\r") \
               P_TEST1
               );

    if (me == 0) {

        _tprintf(_TEXT("Initial thread started.\n\r"));
        (void) CreateAThread(FirstThread, NULL);
        (void) CreateAThread(FirstThread, NULL);
    }

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, CurrentThread());

    /* Barrier */
    AtomicInc((PUINT)&nHere);
    while (nHere < 3)
        ;

    for (i = 0;i < NLOOPS; i++) {
        TIME t;

        Int32ToInt64(t,TIME_RELATIVE(25000)); /* 2.5 milli */
        SleepUntil(t);

        WhoRan[AtomicInc((PUINT)&CurPos)] = me;
/*      _tprintf(_TEXT("%d"), me); */
    }

    /* All threads should be done
     */
    MaybeTerminate();

    /* So how did it go ?
     */
    for (prev = 1; prev <= CurPos; prev++)
        _tprintf(_TEXT("%d"), WhoRan[prev]);
    _tprintf(_TEXT("\n\r"));

    prev = -1;
    for (i = 1; i <= CurPos; i++) {
        if (prev == WhoRan[i]) {
            Passed = FALSE;
            break;
        }
        prev = WhoRan[i];
    }

    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    
    return(0);
}
