/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* This test tests that Condition_Wait returns the proper error code
 * in a number of cases.
 */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

CONDITION cd;
MUTEX mx;

TIME t1 = Int64Initializer(-1, TIME_RELATIVE(TIME_MILLIS(500)));
TIME t2 = Int64Initializer(-1, TIME_RELATIVE(TIME_SECONDS(2)));

void THREAD_LINKAGE MyTimerThread(THREAD_ARGUMENT arg)
{
    /* First test is to see that thread creation did not interfere with
     * condition timeouts below.
     */
    Delay(1000000);
}

void THREAD_LINKAGE MyHelperThread(THREAD_ARGUMENT arg)
{
    /* Wait for main thread to sleep */
    Mutex_Lock(&mx);
    Mutex_Unlock(&mx);

    Condition_Broadcast(&cd);

    SleepUntil(t1);

    Condition_Broadcast(&cd);
}

int _tmain(int argc, _TCHAR **argv)
{
    SCODE sc;
    PIPROCESS prc = CurrentProcess();

    Condition_Init(&cd);
    Mutex_Init(&mx);

    sc = prc->v->CreateThread(prc, 0, MyTimerThread,
                              NULL, 0, NULL, NULL);
    if (sc != S_OK)
        printf("tcond 1 FAILED sc=%x\n", sc);

    sc = Condition_Wait(&cd, NULL, &t1);
    if (sc != E_TIMED_OUT)
        printf("tcond 2 FAILED sc=%x\n", sc);

    Mutex_Lock(&mx);

    sc = Condition_Wait(&cd, &mx, &t1);
    if (sc != E_TIMED_OUT)
        printf("tcond 3 FAILED sc=%x != E_TIMED_OUT\n", sc);

    sc = prc->v->CreateThread(prc, 0, MyHelperThread,
                              NULL, 0, NULL, NULL);
    if (sc != S_OK)
        printf("tcond 4 FAILED sc=%x\n", sc);

    sc = Condition_Wait(&cd, &mx, &t2);
    if (sc != S_OK)
        printf("tcond 5 FAILED sc=%x\n", sc);

    sc = Condition_Wait(&cd, &mx, NULL);
    if (sc != S_OK)
        printf("tcond 6 FAILED sc=%x\n", sc);

    Mutex_Unlock(&mx);

    Condition_Destroy(&cd);
    Mutex_Destroy(&mx);

    printf("tcond main thread done\n");
    return 0;
}


