/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

/* uglyness due to arm compilerbrain */
#if defined(SIM)
#if (SIM == 1)
#define SIMULATED_CPU 1
#endif
#endif

#ifdef SIMULATED_CPU
#define TIME_TO_SLEEP TIME_MILLIS(100)
#else
#define TIME_TO_SLEEP TIME_SECONDS(10)
#endif

BOOL test1(void)
{
    MUTEX Mutex;
    BOOL Passed = TRUE;
    UINT ret;

    _tprintf(_TEXT("Creating a mutex %08x\n\r"), (ADDRESS) &Mutex);
    Mutex_Init(&Mutex);

    ret = Mutex_TryLock(&Mutex);
    _tprintf(_TEXT("TMutex: Trylock returns %d:\n\r"), ret);
    if (ret != 1)
        Passed = FALSE;

    ret = Mutex_TryLock(&Mutex);
    _tprintf(_TEXT("TMutex: Trylock returns %d:\n\r"), ret);
    if (ret != 0)
        Passed = FALSE;

    _tprintf(_TEXT("TMutex: Unlocking mutex\n\r"));
    Mutex_Unlock(&Mutex);

    _tprintf(_TEXT("TMutex: Locking mutex\n\r"));
    Mutex_Lock(&Mutex);

    Mutex_Unlock(&Mutex);
    Mutex_Destroy(&Mutex);

    _tprintf(_TEXT("TMutex: test1 %hs\n\r"),
             (Passed) ? "PASSED" : "FAILED");
    return (Passed);
}

#define MAX_NUM_THREADS         50
#define DEFAULT_NUM_THREADS     5
#define DEFAULT_TOTAL_COUNT     10000

volatile UINT NumThreads = DEFAULT_NUM_THREADS;
UINT TotalCount = DEFAULT_TOTAL_COUNT;
UINT CountPerThread; /* calculated by main() */

volatile UINT Sum;
MUTEX GlobalMutex;
/* PrintMutex isn't really necessary anymore, because the Print
 * implementation serializes correctly, but the extra lock/unlock
 * calls seem to add to the test and trigger interesting behavior
 * on occasion.
 */
MUTEX PrintMutex;

void THREAD_LINKAGE test2_thread(THREAD_ARGUMENT arg)
{
    UINT i;
    PMUTEX Mutex = (PMUTEX) &GlobalMutex;
    MUTEX DummyMutex;
    INT self = (INT) arg;
    INT NumThreadsAtExit;

    Mutex_Lock(&PrintMutex);
    _tprintf(_TEXT("TMutex: thread %d: begin\n\r"), self);
    Mutex_Unlock(&PrintMutex);

    Mutex_Init(&DummyMutex);
    for (i = 0; i < CountPerThread; i++) {
        Mutex_Lock(Mutex);
        Mutex_Lock(&DummyMutex);
        Sum++;
        Mutex_Unlock(&DummyMutex);
        Mutex_Unlock(Mutex);
    }

    Mutex_Lock(Mutex);
    NumThreadsAtExit = --NumThreads;
    Mutex_Unlock(Mutex);
    Mutex_Destroy(&DummyMutex);

    /* The first thread sleeps before calling ThreadExit.
     * The idea here is to have the rest of the test finish,
     * and have the main thread call ProcessExit, while
     * this thread is still sleeping. This tests the case
     * of a callback into a dead process.
     */

    Mutex_Lock(&PrintMutex);
    _tprintf(_TEXT("TMutex: thread %d: %hs (%d threads left)\n\r"), self,
             self == 0 ? "sleeping then end" : "end",
             NumThreadsAtExit);
    Mutex_Unlock(&PrintMutex);
    if (self == 0) {
        TIME t;
        Int32ToInt64(t,TIME_RELATIVE(TIME_TO_SLEEP));
        SleepUntil(t);
        _tprintf(_TEXT("TMutex: thread %d slept long enough.\n\r"), self);
    }
    ThreadExit();
}

void THREAD_LINKAGE CallBack(THREAD_ARGUMENT arg)
{
    INT thd = (INT) arg;

    _tprintf(_TEXT("CallBack: thread %d terminated\n\r"), thd);
}


PITHREAD pThd[MAX_NUM_THREADS];

BOOL test2(void)
{
    UINT i, j, nT;
    PIPROCESS pIPrc;
    UINT OldSum;
    PITHREAD Owner;

    Mutex_Init(&GlobalMutex);
    Mutex_Init(&PrintMutex);
    pIPrc = CurrentProcess();

    /* Cache how many threads we should create.
     * Else we might create too few, and this never ends.
     */
    nT = NumThreads;

    for (i = 0; i < nT; i++) {
        SCODE r;
        r = pIPrc->v->CreateThread(pIPrc, 0, test2_thread,
                                   (THREAD_ARGUMENT)i,
                                   0, NULL, &pThd[i]);
        Mutex_Lock(&PrintMutex);
        _tprintf(_TEXT("TMutex: Back from CreateThread, got %x\n\r"),
                 (ADDRESS) pThd[i] );
        if (E_NOT_ENOUGH_MEMORY == r)
                _tprintf(_TEXT("TMutex: E_NOT_ENOUGH_MEMORY\n\r"));
        Mutex_Unlock(&PrintMutex);

        if (pThd[i] == NULL)
            continue;

    }

    Mutex_Lock(&PrintMutex);
    _tprintf(_TEXT("TMutex: Sum (%d) should be %d\n\r"), Sum, TotalCount);
    Mutex_Unlock(&PrintMutex);


    for (OldSum = Sum, i = 1; Sum < TotalCount ; i++) {

        TIME tsleep;

        _tprintf(_TEXT("TMutex: Sum is %d, sleeping..\n\r"), Sum);
#if arm
        Int32ToInt64(tsleep,TIME_RELATIVE(TIME_MILLIS(100)));
#else
        Int32ToInt64(tsleep,TIME_RELATIVE(TIME_SECONDS(1)));
#endif
        SleepUntil(tsleep);

        Owner = (PITHREAD) (GlobalMutex._mutex_state[0] & ~1);
        for (j = 0; j < nT; j++)
            if (pThd[j] == Owner)
                break;

        if (j == nT)
            j = (UINT) -1;

        Mutex_Lock(&PrintMutex);
        _tprintf(_TEXT("TMutex: Sum = %7d (%6d / sec) owner = %08x (%d)\n\r"),
                 Sum,
#if arm
                 Sum / i,
#else
                 Sum / (INT32) (i * 10),
#endif
                 (ADDRESS) Owner, j);

        Mutex_Unlock(&PrintMutex);
        if ((i % 5) == 0) {
            if (Sum == OldSum) {
                Mutex_Lock(&PrintMutex);
                _tprintf(_TEXT("TMutex: stuck at %d; ending test\n\r"), Sum);
                Mutex_Unlock(&PrintMutex);
                break;
            }
            OldSum = Sum;
        }
    }

    for (i = 0; i < nT; i++) {
        if (pThd[i] == NULL)
            continue;
        pThd[i]->v->Release(pThd[i]);
    }

    Mutex_Lock(&PrintMutex);
    _tprintf(_TEXT("TMutex: Sum = %d; should be %d\n\r"), Sum, TotalCount);
    Mutex_Unlock(&PrintMutex);

    if (Sum != TotalCount) {
        _tprintf(_TEXT("TMutex: test2 FAILED\n\r"));
    } else
        _tprintf(_TEXT("TMutex: test2 PASSED\n\r"));
        /* but what did thread #0 do.. */
    return (Sum == TotalCount);
}

INT _tmain(INT argc, _TCHAR * argv[])
{
    switch (argc) {
    case 3:
        TotalCount = _ttoi(argv[2]);
    case 2:
        NumThreads = _ttoi(argv[1]);
    case 1:
    case 0:
        break;
    default:
        _tprintf(_TEXT("usage: tmutex [NumThreads [TotalCount]]\n\r"));
        exit(EXIT_FAILURE);
    }

    /* NumThreads must satisfy 1 <= NumThreads <= MAX_NUM_THREADS */
    if (NumThreads < 1 || NumThreads > MAX_NUM_THREADS)
        NumThreads = DEFAULT_NUM_THREADS;

    /* the code above assumes that NumThreads divides TotalCount */
    CountPerThread = TotalCount / NumThreads;
    TotalCount = CountPerThread * NumThreads;

    _tprintf(_TEXT("TMutex: using %d threads, counting to %d (%d/thread)\n\r"),
             NumThreads, TotalCount, CountPerThread);

    if (!test1())
        exit(EXIT_FAILURE);
    if (!test2())
        exit(EXIT_FAILURE);
    _tprintf(_TEXT("This test PASSED, exiting\n\r"));
    exit(EXIT_SUCCESS);
}
