/* tree-avl-test.c
 * Tests for AVL tree and AVL node interface defined in include/util/tree-avl.h
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <util/test.h>
#include <util/tree-avl.h>

/* Global variables for testing */

typedef struct _TestStruct {
  int value;
} TestStruct, *pTestStruct;

#define pTS(_pts_) ((pTestStruct)(_pts_))
#define StructGetValue(_pts_) (pTS(_pts_)->value)
#define StructSetValue(_pts_, _value_) (StructGetValue(_pts_) = (_value_))
#define STRUCT_TRANSFORM_VALUE 22

PRIVATE TestStruct ts1;
PRIVATE TestStruct ts2;
PRIVATE TestStruct ts3;
PRIVATE TestStruct ts4;
PRIVATE TestStruct ts5;
PRIVATE TestStruct ts6;
PRIVATE TestStruct ts7;
PRIVATE TestStruct ts8;
PRIVATE pTestStruct pts1 = &ts1;
PRIVATE pTestStruct pts2 = &ts2;
PRIVATE pTestStruct pts3 = &ts3;
PRIVATE pTestStruct pts4 = &ts4;
PRIVATE pTestStruct pts5 = &ts5;
PRIVATE pTestStruct pts6 = &ts6;
PRIVATE pTestStruct pts7 = &ts7;
PRIVATE pTestStruct pts8 = &ts8;

PRIVATE ATNODE atnNode1;
PRIVATE ATNODE atnNode2;
PRIVATE ATNODE atnNode3;
PRIVATE ATNODE atnNode4;
PRIVATE ATNODE atnNode5;
PRIVATE ATNODE atnNode6;
PRIVATE ATNODE atnNode7;
PRIVATE PATNODE patnNode1 = &atnNode1;
PRIVATE PATNODE patnNode2 = &atnNode2;
PRIVATE PATNODE patnNode3 = &atnNode3;
PRIVATE PATNODE patnNode4 = &atnNode4;
PRIVATE PATNODE patnNode5 = &atnNode5;
PRIVATE PATNODE patnNode6 = &atnNode6;
PRIVATE PATNODE patnNode7 = &atnNode7;
PRIVATE AVLTREE at1;
PRIVATE PAVLTREE pat1 = & at1;

PRIVATE PATNODE patnRoot;

/* Function prototypes for tests */

BOOL ATNode_TestInitializer(void);
BOOL Test_ATNode_Init(void);
BOOL Test_ATNode_GetLeft(void);
BOOL Test_ATNode_GetRight(void);
BOOL Test_ATNode_GetData(void);
BOOL Test_ATNode_GetBalance(void);
BOOL Test_ATNode_AdjustLeft(void);
BOOL Test_ATNode_AdjustRight(void);
BOOL Test_ATNode_AdjustDoubleLeft(void);
BOOL Test_ATNode_AdjustDoubleRight(void);
BOOL Test_ATNode_IsLeaf(void);
BOOL Test_ATNode_IsSentinel(void);
BOOL Test_ATNode_IsBalanced(void);
BOOL Test_ATNode_RotateLeft(void);
BOOL Test_ATNode_RotateRight(void);
BOOL Test_ATNode_Search(void);

BOOL AVLTree_TestInitializer(void);
BOOL Test_AVLTree_Init(void);
BOOL Test_AVLTree_IsEmpty(void);
BOOL Test_AVLTree_GetRoot(void);
BOOL Test_AVLTree_GetSortFunc(void);
BOOL Test_AVLTree_GetSentinel(void);
BOOL Test_AVLTree_Insert(void);
BOOL Test_AVLTree_Insert2(void);
BOOL Test_AVLTree_Search(void);
BOOL Test_AVLTree_Remove(void);
BOOL Test_AVLTree_Remove2(void);
BOOL Test_AVLTree_Remove3(void);
BOOL Test_StructGT(void);

/* Prototypes for auxiliary functions */

BOOL StructGT(
    PTR pts1,
    PTR pts2
    );

BOOL TestGT(
    PTR pts1,
    PTR pts2
    );

int main(void) {
  TestFunctionPair ATNodeTestFunctions[] = {
    {ATNode_TestInitializer,          ""},
    {Test_ATNode_Init,                "ATNode_Init"},
    {Test_ATNode_GetLeft,             "ATNode_GetLeft"},
    {Test_ATNode_GetRight,            "ATNode_GetRight"},
    {Test_ATNode_GetData,             "ATNode_GetData"},
    {Test_ATNode_GetBalance,          "ATNode_GetData"},
    {Test_ATNode_AdjustLeft,          "ATNode_GetData"},
    {Test_ATNode_AdjustDoubleLeft,    "ATNode_GetData"},
    {Test_ATNode_AdjustRight,         "ATNode_GetData"},
    {Test_ATNode_AdjustDoubleRight,   "ATNode_GetData"},
    {Test_ATNode_IsLeaf,              "ATNode_IsLeaf"},
    {Test_ATNode_IsSentinel,          "ATNode_IsSentinel"},
    {Test_ATNode_IsBalanced,          "ATNode_IsBalanced"},
    {Test_ATNode_RotateLeft,          "ATNode_RotateLeft"},
    {Test_ATNode_RotateRight,         "ATNode_RotateRight"},
    {Test_ATNode_Search,              "ATNode_Search"}
  };

  TestFunctionPair AVLTreeTestFunctions[] = {
    {AVLTree_TestInitializer,           ""                   },
    {Test_AVLTree_Init,                 "AVLTree_Init"       },
    {Test_AVLTree_IsEmpty,              "AVLTree_IsEmpty"    },
    {Test_AVLTree_GetRoot,              "AVLTree_GetRoot"    },
    {Test_AVLTree_GetSortFunc,          "AVLTree_GetSortFunc"},
    {Test_AVLTree_GetSentinel,          "AVLTree_GetSentinel"},
    {Test_AVLTree_Insert,               "AVLTree_Insert"     },
    {Test_AVLTree_Insert2,              "AVLTree_Insert2"    },
    {Test_AVLTree_Search,               "AVLTree_Search"     },
    {Test_AVLTree_Remove,               "AVLTree_Remove"     },
    {Test_AVLTree_Remove2,              "AVLTree_Remove2"    },
    {Test_AVLTree_Remove3,              "AVLTree_Remove3"    }
 };

  RunTestSuite("AVL Tree Node Test Suite", ATNodeTestFunctions,
	       GetTestFunctionArraySize(ATNodeTestFunctions));
  RunTestSuite("AVL Tree Test Suite", AVLTreeTestFunctions,
	       GetTestFunctionArraySize(AVLTreeTestFunctions));

  return 0;
}

BOOL ATNode_TestInitializer(void)
{
  ATNode_Init(patnNode1);
  ATNode_SetData(patnNode1, pts1);
  ATNode_Init(patnNode2);
  ATNode_SetData(patnNode2, pts2);
  ATNode_Init(patnNode3);
  ATNode_SetData(patnNode3, pts3);
  ATNode_Init(patnNode4);
  ATNode_SetData(patnNode4, pts4);
  ATNode_Init(patnNode5);
  ATNode_SetData(patnNode5, pts5);
  ATNode_Init(patnNode6);
  ATNode_SetData(patnNode6, pts6);
  ATNode_Init(patnNode7);
  ATNode_SetData(patnNode7, pts7);

  StructSetValue(pts1, 4);
  StructSetValue(pts2, 2);
  StructSetValue(pts3, 6);
  StructSetValue(pts4, 1);
  StructSetValue(pts5, 3);
  StructSetValue(pts6, 5);
  StructSetValue(pts7, 7);
  StructSetValue(pts8, 8);

  patnRoot = patnNode1;

  return TRUE;
}

BOOL Test_ATNode_Init(void)
{
  BOOL success = TRUE;

  ATNode_Init(patnNode1);

  success &= compare_integer(NULL, ATNode_GetData(patnNode1),
			     "sets data pointer to NULL");

  success &= compare_integer(NULL, ATNode_GetLeft(patnNode1),
			     "sets left pointer to NULL");

  success &= compare_integer(NULL, ATNode_GetRight(patnNode1),
			     "sets right pointer to NULL");

  return success;
}

BOOL Test_ATNode_GetLeft(void)
{
  BOOL success = TRUE;

  ATNode_SetLeft(patnNode1, patnNode2);

  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode1),
			     "set left == get left");

  return success;
}

BOOL Test_ATNode_GetRight(void)
{
  BOOL success = TRUE;

  ATNode_SetRight(patnNode3, patnNode4);

  success &= compare_integer(patnNode4, ATNode_GetRight(patnNode3),
			     "set right == get right");

  return success;
}

BOOL Test_ATNode_GetData(void)
{
  BOOL success = TRUE;

  ATNode_SetData(patnNode2, (PTR)0x2222);

  success &= compare_integer(0x2222, ATNode_GetData(patnNode2),
			     "set data == get data");

  return success;
}

BOOL Test_ATNode_GetBalance(void)
{
  BOOL success = TRUE;

  ATNode_SetBalance(patnNode2, LEFT_HEAVY);

  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode2),
			     "set balance == get balance");

  return success;
}

BOOL Test_ATNode_AdjustLeft(void)
{
  BOOL success = TRUE;

  ATNode_AdjustLeft(patnNode2);

  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting left works");

  ATNode_AdjustLeft(patnNode2);

  success &= compare_integer(DOUBLE_LEFT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting left twice works");

  return success;
}

BOOL Test_ATNode_AdjustRight(void)
{
  BOOL success = TRUE;

  ATNode_AdjustRight(patnNode2);

  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting right works");

  ATNode_AdjustRight(patnNode2);

  success &= compare_integer(DOUBLE_RIGHT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting right twice works");

  return success;
}

BOOL Test_ATNode_AdjustDoubleLeft(void)
{
  BOOL success = TRUE;

  ATNode_AdjustDoubleLeft(patnNode2);

  success &= compare_integer(DOUBLE_LEFT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting double left works");

  return success;
}

BOOL Test_ATNode_AdjustDoubleRight(void)
{
  BOOL success = TRUE;

  ATNode_AdjustDoubleRight(patnNode2);

  success &= compare_integer(DOUBLE_RIGHT_HEAVY, ATNode_GetBalance(patnNode2),
			     "adjusting double right works");

  return success;
}

BOOL Test_ATNode_IsSentinel(void)
{
  BOOL success = TRUE;

  ATNode_SetLeft(patnNode7, patnNode7);
  ATNode_SetRight(patnNode7, patnNode7);

  success &= compare_integer(TRUE, ATNode_IsSentinel(patnNode7),
			     "node with self-loops is a sentinel");

  return success;
}

BOOL Test_ATNode_IsLeaf(void)
{
  BOOL success = TRUE;

  ATNode_Init(patnNode1);
  success &= compare_integer(TRUE, ATNode_IsLeaf(patnNode1),
			     "initialized node is a leaf");

  TNode_SetLeft(patnNode1, patnNode5);
  success &= compare_integer(FALSE, ATNode_IsLeaf(patnNode1),
			     "node with left is not a leaf");

  TNode_SetRight(patnNode1, patnNode5);
  success &= compare_integer(FALSE, ATNode_IsLeaf(patnNode1),
			     "node with right is not a leaf");

  return success;
}

BOOL Test_ATNode_IsBalanced(void)
{
  BOOL success = TRUE;

  success &= compare_integer(TRUE, ATNode_IsBalanced(patnNode2),
			     "nodes are initially balanced");

  ATNode_AdjustLeft(patnNode2);
  ATNode_AdjustRight(patnNode2);
  success &= compare_integer(TRUE, ATNode_IsBalanced(patnNode2),
			     "node adjusted left and back is balanced");

  return success;
}

BOOL Test_ATNode_RotateLeft(void)
{
  BOOL success = TRUE;

  ATNode_SetLeft(patnNode2, patnNode4);
  ATNode_SetRight(patnNode2, patnNode1);
  ATNode_SetBalance(patnNode2, DOUBLE_RIGHT_HEAVY);

  ATNode_SetLeft(patnNode1, patnNode5);
  ATNode_SetRight(patnNode1, patnNode3);
  ATNode_SetBalance(patnNode1, RIGHT_HEAVY);

  ATNode_SetLeft(patnNode3, patnNode6);
  ATNode_SetRight(patnNode3, patnNode7);

  patnRoot = patnNode2;
  ATNode_RotateLeft(&patnRoot, FALSE);
  success &= compare_integer(patnNode1, patnRoot,
			     "right child is the new top");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode1),
			     "left child of new top is old top");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode1),
			     "right child of new top is same as before");
  success &= compare_integer(patnNode5, ATNode_GetRight(patnNode2),
			     "right child of old top is old left child of new top");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode2),
			     "left child of old top is same as before");
  success &= compare_integer(patnNode6, TNode_GetLeft(patnNode3),
			     "right-left grandchild same as before");
  success &= compare_integer(patnNode7, ATNode_GetRight(patnNode3),
			     "right-right grandchild same as before");

  return success;
}

BOOL Test_ATNode_RotateRight(void)
{
  BOOL success = TRUE;

  ATNode_SetLeft(patnNode3, patnNode1);     /*             patn3(6)       */
  ATNode_SetRight(patnNode3, patnNode7);    /*             /      \       */
  ATNode_SetBalance(patnNode3, DOUBLE_LEFT_HEAVY); /*     /      patn7(7) */
                                            /*        patn1(4)            */
  ATNode_SetLeft(patnNode1, patnNode2);     /*       /        \           */
  ATNode_SetRight(patnNode1, patnNode6);    /*     patn2(2)    patn6(5)   */
  ATNode_SetBalance(patnNode1, LEFT_HEAVY); /*     /       \              */
                                            /* patn4(1)      patn5(3)     */
  ATNode_SetLeft(patnNode2, patnNode4);
  ATNode_SetRight(patnNode2, patnNode5);

  patnRoot = patnNode3;
  ATNode_RotateRight(&patnRoot, FALSE);
  success &= compare_integer(patnNode1, patnRoot,
			     "left child is the new top");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode1),
			     "right child of new top is old top");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode1),
			     "left child of new top is same as before");
  success &= compare_integer(patnNode6, ATNode_GetLeft(patnNode3),
			     "left child of old top is old right child of new top");
  success &= compare_integer(patnNode7, ATNode_GetRight(patnNode3),
			     "right child of old top is same as before");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode2),
			     "left-left grandchild same as before");
  success &= compare_integer(patnNode5, ATNode_GetRight(patnNode2),
			     "left-right grandchild same as before");

  return success;
}

BOOL Test_ATNode_Search(void)
{
  BOOL success = TRUE;

  TNode_SetLeft(patnNode1, patnNode2);   /*            patn1(4)              */
  TNode_SetRight(patnNode1, patnNode3);  /*          /         \             */
                                         /*    patn2(2)      patn3 (6)       */
  TNode_SetLeft(patnNode2, patnNode4);   /*    /      \          /  \        */
  TNode_SetRight(patnNode2, patnNode5);  /* patn4(1) patn5(3)   /    \       */
					 /*                patn6(5) patn7(7) */
  TNode_SetLeft(patnNode3, patnNode6);
  TNode_SetRight(patnNode3, patnNode7);

  success &= compare_integer(patnNode4, ATNode_Search(patnNode1, StructGT, pts4),
			     "node 4 is found");
  success &= compare_integer(NULL, ATNode_Search(patnNode1, StructGT, pts8),
			     "test struct 8 is not found");

  ATNode_Init(patnNode1);
  ATNode_SetData(patnNode1, pts1);
  success &= compare_integer(NULL, ATNode_Search(patnNode1, StructGT, pts4),
			     "node 4 is not found after root is initialized");

  return success;
}

BOOL AVLTree_TestInitializer(void)
{
  AVLTree_Init(pat1, StructGT);
  ATNode_TestInitializer();

  return TRUE;
}

BOOL Test_AVLTree_Init(void)
{
  BOOL success = TRUE;

  AVLTree_Init(pat1, NULL);

  success &= compare_integer(NULL, AVLTree_GetRoot(pat1),
			     "sets root to NULL");

  success &= compare_integer(NULL, AVLTree_GetSortFunc(pat1),
			     "sets SortFunc to NULL");

  return success;
}

BOOL Test_AVLTree_IsEmpty(void)
{
  BOOL success = TRUE;

  AVLTree_SetRoot(pat1, patnNode1);

  success &= compare_integer(FALSE, AVLTree_IsEmpty(pat1),
			     "tree with nodes is not empty");


  AVLTree_Init(pat1, NULL);

  success &= compare_integer(TRUE, AVLTree_IsEmpty(pat1),
			     "initialized tree is empty"); 

  return success;
}

BOOL Test_AVLTree_GetRoot(void)
{
  BOOL success = TRUE;

  AVLTree_SetRoot(pat1, patnNode2);
  success &= compare_integer(patnNode2, AVLTree_GetRoot(pat1),
			     "get root == set root");

  return success;
}

BOOL Test_AVLTree_GetSortFunc(void)
{
  BOOL success = TRUE;

  AVLTree_SetSortFunc(pat1, TestGT);
  success &= compare_integer(TestGT, AVLTree_GetSortFunc(pat1),
			     "get sortfunc == set sortfunc");

  return success;
}

BOOL Test_AVLTree_GetSentinel(void)
{
  BOOL success = TRUE;

  PATNODE pSentinel = AVLTree_GetSentinel(pat1);

  success &= compare_integer(TRUE, ATNode_IsSentinel(pSentinel),
			     "avl tree returns a valid sentinel");

  return success;
}

/* Tests CASE B1 and CASE B2 in rebalancing */
BOOL Test_AVLTree_Insert(void)
{
  BOOL success = TRUE;

  AVLTree_Insert(pat1, patnNode1);

  /*         patn1(4)=
   */
  success &= compare_integer(patnNode1, AVLTree_GetRoot(pat1),
			     "inserting node into empty tree makes it root");

  AVLTree_Insert(pat1, patnNode2);

  /*         patn1(4)/
   *        /
   *     patn2(2)=
   */
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode1),
			     "left child is inserted correctly");

  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode1),
			     "root is left-heavy");
  
  AVLTree_Insert(pat1, patnNode4);

  /*         patn1(4)//                      pstn2(2)=
   *        /                \\              /       \
   *      patn2(2)/       =====>       patn4(1)=    patn1(4)=
   *     /                   //
   *   patn4(1)=
   */
  success &= compare_integer(patnNode2, AVLTree_GetRoot(pat1),
			     "making double-left-heavy rebalances");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode2),
			     "left-left grandchild is inserted correctly");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode1),
			     "old root is balanced");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode2),
			     "new root is balanced");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode4),
			     "new child is balanced");

  AVLTree_Insert(pat1, patnNode3);

  /*         patn2(2)\
   *        /      \
   *   patn4(1)=   patn1(4)\
   *                    \
   *                    patn3(6)=
   */
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode1),
			     "right-right grandchild is inserted correctly");
  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode2),
			     "root is right-heavy");
  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode1),
			     "right-child is right-heavy");

  AVLTree_Insert(pat1, patnNode7);
 
  /*         patn2(2)\?                          patn2(2)\
   *        /      \                             /      \
   *   patn4(1)=   patn1(4)\\         \\    patn4(1)=  patn3(6)=
   *                    \           ====>             /      \
   *                    patn3(6)\     //        patn1(4)=    patn7(7)=
   *                         \
   *                         patn7(7)=
   */
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode2),
			     "making double-right-heavy rebalances");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode3),
			     "rebalanced parent is now balanced");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode1),
			     "rebalanced left child is now balanced");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode7),
			     "newly-inserted child is balanced");
  success &= compare_integer(patnNode2, AVLTree_GetRoot(pat1),
			     "rebalancing keeps same root");
  success &= compare_integer(patnNode7, ATNode_GetRight(patnNode3),
			     "right-right grandchild is inserted correctly");

  return success;

}

/* Tests CASE B3 and CASE B4 in rebalancing */
BOOL Test_AVLTree_Insert2(void)
{
  BOOL success = TRUE;

  AVLTree_Insert(pat1, patnNode1);
  AVLTree_Insert(pat1, patnNode2);
  AVLTree_Insert(pat1, patnNode5);

  success &= compare_integer(patnNode5, AVLTree_GetRoot(pat1),
			     "CASE 3 rebalances correctly");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode5),
			     "new root is balanced");
  success &= compare_integer(patnNode1, ATNode_GetRight(patnNode5),
			     "new root has correct right child");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode5),
			     "new root has correct left child");
  
  AVLTree_Insert(pat1, patnNode3);
  AVLTree_Insert(pat1, patnNode6);

  success &= compare_integer(patnNode5, AVLTree_GetRoot(pat1),
			     "CASE 4 rebalances correctly");
  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode5),
			     "new root is right-heavy");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode6),
			     "new right child is balanced");
  success &= compare_integer(patnNode1, ATNode_GetLeft(patnNode6),
			     "correct right-left grandchild");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode6),
			     "correct right-right grandchild");

  return success;

}

BOOL Test_AVLTree_Search(void)
{
  BOOL success = TRUE;

  AVLTree_Insert(pat1, patnNode1);
  AVLTree_Insert(pat1, patnNode2);
  AVLTree_Insert(pat1, patnNode3);
  AVLTree_Insert(pat1, patnNode4);
  AVLTree_Insert(pat1, patnNode5);
  AVLTree_Insert(pat1, patnNode6);
  AVLTree_Insert(pat1, patnNode7);

  success &= compare_integer(patnNode5, AVLTree_Search(pat1, pts5),
			     "node 5 is found");

  success &= compare_integer(NULL, AVLTree_Search(pat1, pts8),
			     "node 8 not found");

  AVLTree_Init(pat1, StructGT);

  success &= compare_integer(NULL, AVLTree_Search(pat1, pts5),
			     "node 5 is not found after tree is initialized");

  return success;
}

/* Tests CASES 1, 2a, and B1 */
BOOL Test_AVLTree_Remove(void)
{
  BOOL success = TRUE;
  TREE_RESULT_ENUM treeResult;

  AVLTree_Insert(pat1, patnNode1);
  AVLTree_Insert(pat1, patnNode2);
  AVLTree_Insert(pat1, patnNode3);
  AVLTree_Insert(pat1, patnNode4);
  AVLTree_Insert(pat1, patnNode5);
  AVLTree_Insert(pat1, patnNode6);
  AVLTree_Insert(pat1, patnNode7);

  printf("patnNode1 %x\n", patnNode1);
  printf("patnNode2 %x\n", patnNode2);
  printf("patnNode3 %x\n", patnNode3);
  printf("patnNode4 %x\n", patnNode4);
  printf("patnNode5 %x\n", patnNode5);
  printf("patnNode6 %x\n", patnNode6);
  printf("patnNode7 %x\n", patnNode7);

  // Tests CASE 1: removing a leaf
  treeResult = AVLTree_Remove(pat1, patnNode7);
  success &= compare_integer(SUCCESS, treeResult, "key was found");
  success &= compare_integer(FALSE, ATNode_HasRight(patnNode3),
			     "parent no longer has removed right child");
  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode3),
			     "parent is now left-heavy");
  success &= compare_integer(BALANCED, ATNode_GetBalance(patnNode1),
			     "root is still balanced");
  success &= compare_integer(patnNode6, ATNode_GetLeft(patnNode3),
			     "parent still has correct surviving child");
  success &= compare_integer(TRUE, ATNode_IsLeaf(patnNode6),
			     "surviving child is still a leaf");
  success &= compare_integer(FALSE, ATNode_IsLeaf(patnNode3),
			     "parent is not a leaf");
  success &= compare_integer(patnNode6, ATNode_ChildPredecessor(patnNode3),
			     "correct pred for next node to be removed");

  // Tests CASE R2a: removing a left-heavy direct-parent
  treeResult = AVLTree_Remove(pat1, patnNode3);
  success &= compare_integer(SUCCESS, treeResult, "key was found");
  success &= compare_integer(TRUE, ATNode_IsLeaf(patnNode3),
			     "removed node is a leaf");
  success &= compare_integer(patnNode6, ATNode_GetRight(patnNode1),
			     "predecessor replaces removed node");
  success &= compare_integer(TRUE, ATNode_IsLeaf(patnNode6),
			     "predecessor is now a leaf");

  // Tests CASE B1 */
  treeResult =AVLTree_Remove(pat1, patnNode6);
  success &= compare_integer(SUCCESS, treeResult, "key was found");
  success &= compare_integer(patnNode2, AVLTree_GetRoot(pat1),
			     "double-left-heavy removal rebalances correctly");

  /* CASE R3b: removing an interior right-heavy ancestor */
  treeResult = AVLTree_Remove(pat1, patnNode2);
  success &= compare_integer(SUCCESS, treeResult, "key was found");
  success &= compare_integer(patnNode5, AVLTree_GetRoot(pat1),
			     "correct successor found for right-heavy parent");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode5),
			     "correct left child of new root");
  success &= compare_integer(patnNode1, ATNode_GetRight(patnNode5),
			     "correct right child of new root");

  treeResult = AVLTree_Remove(pat1, patnNode5);
  success &= compare_integer(SUCCESS, treeResult, "key was found");
  success &= compare_integer(patnNode1, AVLTree_GetRoot(pat1),
			     "correct successor found for balanced parent");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode1),
			     "correct left child of new root");

  return success;
}

/* Tests CASES R2b */
BOOL Test_AVLTree_Remove2(void)
{
  BOOL success = TRUE;

  AVLTree_Insert(pat1, patnNode1);
  AVLTree_Insert(pat1, patnNode2);
  AVLTree_Insert(pat1, patnNode3);
  AVLTree_Insert(pat1, patnNode4);
  AVLTree_Insert(pat1, patnNode5);

  // Tests CASE R2b: removing an interior left-heavy distant ancestor */
  AVLTree_Remove(pat1, patnNode1);
  success &= compare_integer(patnNode5, AVLTree_GetRoot(pat1),
			     "correct root after swapping");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode5),
			     "correct left-child after swapping");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode5),
			     "correct right-child after swapping");
  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode5),
			     "new root is left-heavy");
  success &= compare_integer(LEFT_HEAVY, ATNode_GetBalance(patnNode2),
			     "left child of root is left heavy");
  success &= compare_integer(patnNode4, ATNode_GetLeft(patnNode2),
			     "left-left grandchild is correct");
  success &= compare_integer(FALSE, ATNode_HasRight(patnNode2),
			     "left-right grandchild is a hole");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode5),
			     "right child is the same");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode5),
			     "new left child is correct");

  return success;
}

/* Tests CASE R3a */
BOOL Test_AVLTree_Remove3(void)
{
  BOOL success = TRUE;

  AVLTree_Insert(pat1, patnNode1);
  AVLTree_Insert(pat1, patnNode2);
  AVLTree_Insert(pat1, patnNode3);
  AVLTree_Insert(pat1, patnNode6);
  AVLTree_Insert(pat1, patnNode7);

  // Tests CASE R3a: removing a right-heavy direct-parent
  AVLTree_Remove(pat1, patnNode1);
  success &= compare_integer(patnNode6, AVLTree_GetRoot(pat1),
			     "correct root after swapping");
  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode6),
			     "new root is right-heavy");
  success &= compare_integer(RIGHT_HEAVY, ATNode_GetBalance(patnNode3),
			     "right child of root is right heavy");
  success &= compare_integer(patnNode2, ATNode_GetLeft(patnNode6),
			     "left child is the same");
  success &= compare_integer(patnNode3, ATNode_GetRight(patnNode6),
			     "new right child is correct");

  return success;
}

/******************************************************************************
 * AUXILIARY FUNCTION DEFINITIONS
 */

BOOL StructGT(
    PTR pts1,
    PTR pts2
    )
{
  int val1 = StructGetValue(pts1);
  int val2 = StructGetValue(pts2);
  if (val1 > val2) {
    return 1;
  }
  else if (val1 == val2) {
    return 0;
  }
  else {
    return -1;
  }
}

BOOL TestGT(
    PTR pts1,
    PTR pts2
    )
{
  return (StructGetValue(pts1) < StructGetValue(pts2));
}


