/* Tests for list and node interface defined in include/base/list.h
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <util/test.h>
#include <util/tree.h>

/* Global variables for testing */

/* Global variables for testing */

typedef struct _TestStruct {
  int value;
} TestStruct, *pTestStruct;

#define pTS(_pts_) ((pTestStruct)(_pts_))
#define StructGetValue(_pts_) (pTS(_pts_)->value)
#define StructSetValue(_pts_, _value_) (StructGetValue(_pts_) = (_value_))
#define STRUCT_TRANSFORM_VALUE 22

PRIVATE TestStruct ts1;
PRIVATE TestStruct ts2;
PRIVATE TestStruct ts3;
PRIVATE TestStruct ts4;
PRIVATE TestStruct ts5;
PRIVATE TestStruct ts6;
PRIVATE TestStruct ts7;
PRIVATE TestStruct ts8;
PRIVATE pTestStruct pts1 = &ts1;
PRIVATE pTestStruct pts2 = &ts2;
PRIVATE pTestStruct pts3 = &ts3;
PRIVATE pTestStruct pts4 = &ts4;
PRIVATE pTestStruct pts5 = &ts5;
PRIVATE pTestStruct pts6 = &ts6;
PRIVATE pTestStruct pts7 = &ts7;
PRIVATE pTestStruct pts8 = &ts8;

PRIVATE TNODE tnNode1;
PRIVATE TNODE tnNode2;
PRIVATE TNODE tnNode3;
PRIVATE TNODE tnNode4;
PRIVATE TNODE tnNode5;
PRIVATE TNODE tnNode6;
PRIVATE TNODE tnNode7;
PRIVATE TNODE tnNode8;
PRIVATE PTNODE ptnNode1 = &tnNode1;
PRIVATE PTNODE ptnNode2 = &tnNode2;
PRIVATE PTNODE ptnNode3 = &tnNode3;
PRIVATE PTNODE ptnNode4 = &tnNode4;
PRIVATE PTNODE ptnNode5 = &tnNode5;
PRIVATE PTNODE ptnNode6 = &tnNode6;
PRIVATE PTNODE ptnNode7 = &tnNode7;
PRIVATE PTNODE ptnNode8 = &tnNode8;
PRIVATE TREE t1;
PRIVATE PTREE pt1 = & t1;

PRIVATE PTNODE ptnRoot;

/* Function prototypes for tests */

BOOL TNode_TestInitializer(void);
BOOL Test_TNode_Init(void);
BOOL Test_TNode_GetLeft(void);
BOOL Test_TNode_GetRight(void);
BOOL Test_TNode_GetData(void);
BOOL Test_TNode_IsLeaf(void);
BOOL Test_TNode_RotateLeft(void);
BOOL Test_TNode_RotateRight(void);
BOOL Test_TNode_Search(void);
BOOL Test_TNode_ChildSuccessor(void);
BOOL Test_TNode_ChildPredecessor(void);
BOOL Test_TNode_Swap(void);

BOOL Tree_TestInitializer(void);
BOOL Test_Tree_Init(void);
BOOL Test_Tree_IsEmpty(void);
BOOL Test_Tree_GetRoot(void);
BOOL Test_Tree_GetSortFunc(void);
BOOL Test_Tree_Search(void);
BOOL Test_StructGT(void);

/* Prototypes for auxiliary functions */

BOOL StructGT(
    PTR pts1,
    PTR pts2
    );

BOOL TestGT(
    PTR pts1,
    PTR pts2
    );

int _tmain(void) {
  TestFunctionPair TNodeTestFunctions[] = {
    {TNode_TestInitializer,          ""                      },
    {Test_TNode_Init,                "TNode_Init"            },
    {Test_TNode_GetLeft,             "TNode_GetLeft"         },
    {Test_TNode_GetRight,            "TNode_GetRight"        },
    {Test_TNode_GetData,             "TNode_IsAlone"         },
    {Test_TNode_IsLeaf,              "TNode_IsLeaf"          },
    {Test_TNode_RotateLeft,          "TNode_RotateLeft"      },
    {Test_TNode_RotateRight,         "TNode_RotateRight"     },
    {Test_TNode_Search,              "TNode_Search"          },
    {Test_TNode_ChildSuccessor,      "TNode_ChildSuccessor"  },
    {Test_TNode_ChildPredecessor,    "TNode_ChildPredecessor"},
    {Test_TNode_Swap,                "TNode_Swap"            }
  };

  TestFunctionPair TreeTestFunctions[] = {
    {Tree_TestInitializer,           ""                      },
    {Test_Tree_Init,                 "Tree_Init"             },
    {Test_Tree_IsEmpty,              "Tree_IsEmpty"          },
    {Test_Tree_GetRoot,              "Tree_GetRoot"          },
    {Test_Tree_GetSortFunc,          "Tree_GetSortFunc"      },
    {Test_Tree_Search,               "Tree_Search"           }
  };

  RunTestSuite("Binary Tree Node Test Suite", TNodeTestFunctions,
	       GetTestFunctionArraySize(TNodeTestFunctions));
  RunTestSuite("Binary Tree Test Suite", TreeTestFunctions,
	       GetTestFunctionArraySize(TreeTestFunctions));

  return 0;
}

BOOL TNode_TestInitializer(void)
{
  TNode_Init(ptnNode1);
  TNode_SetData(ptnNode1, pts1);
  TNode_Init(ptnNode2);
  TNode_SetData(ptnNode2, pts2);
  TNode_Init(ptnNode3);
  TNode_SetData(ptnNode3, pts3);
  TNode_Init(ptnNode4);
  TNode_SetData(ptnNode4, pts4);
  TNode_Init(ptnNode5);
  TNode_SetData(ptnNode5, pts5);
  TNode_Init(ptnNode6);
  TNode_SetData(ptnNode6, pts6);
  TNode_Init(ptnNode7);
  TNode_SetData(ptnNode7, pts7);

  StructSetValue(pts1, 4);
  StructSetValue(pts2, 2);
  StructSetValue(pts3, 6);
  StructSetValue(pts4, 1);
  StructSetValue(pts5, 3);
  StructSetValue(pts6, 5);
  StructSetValue(pts7, 7);
  StructSetValue(pts8, 8);

  TNode_SetLeft(ptnNode1, ptnNode2);
  TNode_SetRight(ptnNode1, ptnNode3);

  TNode_SetLeft(ptnNode2, ptnNode4);
  TNode_SetRight(ptnNode2, ptnNode5);

  TNode_SetLeft(ptnNode3, ptnNode6);
  TNode_SetRight(ptnNode3, ptnNode7);

  ptnRoot = ptnNode1;

  return TRUE;
}

BOOL Test_TNode_Init(void)
{
  BOOL success = TRUE;

  TNode_Init(ptnNode1);

  success &= compare_integer(NULL, TNode_GetData(ptnNode1),
			     "sets data pointer to NULL");

  success &= compare_integer(NULL, TNode_GetLeft(ptnNode1),
			     "sets left pointer to NULL");

  success &= compare_integer(NULL, TNode_GetRight(ptnNode1),
			     "sets right pointer to NULL");

  return success;
}

BOOL Test_TNode_GetLeft(void)
{
  BOOL success = TRUE;

  TNode_SetLeft(ptnNode1, ptnNode2);

  success &= compare_integer(ptnNode2, TNode_GetLeft(ptnNode1),
			     "set left == get left");

  return success;
}

BOOL Test_TNode_GetRight(void)
{
  BOOL success = TRUE;

  TNode_SetRight(ptnNode3, ptnNode4);

  success &= compare_integer(ptnNode4, TNode_GetRight(ptnNode3),
			     "set right == get right");

  return success;
}

BOOL Test_TNode_GetData(void)
{
  BOOL success = TRUE;

  TNode_SetData(ptnNode2, (PTR)0x2222);

  success &= compare_integer(0x2222, TNode_GetData(ptnNode2),
			     "set data == get data");

  return success;
}

BOOL Test_TNode_IsLeaf(void)
{
  BOOL success = TRUE;

  TNode_Init(ptnNode1);

  success &= compare_integer(TRUE, TNode_IsLeaf(ptnNode1),
			     "initialized node is a leaf");

  TNode_SetLeft(ptnNode1, ptnNode5);

  success &= compare_integer(FALSE, TNode_IsLeaf(ptnNode1),
			     "node with left is not a leaf");

  TNode_SetRight(ptnNode1, ptnNode5);

  success &= compare_integer(FALSE, TNode_IsLeaf(ptnNode1),
			     "node with right is not a leaf");

  return success;
}

BOOL Test_TNode_RotateLeft(void)
{
  BOOL success = TRUE;


  TNode_RotateLeft(&ptnRoot);

  success &= compare_integer(ptnNode3, ptnRoot,
			     "right child is the new top");

  success &= compare_integer(ptnNode1, TNode_GetLeft(ptnNode3),
			     "left child of new top is old top");

  success &= compare_integer(ptnNode7, TNode_GetRight(ptnNode3),
			     "right child of new top is same as before");

  success &= compare_integer(ptnNode6, TNode_GetRight(ptnNode1),
			     "right child of old top is old left child of new top");

  success &= compare_integer(ptnNode2, TNode_GetLeft(ptnNode1),
			     "left child of old top is same as before");

  success &= compare_integer(ptnNode4, TNode_GetLeft(ptnNode2),
			     "left-left grandchild same as before");

  success &= compare_integer(ptnNode5, TNode_GetRight(ptnNode2),
			     "left-right grandchild same as before");

  return success;
}

BOOL Test_TNode_RotateRight(void)
{
  BOOL success = TRUE;

  TNode_RotateRight(&ptnRoot);

  success &= compare_integer(ptnNode2, ptnRoot,
			     "left child is the new top");

  success &= compare_integer(ptnNode1, TNode_GetRight(ptnNode2),
			     "right child of new top is old top");

  success &= compare_integer(ptnNode4, TNode_GetLeft(ptnNode2),
			     "left child of new top is same as before");

  success &= compare_integer(ptnNode5, TNode_GetLeft(ptnNode1),
			     "left child of old top is old right child of new top");

  success &= compare_integer(ptnNode3, TNode_GetRight(ptnNode1),
			     "right child of old top is same as before");

  success &= compare_integer(ptnNode6, TNode_GetLeft(ptnNode3),
			     "right-left grandchild same as before");

  success &= compare_integer(ptnNode7, TNode_GetRight(ptnNode3),
			     "right-right grandchild same as before");

  return success;
}

BOOL Test_TNode_Search(void)
{
  BOOL success = TRUE;

  success &= compare_integer(ptnNode4, TNode_Search(ptnNode1, StructGT, pts4),
			     "node 4 is found");

  success &= compare_integer(NULL, TNode_Search(ptnNode1, StructGT, pts8),
			     "test struct 8 is not found");

  TNode_Init(ptnNode1);
  TNode_SetData(ptnNode1, pts1);

  success &= compare_integer(NULL, TNode_Search(ptnNode1, StructGT, pts4),
			     "node 4 is not found after root is initialized");

  return success;
}

BOOL Test_TNode_ChildSuccessor(void)
{
  BOOL success = TRUE;
  PTNODE ptnTemp;

  ptnTemp = TNode_ChildSuccessor(ptnNode1);

  success &= compare_integer(ptnNode6, ptnTemp,
			     "correct distant successor found");

  ptnTemp = TNode_ChildSuccessor(ptnNode3);

  success &= compare_integer(ptnNode7, ptnTemp,
			     "correct near successor found");

  TNode_SetRight(ptnNode7, ptnNode8);

  ptnTemp = TNode_ChildSuccessor(ptnNode7);

  success &= compare_integer(ptnNode8, ptnTemp,
			     "correct sole successor found");

  return success;
}


BOOL Test_TNode_ChildPredecessor(void)
{
  BOOL success = TRUE;
  PTNODE ptnTemp;

  ptnTemp = TNode_ChildPredecessor(ptnNode1);

  success &= compare_integer(ptnNode5, ptnTemp,
			     "correct distant predecessor found");

  ptnTemp = TNode_ChildPredecessor(ptnNode2);

  success &= compare_integer(ptnNode4, ptnTemp,
			     "correct near predecessor found");

  TNode_SetLeft(ptnNode6, ptnNode8);
  ptnTemp = TNode_ChildPredecessor(ptnNode6);

  success &= compare_integer(ptnNode8, ptnTemp,
			     "correct sole predecessor found");

  return success;
}

BOOL Test_TNode_Swap(void)
{
  BOOL success = TRUE;
  
  TNode_SetRight(ptnNode5, ptnNode8);

  printf("  Case S5\n");
  /* CASE S5: swap nodes who are not direct parent-child (patn1 and patn5) */
  TNode_Swap(&ptnRoot, TNode_GetRightPointer(ptnNode2));
  success &= compare_integer(ptnNode5, ptnRoot,
			     "ptn1 is now ptn2");
  success &= compare_integer(ptnNode1, TNode_GetRight(ptnNode2),
			     "ptn2 is now ptn1");
  success &= compare_integer(ptnNode2, TNode_GetLeft(ptnNode5),
			     "ptn2's left is ptn1's old left");
  success &= compare_integer(ptnNode3, TNode_GetRight(ptnNode5),
			     "ptn2's right is ptn1's old right");
  success &= compare_integer(NULL, TNode_GetLeft(ptnNode1),
			     "ptn1's left is ptn2's old left");
  success &= compare_integer(ptnNode8, TNode_GetRight(ptnNode1),
			     "ptn1's right is ptn2's old right");

  printf("  Case S1\n");
  /* CASE S1: swap node and its direct-left-child (ptnNode3 and ptnNode6) */
  TNode_Swap(TNode_GetRightPointer(ptnNode5), TNode_GetLeftPointer(ptnNode3));
  success &= compare_integer(ptnNode6, TNode_GetRight(ptnNode5),
			     "ptn1 is now ptn2");
  success &= compare_integer(ptnNode3, TNode_GetLeft(ptnNode6),
			     "ptn2 is now ptn1");
  success &= compare_integer(ptnNode3, TNode_GetLeft(ptnNode6),
			     "ptn2's left is ptn1's old left");
  success &= compare_integer(ptnNode7, TNode_GetRight(ptnNode6),
			     "ptn2's right is ptn1's old right");
  success &= compare_integer(NULL, TNode_GetLeft(ptnNode3),
			     "ptn1's left is ptn2's old left");
  success &= compare_integer(NULL, TNode_GetRight(ptnNode3),
			     "ptn1's right is ptn2's old right");

  printf("  Case S2\n");
  /* CASE S2: swap node and its direct-right-child (ptnNode2 and ptnNode1) */
  TNode_Swap(TNode_GetLeftPointer(ptnNode5), TNode_GetRightPointer(ptnNode2));
  success &= compare_integer(ptnNode1, TNode_GetLeft(ptnNode5),
			     "ptn1 is now ptn2");
  success &= compare_integer(ptnNode2, TNode_GetRight(ptnNode1),
			     "ptn2 is now ptn1");
  success &= compare_integer(ptnNode4, TNode_GetLeft(ptnNode1),
			     "ptn2's left is ptn1's old left");
  success &= compare_integer(ptnNode2, TNode_GetRight(ptnNode1),
			     "ptn2's right is ptn1's old right");
  success &= compare_integer(NULL, TNode_GetLeft(ptnNode2),
			     "ptn1's left is ptn2's old left");
  success &= compare_integer(ptnNode8, TNode_GetRight(ptnNode2),
			     "ptn1's right is ptn2's old right");

  printf("  Case S3\n");
  /* CASE S3: swap node and its direct-left-parent (ptnNode4 and ptnNode1) */
  TNode_Swap(TNode_GetLeftPointer(ptnNode1), TNode_GetLeftPointer(ptnNode5));
  success &= compare_integer(ptnNode4, TNode_GetLeft(ptnNode5),
			     "ptn1 is now ptn2");
  success &= compare_integer(ptnNode1, TNode_GetLeft(ptnNode4),
			     "ptn2 is now ptn1");
  success &= compare_integer(NULL, TNode_GetLeft(ptnNode1),
			     "ptn2's left is ptn1's old left");
  success &= compare_integer(NULL, TNode_GetRight(ptnNode1),
			     "ptn2's right is ptn1's old right");
  success &= compare_integer(ptnNode1, TNode_GetLeft(ptnNode4),
			     "ptn1's left is ptn2's old left");
  success &= compare_integer(ptnNode2, TNode_GetRight(ptnNode4),
			     "ptn1's right is ptn2's old right");

  printf("  Case S4\n");
  /* CASE S4: swap node and its direct-right-parent (ptnNode7 and ptnNode6) */
  TNode_Swap(TNode_GetRightPointer(ptnNode6), TNode_GetRightPointer(ptnNode5));
  success &= compare_integer(ptnNode7, TNode_GetRight(ptnNode5),
			     "ptn1 is now ptn2");
  success &= compare_integer(ptnNode6, TNode_GetRight(ptnNode7),
			     "ptn2 is now ptn1");
  success &= compare_integer(NULL, TNode_GetLeft(ptnNode6),
			     "ptn2's left is ptn1's old left");
  success &= compare_integer(NULL, TNode_GetRight(ptnNode6),
			     "ptn2's right is ptn1's old right");
  success &= compare_integer(ptnNode3, TNode_GetLeft(ptnNode7),
			     "ptn1's left is ptn2's old left");
  success &= compare_integer(ptnNode6, TNode_GetRight(ptnNode7),
			     "ptn1's right is ptn2's old right");

  return success;
}

BOOL Tree_TestInitializer(void)
{
  Tree_Init(pt1, StructGT);
  Tree_SetRoot(pt1, ptnNode1);
  TNode_TestInitializer();

  return TRUE;
}

BOOL Test_Tree_Init(void)
{
  BOOL success = TRUE;

  Tree_Init(pt1, NULL);

  success &= compare_integer(NULL, Tree_GetRoot(pt1),
			     "sets root to NULL");

  success &= compare_integer(NULL, Tree_GetSortFunc(pt1),
			     "sets SortFunc to NULL");

  return success;
}

BOOL Test_Tree_IsEmpty(void)
{
  BOOL success = TRUE;

  success &= compare_integer(FALSE, Tree_IsEmpty(pt1),
			     "tree with nodes is not empty");


  Tree_Init(pt1, NULL);

  success &= compare_integer(TRUE, Tree_IsEmpty(pt1),
			     "initialized tree is empty"); 

  return success;
}

BOOL Test_Tree_GetRoot(void)
{
  BOOL success = TRUE;

  Tree_SetRoot(pt1, ptnNode2);

  success &= compare_integer(ptnNode2, Tree_GetRoot(pt1),
			     "get root == set root");


  return success;
}

BOOL Test_Tree_GetSortFunc(void)
{
  BOOL success = TRUE;

  Tree_SetSortFunc(pt1, TestGT);

  success &= compare_integer(TestGT, Tree_GetSortFunc(pt1),
			     "get sortfunc == set sortfunc");


  return success;
}

BOOL Test_Tree_Search(void)
{
  BOOL success = TRUE;

  Tree_SetRoot(pt1, ptnNode1);

  success &= compare_integer(ptnNode5, Tree_Search(pt1, pts5),
			     "node 5 is found");

  success &= compare_integer(NULL, Tree_Search(pt1, pts8),
			     "node 8 not found");

  Tree_Init(pt1, StructGT);

  success &= compare_integer(NULL, Tree_Search(pt1, pts5),
			     "node 5 is not found after tree is initialized");

  return success;
}


BOOL StructGT(
    PTR pts1,
    PTR pts2
    )
{
  int val1 = StructGetValue(pts1);
  int val2 = StructGetValue(pts2);
  if (val1 > val2) {
    return 1;
  }
  else if (val1 == val2) {
    return 0;
  }
  else {
    return -1;
  }
}

BOOL TestGT(
    PTR pts1,
    PTR pts2
    )
{
  return (StructGetValue(pts1) < StructGetValue(pts2));
}


