/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * C test case -- memcmp, memcpy, memmove, and memset functions.
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <assert.h>


char c1[64];
char c2[64] =
    "abcdefghijklmnopqrstuvwxyzyxwvutsrqponmlkjihgfedcba";
char c3[64];



int _tmain()
{
    int i, j, k, passes = TRUE;
    void *p;

    /* test memset function */
    for (i = 0; i < 26; ++i) {
        p = memset(c1+i, 'a'+i, 51-2*i);
        assert(p == c1+i);
    }

    if (memcmp(c1, c2, 51)) {
        _tprintf(_TEXT("\n\r  ===> Error in memset test!!!\n\r\n"));
        passes = FALSE;
    }

    /* test memcpy and memcmp functions */
    for (i = 0; i < 4; ++i) {
        p = memset(c3, 0, 64);
        assert(p == c3);
        p = memcpy(c3+i, c2, 51);
        assert(p == c3+i);
        for (j = 0; j < 4; ++j) {
            for (k = 1; k <= 26; ++k) {
                p = memset(c1, ~0, 64);
                assert(p == c1);
                p = memcpy(c1+j, c3+i, k);
                assert(p == c1+j);
                if (memcmp(c1+j, c3+i, k) || !memcmp(c1+j, c3+i, k+1))
                {
                    _tprintf(_TEXT("\n\r  ===> Error in memcpy/memcmp "));
                    _tprintf(_TEXT("test!!!\n\r\n"));
                    passes = FALSE;
                }
            }
        }
    }

    /* test memmove function */
    for (i = 0; i < 4; ++i) {
        for (j = 1; j <= 4; ++j) {
            for (k = 1; k <= 26; ++k) {
                p = memset(c1, ~0, 64);
                assert(p == c1);
                p = memcpy(c1+i, c2, 26);
                assert(p == c1+i);
                p = memmove(c1+i+j, c1+i, k);
                assert(p == c1+i+j);
                if (memcmp(c1+i+j, c2, k) || !memcmp(c1+i+j, c2, k+1))
                {
                    _tprintf(_TEXT("\n\r  ===> Error in memmove test!!!\n\r\n"));
                    passes = FALSE;
                }
            }
        }
    }

    if (passes) {
        _tprintf(_TEXT("Test of memfuncs PASSED.\n\r"));
    } else {
        _tprintf(_TEXT("Test of memfuncs FAILED.\n\r"));
    }

    exit (passes ? EXIT_SUCCESS : EXIT_FAILURE);
}
