/* Tests for src/base/workitem.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define SIM 1

#include <stdio.h>
#include <mmlite.h>
#include <base/condition2.h>
#include <base/workitem.h>
#include <util/test.h>
#include <util/util.h>

PTR THREAD_LINKAGE TestWorkFunc(WORK_ARGUMENT arg);
UINT BarrierCount = 0;

#define WORKITEM_COUNT 22
TIME Deadlines[WORKITEM_COUNT];

int _tmain(void) {

  SCODE sc;
  PISCHEDULER pTheScheduler = CurrentScheduler();
  PIACTIVITY pActCurrent = CurrentActivity();
  PIACTIVITY pNewWorkItem[WORKITEM_COUNT];
  TIME_CONSTRAINT tc;
  int i;

  tc.Start = CurrentTime();
  tc.Estimate = TIME_MILLIS(1);
  tc.Deadline = Int64Add(CurrentTime(), TIME_MILLIS(5));
  sc = pTheScheduler->v->SetConstraint(pTheScheduler, &tc, NULL);
  if (sc != S_OK) {
    printf("Failed to schedule main thread\n");
  }

  for (i = 0; i < WORKITEM_COUNT; i++) {
    sc = pActCurrent->v->CreateWorkItem(pActCurrent, CurrentProcess(),
                                        TestWorkFunc, (PTR)i,
                                        NULL, &pNewWorkItem[i]);
    if (sc == NO_ERROR) {
      printf("Created work item %d 0x%x\n", i, pNewWorkItem[i]);
    }
    else {
      printf("Failed to create work item\n");
    }
    pNewWorkItem[i]->v->Release(pNewWorkItem[i]);

    tc.Deadline = Int64Add(CurrentTime(), TIME_MILLIS(22));
    Deadlines[i] = tc.Deadline;
    sc = pTheScheduler->v->AddActivity(pTheScheduler, pNewWorkItem[i], &tc);
    if (sc == S_OK) {
      printf("Started %d with deadline %ld\n", i, tc.Deadline);
    }
    else {
      printf("Failed to start work item %d\n", i);
    }
    
  }

  while (BarrierCount < WORKITEM_COUNT) {
    MilliSleep(1);
    printf("Waiting for %d\n", WORKITEM_COUNT - BarrierCount);
  }
  
  return 0;
}
    
PRIVATE PTR THREAD_LINKAGE
TestWorkFunc(
      WORK_ARGUMENT arg
      )
{
  UINT num = (UINT)arg;
  int i = 0;
  int j = 0;
  printf("Hello Work Item! Received %d %ld\n",
         num, CurrentTime() - Deadlines[num]);
  for (i = 0; i < 22; i++) {
    j = j * i;
  }
  AtomicInc(&BarrierCount);
  return (PTR)22;
}

