/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*********************************************************************/
/* setvol.c                                                          */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <mmlite.h>

BOOL SetVolume(_TCHAR *AudioName,
               UINT nVolume)
{
    UINT64 Aurgh;
    PIFILE AudioOut;
    SCODE sc;

    sc = BindToObject(CurrentNameSpace(),
                      AudioName,
                      NAME_SPACE_WRITE,
                      &IID_IFile,
                      (void *)&AudioOut);
    if (FAILED(sc)) {
        _tprintf(_T("Output codec %s not found [sc=%x]\r\n"),AudioName,sc);
        return FALSE;
    }

    /* funnything: volume=low8 cmd=56..63 */
    Int64FromHighAndLow(Aurgh,
                        (1<<24),nVolume);

    sc = AudioOut->v->SetSize(AudioOut,Aurgh);
    AudioOut->v->Release(AudioOut);

    if (FAILED(sc)) {
        _tprintf(_T("SetSize[%s] failed with sc=%x\r\n"),AudioName,sc);
        return FALSE;
    }

    return TRUE;
}

/*** main entry point ***/
int _tmain(INT argc,_TCHAR **argv)
{
    UINT                 nVolume;

    argc = 0;
    (void) ParseCmdLine( &argc, &argv);
    if (argc < 3) {
        _tprintf(_T("Usage:%s <codec name> <volume>\n"), argv[0]);
        goto cleanup;
    }

    /* say hello
     */
    nVolume = _tcstoul(argv[2],NULL,10);
    _tprintf(_T("[%s - Set volume of %s to %d]\n"), argv[0], argv[1], nVolume);

    /* do it
     */
    if (!SetVolume(argv[1],nVolume)) {
        _tprintf(_T("not "));
    }
    _tprintf(_T("done.\r\n"));

 cleanup:
    (void) FreeCmdLine( &argc, &argv);
    return 0;
}

