/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>
#include <mmhal.h>
#include <drivers/drivers.h>
#include <tchar.h>
#include <stdlib.h>
#include <string.h>

#define STATISTICS 0

/* wish they all could be..
 */
typedef struct {
    CONDITION WaitPoint;
    MUTEX Lock;
    PUINT16 pBuffer;
    volatile INT Head, Tail;
    UINT Size;
    /* only if statistics are enabled */
    UINT nIn, nOut, nSignalled, nLoopy, nSleepy, nCally;
} BYTE_QUEUE;
typedef struct CUcbDevice {

    struct CDevice CDevice;

    /* Our own state
     */
    IO_PORT PortBase;
    IO_PORT ControlBase;
    UINT16 ChipVersion;
    UINT16 AudioSampleRate, TelcoSampleRate;
    UINT16 AudioGain, AudioAttenuation;
    UINT Errors;
    BOOL Reset, AudioIsActive;

    BYTE_QUEUE AudioOut;

} CUcbDevice, *PCUcbDevice;


void Dump(PIFILE pThis)
{
    PCUcbDevice this = (PCUcbDevice)pThis;

    _tprintf(_T("Dump of UCB device object at x%x\n"), pThis);

    _tprintf(_T("v= x%x\t"), this->CDevice.v );
    _tprintf(_T("RefCnt= x%x\t"), this->CDevice.RefCnt );
    _tprintf(_T("IntrPoint= x%x %x\t"), this->CDevice.IntrPoint );
    _tprintf(_T("SpareForMBufFact= x%x\r\n"), this->CDevice.SpareForMBufFact );
    _tprintf(_T("PortBase= x%x\t"), this->PortBase );
    _tprintf(_T("ControlBase= x%x\t"), this->ControlBase );
    _tprintf(_T("ChipVersion= x%x\r\n"), this->ChipVersion );
    _tprintf(_T("AudioSampleRate= %d\t"), this->AudioSampleRate );
    _tprintf(_T("TelcoSampleRate= %d\t"), this->TelcoSampleRate );
    _tprintf(_T("AudioGain= x%x\t"), this->AudioGain );
    _tprintf(_T("Attenuation= x%x\r\n"), this->AudioAttenuation );
    _tprintf(_T("Errors= %d\t"), this->Errors );
    _tprintf(_T("Reset= x%x\t"), this->Reset );
    _tprintf(_T("AudioIsActive= x%x\r\n"), this->AudioIsActive );
    _tprintf(_T("WaitPoint= x%x %x\t"), this->AudioOut.WaitPoint );
    _tprintf(_T("Lock= x%x\t"), this->AudioOut.Lock );
    _tprintf(_T("pBuffer= x%x\r\n"), this->AudioOut.pBuffer );
    _tprintf(_T("Head= %d\t"), this->AudioOut.Head );
    _tprintf(_T("Tail= %d\t"), this->AudioOut.Tail );
    _tprintf(_T("Size= %d\r\n"), this->AudioOut.Size );
#if STATISTICS
    _tprintf(_T("In= %d\t"), this->AudioOut.nIn );
    _tprintf(_T("IsrOut= %d\t"), this->AudioOut.nOut );
    _tprintf(_T("Loops= %d\t"), this->AudioOut.nLoopy );
    _tprintf(_T("Calls= %d\r\n"), this->AudioOut.nCally );
    _tprintf(_T("Sleeps= %d\t"), this->AudioOut.nSleepy );
    _tprintf(_T("Signals= %d\r\n"), this->AudioOut.nSignalled );
#endif
}

/* Test
 */
#define nSam 256
UINT16 Samples[nSam];

#define DUMPIT(_x_)

void Test(PIFILE AudioOut, int Logorrea)
{
    TIME Now, End;
    int i, j, k;
    SCODE sc;

    /* In those 256 samples we can fit 8 periods of a saw (~~sin)
     * Each zero is one period.  First zero is at 32, the length in samples.
     * At a 11KHz sampling rate (90us period) this is a 347.2 Hz wave.
     * Its audible (and a bit annoying too)
     */
    j = 128;
    i = 0;
    DUMPIT(printf("Zeroes at: ");)
    for (k = 0; k < nSam; k++) {
        DUMPIT(if (i == 0) printf("%d ", k);)
        Samples[k] = i << 4;

        i += j;
        if ((i >= ((1<<11)-1)) || (i <= 0)) {
            j = -j;
        }
    }
    DUMPIT(
           printf("\nSamples: %d %d %d %d ...%d %d\n",
                  Samples[0], Samples[1], Samples[2], Samples[3], 
                  Samples[nSam-2],Samples[nSam-1]);)
    for (k = 0; k < nSam; k++) {
        DUMPIT(printf("%d ", Samples[k]);)
    }
    DUMPIT(printf("\n");)

    Int32ToInt64(End,TIME_SECONDS(5));
    Now = CurrentTime();
    End = TimeAdd(End,Now);

    for (;;) {

        k = 0;
        do {
            i = 0;
            sc = AudioOut->v->WriteAt(AudioOut,
                                      Int64ToUint64(TIME_ORIGIN)/*urgh*/,
                                      ((const BYTE*)Samples)+k,
                                      nSam*2 - k,
                                      (UINT*)&i);
            if (Logorrea) {
                if (Logorrea == 1)
                    printf("Sam%d %d ",k,i);
                else {
                    if (i != nSam*2)
                        printf("Sam%d %d ",k,i);
                }
            }
            k += i;
        } while (k < nSam*2);

        /* Are we done yet ?
         */
        Now = CurrentTime();
        if (TimeLess(End,Now))
            break;
    }
    
}

/* Driving code. 
 */
int _tmain(INT argc, _TCHAR **argv)
{
    PIFILE AudioOut;
    SCODE sc;
    BOOL DoDump;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc) || (argc < 2)) {
        _tprintf(_T("Usage:sinout AudioOutputName\r\n"));
        return 1;
    }

    DoDump = FALSE;
    if (argv[1][0] == _T('-')) {
        /* debug help */
        DoDump = TRUE;
        argv[1]++;
    }

    sc = BindToObject(CurrentNameSpace(),
                      argv[1],
                      NAME_SPACE_WRITE,
                      &IID_IFile,
                      (void *)&AudioOut);

    if (!FAILED(sc)) {
        if (DoDump)
            Dump(AudioOut);
        else
            Test(AudioOut,argc-2);
        AudioOut->v->Release(AudioOut);
    } else {
        _tprintf(_T("Could not find %s [%x]\r\n"),argv[1],sc);
    }

    FreeCmdLine(&argc,&argv);
    return 0;
}
