/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Compute and optionally verify a file's checksum. */
#include <stdio.h>
#include <mmlite.h>
#include <tchar.h>
#include <stdlib.h>
#include <string.h>

UINT8 buf[BUFSIZ];
_TCHAR *Usage = _TEXT("%s [-v decimal-checksum-value] [-m HexAddr DecSize] [filename]\r\n");

SCODE
FileChecksum(
             _TCHAR *filename,
             UINT32 *computed
)
{
    FILE *f;
    size_t n, i;

    /* fopen */
    f = _tfopen(filename, _T("rb"));
    if (f == NULL) {
        _tprintf(_TEXT("IFile %s not found.\n"), filename);
        return E_FILE_NOT_FOUND;
    }

    *computed = 0;
    for (;;) {
        n = fread(buf, 1, sizeof buf, f);
        if (n == 0) break;
        for (i = 0; i < n; i++)
            *computed += buf[i];
    }

    (void) fclose(f);

    _tprintf(_TEXT("Checksum for file %s is x%x (%d decimal).\n\r"),
             filename, *computed, *computed);

    return S_OK;
}

/* for FLASH, usually
 */
SCODE
MemoryChecksum(
               ADDRESS Start,
               UINT nBytes,
               UINT32 *computed
)
{
    UINT8 *pMem = (UINT8 *) Start;
    UINT i;

    *computed = 0;
    for (i = 0; i < nBytes; i++)
        *computed += pMem[i];

    _tprintf(_TEXT("Checksum for memory at [x%x..x%x] is x%x (%d decimal).\n\r"),
             Start, Start+nBytes, *computed, *computed);

    return S_OK;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    UINT32 expected = 0, computed = 0;
    BOOL verify = FALSE;
    ADDRESS StartAddress = ~0;
    UINT nBytes = 0;
    _TCHAR *me;

    sc = ParseCmdLine(&argc, &argv);
    me = argv[0];
    if (argc < 2)
        goto Usage;

    /* In memory ?
     */
    while (argv[1][0] == '-') {
        if (argv[1][1] == 'm') {
            StartAddress = _tcstoul(argv[2],NULL,16);
            nBytes       = _tcstoul(argv[3],NULL,10);
            argv += 3;
            argc -= 3;
        }

        if (argv[1][1] == 'v') {
            verify = TRUE;
            expected = _ttoi(argv[2]);
            argv += 2;
            argc -= 2;
        }
    }

    sc = E_FAIL;
    if (StartAddress != ~0) {
        sc = MemoryChecksum(StartAddress,nBytes,&computed);
        if (FAILED(sc)) goto Done;
    }

    if (argc > 1) {
        sc = FileChecksum(argv[1],&computed);
        if (FAILED(sc)) goto Done;
    }

    if (FAILED(sc)) goto Usage;

    if (verify)
        _tprintf(_TEXT("Computed checksum %d %s match with expected %d.\n\r"),
                 computed, 
                 (computed == expected) ? _TEXT("is a") : _TEXT("does not"),
                 expected);
 Done:
    FreeCmdLine(&argc,&argv);
    return 0;
 Usage:
    _tprintf(Usage,me);
    goto Done;
}
