/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>

#define MAXNOTHREADS    100
IThread *Thread[MAXNOTHREADS];

volatile BOOL KeepGoing = TRUE;
UINT ThreadsLeft;

CONDITION CondDone;

void ThreadFunc(THREAD_ARGUMENT MyId)
{
    UINT ThreadsLeftAfterMe;
    _tprintf(_TEXT("Thread %d starting...\n"),(UINT) MyId);

    while (KeepGoing)
    {
         ;
    }

    ThreadsLeftAfterMe=AtomicDec(&ThreadsLeft);
    _tprintf(_TEXT("%d: Exiting, ThreadsLeft=%d\n"),
             (UINT)MyId,ThreadsLeftAfterMe);
    if (ThreadsLeftAfterMe==0)
        Condition_Signal(&CondDone);
    return;
}

int _tmain(INT argc, _TCHAR *argv[])
{
    UINT NThreads = 50;
    INT TestTime  = 30;
    UINT i;
    SCODE sc;
    PIPROCESS pPrc;
    TIME t;

    sc = ParseCmdLine(&argc,&argv);

    switch (argc) 
    {
        case 3:
            TestTime = _ttoi(argv[2]);
        case 2:
            NThreads = _ttoi(argv[1]);
        case 1:
        case 0:
            break;
        default:
            _tprintf(_TEXT("usage: tctst [NThreads [TestTime]]\n"));
    }

    (void) FreeCmdLine(&argc,&argv);

    if (NThreads > MAXNOTHREADS) 
    {

            _tprintf(_TEXT("Will not create more than %d threads\n"), MAXNOTHREADS);
            exit(EXIT_FAILURE);
    }

    _tprintf(_TEXT("Test %d threads for %d seconds\n"), NThreads, TestTime);

    Condition_Init(&CondDone);

    /* Create all those threads */

    pPrc = CurrentProcess();

    ThreadsLeft=NThreads;
    for (i = 0; i < NThreads; i++) 
    {
        sc = pPrc->v->CreateThread(pPrc, 0, ThreadFunc, (THREAD_ARGUMENT)i,
                                         0, NULL, &Thread[i]);
        if (FAILED(sc)) 
        {
            _tprintf(_TEXT("tctst: error 0x%x creating thread #%d\n"), sc, i);
            exit(EXIT_FAILURE);
        }
    }

    _tprintf(_TEXT("All %d threads created; sleeping for %d seconds\n"), NThreads, TestTime);

    Int32ToInt64(t,TIME_RELATIVE(TIME_SECONDS(TestTime)));
    SleepUntil(t);

    _tprintf(_TEXT("Waiting for threads to exit\n"));
    KeepGoing = FALSE;

    while (ThreadsLeft>0)
    {
        Condition_Wait(&CondDone, NULL, NULL);
    }

    Condition_Destroy(&CondDone);

    _tprintf(_TEXT("Releasing threads\n"));

    for (i = 0; i < NThreads; i++) 
    {
        Thread[i]->v->Release(Thread[i]);
    }
        
    _tprintf(_TEXT("Released all %d threads\n"), NThreads);

    exit(EXIT_SUCCESS);
    return 0;
}
