/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg,
            ADDR_SIZE StackSize,
            /* [out] */ IThread **ppNewThread)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    sc = pPrc->v->CreateThread(pPrc,
                               0,
                               pStart,
                               Arg,
                               StackSize,
                               NULL,
                               ppNewThread);
    return sc;
}


/* Which test(s) should we perform
 */
#define TEST1 1
#define TEST_COND 1
#define TIME_OVERHEADS 1
#define TEST_SLEEP 1
#define TEST_CONSTR 1

/* Print names for tests
 */
#if defined(TEST1)
#define P_TEST1 _T("  TimeShare #1\n")
#else
#define P_TEST1 _T("")
#endif
#if defined(TEST2)
#define P_TEST2 _T("  TimeShare #2\n")
#else
#define P_TEST2 _T("")
#endif
#if defined(TEST3)
#define P_TEST3 _T("  TimeShare #3\n")
#else
#define P_TEST3 _T("")
#endif
#if defined(TEST4)
#define P_TEST4 _T("  SingleThreaded #4\n")
#else
#define P_TEST4 _T("")
#endif
#if defined(TEST_MUTEX)
#define P_TEST_MUTEX _T("  Mutex objects\n")
#else
#define P_TEST_MUTEX _T("")
#endif
#if defined(TEST_COND)
#define P_TEST_COND _T("  Condition variables\n")
#else
#define P_TEST_COND _T("")
#endif
#if defined(TEST_ATOMIC)
#define P_TEST_ATOMIC _T("  AtomicXXX functions\n")
#else
#define P_TEST_ATOMIC _T("")
#endif
#if defined(TIME_OVERHEADS)
#define P_TIME_OVERHEADS _T("  Basic overheads timings\n")
#else
#define P_TIME_OVERHEADS _T("")
#endif
#if defined(TEST_SLEEP)
#define P_TEST_SLEEP _T("  SleepUntil\n")
#else
#define P_TEST_SLEEP _T("")
#endif
#if defined(TEST_CONSTR)
#define P_TEST_CONSTR _T("  Time constraints\n")
#else
#define P_TEST_CONSTR _T("")
#endif


CONDITION fooc, bazc;
MUTEX foom;

int nn = 0;
UINT x1 = 0;

void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me, i;

    UnusedParameter(arg);

    me = nn++;

    if (me == 0)
      _tprintf(_T("Will perform the following tests:\n") \
               P_TIME_OVERHEADS P_TEST_ATOMIC P_TEST_SLEEP \
               P_TEST_COND P_TEST_CONSTR P_TEST_MUTEX \
               P_TEST1 P_TEST2 P_TEST3 P_TEST4
               );

    if (me == 0) {

#if defined(TIME_OVERHEADS)
        TIME t0, t1, tn, t;

        Int32ToInt64(tn,10000000);
        for (i = 0; i < 10; i++) {
            t0 = CurrentTime();
            t1 = CurrentTime();
            t1 = TimeSubtract(t1,t0);
            if (TimeLess(t1,tn)) {
                tn = t1;
            }
            /*_tprintf(_T("CurrentTime() costs %ld (%d-th loop)\n"), t1, i);*/
        }
        _tprintf(_T("Minimum CurrentTime() cost is %ld\n"), tn);

        Int32ToInt64(t,TIME_RELATIVE(10000)); /* 1milli */
        Int32ToInt64(tn,100000); /* 10milli */
        for (i = 0; i < 10; i++) {
            t0 = CurrentTime();
            SleepUntil(t);
            t1 = CurrentTime();
            t1 = TimeSubtract(t1,t0);
            t1 = TimeAdd(t1,t);
            if (TimeLess(t1,tn)) {
                /*_tprintf(_T("SL %ld\n"), t1);*/
                tn = t1;
            }
        }
        _tprintf(_T("Minimum SleepUntil() cost is %ld\n"), tn);
#endif

#if defined(TEST_COND)
        Condition_Init(&fooc);
        Condition_Init(&bazc);
#endif
#if defined(TEST_MUTEX)
        Mutex_Init(&foom);
#endif
#if !defined(TEST4)
        _tprintf(_T("Initial thread started.\n"));
        (void) CreateAThread(FirstThread, NULL, 1024, NULL);
        (void) CreateAThread(FirstThread, NULL, 1024, NULL);
#endif
    }

#if defined(TEST2)
    Delay(3000);
#endif

    _tprintf(_T("Thread %d is x%x\n"), me, CurrentThread());

#if defined(TEST_ATOMIC)
    for (i = 0; i < 1000; i++) {
      AtomicInc(&x1);
      AtomicAdd(&x1,20);
      AtomicMaybeDec(&x1);
      AtomicMaybeInc(&x1);
      AtomicMaybeSub(&x1,10);
      AtomicDec(&x1);
      AtomicSub(&x1,10);
    }
    _tprintf(_T("%d found x1 = %d\n"), me, x1);
#endif

#if defined(TEST_SLEEP)
    /* This tests the IdleThread logic too, since all threads sleep
     */
    {
        TIME t, te;
        for (i = 0; i < 10; i++) {
            Int32ToInt64(t,TIME_RELATIVE(10000)); /* 1milli */
            te = CurrentTime();
            SleepUntil(t);
            t = CurrentTime();
            _tprintf(_T("%d slept from %d.%x to %d.%x.\n"),
                     me, te, t);
        }
    }
#endif

#if defined(TEST_COND)
    if (me == 0) {
        TIME t, t1;
        for (i = 0; i < 3; i++) {
            Int32ToInt64(t,TIME_RELATIVE(20000)); /* 2milli */
            _tprintf(_T("%d Sleeping for %ld\n"), me, t);
            t1 = CurrentTime();
            SleepUntil(t);
            t = CurrentTime();
            _tprintf(_T("%d slept from %ld to %ld\n"), me, t1, t);
            if (i == 2) {
                Condition_Broadcast(&fooc);
                _tprintf(_T("%d signalled all other threads.\n"), me);
            } else {
                Condition_Signal(&fooc);
                _tprintf(_T("%d signalled 1 other thread.\n"), me);
            }
        }
    } else {
        for (i = 0; i < 2; i++) {
            _tprintf(_T("%d into Wait\n"), me);
            Condition_Wait(&fooc,NULL,NULL);
            _tprintf(_T("%d back from Wait\n"), me);
        }
    }

    Delay(10000);

    {
      TIME t, t0;
      Int32ToInt64(t,TIME_RELATIVE(1000*10));
      _tprintf(_T("%d into Wait for %ld\n"), me, t);
      t0 = CurrentTime();
      Condition_Wait(&bazc, NULL, &t);
      _tprintf(_T("%d was in Wait from %ld to %ld\n"),
               me, t0, CurrentTime());
    }

#endif

#if defined(TEST_CONSTR)
    {
        TIME_CONSTRAINT MyCos;
        TIME t0, t1;
        SCODE sc0, sc1;
        MyCos.Criticality = /*_NON*/_CRITICAL;
        /* start time of now plus "me" millisecs */
        Int32ToInt64(MyCos.Start,TIME_RELATIVE(me*1000*10));
        /* run time of 1msec */
        Int32ToInt64(MyCos.Estimate,1000*10);
        /* deadline me+1 millisecs */
        Int32ToInt64(MyCos.Deadline,TIME_RELATIVE((me+1)*1000*10));

        sc0 = BeginConstraint( FALSE, &MyCos, NULL);
        t0 = CurrentTime();
        _tprintf(_T("%d is inside a constraint at %ld\n"), me, t0);
        sc1 = EndConstraint(&t1);

        _tprintf(_T("%d time results are %x %x %ld\n"),
                 me, sc0, sc1, t1);
        if (sc0 != S_OK || sc1 != S_OK)
            _tprintf(_T("    CONSTRAINT TEST FAILED !!\n"));
        Delay(3000);
    }
#endif

#if defined(TEST1)
#define NLOOPS 12
#elif defined(TEST4)
#define NLOOPS 120*3
#else
#define NLOOPS 120
#endif
    for (i = 0;i < NLOOPS; i++) {
#if !defined(TEST3)
        TIME t;
#if defined(TEST_MUTEX)
        Mutex_Lock(&foom);
#endif
        t = CurrentTime();
        _tprintf(_T("%d at %ld\n"), me, t);
#if defined(TEST_MUTEX)
        Mutex_Unlock(&foom);
#endif
#if defined(TEST1)
        Delay(3000/*000*/);
#endif
#if defined(TEST2) || defined(TEST4)
        Delay(3);
#endif
#else /*!TEST3*/
      {
          volatile int *p = (volatile int *)0x20000;
          volatile int *e = p + (64*1024 / 4);

          while (p < e) {
              i = *p;
              p += 8;
          }
      }
#endif
    }

    /* terminate normally */
}

int _tmain(int argc, _TCHAR **argv)
{
    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return 0;
}
