/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Time a test.
 * Runs the command line, taking time at start and end.
 * [stupid name "timer.exe" is for historical reasons]
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#define TAKE_LOAD_TIME 1
extern PIBASERTL _pTheBaseRtl;
typedef INT (*C_ENTRY)(PIBASERTL);

int _tmain(int argc, _TCHAR **argv)
{
    PIPROCESS pPrc = CurrentProcess();
    _TCHAR *Arguments = (_TCHAR*) ProcessArgs();
    _TCHAR *ImageName;
    PIMODULE pMod;
    SCODE sc;
    TIME TOverall;
#if TAKE_LOAD_TIME
    TIME TLoad;
    C_ENTRY ModEntry = NULL;
#endif
    INT RetVal = EXIT_FAILURE;

    /* Need at least something to run
     */
    ImageName = Arguments;
    if (!ImageName || (ImageName[0] == 0)) {
        _tprintf(_TEXT("Usage: %s program [arguments]\r\n"),
                 ProcessImage());
        goto Out;
    }

    /* Should we pass any arguments to it
     */
    Arguments = _tcschr(Arguments, ' ');
    if (Arguments) {
        *Arguments = '\0';
        Arguments++;
    }

#if TAKE_LOAD_TIME
    /* Take the start time
     */
    TOverall = CurrentTime();

    /* Get it started
     */
    sc = pPrc->v->LoadImage(pPrc,
                            ImageName,
                            Arguments,
                            0,
                            &pMod);

    if ((sc != S_OK) || (pMod == NULL)) {
        _tprintf(_TEXT("Could not load %s(%s) sc = x%x pMod=%x\r\n"),
                 ImageName,
                 Arguments,
                 sc,
                 (ADDRESS)pMod);
        goto Out;
    }

    /* Take the load time
     */
    TLoad = TimeSubtract(CurrentTime(), TOverall);

    /* Get the entry point
     */
    sc = pMod->v->GetEntryPoint(pMod, (ADDRESS *)&ModEntry);
    if (sc != S_OK) {
        _tprintf(_TEXT("Could not get entry point for %s, sc = x%x\r\n"),
                 ImageName, sc);
        pMod->v->Release(pMod);
        goto Out;
    }

    /* Re-take the start time
     */
    TOverall = CurrentTime();

    /* Call the entry point directly
     */
    RetVal = ModEntry(_pTheBaseRtl);

    /* Take the execution time
     */
    TOverall = TimeSubtract(CurrentTime(), TOverall);

    /* Normalize times
     */
    TLoad = Int64DividedByInt32(TLoad,10);
    TOverall = Int64DividedByInt32(TOverall,10);

    /* Release last reference
     */
    pMod->v->Release(pMod);

    /* Say what
     */
    _tprintf(_TEXT("Load time %lld usec, "), TLoad);
    _tprintf(_TEXT("Exec time %lld usec, exit code %d\r\n"), 
             TOverall, RetVal);

#else
    /* Take the start time
     */
    TOverall = CurrentTime();

    /* Get it started
     */
    sc = pPrc->v->LoadImage(pPrc,
                            ImageName,
                            Arguments,
                            LOAD_IMAGE_CREATE_THREAD|LOAD_IMAGE_CALL_ENTRY_POINT,
                            &pMod);

    if ((sc != S_OK) || (pMod == NULL)) {
        _tprintf(_TEXT("Could not execute %s(%s) sc = x%x pMod=%x\r\n"),
                 ImageName,
                 Arguments,
                 sc,
                 (ADDRESS)pMod);
        return 1;
    }

    /* Wait for the test to complete.
     * We start off holding one ref.
     */
    pMod->v->WaitFor(pMod);

    /* See how long it took
     */
    TOverall = TimeSubtract(CurrentTime(), TOverall);
    TOverall = Int64DividedByInt32(TOverall,10);

    /* Get the exit code, ignore errors
     */
    (void) pMod->v->GetExitCode(pMod,&RetVal);

    /* Release last reference
     */
    pMod->v->Release(pMod);

    /* Say what
     */
    _tprintf(_TEXT("This took %lld usec total, exit code %d\r\n"), 
             TOverall, RetVal);

#endif

    /* Done
     */
 Out:
    exit(RetVal);
}
