/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>

/* Watches the mouse, stops when anything is typed on the keyboard.
 */
typedef struct _REPORT {
    UINT8 Buttons;
    INT8 x;
    INT8 y;
    INT8 Wheel;
} REPORT, *PREPORT;

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}

PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

volatile int MouseShouldRun = 2;
BOOL MouseDiffsOnly = FALSE;

_TCHAR *Positive(int i)
{
    return (i > 0) ? _T("+") : _T("");
}

void THREAD_LINKAGE MouseThread(THREAD_ARGUMENT arg)
{
    PIFILE Mouse = (PIFILE )arg;
    REPORT R, P;
    SCODE sc;
    UINT nBytes;

    memset(&P,0,sizeof P);
    while (MouseShouldRun > 1) {

        nBytes = 0;
        sc = Mouse->v->ReadAt(Mouse,
                              UINT64_ZERO,
                              (BYTE*)&R,
                              sizeof R,
                              &nBytes);
        if (FAILED(sc) || (nBytes != sizeof R))
            break;

        if (MouseDiffsOnly &&
            (memcmp(&R,&P,sizeof P) == 0))
            continue;
        P = R;

        _tprintf(_T("buttons %x position (%s%d,%s%d) wheel %s%d\n"),
                 R.Buttons,
                 Positive(R.x), R.x,
                 Positive(R.y), R.y,
                 Positive(R.Wheel), R.Wheel);
    }
    MouseShouldRun = 0;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    INT RetVal = EXIT_FAILURE;
    _TCHAR *MouseName = _T("mouse0");
    PIFILE Mouse = NULL;
    int c;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("No command line args (%x), test FAILED\n"), sc);
        exit(RetVal);
    }

    if (argc > 1) {
        MouseName = argv[1];
    }

    if (argc > 2)
        MouseDiffsOnly = ! MouseDiffsOnly;

    sc = BindToObject(CurrentNameSpace(), MouseName,
                      NAME_SPACE_READ,
                      &IID_IFile,
                      (void *)&Mouse);
    if (FAILED(sc)) {
        _tprintf(_T("Could not open mouse file %s\n"), MouseName);
        goto Done;
    }

    _tprintf(_T("Testing %s. Play with it, type anything to terminate\n"),
             MouseName);

    sc = CreateAThread(MouseThread,(THREAD_ARGUMENT)Mouse);
    if (FAILED(sc)) {
        _tprintf(_T("Failed to create a thread, sc=%x\n"), sc);
        goto Done;
    }

    /* Block until something is typed
     */
    c = _gettchar();

    MouseShouldRun = 1;
    while (MouseShouldRun)
        MilliSleep(1000);

    _tprintf(_T("\nTesting of %s complete.\n"), MouseName);

    RetVal = EXIT_SUCCESS;
Done:
    if (Mouse)
        Mouse->v->Release(Mouse);

    (void) FreeCmdLine(&argc,&argv);
    exit(RetVal);
    return 0;
}
