/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>

#include "../../src/web/sampling/builder.h"
#include "../../src/web/sampling/planner.h"

typedef struct INSTIGATOR {
    CONST_VTBL struct IInstigatorVtbl *v;   
    UINT RefCnt;
    /* Instigated task */
    PITASKBUILDER Task;
} INSTIGATOR, *PINSTIGATOR;

#define pInstigator(_i_) ((PINSTIGATOR)((_i_)))
#define iInstigator(_p_) ((PIINSTIGATOR)((_p_)))

/* Counts of tests attempted, succeeded */
int a, s;

PRIVATE SCODE InstigatorQueryInterface (
    PIINSTIGATOR pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_IInstigator);
}
    
PRIVATE UINT InstigatorAddRef (
    PIINSTIGATOR pThis)
{
    return AtomicInc(&pInstigator(pThis)->RefCnt);
}
    
PRIVATE UINT InstigatorRelease (
    PIINSTIGATOR pThis)
{
    PINSTIGATOR This = pInstigator(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    printf("tplanner.c:InstigatorRelease this=0x%x refs=%d\n", 
        This, This->RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */
        This->Task = NULL;
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}
    
const struct IInstigatorVtbl InstVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    InstigatorQueryInterface,
    InstigatorAddRef,
    InstigatorRelease,
    NULL,
    NULL,
    NULL
};    

PRIVATE SCODE MCT Dispatch(
    PTR Arg, PIRESERVATIONBUILDER r)
{
    PIINSTIGATOR iThis = (PIINSTIGATOR) Arg;
    METHOD_DECLARE_THIS("tplanner.c:Dispatch", PINSTIGATOR);

    PWSCONTEXT ctx;

    ++s;
    sc = r->v->CreateContext(r, &ctx);
    CHECKSC("create context object");
    printf("tplanner.c:Dispatch: context=%x\n", ctx);
    ++a;

METHOD_CLEANUP;
    if (ctx)
        free(ctx);

    return sc;
}

extern PIUNKNOWN PlannerCobMain(void);

int Test1(void)
{
    METHOD_DECLARE_NOTHIS("tplanner.c:Test1");
    PINSTIGATOR i;
    PIPLANNER p;
    PTASKBUILDER t;
    PIRESERVATIONBUILDER r = NULL;
    _TCHAR *name = _T("reservation-name");
    PIHEAP pHeap = CurrentHeap();
    TIME Period;

    a = 0;
    s = 0; 

    i = pHeap->v->
        Alloc(pHeap, 
            HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("tplanner.c:Test1")), 
            sizeof(INSTIGATOR), 0);
    if (!i)
        goto Out;

    i->v = &InstVtbl;
    i->RefCnt = 1;
    printf("Created instigator=0x%x v=0x%x refs=%d\n", 
        i, i->v, i->RefCnt);

    ++a;
    p = (PIPLANNER) PlannerCobMain();
    if (!p) goto Out;
    ++s;

    ++a;
    sc = p->v->CreateTask(p, (PIINSTIGATOR) i, &i->Task);
    CHECKSC("create task");
    t = (PTASKBUILDER) i->Task;
    ++s;

#if 1
    ++a;
    sc = i->Task->v->CreateReservation(i->Task, 
        _T("destination URI here"), 
        _T("feedback URI here"), 
        name,
        &r);
    CHECKSC("create reservation");
    ++s;

    ++a;
    sc = r->v->SetParameters(r, 
        CurrentTime(), TIME_RELATIVE(TIME_MILLIS(20)), TIME_RELATIVE(TIME_MILLIS(5)));
    CHECKSC("set reservation parameters");
    ++s;

    ++a;
    sc = r->v->SetDispatch(r, 
        (PTR) Dispatch, (PTR) i);
    CHECKSC("set dispatch function and data");
    ++s;

    ++a;
    Int32ToInt64(Period, TIME_RELATIVE(TIME_SECONDS(2)));
    sc = r->v->SetRecurrence(r, Period, 11);
    CHECKSC("set recurrence");
    ++s;
#else
    UnusedParameter(r);
#endif

    ++a;
    sc = i->Task->v->Activate(i->Task, (PIINSTIGATOR) i);
    {
        TIME Timeout;
        Int32ToInt64(Timeout, TIME_RELATIVE(TIME_SECONDS(2)));
        SleepUntil(Timeout);
    }
    CHECKSC("activate task");
    ++s;

METHOD_CLEANUP;
//    printf("Test1: attempted %d, succeeded with %d\n", a, s);
    i->v->Release((PIINSTIGATOR) i);

    return s - a;
}

int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS("tplanner.c:main");
    INT failed = 0;

    failed -= Test1();

//    printf("Tests completed - %d failed\n", failed);

    sc = S_OK;

METHOD_CLEANUP;
    return failed;
}
