/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <assert.h>
#include <soapif.h>
#include <util/list-double2.h>
#include <sampleif.h>
#include <wsif.h>
#include "../../src/web/sampling/statistics.h"

INT Consumed32[] = { 10000, 10100, 10200, 9900, 9000, 11000, 20000, 10000 };
#define COUNT (sizeof(Consumed32)/sizeof(INT))
INT64 Consumed64[WINDOW_SIZE];

/* For testing... */
#define CORRECT_MEAN 11275
#define CORRECT_VARIANCE 11308571
#define CORRECT_STDDEV 3362

UINT ConfidenceRequests[] = { 
    50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 
    60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 
    70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 
    80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 
    90, 91, 92, 93, 94, 95, 96, 97, 98, 99 };

#define REQUEST_COUNT (sizeof(ConfidenceRequests)/sizeof(UINT))

INT CorrectPredictions32[] = {
    11275, 11353, 11432, 11524, 11603, 11695, 11774, 11865, 11944, 12036, 
    12115, 12207, 12299, 12378, 12470, 12562, 12653, 12745, 12837, 12929, 
    13034, 13126, 13231, 13323, 13428, 13533, 13638, 13757, 13862, 13980,
    14098, 14216, 14348, 14479, 14610, 14755, 14899, 15057, 15214, 15385, 
    15582, 15779, 15989, 16226, 16501, 16803, 17158, 17591, 18169, 19089 };

INT64 CorrectPredictions64[REQUEST_COUNT];

#define DATA_COLLECTION _T("pool1")

PRIVATE INT8 days[2][12] = {{31,28,31,30,31,30, 31,31,30,31,30,31},
                            {31,29,31,30,31,30, 31,31,30,31,30,31}};
PRIVATE char *DayName[7] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

PRIVATE BOOL IsLeap(UINT y)
{
    BOOL is = (!(y%4) && ((y%100) || !(y%400)));
    //DBG(("IsLeap(%d) => %d\n", y, is));
    return is;
}

void TimeFormat(TIME t, _TCHAR *buf, UINT Len)
{
    TIME t2;
    INT leaps, f,s,min,h,d,mo,y, wd, xxx;
    BOOL relative = TimeIsRelative(t) || Int64Equal(t, TIME_ORIGIN);

    if (relative)
        t = Int64Subtract(TIME_ORIGIN, t);

    /* First deal with fractions */
    t2 = Int64ModulusInt32(t,  10000000);
    t = Int64DividedByInt32(t, 10000000);
    f = Int64ToInt32(t2);

    if (relative) {
        _stprintf(buf, _T("P%lld.%07dS"), t,f);
        return;
    }

    /* seconds */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    s = Int64ToInt32(t2);

    /* minutes */
    t2 = Int64ModulusInt32(t, 60);
    t = Int64DividedByInt32(t, 60);
    min = Int64ToInt32(t2);

    /* hours */
    t2 = Int64ModulusInt32(t, 24);
    t = Int64DividedByInt32(t, 24);
    h = Int64ToInt32(t2);
    
#if !defined(arm)               /* arm_arm generates call to __rt_divtest */
    /* days since Jan 1, 1601, Gregorian calendar. */
    xxx = wd = Int64ToInt32(t);
    wd %= 7;
    wd++;
#endif

    /* figure out rough year, then adjust with leap days */
    t2 = Int64ModulusInt32(t, 365);
    t = Int64DividedByInt32(t, 365);
    d = Int64ToInt32(t2);

    y = Int64ToInt32(t);        /* from 1601 */

    leaps = y / 4 - y / 100 + y / 400; /* from 1600, does not include current year */

    d -= leaps;

    //printf("y=%d leaps=%d now=%d d=%d wd=%d [%s] totaldays=%d\n",
             //y, leaps, (y%4 == 0), d, wd, DayName[wd-1], xxx);

    while (d < 0) {
      /* if we are millenia in the future there might be more bogus leaps
       * than one year worth */
        leaps = IsLeap(y+1601-1);       /* last year */
        d += (leaps ? 366 : 365);
        y--;
    }
    leaps = IsLeap(y+1601);       /* this year */

    for (mo = 0; mo < 12; ) {
        if (d < days[leaps][mo])
            break;
        d -= days[leaps][mo];
        mo++;
    }
    /* make them one-based */
    mo++;
    d++;
    y += 1601;

    /* Time zone: UTC (boot time?) */
    _stprintf(buf,_T("%04d-%02d-%02dT%02d:%02d:%02d.%07dZ"), y,mo,d,h,min,s,f);
}

PRIVATE UINT32 RandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}

_TCHAR *CreateUniqueSuffix(void)
{
    TIME current = CurrentTime();
#define BUFLEN 256
    _TCHAR buf[BUFLEN];
    UINT i;
    TimeFormat(current, buf, BUFLEN);
    for (i = 0; i < BUFLEN; ++i)
        if ('\0' == buf[i])
            break;
    if (10 > BUFLEN - i)
        return NULL;
    _stprintf(buf + i, _T("%d"), RandUint32());
    return _tcsdup(buf);
}

int _tmain(INT argc, _TCHAR *argv[])
{
    SCODE sc;
    char *ErrorMessage = NULL;
    PISTATISTICS pStat = NULL;
    UINT i;
    TIME Mean, Variance, Prediction, Test;

    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    printf("%s\n", CreateUniqueSuffix());
    
    Int32ToInt64(Mean, 0);
    Int32ToInt64(Variance, 0);
    Int32ToInt64(Prediction, 0);

    /* Set up test data */
    assert(sizeof(CorrectPredictions32)/sizeof(INT) == REQUEST_COUNT);
    for (i = 0; i < WINDOW_SIZE; ++i) 
        Int32ToInt64(Consumed64[i], Consumed32[i % COUNT]);
    for (i = 0; i < REQUEST_COUNT; ++i)
        Int32ToInt64(CorrectPredictions64[i], CorrectPredictions32[i]);

    /* Load the COB */
    sc = BindToObject(CurrentNameSpace(), _T("COB/statistics.cob"), NAME_SPACE_READ,
                      &IID_IStatistics, (void **) &pStat);
    if (FAILED(sc)) {
        ErrorMessage = "tstat.c:_tmain: No statistics COB\n";
        goto Failed;
    }
DBG(("A\n"));
    /* Store some numbers */
    for (i = 0; i < WINDOW_SIZE; ++i) {
        sc = pStat->v->AddSample(pStat, DATA_COLLECTION, i, Consumed64[i], Consumed64[i]);
        if (FAILED(sc)) {
            ErrorMessage = "tstat.c:_tmain: Couldn't store value\n";
            goto Failed;
        }
    }

    /* Compute the mean and variance */
    sc = pStat->v->GetNormalDistribution(pStat, DATA_COLLECTION, &Mean, &Variance);
    if (FAILED(sc)) {
        DBG(("sc=%x\n", sc));
        ErrorMessage = "tstat.c:_tmain: Failed to compute mean and variance\n";
        goto Failed;
    }

    Int32ToInt64(Test, 11275);
    if (! Int64Equal(Mean, Test)) {
        DBG(("mean=%lld correct=%lld\n", Mean, Test));
        ErrorMessage = "Incorrect mean";
        goto Failed;
    }

    Int32ToInt64(Test, 11308571);
    if (! Int64Equal(Variance, Test)) {
        DBG(("variance=%lld correct=%lld\n", Variance, Test));
        ErrorMessage = "Incorrect variance";
        goto Failed;
    }

    /* Request a prediction */
     for (i = 0; i < REQUEST_COUNT; ++i) {
        UINT Confidence = ConfidenceRequests[i];
        sc = pStat->v->GetPredictionFromNormalDistribution(
            pStat, DATA_COLLECTION, Confidence, &Prediction);
        if (FAILED(sc)) {
            ErrorMessage = "tstat.c:_tmain: Failed to get prediction\n";
            goto Failed;
        }
        DBG(("tstat.c: Testing confidence=%d\n", Confidence));
        if (! Int64Equal(CorrectPredictions64[i], Prediction)) {
            DBG(("FAILED TEST tstat.c: prediction=%lld correct=%lld\n", 
                Prediction, CorrectPredictions64[i]));
//            ErrorMessage = "tstat.c: _tmain: Aborting because of failed test\n";
//            goto Failed;
        }
    }

    printf("%lld\n", Prediction);

    return 0;

Failed:
    DBG(("Statistics tests failed\n"));
    if (ErrorMessage)
        DBG((ErrorMessage));
    return -1;
}

