/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    mkfont.c
 *
 * Purpose:
 *
 *    This program takes a font description file and generates a C source
 *    appropriate for use as kernel font.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * Date:
 *    10/90
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>

char Usage[] = 
"mkfont [+] [-flip] [-8x8] [input-file [first-glyph [last-glyph]]]";

int readln(int fd, char *b, int nlines);

char fn[256];

main(argc,argv)
    char **argv;
{
    int         fd;
    FILE        *fout;
    int         i, j, k, n, l, gsize;
    int         first, last;
    char        *fname, *fs, *pgm = argv[0];
    char        buf[16*9];
    int         verbose = 0;
    int         flippem = 0;
    int         eightby8 = 0;

    if (argc > 1 && argv[1][0] == '+')
        verbose++, argc--, argv++;

    if (argc > 1 && argv[1][0] == '-') {
        if (argv[1][1] == 'f')
            flippem++;
        else
        if (argv[1][1] == '8')
            eightby8++;
        else {
            printf(Usage);
            exit(-1);
        }
        argc--, argv++;
    }

    if (argc > 1)
        fname = argv[1];
    else
        fname = "font8x15";

    fs = (eightby8) ? "8x8" : "8x15";

    first = 0;
    last = 191;    /* 8-bit font, offset by 'space' */
    if (argc > 2) {
        first = atoi(argv[2]);
        last = first + 1;
    }
    if (argc > 3)
        last = atoi(argv[3]) + 1;

    sprintf(fn,"%s.dat",fname);
    fd = open(fn, 0, 0);
    if (fd < 0) {
        printf("Cant open %s for reading\n",fn);
        exit(-1);
    }
    sprintf(fn,"%s.c",fname);
    fout = fopen(fn, "w");
    if (fout == NULL) {
        printf("Cant open %s for writing\n",fn);
        exit(-1);
    }

    fprintf(fout, "/* \n\
 * Copyright (c) Microsoft Corporation. All rights reserved.\n\
 */\n\
/*\n\
 * THIS FILE WAS GENERATED BY %s FROM %s\n\
 * IF YOU NEED TO, BE SURE YOU EDIT THE REAL THING!\n\
 */\n\
\n\
/*\n\
 *      Object:\n\
 *              font_%s                 EXPORTED array\n\
 *\n\
 *              An %s ISO-Latin-1 font\n\
 *\n\
 * The smallest index in this array corresponds to a\n\
 * space. So, we start at 0x20 in the Unicode table.\n\
 * A hole is also left from 0x7f to 0x9f (incl) because non-print.\n\
 * Note that glyphs are mirrored (bitorder)\n\
 * the commented bitmap shows how they really look like\n\
 */\n\
\n\
unsigned char font_%s[] = {\n", pgm, fname, fs, fs, fs);

#define READ(a,b,c) {read(a,b,c);write(1,b,c);}
#if _DEBUG
#define echo(b,c) write(1,b,c)
#else
#define echo(b,c)
#endif

    gsize = (eightby8) ? (8+1) : (15+1);

    if (flippem)
        fprintf(fout, "/* NB: Encoding is bit-flipped */\n");

skip_comments:
    read(fd, buf, 1);
    echo(buf,1);
    if (buf[0] == '#') {
        readln(fd,buf,1);
        goto skip_comments;
    }
    lseek(fd, -1, SEEK_CUR);    /* put char back */

    /* if we must, skip some
     */
    for (l = 0; l < first; l++) {
        readln(fd, buf, gsize);
    }

    /* scan for real now
     */
    if (verbose)
        printf("\t");
    for (i = first; i < last; i++) {
        /* read one full glyph
         */
        if (readln(fd, buf, gsize) < 0)
            break;
        if (verbose)
            printf("Character '%c':\n\n\t", buf[0]);
        /* index and char itself in comments
         */
        fprintf(fout, "/* %3x '%c' */\n", i, 0x20+i);

        /* encode and possibly mirror each one of the 15 scanlines
         */
        for (n = 0; n < ((eightby8) ? 8 : 15); n++) {
            unsigned char    cc[9], swap = 0;
            /* 8 bits per scanline
             */
            cc[8] = 0;
            k = 2+(n*9);
            for (j = 0; j < 8; k++, j++) {
                if (verbose)
                    printf("%c", (buf[k] == '1') ? '@' : ' ');
                if (flippem) {
                    swap = ((buf[k] - '0') << 7) | (swap >> 1);
                } else {
                    swap = (swap << 1) | (buf[k] - '0');
                }
                cc[j] = buf[k];
            }
            fprintf(fout,"\t/* %8s */\t%#2x,\n", cc, (unsigned char)swap);
            if (verbose)
                printf("\n\t");
        }
    }
    fprintf(fout, "};\n");
    fprintf(fout, "\n\
/*\n\
 *      Object:\n\
 *              ffont                           EXPORTED function\n\
 *\n\
 *              Returns the glyph for a character\n\
 *\n\
 * The font file is slightly compressed, this function knows where\n\
 * the holes are. See comment above.\n\
 */\n\
unsigned char *ffont(unsigned int c)\n\
{\n\t\
c &= 0xff;\n\t\
if (c > 0x1f && c < 0x7f)\n\t\
    c = c - ' ';\n\t\
else if (c > 0x9f)\n\t\
    c = c - ' ' - (0xa0 - 0x7f);\n\t\
else\n\t\
    c = 0; /* non-printable --> space */\n\t\
return &font_%s[%s];\n\
}\n", fs, (eightby8) ? "c<<3" : "(c<<4)-c");
    fclose(fout);
}

/* Read canonical lines (e.g. drop CR)
 */

int readln(int fd, char *b, int nlines)
{
    int i = -1;
    while (nlines-- > 0) {
        do {
            i++;
        dont:
            read(fd, b+i, 1);
            if (b[i] == '\r')
                goto dont;
            echo(b+i,1);
        } while (b[i] != '\n');
    }
    return i;
}
