/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *    Encoding of ARM instructions
 *
 *      "ARM Architectural Reference Manual"
 *      Edited by Dave Jaggar.  Prentice Hall ed.
 *      ISBN 0-13-736299-4
 */

#ifndef    _ARM_INSTRUCTION_H_
#define    _ARM_INSTRUCTION_H_ 1

#ifdef _MSC_VER
#pragma pack(push,1)
#endif

typedef union {
    /*
     *    All instructions are 32 bits wide
     */
    unsigned int    bits;

    /*
     *    Jump instruction contain a 24 bit offset,
     *    which is shifted and added to the PC to form
     *    the 32 bit destination address
     */
    struct {
#if    (BYTE_ORDER == BIG_ENDIAN)
        unsigned int
            cond : 4,
            opcode : 4;
        signed int    target : 24;
#else    /* BYTE_NSF */
        signed int
            target : 24;
        unsigned int
            opcode : 4,
            cond : 4;
#endif    /* (BYTE_ORDER == BIG_ENDIAN) */
    } j_format;

    /*
     *    Data processing immediate & friends (cond, shifter, regs)
     */
    struct {
#if    (BYTE_ORDER == BIG_ENDIAN)
        unsigned short
            cond : 4,
            class : 3,
            opcode : 4,
            s_l : 1,
            rn : 4;
        union {
            struct {
                signed short
                    xx : 4,
                    s_val : 12;
            } us;
            struct {
                unsigned short
                    rd : 4,
                    u_val : 12;
            } uu;
            struct {
                unsigned short
                    rd : 4,
                    rotate : 4,
                    immediate : 8;
            } ss;
            struct {
                unsigned short
                    rd : 4,
                    shiftval : 5,
                    shiftmode : 2,
                    xx : 5;
            } ls;
            struct {
                unsigned short
                    rd : 4,
                    shift : 8,
                    rm : 4;
            } ns;
        } shifter;
#else    /* (BYTE_ORDER == BIG_ENDIAN) */
        union {
            struct {
                signed short
                    s_val : 12,
                    xx : 4;
            } us;
            struct {
                unsigned short
                    u_val : 12,
                    rd : 4;
            } uu;
            struct {
                unsigned short
                    immediate : 8,
                    rotate : 4,
                    rd : 4;
            } ss;
            struct {
                unsigned short
                    xx : 5,
                    shiftmode : 2,
                    shiftval : 5,
                    rd : 4;
            } ls;
            struct {
                unsigned short
                    rm : 4,
                    shift : 8,
                    rd : 4;
            } ns;
        } shifter;
        unsigned short
            rn : 4,
            s_l : 1,
            opcode : 4,
            class : 3,
            cond : 4;
#endif    /* (BYTE_ORDER == BIG_ENDIAN) */
    } i_format;

    /*
     *    Generic helper for decode (cond, opcode8, regs, opcode4, rm)
     */
    struct {
#if    (BYTE_ORDER == BIG_ENDIAN)
        unsigned int
            cond : 4,
            opcode : 8,
            rn : 4,
            rd : 4,
            rs : 4,
            op2 : 4,
            rm : 4;
#else    /* (BYTE_ORDER == BIG_ENDIAN) */
        unsigned int
            rm : 4,
            op2 : 4,
            rs : 4,
            rd : 4,
            rn : 4,
            opcode : 8,
            cond : 4;
#endif    /* (BYTE_ORDER == BIG_ENDIAN) */
    } g_format;

    /*
     *    Anal-retentive for hard cases
     */
    struct {
#if    (BYTE_ORDER == BIG_ENDIAN)
        unsigned int
            cond : 4,
            class : 3,
            b24 : 1,
            b23 : 1,
            b22 : 1,
            b21 : 1,
            b20 : 1,
            rn : 4,
            rd : 4,
            rs : 4,
            b7 : 1,
            b6 : 1,
            b5 : 1,
            b4 : 1,
            rm : 4;
#else    /* (BYTE_ORDER == BIG_ENDIAN) */
        unsigned int
            rm : 4,
            b4 : 1,
            b5 : 1,
            b6 : 1,
            b7 : 1,
            rs : 4,
            rd : 4,
            rn : 4,
            b20 : 1,
            b21 : 1,
            b22 : 1,
            b23 : 1,
            b24 : 1,
            class : 3,
            cond : 4;
#endif    /* (BYTE_ORDER == BIG_ENDIAN) */
    } a_format;

} arm_instruction;

#define ARM_SHIFT_LSL 0
#define ARM_SHIFT_LSR 1
#define ARM_SHIFT_ASR 2
#define ARM_SHIFT_ROR 3
#define ARM_SHIFT_RRX 3

/*
 *
 *    Opcodes: 4 bit view (b,bl)
 *
 */

#define op4_lsm      0x8
#define op4_lsmp     0x9
#define op4_b        0xa
#define op4_bl       0xb
#define op4_cp_ls    0xc
#define op4_cp_lsp   0xd
#define op4_cp_dp    0xe
#define op4_swi      0xf

/*
 *
 *    Opcodes: 8 bit view (add/orr/mul..)
 *
 */

#define op_bx        0x12
#define op_swp       0x10
#define op_swpb      0x14

/*
 *
 *    Encoding of floating point instructions
 *
 */

/* TBD */

#ifdef _MSC_VER
#pragma pack(pop)
#endif

#endif    /* _ARM_INSTRUCTION_H_ */
