/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <windows.h>
#include "bo.h"
#include "mips_instruction.h"


static char *mips_op_name[64] = {
/* 0 */ "spec", "bcond","j",    "jal",  "beq",  "bne",  "blez", "bgtz",
/* 8 */ "addi", "addiu","slti", "sltiu","andi", "ori",  "xori", "lui",
/*16 */ "cop0", "cop1", "cop2", "cop3", "beql", "bnel", "blezl","bgtzl",
/*24 */ "op60", "op64", "op68", "op6c", "op70", "op74", "op78", "op7c",
/*32 */ "lb",   "lh",   "lwl",  "lw",   "lbu",  "lhu",  "lwr",  "ld",
/*40 */ "sb",   "sh",   "swl",  "sw",   "opb0", "opb4", "swr",  "sd",
/*48 */ "lwc0", "lwc1", "lwc2", "lwc3", "ldc0", "ldc1", "ldc2", "ldc3",
/*56 */ "swc0", "swc1", "swc2", "swc3", "sdc0", "sdc1", "sdc2", "sdc3"
};

static char *mips_spec_name[64] = {
/* 0 */ "sll",  "spec01","srl", "sra",  "sllv", "spec05","srlv","srav",
/* 8 */ "jr",   "jalr", "joff","pri","syscall","break","spec16","tas",
/*16 */ "mfhi", "mthi", "mflo", "mtlo", "spec24","spec25","spec26","spec27",
/*24 */ "mult", "multu","div",  "divu", "spec34","spec35","spec36","spec37",
/*32 */ "add",  "addu", "sub",  "subu", "and",  "or",   "xor",  "nor",
/*40 */ "spec50","spec51","slt","sltu", "spec54","spec55","spec56","spec57",
/*48 */ "spec60","spec61","spec62","spec63","spec64","spec65","spec66","spec67",
/*56 */ "spec70","spec71","spec72","spec73","spec74","spec75","spec76","spec77"
};

static char *mips_bcond_name[32] = {
/* 0 */ "bltz", "bgez", "bltzl", "bgezl", 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 16 */ "bltzal", "bgezal", "bltzall", "bgezall",
};

static char *mips_cop1_name[64] = {
/* 0 */ "fadd", "fsub", "fmpy", "fdiv", "fsqrt","fabs", "fmov", "fneg",
/* 8 */ "fop08","fop09","fop0a","fop0b","fop0c","fop0d","fop0e","fop0f",
/*16 */ "fop10","fop11","fop12","fop13","fop14","fop15","fop16","fop17",
/*24 */ "fop18","fop19","fop1a","fop1b","fop1c","fop1d","fop1e","fop1f",
/*32 */ "fcvts","fcvtd","fcvte","fop23","fcvtw","fop25","fop26","fop27",
/*40 */ "fop28","fop29","fop2a","fop2b","fop2c","fop2d","fop2e","fop2f",
/*48 */ "fcmp.f","fcmp.un","fcmp.eq","fcmp.ueq","fcmp.olt","fcmp.ult",
        "fcmp.ole","fcmp.ule",
/*56 */ "fcmp.sf","fcmp.ngle","fcmp.seq","fcmp.ngl","fcmp.lt","fcmp.nge",
        "fcmp.le","fcmp.ngt"
};

static char *mips_fmt_name[16] = {
        "s",    "d",    "e",    "fmt3",
        "w",    "fmt5", "fmt6", "fmt7",
        "fmt8", "fmt9", "fmta", "fmtb",
        "fmtc", "fmtd", "fmte", "fmtf"
};

static char *mips_sbregister_name[32] = {
                "zero", "at",   "v0",   "v1",   "a0",   "a1",   "a2",   "a3",
                "t0",   "t1",   "t2",   "t3",   "t4",   "t5",   "t6",   "t7",
                "s0",   "s1",   "s2",   "s3",   "s4",   "s5",   "s6",   "s7",
                "t8",   "t9",   "k0",   "k1",   "gp",   "sp",   "s8",   "ra"
};

static char *mips_c0_opname[32] = {
        "c0op0","tlbr", "tlbwi","c0op3","c0op4","c0op5","tlbwr","c0op7",
        "tlbp", "c0op9","c0op10","c0op11","c0op12","c0op13","c0op14","c0op15",
        "rfe",  "c0op17","c0op18","c0op19","c0op20","c0op21","c0op22","c0op23",
        "c0op24","c0op25","c0op26","c0op27","c0op28","c0op29","c0op30","c0op31"
};

static char *mips_c0_reg[32] = {
        "index","random","tlblo","c0r3","context","c0r5","c0r6","c0r7",
        "badvaddr","c0r9","tlbhi","c0r11","sr", "cause","epc",  "c0r15",
        "c0r16","c0r17","c0r18","c0r19","c0r20","c0r21","c0r22","c0r23",
        "c0r24","c0r25","c0r26","c0r27","c0r28","c0r29","c0r30","c0r31"
};

static int regcount;            /* how many regs used in this inst */
static int regnum[6];           /* which regs used in this inst */


static char *
mips_register_name (ireg)
{
    int i;

    for (i = 0; i < regcount; i++)
        if (regnum[i] == ireg)
            break;
    if (i >= regcount)
        regnum[regcount++] = ireg;
    return (mips_sbregister_name[ireg]);
}

/* Disassemble instruction I, located at IADR.
 * RUNTADDR is normally tha same, it is used to
 * disassemble from object files [in which case
 * IADR is the file offset and RUNTADDR is the
 * address at which the code has been relocated]
 */
void
MipsDisasm(
       unsigned iadr,
       void *pinst,
       int runtaddr)
{
    mips_instruction i = *((mips_instruction*)pinst);
    int bdelay;
    short signed_immediate;
    unsigned unsigned_immediate;

    regcount = 0;
    bdelay = 0;

    switch (i.j_format.opcode) {
    case op_special:
        if (i.bits == 0) {
            printf("nop");
            break;
        }
        else if (i.r_format.func == op_addu && i.r_format.rt == 0) {
            printf("move\t%s,%s",
                      mips_register_name(i.r_format.rd),
                      mips_register_name(i.r_format.rs));
            break;
        }
        printf("%s", mips_spec_name[i.r_format.func]);
        switch (i.r_format.func) {
        case op_sll:
        case op_srl:
        case op_sra:
            printf("\t%s,%s,%d",
                      mips_register_name(i.r_format.rd),
                      mips_register_name(i.r_format.rt),
                      i.r_format.shamt);
            break;
        case op_sllv:
        case op_srlv:
        case op_srav:
            printf("\t%s,%s,%s",
                      mips_register_name(i.r_format.rd),
                      mips_register_name(i.r_format.rt),
                      mips_register_name(i.r_format.rs));
            break;
        case op_mfhi:
        case op_mflo:
            printf("\t%s", mips_register_name(i.r_format.rd));
            break;
        case op_jr:
        case op_jalr:
            bdelay = 1;
            /* fall through */
        case op_mtlo:
        case op_mthi:
            printf("\t%s", mips_register_name(i.r_format.rs));
            break;
        case op_mult:
        case op_multu:
        case op_div:
        case op_divu:
        case op_joff:
            printf("\t%s,%s",
                      mips_register_name(i.r_format.rs),
                      mips_register_name(i.r_format.rt));
            break;

        case op_syscall:
            break;
        case op_break:
            printf("\t%d", i.r_format.rs*32+i.r_format.rt);
            break;
        case op_tas:
            printf("\t%s", mips_register_name(4));
            break;
        default:
            printf("\t%s,%s,%s",
                      mips_register_name(i.r_format.rd),
                      mips_register_name(i.r_format.rs),
                      mips_register_name(i.r_format.rt));
            break;
        };
        break;
    case op_bcond:
        printf("%s\t%s,",
                  mips_bcond_name[i.i_format.rt],
                  mips_register_name(i.i_format.rs));
        goto branch_displacement;
    case op_blez:
    case op_blezl:
    case op_bgtz:
    case op_bgtzl:
        printf("%s\t%s,", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rs));
        goto branch_displacement;
    case op_beq:
    case op_beql:
        if (i.i_format.rs == 0 && i.i_format.rt == 0) {
            printf("b\t");
            goto branch_displacement;
        }
        /* fall through */
    case op_bne:
    case op_bnel:
        printf("%s\t%s,%s,", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rs),
                  mips_register_name(i.i_format.rt));
    branch_displacement:
        signed_immediate = i.i_format.simmediate;
        printlabel(runtaddr+4+(signed_immediate<<2),
                   iadr+4+(signed_immediate<<2));
        bdelay = 1;
        break;
    case op_cop0:
        switch (i.r_format.rs) {
        case op_bc:
            printf("bc0%c\t", "ft"[i.r_format.rt]);
            goto branch_displacement;
        case op_mtc:
            printf("mtc0\t%s,%s",
                      mips_register_name(i.r_format.rt),
                      mips_c0_reg[i.f_format.fs]);
            break;
        case op_mfc:
            printf("mfc0\t%s,%s",
                      mips_register_name(i.r_format.rt),
                      mips_c0_reg[i.f_format.fs]);
            break;
        default:
            printf("c0\t%s", mips_c0_opname[i.f_format.func&31]);
            break;
        };
        break;
    case op_cop1:
        switch (i.r_format.rs) {
        case op_bc:
            printf("bc1%c\t", "ft"[i.r_format.rt]);
            goto branch_displacement;
        case op_mtc:
            printf("mtc1\t%s,f%d",
                      mips_register_name(i.r_format.rt),
                      i.f_format.fs);
            break;
        case op_mfc:
            printf("mfc1\t%s,f%d",
                      mips_register_name(i.r_format.rt),
                      i.f_format.fs);
            break;
        default:
            printf("%s.%s\tf%d,f%d,f%d",
                      mips_cop1_name[i.f_format.func],
                      mips_fmt_name[i.f_format.fmt],
                      i.f_format.fd, i.f_format.fs, i.f_format.ft);
            break;
        };
        break;

    case op_j:
    case op_jal:
        printf("%s\t", mips_op_name[i.j_format.opcode]);
        unsigned_immediate = ((iadr+4)&~((1<<28)-1)) +(i.j_format.target<<2);
        printsym(unsigned_immediate,iadr,TRUE);
        bdelay = 1;
        break;

    case op_swc1:
    case op_lwc1:
        printf("%s\tf%d,", mips_op_name[i.i_format.opcode],
                  i.i_format.rt);
        goto loadstore;

    case op_lb:
    case op_lh:
    case op_lw:
    case op_lbu:
    case op_lhu:
    case op_sb:
    case op_sh:
    case op_sw:
        printf("%s\t%s,", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rt));
    loadstore:
        signed_immediate = i.i_format.simmediate;
        if (signed_immediate < 0) {
            printf("-");
            signed_immediate *= -1;
        }
        printf("%x(%s)", signed_immediate,
                  mips_register_name(i.i_format.rs));
        printsym(i.i_format.uimmediate,iadr,FALSE);
#if 0
        if (showregs && task == TASK_NULL) {
            signed_immediate = i.i_format.simmediate;
            printf(" <%x>", signed_immediate +
                      getreg_val(i.i_format.rs));
        }
#endif
        break;

    case op_ori:
    case op_xori:
        if (i.i_format.rs == 0) {
            printf("li\t%s,%x",
                      mips_register_name(i.i_format.rt),
                      i.i_format.uimmediate);
            break;
        }
        /* fall through */
    case op_andi:
        printf("%s\t%s,%s,%x", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rt),
                  mips_register_name(i.i_format.rs),
                  i.i_format.uimmediate);
        break;
    case op_lui:
        printf("%s\t%s,", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rt),
                  i.i_format.uimmediate);
        printsym(i.i_format.uimmediate,iadr,TRUE);
        break;
    case op_addi:
    case op_addiu:
        if (i.i_format.rs == 0) {
            signed_immediate = i.i_format.simmediate;
            printf("li\t%s,",
                      mips_register_name(i.i_format.rt),
                      signed_immediate);
            printsym(signed_immediate,iadr,TRUE);
            break;
        }
        /* fall through */
    default:
        signed_immediate = i.i_format.simmediate;
        printf("%s\t%s,%s,", mips_op_name[i.i_format.opcode],
                  mips_register_name(i.i_format.rt),
                  mips_register_name(i.i_format.rs));
        if (signed_immediate < 0) {
            printf("-");
            signed_immediate *= -1;
        }
        printsym(signed_immediate,iadr,TRUE);
        break;
    }

#if 0
    /* print out the registers use in this inst */
    if (showregs && regcount > 0 && task == TASK_NULL) {
        printf("\t<");
        for (ireg = 0; ireg < regcount; ireg++) {
            if (ireg != 0)
                printf(",");
            printf("%s=%x",
                      mips_sbregister_name[regnum[ireg]],
                      getreg_val(regnum[ireg]));
        }
        printf(">");
    }
    if (0 && bdelay) {
        printf("\n\t");
        mips_print_instruction(iadr+4, peek_inst(iadr+4, task),
                               showregs, task);
    }
#endif
}

