/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* 
 * Poor Man's Dynamic Link Libraries
 * 
 * ELF version [MIPS ARM]
 */

#include <windows.h>
#include <assert.h>
#include "import.h"
#include "bo.h"

typedef unsigned long  UINT32;
typedef   signed long   INT32;
typedef unsigned short UINT16;
typedef   signed short  INT16;
typedef unsigned char  UINT8;
typedef const char *CTSTR;
typedef void * THREAD_FUNCTION;
#define E_INVALID_PARAMETER \
  MAKE_HRESULT(SEVERITY_ERROR,FACILITY_WIN32,ERROR_INVALID_PARAMETER)
#define E_FILE_NOT_FOUND \
  MAKE_HRESULT(SEVERITY_ERROR,FACILITY_WIN32,ERROR_FILE_NOT_FOUND)

#if !defined(offsetof)
#define offsetof(s,m)  ((UINT)&(((s *)0)->m))
#endif

int nInUse = 0;
#define ALLOCATE(_ret_,_type_,_size_,_msg_) { \
        _ret_ = (_type_)GlobalAlloc(GPTR,_size_); nInUse++; \
        if (fFlags & F_MEM) \
          printf("%x MEMORY Alloc(%d) %s\n",_ret_,_size_,_msg_); \
        } \
        /* if (_ret_) memset(_ret_,0,_size_); Done by GlobalAlloc(GPTR) */

#define REALLOCATE(_ptr_,_type_,_size_,_ret_,_msg_) { \
        void *p = GlobalAlloc(GPTR,_size_); \
        if (p) { \
           memcpy(p,_ptr_,GlobalSize(_ptr_)); GlobalFree(_ptr_);\
        } \
        if (fFlags & F_MEM) \
          printf("%x MEMORY re-Alloc(%d) -> %x %s\n",_ptr_,_size_,p,_msg_); \
        _ret_ = (_type_)p; \
        }

#define FREE(_p_,_msg_) { \
        if (fFlags & F_MEM) printf("%x MEMORY Free %s\n",_p_,_msg_); \
        GlobalFree(_p_); nInUse--; \
        }


/* Types defined in the ELF specification
 */
typedef UINT32 Elf32_Addr, Elf32_Off, Elf32_Word;
typedef INT32 Elf32_Sword;
typedef UINT16 Elf32_Half;


/*
 * This machdep file should provide the following functions
 * (or replacements thereof) and define the type of FILE_HANDLE.
 */
#include "_loader.c"


extern FILE_HANDLE _Open( CTSTR FileName, UINT Mode);
extern INT _Close( FILE_HANDLE File);
extern INT _Read( FILE_HANDLE File, BYTE *Buffer, UINT nBytes);
extern UINT _Seek( FILE_HANDLE File, UINT Position);
extern INT _Stat( FILE_HANDLE File);
/*extern INT _Write( FILE_HANDLE File, BYTE *Buffer, UINT nBytes); */
/*extern INT _Remove( TSTR FileName ); */

/* Random stuff
 */
typedef char *NAME;

#define F_DISASM 0x0001  /* print disassembly (before) */
#define F_SYMS   0x0002  /* print symbol table */
#define F_RELS   0x0004  /* print relocations */
#define F_IMP    0x0008  /* print import section */
#define F_DWARF  0x0010  /* track what happens to ".debug" sections */
#define F_MEM    0x0020  /* trace memory alloc/free */
#define F_EXP    0x0040  /* print export section */
#define F_MUNGE  0x0080  /* force munge()-ing */
#define F_NOISE  0x0100  /* random commentary */
#define F_ALL    0x015f  /* dont track memory */

#define F_STRIP  0x10000

UINT fFlags = 0;
UINT fOptions = F_STRIP;
NAME ScopeFileName = NULL;

/*
 * Describe structure of ELF object file format.
 */

/* Size of file identifier in ELF header */
#define EI_NIDENT       16

/* Identification indexes for e_ident field in ELF header */
#define EI_MAG0     0         /* file ID byte 0 */
#define EI_MAG1     1         /* file ID byte 1 */
#define EI_MAG2     2         /* file ID byte 2 */
#define EI_MAG3     3         /* file ID byte 3 */
#define EI_CLASS    4         /* file class (capacity) */
#define EI_DATA     5         /* data encoding */
#define EI_VERSION  6         /* file version */
#define EI_PAD      7         /* start of padding bytes */
#define EI_NIDENT  16         /* size of e_ident[] */

/* "Magic number" in e_ident field of ELF header */
#define ELFMAG0     '\x7f'    /* e_ident[0] */
#define ELFMAG1     'E'       /* e_ident[1] */
#define ELFMAG2     'L'       /* e_ident[2] */
#define ELFMAG3     'F'       /* e_ident[3] */

/* File class (or capacity) in e_ident[4] of ELF header */
#define ELFCLASSNONE  0       /* invalid class */
#define ELFCLASS32    1       /* 32-bit processor */
#define ELFCLASS64    2       /* 64-bit processor */

/* Data encoding in e_ident[5] of ELF header */
#define ELFDATANONE   0       /* invalid data encoding */
#define ELFDATA2LSB   1       /* little-endian format */
#define ELFDATA2MSB   2       /* big-endian format */

#if (BYTE_ORDER==BIG_ENDIAN)
#define MYBYTEORDER ELFDATA2MSB
#else
#define MYBYTEORDER ELFDATA2LSB
#endif

/* Object file type in e_type field of ELF header */
#define ET_NONE     0         /* no file type */
#define ET_REL      1         /* relocatable file */
#define ET_EXEC     2         /* executable file */
#define ET_DYN      3         /* shared object file */
#define ET_CORE     4         /* core file */
#define ET_LOPROC   0xff00    /* processor-specific */
#define ET_HIPROC   0xffff    /* processor-specific */

/* Required architecture in e_machine field of ELF header */
#define EM_NONE     0         /* no machine */
#define EM_M32      1         /* AT&T WE 32100 */
#define EM_SPARC    2         /* SPARC */
#define EM_386      3         /* Intel 80386 */
#define EM_68K      4         /* Motorola 68000 */
#define EM_88K      5         /* Motorola 88000 */
#define EM_860      7         /* Intel 80860 */
#define EM_MIPS     8         /* MIPS RS3000 */
#define EM_ARM      40        /* Advanced RISC Machines ARM */

/* Object file version in e_version field of ELF header */
#define EV_NONE    0          /* invalid version */
#define EV_CURRENT 1          /* current version */

/* Processor-specific values in e_flags field of ELF header */
#define EF_MIPS_NOREORDER   0x00000001  /* .noreorder directive in source  */
#define EF_MIPS_PIC         0x00000002  /* position independent code       */
#define EF_MIPS_CPIC        0x00000004  /* 'standard' PIC calling sequence */
#define EF_MIPS_ABI2        0x00000020  /* new ABI on Irix6                */
#define EF_MIPS_ARCH_1      0x00000000  /* -mips1 instruction set          */
#define EF_MIPS_ARCH_2      0x10000000  /* -mips2 instruction set          */
#define EF_MIPS_ARCH_3      0x20000000  /* -mips3 instruction set          */
#define EF_MIPS_ARCH_4      0x30000000  /* -mips4 instruction set          */

#define EF_ARM_RELEXEC      0x00000001
#define EF_ARM_HASENTRY     0x00000002
#define EF_ARM_INTERWORK    0x00000004
#define EF_ARM_APCS_26      0x00000008
#define EF_ARM_APCS_FLOAT   0x00000010
#define EF_ARM_PIC          0x00000020
#define EF_ARM_ALIGN8       0x00000040
#define EF_ARM_NEW_ABI      0x00000080
#define EF_ARM_OLD_ABI      0x00000100

/* File header for ELF object file format */
typedef struct {
    UINT8 e_ident[EI_NIDENT];  /* identify file as object file */
    Elf32_Half    e_type;        /* file type */
    Elf32_Half    e_machine;     /* required architecture */
    Elf32_Word    e_version;     /* object file version */
    Elf32_Addr    e_entry;       /* entry point address */
    Elf32_Off     e_phoff;       /* program header table's file offset */
    Elf32_Off     e_shoff;       /* section header table's file offset */
    Elf32_Word    e_flags;       /* processor-specific flags */
    Elf32_Half    e_ehsize;      /* ELF header's size in bytes */
    Elf32_Half    e_phentsize;   /* size of entry in program header table */
    Elf32_Half    e_phnum;       /* no. of entries in program header table */
    Elf32_Half    e_shentsize;   /* size of entry in section header table */
    Elf32_Half    e_shnum;       /* no. of entries in section header table */
    Elf32_Half    e_shstrndx;    /* section hdr index of string table */
} Elf32_Ehdr;

/* Special section indexes (reserved values) */
#define SHN_UNDEF          0     /* undefined section index */
#define SHN_LORESERVE 0xff00     /* LB (lower bound) of ELF reserved indexes */
#define SHN_LOPROC    0xff00     /* LB of processor-specific semantics */
#define SHN_HIPROC    0xff1f     /* UB of processor-specific semantics */
#define SHN_ABS       0xfff1     /* absolute (not relocatable) section */
#define SHN_COMMON    0xfff2     /* C external variables section */
#define SHN_HIRESERVE 0xffff     /* UB (upper bound) of ELF reserved indexes */

/* Section's semantics in sh_type field of ELF section header */
#define SHT_NULL            0    /* no section associated with header */
#define SHT_PROGBITS        1    /* program-defined data */
#define SHT_SYMTAB          2    /* link editing & dynamic linking symbols */
#define SHT_STRTAB          3    /* string table */
#define SHT_RELA            4    /* minimal set of dynamic linking symbols */
#define SHT_HASH            5    /* relocation entries with explicit addends */
#define SHT_DYNAMIC         6    /* symbol hash table (dynamic linking) */
#define SHT_NOTE            7    /* dynamic linking info */
#define SHT_NOBITS          8    /* file-marking information */
#define SHT_REL             9    /* relocation entries without explicit addends */
#define SHT_SHLIB          10    /* reserved */
#define SHT_DYNSYM         11    /* dynamic linking symbol table */
#define SHT_LOPROC 0x70000000    /* LB for processor-specific dynamics */
#define SHT_HIPROC 0x7fffffff    /* UB for processor-specific dynamics */
#define SHT_LOUSER 0x80000000    /* LB for application-specific dynamics */
#define SHT_HIUSER 0x8fffffff    /* UB for application-specific dynamics */

/* Section's attribute flags in sh_flags field of ELF section header */
#define SHF_WRITE             0x1    /* data writable during execution */
#define SHF_ALLOC             0x2    /* occupies memory during execution */
#define SHF_EXECINSTR         0x4    /* executable machine instructions */
#define SHF_MASKPROC   0xf0000000    /* mask for processor-specific semantics */

/* Section header for ELF object file format */
typedef struct {
    Elf32_Word    sh_name;       /* section name (string table index) */
    Elf32_Word    sh_type;       /* section's contents and semantics */
    Elf32_Word    sh_flags;      /* section attribute flag bits */
    Elf32_Addr    sh_addr;       /* address of section's first byte */
    Elf32_Off     sh_offset;     /* file offset to 1st byte in section */
    Elf32_Word    sh_size;       /* section's size in bytes */
    Elf32_Word    sh_link;       /* section header table index link */
    Elf32_Word    sh_info;       /* additional section information */
    Elf32_Word    sh_addralign;  /* address alignment constraints */
    Elf32_Word    sh_entsize;    /* gives size if entry size is fixed */
} Elf32_Shdr;

/* Undefined symbol in symbol (string) table */
#define STN_UNDEF   0

/* Manipulating symbol's type and binding attributes in st_info field */
#define ELF32_ST_BIND(i)     ((i)>>4)
#define ELF32_ST_TYPE(i)     ((i)&0xf)
#define ELF32_ST_INFO(b,t)   (((b)<<4)+((t)&0xf))

/* Symbol binding for ELF32_ST_BIND macro above */
#define STB_LOCAL    0    /* local symbols (scope = source module) */
#define STB_GLOBAL   1    /* global symbols (visible to all) */
#define STB_WEAK     3    /* weak symbols (multiply-defined globals) */
#define STB_LOPROC  13    /* LB for processor-specific semantics */
#define STB_HIPROC  15    /* UB for processor-specific semantics */

/* Symbol binding for ELF32_ST_TYPE macro above */
#define STT_NOTYPE   0    /* symbol's type is unspecified */
#define STT_OBJECT   1    /* symbol for data object (variable, array, etc.) */
#define STT_FUNC     2    /* symbol for function or other executable code */
#define STT_SECTION  3    /* symbol for section */
#define STT_FILE     4    /* name of source file corresp. to this obj. file */
#define STT_LOPROC  13    /* LB for processor-specific semantics */
#define STT_HIPROC  15    /* UB for processor-specific semantics */

/* Symbol table entry in ELF object file */
typedef struct {
    Elf32_Word    st_name;       /* index into symbol string table */
    Elf32_Addr    st_value;      /* value of associated symbol */
    Elf32_Word    st_size;       /* size of object */
    UINT8 st_info;       /* symbol's type and binding attributes */
    UINT8 st_other;      /* =0 (no defined meaning) */
    Elf32_Half    st_shndx;      /* relevant section header table index */
} Elf32_Sym;

/* Manipulating relocation entry's type and binding attributes in r_info field */
#define ELF32_R_SYM(i)     ((i)>>8)
#define ELF32_R_TYPE(i)     ((UINT8)(i))
#define ELF32_R_INFO(b,t)   (((b)<<8)+((UINT8)(t)))

/* Or-ed in to relocation type */
#define R_DLL_TAG   0x80
#define IsaDllReloc(_t_)  (((_t_)&0xe0)==R_DLL_TAG)
#define MaskOffDllTag(_t_) ((_t_)&(R_DLL_TAG-1))
#define AddDllTag(_t_)    ((_t_)|R_DLL_TAG)


/* Relocation entry structures (two types) in ELF object file */
typedef struct {
    Elf32_Addr    r_offset;      /* address offset from section base */
    Elf32_Word    r_info;        /* symbol table index and type */
} Elf32_Rel;

typedef struct {
    Elf32_Addr    r_offset;      /* address offset from section base */
    Elf32_Word    r_info;        /* symbol table index and type */
    Elf32_Sword   r_addend;      /* constant addend to relocatable field */
} Elf32_Rela;

/* Segment type values in p_type field of ELF program header */
#define PT_NULL             0    /* ignore entry */
#define PT_LOAD             1    /* loadable segment */
#define PT_DYNAMIC          2    /* dynamic linking info */
#define PT_INTERP           3    /* interpreter pathname */
#define PT_NOTE             4    /* location and size of auxiliary info */
#define PT_SHLIB            6    /* reserved (not used) */
#define PT_PHDR             7    /* location & size of program header table */
#define PT_LOPROC  0x70000000    /* LB processor specific semantics */
#define PT_HIPROC  0x7fffffff    /* UB processor specific semantics */

#define PT_MIPS_REGINFO 0x70000000 /* register usage information */

#define PTF_IMPORT          1
#define PTF_EXPORT          2
#define PTF_SCRAM           4

/* Program header in ELF object file */
typedef struct {
    Elf32_Word    p_type;       /* kind of segment */
    Elf32_Off     p_offset;     /* file offset of segment */
    Elf32_Addr    p_vaddr;      /* virtual address of segment */
    Elf32_Addr    p_paddr;      /* physical address of segment */
    Elf32_Word    p_filesz;     /* file image size */
    Elf32_Word    p_memsz;      /* memory image size */
    Elf32_Word    p_flags;      /* flags for segment */
    Elf32_Word    p_align;      /* segment alignment */
} Elf32_Phdr;

/* Dynamic array tag values in d_tag field of ELF dynamic structure */
#define DT_NULL             0    /* ignore d_un */
#define DT_NEEDED           1    /* marks end of dynamic array */
#define DT_PLTRELSZ         2    /* total size of relation entries */
#define DT_PLTGOT           3    /* procedure linkage/global offset tbl */
#define DT_HASH             4    /* symbol hash table */
#define DT_STRTAB           5    /* address of string table address */
#define DT_SYMTAB           6    /* address of symbol table */
#define DT_RELA             7    /* address of relocation table */
#define DT_RELASZ           8    /* total size of DT_RELA reloc table */
#define DT_RELAENT          9    /* size of DT_RELA relocation entry */
#define DT_STRSZ           10    /* size of string table */
#define DT_SYMENT          11    /* size of symbol table entry */
#define DT_INIT            12    /* address of initialization function */
#define DT_FINI            13    /* address of termination function */
#define DT_SONAME          14    /* string tbl offset of sh. obj. name */
#define DT_RPATH           15    /* string tbl offset of search path */
#define DT_SYMBOLIC        16    /* start search at shared object */
#define DT_REL             17    /* like DT_RELA, but with addends */
#define DT_RELSZ           18    /* size of DT_REL relocation table */
#define DT_RELENT          19    /* size of DT_REL relocation entry */
#define DT_PLTREL          20    /* size reloc entry from proc link table */
#define DT_DEBUG           21    /* used for debugging */
#define DT_TEXTREL         22    /* mods to nonwritable segment */
#define DT_JMPREL          23    /* procedure linkage table reloc entries */
#define DT_LOPROC  0X70000000    /* LB for processor-specific semantics */
#define DT_HIPROC  0X7FFFFFFF    /* UB for processor-specific semantics */


/* MIPS-related relocation types
 */
#define R_MIPS_16               0x01
#define R_MIPS_32               0x02
#define R_MIPS_REL32            0x03
#define R_MIPS_26               0x04
#define R_MIPS_HI16             0x05
#define R_MIPS_LO16             0x06
#define R_MIPS_GPREL16          0x07
#define R_MIPS_LITERAL          0x08
#define R_MIPS_GOT16            0x09
#define R_MIPS_PC16             0x0a
#define R_MIPS_CALL16           0x0b
#define R_MIPS_GPREL32          0x0c

/* ARM-related relocation types (with GCC extensions)
 */
#define R_ARM_NONE              0x00
#define R_ARM_PC24              0x01
#define R_ARM_ABS32             0x02
#define R_ARM_REL32             0x03
#define R_ARM_PC13              0x04
#define R_ARM_ABS16             0x05
#define R_ARM_ABS12             0x06
#define R_ARM_THM_ABS5          0x07
#define R_ARM_ABS8              0x08
#define R_ARM_SBREL32           0x09
#define R_ARM_THM_PC22          0x0a
#define R_ARM_THM_PC8           0x0b
#define R_ARM_AMP_VCALL9        0x0c
#define R_ARM_SWI24             0x0d
#define R_ARM_THM_SWI8          0x0e
#define R_ARM_XPC25             0x0f
#define R_ARM_THM_XPC22         0x10
#define R_ARM_COPY              0x14       /* copy symbol at runtime */
#define R_ARM_GLOB_DAT          0x15       /* create GOT entry */
#define R_ARM_JUMP_SLOT         0x16       /* create PLT entry */
#define R_ARM_RELATIVE          0x17       /* adjust by program base */
#define R_ARM_GOTOFF            0x18       /* 32 bit offset to GOT */
#define R_ARM_GOTPC             0x19       /* 32 bit PC relative offset to GOT */
#define R_ARM_GOT32             0x1a       /* 32 bit GOT entry */
#define R_ARM_PLT32             0x1b       /* 32 bit PLT address */
#define R_ARM_GNU_VTENTRY       0x64
#define R_ARM_GNU_VTINHERIT     0x65
#define R_ARM_THM_PC11          0x66       /* cygnus extension to abi: thumb unconditional branch */
#define R_ARM_THM_PC9           0x67       /* cygnus extension to abi: thumb conditional branch */
#define R_ARM_RXPC25            0xf9
#define R_ARM_RSBREL32          0xfa
#define R_ARM_THM_RPC22         0xfb
#define R_ARM_RREL32            0xfc
#define R_ARM_RABS32            0xfd
#define R_ARM_RPC24             0xfe
#define R_ARM_RBASE             0xff

#if 0
/* betatest Metaware compiler */
#define R_ARM_NONE              0x0     /* None */
#define R_ARM_ADDR32            0x1     /* 32-bit address  S+A */
#define R_ARM_REL24             0x2     /* 26-bit rel address (S+A-P)>>2 */
#define R_ARM_ADDR8             0x3     /* 8-bit address  S+A */
#define R_ARM_ADDR16            0x4     /* 16-bit address S+A */
#define R_ARM_SDT_OFFSET        0x5     /* S+A */
#define R_ARM_CDT_OFFSET        0x6     /* S+A */
#define R_ARM_SDT_PCREL_OFFSET  0x7     /* S+A-P-8 */
#define R_ARM_CDT_PCREL_OFFSET  0x8     /* S+A-P-8 */
#define R_ARM_ADR               0x9     /* S+A-P-8 */
#define R_ARM_ADRL              0xa     /* S+A-P-8 */
#define R_ARM_SHIFT_IMM         0xb     /* S+A */
#endif

/* Dynamic structure in ELF object file */
typedef struct {
    Elf32_Sword     d_tag;  /* how to interpret d_un */
    union {
        Elf32_Word    d_val;   /* integer value */
        Elf32_Addr    d_ptr;   /* virtual address */
    } d_un;
} Elf32_Dyn;

/* Our extensions
 */
typedef struct {
    UINT8 ih_name[16];         /* name of the DLL we IMPORT from */
    UINT8 ih_uuid[16];         /* UUID of the CInterface         */
} Elf32_Ihdr;

/* For now the .export section has no header
 */

/* Processor-specific functions
 */
typedef struct _ARCHOPS {
  Elf32_Half Machine;
  BOOL (*RelocateEntry)(BOOL MustSwap, int RelaType,
                        Elf32_Addr A, Elf32_Addr PC, char * P);
  void (*Disasm)( unsigned iadr, void *pinst, int runtaddr);
  NAME (*RelName)(UINT t);
  Elf32_Off (*RemapReloc)(int RelaType, Elf32_Off Val);
} ARCHOPS;


/* Linkedit state for an image
 */
typedef struct {
  Elf32_Ehdr Elf;
#define NMYPH 5
#define PGH_BITS 0
#define PGH_DYNA 1
#define PGH_IMPT 2
#define PGH_EXPT 3
#define PGH_SCRAM 4
  Elf32_Phdr Ph[NMYPH];
#define MYDYN 4
  Elf32_Dyn  Dyn[MYDYN];
  Elf32_Shdr Sh0;   /* silly zero section */
  Elf32_Shdr Sh[1]; /* varsize */
} MYHEADERS;

typedef struct {
  Elf32_Shdr *pShdr;
  char *SecBits;
  Elf32_Addr SecAddr;
  Elf32_Off SecOffLimit;
  Elf32_Off SecOffDelta;
  int SecWas;
} SECMAP;

typedef struct {
  Elf32_Addr r_offset;
  char *Name;
  UINT16 r_sym;
  UINT8 r_type;
  UINT8 r_section;
}RELINFO;
/* keep these two guys in synch for sharing of search code */
typedef struct {
  Elf32_Addr Value;
  char *Name;
  UINT16 Size;
  UINT8 Binding;
  UINT8 Type;
} SYMINFO;

typedef struct {
  UINT nSymbols;
  UINT Hint;
  SYMINFO *pSymbols;
  Elf32_Addr SecBase;
} SPLITINFO, *PSPLITINFO;

typedef struct {
  SPLITINFO CodeRelocs;
  UINT nSections;
  SPLITINFO Secinfo[1];/*varsize*/
} SPLIT, *PSPLIT;

typedef struct {
  ARCHOPS *v;
  char *ImageName;
  char *FileBase;
  Elf32_Addr NewLoadAddress;
  Elf32_Off  NewLoadOffset;
  Elf32_Addr OldLoadAddress;
  Elf32_Off  OldLoadOffset;
  Elf32_Ehdr *pElf;
  UINT ElfSize;
  SECMAP *pSectionMap;
  UINT nSectionBits;
  BOOL MustSwap;
  UINT BitsSize;
  PRTL *pRtl;
  UINT nRtl;
  PRTL pExportRtl;
  Elf32_Shdr *pExportSection;
  Elf32_Off ExportAdjustment;
  Elf32_Rela *pRela;
  UINT RelaSize;
  Elf32_Shdr *pImport;
  UINT ImportSize;
  MYHEADERS *pHdr;

  /* Disassembler state */
  PSPLIT pSplit;
  int    CodeIndex;
  SPLITINFO *Codinfo; /* cache == &pSplit->Secinfo[CodeIndex] */
} IMAGE, *PIMAGE;

/* We use this also to hold all headers */
#define FILEOFF 0x1000

/* Utilities
 */
UINT32 SwapWord(UINT32 w)
{
   return (((w & 0xff) << 24) |
           ((w & 0xff00) << 8) |
           ((w & 0xff0000) >> 8) |
           ((w & 0xff000000) >> 24));
}

UINT16 SwapHalf(UINT16 h)
{
   return (((h & 0xff) << 8) |
           ((h & 0xff00) >> 8));
}

#define GetWord(_swap_,_w_)  ((_swap_) ? SwapWord(_w_) : (_w_))
#define GetHalf(_swap_,_h_)  ((_swap_) ? SwapHalf(_h_) : (_h_))
#define SetWord(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapWord(_v_) : (_v_))
#define SetHalf(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapHalf(_v_) : (_v_))


char *STypeName(UINT t);

/* Disassembler support stuff
 */
PIMAGE      TheImage;

#define TANY 0xff

SYMINFO *srchsec(SPLITINFO *pSrch, Elf32_Addr Value, UINT8 Type, UINT8 Bind)
{
  int i, j, k = -1;
  SYMINFO *pI;

  i  = pSrch->Hint;
  j  = pSrch->nSymbols;
  pI = pSrch->pSymbols;

  if (pI == NULL)
      return NULL;

  /* Make sure we are close by
   */
  while ((pI[i].Value >= Value) && (i > 0))
    i--;
  while ((pI[i].Value < Value) && (i < j))
    i++;
  /* Hit ?
   */
  while ((i < j) && (pI[i].Value == Value)) {
    k = i;
    /* But we want the right one.. */
    if (pI[i].Type == Type)
      break;
    if ((Bind != TANY) && (pI[i].Binding == Bind))
      break;
    i++;
  }
  if (k == -1)
    return NULL;
  pSrch->Hint = k;
  return pI+k;
}

void printaddr(Elf32_Addr Offset, Elf32_Addr Address)
{
  SYMINFO *pI = NULL;

  if (TheImage) {
    pI = srchsec(TheImage->Codinfo,Offset,STT_FUNC,TANY);
    if (pI == NULL)
        pI = srchsec(TheImage->Codinfo,Address,STT_FUNC,TANY);
  }

  if (pI)
    printf("%s:\n",pI->Name);
}

void printlabel(Elf32_Addr LabValue, Elf32_Off LabOffset)
{
  printf("%x",LabValue);
  if (TheImage) {
    SYMINFO *pI;
    pI = srchsec(TheImage->Codinfo,LabOffset,STT_NOTYPE,STB_LOCAL);
    if (pI == NULL)
        pI = srchsec(TheImage->Codinfo,LabValue,STT_NOTYPE,STB_LOCAL);
    if (pI)
      printf(" <%s>", pI->Name);
  }
}

void printsym(Elf32_Off SymVal, Elf32_Addr InstOff, BOOL PrintVal)
{
  RELINFO *pR;
  SYMINFO *pI;
  SPLITINFO *pS;
  char *Name = NULL;

  if (PrintVal)
    printf("%x",SymVal);

  if (TheImage) {
    pR = (RELINFO*)srchsec(&TheImage->pSplit->CodeRelocs,InstOff,TANY,TANY);
    if (pR == NULL)
        pR = (RELINFO*)srchsec(TheImage->Codinfo,SymVal,STT_FUNC,TANY);
    if (pR) {
      /* Dont print if its a MIPS HI16.
       */
      if (!((TheImage->v->Machine == EM_MIPS) &&
            (pR->r_type == R_MIPS_HI16))) {
        Name = pR->Name;
        if (Name == NULL) {
          /* bsearch() is probably a better idea here
           */
          pS = &TheImage->pSplit->Secinfo[pR->r_section];
          if (SymVal > pS->SecBase)
              SymVal -= pS->SecBase;
          else {
              if (SymVal > TheImage->NewLoadAddress)
                  SymVal -= TheImage->NewLoadAddress;
          }
          pI = srchsec(pS, SymVal,TANY,STB_GLOBAL);
          if (pI == NULL) {
            pI = srchsec(pS, SymVal,TANY,STB_LOCAL);
          }
          if (pI)
            Name = pI->Name;
        }
      }
    }
  }
  if (Name)
    printf(" <%s>",Name);

#if 0 //needs work
  Info = GetWord(pImg->MustSwap,TheSymbols[Sym].st_value);
  if (Info != SymVal) {
    INT32 diff = SymVal - Info;
    char Sign = '+';
    if (diff < 0) {
      Sign = '-';
      diff = -diff;
    }
    printf("%c%x",Sign,diff);
  }
#endif
}

/* Disassemblers proper
 */
extern void
ArmDisasm(
       unsigned iadr,
       void *pinst,
       int runtaddr);

extern void
MipsDisasm(
       unsigned iadr,
       void *pinst,
       int runtaddr);

void
NoDisasm(
       unsigned iadr,
       void *pinst,
       int runtaddr)
{
}

/* Knows about the split
 */
void PrintSortedFunctions(PIMAGE pImg, char *FileName)
{
    int nSec, SecWas, i;
    PSPLITINFO p;
    FILE *f;
    Elf32_Addr SecBase;

    if (TheImage != pImg) {
        printf("I'm sooo confused.\n");
        exit(1);
    }

    f = fopen(FileName,"w");
    if (f == NULL) {
        printf("Could not (re)create %s for writing\n", FileName);
        return;
    }

    /* This is very messy because of the section remapping mistake
     */
    for (nSec = 0; nSec < pImg->pSplit->nSections; nSec++) {

#if 0
        fprintf(f,"SecWas no.%d for %d\n",
                pImg->pSectionMap[nSec].SecWas, nSec);
#endif        
        /* Was it allocated
         */
        p = &pImg->pSplit->Secinfo[nSec];
        if (p->SecBase == 0)
            continue;
        SecBase = p->SecBase;

        /* Where did it go
         */
        SecWas = pImg->pSectionMap[nSec].SecWas;
        if (SecWas < 0)
            continue;
        
        /* Does it have any symbols
         */
        p = &pImg->pSplit->Secinfo[SecWas];
        if (p->nSymbols == 0)
            continue;
#if 0
        fprintf(f,"Sorted symbols for section no.%d [aka %d], SecBase = %x\n",
                nSec, SecWas, SecBase);
#endif
        for (i = 0; i < p->nSymbols; i++) {
            if ((p->pSymbols[i].Name[0]) &&
                ((p->pSymbols[i].Type == STT_FUNC) ||
                 (p->pSymbols[i].Binding == STB_GLOBAL))
                )
                fprintf(f,
                    "%8.8x %c %s\n",
                    p->pSymbols[i].Value + SecBase,
                    STypeName(p->pSymbols[i].Type)[0],
                    p->pSymbols[i].Name);
        }
    }
    fclose(f);
}

/* Other printing related functions
 */
char *PgmTypeName(UINT t)
{
  static char bf[128];
  switch (t) {
  case PT_NULL: return "PT_NULL";
  case PT_LOAD: return "PT_LOAD";
  case PT_DYNAMIC: return "PT_DYNAMIC";
  case PT_INTERP: return "PT_INTERP";
  case PT_NOTE: return "PT_NOTE";
  case PT_SHLIB: return "PT_SHLIB";
  case PT_PHDR: return "PT_PHDR";
  default: 
    sprintf(bf, "%s x%x", ((t >= PT_LOPROC) && (t < PT_HIPROC)) ?
            "ProcessorSpecific PgmHeader Type" :
            "?Unknown PgmHeader Type?",
            t);
    return bf;
  }
}

char *ScnTypeName(UINT t)
{
  static char bf[128];
  switch (t) {
  case SHT_NULL: return "SHT_NULL";
  case SHT_PROGBITS: return "SHT_PROGBITS";
  case SHT_SYMTAB: return "SHT_SYMTAB";
  case SHT_STRTAB: return "SHT_STRTAB";
  case SHT_RELA: return "SHT_RELA";
  case SHT_HASH: return "SHT_HASH";
  case SHT_DYNAMIC: return "SHT_DYNAMIC";
  case SHT_NOTE: return "SHT_NOTE";
  case SHT_NOBITS: return "SHT_NOBITS";
  case SHT_REL: return "SHT_REL";
  case SHT_SHLIB: return "SHT_SHLIB";
  case SHT_DYNSYM: return "SHT_DYNSYM";
  default: 
    sprintf(bf, "%s x%x",
            ((t >= SHT_LOPROC) && (t < SHT_HIPROC)) ?
            "ProcessorSpecific Section Type" :
            (((t >= SHT_LOUSER) && (t < SHT_HIUSER)) ?
            "ApplicationSpecific Section Type" :
             "?Unknown Section Type?"),
            t);
    return bf;
  }
}

char SBindName(UINT b)
{
    switch (b) {
    case STB_LOCAL: return 'l';
    case STB_GLOBAL: return 'g';
    case STB_WEAK: return 'w';
    default: if (b < 10) return '0'+b;
        else return 'A'+b;
    }
}

char *STypeName(UINT t)
{
    switch (t) {
    case STT_NOTYPE: return "NoType";
    case STT_OBJECT: return "Object";
    case STT_FUNC: return "Function";
    case STT_SECTION: return "Section";
    case STT_FILE: return "File";
    default: return "UnkType";
    }
}

char *SectionName(PIMAGE pImg, UINT NameIx, UINT Ix)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    char *StringTable;
    Elf32_Shdr *pShdr;
    UINT t;
    static char bf[128];

    t = GetHalf(MustSwap,pElf->e_shstrndx);
    if (t == SHN_UNDEF)
      goto Numeric;

    /* Point to the proper section header and sanity check it
     */
    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    pShdr += t;
    if (GetWord(MustSwap,pShdr->sh_type) != SHT_STRTAB)
      goto Numeric;

    StringTable = pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset);
    return &StringTable[NameIx];

Numeric:
    sprintf(bf,"%d",Ix);
    return bf;
}


char *FindStringSection(PIMAGE pImg)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    Elf32_Shdr *pShdr;
    UINT i,t,tag;

    t = GetHalf(MustSwap,pElf->e_shnum);
    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    for (i = 0; i < t; ++i, ++pShdr) {
        tag = GetWord(MustSwap,pShdr->sh_type);
        if ((tag == SHT_STRTAB) &&
            (memcmp(SectionName(pImg,
                                GetWord(MustSwap,pShdr->sh_name),
                                i),
                    ".strtab",
                    8) == 0)) {
            return pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset);
        }
    }
    return NULL;
}

void PrintElfHeader(BOOL MustSwap, Elf32_Ehdr *pElf)
{
    int i;

    printf("ElfHeader @%x: ", pElf);

    for (i = 0; i < EI_NIDENT;i++)
        printf("%2x ", pElf->e_ident[i]);

    printf("\n  type=%x machine=%x vers=x%x (%d) entry=%x phoff=%x shoff=%x flg=%x\n",
           GetHalf(MustSwap,pElf->e_type),
           GetHalf(MustSwap,pElf->e_machine),
           GetWord(MustSwap,pElf->e_version),
           GetWord(MustSwap,pElf->e_version),
           GetWord(MustSwap,pElf->e_entry),
           GetWord(MustSwap,pElf->e_phoff),
           GetWord(MustSwap,pElf->e_shoff),
           GetWord(MustSwap,pElf->e_flags));

    printf("  ehsize=%x phsize=%x phnum=%x shsize=%x shnum=%x six=%x\n",
           GetHalf(MustSwap,pElf->e_ehsize),
           GetHalf(MustSwap,pElf->e_phentsize),
           GetHalf(MustSwap,pElf->e_phnum),
           GetHalf(MustSwap,pElf->e_shentsize),
           GetHalf(MustSwap,pElf->e_shnum),
           GetHalf(MustSwap,pElf->e_shstrndx));

}

NAME UnkRel = "UNKNOWN_RELOCATION_TYPE";

NAME MipsRelName(UINT t)
{
    static NAME RelName[] = {
        "?0?", "16", "32", "REL32",
        "26","HI16","LO16", "GPREL16",
        "LITERAL", "GOT16", "PC16", "CALL16",
        "GPREL32"
    };

    return (t < (sizeof RelName / sizeof RelName[0])) ?
        RelName[t] : UnkRel;
}

NAME ArmRelName(UINT t)
{
    static NAME RelName0[] = {
        "NONE", "PC24", "ABS32", "REL32",
        "PC13", "ABS16", "ABS12", "THM_ABS5",
        "ABS8", "SBREL32", "THM_PC22", "THM_PC8",
        "AMP_VCALL9", "SWI24", "THM_SWI8", "XPC25",
        "THM_XPC22", "COPY", "GLOB_DAT", "JUMP_SLOT",
        "RELATIVE", "GOTOFF", "GOTPC", "GOT32",
        "PLT32"
    };
    static NAME RelName1[] = {
        "GNU_VTENTRY", "GNU_VTINHERIT", "THM_PC11", "THM_PC9"
    };
    static NAME RelName2[] = { /* grows backward */
        "RXPC25", "RSBREL32", "THM_RPC22",
        "RREL32", "RABS32", "RPC24", "RBASE"
    };

    if (t < (sizeof RelName0 / sizeof RelName0[0]))
        return RelName0[t];
    if ((t >= R_ARM_GNU_VTENTRY) && (t <= R_ARM_THM_PC9))
        return RelName1[t-R_ARM_GNU_VTENTRY];
    if ((t >= R_ARM_RXPC25) && (t <= R_ARM_RBASE))
        return RelName2[t-R_ARM_RXPC25];

    return UnkRel;

}

NAME NoRelName(UINT t)
{
    static char xxx[24];
    sprintf(xxx,"x%x",t);
    return xxx;
}

void PrintRelas(PIMAGE pImg, Elf32_Shdr *pShdr)
{
    BOOL MustSwap = pImg->MustSwap;
    UINT i, t, s, ns, PgmNum, SymNum;
    BOOL Code;
    Elf32_Rela *pRela;
    Elf32_Phdr *pPgmHdr;
    char *SecBase;

    t = GetWord(MustSwap,pShdr->sh_size);
    ns = t / sizeof *pRela;

    /* Our relas are funny, see if we actually got a pgh header
     */
    i = GetWord(MustSwap,pImg->pElf->e_phoff);
    PgmNum = GetWord(MustSwap,pShdr->sh_info);
    if (i) {
      /* yes, its our stuff
       */
      pPgmHdr = (Elf32_Phdr *)(pImg->FileBase + i);
      pPgmHdr += PgmNum;
      SecBase = pImg->FileBase + GetWord(MustSwap,pPgmHdr->p_offset);
      /* This should be null */
      SymNum = GetWord(MustSwap,pShdr->sh_link);
           
    } else {
      /* BFD stuff
       */
      SecBase = pImg->pSectionMap[PgmNum].SecBits;
    }

    printf(" Found %d(x%x) Relas [for pgh %d at %x]:\n",
           ns, t, PgmNum, SecBase);

    pRela = (Elf32_Rela *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    for (i = 0; i < ns; i++, pRela++) {
        Elf32_Addr Off = GetWord(MustSwap,pRela->r_offset);
        Elf32_Word Info = GetWord(MustSwap,pRela->r_info);
        Elf32_Sword Add = GetWord(MustSwap,pRela->r_addend);
        UINT32 Val;
        char *pfx;


        t = ELF32_R_TYPE(Info);
        s = ELF32_R_SYM(Info);
        pfx = (IsaDllReloc(t)) ? "DLL+" : "";
        t = MaskOffDllTag(t);

        printf("  [%d] Off=%x Sym=%x Rel=%s%s Add=%x", i,
                   Off,s,pfx,pImg->v->RelName(t),Add);
               
        Val = *(UINT *)(SecBase + Off);
        Val = GetWord(MustSwap,Val);
        printf(" Val=%x ", Val);

#if 0
        Code = <compare offset against text section>;
#else
        Code = TRUE;
#endif
        if (Code) {
            pImg->v->Disasm(Off,&Val,Off);
        }
        printf("\n");
    }

}
void PrintRels(PIMAGE pImg, Elf32_Shdr *pShdr)
{
    BOOL MustSwap = pImg->MustSwap;
    UINT i, t, s, ns, SecNum, SymNum;
    BOOL Code;
    Elf32_Rel *pRel;
    Elf32_Sym *pSym;
    Elf32_Shdr *pSecHdr, *pSymHdr;
    char *SecBase, *SymBase;

    t = GetWord(MustSwap,pShdr->sh_size);
    ns = t / sizeof *pRel;
    SecNum = GetWord(MustSwap,pShdr->sh_info);
    SymNum = GetWord(MustSwap,pShdr->sh_link);
           
    pSecHdr = (Elf32_Shdr *)(pImg->FileBase +
                             GetWord(MustSwap,pImg->pElf->e_shoff));
    pSymHdr = pSecHdr + SymNum;
    pSecHdr = pSecHdr + SecNum;
    SecBase = pImg->FileBase + GetWord(MustSwap,pSecHdr->sh_offset);
    SymBase = pImg->FileBase + GetWord(MustSwap,pSymHdr->sh_offset);

    Code = (GetWord(MustSwap,pSecHdr->sh_flags) & SHF_EXECINSTR);

    printf(" Found %d(x%x) Rels for scn %d @%x, symbols in Scn %d @%x :\n",
           ns, t, SecNum, SecBase, SymNum, SymBase);

    pRel = (Elf32_Rel *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    pSym = (Elf32_Sym *)SymBase;
    for (i = 0; i < ns; i++, pRel++) {
        UINT Off = GetWord(MustSwap,pRel->r_offset);
        UINT Info = GetWord(MustSwap,pRel->r_info);
        UINT32 Val, Six;


        t = ELF32_R_TYPE(Info);
        s = ELF32_R_SYM(Info);

        printf("  [%d] Off=%x Sym=%x Rel=%s", i, Off,
                 s,pImg->v->RelName(t));
               
        Val = *(UINT *)(SecBase + Off);
        Val = GetWord(MustSwap,Val);
        printf(" Val=%x", Val);
               
        Six = GetWord(MustSwap,pSym[s].st_name);
        printf(" St=[%x %x %x %x %x %x]",
                   Six,
                   GetWord(MustSwap,pSym[s].st_value),
                   GetWord(MustSwap,pSym[s].st_size),
                   pSym[s].st_info, pSym[s].st_other,
                   GetHalf(MustSwap,pSym[s].st_shndx));

        if (Code) {
            pImg->v->Disasm(Off,&Val,Off);
        }
        printf("\n");
    }

}

void PrintAdsRel(PIMAGE pImg, Elf32_Word Offset, UINT Count)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Rel *pRel;
    UINT i, t, s;

    printf("Ads %d REL at x%x\n", Count, Offset);

    pRel = (Elf32_Rel *)(pImg->FileBase + Offset);
    for (i = 0; i < Count; i++, pRel++) {
        UINT Off = GetWord(MustSwap,pRel->r_offset);
        UINT Info = GetWord(MustSwap,pRel->r_info);
        UINT32 Val, Six;


        t = ELF32_R_TYPE(Info);
        s = ELF32_R_SYM(Info);

        printf("  [%d] Off=%x Info=%x {Sym=%x Rel=%x %s}", i, Off, Info,
                 s,t,pImg->v->RelName(t));

        printf("\n");
    }
}

void PrintDt(Elf32_Word t, Elf32_Word v)
{
    static NAME DtName[] = {
        "DT_NULL", "DT_NEEDED","DT_PLTRELSZ","DT_PLTGOT","DT_HASH",
        "DT_STRTAB","DT_SYMTAB","DT_RELA","DT_RELASZ","DT_RELAENT",
        "DT_STRSZ","DT_SYMENT","DT_INIT","DT_FINI","DT_SONAME",
        "DT_RPATH","DT_SYMBOLIC","DT_REL","DT_RELSZ","DT_RELENT",
        "DT_PLTREL","DT_DEBUG","DT_TEXTREL","DT_JMPREL"
    };

    if (t <= DT_JMPREL)
        printf("%s=x%x\n", DtName[t], v);
    else
        printf("DT_x%x=x%x\n", t, v);
}

void PrintDynamic(PIMAGE pImg, Elf32_Phdr *pPhdr)
{
    Elf32_Dyn *pDyn, *pEnd;
    Elf32_Word t;
    UINT DynOffset, RelOffset = 0, Count = 0, RelSize = 0;
    BOOL MustSwap = pImg->MustSwap;
    char *Str = NULL, *DynBits;
    UINT StrSize = 0;

    DynOffset = GetWord(MustSwap,pPhdr->p_offset);
    DynBits = pImg->FileBase+ DynOffset;

    pDyn = (Elf32_Dyn *)DynBits;
    pEnd = (Elf32_Dyn *)(DynBits + GetWord(MustSwap,pPhdr->p_filesz));

    t = GetWord(MustSwap,pDyn->d_tag);
    while (t != DT_NULL) {
        PrintDt(t,GetWord(MustSwap,pDyn->d_un.d_val));
        switch (t) {
        case DT_REL:
            RelOffset =  GetWord(MustSwap,pDyn->d_un.d_ptr);
            break;
        case DT_RELSZ:
            Count = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
        case DT_RELENT:
            RelSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;

        case DT_STRTAB:
            Str = DynBits+GetWord(MustSwap,pDyn->d_un.d_val);
            break;
        case DT_STRSZ:
            StrSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
        default:
            break;
        }
        ++pDyn;
        t = GetWord(MustSwap,pDyn->d_tag);
    }

    while (++pDyn < pEnd)
        printf("  x%x x%x\n",
               GetWord(MustSwap,pDyn->d_tag),
               GetWord(MustSwap,pDyn->d_un.d_val));

    if (Str != NULL) {
        char c = 0;
        printf("Strings:");
        while (StrSize-- != 0) {
            if (c == 0)
                printf("\n'");
            c = *Str++;
            if (c == 0)
                printf("'");
            else
                printf("%c",c);
        }        
        printf("\n");
    }

    if ((RelSize == sizeof(Elf32_Rel)) &&
        (RelOffset != 0) && (Count != 0)) {
        Count = Count / RelSize;
        PrintAdsRel(pImg,DynOffset+RelOffset,Count);
    }
}

void PrintInsts(PIMAGE pImg, Elf32_Shdr *pShdr, UINT Six)
{
    BOOL MustSwap = pImg->MustSwap;
    UINT i, t, ns;
    UINT *pInst;
    Elf32_Addr Addr;
    UINT32 Val;

    /* Make sure its code */
    if ((GetWord(MustSwap,pShdr->sh_flags) & SHF_EXECINSTR) == 0)
        return;

    t = GetWord(MustSwap,pShdr->sh_size);
    ns = t / sizeof Val;
           
    printf(" Disassembling %d instructions for section %s:\n",
           ns,
           SectionName(pImg,GetWord(MustSwap,pShdr->sh_name),Six));

    printf("Address  [ Offset ]  Content   Instruction\n");
    Addr = GetWord(MustSwap,pShdr->sh_addr);
    pInst = (UINT *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    for (i = 0; i < ns; i++, pInst++) {

        Val = *pInst;
        Val = GetWord(MustSwap,Val);

        printaddr(i<<2, Addr);
        printf("%8x [%8x] %8x   ", Addr, i<<2, Val);
        pImg->v->Disasm(i<<2/*Addr*/,&Val,Addr);
        printf("\n");

        Addr += 4;
    }
}


int PrintSyms(PIMAGE pImg, Elf32_Shdr *pShdr, UINT Six)
{
    UINT i, t, ns, sx, bnd, typ;
    Elf32_Sym *pSym;
    char *Names = FindStringSection(pImg);
    BOOL MustSwap = pImg->MustSwap;

    t = GetWord(MustSwap,pShdr->sh_size);
    ns = t / sizeof *pSym;
           
    if ((fFlags & F_SYMS) == 0)
      return ns;

    printf(" Section %s holds %d symbols:\n",
           SectionName(pImg,GetWord(MustSwap,pShdr->sh_name),Six),
           ns);

    pSym = (Elf32_Sym *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    for (i = 0; i < ns; i++, pSym++) {
      t = GetWord(MustSwap,pSym->st_name);
      sx = GetHalf(MustSwap,pSym->st_shndx);
      typ = ELF32_ST_TYPE(pSym->st_info);
      bnd = ELF32_ST_BIND(pSym->st_info);
      printf("  [%x] nam=%x val=%x sz=%x t=%s b=%c o=%x s=%x", i,
             t,
             GetWord(MustSwap,pSym->st_value),
             GetWord(MustSwap,pSym->st_size),
             STypeName(typ), SBindName(bnd), pSym->st_other,
             sx);
      if (t && Names)
          printf(" sname=%s", Names + t);
      printf("\n");
    }
    return ns;
}


void PrintImports(PIMAGE pImg, Elf32_Shdr *pShdr, UINT Six)
{
    UINT i, ns;
    Elf32_Ihdr *pIhdr;
    BOOL MustSwap = pImg->MustSwap;

    i = GetWord(MustSwap,pShdr->sh_size);
    ns = i / sizeof *pIhdr;
           
    printf(" Section %s holds %d imports:\n",
           SectionName(pImg,GetWord(MustSwap,pShdr->sh_name),Six),
           ns);

    pIhdr = (Elf32_Ihdr *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    for (i = 0; i < ns; i++, pIhdr++)
      printf("  [%x] dll=%s uuid=%s\n", i,
             pIhdr->ih_name, pIhdr->ih_uuid);
    
}

void PrintExports(PIMAGE pImg, Elf32_Shdr *pShdr, UINT Six)
{
    UINT i, ns;
    Elf32_Addr *pExp;
    BOOL MustSwap = pImg->MustSwap;

    i = GetWord(MustSwap,pShdr->sh_size);
    ns = i / sizeof *pExp;
           
    printf(" Section %s holds %d exports:\n",
           SectionName(pImg,GetWord(MustSwap,pShdr->sh_name),Six),
           ns);

    pExp = (Elf32_Addr *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
    for (i = 0; i < ns; i++, pExp++)
      printf("  Ordinal=%d Address=%x\n", i, GetWord(MustSwap,*pExp));
}

void PrintSectionMap(PIMAGE pImg)
{
    int i;

    printf("SectionMap[%d] @%x:\n",pImg->nSectionBits,pImg->pSectionMap);
    for (i = 0; i < pImg->nSectionBits; i++) {
        printf("\t@%x %x %x %x %d(dec)\n",
               pImg->pSectionMap[i].SecBits,
               pImg->pSectionMap[i].SecAddr,
               pImg->pSectionMap[i].SecOffLimit,
               pImg->pSectionMap[i].SecOffDelta,
               pImg->pSectionMap[i].SecWas);
    }
}

/* Recompute a relocation value
 */
Elf32_Off ArmRemap(int RelaType, Elf32_Off Val)
{
    switch (RelaType) {
    case R_ARM_ABS32:
        /* example:  .long foo+0xNNN
         */
        return Val;

    case R_ARM_PC24:
        /* example:  bl foo
         */
        return Val & 0xffffff;

    }
    printf("Internal error -- ArmRemap %x\n", RelaType);
    return 0;   /* error */
}

Elf32_Off NoRemap(int RelaType, Elf32_Off Val)
{
    return Val;
}



/* Relocate a single storage location in the ELF movable executable.
 */
BOOL RelocateMipsEntry(BOOL MustSwap, int RelaType,
                       Elf32_Addr A, Elf32_Addr PC, char * P)
{
    /* DLL ? Leave it alone
     */
    if (IsaDllReloc(RelaType))
        return TRUE;

    switch (RelaType) {
    case R_MIPS_PC16:
        /* example:  bgtz   a0,foo
         */
        return TRUE;
    case R_MIPS_16:
        /* example:  .short foo
         */
#if 0
        /* this is one way of dealing with sex */
        *(UINT16 *)P = GetHalf(MustSwap,(UINT16)A);
#else
        /* and this is another, perhaps less confusing */
        SetHalf(MustSwap,*(UINT16 *)P,(UINT16)A);
#endif
        break;

    case R_MIPS_32:
        /* example:  .long foo
         */
        SetWord(MustSwap,*(UINT32 *)P,A);
        break;

    case R_MIPS_26:
        /* example:  jal foo
         */
        A = (A /*- 4*/) >> 2;
        A = (A & 0x03ffffff) | (GetWord(MustSwap,*(UINT32 *)P) & 0xfc000000);
        SetWord(MustSwap,*(UINT32 *)P,A);
        break;

    case R_MIPS_GOT16: /*wrong! only use for test */
    case R_MIPS_HI16:
        /* example:  lui a0,foo
         */
        /* beware of sign extension */
        if (A & 0x8000)
            A = (A >> 16) + 1;
        else
            A = (A >> 16);
        /* fall through... */

    case R_MIPS_LO16:
        /* example:  addiu a0,a0,foo
         */
        if (MustSwap) {
            P += 2;
            A = SwapHalf((UINT16)A);
        }
        *(UINT16 *)P = (UINT16)A;
        break;

    default:
        return FALSE;   /* error */
    }

    return TRUE;   /* okay */
}

BOOL RelocateArmEntry(BOOL MustSwap, int RelaType,
                      Elf32_Addr A, Elf32_Addr PC, char * P)
{
    //printf("RelocArmEntry: %x %x %x %x [%x]\n", RelaType, PC, A, P, ((UINT32*)P)[0]);

    /* DLL ? Leave it alone
     */
    if (IsaDllReloc(RelaType))
        return TRUE;

    switch (RelaType) {

    case R_ARM_ABS32:
        /* example:  .long foo+0xNNN
         */
        SetWord(MustSwap,*(UINT32*)P,A);
        break;

    case R_ARM_PC24:
        /* example:  bl foo
         */
        A = A - PC; /* make PC-relative */
        A = (A - 8) >> 2; /* adjust for ARM's two-ahead PC value */
        A = (A & 0x00ffffff) | (GetWord(MustSwap,*(UINT32 *)P) & 0xff000000);
        SetWord(MustSwap,*(UINT32 *)P,A);
        break;

    default:
        return FALSE;   /* error */
    }
    return TRUE;
}

BOOL DontRelocate(BOOL MustSwap, int RelaType,
                  Elf32_Addr A, Elf32_Addr PC, char * P)
{
    return FALSE;
}



/* Perform all relocations specified in the ELF file pointed to by PIMG.
 * Loadable sections have been loaded in memory at PBITS.
 * Those sections will execute at address DESTINATION.
 * This file should be a movable executable with program headers,
 * relocation tables, and DYNSYM simplified dynamic symbols.
 * We make some simplifying assumtions on the file layout.
 */
BOOL DoElfRelocations(FILE_HANDLE f,
                      PIMAGE pImg)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    UINT i;
    BOOL sc;
    UINT32 numPhdrs;
    Elf32_Word RelaSize, RelaEntSize;
    Elf32_Word RelSize, RelEntSize;
    Elf32_Addr delta;
    Elf32_Off  offset, adjust;
    Elf32_Phdr *pPhdr, *pPhdrBase;
    Elf32_Dyn  *pDyn;
    Elf32_Rela *pRela, *pRelaEnd;
    Elf32_Rel  *pRel;
    Elf32_Word w;
    Elf32_Off  Offlimit, n;

    /* Find and extract information from the PT_DYNAMIC section
     * A somewhat ugly side effect: Munge() changes the pgm headers
     * and we must now look at the new ones. Hence the use of pElf
     * in places where pImg->FileBase would be best choice.
     */
    pPhdrBase = (Elf32_Phdr *)(((char*)pElf)+ GetWord(MustSwap,pElf->e_phoff));
    numPhdrs = GetHalf(MustSwap,pElf->e_phnum);
    pRela = NULL;
    RelaSize = RelaEntSize = 0;
    pRel = NULL;
    RelSize = RelEntSize = 0;
    for (i = 0, pPhdr = pPhdrBase; i < numPhdrs; ++i, ++pPhdr) {
      if (GetWord(MustSwap,pPhdr->p_type) == PT_DYNAMIC) {
        pDyn = (Elf32_Dyn *)(((char*)pElf)+ GetWord(MustSwap,pPhdr->p_offset));
        w = GetWord(MustSwap,pDyn->d_tag);
        while (w != DT_NULL) {
          switch (w) {
          case DT_RELA:
            offset = GetWord(MustSwap,pDyn->d_un.d_ptr);
            break;
          case DT_RELASZ:
            RelaSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
          case DT_RELAENT:
            RelaEntSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
          case DT_REL:
            offset = GetWord(MustSwap,pDyn->d_un.d_ptr) +
                     GetWord(MustSwap,pPhdr->p_offset);
            break;
          case DT_RELSZ:
            RelSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
          case DT_RELENT:
            RelEntSize = GetWord(MustSwap,pDyn->d_un.d_val);
            break;
          default:
            break;
          }
          ++pDyn;
          w = GetWord(MustSwap,pDyn->d_tag);
        }
      }
    }

    /* Calculate difference between base address of actual memory image
     * and base address to which file was originally linked.
     */
    delta = pImg->NewLoadAddress - pImg->OldLoadAddress;

    /* If there is a difference we must have some valid relocations
     * But anyways we are still not done: -r linking leaves holes.
     */
    if (((RelaSize+RelSize) == 0) || 
        (RelaEntSize && (RelaEntSize != sizeof *pRela)) ||
        (RelEntSize && (RelEntSize != sizeof *pRel))) {
        printf("Error -- No valid relocations!\n");
        return FALSE;
    }

    /* Patch up entry-point address in ELF header.
     */
    adjust = pImg->ExportAdjustment;
    pElf->e_entry = 
            GetWord(MustSwap,pElf->e_entry) + delta + adjust;
    if (fFlags & F_NOISE)
        printf("Entry gets %x\n", pElf->e_entry);
    pElf->e_entry = GetWord(MustSwap,pElf->e_entry);

    /* Check pointer to relocation entries
     * Read them in if we have not created them.
     */
    n = 0;
    if (pImg->pRela) {
        pRela = pImg->pRela;
        Offlimit = pImg->pSectionMap[n].SecOffLimit;
    } else 
    if (RelaSize) {
        /* Allocate and read in the RELA section
         */
        ALLOCATE(pRela,Elf32_Rela *,RelaSize,"RelaSec");
        if (pRela == NULL) {
            return FALSE;
        }
        _Seek(f, offset);
        if (RelaSize != (Elf32_Word)_Read(f, (PBYTE)pRela, RelaSize)) {
            FREE(pRela,"RelaSec");
            return FALSE;
        }
        pImg->pRela = pRela;
        Offlimit = (Elf32_Off)~0;
    } else
    if (RelSize) {
        /* Allocate, read in and convert (to RELA) the REL section
         */
        int Count = (RelSize / sizeof(*pRel));

        RelaSize = sizeof(*pRela) * Count;
        ALLOCATE(pRela,Elf32_Rela *,RelaSize,"RelaSec");
        if (pRela == NULL) {
            return FALSE;
        }

        ALLOCATE(pRel,Elf32_Rel *,RelSize,"RelSec");
        if (pRel == NULL) {
            FREE(pRela,"RelaSec");
            return FALSE;
        }
        _Seek(f, offset);
        if (RelSize != (Elf32_Word)_Read(f, (PBYTE)pRel, RelSize)) {
            FREE(pRela,"RelaSec");
            FREE(pRel,"RelSec");
            return FALSE;
        }
        /*  convert
         */
        for (i = 0; i < Count; i++) {
            pRela[i].r_offset = pRel[i].r_offset;
            pRela[i].r_info = pRel[i].r_info;
            pRela[i].r_addend = 0;
        }

        FREE(pRel,"RelSec");
        pImg->pRela = pRela;
        Offlimit = (Elf32_Off)~0;
    }

    /* Perform the relocations.
     */
    pRelaEnd = (Elf32_Rela *)(((char *)pRela) + RelaSize);

    if (fFlags & F_DWARF)
    PrintSectionMap(pImg);

    /* Perform all relocations specified in table.
     */
    i = 0;
    for ( ; pRela < pRelaEnd; ++pRela, i++) {
        UINT t = GetWord(MustSwap,pRela->r_info);
        Elf32_Addr A, PC;
        char *P;

        t = ELF32_R_TYPE(t);

        /* r_offset is a file offset as usual.
         * r_addend is from link base
         */
        A = ((Elf32_Sword)GetWord(MustSwap,pRela->r_addend));
          /* - pImg->NewLoadAddress; */
        offset = GetWord(MustSwap,pRela->r_offset);
        PC = offset+pImg->NewLoadAddress;
        while (offset > Offlimit) {
            if (++n >= pImg->nSectionBits) {
                printf("Internal: exceeding section count ? %d %x\n",
                       i, offset);
                n--;
                break;
            }
          Offlimit = pImg->pSectionMap[n].SecOffLimit;
        }
        offset += pImg->pSectionMap[n].SecOffDelta;
        /*printf("[%d] O=%x V=%x PC=%x\n",i++,pImg->OldLoadOffset+offset,A,PC);*/
        P = pImg->FileBase+pImg->OldLoadOffset/*-adjust*/+offset;

        sc = pImg->v->RelocateEntry(MustSwap, t, A, PC, P);
        if (!sc) {
            SetLastError(E_FAIL);
            return FALSE;   /* error */
        }
    }
    return TRUE;   /* okay */
}

BOOL FindRelocsForSection(PIMAGE pImg,
                          Elf32_Shdr *pShdr,
                          UINT SectionIndex,
                          UINT *pRelSectionIndex)
{
  BOOL MustSwap = pImg->MustSwap;
  Elf32_Shdr *pRhdr;
  UINT SecNum, n, rs;

  /* Its usually the next one */
  pRhdr = pShdr + SectionIndex + 1;
  SecNum = GetWord(MustSwap,pRhdr->sh_info);
  if (SecNum == SectionIndex) {
    *pRelSectionIndex = SectionIndex + 1;
    return TRUE;
  }
  return FALSE;
}

int CompareSyminfosByValue(const void *node1, const void *node2)
{
  return ((SYMINFO*)node1)->Value - ((SYMINFO*)node2)->Value;
}

int CompareRelinfosByOffset(const void *node1, const void *node2)
{
  return ((RELINFO*)node1)->r_offset - ((RELINFO*)node2)->r_offset;
}

PSPLIT Desplit(PSPLIT pSplit)
{
  UINT i;
  if (pSplit) {
    for (i = 0; i < pSplit->nSections; i++)
      if (pSplit->Secinfo[i].pSymbols) {
        FREE(pSplit->Secinfo[i].pSymbols,"SYminfos");
      }
    if (pSplit->CodeRelocs.pSymbols)
      FREE(pSplit->CodeRelocs.pSymbols,"Disrelocs");
    FREE(pSplit,"Split");
  }
  return NULL;
}

PSPLIT SplitSectionSymbols(PIMAGE pImg,
                           UINT nSections,
                           Elf32_Sym *pSym,
                           UINT SymCount,
                           char *Names,
                           Elf32_Shdr *pRhdr)
{
    BYTE *pRelBase;
    Elf32_Rel *pRel;
    Elf32_Off nRels;
    UINT i, sx, bnd, typ, rsize;
    BOOL MustSwap = pImg->MustSwap;
    PSPLIT pSplit;
    SYMINFO *pI;
    RELINFO *pR;

    /* Sanity check
     */
    if (pSym == NULL)
        return NULL;

    /* Relocations could be RELs or RELAs
     */
    pRelBase = (BYTE *)(pImg->FileBase + 
                        GetWord(MustSwap,pRhdr->sh_offset));
    i = GetWord(MustSwap,pRhdr->sh_size);
    rsize = GetWord(MustSwap,pRhdr->sh_entsize);
    /* sanity */
    if (rsize == 0)
        rsize = sizeof(Elf32_Rel);/* and good luck */
    nRels = i / rsize;


    ALLOCATE(pSplit,PSPLIT,
             sizeof(SPLIT) + nSections * sizeof(SPLITINFO),"Split");
    if (pSplit == NULL)
      return NULL;
    pSplit->nSections = nSections;

    /* First quick pass to count how many symbols per section
     */
    for (i = 0; i < SymCount; i++) {
      sx = GetHalf(MustSwap,pSym[i].st_shndx);
      /* beware of oddballs */
      if (sx < nSections)
        pSplit->Secinfo[sx].nSymbols++;
    }

    /* Allocate room for the symbol arrays
     */
    for (i = 0; i < nSections; i++) {
      if (pSplit->Secinfo[i].nSymbols) {
        ALLOCATE(pI,SYMINFO*,
                 pSplit->Secinfo[i].nSymbols*sizeof *pI,
                 "SYminfos");
        if (pI == NULL)
          return Desplit(pSplit);
        pSplit->Secinfo[i].pSymbols = pI;
        pSplit->Secinfo[i].Hint = 0;
      }
    }

    /* Now copy them symbols
     */
    for (i = 0; i < SymCount; i++) {
      sx = GetHalf(MustSwap,pSym[i].st_shndx);
      /* beware of oddballs */
      if (sx < nSections) {
        pI = pSplit->Secinfo[sx].pSymbols + pSplit->Secinfo[sx].Hint++;
        pI->Name = Names + GetWord(MustSwap,pSym[i].st_name);
        pI->Type = ELF32_ST_TYPE(pSym[i].st_info);
        pI->Binding = ELF32_ST_BIND(pSym[i].st_info);
        pI->Value = GetWord(MustSwap,pSym[i].st_value);
        pI->Size = (UINT16) GetWord(MustSwap,pSym[i].st_size);
      }
    }

    /* And finally, reorder the symbols by value.
     */
    for (i = 0; i < nSections; i++) {
      if (pSplit->Secinfo[i].nSymbols) {
        qsort((void *)pSplit->Secinfo[i].pSymbols,
              pSplit->Secinfo[i].nSymbols,
              sizeof(SYMINFO),
              CompareSyminfosByValue);
        pSplit->Secinfo[i].Hint = 0;

        if (fFlags & F_DISASM) {
          UINT s;
          printf("Sorted symbols for section %d:\n", i);
          pI = pSplit->Secinfo[i].pSymbols;
          for (s = 0; s < pSplit->Secinfo[i].nSymbols;s++)
            printf(" [%d] %s %x %c %s %x\n", s,
                   pI[s].Name, pI[s].Value, 
                   SBindName(pI[s].Binding),
                   STypeName(pI[s].Type),
                   pI[s].Size);
        }
      }
    }    

    /* Next the relocs.  Unswap them, and reorder by offset.
     */
    ALLOCATE(pR,RELINFO*,nRels*sizeof *pR,"Disrelocs");
    if (pR == NULL)
      return Desplit(pSplit);

    pSplit->CodeRelocs.pSymbols = (SYMINFO*)pR;
    pSplit->CodeRelocs.nSymbols = nRels;

    for (i = 0; i < nRels; i++, pR++) {
      pRel = (Elf32_Rel *)(pRelBase + i * rsize);
      pR->r_offset = GetWord(MustSwap,pRel->r_offset);
      typ = GetWord(MustSwap,pRel->r_info);
      pR->r_type = ELF32_R_TYPE(typ);
      bnd = ELF32_R_SYM(typ);
      pR->Name = Names + GetWord(MustSwap,pSym[bnd].st_name);
      if (bnd < nSections) {
        pR->r_section = GetHalf(MustSwap,pSym[bnd].st_shndx);
        pR->r_sym = 0;/* happens to be right ;-) */
      } else {
        pR->r_section = pImg->CodeIndex;
        pR->r_sym = bnd;
      }
      if (pR->Name == Names)
        pR->Name = NULL;
    }

    qsort((void *)pSplit->CodeRelocs.pSymbols,
          pSplit->CodeRelocs.nSymbols,
          sizeof(RELINFO),
          CompareRelinfosByOffset);

    if (fFlags & F_DISASM) {
      pR = (RELINFO*)pSplit->CodeRelocs.pSymbols;
      printf("Sorted relocs for section %d:\n",pImg->CodeIndex);
      for (i = 0; i < nRels; i++, pR++) {
        printf(" [%d] Off=%x Name=%s Sym=%x t=%x Sec=%x\n", i,
               pR->r_offset, pR->Name,
               pR->r_sym, pR->r_type, pR->r_section);
      }
    }

    /* Done.
     */
    return pSplit;
}


BOOL MatchExportSymbols(PIMAGE pImg, PRTL pRtl)
{
    Elf32_Shdr *pShdr;
    UINT i, j, k, t, ns, sx, bnd, typ;
    UINT nSections;
    Elf32_Sym *pSym;
    char *Names;
    PRTL_ENTRY pEntry;
    UINT nMatches;
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;

    if (fFlags & F_EXP)
    printf("Checking symbols exported by %s...\n",pRtl->DllName);

    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    nSections = GetHalf(MustSwap,pElf->e_shnum);
    Names = FindStringSection(pImg);

    if (Names == NULL) {
        printf("Fatal -- strings missing \n");
        SetLastError((UINT)E_INVALID_PARAMETER);
        return FALSE;
    }

    nMatches = 0;

    /* Pessimistically assume there is more than one symtab
     */
    for (i = 0; i < nSections; ++i, ++pShdr) {

        if (GetWord(MustSwap,pShdr->sh_type) == SHT_SYMTAB) {


            t = GetWord(MustSwap,pShdr->sh_size);
            ns = t / sizeof *pSym;
           
            pSym = (Elf32_Sym *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
            j = 1; pSym++;
            for (; j < ns; j++, pSym++) {
                sx = GetHalf(MustSwap,pSym->st_shndx);
                typ = ELF32_ST_TYPE(pSym->st_info);
                bnd = ELF32_ST_BIND(pSym->st_info);
                /* Only take defined, global, function or data symbols
                 */
                if ((sx != SHN_UNDEF) &&
                    (bnd == STB_GLOBAL) &&
                    ((typ == STT_FUNC) || (typ == STT_OBJECT))) {

                    t = GetWord(MustSwap,pSym->st_name);

                    pEntry = IsaRtl(pRtl,Names+t);
                    if (pEntry) {
                        if (fFlags & F_EXP)
                        printf(" [%x] nam=%s val=%x sz=%x t=%s b=%c o=%x s=%x",
                               j, Names+t,
                               GetWord(MustSwap,pSym->st_value),
                               GetWord(MustSwap,pSym->st_size),
                               STypeName(typ), SBindName(bnd), pSym->st_other,
                               sx);
                        /* The same symbol could export as multiple ordinals
                         * NB: qsort by name + bsearch means we find ONE of
                         * them, not necessarily the first.  Sigh.
                         */
                        k = pEntry - pRtl->pEntries;
                        while ((k > 0)
                               && (strcmp(pEntry[-1].RtlName,Names+t)==0)) {
                            pEntry--;
                            k--;
                        }
                        for (;
                             k < pRtl->nEntries;
                             k++) {
                            if (strcmp(Names+t,pEntry->RtlName) != 0)
                                break;
                            if (fFlags & F_EXP)
                            printf(" Resolved %s[%d] with %s\n",
                                   pRtl->DllName,pEntry->Ordinal,Names+t);
                            if (pEntry->RefAsSymbol) {
                                printf("Warning -- Exported symbol %s is multiply defined\n",
                                       pEntry->RtlName);
                            } else {
                                pEntry->RefAsSymbol = (char *)pSym;
                                nMatches++;
                            }
                            pEntry++;
                        }
                    }
                }
            }
        }
    }
    /* We must have found all exported symbols or else.
     */
    if (nMatches != pRtl->nEntries) {
      printf("Fatal -- Could not resolve %d exported symbol(s):\n",
             pRtl->nEntries - nMatches);
      /* Oh well, tell them what it is..
       */
      pEntry = pRtl->pEntries;
      for (i = 0; i < pRtl->nEntries; i++, pEntry++)
        if (pEntry->RefAsSymbol == NULL)
          printf("%d:%s\n",pEntry->Ordinal,pEntry->RtlName);
      SetLastError((UINT)E_INVALID_PARAMETER);
      return FALSE;
    }
    return TRUE;
}

/* Build the export section.
 * NB: Takes values from symbols
 */
BOOL BuildExportSection(PIMAGE pImg, Elf32_Addr Address, Elf32_Off Offset)
{
    BOOL MustSwap = pImg->MustSwap;
    PRTL pRtl;
    PRTL_ENTRY pEntry;
    UINT nExport, i;
    Elf32_Shdr *pShdr;
    Elf32_Addr *pExp;
    Elf32_Sym  *pSym;

    if (fFlags & F_EXP)
    printf("Building export section...\n");

    /* Got to find how many ordinals we need.
     * Sort the array by ordinals and look at the last entry.
     */
    pRtl = pImg->pExportRtl;
    qsort((void *)pRtl->pEntries, pRtl->nEntries, sizeof(RTL_ENTRY), CompareRtlByOrdinal);
    pEntry = pRtl->pEntries + (pRtl->nEntries - 1);
    nExport = pEntry->Ordinal+1;
    if (fFlags & F_EXP)
    printf("Got %d ordinals\n", nExport);

    /* Allocate and zero one big chunk for header and section
     */
    i = sizeof(Elf32_Shdr) + (nExport * sizeof(Elf32_Addr));
    ALLOCATE(pShdr,Elf32_Shdr *,i,"XportSec");
    if (pShdr == NULL)
        return FALSE;

    pShdr->sh_entsize = GetWord(MustSwap,sizeof(Elf32_Addr));
    i = nExport * sizeof(Elf32_Addr);
    pShdr->sh_size = GetWord(MustSwap,i);

    /* Round up size, we cant have holes
     */
    i = (i + 15) & ~15;
    pImg->ExportAdjustment = i;

    pShdr->sh_name = 0; /* BUGBUG */
    pShdr->sh_type = GetWord(MustSwap,SHT_DYNSYM);
    pShdr->sh_flags = GetWord(MustSwap,SHF_ALLOC);
    pShdr->sh_addr = GetWord(MustSwap,Address);
    pShdr->sh_offset = GetWord(MustSwap,Offset);
    pShdr->sh_link = 0;
    pShdr->sh_info = 0;
    pShdr->sh_addralign = GetWord(MustSwap,4);


    pExp = (Elf32_Addr *)(pShdr + 1);
    pEntry = pImg->pExportRtl->pEntries;
    for (i = 0; i < nExport; i++) {
        if (i == (UINT) pEntry->Ordinal) {
            pSym = (Elf32_Sym *)pEntry->RefAsSymbol;
#if 1  //quick
            *pExp = GetWord(MustSwap,pSym->st_value) +
                pImg->NewLoadAddress + pImg->ExportAdjustment;
            if (fFlags & F_EXP)
            printf("  %s gets x%x\n", pEntry->RtlName, *pExp);
            *pExp = GetWord(MustSwap,*pExp);
#endif
            pEntry++;
        } else {
            if (fFlags & F_EXP)
            printf("  %d is unused\n", i);
            *pExp = (Elf32_Addr)~0;
        }
        pExp++;
    }

    pImg->pExportSection = pShdr;
    return TRUE;
}

/* It seems we cannot trust the "addend" field of relocations,
 * which we need at relocation time.  In the .REL case
 * it is just plain missing, in the .RELA case its often bogus.
 *
 * For MIPS we have to pick up our info from the code.  This means
 * reassembling the hi16 and lo16 pieces of an address, because
 * the corresponding symbol is often missing.
 * It is solidly rooted into the linker that these relocations
 * must be contiguous, and the lo16 follows any hi16 it "completes".
 * There are typically one or two hi16 per lo16.
 */
typedef struct {
    Elf32_Rela *Rela;
    INT16 Value;
} MIPS_FIXUP, *PMIPS_FIXUP;
typedef struct {
    UINT nFixup;
    UINT FixCount;
    PMIPS_FIXUP pFixup;
} SECFIXUP, *PSECFIXUP;

BOOL MapRelocs(PIMAGE pImg, Elf32_Rela *pRela, SECMAP *NewpSectionMap)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    UINT nRela;
    Elf32_Shdr *pShdr;
    Elf32_Shdr *pSecHdr, *pSymHdr;
    char *SecBase;
    Elf32_Rela *pRelaIn;
    Elf32_Sym *pSym;
    UINT i, j, nSections, nRel, tag, SecNum, NewSecNum, SymNum;
    Elf32_Off Off, SecOff;
    UINT Info, sym, Val, Add;
    Elf32_Addr SecValue;
    PSECFIXUP pSecFix, pFix;
    MYHEADERS *pHdr = pImg->pHdr;
    BOOL Ret = FALSE;

    nSections = GetHalf(MustSwap,pElf->e_shnum);
    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));

    /* Allocate a small fixup array, it will grow on demand
     */
    ALLOCATE(pSecFix,PSECFIXUP,nSections*sizeof(SECFIXUP),"SecFixups");
    if (pSecFix == NULL) {
        printf("Fatal -- Out of memory\n");
        return FALSE;
    }
    for (i = 0; i < nSections; ++i) {
        pSecFix[i].FixCount = 0;
        pSecFix[i].nFixup = 8;
        ALLOCATE(pSecFix[i].pFixup,MIPS_FIXUP *,8*sizeof(MIPS_FIXUP),"Fixups");
        if (!pSecFix[i].pFixup) {
            printf("Fatal -- Out of memory\n");
            return FALSE;
        }
    }

    /* Now go back and "copy" the relocations
     */
    nRela = 0;
    for (i = 0; i < nSections; ++i) {

        /* Ignore dwarf relocations
         */
        char *n = SectionName(pImg,GetWord(MustSwap,pShdr[i].sh_name),i);
        if ((strncmp(n,".rel.debug_",4+7) == 0) ||
            (strncmp(n,".rel.stab",4+5) == 0)) {
            if (fFlags & F_DWARF)
                printf("Dropping section %s\n", n);
            continue;
        }

        /* Share code between REL and RELAs, they only differ
         * in the trailing r_addend field (which is clobbered).
         */
        Add = GetWord(MustSwap,pShdr[i].sh_type);
        if (Add == SHT_REL) Add = sizeof(Elf32_Rel);
        else if (Add == SHT_RELA) Add = sizeof(Elf32_Rela);
        else Add = 0;
        
        if (Add) {

            /* Get proper count of relocations
             */
            nRel = GetWord(MustSwap,pShdr[i].sh_size) / Add;

            /* Get ptrs to section bits to relocate
             */
            SecNum = GetWord(MustSwap,pShdr[i].sh_info);
            pSecHdr = pShdr + SecNum;
            SecBase = pImg->pSectionMap[SecNum].SecBits;

            /* Get ptrs to symbols
             */
            SymNum = GetWord(MustSwap,pShdr[i].sh_link);
            pSymHdr = pShdr + SymNum;
            pSym = (Elf32_Sym*)pImg->pSectionMap[SymNum].SecBits;

            /* We are flattening all sections, Off must change
             * NB: It becomes the offset from the start of the load-section,
             * and NOT anymore from the start of the file.
             */
#if 0
            SecOff = GetWord(MustSwap,pHdr->Sh[SecNum].sh_offset);
            SecOff = SecOff - pImg->NewLoadOffset;
            /* Doesnt work because section indexes have changed */
#endif
#if 0
            SecOff = GetWord(MustSwap,pShdr[i].sh_offset);
            SecOff = SecOff - pImg->OldLoadOffset + pImg->NewLoadOffset;
            /* Doesnt work because we might add stuff upfront */
#endif
            /* So it seems we cant escape a section remapping table
             * [which was built by Munge]
             */
            SecOff = pShdr[SecNum].sh_addralign;
            SecOff = GetWord(MustSwap,pHdr->Sh[SecOff].sh_offset);
            SecOff = SecOff - pImg->NewLoadOffset;

            /* Map REL records to RELA records (see comment above)
             */
            for (j = 0; j < nRel; j++) {
                pRelaIn = (Elf32_Rela *)(pImg->pSectionMap[i].SecBits + (j * Add));

                Off = GetWord(MustSwap,pRelaIn->r_offset);
                Info = GetWord(MustSwap,pRelaIn->r_info);

                tag = ELF32_R_TYPE(Info);
                sym = ELF32_R_SYM(Info);

                /* Our DLL rels?
                 */
                if (IsaDllReloc(tag)) {
                    /* Only offset can change */
                    Val = GetWord(MustSwap,pRelaIn->r_addend);
                    goto SetRela;
                }

                /* Must follow symtab remapping of section numbers
                 */
                SecNum = GetHalf(MustSwap,pSym[sym].st_shndx);

                /* So we get these ABS symbols that are not absolutes.
                 * BUGBUG check the list for "__bss_start" & co.
                 */
                if (SecNum >= SHN_LORESERVE) {
                    if (SecNum != SHN_ABS) {
                        printf("Fatal -- Relocation for bad symbol (%x)\n",SecNum);
                        goto Exit;
                    }
                    Val = GetWord(MustSwap,pSym[sym].st_value);
                    Val = Val   /*- pImg->OldLoadAddress*/ + pImg->ExportAdjustment;
                    /*printf("bss-like symbol gets %x\n", Val);*/
                    Off = Off;
                    goto SetRela;
                }

                pFix = pSecFix + SecNum;
                NewSecNum = pShdr[SecNum].sh_addralign;

                /* Sections themselves act as symbols (their base)
                 * Get the new value for the section start.
                 */
                if (SecNum != SHN_UNDEF)
                    SecValue = NewpSectionMap[NewSecNum].SecAddr;
                else
                    SecValue = 0;

                /* Value might come from an instruction
                 */
                Val = GetWord(MustSwap,*((UINT*)(SecBase + Off)));

                /* This should be in a separate function but..
                 * ..the MIPS case is so freaking messy I am not 
                 * going to try.  Unless its simple.
                 */

                if (pImg->v->Machine != EM_MIPS) {
                    Val = pImg->v->RemapReloc(tag,Val);
                } else {
                    switch (tag) {
                    case R_MIPS_PC16:
                        /* PC-relative, no need to do anything */
                        Val &= 0xffff;
                        Info = ELF32_R_INFO(NewSecNum,tag);
                        goto SetRela;
                    case R_MIPS_16:
                        Val &= 0xffff;
                        break;
                    case R_MIPS_32:
                        break;
                    case R_MIPS_26:
                        Val &= 0x03ffffff;
                        Val <<= 2;
                        break;
                    case R_MIPS_GOT16: /* not really but.. */
                    case R_MIPS_HI16:

                        Val <<= 16;
                        /* If we have a symbol for it we dont really need
                         * to wait until the low part shows up.
                         */
                        if (sym <= nSections) {
                            /* One of the bad ones
                             */
                            if (pFix->FixCount == pFix->nFixup) {
                                REALLOCATE(pFix->pFixup,
                                           MIPS_FIXUP *,
                                           2*pFix->nFixup*sizeof(MIPS_FIXUP),
                                           pFix->pFixup,
                                           "Fixups");
                                if (pFix->pFixup == NULL) {
                                    printf("Fatal -- Out of memory (%d nfix) err=x%x\n",
                                           pFix->nFixup, GetLastError());
                                    return FALSE;
                                }
                                pFix->nFixup = 2*pFix->nFixup;
                            }

                            /* Shift back, make it fit */
                            Val >>= 16;

                            /*printf("BadHi[%d] %d@%x\n",pFix->FixCount,j,Off);*/
                            pFix->pFixup[pFix->FixCount].Rela = pRela+nRela;
                            pFix->pFixup[pFix->FixCount].Value = Val;
                            pFix->FixCount++;
                            /* Proceed to compute the other fields
                             */
                        }
                        break;

                    case R_MIPS_LO16:
                        if (sym <= nSections) {

                            INT16 sVal;

                            /* We better have a (few) Hi16 pending.
                             * But if we dont its ok, just reuse the previous ones.
                             */
                            /*printf("GoodLo (%d) %d@%x\n",pFix->FixCount,j,Off);*/
                            if (pFix->FixCount == 0) {
                                if (fFlags & F_RELS)
                                    printf("Warning -- Lo16[%d] reuse, section #%d -> %d\n",
                                           j, i, SecNum);
                            }

                            /* They are all just like the first one
                             */
                            sVal = (INT16)Val;
                            Val = (UINT)(sVal + (INT32)(pFix->pFixup[0].Value << 16));

                            /* Adjust the addend
                             */
                            Val = Val + SecValue /* - pImg->NewLoadAddress + pImg->ExportAdjustment*/;
                            /* Fixup all the Hi16
                             */
                            while (pFix->FixCount-- != 0) {
                                pFix->pFixup[pFix->FixCount].Rela->r_addend = 
                                    GetWord(MustSwap,Val);
                            }
                            pFix->FixCount = 0;
                            goto SetRela;
                        }
                        /* otherwise the Val is quickly clobbered */
                        break;
                    default:
                        printf("Fatal -- Bad relocation type x%x\n", tag);
                        goto Exit;
                    }

                }
                if (sym > nSections) {
                    /* Val changes (its zero anyways), comes from symbol
                     */
//if (Val != 0) printf("Val=%x rel=%d tag=%d sym=x%x\n", Val, j, tag, sym);
                    Val += GetWord(MustSwap,pSym[sym].st_value);

                    /* Is this a DLL reference ?  Or an undefined symbol.
                     */
                    if (SecNum == SHN_UNDEF) {
                        int DllIx = pSym[sym].st_other;
                        if (DllIx == 0) {
                            printf("Fatal -- Undefined symbol ix=x%x in image\n",sym);
                            goto Exit;
                        }
                        /* Perform DLL index mapping
                         */
                        DllIx = pImg->pRtl[DllIx-1]->Index;

                        /* Turn on high bit in type to mark it as DLL import
                         */
                        Info = ELF32_R_INFO(DllIx,AddDllTag(tag));

                        /* Addend is the ordinal (in st_value)
                         */
                        Val = Val;

                        if (fFlags)
                            printf("Dllrel: %x %x %x\n", Off, Info, Val);
                        goto SetRela;
                    }
                }

                /* Compute the addend now */
                Val = Val + SecValue /* - pImg->NewLoadAddress + pImg->ExportAdjustment*/;
                Info = ELF32_R_INFO(NewSecNum,tag);

            SetRela:
                Off += SecOff;

                pRela[nRela].r_addend = GetWord(MustSwap,Val);
                pRela[nRela].r_offset = GetWord(MustSwap,Off);
                pRela[nRela].r_info   = GetWord(MustSwap,Info);

                nRela++;
            }

            /* Done with this section
             */
        } /* if we have RELs in this section */

    } /* forall sections */
    Ret = TRUE;
 Exit:
    for (i = 0; i < nSections; ++i) {
        if (pSecFix[i].FixCount != 0) {
            printf("Something rotten in Denmark[%d]\n",i);
            Ret = FALSE;
        }
        FREE(pSecFix[i].pFixup,"Fixups");
    }
    FREE(pSecFix,"SecFixups");
    return Ret;
}

/* Main control code for building MMLite images.
 * Takes the linker output and adds import and export sections.
 * Rebuilds the program headers.  Re-creates the relocations.
 * Re-assigns section addresses, possibly dropping useless sections.
 */
BOOL Munge(PIMAGE pImg)
{
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    Elf32_Addr Base = pImg->NewLoadAddress;
    UINT RelaCnt, RelaSize;
    Elf32_Shdr *pShdr;
    Elf32_Rela *pRela;
    UINT i, j, nSections, nMySections,  tag, fl, NoLoad;
    UINT BitsSize;
    Elf32_Addr Addr, FileBase;
    Elf32_Off Off;
    Elf32_Dyn *pDyn;
    MIPS_FIXUP *pFixHi, *pFixLo;
    MYHEADERS *pHdr;
    SECMAP *pNewSectionMap;

    if (fFlags & F_DWARF)
    PrintSectionMap(pImg);

    pFixHi = pFixLo = NULL;

    /* Find start of section headers
     */
    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    nSections = GetHalf(MustSwap,pElf->e_shnum);

    /* Make a first pass to decide which section we'll keep and which we'll drop
     * Since we are dropping a few sections we need to redo the
     * scatter array of section pointers.  Allocate the new section map
     * We'll be adding at most three new sections: rela, im/export.
     * But most likely we'll squash quite a few down below.
     */
    ALLOCATE(pNewSectionMap,SECMAP *,(nSections+3)*sizeof(SECMAP),"pSecMap");
    if (pNewSectionMap == NULL)
        return FALSE;


    NoLoad = 0;
    j = 0;
    for (i = 0; i < nSections; ++i) {

        char *n = SectionName(pImg,GetWord(MustSwap,pShdr[i].sh_name),i);

        fl = GetWord(MustSwap,pShdr[i].sh_flags);
        tag = GetWord(MustSwap,pShdr[i].sh_type);

        switch (tag) {

        case SHT_NOBITS: /* .bss .sbss */
          NoLoad += GetWord(MustSwap,pShdr[i].sh_size);
          break;

        case SHT_PROGBITS:
          /* keep it, unless its one of the dwarf debug ones
           */
          if (strncmp(n,".debug_",7) == 0) {
              if (fFlags & F_DWARF)
                printf("Dropping %s section (DWARF debug methinks)\n",n);
              fl = 0;
          }
          break;

        case SHT_SYMTAB:
        case SHT_STRTAB:
          if ((fOptions & F_STRIP) == 0)
              fl = SHF_ALLOC;
          /* else its already 0 */
          break;
        case 0x70000006: /* .reginfo */
        case SHT_RELA:
        case SHT_HASH:
        case SHT_DYNAMIC:
        case SHT_NOTE:
        case SHT_REL:
        case SHT_SHLIB:
        case SHT_DYNSYM:
        default:
          /* discard it */
          fl = 0;
          break;
        }

        /* Keep section slot so that indeces for symtab stay ok.
         */
        if ((fl & SHF_ALLOC) ||((fOptions & F_STRIP) == 0)) {
          /* keep it */
          pNewSectionMap[i].pShdr = pShdr+i;
          j++;
        }
    }

    if (pImg->pImport)
        j++; /* we'll add an import section */
    if (pImg->pExportRtl)
        j++; /* we'll add an export section */

    if (fFlags & F_NOISE)
      printf("Will keep %d+1 sections out of %d\n", j, nSections);

    /* Allocate and zero proto hdrs
     * We'll need an extra one for RELAs
     */
     j++;
#define LOGALIGN 4
#if 0
    FileBase = 0x1000;/* old FILEOFF hack */
#else
    FileBase = sizeof(MYHEADERS) + (j*sizeof(Elf32_Shdr));
    FileBase = (FileBase + (2<<LOGALIGN) - 1) & ~((2<<LOGALIGN) - 1);
#endif
    ALLOCATE(pHdr,MYHEADERS*,FileBase,"pHdr");
    if (pHdr == NULL)
        return FALSE;
    memset(pHdr,0,FileBase);
    
    /* This will be the new base offset
     */
    pImg->NewLoadOffset = FileBase;
    pImg->ElfSize = FileBase;

    /* Initialize ELF hdr
     */
    pImg->pHdr = pHdr;

    pHdr->Elf = *pElf;/* defaults */
    pHdr->Elf.e_type = GetHalf(MustSwap,ET_EXEC);
    pHdr->Elf.e_phoff = GetWord(MustSwap,sizeof(Elf32_Ehdr));
    pHdr->Elf.e_shoff = GetWord(MustSwap,offsetof(MYHEADERS,Sh0));
    pHdr->Elf.e_phentsize = GetHalf(MustSwap,sizeof(Elf32_Phdr));
    pHdr->Elf.e_phnum = GetHalf(MustSwap,NMYPH);
    pHdr->Elf.e_shstrndx = GetHalf(MustSwap,1);

    /* Initialize pgm headers
     */
    pHdr->Ph[PGH_BITS].p_type   = GetWord(MustSwap,PT_LOAD);
    pHdr->Ph[PGH_BITS].p_offset = GetWord(MustSwap,FileBase);
    pHdr->Ph[PGH_BITS].p_vaddr  = GetWord(MustSwap,Base);
    pHdr->Ph[PGH_BITS].p_paddr  = GetWord(MustSwap,Base);
    pHdr->Ph[PGH_BITS].p_flags  = GetWord(MustSwap,0x7);
    pHdr->Ph[PGH_BITS].p_align  = GetWord(MustSwap,LOGALIGN);

    /* Add .export section at the front, if we need one
     * Actually.. we must put the bits upfront but the
     * section header we put it at the end to minimize
     * the damage to ELF compatibility.
     */
    nMySections = 0;
    Addr = Base;
    Off = pImg->NewLoadOffset;
    if (pImg->pExportRtl) {

        /* Build the export section
         */
        BuildExportSection(pImg, Addr, Off);

        /* Round up size, we cant have holes
         */
        i = GetWord(MustSwap,pImg->pExportSection->sh_size);
        i = (i + 15) & ~15;
        Off += i;
        Addr += i;
    }

    /* Select sections we'll keep
     * Assign them both a file offset and a load address
     */
    for (i = 0; i < nSections; ++i) {

        char *n = SectionName(pImg,GetWord(MustSwap,pShdr[i].sh_name),i);
        BOOL IsHot = FALSE;

        /* Did we decide to drop it?
         */
        if (pNewSectionMap[i].pShdr == NULL)
            continue;
        pNewSectionMap[i].pShdr = NULL;

        fl = GetWord(MustSwap,pShdr[i].sh_flags);
        tag = GetWord(MustSwap,pShdr[i].sh_type);

        if (n && (strncmp(n,".hot",5) == 0)) {
            printf("Got the hot one at %d\n",i);
            IsHot = TRUE;
        }
            
        if (fl & SHF_ALLOC) {
          /* copy it */
          pHdr->Sh[nMySections] = pShdr[i];
    	  pNewSectionMap[nMySections].pShdr = &pHdr->Sh[nMySections];
          pNewSectionMap[nMySections].SecBits = pImg->pSectionMap[i].SecBits;

          pNewSectionMap[nMySections].SecWas = i;

          /* alignment */
          j = GetWord(MustSwap,pShdr[i].sh_addralign);
          if (j) {
              Off = (Off + j - 1) & ~(j - 1);
              Addr = (Addr + j - 1) & ~(j - 1);
          }

          pHdr->Sh[nMySections].sh_offset = GetWord(MustSwap,Off);
          pNewSectionMap[nMySections].SecOffDelta =
            (GetWord(MustSwap,pShdr[i].sh_offset) - pImg->OldLoadOffset) -
            (Off - pImg->NewLoadOffset);

          j = GetWord(MustSwap,pShdr[i].sh_size);

          /* Where will it land.
           */
          if (! IsHot) {

              pHdr->Sh[nMySections].sh_addr = GetWord(MustSwap,Addr);
              pNewSectionMap[nMySections].SecAddr = Addr;

          } else {
              /* The .hot section is at a fixed address,
               * dont change it.
               */
              pNewSectionMap[nMySections].SecAddr = 
                  GetWord(MustSwap,pHdr->Sh[nMySections].sh_addr);

              /* Needs a pgm header, too
               */
              pHdr->Ph[PGH_SCRAM].p_type   = GetWord(MustSwap,PT_LOAD);
              pHdr->Ph[PGH_SCRAM].p_offset = pHdr->Sh[nMySections].sh_offset;
              pHdr->Ph[PGH_SCRAM].p_vaddr  = pHdr->Sh[nMySections].sh_addr;
              pHdr->Ph[PGH_SCRAM].p_paddr  = pHdr->Sh[nMySections].sh_addr;
              pHdr->Ph[PGH_SCRAM].p_filesz = pHdr->Sh[nMySections].sh_size;
              pHdr->Ph[PGH_SCRAM].p_memsz  = pHdr->Sh[nMySections].sh_size;
              pHdr->Ph[PGH_SCRAM].p_flags  = GetWord(MustSwap,PTF_SCRAM);
              pHdr->Ph[PGH_SCRAM].p_align  = GetWord(MustSwap,4);

          }

          if (fFlags & F_NOISE)
          printf("SecAddr = %x\n", pNewSectionMap[nMySections].SecAddr);

          if (fFlags)
          printf("Section %d [was %d] is x%x bytes @ %x [%x]\n",
                 nMySections, i, j, Addr, Off);
          Addr += j;
          if (tag != SHT_NOBITS)
              Off += j;
          pNewSectionMap[nMySections].SecOffLimit = Off - pImg->NewLoadOffset;

          /* We have killed the .shstrtab section */
          if (fOptions & F_STRIP)
            pHdr->Sh[nMySections].sh_name = SHN_UNDEF;

          /* Down below we need to remember which section came from which.
           * Since I am about to discard the old section table I might
           * as well use it for this purpose.  I am stealing the sh_addralign
           * field (instead of building a separate table) because it
           * seems the least likely to be still useful.
           */
          pShdr[i].sh_addralign = nMySections;

          nMySections++;
        } else if ((fOptions & F_STRIP) == 0) {
          /* Keep section slot so that indeces (rela,symtab) stay ok.
           */
          memset(&pHdr->Sh[nMySections],0,sizeof(Elf32_Shdr));
          pNewSectionMap[nMySections].pShdr = &pHdr->Sh[nMySections];
          nMySections++;
printf("Section %d -> zeroed %d\n", i, nMySections);
        }
    }

    /* Basic info for the OS loader
     */
    pHdr->Ph[PGH_BITS].p_memsz = GetWord(MustSwap,Addr-Base);
    pHdr->Ph[PGH_BITS].p_filesz = GetWord(MustSwap,Addr-Base-NoLoad);

    BitsSize = Addr-Base-NoLoad;

    /* Find how many REL/RELA entries we got (we'll collapse them all)
     */
    RelaCnt = 0;
    for (i = 0; i < nSections; ++i) {
        char *n = SectionName(pImg,GetWord(MustSwap,pShdr[i].sh_name),i);
        tag = GetWord(MustSwap,pShdr[i].sh_type);
        /* Drop dwarf relocations, we dropped them sections
         */
        if ((strncmp(n,".rel.debug_",4+7) == 0) ||
            (strncmp(n,".rel.stab",4+5) == 0)) {
            if (fFlags & F_DWARF)
            printf("Dropping %d RELs from section %s\n", 
                   GetWord(MustSwap,pShdr[i].sh_size) / sizeof(Elf32_Rel),
                   n);
            continue;
        }
        if (tag == SHT_REL) {
            RelaCnt += GetWord(MustSwap,pShdr[i].sh_size) / sizeof(Elf32_Rel);
        }
        if (tag == SHT_RELA) {
            RelaCnt += GetWord(MustSwap,pShdr[i].sh_size) / sizeof(Elf32_Rela);
        }
    }
    RelaSize = RelaCnt * sizeof *pRela;
    if (fFlags)
    printf("Total of %d REL entries\n", RelaCnt);
    if (RelaCnt == 0) {
        SetLastError((UINT)E_INVALID_PARAMETER);
        return FALSE; /* sanity */
    }
    
    /* Fill in PT_DYNAMIC program header
     */
    pHdr->Ph[PGH_DYNA].p_type = GetWord(MustSwap,PT_DYNAMIC);
    pHdr->Ph[PGH_DYNA].p_offset = GetWord(MustSwap,offsetof(MYHEADERS,Dyn));
    pHdr->Ph[PGH_DYNA].p_vaddr = 0;
    pHdr->Ph[PGH_DYNA].p_paddr = 0;
    pHdr->Ph[PGH_DYNA].p_filesz = GetWord(MustSwap,(MYDYN*sizeof *pDyn));
    pHdr->Ph[PGH_DYNA].p_memsz = 0;
    pHdr->Ph[PGH_DYNA].p_flags = 0;
    pHdr->Ph[PGH_DYNA].p_align = GetWord(MustSwap,0x4);

    /* And the bits it points to
     */
    pDyn = pHdr->Dyn;
    pDyn[0].d_tag = GetWord(MustSwap,DT_RELA);
    pDyn[0].d_un.d_ptr = GetWord(MustSwap,FileBase + BitsSize); /* uhu? */
    pDyn[1].d_tag = GetWord(MustSwap,DT_RELASZ);
    pDyn[1].d_un.d_val = GetWord(MustSwap,RelaSize);
    pDyn[2].d_tag = GetWord(MustSwap,DT_RELAENT);
    pDyn[2].d_un.d_val = GetWord(MustSwap,sizeof *pRela);
    pDyn[3].d_tag = GetWord(MustSwap,DT_NULL);
    pDyn[3].d_un.d_val = 0;


    /* Allocate room for the RELA bits
     */
    ALLOCATE(pRela,Elf32_Rela *,RelaSize,"RELAs");
    if (pRela == NULL) {
        return FALSE;
    }

    /* Update link state
     */
    pImg->pRela = pRela;
    pImg->RelaSize = RelaSize;

    /* Fill RELA section header [strip it ?]
     */
    Off = FileBase + BitsSize;
    pNewSectionMap[nMySections].SecBits = (char*)pRela;
    pNewSectionMap[nMySections].SecAddr = 0;
    pNewSectionMap[nMySections].SecWas = -1;
    pHdr->Sh[nMySections].sh_name = 0; /* BUGBUG */
    pHdr->Sh[nMySections].sh_type = GetWord(MustSwap,SHT_RELA);
    pHdr->Sh[nMySections].sh_flags = 0;
    pHdr->Sh[nMySections].sh_addr = 0;
    pHdr->Sh[nMySections].sh_offset = GetWord(MustSwap,Off);
    pHdr->Sh[nMySections].sh_size = GetWord(MustSwap,RelaSize);
    pHdr->Sh[nMySections].sh_link = 0;
    pHdr->Sh[nMySections].sh_info = GetWord(MustSwap,PGH_BITS);
    pHdr->Sh[nMySections].sh_addralign = GetWord(MustSwap,4);
    pHdr->Sh[nMySections].sh_entsize = GetWord(MustSwap,sizeof *pRela);
    if (fFlags)
    printf("Section %d [is new .rela] is x%x bytes @ [%x] %x\n",
           nMySections, RelaSize, Addr, Off);
    nMySections++;
    Off += RelaSize;

    /* Fill IMPORT section iff
     */
    if (pImg->pImport) {
      pHdr->Sh[nMySections] = *pImg->pImport;
      pHdr->Sh[nMySections].sh_offset = GetWord(MustSwap,Off);
      pNewSectionMap[nMySections].SecBits = (char*)(pImg->pImport+1);
      pNewSectionMap[nMySections].SecAddr = 0;
      pNewSectionMap[nMySections].SecWas = -1;

      /* Needs a pgm header, too
       */
      i = GetWord(MustSwap,pHdr->Sh[nMySections].sh_size);
      pHdr->Ph[PGH_IMPT].p_type   = GetWord(MustSwap,PT_SHLIB);
      pHdr->Ph[PGH_IMPT].p_offset = pHdr->Sh[nMySections].sh_offset;
      pHdr->Ph[PGH_IMPT].p_vaddr  = 0;
      pHdr->Ph[PGH_IMPT].p_paddr  = 0;
      pHdr->Ph[PGH_IMPT].p_filesz = pHdr->Sh[nMySections].sh_size;
      pHdr->Ph[PGH_IMPT].p_memsz  = 0;
      pHdr->Ph[PGH_IMPT].p_flags  = GetWord(MustSwap,PTF_IMPORT);
      pHdr->Ph[PGH_IMPT].p_align  = GetWord(MustSwap,0x1);

      if (fFlags)
      printf("Section %d [is new .import] is x%x bytes @ [%x] %x\n",
             nMySections, i, Addr, Off);
      nMySections++;
    }

    /* Fill in export section iff
     */
    if (pImg->pExportRtl) {

        /* Just copy it, it was built right.
         */
        pHdr->Sh[nMySections] = *pImg->pExportSection;

        /* Set section Map stuff
         */
        pNewSectionMap[nMySections].SecBits = (char *)(pImg->pExportSection+1);
        pNewSectionMap[nMySections].SecAddr = GetWord(MustSwap,pImg->pExportSection->sh_addr);
        pNewSectionMap[nMySections].SecWas = -1;
        pNewSectionMap[nMySections].SecOffLimit = 0;
        pNewSectionMap[nMySections].SecOffDelta = 0;

        /* Round up size, we cant have holes
         */
        i = GetWord(MustSwap,pImg->pExportSection->sh_size);
        i = (i + 15) & ~15;

        /* And its pgm header
         */
        pHdr->Ph[PGH_EXPT].p_type   = GetWord(MustSwap,PT_SHLIB);
        pHdr->Ph[PGH_EXPT].p_offset = pHdr->Sh[nMySections].sh_offset;
        pHdr->Ph[PGH_EXPT].p_vaddr  = pHdr->Sh[nMySections].sh_addr;
        pHdr->Ph[PGH_EXPT].p_paddr  = pHdr->Sh[nMySections].sh_addr;
        pHdr->Ph[PGH_EXPT].p_filesz = pHdr->Sh[nMySections].sh_size;
        pHdr->Ph[PGH_EXPT].p_memsz  = GetWord(MustSwap,i);
        pHdr->Ph[PGH_EXPT].p_flags  = GetWord(MustSwap,PTF_EXPORT);
        pHdr->Ph[PGH_EXPT].p_align  = GetWord(MustSwap,0x1);


        if (fFlags)
        printf("Section %d [is new .export] is x%x bytes @ %x [%x]\n",
               nMySections, i,
               pNewSectionMap[nMySections].SecAddr,
               GetWord(MustSwap,pImg->pExportSection->sh_offset));

        nMySections++;
    }

    pHdr->Elf.e_shnum = GetHalf(MustSwap,(UINT16)nMySections);
    if (fFlags)
    printf("Got %d sections total\n", nMySections);

    /* Now go back and "copy" the relocations
     */
    if (!MapRelocs(pImg,pRela,pNewSectionMap)) {
      SetLastError((UINT)E_INVALID_PARAMETER);
      return FALSE;
    }
    pHdr->Elf.e_shnum = GetHalf(MustSwap,(UINT16)nMySections+1); /* silly zero section */

    /* We should not be needing the old stuff anymore
     */
    FREE(pImg->pSectionMap,"pSecMap");
    pImg->pSectionMap = pNewSectionMap;
    pImg->nSectionBits = nMySections;

    pImg->BitsSize = BitsSize;
    return TRUE;

}

BOOL ResolveUndefinedSymbols(PIMAGE pImg, PRTL pRtl)
{
    Elf32_Shdr *pShdr;
    UINT i, j, t, ns, sx, bnd, typ;
    UINT nSections;
    Elf32_Sym *pSym;
    char *Names;
    PRTL_ENTRY pEntry;
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;

    if (fFlags)
    printf("Resolving undefined symbol using %s...\n",pRtl->DllName);

    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    nSections = GetHalf(MustSwap,pElf->e_shnum);
    Names = FindStringSection(pImg);

    if (Names == NULL) {
        printf("Warning -- strings missing \n");
        return FALSE;
    }

    /* Pessimistically assume there is more than one symtab
     */
    for (i = 0; i < nSections; ++i, ++pShdr) {

        if (GetWord(MustSwap,pShdr->sh_type) == SHT_SYMTAB) {


            t = GetWord(MustSwap,pShdr->sh_size);
            ns = t / sizeof *pSym;
           
            pSym = (Elf32_Sym *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
            j = 1; pSym++;
            for (; j < ns; j++, pSym++) {
                sx = GetHalf(MustSwap,pSym->st_shndx);
                if (sx == SHN_UNDEF) {
                    t = GetWord(MustSwap,pSym->st_name);
                    typ = ELF32_ST_TYPE(pSym->st_info);
                    bnd = ELF32_ST_BIND(pSym->st_info);

                    if (fFlags) {
                    printf(" [%x] nam=%x val=%x sz=%x t=%s b=%c o=%x s=%x", j,
                       t,
                       GetWord(MustSwap,pSym->st_value),
                       GetWord(MustSwap,pSym->st_size),
                       STypeName(typ), SBindName(bnd), pSym->st_other,
                       sx);
                    printf(" sname=%s\n", Names + t);
                    }

                    pEntry = IsaRtl(pRtl,Names+t);
                    if (pEntry) {
                        if (fFlags)
                        printf(" Resolved %s as %s[%d]\n",
                               Names+t,pRtl->DllName,pEntry->Ordinal);
                        /* BUGBUG should use a larger field */
                        pSym->st_other = pRtl->Index+1;
                        pSym->st_value = GetWord(MustSwap,pEntry->Ordinal);
                        pRtl->Referenced = TRUE;
                    }
                }
            }
        }
    }
    return TRUE;
}

BOOL CheckUndefinedSymbols(PIMAGE pImg)
{
    Elf32_Shdr *pShdr;
    UINT i, j, t, ns, sx, bnd, typ;
    UINT nSections;
    Elf32_Sym *pSym;
    char *Names;
    BOOL MustSwap = pImg->MustSwap;
    Elf32_Ehdr *pElf = pImg->pElf;
    BOOL Undf = FALSE;

    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    nSections = GetHalf(MustSwap,pElf->e_shnum);
    Names = FindStringSection(pImg);

    /* Check all symtables.
     */
    for (i = 0; i < nSections; ++i, ++pShdr) {

        if (GetWord(MustSwap,pShdr->sh_type) == SHT_SYMTAB) {

            t = GetWord(MustSwap,pShdr->sh_size);
            ns = t / sizeof *pSym;
           
            pSym = (Elf32_Sym *)(pImg->FileBase + GetWord(MustSwap,pShdr->sh_offset));
            j = 1; pSym++;
            for (; j < ns; j++, pSym++) {
                sx = GetHalf(MustSwap,pSym->st_shndx);
                /* Ignore those we matched with a DLL
                 */
                if ((sx == SHN_UNDEF) &&
                    (pSym->st_other == 0)) {
                    t = GetWord(MustSwap,pSym->st_name);
                    typ = ELF32_ST_TYPE(pSym->st_info);
                    bnd = ELF32_ST_BIND(pSym->st_info);

                    printf("Undefined symbol");
                    if (Names)
                        printf(" '%s'", Names + t);

                    printf(": [%x] name=%x size=%x type=%s binding=%c\n", j,
                       t,
                       GetWord(MustSwap,pSym->st_size),
                       STypeName(typ), SBindName(bnd), pSym->st_other,
                       sx);
//                    Undf = TRUE;
                }
            }
        }
    }
    return !Undf;
}

BOOL BuildImportSection(PIMAGE pImg)
{
    BOOL MustSwap = pImg->MustSwap;
    PRTL pRtl;
    UINT nImport, i;
    Elf32_Shdr *pShdr;
    Elf32_Ihdr *pIhdr;

    /* Allocate and zero one big chunk for header and section
     */
    nImport = pImg->nRtl;
    i = sizeof(Elf32_Shdr) + (nImport * sizeof(Elf32_Ihdr));
    ALLOCATE(pShdr,Elf32_Shdr*,i,"ImprtSec");
    if (pShdr == NULL)
        return FALSE;

    pShdr->sh_entsize = GetWord(MustSwap,sizeof(Elf32_Ihdr));
    pShdr->sh_size = GetWord(MustSwap,nImport * sizeof(Elf32_Ihdr));
    pShdr->sh_offset = GetWord(MustSwap,sizeof(Elf32_Shdr));

    pShdr->sh_name = 0; /* BUGBUG */
    pShdr->sh_type = GetWord(MustSwap,SHT_SHLIB);
    pShdr->sh_flags = 0;
    pShdr->sh_addr = 0;
    pShdr->sh_link = 0;
    pShdr->sh_info = 0;
    pShdr->sh_addralign = GetWord(MustSwap,4);


    pIhdr = (Elf32_Ihdr *)(pShdr + 1);
    for (i = 0; i < nImport; i++) {
      pRtl = pImg->pRtl[i];
      if (pRtl->DllName)
        strcpy(pIhdr[i].ih_name,pRtl->DllName);
      if (pRtl->Uuid)
        memcpy(pIhdr[i].ih_uuid,pRtl->Uuid,16);
    }
    pImg->pImport = pShdr;
    pImg->ImportSize = GetWord(MustSwap,pShdr->sh_size);

    return TRUE;
}

void SafeWrite(FILE *f, char *Buf, size_t nBytes)
{
  size_t n = fwrite(Buf,1,nBytes,f);
  if (n != nBytes) {
    printf("Short write (x%x) !! [%d bytes instead of %d]\n",
           GetLastError(), n, nBytes);
    exit(1);
  }
  fflush(f);/*uhu?*/
}

void WriteSections(FILE *f, PIMAGE pImg)
{
  UINT i, nSections, tag;
  Elf32_Shdr *pShdr;
  Elf32_Off Offset, Size;

  pShdr = pImg->pHdr->Sh;
  nSections = pImg->nSectionBits;
  if ((nSections+1) != GetHalf(pImg->MustSwap,pImg->pHdr->Elf.e_shnum))
  {
    printf("WriteSections?? %d != %d\n", nSections, pImg->nSectionBits);
  }
  for (i = 0; i < nSections; i++) {
      Offset = GetWord(pImg->MustSwap,pShdr[i].sh_offset);
      Size = GetWord(pImg->MustSwap,pShdr[i].sh_size);
      tag = GetWord(pImg->MustSwap,pShdr[i].sh_type);
      if (Size && (tag != SHT_NOBITS)) {
          if (fFlags)
              printf("WriteSection[%d] x%x bytes addr %x from x%x [@ %x]\n",
                     i,Size,
                     pImg->pSectionMap[i].SecAddr,
                     pImg->pSectionMap[i].SecBits,
                     Offset);
          fseek(f,Offset,SEEK_SET);
          SafeWrite(f,pImg->pSectionMap[i].SecBits,Size);
      } else if (fFlags)
          printf("-----Section[%d] x%x bytes addr %x from x%x [@ %x]\n",
                 i,Size,
                 pImg->pSectionMap[i].SecAddr,
                 pImg->pSectionMap[i].SecBits,
                 Offset);
  }

}

void WriteBits(FILE *f, PIMAGE pImg)
{
  /* Just relocation ?
   */
  if (pImg->pHdr == NULL) {
      SafeWrite(f,(char*)pImg->pElf,pImg->ElfSize);
      return;
  }

  SafeWrite(f,(char*)pImg->pHdr,pImg->ElfSize);

  WriteSections(f,pImg);

}

/*
 * ARCHOPS v-tables
 */
struct _ARCHOPS VTables[] = 
{
 { /* MIPS */
   EM_MIPS,
   RelocateMipsEntry,
   MipsDisasm,
   MipsRelName,
   NoRemap
 },
 { /* ARM */
   EM_ARM,
   RelocateArmEntry,
   ArmDisasm,
   ArmRelName,
   ArmRemap
 },
 { /* unknown, doesnt work [NB: MUST be last] */
   EM_NONE,
   DontRelocate,
   NoDisasm,
   NoRelName,
   NoRemap
 }
};
#define nARCHOPS (sizeof VTables / sizeof VTables[0])

/*
 * Load an ELF-formatted program image into memory.
 */
SCODE LoadElfImage(PIMAGE pImg)
{
    CTSTR pImage = (CTSTR)pImg->ImageName;
    SCODE sc;
    FILE_HANDLE f;
    INT MemorySize;
    UINT i;
    Elf32_Word t, nSections;
    Elf32_Off o, BaseOffset;
    Elf32_Addr LinkBase;
    THREAD_FUNCTION EntryPoint;
    Elf32_Ehdr *pElf;
    Elf32_Phdr *pPhdr;
    Elf32_Shdr *pShdr;
    char *s;
    BYTE *pMem = NULL;
    BOOL MustSwap;
    Elf32_Sym  *Symbols = NULL;
    Elf32_Off   SymbolsCount = 0;
    Elf32_Rel  *Relocations;
    Elf32_Off   RelocationsCount;

    /* Check the file exists, see how big it is
     */
    f = _Open( pImage, 0x1 );
    if (f == INVALID_FILE_HANDLE) {
        printf("Cannot open file %s\n", pImage);
        return E_FILE_NOT_FOUND;
    }

    MemorySize = _Stat( f );
    if (MemorySize < sizeof(Elf32_Ehdr)) {
        goto Exit;
    }

    /* Read in all the file.
     */
    ALLOCATE(pMem,BYTE *,MemorySize,"AllFile");
    if (pMem == NULL) {
        goto Exit;
    }

    if (MemorySize != _Read( f, pMem, MemorySize)) {
        goto Exit;
    }
    if (fFlags)
    printf("Read in x%x bytes for file %s\n", MemorySize, pImage);

    /* Is this a valid ELF file?
     */
    pElf = (Elf32_Ehdr *)pMem;
    MustSwap = (pElf->e_ident[EI_DATA] != MYBYTEORDER);
    s = (char *)pElf->e_ident;
    if ((*s != ELFMAG0) || 
        (memcmp(s + 1, "ELF", 3) != 0)) {
        printf("Error -- Not a valid ELF file!\n");
        SetLastError((UINT)E_INVALID_PARAMETER);
        goto Exit;
    }

    /* Do we know this architecture ?
     */
    for (i = 0; i < nARCHOPS; i++) {
        if (GetHalf(MustSwap,pElf->e_machine) == VTables[i].Machine) {
            pImg->v = VTables+i;
            break;
        }
    }
    if (i == nARCHOPS) {
        printf("Unknown machine type [x%x] ..this wont go far\n",
               GetHalf(MustSwap,pElf->e_machine));
        pImg->v = VTables+i-1;
    }
    

    pImg->FileBase = pMem;
    pImg->pElf = pElf;
    pImg->ElfSize = MemorySize;
    pImg->MustSwap = MustSwap;

    /* Print the ELF header
     */
    if (fFlags)
        PrintElfHeader(MustSwap,pElf);

    /* Print all program headers
     */
    pPhdr = (Elf32_Phdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_phoff));

    t = GetHalf(MustSwap,pElf->e_phnum);
    if ((t == 0) && (fFlags)) {
        printf("Info -- Input file has no pgm headers (ld -r ?).\n");
    }

    if (fFlags)
    for (i = 0; i < t; ++i) {
        printf("PgmHeader %d: ",i);
        printf("type=%s off=%x vaddr=%x paddr=%x fs=%x ms=%x flg=%x aln=%x\n",
               PgmTypeName(GetWord(MustSwap,pPhdr[i].p_type)),
               GetWord(MustSwap,pPhdr[i].p_offset),
               GetWord(MustSwap,pPhdr[i].p_vaddr),
               GetWord(MustSwap,pPhdr[i].p_paddr),
               GetWord(MustSwap,pPhdr[i].p_filesz),
               GetWord(MustSwap,pPhdr[i].p_memsz),
               GetWord(MustSwap,pPhdr[i].p_flags),
               GetWord(MustSwap,pPhdr[i].p_align) );

        if (GetWord(MustSwap,pPhdr[i].p_type) == PT_DYNAMIC)
            PrintDynamic(pImg,pPhdr+i);
    }

    /* Print all section headers
     */
    t = GetHalf(MustSwap,pElf->e_shentsize);
    if (t != (sizeof *pShdr)) {
        printf("Fatal -- section headers have the wrong size (%d != %d)\n",
               t, sizeof *pShdr);
        SetLastError((UINT)E_INVALID_PARAMETER);
        goto Exit;
    }

    nSections = GetHalf(MustSwap,pElf->e_shnum);
    if (nSections == 0) {
        /* this thing is really bogus */
        printf("Fatal -- no section headers\n");
        SetLastError((UINT)E_INVALID_PARAMETER);
        goto Exit;
    }

    /* Allocate and zero the section bit pointers. Make room for additions
     */
    pImg->nSectionBits = nSections + 2;
    ALLOCATE(pImg->pSectionMap,SECMAP*,sizeof(SECMAP)*pImg->nSectionBits,
             "pSecMap");
    if (pImg->pSectionMap == NULL) goto Exit;

    /* Find the base load address and offset while we are at it
     */
    BaseOffset = (Elf32_Off)~0;
    LinkBase = (Elf32_Addr)~0;
    pShdr = (Elf32_Shdr *)(pImg->FileBase + GetWord(MustSwap,pElf->e_shoff));
    if (fFlags) printf("\n");
    for (i = 0; i < nSections; ++i) {
        UINT tag = GetWord(MustSwap,pShdr[i].sh_type);
        UINT flg = GetWord(MustSwap,pShdr[i].sh_flags);
        o = GetWord(MustSwap,pShdr[i].sh_offset);
        if (fFlags) {
        printf("ScnHeader %d [%s]: ", i,
               SectionName(pImg,GetWord(MustSwap,pShdr[i].sh_name),i));
        printf(" n=%x t=%s f=%x a=%x o=%x sz=%x l=%x i=%x gn=%x es=%x\n",
               GetWord(MustSwap,pShdr[i].sh_name),
               ScnTypeName(tag),
               flg,
               GetWord(MustSwap,pShdr[i].sh_addr),
               GetWord(MustSwap,pShdr[i].sh_offset),
               GetWord(MustSwap,pShdr[i].sh_size),
               GetWord(MustSwap,pShdr[i].sh_link),
               GetWord(MustSwap,pShdr[i].sh_info),
               GetWord(MustSwap,pShdr[i].sh_addralign),
               GetWord(MustSwap,pShdr[i].sh_entsize));
        }

        if (o) {
          pImg->pSectionMap[i].SecBits = pImg->FileBase + o;
          if (o < BaseOffset)
            BaseOffset = o;
          pImg->pSectionMap[i].SecAddr = GetWord(MustSwap,pShdr[i].sh_addr);
        }

        if ((tag == SHT_REL) && (fFlags & F_RELS))
            PrintRels(pImg,&pShdr[i]);

        if ((tag == SHT_RELA) && (fFlags & F_RELS))
            PrintRelas(pImg,&pShdr[i]);

        if (tag == SHT_PROGBITS) {
          o = GetWord(MustSwap,pShdr[i].sh_addr);
          /* Beware of .debug sections (which we'll discard) */
          if ((o < LinkBase) && (flg != 0))
            LinkBase = o;
          if (fFlags & F_DISASM)
            PrintInsts(pImg,&pShdr[i],i);
          if ((flg & SHF_EXECINSTR) &&
              (pImg->CodeIndex == 0)) /* pick the first one, sigh */
            pImg->CodeIndex = i;
        }

        if (tag == SHT_SYMTAB) {
          /* checked inside, needs sideffects if (fFlags & F_SYMS)*/
          SymbolsCount = PrintSyms(pImg,&pShdr[i],i);
          Symbols = (Elf32_Sym *)pImg->pSectionMap[i].SecBits;
        }

        if ((tag == SHT_SHLIB) && (fFlags & F_IMP))
            PrintImports(pImg,&pShdr[i],i);

        if ((tag == SHT_DYNSYM) && (fFlags & F_EXP))
            PrintExports(pImg,&pShdr[i],i);
    }

    pImg->OldLoadOffset = BaseOffset;
    pImg->OldLoadAddress = LinkBase;
    /* Unless overruled... */
    pImg->NewLoadOffset = BaseOffset;

    pImg->BitsSize = MemorySize - BaseOffset;

    /* For later..
     */
    if (fOptions & F_DISASM) {
      UINT Relix = 0;
      if (FindRelocsForSection(pImg,pShdr,pImg->CodeIndex, &Relix) || 1) {
        pImg->pSplit  = SplitSectionSymbols(pImg, nSections,
                                            Symbols,SymbolsCount,
                                            FindStringSection(pImg),
                                            pShdr + Relix);
        if (pImg->pSplit) {
          pImg->Codinfo = &pImg->pSplit->Secinfo[pImg->CodeIndex];
          TheImage = pImg;
        }
      }
      /* Otherwise no symbols */
    }

    /* Find all exported symbols, make sure they are defined.
     * NB: could get fancy and do it after the imports...
     */
    if (pImg->pExportRtl) {
      if (!MatchExportSymbols(pImg,pImg->pExportRtl))
        goto Exit;
    }

    /* Walk all undefined symbols and fixem up from the RTLs
     */
    if (pImg->pRtl) {
      BOOL GotRefs = FALSE;
      PRTL pRtl;
      for (i = 0; i < pImg->nRtl; i++) {

        pRtl = pImg->pRtl[i];
        if (!ResolveUndefinedSymbols(pImg,pRtl))
          continue;
        GotRefs = GotRefs || pRtl->Referenced;
      }

      /* Build an import section if necessary
       */
      if (GotRefs)
        BuildImportSection(pImg);
    }

    /* Make sure we have no undefined symbols
     */
    if (!CheckUndefinedSymbols(pImg)) {
        printf("Fatal -- Not all symbols are defined\n");
        SetLastError((UINT)E_INVALID_PARAMETER);
        goto Exit;
    }

    /* If there are no program headers that means we are munging a -r image
     */
    t = GetHalf(MustSwap,pElf->e_phnum);
    if ((t == 0) || (fFlags & F_MUNGE)) {
        if (!Munge(pImg)) {
            goto Exit;
        }
        pMem = NULL;
        pImg->pElf = &pImg->pHdr->Elf;
    }

    /* Do the relocation now
     */
    if (fFlags)
    printf("Relocating at 0x%x\n", pImg->NewLoadAddress);
    if (!DoElfRelocations(f, pImg)) {
        /* Might just want to disassemble a regular image ?*/
        if ((fOptions & F_DISASM) == 0)
            goto Exit;
    }

    /* Disassemble output if required
     */
    if (fOptions & F_DISASM) {
      if (pImg->pSplit)
        for (i = 0; i < nSections; i++) {
          pImg->pSplit->Secinfo[i].SecBase = pImg->pSectionMap[i].SecAddr;
        }
      /* xxx not really xxx */
      pShdr[pImg->CodeIndex].sh_addr = 
        GetWord(MustSwap,pImg->NewLoadAddress+pImg->ExportAdjustment);

      for (i = 0; i < nSections; i++) {
        if (GetWord(MustSwap,pShdr[i].sh_flags) & SHF_EXECINSTR) {
            pImg->CodeIndex = i;
            pImg->Codinfo = &pImg->pSplit->Secinfo[pImg->CodeIndex];
            PrintInsts(pImg,pShdr+i,i);
        }
      }

      if (fOptions & F_SYMS)
          PrintSortedFunctions(pImg, ScopeFileName);

    }

    /* Done with file handle
     */
    _Close(f);
    f = INVALID_FILE_HANDLE;

    EntryPoint = (THREAD_FUNCTION)(GetWord(MustSwap,pImg->pElf->e_entry));
    if (fFlags & F_NOISE) {
        printf("Image %hs will use x%x bytes at x%x, entry = %x\n", 
               pImage, pImg->BitsSize, pImg->NewLoadAddress, EntryPoint);
        printf("Done loading\n");
    }
    SetLastError( NO_ERROR );

    pMem = NULL;

Exit:
    sc = GetLastError();
    if (pMem)
        FREE(pMem,"AllFile");
    if (f != INVALID_FILE_HANDLE)
        _Close(f);
    return sc;
}



/* Runtime functions
 */
__inline int xtoi(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    if (c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    return -1;
}

int atox(const char *str)
{
    int i,j;
    char c;

    i = 0;
    while (str && (c = *str++)) {
      j = xtoi(c);
      if (j < 0) break;
      i = (i << 4) | j;
    }
    return i;
}

#define MAXDLL 255
BOOL AddDll(PIMAGE pImg, char *Filename)
{
  int ix;
  PRTL pRtl;

  if (pImg->pRtl == NULL) {
    ALLOCATE(pImg->pRtl,PRTL *,MAXDLL*sizeof(PRTL),"pRtl");
    if (pImg->pRtl == NULL) return FALSE;
    pImg->nRtl = 0;
  }
  ix = pImg->nRtl++;
  if (ix >= MAXDLL) {
    printf("Fatal -- Too many DLLs\n");
    return FALSE;
  }
  printf(".import %s: ", Filename);
  pRtl = ParseDllDeffile(Filename);
  if (!pRtl)
    return FALSE;
  pImg->pRtl[ix] = pRtl;
  pRtl->Index = ix;
  return TRUE;
}

/* BUGBUG Only one for now */
BOOL AddExport(PIMAGE pImg, char *Filename)
{
  PRTL pRtl;

  if (pImg->pExportRtl) {
    printf("Fatal -- Too many .export directives\n");
    return FALSE;
  }
  printf(".export %s: ", Filename);
  pRtl = ParseDllDeffile(Filename);
  if (!pRtl)
    return FALSE;
  pImg->pExportRtl = pRtl;
  return TRUE;
}

/* */
void Cleanup(PIMAGE pImg)
{
  if (pImg->pRtl) {
    while (pImg->nRtl--)
      DestroyRtl(pImg->pRtl[pImg->nRtl]);
    FREE(pImg->pRtl,"pRtl");
  }
  if (pImg->pSectionMap)
    FREE(pImg->pSectionMap,"pSecMap");
  if (pImg->FileBase)
    FREE(pImg->FileBase,"AllFile");
  if (pImg->pImport)
    FREE(pImg->pImport,"ImprtSec");
  if (pImg->pExportSection)
    FREE(pImg->pExportSection,"XportSec");
  if (pImg->pRela)
    FREE(pImg->pRela,"RELAs");
  if (pImg->pHdr)
    FREE(pImg->pHdr,"pHdr");
  if (pImg->pSplit)
    Desplit(pImg->pSplit);
    
  memset(pImg,0,sizeof *pImg);
  if (nInUse)
    printf("MEMORY LEAK? nInUse == %d\n", nInUse);
}

int main(int argc, char **argv)
{
  Elf32_Addr Where = 0x4000;
  SCODE sc;
  char *pgm = argv[0];
  IMAGE Img;
  char *ImageType = "ELF";
  UINT32 ImageVersion = EV_CURRENT;


  memset(&Img,0,sizeof Img);

  while ((argc > 1) && (argv[1][0] == '-')) {
    switch (argv[1][1]) {
    case 'd':
      fOptions |= F_DISASM;
      argv += 1;
      argc -= 1;
      break;
    case 's':
      fOptions |= F_SYMS;
      ScopeFileName = argv[2];
      argv += 2;
      argc -= 2;
      break;
    case 'z':
      fOptions = fOptions ^ F_STRIP;
      argv += 1;
      argc -= 1;
      break;
    case 'T':
      Where = atox(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'v':
      fFlags = atox(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'V':
      ImageVersion = atoi(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'I':
      if (!AddDll(&Img,argv[2]))
        return 1;
      argv += 2;
      argc -= 2;
      break;
    case 'E':
      if (!AddExport(&Img,argv[2]))
        return 1;
      argv += 2;
      argc -= 2;
      break;
    default:
      printf("Unknown option '-%c'\n",argv[1][1]);
      goto usage;
    }
  }

  if (argc == 1) {
  usage:
    printf(
"Usage: [-d] [-z] [-T hexAddr] [-v Flags] [-V DecimalVersionNumber] [-Import DefFile]* [-Export DefFile] "\
           "InFile [OutFile]\n", pgm);
    printf(
" -d disassembly of output file (after relocation), with symbols.\n");
    printf(
" -z strip output file to the minimum, only keeps the relocations.\n");
    printf(
" -T is the hexadecimal address to relocate the image to, it could\n"\
"    also be used for statically rebasing an image. Default: %x.\n",
           Where);
    printf(
" -V sets the version number of the output image. [Default: %d]\n"\
" -v depends on the or-ed bits of the (hex) Flags argument [Default: %x]:\n"\
"\t 1 print raw disassembly\n"\
"\t 2 print symbols\n"\
"\t 4 print relocations\n"\
"\t 8 print import section\n"\
"\t10 trace .REL matching code\n"\
"\t20 trace internal memory allocations\n"\
"\t40 print export section\n",  EV_CURRENT, fFlags
           );
    printf(
" -I[mport] defines functions and UUIDs imported from other components.\n");
    printf(
" -E[xport] defines functions and UUIDs that OutFile will export.\n");
    printf(
" InFile is the input %s file.  It is an object file, or the output of\n"\
"    'ld -r -d' or of a previous run of %s itself.\n"\
"    Use ld to bind multiple objects in one.\n",
          ImageType, pgm);
    printf(
" OutFile is the output image. If missing %s just dumps info.\n",
           pgm);
    return 1;
  }


  Img.ImageName = argv[1];
  Img.NewLoadAddress = Where;
  sc = LoadElfImage(&Img);
  if (FAILED(sc)) {
    printf("Link failed (x%x)\n", sc);
    return 1;
  }

  if (argc > 2) {
    FILE *f;

    f = fopen(argv[2],"wb+");
    if (f == NULL) {
      printf("Could not create output file %s\n", argv[2]);
      return 1;
    }

    Img.pElf->e_version = GetWord(Img.MustSwap,ImageVersion);

    WriteBits(f,&Img);

    fclose(f);
  }

  Cleanup(&Img);

  return 0;
}
